/* tf2ss_c subroutine
 * Transfer function to minimal state space model form computation
 *
 * Copyright (C) 2007-2015 Alan Layec
 *
 * This file is part of modnumlib.
 *
 * modnumlib is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * modnumlib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with modnumlib; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

/* REVISION HISTORY :
 * $Log: tf2ss_c.c,v $
 * Revision 1.1  2016-07-20 14:00:20  jpc
 * updates
 *
 * Revision 1.1  2015-12-25 15:47:58  layec
 * add
 *
 */

#include "modnum_lib.h"

/* 
 *       b0*s^n + b1*s^n-1 + .... + bn-1*s + bn
 * T(s)= ---------------------------------------
 *       a0*s^n + a1*s^n-1 + .... + an-1*s + an
 * 
 * dx/dt=Ax+Bu
 * y=Cx+Du
 * 
 * num : [b0 b1 ... bn]
 * den : [a0 a1 ... an]
 * 
 */
void tf2ss_c(int *index,int *typ,double *num,double *den,double *A,double *B,double *C,double *D)
{
 /* Déclaration des variables compteurs */
 int i,j,k;
 int ii,jj;
 
 double zero=0.;
 double norm;
 
 /* Initialisation des matrices d'états */
 setd_c((i=(*index)*(*index),&i),&zero,A);
 setd_c(index,&zero,B);
 setd_c(index,&zero,C);
 D[0]=0.;
 
 /* Sélection du type de système linéaire */
 switch((*typ))
 {
   /* Forme canonique commandable */
   case 1 : /* normalisation */
            norm=den[0];
            for(i=0;i<((*index)+1);i++) {
              num[i] = num[i] / norm;
              den[i] = den[i] / norm;
            }
            /* Matrice A */
            /*
             *    | 0   1     0    ...... 0  |
             *    | 0   0     1    ...... 0  |
             * A= | 0   0     0    ...... 0  |
             *    | 0   0     0    ...... 1  |
             *    |-an -an-1 -an-1 ..... -a1 |
             * 
             */
            /* identité décalée d'1 colonne */
            ii=0;jj=1;
            for(i=0;i<(*index);i++) {   /* ligne */
              for(j=0;j<(*index);j++) { /* colonne */
                k = i+j*(*index);
                if ((j==jj) && (i==ii)) {
                  A[k]=1.;
                  ii++;jj++;
                }
                if (i==(*index)-1) {
                  A[k]=-den[(*index)-j];
                }
              }
            }

            /* Matrice B */
            /*
             *    | 0 |
             *    | 0 |
             * B= | 0 |
             *    | 0 |
             *    | 1 |
             * 
             */
            B[(*index)-1]=1.;

            /* Matrice C */
            /*
             * C = | bn-an*b0 bn-1-an-1*b0 ..... b1-a1b0 |
             * 
             */
            for(i=0;i<(*index);i++) {   /* ligne */
              C[i]=num[(*index)-i]-den[(*index)-i]*num[0];
            }

            /* Matrice D
             *
             * D = | b0 |
             * 
             */
            D[0]=num[0];
            break;

  /* do nothing */
  default : break;
 }
}
