/* Modnumlib Scicos interfacing function
 * Copyright (C) 2009-2015 Alan Layec
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/* vectorscope Scicos Time domain vectorial scope block
 *
 * - Based on Scicoslab NG routines
 * - Based on Nsp NG routines
 *
 * Type 4 simulation function ver 1.0 - scilab-4.1
 * 24 mai 2007 - INRIA - Author : A.Layec
 * 
 * Type 4 simulation function ver 1.1 - scicos-4.4
 * 3 avril 2013 - ENPC - Author : A.Layec
 *
 */

/* REVISION HISTORY :
 * $Log: vectorscope41_nsp.c,v $
 * Revision 1.1  2016-07-20 14:00:21  jpc
 * updates
 *
 * Revision 1.1  2015-12-25 15:47:52  layec
 * add
 *
 */

/*Standarc C include*/
#include <stdio.h>
#include <string.h>

#include <nsp/axes.h>
#include <nsp/qcurve.h>

/*Scicos block st include*/
#include <scicos/scicos_block.h>
#include <modnum_lib.h>

extern BCG *scicos_set_win(int wid, int *oldwid);

/*work structure of that block*/
typedef struct work_st {
  NspAxes *Axes;    /*hdl of the NG scope*/
  int    wid;    /*wid of the NG scope*/
  /*y axis autoscale variables*/
  double y_min;
  double y_max;
  double y_up;
  double y_down;
  double y_delta;
} work_struct;

/* vectorscope_draw :
 *  Initialization and re-initialization
 *   of the vectorscope block
 */
void vectorscope_draw(scicos_block *block)
{
 BCG *Xgc;
 NspFigure *F;
 NspAxes *Axe,*Axe1;

 /* counter variables */
 int i,j,l;

 /* variables to get rpar/ipar ptr */
 double *rpar;
 int    *ipar;

 /* variables to store axes bound parameters */
 double xmin, xmax;
 double ymin, ymax;

 double Tse; /*sample unit time*/

 /* various graphic variables */
 double frect[4];
 char wnam[100];
 char x_label[30];
 char y_label[30];

 int win_dim[2] = {-1,-1};

 int wid,oldwid; /*the windows id*/
 int nvec; /*number of vectors to display*/
 int insz; /*length of one vector */

 /* ptr of the struct of that block */
 work_struct *work;

 /* get size of input ports */
 insz = block->insz[0];

 /* get the struct of that block */
 work = (work_struct*) *block->work;

 /* get rpar/ipar */
 rpar = block->rpar;
 ipar = block->ipar;

 wid  = ipar[0];
 nvec = ipar[1];

 win_dim[0] = ipar[3];
 win_dim[1] = ipar[4];

 Tse     = rpar[0];
 ymin    = rpar[1];
 ymax    = rpar[2];

 /* compute xmin/xmax bound of the scope */
 xmin = 0;
 xmax = nvec*insz*Tse-Tse;

 /* get wid of the scope and store it in my_st*/
 if (wid<0) {
   work->wid = 20000 + get_block_number();
 }
 else {
   work->wid = wid;
 }

 /* set current window to id number wid */
 Xgc = scicos_set_win(work->wid, &oldwid);

 /* set window dims */
 if (win_dim[0] >= 0) {
   Xgc->graphic_engine->xset_windowdim(Xgc, win_dim[0], win_dim[1]);
 }

 /* set window title */
 sprintf(wnam, "Vector Scope (%d)", work->wid);
 Xgc->graphic_engine->setpopupname (Xgc, wnam);

 if ((F = nsp_check_for_figure(Xgc,FALSE)) == NULL) {
   scicos_set_block_error(-16);
   return;
 }

 /* clean the figure */
 l = nsp_list_length(F->obj->children);
 for (i=0;i<l;i++) {
   nsp_list_remove_first(F->obj->children);
 }

 /* create a new axe */
 if ((Axe = nsp_check_for_axes(Xgc, NULL)) == NULL) {
   scicos_set_block_error(-16);
   return;
 }

 /* create a set of qcurves and insert them in axe */
 for (i=0;i<nvec;i++) {
   double *qx;
   double *qy;
   NspQcurve *curve;

   NspMatrix *Pts = nsp_matrix_create("Pts", 'r', insz, 2);
   if (Pts == NULL) {
     scicos_set_block_error(-16);
     return;
   }

   qx = Pts->R;
   qy = Pts->R + insz;

   /* initial values */
   for (j=0;j<insz;j++) {
     qx[j] = (j+i*insz)*Tse;
     qy[j] = 0.;
   }

   curve = nsp_qcurve_create("curve", -1, 0, 0, i+1, qcurve_std, Pts, NULL, -1, -1, NULL);
   if (curve == NULL) {
     scicos_set_block_error(-16);
     return;
   }

   curve->obj->start = 0;
   curve->obj->last  = insz-1;

   /* insert the new curve */
   if (nsp_axes_insert_child(Axe, (NspGraphic *)curve, FALSE) == FAIL) {
     scicos_set_block_error(-16);
     return;
   }
 }

 /* updates the axes scale */
 frect[0] = xmin; frect[1] = ymin;
 frect[2] = xmax; frect[3] = ymax;
 nsp_strf_axes(Axe, frect, '5');
 memcpy(Axe->obj->frect->R, frect, 4*sizeof(double));
 memcpy(Axe->obj->rect->R, frect, 4*sizeof(double));

 /* set grid, log scale and etc */
 Axe->obj->grid  = 0;
 Axe->obj->axes  = 1;
 Axe->obj->xlog  = FALSE;
 Axe->obj->ylog  = FALSE;
 Axe->obj->fixed = TRUE;

 /* axis labels */
 sprintf(x_label,"Time [s]"); /*x axis*/
 Axe->obj->x = nsp_string_copy(x_label);

 sprintf(y_label,"Level"); /*y axis*/
 Axe->obj->y = nsp_string_copy(y_label);

 /* keep a copy in case Axes is destroyed during simulation
  * axe is a by reference object
  */
 Axe1 = nsp_axes_copy(Axe);
 if (Axe1 == NULL) {
   scicos_set_block_error(-16);
   return;
 }

 /* store Axes1 ptr in work st */
 work->Axes=Axe1;

 /* draw initial axe */
 nsp_axes_invalidate((NspGraphic *)work->Axes);
}

/* main routine */
void vectorscope41(scicos_block *block,int flag)
{
 /* graphic ptr variables */
 Cell *Loc;
 NspList *L;

 /* counter variables */
 int i,k;

 /* variables to get block ptr */
 int *ipar;
 double *u1;

 /* ptr of the struct of that block */
 work_struct *work;

 /* counters and indices variables */
 int nvec;
 int cnt_nvec;
 int insz;

 /* some scope parameters */
 int scaltyp; /*0:ymin/ymax/1:autoscale*/

 /* autoscale variables */
 double y_height;
 double y_delta;
 double y_up;
 double y_down;
 double fact=0.4;

 /* get ptr and size of input ports */
 u1   = block->inptr[0];
 insz = block->insz[0];

 /* get ipar variables */
 ipar    = block->ipar;
 nvec    = ipar[1]; /*number of vector*/
 scaltyp = ipar[2];

 /* get counters value */
 cnt_nvec = (int)block->z[0];

 /* ptr of the struct of that block */
 work = (work_struct*) *block->work;

 if (flag==4) { /*flag 4*/

   /* my_st */
   if ((work=(work_struct*) scicos_malloc(sizeof(work_struct))) == NULL) {
     set_block_error(-16);
     return;
   }

   /* init. autoscale y axis bound */
   work->y_delta = 0;

   *block->work=(void *)work; /*store my_st in *block->work*/

   vectorscope_draw(block); /*call vectorscope_draw for initialization*/
 }

 else if (flag==2) { /*flag 2*/

   /* Axes was destroyed during simulation */
   if (work->Axes->obj->ref_count <= 1) {
     return;
   }

   /* increase vector counter value */
   cnt_nvec++;

   L   = work->Axes->obj->children; /* qcurves list */
   Loc = L->first;
   i   = 0; /* nvec counter */

   /* choose the good polyline */
   while ((i+1) != cnt_nvec) {
     Loc = Loc->next;
     i++;
   }

   if (Loc->O != NULLOBJ) {
     NspQcurve *curve =(NspQcurve *) Loc->O;
     NspMatrix *M = curve->obj->Pts;
     double *qy   = M->R + M->m;

     if (i == 1) {
       work->y_min = u1[0];
       work->y_max = u1[0];
     }

     for (k=0;k<insz;k++) {
       qy[k]=u1[k];
       work->y_min=min(work->y_min,u1[k]);
       work->y_max=max(work->y_max,u1[k]);
     }
   }

   /* draw qcurve(s) */
   if (cnt_nvec == nvec) {

     /* autoscale */
     if (scaltyp == 1) {
       y_height = work->y_max - work->y_min;
       y_delta  = work->y_min + y_height/2; /*also equal to max_y-y_height/2*/
       y_down   = y_delta - (1+fact) * y_height/2;
       y_up     = y_delta + (1+fact) * y_height/2;

       if ( (((1+fact) * y_height/2) > (1+fact) * work->y_delta) || \
            (((1+fact) * y_height/2) < (1-fact) * work->y_delta) ) {
          work->y_delta = ( (1+fact) * y_height/2);
          work->y_up    = y_up;
          work->y_down  = y_down;
       }

       work->Axes->obj->rect->R[1] = y_down;
       work->Axes->obj->rect->R[3] = y_up;
     }

     nsp_axes_invalidate((NspGraphic *)work->Axes);

     /* reset counter */
     cnt_nvec = 0;
   }

   /* save vector counter value in dstate */
   block->z[0] = (double)cnt_nvec;
 }

 else if (flag==5) { /*flag 5*/
   /* Axes was not destroyed during simulation */
   if (work->Axes->obj->ref_count >= 1) {
     nsp_axes_invalidate((NspGraphic *)work->Axes);
     nsp_axes_destroy(work->Axes);
   }

   scicos_free(work);
 }

}

