#######################################################################
#                                                                     #
#                               Simport                               #
#                                                                     #
#                   Pierre Weis, INRIA Rocquencourt                   #
#                                                                     #
#  Copyright 2010-2019,                                               #
#  Institut National de Recherche en Informatique et en Automatique.  #
#  All rights reserved.                                               #
#                                                                     #
#  This file is distributed under the terms of the BSD License.       #
#                                                                     #
#######################################################################

# $Id: Caml.mk,v 1.1 2019-05-10 13:58:19 jpc Exp $

# This Makefile defines generic rules to compile various kinds of Caml source
# files and to automatically handle their dependencies.

# This Makefile also defines two entries:
# - clean (to clean up the compiled files)
# - depend (to recompute the dependency order).

# This Makefile should be included at the end of the Makefile that handles a
# set of Caml files (to build a library or an application).
# Simpy write at the end of your Makefile:
# include path_to_Caml.mk/Caml.mk

.PHONY: default all generation byt bin
.PHONY: clean clean-all clean-spurious clean-compiled clean-local
.PHONY: clean-generated clean-caml-generated clean-depend dist-clean
.PHONY: depend before-depend common-depend local-depend internal-depend
.PHONY: depend-printer before-depend-printer common-depend-printer
.PHONY: local-depend-printer internal-depend-printer
.PHONY: test test-all

# Main targets
all: generation byt bin

generation: $(CONFIG_TEXT_FILES) $(CAML_FILES)

# Generic clean up
COMPILED_EXT_TO_CLEAN=\
 .cm* .o .a .annot .output .obj .lib .so .byt .bin .dll .exe

clean-compiled:
	@echo "Cleaning compiled files $(COMPILED_EXT_TO_CLEAN)" && \
	for EXT in $(COMPILED_EXT_TO_CLEAN); do \
	  $(FIND) . -name "*$$EXT" -exec $(RM) "{}" \; ; \
	done

clean-local:

SPURIOUS_EXT_TO_CLEAN=*~ .*~ a.out ".\#*" "\#*"

clean-spurious:
	@echo "Cleaning spurious files $(SPURIOUS_EXT_TO_CLEAN)" && \
	for EXT in $(SPURIOUS_EXT_TO_CLEAN); do \
	  $(FIND) . -name "$$EXT" -exec $(RM) "{}" \; ; \
	done

clean-generated: clean-caml-generated

clean-caml-generated:
	@echo "Cleaning Caml generated files $(CAML_GENERATED_FILES)" && \
	$(RM) $(CAML_GENERATED_FILES)

clean: clean-compiled clean-local clean-spurious

clean-all: clean clean-caml-generated
	$(MAKE) depend

dist-clean: clean

# Rebuilding dependencies
clean-depend:
	$(RM) .depend && $(TOUCH) .depend

before-depend: generation

common-depend:
	@echo "Computing Caml file dependencies" && \
	$(CAMLC_DEP) $(CAML_INCLUDES) $(CAML_FILES) | $(SORT) - >> .depend

local-depend:

internal-depend:

depend:\
 clean-depend depend-printer\
 before-depend common-depend\
 local-depend internal-depend\

# Rebuilding printer rules
before-depend-printer:

common-depend-printer:
	@echo "Computing OCamlgen printer generation rules" && \
	if test -n "$(CAML_PRINTER_SRC_FILES)"; then \
	  (for file in $(CAML_PRINTER_SRC_FILES); do \
	    echo "$$(dirname $$file)/$$(basename $$file .mli)_print.ml : \\" && \
	    echo "    $$file" && \
	    echo "	@echo \"Generating printer for $$file\" && \\" && \
	    echo "	\$$(CAMLC_PPR) $$file"; \
	  done) >> .depend; \
	else echo "Nothing to be done for printers"; \
	fi

local-depend-printer:

internal-depend-printer:

depend-printer:\
 before-depend-printer common-depend-printer\
 local-depend-printer internal-depend-printer\

# Compilation rules
# Note: order of suffixes is significant, object file extension should
# precede source file extensions!
.SUFFIXES: .cmxa .cmx .cma .cmo .cmi .obj
.SUFFIXES: .ml .mli .mlin .mliin .mll .mly .mlm .mlms
.SUFFIXES: .a .o .lib .so .c .h
.SUFFIXES: .in .indata

.ml.cmo:
	@echo "Byte compiling $<" && \
	$(CAMLC_BYT) -c $<

.mli.cmi:
	@echo "Compiling interface $<" && \
	$(CAMLC_BYT) -c $<

.ml.cmx:
	@echo "Binary compiling $<" && \
	$(CAMLC_BIN) -c $<

.mlin.ml:
	@echo "Expanding $< into $@" && \
	$(HTMLC) -i $< -o $@

.mliin.mli:
	@echo "Expanding $< into $@" && \
	$(HTMLC) -i $< -o $@

.mll.ml:
	@echo "Generating lexer $<" && \
	$(CAMLC_LEX) $<

.mly.ml:
	@echo "Generating parser $<" && \
	$(CAMLC_YAC) $<

.mly.mli:
	@echo "Generating parser $<" && \
	$(CAMLC_YAC) $<

.mlm.ml:
	@echo "Generating implementation for moca file $<" && \
	$(MOCAC) -c $<

.mlm.mli:
	@echo "Generating interface for moca file $<" && \
	$(MOCAC) -i $<

.mlms.ml:
	@echo "Generating implementation for moca file $<" && \
	$(MOCAC) -c $<

.mlms.mli:
	@echo "Generating interface for moca file $<" && \
	$(MOCAC) -i $<

.c.o:
	@echo "C compiling $<" && \
	$(CAMLC_CC) -c $<

.indata.in:
	@echo "Expanding $< into $@" && \
	$(HTMLC) -i $< -o $@

banner: banner.in
	@echo "Expanding $< into $@" && \
	$(HTMLC) -i banner.in -o banner

include .depend

# Caml.mk does not define make file targets and rules to build and update the
# .depend file automatically.
#
# Instead, Caml.mk defines the PHONY target depend, to build the depend .file
# on demand; target depend is explicitely called each time file .depend must
# be updated, that is either on user request or as the last action of the
# clean-all target.
#
# Note: the existence of file .depend is mandatory to support the make
# machinery used in this set of *.mk files. Hence, file .depend is reset to
# pristine before filling, i.e. touched just after removal then filled up
# again (see targets common-depend and clean-depend above).
#
# The make rule to build file .depend is defined by initial make file
# Config. mk. It explicitely states that .depend is always up to date and has
# not to be handled by implicit make fules.
#
#.depend: ;

# This file has not to be recursively rebuilt by implicit make rules.
$(CONFIG_DIR)/Caml.mk: ;
