(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_compile.ml,v 1.1 2019-05-10 14:09:17 jpc Exp $ *)

let read_simulink_implementation_marshaled_file,
    write_simulink_implementation_marshaled_file =
  Umarshal.read_write_file ()
;;

(** From Simulink interface file to compiled interface file. *)
(* Should we warn if option -c was not given ?
   if not (Configuration.get_compile_only ()) then
   else Say.warning ...
*)
let compile_simulink_interface efname =
  let intf =
    Mdl_interface_to_explicit_ast.
    load_simulink_interface_source_file efname in
  let dst_efname =
    Configuration.simulink_interface_marshaled_file_name efname in
  Mdl_interface_to_explicit_ast.write_simulink_interface_marshaled_file
    dst_efname intf
;;

(** From Simulink implementation file to compiled implementation file. *)
let write_translated_source_file dst_efname code =
  Path.with_explicit_out_file dst_efname (fun oc ->
    let ppf = Format.formatter_of_out_channel oc in
    Format.fprintf ppf "%a@."
      Code_to_host.print_translated_file code)
;;

(** From Simulink compiled implementation file to scicos host source file. *)
let translate_simulink_implementation_marshaled_file src_efname =
  let code = read_simulink_implementation_marshaled_file src_efname in
  let dst_efname =
    Configuration.simulink_implementation_translated_file_name src_efname in
  Say.fdebug (fun epf ->
    Format.fprintf epf
      "@[<v 2>translate_simulink_implementation_marshaled_file:@,\
         src_efname = %s@ \
         dst_efname = %s@]"
      (Path.of_explicit_file_name src_efname)
      (Path.of_explicit_file_name dst_efname));
  write_translated_source_file dst_efname code
;;

let read_mdl_ast src_efname =
  Mdl_to_ast.read_mdl_file src_efname
;;

let read_xml_ast src_efname =
  Slx_to_mdl_ast.read_xml_file src_efname
;;

let read_slx_ast src_efname =
  Slx_to_mdl_ast.read_slx_file src_efname
;;

let read_simulink_ast src_efname =
  if Configuration.
     is_mdl_implementation_source_explicit_file_name src_efname
    then read_mdl_ast src_efname else
  if Configuration.
     is_xml_implementation_source_explicit_file_name src_efname
    then read_xml_ast src_efname else
  if Configuration.
     is_slx_implementation_source_explicit_file_name src_efname
    then read_slx_ast src_efname else
  assert false
;;

let write_ast_to_temp_file src_efname ast =
  let efname = Path.remove_explicit_extension src_efname in
  let base_name = Path.base_name efname in
  let prefix = Printf.sprintf "%s-mdl-ast-" base_name in
  let fname, oc = Filename.open_temp_file prefix ".log" in
  try
    let ppf = Format.formatter_of_out_channel oc in
    Format.fprintf ppf "%a@." Mdl_ast_print.print_simulink_file ast;
    close_out oc;
    Say.fdebug (fun epf ->
      Format.fprintf epf "mdl_compile: wrote temporary mdl ast to %s" fname);
  with
  | x ->
    close_out oc;
    raise x
;;

let register_block_translators () =
  if Configuration.get_verbose () then
    Format.fprintf Format.err_formatter
      "%,Registering simulink block translators.@.";

  let target_model_system = Configuration.get_target_modeler () in
  Genr_block.init target_model_system;
  Genr_utils.init target_model_system;
;;

let compile_simulink_implementation src_efname =

  register_block_translators ();

  let ast = read_simulink_ast src_efname in
  if Configuration.get_debug () then
    write_ast_to_temp_file src_efname ast;
  let code = Mdl_to_code.treat_simulink_ast ast in
  let dst_efname =
    Configuration.simulink_implementation_marshaled_file_name src_efname in
  write_simulink_implementation_marshaled_file dst_efname code;
  if not (Configuration.get_compile_only ()) then
    translate_simulink_implementation_marshaled_file dst_efname
;;

let write_companion_translated_source_file dst_efname src_efname =
  Path.with_explicit_out_file dst_efname (fun oc ->
    let ppf = Format.formatter_of_out_channel oc in
    Format.fprintf ppf "%a@."
      M_to_host.translate_file src_efname)
;;

let translate_matlab_source_file src_efname =
  let dst_efname = Configuration.matlab_translated_file_name src_efname in
  write_companion_translated_source_file dst_efname src_efname
;;

let translate_mtlb_source_file src_efname =
  let dst_efname = Configuration.mtlb_translated_file_name src_efname in
  write_companion_translated_source_file dst_efname src_efname
;;

let write_sime_translated_source_file dst_efname src_efname =
  Path.with_explicit_out_file dst_efname (fun oc ->
    let ppf = Format.formatter_of_out_channel oc in
    Format.fprintf ppf "%a@."
      Mbd_to_target.print src_efname)
;;

let translate_sime_source_file src_efname =
  let dst_efname = Configuration.sime_translated_file_name src_efname in
  write_sime_translated_source_file dst_efname src_efname
;;

let compile src_fname =
  let src_efname = Path.find src_fname in

  if Configuration.is_simulink_interface_source_file_name src_fname
    then compile_simulink_interface src_efname else
  if Configuration.is_simulink_implementation_source_file_name src_fname
    then compile_simulink_implementation src_efname else
  if Configuration.is_simulink_implementation_marshaled_file_name src_fname
    then translate_simulink_implementation_marshaled_file src_efname else

  if Configuration.is_mtlb_source_file_name src_fname
    then translate_mtlb_source_file src_efname else
  if Configuration.is_sime_source_file_name src_fname
    then translate_sime_source_file src_efname else

  if Configuration.is_matlab_source_file_name src_fname
    then translate_matlab_source_file src_efname else

  Say.fatal_error
    (Printf.sprintf "do not know what to do with file %S" src_fname)
;;

(*
 Local Variables:
  compile-command: "cd ../..; make"
  End:
*)
