(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: arguments.ml,v 1.1 2019-05-10 14:09:17 jpc Exp $ *)

let get_usage () =
  let soft = Configuration.get_software_name () in
  Printf.sprintf
    "Usage: %s [options] <filename>\
   \n  options are:"
    soft
;;

let print_version () =
  let soft = Configuration.get_software_name ()
  and version = Configuration.get_software_version () in
  Printf.eprintf "%s, version %s\n"
    soft version;
  exit 0;
;;

let print_full_version () =
  let full_version = Configuration.get_software_full_version () in
  Printf.eprintf "%s\n" full_version;
  exit 0;
;;

let print_interface_version () =
  Printf.eprintf "%s\n" (Configuration.get_interface_version ());
  exit 0;
;;

let set_general_debug () =
  Configuration.set_debug ();
  Configuration.set_debug_parsing ();
  Xml_configuration.set_debug ();
  Configuration.set_debug_slx ()
;;

let executable_file = Sys.argv.(0);;
let which_dir = Filename.dirname executable_file;;
let install_root_dir = Filename.concat which_dir Filename.parent_dir_name;;

let print_which () =
  Printf.eprintf "%s\n" which_dir;
  exit 0;
;;
let print_where () =
  Printf.eprintf "%s\n" (Filename.concat install_root_dir "lib");
  exit 0;
;;

let parse_args () =
  Arg.parse_argv Sys.argv [
    ("-c", Arg.Unit Configuration.set_compile_only,
     ": set the compile only mode");
    ("-i", Arg.String Configuration.set_source_file,
     ": set the source file name");
    ("-o", Arg.String Configuration.set_target_file,
     ": set the target file name");
    ("-I", Arg.String Path.push,
     "<directory-name>:\
    \n   add the directory name argument to the search path for files");
    ("-tl", Arg.String Configuration.set_target_language,
     "<lang>: shorthand for the -target-language option.");
    ("-target-language", Arg.String Configuration.set_target_language,
     "<lang>: translate compiled file to <lang> source file\
    \n  (the default target language is Nsp).\
    \n   Available target languages are:\
    \n   - ``nsp'' to translate to the Nsp language,\
    \n   - ``scicoslab'' (or ``sci'') to translate to the ScicosLab language\
    \n   - ``hypermath'' (or ``hml'') to translate to HyperMath,\
    \n   - ``octave'' (or ``oct'') to translate to Octave,\
    \n   - ``matlab'' (or ``m'') to translate to Matlab,\
    \n   - ``e-scicos'' (or ``cose'') to translate to\
    \n       Scicos embedded C-like expressions.");
    ("-tr", Arg.String Configuration.set_translator,
     "<trans>: shorthand for the -translator option");
    ("-translator", Arg.String Configuration.set_translator,
     "<trans>: use translator <trans> to compile the source file argument\
    \n  (the default translator is mdl translator).\
    \n   Available translators are:\
    \n   - mdl (to translate Simulink models to Scicos or Simulate),\
    \n   - mtlb (to translate MatLab programs to Nsp, ScicosLab, etc),\
    \n   - sime (to translate Simulink embedded C-like expressions\
    \n     to Nsp, ScicosLab, etc.)");
    ("-tms", Arg.String Configuration.set_target_modeler,
     "<modeling-system>: shorthand for the -target-modeling-system option");
    ("-target-modeling-system", Arg.String Configuration.set_target_modeler,
     "<modeling-system>: set <modeling-system> as target block modeling system\
    \n  (the default target block modeling system is scicos)\
    \n   Available block modeling systems are:\
    \n   - scicos (to translate to traditional Scicos),\
    \n   - scicos-pro, vss, or activate (to translate to Activate),\
    \n   - modelicos (to translate to a modelica block library).");
    ("-sv", Arg.String Configuration.set_target_modeler,
     "<scicos-variant>: OBSOLETE OPTION, use option -tms instead.");
    ("-ccf", Arg.String Configuration.add_companion_source_file,
     ": add a Matlab companion source file name");
    ("-ocf", Arg.String Configuration.add_companion_target_file,
     ": add a Matlab companion target file name");
    ("-mn", Arg.String Configuration.set_slx_model_ident,
     "<model-name>: a shorthand for the -model-name option.");
    ("-model-name", Arg.String Configuration.set_slx_model_ident,
     "<name>: set the model name of an XML or slx block diagram source file");
    ("-ignore-encoding", Arg.Unit Configuration.set_ignore_encoding,
     ": disable characters conversion\
    \n  (source and target encoding are considered identical). ");
    ("-source-encoding", Arg.String Configuration.set_source_encoding,
     "<encoding>: set encoding for the source file argument\
    \n  (the default source encoding is ISO-8859-1).\
    \n  Available source encodings are:\
    \n  - US-ASCII (American Standard Code for Information Interchange),\
    \n  - ISO-8859-15 (Latin-9),\
    \n  - ISO-8859-1 (Latin-1),\
    \n  - Windows-1252 (a.k.a ANSI),\
    \n  - UTF-8 (Universal character set Transformation Format 8-bit).");
    ("-target-encoding", Arg.String Configuration.set_target_encoding,
     "<encoding>: set encoding for the target file.\
    \n  Default target encodings are:\
    \n  - ISO-8859-1 (Latin-1) for ScicosLab,\
    \n  - UTF-8 for HyperMath and Nsp.");
    ("-verbose", Arg.Unit Configuration.set_verbose,
     ": set the verbose mode");
    ("-v", Arg.Unit Configuration.set_verbose,
     ": set the verbose mode");
    ("-nw", Arg.Unit Configuration.set_no_warnings,
     ": a shorthand for the -no-warnings option");
    ("-no-warnings", Arg.Unit Configuration.set_no_warnings,
     ": set the no warnings mode");
    ("-d", Arg.Unit Configuration.set_debug,
     ": fire debugging mode");
    ("-debug", Arg.Unit set_general_debug,
     ": trigger all debugging flags");
    ("-dparse", Arg.Unit Configuration.set_debug_parsing,
     ": trigger file parsing debugging");
    ("-dxml", Arg.Unit Xml_configuration.set_debug,
     ": trigger XML debugging (parsing and expanding)");
    ("-dslx", Arg.Unit Configuration.set_debug_slx,
     ": trigger SLX debugging; in particular, all files extracted from slx\
    \n   source archive are kept in a separate directory");
    ("-where", Arg.Unit print_where,
     ": print absolute directory of simport library.");
    ("-which", Arg.Unit print_which,
     ": print absolute directory of simport executable.");
    ("-version", Arg.Unit print_version,
     ": print the simport version");
    ("--version", Arg.Unit print_full_version,
     ": print the full simport version");
    ("-interface-version", Arg.Unit print_interface_version,
     ": print the version of model interfaces compatible with\
    \n  this version of simport");
  ]
  Configuration.set_source_file
  (get_usage ())
;;

let parse_and_get_source_file () =
  Path.init (Configuration.get_search_path ());
  parse_args ();
  Configuration.init_and_check ();
  Configuration.get_source_file ()
;;

(*
 Local Variables:
  compile-command: "cd ../..; make"
  End:
*)
