(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: configuration.ml,v 1.1 2019-05-10 14:09:17 jpc Exp $ *)

open Simport_configuration;;

type message = string;;

type error =
  | Command_line_error of message
;;

exception Error of error;;

let command_line_error message = raise (Error (Command_line_error message));;

let print_version v = Format.fprintf Format.std_formatter "%s" v;;
let print_full_version = print_version;;
let print_interface_version = print_version;;

let simport_name = "simport";;
let simport_version = "1.4";;
let simport_full_version = "1.4.0";;
let simport_interface_version = "1.0.0";;

let mdl_source_file_extension = Path.to_file_extension ".mdl"
and slx_source_file_extension = Path.to_file_extension ".slx"

and nsp_source_file_extension = Path.to_file_extension ".nsp"
and scicoslab_source_file_extension = Path.to_file_extension ".sci"
and matlab_source_file_extension = Path.to_file_extension ".m"
and octave_source_file_extension = Path.to_file_extension ".oct"
and hypermath_source_file_extension = Path.to_file_extension ".hml"
and oml_source_file_extension = Path.to_file_extension ".oml"
and mtlb_source_file_extension = Path.to_file_extension ".mtlb"
and modelica_source_file_extension = Path.to_file_extension ".mo"

and scicos_mbd_source_file_extension = Path.to_file_extension ".cose"
and sime_source_file_extension = Path.to_file_extension ".sime"

and xml_source_file_extension = Path.to_file_extension ".xml"

and mdl_interface_source_file_extension = Path.to_file_extension ".mdli"
;;

let implementation_target_file_extension = Path.to_file_extension ".mdlo"
and interface_target_file_extension = Path.to_file_extension ".mdlio"
;;

let mdl_implementation_source_file_extension = mdl_source_file_extension
and slx_implementation_source_file_extension = slx_source_file_extension
and xml_implementation_source_file_extension = xml_source_file_extension

and simulink_interface_source_file_extension =
  mdl_interface_source_file_extension
and simulink_companion_source_file_extension =
  matlab_source_file_extension
;;

let simulink_implementation_marshaled_file_extension =
  implementation_target_file_extension
and simulink_interface_marshaled_file_extension =
  interface_target_file_extension
;;

let sime_translated_file_extension = scicos_mbd_source_file_extension;;

let sime_target_file_extension = sime_translated_file_extension;;

let simulink_implementation_translated_file_extensions =
  ".nsp, .sci, .hml, .oml, .oct, or .mo"
and simulink_source_file_extensions =
  ".mdl, .mdli, .mdlo, .xml, .slx or .m"
and mtlb_translated_file_extensions =
  ".sci, .nsp, .hml, .oml, .oct, .mo, or .m"
and mtlb_source_file_extensions = ".mtlb, or .m"
;;

let source_file_extension_of_mdl_target_language = function
  | ScicosLab -> scicoslab_source_file_extension
  | Nsp -> nsp_source_file_extension
  | Modelica -> modelica_source_file_extension
  | Hypermath -> hypermath_source_file_extension
  | Oml -> oml_source_file_extension
  | Octave -> octave_source_file_extension
;;

let source_file_extension_of_mtlb_target_language = function
  | Matlab -> matlab_source_file_extension
  | Scicos_host lang -> source_file_extension_of_mdl_target_language lang
;;

let source_file_extension_of_sime_target_language = function
  | E_scicos -> scicos_mbd_source_file_extension
;;

let source_file_extension_of_target_language = function
  | Target_simulink lang ->
    source_file_extension_of_mdl_target_language lang
  | Target_mtlb lang ->
    source_file_extension_of_mtlb_target_language lang
  | Target_sime lang ->
    source_file_extension_of_sime_target_language lang
;;

let source_file_extension_of_mdl_source_language = function
  | Mdl -> mdl_implementation_source_file_extension
  | Slx -> slx_implementation_source_file_extension
  | Xml -> xml_implementation_source_file_extension
;;

let source_file_extension_of_mtlb_source_language = function
  | Mtlb -> octave_source_file_extension
;;

let source_file_extension_of_sime_source_language = function
  | E_simulink -> sime_source_file_extension
;;

let source_file_extension_of_source_language = function
  | Source_simulink lang ->
    source_file_extension_of_mdl_source_language lang
  | Source_mtlb lang ->
    source_file_extension_of_mtlb_source_language lang
  | Source_sime lang ->
    source_file_extension_of_sime_source_language lang
;;

let source_language_of_string =
  let simulink_source_language_of_string = function
    | "mdl" | "simulink" -> Mdl
    | "xml" -> Xml
    | "slx" -> Slx
    | _ -> raise Not_found

  and sime_source_language_of_string = function
    | "e-simulink" | "embedded-simulink" | "sime" -> E_simulink
    | _ -> raise Not_found

  and mtlb_source_language_of_string = function
    | "mtlb" -> Mtlb
    | _ -> raise Not_found in

  (function lang ->
   try Source_simulink (simulink_source_language_of_string lang) with
   | Not_found ->
     try Source_sime (sime_source_language_of_string lang) with
     | Not_found ->
       try Source_mtlb (mtlb_source_language_of_string lang) with
       | Not_found ->
         command_line_error (Printf.sprintf
            "Unknown source language %s"
            lang))
;;

let string_of_simulink_source_language = function
  | Mdl -> "mdl"
  | Xml -> "xml"
  | Slx -> "slx"
;;

let string_of_mtlb_source_language = function
  | Mtlb -> "mtlb"
;;

let string_of_sime_source_language = function
  | E_simulink -> "embedded-simulink"
;;

let string_of_source_language = function
  | Source_simulink simulink_source_language ->
    string_of_simulink_source_language simulink_source_language
  | Source_mtlb mtlb_source_language ->
    string_of_mtlb_source_language mtlb_source_language
  | Source_sime sime_source_language ->
    string_of_sime_source_language sime_source_language
;;

let target_language_of_string =

  let simulink_target_language_of_string = function
    | "scicoslab" | "sci" -> ScicosLab
    | "nsp" -> Nsp
    | "modelica" -> Modelica
    | "hypermath" | "hml" -> Hypermath
    | "oml" -> Oml
    | "octave" | "oct" -> Octave
    | _ -> raise Not_found

  and sime_target_language_of_string = function
    | "e-scicos" | "embedded-scicos" | "cose" -> E_scicos
    | _ -> raise Not_found in

  let mtlb_target_language_of_string = function
    | "matlab" | "m" -> Matlab
    | lang ->
      try Scicos_host (simulink_target_language_of_string lang) with
      | Not_found ->
        command_line_error (Printf.sprintf
          "``%s'' is not a valid target language for matlab translation"
          lang) in

  (function lang ->
   try Target_simulink (simulink_target_language_of_string lang) with
   | Not_found ->
     try Target_sime (sime_target_language_of_string lang) with
     | Not_found ->
       Target_mtlb (mtlb_target_language_of_string lang))
;;

let string_of_simulink_target_language = function
  | ScicosLab -> "scicoslab"
  | Nsp -> "nsp"
  | Modelica -> "modelica"
  | Hypermath -> "hypermath"
  | Oml -> "oml"
  | Octave -> "octave"
;;

let string_of_mtlb_target_language = function
  | Matlab -> "matlab"
  | Scicos_host lang -> string_of_simulink_target_language lang
;;

let string_of_sime_target_language = function
  | E_scicos -> "e-scicos"
;;

let string_of_target_language = function
  | Target_simulink lang ->
    string_of_simulink_target_language lang
  | Target_mtlb lang ->
    string_of_mtlb_target_language lang
  | Target_sime lang ->
    string_of_sime_target_language lang
;;

let string_of_target_modeler = function
  | Simport_configuration.Modeler_scicos ->
    "scicos"
  | Simport_configuration.Modeler_scicos_pro ->
    "scicos_pro"
  | Simport_configuration.Modeler_modelicos ->
    "modelicos"
;;

let simulink_pervasives_interface = Ident.Ident "simport_pervasives";;

(** Defaults *)
let default_slx_model_ident = Ast_types.ident_from_string "BlockDiagram";;

let default_source_language = Source_simulink Mdl;;
let default_target_modeler = Modeler_scicos;;
let default_source_encoding = Uencoding_types.ISO_8859_1;;
let available_encodings = "\
  \n  - US-ASCII (American Standard Code for Information Interchange),\
  \n  - ISO-8859-15 (Latin-9),\
  \n  - ISO-8859-1 (Latin-1),\
  \n  - Windows-1252 (a.k.a ANSI),\
  \n  - UTF-8 (UCS transformation format 8 bits).\
  \n  - Shift-JIS (Shift Japanese Industrial Standards)."
;;

let default_scicos_host_language = Nsp;;
let default_scicos_pro_host_language = Oml;;
let default_modelicos_host_language = Modelica;;

let default_target_language = Target_simulink ScicosLab;;

let default_translator = T_simulink;;

let natural = Unatural_types.to_natural;;
let length = natural;;

let default_mean_blocks_by_system = natural 59
and default_mean_block_bindings = Unatural._7_
and default_mean_block_parameters = Unatural._3_
and default_max_block_types = natural 101
;;

let default_slx_simulink_dir_name = "simulink"
and default_slx_blockdiagram_base_name = "blockdiagram.xml"
and default_slx_block_diagram_defaults_base_name = "bddefaults.xml"
and default_slx_config_set_info_base_name = "configSetInfo.xml"
and default_slx_graphical_interface_base_name = "graphicalInterface.xml"
and default_slx_stateflow_base_name = "stateflow.xml"
;;

let mk_slx_simulink_path base_name =
  Filename.concat
    default_slx_simulink_dir_name base_name
;;

let default_slx_blockdiagram_file_name =
  mk_slx_simulink_path default_slx_blockdiagram_base_name
and default_slx_block_diagram_defaults_file_name =
  mk_slx_simulink_path default_slx_block_diagram_defaults_base_name
and default_slx_config_set_info_file_name =
  mk_slx_simulink_path default_slx_config_set_info_base_name
and default_slx_graphical_interface_file_name =
  mk_slx_simulink_path default_slx_graphical_interface_base_name
and default_slx_stateflow_file_name =
  mk_slx_simulink_path default_slx_stateflow_base_name
;;

let default_font = {
  Ast_types.
  mdl_font_name = "Helvetica";
  mdl_font_angle = "normal";
  mdl_font_size = natural 10;
  mdl_font_weight = "normal";
}
;;

let default_source_file = Path.builtin_input_file
and default_target_file = Path.builtin_output_file
and default_path = Path.builtin_search_path
;;

let default_target_model_ident = Ast_types.ident_from_string "scsm";;

let conf = {
  Simport_configuration.
  software_name = simport_name;
  software_version = simport_version;
  software_full_version = simport_full_version;
  interface_version = simport_interface_version;

  compile_only = false;

  verbose = false;
  debug = false;
  debug_slx = false;
  warnings = true;

  search_path = default_path;

  slx_model_ident = None;

  ignore_encoding = false;

  source_file = None;
  source_encoding = None;
  companion_source_files = [];

  target_file = None;
  target_encoding = None;
  companion_target_files = [];

  translator = None;

  interfaces = [];

  source_language = None;
  target_language = None;
  target_modeler = None;

(* FIXME: to be removed ?
  window_width = length 1920;
  window_height = length 1200;

  diagram_left_margin = length 10;
  diagram_right_margin = length 450;
  diagram_bottom_margin = length 10;
*)
  default_font = default_font;

  mean_blocks_by_system = default_mean_blocks_by_system;
  mean_block_bindings = default_mean_block_bindings;
  max_block_types = default_max_block_types;
  mean_block_parameters = default_mean_block_parameters;

  block_vertical_spacing = length 50;
  block_port_width = length 15;
  block_port_height = length 10;
  block_event_width = length 20;
  block_event_height = length 15;
  block_geometry_magnification = natural 100;

  not_yet_translated_mark = "?";

  target_model_ident = default_target_model_ident;
}
;;

let translator_of_string = function
  | "mdl" | "simulink" | "Simulink"
  | "xml" | "slx" -> T_simulink
  | "mtlb" | "matlab" | "Matlab" -> T_mtlb
  | "sime" | "e-simulink" | "simulink-embedded-expression" -> T_sime
  | trans ->
    command_line_error (Printf.sprintf
      "translator %S is unknown" trans)
;;

let string_of_translator = function
  | T_simulink -> "simulink"
  | T_mtlb -> "mtlb"
  | T_sime -> "sime"
;;

let get_option_with_default default = function
  | None -> default
  | Some x -> x
;;

let get_translator () =
  get_option_with_default default_translator conf.translator

and set_translator trans =
  match conf.translator with
  | Some tr ->
    command_line_error (Printf.sprintf
      "translator has already been set to \"%s\";\
     \n  could not set it again to \"%s\""
      (string_of_translator tr) trans)
  | None ->
    let tr = translator_of_string trans in
    conf.translator <- Some tr
;;

let translator_of_source_file_extension ext =
  match Path.of_file_extension ext with
  | ".mdl" | ".mdli" | ".mdlo"
  | ".xml" | ".slx" -> Some T_simulink
  | ".mtlb" | ".m" -> Some T_mtlb
  | ".sime" -> Some T_sime
  | _ -> None
;;

let translator_of_target_file_extension ext =
  match Path.of_file_extension ext with
  | ".sci" | ".nsp" | ".mo" | ".hml"
  | ".mdlo" | ".mdlio" -> Some T_simulink
  | ".m" | ".oct" -> Some T_mtlb
  | ".cose" -> Some T_sime
  | _ -> None
;;

let translator_of_companion_file_extension ext =
  match Path.of_file_extension ext with
  | ".m" -> Some T_mtlb
  | _ -> None
;;

let translator_of_source_file_name efname =
  translator_of_source_file_extension (Path.get_extension efname)
;;

let translator_of_target_file_name efname =
  translator_of_target_file_extension (Path.get_extension efname)
;;

let translator_of_companion_file_name efname =
  translator_of_companion_file_extension (Path.get_extension efname)
;;

let translator_of_companion_files = function
  | [] -> None
  | fname :: _ ->
    translator_of_companion_file_name fname
;;

let get_target_language () =
  get_option_with_default default_target_language conf.target_language

and set_target_language lang =
  match conf.target_language with
  | Some tl ->
    command_line_error (Printf.sprintf
      "target language has already been set to \"%s\";\
     \n  could not set it again to \"%s\""
      (string_of_target_language tl) lang)
  | None ->
    let tl = target_language_of_string lang in
    conf.target_language <- Some tl
;;

let get_simulink_target_language () =
  match get_target_language () with
  | Target_simulink simulink_target_language -> simulink_target_language
  | Target_mtlb _
  | Target_sime _ ->
    command_line_error "cannot find the Simulink target language"
;;

let get_source_language () =
  get_option_with_default default_source_language conf.source_language

and set_source_language lang =
  match conf.source_language with
  | Some lg ->
    command_line_error (Printf.sprintf
      "source language has already been set to \"%s\";\
     \n  could not set it again to \"%s\""
      (string_of_source_language lg) lang)
  | None ->
    let lg = source_language_of_string lang in
    conf.source_language <- Some lg
;;

let get_target_modeler () =
  get_option_with_default
    default_target_modeler conf.target_modeler

and set_target_modeler =

  let target_modeler_of_string = function
    | "scicos" -> Modeler_scicos
    | "activate" | "scicos-pro" | "vss" -> Modeler_scicos_pro
    | "modelicos" -> Modeler_modelicos
    | _ -> raise Not_found

  and string_of_target_modeler = function
  | Modeler_scicos -> "scicos"
  | Modeler_scicos_pro -> "activate"
  | Modeler_modelicos -> "modelicos" in

  (function modeler_name ->
    match conf.target_modeler with
    | Some tmod ->
      command_line_error (Printf.sprintf
        "target modeling system language has already been set to \"%s\";\
       \n  could not set it again to \"%s\""
        (string_of_target_modeler tmod) modeler_name)
    | None ->
      let tmod =
        try target_modeler_of_string modeler_name with
        | Not_found ->
          command_line_error (Printf.sprintf
            "Unknown target modeling system %s"
            modeler_name) in
      conf.target_modeler <- Some tmod)
;;

let get_interfaces () = List.rev conf.interfaces;;
let add_interface int_id = conf.interfaces <- int_id :: conf.interfaces;;

let add_simulink_pervasives_interface () =
  add_interface simulink_pervasives_interface
;;

let get_search_path () = conf.search_path
and set_search_path p = conf.search_path <- p
;;

let set_compile_only, get_compile_only =
  (fun () -> conf.compile_only <- true),
  (fun () -> conf.compile_only)
;;

let set_verbose, get_verbose =
  (fun () -> conf.verbose <- true),
  (fun () -> conf.verbose)
;;

let set_debug, get_debug =
  (fun () -> conf.debug <- true),
  (fun () -> conf.debug)
;;

let set_debug_slx, get_debug_slx =
  (fun () -> conf.debug_slx <- true),
  (fun () -> conf.debug_slx)
;;

let set_warnings, get_warnings =
  (fun () -> conf.warnings <- true),
  (fun () -> conf.warnings)
;;

let set_no_warnings, get_no_warnings =
  (fun () -> conf.warnings <- false),
  (fun () -> not conf.warnings)
;;

let set_default_font, get_default_font =
  (fun fnt -> conf.default_font <- fnt),
  (fun () -> conf.default_font)
;;

let set_debug_parsing () = ignore (Parsing.set_trace true);;

let get_software_name () = conf.software_name;;
let get_software_version () = conf.software_version;;
let get_software_full_version () = conf.software_full_version;;
let get_interface_version () = conf.interface_version;;

let warning s =
  if get_warnings () then
    Format.eprintf "@[<hov 2>%s warning:@ %s@]@."
      (get_software_name ()) s
;;

let fdebug f =
  if get_debug () then
    Format.eprintf "@[<hv 2>%s debug:@ %t@]@."
      (get_software_name ()) f
;;

let fdebug_call fname f x =
  fdebug (fun epf -> Format.fprintf epf "Starting %s" fname);
  let y = f x in
  fdebug (fun epf -> Format.fprintf epf "Ending %s" fname);
  y
;;

let debug s = fdebug (fun epf -> Format.fprintf epf "%s" s);;

let get_ignore_encoding () = conf.ignore_encoding

and set_ignore_encoding () = conf.ignore_encoding <- true
;;

let is_source_encoding_set () =
  match conf.source_encoding with
  | None -> false
  | Some _ -> true
;;

let get_source_encoding () =
  get_option_with_default default_source_encoding conf.source_encoding
;;

let is_command_line_target_file_set () =
  match conf.target_file with
  | None -> false
  | Some _ -> true
;;

let is_target_encoding_set () = conf.target_encoding = None;;

let get_encoding_for_simulink_target_language = function
  | ScicosLab | Octave -> Uencoding_types.ISO_8859_1
  | Nsp | Modelica | Hypermath | Oml -> Uencoding_types.UTF_8
;;

let get_encoding_for_mtlb_target_language = function
  | Matlab -> Uencoding_types.ISO_8859_1
  | Scicos_host lang -> get_encoding_for_simulink_target_language lang
;;

let get_encoding_for_sime_target_language = function
  | E_scicos -> Uencoding_types.ISO_8859_1
;;

let get_encoding_for_target_language () =
  match get_target_language () with
  | Target_simulink lang -> get_encoding_for_simulink_target_language lang
  | Target_mtlb lang -> get_encoding_for_mtlb_target_language lang
  | Target_sime lang -> get_encoding_for_sime_target_language lang
;;

let get_target_encoding () =
  match conf.target_encoding with
  | None -> get_encoding_for_target_language ()
  | Some cli_encoding -> cli_encoding
;;

let get_slx_model_ident () =
  match conf.slx_model_ident with
  | None ->
    begin
      match conf.source_file with
      | None -> default_slx_model_ident
      | Some fname ->
        let extension = Path.get_extension fname in
        Ast_types.ident_from_any_string
          (Path.chop_extension (Filename.basename fname) extension)
    end
  | Some slx_model_ident -> slx_model_ident
;;

let get_companion_source_files () = conf.companion_source_files;;

let get_companion_target_files () = conf.companion_target_files;;

let set_slx_model_ident slx_model_ident =
  match conf.slx_model_ident with
  | Some old_slx_model_ident ->
    command_line_error (Printf.sprintf
      "model ident has already been set to \"%s\";\
     \n  could not set it again to \"%s\""
      (Ast_types.string_of_ident old_slx_model_ident)
      slx_model_ident)
  | None ->
    conf.slx_model_ident <- Some (Ast_types.ident_from_string slx_model_ident)
;;

let get_target_file () =
  get_option_with_default default_target_file conf.target_file
;;

let get_source_file () =
  get_option_with_default default_source_file conf.source_file

and set_source_file fname =
  match conf.source_file with
  | Some src_name ->
    command_line_error (Printf.sprintf
      "source file has already been set to \"%s\";\
     \n  could not set it again to \"%s\""
      src_name fname)
  | None ->
    let tgt_file = get_target_file () in
    if fname = tgt_file && tgt_file <> Path.builtin_output_file then
      command_line_error (Printf.sprintf
        "target file \"%s\" \
       \n  would overwrite source file \"%s\""
        tgt_file fname) else
    conf.source_file <- Some fname
;;

let set_source_encoding encoding_name =
  match conf.source_encoding with
  | Some encoding ->
    command_line_error (Printf.sprintf
      "source encoding has already been set to \"%s\";\
     \n  could not set it again to \"%s\""
      (Wstring.encoding_name_of_character_encoding encoding)
      encoding_name)
  | None ->
    try
      let encoding =
        Wstring.character_encoding_of_encoding_name encoding_name in
      conf.source_encoding <- Some encoding with
    | Failure _ ->
      command_line_error (Printf.sprintf
        "unknown source encoding \"%s\";\
       \n  available encodings are:\
       \n  %s"
        encoding_name available_encodings)
;;

let set_target_encoding encoding_name =
  match conf.target_encoding with
  | Some encoding ->
    command_line_error (Printf.sprintf
      "target encoding has already been set to \"%s\";\
     \n  could not set it again to \"%s\""
      (Wstring.encoding_name_of_character_encoding encoding) encoding_name)
  | None ->
    try
      let encoding =
        Wstring.character_encoding_of_encoding_name encoding_name in
      conf.target_encoding <- Some encoding with
    | Failure _ ->
      command_line_error (Printf.sprintf
        "unknown target encoding \"%s\";\
       \n  available encodings are:\
       \n  %s"
        encoding_name available_encodings)
;;

let encoding_of_encoding_name encoding_name =
  let encoding =
    match encoding_name with
    | "ISO-8859-1" -> Uencoding_types.ISO_8859_1
    | "ISO-8859-15" -> Uencoding_types.ISO_8859_15
    | "US-ASCII" -> Uencoding_types.US_ASCII
    | "UTF-8" -> Uencoding_types.UTF_8
    | "Shift_JIS" -> Uencoding_types.SHIFT_JIS
    | "windows-1252" -> Uencoding_types.CP_1252
    | ("windows-1250" |
       "windows-1251" |
       "windows-1254" |
       "windows-1255" |
       "windows-1257" |
       "windows-874"
      ) ->
      warning (Printf.sprintf
        "encoding ``%s'' is not supported. Using Windows-1252"
        encoding_name);
      Uencoding_types.CP_1252
    | ("Big5" |
       "GBK" |
       "ibm-1386_P100-2002" |
       "ibm-5347_P100-1998" |
       "ibm-5348_P100-1997" |
       "ibm-5352_P100-1998" |
       "ibm-943_P14A-2000" |
       _
      ) ->
      warning (Printf.sprintf
        "encoding ``%s'' is not supported. Using ISO_8859_15"
        encoding_name);
      Uencoding_types.ISO_8859_15 in
   debug (Printf.sprintf
     "Using encoding ``%s''"
     (Wstring.encoding_name_of_character_encoding encoding));
   encoding
;;

let fix_source_encoding encoding_name =
  debug (Printf.sprintf "fix_source_encoding %s" encoding_name);
  match conf.source_encoding with
  | Some cli_encoding ->
    let source_encoding = encoding_of_encoding_name encoding_name in
    if cli_encoding <> source_encoding then
    warning (Printf.sprintf
      "source file declared character encoding ``%s'' was ignored\
     \nsince it conflicts with command line option encoding ``%s''."
      encoding_name
      (Wstring.encoding_name_of_character_encoding cli_encoding));
  | None ->
    conf.source_encoding <- Some (encoding_of_encoding_name encoding_name)
;;

let fix_target_encoding target_language_encoding_name =
  let target_language_encoding =
    encoding_of_encoding_name target_language_encoding_name in
  debug (Printf.sprintf
    "fix_target_encoding %s" target_language_encoding_name);
  match conf.target_encoding with
  | Some cli_encoding ->
    if cli_encoding <> target_language_encoding then
      warning (Printf.sprintf
        "target file declared character encoding ``%s'' was ignored\
       \n  since it conflicts with command line option encoding ``%s''."
        target_language_encoding_name
        (Wstring.encoding_name_of_character_encoding cli_encoding));
  | None -> conf.target_encoding <- Some target_language_encoding
;;

let set_target_file fname =
  match conf.target_file with
  | Some tgt_name ->
    command_line_error (Printf.sprintf
      "target file has already been set to \"%s\";\
     \n  could not set it again to \"%s\""
      tgt_name fname)
  | None ->
    let src_file = get_source_file () in
    if fname = src_file && src_file <> Path.builtin_input_file then
      command_line_error (Printf.sprintf
        "target file \"%s\" \
       \n  would overwrite source file \"%s\""
        src_file fname) else
    conf.target_file <- Some fname;
;;

let add_companion_source_file fname =
  let source_fnames = get_companion_source_files () in
  if List.mem fname source_fnames then
    command_line_error (Printf.sprintf
      "file ``%s'' has already been recorded as a companion source file;\
     \n  cannot add it again"
      fname) else
  let target_fnames = get_companion_target_files () in
  if List.mem fname target_fnames then
    command_line_error (Printf.sprintf
      "file ``%s'' has already been recorded as a companion target file;\
     \n  cannot add it as companion source file"
      fname) else
  conf.companion_source_files <- fname ::  source_fnames
;;

let add_companion_target_file fname =
  let source_fnames = get_companion_source_files () in
  if List.mem fname source_fnames then
    command_line_error (Printf.sprintf
      "file ``%s'' has already been recorded as a companion source file;\
     \n  cannot add it as a target companion file"
      fname) else
  let target_fnames = get_companion_target_files () in
  if List.mem fname target_fnames then
    command_line_error (Printf.sprintf
      "file ``%s'' has already been recorded as a companion target file;\
     \n  cannot add it again"
      fname) else
  conf.companion_target_files <- fname :: target_fnames
;;

let get_target_model_ident () = conf.target_model_ident
and set_target_model_ident s =
  let id = Ast_types.ident_from_string s in
  conf.target_model_ident <- id
;;

let get_mean_blocks_by_system () = conf.mean_blocks_by_system;;
let get_mean_block_bindings () = conf.mean_block_bindings;;
let get_max_block_types () = conf.max_block_types;;
let get_mean_block_parameters () = conf.mean_block_parameters;;

(* FIXME: should be suppressed

let get_window_width () = conf.window_width;;
let get_window_height () = conf.window_height;;

let get_diagram_left_margin () = conf.diagram_left_margin;;
let get_diagram_right_margin () = conf.diagram_right_margin;;
let get_diagram_bottom_margin () = conf.diagram_bottom_margin;;

let get_block_vertical_spacing () = conf.block_vertical_spacing;;

FIXME: end should be suppressed *)

let get_block_port_width () = conf.block_port_width;;
let get_block_port_height () = conf.block_port_height;;
let get_block_geometry_magnification () = conf.block_geometry_magnification;;
let get_block_event_width () = conf.block_event_width;;
let get_block_event_height () = conf.block_event_height;;

let set_block_port_width w = conf.block_port_width <- w;;
let set_block_port_height h = conf.block_port_height <- h;;
let set_block_event_width w = conf.block_event_width <- w;;
let set_block_event_height h = conf.block_event_height <- h;;
let set_block_geometry_magnification m =
  conf.block_geometry_magnification <- m
;;

let get_not_yet_translated_mark () = conf.not_yet_translated_mark;;

let add_source_file_extension_of_target_language efname =
  let target_extension =
    source_file_extension_of_target_language (get_target_language ()) in
  Path.change_explicit_extension efname target_extension
;;

(*let add_mdl_implementation_source_extension fname =
  Path.add_extension mdl_implementation_source_file_extension fname
;;

let add_target_language_source_file_extension =
  Path.add_extension conf.target_language_source_file_extension
;;
let add_implementation_translated_file_extension =
  add_target_language_source_file_extension
;;

let add_mtlb_translated_file_extension =
  add_target_language_source_file_extension
;;
let add_sime_translated_file_extension =
  add_target_language_source_file_extension
;;
*)

let is_mdl_implementation_source_file_name fname =
  Path.check_extension fname mdl_implementation_source_file_extension
and is_xml_implementation_source_file_name fname =
  Path.check_extension fname xml_implementation_source_file_extension
and is_slx_implementation_source_file_name fname =
  Path.check_extension fname slx_implementation_source_file_extension
and is_mdl_implementation_source_explicit_file_name efname =
  Path.check_explicit_extension efname mdl_implementation_source_file_extension
and is_xml_implementation_source_explicit_file_name efname =
  Path.check_explicit_extension efname xml_implementation_source_file_extension
and is_slx_implementation_source_explicit_file_name efname =
  Path.check_explicit_extension efname slx_implementation_source_file_extension

and is_simulink_interface_source_file_name fname =
  Path.check_extension fname simulink_interface_source_file_extension
and is_simulink_companion_source_file_name fname =
  Path.check_extension fname matlab_source_file_extension

and is_matlab_source_file_name fname =
  Path.check_extension fname matlab_source_file_extension
and is_matlab_source_explicit_file_name efname =
  Path.check_explicit_extension efname matlab_source_file_extension
and is_mtlb_source_file_name fname =
  Path.check_extension fname mtlb_source_file_extension
and is_mtlb_source_explicit_file_name efname =
  Path.check_explicit_extension efname mtlb_source_file_extension
and is_sime_source_file_name fname =
  Path.check_extension fname sime_source_file_extension
and is_sime_source_explicit_file_name efname =
  Path.check_explicit_extension efname sime_source_file_extension
;;

let is_simulink_implementation_source_file_name fname =
  is_mdl_implementation_source_file_name fname ||
  is_xml_implementation_source_file_name fname ||
  is_slx_implementation_source_file_name fname
and is_simulink_implementation_source_explicit_file_name efname =
  is_mdl_implementation_source_explicit_file_name efname ||
  is_xml_implementation_source_explicit_file_name efname ||
  is_slx_implementation_source_explicit_file_name efname
;;

let is_simport_source_file_name fname =
  is_simulink_interface_source_file_name fname ||
  is_simulink_companion_source_file_name fname ||
  is_simulink_implementation_source_file_name fname

and is_mtlb_translator_source_file_name fname =
  is_matlab_source_file_name fname ||
  is_mtlb_source_file_name fname
and is_mtlb_translator_source_explicit_file_name efname =
  is_matlab_source_explicit_file_name efname ||
  is_mtlb_source_explicit_file_name efname
;;

let is_simulink_implementation_marshaled_file_name fname =
  Path.check_extension fname simulink_implementation_marshaled_file_extension
and is_simulink_implementation_marshaled_explicit_file_name efname =
  Path.check_explicit_extension
    efname simulink_implementation_marshaled_file_extension
and is_simulink_interface_marshaled_file_name fname =
  Path.check_extension fname simulink_interface_marshaled_file_extension
;;

let is_simulink_implementation_translated_file_name fname =
  Path.check_extension fname scicoslab_source_file_extension ||
  Path.check_extension fname nsp_source_file_extension ||
  Path.check_extension fname modelica_source_file_extension ||
  Path.check_extension fname hypermath_source_file_extension ||
  Path.check_extension fname oml_source_file_extension ||
  Path.check_extension fname octave_source_file_extension

and is_sime_translated_file_name fname =
  Path.check_extension fname sime_translated_file_extension

and is_mtlb_translated_file_name fname =
  Path.check_extension fname scicoslab_source_file_extension ||
  Path.check_extension fname nsp_source_file_extension ||
  Path.check_extension fname modelica_source_file_extension ||
  Path.check_extension fname hypermath_source_file_extension ||
  Path.check_extension fname oml_source_file_extension ||
  Path.check_extension fname octave_source_file_extension ||
  Path.check_extension fname matlab_source_file_extension
;;

let get_output_file_name find_output src_efname =
  match conf.target_file with
  | None -> find_output src_efname
  | Some fname ->
    let fname =
      if Path.is_explicit fname then
        fname
      else
        Filename.concat "." fname in
    Path.to_explicit_file_name fname
;;

let simulink_implementation_translated_file_name =
  get_output_file_name (fun efname ->
    if not (is_simulink_implementation_source_explicit_file_name efname) &&
       not (is_simulink_implementation_marshaled_explicit_file_name efname)
    then
      failwith (Printf.sprintf
        "%s is not a valid Simulink implementation file name"
        (Path.of_explicit_file_name efname));
    add_source_file_extension_of_target_language efname)
;;

let matlab_translated_file_name =
  get_output_file_name (fun efname ->
    if not (is_matlab_source_explicit_file_name efname) then
      failwith (Printf.sprintf
        "%s is not a valid matlab source file name"
        (Path.of_explicit_file_name efname));
    add_source_file_extension_of_target_language efname)
;;

let mtlb_translated_file_name =
  get_output_file_name (fun efname ->
    if not (is_mtlb_translator_source_explicit_file_name efname) then
      failwith (Printf.sprintf
        "%s is not a valid mtlb source file name"
        (Path.of_explicit_file_name efname));
    add_source_file_extension_of_target_language efname)
;;

let sime_translated_file_name =
  get_output_file_name (fun efname ->
    if not (is_sime_source_explicit_file_name efname) then
      failwith (Printf.sprintf
        "%s is not a valid simulink_expression source file name"
        (Path.of_explicit_file_name efname));
    Path.change_explicit_extension efname sime_translated_file_extension)
;;

let simulink_interface_marshaled_file_name efname =
  Path.change_explicit_extension
    efname simulink_interface_marshaled_file_extension
;;

let simulink_implementation_marshaled_file_name efname =
  Path.change_explicit_extension
    efname simulink_implementation_marshaled_file_extension
;;

(* Consistency tests. *)

(* Target languages. *)

let default_sime_target_language = E_scicos;;
let default_simulink_target_language = Nsp;;
let default_mtlb_target_language = Scicos_host Nsp;;

let check_sime_target_language tl_option =
  match tl_option with
  | None ->
    conf.target_language <- Some (Target_sime default_sime_target_language)
  | Some tl ->
    match tl with
    | Target_simulink _ | Target_mtlb _ ->
      command_line_error (Printf.sprintf
        "sime target language should be E_scicos, \
       \n  it can not be %s"
        (string_of_target_language tl))
    | Target_sime E_scicos -> ()
;;

let check_simulink_target_language tm_option tl_option =
  let simulink_target_language_for_modeler tm_option =
    match tm_option with
    | None -> default_simulink_target_language
    | Some tm ->
      match tm with
      | Modeler_scicos -> default_scicos_host_language
      | Modeler_scicos_pro -> default_scicos_pro_host_language
      | Modeler_modelicos -> default_modelicos_host_language in
  match tl_option with
  | None ->
    let tl = simulink_target_language_for_modeler tm_option in
    conf.target_language <- Some (Target_simulink tl)
  | Some tl ->
    match tl with
    | Target_simulink _ -> ()
    | Target_mtlb (
        Matlab
      | Scicos_host (
          ScicosLab | Nsp | Modelica | Hypermath | Oml | Octave)
      ) -> ()
    | Target_sime _ ->
      command_line_error (Printf.sprintf
        "simulink target language can not be %s"
        (string_of_target_language tl))
;;

let rec check_mtlb_target_language tl_option =
  match tl_option with
  | None ->
    conf.target_language <- Some (Target_mtlb default_mtlb_target_language)
  | Some tl ->
    match tl with
    | Target_simulink lang ->
      check_mtlb_target_language (Some (Target_mtlb (Scicos_host lang)))
    | Target_sime _ ->
      command_line_error (Printf.sprintf
        "mtlb target language should be one of Nsp, HyperMath, Oml, \
       \n  ScicosLab, Matlab, or Octave; it can not be %s"
        (string_of_target_language tl))
    | Target_mtlb _ -> ()
;;

(* Source and target files consistency checks. *)

let check_sime_src_tgt_files sf tf =
  if not (is_sime_source_file_name sf) then
    command_line_error (Printf.sprintf
      "sime source file should have extension %s"
      (Path.string_of_extension sime_source_file_extension)) else
  if not (is_command_line_target_file_set ()) &&
     not (is_sime_translated_file_name tf) then
    command_line_error (Printf.sprintf
      "sime target file should have extension %s"
      (Path.string_of_extension sime_target_file_extension))
;;

let check_simulink_src_tgt_files (sf : Path.file_name) (tf : Path.file_name) =
  if is_simulink_interface_source_file_name sf then
    if not (is_simulink_interface_marshaled_file_name tf) then
      command_line_error (Printf.sprintf
        "mdl compiled interface file should have extension %s"
        (Path.string_of_extension
           simulink_interface_marshaled_file_extension))
    else () else
  if is_simulink_implementation_source_file_name sf ||
    is_simulink_implementation_marshaled_file_name sf ||
    is_simulink_companion_source_file_name sf then
    if not (is_command_line_target_file_set ()) &&
       not (is_simulink_implementation_translated_file_name tf) then
      command_line_error (Printf.sprintf
        "%s is not a valid simulink compiled implementation file name;\
       \n  it should have extension %s"
        tf simulink_implementation_translated_file_extensions) else ()
  else
    command_line_error (Printf.sprintf
      "%s is not a valid simport source file;\
     \n  its extension should be one of %s"
      sf simulink_source_file_extensions)
;;

let check_mtlb_src_tgt_files sf tf =
  if is_mtlb_source_file_name sf ||
     is_matlab_source_file_name sf then
    if not (is_command_line_target_file_set ()) &&
       not (is_mtlb_translated_file_name tf) then
      command_line_error (Printf.sprintf
        "%s is not a valid mtlb translated file name;\
       \n  its extension should be one of %s."
        tf mtlb_translated_file_extensions) else ()
  else
    command_line_error (Printf.sprintf
      "%s is not a valid mtlb source file;\
     \n  its extension should be one of %s."
      sf mtlb_source_file_extensions)
;;

let fill_simulink_src_tgt_files _sf _tf = ();;

let fill_mtlb_src_tgt_files _sf _tf = ();;

let fill_sime_src_tgt_files _sf _tf = ();;

let set_translation_option opt =
  match conf.translator with
  | None -> conf.translator <- opt
  | Some _ -> assert false
;;

let check_and_fill_translator tr_option =

  match tr_option with
  | None ->
    let tr =
      let src_file_opt = conf.source_file in
      let tgt_file_opt = conf.target_file in
      let companion_source_files = get_companion_source_files () in
      let translator_of_companion_file =
        translator_of_companion_files companion_source_files in
      (match src_file_opt, tgt_file_opt with
       | None, None -> translator_of_companion_file
       | None, Some tgt_name ->
         (match translator_of_target_file_name tgt_name with
          | None -> translator_of_companion_file
          | Some _ as tr -> tr
         )
       | Some src_name, None ->
         (match translator_of_source_file_name src_name with
          | None -> translator_of_companion_file
          | Some _ as tr -> tr
         )
       | Some src_name, Some tgt_name ->
         (match translator_of_source_file_name src_name with
          | None ->
            (match translator_of_target_file_name tgt_name with
             | None -> translator_of_companion_file
             | Some _ as tr -> tr
            )
          | Some tr_src as tr ->
            (match translator_of_target_file_name tgt_name with
             | None -> translator_of_companion_file
             | Some tr_tgt ->
               if tr_src = tr_tgt then tr else translator_of_companion_file
            )
         )
      ) in
    set_translation_option tr
  | Some _tr -> ()
;;

let convert_source_file_string s =
  if is_source_encoding_set () then
    let source_encoding = get_source_encoding () in
    Wstring.string_to_wstring source_encoding s else
  Wstring.string_to_wstring Uencoding_types.UTF_8 s
;;

let init_and_check () =

  add_simulink_pervasives_interface ();

  check_and_fill_translator conf.translator;

  let target_encoding_name =
    let target_encoding = get_encoding_for_target_language () in
    Wstring.encoding_name_of_character_encoding target_encoding in
  fix_target_encoding target_encoding_name;

  let tr = get_translator ()
  and sf = conf.source_file
  and tf = conf.target_file in

  match tr with
  | T_simulink ->
    let tm_option = conf.target_modeler
    and tl_option = conf.target_language in
    check_simulink_target_language tm_option tl_option;
    fill_simulink_src_tgt_files sf tf;
    check_simulink_src_tgt_files (get_source_file ()) (get_target_file ())
  | T_mtlb ->
    check_mtlb_target_language conf.target_language;
    fill_mtlb_src_tgt_files sf tf;
    check_mtlb_src_tgt_files (get_source_file ()) (get_target_file ())
  | T_sime ->
    check_sime_target_language conf.target_language;
    fill_sime_src_tgt_files sf tf;
    check_sime_src_tgt_files (get_source_file ()) (get_target_file ())
;;

let report_error ppf = function
  | Command_line_error s ->
    let soft = get_software_name () in
    Format.fprintf ppf
      "@[<v 5>%s: command line options error %s@]@." soft s
;;

(*
 Local Variables:
  compile-command: "cd ../..; make"
  End:
*)
