(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: configuration.mli,v 1.1 2019-05-10 14:09:17 jpc Exp $ *)

open Ast_types;;

type message = string;;

type error =
  | Command_line_error of message
;;

exception Error of error;;

val report_error : Format.formatter -> error -> unit;;

val print_version : Simport_configuration.version -> unit;;
val print_full_version : Simport_configuration.full_version -> unit;;
val print_interface_version :
  Simport_configuration.interface_version -> unit
;;

val get_target_language : unit -> Simport_configuration.target_language;;
val set_target_language : string -> unit;;

val get_source_language : unit -> Simport_configuration.source_language;;
val set_source_language : string -> unit;;

val get_simulink_target_language :
  unit -> Simport_configuration.simulink_target_language
;;

val get_target_modeler : unit -> Simport_configuration.target_modeler;;
val set_target_modeler : string -> unit;;

val translator_of_source_file_extension :
  Path.file_extension -> Simport_configuration.translator option
;;

val get_translator : unit -> Simport_configuration.translator;;
val set_translator : string -> unit;;

val get_interfaces : unit -> Ident.ident list;;
val add_interface : Ident.ident -> unit;;

val add_simulink_pervasives_interface : unit -> unit;;

val set_compile_only : unit -> unit
(** Set the [compile_only] mode on: the compiler only generates a compiled
  version of the Simulink model after semantic analysis but phase has been
  completed. *)
;;

val get_compile_only : unit -> bool;;

val set_verbose : unit -> unit
(** Set the [verbose] flag on: the program outputs more material. *)
;;
val get_verbose : unit -> bool;;

val get_debug : unit -> bool;;
val set_debug : unit -> unit
(** Set the debug flag on: functions {!debug} and {!fdebug} will output
  debugging material. *)
;;

val get_debug_slx : unit -> bool;;
val set_debug_slx : unit -> unit;;

val get_no_warnings : unit -> bool;;
val set_no_warnings : unit -> unit;;

val get_warnings : unit -> bool;;
val set_warnings : unit -> unit;;

val warning : message -> unit;;

val debug : message -> unit;;

val fdebug : (Format.formatter -> unit) -> unit
(** [fdebug f] apply [f] to a formatter only if debug is set. *)
;;

val fdebug_call : string -> ('a -> 'b) -> 'a -> 'b
(** [fdebug_call f_name f x] returns [f x] while surrounding the call with
  debug messages indicating that function [f_name] is called. *)
;;

val set_debug_parsing : unit -> unit;;

val get_slx_model_ident : unit -> Ident.ident;;
val set_slx_model_ident : string -> unit;;

val get_ignore_encoding : unit -> bool;;
val set_ignore_encoding : unit -> unit;;

val get_source_file : unit -> Path.file_name;;
val set_source_file : Path.file_name -> unit;;

val get_source_encoding : unit -> Uencoding_types.character_encoding;;
val set_source_encoding : Uencoding_types.character_encoding_name -> unit;;

val is_source_encoding_set : unit -> bool;;
val fix_source_encoding : Uencoding_types.character_encoding_name -> unit;;

val get_companion_source_files : unit -> Path.file_name list;;
val add_companion_source_file : Path.file_name -> unit;;

val get_target_file : unit -> Path.file_name;;
val set_target_file : Path.file_name -> unit;;

val get_target_encoding : unit -> Uencoding_types.character_encoding;;
val set_target_encoding : Uencoding_types.character_encoding_name -> unit;;

val is_target_encoding_set : unit -> bool;;
val fix_target_encoding : Uencoding_types.character_encoding_name -> unit;;

val get_companion_target_files : unit -> Path.file_name list;;
val add_companion_target_file : Path.file_name -> unit;;

val simulink_interface_marshaled_file_name :
  Path.explicit_file_name ->
  Simport_configuration.interface_marshaled_file_name
;;

val simulink_implementation_marshaled_file_name :
  Path.explicit_file_name ->
  Simport_configuration.implementation_marshaled_file_name
;;
val simulink_implementation_translated_file_name :
  Simport_configuration.implementation_source_file_name ->
  Simport_configuration.implementation_translated_file_name
;;

val matlab_translated_file_name :
  Simport_configuration.matlab_source_file_name ->
  Simport_configuration.matlab_translated_file_name
;;

val mtlb_translated_file_name :
  Simport_configuration.mtlb_source_file_name ->
  Simport_configuration.mtlb_translated_file_name
;;

val sime_translated_file_name :
  Simport_configuration.sime_source_file_name ->
  Simport_configuration.sime_translated_file_name
;;

val is_simulink_interface_source_file_name : Path.file_name -> bool;;
val is_mdl_implementation_source_file_name : Path.file_name -> bool;;
val is_xml_implementation_source_file_name : Path.file_name -> bool;;
val is_slx_implementation_source_file_name : Path.file_name -> bool;;
val is_mdl_implementation_source_explicit_file_name :
  Path.explicit_file_name -> bool;;
val is_xml_implementation_source_explicit_file_name :
  Path.explicit_file_name -> bool;;
val is_slx_implementation_source_explicit_file_name :
  Path.explicit_file_name -> bool;;
val is_simulink_implementation_source_file_name : Path.file_name -> bool;;
val is_simulink_companion_source_file_name : Path.file_name -> bool;;
val is_matlab_source_file_name : Path.file_name -> bool;;
val is_mtlb_source_file_name : Path.file_name -> bool;;
val is_sime_source_file_name : Path.file_name -> bool;;

val is_simport_source_file_name : Path.file_name -> bool;;
val is_mtlb_translator_source_file_name : Path.file_name -> bool;;

val is_simulink_implementation_marshaled_file_name :
  Path.file_name -> bool;;
val is_simulink_interface_marshaled_file_name : Path.file_name -> bool;;

val get_default_font : unit -> mdl_font;;
val set_default_font : Ast_types.mdl_font -> unit;;

val get_software_name : unit -> string;;
val get_software_version : unit -> Simport_configuration.version;;
val get_software_full_version : unit -> Simport_configuration.full_version;;
val get_interface_version : unit -> Simport_configuration.interface_version;;

val get_search_path : unit -> Path.search_path;;
val set_search_path : Path.search_path -> unit;;

(* FIXME: should be suppressed
val get_window_width : unit -> Ast_types.natural;;
val get_window_height : unit -> natural;;

val get_diagram_left_margin : unit -> natural;;
val get_diagram_right_margin : unit -> natural;;

val get_diagram_bottom_margin : unit -> natural;;
val get_block_vertical_spacing : unit -> natural;;
 FIXME: end should be suppressed *)

val get_block_port_width : unit -> natural;;
val set_block_port_width : natural -> unit;;

val get_block_port_height : unit -> natural;;
val set_block_port_height : natural -> unit;;

val get_block_geometry_magnification : unit -> natural;;
val set_block_geometry_magnification : natural -> unit;;

val get_block_event_width : unit -> natural;;
val set_block_event_width : natural -> unit;;
val get_block_event_height : unit -> natural;;
val set_block_event_height : natural -> unit;;

val get_not_yet_translated_mark : unit -> string;;

val get_target_model_ident : unit -> Ident.ident;;
val set_target_model_ident : string -> unit;;

val string_of_target_language :
  Simport_configuration.target_language -> string
;;

val string_of_target_modeler :
  Simport_configuration.target_modeler -> string
;;

val default_slx_simulink_dir_name : Path.dir_name;;
val default_slx_blockdiagram_base_name : Path.file_name;;
val default_slx_block_diagram_defaults_base_name : Path.file_name;;
val default_slx_config_set_info_base_name : Path.file_name;;
val default_slx_graphical_interface_base_name : Path.file_name;;
val default_slx_stateflow_base_name : Path.file_name;;

val default_slx_blockdiagram_file_name : Path.file_name;;
val default_slx_block_diagram_defaults_file_name : Path.file_name;;
val default_slx_config_set_info_file_name : Path.file_name;;
val default_slx_graphical_interface_file_name : Path.file_name;;
val default_slx_stateflow_file_name : Path.file_name;;

(** Internal usage. *)
val get_mean_blocks_by_system : unit -> natural;;
val get_mean_block_bindings : unit -> natural;;
val get_max_block_types : unit -> natural;;
val get_mean_block_parameters : unit -> natural;;

val xml_source_file_extension : Path.file_extension;;

val simulink_interface_source_file_extension : Path.file_extension;;
val mdl_implementation_source_file_extension : Path.file_extension;;
val xml_implementation_source_file_extension : Path.file_extension;;
val slx_implementation_source_file_extension : Path.file_extension;;
val simulink_companion_source_file_extension : Path.file_extension;;

val matlab_source_file_extension : Path.file_extension;;
val mtlb_source_file_extension : Path.file_extension;;
val sime_source_file_extension : Path.file_extension;;

val simulink_interface_marshaled_file_extension : Path.file_extension;;
val simulink_implementation_marshaled_file_extension : Path.file_extension;;

val source_file_extension_of_source_language :
  Simport_configuration.source_language -> Path.file_extension
;;

(* That function could be used in any converter (Matlab, embedded etc.).
  Hence its definition next to the root of the module dependency graph. *)
val convert_source_file_string : string -> Wstring.wstring
(** [convert_source_file_string str] converts [str] found in a simport source
  file to a string for the target language according to source file and
  target language encodings. *)
;;

val init_and_check : unit -> unit
(** [init_and_check ()] initializes the simport compiler internal variables
  and checks their consistency.
  In particular, it checks that the source file and its language can be
  translated to the specified target file and its language.
  [Configuration.init_and_check] must be called after the command line
  treatment. *)
;;

(*
 Local Variables:
  compile-command: "cd ../..; make"
  End:
*)
