(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: ast_types.ml,v 1.1 2019-05-10 14:09:18 jpc Exp $ *)

type natural = Unatural.t;;

(* Private types *)
type mdl_name =
  | Mdl_name of string
;;

type mdl_path =
  | Mdl_path of string
;;

type mdl_bool =
  | Mdl_bool of string
;;

type mdl_text =
  | Mdl_text of string
;;

type mdl_shallow_code =
  | Mdl_shallow_code of mdl_text
;;

type mdl_code =
  | Mdl_code of mdl_text
;;

type mdl_port_number =
  | Mdl_port_number of natural
;;

(* Public types *)
type mdl_font = {
  mdl_font_name : mdl_font_name;
  mdl_font_angle : string;
  mdl_font_size : natural;
  mdl_font_weight : string;
}

and mdl_font_name = string
;;

let mdl_name_of_string s = Mdl_name s;;

let crlf_normalization s =
  let buffer = Buffer.create (String.length s) in
  let ib = Scanf.Scanning.from_string s in
  let rec loop ib =
    if Scanf.Scanning.end_of_input ib then
      Buffer.contents buffer
    else
      Scanf.bscanf ib "%c" (function
      | '\r' -> loop ib
      | c ->
        Buffer.add_char buffer c;
        loop ib) in
  loop ib
;;

let mdl_path_of_string s = Mdl_path (crlf_normalization s);;

let mdl_bool_of_string = function
  | "on" | "off" as s -> Mdl_bool s
  | s -> failwith (Printf.sprintf "``%s'' is not a valid boolean" s)
;;

let mdl_shallow_code_of_string s = Mdl_shallow_code (Mdl_text s);;
let mdl_code_of_string s = Mdl_code (Mdl_text s);;
let mdl_text_of_string s = Mdl_text s;;

let mdl_port_number_of_natural n =
  if n = Unatural._0_ then
    failwith (Printf.sprintf
      "%d is not a valid mdl port number" (Unatural.to_int n))
  else
    Mdl_port_number n
;;

let mdl_port_number_of_int i =
  (* Could simply be mdl_port_number_of_natural (Unatural.of_int i), but the
     error message would be unclear in case of negative integer. *)
  if i < 1 then
    failwith (Printf.sprintf
      "non positive integer %d is not a valid mdl port number" i)
  else
    Mdl_port_number (Unatural.of_int i)
;;

let bool_of_mdl_string = function
  | "on" -> true
  | "off" -> false
  | s ->
    failwith (Printf.sprintf "``%s'' is not a valid boolean" s)
;;

let bool_of_mdl_bool = function
  | Mdl_bool ("on" | "off" as s) -> bool_of_mdl_string s
  | Mdl_bool _ -> assert false
;;

let string_of_mdl_name (Mdl_name s) = s;;
let string_of_mdl_path (Mdl_path s) = s;;
let string_of_mdl_bool (Mdl_bool s) = s;;
let string_of_mdl_text (Mdl_text s) = s;;
let string_of_ident (Ident.Ident s) = s;;

let natural_of_mdl_port_number (Mdl_port_number n) = n;;
let int_of_mdl_port_number port =
  Unatural.to_int (natural_of_mdl_port_number port)
;;
let string_of_mdl_port_number port =
  string_of_int (int_of_mdl_port_number port)
;;

let mdl_text_of_ident = function
  | Ident.Ident s -> Mdl_text s
;;

let mdl_text_of_mdl_name = function
  | Mdl_name s -> Mdl_text s
;;

let mdl_text_of_mdl_code = function
  | Mdl_code text -> text
;;

let string_of_mdl_shallow_code (Mdl_shallow_code (Mdl_text s)) = s;;
let string_of_mdl_code code =
  string_of_mdl_text (mdl_text_of_mdl_code code)
;;

let mdl_name_of_ident = function
  | Ident.Ident s -> Mdl_name s
;;

let valid_ident_name s =
  let len = String.length s in
  let b = Buffer.create len in
  let rec loop i =
    if i >= len then Buffer.contents b else
    let c =
      match s.[i] with
      | 'a'..'z'
      | 'A'..'Z'
      | '0'..'9'
      | '_' as c -> c
      | _ -> '_' in
    Buffer.add_char b c; loop (succ i) in
  loop 0
;;

let ident_from_string id =
  let ib = Scanf.Scanning.from_string id in
  try
    Scanf.bscanf ib " %[a-zA-Z_0-9] %!" (fun id -> Ident.Ident id)
  with
  | Scanf.Scan_failure s | Failure s ->
    failwith (Printf.sprintf "``%s'' is not a valid ident: %s" id s)
;;

let ident_from_mdl_name (Mdl_name s) = ident_from_string s;;
let ident_from_any_string s = ident_from_string (valid_ident_name s);;

(*
 Local Variables:
  compile-command: "cd ../..; make"
  End:
*)
