(***********************************************************************)
(*                                                                     *)
(*                          Ocaml-xml Library                          *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2013-2018,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: lib_strings.ml,v 1.1 2019-05-10 14:00:49 jpc Exp $ *)

(** {3 A collection of functions for string manipulation} *)

(** {6 Case manipulation for ISO-latin characters characters and string} *)

let lowercase_char c =
  if c >= 'A' && c <= 'Z'
  || c >= '\192' && c <= '\214'
  || c >= '\216' && c <= '\222'
  then char_of_int (int_of_char c + 32)
  else c
;;

let uppercase_char c =
  if c >= 'a' && c <= 'z'
  || c >= '\224' && c <= '\246'
  || c >= '\248' && c <= '\254'
  then char_of_int (int_of_char c - 32)
  else c
;;

let lowercase = String.map lowercase_char
and uppercase = String.map uppercase_char
;;

let capitalize s =
  let len = String.length s in
  if len = 0 then s else
  let c = s.[0] in
  let uc = uppercase_char c in
  if uc = c then s else
  String.make 1 uc ^ String.sub s 1 (len - 1)

and uncapitalize s =
  let len = String.length s in
  if len = 0 then s else
  let c = s.[0] in
  let lc = lowercase_char c in
  if lc = c then s else
  String.make 1 lc ^ String.sub s 1 (len - 1)
;;

(** {6 String searching} *)

(** Test if, starting from index [idx], the string [s] matches the
   pattern [pat] *)
let starts_by s idx pat =
  let ls = String.length s
  and lpat = String.length pat in
  let lim = idx + lpat in
  if idx < 0 || ls > 0 && idx >= ls then invalid_arg "starts_by" else
  let rec starts i j =
    i >= lim || s.[i] = pat.[j] && starts (i + 1) (j + 1) in
  lim <= ls && starts idx 0
;;

(** Find the index of character [c] in string [s] starting from index [idx],
   which should be a valid index into [s], and looking forward to the
   end of string [s]. *)
let pos_string s idx c =
  let lim = String.length s - 1 in
  if idx < 0 || idx > lim then invalid_arg "pos_string" else
  let rec search i =
    if i > lim then raise Not_found else
    if s.[i] = c then i else search (i + 1) in
  search idx
;;

(** Find the index of character [c] in string [s] starting from index [idx],
   which should be a valid index into [s], and looking backward to the
   beginning of string [s]. *)
let pos_string_rev s idx c =
  let lim = String.length s - 1 in
  if idx < 0 || idx > lim then invalid_arg "pos_string_rev" else
  let rec search i =
    if i < 0 then raise Not_found else
    if s.[i] = c then i else search (i - 1) in
  search idx
;;

(** {6 String extraction} *)

(** Returns the sub-string of [s] starting at index [i_start]
   (inclusively) to index [i_stop] (excluded). *)
let extract_string s i_start i_stop =
  let len = i_stop - i_start in
  String.sub s i_start len
;;

(** Find the sub string of [s] that starts by character [starter] and
   ends by [stopper], starting from valid index [idx]. *)
let sub_string_from_to s idx starter stopper =
  let i_start =
    try pos_string s idx starter with
    | Not_found ->
      failwith (Printf.sprintf "Cannot find %c in %s" starter s) in
  let i_stop =
    try pos_string s (i_start + 1) stopper with
    | Not_found ->
      failwith (Printf.sprintf "Cannot find %c in %s" stopper s) in
  extract_string s (i_start + 1) i_stop
;;

(** Returns the sub string starting from the position following the
   last occurrence of character [c] to the end of [s].
   If [c] is not found, [s] is returned *)
let from_char_to s c idx =
  let len = String.length s in
  if idx >= len then invalid_arg "from_char_to" else
  let rec search i =
    if i < 0 then s else
    if s.[i] = c then extract_string s (i + 1) len
    else search (i - 1) in
  search (idx - 1)
;;

let from_char_to_end s c = from_char_to s c (String.length s - 1);;

let from_to_char s idx c =
  let len = String.length s in
  if idx >= len || idx < 0 then invalid_arg "from_char_to" else
  let rec search i =
    if i >= len then String.sub s idx (len - idx) else
    if s.[i] = c then extract_string s idx i
    else search (i + 1) in
  search idx
;;

let split_string_at_comma = fun s ->
  let ib = Scanf.Scanning.from_string s in
  let rec loop accu =
    if Scanf.Scanning.end_of_input ib then List.rev accu else
    Scanf.bscanf ib ",%[^,]"
      (fun string_part -> loop (string_part :: accu)) in
  try
    Scanf.bscanf ib "%[^,]"
      (fun string_part -> loop [ string_part ])
  with
  | Scanf.Scan_failure error
  | Failure error ->
    failwith
      (Printf.sprintf
         "split_string_at_comma, \
          while scanning ``%s'', encountered error %s"
         s
         error)
;;

(** {6 String alignment} *)

let not_enough_room funname len_s len_in =
  failwith (Printf.sprintf
    "Lib_strings.%s: cannot fit a string of length %d \
     into a string of length %d"
    funname len_s len_in)
;;

let buffer_enclose_string b c s idx len =
  for _i = 0 to pred idx do
    Buffer.add_char b c;
  done;
  Buffer.add_string b s;
  for _i = idx + String.length s to pred len do
    Buffer.add_char b c;
  done;
;;

let string_enclose_string c s idx len =
  let b = Buffer.create len in
  buffer_enclose_string b c s idx len;
  Buffer.contents b
;;

let center_gen c s len_in =
  let len_s = String.length s in
  if len_in < len_s then not_enough_room "center" len_s len_in else
  let idx = (len_in - len_s) / 2 in
  string_enclose_string c s idx len_in
;;

let center s = center_gen ' ' s;;

let flush_left_gen c s len_in =
  let len_s = String.length s in
  if len_in < len_s then not_enough_room "flush_left" len_s len_in else
  let idx = 0 in
  string_enclose_string c s idx len_in
;;

let flush_left s = flush_left_gen ' ' s;;

let flush_right_gen c s len_in =
  let len_s = String.length s in
  if len_in < len_s then not_enough_room "flush_right" len_s len_in else
  let idx = len_in - len_s in
  string_enclose_string c s idx len_in
;;

let flush_right s = flush_right_gen ' ' s;;

let spaces n = String.make n ' ';;

(*
Replace all occurrences of multiple spaces by only one space.

let suppress_spurious_spaces txt =
  let ib = Scanf.Scanning.from_string txt in
  let rec loop accu =
    if Scanf.Scanning.end_of_input ib then List.rev accu else
    Scanf.bscanf ib "%[^ ]%[ ]"
      (fun no_spaces spaces ->
       let l  = String.length spaces in
       if l >= 1
       then loop (" " :: no_spaces :: accu)
       else loop (no_spaces :: accu)) in
  String.concat "" (loop [])
;;
*)

let suppress_spaces txt =
  let ib = Scanf.Scanning.from_string txt in
  let rec loop accu =
    if Scanf.Scanning.end_of_input ib then List.rev accu else
    Scanf.bscanf ib "%[^ \t]%_[ \t]"
      (fun no_spaces -> loop (no_spaces :: accu)) in
  String.concat "" (loop [])
;;

(* Input to string. *)

let input_string_fmt len =
  Scanf.format_from_string
    (Printf.sprintf "%%%d%,[\000-\255]" len)
    "%s"
;;

let input_string len =
  let fmt = input_string_fmt len in
  fun ic ->
    let ib = Scanf.Scanning.from_channel ic in
    if Scanf.Scanning.end_of_input ib then raise End_of_file else
    Scanf.bscanf ib fmt (fun x -> x)
;;

let default_input_size = 2048;;
let default_input_fmt = input_string_fmt default_input_size;;

let iter_channel f ic =
  let ib = Scanf.Scanning.from_channel ic in
  let rec loop () =
    if not (Scanf.Scanning.end_of_input ib) then
      begin
        Scanf.bscanf ib default_input_fmt f;
        loop ();
      end; in
  loop ();
  Scanf.Scanning.close_in ib;
;;

let buffer_add_channel b ic = iter_channel (Buffer.add_string b) ic;;

let buffer_of_ic ic =
  let len = max default_input_size (in_channel_length ic) in
  if len > Sys.max_string_length then
    failwith "buffer_of_ic: too long input channel" else
  let b = Buffer.create len in
  buffer_add_channel b ic;
  b
;;

let string_of_ic ic = Buffer.contents (buffer_of_ic ic);;
