(***********************************************************************)
(*                                                                     *)
(*                          Ocaml-xml Library                          *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2013-2018,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: location.mli,v 1.1 2019-05-10 14:00:49 jpc Exp $ *)

type file_name = Path.file_name;;

type position = Lexing.position = {
  pos_fname : file_name; (* File name *)
  pos_lnum : int; (* Line number *)
  pos_bol : int; (* Beginning of lexem character position *)
  pos_cnum : int; (* Current lexing character position *)
}
;;

type t = {
 loc_start : position;
 loc_end : position;
}
(** The location of an AST node,
  beginning and ending position of its corresponding source text. *)
;;

val dummy : t
(** A dummy location. *)
;;

val mk_loc : position -> position -> t
(** [mk_loc loc_start loc_end] returns the location of the text source between
  [loc_start] and [loc_end]. *)
;;

val get_lexbuf_loc : Lexing.lexbuf -> t
(** [get_lexbuf_loc lexbuf] gets the current location of [lexbuf]. *)
;;

val get_symbol_loc : unit -> t
(** [get_symbol_loc ()] returns the location of current matched grammar rule.
  See also {!Parsing.symbol_start_pos}. *)
;;

val get_rhs_loc : int -> t
(** [get_rhs_loc n] returns the location of the [n]th item on the right-hand
  side of the current matched grammar rule.
  [n] is [1] for the leftmost item.
  See also {!Parsing.rhs_start_pos}. *)
;;

val set_lexbuf_filename_positions : Lexing.lexbuf -> file_name -> unit
(** [set_lexbuf_filename_positions lexbuf fname] set the file name component of
  [lexbuf] to [fname]. *)
;;

val init : Lexing.lexbuf -> Path.file_name -> unit
(** [init lexbuf fname] initializes the lexing buffer positions to the
  beginning of file [fname]. *)
;;

val lexing_from_file : file_name -> Pervasives.in_channel * Lexing.lexbuf
(** Open a file and return an input channel and a lexbuf reading from it. *)
;;

val lexing_from_explicit_file :
  Path.explicit_file_name -> Pervasives.in_channel * Lexing.lexbuf
(** Open an explicit file and return an input channel and a lexbuf reading from
  it. *)
;;

val print_token_location : Format.formatter -> (position * position) -> unit
(** [print_token_location ppf start_pos end_pos] print the location of the
  characters between [start_pos] (inclusive) and [end_pos] (exclusive). *)
;;

val print_location : Format.formatter -> t -> unit
(** [print_location ppf loc]  prints the location [loc], including the file
  name of the location, in the format required by emacs for error messages.
*)
;;

(* Location_print.print_t is
val print_t : Format.formatter -> t -> unit
(** [print_t ppf loc] print the location of the
  characters between [start_pos] (inclusive) and [end_pos] (exclusive),
  including file name. *)
;;
*)

(*
 Local Variables:
  compile-command: "cd ../..; make"
  End:
*)
