(***********************************************************************)
(*                                                                     *)
(*                          Ocaml-xml Library                          *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2013-2018,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: uint8.ml,v 1.1 2019-05-10 14:00:49 jpc Exp $ *)

(** Implementing one byte unsigned natural numbers. *)

type uint8 = char;;

type saturation =
   | Bounds
   | Ring
   | Error
;;

let get_saturation, set_saturation =
  let sat = ref Bounds in
  (fun () -> !sat),
  (fun saturation -> sat := saturation)
;;

let size = 8;;
let int_max_int = 0xFF;;
let int_min_int = 0;;

let max_int = char_of_int int_max_int;;
let min_int = char_of_int int_min_int;;

let zero = min_int;;
let one = '\001';;

let check_range i =
  if i >= 0 && i < size then
  failwith (Printf.sprintf "Uint8: index %d is not in range [0..%d]" i size)
;;

(*
let pmodulo n m =
  let p = n mod m in
  if p >= 0 then p else p + m
;;
*)

let saturate n =
  match get_saturation () with
  | Bounds ->
    if n < int_min_int then min_int else
    if n > int_max_int then max_int else char_of_int n
  | Ring ->
    char_of_int (n land int_max_int)
  | Error ->
    if n < int_min_int || n > int_max_int
    then failwith (Printf.sprintf "Uint8: result %d out of bound" n)
    else char_of_int n
;;

let nth c i = (1 lsl i) land (int_of_char c);;

let nth_bit c i =
  check_range i;
  nth c i
;;

let bits_of c =
  let rec loop accu i =
    if i >= size then accu else
    loop (nth c i :: accu) (i + 1) in
  loop [] 0
;;

let add c1 c2 =
  let i1 = int_of_char c1 and i2 = int_of_char c2 in
  saturate (0xFF land (i1 + i2))
;;

let sub c1 c2 =
  let i1 = int_of_char c1 and i2 = int_of_char c2 in
  char_of_int (0xFF land (i1 - i2))
;;

let mul c1 c2 =
  let i1 = int_of_char c1 and i2 = int_of_char c2 in
  char_of_int (0xFF land (i1 * i2))
;;

let div c1 c2 =
  let i1 = int_of_char c1 and i2 = int_of_char c2 in
  char_of_int (i1 / i2)
;;

let rem c1 c2 =
  let i1 = int_of_char c1 and i2 = int_of_char c2 in
  char_of_int (i1 mod i2)
;;

let succ c = add c one;;
let pred c = sub c one;;

let logand c1 c2 =
  let i1 = int_of_char c1 and i2 = int_of_char c2 in
  char_of_int (i1 land i2)
;;

let logor c1 c2 =
  let i1 = int_of_char c1 and i2 = int_of_char c2 in
  char_of_int (i1 lor i2)
;;

let lognot c = char_of_int (lnot (int_of_char c))
;;

let shift_left c s =
  let i = int_of_char c in
  char_of_int (i lsl s)
;;

(*
let shift_right c s =
  let i = int_of_char c in
  char_of_int (i asr s)
;;
*)

let shift_right_logical c s =
  let i = int_of_char c in
  char_of_int (i lsr s)
;;

let of_int = char_of_int;;
let to_int = int_of_char;;

let of_float f = char_of_int (int_of_float f);;
let to_float c = float_of_int (int_of_char c);;

let of_string s = char_of_int (int_of_string s);;
let to_string c = string_of_int (int_of_char c);;

let bits_of_float f = char_of_int (Int32.to_int (Int32.bits_of_float f));;
let float_of_bits c = Int32.float_of_bits (Int32.of_int (int_of_char c));;

type t = uint8;;

let compare c1 c2 = int_of_char c1 - int_of_char c2;;

(*
 Local Variables:
  compile-command: "cd ../..; make"
  End:
*)
