(***********************************************************************)
(*                                                                     *)
(*                          Ocaml-xml Library                          *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2013-2018,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: ulist.mli,v 1.1 2019-05-10 14:00:49 jpc Exp $ *)

(** Additional functionality for lists. *)

type 'a t = 'a list;;

val flat_map : ('a -> 'b list) -> 'a list -> 'b list
(** [ flat_map f l ] maps function [f] on list [l], and flatten the
    list of results. [flat_map f l] is thus equivalent to
    [List.flatten (List.map f l)], although this implementation of
    [flat_map] is tail recursive.
*)
;;

val rev_flat_map : ('a -> 'b list) -> 'a list -> 'b list
(** Same as [ flat_map ] but the result is in reverse order.
    [rev_flat_map] is tail recursive. *)
;;

val rev_flat_map_end : ('a -> 'b list) -> 'a list -> 'b list -> 'b list
(** Same as [ flat_map ], but [ rev_flat_map_end f l accu ]
  accumulates the results in reverse order in list [ accu ].
  [rev_flat_map_end] is tail recursive. *)
;;

val fold_map :
  ('env -> 'a -> 'env * 'b) ->
  ('env -> 'a list -> 'env * 'b list)
(** [fold_map f env l] maps [f] on list [l] while accumulating data in
  [env]. More precisely,
  [ fold_map f env [a1; ...; an] ] is equivalent to
  [
    let env, b1 = f env a1 in
    ...
    let env, bn = f env an in

    env, [ b1; ...; bn]
  ]
*)
;;

val fold_flat_map :
  ('env -> 'a -> 'env * 'b list) ->
  ('env -> 'a list -> 'env * 'b list)
(** [fold_flat_map f env l] collects list results obtained by
  flat mapping [f] on list [l] while accumulating data in [env].
  More precisely,
  [ fold_flat f (env, [a1; ...; an]) ] is equivalent to
  [
    let env, b1 = f (env, a1) in
    ...
    let env, bn = f (env, an) in

    env, (b1 @ ... @ bn)
  ]
*)
;;

val rev_fold_flat_map :
  ('env -> 'a -> 'env * 'b list) ->
  ('env -> 'a list -> 'env * 'b list)

val rev_fold_flat_map_end :
  ('env -> 'a -> 'env * 'b list) ->
  ('env -> 'a list -> 'b list -> 'env * 'b list)
;;

(*
val array_fold_map :
  ('env * 'a -> 'env * 'b) ->
  ('env * 'a array -> 'env * 'b array)
(** [array_fold_map] is the equivalent of [fold_map] for arrays.
  More precisely,
  [ array_fold_map f (env * v) ] is equivalent to
  [
    let env, list = fold_map f (env, Array.to_list v) in
    env, Array.of_list list
  ]
*)
;;

val array_fold_flat :
  ('env -> 'a -> 'env * 'b list) ->
  ('env * 'a array -> 'env * 'b array)
(** [array_fold_flat] is the equivalent of [fold_flat] for arrays.
  More precisely,
  [ array_fold_flat f (env * v) ] is equivalent to
  [
    let env, list = fold_flat f (env, Array.to_list v) in
    env, Array.of_list list
  ]
*)
;;
*)

(*
 Local Variables:
  compile-command: "cd ../..; make"
  End:
*)
