(***********************************************************************)
(*                                                                     *)
(*                          Ocaml-xml Library                          *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2013-2018,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: ulist2.mli,v 1.1 2019-05-10 14:00:49 jpc Exp $ *)

(** The module defining lists with at least two elements. *)

type 'a list2 = 'a * 'a Ulist1.t;;

type 'a t = 'a list2;;

val length : 'a list2 -> int
(** Return the length (number of elements) of the given list. *)
;;

val cons : 'a -> 'a list2 -> 'a list2;;

val to_list : 'a list2 -> 'a list
(** [to_list l] returns the (usual) list of all the elements of the
  nonempty list [l]. *)
;;

val of_list : 'a list -> 'a list2
(** [of_list l] returns the [list2] list containing the elements of
   [l].
   Raise [Invalid_argument] if the list has not at least two elements. *)
;;

val of_list1 : 'a Ulist1.t -> 'a list2
(** [of_list1 l] returns the [list2] list containing the elements of
   [l].
   Raise [Invalid_argument] if the list has not at least two elements. *)
;;

val rev_append : 'a list2 -> 'a list -> 'a list2
(** Concatenate a nonempty list to a list in reverse order.
   Tail-recursive
   (length of the first argument). *)
;;

val rev : 'a list2 -> 'a list2
(** [rev (a1, [a2; ...; an])] is [ (an, [...; a2; a1])] *)
;;

val iter : ('a -> unit) -> 'a list2 -> unit
(** [iter f [a1; ...; an]] applies function [f] in turn to
   [a1; ...; an]. It is equivalent to
   [begin f a1; f a2; ...; f an; () end]. *)
;;

val map : ('a -> 'b) -> 'a list2 -> 'b list2
(** [map f [a1; ...; an]] applies function [f] to [a1, ..., an],
   and builds the list [[f a1; ...; f an]]
   with the results returned by [f]. Not tail-recursive. *)
;;

val fold_left : ('a -> 'b -> 'a) -> 'a -> 'b list2 -> 'a
(** [fold_left f a (b1, [b2; ...; bn])] is
   [f (... (f (f a b1) b2) ...) bn]. *)
;;

(*
val foldl : ('a -> 'a -> 'a) -> 'a list2 -> 'a
(** [foldl f (a1, [a2; ...; an])] is
   [f (... (f (f a1 a2) a3 ...)) an].
    [foldl f (a1, [])] is [a1]. *)
;;
*)

val fold_right : ('a -> 'b -> 'b) -> 'a list2 -> 'b -> 'b
(** [fold_right f (a1, [a2; ...; an]) b] is
   [f a1 (f a2 (... (f an b) ...))]. Not tail-recursive. *)
;;

(*
val foldr : ('a -> 'a -> 'a) -> 'a * 'a list2 -> 'a;;
(**
 [foldr f (a, (a1, a2, [a3; ...; an-1; an]))] is
   [f a (f a1 (f a2 (... (f an-1 an) ...)))].
  Not tail-recursive. *)
*)

val fold_map : ('a -> 'b -> 'a * 'c) ->
  ('a -> 'b list2 -> 'a * 'c list2)
;;

val fold_flat_map : ('a -> 'b -> 'a * 'c list) ->
  ('a -> 'b list2 -> 'a * 'c list)
;;

val iter2 : ('a -> 'b -> unit) -> 'a list2 -> 'b list2 -> unit
(** [iter2 f [a1; ...; an] [b1; ...; bn]] calls in turn
   [f a1 b1; ...; f an bn].
   Raise [Invalid_argument] if the two lists have
   different lengths. *)
;;

val map2 : ('a -> 'b -> 'c) -> 'a list2 -> 'b list2 -> 'c list2
(** [map2 f [a1; ...; an] [b1; ...; bn]] is
   [[f a1 b1; ...; f an bn]].
   Raise [Invalid_argument] if the two lists have
   different lengths.  Not tail-recursive. *)
;;

val find : ('a -> bool) -> 'a list2 -> 'a
(** [find p l] returns the first element of the list [l]
   that satisfies the predicate [p].
   Raise [Not_found] if there is no value that satisfies [p] in the
   list [l]. *)
;;

val filter : ('a -> bool) -> 'a list2 -> 'a list
(** [filter p l] returns all the elements of the list [l]
   that satisfy the predicate [p].  The order of the elements
   in the input list is preserved.  *)
;;

val partition : ('a -> bool) -> 'a list2 -> 'a list * 'a list
(** [partition p l] returns a pair of lists [(l1, l2)], where
   [l1] is the list of all the elements of [l] that satisfy the predicate [p],
   and [l2] is the list of all the elements of [l] that do not satisfy [p].
   The order of the elements in the input list is preserved. *)
;;

(** {6 Association lists} *)

val max_snd_assoc : ('a -> 'a -> int) -> ('b * 'a) list2 -> 'b * 'a
(** [max_snd_assoc l] returns the leftmost element of the list of pairs
   [l] with the maximum second component. *)
;;

val split : ('a * 'b) list2 -> 'a list2 * 'b list2
(** Transform a list2 of pairs into a pair of list2s:
   [split [(a1, b1); ...; (an, bn)]] is [([a1; ...; an], [b1; ...; bn])].
   Not tail-recursive. *)
;;

val combine : 'a list2 -> 'b list2 -> ('a * 'b) list2
(** Transform a pair of list2s into a list2 of pairs:
   [combine [a1; ...; an] [b1; ...; bn]] is
   [[(a1, b1); ...; (an, bn)]].
   Raise [Invalid_argument] if the two list2s
   have different lengths.  Not tail-recursive. *)
;;

(*
 Local Variables:
  compile-command: "cd ../..; make"
  End:
*)
