(***********************************************************************)
(*                                                                     *)
(*                          Ocaml-xml Library                          *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2013-2018,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: xml_ast.mli,v 1.1 2019-05-10 14:00:49 jpc Exp $ *)

(* The definition of Abstraxt Syntax Trees for XML documents. *)

open Xml_basics_ast;;

type space = string;;
type comment = string;;
type system_literal = string;;
type pubid_literal = string;;
type cdata = string;;
type version_number = string;;
type encoding_name = string;;
type 'a list2 = 'a * 'a Ulist1.t;;

type raw_document =
  {
    prolog : prolog;
    element : element;
    misc_star : misc_star;
  }

and prolog =
  {
    xml_decl : xml_decl option;
    xml_misc_star : misc_star;
    xml_doc_type_decl : doc_type_decl option;
    xml_misc_star_trailing : misc_star;
  }

and xml_decl =
  {
    xml_version_info : version_number;
    xml_encoding_decl : encoding_name option;
    xml_standalone_decl : standalone_decl option;
  }

and standalone_decl =
  | Standalone
  | Not_standalone

and element =
  | Empty_element_tag of tag
  | Content_element_tag of tag * element_content

and tag =
  {
    tag_name : tag_name;
    tag_attributes : tag_attribute list;
  }
and tag_attribute =
  {
    tag_attribute_name : tag_attribute_name;
    tag_attribute_value : att_value;
  }

and char_data =
  | Char_data of char_data_text

and element_content =
  | Element_content of (char_data, tag_child) Intercal.t

and tag_child =
  | Child_element of element
  | Child_cd_sect of cdata
  | Child_pi of pi
  | Child_comment of comment
  | Child_reference of reference

and misc_star = misc list
and misc =
  | Misc_comment of comment
  | Misc_pi of pi
  | Misc_space of space

and pi =
  {
    pi_target : pi_target;
    pi_instructions : Xml_basics_ast.pi_instructions;
  }

and doc_type_decl =
  {
    doctype_name : tag_name;
    external_id : external_id option;
    int_subset : int_subset option;
  }

and external_id =
  | System of system_literal
  | Public of pubid_literal * system_literal

and public_id = pubid_literal

and int_subset = int_subset_elem list
and int_subset_elem =
  | Markup_decl of markup_decl
  | Sep_decl of sep_decl

and sep_decl =
  | Pe_reference_decl of entity_name
  | Space_decl of space

and notation_id =
  | External_id of external_id
  | Public_id of public_id

and markup_decl =
  | Element_decl of tag_name * content_spec
  | Attlist_decl of tag_attribute_name * att_def list
  | Entity_decl of entity_decl
  | Notation_decl of notation_name * notation_id
  | Pi_decl of pi
  | Comment_decl of comment

and att_def =
  {
    att_def_name : tag_attribute_name;
    att_def_type : att_type;
    att_def_default : att_default;
  }

and att_type =
  | String_type
  | Tokenized_type of tokenized_type
  | Enumerated_type of enumerated_type

and tokenized_type =
  | Id_type
  | Idref_type
  | Idrefs_type
  | Entity_type
  | Entities_type
  | Nmtoken_type
  | Nmtokens_type

and enumerated_type =
  | Notation_type of notation_name Ulist1.t
  | Enumeration_type of Xml_basics_ast.nm_token Ulist1.t

and att_default =
  | Required
  | Implied
  | Att_value of Xml_basics_ast.att_value
  | Fixed_att_value of Xml_basics_ast.att_value

and entity_decl =
  | Ge_decl of entity_name * gentity_def
  | Pe_decl of entity_name * pentity_def

and gentity_def =
  | Gentity_value of Xml_basics_ast.entity_value
  | Gentity_external_id of external_id * ndata_decl option

and ndata_decl = notation_name

and pentity_def =
  | Pentity_value of entity_value
  | Pentity_external_id of external_id

and content_spec =
  | Content_spec_empty
  | Content_spec_any
  | Content_spec_mixed of content_spec_mixed
  | Content_spec_children of content_spec_children

and content_particle_choice = content_spec_mod * content_spec_particle list2
and content_particle_seq = content_spec_mod * content_spec_particle Ulist1.t
and content_particle_name = content_spec_mod * tag_name

and content_spec_children =
  | Content_spec_choice of content_particle_choice
  | Content_spec_seq of content_particle_seq

and content_spec_particle =
  | Content_particle_name of content_particle_name
  | Content_particle_choice of content_particle_choice
  | Content_particle_seq of content_particle_seq

and content_spec_mixed =
  | PCData
  | PCData_empty
  | PCData_alternatives of tag_name list
(* TODO: it looks like that it's not a valid rule for XML documents
   | PCData_alternatives_PEReferences of entity_name list
*)

and content_spec_mod =
  | None_mod
  | Opt_mod
  | Star_mod
  | Plus_mod


and raw_ext_subset = text_decl option * ext_subset_decl
and ext_subset_decl = ext_subset_decl_elem list

and ext_subset_decl_elem =
  | Ext_markup_decl of markup_decl
  | Ext_conditional_sect of conditional_sect
  | Ext_sep_decl of sep_decl

and text_decl =
  {
    text_version_info : version_number option;
    text_encoding_decl : encoding_name;
  }

and conditional_sect =
  | Include_sect of ext_subset_decl
  | Ignore_sect of ignore_text
and ignore_text = string
;;

(*

 Local Variables:
  compile-command: "cd ../../..; make"
  End:

*)
