(***********************************************************************)
(*                                                                     *)
(*                          Ocaml-xml Library                          *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2013-2018,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: xml_expand.ml,v 1.1 2019-05-10 14:02:41 jpc Exp $ *)

(** Module to handle external entities *)

let _normalize_public_id pubid =
  let ib = Scanf.Scanning.from_string pubid in
  let b = Buffer.create (String.length pubid) in
  let rec loop first =
    if Scanf.Scanning.end_of_input ib then
      Buffer.contents b
    else
      Scanf.bscanf ib " %[^ \r\t\n]" (fun s ->
        if s = "" then loop first else
          begin
            if not first then
              Buffer.add_char b ' ';
            Buffer.add_string b s;
            loop false
          end) in
  loop true
(** Replace any number of spaces chars by a single space (&#20;).
    And chomp leading and trailing spaces. *)
;;

(* End of module for external entities *)

open Xml_basics_ast;;
open Xml_ast;;

type document = Xml_ast.raw_document;;

type ext_subset = Xml_ast.raw_ext_subset;;

type entity_decl_value =
| Internal of Xml_basics_ast.text
| External of Xml_basics_ast.text option
;;

type entity_env = {
  parameter_env : entity_decl_value Xml_env.Expand_env.t;
  general_env : entity_decl_value Xml_env.Expand_env.t;
  validating : bool;
}
;;

let fatal_error s = failwith (Printf.sprintf "XML expand: %s" s);;

let warning s = Xml_say.warning (Printf.sprintf "XML expand: %s" s);;

let debug s =
  Xml_say.debug ("Xml_expand: " ^ s)
;;

let fdebug f =
  Xml_say.fdebug
    (fun ppf ->
     Format.fprintf ppf "xml_expand: %a" (fun ppf () -> f ppf) ())
;;

let add_parameter_entity env entity_name entitity_decl_value =
  begin
    match entitity_decl_value with
    | Internal _
    | External (Some _) -> ()
    | External None ->
      warning (Printf.sprintf
        "registering external parameter entity %S without its replacement text"
        (Xml_basics_ast.of_xml_name entity_name))
  end;
  let parameter_env =
    Xml_env.Expand_env.add entity_name entitity_decl_value env.parameter_env in
  {
    env with
      parameter_env = parameter_env;
  }
;;

let add_general_entity env entity_name entitity_decl_value =
  begin
    match entitity_decl_value with
    | Internal _
    | External (Some _) -> ()
    | External None ->
      warning (Printf.sprintf
        "registering external general entity %S without its replacement text"
        (Xml_basics_ast.of_xml_name entity_name))
  end;
  let general_env =
    Xml_env.Expand_env.add entity_name entitity_decl_value env.general_env in
  {
    env with
      general_env = general_env;
  }
;;

(*
  insert the predefined entities for all XML documents
  <!ENTITY amp    "&#38;#38;">
  <!ENTITY apos   "&#39;">
  <!ENTITY gt     "&#62;">
  <!ENTITY lt     "&#38;#60;">
  <!ENTITY quot   "&#34;">
*)
let fill_predefined_entities env =
  let add_entitity env (entity_name, entity_value) =
    let entity_name = Xml_basics_ast.to_xml_name entity_name
    and entity_decl_value = Internal entity_value in
    add_general_entity env entity_name entity_decl_value in
  let predefined =
    [
      "amp", "&#38;";
      "apos", "\'";
      "gt", ">";
      "lt", "&#60;";
      "quot", "\"";
    ] in
  List.fold_left add_entitity env predefined
;;

let init_env validating =
  let env = {
    parameter_env = Xml_env.Expand_env.empty;
    general_env = Xml_env.Expand_env.empty;
    validating = validating;
  } in
  fill_predefined_entities env
;;

let find_general_entity_ref env e =
  debug (Printf.sprintf
    "find_general_entity_ref %S" (Xml_basics_ast.of_xml_name e));
  if Xml_env.Expand_env.mem e env.general_env then
    match Xml_env.Expand_env.find e env.general_env with
    | Internal text
    | External (Some text) -> text
    | External None ->
      fatal_error (Printf.sprintf
        "trying to retrieve entity replacement text for \
         general entity %S"
        (Xml_basics_ast.of_xml_name e))
  else
    fatal_error (Printf.sprintf
      "unknown general entity reference %S" (Xml_basics_ast.of_xml_name e))
;;

let find_parameter_entity_ref env e =
  if Xml_env.Expand_env.mem e env.parameter_env then
    match Xml_env.Expand_env.find e env.parameter_env with
    | Internal text
    | External (Some text) -> text
    | External None ->
      fatal_error (Printf.sprintf
        "trying to retrieve entity replacement text for \
         parameter entity %S"
        (Xml_basics_ast.of_xml_name e))
  else
    fatal_error (Printf.sprintf
      "unknown parameter entity reference %S" (Xml_basics_ast.of_xml_name e))
;;

let concat_char_data (Char_data fst_char_data) (Char_data snd_char_data) =
  Char_data (fst_char_data ^ snd_char_data)
;;

let add_char_data_to_element_content char_data = function
  | None, [] -> Some char_data, []
  | Some previous_char_data, [] ->
    Some (concat_char_data previous_char_data char_data), []
  | fst_child, (child, None) :: children ->
    fst_child, (child, Some char_data) :: children
  | fst_child, (child, Some previous_char_data) :: children ->
    fst_child,
    (child, Some (concat_char_data previous_char_data char_data)) :: children
and add_tag_child_to_element_content =
  Intercal.cons_beta
;;

let expand_quoted_string _env s = s;;

let expand_list exp_a env l =
  let env, l =
    List.fold_left
      (fun (env, accu) a -> let env, a = exp_a env a in env, a :: accu)
      (env, []) l in
  let l = List.rev l in
  env, l
;;

let map_ro_to_rw exp exp_a env x =
  let _, x = exp (fun env x -> env, exp_a env x) env x in
  x
;;

let ro_expand_list exp_a env l =
  map_ro_to_rw expand_list exp_a env l
;;

let expand_ulist1 exp_a env (a, l) =
  let env, a = exp_a env a in
  let env, l = expand_list exp_a env l in
  env, (a, l)
;;

let ro_expand_ulist1 exp_a env u_l =
  map_ro_to_rw expand_ulist1 exp_a env u_l
;;

let expand_list2 exp_a env = function
  | (a, a_u_t) ->
    let a = exp_a env a in
    let a_u_t = ro_expand_ulist1 exp_a env a_u_t in
    a, a_u_t
;;

let expand_option exp_a env = function
  | None -> env, None
  | Some a ->
    let env, a = exp_a env a in
    env, Some a
;;

(* Xml_basic_ast *)

let expand_text env s = expand_quoted_string env s;;

let expand_char_data_text env t = expand_text env t;;

let expand_nm_token env s = expand_quoted_string env s;;

let expand_xml_name env x = expand_nm_token env x;;

let expand_pi_target env p = expand_xml_name env p;;

let expand_pi_instructions env t = expand_text env t;;

let expand_tag_name env x = expand_xml_name env x;;

let expand_tag_attribute_name env x = expand_xml_name env x;;

(* FIXME: maybe the core of the expand process *)

(* Not yet used
let get_entity_replacement_text env name =
  let s =
    if Xml_env.Expand_env.mem name env then
      Xml_env.Expand_env.find name env
    else
      fatal_error (
        Printf.sprintf "entity name %S not found"
          (Xml_basics_ast.of_xml_name name)) in
  (s : Xml_basics_ast.text)
;;
*)

(* TODO: We have expand_entity_ref function. Probably we will remove that
   function.

let expand_entity_name mode env entity_name =
  match mode with
  | Pe_ref_Included_in_literal ->
    let replacement_text = get_entity_replacement_text env entity_name in
    (* Should parse and expanse again *)
    failwith replacement_text
  | Pe_ref_Included_as_pe -> failwith "Not yet"
(* Not yet used
  | Pe_ref_Not_recognized ->
    assert false
*)
;;
*)

(* TODO:; maybe it is here where we must warn the user about an unparsed
   entity or register the entity as unparsed. *)
let expand_notation_name env x = env, expand_xml_name env x;;

let expand_att_value_text _env t = t;;

(* TODO: may be we must expand the character reference according to the
   charset. *)
let expand_char_ref _env i =
  let i =
    if i > Int32.of_int 0xFF then
      failwith (Printf.sprintf "out of range character reference (%ld)" i)
    else
      Int32.to_int i in
  let c = char_of_int i in
  String.make 1 c
;;

let intercal_of_char_data_intercal_token = function
  | Xml_parser.CharDataIntercal intercal -> intercal
  | Xml_parser.Eof
  | Xml_parser.OpenConditionalSect
  | Xml_parser.CloseConditionalSect
  | Xml_parser.OpenBracket
  | Xml_parser.CloseBracket
  | Xml_parser.OpenParen
  | Xml_parser.CloseParen
  | Xml_parser.CloseParenQuestion
  | Xml_parser.CloseParenStar
  | Xml_parser.CloseParenPlus
  | Xml_parser.GT
  | Xml_parser.Question
  | Xml_parser.Star
  | Xml_parser.Plus
  | Xml_parser.Eq
  | Xml_parser.Comma
  | Xml_parser.Bar
  | Xml_parser.XmlDeclStart
  | Xml_parser.XmlDeclEnd
  | Xml_parser.ExternalIDSystem _
  | Xml_parser.ExternalIDPublic _
  | Xml_parser.PublicID _
  | Xml_parser.PCDATA
  | Xml_parser.REQUIRED
  | Xml_parser.IMPLIED
  | Xml_parser.FIXED
  | Xml_parser.INCLUDE
  | Xml_parser.Ignore _
  | Xml_parser.VersionInfo _
  | Xml_parser.EncodingDecl _
  | Xml_parser.SDDecl _
  | Xml_parser.NDataDecl _
  | Xml_parser.AttlistName _
  | Xml_parser.NotationTypeName _
  | Xml_parser.Space _
  | Xml_parser.Comment _
  | Xml_parser.NotationName _
  | Xml_parser.GEDeclName _
  | Xml_parser.PEDeclName _
  | Xml_parser.AttValue _
  | Xml_parser.PEReference _
  | Xml_parser.CharRef _
  | Xml_parser.EntityRef _
  | Xml_parser.CData _
  | Xml_parser.FirstSTag _
  | Xml_parser.FirstEmptyElemTag _
  | Xml_parser.STag _
  | Xml_parser.EmptyElemTag _
  | Xml_parser.ETag _
  | Xml_parser.Nmtoken _
  | Xml_parser.DocTypeName _
  | Xml_parser.ElementName _
  | Xml_parser.PI _ ->
    fatal_error "char data expected, but another kind of token read"
;;

let expand_general_entity_ref env e =
  let buffer = Buffer.create 250 in
  let debug_token token =
    fdebug
      (fun ppf ->
       Format.fprintf ppf
         "expand_general_entity_ref: token: %a"
         Xml_parser_print.print_token token) in
  let rec expand_reference = function
    | Char_ref i ->
       let text = expand_char_ref env i in
       begin
         try
           Buffer.add_string buffer text
         with
         | x ->
            prerr_endline (Printf.sprintf
              "trying to add string %S to buffer of entity ref failed: %s"
              text (Printexc.to_string x));
            prerr_endline (Printf.sprintf
              "current buffer length: %d" (Buffer.length buffer));
            prerr_endline (Printf.sprintf
              "current buffer content");
            prerr_endline (Buffer.contents buffer);
            raise x
       end
    | Entity_ref entity_name -> expand_entity_name entity_name
  and expand_entity_name entity_name =
    let text = find_general_entity_ref env entity_name in
    expand_text text
  and expand_text text =
    debug (Printf.sprintf
      "trying to lex %S to find nested entities" text);
    let lexbuf = Lexing.from_string text in
    let intercal =
      let intercal_token =
        Xml_lexer.in_char_data_or_references
          Xml_lexer.Standalone_char_data Intercal.empty lexbuf in
      debug_token intercal_token;
      intercal_of_char_data_intercal_token intercal_token in
    Intercal.iter
      (fun char_data ->
       try
         Buffer.add_string buffer char_data
       with
       | x ->
          prerr_endline (Printf.sprintf
            "trying to add string %S to buffer of entity ref failed: %s"
            char_data (Printexc.to_string x));
          prerr_endline (Printf.sprintf
            "current buffer length: %d" (Buffer.length buffer));
          prerr_endline (Printf.sprintf
            "current buffer content");
          prerr_endline (Buffer.contents buffer);
          raise x)
      expand_reference
      intercal in
  expand_entity_name e;
  Buffer.contents buffer
;;

(* TODO: we have specific functions for each reference type. We don't need a
   general function.

let _expand_reference env mode = function
  | Xml_basics_ast.Char_ref i ->
    env, expand_char_ref mode i
  | Xml_basics_ast.Entity_ref e ->
    expand_entity_ref env mode e
;;

let expand_parameter_entity_ref _env _e : Xml_basics_ast.text =
  failwith "Not yet"
;;
*)

type entity_type =
   | Internal_general
   | External_parsed_general
(* Not yet used
   | Unparsed
*)
;;

let get_entity_type env e =
(* Could also be External_parsed_general or Unparsed
   according to where the entity is bound *)
  if Xml_env.Expand_env.mem e env.general_env then
    match Xml_env.Expand_env.find e env.general_env with
    | Internal _ -> Internal_general
    | External _ -> External_parsed_general
  else
    fatal_error (Printf.sprintf
      "unknown general entity reference %S" (Xml_basics_ast.of_xml_name e))
;;

let rec expand_att_value env a_l =
  let a_l = ro_expand_list expand_att_value_elem env a_l in
  let a_l =
    match a_l with
    | [] -> []
    | _ ->
      let rec loop accu = function
        | [] -> accu
        | Xml_basics_ast.Att_value_text att_value_text :: att_values ->
          loop (accu ^ att_value_text) att_values
        | Xml_basics_ast.Att_value_reference _ :: _ ->
          assert false in
      let att_value_text = loop "" a_l in
      [Xml_basics_ast.Att_value_text att_value_text] in
  a_l

and expand_att_value_elem env = function
  | Att_value_text a ->
    let a = expand_att_value_text env a in
    Att_value_text a
  | Att_value_reference (Char_ref i) ->
    Att_value_text (expand_char_ref env i)
  | Att_value_reference (Entity_ref e) ->
    (* According to the entity type of e
       internal general, external parsed general, or unparsed. *)

    let replacement_text =
      match get_entity_type env e with
      | Internal_general -> expand_general_entity_ref env e
      | External_parsed_general ->
        failwith "reference to external parsed general entity is forbidden \
                  in attribute value"
    (* Not yet used
       | Unparsed -> Entity_ref_Forbidden
    *)
    in
    Att_value_text replacement_text
;;

(* end of Xml_basic_ast *)

let expand_space env s = expand_quoted_string env s;;

let expand_comment env s = expand_quoted_string env s;;

let expand_system_literal env s = expand_quoted_string env s;;

(* FIXME: is there something to do ? *)
let expand_unparsed_att_value _env _s =
  failwith "Not yet"
;;

let expand_pubid_literal env s = expand_quoted_string env s;;

let expand_cdata env s = expand_quoted_string env s;;

(* TODO: the version may be an important information used to expand the XML
   file. It could be added to the environment. *)
let expand_version_number env s = env, expand_quoted_string env s;;

(* TODO: the encoding may be an important information used to expand the XML
   file. It could be added to the environment. *)
let expand_encoding_name env s = env, expand_quoted_string env s;;

let expand_nm_token env s = expand_quoted_string env s;;

let rec expand_document env = function
  | { Xml_ast.prolog = p; Xml_ast.element = e; Xml_ast.misc_star = m; } ->
    let env, p = expand_prolog env p in
    let e = expand_element env e in
    let m = expand_misc_star env m in
    env, { Xml_ast.prolog = p; Xml_ast.element = e; Xml_ast.misc_star = m; }

and expand_prolog env = function
  | {
      Xml_ast.xml_decl = x_o;
      Xml_ast.xml_misc_star = m;
      Xml_ast.xml_doc_type_decl = d_o;
      Xml_ast.xml_misc_star_trailing = m0;
    } ->
    let env, x_o = expand_option expand_xml_decl env x_o in
    let m = expand_misc_star env m in
    let env, d_o = expand_option expand_doc_type_decl env d_o in
    let m0 =  expand_misc_star env m0 in
    env,
    {
      Xml_ast.xml_decl = x_o;
      Xml_ast.xml_misc_star = m;
      Xml_ast.xml_doc_type_decl = d_o;
      Xml_ast.xml_misc_star_trailing = m0;
    }

and expand_xml_decl env = function
  | {
      Xml_ast.xml_version_info = v;
      Xml_ast.xml_encoding_decl = e_o;
      Xml_ast.xml_standalone_decl = s_o;
    } ->
    let env, v = expand_version_number env v in
    let env, e_o = expand_option expand_encoding_name env e_o  in
    let env, s_o = expand_option expand_standalone_decl env s_o in
    env,
    {
      Xml_ast.xml_version_info = v;
      Xml_ast.xml_encoding_decl = e_o;
      Xml_ast.xml_standalone_decl = s_o;
    }

and expand_standalone_decl env = function
  | Xml_ast.Standalone -> env, Standalone
  | Xml_ast.Not_standalone -> env, Not_standalone

and expand_element env = function
  | Xml_ast.Empty_element_tag t ->
    let t = expand_tag env t in
    Empty_element_tag t
  | Xml_ast.Content_element_tag (t, e) ->
    let t = expand_tag env t in
    let e = expand_element_content env e in
    Content_element_tag (t, e)

and expand_tag env = function
  | { Xml_ast.tag_name = t; Xml_ast.tag_attributes = t_l; } ->
    let t = expand_tag_name env t in
    let t_l = ro_expand_list expand_tag_attribute env t_l in
    { Xml_ast.tag_name = t; Xml_ast.tag_attributes = t_l; }

and expand_tag_attribute env = function
  | { Xml_ast.tag_attribute_name = t; Xml_ast.tag_attribute_value = a; } ->
    let t = expand_tag_attribute_name env t in
    let a = expand_att_value env a in
    { Xml_ast.tag_attribute_name = t; Xml_ast.tag_attribute_value = a; }

and expand_char_data env = function
  | Xml_ast.Char_data c ->
    let c = expand_char_data_text env c in
    Xml_ast.Char_data c

and expand_element_content env = function
  | Xml_ast.Element_content c_t_i_t ->
    let char_data_folder accu char_data =
      let char_data = expand_char_data env char_data in
      let accu = add_char_data_to_element_content char_data accu in
      accu in
    let tag_child_folder accu tag_child =
      expand_tag_child env accu tag_child in
    let c_t_i_t =
      Intercal.fold_left
        char_data_folder
        tag_child_folder
        Intercal.empty
        c_t_i_t in
    let c_t_i_t = Intercal.rev c_t_i_t in
    Xml_ast.Element_content c_t_i_t

and expand_tag_child env accu = function
  | Xml_ast.Child_element e ->
    let e = expand_element env e in
    let accu =
      add_tag_child_to_element_content (Xml_ast.Child_element e) accu in
    accu
  | Xml_ast.Child_cd_sect c ->
    let c = expand_cdata env c in
    let accu =
      add_tag_child_to_element_content (Xml_ast.Child_cd_sect c) accu in
    accu
  | Xml_ast.Child_pi p ->
    let p = expand_pi env p in
    let accu =
      add_tag_child_to_element_content (Xml_ast.Child_pi p) accu in
    accu
  | Xml_ast.Child_comment c ->
    let c = expand_comment env c in
    let accu =
      add_tag_child_to_element_content (Xml_ast.Child_comment c) accu in
    accu
  | Xml_ast.Child_reference (Char_ref i) ->
    debug (Format.sprintf "expanding character reference %ld" i);
    let text = expand_char_ref env i in
    let accu =
      add_char_data_to_element_content (Xml_ast.Char_data text) accu in
    accu
  | Xml_ast.Child_reference ((Entity_ref e) as er) ->
    fdebug
      (fun ppf ->
       Format.fprintf ppf
         "expanding reference %a"
         Xml_basics_ast_print.print_reference er);
    let text = expand_general_entity_ref env e in
    debug (Printf.sprintf "expanded reference: %S:" text);
    let accu =
      add_char_data_to_element_content (Xml_ast.Char_data text) accu in
    accu

and expand_misc_star env m_l = ro_expand_list expand_misc env m_l

and expand_misc env = function
  | Xml_ast.Misc_comment c ->
    let c = expand_comment env c in
    Xml_ast.Misc_comment c
  | Xml_ast.Misc_pi p ->
    let p = expand_pi env p in
    Xml_ast.Misc_pi p
  | Xml_ast.Misc_space s ->
    let s = expand_space env s in
    Xml_ast.Misc_space s

and expand_pi env = function
  | { Xml_ast.pi_target = p; Xml_ast.pi_instructions = p0; } ->
    let p = expand_pi_target env p in
    let p0 = expand_pi_instructions env p0 in
    { Xml_ast.pi_target = p; Xml_ast.pi_instructions = p0; }

and expand_doc_type_decl env = function
  | {
      Xml_ast.doctype_name = t;
      Xml_ast.external_id = e_o;
      Xml_ast.int_subset = i_o;
    } ->
    let t = expand_tag_name env t in
    let env, e_o = expand_option expand_external_id env e_o in
    let env, i_o = expand_option expand_int_subset env i_o in
    env, {
      Xml_ast.doctype_name = t;
      Xml_ast.external_id = e_o;
      Xml_ast.int_subset = i_o;
    }

and expand_external_id env = function
  | Xml_ast.System s ->
    let s = expand_system_literal env s in
    env, Xml_ast.System s
  | Xml_ast.Public (p, s) ->
    let p = expand_pubid_literal env p in
    let s = expand_system_literal env s in
    env, Xml_ast.Public (p, s)

and expand_public_id env p = expand_pubid_literal env p

and expand_int_subset env i_l = expand_list expand_int_subset_elem env i_l

and expand_int_subset_elem env = function
  | Xml_ast.Markup_decl m ->
    let env, m = expand_markup_decl env m in
    env, Xml_ast.Markup_decl m
  | Xml_ast.Sep_decl s ->
    let env, s = expand_sep_decl env s in
    env, Xml_ast.Sep_decl s

and expand_sep_decl env = function
  | Xml_ast.Pe_reference_decl e ->
    let _replacement_text = find_parameter_entity_ref env e in
    (* TODO: we must parse the replacement_text with its surrounding text *)
    failwith "Not yet", Xml_ast.Pe_reference_decl e
  | Xml_ast.Space_decl s ->
    let s = expand_space env s in
    env, Xml_ast.Space_decl s

and expand_notation_id env = function
  | Xml_ast.External_id e ->
    let env, e = expand_external_id env e in
    env, Xml_ast.External_id e
  | Xml_ast.Public_id p ->
    let p = expand_public_id env p in
    env, Xml_ast.Public_id p

and expand_markup_decl env = function
  | Xml_ast.Element_decl (t, c) ->
    let t = expand_tag_name env t in
    let c = expand_content_spec env c in
    env, Xml_ast.Element_decl (t, c)
  | Xml_ast.Attlist_decl (t, a_l) ->
    let t = expand_tag_attribute_name env t in
    let env, a_l = expand_list expand_att_def env a_l in
    env, Xml_ast.Attlist_decl (t, a_l)
  | Xml_ast.Entity_decl e ->
    let env, e = expand_entity_decl env e in
    env, Xml_ast.Entity_decl e
  | Xml_ast.Notation_decl (n, n0) ->
    let env, n = expand_notation_name env n in
    let env, n0 = expand_notation_id env n0 in
    env, Xml_ast.Notation_decl (n, n0)
  | Xml_ast.Pi_decl p ->
    let p = expand_pi env p in
    env, Xml_ast.Pi_decl p
  | Xml_ast.Comment_decl c ->
    let c = expand_comment env c in
    env, Xml_ast.Comment_decl c

and expand_att_def env = function
  | {
      Xml_ast.att_def_name = t;
      Xml_ast.att_def_type = a;
      Xml_ast.att_def_default = a0;
    } ->
    let t = expand_tag_attribute_name env t in
    let env, a = expand_att_type env a in
    let env, a0 = expand_att_default env a0 in
    env, {
      Xml_ast.att_def_name = t;
      Xml_ast.att_def_type = a;
      Xml_ast.att_def_default = a0;
    }

and expand_att_type env = function
  | Xml_ast.String_type -> env, Xml_ast.String_type
  | Xml_ast.Tokenized_type t ->
    let env, t = expand_tokenized_type env t in
    env, Xml_ast.Tokenized_type t
  | Xml_ast.Enumerated_type e ->
    let env, e = expand_enumerated_type env e in
    env, Xml_ast.Enumerated_type e

and expand_tokenized_type env = function
  | Xml_ast.Id_type -> env, Xml_ast.Id_type
  | Xml_ast.Idref_type -> env, Xml_ast.Idref_type
  | Xml_ast.Idrefs_type -> env, Xml_ast.Idrefs_type
  | Xml_ast.Entity_type -> env, Xml_ast.Entity_type
  | Xml_ast.Entities_type -> env, Xml_ast.Entity_type
  | Xml_ast.Nmtoken_type -> env, Xml_ast.Nmtoken_type
  | Xml_ast.Nmtokens_type -> env, Xml_ast.Nmtokens_type

and expand_enumerated_type env = function
  | Xml_ast.Notation_type n_u_t ->
    let env, n_u_t = expand_ulist1 expand_notation_name env n_u_t in
    env, Xml_ast.Notation_type n_u_t
  | Xml_ast.Enumeration_type n_u_t ->
    let n_u_t = ro_expand_ulist1 expand_nm_token env n_u_t in
    env, Xml_ast.Enumeration_type n_u_t

and expand_att_default env = function
  | Xml_ast.Required -> env, Xml_ast.Required
  | Xml_ast.Implied -> env, Xml_ast.Implied
  | Xml_ast.Att_value u ->
    let env, u = expand_unparsed_att_value env u in
    env, Xml_ast.Att_value u
  | Xml_ast.Fixed_att_value u ->
    let env, u = expand_unparsed_att_value env u in
    env, Xml_ast.Fixed_att_value u

and expand_entity_decl env = function
  | Xml_ast.Ge_decl (e, (Gentity_value entity_value)) as entity_decl ->
    let replacement_text = expand_entity_value env entity_value in
    let env = add_general_entity env e (Internal replacement_text) in
    env, entity_decl
  | Xml_ast.Ge_decl
      (e,
       Gentity_external_id (_external_id, _ndata_decl_opt)) as entity_decl ->
    let env = add_general_entity env e (External None) in
    env, entity_decl
  | Xml_ast.Pe_decl (e, Pentity_value entity_value) as entity_decl ->
    let replacement_text = expand_entity_value env entity_value in
    let env = add_parameter_entity env e (Internal replacement_text) in
    env, entity_decl
  | Xml_ast.Pe_decl (e, Pentity_external_id _external_id) as entity_decl ->
    let env = add_parameter_entity env e (External None) in
    env, entity_decl

and expand_entity_value_elem env accu = function
  | Entity_value_text text ->
    Buffer.add_string accu text
  | Entity_value_reference (Xml_basics_ast.Char_ref r) ->
    let text = expand_char_ref env r in
    Buffer.add_string accu text
  | Entity_value_reference (Xml_basics_ast.Entity_ref e) ->
    let text = Printf.sprintf "&%s;" (Xml_basics_ast.of_xml_name e) in
    Buffer.add_string accu text
  | Entity_value_pe_reference e ->
    let text = find_parameter_entity_ref env e in
    Buffer.add_string accu text

and expand_entity_value env e =
  let b = Buffer.create 127 in
  List.iter (expand_entity_value_elem env b) e;
  Buffer.contents b

and _expand_ndata_decl env n = env, n

and _expand_pentity_def env = function
  | Xml_ast.Pentity_value e as def ->
    def, expand_entity_value env e
  | Xml_ast.Pentity_external_id _e ->
    fatal_error "expand of external parameter entity is not supported"
(*
    let env, e = expand_external_id env e in
    env, Xml_ast.Pentity_external_id e
*)

and expand_content_spec env = function
  | Xml_ast.Content_spec_empty -> Xml_ast.Content_spec_empty
  | Xml_ast.Content_spec_any -> Xml_ast.Content_spec_any
  | Xml_ast.Content_spec_mixed c ->
    let c = expand_content_spec_mixed env c in
    Xml_ast.Content_spec_mixed c
  | Xml_ast.Content_spec_children c ->
    let c = expand_content_spec_children env c in
    Xml_ast.Content_spec_children c

and expand_content_particle_choice env = function
  | (c, c_l) ->
    let c = expand_content_spec_mod env c in
    let c_l = expand_list2 expand_content_spec_particle env c_l in
    c, c_l

and expand_content_particle_seq env = function
  | (c, c_u_t) ->
    let c = expand_content_spec_mod env c in
    let c_u_t = ro_expand_ulist1 expand_content_spec_particle env c_u_t in
    c, c_u_t

and expand_content_particle_name env = function
  | (c, t) ->
    let c = expand_content_spec_mod env c in
    let t = expand_tag_name env t in
    c, t

and expand_content_spec_children env = function
  | Xml_ast.Content_spec_choice c ->
    let c = expand_content_particle_choice env c in
    Xml_ast.Content_spec_choice c
  | Xml_ast.Content_spec_seq c ->
    let c = expand_content_particle_seq env c in
    Xml_ast.Content_spec_seq c

and expand_content_spec_particle env = function
  | Xml_ast.Content_particle_name c ->
    let c = expand_content_particle_name env c in
    Xml_ast.Content_particle_name c
  | Xml_ast.Content_particle_choice c ->
    let c = expand_content_particle_choice env c in
    Xml_ast.Content_particle_choice c
  | Xml_ast.Content_particle_seq c ->
    let c = expand_content_particle_seq env c in
    Xml_ast.Content_particle_seq c

and expand_content_spec_mixed env = function
  | Xml_ast.PCData -> Xml_ast.PCData
  | Xml_ast.PCData_empty -> Xml_ast.PCData_empty
  | Xml_ast.PCData_alternatives t_l ->
    let t_l = ro_expand_list expand_tag_name env t_l in
    Xml_ast.PCData_alternatives t_l
(* TODO: that rule looks like it's not valid
   | Xml_ast.PCData_alternatives_PEReferences e_l ->
    (* FIXME *)
   let env, e_l = expand_list find_parameter_entity_ref env e_l in
   env, Xml_ast.PCData_alternatives_PEReferences e_l
*)

and expand_content_spec_mod _env = function
  | Xml_ast.None_mod -> Xml_ast.None_mod
  | Xml_ast.Opt_mod -> Xml_ast.Opt_mod
  | Xml_ast.Star_mod -> Xml_ast.Star_mod
  | Xml_ast.Plus_mod -> Xml_ast.Plus_mod

and expand_ext_subset env = function
  | (t_o, e) ->
    let env, t_o = expand_option expand_text_decl env t_o in
    let env, e = expand_ext_subset_decl env e in
    env, (t_o, e)

and expand_ext_subset_decl env = function
  | e_l -> expand_list expand_ext_subset_decl_elem env e_l

and expand_ext_subset_decl_elem env = function
  | Xml_ast.Ext_markup_decl m ->
    let env, m = expand_markup_decl env m in
    env, Xml_ast.Ext_markup_decl m
  | Xml_ast.Ext_conditional_sect c ->
    let env, c = expand_conditional_sect env c in
    env, Xml_ast.Ext_conditional_sect c
  | Xml_ast.Ext_sep_decl s ->
    let env, s = expand_sep_decl env s in
    env, Xml_ast.Ext_sep_decl s

and expand_text_decl env = function
  | { Xml_ast.text_version_info = v_o; Xml_ast.text_encoding_decl = e; } ->
    let env, v_o = expand_option expand_version_number env v_o in
    let env, e = expand_encoding_name env e in
    env, { Xml_ast.text_version_info = v_o; Xml_ast.text_encoding_decl = e; }

and expand_conditional_sect env = function
  | Xml_ast.Include_sect e ->
    let env, e = expand_ext_subset_decl env e in
    env, Xml_ast.Include_sect e
  | Xml_ast.Ignore_sect i ->
    let i = expand_ignore_text env i in
    env, Xml_ast.Ignore_sect i

and expand_ignore_text env s = expand_quoted_string env s;;

  (* TODO: no validation? never? *)
let default_validating = false;;

let to_document d =
  debug "to_document";
  let env = init_env default_validating in
  let _, d = expand_document env d in
  debug "xml: expanding phase done";
  d
;;

external of_document :
  document -> Xml_ast.raw_document = "%identity";;

let to_ext_subset es =
  let _, es = expand_ext_subset (init_env default_validating) es in
  es
;;

external of_ext_subset :
  ext_subset -> Xml_ast.raw_ext_subset = "%identity";;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
