(***********************************************************************)
(*                                                                     *)
(*                           Matlab Parser                             *)
(*                                                                     *)
(*          J.Ph Chancelier, Enpc/Cermics                              *)
(*                                                                     *)
(*  Copyright 2012-2019,                                               *)
(*  Ecole Nationale des ponts et chaussées                             *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mtlb_ast.mli,v 1.1 2019-05-10 14:09:18 jpc Exp $ *)

open Ast_node;;

type parsing =
   | ParsingMtlb
;;

(* Abstract syntax tree produced by parsing *)

type mtlb_constant =
  | Num of string
  | Imag_num of string
  | Dq_string of string
  | Sq_string of string
;;

type 'info mtlb_ast = ('info mtlb_ast_desc, 'info) ast_node
and 'info mtlb_ast_desc =
  | Ast of 'info exprs list

and 'info exprs = ('info exprs_desc, 'info) ast_node
and 'info exprs_desc = 'info exprsep list

and 'info exprsep = ('info exprsep_desc, 'info) ast_node
and 'info exprsep_desc =
  | Exprsep of 'info expr * 'info sep

and 'info sep = ('info sep_desc, 'info) ast_node
and 'info sep_desc =
  | Comma of int
  | Newline of int
  | Semi of int
  | Sep_None of int

and 'info expr = ('info expr_desc, 'info) ast_node
and 'info expr_desc =
  | Expression of 'info expression
  | Command of 'info command
  | WordListCmd of 'info identifier * 'info word_list
  | Comments of string
  | Pcomments of string
  | Dcomments of string
  | Empty

and 'info word_list = ('info word_list_desc, 'info) ast_node
and 'info word_list_desc = 'info wcmds_desc list
and 'info wcmds_desc =
  | SQ of string
  | DQ of string

and 'info expression = ('info expression_desc, 'info) ast_node
and 'info expression_desc =
  | SimpleExpr of 'info simple_expr
  | AssignExpr of 'info assign_expr
  | AnonFcnHandle of 'info param_list * 'info expression

and 'info command = ('info command_desc, 'info) ast_node
and 'info command_desc =
  | Declaration of 'info declaration
  | Switch of 'info expression * 'info case_list
  | IfCommand of 'info if_command
  | LoopCommand of 'info loop_command
  | JumpCommand of 'info jump_command
  | ExceptCommand of 'info except_command
  | SciptFile of 'info opt_list
  | MFunction of 'info mfunction

and 'info jump_command = ('info jump_command_desc, 'info) ast_node
and 'info jump_command_desc =
  | Break | FReturn of 'info return_list | Continue

and 'info primary_expr = ('info primary_expr_desc, 'info) ast_node
and 'info primary_expr_desc =
  | PexpIdentifier of 'info identifier
  | PexpConstant of 'info constant
  | PexpFcnHandle of 'info identifier
  | PexpMatrix of 'info exprs list
  | PexpCell of 'info exprs list
  | PexpMetaIdentifier of 'info meta_identifier
  | PexpSuperIdentifier of 'info superclass_identifier
  | PexpExpression of 'info expression

and 'info identifier = ('info identifier_desc, 'info) ast_node
and 'info identifier_desc = string

and 'info meta_identifier = ('info meta_identifier_desc, 'info) ast_node
and 'info meta_identifier_desc = string

and 'info superclass_identifier = ('info superclass_identifier_desc, 'info) ast_node
and 'info superclass_identifier_desc = string

and 'info arg_list = ('info arg_list_desc, 'info) ast_node
and 'info arg_list_desc = 'info  arg list

and 'info arg = ('info arg_desc, 'info) ast_node
and 'info arg_desc =
   | ArgExpression of 'info expression
   | ArgColon
   | ArgTilde

and 'info oper_expr = ('info oper_expr_desc, 'info) ast_node
and 'info oper_expr_desc =
  | PrimaryExpr of 'info primary_expr
  | PostfixOp of string * 'info oper_expr
  | IndexExpr of 'info oper_expr * 'info arg_list * string * string
  | IndirectRef of 'info oper_expr * 'info expression
  | IndirectRefU of 'info oper_expr * 'info identifier
  | PrefixOpUnary of string * 'info oper_expr
  | BinaryOp of string * 'info oper_expr * 'info oper_expr
  | Op_stars of 'info stars * 'info oper_expr * 'info oper_expr
  | Op_divs  of 'info divs * 'info oper_expr * 'info oper_expr

and 'info simple_expr= ('info simple_expr_desc, 'info) ast_node
and 'info simple_expr_desc =
  | ColonExpr of 'info oper_expr list
  | Op_compare of 'info compare  * 'info simple_expr *  'info simple_expr
  | SexpBinaryOp of string  * 'info simple_expr *  'info simple_expr
  | BooleanOp of string * 'info simple_expr * 'info  simple_expr

and 'info constant = ('info constant_desc, 'info) ast_node
and 'info constant_desc =
  | Constant of mtlb_constant

and 'info param_list = ('info param_list_desc, 'info) ast_node
and 'info param_list_desc =
  | EmptyParamList
  | ParamList of 'info decl2 list

and 'info assign_lhs = ('info assign_lhs_desc, 'info) ast_node
and 'info assign_lhs_desc =
  | LhsSimple of 'info simple_expr
  | LhsArgs of 'info exprs

and 'info assign_expr = ('info  assign_expr_desc, 'info) ast_node
and 'info assign_expr_desc =
  | Assign of string * 'info assign_lhs * 'info expression

and 'info declaration = ('info declaration_desc, 'info) ast_node
and 'info declaration_desc =
  | Global of 'info decl2 list
  | Static of 'info decl2 list

and 'info decl2 = ('info decl2_desc, 'info) ast_node
and 'info decl2_desc =
  | Declare of 'info identifier
  | DeclareSet of 'info identifier * 'info expression
  | MagicTilde

(* switch statement *)

and 'info case_list = ('info case_list_desc, 'info) ast_node
and 'info case_list_desc =
    | Nocase
    | DefaultCase of 'info default_case
    | Cases of 'info switch_case list
    | CasesOther of ('info switch_case list ) * 'info default_case

and 'info switch_case = ('info switch_case_desc, 'info) ast_node
and 'info switch_case_desc =
  | Case of 'info opt_list * 'info opt_list * 'info opt_list

and 'info default_case = ('info default_case_desc, 'info) ast_node
and 'info default_case_desc =
  | Otherwize of 'info opt_list

and 'info opt_list = ('info opt_list_desc, 'info) ast_node
and 'info opt_list_desc = 'info exprs list

(* if *)

and 'info if_command= ('info if_command_desc, 'info) ast_node
and 'info if_command_desc =
  | IfCmd of 'info if_cmd list
  | IfCmdWithElse of ('info if_cmd list) * 'info else_clause

and 'info if_cmd = ('info if_cmd_desc, 'info) ast_node
and 'info if_cmd_desc =
  | If of     'info opt_list * 'info opt_list * 'info opt_list
  | ElseIf of 'info opt_list * 'info opt_list * 'info opt_list

and 'info else_clause= ('info else_clause_desc, 'info) ast_node
and 'info else_clause_desc =
  | Else of 'info opt_list

(* loop *)

and 'info loop_command = ('info loop_command_desc, 'info) ast_node
and 'info loop_command_desc =
  | WhileCommand of 'info opt_list * 'info opt_list * 'info opt_list
  |  DoUntilCommand of 'info opt_list * 'info expression
  |  ForCommand of 'info opt_list * 'info opt_list * 'info opt_list

(* exceptions *)

and 'info except_command= ('info except_command_desc, 'info) ast_node
and 'info except_command_desc =
  | UnwindCommand of 'info opt_list * 'info opt_list
  | TryCatchCommand of 'info opt_list * 'info opt_list
  | TryCommand of 'info opt_list

and 'info mfunction= ('info mfunction_desc, 'info) ast_node
and 'info mfunction_desc =
  | FunctionNoRet of 'info identifier * 'info body
  | Function of 'info return_list * 'info identifier * 'info  body

and 'info body = ('info body_desc, 'info) ast_node
and 'info body_desc =
  | Body of 'info param_list * 'info opt_list * 'info endfun
  | SimpleBody of 'info opt_list * 'info endfun

and 'info endfun = ('info endfun_desc, 'info) ast_node
and 'info endfun_desc =
 | End_fun | Eof_fun

and 'info return_list = ('info return_list_desc, 'info) ast_node
and 'info return_list_desc =
   | EmptyReturn
   | Return of 'info identifier list

and 'info compare = ('info compare_desc, 'info) ast_node
and 'info compare_desc = 'info op_compare_desc * string
and 'info op_compare_desc =
  | Lt | Le | Eq | Ne | Ge | Gt

and 'info stars = ('info stars_desc, 'info) ast_node
and 'info stars_desc = 'info op_stars_desc * string
and 'info op_stars_desc =
  | Star | Dot_star

and 'info divs = ('info divs_desc, 'info) ast_node
and 'info divs_desc =  'info op_divs_desc * string
and 'info op_divs_desc =
  | Slash | Bslash | Dot_slash | Dot_bslash
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
