(***********************************************************************)
(*                                                                     *)
(*                           Matlab Parser                             *)
(*                                                                     *)
(*          J.Ph Chancelier, Enpc/Cermics                              *)
(*                                                                     *)
(*  Copyright 2012-2019,                                               *)
(*  Ecole Nationale des ponts et chaussées                             *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mtlb_ast_funs.ml,v 1.1 2019-05-10 14:09:18 jpc Exp $ *)

open Ast_node;;
open Ast_node_utils;;
open Mtlb_ast;;

(* val mk_dummy_ast_node :
   'a -> ('a, Ast_node.ast_node_info) Ast_node.ast_node;; *)

let mk_dummy_ast_node desc =
  Ast_node_utils.make_dummy_ast_node desc Mtlb_ast.ParsingMtlb
;;

(** val mk : 'a -> ('a, Ast.ast_node_info) Ast_node.ast_node;; *)

let mk desc =
  Ast_node_utils.make_ast_node (get_symbol_loc ()) desc Mtlb_ast.ParsingMtlb
;;

let change_ast_desc node desc =
  Ast_node_utils.change_ast_node_contents node desc Mtlb_ast.ParsingMtlb
;;

(* test if an Expression (e) is an assign *)

let rec get_assign = function
  | Expression e ->
      let e1 = match e.Ast_node.ast_desc with
      | AssignExpr assign -> [assign]
      | SimpleExpr e2 ->
          let e3 = match e2.ast_desc with
          | ColonExpr [e4] ->
              let e4 = match e4.ast_desc with
              | PrimaryExpr e5 ->
		  let e6 = match e5.ast_desc with
		  | PexpExpression expression ->
		      get_assign (Expression expression)
		  | PexpIdentifier _
		  | PexpConstant _
		  | PexpFcnHandle _
		  | PexpMatrix _
		  | PexpCell _
		  | PexpMetaIdentifier _
		  | PexpSuperIdentifier _ -> [] in
		  e6
	      | PostfixOp _
	      | IndexExpr _
	      | IndirectRef _
	      | IndirectRefU _
	      | PrefixOpUnary _
	      | BinaryOp _
	      | Op_stars _
	      | Op_divs  _ -> [] in
              e4
	  | ColonExpr _
	  | Op_compare _
	  | SexpBinaryOp _
	  | BooleanOp _ -> [] in
          e3
      | AnonFcnHandle _ -> [] in
      e1
  | Command _
  | WordListCmd _
  | Comments _
  | Pcomments _
  | Dcomments _ 
  | Empty -> [] 

let set_assign x assign =
  let desc2 = AssignExpr assign in
  let node2 = { x with ast_desc = desc2;} in
  { x with ast_desc = Expression node2;}

(* checks that a list of exprseps node are comments *)

let rec exprseps_are_comments exprseps =
  match exprseps with
  | [] -> true
  | exprsep :: exprseps ->
      ( match exprsep.ast_desc with
      | Exprsep ({ ast_desc = e; _;}, _) ->
          ( match e with
          | Pcomments(_)
          | Dcomments(_) -> exprseps_are_comments exprseps
	  | Expression _
	  | Command _
	  | WordListCmd _
	  | Comments  _ 
	  | Empty -> false))
;;

let exprs_are_comments exprs =
  exprseps_are_comments exprs.ast_desc
;;

(* test if an Exprsep is an expression *)

let get_expr = function
  | Expression e ->
      let e1 = match e.ast_desc with
      | SimpleExpr _expr -> [Expression (e)]
      | AssignExpr _
      | AnonFcnHandle _ -> [] in e1
  | Command _
  | WordListCmd _
  | Comments _
  | Pcomments _
  | Dcomments  _ 
  | Empty -> []
;;
(* test if an Exprsep is an assign followed bycomments *)

(* splits an exprseps in two expseps
 * the first one contains an expseps starting with assign and followed by possible comments
 * the second one contains exprseps which are not assign ops
 *)

let exprseps_assign_split exprs =
  match  exprs.ast_desc with
  | [] -> (exprs, exprs)
  | exprsep :: exprseps ->
      let (exprs_desc1, exprs_desc2) =
        match exprsep.ast_desc with
        | Exprsep (x, y) ->
            let x1 = x.ast_desc in
            let new_assign = get_assign x1 in
            if new_assign <> [] then
              let new_x = set_assign x (List.hd new_assign) in
              let new_exprsep = { exprsep with ast_desc = Exprsep (new_x, y);} in
              if exprseps_are_comments (exprseps) then
                (new_exprsep :: exprseps, [])
              else
                ([new_exprsep], exprseps)
            else
              ([], exprsep :: exprseps) in
      ({ exprs with ast_desc = exprs_desc1;},
       { exprs with ast_desc = exprs_desc2;})
;;

(* splits a list of exprseps: gather the leading comments
 * then an assign then the rest *)

let rec exprseps_list_assign_split flag co assign rest = function
  | [] -> (co, assign, rest)
  | x :: xs ->
      if flag && exprs_are_comments x then
        exprseps_list_assign_split flag (x :: co) assign rest xs
      else
        (* first expression which is not a comment *)
        let (assign, y) = exprseps_assign_split x in
        (co, [assign], if y.ast_desc = [] then xs else y :: xs)
;;

(* UTILITY 1 *)


(* checks that an expr is of type Expression ({SimpleExpr (..) } *)

let check_expr expr =
  match expr.ast_desc with
  | Expression expr -> 
      (match expr.ast_desc with
      | SimpleExpr _-> true 
      | AssignExpr _
      | AnonFcnHandle _ -> false)
  | Command _
  | WordListCmd _
  | Comments _
  | Pcomments _
  | Dcomments _ 
  | Empty -> false
;;

(* checks that an exprsep contains an expression *)

let check_exprsep exprsep =
  match exprsep.ast_desc with
  | Exprsep (x, _) -> check_expr x
;;

(* splits an exprs in two exprs:
 * the first one contains an expression followed by possible comments
 * the second one contains the rest.
 *)
let exprs_split exprs =
  match exprs.ast_desc with
  | [] -> (exprs, exprs)
  | exprsep :: exprseps ->
      let ( exprs_desc1, exprs_desc2 ) =
        if check_exprsep exprsep then
          if exprseps_are_comments(exprseps) then
            ( exprsep :: exprseps, [])
          else
            ( [exprsep], exprseps)
        else
          ([], exprsep :: exprseps) in
      ({ exprs with ast_desc = exprs_desc1;},
       { exprs with ast_desc = exprs_desc2;})
;;

(* splits a list of exprs: gather the leading comments
 * then an expression then the rest
 *)
let rec exprs_list_split flag co assign rest exprs_list =
  match exprs_list with
  | [] -> (co, assign, rest)
  | x :: xs ->
      if flag && (exprs_are_comments x) then
        exprs_list_split flag (x :: co) assign rest xs
      else
        (* first expression which is not a comment *)
        let (assign, y) = exprs_split x in
        (co, [assign], (if y.ast_desc = [] then xs else y :: xs))
;;

(* UTILITY 2 *)

(* get or set the opt_list part of a Function body *)
let function_get_body_opt_list f =
  match f.ast_desc with
  | FunctionNoRet (_, {ast_desc = Body (_, opt_list, _ );_}) -> opt_list
  | FunctionNoRet (_, {ast_desc = SimpleBody (opt_list, _);_}) -> opt_list
  | Function (_, _, {ast_desc = Body (_, opt_list, _);_}) -> opt_list
  | Function (_, _, {ast_desc = SimpleBody (opt_list, _);_} ) -> opt_list
;;

let function_set_body_opt_list f opt_list_desc =
  let body_set_body_opt_list body opt_list_desc =
    match body.ast_desc with
    | Body (plist, opt_list, endfun) ->
        (change_ast_desc body
           (Body (plist, change_ast_desc opt_list opt_list_desc, endfun)))
    | SimpleBody (opt_list, endfun) ->
        (change_ast_desc body
           (SimpleBody (change_ast_desc opt_list opt_list_desc, endfun)))
  in
  match f.ast_desc with
  | FunctionNoRet (id, body ) ->
      (change_ast_desc f
         (FunctionNoRet (id, body_set_body_opt_list body opt_list_desc)))
  | Function (rlist, id, body) ->
      (change_ast_desc f
         (Function(rlist, id, body_set_body_opt_list body opt_list_desc)))
;;

(* check that a exprsep if a Exprsep (Command(MFunction(x))) *)

let exprsep_is_mfunction exprsep =
  match exprsep.ast_desc with
  | Exprsep (e, _) ->
      (match e.ast_desc with
      | Command cmd -> 
          ( match cmd.ast_desc with
          | MFunction _ -> true 
	  | Declaration _
	  | Switch _
	  | IfCommand _
	  | LoopCommand _
	  | JumpCommand _
	  | ExceptCommand _
	  | SciptFile _ -> false )
      | Expression _
      | WordListCmd _
      | Comments _
      | Pcomments _
      | Dcomments _ 
      | Empty -> false)
;;

(* get x in Exprsep (Command(MFunction(x))) *)

let exprsep_get_mfunction exprsep =
  if not (exprsep_is_mfunction exprsep) then None else
  match exprsep.ast_desc with
  | Exprsep (node1, _) ->
      ( match node1.ast_desc with
      | Command node2 ->
          ( match node2.ast_desc with
          | MFunction x -> Some x
	  | Declaration _
	  | Switch _
	  | IfCommand _
	  | LoopCommand _
	  | JumpCommand _
	  | ExceptCommand _
	  | SciptFile _ -> None)
      | Expression _
      | WordListCmd _
      | Comments _
      | Pcomments _
      | Dcomments _ 
      | Empty -> None)
;;

(* set x in Exprsep (Command(MFunction(x))) *)

let exprsep_set_mfunction mfun exprsep =
  match exprsep.ast_desc with
  | Exprsep (node1, sep) ->
      ( match node1.ast_desc with
      | Command node2 ->
          ( match node2.ast_desc with
          | MFunction x ->
              let new_node2 =
                {node2 with ast_desc =
                   MFunction ({ x with ast_desc = mfun; }); } in
              let new_node1 =
                { node1 with
                  ast_desc = Command new_node2; } in
              Some {exprsep with
                    ast_desc = Exprsep (new_node1, sep); }
	  | Declaration _
	  | Switch _
	  | IfCommand _
	  | LoopCommand _
	  | JumpCommand _
	  | ExceptCommand _
	  | SciptFile _ -> None)
      | Expression _
      | WordListCmd _
      | Comments _
      | Pcomments _
      | Dcomments _ 
      | Empty -> None)
;;


(* UTILITY 3 *)

(* counts the EOF ended functions and the end ended functions *)

let rec fct_count_exprs_list = function
 | [] -> (0, 0)
 | exprs :: exprs_l ->
  let (a, b) = fct_count_exprs exprs in
  let (c, d) = fct_count_exprs_list exprs_l in
  (a + c, b + d)

and fct_count_exprs exprs =
  fct_count_exprs_desc exprs.ast_desc

and  fct_count_exprs_desc = function
  | [] -> (0, 0)
  | exprsep :: exprs ->
      let (a, b) = fct_count_exprsep_desc exprsep.ast_desc in
      let (c, d) = fct_count_exprs_desc exprs in
      (a + c, b + d);

and fct_count_exprsep_desc = function
 | Exprsep (expr, _) ->  fct_count_expr_desc expr.ast_desc

and fct_count_expr_desc = function
  | Command c  -> fct_count_command_desc c.ast_desc
  | Expression _
  | WordListCmd _
  | Comments _
  | Pcomments _
  | Dcomments _ 
  | Empty -> (0, 0)

and fct_count_command_desc = function
   | MFunction mfun -> fct_count_mfunction_desc mfun.ast_desc
   | Declaration _
   | Switch _
   | IfCommand _
   | LoopCommand _
   | JumpCommand _
   | ExceptCommand _
   | SciptFile _ -> (0, 0)

and fct_count_mfunction_desc = function
  | FunctionNoRet (_, body) -> fct_count_body_desc body.ast_desc
  | Function (_, _, body) -> fct_count_body_desc body.ast_desc

and fct_count_body_desc = function
  | Body (_, opt_list, end_fun) ->
      fct_count_opt_list_desc opt_list.ast_desc end_fun.ast_desc
  | SimpleBody (opt_list, end_fun) ->
     fct_count_opt_list_desc opt_list.ast_desc end_fun.ast_desc

and fct_count_opt_list_desc opt_list end_fun =
  let (a, b) = fct_count_exprs_list opt_list in
  let (c, d) = fct_count_end_fun_desc end_fun in
  (a + c, b + d)

and fct_count_end_fun_desc = function
  | End_fun -> (0, 1)
  | Eof_fun -> (1, 0)

;;

(* extract last function *)

(* liste d'exprsep -> (list d'exprsep, option exprsep)  *)

let rec fct_extract_exprs_desc = function
  | [] -> ([], None)
  | elt :: elts ->
      if exprsep_is_mfunction elt then
        let (a, _) =  fct_extract_exprs_desc elts in
        (a, Some elt)
      else
        let (a, b) =  fct_extract_exprs_desc elts in
        (elt :: a, b)
;;

(* une exprs_list -> (exprs_list, exprs_list with one exprsep) *)

let rec fct_extract_exprs_list = function
  | [] ->  ([], [])
  | exprs :: exprs_l ->
      let (a, b) = fct_extract_exprs_desc exprs.ast_desc in
      match b with
      | Some elt ->
          ([ { exprs with ast_desc = a; } ],
           [ { exprs with ast_desc = [elt] } ])
      | None ->
          let (c, d) = fct_extract_exprs_list exprs_l  in
          ({ exprs with ast_desc = a; } :: c, d)
;;

(* extract from function: fun -> (fun, exprsep_list ) *)

let fct_extract_from_fun1 node =
  let opt_list = function_get_body_opt_list node in
  let (a, b) = fct_extract_exprs_list opt_list.ast_desc  in
  (function_set_body_opt_list node a, b)
;;

(* exprsep -> ( [] ou [exprsep], list exprsep )  *)

let fct_extract_from_fun exprsep =
  let mfun_opt = exprsep_get_mfunction exprsep in
  match mfun_opt with
  | None -> ([], [])
  | Some mfun ->
      let (a, b) = fct_extract_from_fun1 mfun in
      match exprsep_set_mfunction a.ast_desc exprsep with
      | None -> ([], b)
      | Some x -> ([x], b)
;;

let fct_extract_r = function
  | [] -> ([], [])
  | [exprsep] -> fct_extract_from_fun exprsep
  | _ -> assert false
;;

let rec fct_extract exprs =
  let (a, b) = fct_extract_exprs_list exprs in
  match b with
  | [] -> a
  | [exprsep] ->
      let (c, d) = fct_extract_r exprsep.ast_desc in
      a @ [ { exprsep with ast_desc = c; } ] @ fct_extract d
  | _ -> assert false
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
