(***********************************************************************)
(*                                                                     *)
(*                           Matlab Parser                             *)
(*                                                                     *)
(*          J.Ph Chancelier, Enpc/Cermics                              *)
(*                                                                     *)
(*  Copyright 2012-2019,                                               *)
(*  Ecole Nationale des ponts et chaussées                             *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mtlb_lexer.mll,v 1.1 2019-05-10 14:09:18 jpc Exp $ *)

{

(* Prelude part: this is pure Caml. *)

open Mtlb_ast;;
open Mtlb_parser;;
open Lexing;;

(** {6 Lexing errors} *)

type error =
  | Illegal_character of char
  | Unterminated_comment
  | Unterminated_string
  | Test_error
;;

(** The various errors when lexing. *)

exception Error of error * Lexing.position * Lexing.position;;

let error (reason, start_p, curr_p) =
  raise (Error (reason, start_p, curr_p))
;;

(** {6 Explaining lexing errors} *)

let report_error ppf = function
  | Illegal_character c ->
      Format.fprintf ppf "Illegal character %C" c
  | Unterminated_comment ->
      Format.fprintf ppf "Unterminated comment"
  | Unterminated_string ->
      Format.fprintf ppf "Unterminated string"
  | Test_error -> Format.fprintf ppf "Test location"
;;

let report_lexical_error ppf = function
  | (Error (r, sp, ep) : exn) ->
    let loc = Mtlb_location.mk_loc sp ep in
    Format.fprintf ppf
      "%a@.Lexical error: %a@."
      Mtlb_location.print loc
      report_error r
  | exn -> raise exn
;;

let debug = false;;

(* Push the character `c' back in the `lexbuf'.  Since ocamllex does not
 * implement this functionality, I made it myself.  I used the `dump_lex'
 * function above to track the internal state of the lexer and see how it
 * was used and implemented push_back accordingly.  But since I'm not sure
 * this reverse engineering process was 100% right, I left some assertions
 * and the code of dump_lex above.
 * This code will b0rk everything up if you push_back something different
 * then what has been lexed by the last semantic action of ocamllex.  You
 * were warned.
 * from revcpp Quentin Hocquet   <mefyl@lrde.epita.fr>
 * Benoit Sigoure    <tsuna@lrde.epita.fr>
 *)

let push_back lexbuf c =
  if debug then
    prerr_endline (Printf.sprintf
      "push back a character =[%s]" (String.make 1 c));
  let fail msg = print_string msg in

  if lexbuf.lex_curr_pos = 0 then
    (* The buffer was flushed so we need to really push back `c' in the
       buffer. *)
(*
      let new_lex_buffer = Bytes.extend lexbuf.lex_buffer 1 0 in
      Bytes.set new_lex_buffer 0 c;
      lexbuf.lex_buffer <- new_lex_buffer;
*)
    (let new_lex_buffer =
      (* In order to be able to compile with the -safe-string flag,
         we simply forget about safety here: in the first place, we already
         broke the Lexing module abstraction barreer and forgot about
         considering to respect its invariants. *)
       Obj.magic (String.make 1 c ^ Obj.magic lexbuf.lex_buffer) in
     lexbuf.lex_buffer <- new_lex_buffer;
     lexbuf.lex_buffer_len <- lexbuf.lex_buffer_len + 1);

  (* Now `c' is in the buffer, simply change the current positions. *)
  (* Perform some sanity checks:
     1. The buffer_len can't be 0 since our `c' must still be in it.
     2. start_pos, curr_pos and last_pos must not be 0 since we're going
        to decrement them.
     3. Ditto for start_p.pos_cnum and curr_p.pos_cnum.
  *)
  if lexbuf.lex_buffer_len = 0 then
    fail ("Assertion failed: lex_curr_pos = "
          ^ string_of_int lexbuf.lex_curr_pos
          ^ " and lex_buffer_len = 0");
  if lexbuf.lex_curr_pos <= 0 then
    fail ("Assertion failed: lex_curr_pos <= 0");
  if lexbuf.lex_curr_p.pos_cnum <= 0 then
    fail ("Assertion failed: lex_curr_p.pos_cnum <= 0");

  lexbuf.lex_curr_pos <- lexbuf.lex_curr_pos - 1;
  lexbuf.lex_last_pos <- lexbuf.lex_last_pos - 1;
  lexbuf.lex_curr_p <- {
    lexbuf.lex_curr_p with
    pos_cnum = lexbuf.lex_curr_p.pos_cnum - 1;
  }
;;
(* Push the string `s' back in the `lexbuf'.  The comments of push_back
 *  above also apply here.  You were warned.
 *)

let push_back_strings lexbuf = function
  | [] -> ()
  | elt :: elts ->
    String.iter (push_back lexbuf) elt;
    List.iter (function arg -> String.iter (push_back lexbuf) arg) elts
;;

(* The table of keywords *)

let keyword_table =
  Hash_table.of_bindings [
  "break", BREAK;
  "case" , CASE;
  "catch", CATCH;
  "cleanup", CLEANUP;
  "continue", CONTINUE;
  "else", ELSE;
  "elseif", ELSEIF;
  "end", END;
  "endif", END; (* octave extension *)
  "endfor", END; (* octave extension *)
  "endwhile", END; (* octave extension *)
  "end_try_catch", END; (* octave extension *)
  "endswitch", END; (* octave extension *)
  "endfunction", END;
  "return", FUNC_RET;
  "for", FOR;
  "function", FUNCTION;
  "global", GLOBAL;
  "if", IF;
  "otherwise", OTHERWISE;
  "persistent", STATIC;
  "switch", SWITCH;
  "try", TRY;
  "unwind", UNWIND;
  "while", WHILE;
  "unwind_protect", UNWIND;
  "unwind_protect_cleanup" , CLEANUP;
  "end_unwind_protect", END;
  "do", DO; (* octave extension *)
  "until", UNTIL;(* octave extension *)
]
;;

(** {6 Keeping the internal buffer locations up to date} *)

let update_loc lexbuf fname line absolute chars =
  let pos = lexbuf.lex_curr_p in
  let new_fname =
    match fname with
    | None -> pos.pos_fname
    | Some s -> s in
  lexbuf.lex_curr_p <- {
    pos with
    pos_fname = new_fname;
    pos_lnum = if absolute then line else pos.pos_lnum + line;
    pos_bol = pos.pos_cnum - chars;
  }
;;

(** Add one to the current line counter of the file being lexed. *)

let incr_line_num lexbuf =
  update_loc lexbuf None 1 false 0
;;

(** Add number of newlines contained in string to the current line
    counter of the file being lexed. *)

let rec count_newlines s =
  let n = (String.length s) in
  if n = 0 then 0
  else
    let rest = count_newlines (String.sub s 1 (n -1)) in
    if s.[0] == '\n' then
      rest + 1
    else
      rest
;;

let incr_lines lexbuf str =
  let count = (count_newlines str) in
  update_loc lexbuf None count false 0
;;

(* debug *)

let show_location str lexbuf =
  let loc = Mtlb_location.mk_loc  lexbuf.lex_start_p lexbuf.lex_curr_p in
  Format.fprintf Format.std_formatter "%a@.Test Lexical error: %s %a@."
    Mtlb_location.print loc str report_error Test_error

(* To buffer string literals *)
let string_buffer = Buffer.create 256;;

let store_string_char c =
  Buffer.add_char string_buffer c
;;

let get_stored_string () =
  let s = Buffer.contents string_buffer in
  Buffer.reset string_buffer;
  s
;;

(* To store the position of the beginning of a string and comment *)
let string_start_pos = ref None;;
let comment_start_pos = ref None;;

(* to separate strings starting with ' from transpose .... *)

let is_transpose = ref false;;

let is_ident = function
  | IDENT _ -> true
  | (NEWLINE_WCMD _|WCMD _|EQUAL|COLON|MINUS|MINUSW|ADD|ADDW|AND_AND|OR_OR|AND|
     OR|NOT|COMPARE _|STARS _|DIVS _|DOT_ADD|DOT_MINUS|QUOTE|TRANSPOSE|ADD_ADD|
     MINUS_MINUS|POW|DOT_POW|COMMENTS _|PCOMMENT _|DCOMMENT _|NUM _|IMAG_NUM _|
     STRUCT_ELT _|END|DQ_STRING _|SQ_STRING _|FOR|WHILE|DO|UNTIL|IF|ELSEIF|ELSE|
     SWITCH|CASE|OTHERWISE|BREAK|CONTINUE|FUNC_RET|UNWIND|CLEANUP|TRY|CATCH|
     GLOBAL|STATIC|COLONEQUAL|FUNCTION_HANDLE _|PROPERTIES _|METHODS _|EVENTS _|
     ENUMERATION _|METAQUERY _|SUPERCLASSREF _|GET _|SET _|NEWLINE _|COMMA|
     SEMICOLUMN|NEWLINE_SEMICOLUMN|NEWLINE_G|LPAREN|RPAREN|LBRACE|RBRACE|LBRACKET|
     RBRACKET|RBRACKET_EQUAL|END_OF_INPUT|FUNCTION|CLASSDEF|EOF|ATMARK|
     EQUAL_EXT _|RBRACKET_EQUAL_EXT _|DOT|DOTIDENT _|UADD|UMINUS|UADD_ADD|
     UMINUS_MINUS|SPACE) -> false
;;

let transpose_activate token =
  is_transpose := is_ident token;
  token
;;

(* flags to decide if white spaces have meaning *)

let flag_stack_bottom = [ (false, false) ];;
let flag_stack = ref flag_stack_bottom;;
let push_flag x = flag_stack := x :: !flag_stack;;
let pop_flag () =
  if !flag_stack <> flag_stack_bottom then flag_stack := List.tl !flag_stack
;;
let top_stack () =  List.hd !flag_stack;;
let clear_flag_stack () = flag_stack := flag_stack_bottom;;
let flag_stack_empty () = !flag_stack = flag_stack_bottom;;
(* true when a {} is not to be considered as a matrix case *)
let flag_accol = ref false;;

(* utility function for keywords and idents *)

let  mk_keyword_or_ident lexbuf s bl =
  if debug then prerr_endline  (Printf.sprintf "ident or key=[%s]" s);
  try
    let s1 = Hashtbl.find keyword_table s in
    let (x, _) = top_stack () in
    if x && s = "end" then push_back_strings lexbuf [bl];
    transpose_activate s1;
  with Not_found ->
    push_back_strings lexbuf [bl];
    transpose_activate (IDENT s)
;;

(*
let mk_ident_check_accol_type lexbuf bl_flag bl =
  let (mat,_) = top_stack () in
  if ( not mat || (String.length bl) = 0) then
    ( (* we are in the matrix case blanks are ignored in this case *)
      flag_accol := true;
     )
  else
    (
     flag_accol := false;
     if bl_flag then push_back_strings lexbuf [bl];
    )
;;
*)
let mk_ident_check_accol_type lexbuf bl_flag bl =
  let mat, _ = top_stack () in
  flag_accol := not mat || String.length bl = 0;
  if not !flag_accol && bl_flag then push_back_strings lexbuf [bl]
;;

let test_wcmd str f =
  let check_wcmd = function
    | [] -> true
    | "par" :: _ -> false
    | "accol" :: _ -> false
    | "op" :: _ -> false
    | _ -> true in
  let lexbuf = Lexing.from_string str in
  let strl = f [] lexbuf in
  check_wcmd (List.rev strl)
;;

let print_lexemes = function
  | [] -> ()
  | elt :: elts ->
    prerr_endline (Printf.sprintf  "%s" elt);
    List.iter (function arg -> prerr_endline (Printf.sprintf "%s" arg)) elts
;;

let split_wcmd str f =
  let lexbuf = Lexing.from_string (" " ^ str) in
  if debug then prerr_endline (Printf.sprintf "tosplit=[%s]" str);
  List.rev (f [] lexbuf)
;;


}

(* start of lexer rules *)

let blank = [ ' ' '\t' ]
let eol_char = ('\n' | ('\r' '\n') | '\r')
let not_eol_char = [^ '\n' '\r' ]
let newline = blank* eol_char ( blank | eol_char ) *

(* matlab line continuation *)

let continuation1 = blank * "..." not_eol_char * eol_char +
let continuation2 = blank * '\\' eol_char (blank * eol_char) * blank *
let blanks = blank * (continuation1 | continuation2)? blank *
let blanksp =
   blank + (continuation1 | continuation2)? blank *
 | blank * (continuation1 | continuation2)? blank +


(* comments up to end of line, block comments are defined latter *)

let comment_char = ( '#' | '%')
let char_start_comment = ([^ '{' '\n' '\r' ] | ('\r' [^ '\n' ]))
let line_comment = (eol_char | eof | (char_start_comment (not_eol_char * )))
let block_comment = comment_char '{'

let lowercase = [ 'a' - 'z' '_' ]
let uppercase = [ 'A' - 'Z' ]
let ident = [ 'A' - 'Z' 'a' - 'z' '_' '$' ]
            ([ '0' - '9' ] | [ 'A' - 'Z' 'a' - 'z' '_' '$' ]) *
let identchar = [ 'A' - 'Z' 'a' - 'z' '_' '$' ]
let digit = [ '0' - '9' ]
let decimal_literal = digit+

(* numbers: take care that 1./ is 1 ./ *)

let float_literal =
  ((['0'-'9']+ ('.' ['0'-'9']* )? (['e' 'E' 'd' 'D'] ['+' '-']? ['0'-'9']+)?)
   |(('.' ['0'-'9']+)? (['e' 'E' 'd' 'D'] ['+' '-']? ['0'-'9']+)? ))
let imag_literal =
  ((['0'-'9']+ ('.' ['0'-'9']* )? (['e' 'E' 'd' 'D'] ['+' '-']? ['0'-'9']+)?['i'-'j'])
   |(('.' ['0'-'9']+)? (['e' 'E' 'd' 'D'] ['+' '-']? ['0'-'9']+)?['i'-'j']))
let hexa = ('0'['x' 'X']['0'-'9' 'a'-'f' 'A'-'F']+)

(* extended affectations, no optional extra character here *)

let equal_ops = ('+'|'-'| '*'| '/' | '\\' | '^' | "**" )
let equal_ext = (equal_ops | '.' equal_ops | ("<<" | ">>" | '&'| '|') ) '='
let lhs_ext = "]" blanks equal_ext


(* affectations *)

let lhs = (']'  blanks  '=')
let lhs_equal_equal = ( ']' blanks  "==")

(* comma, semicolumn followed by continuation *)

let comma = ','
let semicolumn = ';'

(* detect when space can be replaced by comma *)

let mat_blank = (blank +) (continuation1)? (blank*)
(* unary + when inside matrix, comma to be inserted *)
let iscomma1 = mat_blank + '+' [^ ' ' '\t']
(* unary - when inside matrix, comma to be inserted *)
let iscomma2 = mat_blank + '-' [^ ' ' '\t']
(* unary ! or ~ * when inside matrix, comma to be inserted *)
let iscomma21 = mat_blank + ('!'|'~') [^ '=' ]
(* when inside matrix, comma to be inserted *)
let iscomma3 = mat_blank (identchar | '\"' | '\'' | digit | '@' )
let iscomma31 = mat_blank ('[' | '{' | '(' )
let iscomma4 = mat_blank ( '.'  digit )

(* used for word commands *)

let compare = "<" | "<=" | ">" | ">=" | "==" | "~=" | "!="
let operators = "*" | ".*" | "+" | ".+" | "-" |".-" |"--" |
  "/" |"./" |"\\" |".\\" |"^" |".^"
let wcmd_char = [^ '\n' ',' ';' '%' '#' '{'] | ('\r' [^ '\n' ])

(* tokenizers *)

rule token = parse

| blanks ":" blanks
    {
     incr_lines lexbuf (Lexing.lexeme lexbuf);
     if debug then prerr_endline  "colon={}";
     if debug then prerr_endline (Printf.sprintf "[%s]" (Lexing.lexeme lexbuf));
     is_transpose := false; COLON;
   }
|  blank* (equal_ext as s) blanks
    {
     incr_lines lexbuf (Lexing.lexeme lexbuf);
     if debug then prerr_endline (Printf.sprintf  "equal_ext={%s}" s);
     is_transpose := false;
     EQUAL_EXT(s)
   }
| blank* "++" {is_transpose:= false; ADD_ADD }
| blank* "--" {is_transpose:= false; MINUS_MINUS }
| iscomma1 as s
    {
     incr_lines lexbuf s;
      (* a unary + after a blank *)
      is_transpose:= false;
      if debug then prerr_endline (Printf.sprintf  "comma1=[%s]" s);
      push_back lexbuf (String.get s ((String.length s) -1));
      let (mat,_) = top_stack () in
      if (mat) then
        (
         push_back lexbuf (String.get s ((String.length s) -2));
         COMMA)
      else
        ( ADD )
    }
| iscomma2 as s
    {
     incr_lines lexbuf s;
      (* a unary - after a blank *)
      is_transpose:= false;
      if debug then prerr_endline (Printf.sprintf  "comma2=[%s]" s);
      push_back lexbuf (String.get s ((String.length s) -1));
      let (mat,_) = top_stack () in
      if (mat) then
        (
         push_back lexbuf (String.get s ((String.length s) -2));
         COMMA)
      else
        MINUS
    }
| ( iscomma21 as s) (blanks as x)
    {
     incr_lines lexbuf (Lexing.lexeme lexbuf); (* XXX: Pb si push_back *)
      (* a unary ! or ~ *)
      if debug then prerr_endline (Printf.sprintf  "comma21=[%s]" s);
      let (mat,_) = top_stack () in
      if (mat) then
        (
         push_back_strings lexbuf [x];
         push_back lexbuf (String.get s ((String.length s) -1));
         push_back lexbuf (String.get s ((String.length s) -2));
         COMMA)
      else
        (
         let s = (String.get s ((String.length s) -1)) in
         if (s= ' ' || s = '\t') then   (NOT)
         else
           ( push_back_strings lexbuf [x];
             push_back lexbuf s;
             NOT )
        )
    }
| iscomma3 as s
    {
     incr_lines lexbuf s;
      let c = (String.get s ((String.length s) -1)) in
      push_back lexbuf c;
      if debug then prerr_endline (Printf.sprintf  "comma3=[%s]" s);
      let (mat,_) = top_stack () in
      if c= '\'' then is_transpose:= false;
      if (mat) then
        (
          COMMA)
      else
        (
         SPACE
        )
    }
| (iscomma31 as s)
   {
     incr_lines lexbuf s;
     let c = (String.get s ((String.length s) -1)) in
     push_back lexbuf c;
     if debug then prerr_endline (Printf.sprintf  "comma31=[%s]" s);
     let (mat,_) = top_stack () in
     if c= '\'' then is_transpose:= false;
     if (mat ) then
       (
         COMMA
        )
     else
       (
        token lexbuf
       )
  }
| iscomma4 as s
    {
     incr_lines lexbuf s;
      if debug then prerr_endline (Printf.sprintf  "comma4=[%s]" s);
      push_back lexbuf (String.get s ((String.length s) -1));
      push_back lexbuf (String.get s ((String.length s) -2));
      let (mat,_) = top_stack () in
      if (mat) then
        (
          is_transpose:= false;COMMA)
      else
        (token lexbuf)
    }
| continuation1 as s
    {
     incr_lines lexbuf s;
     if debug then prerr_endline (Printf.sprintf  "continuation1=[%s]" s);
     is_transpose:= false; token lexbuf
   }
| continuation2 as s
    {
     incr_lines lexbuf s;
     if debug then prerr_endline (Printf.sprintf  "continuation=[%s]" s);
     is_transpose:= false; token lexbuf }

| blanks ']'
    {
     incr_lines lexbuf (Lexing.lexeme lexbuf);
     if debug then prerr_endline  "rbracket"; is_transpose:= true; pop_flag();
     if debug then
       (
        prerr_endline (Printf.sprintf  "pop_flag %d" (List.length !flag_stack));
        if flag_stack_empty () then
          prerr_endline  "pop_flag is (false,false)"
        else
          prerr_endline  "pop_flag is wrong");
     RBRACKET}
| blanks '}'
    {
     incr_lines lexbuf (Lexing.lexeme lexbuf);
     if debug then prerr_endline  "rbrace";   is_transpose:= true; pop_flag();
     RBRACE}

| (ident as s) (blank+ as bl) (wcmd_char + as rest)
    {
     (* trying to detect a word command list at the start of a file *)
     if debug then prerr_endline (Printf.sprintf  "wlc tested=[%s,%s,%s]" s bl rest);
     if lexbuf.lex_start_pos <> 0 ||
        not (flag_stack_empty ()) ||
        Hashtbl.mem keyword_table s ||
        not (test_wcmd rest wcmdl)
     then
       (
         (* a wcmd cannot be here *)
         if debug then prerr_endline  "not a wcmd";
         push_back_strings lexbuf [rest];
         is_transpose:= false;
         incr_line_num lexbuf;
         mk_keyword_or_ident lexbuf s bl;
        )
     else
       (
        incr_lines lexbuf (Lexing.lexeme lexbuf);
        if debug then prerr_endline (Printf.sprintf  "wlc found=[%s,%s]" s rest);
        NEWLINE_WCMD(s, split_wcmd rest split_wcmdl , Newline 1);
       )
   }

| (lowercase + as s)  (blanks as bl) '{'
    {
     incr_lines lexbuf bl;
     let rep = mk_keyword_or_ident lexbuf s bl in
     if is_ident rep then mk_ident_check_accol_type lexbuf false bl;
     push_back lexbuf '{';
     rep
   }
| ( '~' as c)  (blanks as bl) ']'
    {
     let s = String.make 1 c in
     incr_lines lexbuf bl;
     let rep = mk_keyword_or_ident lexbuf s bl in
     push_back lexbuf ']';
     rep
    }
| ( '~' as c)  (blanks as bl) ','
    {
     let s = String.make 1 c in
     incr_lines lexbuf bl;
     let rep = mk_keyword_or_ident lexbuf s bl in
     push_back lexbuf ',';
     rep
    }

| (lowercase + as s)  (blanks as bl)
    {
     incr_lines lexbuf bl;
     mk_keyword_or_ident lexbuf s bl
(*
     let rep = mk_keyword_or_ident lexbuf s bl in
     match rep with
     | IDENT _ -> rep
     | _ -> rep
*)
    }

| ( ident as s ) (blanks as bl) '{'
    {
     (* special case where a '{' should not be treated as inside cell *)
     if debug then prerr_endline (Printf.sprintf  "ident=[%s]" s);
     (* check the status of white space in this zone *)
     mk_ident_check_accol_type lexbuf true bl;
     push_back lexbuf '{';
     transpose_activate (IDENT(Lexing.lexeme lexbuf))
   }

| ident as s
    {
     if debug then prerr_endline (Printf.sprintf  "ident=[%s]" s);
     transpose_activate (IDENT(Lexing.lexeme lexbuf))
    }

| (newline as nl) (ident as s) (blank+ as bl) (wcmd_char + as rest)
    {
     (* trying to insert a word command list *)
     if debug then
       prerr_endline (Printf.sprintf  "wlc tested=[%s,%s,%s]" s bl rest);
     if not (flag_stack_empty ()) ||
        Hashtbl.mem keyword_table s ||
        not (test_wcmd rest wcmdl)
     then
       (
         (* a wcmd cannot be here *)
         if debug then prerr_endline  "not a wcmd";
         push_back_strings lexbuf (rest::bl::[s]);
         is_transpose:= false;
         incr_line_num lexbuf;
         match top_stack () with
         | (true,_)  -> NEWLINE_SEMICOLUMN
         | (_,true)  -> token lexbuf
         | (_,false) -> 
	     if debug then prerr_endline (Printf.sprintf  "newline-found=[%d]" (count_newlines nl));
	     NEWLINE(count_newlines nl)
        )
     else
       (
        incr_lines lexbuf (Lexing.lexeme lexbuf);
        if debug then
          prerr_endline (Printf.sprintf  "wlc found=[%s,%s]" s rest);
        NEWLINE_WCMD (s, split_wcmd rest split_wcmdl, Newline 1);
       )
   }

| blanks (semicolumn | comma as sep) blanks (ident as s) (blank+ as bl) (wcmd_char + as rest)
    {
     incr_lines lexbuf (Lexing.lexeme lexbuf);
     (* trying to insert a word command list *)
     if debug then prerr_endline (Printf.sprintf  "wlc tested=[%s,%s,%s]" s bl rest);
     if not (flag_stack_empty ()) ||
        Hashtbl.mem keyword_table s ||
        not (test_wcmd rest wcmdl)
     then
       (
         (* a wcmd cannot be here *)
         if debug then prerr_endline  "not a wcmd";
         push_back_strings lexbuf (rest::bl::[s]);
         is_transpose:= false;
         if sep = ';' then
           SEMICOLUMN
         else
           COMMA
        )
     else
       (
        if debug then prerr_endline (Printf.sprintf  "wlc found=[%s,%s]" s rest);
        WCMD (s, split_wcmd rest split_wcmdl,if sep = ';' then (Semi 1) else (Comma 1));
       )
   }

| ( (newline | newline semicolumn+ blanks ) as nl)
    {
        incr_lines lexbuf (Lexing.lexeme lexbuf);
        if debug then  prerr_endline "a newline=[]";
        is_transpose:= false;
        match top_stack () with
        | (true,_)  -> NEWLINE_SEMICOLUMN
        | (_,true)  -> token lexbuf
        | (_,false) -> 
	    if debug then prerr_endline (Printf.sprintf  "newline-found=[%d]" (count_newlines nl));
	    NEWLINE(count_newlines nl)
    }

| decimal_literal  { is_transpose:= true; NUM(Lexing.lexeme lexbuf) }
| hexa             { is_transpose:= true; NUM(Lexing.lexeme lexbuf) }
| (['0'-'9']+ as num) ('.' '/')
    { is_transpose:= true;
      push_back lexbuf '/';
      push_back lexbuf '.';
      NUM(num) }
| float_literal    { is_transpose:= true; NUM(Lexing.lexeme lexbuf) }
| imag_literal     { is_transpose:= true; IMAG_NUM(Lexing.lexeme lexbuf) }
| "\""
    { let string_istart = Lexing.lexeme_start lexbuf in
      let string_start = Some (lexbuf.lex_start_p, lexbuf.lex_curr_p) in
      string_start_pos := string_start;
      string lexbuf;
      lexbuf.Lexing.lex_start_pos <-
        string_istart - lexbuf.Lexing.lex_abs_pos;
      is_transpose:= true;
      let str = get_stored_string() in
      if debug then prerr_endline (Printf.sprintf  "string=[%s]" str);
      DQ_STRING (str);
    }

| block_comment
    {
      (* nested comments in matlab 7 and octave *)
      is_transpose:= false;
      (* let string_istart = Lexing.lexeme_start lexbuf in *)
      let string_start = Some (lexbuf.lex_start_p, lexbuf.lex_curr_p) in
      string_start_pos := string_start;
      comment_start_pos := string_start;
      store_string_char('%');
      store_string_char('{');
      comment 0 lexbuf;
      (*
      lexbuf.Lexing.lex_start_pos <-
         string_istart - lexbuf.Lexing.lex_abs_pos;
       *)
      let str = get_stored_string() in
      if debug then prerr_endline (Printf.sprintf  "block_comment=[%s]" str);
      COMMENTS(str);
    }

| blanks (comment_char as c) (line_comment as s)
    {
     let s1 =
       if s = "\n" then
         ( push_back lexbuf '\n'; "" )
       else
       if s = "\r\n" then
         ( push_back lexbuf '\r'; push_back lexbuf '\n'; "" )
       else
         s in
     if debug then prerr_endline (Printf.sprintf  "comment={%c,%s}" c s1);
     is_transpose:= false;
     if c = '#' then
       DCOMMENT(s1)
     else
       PCOMMENT(s1)
   }

| blanks (lhs_equal_equal as s) blanks
    {
     incr_lines lexbuf (Lexing.lexeme lexbuf);
     (* close a lhs and == *)
     if debug then prerr_endline (Printf.sprintf  "lhs={%s}" s);
     pop_flag ();
     is_transpose:= false;
     RBRACKET_EQUAL  (* rajouter un nouvel oprateur *)
    }
| blanks (lhs as s ) blanks
    {
     incr_lines lexbuf (Lexing.lexeme lexbuf);
     (* close a lhs and = *)
     if debug then prerr_endline (Printf.sprintf  "lhs={%s}" s);
     pop_flag ();
     is_transpose:= false;
     RBRACKET_EQUAL
   }
| blanks (lhs_ext as s) blanks
    {
     incr_lines lexbuf (Lexing.lexeme lexbuf);
     if debug then prerr_endline (Printf.sprintf  "lhs_ext={%s}" s);
     pop_flag ();
     is_transpose:= false;
     RBRACKET_EQUAL_EXT(s)
   }
| blanks "(" blanks  {     incr_lines lexbuf (Lexing.lexeme lexbuf);
           is_transpose:= false; push_flag (false, true); LPAREN }
| blanks ")"  {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= true; pop_flag (); RPAREN }
| blanks "*" blanks as s
    {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; STARS((Star,s)) }
| blanks ".*" blanks as s {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; STARS((Dot_star,s)) }
| blanks '+' blanksp  {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; ADDW } (* this one is binary *)
| blanks "+" blanks {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; ADD } (* this one can be unary *)
| blanks ".+" blanks {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; DOT_ADD }
| blanks '-' blanksp    {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; MINUSW }
| blanks "-" blanks {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; MINUS }
| blanks ".-" blanks {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; DOT_MINUS }
| blanks "--" {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; MINUS_MINUS }
| blanks "/" blanks as s {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; DIVS((Slash ,s))}
| blanks "./" blanks as s {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; DIVS((Dot_slash ,s))}
| blanks "\\" blanks as s {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; DIVS((Bslash ,s))}
| blanks ".\\" blanks as s {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; DIVS((Dot_bslash ,s))}
| blanks "^" blanks {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; POW }
| blanks ".^" blanks {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; DOT_POW }
| blanks semicolumn blanks
    {
     if debug then prerr_endline  "semi=[;]";
     incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; SEMICOLUMN }
| blanks comma blanks
    {
     if debug then prerr_endline  "comma=[,]";
     incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; COMMA }
| blanks "<"  blanks as s
    {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false;  COMPARE(Lt,s) }
| blanks "<=" blanks as s
    {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false;  COMPARE((Le,s)) }
| blanks ">"  blanks as s
    {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; COMPARE((Gt,s)) }
| blanks ">=" blanks as s
    {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; COMPARE((Ge,s)) }
| blanks "==" blanks as s
    {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; COMPARE((Eq,s)) }
| blanks ("~=" | "!=" ) blanks as s
    {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; COMPARE((Ne,s)) }
| blanks "=" blanks
    {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false;EQUAL}
| "!"  blanks
    {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; NOT } (* octave *)
| "~"  blanks
    {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; NOT } (* matlab *)
| ("[" blanks ) | ("[" newline)
   {
     incr_lines lexbuf (Lexing.lexeme lexbuf);
     if debug then prerr_endline (Printf.sprintf  "lbracket flag length=%d" (List.length !flag_stack));
     is_transpose:= false; push_flag (true,false); LBRACKET
   }
| blanks "]"
    {
     incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= true; pop_flag(); RBRACKET }
| ("{"  blanks ) | ("{" newline)
    {
     incr_lines lexbuf (Lexing.lexeme lexbuf);
     is_transpose:= false;
     if !flag_accol then push_flag (false, false) else push_flag (true, false);
     flag_accol := false;
     LBRACE
   }
| (blanks|newline) "}"
    {
     incr_lines lexbuf (Lexing.lexeme lexbuf);
     is_transpose:= true; pop_flag(); RBRACE }
| (blanks|newline) "||" blanks
    {
     incr_lines lexbuf (Lexing.lexeme lexbuf);
     is_transpose:= false; OR_OR
    }
| blanks "|"  blanks {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; OR }
| blanks "&&" blanks {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; AND_AND }
| blanks "&"  blanks {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; AND }
| blanks "@"  {incr_lines lexbuf (Lexing.lexeme lexbuf);is_transpose:= false; ATMARK }
| "'"
    {
     if !is_transpose then (
       if debug then prerr_endline  "quote";
       QUOTE )
     else
       (
         if debug then prerr_endline  "string start";
         let string_istart = Lexing.lexeme_start lexbuf in
         let string_start = Some (lexbuf.lex_start_p, lexbuf.lex_curr_p) in
         string_start_pos := string_start;
         string_sq lexbuf;
         lexbuf.Lexing.lex_start_pos <-
           string_istart - lexbuf.Lexing.lex_abs_pos;
         is_transpose:= true;
         let str = get_stored_string() in
         if debug then prerr_endline (Printf.sprintf  "string={%s}" str);
         SQ_STRING (str);
        )
   }
| ".\'" { TRANSPOSE }
| "." blank* (ident as s)
  {
    if debug then prerr_endline (Printf.sprintf "dot={%s}" s);
    DOTIDENT(s)
  }
| "." blank*  { DOT }
| blank +     { SPACE }
| blank * (semicolumn | comma) * blank * (comment_char line_comment) ?
    eof { EOF }
| _
    { error
        (Illegal_character (Lexing.lexeme_char lexbuf 0),
         lexbuf.lex_start_p,
         lexbuf.lex_curr_p) }

(* comments *)

and comment level = parse
  | "%}"
      {
        store_string_char('%');
        store_string_char('}');
        if debug then prerr_endline (Printf.sprintf  "comments (%d) end" level);
          if level = 0 then ()
        else comment (level-1) lexbuf
      }
  | "%{"
      {
       store_string_char('%');
       store_string_char('{');
       comment (level+1) lexbuf
      }

  | eof
    { match !comment_start_pos with
      | Some (start_pos, end_pos) ->
        error (Unterminated_comment, start_pos, end_pos)
      | _ -> assert false }

  | _
      { store_string_char(Lexing.lexeme_char lexbuf 0);
        comment level lexbuf }

(* strings with double quote *)

and string = parse  (* strings: a completer pour nsp *)
   '\"' '\"'
    { store_string_char('\"');
      store_string_char('\"');
      string lexbuf
    }
  |
    '\"'
      { (* stop *)  () }
  | '\\' ("\010" | "\013" | "\013\010") [' ' '\009'] *   (* octave string continuation *)
      { string lexbuf }
  | '\\' ['\\' '\"' '?' '\'' 'a' 'n' 't' 'b' 'r' 'v' 'f']
      {
        store_string_char(Lexing.lexeme_char lexbuf 0);
        store_string_char(Lexing.lexeme_char lexbuf 1);
        string lexbuf }
  | eof
    { match !string_start_pos with
      | Some (start_pos, end_pos) ->
        error (Unterminated_string, start_pos, end_pos)
      | _ -> assert false }

  | _
      { store_string_char(Lexing.lexeme_char lexbuf 0);
        string lexbuf }

(* strings with single quote *)

and string_sq = parse  (* strings: a completer pour nsp *)
  | "\'\'"
    { store_string_char('\'');
      string_sq lexbuf
    }
  | '\''
      { (* stop *)
        () }
  | '\\' ("\010" | "\013" | "\013\010") [' ' '\009'] *
      { string_sq lexbuf }
  | eof
    { match !string_start_pos with
      | Some (start_pos, end_pos) ->
        error (Unterminated_string, start_pos, end_pos)
      | _ -> assert false }

  | _
      { store_string_char(Lexing.lexeme_char lexbuf 0);
        string_sq lexbuf }

(* rules for wcmdl :
 * this lexer is used by test_wcmd to check for
 * matlab word command lists
 *)

and wcmdl str_list = parse
  | blank* '{' as s
    {
     if debug then prerr_endline (Printf.sprintf  "accol={%s}" s);
     wcmdl ( "accol" :: str_list ) lexbuf
   }
  | blank* "(" as s
    {
     if debug then prerr_endline (Printf.sprintf  "par={%s}" s);
     wcmdl ( "par" :: str_list ) lexbuf
   }

  | blank* ( compare as s ) blank*
    {
      if debug then prerr_endline (Printf.sprintf  "compare={%s}" s);
      wcmdl ( "op" :: str_list) lexbuf
    }
  | blank* ( equal_ext as s) blank*
    {
     if debug then prerr_endline (Printf.sprintf  "equalext={%s}" s);
     wcmdl ( "op" :: str_list) lexbuf
    }

  | blank* ( "=" as s ) blank*
    {
     if debug then prerr_endline (Printf.sprintf  "equal={%c}" s);
     wcmdl ( "op" :: str_list) lexbuf
    }

  | blank* ( operators as s ) blank+
    {
     if debug then prerr_endline (Printf.sprintf  "equal={%s}" s);
     wcmdl ( "op" :: str_list) lexbuf
    }

  | _ as s
    {
     if debug then prerr_endline (Printf.sprintf  "misc={%c}" s);
     wcmdl ( "misc" :: str_list) lexbuf
    }
  | eof { str_list  }

(* rules for split_wcmdl
 * this lexer is used by split_wcmd
 *)

and split_wcmdl str_list = parse
  | blank+ '\"'
    {
      let string_istart = Lexing.lexeme_start lexbuf in
      let string_start = Some (lexbuf.lex_start_p, lexbuf.lex_curr_p) in
      string_start_pos := string_start;
      wcmd_string lexbuf;
      lexbuf.Lexing.lex_start_pos <-
        string_istart - lexbuf.Lexing.lex_abs_pos;
      let str = get_stored_string() in
      if debug then prerr_endline (Printf.sprintf  "string_in_split1={%s}" str);
      split_wcmdl ( DQ(str) :: str_list) lexbuf;
    }

  | blank+ '\''
    {
     let string_istart = Lexing.lexeme_start lexbuf in
     let string_start = Some (lexbuf.lex_start_p, lexbuf.lex_curr_p) in
     string_start_pos := string_start;
     wcmd_string_sq lexbuf;
     lexbuf.Lexing.lex_start_pos <-
       string_istart - lexbuf.Lexing.lex_abs_pos;
     let str = get_stored_string() in
     if debug then prerr_endline (Printf.sprintf  "string_in_split2={%s}" str);
     split_wcmdl ( SQ(str) :: str_list) lexbuf;
   }
  | blank+ ([^ '\'' '\"'] as c)
    {
     let string_istart = Lexing.lexeme_start lexbuf in
     let string_start = Some (lexbuf.lex_start_p, lexbuf.lex_curr_p) in
     string_start_pos := string_start;
     store_string_char c;
     wcmd_string_nd lexbuf;
     lexbuf.Lexing.lex_start_pos <-
       string_istart - lexbuf.Lexing.lex_abs_pos;
     let str = get_stored_string() in
     if debug then prerr_endline (Printf.sprintf  "string_in_split3={%s}" str);
     if c != ' ' then
       split_wcmdl ( SQ(str) :: str_list) lexbuf
     else
       split_wcmdl str_list lexbuf;
    }
  | _ as c
    {
     if debug then prerr_endline (Printf.sprintf  "lost_in_split={%c}" c);
     split_wcmdl str_list lexbuf
    }
  | eof { str_list }

(* strings with double quote *)

and wcmd_string = parse
  | '\"' '\"'  { store_string_char('\"'); store_string_char('\"');  wcmd_string lexbuf  }
  | '\"'     { (* stop *)  () }
  | '\\' ("\010" | "\013" | "\013\010") [' ' '\009'] *   { wcmd_string lexbuf }
  | '\\' ['\\' '\"' '?' '\'' 'a' 'n' 't' 'b' 'r' 'v' 'f']
      { store_string_char(Lexing.lexeme_char lexbuf 1);   wcmd_string lexbuf }
  | eof  { (* stop *)  () }
  | '\r' eof { (* stop *)  () }
  | _  { store_string_char(Lexing.lexeme_char lexbuf 0);  wcmd_string lexbuf }

(* strings with single *)

and wcmd_string_sq = parse  (* strings: a completer pour nsp *)
  | "\'\'"  { store_string_char('\'');  wcmd_string_sq lexbuf  }
  | '\''  { (* stop *)        () }
  | '\\' ("\010" | "\013" | "\013\010") [' ' '\009'] * { wcmd_string_sq lexbuf }
  | eof   {  (* stop *) () }
  | '\r' eof   {  (* stop *) () }
  | _
      { store_string_char(Lexing.lexeme_char lexbuf 0);   wcmd_string_sq lexbuf }

(* up to next blank as str *)

and wcmd_string_nd = parse
  | (' ' | '\t') as c  {  push_back lexbuf c; ()}
  | eof   {  (* stop *) () }
  | '\r' eof   {  (* stop *) () }
  | _  { store_string_char(Lexing.lexeme_char lexbuf 0);   wcmd_string_nd lexbuf }

{

(* zone for extra functions *)

let init () = is_transpose := false; clear_flag_stack ();;

(*

let main () =
  let cin =
    if Array.length Sys.argv > 1
    then open_in Sys.argv.(1)
    else stdin
  in
  let lexbuf = Lexing.from_channel cin in
  let str_list = wcmdl [] lexbuf in
  if (check_wcmd (List.rev str_list)) then
    prerr_endline (Printf.sprintf  "wcmd ok"
  else
    prerr_endline (Printf.sprintf  "wcmd false";
  prerr_endline (Printf.sprintf  "%s" "res=";
  print_lexemes (List.rev str_list)
  let _ = Printexc.print main ()

*)

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)

}
