/***********************************************************************/
/*                                                                     */
/*                           Matlab Parser                             */
/*                                                                     */
/*          J.Ph Chancelier, Enpc/Cermics                              */
/*                                                                     */
/*  Copyright 2012-2019,                                               */
/*  Ecole Nationale des ponts et chaussées                             */
/*  All rights reserved.                                               */
/*                                                                     */
/*  This file is distributed under the terms of the BSD License.       */
/*                                                                     */
/***********************************************************************/
%{

(* $Id: mtlb_parser.mly,v 1.1 2019-05-10 14:09:18 jpc Exp $ *)

open Mtlb_ast;;
open Mtlb_ast_funs;;

type 'info expr_or_sep_desc =
  | Expr of 'info Mtlb_ast.expr
  | Sep of 'info Mtlb_ast.sep
;;

(* checks if an Expr is a global or persistent statement *)

let expr_is_global_declaration exp =
  match exp.Ast_node.ast_desc with
  | Command c ->
      (match c.Ast_node.ast_desc with
      | Declaration _ -> true
      | Switch _
      | IfCommand  _ 
      | LoopCommand _ 
      | JumpCommand _ 
      | ExceptCommand _
      | SciptFile _
      | MFunction _ -> false 
      )
  | Expression _
  | WordListCmd _
  | Comments _
  | Pcomments _
  | Dcomments _
  | Empty -> false

(* recombine a list of Sep,Expr to a list of ExprSep
 *  (Sep(sep1),Expr(exp1),Sep(sep2),Expr(exp2),...) -> 
 *    (Exprsep(exp1,sep1),Exprsep(exp2,sep2),...)
 * There's a special case when Expr is a Declaration 
 * in that case the separator is modified to force a SEMI
 *)

let rec mk_expr_list_no_rev  = function
  | [] ->  []
  | sep :: (expr as e2) :: elts ->
      (match sep.Ast_node.ast_desc with
      | Sep y ->
          (match expr.Ast_node.ast_desc with
          | Expr x ->
	      let expr =
		if  expr_is_global_declaration x then
		  let newsep =
		    match y.Ast_node.ast_desc with
		    | Comma _ 
		    | Semi  _ 
		    | Sep_None _ -> y
		    | Newline n -> { y with Ast_node.ast_desc = Semi n;} in
		  { e2 with Ast_node.ast_desc = Exprsep (x, newsep); }
		else
		  { e2 with Ast_node.ast_desc = Exprsep (x, y); } in
	      expr :: mk_expr_list_no_rev elts
          | Sep _ -> [] )
      | Expr _ -> [] )
  | _ -> []
;;

let mk_expr_list x = List.rev (mk_expr_list_no_rev x);;

(* get the k value in [mk [mk (Exprsep(mk Empty,mk (Newline k)))]]  *)

let get_newlines_number expr =
  let get_newlines_number_in_sep sep =
    match sep with
    | ( Comma _ | Semi _ | Sep_None _ ) -> 0
    | Newline n -> n
  in
  match expr with
  | [] -> 0
  | elt :: [] ->
      ( match elt.Ast_node.ast_desc with
      | [] -> 0
      | e :: [] ->
          (match e.Ast_node.ast_desc with
          | Exprsep(x,sep) ->
              (match x.Ast_node.ast_desc with
              | Empty ->
          get_newlines_number_in_sep sep.Ast_node.ast_desc
              | Expression _
              | Command _
              | WordListCmd _
              | Comments _
              | Pcomments _
              | Dcomments _ -> 0
              )
          )
      | _ -> 0)
  | _ -> 0
;;

(* checks that a lsep is just a coded newline not a Wcmd *)

let lsep_is_newlines expr =
  match expr with
  | [] -> false
  | elt :: [] ->
      ( match elt.Ast_node.ast_desc with
      | [] -> false
      | e :: [] ->
          (match e.Ast_node.ast_desc with
          | Exprsep(x,_sep) ->
              (match x.Ast_node.ast_desc with
              | Empty -> true
              | Expression _
              | Command _
              | WordListCmd _
              | Comments _
              | Pcomments _
              | Dcomments _ -> false
              )
          )
      | _ -> false)
  | _ -> false
;;

let exprsep_list_set_newlines n = function
  | [] -> []
  | e :: es ->
      (match e.Ast_node.ast_desc with
      | Exprsep(x,sep) ->
          let newsep =
            match sep.Ast_node.ast_desc with
            | Comma n1 -> {sep with Ast_node.ast_desc = Comma (n1 + n -1);}
            | Semi  n1 -> {sep with Ast_node.ast_desc = Semi  (n1 + n -1);}
            | Sep_None n1 -> {sep with Ast_node.ast_desc = Sep_None (n1 + n -1);}
            | Newline n1 -> {sep with Ast_node.ast_desc = Newline (n1 + n -1);}
          in
          let newe = { e  with Ast_node.ast_desc = Exprsep(x,newsep);} in
          newe :: es;
      )

%}

/* Tokens */

%token <string * (Mtlb_ast.parsing Mtlb_ast.wcmds_desc list) * (Mtlb_ast.parsing Mtlb_ast.sep_desc) > NEWLINE_WCMD
%token <string * (Mtlb_ast.parsing Mtlb_ast.wcmds_desc list) * (Mtlb_ast.parsing Mtlb_ast.sep_desc) > WCMD

/* MINUS and ADD can be unary MINUSW and ADDW are always binary  */
%token EQUAL COLON MINUS MINUSW ADD ADDW

/* logicals */
%token AND_AND OR_OR AND OR NOT

/* comparisons: LT LE EQ NE GE GT */
%token <Mtlb_ast.parsing Mtlb_ast.op_compare_desc * string > COMPARE

/* multiply and like: STAR DOT_STAR */
%token <Mtlb_ast.parsing Mtlb_ast.op_stars_desc * string> STARS
/* divide and like:  SLASH BSLASH DOT_SLASH DOT_BSLASH */
%token <Mtlb_ast.parsing Mtlb_ast.op_divs_desc * string> DIVS

%token DOT_ADD DOT_MINUS
/* transposition: quote ' and transpose .' */
%token QUOTE TRANSPOSE
%token ADD_ADD MINUS_MINUS
%token POW DOT_POW

/*block, %comments and #comments */
%token <string>  COMMENTS PCOMMENT DCOMMENT
%token <string>  NUM IMAG_NUM
%token <string>  STRUCT_ELT
%token <string>  IDENT
/* all the ends */
%token END
/* double and single quote strings */
%token <string>  DQ_STRING SQ_STRING
%token FOR WHILE DO UNTIL IF ELSEIF ELSE SWITCH CASE OTHERWISE
%token BREAK CONTINUE FUNC_RET
%token UNWIND CLEANUP TRY CATCH
%token GLOBAL STATIC
%token COLONEQUAL

/* */
%token <string>  FUNCTION_HANDLE
%token <string>  PROPERTIES METHODS EVENTS ENUMERATION
%token <string>  METAQUERY
%token <string>  SUPERCLASSREF
%token <string>  GET SET

%token <int> NEWLINE
%token COMMA SEMICOLUMN NEWLINE_SEMICOLUMN NEWLINE_G
%token LPAREN RPAREN LBRACE RBRACE LBRACKET RBRACKET RBRACKET_EQUAL
%token END_OF_INPUT
%token FUNCTION CLASSDEF
%token EOF
%token ATMARK

%token <string> EQUAL_EXT
%token <string> RBRACKET_EQUAL_EXT
%token DOT
%token <string> DOTIDENT
%token UADD UMINUS UADD_ADD UMINUS_MINUS
%token SPACE

// Precedence and associativity.
%left SPACE
%right EQUAL_EXT
%left OR_OR
%left AND_AND
%left OR
%left AND
%left COMPARE // LT LE EQ NE GE GT
%left LSHIFT RSHIFT
%left COLON
%left MINUS MINUSW ADD ADDW DOT_ADD DOT_MINUS
%left STARS DIVS
%right NOT
%left POW DOT_POW QUOTE TRANSPOSE
%right ADD_ADD MINUS_MINUS
%nonassoc UADD_ADD UMINUS_MINUS
%nonassoc UADD UMINUS
%left LPAREN DOT DOTIDENT LBRACE

// Where to start.

%start mtlb_ast
%type <Mtlb_ast.parsing Mtlb_ast.mtlb_ast> mtlb_ast

%%

mtlb_ast :
  | exprs1 EOF
    {
     let exprs = List.rev $1 in
     let (a,b) = (fct_count_exprs_list exprs) in
     if ( a > 1 && b = 0) then
       mk (Ast(fct_extract exprs))
     else
       mk (Ast(exprs))
    }
  | SPACE exprs1 EOF
    {
     let exprs = List.rev $2 in
     let (a,b) = (fct_count_exprs_list exprs) in
     if ( a > 1 && b = 0) then
       mk (Ast(fct_extract exprs))
     else
       mk (Ast(exprs))
   }

/* block expressions
 *  exprs1 is very similar to exprs but can be ended by END or be empty.
 * It is used in block (if, for, while,...).
 */

exprs1 : {[]}
  | lexpr3 sep1 { [(mk (mk_expr_list ($2 @ $1))) ] }
  | lexpr3     { [mk (mk_expr_list  ((mk (Sep(mk (Newline 1)))) :: $1))] }
  | sep1 lexpr3 { [mk (mk_expr_list ((mk (Sep(mk (Newline 1)))) :: $2))] }
  | sep1 lexpr3 sep1 { [mk (mk_expr_list ($3 @ $2))] }
  | exprs2             { $1 }
  | exprs2 lexpr3 sep1 { (mk (mk_expr_list ($3 @ $2)))     :: $1}
  | exprs2 lexpr3      { (mk (mk_expr_list ((mk (Sep(mk (Newline 1)))) :: $2)))::$1}
;

exprs2 :
  | lsep
      {
       if lsep_is_newlines $1 then  [] else $1
      }
  | lexpr2     { $1 }
  | exprs2 lexpr2  { $2 @ $1 }
;

lexpr2 :
  | sep1 lsep { $2 }
  | lexpr3 lsep
      {
       if lsep_is_newlines $2
       then
         let n = get_newlines_number $2 in
         [mk (mk_expr_list ((mk (Sep(mk (Newline n)))) :: $1))]
       else
         ($2) @ [mk (mk_expr_list ((mk (Sep(mk (Newline 1)))) :: $1))]
     }
  | lexpr3 sep1 lsep
      {
       if  lsep_is_newlines $3
       then
         let n = get_newlines_number $3 in
         let exprsep_list =
           List.rev (exprsep_list_set_newlines n (mk_expr_list_no_rev ($2 @ $1))) in
         [mk exprsep_list]
       else
         ($3) @ [mk (mk_expr_list ($2 @ $1))]
     }
  | sep1 lexpr3 lsep
      {
       if  lsep_is_newlines $3
       then
         let n = get_newlines_number $3 in
         [mk (mk_expr_list ((mk (Sep(mk (Newline n)))) :: $2))]
       else
         ($3) @ [mk (mk_expr_list ((mk (Sep(mk (Newline 1)))) :: $2))]
       }

  | sep1 lexpr3 sep1 lsep
      {
       if  lsep_is_newlines $4
       then
         let n = get_newlines_number $4 in
         let exprsep_list =
           List.rev (exprsep_list_set_newlines n (mk_expr_list_no_rev ($3 @ $2))) in
         [mk exprsep_list]
       else
         ($4) @ [mk (mk_expr_list ($3 @ $2))]
     }
;

lexpr3:
  | expr                   { [$1] }
  | expr1                  { [$1] }
  | WCMD
    {
     let (a,b,c)= $1 in
     (mk (Expr(mk (WordListCmd(mk a,mk b)))))::[ (mk (Sep(mk c)))]
    }
  | lexpr3 WCMD
    {
     let (a,b,c)= $2 in
     ((mk (Expr(mk (WordListCmd(mk a,mk b)))))::[ (mk (Sep(mk c)))]) @ $1
    }
  | lexpr3 sep1 expr       { $3 :: ($2 @ $1) }
  | lexpr3 sep1 expr1      { $3 :: ($2 @ $1) }
  | lexpr3 sep1 WCMD       { let (a, b, _) = $3 in
                             (mk (Expr (mk (WordListCmd (mk a, mk b))))) ::
                             ($2 @ $1) }
  | lexpr3 expr1           { $2 :: mk (Sep (mk (Sep_None 1))) :: $1 }
;

/*
 * separateurs
 */

sep1:
  | COMMA             { [ mk (Sep (mk (Comma 1))) ] }
  | SEMICOLUMN        { [ mk (Sep (mk (Semi 1))) ]  }
  | sep1 COMMA        { $1 }
  | sep1 SEMICOLUMN   { $1 }
;

lsep :
  | NEWLINE
      {
       let rep = [mk [mk (Exprsep(mk Empty,mk (Newline $1)))]] in
       (* just a test to check that get_newlines_number is properly typed *)
       (* let _n = get_newlines_number rep in *)
       rep
     }
  | NEWLINE_WCMD
      {
        let (a, b, _) = $1 in
        [mk [mk (Exprsep(mk (WordListCmd(mk a,mk b)),mk (Newline 1)))]]
      }
  | lsep NEWLINE
      {
       mk [mk (Exprsep(mk Empty,mk (Newline 1)))] :: $1
      }

  | lsep NEWLINE_WCMD
    {
     let (a,b,_) = $2 in
     mk [mk (Exprsep(mk (WordListCmd(mk a,mk b)),mk (Newline 1)))]:: $1
   }
;

/* a line expression : ie a list of expression separated by comma semicolum
 * and possibly ended by a comment
 */

expr1:
  | PCOMMENT  { mk ( Expr(mk (Pcomments($1)))); }
  | DCOMMENT  { mk ( Expr(mk (Dcomments($1)))); }
;

expr:
  | expression  { mk ( Expr(mk (Expression($1)))); }
  | command     { mk ( Expr(mk (Command($1)))); }
  | COMMENTS    { mk ( Expr(mk (Comments($1)))); }
;

/*
 * Expressions
 */

identifier: IDENT      { mk $1; }
;

superclass_identifier: SUPERCLASSREF    { mk $1; }
;

meta_identifier: METAQUERY  { mk $1; }
;

constant:
  | NUM        { mk (Constant (Num $1)); }
  | IMAG_NUM   { mk (Constant (Imag_num $1)); }
  | DQ_STRING  { mk (Constant (Dq_string $1)); }
  | SQ_STRING  { mk (Constant (Sq_string $1)); }
;

/* matrix or cell expressions */

matrix: LBRACKET mexprs1 RBRACKET { mk (PexpMatrix (List.rev $2)) }

cell: LBRACE mexprs1 RBRACE { mk (PexpCell (List.rev $2)); }

mexprs1 : {[]}
  | mexpr3 sep2 { [mk (mk_expr_list ((mk (Sep($2))) :: $1))] }
  | mexpr3     { [(mk (mk_expr_list ((mk (Sep(mk (Newline 1)))) :: $1)))] }
  | sep2 mexpr3 { [(mk (mk_expr_list ((mk (Sep(mk (Newline 1)))) :: $2)))] }
  | sep2 mexpr3 sep2 { [(mk (mk_expr_list ((mk (Sep($3))) :: $2)))] }
  | mexprs2 { $1 }
  | mexprs2 mexpr3 sep2 { (mk (mk_expr_list ((mk (Sep($3))) :: $2))) ::$1}
  | mexprs2 mexpr3  { (mk (mk_expr_list ((mk (Sep(mk (Newline 1)))) :: $2))) ::$1 }

mexprs2 :
  | lsep1           { [] }
  | mexpr2          { [($1)] }
  | mexprs2 mexpr2  { $2 :: $1 }
;

mexpr2 :
  | sep2 lsep1              { (mk [])}
  | mexpr3 lsep1            { mk (mk_expr_list ((mk (Sep($2))) :: $1)) }
  | sep2 mexpr3 lsep1       { mk (mk_expr_list ((mk (Sep($3))) :: $2)) }
  | mexpr3 sep2 lsep1       { mk (mk_expr_list ((mk (Sep($2))) :: $1)) }
  | sep2 mexpr3 sep2 lsep1  { mk (mk_expr_list ((mk (Sep($3))) :: $2)) }
;

mexpr3:
  | mexpr              { [$1] }
  | expr1              { [$1] }
  | mexpr3 msep1 mexpr { $3 :: (mk (Sep($2))) :: $1 }
  | mexpr3 msep1 expr1 { $3 :: (mk (Sep($2))) :: $1 }
  | mexpr3 expr1       { $2 :: (mk (Sep(mk (Sep_None 1)))) :: $1 }

mexpr:
  | zexpression
      { mk (Expr((mk (Expression($1))))); }
  | COMMENTS
      { mk (Expr((mk (Comments($1))))); }
;

msep1:
  | COMMA          { mk (Comma 1); }
  | SEMICOLUMN     { mk (Semi 1); }
;

lsep1:
  | NEWLINE { mk (Newline 1); }
  | NEWLINE_SEMICOLUMN { mk (Newline 1); }
  | lsep1 NEWLINE { mk (Newline 1); }
  | lsep1 NEWLINE  NEWLINE_SEMICOLUMN { mk (Newline 1); }

sep2:
  | COMMA          { mk (Comma 1); }
  | SEMICOLUMN     { mk (Semi 1); }
  | sep2 COMMA     { $1 }
  | sep2 SEMICOLUMN{ $1 }

/* */

/*  ( exp, exp, exp,... ) with optional newlines between exps
 *  this is an octave extension to authorize newlines
 */

paren_arg_list:
  | parg         { $1 }
  | paren_arg_list opt_oper_sep COMMA opt_oper_sep parg { $1 @ $5 }

parg:
  | zexpression { [(mk (ArgExpression $1))] }
  | COLON       { [(mk ArgColon)] }
;

primary_expr:
  | identifier                { mk ( PexpIdentifier $1 ); }
  | constant                  { mk ( PexpConstant $1 ); }
  | ATMARK identifier         { mk ( PexpFcnHandle $2 ); }
  | matrix                    { $1 }
  | cell                      { $1 }
  | meta_identifier           { mk ( PexpMetaIdentifier $1 ); }
  | superclass_identifier     { mk ( PexpSuperIdentifier $1 ); }
  | LPAREN zexpression RPAREN { mk ( PexpExpression $2 ); }
;

oper_expr:
  | primary_expr                  { mk ( PrimaryExpr $1 ); }
  | oper_expr ADD_ADD        %prec UADD_ADD  { mk ( PostfixOp ("++", $1); ); }
  | oper_expr MINUS_MINUS    %prec UMINUS_MINUS  { mk ( PostfixOp ("--", $1); ); }
  | oper_expr LPAREN RPAREN        { mk ( IndexExpr ($1,(mk []) , "(",")" ); ); }
  | oper_expr LPAREN paren_arg_list RPAREN   { mk ( IndexExpr ($1, (mk $3), "(",")" ); ); }
  | oper_expr LBRACE RBRACE            { mk ( IndexExpr ($1,(mk []) , "{","}"); ); }
  | oper_expr LBRACE paren_arg_list RBRACE   { mk ( IndexExpr ($1,(mk $3), "{","}"); ); }
  | oper_expr QUOTE                    { mk ( PostfixOp ("'", $1); ); }
  | oper_expr TRANSPOSE                { mk ( PostfixOp (".'", $1); ); }
  | oper_expr DOTIDENT                 { mk ( IndirectRefU ($1,(mk $2)); ); }
  | oper_expr DOT LPAREN zexpression RPAREN { mk ( IndirectRef ($1, $4); ); }
  | ADD_ADD oper_expr %prec  UADD_ADD  { mk ( PrefixOpUnary ("++", $2); ); }
  | MINUS_MINUS oper_expr %prec UMINUS_MINUS  { mk ( PrefixOpUnary ("--", $2); ); }
  | NOT oper_expr                      { mk ( PrefixOpUnary ("!", $2); ); }
  | ADD oper_expr %prec UADD           { mk ( PrefixOpUnary ("+", $2); ); }
  | ADDW oper_expr %prec UADD          { mk ( PrefixOpUnary ("+", $2); ); }
  | MINUS oper_expr %prec UMINUS       { mk ( PrefixOpUnary ("-", $2); ); }
  | MINUSW oper_expr %prec UMINUS      { mk ( PrefixOpUnary ("-", $2); ); }
  | oper_expr POW oper_expr            { mk ( BinaryOp ("^", $1,  $3); ); }
  | oper_expr DOT_POW oper_expr        { mk ( BinaryOp (".^", $1,  $3); ); }
  | oper_expr ADD oper_expr            { mk ( BinaryOp ("+", $1,  $3); ); }
  | oper_expr ADDW oper_expr            { mk ( BinaryOp ("+", $1,  $3); ); }
  | oper_expr DOT_ADD oper_expr        { mk ( BinaryOp (".+", $1,  $3); ); }
  | oper_expr MINUS oper_expr          { mk ( BinaryOp ("-", $1,  $3); ); }
  | oper_expr MINUSW oper_expr            { mk ( BinaryOp ("-", $1,  $3); ); }
  | oper_expr DOT_MINUS oper_expr      { mk ( BinaryOp (".-", $1,  $3); ); }
  | oper_expr STARS oper_expr           { mk ( Op_stars (mk $2, $1,  $3); ); }
  | oper_expr DIVS oper_expr           { mk ( Op_divs (mk $2, $1,  $3); ); }
;

colon_expr:
  | oper_expr  { [$1] }
  | oper_expr COLON oper_expr  { $3 :: [$1] }
  | oper_expr COLON oper_expr COLON  oper_expr { $5:: $3 :: [$1] }

;

simple_expr:
  | colon_expr                        { mk ( ColonExpr(List.rev $1) ); }
  | simple_expr LSHIFT simple_expr    { mk ( SexpBinaryOp ("<<", $1,  $3); ); }
  | simple_expr RSHIFT simple_expr    { mk ( SexpBinaryOp (">>", $1,  $3); ); }
  | simple_expr COMPARE simple_expr   { mk ( Op_compare(mk $2, $1,  $3); ); }
  | simple_expr AND simple_expr       { mk ( SexpBinaryOp ("&", $1,  $3); ); }
  | simple_expr OR simple_expr        { mk ( SexpBinaryOp ("|", $1,  $3); ); }
  | simple_expr AND_AND simple_expr   { mk ( BooleanOp ("&&", $1, $3); ); }
  | simple_expr OR_OR simple_expr     { mk ( BooleanOp ("||", $1, $3); ); }
;

assign_expr:
  | simple_expr EQUAL expression
    { mk ( Assign("=",(mk (LhsSimple($1))), $3) ); }
  | LBRACKET mexpr3 RBRACKET_EQUAL expression
    {
     let l = (mk_expr_list ((mk (Sep(mk (Newline 1)))) :: $2)) in
     mk (Assign("=",(mk (LhsArgs(mk l))), $4));
    }

assign_ext_expr:
  | simple_expr EQUAL_EXT expression
    { mk ( Assign($2, (mk (LhsSimple($1))),$3) ); }
  | LBRACKET mexpr3 RBRACKET_EQUAL_EXT expression
    {
     let l = (mk_expr_list ((mk (Sep(mk (Newline 1)))) :: $2)) in
     mk (Assign($3,(mk (LhsArgs(mk l))), $4));
   }

expression:
  | simple_expr                   { mk ( SimpleExpr $1 ); }
  | assign_expr                   { mk ( AssignExpr $1 ); }
  | assign_ext_expr               { mk ( AssignExpr $1 ); }
  | ATMARK param_list SPACE expression  { mk ( AnonFcnHandle($2,$4) ); }
  | ATMARK param_list COMMA expression  { mk ( AnonFcnHandle($2,$4) ); }
  | ATMARK param_list SEMICOLUMN expression  { mk ( AnonFcnHandle($2,$4) ); }
  | ATMARK param_list expression  { mk ( AnonFcnHandle($2,$3) ); }


/* special expression which can contain end
 */

zprimary_expr:
/* Was
  | identifier               { mk ( PexpIdentifier $1 ); }
*/
  | param                    { mk ( PexpIdentifier $1 ); }
  | DO                       { mk ( PexpIdentifier (mk "do") ); }
  | END                      { mk ( PexpIdentifier (mk "end") ); }
  | constant                 { mk ( PexpConstant $1 ); }
  | ATMARK identifier        { mk ( PexpFcnHandle $2 ); }
  | matrix                   { $1 }
  | cell                     { $1 }
  | meta_identifier          { mk ( PexpMetaIdentifier $1 ); }
  | superclass_identifier    { mk ( PexpSuperIdentifier $1 ); }
  | LPAREN zexpression RPAREN { mk ( PexpExpression $2 ); }
;

zoper_expr:
  | zprimary_expr                { mk (PrimaryExpr $1); }
  | zoper_expr  ADD_ADD          %prec UADD_ADD     { mk (PostfixOp ("++", $1)); }
  | zoper_expr  MINUS_MINUS      %prec UMINUS_MINUS { mk (PostfixOp ("--", $1)); }
  | zoper_expr  LPAREN RPAREN        { mk (IndexExpr ($1,mk [] , "(",")" )); }
  | zoper_expr  LPAREN paren_arg_list RPAREN    { mk (IndexExpr ($1,mk $3, "(",")" )); }
  | zoper_expr  LBRACE RBRACE            { mk (IndexExpr ($1,mk [] , "{","}")); }
  | zoper_expr  LBRACE paren_arg_list RBRACE     { mk (IndexExpr ($1,mk $3, "{","}")); }
  | zoper_expr  QUOTE                    { mk (PostfixOp ("'", $1)); }
  | zoper_expr  TRANSPOSE                  { mk (PostfixOp (".'", $1)); }
  | zoper_expr  DOTIDENT            { mk (IndirectRefU ($1, mk $2)); }
  | zoper_expr  DOT LPAREN zexpression RPAREN  { mk (IndirectRef ($1, $4)); }
  | ADD_ADD  zoper_expr  %prec UADD_ADD      { mk (PrefixOpUnary ("++", $2)); }
  | MINUS_MINUS  zoper_expr  %prec UMINUS_MINUS  { mk (PrefixOpUnary ("--", $2)); }
  | NOT  zoper_expr         { mk (PrefixOpUnary ("!", $2)); }

  | ADD  zoper_expr %prec UADD          { mk (PrefixOpUnary ("+", $2)); }
  | ADDW zoper_expr %prec UADD          { mk (PrefixOpUnary ("+", $2)); }

  | MINUS  zoper_expr %prec UMINUS          { mk (PrefixOpUnary ("-", $2)); }
  | MINUSW zoper_expr %prec UMINUS          { mk (PrefixOpUnary ("-", $2)); }

  | zoper_expr  POW   zoper_expr { mk (BinaryOp ("^", $1,  $3)); }
  | zoper_expr  DOT_POW  zoper_expr       { mk (BinaryOp (".^", $1,  $3)); }

  | zoper_expr ADD  zoper_expr           { mk (BinaryOp ("+", $1,  $3)); }
  | zoper_expr ADDW  zoper_expr          { mk (BinaryOp ("+", $1,  $3)); }

  | zoper_expr MINUS  zoper_expr           { mk (BinaryOp ("-", $1,  $3)); }
  | zoper_expr MINUSW  zoper_expr          { mk (BinaryOp ("-", $1,  $3)); }

  | zoper_expr  STARS   zoper_expr        { mk (Op_stars (mk $2, $1,  $3 )); }
  | zoper_expr  DIVS   zoper_expr        { mk (Op_divs (mk $2, $1,  $3 )); }
  | zoper_expr  DOT_ADD  zoper_expr      { mk (BinaryOp (".+", $1,  $3)); }
  | zoper_expr  DOT_MINUS  zoper_expr    { mk (BinaryOp (".-", $1,  $3)); }
;

zcolon_expr:
  | zoper_expr  { [$1] }
  | zcolon_expr COLON zoper_expr  { $3 :: $1 }
;

zsimple_expr:
  | zcolon_expr                          { mk (ColonExpr(List.rev $1)); }
  | zsimple_expr  LSHIFT  zsimple_expr   { mk (SexpBinaryOp ("<<", $1,  $3)); }
  | zsimple_expr  RSHIFT  zsimple_expr   { mk (SexpBinaryOp (">>", $1,  $3)); }
  | zsimple_expr  COMPARE  zsimple_expr  { mk (Op_compare(mk $2, $1,  $3)); }
  | zsimple_expr  AND  zsimple_expr      { mk (SexpBinaryOp ("&", $1,  $3)); }
  | zsimple_expr  OR  zsimple_expr       { mk (SexpBinaryOp ("|", $1,  $3)); }
  | zsimple_expr  AND_AND  zsimple_expr  { mk (BooleanOp ("&&", $1, $3)); }
  | zsimple_expr  OR_OR zsimple_expr     { mk (BooleanOp ("||", $1, $3)); }
;

zassign_expr:
  |  zsimple_expr EQUAL zexpression
      { mk (Assign("=",mk (LhsSimple($1)), $3)) }
  | LBRACKET mexpr3 RBRACKET_EQUAL zexpression
      {
       let l = (mk_expr_list ((mk (Sep(mk (Newline 1)))) :: $2)) in
       mk (Assign("=",mk (LhsArgs(mk l)), $4))
     }

zassign_ext_expr:
  | zsimple_expr EQUAL_EXT zexpression
    { mk ( Assign($2,(mk (LhsSimple($1))),$3)) }
  | LBRACKET mexpr3 RBRACKET_EQUAL_EXT zexpression
    {
     let l = (mk_expr_list ((mk (Sep(mk (Newline 1)))) :: $2)) in
     mk (Assign($3,(mk (LhsArgs(mk l))), $4))
   }

zexpression:
  | zsimple_expr                   { mk (SimpleExpr $1) }
  | zassign_expr                   { mk (AssignExpr $1) }
  | zassign_ext_expr               { mk (AssignExpr $1) }
  | ATMARK param_list SPACE zexpression
    { mk (AnonFcnHandle($2,$4)) }
  | ATMARK param_list COMMA zexpression
    { mk (AnonFcnHandle($2,$4)) }
  | ATMARK param_list SEMICOLUMN zexpression
    { mk (AnonFcnHandle($2,$4)) }
  | ATMARK param_list zexpression
    { mk (AnonFcnHandle($2,$3)) }
;

/*  Word-list command, ugly matlab stuff (not finished)  */

/*
word_list_cmd:
  | identifier SPACE word_list
    { (mk (WordListCmd($1,List.rev $3))); }
;
word_list:
  | identifier           { [$1] }
  | word_list SPACE identifier { $3 :: $1 }
;

*/

/* commands  */

command:
  | declaration                 { mk (Declaration($1)) }
  | if_command                  { $1; }
  | switch_command              { $1; }
  | loop_command                { mk (LoopCommand($1)) }
  | jump_command                { mk (JumpCommand($1)) }
  | except_command              { mk (ExceptCommand($1)) }
  | mfunction                   { mk (MFunction($1)) }
;

/* Declarations */

declaration:
  | GLOBAL decl1         { mk (Global(List.rev $2)) }
  | STATIC decl1         { mk (Static(List.rev $2)) }
;

decl1:
  | identifier          { [(mk (Declare($1)))] }
  | identifier EQUAL simple_expr
    { [(mk (DeclareSet($1,(mk (SimpleExpr $3)))))] }
  | decl1 SPACE identifier
    { (mk (Declare($3))) :: $1 }
  | decl1 SPACE identifier EQUAL simple_expr
    { (mk (DeclareSet($3,(mk (SimpleExpr($5)))))) :: $1 }
;

/*  switch  */

switch_command:
  | SWITCH expression END
    { mk (Switch($2,(mk (Cases([]))))); }
  | SWITCH expression case_list END
    { mk (Switch($2,$3)); }
  | SWITCH expression switch_comments case_list END
    { mk (Switch($2,$4)); }

/*  XXXXX should be kept after parsing  */

switch_comments:
  | COMMA                         { "," }
  | SEMICOLUMN                    { ";" }
  | NEWLINE                       { "\n" }
  | expr1                         { " " }
  | switch_comments COMMA         { "," }
  | switch_comments SEMICOLUMN    { ";" }
  | switch_comments NEWLINE       { "\n" }
  | switch_comments expr1         { "\n" }
;

case_list:
  | default_case                  { mk (DefaultCase $1); }
  | case_list1                    { mk (Cases(List.rev $1)); }
  | case_list1 default_case       { mk (CasesOther(List.rev($1),$2)); }
;

case_list1:
  | switch_case                   { [$1] }
  | case_list1 switch_case        { $2 :: $1 }
;

switch_case:
  | CASE exprs1
    {
     (* $2 is Mtlb_ast.parsing Mtlb_ast.exprs list *)
     let (co,expr,rest) = exprs_list_split true [] [] [] (List.rev $2) in
     mk (Case(mk co,mk expr,mk rest));
    }
  | CASE expression SPACE  exprs1
    {
     mk (Case(mk [],mk [mk [mk (Exprsep(mk (Expression($2)),mk (Newline 1)))]],mk $4))
    }
;

default_case:
  | OTHERWISE exprs1 { mk (Otherwize(mk (List.rev $2))); }
;

/* If command */

if_command:
  | IF if_cmd_list END
    { mk (IfCommand($2)) }
;

if_cmd_list:
  | if_cmd_list1              { mk (IfCmd(List.rev $1)) }
  | if_cmd_list1 else_clause  { mk (IfCmdWithElse(List.rev $1,$2)) }
;

if_cmd_list1:
  | exprs1
    {
     let (co,expr,rest) = exprs_list_split true [] [] []
         (List.rev $1) in
     [(mk (If(mk co,mk expr,mk rest)))]
    }
  | expression SPACE exprs1
    {
     (* ici expression ne peut tre un commentaire *)
     [(mk (If(mk [], mk [mk [mk (Exprsep(mk (Expression($1)),mk (Newline 1)))]],mk (List.rev $3))))]
    }
  | if_cmd_list1 elseif_clause
    { $2 :: $1 }
;

elseif_clause:
  | ELSEIF expression SPACE exprs1
      {
       (* ici expression ne peut tre un commentaire *)
       (mk (ElseIf(mk [], mk [mk [mk (Exprsep(mk (Expression($2)),mk (Newline 1)))]],mk (List.rev $4))))
       (*
     let (co,expr,rest) = exprs_list_split true [] [] []
         (List.rev $4) in
     mk (ElseIf(mk co,mk expr,mk rest)); *)
     }
  | ELSEIF exprs1  /* XXXXX */
    {
     let (co,expr,rest) = exprs_list_split true [] [] []
         (List.rev $2) in
     mk (ElseIf(mk co,mk expr,mk rest))
    }

else_clause:
  |  ELSE exprs1 { mk (Else(mk (List.rev $2))); }
;

/* for loops */

loop_command:
  | WHILE exprs1 END
    {
     let (co,expr,rest) = exprs_list_split true [] [] []
         (List.rev $2) in
     mk (WhileCommand(mk co,mk expr,mk rest));
    }
  | WHILE expression SPACE exprs1 END
    {
     (* ici expression ne peut tre un commentaire *)
     mk (WhileCommand(mk [], mk [mk [mk (Exprsep(mk (Expression($2)),mk (Newline 1)))]],mk $4));
    }
  | DO exprs1 UNTIL expression
    { mk (DoUntilCommand(mk (List.rev $2), $4));  }
  | FOR exprs1 END
    {
     let (co,assign,rest) = exprseps_list_assign_split true [] [] []
         (List.rev $2) in
     (mk (ForCommand(mk co,mk assign,mk rest)));  (* XXXX *)
    }
  | FOR expression SPACE exprs1 END
    {
     (* ici expression ne peut tre un commentaire *)
     (mk (ForCommand(mk [], mk [mk [mk (Exprsep(mk (Expression($2)),mk (Newline 1) ))]],mk $4)));
    }

/* jumps */

jump_command:
  | BREAK            { mk Break }
  | CONTINUE         { mk Continue }
  | FUNC_RET         { mk (FReturn (mk EmptyReturn)) }
;

/* Exceptions */

except_command:
  | UNWIND exprs1 CLEANUP exprs1 END
    {  mk (UnwindCommand (mk (List.rev $2),mk (List.rev $4))); }
  | TRY exprs1 CATCH exprs1 END
    {  mk (TryCatchCommand (mk (List.rev $2), mk (List.rev $4))); }
  | TRY exprs1 END
    {  mk (TryCommand(mk (List.rev $2))); }
;

/* function parameters */

param_list:
  | LPAREN param_list1 RPAREN {$2 }

param_list1:
  | { mk EmptyParamList; }
  | param_list2 { mk (ParamList(List.rev $1)); }
;

param:
  | identifier { $1 }
  | NOT { mk "~" }
;

param_list2:
  | param opt_oper_sep
    { [(mk (Declare($1)))] }
  | param EQUAL simple_expr opt_oper_sep
    { [(mk (DeclareSet($1,(mk (SimpleExpr($3))))))] }
  | param_list2 COMMA opt_oper_sep param
    { (mk (Declare($4))) :: $1 }
  | param_list2 COMMA opt_oper_sep param EQUAL simple_expr
    { (mk (DeclareSet($4,(mk (SimpleExpr($6)))))) ::$1 }
;

/* returned values by a function (return_list in function return_list=f(...) )
 */

return_list:
  | identifier EQUAL { mk (Return([$1]));  }
  | LBRACKET return_list1 RBRACKET_EQUAL
    { mk (Return(List.rev $2)); }
;

return_list1:
  | { [] } // empty
  | identifier { [$1] }
  | return_list1 COMMA identifier { $3 :: $1 }
;

/* function ... endfunction  */

mfunction:
  | FUNCTION fcn_name function2             { mk (FunctionNoRet($2,$3)); }
  | FUNCTION return_list fcn_name function2 { mk (Function($2,$3,$4));  }
;

fcn_name:
  | identifier { $1 }
  | GET DOT identifier { $3 }
  | SET DOT identifier { $3 }
;

function2:
  | LPAREN param_list1 RPAREN exprs1 endfun
    { mk (Body($2,mk (List.rev $4),$5)); }
  | fsep exprs1 endfun
    { mk (SimpleBody(mk (List.rev $2),$3)); }
;

fsep:
  | COMMA       { "," }
  | SEMICOLUMN  { ";" }
  | NEWLINE     { "\n" }
  | PCOMMENT    { "\n" } //(* XXX le garder *)
  | DCOMMENT    { "\n" } //(* XXX le garder *)
;

/* function can be ended by EOF */

endfun:
  | END { mk End_fun }
  | EOF { mk Eof_fun }

/* misc */

opt_oper_sep:
  | { "void" }
  | NEWLINE { "\n" }
;

/*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*/
