(***********************************************************************)
(*                                                                     *)
(*                           Matlab Parser                             *)
(*                                                                     *)
(*          J.Ph Chancelier, Enpc/Cermics                              *)
(*                                                                     *)
(*  Copyright 2012-2019,                                               *)
(*  Ecole Nationale des ponts et chaussées                             *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mtlb_make_host_print.ml,v 1.1 2019-05-10 14:09:18 jpc Exp $ *)

open Mtlb_ast;;
open Format;;

(** The functor to print code, given an host specific translation module. *)

module Make_print
  (Host : Mtlb_make_host_print_spec.HOST) = struct

  (* utility function: checks that expression is an implicit vector
   * could be moved in ast_funs
   *)

  let rec check_implicit_vector arg =
    match arg.Ast_node.ast_desc with
    | [] -> None
    | elt :: [] ->
        let desc = elt.Ast_node.ast_desc in
        check_implicit_vector_exprsep_list_1 desc
    | _elt :: _elts -> None

  and check_implicit_vector_exprsep_list_1 = function
    | [] -> None
    | elt :: [] -> check_implicit_vector_exprsep_1 elt
    | _elt :: _elts -> None   (* on peut autoriser des commentaires *)

  and check_implicit_vector_exprsep_1 elt =
    match elt.Ast_node.ast_desc  with
    | Exprsep(x,_sep) -> check_implicit_vector_expr x

  and check_implicit_vector_expr expr =
    match expr.Ast_node.ast_desc with
    | Expression exp -> check_implicit_vector_expression_assign exp
    | Command _
    | WordListCmd _
    | Comments _
    | Pcomments _
    | Dcomments _
    | Empty -> None

  and check_implicit_vector_expression_assign expression =
    match expression.Ast_node.ast_desc with
    | SimpleExpr _
    | AnonFcnHandle _ -> None
    | AssignExpr assign_expr ->
        match assign_expr.Ast_node.ast_desc with
        | Assign (string, assign_lhs, expression) ->
            match check_implicit_vector_expression_simple expression with
            | None -> None
            | Some x -> Some ( assign_lhs, string, x)

  and check_implicit_vector_expression_simple expression =
    match expression.Ast_node.ast_desc with
    | SimpleExpr simple_expr -> check_implicit_vector_simple_expr simple_expr
    | AnonFcnHandle _
    | AssignExpr _ -> None

  and check_implicit_vector_simple_expr simple_expr =
    match simple_expr.Ast_node.ast_desc with
    | ColonExpr oper_expr_list -> Some oper_expr_list
    | SexpBinaryOp _
    | Op_compare _
    | BooleanOp _  -> None
  ;;

  (* utility function: returns the last separator used in a exprsep list *)

  let rec get_last_sep_of_exprsep_list = function
    | [] -> None
    | elt :: [] ->
        Some
          (match elt.Ast_node.ast_desc  with
          | Exprsep(_x,sep) -> sep.Ast_node.ast_desc)
    | _elt :: elts -> get_last_sep_of_exprsep_list  elts
  ;;
  (* printing: print_exprs_list is enclosed in a vertical box *)

  let rec print_exprs_list ppf arg =
    print_exprs_list_last_sep true ppf arg

  and print_exprs_list_no_last_sep ppf arg =
    print_exprs_list_last_sep false ppf arg;

  and print_exprs_list_last_sep use_last_sep ppf =
    print_exprs_list_last_sep_1 None use_last_sep ppf

  and print_sep_newlines ppf n =
    if n >= 0 then
      (fprintf ppf "@,";
       print_sep_newlines ppf (n - 1);)

  and print_exprs_list_last_sep_1 previous_sep use_last_sep ppf = function
    | [] -> ()
    | elt :: elts ->
        let print_el ppd el =
          let _last = print_exprsep_list_last_sep use_last_sep ppd el in
          () in
        let desc = elt.Ast_node.ast_desc in
        (match previous_sep with
        | None -> fprintf ppf "@[<hov>%a@]" print_el desc
        | Some (Comma n)
        | Some (Newline n)
        | Some (Semi n)
        | Some (Sep_None n) ->
            (* fprintf ppf "%s%d%s" "<" n ">"; *)
            print_sep_newlines ppf (n -1);
            fprintf ppf "@[<hov>%a@]" print_el desc);
        print_exprs_list_last_sep_1
          (get_last_sep_of_exprsep_list desc) use_last_sep ppf elts

  (* print_exprsep_list and print_exprsep_list_last_sep:
   * used to print a list of expression on the same line,
   * print_exprsep_list_last_sep is used when printing if conditions
   * and adapt the printing to the existence or not or a then_delimiter.
   * When the then_delimiter is not "", the last separator
   * when printing the list must be omited and if the list ends
   * with a comment the comment must be omited to
   *)

  and print_exprsep_list ppf arg =
    print_exprsep_list_last_sep true ppf arg

  and print_exprsep_list_last_sep use_last_sep ppf arg =
    let rec print_exprsep_list_last_sep_1
            use_last_sep ppf previous_sep = function
      | [] -> ()
      | elt :: [] ->
          let _last1 = print_exprsep ppf use_last_sep previous_sep elt
          in ()
      | elt :: elts ->
          let last1 = print_exprsep ppf use_last_sep previous_sep elt in
          print_exprsep_list_last_sep_1 use_last_sep ppf (Some last1) elts
    in
    print_exprsep_list_last_sep_1 use_last_sep ppf None arg

  (* print_exprsep:
   * used to print an Exprsep(x,sep) which is an x followed by a separtor
   *)

  and _print_exprsep_debug ppf use_last_sep previous_sep elt =
    fprintf ppf "%s" "[|exprsep:";
    let sep = print_exprsep ppf use_last_sep previous_sep elt in
    fprintf ppf "%s" "|]";
    sep

  and print_exprsep ppf use_last_sep previous_sep elt =
    match elt.Ast_node.ast_desc  with
    | Exprsep(x,sep) ->
        match x.Ast_node.ast_desc with
        | Empty ->
            (
              (* fprintf ppf "%s" "<empty>"; *)
              match previous_sep with
              | None -> Sep_None (-1)
              | Some sep -> sep
             )
        | Expression _
        | Command _
        | WordListCmd _
        | Comments _
        | Pcomments _
        | Dcomments _  ->
            if previous_sep <> None then fprintf ppf "@,";
            let sep_desc = sep.Ast_node.ast_desc in
            if use_last_sep then
              fprintf ppf "%a%a" print_expr x print_sep sep_desc
            else
              fprintf ppf "%a" print_expr_no_comments x;
            (* returns the separator *)
            sep_desc

  and print_sep ppf sep =
    if true then
      match sep with
      | Comma _ -> fprintf ppf "%s" ",";
      | Newline _ -> fprintf ppf "%s" "";
      | Semi _ -> fprintf ppf "%s" ";";
      | Sep_None _ -> fprintf ppf "%s" "" (* should not go there *);
    else
      (* debug version *)
      match sep with
      | Comma n -> fprintf ppf "%s<%d>" "<sep-c>," n;
      | Newline n -> fprintf ppf "%s<%d>" "<sep-n>" n;
      | Semi n -> fprintf ppf "%s<%d>" "<sep-s>;" n;
      | Sep_None n -> fprintf ppf "%s<%d>" "<sep-e>" n;

  (* print_exprs_list_matrix:
   * used to print matrix contents taking care of the fact that box printing
   * will break lines and that we have to preserve the matrix parsing rules.
   * the function is thus quite complex.
   *)

  and print_exprs_list_matrix ppf =
    if true then
      print_exprs_list_matrix_check_first ppf true (* new definition *)
    else
      print_exprs_list ppf (* old definition *)

  and print_exprs_list_matrix_check_first ppf is_first = function
    | [] -> ()
    | elt :: [] ->
      print_exprsep_list_matrix_elt ppf is_first true elt
    | elt :: elts ->
      print_exprsep_list_matrix_elt ppf is_first false elt;
      print_exprs_list_matrix_check_first ppf false elts

  and print_exprsep_list_matrix_elt ppf is_first is_last elt =
     let desc = elt.Ast_node.ast_desc in
     let printer ppf =
       print_exprsep_list_matrix_last_sep ppf (false,None) is_last in
     if is_first then
       fprintf ppf "@[<hov>%a@]" printer desc
     else
       fprintf ppf "@{<n>@,@}@[<hov>%a@]" printer desc

  and print_exprsep_list_matrix_last_sep
      ppf previous_sep is_last_row = function
    | [] -> ()
    | elt :: [] ->
      let _last1 =
        print_exprsep_matrix ppf previous_sep true is_last_row false elt in
      ()
    | elt :: elts ->
        let next_is_comment = match elts with
        | [] -> false
        | eltn :: [] -> elt_is_comment eltn
        | _elts -> false
        in
        let last1 =
          print_exprsep_matrix
            ppf previous_sep false is_last_row next_is_comment elt in
        print_exprsep_list_matrix_last_sep ppf last1 is_last_row elts

  and elt_is_comment elt =
    match elt.Ast_node.ast_desc  with
    | Exprsep(x, _sep) -> expr_is_comments x

  and expr_is_comments expr =
    match expr.Ast_node.ast_desc with
    | Expression _
    | Command _
    | WordListCmd _ -> false
    | Comments _
    | Pcomments _
    | Dcomments _
    | Empty -> true

  and print_exprsep_matrix
        ppf previous_sep is_last_row_elt is_last_row next_is_comment elt =
    match elt.Ast_node.ast_desc  with
    | Exprsep(x,sep) ->
      let is_comment = expr_is_comments x in
      let print_separator =
        if is_comment || is_last_row && is_last_row_elt then
          print_no_sep
        else
          if is_last_row_elt then
            print_last_row_elt_sep
          else
            print_matrix_sep
      in
      let sep_desc = sep.Ast_node.ast_desc in
      let sep_desc = if next_is_comment then (Semi 1) else sep_desc in
      let _previous_is_comment, previous_sep = previous_sep in
      if previous_sep <> None then
        fprintf ppf "@,%a%a" print_expr x print_separator sep_desc
      else
        fprintf ppf "%a%a" print_expr x print_separator sep_desc;
      (is_comment, Some sep_desc)

  and print_no_sep _ppf = function
    | Comma _
    | Newline _
    | Semi _
    | Sep_None _ -> ()

  and print_last_row_elt_sep ppf sep =
    if true then
      fprintf ppf "%s" ";"
    else
      (* debug version *)
      match sep with
      | Comma _ -> fprintf ppf "%s" "<l-c>;";
      | Newline _ -> fprintf ppf "%s" "<l-n>;";
      | Semi _ -> fprintf ppf "<l-s>;";
      | Sep_None _ -> fprintf ppf "%s" "<l-e>;";

  and print_matrix_sep ppf sep =
    if true then
      match sep with
      | Comma _ -> fprintf ppf "%s" ",";
      | Newline _ -> fprintf ppf "%s" ";";
      | Semi _ -> fprintf ppf ";";
      | Sep_None _ -> fprintf ppf "%s" ";";
    else
      (* debug version *)
      match sep with
      | Comma _ -> fprintf ppf "%s" "<e-c>,";
      | Newline _ -> fprintf ppf "%s" "<e-n>;";
      | Semi _ -> fprintf ppf "<e-s>;";
      | Sep_None _ -> fprintf ppf "%s" "<e-e>;";

  and print_opt_list_with_delim ppf =
    if Host.do_delim = ""
    then print_opt_list ppf
    else print_opt_list_no_last_sep ppf

  (* expressions *)

  and print_expr_no_comments ppf expr =
    match expr.Ast_node.ast_desc with
    | Expression exp -> fprintf ppf "%a" print_expression exp;
    | Command cmd -> fprintf ppf "%a" print_command cmd;
    | WordListCmd (identifier, word_list) ->
      fprintf ppf "%a(%a)" print_identifier identifier
        print_word_list word_list
    | Comments _string
    | Pcomments _string
    | Dcomments _string -> ()
    | Empty -> () (* fprintf ppf "%s" "<empty>" *)

  and print_expr ppf expr =
    match expr.Ast_node.ast_desc with
    | Expression exp -> fprintf ppf "%a" print_expression exp;
    | Command cmd -> fprintf ppf "%a" print_command cmd;
    | WordListCmd (identifier, word_list) ->
      fprintf ppf "%a(%a)" print_identifier identifier
        print_word_list word_list
    | Comments string -> fprintf ppf "%s" string
    | Pcomments string -> fprintf ppf "%s%s"
      (Host.comment_char_dict_fun '%') string
    | Dcomments string -> fprintf ppf "%s%s"
      (Host.comment_char_dict_fun '#') string
    | Empty -> () (* fprintf ppf "%s" "<empty>" *)

  and print_word_list ppf word_list =
    match word_list.Ast_node.ast_desc with
    | [] -> ()
    | arg :: args ->
      print_word ppf arg;
      List.iter (function arg -> fprintf ppf ", %a" print_word arg) args

  and print_word ppf word =
    match word with
    | SQ x -> fprintf ppf "'%s'" x (* print_xstring ppf x *)
    | DQ x -> fprintf ppf "\"%s\"" x (* print_xstring ppf x *)

  (* strings *)

  and print_dq_string ppf s =
    fprintf ppf "\"";
    String.iter (print_dq_char ppf) s;
    fprintf ppf "\"";

  and print_sq_string ppf s =
    fprintf ppf "%c" Host.sq_string_delimiter;
    String.iter (print_sq_char ppf) s;
    fprintf ppf "%c" Host.sq_string_delimiter;

  and print_dq_char ppf = function
    | '\n' -> fprintf ppf "%s" "\\n"
    | '\r' -> ()
    | x -> fprintf ppf "%c" x

  and print_sq_char ppf = function
    | '\r' -> ()
    | c -> fprintf ppf "%s" (Host.sq_char_dict_fun c)

  and print_expression ppf expression =
    match expression.Ast_node.ast_desc with
    | SimpleExpr simple_expr ->
        fprintf ppf "@[<h>%a@]" print_simple_expr simple_expr
    | AssignExpr assign_expr -> print_assign_expr ppf assign_expr;
    | AnonFcnHandle (param_list, statement) ->
      fprintf ppf "@@%a %a"
        print_param_list param_list print_expression statement;

  and print_command ppf command =
    match command.Ast_node.ast_desc with
    | Declaration declaration ->
        fprintf ppf "%a" print_declaration declaration
    | Switch (expression, case_list) ->
      fprintf ppf "@[<v>@[<v 2>%s %s%a%s %s@,%a@]@,%s@]"
        Host.switch_keyword
        Host.switch_condition_open_marker
        print_expression expression
        Host.switch_condition_close_marker
        Host.begin_switch_marker
        print_case_list case_list
        Host.end_switch_marker
    | IfCommand if_command -> print_if_command ppf if_command
    | LoopCommand loop_command -> print_loop_command ppf loop_command;
    | JumpCommand jump_command -> print_jump_command ppf jump_command;
    | ExceptCommand except_command ->
        fprintf ppf "%a" print_except_command except_command;
    | SciptFile _ -> ()
    | MFunction mfunction -> print_mfunction ppf mfunction;

  and print_jump_command ppf jump_command=
    match jump_command.Ast_node.ast_desc with
    | Break -> fprintf ppf "%s" "break"
    | FReturn return_list -> if Host.return_list_delim then
        fprintf ppf "%s" "return"
      else
        fprintf ppf "return %a"
        print_return_list return_list
    | Continue -> fprintf ppf "%s" "continue"

  and print_primary_expr ppf = function
    | PexpIdentifier identifier -> print_identifier ppf identifier;
    | PexpConstant constant -> print_constant ppf constant;
    | PexpFcnHandle fcn_handle ->
      fprintf ppf "@@%a" print_identifier fcn_handle;
    | PexpMatrix matrix ->
      if check_matrix_isrow matrix then
        print_rowmat ppf matrix
      else
        if check_exprs_list matrix then
          (* add an extra line-break at the end before closing the matrix
             if last matrix row is ended by a comment *)
          fprintf ppf "@[<h>[@[<v>@{<m>%a@}@,]@]@]"
            print_exprs_list_matrix matrix
        else
          fprintf ppf "@[<h>[@[<v>@{<m>%a@}@]]@]"
            print_exprs_list_matrix matrix
    | PexpCell cells ->
      if check_exprs_list cells then
        fprintf ppf "@[<h>%s@[<v>@{<m>%a@}@,%s@]@]"
          "{" print_exprs_list_matrix cells "}"
      else
        fprintf ppf "@[<h>%s@[<v>@{<m>%a@}@]%s@]"
          "{" print_exprs_list_matrix cells "}"
    | PexpMetaIdentifier meta_identifier ->
      fprintf ppf "%s" meta_identifier.Ast_node.ast_desc
    | PexpSuperIdentifier superclass_identifier ->
      fprintf ppf "%s"  superclass_identifier.Ast_node.ast_desc
    | PexpExpression ( expression ) ->
      (* we can print expression inside parentheses
       * with continuation on matlab or just newline on octave
       *)
      fprintf ppf "(%a)"  print_expression expression;

  and print_identifier ppf x =
    fprintf ppf "%s" (Host.dict_fun x.Ast_node.ast_desc)

  and print_arg_list ppf arg_list =
    match arg_list.Ast_node.ast_desc with
    | [] -> ()
    | arg :: args ->
      print_arg ppf arg;
      List.iter (function arg -> fprintf ppf ", %a" print_arg arg) args

  and print_arg ppf arg =
    match arg.Ast_node.ast_desc with
    | ArgExpression expression -> print_expression ppf expression;
    | ArgColon -> fprintf ppf  ":"
    | ArgTilde -> fprintf ppf  "~"

  and print_oper_expr ppf oper_expr =
    match oper_expr.Ast_node.ast_desc with
    | PrimaryExpr primary_expr ->
      print_primary_expr ppf primary_expr.Ast_node.ast_desc;
    | PostfixOp (string, oper_expr) ->
      fprintf ppf "%a%s" print_oper_expr
        oper_expr (Host.postfix_op_dict_fun string)
    | IndexExpr (oper_expr, arg_list, string1, string2) ->
      (* using a matlab box *)
      fprintf ppf "%a@[@{<m>%s%a%s@}@]"
        print_oper_expr oper_expr
          (Host.extract_delim_start string1) print_arg_list arg_list
          (Host.extract_delim_end string2)
    | IndirectRef (oper_expr, expression) ->
      fprintf ppf ("%a.(%a)")
        print_oper_expr oper_expr print_expression expression;
    | IndirectRefU ( oper_expr,identifier ) ->
      fprintf ppf ("%a.%a")
        print_oper_expr oper_expr print_identifier identifier;
    | PrefixOpUnary (string,  oper_expr) ->
      (match string with
       | "!" -> fprintf ppf "%s%a" Host.not_op
         print_oper_expr oper_expr
       | _  -> fprintf ppf "%s%a" string print_oper_expr oper_expr)
    | BinaryOp (string, oper_expr, oper_expr1) ->
      fprintf ppf ("%a %s@, %a") print_oper_expr oper_expr string
        print_oper_expr oper_expr1;
    | Op_stars (stars, oper_expr, oper_expr1) ->
      fprintf ppf ("%a %a@, %a") print_oper_expr oper_expr
        print_stars stars
        print_oper_expr oper_expr1;
    | Op_divs (divs, oper_expr, oper_expr1) ->
      fprintf ppf ("%a %a@, %a") print_oper_expr oper_expr
        print_divs divs
        print_oper_expr oper_expr1;

  and print_stars ppf stars =
    match stars.Ast_node.ast_desc with
    | (Star, _) -> fprintf ppf "*";
    | (Dot_star, _) -> fprintf ppf " .*"

  and print_divs  ppf divs =
    match divs.Ast_node.ast_desc with
    | Slash, _ -> fprintf ppf "/"
    | Bslash, _ -> fprintf ppf "\\"
    | Dot_slash, _ -> fprintf ppf " ./"
    | Dot_bslash, _ -> fprintf ppf " .\\"

  and print_simple_expr ppf simple_expr =
    match simple_expr.Ast_node.ast_desc with
    | ColonExpr oper_expr_list ->
      print_colon_list_protect ppf oper_expr_list
    | SexpBinaryOp (op, simple_expr, simple_expr2) ->
      fprintf ppf "%a %s@, %a"
        print_simple_expr simple_expr
        op
        print_simple_expr simple_expr2
    | Op_compare (compare, simple_expr, simple_expr2)  ->
      fprintf ppf "%a %a@, %a"
        print_simple_expr simple_expr
        print_compare compare
        print_simple_expr simple_expr2
    | BooleanOp (op, simple_expr, simple_expr2) ->
      fprintf ppf "%a %s@, %a"
        print_simple_expr simple_expr
        op
        print_simple_expr simple_expr2

  and print_compare ppf compare =
    match compare.Ast_node.ast_desc with
    | Lt, _ -> fprintf ppf "<"
    | Le, _ -> fprintf ppf "<="
    | Eq, _ -> fprintf ppf "=="
    | Ne, _ -> fprintf ppf "%s" Host.ne_op
    | Ge, _ -> fprintf ppf ">="
    | Gt, _ -> fprintf ppf ">"

  and print_colon_list_protect ppf oper_expr_list =
    if List.length oper_expr_list > 1 && Host.protect_colon_expr then
      fprintf ppf "%s%a%s" "[" print_colon_list oper_expr_list "]"
    else
      fprintf ppf "%a" print_colon_list_with_parenth oper_expr_list

  and print_colon_list ppf = function
    | [] -> ()
    | arg :: args ->
      print_oper_expr ppf arg;
      List.iter (function arg -> fprintf ppf ": %a" print_oper_expr arg) args

  (* x ~= 1:5 should be translated as ->  x ~= (1:5) for nsp/scicoslab
   * since priorities are not the same
   *)

  and print_colon_list_with_parenth ppf expr =
    match expr with
    | [] -> ()
    | arg :: [] ->
        fprintf ppf "%a" print_oper_expr arg;
    | arg1 :: arg2 :: [] ->
          fprintf ppf "%s%a : %a%s"
          "("
          print_oper_expr arg1
          print_oper_expr arg2
          ")"
    | arg1 :: arg2 :: arg3 :: []  ->
        fprintf ppf "%s%a : %a : %a%s"
          "("
          print_oper_expr arg1
          print_oper_expr arg2
          print_oper_expr arg3
          ")"
    | _ ->  print_colon_list ppf expr

  and print_colon_list_for_special_do_loop ppf expr =
    if Host.protect_colon_expr then
      match expr with
      | [] -> ()
      | arg :: [] ->
          fprintf ppf "%a" print_oper_expr arg;
      | arg1 :: arg2 :: [] ->
          fprintf ppf "%a, %a"
            print_oper_expr arg1
            print_oper_expr arg2
      | arg1 :: arg2 :: arg3 :: []  ->
          fprintf ppf "%a, %a, %a"
            print_oper_expr arg1
            print_oper_expr arg3
            print_oper_expr arg2
      | _ ->  print_colon_list ppf expr
    else
      print_colon_list ppf expr

  and print_constant ppf constant =
    match constant.Ast_node.ast_desc with
    | Constant c ->
      match c with
      | Num s      -> fprintf ppf "%s" s
      | Imag_num s -> fprintf ppf "%s" (Host.imag s)
      | Dq_string s -> print_dq_string ppf s
      | Sq_string s -> print_sq_string ppf s

  (* arguments of a function declaration or of an anonymous fonction handle *)

  and print_decl2_list fmt sep ppf = function
    | [] -> ()
    | arg :: args ->
      print_decl2 ppf arg;
      List.iter (function arg ->
        fprintf ppf fmt sep print_decl2 arg) args

  and print_param_list ppf param_list =
    let print_cut_decl2_list = print_decl2_list "%s@,%a" "," in
    match param_list.Ast_node.ast_desc with
    | EmptyParamList -> fprintf ppf "()"
    | ParamList decl2_list ->
      fprintf ppf "@[@{<m>(%a)@}@]" print_cut_decl2_list decl2_list

  (* assignments *)

  and print_assign_lhs ppf assign_lhs =
    match assign_lhs.Ast_node.ast_desc with
    | LhsSimple simple_expr ->
      fprintf ppf "%a"  print_simple_expr simple_expr
    | LhsArgs exprsep_list ->
      let protect = Host.protect_lhs in
      if protect then
        fprintf ppf "[%a]" print_exprsep_list exprsep_list.Ast_node.ast_desc
      else
        fprintf ppf "%a" print_exprsep_list exprsep_list.Ast_node.ast_desc

  and print_assign_expr ppf assign_expr =
    match assign_expr.Ast_node.ast_desc with
    | Assign (string, assign_lhs, expression) ->
      fprintf ppf "@[<h>%a %s %a@]" print_assign_lhs assign_lhs
        string print_expression expression

  and print_declaration ppf declaration =
    let print_comma_declarations =
      print_decl2_list "%s%a" Host.persistent_declaration_sep in
    match declaration.Ast_node.ast_desc with
    | Global decl2_list ->
      fprintf ppf "@[<h>global %a@]" print_comma_declarations decl2_list
    | Static decl2_list ->
      fprintf ppf "@[<h>%s %a@]"
        Host.persistent_keyword
        print_comma_declarations decl2_list

  and print_decl2 ppf decl2 =
    match decl2.Ast_node.ast_desc with
    | Declare identifier -> print_identifier ppf identifier;
    | DeclareSet (identifier, expression) ->
      fprintf ppf "%a=%a"
        print_identifier identifier
        print_expression expression;
    | MagicTilde -> fprintf ppf "~";

  (* switch statement *)

  and print_case_list ppf case_list =
    match case_list.Ast_node.ast_desc with
    | Nocase -> ()
    | DefaultCase default_case -> print_default_case ppf default_case
    | Cases switch_case_list -> print_switch_case_list ppf switch_case_list;
    | CasesOther (switch_case_list, default_case) ->
      print_switch_case_list ppf switch_case_list;
      print_default_case ppf default_case

  and print_switch_case_list ppf = function
    | [] -> ()
    | arg :: args ->
      print_switch_case ppf arg;
      List.iter (function arg -> fprintf ppf "@,%a" print_switch_case arg) args

  and print_break_case_keyword ppf =
    let kwd = Host.break_case_keyword in
    if kwd <> "" then fprintf ppf "@ %s" kwd

  and print_switch_case ppf switch_case =
    match switch_case.Ast_node.ast_desc with
    | Case (comments, expr, opt_list) ->
      let comments_desc = comments.Ast_node.ast_desc in
      if comments_desc = [] then
        fprintf ppf "@[<v>@[<v 2>case %s%a%s%s@,%a%t@]@]"
          Host.switch_condition_open_marker
          print_opt_list expr
          Host.switch_condition_close_marker Host.case_condition_sep_marker
          print_opt_list opt_list
          print_break_case_keyword
      else
        fprintf ppf "@[<v>@[<v 2>case %a@,%s%a%s%s@,%a%t@]@]"
          print_opt_list comments
          Host.switch_condition_open_marker
          print_opt_list expr
          Host.switch_condition_close_marker Host.case_condition_sep_marker
          print_opt_list opt_list
          print_break_case_keyword

  and print_default_case ppf  default_case =
    match default_case.Ast_node.ast_desc with
    | Otherwize opt_list ->
      fprintf ppf "@,@[<v 2>%s @,%a%t@]"
        Host.otherwise_keyword
        print_opt_list opt_list
        print_break_case_keyword

  (* sequence of statements in a for select while etc... *)

  and print_opt_list ppf = function
    | x -> print_exprs_list ppf x.Ast_node.ast_desc
  and print_opt_list_no_last_sep ppf = function
    | x -> print_exprs_list_no_last_sep ppf x.Ast_node.ast_desc

  and print_condition ppf = function
    | x ->
        if Host.then_delim = "" then
          print_exprs_list ppf x.Ast_node.ast_desc
        else
          fprintf ppf "%a%s"
            print_exprs_list_no_last_sep x.Ast_node.ast_desc
            Host.then_delim

  (* if with octave endif *)

  and print_if_command  ppf if_command =
    match if_command.Ast_node.ast_desc with
    | IfCmd if_cmd_list ->
      fprintf ppf "@[<v>%a@,%s@]"
        print_if_cmd_list if_cmd_list
        (Host.close_block_dict_fun "endif")
    | IfCmdWithElse (if_cmd_list, else_clause) ->
      fprintf ppf "@[<v>%a%a@,%s@]"
        print_if_cmd_list if_cmd_list
        print_else_clause else_clause
        (Host.close_block_dict_fun "endif")

  and print_if_cmd_list ppf = function
    | [] -> ()
    | arg :: args ->
      print_if_cmd ppf arg;
      List.iter (function arg -> fprintf ppf "%a" print_if_cmd arg) args

  and print_if_cmd ppf if_cmd =
    match if_cmd.Ast_node.ast_desc with
    | If (comments, expr, opt_list) ->
      let comments_desc = comments.Ast_node.ast_desc in
      if comments_desc = [] then
        fprintf ppf "@[<v 2>@[<h>if %a@]@,%a@]"
          print_condition expr print_opt_list opt_list
      else
        fprintf ppf "@[<v 2>if %a@,%a@,%a@]"
          print_opt_list comments
          print_condition expr
          print_opt_list opt_list
    | ElseIf (comments, expr, opt_list) ->
      let comments_desc = comments.Ast_node.ast_desc in
      if comments_desc = [] then
        fprintf ppf "@,@[<v 2>@[<h>elseif %a@]@,%a@]"
          print_condition expr print_opt_list opt_list
      else
        fprintf ppf "@,@[<v 2>elseif %a@,%a@,%a@]"
          print_opt_list comments
          print_condition expr
          print_opt_list opt_list

  and print_else_clause ppf else_clause =
    match else_clause.Ast_node.ast_desc with
    | Else opt_list ->
      fprintf ppf "@,@[<v 2>else @,%a@]" print_opt_list opt_list

  (* loop *)

  and print_loop_command  ppf loop_command =
    match loop_command.Ast_node.ast_desc with
    | WhileCommand (comments, expr, opt_list) ->
      let comments_desc = comments.Ast_node.ast_desc in
      if comments_desc = [] then
        fprintf ppf "@[<v>@[<v 2>while %a%s@,%a@]@,%s@]"
          print_opt_list_with_delim expr
          Host.do_delim
          print_opt_list opt_list
          (Host.close_block_dict_fun "endwhile")
      else
        fprintf ppf "@[<v>@[<v 2>while %a@,%a%s@,%a@]@,%s@]"
          print_opt_list comments
          print_opt_list_with_delim expr
          Host.do_delim
          print_opt_list opt_list
          (Host.close_block_dict_fun "endwhile")

    | DoUntilCommand (opt_list, expression) ->
      fprintf ppf "@[<v>@[<v 2>do @,%a@]@,until %a@]"
        print_opt_list opt_list print_expression expression;

    | ForCommand (comments, assign_expr, opt_list) ->
      let comments_desc = comments.Ast_node.ast_desc in
      match check_implicit_vector assign_expr with
      | None ->
          if comments_desc = [] then
            fprintf ppf "@[<v>@[<v 2>for %a%s@,%a@]@,%s@]"
              print_opt_list_with_delim assign_expr
              Host.do_delim
              print_opt_list opt_list
              (Host.close_block_dict_fun "endfor")
          else
            fprintf ppf "@[<v>@[<v 2>for %a@,%a%s@,%a@]@,%s@]"
              print_opt_list comments
              print_opt_list_with_delim assign_expr
              Host.do_delim
              print_opt_list opt_list
              (Host.close_block_dict_fun "endfor")
      | Some (lhs ,string, rhs) ->
          let print_assign_do ppf expr =
            let lhs, string, rhs = expr in
            fprintf ppf "@[<h>%a %s %a@]"
              print_assign_lhs lhs
              string
              print_colon_list_for_special_do_loop rhs
          in
          if comments_desc = [] then
            fprintf ppf "@[<v>@[<v 2>for %a%s@,%a@]@,%s@]"
              print_assign_do (lhs , string, rhs)
              (* print_opt_list_with_delim assign_expr *)
              Host.do_delim
              print_opt_list opt_list
              (Host.close_block_dict_fun "endfor")
          else
            fprintf ppf "@[<v>@[<v 2>for %a@,%a%s@,%a@]@,%s@]"
              print_opt_list comments
              print_assign_do (lhs, string, rhs)
              (* print_opt_list_with_delim assign_expr *)
              Host.do_delim
              print_opt_list opt_list
              (Host.close_block_dict_fun "endfor")


  (* exceptions *)

  and print_except_command ppf except_command =
    match except_command.Ast_node.ast_desc with
    | UnwindCommand (opt_list, opt_list2) ->
      fprintf ppf "@[<v>@[<v 2>%s@,%a@]@,@[<v 2>%s@,%a@]@,%s@]"
        "unwind_protect" print_opt_list opt_list
        "unwind_protect_cleanup" print_opt_list opt_list2
        "end_unwind_protect"
    | TryCatchCommand (opt_list, opt_list2) ->
      fprintf ppf "@[<v>@[<v 2>try@,%a@]@,@[<v 2>catch@,%a@]@,%s@]"
        print_opt_list opt_list print_opt_list opt_list2
        (Host.close_block_dict_fun "end_try_catch")
    | TryCommand (opt_list) ->
      fprintf ppf "@[<v>@[<v 2>try@,%a@]@,%s@]"
        print_opt_list opt_list
        (Host.close_block_dict_fun "end_try_catch")

  (* function *)

  and print_mfunction ppf mfunction =
    match mfunction.Ast_node.ast_desc with
    | FunctionNoRet (identifier,
        {Ast_node.ast_desc = Body (param_list, opt_list, _); _}) ->
      fprintf ppf "@[<v>@[<v 2>@{<m>function %a%a@}@,%a@]@,%s@]"
        print_identifier identifier
        print_param_list param_list
        print_opt_list opt_list
        (Host.close_fun_dict_fun "endfunction")
    | FunctionNoRet (identifier,
        {Ast_node.ast_desc = SimpleBody (opt_list, _); _}) ->
      fprintf ppf "@[<v>@[<v 2>function %a@,%a@]@,%s@]"
        print_identifier identifier
        print_opt_list opt_list
        (Host.close_fun_dict_fun "endfunction")
    | Function (return_list, identifier,
         {Ast_node.ast_desc = Body (param_list, opt_list, _); _}) ->
      if Host.return_list_delim then
        fprintf ppf "@[<v>@[<v 2>@{<m>function %a=%a%a@}@,%a@]@,%s@]"
          print_return_list return_list
          print_identifier identifier
          print_param_list param_list
          print_opt_list opt_list
          (Host.close_fun_dict_fun "endfunction")
        else
        fprintf ppf "@[<v>@[<v 2>@{<m>function %a%a@}@,%a@]@,return %a\n%s@]"
          print_identifier identifier
          print_param_list param_list
          print_opt_list opt_list
          print_return_list return_list
          (Host.close_fun_dict_fun "endfunction")
    | Function (return_list, identifier,
         {Ast_node.ast_desc = SimpleBody (opt_list, _); _}) ->
      if Host.return_list_delim then
        fprintf ppf "@[<v>@[<v 2>@{<m>function %a=%a@}@,%a@]@,%s@]"
          print_return_list return_list
          print_identifier identifier
          print_opt_list opt_list
          (Host.close_fun_dict_fun "endfunction")
      else
        fprintf ppf "@[<v>@[<v 2>function %a@,%a@]@,return %a\n%s@]"
          print_identifier identifier
          print_opt_list opt_list
          print_return_list return_list
          (Host.close_fun_dict_fun "endfunction")

  (* list of returned values by the function *)

  and print_return_list ppf return_list =
    match return_list.Ast_node.ast_desc with
    | EmptyReturn -> fprintf ppf "%s" ""
    | Return identifier_list ->
        if List.length identifier_list = 1 ||
        not Host.return_list_delim then
          fprintf ppf "%a" print_identifier_list identifier_list
        else
          fprintf ppf "[%a]" print_identifier_list identifier_list

  and print_identifier_list ppf = function
    | [] -> ()
    | arg :: args ->
      print_identifier ppf arg;
      List.iter (function arg -> fprintf ppf ", %a" print_identifier arg) args

  (* checks if exprs are endend by comments *)

  and check_exprs_list = function
    | [] -> false
    | [elt] -> check_exprsep_list elt.Ast_node.ast_desc
    | _ :: elts -> check_exprs_list elts

  and check_exprsep_list = function
    | [] -> false
    | [elt] ->
        (match elt.Ast_node.ast_desc with
        | Exprsep (x, _) -> check_expr x)
    | _ :: elts -> check_exprsep_list elts

  and check_expr expr =
    match expr.Ast_node.ast_desc with
    | Expression _
    | Command _
    | WordListCmd _
    | Comments _ -> false
    | Pcomments _
    | Dcomments _ -> true
    | Empty -> false

  (* checks that exprs is a row vector *)

  and check_matrix_isrow exprs =
    List.length exprs = 1 &&
    check_matrix_isrow_exprsep_list (List.hd exprs).Ast_node.ast_desc

  and check_matrix_isrow_exprsep_list = function
    | [] -> true
    | [x] ->
      (match x.Ast_node.ast_desc with
       | Exprsep(_,_) -> true)
    | x :: elts ->
      check_matrix_isrow_exprsep_list elts &&
      match x.Ast_node.ast_desc with
      | Exprsep(_,sep) ->
          match sep.Ast_node.ast_desc with
          | Comma _ -> true
          | Newline _
          | Semi _
          | Sep_None _ -> false

  (* print a row vector *)

  and print_rowmat ppf exprs =
    fprintf ppf "@[@{<m>[%a]@}@]"
      print_rowmat1 (List.hd exprs).Ast_node.ast_desc;

  and print_rowmat1 ppf  = function
    | [] -> ()
    | arg :: args ->
      print_rowelt ppf arg;
      List.iter (function arg -> fprintf ppf "%a" print_rowelt arg) args

  and print_rowelt ppf rowelt =
    match rowelt.Ast_node.ast_desc with
    | Exprsep (x, sep) ->
      fprintf ppf "%a%a@," print_expr x print_sep sep.Ast_node.ast_desc;
  ;;

  let print_ast_desc ppf = function
    | Ast exprs -> fprintf ppf "@[<v>%a@]" print_exprs_list exprs;
  ;;

  let print_matlab_ast ppf ast =
    print_ast_desc ppf ast.Ast_node.ast_desc
  ;;

  (* Translating a Matlab file *)
  let translate_file ppf efname =
    let m_ast = Mtlb_to_rewritten_ast.read_mtlb_file efname in
    Mtlb_ppf.matlab_printer print_matlab_ast Host.tnlm ppf m_ast;
  ;;

  (* Translating a Matlab expression *)
  let translate_expression txt =
    let s = Ast_types.string_of_mdl_text txt in
    Say.debug (Printf.sprintf "<-- translate_expression @%s@" s);
    let m_ast = Mtlb_to_rewritten_ast.read_mtlb_string s in
    let res =
      Mtlb_ppf.make_single_line_sprinter print_matlab_ast m_ast in
    Say.debug (Printf.sprintf "--> translate_expression @%s@" res);
    Ast_types.mdl_text_of_string res
  ;;

  (* Translating a Matlab script *)
  let translate_script txt =
    let s = Ast_types.string_of_mdl_text txt in
    Say.debug (Printf.sprintf "<-- translate_script @%s@" s);
    let m_ast = Mtlb_to_rewritten_ast.read_mtlb_string s in
    let res =
      Mtlb_ppf.matlab_sprinter print_matlab_ast Host.tnlm m_ast in
    Say.debug (Printf.sprintf "--> translate_script @%s@" res);
    Ast_types.mdl_text_of_string res
  ;;

  let quote_text txt =
    let s = Ast_types.string_of_mdl_text txt in
    Say.debug (Printf.sprintf "<-- quote_text @%s@" s);
    let delimiters = Host.text_delimiters txt in
    let add_sq_string b s =
      Buffer.add_string b delimiters.Mtlb_make_host_print_spec.open_text;
      String.iter
        (function c -> Buffer.add_string b (Host.text_char_dict_fun c))
        s;
      Buffer.add_string b delimiters.Mtlb_make_host_print_spec.close_text in

    let res =
      let b = Buffer.create (String.length s) in
      add_sq_string b s;
      Buffer.contents b in
    Say.debug (Printf.sprintf "--> quote_text @%s@" res);
    Ast_types.mdl_text_of_string res
  ;;

end
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
