(***********************************************************************)
(*                                                                     *)
(*                           Matlab Parser                             *)
(*                                                                     *)
(*          J.Ph Chancelier, Enpc/Cermics                              *)
(*                                                                     *)
(*  Copyright 2012-2019,                                               *)
(*  Ecole Nationale des ponts et chaussées                             *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mtlb_make_host_print_spec.mli,v 1.1 2019-05-10 14:09:18 jpc Exp $ *)

(** The format strings and formatters specific to a given host translation. *)

type text_delimiters =
  {
    open_text : string;
    close_text : string;
  }
  (** The delimiters for text in the host language.
    open_text and close_text are typically equal,
    either to "\'" or "\"" for regular strings.
    In some language extra delimiting characters can be added to get
    rid of strings with embedded newlines...
    (E.g. as a matrix of strings with no newlines inside each string.)
  *)
;;

module type HOST = sig

  val then_delim : string
  (** Translation of the then delimiter *)
  ;;

  val return_list_delim : bool
  (** Translation of the return delimiter *)
  ;;

  val do_delim : string
  (** Translation of the do delimiter *)
  ;;

  val persistent_keyword : string
  (** Translation of the persistent keyword *)
  ;;

  val break_case_keyword : string
  (** The break keyword for case of switch *)
  ;;

  val switch_keyword : string
  (** The switch keyword *)
  ;;

  val begin_switch_marker : string
  (** The marker for begin switch *)
  ;;

  val switch_condition_open_marker : string;;
  val switch_condition_close_marker : string
  (** Open and close of case conditional expression *)
  ;;

  val case_condition_sep_marker : string
  (** The separator following case condition *)
  ;;

  val otherwise_keyword : string
  (** The otherwise keyword *)
  ;;

  val end_switch_marker : string
  (** The marker for end switch *)
  ;;

  val persistent_declaration_sep : string
  (** Translation of persistent/global declaration separator *)
  ;;

  val extract_delim_start : string -> string
  (** Specific translation for extraction start delimiter *)
  ;;

  val extract_delim_end : string -> string
  (** Specific translation for extraction end delimiter *)
  ;;

  val imag : string -> string
  (** Specific translation for imaginary constant *)
  ;;

  val dict_fun : string -> string
  (** Specific translation for idents *)
  ;;

  val close_block_dict_fun : string -> string
  (** String marker to close syntactic blocks *)
  ;;

  val close_fun_dict_fun :  string -> string
  (** String marker to close a function ... endfunction definition *)
  ;;

  val comment_char_dict_fun : char -> string
  (** String marker to start comments *)
  ;;

  val postfix_op_dict_fun :  string -> string
  (** Translation for postfix operators
     (in particular the transpose operator) *)
  ;;

  val sq_char_dict_fun : char -> string
  (** Translation for characters in matlab string tokens
     (escaping characters according to host
     conventions).

     Note: no matlab string tokens can contain a \n (ASCII 10) or char (10)
     character. *)
  ;;

  val text_delimiters : Ast_types.mdl_text -> text_delimiters
  (** Returns opening and closing delimiters for quoted texts. *)
  ;;

  val text_char_dict_fun : char -> string
  (** Translation for characters in text from double quoted Mdl strings such
    like block Ids or Names (escaping characters according to host
    conventions).
.
    Note: text tokens may contain arbitrary characters including a \n (ASCII
    10) or char (10) character. *)
  ;;

  val sq_string_delimiter : char
  (** Char delimiter to start and close a single quoted string *)
  ;;

  val not_op : string
  (** Translation of the not logical operator *)
  ;;

  val ne_op : string
  (** Translation of the not equal comparison operator *)
  ;;

  val protect_colon_expr : bool
  (** If true, colon operator expressions must be enclosed in brackets.
     E.g. exp:exp -> [exp:exp]. *)
  ;;

  val protect_lhs : bool
  (** If true, lhs arguments must be enclosed in brackets *)
  ;;

  val tnlm : Mtlb_ppf.tag_new_line_markers
  (** Specify the line continuation and end of line markers
    for the host language.
    E.g. for Nsp the line continuation marker for expressions or commands
    is "...\n". For Oml the line continuation marker is simply "\n", as is
    usual in 'standard' languages. *)
  ;;

(*
  val print_quoted_text :  Format.formatter -> string -> unit
  ;;
*)
end
;;

module type PRINT_TO_HOST = sig

  val translate_file : Format.formatter -> Path.explicit_file_name -> unit
  (** Translating a Matlab file *)
  ;;

  val translate_expression :
    Ast_types.mdl_text -> Ast_types.mdl_text
  (** Translate a Matlab string *)
  ;;

  val translate_script :
    Ast_types.mdl_text -> Ast_types.mdl_text
  (** Translate a Matlab script *)
  ;;

  val quote_text : Ast_types.mdl_text -> Ast_types.mdl_text
  (** Quote a text for the host *)
  ;;

end
;;

module type MAKE_PRINT = functor (Host : HOST) -> PRINT_TO_HOST;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
