(***********************************************************************)
(*                                                                     *)
(*                           Matlab Parser                             *)
(*                                                                     *)
(*          J.Ph Chancelier, Enpc/Cermics                              *)
(*                                                                     *)
(*  Copyright 2012-2019,                                               *)
(*  Ecole Nationale des ponts et chaussées                             *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mtlb_ppf.ml,v 1.1 2019-05-10 14:09:18 jpc Exp $ *)

(* BEGIN: From the 4.01 Format module *)

type formatter_out_functions = {
  out_string : string -> int -> int -> unit;
  out_flush : unit -> unit;
  out_newline : unit -> unit;
  out_spaces : int -> unit;
}
;;

let pp_set_formatter_out_functions ppf out_funs =
  match out_funs with
  | {
      out_string = pp_out_string;
      out_flush = pp_out_flush;
      out_newline = pp_out_newline;
      out_spaces = pp_out_spaces;
    } ->
    Format.pp_set_all_formatter_output_functions ppf
      ~out:pp_out_string ~flush:pp_out_flush
      ~newline:pp_out_newline ~spaces:pp_out_spaces
;;

let pp_get_formatter_out_functions ppf () =
  let (pp_out_string, pp_out_flush,
       pp_out_newline, pp_out_spaces) =
    Format.pp_get_all_formatter_output_functions ppf () in
  {
    out_string = pp_out_string;
    out_flush = pp_out_flush;
    out_newline = pp_out_newline;
    out_spaces = pp_out_spaces;
  }
;;

(* END: From the 4.01 Format module *)

type tag_new_line_markers = {
  m_tag_new_line : string;
  n_tag_new_line : string;
}
;;

type tag_handler = {
  tag_funs : Format.formatter_tag_functions;
  out_funs : Format.formatter_out_functions;
  mark_tags : bool;
}
;;

let set_tag_handler ppf tag_handler =
(*  prerr_endline "Setting the tag handler to a ppf argument";*)
  Format.pp_set_formatter_tag_functions ppf tag_handler.tag_funs;
  Format.pp_set_formatter_out_functions ppf tag_handler.out_funs;
  Format.pp_set_mark_tags ppf tag_handler.mark_tags;
;;

let get_tag_handler ppf =
(*  prerr_endline "Getting the tag handler of a ppf argument";*)
 {
   tag_funs = Format.pp_get_formatter_tag_functions ppf ();
   out_funs = Format.pp_get_formatter_out_functions ppf ();
   mark_tags = Format.pp_get_mark_tags ppf ();
 }
;;

let m_tag_tag_handlers = ref [];;
let push_tag_handler tag_handler =
(*  prerr_endline "Adding a tag handler to the stack";*)
  m_tag_tag_handlers := tag_handler :: !m_tag_tag_handlers
and top_tag_handler () =
(*  prerr_endline "Getting top tag handler";*)
  match !m_tag_tag_handlers with
  | [] -> failwith "top_tag_handler"
  | tag_handler :: _ -> tag_handler
and pop_tag_handler () =
  match !m_tag_tag_handlers with
  | [] -> failwith "pop_tag_handler"
  | _ :: tag_handlers ->
(*    prerr_endline "Removing a tag handler from the stack";*)
    m_tag_tag_handlers := tag_handlers
;;

let make_tag_ppf tnlm ppf =

  let ppf_tag_handler = get_tag_handler ppf in

  let m_tag_out_newline () =
    ppf_tag_handler.out_funs.Format.out_string
      tnlm.m_tag_new_line 0 (String.length tnlm.m_tag_new_line) in

  let n_tag_out_newline () =
    ppf_tag_handler.out_funs.Format.out_string
      tnlm.n_tag_new_line 0 (String.length tnlm.n_tag_new_line) in

  let rec m_tag_mark_open_tag = function
    | "m" | "" ->
      push_tag_handler (get_tag_handler ppf);
      set_tag_handler ppf m_tag_tag_handler;
      ""
    | "n" ->
      push_tag_handler (get_tag_handler ppf);
      set_tag_handler ppf n_tag_tag_handler;
      ""
    | tag -> ppf_tag_handler.tag_funs.Format.mark_open_tag tag

  and m_tag_mark_close_tag = function
    | "m" | "" ->
      let ppf_tag_handler = top_tag_handler () in
      set_tag_handler ppf ppf_tag_handler;
      pop_tag_handler ();
      ""
    | "n" ->
      let ppf_tag_handler = top_tag_handler () in
      set_tag_handler ppf ppf_tag_handler;
      pop_tag_handler ();
      ""
    | tag -> ppf_tag_handler.tag_funs.Format.mark_close_tag tag

  and m_tag_tag_handler = {
    tag_funs = {
      ppf_tag_handler.tag_funs with
      Format.mark_open_tag = m_tag_mark_open_tag;
      Format.mark_close_tag = m_tag_mark_close_tag;
    };
    out_funs = {
      ppf_tag_handler.out_funs with
      Format.out_newline = m_tag_out_newline;
    };
    mark_tags = true;
  }

  and n_tag_tag_handler = {
    tag_funs = {
      ppf_tag_handler.tag_funs with
      Format.mark_open_tag = m_tag_mark_open_tag;
      Format.mark_close_tag = m_tag_mark_close_tag;
    };
    out_funs = {
      ppf_tag_handler.out_funs with
      Format.out_newline = n_tag_out_newline;
    };
    mark_tags = true;
  } in

  Format.pp_set_formatter_tag_functions ppf m_tag_tag_handler.tag_funs;
  Format.pp_set_mark_tags ppf m_tag_tag_handler.mark_tags;
  ppf
;;

let make_sprinter mk_ppf printer ast =
  let b = Buffer.create 1024 in
  let ppf = Format.make_formatter (Buffer.add_substring b) ignore in
  printer (mk_ppf ppf) ast;
  Format.pp_print_flush ppf ();
  Buffer.contents b;
;;

(* Single line ppf. *)
let make_single_line_ppf ppf =
  let out_functions = pp_get_formatter_out_functions ppf () in
  pp_set_formatter_out_functions ppf
    { out_functions with
      out_newline = ignore;
    };
  ppf
;;

let make_tag_printer tnlm printer ppf ast =
  let mark_tags = Format.pp_get_mark_tags ppf () in
  printer (make_tag_ppf tnlm ppf) ast;
  Format.pp_set_mark_tags ppf mark_tags;
;;

let make_single_line_printer printer ppf ast =
  let out_functions = pp_get_formatter_out_functions ppf () in
  printer (make_single_line_ppf ppf) ast;
  pp_set_formatter_out_functions ppf out_functions
;;

let make_single_line_tag_printer tnlm printer ppf ast =
  let mark_tags = Format.pp_get_mark_tags ppf () in
  let out_functions = pp_get_formatter_out_functions ppf () in

  let ppf = make_tag_ppf tnlm (make_single_line_ppf ppf) in
  printer ppf ast;

  pp_set_formatter_out_functions ppf out_functions;
  Format.pp_set_mark_tags ppf mark_tags;
;;

let make_single_line_sprinter printer ast =
  make_sprinter make_single_line_ppf printer ast
;;

let make_tag_sprinter tnlm printer ast =
  make_sprinter (make_tag_ppf tnlm) printer ast
;;

let make_single_line_tag_sprinter tnlm printer ast =
  make_sprinter
   (fun ppf -> make_tag_ppf tnlm (make_single_line_ppf ppf))
   printer ast
;;

let matlab_printer printer tnlm ppf ast =
  let mark_tags = Format.pp_get_mark_tags ppf () in
  printer (make_tag_ppf tnlm ppf) ast;
  Format.pp_set_mark_tags ppf mark_tags;
;;

let matlab_sprinter printer tnlm ast =
  make_tag_sprinter tnlm printer ast
;;

(* Examples:

let ppf = make_matlab_ppf (Format.formatter_of_out_channel stdout);;

let p fmt = Format.fprintf ppf fmt;;

p "@[<v>@{@[<v>poi@{<n>@ @}@ poi@]@}toto@ @]@."
;;

p "@{<n>\
    toto;@{\
           @[<v>function@ \
           @{<n>\
             poipoi.@ \
             poipoi.\
           @}@ \
           toto\
           @]\
         @}\
   @}\
   @."
;;
*)

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
