(***********************************************************************)
(*                                                                     *)
(*                           Matlab Parser                             *)
(*                                                                     *)
(*          J.Ph Chancelier, Enpc/Cermics                              *)
(*                                                                     *)
(*  Copyright 2012-2019,                                               *)
(*  Ecole Nationale des ponts et chaussées                             *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mtlb_rewrite_ast.ml,v 1.1 2019-05-10 14:09:18 jpc Exp $ *)

open Ast_node;;
open Mtlb_ast;;

type rewriting =
   | RewritingMtlb
;;

let rewrite_ast_node f node =
  {
    node with
    ast_desc = f node.ast_desc;
    ast_info = RewritingMtlb;
  }
;;

let rewrite_mtlb_constant = function
  | Num _s
  | Imag_num _s
  | Dq_string _s
  | Sq_string _s as desc -> desc
;;

let rewrite_constant_desc = function
  | Constant m ->
    Constant (rewrite_mtlb_constant m)
;;

let rewrite_constant = function
  | i_c_i_a ->
    rewrite_ast_node rewrite_constant_desc i_c_i_a
;;

let rewrite_sep_desc = function
  | Comma _i
  | Newline _i
  | Semi _i
  | Sep_None _i as desc -> desc
;;

let rewrite_sep = function
  | i_s_i_a ->
    rewrite_ast_node rewrite_sep_desc i_s_i_a
;;

let make_empty_return_list node =
  {
    node with
    ast_desc = EmptyReturn;
    ast_info = RewritingMtlb;
  }
;;

let rewrite_wcmds_desc = function
  | SQ _s
  | DQ _s as desc -> desc
;;

let rewrite_word_list_desc = function
  | i_w_l ->
    List.map rewrite_wcmds_desc i_w_l
;;

let rewrite_word_list = function
  | i_w_i_a ->
    rewrite_ast_node rewrite_word_list_desc i_w_i_a
;;

let rec rewrite_identifier = function
  | i_i_i_a ->
    rewrite_ast_node rewrite_identifier_desc i_i_i_a

and rewrite_identifier_desc = function
  | id -> id

and rewrite_meta_identifier = function
  | i_m_i_a ->
    rewrite_ast_node rewrite_meta_identifier_desc i_m_i_a

and rewrite_meta_identifier_desc = function
  | id -> id

and rewrite_superclass_identifier = function
  | i_s_i_a ->
    rewrite_ast_node rewrite_superclass_identifier_desc i_s_i_a

and rewrite_superclass_identifier_desc = function
  | id -> id
;;

let rewrite_endfun_desc = function
  | End_fun
  | Eof_fun as desc -> desc

and rewrite_return_list_desc = function
  | EmptyReturn as desc -> desc
  | Return i_i_l ->
    Return (List.map rewrite_identifier i_i_l)

and rewrite_op_compare_desc = function
  | Lt
  | Le
  | Eq
  | Ne
  | Ge
  | Gt as desc -> desc

and rewrite_op_stars_desc = function
  | Star
  | Dot_star as desc -> desc

and rewrite_op_divs_desc = function
  | Slash
  | Bslash
  | Dot_slash
  | Dot_bslash as desc -> desc
;;

let rewrite_divs_desc = function
  | (i_o, s) ->
    rewrite_op_divs_desc i_o, s
;;

let rewrite_stars_desc = function
  | (i_o, s) ->
    rewrite_op_stars_desc i_o, s
;;

let rewrite_compare_desc = function
  | (i_o, s) ->
    rewrite_op_compare_desc i_o, s
;;

let rewrite_endfun = function
  | i_e_i_a ->
    rewrite_ast_node rewrite_endfun_desc  i_e_i_a

and rewrite_return_list = function
  | i_r_i_a ->
    rewrite_ast_node rewrite_return_list_desc i_r_i_a

and rewrite_compare  = function
  | i_c_i_a ->
    rewrite_ast_node rewrite_compare_desc i_c_i_a

and rewrite_stars  = function
  | i_s_i_a ->
    rewrite_ast_node rewrite_stars_desc i_s_i_a

and rewrite_divs  = function
  | i_d_i_a ->
    rewrite_ast_node rewrite_divs_desc i_d_i_a
;;

let rec rewrite_mtlb_ast = function
  | i_i_i_a ->
    let return_list =
      make_empty_return_list i_i_i_a in
    rewrite_ast_node (rewrite_mtlb_ast_desc return_list) i_i_i_a

and rewrite_mtlb_ast_desc return_list = function
  | Ast i_e_l ->
    Ast (List.map (rewrite_exprs return_list) i_e_l);

and rewrite_exprs return_list = function
  | i_e_i_a -> rewrite_ast_node (rewrite_exprs_desc return_list) i_e_i_a

and rewrite_exprs_desc return_list = function
  | i_e_l -> List.map (rewrite_exprsep return_list) i_e_l

and rewrite_exprsep return_list = function
  | i_e_i_a ->
    rewrite_ast_node (rewrite_exprsep_desc return_list) i_e_i_a

and rewrite_exprsep_desc return_list = function
  | Exprsep (i_e, i_s) ->
    Exprsep (rewrite_expr return_list i_e, rewrite_sep i_s)

and rewrite_expr return_list = function
  | i_e_i_a ->
    rewrite_ast_node (rewrite_expr_desc return_list) i_e_i_a

and rewrite_expr_desc return_list = function
  | Expression i_e ->
    Expression (rewrite_expression return_list i_e)
  | Command i_c ->
    Command (rewrite_command return_list i_c)
  | WordListCmd (i_i, i_w) ->
    WordListCmd (rewrite_identifier i_i, rewrite_word_list i_w)
  | Comments _s
  | Pcomments _s
  | Dcomments _s as desc -> desc
  | Empty as desc -> desc

and rewrite_expression return_list = function
  | i_e_i_a ->
    rewrite_ast_node (rewrite_expression_desc return_list) i_e_i_a

and rewrite_expression_desc return_list = function
  | SimpleExpr i_s ->
    SimpleExpr (rewrite_simple_expr return_list i_s)
  | AssignExpr i_a ->
    AssignExpr (rewrite_assign_expr return_list i_a)
  | AnonFcnHandle (i_p, i_e) ->
    AnonFcnHandle (
      rewrite_param_list return_list i_p,
      rewrite_expression return_list i_e)

and rewrite_command return_list = function
  | i_c_i_a ->
    rewrite_ast_node (rewrite_command_desc return_list) i_c_i_a

and rewrite_command_desc return_list = function
  | Declaration i_d ->
    Declaration (rewrite_declaration return_list i_d)
  | Switch (i_e, i_c) ->
    Switch (
      rewrite_expression return_list i_e,
      rewrite_case_list return_list i_c)
  | IfCommand i_i ->
    IfCommand (rewrite_if_command return_list i_i)
  | LoopCommand i_l ->
    LoopCommand (rewrite_loop_command return_list i_l)
  | JumpCommand i_j ->
    JumpCommand (rewrite_jump_command return_list i_j)
  | ExceptCommand i_e ->
    ExceptCommand (rewrite_except_command return_list i_e)
  | SciptFile i_o ->
    SciptFile (rewrite_opt_list return_list i_o)
  | MFunction i_m ->
    MFunction (rewrite_mfunction i_m)

and rewrite_jump_command return_list = function
  | i_j_i_a ->
    rewrite_ast_node (rewrite_jump_command_desc return_list) i_j_i_a

and rewrite_jump_command_desc return_list = function
  | Break
  | Continue as desc -> desc
  | FReturn _i_r ->
    FReturn return_list

and rewrite_primary_expr return_list = function
  | i_p_i_a ->
    rewrite_ast_node (rewrite_primary_expr_desc return_list) i_p_i_a

and rewrite_primary_expr_desc return_list = function
  | PexpIdentifier i_i ->
    PexpIdentifier (rewrite_identifier i_i)
  | PexpConstant i_c ->
    PexpConstant (rewrite_constant i_c)
  | PexpFcnHandle i_i ->
    PexpFcnHandle (rewrite_identifier i_i)
  | PexpMatrix i_e_l ->
    PexpMatrix (List.map (rewrite_exprs return_list) i_e_l)
  | PexpCell i_e_l ->
    PexpCell (List.map (rewrite_exprs return_list) i_e_l)
  | PexpMetaIdentifier i_m ->
    PexpMetaIdentifier (rewrite_meta_identifier i_m)
  | PexpSuperIdentifier i_s ->
    PexpSuperIdentifier (rewrite_superclass_identifier i_s)
  | PexpExpression i_e ->
    PexpExpression (rewrite_expression return_list i_e)

and rewrite_arg_list return_list = function
  | i_a_i_a ->
    rewrite_ast_node (rewrite_arg_list_desc return_list) i_a_i_a

and rewrite_arg_list_desc return_list = function
  | i_a_l -> List.map (rewrite_arg return_list) i_a_l

and rewrite_arg return_list = function
  | i_a_i_a ->
    rewrite_ast_node (rewrite_arg_desc return_list) i_a_i_a

and rewrite_arg_desc return_list = function
  | ArgExpression i_e ->
    ArgExpression (rewrite_expression return_list i_e)
  | ArgColon
  | ArgTilde as desc -> desc

and rewrite_oper_expr return_list = function
  | i_o_i_a ->
    rewrite_ast_node (rewrite_oper_expr_desc return_list) i_o_i_a

and rewrite_oper_expr_desc return_list = function
  | PrimaryExpr i_p ->
    PrimaryExpr (rewrite_primary_expr return_list i_p)
  | PostfixOp (s, i_o) ->
    PostfixOp (s, rewrite_oper_expr return_list i_o)
  | IndexExpr (i_o, i_a, s, s0) ->
    IndexExpr (
      rewrite_oper_expr return_list i_o,
      rewrite_arg_list return_list i_a, s, s0)
  | IndirectRef (i_o, i_e) ->
    IndirectRef (
      rewrite_oper_expr return_list i_o,
      rewrite_expression return_list i_e)
  | IndirectRefU (i_o, i_i) ->
    IndirectRefU (
      rewrite_oper_expr return_list i_o,
      rewrite_identifier i_i)
  | PrefixOpUnary (s, i_o) ->
    PrefixOpUnary (
      s,
      rewrite_oper_expr return_list i_o)
  | BinaryOp (s, i_o, i_o0) ->
    BinaryOp (
      s,
      rewrite_oper_expr return_list i_o,
      rewrite_oper_expr return_list i_o0)
  | Op_stars (i_s, i_o, i_o0) ->
    Op_stars (
      rewrite_stars i_s,
      rewrite_oper_expr return_list i_o,
      rewrite_oper_expr return_list i_o0)
  | Op_divs (i_d, i_o, i_o0) ->
    Op_divs (
      rewrite_divs i_d,
      rewrite_oper_expr return_list i_o,
      rewrite_oper_expr return_list i_o0)

and rewrite_simple_expr return_list = function
  | i_s_i_a ->
    rewrite_ast_node (rewrite_simple_expr_desc return_list) i_s_i_a

and rewrite_simple_expr_desc return_list = function
  | ColonExpr i_o_l ->
    ColonExpr (List.map (rewrite_oper_expr return_list) i_o_l)
  | Op_compare (i_c, i_s, i_s0) ->
    Op_compare (
      rewrite_compare i_c,
      rewrite_simple_expr return_list i_s,
      rewrite_simple_expr return_list i_s0)
  | SexpBinaryOp (s, i_s, i_s0) ->
    SexpBinaryOp (
      s, rewrite_simple_expr return_list i_s,
      rewrite_simple_expr return_list i_s0)
  | BooleanOp (s, i_s, i_s0) ->
    BooleanOp (
      s,
      rewrite_simple_expr return_list i_s,
      rewrite_simple_expr return_list i_s0)

and rewrite_param_list return_list = function
  | i_p_i_a ->
    rewrite_ast_node (rewrite_param_list_desc return_list) i_p_i_a

and rewrite_param_list_desc return_list = function
  | EmptyParamList as desc -> desc
  | ParamList i_d_l ->
    ParamList (List.map (rewrite_decl2 return_list) i_d_l)

and rewrite_assign_lhs return_list = function
  | i_a_i_a ->
    rewrite_ast_node (rewrite_assign_lhs_desc return_list) i_a_i_a

and rewrite_assign_lhs_desc return_list = function
  | LhsSimple i_s ->
    LhsSimple (rewrite_simple_expr return_list i_s)
  | LhsArgs i_e ->
    LhsArgs (rewrite_exprs return_list i_e)

and rewrite_assign_expr return_list = function
  | i_a_i_a ->
    rewrite_ast_node (rewrite_assign_expr_desc return_list) i_a_i_a

and rewrite_assign_expr_desc return_list = function
  | Assign (s, i_a, i_e) ->
    Assign (
      s,
      rewrite_assign_lhs return_list i_a,
      rewrite_expression return_list i_e)

and rewrite_declaration return_list = function
  | i_d_i_a ->
    rewrite_ast_node (rewrite_declaration_desc return_list) i_d_i_a

and rewrite_declaration_desc return_list = function
  | Global i_d_l ->
    Global (List.map (rewrite_decl2 return_list) i_d_l)
  | Static i_d_l ->
    Static (List.map (rewrite_decl2 return_list) i_d_l)

and rewrite_decl2 return_list = function
  | i_d_i_a ->
    rewrite_ast_node (rewrite_decl2_desc return_list) i_d_i_a

and rewrite_decl2_desc return_list = function
  | Declare i_i ->
    Declare (rewrite_identifier i_i)
  | DeclareSet (i_i, i_e) ->
    DeclareSet (rewrite_identifier i_i, rewrite_expression return_list i_e)
  | MagicTilde as desc -> desc

and rewrite_case_list return_list = function
  | i_c_i_a ->
    rewrite_ast_node (rewrite_case_list_desc return_list) i_c_i_a

and rewrite_case_list_desc return_list = function
  | Nocase as desc -> desc
  | DefaultCase i_d ->
    DefaultCase (rewrite_default_case return_list i_d)
  | Cases i_s_l ->
    Cases (List.map (rewrite_switch_case return_list) i_s_l)
  | CasesOther (i_s_l, i_d) ->
    CasesOther (
      List.map (rewrite_switch_case return_list) i_s_l,
      rewrite_default_case return_list i_d)

and rewrite_switch_case return_list = function
  | i_s_i_a ->
    rewrite_ast_node (rewrite_switch_case_desc return_list) i_s_i_a

and rewrite_switch_case_desc return_list = function
  | Case (i_o, i_o0, i_o1) ->
    Case (
      rewrite_opt_list return_list i_o,
      rewrite_opt_list return_list i_o0,
      rewrite_opt_list return_list i_o1)

and rewrite_default_case return_list = function
  | i_d_i_a ->
    rewrite_ast_node (rewrite_default_case_desc return_list) i_d_i_a

and rewrite_default_case_desc return_list = function
  | Otherwize i_o ->
    Otherwize (rewrite_opt_list return_list i_o)

and rewrite_opt_list return_list = function
  | i_o_i_a ->
    rewrite_ast_node (rewrite_opt_list_desc return_list) i_o_i_a

and rewrite_opt_list_desc return_list = function
  | i_e_l -> List.map (rewrite_exprs return_list) i_e_l

and rewrite_if_command return_list  = function
  | i_i_i_a ->
    rewrite_ast_node (rewrite_if_command_desc return_list) i_i_i_a

and rewrite_if_command_desc return_list = function
  | IfCmd i_i_l ->
    IfCmd (List.map (rewrite_if_cmd return_list) i_i_l)
  | IfCmdWithElse (i_i_l, i_e) ->
    IfCmdWithElse (
      List.map (rewrite_if_cmd return_list) i_i_l,
      rewrite_else_clause return_list i_e)

and rewrite_if_cmd return_list = function
  | i_i_i_a ->
    rewrite_ast_node (rewrite_if_cmd_desc return_list) i_i_i_a

and rewrite_if_cmd_desc return_list = function
  | If (i_o, i_o0, i_o1) ->
    If (
      rewrite_opt_list return_list i_o,
      rewrite_opt_list return_list i_o0,
      rewrite_opt_list return_list i_o1)
  | ElseIf (i_o, i_o0, i_o1) ->
    ElseIf (
      rewrite_opt_list return_list i_o,
      rewrite_opt_list return_list i_o0,
      rewrite_opt_list return_list i_o1)

and rewrite_else_clause return_list = function
  | i_e_i_a ->
    rewrite_ast_node (rewrite_else_clause_desc return_list) i_e_i_a

and rewrite_else_clause_desc return_list = function
  | Else i_o ->
    Else (rewrite_opt_list return_list i_o)

and rewrite_loop_command return_list = function
  | i_l_i_a ->
    rewrite_ast_node (rewrite_loop_command_desc return_list) i_l_i_a

and rewrite_loop_command_desc return_list = function
  | WhileCommand (i_o, i_o0, i_o1) ->
    WhileCommand (
      rewrite_opt_list return_list i_o,
      rewrite_opt_list return_list i_o0,
      rewrite_opt_list return_list i_o1)
  | DoUntilCommand (i_o, i_e) ->
    DoUntilCommand (
      rewrite_opt_list return_list i_o,
      rewrite_expression return_list i_e)
  | ForCommand (i_o, i_o0, i_o1) ->
    ForCommand (
      rewrite_opt_list return_list i_o,
      rewrite_opt_list return_list i_o0,
      rewrite_opt_list return_list i_o1)

and rewrite_except_command return_list = function
  | i_e_i_a ->
    rewrite_ast_node (rewrite_except_command_desc return_list) i_e_i_a

and rewrite_except_command_desc return_list = function
  | UnwindCommand (i_o, i_o0) ->
    UnwindCommand (
      rewrite_opt_list return_list i_o,
      rewrite_opt_list return_list i_o0)
  | TryCatchCommand (i_o, i_o0) ->
    TryCatchCommand (
      rewrite_opt_list return_list i_o,
      rewrite_opt_list return_list i_o0)
  | TryCommand i_o ->
    TryCommand (rewrite_opt_list return_list i_o)

and rewrite_mfunction = function
  | i_m_i_a ->
    rewrite_ast_node rewrite_mfunction_desc i_m_i_a

and rewrite_mfunction_desc = function
  | FunctionNoRet (i_i, i_b) ->
    let return_list = make_empty_return_list i_i in
    FunctionNoRet (rewrite_identifier i_i, rewrite_body return_list i_b)
  | Function (i_r, i_i, i_b) ->
    let return_list = rewrite_return_list i_r in
    Function (
      return_list,
      rewrite_identifier i_i,
      rewrite_body return_list i_b)

and rewrite_body return_list = function
  | i_b_i_a ->
    rewrite_ast_node (rewrite_body_desc return_list) i_b_i_a

and rewrite_body_desc return_list = function
  | Body (i_p, i_o, i_e) ->
    Body (
      rewrite_param_list return_list i_p,
      rewrite_opt_list return_list i_o,
      rewrite_endfun i_e)
  | SimpleBody (i_o, i_e) ->
    SimpleBody (rewrite_opt_list return_list i_o, rewrite_endfun i_e)
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
