(***********************************************************************)
(*                                                                     *)
(*                           Matlab Parser                             *)
(*                                                                     *)
(*          J.Ph Chancelier, Enpc/Cermics                              *)
(*                                                                     *)
(*  Copyright 2012-2019,                                               *)
(*  Ecole Nationale des ponts et chaussées                             *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: m_to_host.ml,v 1.1 2019-05-10 14:09:18 jpc Exp $ *)

(** The procedures to translate MatLab code to Scicoslab code. *)

open Simport_configuration;;

module Print_to_octave =
  Mtlb_make_host_print.Make_print (Mtlb_octave_host.Host)
;;

module Print_to_matlab =
  Mtlb_make_host_print.Make_print (Mtlb_matlab_host.Host)
;;

module Print_to_nsp =
  Mtlb_make_host_print.Make_print (Mtlb_nsp_host.Host)
;;

module Print_to_modelica =
  Mtlb_make_host_print.Make_print (Mtlb_modelica_host.Host)
;;

module Print_to_hypermath =
  Mtlb_make_host_print.Make_print (Mtlb_hypermath_host.Host)
;;

module Print_to_oml =
  Mtlb_make_host_print.Make_print (Mtlb_oml_host.Host)
;;

module Print_to_scicoslab =
  Mtlb_make_host_print.Make_print (Mtlb_scicoslab_host.Host)
;;

type 'a translators = {
  octave : 'a;
  matlab : 'a;
  nsp : 'a;
  modelica : 'a;
  hypermath : 'a;
  oml : 'a;
  scicoslab : 'a;
}
;;

let rec translate_with translators lang =
  match lang with
  | Simport_configuration.Target_mtlb lang ->
    (match lang with
    | Simport_configuration.Matlab -> translators.matlab
    | Scicos_host Octave -> translators.octave
    | Scicos_host ScicosLab -> translators.scicoslab
    | Scicos_host Nsp -> translators.nsp
    | Scicos_host Modelica -> translators.modelica
    | Scicos_host Hypermath -> translators.hypermath
    | Scicos_host Oml -> translators.oml
    )

  | Simport_configuration.Target_simulink lang ->
    translate_with translators (Target_mtlb (Scicos_host lang))
  | Simport_configuration.Target_sime _ ->
    Say.fatal_error
      "Cannot translate matlab file to a scicos embedded expression"
;;

(** {6 Translating files} *)

let file_translators = {
  octave = Print_to_octave.translate_file;
  matlab = Print_to_matlab.translate_file;
  nsp = Print_to_nsp.translate_file;
  modelica = Print_to_modelica.translate_file;
  hypermath = Print_to_hypermath.translate_file;
  oml = Print_to_oml.translate_file;
  scicoslab = Print_to_scicoslab.translate_file;
}
;;

let translate_file_to = translate_with file_translators
(** Reads a MatLab file, translates it to the target language, and prints
    the translation onto the formatter argument. *)
;;

let translate_file ppf =
  translate_file_to (Configuration.get_target_language ()) ppf
(** Reads a MatLab file, translates it to the current target language, and
  prints the translation onto the formatter argument. *)
;;

(** {6 Translating expressions} *)

let expression_translators = {
  octave = Print_to_octave.translate_expression;
  matlab = Print_to_matlab.translate_expression;
  nsp = Print_to_nsp.translate_expression;
  modelica = Print_to_modelica.translate_expression;
  hypermath = Print_to_hypermath.translate_expression;
  oml = Print_to_oml.translate_expression;
  scicoslab = Print_to_scicoslab.translate_expression;
}
;;

let translate_expression_to = translate_with expression_translators
(** Reads a string that contains a MatLab expression and
  returns its translation to the host language as a single line string. *)
;;

let translate_expression txt =
  translate_expression_to (Configuration.get_target_language ()) txt
(** Reads a string that contains a MatLab expression and
  returns its translation to the current host language as a single line
  string. *)
;;

(** {6 Translating scripts} *)

let script_translators = {
  octave = Print_to_octave.translate_script;
  matlab = Print_to_matlab.translate_script;
  nsp = Print_to_nsp.translate_script;
  modelica = Print_to_modelica.translate_script;
  hypermath = Print_to_hypermath.translate_script;
  oml = Print_to_oml.translate_script;
  scicoslab = Print_to_scicoslab.translate_script;
}
;;

let translate_script_to = translate_with script_translators
(** Reads a string that contains a MatLab script and
  returns its translation to the host language as a (multi-line) string. *)
;;

let translate_script s =
  translate_script_to (Configuration.get_target_language ()) s
(** Reads a string that contains a MatLab script and
  returns its translation to the current host language as a (multi-line) string. *)
;;

(** {6 Quoting texts} *)

type 'a quoted = Ast_types.mdl_text;;

let quote_translators = {
  octave = Print_to_octave.quote_text;
  matlab = Print_to_matlab.quote_text;
  nsp = Print_to_nsp.quote_text;
  modelica = Print_to_modelica.quote_text;
  hypermath = Print_to_hypermath.quote_text;
  oml = Print_to_oml.quote_text;
  scicoslab = Print_to_scicoslab.quote_text;
}
;;

let quote_text_to = translate_with quote_translators
(** Reads a text and returns its proper quotation to the given language
  as a text containing a valid string token for the language. *)

let quote_quoted_text_to = quote_text_to;;

let quote_text txt =
  quote_text_to
    (Configuration.get_target_language ()) txt
(** Reads a text and returns its proper quotation to the host language
  as a text containing a valid string token for the host language. *)
;;

let quote_quoted_text = quote_text
(** Quote again an already quoted text. *)
;;

(** {6 Printing quoted texts} *)
let print_quoted_text ppf quoted_text =
  let txt = Ast_types.string_of_mdl_text quoted_text in
  Format.fprintf ppf "%s" txt
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
