(***********************************************************************)
(*                                                                     *)
(*                           Matlab Parser                             *)
(*                                                                     *)
(*          J.Ph Chancelier, Enpc/Cermics                              *)
(*                                                                     *)
(*  Copyright 2012-2019,                                               *)
(*  Ecole Nationale des ponts et chaussées                             *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: m_to_host.mli,v 1.1 2019-05-10 14:09:18 jpc Exp $ *)

(** {3 Translating Matlab code to host language code. *)

open Ast_types;;

(** {6 Quoting a text as a valid string token} *)

type 'a quoted = private Ast_types.mdl_text
(** The private type of texts that have been quoted as valid string tokens of
  some language.
  Also the result type of the various [quote_*] functions below. *)
;;

val quote_text_to : Simport_configuration.target_language ->
  mdl_text -> mdl_text quoted
(** Reads a text and returns its proper quotation as a text containing a
  valid string token of the given target language. *)
;;

val quote_quoted_text_to : Simport_configuration.target_language ->
  mdl_text quoted -> mdl_text quoted quoted
(** Reads a text containing a valid string token and returns its proper
  quotation as a text containing a valid string token of the given
  language. *)
;;

val print_quoted_text : Format.formatter -> 'a quoted -> unit
(** Prints a quoted text of any level.

  Note: this function does not quote the text argument.
  The text should be quoted before using the relevant [quote_text] function,
  according to the relevant host.
  Note: this function is host independent, since it simply prints the quoted
  text whatever the host could be. *)
;;

(** {6 Translating Matlab to a target language} *)
val translate_file_to :
  Simport_configuration.target_language ->
  Format.formatter -> Path.explicit_file_name -> unit
(** Reads a MatLab file, translates it to the given language,
  and prints the translation onto the formatter argument. *)
;;

val translate_expression_to :
  Simport_configuration.target_language ->
  Ast_types.mdl_text -> mdl_text
(** Reads a text that contains a MatLab expression, and returns its
  translation to the curernt Scicos host language as a single line text. *)
;;

val translate_script_to :
  Simport_configuration.target_language ->
  mdl_text -> mdl_text
(** Reads a text that contains a MatLab script, and returns its translation
  to the given language as a possibly multiline text. *)
;;

(** {6 Translating Matlab to the current Scicos host language} *)

val translate_file :
  Format.formatter -> Path.explicit_file_name -> unit
(** Reads a MatLab file, translates it to the current Scicos host language
  and prints the translation onto the formatter argument. *)
;;

val translate_expression : Ast_types.mdl_text -> mdl_text
(** Reads a text that contains a MatLab expression, and returns its
  translation to the curernt Scicos host language as a single line text. *)
;;

val translate_script : mdl_text -> mdl_text
(** Reads a text that contains a MatLab script, and returns its translation
  to the current Scicos host language as a possibly multiline text. *)
;;

val quote_text : mdl_text -> mdl_text quoted
(** Same functionality as [quote_text_to] specialized to the current
  Scicos host language. *)
;;

val quote_quoted_text : mdl_text quoted -> mdl_text quoted quoted
(** Same functionality as [quote_quoted_text_to] specialized to the current
  Scicos host language. *)
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
