(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_action.ml,v 1.1 2019-05-10 14:07:13 jpc Exp $ *)

(** {3 Action phase} *)

(** From models in [Mld_split_ast] syntax to models using [Mdl_action_ast]
  syntax. *)

open Mdl_implementation_explicit_ast
open Mdl_equiped_block
open Mdl_basics_deep_ast
open Mdl_action_ast
;;

let string_of_block_name = Mdl_explicit_block_utils.string_of_block_name;;

(** Map line connectors to directed connectors or fail.

  After this 'action' phase, all activation lines connect to normal ports via
  directed connectors: all connectors to special Simulink ``activation
  ports'' have been rerouted to normal ports.

  In effect, activation ports belong to Simulink block of action and this
  phase translates blocks of action to so-called 'activated' blocks.  An
  'activated' block is a compound of a regular block and an extra companion
  event generator block; so lines to the original Simulink activation port
  now connect to the corresponding event generator block of the activated
  block.

 *)

open Mdl_split_ast;;

(* Given a block of action [boa_block_name] and its corresponding
   block event generator [egb_block_name],
   create an event link from [egb_block_name] (port 1)
   to [boa_block_name] (port 1) *)
let create_link make_link_desc src_name src_port dst_name dst_port =
  let make_link_connector block_name port_number =
    {
      explicit_connector_port_block = block_name;
      explicit_connector_port_desc = port_number;
    } in

  Say.debug "Mdl_action: computing port numbers";

  let src_port_number = Ast_types.mdl_port_number_of_natural src_port
  and dst_port_number = Ast_types.mdl_port_number_of_natural dst_port in

  Say.debug "Mdl_action: port numbers computed";

  let link_source = make_link_connector src_name src_port_number
  and link_destination = make_link_connector dst_name dst_port_number in

  let link_desc = make_link_desc link_source link_destination
  and link_name =
    Ast_types.mdl_name_of_string
      (Printf.sprintf
         "Link from %s to %s"
         (string_of_block_name src_name)
         (string_of_block_name dst_name)) in

  {
    Mdl_split_ast.
    split_link_name = Some link_name;
    split_link_font = Mdl_make_block.dummy_explicit_font;
    split_link_labels = None;
    split_link_points = None;
    split_link_desc = link_desc;
  }
;;

let make_event_link_desc link_source link_destination =
  let action_event_connectors =
    {
      split_link_src = link_source;
      split_link_dst = link_destination;
    } in
  Mdl_action_ast.Action_event_link action_event_connectors

and make_regular_link_desc link_source link_destination =
  let action_directed_connectors =
    {
      split_link_src = link_source;
      split_link_dst = link_destination;
    } in
  Mdl_action_ast.Action_directed_link action_directed_connectors
;;

let create_event_link = create_link make_event_link_desc;;
let create_regular_link = create_link make_regular_link_desc;;

let create_activated_event_link boa_block_name ab =
  let egb_block_name =
    ab.Mdl_equiped_block.actioned_block.
    event_generator.explicit_block_name in
  create_event_link
    egb_block_name Unatural._1_
    boa_block_name Unatural._1_
;;

(* Given a block of action [boa_block_name]
   and its corresponding [ab], an [activated_block] value,
   create zero, one or two regular links:
   If [ab] has kind
   - No_link then do nothing,
   - One_link then
     * create a regular link from [egb_block_name] (port 1)
     to [boa_block_name] (last port)
   - Two_link then
     * create a regular link from [egb_block_name] (out port 1)
     to [boa_block_name] (last port - 1)
     * create another regular link from [egb_block_name] (out port 2)
     to [boa_block_name] (last port)

  Note: in this phase the [boa] block has its nin already set to the right
  number of ports.
*)
let create_activated_regular_links boa_block_name ab action_links =
  match ab with
  | {
      actioned_block = {
        block_of_action = boa;
        event_generator = egb;
      };
      activated_kind = activated_kind;
      boa_kind = _boa_kind;
    } ->
    let egb_block_name = egb.explicit_block_name in

    let nin = Mdl_explicit_block_utils.get_block_regular_port_nin boa in
    Say.debug (Printf.sprintf "boa port nin is %d" (Unatural.to_int nin));

    match activated_kind with
    | No_link -> action_links
    | One_link ->
      let link1 =
        create_regular_link
          egb_block_name Unatural._1_
          boa_block_name nin in
      link1 :: action_links
    | Two_link ->
      let link1 =
        create_regular_link
          egb_block_name Unatural._1_
          boa_block_name (Unatural.pred nin) in
      let link2 =
        create_regular_link
          egb_block_name Unatural._2_
          boa_block_name nin in
      link1 :: link2 :: action_links
;;

(**
  In order to activate the Scicos translation of a block of action
  sub-system, we need to add a fake Simulink AtomicPort block to the
  [ebss] list of blocks of the sub-system [eb],
  We compute the position of this new AtomicPort block:
  it is inside the block of action sub-system, centered at the north
  position.
*)
let compute_atomic_port_position ebs =
  let (xmin, xmax, ymin) =
    List.fold_left
      (fun (xmin, xmax, ymin) eb ->
       let position = Mdl_explicit_block_utils.get_block_position eb in
       (min xmin position.xmin,
        max xmax position.xmax,
        min ymin position.ymin))
      (max_int, min_int, max_int)
      ebs in
  let x_center = (xmin + xmax) / 2 in
  let block_size = 20 in
  {
    xmin = x_center - block_size / 2;
    xmax = x_center + block_size / 2;
    ymin = ymin - 2 * block_size;
    ymax = ymin - block_size;
  }
;;

let deep_value_of_natural n =
  Mdl_basics_deep_ast.Deep_int (Unatural.to_string n)
and deep_value_of_int n =
  Mdl_basics_deep_ast.Deep_int (string_of_int n)
;;

(* Build the parameters (that is a binding table) of an input port block
   (an [Inport] Simulink block),
   binding the [Port] ident to the given [port_number],
   and the [PortDimensions] ident to the given [port_dimensions]. *)
let make_inport_block_parameters port_number port_dimensions =
  let table = Hashtbl.create 2 in
  let port_ident =
    Mdl_ast_functions.port_ident in
  let port_binding =
    Mdl_ast_functions.
    make_binding port_ident
      (deep_value_of_natural port_number) in
  Hashtbl.add table
    port_ident port_binding;
  let port_dimensions_ident =
    Mdl_ast_functions.port_dimensions_ident in
  let port_dimensions_binding =
    Mdl_ast_functions.
    make_binding port_dimensions_ident
      (deep_value_of_int port_dimensions) in
  Hashtbl.add table
    port_dimensions_ident port_dimensions_binding;
  table
;;

(** The inherited port dimensions value is conventionally -1. *)
let port_dimensions_inherited = -1;;

(* Build the parameters (a binding table) of an input port block
   (an Inport Simulink block),
   binding its port dimensions to inherited. *)
let inport_block_parameters port_number =
  let port_dimensions = port_dimensions_inherited in
  make_inport_block_parameters port_number port_dimensions
;;

(* Translate a given [poab] activation port block into a regular Simulink
  input port (an Inport Simulink block).
  The resulting input port has the port number given by argument
  [input_port_number]. *)
let transl_port_of_action_block poab input_port_number =
  let poab_body =
    Mdl_explicit_block_utils.get_block_body poab
  and poab_all_bindings =
    Mdl_explicit_block_utils.get_block_all_bindings poab in
  let poab_all_bindings = {
    poab_all_bindings with
    explicit_block_parameters =
      inport_block_parameters input_port_number;
  } in
  {
    poab with
    explicit_block_body = {
      poab_body with
      explicit_block_all_bindings = poab_all_bindings;
    };
    explicit_block_known = true;
    explicit_block_desc =
      Explicit_block_regular {
        explicit_simple_block_type = Mdl_block_types.Inport;
        explicit_simple_block_activation = Simple_block_plain;
      };
  }
;;

(** Find and replace a given [poab] activation port block in a list of
  blocks (presumably the list of blocks of a Simulink sub-system block).
  The activation port block argument is then replaced by an input port block
  with the provided port number.
  Note: the resulting input port block has the same bindings as the original
  [poab] argument. *)
let replace_port_of_action_block_in_blocks
    poab input_port_number ebss_blocks =
  (* Compute new input port block *)
  let input_port_block =
    transl_port_of_action_block poab input_port_number in
  let rec loop accu = function
    | [] -> List.rev accu
    | eb :: ebs ->
      if eb == poab then List.rev_append accu (input_port_block :: ebs)
      else loop (eb :: accu) ebs in
  loop [] ebss_blocks
;;

let remove_port_of_action_block_in_blocks poab
    _input_port_number ebss_blocks =
  (* Remove the ActionPort block *)
  let rec loop accu = function
    | [] -> List.rev accu
    | eb :: ebs ->
      if eb == poab then List.rev_append accu ebs
      else loop (eb :: accu) ebs in
  loop [] ebss_blocks
;;

let change_port_of_action_block_in_ebss
    change_port_of_action_block_in_blocks poab input_port_number ebss =
  let ebss_blocks = Mdl_explicit_block_utils.get_ebss_blocks ebss in
  let ebss_blocks =
    change_port_of_action_block_in_blocks
      poab input_port_number ebss_blocks in
  Mdl_explicit_block_utils.
  set_ebss_blocks ebss ebss_blocks
;;

let replace_port_of_action_block_in_ebss =
  change_port_of_action_block_in_ebss replace_port_of_action_block_in_blocks
;;
let remove_port_of_action_block_in_ebss =
  change_port_of_action_block_in_ebss remove_port_of_action_block_in_blocks
;;

let get_port_of_action_blocks_in_ebss port_nin ebss =
  match ebss.explicit_block_sub_system_activation with
  | Sub_system_plain -> []
  | Sub_system_activated sub_system_activation ->
    match sub_system_activation with
    | Sub_system_plain_atomic_unit _ -> []
    | Sub_system_enabled explicit_block
    | Sub_system_if_action (_, explicit_block)
    | Sub_system_triggered (_, explicit_block) ->
      [ explicit_block, port_nin; ]
    | Sub_system_triggered_and_enabled (
        (_explicit_trigger_type, explicit_block1), explicit_block2) ->
      [ explicit_block1, Unatural.pred port_nin;
        explicit_block2, port_nin; ]
;;

let change_activated_block_ebss ab ebss =
  let actioned_block = ab.actioned_block in
  let boa = actioned_block.block_of_action in
  let boa =
    Mdl_explicit_block_utils.
    set_block_ebss boa ebss in
  let boa_kind = Boa_sub_system ebss in
  Active {
    ab with
    actioned_block = {
      actioned_block with
      block_of_action = boa;
    };
    boa_kind = boa_kind;
  }
;;

let transl_port_of_action_block_to_equiped_block equiped_block =
  match equiped_block with
  | Mdl_equiped_block.
    Passive _ -> equiped_block
  | Active ({
      actioned_block = {
        block_of_action = boa;
        event_generator = _egb;
      };
      activated_kind = activated_kind;
      boa_kind = boa_kind;
    } as ab) ->
    match boa_kind with
    | Boa_simple -> equiped_block
    | Boa_sub_system ebss ->
      let regular_port_nin =
        Mdl_explicit_block_utils.get_block_regular_port_nin boa in
      let poab_regular_ports_numbers =
        get_port_of_action_blocks_in_ebss regular_port_nin ebss in
      (* No poab can be found: just return original equiped block. *)
      if poab_regular_ports_numbers = [] then equiped_block else
      let change_port_of_action_block_in_ebss =
        match activated_kind with
        | No_link ->
          (* If the block is an IfAction block, then we must remove from the
             sub-system blocks the [ActionPort] block that qualified the
             sub-system as an action block. *)
          remove_port_of_action_block_in_ebss
        | One_link | Two_link ->
          replace_port_of_action_block_in_ebss in
      let ebss =
        List.fold_right
          (fun (poab, regular_ports_number) ebss ->
           change_port_of_action_block_in_ebss
             poab regular_ports_number ebss)
          poab_regular_ports_numbers ebss in
      change_activated_block_ebss ab ebss
;;

(**
  Add a fake Simulink AtomicPort block to the [ebss] list of blocks of a
  sub-system block of action [eb], in order to activate the Scicos block
  translation of the given sub-system block of action.
*)
let add_ebss_event_input_port_block ebss =

  let blocks = Mdl_explicit_block_utils.get_ebss_blocks ebss in
  let atomic_port_block_position = compute_atomic_port_position blocks in
  let atomic_port_block_name =
    Ast_types.mdl_name_of_string "AtomicPort" in

  let atomic_port_block =
    Mdl_make_block.
    make_atomic_port_block
      atomic_port_block_position
      None
      atomic_port_block_name in

  Mdl_explicit_block_utils.
  add_ebss_block ebss atomic_port_block
;;

let add_equiped_block_event_input_port_block equiped_block =
  match equiped_block with
  | Mdl_equiped_block.
    Passive _ -> equiped_block
  | Active ({
      actioned_block = ({
        block_of_action = boa;
        event_generator = _egb;
      } as actioned_block);
      activated_kind = _;
      boa_kind = boa_kind;
    } as ab) ->
    match boa_kind with
      (* No need to add an event_input port block *)
    | Boa_simple -> equiped_block
      (* Need to add an event_input port block to the sub-system *)
    | Boa_sub_system ebss ->
      let ebss = add_ebss_event_input_port_block ebss in
      let boa =
        Mdl_explicit_block_utils.
        set_block_ebss boa ebss in
      let boa_kind = Boa_sub_system ebss in
      Active {
        ab with
        actioned_block = {
          actioned_block with
          block_of_action = boa;
        };
        boa_kind = boa_kind;
      }
;;

(** Reroute the target of a link to a Simulink special port
  (Enable/Trigger/If_action) of a block of action to the corresponding input
  port of the block of action. *)
let reroute_target_activation_port activation_port =
  let port_number =
    match activation_port with
    | Enable_port -> 1
    | Trigger_port -> 1
    | Enable_trigger_port -> 2
    | Trigger_enable_port -> 1
    | If_action_port -> 1 in
  Ast_types.mdl_port_number_of_int port_number
;;

(**
  Redirect split link connectors to action ports.
  If link is targeted to block of action sub-system [boa]
  it is redirected to the event generator block [egb].
*)
let reroute_split_link activated_table split_link =
  match split_link with
  | {
      Mdl_split_ast.
      split_link_name = _;
      split_link_font = _;
      split_link_labels = _;
      split_link_points = _;
      split_link_desc = desc;
    } ->
    match desc with
    | Split_connection_link_desc split_connection_link_connectors ->
      (* Nothing to do: there is no connection line to action ports.
         Simply return the connection line as a connection link. *)
(*
      begin match
        split_connection_link_connectors.
          split_link_dst with
      | {
          explicit_connector_port_block = block_name;
          explicit_connector_port_desc =
            Split_directed_connector_desc directed_connector_desc
        } ->
*)
      let action_link_desc =
        let action_connection_connectors =
          split_connection_link_connectors in
        Action_connection_link action_connection_connectors in
     {
       split_link with
       Mdl_split_ast.
       split_link_desc = action_link_desc;
     }

    | Split_directed_link_desc split_directed_link_connectors ->
      match
        split_directed_link_connectors.
          split_link_dst with
      | {
          explicit_connector_port_block = _;
          explicit_connector_port_desc =
            Split_directed_connector_desc directed_connector_desc
        } ->
        let split_link_dst =
          {
            split_directed_link_connectors.split_link_dst with
            explicit_connector_port_desc = directed_connector_desc;
          } in
        let action_directed_connectors =
          {
            split_directed_link_connectors with
            Mdl_split_ast.
            split_link_dst = split_link_dst;
          } in
        let action_link_desc =
          Action_directed_link action_directed_connectors in
        {
          split_link with
          Mdl_split_ast.
          split_link_desc = action_link_desc;
        }
      | {
          explicit_connector_port_block = boa_block_name;
          explicit_connector_port_desc =
            Split_activation_connector_desc action_connector_desc
        } ->
        (* [boa_block_name] is the block of action.
           [egb_block_name] is the event generator block
           associated to [boa_block_name].
           Reroute the link with destination the activation port of
           [boa_block_name] to the corresponding block port number 1 or 2
           of [egb_block_name] (according to the action port
           (i.e. the activation mode)). *)
        let egb_block_name =
          if not (Hashtbl.mem activated_table boa_block_name) then
            Say.fatal_error (Printf.sprintf
              "failed to find event generator block associated \
               with block of action %S"
              (string_of_block_name boa_block_name)) else
            Hashtbl.find activated_table boa_block_name in
        let dst_link_connector =
          let dst_connector_desc =
            reroute_target_activation_port action_connector_desc in
          {
            explicit_connector_port_block = egb_block_name;
            explicit_connector_port_desc = dst_connector_desc;
          } in
        let action_directed_connectors =
          {
            split_directed_link_connectors with
            Mdl_split_ast.
            split_link_dst = dst_link_connector;
          } in
        let action_link_desc =
          Action_directed_link action_directed_connectors in
        {
          split_link with
          Mdl_split_ast.
          split_link_desc = action_link_desc;
        }
;;

let reroute_split_links activated_table split_links =
  (* List.rev_map *)
  List.map (reroute_split_link activated_table) split_links
;;

let equip_split_blocks blks =
  List.rev_map Mdl_equip_block.equip_block blks
;;

let collect_split_blocks equiped_blks =
  List.fold_left
    (fun blks equiped_block ->
     match equiped_block with
     | Mdl_equiped_block.Passive eb ->
       (eb :: blks)
     | Active {
         actioned_block = {
           block_of_action = boa;
           event_generator = egb;
         };
         activated_kind = _;
         boa_kind = _;
       } ->
       boa :: egb :: blks)
    [] equiped_blks
;;

type activated_tables = {
  event_generator_table :
    (Mdl_ast.block_name, Mdl_ast.block_name) Hashtbl.t;
  activated_block_table :
    (Mdl_ast.block_name, Mdl_equiped_block.activated_block) Hashtbl.t;
}
;;

let activated_tables_of_activated_blocks equiped_blocks =
  let boa_egb_tbl = Hashtbl.create 10
  and boa_ab_tbl = Hashtbl.create 10 in
  List.iter
    (function
     | Mdl_equiped_block.Passive _ -> ()
     | Active ({
         Mdl_equiped_block.
         actioned_block = {
           block_of_action = boa;
           event_generator = egb;
         };
         activated_kind = _activated_kind;
         boa_kind = _boa_kind;
       } as ab) ->
       let boa_name = boa.explicit_block_name
       and egb_name = egb.explicit_block_name in
       Say.fdebug (fun ppf ->
         Format.fprintf ppf
           "Adding association from %S to %S\n"
           (string_of_block_name boa_name)
           (string_of_block_name egb_name));
       Hashtbl.add boa_egb_tbl boa_name egb_name;
       Hashtbl.add boa_ab_tbl boa_name ab)
    equiped_blocks;
  {
    event_generator_table = boa_egb_tbl;
    activated_block_table = boa_ab_tbl;
  }
;;

let rec make_action_blocks split_links split_blocks =
  let equiped_blocks = equip_split_blocks split_blocks in
  let activated_tables =
    activated_tables_of_activated_blocks equiped_blocks in
  (* Association of boa_name to their event generator name *)
  let event_generator_table = activated_tables.event_generator_table
  (* Association of boa_name to the corresponding [activated_block] value *)
  and activated_block_table = activated_tables.activated_block_table in
  (* Redirect target of links to action ports of Simulink block of action
     to corresponding event generator block of the activated block of
     the Simulink block of action. *)
  let action_links =
    reroute_split_links event_generator_table split_links in

  (* Add links from event generator blocks
     to corresponding translated block of action of each activated block. *)

  (* For each activated block:
     - add an event link from the event generator block to
       the block of action block,
     - if needed, add regular links between out ports of the event generator
       block to the newly added input ports of the translated block of
       action. *)
  let add_activated_links boa_block_name ab action_links =
    create_activated_event_link boa_block_name ab ::
    create_activated_regular_links boa_block_name ab action_links in
  let action_links =
    Hashtbl.fold add_activated_links activated_block_table action_links in

  (* Add a new event input port block for all sub-systems block of action *)
  let equiped_blocks =
    List.rev_map add_equiped_block_event_input_port_block equiped_blocks in

  (* Translate activation port blocks into regular input port blocks
    for all sub-systems block of action *)
  let equiped_blocks =
    List.rev_map
      transl_port_of_action_block_to_equiped_block equiped_blocks in

  (* Collect all blocks out of equiped blocks *)
  let split_blocks = collect_split_blocks equiped_blocks in

  (* Recurse in sub-systems to turn out all split blocks to action blocks. *)
  let action_blocks =
    List.rev_map make_action_block split_blocks in

  action_links, action_blocks

and make_action_block
  (split_block : Mdl_split_ast.split_block) :
  Mdl_action_ast.action_block =
  match split_block with
  | {
      explicit_block_desc =
        (Explicit_block_regular _ |
         Explicit_block_reference _ |
         Explicit_block_s_function _);
      _;
    } as blk -> blk
  | {
      explicit_block_desc =
        Explicit_block_sub_system explicit_block_sub_system;
      _;
    } as blk ->
    let action_block_desc =
      let action_block_sub_system =
        make_action_block_sub_system explicit_block_sub_system in
      Explicit_block_sub_system action_block_sub_system in
    {
      blk with
      explicit_block_desc = action_block_desc;
    }

and make_action_block_sub_system ebss_desc =
  let ebss_system = ebss_desc.explicit_block_sub_system_body in
  let action_system = make_action_system ebss_system in
  (* FIXME: If the sub-system was a block of action then,
    now that it is equipped, should it still be considered as being a block
    of action ???
    (for the time being we change the field to plain i.e. Sub_system_plain)
    FIXME: What to to with field
    ebss_desc.explicit_block_sub_system_treat_as_atomic_unit ???
    should we force treat_as_atomic_unit to be something ???
  *)
  let action_activation =
    make_action_activation
      ebss_desc.explicit_block_sub_system_activation in
  {
    ebss_desc with
    explicit_block_sub_system_activation = action_activation;
    explicit_block_sub_system_body = action_system;
  }

and make_action_activation ea =
(* FIXME: should compute the relevant explicit_activation
   Or we can skip sub_system_activation field since the
   sub_system_activation field may not be needed after this phase ??? *)
  match ea with
  | Sub_system_plain -> Sub_system_plain
  | Sub_system_activated _sub_system_activation ->
    Sub_system_plain

and make_action_system
  (split_system : Mdl_split_ast.split_system) :
  Mdl_action_ast.action_system =
  let split_links =
    split_system.explicit_system_lines
  and split_blocks =
    split_system.explicit_system_blocks in

  let action_links, action_blocks =
    make_action_blocks split_links split_blocks in

  {
    split_system with
    explicit_system_blocks = action_blocks;
    explicit_system_lines = action_links;
  }
;;

let action_model_body split_model_body =
  {
    split_model_body with
    explicit_model_system =
      make_action_system
        split_model_body.explicit_model_system;
  }
;;

let action_model split_model =
  {
    split_model with
    explicit_model_body =
      action_model_body split_model.explicit_model_body;
  }
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
