(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_action_ast.mli,v 1.1 2019-05-10 14:07:13 jpc Exp $ *)

(** {3 Action phase} *)

(** From [Mld_split_ast] syntax to [Mdl_action_ast] syntax.

  From split link kind models to action link kind models, i.e.
  from models using Scicos-style lines featuring split blocks instead of line
  branches ([Mdl_action_ast] syntax) to models using almost true Scicos links
  with directed and event links ([Mdl_action_ast] syntax).

*)

(** {6 Action link kind models} *)

(**
  A model with action link connections is an [action_model] (as defined
  below).

  An [action_model] is the result of handling Simulink activation block and
  Simulink activation lines of a Simulink model.

  Activation links are turned into devoted 'event links' and activation
  blocks disappear from the model: activation blocks become special purpose
  event generator blocks.

  Event links are particular directed links. Directed links are made of
  directed connectors (as defined below as a [directed_connector
  poly_explicit_link]), either binding regular port to port connector or
  event connector.

  There are no more activation port blocks in an [action_model]: all
  activation port blocks (EnablePort, TriggerPort, ...) have been replaced by
  regular input ports. In short, Simulink activation mechanism for blocks is
  translated via extra external blocks and links. External blocks are event
  generators activating blocks through event links (in practice, event
  generators are introduced in the model as fake Simulink blocks).

*)

(** {6 Action type definitions} *)

(** {7 Borrowed from explicit phase} *)

type 'a explicit_block = 'a Mdl_implementation_explicit_ast.explicit_block
and 'a explicit_system = 'a Mdl_implementation_explicit_ast.explicit_system
and 'a explicit_model = 'a Mdl_implementation_explicit_ast.explicit_model
and 'a explicit_simulink_file = 'a Mdl_explicit_ast.explicit_simulink_file
;;

(** {7 Borrowed from split phase} *)

type ('a, 'b) split_link_connectors =
  ('a, 'b) Mdl_split_ast.split_link_connectors
;;

(**
  After the action phase, ports are limited to simple numbers:
  - no more Port_trigger, Port_enable, ..., that have been suppressed by event
    generator and event links;
  - nor Port_lconn, Port_lconn that have been reserved to connection links.
*)
type directed_connector_desc = Mdl_split_ast.directed_connector_desc
  (* = port_number
     = Ast_types.mdl_port_number
     = private Mdl_port_number of Unatural.t
  *)
;;

type 'a split_connector_port = 'a Mdl_split_ast.split_connector_port;;

type 'a poly_split_link = 'a Mdl_split_ast.poly_split_link;;

(** {7 Defining action links} *)

type 'a action_connector_port = 'a split_connector_port;;

type 'link_desc poly_action_link = 'link_desc poly_split_link;;

type 'a action_link_connectors = ('a, 'a) split_link_connectors
(** After action phase source and target connectors get the same types. *)
;;

type action_directed_connector_desc = directed_connector_desc;;
type action_event_connector_desc = action_directed_connector_desc
and action_connection_connector_desc =
  Mdl_split_ast.split_connection_connector_desc
;;

(** Abbreviation only useful for translation phase (instructions.mli) *)
type action_directed_connector =
  action_directed_connector_desc action_connector_port
;;
type action_connection_connector =
  action_connection_connector_desc action_connector_port
;;

type action_directed_connectors =
  action_directed_connector_desc action_link_connectors
;;

type action_event_connectors =
  action_event_connector_desc action_link_connectors
;;

type action_connection_connectors =
  action_connection_connector_desc action_link_connectors
;;

type action_link_desc =
  | Action_directed_link of action_directed_connectors
    (* The ancestor of a regular Scicos link
       (unfortunately this is called an 'explicit' link in Scicos):
       the result of the analysis of Simulink lines
       after explicitation of line branches using
       Scicos split blocks and Scicos links. *)
  | Action_event_link of action_event_connectors
    (* The ancestor of a Scicos event link:
       the result of the analysis of Simulink action lines,
       found as source of output ports of Simulink If and Switch blocks
       and as target of Simulink action sub-systems having
       ActionPort or TriggerPort or EnablePort. *)
  | Action_connection_link of action_connection_connectors
    (* The ancestor of a Scicos connection link
       (unfortunately this is called 'implicit link in Scicos):
       the result of the analysis of Simulink connection lines,
       found as source of output or input ports of Simulink Stateflow
       blocks. *)
;;

type action_links = action_link list
and action_link = action_link_desc poly_action_link
;;

(** {7 Defining action models} *)

type action_block = action_link explicit_block
and action_system = action_link explicit_system
and action_model = action_link explicit_model
and action_simulink_file = action_link explicit_simulink_file
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
