(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_equip_block.ml,v 1.1 2019-05-10 14:07:13 jpc Exp $ *)

(* From explicit syntax with Scicos-style links with
  splitting blocks (mld_split_explicit_syntax) to true Scicos links
  (mdl_action_explicit_syntax). *)

open Mdl_basics_deep_ast
open Mdl_implementation_explicit_ast
open Mdl_equiped_block
;;

type activated_block_positions = {
  block_of_action_position : Mdl_basics_deep_ast.deep_position;
  event_generator_position : deep_position;
}
;;

(**

  Given [boa_position], the position of a Simulink block of action, i.e. a
  block that needs an additional event generator block,
  [compute_activated_block_positions] computes the geometry of the translated
  block and the position of its additional event generator block.

  The two blocks together occupy the same geometrical area of the diagram as
  the original block of action.

  The width of the original block of action is not modified but its height is
  slightly shrinked to give room to the additional event generator block.
  The event generator block is centered at the north of the original active
  block. The translation is suggested as the following ASCII art:

  Original block of action     Translation of the block into two blocks

  |-----------------|          |-----------------|
  |                 |          | Event generator |
  |                 |          |-----------------|
  |                 |          |-----------------|
  | Block of action |          |                 |
  |                 |          | Block of action |
  |                 |          |  (translated)   |
  |                 |          |                 |
  |-----------------|          |-----------------|

*)
let compute_activated_block_positions boa_position =

  let boa_height = boa_position.ymax - boa_position.ymin in
  let vertical_spacing = boa_height / 10 in
  let event_generator_height = max (boa_height / 5) 5 in

  (* Compute the position of the event generator block. *)
  let event_generator_position =
    let ymin = boa_position.ymin in
    let ymax = ymin + event_generator_height in
    let xmin = boa_position.xmin in
    let xmax = boa_position.xmax in
    { xmin = xmin; xmax = xmax; ymin = ymin; ymax = ymax; } in

  (* The new position of the translation of the original Simulink block of
     action: slighly lower than before to give room to the event generator
     block.
     Namely, remove the event generator block height and the vertical
     spacing for blocks to the original block [ymax] maximum extension.
     Other than [ymax], the position of the original block is unchanged. *)
  let boa_position =
    { boa_position with
      ymin =
        boa_position.ymin +
        vertical_spacing + event_generator_height;
    } in

  {
    block_of_action_position = boa_position;
    event_generator_position = event_generator_position;
  }
;;

(**
   Given:

   - A Simulink block of action [boa] that needs an event generator block.

   We build the new event generator block and a new block of action
   by modifying [boa] to reflect the new geometry of [boa] (since the
   modified block [boa] and its event generator should occupy the same room
   as the original block of action).

   Return an activated block, i.e. the new block of action and the newly
   created event generator block.
*)

let get_simple_block_activated_kind explicit_simple_block_desc =
  match explicit_simple_block_desc.explicit_simple_block_activation with
  | Simple_block_plain -> No_link
  | Simple_block_activated simple_block_activation ->
    match simple_block_activation with
    | Simple_block_if_action -> No_link
    | Simple_block_enabled
    | Simple_block_triggered -> One_link
    | Simple_block_triggered_and_enabled -> Two_link
;;

let activated_block_of_block_of_action boa =
  (* Find the position of argument [boa]. *)
  let boa_body = Mdl_explicit_block_utils.get_block_body boa in
  let boa_all_bindings = boa_body.explicit_block_all_bindings in
  let boa_graphical = boa_all_bindings.explicit_block_graphical in
  let boa_position = boa_graphical.explicit_block_position in

  let {
        block_of_action_position = boa_position;
        event_generator_position = egb_position;
      } =
    compute_activated_block_positions boa_position in

  let boa =
    Mdl_explicit_block_utils.set_block_position boa boa_position in

  let actioned_block =
    Mdl_event_generator.make_actioned_block boa egb_position in

  let boa_kind, activated_kind =
    match boa.explicit_block_desc with
    | Explicit_block_regular explicit_simple_block_desc ->
      Boa_simple, get_simple_block_activated_kind explicit_simple_block_desc
    | Explicit_block_reference explicit_simple_block_desc ->
      Boa_simple, get_simple_block_activated_kind explicit_simple_block_desc
    | Explicit_block_s_function _ ->
      Boa_simple, No_link
   | Explicit_block_sub_system ebss ->
     let boa_kind = Boa_sub_system ebss in
     let activated_kind =
       match ebss.explicit_block_sub_system_activation with
       | Sub_system_plain -> No_link
       | Sub_system_activated sub_system_activation ->
         match sub_system_activation with
         | Sub_system_if_action _
         | Sub_system_plain_atomic_unit _ -> No_link
         | Sub_system_enabled _
         | Sub_system_triggered _ -> One_link
         | Sub_system_triggered_and_enabled _ -> Two_link in
     boa_kind, activated_kind in

  {
    Mdl_equiped_block.
    actioned_block = actioned_block;
    activated_kind = activated_kind;
    boa_kind = boa_kind;
  }
;;

let equip_block_sub_system eb ebss =
  match ebss.explicit_block_sub_system_activation with
  | Sub_system_plain ->
    (* Nothing to do: it is a regular sub-system block *)
    Passive eb
  | Sub_system_activated _sub_system_activation ->
    (* This is a block of action, it must be equipped *)
    Active (activated_block_of_block_of_action eb)
;;

(**
  Computes the [equiped_block] value corresponding to reference block [eb]
  with [simple_block_desc] descriptor value [eb_desc].
*)
let equip_block_reference eb eb_desc =
  match eb_desc.explicit_simple_block_activation with
  | Simple_block_plain ->
    (* Nothing to do: it is a regular simple block *)
    Passive eb
  | Simple_block_activated _activation ->
    (* This is a block of action, it must be equipped *)
    Active (activated_block_of_block_of_action eb)
;;

(** Regular blocks are treated the same as reference blocks *)
let equip_block_regular = equip_block_reference
;;

let equip_block_s_function eb = Passive eb
;;

(* Given block [eb], equip_block builds the [equiped_block] value
   corresponding to [eb].

   - if [eb] is not a block of action, the resulting [equiped_block] value is
   simply Passive [eb].

   - otherwise, [eb] is a block of action and the resulting [equiped_block]
   value is Active [ab],
     where [ab] is an [activated_block] value, i.e. a modified version of block
     [eb] to gether with an event generator corresponding to the kind of action
     required for [eb]. *)
let equip_block eb =
  match eb.explicit_block_desc with
    (* We may have to equip this block *)
  | Explicit_block_regular eb_desc -> equip_block_regular eb eb_desc
    (* We may have to equip this block *)
  | Explicit_block_reference eb_desc -> equip_block_reference eb eb_desc
  | Explicit_block_s_function _ -> equip_block_s_function eb
  | Explicit_block_sub_system eb_desc -> equip_block_sub_system eb eb_desc
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
