(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_equiped_block.mli,v 1.1 2019-05-10 14:07:13 jpc Exp $ *)

(** {3 Equiped block definition} *)

(**
  An equiped block is a Simulink explicit block that is either

  - passive: the block is not a block of action

  - active: the block is a block of action and it should be translated into
    two blocks: the translation of the original block and an event generator
    block according to the kind of activation event appropriate to the block
    of action.

  An active equiped block is an activated block that specifies the kind of
  block of action and an actioned block, i.e. the pair of blocks that
  will implement in Scicos the Simulink block of action.

*)

open Mdl_split_ast;;

(** Activated blocks kinds: an activated block is either:
  - sample: there is no output to its event generator block
      ([No_link] case),
  - single: there is one output to its event generator block
      ([One_link] case),
  - double: there are two outputs to its event generator block
      ([Two_link] case).
*)

type activated_kind =
  | No_link
    (* No need to add links to the equiped block *)
  | One_link
    (* Need to add one link from event generator to translated block of action *)
  | Two_link
    (* Need to add two links from event generator to translated block of action *)
;;

(** Block of action kinds: a block of action is either a simple block or a
  sub-system. *)

type boa_kind =
  | Boa_simple
    (* The boa is not a sub-system *)
  | Boa_sub_system of
      split_link
      Mdl_implementation_explicit_ast.explicit_block_sub_system_desc
    (* The boa is a sub-system, and here is its descriptor *)
;;

type actioned_block = {
  block_of_action : Mdl_split_ast.split_block;
  event_generator : Mdl_split_ast.split_block;
}
(** An actioned block is the pair of a block of action with its
  associated event generator block. *)
;;


type activated_block = {
  actioned_block : actioned_block;
  activated_kind : activated_kind;
  boa_kind : boa_kind;
}
(** An activated block is an actioned block with all its associated properties:
  - is the block simple or is it a sub-system ?
  - does the block event generator needs extra links and how many ?
*)
;;

type equiped_block =
  | Passive of split_block
  | Active of activated_block
(** After equipment a block is classified either as
  - a passive block that needs no further action treatment
  - an active block that needs further modifications for links
    and internal blocks.
*)
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
