(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_event_generator.ml,v 1.1 2019-05-10 14:07:13 jpc Exp $ *)

(** {3 Creating event generator blocks for Simulink blocks of action} *)

open Mdl_implementation_explicit_ast
open Mdl_explicit_block_utils
;;

type event_generator_desc =
  | Eg_sample of Mdl_deep_ast.deep_value
  | Eg_enable
  | Eg_if_action of Mdl_implementation_explicit_ast.explicit_action_type
  | Eg_trigger of Mdl_implementation_explicit_ast.explicit_trigger_type
  | Eg_trigger_and_enable of explicit_trigger_type
;;

let get_event_generator_reference_type event_generator_desc =
  match event_generator_desc with
  | Eg_sample _deep_value ->
    Mdl_fake_block_types.atomic_event_generator_block_reference_type
  | Eg_enable ->
    Mdl_fake_block_types.enable_event_generator_block_reference_type
  | Eg_if_action _action_type ->
    Mdl_fake_block_types.if_action_event_generator_block_reference_type
  | Eg_trigger trigger_type ->
    begin match trigger_type with
    | Trigger_either ->
      Mdl_fake_block_types.
      trigger_either_event_generator_block_reference_type
    | Trigger_falling ->
      Mdl_fake_block_types.
      trigger_falling_event_generator_block_reference_type
    | Trigger_function_call ->
      Mdl_fake_block_types.
      trigger_function_call_generator_block_reference_type
    | Trigger_rising ->
      Mdl_fake_block_types.
      trigger_rising_event_generator_block_reference_type
    | Trigger_rising_edge
    | Trigger_unset ->
      Say.warning
        "Cannot translate \'rising edge\', \'function call\', or \'unset\'\
       \ntrigger types. Using rising trigger type instead.";
      Mdl_fake_block_types.
      trigger_rising_event_generator_block_reference_type
    end
  | Eg_trigger_and_enable trigger_type ->
    begin match trigger_type with
    | Trigger_either ->
      Mdl_fake_block_types.
      trigger_and_enable_either_event_generator_block_reference_type
    | Trigger_falling ->
      Mdl_fake_block_types.
      trigger_and_enable_falling_event_generator_block_reference_type
    | Trigger_function_call ->
      Mdl_fake_block_types.
      trigger_and_enable_function_call_generator_block_reference_type
    | Trigger_rising ->
      Mdl_fake_block_types.
      trigger_and_enable_rising_event_generator_block_reference_type
    | Trigger_rising_edge
    | Trigger_unset ->
      Say.warning
        "Cannot translate \'rising edge\', \'function call\', or \'unset\'\
       \ntrigger types. Using rising trigger type instead.";
      Mdl_fake_block_types.
      trigger_and_enable_rising_event_generator_block_reference_type
    end
;;

let regular_ports_number_of_egb event_generator_desc =
  match event_generator_desc with
  | Eg_sample _deep_value ->
    { Mdl_basics_deep_ast.port_nin = Unatural._0_; port_nout = Unatural._0_ }
  | Eg_enable ->
    { Mdl_basics_deep_ast.port_nin = Unatural._1_; port_nout = Unatural._1_ }
  | Eg_if_action _action_type ->
    { Mdl_basics_deep_ast.port_nin = Unatural._1_; port_nout = Unatural._0_ }
  | Eg_trigger _trigger_type ->
    { Mdl_basics_deep_ast.port_nin = Unatural._1_; port_nout = Unatural._1_ }
  | Eg_trigger_and_enable _trigger_type ->
    { Mdl_basics_deep_ast.port_nin = Unatural._2_; port_nout = Unatural._2_; }
;;

let eg_if_action_unset = Eg_if_action Action_unset
and eg_trigger_unset = Eg_trigger Trigger_unset
and eg_trigger_and_enable_unset = Eg_trigger_and_enable Trigger_unset
;;

let make_event_generator_desc_option eb =
  let event_generator_desc_of_simple_block eb_desc =
    begin match eb_desc.explicit_simple_block_activation with
    | Simple_block_plain -> None
    | Simple_block_activated simple_block_activation ->
      match simple_block_activation with
      | Simple_block_enabled -> Some Eg_enable
      | Simple_block_if_action -> Some eg_if_action_unset
      | Simple_block_triggered -> Some eg_trigger_unset
      | Simple_block_triggered_and_enabled -> Some eg_trigger_and_enable_unset
    end in
  match eb.explicit_block_desc with
  | Explicit_block_regular eb_desc ->
    event_generator_desc_of_simple_block eb_desc
  | Explicit_block_reference eb_desc ->
    event_generator_desc_of_simple_block eb_desc
  | Explicit_block_s_function _ ->
    None
  | Explicit_block_sub_system explicit_block_sub_system_desc ->
    match
      explicit_block_sub_system_desc.
      explicit_block_sub_system_activation
    with
    | Sub_system_plain -> None
    | Sub_system_activated sub_system_activation ->
      match sub_system_activation with
      | Sub_system_plain_atomic_unit sample ->
        Some (Eg_sample sample)
      | Sub_system_enabled _ ->
        Some Eg_enable
      | Sub_system_if_action (action_type, _) ->
        Some (Eg_if_action action_type)
      | Sub_system_triggered (trigger_type, _) ->
        Some (Eg_trigger trigger_type)
      | Sub_system_triggered_and_enabled ((trigger_type, _), _) ->
        Some (Eg_trigger_and_enable trigger_type)
;;

let fix_block_regular_port_nin boa egb_nout =
  let regular_port_nin =
    Unatural.add (get_block_regular_port_nin boa) egb_nout in
  Mdl_explicit_block_utils.
  set_block_regular_port_nin boa regular_port_nin
;;

(** Invert the mirror property argument. *)
let invert_mirror
    (mirror : Mdl_basics_explicit_ast.explicit_mirror) :
    Mdl_basics_explicit_ast.explicit_mirror =
  not mirror
;;

let hashtbl_copy tbl =
  let res = Hashtbl.create (Hashtbl.length tbl) in
  Hashtbl.iter (Hashtbl.add res) tbl;
  res
;;

let get_ebss_parameters_of_ports_of_action ebss =
  match ebss.explicit_block_sub_system_activation with
  | Sub_system_plain ->
    Say.fatal_error
       "Mdl_event_generator: \
        Plain sub-system block should not be equiped with an event generator"
  | Sub_system_activated sub_system_activation ->
    match sub_system_activation with
    | Sub_system_plain_atomic_unit system_sample_time ->
      let block_parameters =
        let table = Hashtbl.create 1 in
        let system_sample_time_ident =
          Mdl_ast_functions.system_sample_time_ident in
        let system_sample_time_binding =
          Mdl_ast_functions.
          make_binding system_sample_time_ident system_sample_time in
        Hashtbl.add table
          system_sample_time_ident system_sample_time_binding;
        table in
      block_parameters
    | Sub_system_enabled explicit_block
    | Sub_system_if_action (_, explicit_block)
    | Sub_system_triggered (_, explicit_block) ->
      Mdl_explicit_block_utils.get_block_parameters explicit_block
    | Sub_system_triggered_and_enabled (
        (_explicit_trigger_type, explicit_block1), explicit_block2) ->
      let block_parameters =
        hashtbl_copy (get_block_parameters explicit_block2) in
      Hashtbl.iter
        (Hashtbl.add block_parameters)
        (get_block_parameters explicit_block1);
      block_parameters
;;

let get_boa_event_generator_parameters boa =
  match boa.explicit_block_desc with
    (* No parameters to get for the event generator block *)
  | Explicit_block_regular _
  | Explicit_block_reference _
  | Explicit_block_s_function _ ->
    Mdl_make_block.empty_block_parameters
    (* Get parameters of sub-system port of action as
       parameters for the event generator block *)
  | Explicit_block_sub_system ebss ->
    get_ebss_parameters_of_ports_of_action ebss
;;

let string_of_action_type action_type =
  match action_type with
  | Action_case ->
    "ActionCase"
  | Action_default ->
    "ActionDefault"
  | Action_else ->
    "ActionElse"
  | Action_elseif ->
    "ActionElseif"
  | Action_then ->
    "ActionThen"
  | Action_unset ->
    "ActionUnset"
;;

let string_of_trigger_type trigger_type =
  match trigger_type with
  | Trigger_either ->
    "TriggerEither"
  | Trigger_falling ->
    "TriggerFalling"
  | Trigger_function_call ->
    "TriggerFunctionCall"
  | Trigger_rising ->
    "TriggerRising"
  | Trigger_rising_edge ->
    "TriggerRisingEdge"
  | Trigger_unset ->
    "TriggerUnset"
;;

let prefix_of_event_generator_desc egb_desc =
  match egb_desc with
  | Eg_sample sample_time ->
    Printf.sprintf "Sample(%s)"
      (Mdl_deep_basics.string_of_deep_value
         "sample time event generator" sample_time)
  | Eg_enable -> "Enable"
  | Eg_if_action action_type ->
    Printf.sprintf "IfAction(%s)" (string_of_action_type action_type)
  | Eg_trigger trigger_type ->
    Printf.sprintf "Trigger(%s)" (string_of_trigger_type trigger_type)
  | Eg_trigger_and_enable trigger_type ->
    Printf.sprintf "EnableAndTrigger(%s)" (string_of_trigger_type trigger_type)
;;

let get_boa_event_generator_name boa_name egb_desc =
  Ast_types.mdl_name_of_string
    (Printf.sprintf "%s_%s"
       (prefix_of_event_generator_desc egb_desc) boa_name)
;;

let make_actioned_block boa egb_position =
  let event_generator_desc =
    match make_event_generator_desc_option boa with
    | None ->
      Say.fatal_error
        "Trying to get an event generator block \
         for a block which is not a block of action"
    | Some eg_desc -> eg_desc in
  let egb_ty = get_event_generator_reference_type event_generator_desc in
  let egb_parameters = get_boa_event_generator_parameters boa in
  let egb_mirror = invert_mirror (get_block_mirror boa) in
  let egb_name =
    get_boa_event_generator_name
      (Mdl_explicit_block_utils.get_string_block_name boa)
      event_generator_desc in
  let egb_regular_ports_number =
    regular_ports_number_of_egb event_generator_desc in
  let egb =
    Mdl_make_block.make_event_generator_block
     egb_ty
     egb_parameters
     egb_mirror egb_position
     egb_name
     egb_regular_ports_number in
  let egb_nout = egb_regular_ports_number.Mdl_basics_deep_ast.port_nout in
  Say.debug (Printf.sprintf
    "Before fixing, port_nin is %d"
    (Unatural.to_int
      (Mdl_explicit_block_utils.get_block_regular_port_nin boa)));
  let boa = fix_block_regular_port_nin boa egb_nout in
  Say.debug (Printf.sprintf
    "After fixing (adding %d), port_nin is %d"
    (Unatural.to_int egb_nout)
    (Unatural.to_int
      (Mdl_explicit_block_utils.get_block_regular_port_nin boa)));
  {
    Mdl_equiped_block.
    block_of_action = boa;
    event_generator = egb;
  }
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
