(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_explicit_block_utils.ml,v 1.1 2019-05-10 14:07:13 jpc Exp $ *)

(** {3 Access and setting functions for various explicit block fields} *)

open Mdl_implementation_explicit_ast
;;

let string_of_block_name = Mdl_ast_to_deep_ast.string_of_block_name;;

(** {Accessing fields of explicit blocks} *)
let get_block_name eb = eb.explicit_block_name;;

let get_string_block_name eb = string_of_block_name (get_block_name eb);;

let get_block_body eb = eb.explicit_block_body;;

let get_block_all_bindings eb =
  (get_block_body eb).explicit_block_all_bindings
;;

let get_block_parameters eb =
  (get_block_all_bindings eb).explicit_block_parameters
;;

let get_block_parameter_value eb id =
  let block_parameters = get_block_parameters eb in
  try
    (Hashtbl.find block_parameters id).
    Binding.bound_value with
  | Not_found ->
    Say.fatal_error (Printf.sprintf
      "missing parameter %s in block %s"
      (Ast_types.string_of_ident id)
      (get_string_block_name eb)
    )
;;

let get_block_graphical eb =
  (get_block_all_bindings eb).explicit_block_graphical
;;

let get_block_mirror eb =
  (get_block_graphical eb).explicit_block_mirror
;;

let get_block_rotation eb =
  (get_block_graphical eb).explicit_block_rotation
;;

let get_block_ports_description eb =
  (get_block_graphical eb).explicit_block_ports_description
;;

let get_block_regular_port_nin eb =
  Mdl_ast_to_deep_ast.get_regular_port_nin (get_block_ports_description eb)
;;

let get_block_regular_port_nout eb =
  Mdl_ast_to_deep_ast.get_regular_port_nout (get_block_ports_description eb)
;;

let get_block_implicit_port_nin eb =
  Mdl_ast_to_deep_ast.get_implicit_port_nin (get_block_ports_description eb)
;;

let get_block_implicit_port_nout eb =
  Mdl_ast_to_deep_ast.get_implicit_port_nout (get_block_ports_description eb)
;;

let get_block_position eb =
  (get_block_graphical eb).explicit_block_position
;;

(** {Setting fields of explicit blocks} *)

let add_block_parameters eb parameters =
  let t = get_block_parameters eb in
  List.iter (fun (id, v) ->
    let bdg =
      {
        Binding.bound_ident = id;
        Binding.bound_value = v;
      } in
    Hashtbl.add t id bdg) parameters
;;

let set_block_body eb block_body =
  {
    eb with
    explicit_block_body = block_body;
  }
;;

let set_block_all_bindings eb all_bindings =
  let block_body = get_block_body eb in
  let block_body =
    {
      block_body with
      explicit_block_all_bindings = all_bindings;
    } in
  set_block_body eb block_body
;;

let set_block_graphical eb block_graphical =
  let all_bindings = get_block_all_bindings eb in
  let all_bindings =
    {
      all_bindings with
      explicit_block_graphical = block_graphical;
    } in
   set_block_all_bindings eb all_bindings
;;

let map_block_graphical f eb =
  let block_graphical = f (get_block_graphical eb) in
  set_block_graphical eb block_graphical
;;

let map_block_position f eb =
  let f block_graphical =
    {
      block_graphical with
      explicit_block_position =
        f block_graphical.explicit_block_position;
    } in
  map_block_graphical f eb
;;

let set_block_position eb position =
  map_block_position (fun _ -> position) eb
;;

let map_block_rotation f eb =
  let f block_graphical =
    {
      block_graphical with
      explicit_block_rotation =
        f block_graphical.explicit_block_rotation;
    } in
  map_block_graphical f eb
;;

let set_block_rotation eb rotation =
  map_block_rotation (fun _ -> rotation) eb
;;

let map_block_mirror f eb =
  let f block_graphical =
    {
      block_graphical with
      explicit_block_mirror =
        f block_graphical.explicit_block_mirror;
    } in
  map_block_graphical f eb
;;

let set_block_mirror eb mirror =
  map_block_mirror (fun _ -> mirror) eb
;;

let map_block_ports_description f eb =
  let f block_graphical =
    {
      block_graphical with
      explicit_block_ports_description =
        f block_graphical.explicit_block_ports_description;
    } in
  map_block_graphical f eb
;;

let set_block_ports_description eb ports_description =
  map_block_ports_description (fun _ -> ports_description) eb
;;

let set_block_regular_port_nin eb nin =
  let ports_description = get_block_ports_description eb in
  let ports_description =
    Mdl_ast_to_deep_ast.set_regular_port_nin ports_description nin in
  set_block_ports_description eb ports_description
;;

let set_block_regular_port_nout eb nout =
  let ports_description = get_block_ports_description eb in
  let ports_description =
    Mdl_ast_to_deep_ast.set_regular_port_nout ports_description nout in
  set_block_ports_description eb ports_description
;;

let set_block_implicit_port_nin eb nin =
  let ports_description = get_block_ports_description eb in
  let ports_description =
    Mdl_ast_to_deep_ast.set_implicit_port_nin ports_description nin in
  set_block_ports_description eb ports_description
;;

let set_block_implicit_port_nout eb nout =
  let ports_description = get_block_ports_description eb in
  let ports_description =
    Mdl_ast_to_deep_ast.set_implicit_port_nout ports_description nout in
  set_block_ports_description eb ports_description
;;

(** Access/setting in ebss component of a sub-system block. *)
let get_ebss_body ebss = ebss.explicit_block_sub_system_body;;

let get_ebss_body_blocks ebss_body = ebss_body.explicit_system_blocks;;

let get_ebss_blocks ebss = get_ebss_body_blocks (get_ebss_body ebss);;

let set_ebss_body ebss ebss_body =
  {
    ebss with
    explicit_block_sub_system_body = ebss_body;
  }
;;

let set_ebss_body_blocks ebss_body blocks =
  {
    ebss_body with
    explicit_system_blocks = blocks;
  }
;;

let set_ebss_blocks ebss blocks =
  set_ebss_body ebss
    (set_ebss_body_blocks (get_ebss_body ebss) blocks)
;;

(** Add a block to the list of blocks of a sub-system block. *)
let add_ebss_block ebss blk =
  set_ebss_blocks ebss (blk :: get_ebss_blocks ebss)
;;

(** Set ebss of a sub-system or fail *)
let set_block_ebss eb ebss =
  match eb.explicit_block_desc with
  | Explicit_block_regular _
  | Explicit_block_reference _
  | Explicit_block_s_function _ -> assert false
  | Explicit_block_sub_system _sub_system_desc ->
    { eb with
      explicit_block_desc =
        Explicit_block_sub_system ebss;
    }
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
