(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_ast.mli,v 1.1 2019-05-10 14:07:13 jpc Exp $ *)

(** The Abstract syntax tree for MDL files.
   The stateflow section has a tree shape. *)

open Ident;;
open Ast_types;;
open Binding;;
open Mdl_block_types;;

(** Shorthands borrowed from basic modules *)
type model_ident = Ident.ident
and configuration_ident = Ident.ident

and tag_ident = Ident.ident
;;

type prop_name = Ident.ident
and class_name = Ast_types.mdl_name (* In fact "Ident.ident" *)
;;

(** Borrowed from Ast_types *)
type system_name = Ast_types.mdl_name

and block_name = Ast_types.mdl_name
and block_source_path = Ast_types.mdl_path
and block_source_type = Ast_types.mdl_name

and line_name = Ast_types.mdl_name
and annotation_name = Ast_types.mdl_text
(* FIXME: seem not required (all tests compile without it).
and annotation_tag = Ast_types.mdl_text
*)

(* FIXME: not really verified. *)
and port_name = Ast_types.mdl_name
and port_number = Ast_types.mdl_port_number

and object_name = Ast_types.mdl_name
and backup_class_name = Ast_types.mdl_name
and array_name = Ast_types.mdl_name
and ui_name = Ast_types.mdl_name
and matstruct_name = Ast_types.mdl_name
;;

(* Proper type definitions *)

type 'a ast_node =
  {
    ast_loc : Location.t;
    ast_desc : 'a;
  }
;;

type model_version =
  | Mdl_model_version of string
  | Slx_model_version of string
;;

type enum = Unatural.t;;

type configuration_longident =
  | Configuration_lident of configuration_ident
  | Configuration_ldot of configuration_longident * configuration_ident
;;

type line_type =
  | Connection

and connection_type =
  | Dest_src
  | Dest_dest
  | Src_dest
  | Src_src

and array_type =
  | Struct
  | Cell
  | Handle
  | Array_type of string

and object_type = mdl_name

and matstruct_type = mdl_name

and interface_block_parameter_type = Ident.ident
;;

type block_sf_block_type =
  | Block_sf_block_type_none
  | Block_sf_block_type_matlab_function
  | Block_sf_block_type_chart
;;

type custom_property_type =
  | Custom_property_type_uint8
  | Custom_property_type_uint16
  | Custom_property_type_uint32
  | Custom_property_type_uint64
  | Custom_property_type_int8
  | Custom_property_type_int16
  | Custom_property_type_int32
  | Custom_property_type_int64
  | Custom_property_type_slbool
  | Custom_property_type_string
  | Custom_property_type_custom of Ast_types.mdl_name
;;

type naturals = Ast_types.natural list;;

(* FIXME: to be suppressed ?

type mdl_file_name =
  | Mdl_file_name of string
;;
*)

type integer_matrix = int array array;;

type object_id = Path.file_name * natural;;

type zorder = string;;

type simple_values = simple_value list
(* FIXME: simple_values = simple_value array ??? *)
and simple_value =
  | Bool of string
  | Float of string
  (* FIXME: uncomment Nat?
    | Nat of string *)
  | Int of string
  | String of string
  | Ident of Ident.ident
;;

(* Polymorphic [ident_binding] type constructor is defined in
  module basics/[Binding] *)
type simple_binding = simple_value Binding.ident_binding;;

type model_values = model_value list
and model_value =
  | Value_simple of simple_value
  | Value_vector of simple_values
(* FIXME:
  | Value_matrix of simple_values list
  Or more precisely, after careful examination of the parser:
  | Value_matrix of simple_value_rows
  and simple_value_rows = simple_value_row list
  and simple_value_row =
    | Value_matrix_row of simple_values
  or even better, since no row can be empty in MDL:
    | Value_matrix_row of simple_values1
*)
  | Value_matrix of model_values
  | Value_bindings of bindings

and bindings = binding list
and binding = model_value Binding.ident_binding
;;

type 'font_and_color poly_annotation_graphical_default =
  | Font_and_color of 'font_and_color
  | Horizontal_alignment of horizontal_alignment
  | Vertical_alignment of vertical_alignment
  | Text_as_click_cb of mdl_bool
  | Click_fcn of mdl_code
  | Tex_mode of mdl_bool
  | Interpreter of string
  | Annotation_graphical_binding of binding

and horizontal_alignment =
  | Alignment_Center
  | Alignment_Left
  | Alignment_Right

and vertical_alignment =
  | Alignment_Top
  | Alignment_Middle
  | Alignment_Bottom
;;

type font_items = font_item list
and font_item =
  | Font_angle of string
  | Font_name of mdl_font_name
  | Font_size of natural
  | Font_weight of string
;;

type font_and_color =
  | Background of color
  | Drop_shadow of mdl_bool
  | Font_item of font_item
  | Foreground of color

and color = string
;;

type block_graphical_defaults = block_graphical_default list
and block_graphical_default =
  | Block_mirror of mdl_bool
  | Block_name_placement of string
  | Block_orientation of string
  | Block_rotation of natural
  | Block_show_name of mdl_bool
  | Block_hide_automatic_name of mdl_bool
  | Block_zorder of zorder
  | Block_font_and_color of font_and_color
;;

type block_parameter_defaults_list = block_parameter_defaults list
and block_parameter_defaults =
  {
    block_type : block_type;
    parameter_defaults : bindings;
  }
;;

type line_graphical_defaults = line_graphical_default list
and line_graphical_default =
  | Line_font_item of font_item
  | Line_zorder_item of zorder
;;

type mask_graphical_defaults = mask_graphical_default list
and mask_graphical_default = binding
and mask_parameter_defaults = mask_parameter_default list
and mask_parameter_default = binding
;;

type annotation_graphical_defaults = annotation_graphical_default list
and annotation_graphical_default =
  font_and_color poly_annotation_graphical_default
;;

(* FIXME:

This should be here

type defaults_section =
  | Block_graphical_defaults_section of block_graphical_defaults
  | Block_parameter_defaults_section of block_parameter_defaults_list
  | Annotation_graphical_defaults_section of annotation_graphical_defaults
  | Line_graphical_defaults_section of line_graphical_defaults
  | Mask_graphical_defaults_section of mask_graphical_defaults
  | Mask_parameter_defaults_section of mask_parameter_defaults

and defaults_sections = defaults_section list
;;

type configuration_set_parameter =
  | Configuration_parameter of Mdl_ast.configuration_set_parameter
  | Model_configuration_section of Mdl_ast.model_configuration
;;
*)

type lines = line list
and line =
  | Line of line_items

and line_items = line_item list
and line_item =
  | Line_name of line_name
  | Line_source of line_source
  | Line_target of line_target
  | Line_type of line_type
  | Line_labels of labels
  | Line_points of points
  | Line_graphical_default of line_graphical_default

(** Line source *)
and line_source =
  | Line_source_connector of line_connector

and line_connector =
  {
    line_connector_block : block_name;
    line_connector_port : port_value;
  }

and port_value =
  | Port_trigger
  | Port_enable
  | Port_state
  | Port_ifaction
  | Port_number of port_number
  | Port_lconn of port_number
  | Port_rconn of port_number

(** Line target *)
and line_target =
  | Line_target_connector of line_connector
  | Line_target_branch of line_target_branch_items

and line_target_branch_items = line_target_branch_item list
and line_target_branch_item =
  | Line_target_branch_connection_type of connection_type
  | Line_target_branch_labels of labels
  | Line_target_branch_points of points
  | Line_target_branch_source of line_source
  | Line_target_branch_target of line_target
  | Line_target_branch_zorder of zorder

and points = integer_matrix

and labels = integer_matrix
;;

type annotations = annotation list
and annotation =
  | Annotation of annotation_items

and annotation_items = annotation_item list
and annotation_item =
  | Annotation_name of annotation_name
  | Annotation_sid of block_sid
(*
  | Annotation_tag of annotation_tag
*)
  | Annotation_position of position
  | Annotation_auto_size of mdl_bool
  | Annotation_word_wrap of mdl_bool
  | Annotation_zorder of zorder
  | Annotation_graphical_default of annotation_graphical_default
  | Annotation_binding of binding

and block_sid =
  | Sid_compound of block_sid_compound
  | Sid_string of string

and block_sid_compound =
  | Sid_natural of natural
  | Sid_ref of natural * block_sid_compound
  | Sid_child of natural * block_sid_compound

and position = int list
;;

type system_connectors = system_connector list
and system_connector = bindings
;;

type array_description =
  {
    array_type : array_type;
    array_dimension : dimension;
    array_prop_name : prop_name;
    array_items : array_items;
  }

and dimension = simple_value

and array_items = array_item list
and array_item =
  | Array_name of array_name
  | Array_sub_array of array_description
  | Array_custom_property of custom_property
  | Array_matstruct of matstruct_items
  | Array_object of object_description
  | Array_vendor_binding of vendor_binding
  | Array_vendor_configset of vendor_ident * vendor_items
  | Array_binding of binding

and vendor_binding =
  {
    vendor_ident : vendor_ident;
    vendor_configuration_longident : configuration_longident;
    vendor_items : vendor_items;
  }
and block_vendor = vendor_binding

and vendor_items = vendor_item list
and vendor_item =
  | Vendor_binding of object_item
  | Vendor_dollar_binding of dollar_binding
  | Vendor_vendor_binding of vendor_binding

and dollar_binding =
  | Dollar_object_id of object_id
  | Dollar_class_name of class_name
  | Dollar_backup_class of backup_class_name
  | Dollar_prop_name of prop_name

and vendor_ident =
  | Simulink
  | Scicos
  | Rtw
  | Ui of Ident.ident

and custom_property =
  {
    custom_property_type : custom_property_type;
    custom_property_enum_str_pairs : enum_str_pairs;
    custom_property_ident : prop_name;
    custom_property_value : simple_value;
  }

and matstruct_items = matstruct_item list
and matstruct_item =
  | Matstruct_type of matstruct_type
  | Matstruct_name of matstruct_name
  | Matstruct_array of array_description
  | Matstruct_binding of binding

and enum_str_pairs = enum_str_pair list
and enum_str_pair = (enum * mdl_text)

and object_description = object_item list

and object_item =
  | Object_type of object_type
  | Object_name of object_name
  | Object_dollar_binding of dollar_binding
  | Object_array of array_description
  | Object_custom_property of custom_property
  | Object_object of object_description
  | Object_binding of binding
;;

type block_mask_item =
    (* MaskCallbackString "||powergui('HookPort',[],[],gcb);|" *)
    (* MaskCallbackString "|validateInstance|" *)
  | Mask_callback_string of mdl_pipe_expressions
    (* MaskCapabilities "slmaskedcaps(gcbh)" *)
  | Mask_capabilities of mdl_expression
    (* MaskDescription "Matrix Gain." *)
  | Mask_description of mdl_text
    (* MaskDisplay "plot(0,0,100,100,[33,47])" *)
  | Mask_display of mdl_code
    (* Mask_display_string "Pilote Teta" *)
  | Mask_display_string of mdl_text
    (* Mask_enable_string "on,on" *)
  | Mask_enable_string of mdl_comma_bools
    (* MaskHelp "This block is part of ..." *)
  | Mask_help of mdl_text
  | Mask_hide_contents of mdl_bool
  | Mask_icon_frame of mdl_bool
  | Mask_icon_opaque of mdl_bool
    (* MaskIconRotate "port" or "none" *)
  | Mask_icon_rotate of Ident.ident
    (* Mask_icon_units "pixels" or "autoscale" or "normalized" *)
  | Mask_icon_units of Ident.ident
    (* MaskInitialization "Select = 4 - Select;" *)
  | Mask_initialization of mdl_shallow_code
    (* MaskPortRotate "default" *)
  | Mask_port_rotate of Ident.ident
    (* MaskPromptString "ID|Block Title" *)
  | Mask_prompt_string of mdl_pipe_texts
  | Mask_run_init_for_icon_redraw of mdl_bool
  | Mask_self_modifiable of mdl_bool
    (* MaskStyleString "popup(yes|no),edit,popup(blocks|steady|zero)" *)
    (* MaskStyleString "popup(==|~=|<|<=|>=|>),edit" *)
  | Mask_style_string of mdl_comma_texts
    (* MaskTabNameString ",,," *)
  | Mask_tab_name_string of mdl_text
    (* MaskToolTipString "on,on,on" *)
  | Mask_tooltip_string of mdl_comma_bools
    (* MaskTunableValueString "off,on" *)
  | Mask_tunable_value_string of mdl_comma_bools
    (* MaskType "SRS Requirements" *)
  | Mask_type of mdl_text
    (* MaskValueString "SystemReq|SRS Requirements" *)
    (* MaskValueString "0.05||d0d79838-f694-4c67-8e60" *)
  | Mask_value_string of mdl_pipe_texts
    (* MaskVarAliasString ",," *)
  | Mask_var_alias_string of mdl_text
    (* MaskVariables "label=&1;index=@2;" *)
  | Mask_variables of mdl_semi_texts
  | Mask_visibility_string of mdl_comma_bools

and mdl_expression =
  | Expression of mdl_text
and mdl_comma_bools =
  | Comma_bools of mdl_text
and mdl_pipe_expressions =
  | Pipe_expressions of mdl_text
and mdl_pipe_texts =
  | Pipe_texts of mdl_text
and mdl_comma_texts =
  | Comma_texts of mdl_text
and mdl_semi_texts =
  | Semi_texts of mdl_text
;;

type blocks = block list
and block =
  | Block of block_items
  | Block_vendor of block_vendor

and block_items = block_item list
and block_item =
  | Block_type of block_type
  | Block_name of block_name
  | Block_source_path of block_source_path
  | Block_source_type of block_source_type
  | Block_port of block_port
  | Block_ports_description of block_ports_description
  | Block_position of position
  | Block_tag of mdl_text
  | Block_sid of block_sid
  | Block_state of simple_value
  | Block_sf_block_type of block_sf_block_type
  | Block_array of array_description
  | Block_object of object_description
  | Block_system of system
  | Block_graphical_default of block_graphical_default
  | Block_mask_item of block_mask_item
  | Block_mask of mask_description
  | Block_vendor_binding of vendor_binding
  | Block_binding of binding
  | Block_instance_data of bindings

and mask_description =
  {
    mask_type : mdl_text option;
    mask_display : mask_display;
    mask_description : mdl_text option;
    mask_initialisation : mdl_code option;
    mask_parameters : mask_parameter list;
    mask_bindings : bindings;
  }

and mask_parameter =
  {
    mask_parameter_ident : Ident.ident;
    mask_parameter_style : mdl_text;
    mask_parameter_prompt : mdl_text;
    mask_parameter_callback : mdl_code option;
    mask_parameter_tunable : mdl_bool option;
    mask_parameter_value : mdl_text;
    mask_parameter_evaluation : mdl_bool option;
    mask_parameter_enable : mdl_bool option;
    mask_parameter_visible : mdl_bool option;
    mask_parameter_internal : mdl_bool option;
    mask_parameter_never_save : mdl_bool option;
    mask_parameter_read_only : mdl_bool option;
  }

and mask_display =
  {
    mask_display_source : mdl_code option;
    mask_display_bindings : bindings;
(* todo: type some bindings
   "IconOpaque"
    mask_display_icon_opaque : mdl_text option;
    "IconRotate"
    mask_display_icon_rotate : mdl_text option;
    "IconUnits"
    mask_display_icon_units : mdl_text option;
    "RunInitForIconRedraw"
    mask_display_run_init_for_icon_redraw : bool option;
 *)
  }

and block_port =
  {
    block_port_number : port_number;
    block_port_name : port_name option;
    block_port_bindings : block_port_bindings;
  }

and block_port_bindings = block_port_binding list
and block_port_binding =
  | Block_port_binding_array of array_description
  | Block_port_binding_vendor of vendor_binding
  | Block_port_binding of binding

and block_ports_description = naturals

and system =
  {
    system_name : system_name;
    system_position : position;
    system_bindings : bindings;
    system_blocks : blocks;
    system_lines : lines;
    system_annotations : annotations;
    system_connectors : system_connectors;
  }
;;

type model_section =
  {
    model_ident : model_ident;
    model_version : model_version;
    model_configuration_parameters : model_configuration_parameters;
    model_configurations : model_configurations;
    model_set_configuration_option : model_set_configuration_option;
    model_workspace : model_workspace;
  (* FIXME:
     Should disappear as soon as possible,
     and replaced by

    code_generation_section : bindings;
    model_defaults_sections: model_defaults_sections;

    and model_defaults_sections = model_defaults_section list

    and model_defaults_section =
      | block_graphical_defaults_section : block_graphical_defaults;
      | block_parameter_defaults_section : block_parameter_defaults list;
      | annotation_graphical_defaults_section : annotation_graphical_defaults;
      | line_graphical_defaults_section : line_graphical_defaults;
      | mask_graphical_defaults_section : mask_graphical_defaults;
      | mask_parameter_defaults_section : mask_parameter_defaults;

    have a look at Mdl_parser.make_model,
    that should be moved into deep phase
   *)

    block_graphical_defaults_section : block_graphical_defaults;
    block_parameter_defaults_section : block_parameter_defaults list;
    annotation_graphical_defaults_section : annotation_graphical_defaults;
    line_graphical_defaults_section : line_graphical_defaults;
    mask_graphical_defaults_section : mask_graphical_defaults;
    mask_parameter_defaults_section : mask_parameter_defaults;
  (*
    model_defaults_sections: model_defaults_sections;
  *)
    model_system : system;
  }

and model_configuration_parameters = model_configuration_parameter list
and model_configuration_parameter =
  | Configuration_graphical_interface of graphical_interface_items
  | Configuration_object of object_description
  | Configuration_vendor_binding of vendor_binding
  | Configuration_binding of simple_binding

and graphical_interface_items = graphical_interface_item list
and graphical_interface_item =
  | Graphical_enableport of io_port_binding_items
  | Graphical_inport of io_port_binding_items
  | Graphical_outport of io_port_binding_items
  | Graphical_trigport of io_port_binding_items
  | Graphical_inports_number of natural
  | Graphical_outports_number of natural
  | Graphical_trigports_number of natural
  | Graphical_binding of binding

and io_port_binding_items = io_port_binding_item list
and io_port_binding_item =
  | Io_port_name of port_name
  | Io_port_binding of binding

and model_configurations = model_configuration list
and model_configuration = array_description

and model_set_configuration_option = configuration_set option

and configuration_set =
  {
    configuration_vendor : vendor_ident;
    configuration_longident : configuration_longident;
    configuration_prop_name : prop_name;
    configuration_object_id : object_id;
  }

and model_workspace = model_workspace_datas option

and model_workspace_datas = model_workspace_data list
and model_workspace_data =
  | Workspace_data_source_type of workspace_data_source_type
  | Workspace_data_mat_file_name of Path.file_name
  | Workspace_data_source_file_name of Path.file_name
  | Workspace_data_mcode of simple_value
  | Workspace_data_mdl_file of tag_ident

and workspace_data_source_type =
  | M_Code
  | MATLAB_Code
  | MATLAB_File
  | MAT_File
;;

type matdata_section =
  | Matdata_section of matdata_section_item

and matdata_section_item =
  {
    matdata_num_records : natural;
    matdata_records : matdata_records;
  }

and matdata_records = matdata_record list
and matdata_record =
  {
    matdata_record_tag : tag_ident;
    matdata_record_data : mdl_text;
  }
;;

(** {6 Stateflow types section} *)

(* A stateflow id is a unique natural number. *)
type stateflow_id = Ast_types.natural;;

(* A stateflow reference to a stateflow id is a stateflow id. *)
type stateflow_ref = stateflow_id;;

type stateflow_numbers = stateflow_number list
and stateflow_number =
  | Stateflow_int of string
  | Stateflow_float of string
;;

type stateflow_src_dst_items = stateflow_src_dst_item list
and stateflow_src_dst_item =
  | Stateflow_src_dst_ref of stateflow_ref
  | Stateflow_intersection of stateflow_numbers
;;

type stateflow_eml_items = stateflow_eml_item list
and stateflow_eml_item =
  | Stateflow_eml_name of mdl_name
  | Stateflow_eml_script of mdl_code
  | Stateflow_eml_binding of binding
;;

type stateflow_props_array_items = stateflow_props_array_item list
and stateflow_props_array_item =
  | Stateflow_props_array_item_size of mdl_text
  | Stateflow_props_array_item_first_index of mdl_text
  | Stateflow_props_array_item_binding of binding
;;

type stateflow_props_type_items = stateflow_props_type_item list
and stateflow_props_type_item = binding
;;

type stateflow_props_items = stateflow_props_item list
and stateflow_props_item =
  | Stateflow_props_array of stateflow_props_array_items
  | Stateflow_props_type of stateflow_props_type_items
  | Stateflow_props_binding of binding
;;

type stateflow_items = stateflow_item list
and stateflow_item =
  | Stateflow_binding of binding
  | Stateflow_eml of stateflow_eml_items
  | Stateflow_execution_order of natural
  | Stateflow_label_string of mdl_text
  | Stateflow_chart_file_number of natural
  | Stateflow_name of mdl_name
  | Stateflow_scope of ident
  | Stateflow_props of stateflow_props_items
  | Stateflow_ssid_number of natural
  | Stateflow_trigger of ident
  | Stateflow_type of ident
  | Stateflow_src of stateflow_src_dst_items
  | Stateflow_dst of stateflow_src_dst_items
  | Stateflow_plant_modeling_info
;;

type stateflow_junctions = stateflow_junction list
and stateflow_junction =
  {
    stateflow_junction_id : stateflow_id;
    stateflow_junction_items : stateflow_items;
  }
;;

type stateflow_transitions = stateflow_transition list
and stateflow_transition =
  {
    stateflow_transition_id : stateflow_id;
    stateflow_transition_items : stateflow_items;
  }
;;

type stateflow_events = stateflow_event list
and stateflow_event = {
  stateflow_event_id : stateflow_id;
  stateflow_event_items : stateflow_items;
}
;;

type stateflow_datas = stateflow_data list
and stateflow_data = {
  stateflow_data_id : stateflow_id;
  stateflow_data_items : stateflow_items;
}
;;

type stateflow_states = stateflow_state list
and stateflow_state =
  {
    stateflow_state_id : stateflow_id;
    stateflow_state_datas : stateflow_datas;
    stateflow_state_events : stateflow_events;
    stateflow_state_junctions : stateflow_junctions;
    stateflow_state_states : stateflow_states;
    stateflow_state_transitions : stateflow_transitions;
    stateflow_state_items : stateflow_items;
  }
;;

type stateflow_targets = stateflow_target list
and stateflow_target =
  {
    stateflow_target_id : stateflow_id;
    stateflow_target_items : stateflow_items;
  }
;;

type stateflow_charts = stateflow_chart list
and stateflow_chart =
  {
    stateflow_chart_id : stateflow_id;
    stateflow_chart_datas : stateflow_datas;
    stateflow_chart_events : stateflow_events;
    stateflow_chart_junctions : stateflow_junctions;
    stateflow_chart_states : stateflow_states;
    stateflow_chart_transitions : stateflow_transitions;
    stateflow_chart_items : stateflow_items;
   }
;;

type stateflow_machine =
  {
    stateflow_machine_id : stateflow_id;
    stateflow_machine_charts : stateflow_charts;
    stateflow_machine_datas : stateflow_datas;
    stateflow_machine_events : stateflow_events;
    stateflow_machine_targets : stateflow_targets;
    stateflow_machine_items : stateflow_items;
  }
;;

type stateflow_instance =
  {
    stateflow_instance_id : stateflow_id;
    stateflow_instance_name : mdl_name;
    stateflow_instance_machine_ref : stateflow_ref;
    stateflow_instance_chart_ref : stateflow_ref;
  }
;;

type stateflow_section_items = stateflow_section_item list
and stateflow_section_item =
  | Stateflow_section_machine of stateflow_machine
  | Stateflow_section_instance of stateflow_instance
;;

type stateflow_section =
  | Stateflow_section of stateflow_section_items
;;

(* End of stateflow section *)

type interface_block_parameters = interface_block_parameter list
and interface_block_parameter =
  interface_block_parameter_type ident_binding
;;

type interface_blocks = interface_block list
and interface_block =
  {
    interface_block_type : block_type;
    interface_block_source_type : block_source_type option;
    interface_block_source_path : block_source_path option;
    interface_block_ports_description : block_ports_description;
    interface_block_parameters : interface_block_parameters;
  }
;;

type interface =
  {
    interface_ident : Ident.ident;
    interface_blocks : interface_blocks;
  }
;;

type 'stateflow poly_implementation_section_items =
  'stateflow poly_implementation_section_item list
and 'stateflow poly_implementation_section_item =
  | Implementation_library of model_section
  | Implementation_model of model_section
  | Implementation_matdata of matdata_section
  | Implementation_stateflow of 'stateflow
;;

type 'stateflow poly_implementation =
  'stateflow poly_implementation_section_items
;;

type 'stateflow poly_simulink_file_contents =
  | Implementation of 'stateflow poly_implementation
  | Interface of interface
;;

type 'stateflow poly_simulink_file =
  Path.explicit_file_name * 'stateflow poly_simulink_file_contents
;;

type simulink_file_contents = stateflow_section poly_simulink_file_contents;;

type simulink_file = stateflow_section poly_simulink_file
(** A [simulink_file] is indeed a file name and a Simulink file contents
  containing a stateflow section. *)
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
