(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: code_make_host_print.ml,v 1.1 2019-05-10 14:07:14 jpc Exp $ *)

(** The functor to print code, given an host specific translation module. *)

open Mdl_basics_deep_ast;;
open Mdl_deep_ast;;
open Mdl_basics_explicit_ast;;
open Instructions;;
open Concat_list;;
open Binding;;

open Unatural_types_print;;

let string_print f v =
  let b = Buffer.create 512 in
  let ppf = Format.formatter_of_buffer b in
  Format.fprintf ppf "%a@?" f v;
  Buffer.contents b
;;

module Make_print (Host : Code_make_host_print_spec.HOST) = struct

  type commented =
    | Commented
    | Not_commented
  ;;

  let print_text_as_raw_string_token =
    Mdl_translation.print_text_as_raw_string_token
  ;;
  (* This is:
     M_to_host.print_quoted_text ppf
       (M_to_host.quote_text txt)
  *)

  let print_text_as_encoded_string_token =
    Mdl_translation.print_text_as_encoded_string_token
  ;;

  let print_name ppf name =
    print_text_as_encoded_string_token ppf
      (Ast_types.mdl_text_of_mdl_name name)
  ;;

  let print_block_name = print_name;;

  let print_block_icon_text = print_text_as_encoded_string_token;;

  let print_diagram_name = print_name;;

  let print_mask_description ppf desc =
    let txt = match desc with
      | None -> Ast_types.mdl_text_of_string ""
      | Some desc -> desc in
    print_text_as_encoded_string_token ppf txt
  ;;

  let print_val ppf txt =
(*
prerr_endline
  (Printf.sprintf "print_val <- %s" (Ast_types.string_of_mdl_text txt));
let res = M_to_host.translate_expression txt in
prerr_endline
  (Printf.sprintf "print_val -> %s" (Ast_types.string_of_mdl_text res));
*)
(* That's print_text from Mdl_translation. *)
    Format.fprintf ppf "%s"
      (Ast_types.string_of_mdl_text (M_to_host.translate_expression txt))
  ;;

  let print_code ppf txt =
    Format.fprintf ppf "%a"
      M_to_host.print_quoted_text
      (M_to_host.quote_text
        (M_to_host.translate_script txt))
  ;;

  let print_host_bool = Host.print_host_bool;;

  let rec print_code_value ppf cv =
    Say.fdebug (fun epf ->
      Format.fprintf epf "print_code_value <- |%a|"
        Mdl_deep_ast_print.print_deep_value cv);

    let txt = mdl_text_of_code_value cv in
    let quoted_txt = M_to_host.quote_text txt in

    Say.fdebug (fun epf ->
      Format.fprintf epf "print_code_value -> |%a|"
        M_to_host.print_quoted_text quoted_txt);

    Format.fprintf ppf "%a" M_to_host.print_quoted_text quoted_txt

  and print_deep_value ppf dv =

    match dv with
    | Deep_bool b ->
      print_host_bool ppf b
    | Deep_float s
    | Deep_int s ->
      print_val ppf (Ast_types.mdl_text_of_string s)
    | Deep_ident id ->
      print_val ppf (Ast_types.mdl_text_of_ident id)

    | Deep_string s ->
      print_text_as_encoded_string_token ppf (Ast_types.mdl_text_of_string s)
    | Deep_wstring ws ->
      (* FIXME: Is that of_wstring correct? *)
      print_text_as_encoded_string_token ppf
        (Ast_types.mdl_text_of_string (Wstring.of_wstring ws))

    | Deep_expression txt -> print_val ppf txt
    | Deep_code (Ast_types.Mdl_code txt) -> print_code ppf txt

    | Deep_vector dvs ->
      print_val ppf (text_of_vector dvs)

    | Deep_matrix dvs ->
      print_val ppf (text_of_matrix dvs)

    | Deep_bindings bds ->
      print_val ppf (text_of_deep_bindings bds)

  and text_of_vector dvs = text_of_deep_values ", " dvs
  and text_of_matrix dvs = text_of_deep_values "; " dvs

  and text_of_deep_values sep dvs =
    Ast_types.mdl_text_of_string
      (Printf.sprintf "[ %s ]"
        (String.concat sep (List.map string_of_deep_value dvs)))

  and matlab_string_of_bool b = if b then "1" else "0"

  and string_of_deep_value dv =
    match dv with
    | Deep_bool b -> matlab_string_of_bool b

    | Deep_float s
    | Deep_string s
    | Deep_int s -> s
    | Deep_ident id -> Ast_types.string_of_ident id

    | Deep_wstring ws -> Wstring.of_wstring ws

    | Deep_expression txt
    | Deep_code (Ast_types.Mdl_code txt) -> Ast_types.string_of_mdl_text txt

    | Deep_vector dvs -> Ast_types.string_of_mdl_text (text_of_vector dvs)

    | Deep_matrix dvs -> Ast_types.string_of_mdl_text (text_of_matrix dvs)

    | Deep_bindings bds ->
      Ast_types.string_of_mdl_text (text_of_deep_bindings bds)

  and print_deep_binding ppf bdg =
    Format.fprintf ppf "%a, %a"
      print_deep_value
        (Deep_string (Ast_types.string_of_ident bdg.Binding.bound_ident))
      print_deep_value
        (bdg.Binding.bound_value)

  and print_deep_bindings ppf = function
    | [] -> Format.fprintf ppf "list (%s)" ""
    | bdg :: bdgs ->
      Format.fprintf ppf
       "list (%a%a)"
       print_deep_binding bdg
       (fun ppf ->
        List.iter
          (fun bdg -> Format.fprintf ppf ", %a" print_deep_binding bdg))
       bdgs

  and text_of_deep_bindings bdgs =
    Ast_types.mdl_text_of_string (string_print print_deep_bindings bdgs)

  and mdl_text_of_code_value cv =
    Say.fdebug (fun epf ->
      Format.fprintf epf "mdl_text_of_code_value <- %a"
        Mdl_deep_ast_print.print_deep_value cv);
    let r = Ast_types.mdl_text_of_string (string_print print_deep_value cv) in
    Say.fdebug (fun epf ->
      Format.fprintf epf "mdl_text_of_code_value -> %s"
        (Ast_types.string_of_mdl_text r));
    r
  ;;

  let print_code_values ppf = function
    | [] -> ()
    | v :: vs ->
      Format.fprintf ppf "%a%a"
        print_code_value v
        (fun ppf ->
           List.iter
             (fun v -> Format.fprintf ppf ",@ %a" print_code_value v))
        vs
  ;;

  let print_block_parameter_value ppf = function
    | Parameter_value v -> print_code_value ppf v
    | Parameter_struct fields ->
      let print_field ppf
          {
            Binding.bound_ident = label;
            Binding.bound_value = code_values;
          } =
        Format.fprintf ppf Host.struct_item_format
          (Ast_types.string_of_ident label)
          print_code_values code_values in

      let print_fields ppf (field, fields) =
        Format.fprintf ppf Host.struct_create_format
          print_field field
          (fun ppf ->
           List.iter
             (fun field -> Format.fprintf ppf ",@ %a" print_field field))
          fields in

      print_fields ppf fields
  ;;

  let print_block_parameter_binding binding_format ppf
      { bound_ident = id; bound_value = v; } =
    Format.fprintf ppf ("@{<m>" ^^ binding_format ^^ "@}")
       (Ast_types.string_of_ident id)
       print_block_parameter_value v
  ;;

  let print_block_parameter_bindings binding_format ppf (binding, bindings) =
    Format.fprintf ppf "%a%a"
      (print_block_parameter_binding binding_format) binding
      (fun ppf ->
       List.iter
         (fun binding ->
          Format.fprintf ppf ";@ %a"
            (print_block_parameter_binding binding_format)
            binding)
      )
      bindings
  ;;

  let print_host_block_parameters = print_block_parameter_bindings;;

  let print_mask_parameter_value ppf = function
    | Mdl_basics_explicit_ast.Explicit_mask_parameter_quote txt ->
      print_code_value ppf (Deep_string (Ast_types.string_of_mdl_text txt))
    | Mdl_basics_explicit_ast.Explicit_mask_parameter_eval txt ->
      print_code_value ppf (Deep_expression txt)
  ;;

  let print_matlab_script ppf (Matlab_script script) =
    let mdl_text = M_to_host.translate_script script in
    M_to_host.print_quoted_text ppf (M_to_host.quote_text mdl_text)
  ;;

  let print_context ppf =

    let print_binding commented ppf binding =
      match commented with
      | Not_commented ->
        print_block_parameter_binding Host.binding_format ppf binding
      | Commented ->
        (* FIXME: a brave soul should investigate why there was such an
           assertion and what is the problem to suppress it.

        match binding with
        | {
            Binding.bound_ident = _ident;
            Binding.bound_value = v;
          } ->
          let is_matlab_string = function
            | Parameter_struct _ -> false
            | Parameter_value v ->
              begin match v with
              | Deep_string _
              | Deep_wstring _ -> true
              |(Deep_bool _ | Deep_float _ | Deep_int _ | Deep_ident _ |
                Deep_code _ | Deep_expression _| Deep_vector _ |
                Deep_matrix _ | Deep_bindings _) -> false
              end in
          Say.fdebug
            (fun epf ->
             Format.fprintf epf "print_binding commented %a =@ %a"
               Ident_print.print_ident _ident
               Instructions_print.print_scicos_block_parameter_value v);

          assert (is_matlab_string v);
        *)

        let line =
          string_print
            (print_block_parameter_binding
               Host.commented_binding_format)
            binding in

        Format.fprintf ppf "%s" (Host.comment_string line) in

    let pr_bindings commented ppf (binding, bindings) =
      let pr_binding = print_binding commented in
      Format.fprintf ppf "%a%a"
        pr_binding binding
        (fun ppf ->
         List.iter
           (fun binding ->
            Format.fprintf ppf "%s@ %a"
              Host.row_separator
              pr_binding binding))
        bindings in

    let print_bindings commented ppf bindings =
      Format.fprintf ppf "@[<v>%a@]"
        (pr_bindings commented)
        bindings in

    function
    | Context_code script ->
      print_matlab_script ppf script
    | Context_bindings bindings ->
      print_bindings Not_commented ppf bindings
    | Context_commented bindings ->
      print_bindings Commented ppf bindings
  ;;

  let print_workspace ppf workspace =

    let print_workspace_item ppf wi =
      let workspace_line =
        match wi with
        | Mdl_deep_ast.
          Deep_code_workspace_data {
            Mdl_deep_ast.
            workspace_data_source_type =
              (None | Some Deep_M_Code | Some Deep_MATLAB_Code);
            workspace_data_source_code = code;
          } ->
          begin match code with
          | Deep_workspace_data_source_code code -> code
          end
        | Deep_code_workspace_data {
            workspace_data_source_type = Some (Deep_MATLAB_File wssfn);
            workspace_data_source_code = Deep_workspace_data_source_code code;
          } ->
          (* FIXME: we should verify that a corresponding target language
             source file exists orelse warn the user *)
          Say.warning (Printf.sprintf
            "matlab file %s must be converted to %s"
            wssfn
            (Configuration.string_of_target_language
               (Configuration.get_target_language ())));
          code

        | Deep_mat_file_workspace_data wssfn ->
          let code_string =
            Printf.sprintf "load (%s)"
              (Genr_utils.matlab_string_of_file_name wssfn) in
          Ast_types.mdl_code_of_string code_string in

      let script =
        Matlab_script
          (Ast_types.mdl_text_of_mdl_code workspace_line) in
      Format.fprintf ppf "%a"
        print_matlab_script script in

    let rec pr_workspace_items ppf = function
      | [] -> ()
      | [ workspace_item ] ->
        Format.fprintf ppf "%a" print_workspace_item workspace_item
      | workspace_item :: workspace_items ->
        Format.fprintf ppf
          "%a%s@ %a"
          print_workspace_item workspace_item
          Host.row_separator
          pr_workspace_items workspace_items in

    Format.fprintf ppf "@[<v>%a@]" pr_workspace_items workspace
  ;;

  let print_matlab_script_as_value ppf matlab_code =
    Format.fprintf ppf "@[<v>%a@]"
      print_matlab_script matlab_code
  ;;

  let print_host_ident ppf = function
    | Hi s -> Format.fprintf ppf "%s" s
  ;;

  let print_host_block_ident ppf = function
    | Hbi s -> Format.fprintf ppf "%s" s
  ;;

  let print_host_block_tag_ident ppf = function
    | Hbti s -> Format.fprintf ppf "%s" s
  ;;

  let print_host_diagram_ident ppf = function
    | Hdi s -> Format.fprintf ppf "%s" s
  ;;

  let print_host_model_ident ppf = function
    | Hmi s -> Format.fprintf ppf "%s" s
  ;;

  let print_host_function_ident ppf = function
    | Hfi s -> Format.fprintf ppf "%s" s
  ;;

  let print_host_idents ppf host_idents =

    let rec loop ppf = function
      | [] -> ()
      | [ id ] -> Format.fprintf ppf "%a" print_host_ident id
      | id :: idents ->
        Format.fprintf ppf "%a,@ %a"
          print_host_ident id
          loop idents in

    Format.fprintf ppf "@[%a@]" loop host_idents
  ;;

  let print_host_point ppf { sp_x = x; sp_y = y; } =
    Format.fprintf ppf "[%g, %g]" x y
  ;;

  let print_host_size ppf { ss_w = w; ss_h = h; } =
    Format.fprintf ppf "[%g, %g]" w h
  ;;

  let print_block_type ppf = function
    | Sbt s ->
      let txt = Ast_types.mdl_text_of_string s in
      print_text_as_encoded_string_token ppf txt
  ;;

  let print_host_color ppf = function
    | Sc { red = r; green = g; blue = b; } ->
      Format.fprintf ppf "[%g, %g, %g]" r g b
  ;;

  let print_host_port_value ppf = function
    | Ast_types.Mdl_port_number _ as port_number ->
      let txt =
        Ast_types.mdl_text_of_string
          (Ast_types.string_of_mdl_port_number port_number) in
      print_text_as_raw_string_token ppf txt
  ;;

  let print_host_directed_connector ppf (block_tag, port_value) =
    Format.fprintf ppf "%s%a, %a%s"
      Host.strs_begin
      print_host_block_tag_ident block_tag
      print_host_port_value port_value
      Host.strs_end
  ;;

  let print_host_directed_connector_src ppf = function
    | Dcs (block_tag, c) ->
      let port_value =
        c.Mdl_implementation_explicit_ast.explicit_connector_port_desc in
      Format.fprintf ppf "%a"
        print_host_directed_connector
        (block_tag, port_value)
  ;;

  let print_host_directed_connector_dst ppf = function
    | Dcd (block_tag, c) ->
      let port_value =
        c.Mdl_implementation_explicit_ast.explicit_connector_port_desc in
      Format.fprintf ppf "%a"
        print_host_directed_connector
        (block_tag, port_value)
  ;;

  let print_host_port_side ppf side =
    print_deep_value ppf (Deep_string side)
  ;;

  let print_host_connection_connector ppf (block_tag, port_value, side) =
    Format.fprintf ppf "%s%a, %a, %a%s"
      Host.strs_begin
      print_host_block_tag_ident block_tag
      print_host_port_value port_value
      print_host_port_side side
      Host.strs_end
  ;;

  let print_host_connection_connector ppf
      (block_tag, rename_connection_connector) =
    match rename_connection_connector with
    |  {
         Mdl_implementation_explicit_ast.
         explicit_connector_port_block = _block_name;
         explicit_connector_port_desc = rename_connection_connector_desc;
       } ->
(*
 c = rename_connection_connector
 port_type =
   c.Mdl_implementation_explicit_ast.explicit_connector_port_desc
*)
(* TODO write the code for get_port_value and get_side! *)
       let make_connection_port_number
           (Ast_types.Mdl_port_number port_number) offset =
         Ast_types.mdl_port_number_of_natural (
           Unatural.add port_number offset) in

       let get_port_value_port_side rename_connection_connector_desc =
         match rename_connection_connector_desc with
         | Mdl_rename_ast.
           Rename_left_connector ({
             Mdl_rename_ast.
             connector_block_nin = nin;
             connector_block_nout = _nout;
             connection_port_value =
               Mdl_implementation_explicit_ast.
               Connection_port_value port_number;
           } as _connection_port_info) ->
           make_connection_port_number port_number nin, "input"
         | Mdl_rename_ast.
           Rename_right_connector  ({
             Mdl_rename_ast.
             connector_block_nin = _nin;
             connector_block_nout = nout;
             connection_port_value =
               Mdl_implementation_explicit_ast.
               Connection_port_value port_number;
           } as _connection_port_info) ->
           make_connection_port_number port_number nout, "output" in

       let port_value, port_side =
         get_port_value_port_side rename_connection_connector_desc in

       Format.fprintf ppf "%a"
         print_host_connection_connector
         (block_tag, port_value, port_side)
  ;;

  let print_host_connection_connector_src ppf = function
    | Ccs (block_tag, rename_connection_connector) ->
      print_host_connection_connector ppf
        (block_tag, rename_connection_connector)
  ;;

  let print_host_connection_connector_dst ppf = function
    | Ccd (block_tag, rename_connection_connector) ->
      print_host_connection_connector ppf
        (block_tag, rename_connection_connector)
  ;;

  let print_host_position ppf {
      sp_xmin = xmin; sp_ymin = ymin;
      sp_xmax = xmax; sp_ymax = ymax;
    } =
    Format.fprintf ppf "[%g, %g, %g, %g]" xmin ymin xmax ymax
  ;;

  let print_host_mask_parameters ppf mask_parameters =

    let print_mask_parameter ppf =
      Format.fprintf ppf "@{<m>%a@}"
         (Host.print_mask_parameter
          print_mask_parameter_value) in

    let rec pr_mask_parameters ppf = function
      | [] -> ()
      | [ mp ] ->
        Format.fprintf ppf "%a" print_mask_parameter mp
      | mp :: mps ->
        Format.fprintf ppf "%a;@ %a"
          print_mask_parameter mp
          pr_mask_parameters mps in

    pr_mask_parameters ppf mask_parameters
  ;;

  let print_host_points ppf points =
    let print_point ppf { sp_x = x; sp_y = y; } =
      Format.fprintf ppf "%g, %g" x y in

    let rec pr_points ppf = function
      | [] -> ()
      | [ point ] -> Format.fprintf ppf "%a" print_point point
      | point :: points ->
        Format.fprintf ppf "%a; %a" print_point point pr_points points in

    Format.fprintf ppf "[%a]" pr_points points
  ;;

  let print_script pp_a pp_b ppf (a, b) =
    Format.fprintf ppf
      "@{<m>\
       @[<v 2>\
         %a =@ \
           %s%a%s\
       @]\
       @}"
      pp_a a
      Host.strs_begin
      pp_b b
      Host.strs_end
  ;;

  let print_model_script_binding mod_id bound_id
      host_binding_function ppf matlab_code =
    let bi = Hi bound_id in
    Format.fprintf ppf "%a;@ "
      (print_script print_host_ident print_matlab_script_as_value)
      (bi, matlab_code);
    Format.fprintf ppf "%a = %s (%a, %a)"
      print_host_model_ident mod_id
      host_binding_function
      print_host_model_ident mod_id
      print_host_ident bi
  ;;

  let print_scicos_simulation_parameters ppf
    {
      ssp_abs_tol = abs_tol;
      ssp_rel_tol = rel_tol;
      ssp_tim_tol = tim_tol;
      ssp_delta_t = delta_t;
      ssp_rt_scale = rt_scale;
      ssp_solver = solver;
      ssp_max_step = max_step;
    } =
    Format.fprintf ppf
      "%s \
       %a%s \
       %a%s \
       %a%s \
       %a%s \
       %a%s \
       %a%s \
       %a \
       %s"
      Host.strs_begin
      print_code_value abs_tol Host.row_separator
      print_code_value rel_tol Host.row_separator
      print_code_value tim_tol Host.row_separator
      print_code_value delta_t Host.row_separator
      print_code_value rt_scale Host.row_separator
      print_code_value solver Host.row_separator
      print_code_value max_step
      Host.strs_end
  ;;

  let is_comment_instruction = function
    | Comment _ -> true
    | (Add_block (_, _, _, _) |
       Add_event_link (_, _, _, _) | Add_directed_link (_, _, _, _) |
       Add_connection_link (_, _, _, _) |
       Define_function (_, _, _, _) |
       Evaluate_model (_, _) | Fill_super_block (_, _) |
       Instantiate_block (_, _) | Instantiate_super_block _ |
       Instantiate_diagram _ | Mask_block _ |
       Set_block_bg_color (_, _) | Set_block_fg_color (_, _) |
       Set_block_evtnin (_, _) | Set_block_evtnout (_, _) |
       Set_block_flip (_, _) |
       Set_block_icon (_, _) | Set_block_icon_text (_, _) |
       Set_block_ident (_, _) | Set_block_mask (_, _) |
       Set_block_nin (_, _) | Set_block_nout (_, _) |
       Set_block_origin (_, _) | Set_block_parameters (_, _) |
       Set_block_size (_, _) | Set_block_theta (_, _) |
       Set_diagram_3d (_, _) | Set_diagram_bg_color (_, _) |
       Set_diagram_context (_, _) | Set_diagram_position (_, _) |
       Set_diagram_name (_, _) | Set_diagram_zoom (_, _) |
       Set_function_call (_, _, _) | Set_model_close_fcn (_, _) |
       Set_model_final_simulation_time (_, _) | Set_model_init_fcn (_, _) |
       Set_model_initial_simulation_time (_, _) |
       Set_model_postload_fcn (_, _) |
       Set_model_postlude (_, _) | Set_model_postsave_fcn (_, _) |
       Set_model_preload_fcn (_, _) | Set_model_prelude (_, _) |
       Set_model_presave_fcn (_, _) | Set_model_start_fcn (_, _) |
       Set_model_stop_fcn (_, _) | Set_model_workspace (_, _) |
       Set_model_solver_parameters (_, _) |
       Set_super_block_atomic_property (_, _) |
       Set_target_modeler _) -> false
  ;;

  let is_multiline_instruction = function
    | (Comment _ |
       Add_block (_, _, _, _) |
       Add_event_link (_, _, _, _) | Add_directed_link (_, _, _, _) |
       Add_connection_link (_, _, _, _) |
       Fill_super_block (_, _) |
       Instantiate_block (_, _) | Instantiate_super_block _ |
       Instantiate_diagram _ |
       Set_block_bg_color (_, _) | Set_block_fg_color (_, _) |
       Set_block_evtnin (_, _) | Set_block_evtnout (_, _) |
       Set_block_flip (_, _) |
       Set_block_icon (_, _) | Set_block_icon_text (_, _) |
       Set_block_ident (_, _) |
       Set_block_nin (_, _) | Set_block_nout (_, _) |
       Set_block_origin (_, _) |
       Set_block_size (_, _) | Set_block_theta (_, _) |
       Set_diagram_3d (_, _) | Set_diagram_bg_color (_, _) |
       Set_diagram_position (_, _) |
       Set_diagram_name (_, _) | Set_diagram_zoom (_, _) |
       Set_function_call (_, _, _) |
       Set_model_final_simulation_time (_, _) |
       Set_model_initial_simulation_time (_, _) |
       Set_super_block_atomic_property (_, _) |
       Set_target_modeler _) -> false
    | Define_function (_, _, _, _) |
      Evaluate_model (_, _) |
      Set_diagram_context (_, _) |
      Mask_block _ |
      Set_block_mask (_, _) |
      Set_model_close_fcn (_, _) |
      Set_model_init_fcn (_, _) |
      Set_model_postload_fcn (_, _) |
      Set_model_postlude (_, _) | Set_model_postsave_fcn (_, _) |
      Set_model_preload_fcn (_, _) | Set_model_prelude (_, _) |
      Set_model_presave_fcn (_, _) | Set_model_start_fcn (_, _) |
      Set_model_stop_fcn (_, _) |
      Set_model_solver_parameters (_, _) |
      Set_model_workspace (_, _) |
      Set_block_parameters (_, _)
      -> true
  ;;

  (* Kind is "event" (for event links) or "explicit" (for regular links) *)
  let add_directed_link kind ppf scsm_id connector_src points connector_dst =
    Format.fprintf ppf "%a = add_%s_link (%a, %a, %a, %a)"
      print_host_diagram_ident scsm_id
      kind
      print_host_diagram_ident scsm_id
      print_host_directed_connector_src connector_src
      print_host_directed_connector_dst connector_dst
      print_host_points points
  ;;

  (* Kind is "implicit" (for all connection event links) *)
  let add_connection_link kind ppf scsm_id connector_src points connector_dst =
    Format.fprintf ppf "%a = add_%s_link (%a, %a, %a, %a)"
      print_host_diagram_ident scsm_id
      kind
      print_host_diagram_ident scsm_id
      print_host_connection_connector_src connector_src
      print_host_connection_connector_dst connector_dst
      print_host_points points
  ;;

  let print_host_target_modeler = Host.print_host_target_modeler;;

  let rec print_instruction ppf = function
    | Add_block (scsm_id, block_tag, block_id, block_name) ->
      Format.fprintf ppf "%s%a, %a%s = add_block (%a, %a, %a)"
        Host.lhs_begin
        print_host_diagram_ident scsm_id
        print_host_block_tag_ident block_tag
        Host.lhs_end
        print_host_diagram_ident scsm_id
        print_host_block_ident block_id
        print_block_name block_name;
    | Add_event_link (scsm_id, connector_src, connector_dst, points) ->
      add_directed_link "event" ppf
        scsm_id connector_src points connector_dst
    | Add_directed_link (scsm_id, connector_src, connector_dst, points) ->
      add_directed_link "explicit" ppf
        scsm_id connector_src points connector_dst
    | Add_connection_link (scsm_id, connector_src, connector_dst, points) ->
      add_connection_link "implicit" ppf
        scsm_id connector_src points connector_dst
    | Evaluate_model (mod_id, sci) ->
      Format.fprintf ppf
        ("@ @{<m>" ^^ Host.evaluate_model_format ^^ "@};@ ")
        print_host_model_ident mod_id
        print_host_model_ident mod_id;
      Format.fprintf ppf "@ @{<m>%a = %a@}"
        print_host_diagram_ident sci
        print_host_model_ident mod_id
    | Fill_super_block (block_id, scsm_id) ->
      Format.fprintf ppf "%a = fill_super_block (%a, %a)"
        print_host_block_ident block_id
        print_host_block_ident block_id
        print_host_diagram_ident scsm_id
    | Instantiate_block (block_id, block_type) ->
      Format.fprintf ppf "%a = instantiate_block (%a)"
        print_host_block_ident block_id
        print_block_type block_type
    | Instantiate_super_block block_id ->
      Format.fprintf ppf "%a = instantiate_super_block ()"
        print_host_block_ident block_id
    | Instantiate_diagram scsm_id ->
      Format.fprintf ppf
        "%a = instantiate_diagram ()"
        print_host_diagram_ident scsm_id
    | Mask_block (block_id, context, mask) ->
      let ctx = Hi "ctx" in
      Format.fprintf ppf "%a;@ "
        (print_script print_host_ident print_context)
          (ctx, context);
      Format.fprintf ppf "%s" Host.set_block_mask_header;
      begin
        match mask.Mdl_basics_explicit_ast.explicit_mask_parameters with
        | [] -> Format.fprintf ppf "@;"
        | _ :: _ ->
          Format.fprintf ppf "@;<1 2>@[<v>%a@];@ "
            print_host_mask_parameters mask.explicit_mask_parameters;
      end;
      Format.fprintf ppf "%a = mask_block (%a, %a, params, %a)"
        print_host_block_ident block_id
        print_host_block_ident block_id
        print_host_ident ctx
        print_mask_description
        mask.Mdl_basics_explicit_ast.explicit_mask_description
    | Set_block_bg_color (block_id, scicos_color) ->
      Format.fprintf ppf "%a = set_block_bg_color (%a, %a)"
        print_host_block_ident block_id
        print_host_block_ident block_id
        print_host_color scicos_color
    | Set_block_fg_color (block_id, scicos_color) ->
      Format.fprintf ppf "%a = set_block_fg_color (%a, %a)"
        print_host_block_ident block_id
        print_host_block_ident block_id
        print_host_color scicos_color
    | Set_block_evtnin (block_id, natural) ->
      Format.fprintf ppf
        "%a = set_block_evtnin (%a, %a)"
        print_host_block_ident block_id
        print_host_block_ident block_id
        Unatural_types_print.print_natural natural
    | Set_block_evtnout (block_id, natural) ->
      Format.fprintf ppf
        "%a = set_block_evtnout (%a, %a)"
        print_host_block_ident block_id
        print_host_block_ident block_id
        print_natural natural
    | Set_block_flip (block_id, bool) ->
      Format.fprintf ppf "%a = set_block_flip (%a, %a)"
        print_host_block_ident block_id
        print_host_block_ident block_id
        print_host_bool bool
    | Set_block_icon (block_id, matlab_code) ->
      Format.fprintf ppf "%a = set_block_icon (%a, %a)"
        print_host_block_ident block_id
        print_host_block_ident block_id
        print_matlab_script matlab_code
    | Set_block_icon_text (block_id, txt) ->
      Format.fprintf ppf "%a = set_block_icon_text (%a, %a)"
        print_host_block_ident block_id
        print_host_block_ident block_id
        print_block_icon_text txt
    | Set_block_ident (block_id, block_name) ->
      Format.fprintf ppf "%a = set_block_ident (%a, %a)"
        print_host_block_ident block_id
        print_host_block_ident block_id
        print_block_name block_name;
    | Set_block_mask (block_id, mask) ->
      Format.fprintf ppf "%s" Host.set_block_mask_header;
      begin
        match mask.Mdl_basics_explicit_ast.explicit_mask_parameters with
        | [] -> Format.fprintf ppf "@;"
        | _ :: _ ->
          Format.fprintf ppf "@;<1 2>@[<v>%a@];@ "
            print_host_mask_parameters mask.explicit_mask_parameters;
      end;
      Format.fprintf ppf "%a = set_block_mask (%a, params, %a)"
        print_host_block_ident block_id
        print_host_block_ident block_id
        print_mask_description
        mask.Mdl_basics_explicit_ast.explicit_mask_description
    | Set_block_nin (block_id, natural) ->
      Format.fprintf ppf "%a = set_block_nin (%a, %a)"
        print_host_block_ident block_id
        print_host_block_ident block_id
        print_natural natural
    | Set_block_nout (block_id, natural) ->
      Format.fprintf ppf "%a = set_block_nout (%a, %a)"
        print_host_block_ident block_id
        print_host_block_ident block_id
        print_natural natural
    | Set_block_origin (block_id, scicos_point) ->
      Format.fprintf ppf "%a = set_block_origin (%a, %a)"
        print_host_block_ident block_id
        print_host_block_ident block_id
        print_host_point scicos_point
    | Set_block_parameters (block_id, scicos_block_parameters) ->
      Format.fprintf ppf
        (Host.set_block_parameters_format ^^ ";@ ")
        (print_host_block_parameters Host.binding_format)
        scicos_block_parameters;
      Format.fprintf ppf "%a = set_block_parameters (%a, params)"
        print_host_block_ident block_id
        print_host_block_ident block_id
    | Set_block_size (block_id, scicos_size) ->
      Format.fprintf ppf "%a = set_block_size (%a, %a)"
        print_host_block_ident block_id
        print_host_block_ident block_id
        print_host_size scicos_size
    | Set_block_theta (block_id, float) ->
      Format.fprintf ppf "%a = set_block_theta (%a, %g)"
        print_host_block_ident block_id
        print_host_block_ident block_id
        float
    | Set_diagram_3d (scsm_id, bool) ->
      Format.fprintf ppf "%a = set_diagram_3d (%a, %a)"
        print_host_diagram_ident scsm_id
        print_host_diagram_ident scsm_id
        print_host_bool bool
    | Set_diagram_bg_color (scsm_id, scicos_color) ->
      Format.fprintf ppf "%a = set_diagram_bg_color (%a, %a)"
        print_host_diagram_ident scsm_id
        print_host_diagram_ident scsm_id
        print_host_color scicos_color
    | Set_diagram_context (scsm_id, context) ->
      let ctx = Hi "ctx" in
      Format.fprintf ppf "%a;@ "
        (print_script print_host_ident print_context)
          (ctx, context);
      Format.fprintf ppf "%a = set_diagram_context (%a, %a)"
        print_host_diagram_ident scsm_id
        print_host_diagram_ident scsm_id
        print_host_ident ctx
    | Set_diagram_position (scsm_id, scsm_loc) ->
      Format.fprintf ppf "%a = set_diagram_location (%a, %a)"
        print_host_diagram_ident scsm_id
        print_host_diagram_ident scsm_id
        print_host_position scsm_loc
    | Set_diagram_name (scsm_id, system_name) ->
      Format.fprintf ppf "%a = set_diagram_name (%a, %a)"
        print_host_diagram_ident scsm_id
        print_host_diagram_ident scsm_id
        print_diagram_name system_name
    | Set_diagram_zoom (scsm_id, zoom_factor) ->
      Format.fprintf ppf "%a = set_diagram_zoom (%a, %g)"
        print_host_diagram_ident scsm_id
        print_host_diagram_ident scsm_id
        zoom_factor
    | Set_model_initial_simulation_time (mod_id, sv) ->
      Format.fprintf ppf "%a = set_initial_time (%a, %a)"
        print_host_model_ident mod_id
        print_host_model_ident mod_id
        print_code_value sv
    | Set_model_final_simulation_time (mod_id, sv) ->
      Format.fprintf ppf "%a = set_final_time (%a, %a)"
        print_host_model_ident mod_id
        print_host_model_ident mod_id
        print_code_value sv
    | Set_model_init_fcn (mod_id, matlab_code) ->
      print_model_script_binding mod_id
        "init_fcn" "set_model_init_fcn" ppf matlab_code
    | Set_model_close_fcn (mod_id, matlab_code) ->
      print_model_script_binding mod_id
        "close_fcn" "set_model_close_fcn" ppf matlab_code
    | Set_model_postload_fcn (mod_id, matlab_code) ->
      print_model_script_binding mod_id
        "postload_fcn" "set_model_postload_fcn" ppf matlab_code
    | Set_model_preload_fcn (mod_id, matlab_code) ->
      print_model_script_binding mod_id
        "preload_fcn" "set_model_preload_fcn" ppf matlab_code
    | Set_model_postsave_fcn (mod_id, matlab_code) ->
      print_model_script_binding mod_id
        "postsave_fcn" "set_model_postsave_fcn" ppf matlab_code
    | Set_model_presave_fcn (mod_id, matlab_code) ->
      print_model_script_binding mod_id
        "presave_fcn" "set_model_presave_fcn" ppf matlab_code
    | Set_model_postlude (mod_id, matlab_code) ->
      print_model_script_binding mod_id
        "postlude" "set_model_postlude" ppf matlab_code
    | Set_model_prelude (mod_id, matlab_code) ->
      print_model_script_binding mod_id
        "prelude" "set_model_prelude" ppf matlab_code
    | Set_model_start_fcn (mod_id, matlab_code) ->
      print_model_script_binding mod_id
        "start_fcn" "set_model_start_fcn" ppf matlab_code
    | Set_model_stop_fcn (mod_id, matlab_code) ->
      print_model_script_binding mod_id
        "stop_fcn" "set_model_stop_fcn" ppf matlab_code
    | Set_model_workspace (mod_id, workspace) ->
      let wks = Hi "wks" in
      Format.fprintf ppf "%a;@ "
        (print_script print_host_ident print_workspace)
        (wks, workspace);
      Format.fprintf ppf "%a = set_model_workspace (%a, %a)"
        print_host_model_ident mod_id
        print_host_model_ident mod_id
        print_host_ident wks
    | Set_model_solver_parameters (mod_id, scicos_simulation_parameters) ->
      let tol = Hi "tol" in
      Format.fprintf ppf "@{<m>%a = %a@};@ "
        print_host_ident tol
        print_scicos_simulation_parameters scicos_simulation_parameters;
      Format.fprintf ppf "@{<m>%a = set_solver_parameters (%a, %a)@}"
        print_host_model_ident mod_id
        print_host_model_ident mod_id
        print_host_ident tol

    | Set_super_block_atomic_property (block_id, scicos_bool) ->
      Format.fprintf ppf "%a = set_atomic_property (%a, %a)"
        print_host_block_ident block_id
        print_host_block_ident block_id
        print_host_bool scicos_bool

    | Define_function (scsm_id, fun_id, host_idents, code) ->
      Format.fprintf ppf "@ function %a%a (%a)@ @;<1 2>@[<v>%a%a@]@ @ %s"
        (Host.print_define_function_lhs print_host_diagram_ident) scsm_id
        print_host_function_ident fun_id
        print_host_idents host_idents
        print_code code
        (Host.print_function_code_footer print_host_diagram_ident) scsm_id
        Host.function_footer
    | Set_function_call (id, fun_id, host_idents) ->
      Format.fprintf ppf
        "@[<v>%a = %a (%a)@]"
        print_host_diagram_ident id
        print_host_function_ident fun_id
        print_host_idents host_idents
    | Set_target_modeler modeler ->
      Format.fprintf ppf "set_target_modeler (%a)"
        print_host_target_modeler
        modeler
    | Comment s ->
      Format.fprintf ppf Host.comment_format s

  and print_multiline_instruction ppf instruct =
    if is_multiline_instruction instruct then
      Format.fprintf ppf "%a" print_instruction instruct
    else
      Format.fprintf ppf "@{<m>%a@}" print_instruction instruct

  and print_code ppf code =

    let rec pr_instructions ppf = function
      | [] -> ()
      | [ instruct ] ->
        print_multiline_instruction ppf instruct
      | instruct :: instructs ->
        let fmt =
          if is_comment_instruction instruct
          then format_of_string "%a@ %a"
          else format_of_string "%a;@ %a" in
        Format.fprintf ppf fmt
          print_multiline_instruction instruct
          pr_instructions instructs in

    let rec pr_code ppf = function
      | List l -> pr_instructions ppf l
      | Concat (c1, c2) ->
        Format.fprintf ppf "%a;@ %a"
          pr_code c1
          pr_code c2 in

    Format.fprintf ppf "@[<v>%a;@]" pr_code code
  ;;

  let print_translated_file ppf code =

    let banner =
      Comment (
        Printf.sprintf " Automatically generated by %s version %s"
          (Configuration.get_software_name ())
          (Configuration.get_software_full_version ()))
    and source_file =
      Comment (
        Printf.sprintf " from model source file %s"
          (Configuration.get_source_file ()))

    and set_target_modeler =
      let modeler =
        Configuration.get_target_modeler () in
      Set_target_modeler modeler in

    let code =
      Concat_list.cons banner
        (Concat_list.cons source_file
           (Concat_list.cons set_target_modeler code)) in

    Format.fprintf ppf "@{<n>@[<v>%a@]@}"
      print_code code
  ;;

  let print_translated_file =
    Mtlb_ppf.make_tag_printer Host.tnlm print_translated_file
  ;;

end
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
