(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: code_nsp_host.ml,v 1.1 2019-05-10 14:07:14 jpc Exp $ *)

(** The specific format strings and formatters for Nsp translation. *)

module Host : Code_make_host_print_spec.HOST = struct

  let tnlm = Mtlb_nsp_host.Host.tnlm;;

  (*
     Escape what should be escaped to translate a text into a valid string for
     the host language.
     (It turns the string text argument into a valid host language string
     lexem).
     For instance (for Nsp):
     - Caml string "toto \n'\"" (8 character long string) becomes
     the nsp string with contents (11 character long string)
     toto \n''"" i.e. the Nsp lexem "toto \n''"""
     i.e. the Caml string
     \"toto \n\'\'\"\"\" (12 character long string)
     - Caml string "\\" (1 character long) becomes the nsp string
     with contents \ i.e. the Nsp lexem "\"
     i.e. the Caml string
     \"\\\" (3 character long).
     " (to help the Caml parser.)
  *)
  let string_to_host_string text =
    (*prerr_endline (Printf.sprintf "string_to_host_string |%s|" text);*)
    let lim = String.length text - 1 in
    let accu = Buffer.create (lim + 3) in

    let rec loop i =
      if i <= lim then
        begin
          (match text.[i] with
           | '\"' -> Buffer.add_string accu "\"\""
           | '\'' -> Buffer.add_string accu "\'\'"
           | '\n' -> Buffer.add_string accu "\\n"
           | c -> Buffer.add_char accu c
          );
          loop (i + 1);
        end in

    Buffer.add_char accu '\"';
    loop 0;
    Buffer.add_char accu '\"';
    Buffer.contents accu
  ;;

  let comment_format = format_of_string "//%s";;

  let comment_string text =
    let text =
      Format.fprintf Format.str_formatter comment_format text;
      Format.flush_str_formatter () in
    string_to_host_string text
  ;;

  let print_host_target_modeler ppf modeler =
    Format.fprintf ppf "%s"
      (string_to_host_string
        (Configuration.string_of_target_modeler modeler))
  ;;

  let print_host_bool ppf b =
    let text = if b then "%t" else "%f" in
    Format.fprintf ppf "%s" text
  ;;

  let print_host_int ppf i =
    Format.fprintf ppf "%d" i
  ;;

  let print_host_float ppf f =
    Format.fprintf ppf "%f" f
  ;;

  let binding_format =
    format_of_string "\
      @[<hv 2>\
      params.concatd [@ \
        @[<hv 2>{@ \
          \"%s\",@ \
          %a@;<1 -2>\
        }\
        @]\
        @;<1 -2>\
      ]\
      @]\
      "
  ;;

  let commented_binding_format =
    format_of_string "%s = %a"
  ;;

  let print_mask_parameter print_mask_parameter_value ppf mp =
    match mp with
    | {
        Mdl_basics_explicit_ast.
        explicit_mask_parameter_ident = parameter_ident;
        explicit_mask_parameter_rank = _parameter_rank;
        explicit_mask_parameter_style = parameter_style;
        explicit_mask_parameter_prompt = prompt;
        explicit_mask_parameter_value = v;
        explicit_mask_parameter_evaluation = eval_mode;
        explicit_mask_parameter_enable = enable_mode_opt;
        explicit_mask_parameter_visible = visible_mode_opt;
      } ->
      let parameter_ident = Ast_types.string_of_ident parameter_ident
      and print_text = Mdl_translation.print_text_as_encoded_string_token
      and eval_mode = Mdl_translation.bool_of_evaluation_mode eval_mode
      and enable_mode_opt =
        Mdl_translation.bool_of_enable_mode_opt enable_mode_opt
      and visible_mode_opt =
        Mdl_translation.bool_of_visible_mode_opt visible_mode_opt
      and target_style =
        Mdl_translation.target_style_of_parameter_style parameter_style in

      Format.fprintf ppf
        "params.concatd [ { \"%s\", %a, %a, [ %a, %a, %a ], [ %a, [ %a ] ] } ]"
        parameter_ident print_mask_parameter_value v
        print_text prompt
        print_host_bool eval_mode
        print_host_bool enable_mode_opt
        print_host_bool visible_mode_opt
        print_text target_style
        Mdl_translation.print_parameter_style_items parameter_style
  ;;

  let evaluate_model_format =
    format_of_string "%a = evaluate_model (%a)"
  ;;

  let set_block_mask_header = "params = cell (0, 5);";;

  let set_block_parameters_format =
    format_of_string "\
      params = cell (0, 2);@;<1 2>\
        @[<v>%a@]"
  ;;

  let print_function_code_footer _print_host_diagram_ident _ppf = ignore;;

  let print_define_function_lhs print_host_diagram_ident ppf scsm_id =
    Format.fprintf ppf "%a = " print_host_diagram_ident scsm_id
  ;;

  let function_footer = "endfunction";;

  let lhs_begin = "[";;
  let lhs_end = "]";;

  let strs_begin = lhs_begin;;
  let strs_end = lhs_end;;

  let row_separator = ";";;

  let struct_item_format =
    format_of_string "\
      @[<hv 2>\
      %s =@ \
        @[<hv 2>\
        {@,\
          @[<hov>%a@]\
          @;<0 -2>\
        }\
        @]\
      @]"
  ;;

  let struct_create_format =
    format_of_string "\
      @[<hv 2>\
      hash (@,\
        @[<b>%a%a\
        @]@;<0 -2>\
      )\
      @]"
  ;;

end
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
