(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: code_scicoslab_host.ml,v 1.1 2019-05-10 14:07:14 jpc Exp $ *)

(** The specific format strings and formatters for Scicoslab translation. *)

module Host : Code_make_host_print_spec.HOST = struct

  let tnlm = Mtlb_scicoslab_host.Host.tnlm;;

  let string_to_host_string text =
    let lim = String.length text - 1 in
    let accu = Buffer.create (lim + 3) in

    let rec loop i =
      if i <= lim then
        begin
          (match text.[i] with
           | '\"' -> Buffer.add_string accu "\"\""
           | '\'' -> Buffer.add_string accu "\'\'"
           | '\n' -> Buffer.add_string accu "\\n"
           | c -> Buffer.add_char accu c
          );
          loop (i + 1);
        end in

    Buffer.add_char accu '\'';
    loop 0;
    Buffer.add_char accu '\'';
    Buffer.contents accu
  ;;

  let comment_format = Code_nsp_host.Host.comment_format;;
  let comment_string = Code_nsp_host.Host.comment_string;;

  let print_host_target_modeler ppf modeler =
    Format.fprintf ppf "%s"
      (string_to_host_string
        (Configuration.string_of_target_modeler modeler))
  ;;

  let print_host_bool ppf b =
    let s = if b then "1" else "0" in
    Format.fprintf ppf "%s" s
  ;;

  let print_host_int = Code_nsp_host.Host.print_host_int;;
  let print_host_float = Code_nsp_host.Host.print_host_float;;

  let print_host_rank = Unatural_print.print_t;;

  let binding_format =
    format_of_string "params.%s = %a"
  ;;

  let commented_binding_format =
    Code_nsp_host.Host.commented_binding_format
  ;;

  let print_mask_parameter print_mask_parameter_value ppf mp =
    match mp with
    | {
        Mdl_basics_explicit_ast.
        explicit_mask_parameter_ident = parameter_ident;
        explicit_mask_parameter_rank = parameter_rank;
        explicit_mask_parameter_style = parameter_style;
        explicit_mask_parameter_prompt = prompt;
        explicit_mask_parameter_value = v;
        explicit_mask_parameter_evaluation = eval_mode;
        explicit_mask_parameter_enable = enable_mode_opt;
        explicit_mask_parameter_visible = visible_mode_opt;
      } ->
      let parameter_ident = Ast_types.string_of_ident parameter_ident in

      let params = Printf.sprintf "params.%s" parameter_ident
      and print_text = Mdl_translation.print_text_as_encoded_string_token
      and eval_mode = Mdl_translation.bool_of_evaluation_mode eval_mode
      and enable_mode_opt =
        Mdl_translation.bool_of_enable_mode_opt enable_mode_opt
      and visible_mode_opt =
        Mdl_translation.bool_of_visible_mode_opt visible_mode_opt
      and target_style =
        Mdl_translation.target_style_of_parameter_style parameter_style in

      Format.fprintf ppf "\
        %s.rank = %a;@ \
        %s.value = %a;@ \
        %s.prompt = %a;@ \
        %s.evaluate = %a;@ \
        %s.enable = %a;@ \
        %s.visible = %a;@ \
        %s.parameter_style = [ %a, [ %a ] ]"

        params print_host_rank parameter_rank
        params print_mask_parameter_value v
        params print_text prompt
        params print_host_bool eval_mode
        params print_host_bool enable_mode_opt
        params print_host_bool visible_mode_opt
        params print_text target_style
        Mdl_translation.print_parameter_style_items parameter_style;
  ;;

  let evaluate_model_format =
    format_of_string "%a = evaluate_model (%a, %%scicos_context)"
  ;;

  let set_block_mask_header = "params = struct ();";;

  let set_block_parameters_format =
    format_of_string "params = struct ();@;<1 2>@[<v>%a@]"
  ;;

  let print_function_code_footer =
    Code_nsp_host.Host.print_function_code_footer
  ;;

  let print_define_function_lhs =
    Code_nsp_host.Host.print_define_function_lhs
  ;;

  let function_footer = Code_nsp_host.Host.function_footer;;
  let lhs_begin = Code_nsp_host.Host.lhs_begin;;
  let lhs_end = Code_nsp_host.Host.lhs_end;;

  let strs_begin = Code_nsp_host.Host.strs_begin;;
  let strs_end = Code_nsp_host.Host.strs_end;;

  let row_separator = Code_nsp_host.Host.row_separator;;

  let struct_item_format =
    format_of_string "\'%s\',@ {%a}"
  ;;

  let struct_create_format =
    format_of_string "struct (@[<hv>%a%a@])"
  ;;

end
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
