(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_ast_to_deep_ast.ml,v 1.1 2019-05-10 14:07:14 jpc Exp $ *)

(* From parse tree to deep syntax:
   - we perform various validity checks on the tree obtained from the parser,
   - we build explicit record representations of data instead of simple
     values with associated so-called property lists.
*)

open Ast_types;;
open Mdl_ast;;

let deep_error = Mdl_deep_basics.deep_error;;
let deep_warning = Mdl_deep_basics.deep_warning;;
let deep_bool_of_natural = Mdl_deep_basics.deep_bool_of_natural;;
let deep_binding = Mdl_deep_basics.deep_binding
and deep_bindings = Mdl_deep_basics.deep_bindings
;;
let natural_of_deep_value = Mdl_deep_basics.natural_of_deep_value;;
let deep_simple_model_value = Mdl_deep_basics.deep_simple_model_value;;
let make_deep_block_reference_type =
  Mdl_deep_basics.make_deep_block_reference_type
;;

type block_fields_classifier =
  {
    block_type_option : Mdl_block_types.block_type option;
    block_name_option : Ast_types.mdl_name option;
    block_sid_option : Mdl_ast.block_sid option;
    block_source_type_option : Mdl_ast.block_source_type option;
    block_source_path_option : Mdl_ast.block_source_path option;
    block_tag_option : Ast_types.mdl_text option;
    block_mask_option : Mdl_ast.mask_description option;
  }
;;

type block_array_system_classifier =
  {
    array_description_option : array_description option;
    system_option : system option;
    object_description_option : object_description option;
    block_mask_item_list : block_mask_item list;
    bindings_option : bindings option;
    block_sf_block_type_option : block_sf_block_type option;
    block_item_list : block_item list;
  }
;;

let empty_block_fields_classifier =
  {
    block_type_option = None;
    block_name_option = None;
    block_sid_option = None;
    block_source_type_option = None;
    block_source_path_option = None;
    block_tag_option = None;
    block_mask_option = None;
  }
and empty_block_array_system_classifier =
  {
    array_description_option = None;
    system_option = None;
    object_description_option = None;
    block_mask_item_list = [];
    bindings_option = None;
    block_sf_block_type_option = None;
    block_item_list = [];
  }
;;

let deep_points a =
  let deep_point v =
    let len = Array.length v in
    if len <> 2 then
      deep_error
        (Printf.sprintf "Invalid point definition %d" len) else
    {
      Mdl_basics_deep_ast.
      deep_x = v.(0);
      deep_y = v.(1);
    } in
  Array.to_list (Array.map deep_point a)
;;

let rec string_of_block_sid bl_sid =
  match bl_sid with
  | Sid_compound sc -> string_of_block_sid_compound sc
  | Sid_string s -> s

and string_of_block_sid_compound = function
  | Sid_natural n -> Unatural.to_string n
  | Sid_ref (n, sc) ->
    Printf.sprintf "%s:%s"
      (Unatural.to_string n) (string_of_block_sid_compound sc)
  | Sid_child (n, sc) ->
    Printf.sprintf "%s::%s"
      (Unatural.to_string n) (string_of_block_sid_compound sc)
;;

let string_of_block_sf_block_type = function
  | Block_sf_block_type_none -> "NONE"
  | Block_sf_block_type_matlab_function -> "MATLAB Function"
  | Block_sf_block_type_chart -> "Chart"
;;

let string_of_block_name = Ast_types.string_of_mdl_name;;
let string_of_block_type = Mdl_block_types_tables.string_of_block_type;;
let string_of_block_source_type = Ast_types.string_of_mdl_name;;
let string_of_block_source_path = Ast_types.string_of_mdl_path;;
let string_of_block_tag = Ast_types.string_of_mdl_text;;

let deep_port_enable = deep_bool_of_natural
and deep_port_trigger = deep_bool_of_natural
and deep_port_state = deep_bool_of_natural
and deep_port_ifaction = deep_bool_of_natural
;;

let deep_zorder z = int_of_string z;;

(** Regular ports *)
let get_number_of_regular_ports deep_ports_description =
  deep_ports_description.Mdl_basics_deep_ast.number_of_regular_ports

and set_number_of_regular_ports deep_ports_description number_of_regular_ports =
  {
    deep_ports_description with
    Mdl_basics_deep_ast.
    number_of_regular_ports = number_of_regular_ports;
  }
;;

let get_regular_port_nin deep_ports_description =
  (get_number_of_regular_ports deep_ports_description).
  Mdl_basics_deep_ast.port_nin
;;

let get_regular_port_nout deep_ports_description =
  (get_number_of_regular_ports deep_ports_description).
  Mdl_basics_deep_ast.port_nout

and set_regular_port_nin deep_ports_description port_nin =
  {
    deep_ports_description with
    Mdl_basics_deep_ast.
    number_of_regular_ports = {
      (get_number_of_regular_ports deep_ports_description) with
      Mdl_basics_deep_ast.
      port_nin = port_nin;
    };
  }
and set_regular_port_nout deep_ports_description port_nout =
  {
    deep_ports_description with
    Mdl_basics_deep_ast.
    number_of_regular_ports = {
      (get_number_of_regular_ports deep_ports_description) with
      Mdl_basics_deep_ast.
      port_nout = port_nout;
    };
  }
;;

(* Building deep_ports_description values. *)
let make_number_of_regular_ports nin nout =
  {
    Mdl_basics_deep_ast.
    port_nin = nin;
    port_nout = nout;
  }
;;

(* Building number of implicit ports. *)
let make_number_of_implicit_ports niin niout =
  {
    Mdl_basics_deep_ast.
    port_niin = niin;
    port_niout = niout;
  }
;;

(* Default regular port number is no ports at all. *)
let default_number_of_regular_ports =
  make_number_of_regular_ports Unatural._0_ Unatural._0_
;;

(* Default implicit port number is no ports at all. *)
let default_number_of_implicit_ports =
  make_number_of_implicit_ports Unatural._0_ Unatural._0_
;;

(** Implicit ports. *)

let get_number_of_implicit_ports deep_ports_description =
  deep_ports_description.Mdl_basics_deep_ast.number_of_implicit_ports

and set_number_of_implicit_ports
    deep_ports_description number_of_implicit_ports =
  {
    deep_ports_description with
    Mdl_basics_deep_ast.
    number_of_implicit_ports = number_of_implicit_ports;
  }
;;

let get_implicit_port_nin deep_ports_description =
  (get_number_of_implicit_ports deep_ports_description).
  Mdl_basics_deep_ast.port_niin
and get_implicit_port_nout deep_ports_description =
  (get_number_of_implicit_ports deep_ports_description).
  Mdl_basics_deep_ast.port_niout

and set_implicit_port_nin deep_ports_description port_nin =
  {
    deep_ports_description with
    Mdl_basics_deep_ast.
    number_of_implicit_ports = {
      (get_number_of_implicit_ports deep_ports_description) with
      Mdl_basics_deep_ast.
      port_niin = port_nin;
    };
  }
and set_implicit_port_nout deep_ports_description port_nout =
  {
    deep_ports_description with
    Mdl_basics_deep_ast.
    number_of_implicit_ports = {
      (get_number_of_implicit_ports deep_ports_description) with
      Mdl_basics_deep_ast.
      port_niout = port_nout;
    };
  }
;;

(* Default [default_deep_ports_description] is for passive blocks
   with provided default or regular and implicit port number. *)
let make_default_deep_ports_description
    number_of_regular_ports number_of_implicit_ports =
  {
    Mdl_basics_deep_ast.
    number_of_regular_ports = number_of_regular_ports;
    port_enable = false;
    port_trigger = false;
    port_ifaction = false;
    port_state = false;
    number_of_implicit_ports = number_of_implicit_ports;
  }
;;

(* Default [default_ports_description] is for passive blocks withtout regular
  ports, or an initial guess for a more complex block. *)
let default_deep_ports_description =
  make_default_deep_ports_description
    default_number_of_regular_ports
    default_number_of_implicit_ports
;;

(* If port_state is true (meaning flag ShowPortState is on)
   then we need an additional output port. *)
let correct_port_nout_for_port_state deep_ports_description =
  if not deep_ports_description.Mdl_basics_deep_ast.port_state
    then deep_ports_description else
  let nout =
    Unatural.succ (get_regular_port_nout deep_ports_description) in
  set_regular_port_nout deep_ports_description nout
;;

let deep_ports_description ports_description =

  let deep_ports_description =
    match ports_description with
    | [] -> default_deep_ports_description
    | port_nin :: [] ->
      set_regular_port_nin default_deep_ports_description port_nin
    | port_nin :: port_nout :: other_descriptions ->
      let deep_ports_description =
        set_number_of_regular_ports
          default_deep_ports_description
          (make_number_of_regular_ports port_nin port_nout) in
      match other_descriptions with
      | [] -> deep_ports_description
      | [ port_enable; ] ->
        {
          deep_ports_description with
          Mdl_basics_deep_ast.
          port_enable = deep_port_enable port_enable; }
      | [ port_enable; port_trigger; ] ->
        {
          deep_ports_description with
          Mdl_basics_deep_ast.
          port_enable = deep_port_enable port_enable;
          port_trigger = deep_port_trigger port_trigger;
        }
      | [ port_enable; port_trigger; port_state; ] ->
        {
          deep_ports_description with
          Mdl_basics_deep_ast.
          port_enable = deep_port_enable port_enable;
          port_trigger = deep_port_trigger port_trigger;
          port_state = deep_port_state port_state;
        }
      | [ port_enable; port_trigger; port_state;
          port_niin; ] ->
        {
          deep_ports_description with
          Mdl_basics_deep_ast.
          port_enable = deep_port_enable port_enable;
          port_trigger = deep_port_trigger port_trigger;
          port_state = deep_port_state port_state;
          number_of_implicit_ports = {
            deep_ports_description.
              Mdl_basics_deep_ast.number_of_implicit_ports with
            Mdl_basics_deep_ast.
            port_niin = port_niin;
          };
        }
      | [ port_enable; port_trigger; port_state;
          port_niin; port_niout; ] ->
        {
          deep_ports_description with
          Mdl_basics_deep_ast.
          port_enable = deep_port_enable port_enable;
          port_trigger = deep_port_trigger port_trigger;
          port_state = deep_port_state port_state;
          number_of_implicit_ports =
            make_number_of_implicit_ports port_niin port_niout;
        }
      | [ port_enable; port_trigger;
          port_state; port_niin; port_niout; port_ifaction; ] ->
        {
          deep_ports_description with
          Mdl_basics_deep_ast.
          port_enable = deep_port_enable port_enable;
          port_trigger = deep_port_trigger port_trigger;
          port_ifaction = deep_port_ifaction port_ifaction;
          port_state = deep_port_state port_state;
          number_of_implicit_ports =
            make_number_of_implicit_ports port_niin port_niout;
        }
      | _ -> assert false in
  correct_port_nout_for_port_state deep_ports_description
;;

let deep_state_item = Mdl_deep_basics.deep_simple_model_value
and deep_tag_text tag_text =
  let tag_string = string_of_mdl_text tag_text in
  Scanf.sscanf tag_string " Stateflow S-Function %s %d %!"
    (fun mdl_name deep_chart_file_number ->
     Ast_types.mdl_name_of_string mdl_name,
     Unatural.of_int deep_chart_file_number)

and deep_block_binding = deep_binding
;;

let deep_sf_block_type x = x;;

let deep_placement = function
  | "normal" -> Mdl_basics_deep_ast.Deep_placement_normal
  | "alternate" -> Mdl_basics_deep_ast.Deep_placement_alternate
  | s -> deep_error (Printf.sprintf "unknown placement \'%s\'" s)
;;

let deep_block_port_bindings bds =

  let rec loop accu = function
  | [] -> accu
  | Block_port_binding_array _array :: l ->
    deep_warning
      "Dropping an array description the bindings of a block port.";
    loop accu l
  | Block_port_binding_vendor _vendor_binding :: l ->
    deep_warning
      "Dropping a vendor binding in the bindings of a block port.";
    loop accu l
  | Block_port_binding binding :: l ->
    loop (deep_binding binding :: accu) l in

  loop [] bds
;;

let deep_block_port = function
  | {
      block_port_number = port_number;
      block_port_name = port_name_opt;
      block_port_bindings = block_port_bindings;
    } ->
    {
      Mdl_basics_deep_ast.
      deep_block_port_number = port_number;
      deep_block_port_name = port_name_opt;
      deep_block_port_bindings =
        deep_block_port_bindings block_port_bindings;
    }
;;

let rgb_color_of_string s =
  let ib = Scanf.Scanning.from_string s in
  Scanf.kscanf ib
    (fun _ _ ->
    (* We can not understand the color specification,
       either because we do not know the color name or
       becasue we can not parse its specification:
       we then emit a warning and go on with white. *)
     deep_warning
       (Printf.sprintf "unknown color \'%s\', using white" s);
     Mdl_basics_deep_ast.Deep_white
    )
    " [ %f , %f , %f ] "
    (fun r g b ->
     Mdl_basics_deep_ast.
     Deep_rgb { Mdl_basics_deep_ast.red = r; green = g; blue = b; })
;;

let deep_color = function
  | "red" -> Mdl_basics_deep_ast.Deep_red
  | "green" -> Mdl_basics_deep_ast.Deep_green
  | "blue" -> Mdl_basics_deep_ast.Deep_blue
  | "cyan" -> Mdl_basics_deep_ast.Deep_cyan
  | "magenta" -> Mdl_basics_deep_ast.Deep_magenta
  | "yellow" | "paleyellow" -> Mdl_basics_deep_ast.Deep_yellow
  | "white" -> Mdl_basics_deep_ast.Deep_white
  | "black" -> Mdl_basics_deep_ast.Deep_black
  | "grey" | "gray" -> Mdl_basics_deep_ast.Deep_grey
  | "orange" -> Mdl_basics_deep_ast.Deep_orange
  | "lightRed" -> Mdl_basics_deep_ast.Deep_lightred
  | "lightGreen" -> Mdl_basics_deep_ast.Deep_lightgreen
  | "lightBlue" -> Mdl_basics_deep_ast.Deep_lightblue
  | "lightGrey" | "lightGray" -> Mdl_basics_deep_ast.Deep_lightgrey
  | "darkRed" -> Mdl_basics_deep_ast.Deep_darkred
  | "darkGreen" -> Mdl_basics_deep_ast.Deep_darkgreen
  | "darkBlue" -> Mdl_basics_deep_ast.Deep_darkblue
  | "darkGrey" | "darkGray" -> Mdl_basics_deep_ast.Deep_darkgrey
  | "purple" -> Mdl_basics_deep_ast.Deep_purple
  | s ->
    Say.fdebug_call
      (Printf.sprintf "rgb_color_of_string %S" s)
      rgb_color_of_string s
;;

let deep_position = function
  | [ x; y; z; t; ] ->
    { Mdl_basics_deep_ast.
      xmin = x; ymin = y; xmax = z; ymax = t;
    }
  | l ->
    deep_error
      (Printf.sprintf
         "incorrect number of coordinates:\
       \n need 4 natural numbers, but got %d"
         (List.length l))
;;

let deep_annotation_position = function
  | [ x; y; z; t; ] ->
    { Mdl_basics_deep_ast.
      xpos = x; ypos = y;
      maxpos =
        Some {
          Mdl_basics_deep_ast.
          xmaxpos = z;
          ymaxpos = t;
        };
    }
  | [ x; y; ] ->
    { Mdl_basics_deep_ast.
      xpos = x; ypos = y; maxpos = None;
    }
  | l ->
    deep_error
      (Printf.sprintf
         "incorrect number of coordinates for annotation position:\
       \n need 2 natural numbers, but got %d"
         (List.length l))
;;

let deep_font_angle = function
  | "normal" -> Mdl_basics_deep_ast.Angle_normal
  | "italic" -> Mdl_basics_deep_ast.Angle_italic
  | "oblique" -> Mdl_basics_deep_ast.Angle_oblique
  | s -> deep_error (Printf.sprintf "unknown font angle \'%s\'" s)

and deep_font_weight = function
  | "normal" -> Mdl_basics_deep_ast.Weight_normal
  | "bold" -> Mdl_basics_deep_ast.Weight_bold
  | "light" -> Mdl_basics_deep_ast.Weight_light
  | s -> deep_error (Printf.sprintf "unknown font weight \'%s\'" s)
;;

let deep_font = function
  | {
      Ast_types.
      mdl_font_name = fn;
      mdl_font_angle = fa;
      mdl_font_size = fs;
      mdl_font_weight = fw;
    }  ->
    {
      Mdl_basics_deep_ast.
      deep_font_name = Some fn;
      deep_font_angle = Some (deep_font_angle fa);
      deep_font_size = Some fs;
      deep_font_weight = Some (deep_font_weight fw);
    }
;;

let deep_font_and_color = function
  | Background c ->
    Mdl_basics_deep_ast.Deep_background (deep_color c)
  | Drop_shadow mdl_bool ->
    Mdl_basics_deep_ast.Deep_drop_shadow (Ast_types.bool_of_mdl_bool mdl_bool)
  | Font_item _fit ->
    deep_error "deep_font_and_color: spurious font_item"
  | Foreground c ->
    Mdl_basics_deep_ast.Deep_foreground (deep_color c)
;;

let string_of_natural = Unatural.to_string;;

let deep_font_of_font_items fis =
  let df = {
    Mdl_basics_deep_ast.
    deep_font_name = None;
    deep_font_angle = None;
    deep_font_size = None;
    deep_font_weight = None;
  } in
  let rec loop df = function
  | [] -> df
  | fi :: fis ->
    match fi, df with
    | Font_angle fta,
      {
        Mdl_basics_deep_ast.
        deep_font_angle = None;
        deep_font_name = _;
        deep_font_size = _;
        deep_font_weight = _; } ->
      loop
        {
          df with
          Mdl_basics_deep_ast.
          deep_font_angle = Some (deep_font_angle fta);
        } fis
    | Font_angle fta, _ ->
      deep_error (
        Printf.sprintf
          "duplicated font angle specification (%s)"
          fta)

    | Font_name ftn,
      {
        Mdl_basics_deep_ast.
        deep_font_angle = _; deep_font_name = None;
        deep_font_size = _; deep_font_weight = _; } ->
      loop
        {
          df with
          Mdl_basics_deep_ast.
          deep_font_name = Some ftn;
        } fis
    | Font_name ftn, _ ->
      deep_error (
        Printf.sprintf
          "duplicated font name specification (%s)"
          ftn)

    | Font_size fts,
      {
        Mdl_basics_deep_ast.
        deep_font_angle = _; deep_font_name = _; deep_font_size = None;
        deep_font_weight = _;
      } ->
      loop { df with Mdl_basics_deep_ast.deep_font_size = Some fts; } fis
    | Font_size fts, _ ->
      deep_error (
        Printf.sprintf "duplicated font size specification (%s)"
        (string_of_natural fts))

    | Font_weight ftw,
      {
        Mdl_basics_deep_ast.
        deep_font_angle = _; deep_font_name = _; deep_font_size = _;
        deep_font_weight = None;
      } ->
      loop
        {
          df with
          Mdl_basics_deep_ast.
          deep_font_weight = Some (deep_font_weight ftw);
        } fis
    | Font_weight ftw, _ ->
      deep_error (
        Printf.sprintf "duplicated font weight specification (%s)"
        ftw) in
  loop df fis
;;

let deep_orientation = function
  | "right" -> Mdl_basics_deep_ast.Orientation_right
  | "left" -> Mdl_basics_deep_ast.Orientation_left
  | "up" -> Mdl_basics_deep_ast.Orientation_up
  | "down" -> Mdl_basics_deep_ast.Orientation_down
  | s -> deep_error (Printf.sprintf "unknown orientation \'%s\'" s)
;;

let from_option name = function
  | Some v -> v
  | None -> deep_error (name ^ " required but not provided")
;;

let classify_block_fields blis =

  (* Each set_\(.*\) function here checks $1 w.r.t $1 option to verify
     that $1 option is only set once. If $1 option is set twice the function
     emits an error, excpet for set_block_source_type that emits a warning:
     this option may occur more than once (see clause Block_source_type
     above)!
  *)
  let fail_defined_twice message appendix =
    deep_error
      (Printf.sprintf "block %s defined twice%s"
         message
         appendix)in
  let set_option message string_of v = function
    | None -> Some v
    | Some v_opt ->
      let appendix =
        Printf.sprintf " as \'%s\' and \'%s\'"
          (string_of v_opt)
          (string_of v) in
      fail_defined_twice message appendix in
   let set_block_type = set_option "type" string_of_block_type
   and set_block_name = set_option "name" string_of_block_name
   and set_block_sid = set_option "sid" string_of_block_sid
   and set_block_source_type bl_src_type = function
     | None -> Some bl_src_type
     | Some bl_ty as opt ->
       deep_warning
         (Printf.sprintf
            "block source type defined twice as \'%s\' and \'%s\'"
            (string_of_block_source_type bl_ty)
            (string_of_block_source_type bl_src_type));
       opt
   and set_block_source_path =
     set_option "source path" string_of_block_source_path
   and set_block_tag = set_option "tag" string_of_block_tag
   and set_block_mask mask = function
     | None -> Some mask
     | Some _ -> fail_defined_twice "mask" "" in

  let rec loop block_fields_classifier accu = function
  | [] -> block_fields_classifier, accu
  | bli :: blis ->
    let block_fields_classifier, accu =
      match bli with
      | Block_type bl_type ->
        { block_fields_classifier with
          block_type_option =
            set_block_type
              bl_type block_fields_classifier.block_type_option;
        }, accu
      | Block_name bl_name ->
        { block_fields_classifier with
          block_name_option =
            set_block_name
              bl_name block_fields_classifier.block_name_option;
        }, accu
      | Block_sid bl_sid ->
        { block_fields_classifier with
          block_sid_option =
            set_block_sid
              bl_sid block_fields_classifier.block_sid_option;
        }, accu
      | Block_source_type bl_src_type ->
        (* If Block source type is specified twice,
           it is NOT an error :
           the first occurrence indeed defines the block source type,
           while the second occurrence defines a binding for the block.
           What a strange and error prone semantics! *)
        let src_ty_opt =
          block_fields_classifier.block_source_type_option in
        let src_block_type_opt =
          set_block_source_type bl_src_type src_ty_opt in
        let accu =
          match src_ty_opt with
          | None -> accu
          | Some _ ->
            let make_source_type_block_binding bl_src_type =
              let Mdl_name bl_src_type = bl_src_type in
              Block_binding {
                Binding.bound_ident = Ident.Ident "SourceType";
                bound_value = Value_simple (String bl_src_type);
              } in
            make_source_type_block_binding bl_src_type :: accu in
        let block_fields_classifier =
          { block_fields_classifier with
            block_source_type_option = src_block_type_opt;
          } in
        block_fields_classifier, accu
      | Block_source_path bl_source_path ->
        { block_fields_classifier with
          block_source_path_option =
            set_block_source_path
              bl_source_path
              block_fields_classifier.block_source_path_option;
        }, accu
      | Block_tag bl_tag ->
        { block_fields_classifier with
          block_tag_option =
            set_block_tag
              bl_tag
              block_fields_classifier.block_tag_option;
        }, accu
      | Block_mask bl_mask ->
        { block_fields_classifier with
          block_mask_option =
            set_block_mask
              bl_mask
              block_fields_classifier.block_mask_option;
        }, accu
      | Block_array _
      | Block_object _
      | Block_system _
      | Block_port _
      | Block_ports_description _
      | Block_position _
      | Block_state _
      | Block_sf_block_type _
      | Block_graphical_default _
      | Block_mask_item _
      | Block_vendor_binding _
      | Block_binding _
      | Block_instance_data _ as bli ->
        block_fields_classifier, bli :: accu in
    loop block_fields_classifier accu blis in

  loop empty_block_fields_classifier [] blis
;;

let deep_model_block_graphical_defaults bds =

  let rec loop (font_items as fis) accu = function
  | [] ->
    let deep_font_and_color =
      Mdl_basics_deep_ast.Deep_font (deep_font_of_font_items font_items) in
    Mdl_basics_deep_ast.
    Deep_block_font_and_color deep_font_and_color :: accu
  | bd :: bds ->
    match bd with
    | Block_mirror mdl_bool ->
      loop fis
        (Mdl_basics_deep_ast.
         Deep_block_mirror (bool_of_mdl_bool mdl_bool) :: accu)
        bds
    | Block_name_placement string ->
      loop fis
        (Mdl_basics_deep_ast.
         Deep_block_name_placement (deep_placement string) :: accu)
        bds
    | Block_orientation string ->
      loop fis
        (Mdl_basics_deep_ast.
         Deep_block_orientation (deep_orientation string) :: accu)
        bds
    | Block_rotation natural ->
      loop fis (Mdl_basics_deep_ast.Deep_block_rotation natural :: accu) bds
    | Block_show_name mdl_bool ->
      loop fis
        (Mdl_basics_deep_ast.
         Deep_block_show_name (bool_of_mdl_bool mdl_bool) :: accu) bds
    | Block_hide_automatic_name mdl_bool ->
      loop fis
        (Mdl_basics_deep_ast.
         Deep_block_hide_automatic_name
           (bool_of_mdl_bool mdl_bool) :: accu) bds
    | Block_zorder zo ->
      loop fis
        (Mdl_basics_deep_ast.Deep_block_zorder (deep_zorder zo) :: accu) bds
    | Block_font_and_color fc ->
      begin match fc with
      | Font_item fitem ->
        loop (fitem :: fis) accu bds
      | Background _ | Drop_shadow _ | Foreground _ as fc ->
        loop fis
          (Mdl_basics_deep_ast.
           Deep_block_font_and_color (deep_font_and_color fc) :: accu)
          bds
      end in

  loop [] [] bds
;;

let deep_block_graphical_default bl_graphical_default =
  match bl_graphical_default with
  | Block_mirror mdl_bool ->
    Mdl_basics_deep_ast.Deep_block_mirror (bool_of_mdl_bool mdl_bool)
  | Block_name_placement string ->
    Mdl_basics_deep_ast.Deep_block_name_placement (deep_placement string)
  | Block_orientation string ->
    Mdl_basics_deep_ast.Deep_block_orientation (deep_orientation string)
  | Block_rotation natural ->
    Mdl_basics_deep_ast.Deep_block_rotation natural
  | Block_show_name mdl_bool ->
    Mdl_basics_deep_ast.Deep_block_show_name (bool_of_mdl_bool mdl_bool)
  | Block_hide_automatic_name mdl_bool ->
    Mdl_basics_deep_ast.Deep_block_hide_automatic_name (bool_of_mdl_bool mdl_bool)
  | Block_zorder zo ->
    Mdl_basics_deep_ast.Deep_block_zorder (deep_zorder zo)
  | Block_font_and_color fc ->
    begin match fc with
    | Font_item _ ->
      deep_error "deep_block_graphical_default: spurious font item"
    | Background _ | Drop_shadow _ | Foreground _ as fc ->
      Mdl_basics_deep_ast.Deep_block_font_and_color (deep_font_and_color fc)
    end
;;

let deep_block_items blis =

  let rec loop (font_items as fis) port_bodies accu = function
  | [] ->
    let accu =
      if port_bodies <> [] then
        Mdl_basics_deep_ast.
        Deep_block_graphical_default (
          Mdl_basics_deep_ast.Deep_block_ports port_bodies) :: accu
      else accu in
    let deep_font_and_color =
      Mdl_basics_deep_ast.Deep_font (deep_font_of_font_items font_items) in
    Mdl_basics_deep_ast.
    Deep_block_graphical_default
      (Mdl_basics_deep_ast.
       Deep_block_font_and_color deep_font_and_color) ::
    accu

  | bli :: blis ->
    match bli with

    (* Filtering out erroneous block items *)
    | Block_type bl_type ->
      deep_error (
        Printf.sprintf "spurious block type \'%s\'"
          (string_of_block_type bl_type))
    | Block_name bl_name ->
      deep_error (
        Printf.sprintf "spurious block name \'%s\'"
          (string_of_block_name bl_name))
    | Block_source_path bl_source_path ->
      deep_error (
        Printf.sprintf "spurious source block path \'%s\'"
          (string_of_block_source_path bl_source_path))
    | Block_source_type bl_source_type ->
      deep_error (
        Printf.sprintf "spurious block type \'%s\'"
          (string_of_block_source_type bl_source_type))
    | Block_array _bl_array_description ->
      deep_error
        "spurious array description in block parameter defaults"
    | Block_object _ ->
      deep_error
        "spurious object description in block parameter defaults"
    | Block_system _bl_system ->
      deep_error
        "spurious system section in block parameter defaults"
    | Block_mask_item _bm_item ->
      deep_error
        "spurious mask item in block parameter defaults"
    | Block_mask _bm ->
      deep_error
        "spurious mask in block parameter defaults"
    | Block_instance_data _bl_instance_data ->
      deep_error
        "spurious instance data in block parameter defaults"
    | Block_tag _bl_tag_text ->
      deep_error
        "spurious block tag in block parameter defaults"
    | Block_sf_block_type block_sf_block_type ->
      deep_error (
        Printf.sprintf
          "spurious sf block type \'%s\' in block parameter defaults"
          (string_of_block_sf_block_type block_sf_block_type))

    (* Filtering items to the appropriate item category *)
    | Block_port bl_port ->
      loop fis (deep_block_port bl_port :: port_bodies) accu blis
    | Block_ports_description bl_ports_description ->
      loop fis port_bodies
        (Mdl_basics_deep_ast.
         Deep_block_graphical_default
           (Mdl_basics_deep_ast.
            Deep_block_ports_description
              (deep_ports_description bl_ports_description)) :: accu)
        blis
    | Block_position bl_pos ->
      loop fis port_bodies
        (Mdl_basics_deep_ast.
         Deep_block_position (deep_position bl_pos) :: accu)
        blis
    | Block_sid sid ->
      loop fis port_bodies
        (Mdl_basics_deep_ast.
         Deep_block_sid sid :: accu)
        blis
    | Block_state bl_simple_model_value ->
      loop fis port_bodies
        (Mdl_basics_deep_ast.
         Deep_block_state (deep_state_item bl_simple_model_value) :: accu)
        blis
    | Block_graphical_default bl_graphical_default ->
      begin match bl_graphical_default with
      | Block_font_and_color fc ->
        begin match fc with
        | Font_item fitem ->
          loop (fitem :: fis) port_bodies accu blis
        | Background _ | Drop_shadow _ | Foreground _ ->
          loop fis port_bodies
            (Mdl_basics_deep_ast.
             Deep_block_graphical_default
               (deep_block_graphical_default bl_graphical_default) :: accu)
            blis
        end
      | Block_mirror _
      | Block_name_placement _
      | Block_orientation _
      | Block_rotation _
      | Block_show_name _
      | Block_hide_automatic_name _
      | Block_zorder _ as bl_graphical_default ->
        loop fis port_bodies
          (Mdl_basics_deep_ast.
           Deep_block_graphical_default
             (deep_block_graphical_default bl_graphical_default) :: accu)
          blis
      end

    (* Do not know how to treat vendor bindings *)
    | Block_vendor_binding _vendor_binding ->
      Say.warning (Printf.sprintf "discarding block vendor binding item");
      loop fis port_bodies accu blis

    | Block_binding bl_binding ->
      loop fis port_bodies
        (Mdl_basics_deep_ast.
         Deep_block_binding
           (deep_block_binding bl_binding) :: accu)
        blis in

  loop [] [] [] blis
;;

(* From the unclassified block item list description of a block, extract
   an optional array description,
   an optional (sub)-system description,
   an optional object description,
   an optional mask item list description for the block.
   Other block items are recorded as is. *)
let classify_block_array_system blis =

  let rec loop classifier = function
  | [] -> classifier
  | bli :: blis ->
    let classifier =
      match bli with
      | Block_array bl_array ->
        {
          classifier with
          array_description_option =
            set_array bl_array classifier.array_description_option;
        }
      | Block_system bl_system ->
        {
          classifier with
          system_option =
            set_system bl_system classifier.system_option;
        }
      | Block_object bl_obj ->
        {
          classifier with
          object_description_option =
            set_object bl_obj classifier.object_description_option;
        }
      | Block_mask_item mask_item ->
        {
          classifier with
          block_mask_item_list =
            mask_item :: classifier.block_mask_item_list;
        }
      | Block_instance_data instance_data ->
        {
          classifier with
          bindings_option =
            set_instance_data instance_data classifier.bindings_option;
        }
      | Block_sf_block_type sf_block_type ->
        {
          classifier with
          block_sf_block_type_option =
            set_sf_block_type
              sf_block_type classifier.block_sf_block_type_option;
        }

      | Block_type _
      | Block_name _
      | Block_source_path _
      | Block_source_type _
      | Block_port _
      | Block_ports_description _
      | Block_position _
      | Block_sid _
      | Block_state _
      | Block_tag _
      | Block_graphical_default _
      | Block_vendor_binding _
      | Block_mask _
      | Block_binding _ as bli ->
        {
          classifier with
          block_item_list = bli :: classifier.block_item_list;
        } in
    loop classifier blis

  and set_array bl_array = function
    | None -> Some bl_array
    | Some _bl_ar -> deep_error "block array defined twice"
  and set_system bl_system = function
    | None -> Some bl_system
    | Some _bl_sy -> deep_error "block system defined twice"
  and set_object bl_object = function
    | None -> Some bl_object
    | Some _bl_sy -> deep_error "block object defined twice"
  and set_instance_data bl_instance_data = function
    | None -> Some bl_instance_data
    | Some _bl_instance_data ->
      deep_error "block instance_data defined twice"
  and set_sf_block_type sf_block_type = function
    | None -> Some sf_block_type
    | Some _sf_block_type -> deep_error "sf block type defined twice" in
  loop empty_block_array_system_classifier blis
;;

let rec deep_array_description = function
  | {
      array_type = ty;
      array_dimension = d;
      array_items = array_items;
      array_prop_name = prop_name;
    } ->
    let d =
      natural_of_deep_value
        "array dimension is not a natural"
        (deep_simple_model_value d) in
    let array_items =
      Array.of_list (List.map deep_array_item array_items) in
    let l = Unatural_types.to_natural (Array.length array_items) in
    if d <> l then
      deep_error (Printf.sprintf
        "wrong array length: %d items within array of length %d"
        (Unatural_types.of_natural l) (Unatural_types.of_natural d));
    {
      Mdl_deep_ast.
      deep_array_type = ty;
      deep_array_items = array_items;
      deep_array_prop_name = prop_name;
    }

and deep_array_item = function
  | Array_name array_name ->
    Mdl_deep_ast.
    Deep_array_name array_name
  | Array_sub_array array_description ->
    Mdl_deep_ast.
    Deep_array_sub_array (deep_array_description array_description)
  | Array_custom_property array_custom_property ->
    Mdl_deep_ast.
    Deep_array_custom_property array_custom_property
  | Array_matstruct matstruct_items ->
    Mdl_deep_ast.
    Deep_array_matstruct matstruct_items
  | Array_object object_description ->
    Mdl_deep_ast.
    Deep_array_object (deep_object_description object_description)
  | Array_vendor_binding vendor_binding ->
    Mdl_deep_ast.
    Deep_array_vendor_binding vendor_binding
  | Array_vendor_configset (vendor_name, vendor_items) ->
    Mdl_deep_ast.
    Deep_array_vendor_configset (vendor_name, vendor_items)
  | Array_binding binding ->
    Mdl_deep_ast.
    Deep_array_binding (deep_binding binding)

and deep_object_items items = List.map deep_object_item items
and deep_object_item = function
  | Object_type object_type ->
    Mdl_deep_ast.
    Deep_object_type object_type
  | Object_name object_name ->
    Mdl_deep_ast.
    Deep_object_name object_name
  | Object_dollar_binding dollar_binding ->
    Mdl_deep_ast.
    Deep_object_dollar_binding dollar_binding
  | Object_array array_description ->
    Mdl_deep_ast.
    Deep_object_array (deep_array_description array_description)
  | Object_custom_property custom_property ->
    Mdl_deep_ast.
    Deep_object_custom_property custom_property
  | Object_object object_description ->
    Mdl_deep_ast.
    Deep_object_object (deep_object_description object_description)
  | Object_binding binding ->
    Mdl_deep_ast.
    Deep_object_binding (deep_binding binding)

and deep_object_description items =
  let rec classify accu = function
    | idopt, classopt, propopt, [] -> idopt, classopt, propopt, List.rev accu
    | idopt, classopt, propopt,
      Object_dollar_binding (Dollar_object_id id) :: items ->
      if idopt <> None then
        Say.fatal_error "object id multiply defined" else
      classify accu (Some id, classopt, propopt, items)
    | idopt, classopt, propopt,
      Object_dollar_binding (Dollar_class_name class_name) :: items ->
      if classopt <> None then
        Say.fatal_error "object class name multiply defined" else
      classify accu (idopt, Some class_name, propopt, items)
    | idopt, classopt, propopt,
      Object_dollar_binding (Dollar_prop_name prop_name) :: items ->
      if propopt <> None then
        Say.fatal_error "object prop name multiply defined" else
      classify accu (idopt, classopt, Some prop_name, items)
    | idopt, classopt, propopt,
      (Object_dollar_binding (Dollar_backup_class _) as item) :: items ->
      classify (item :: accu) (idopt, classopt, propopt, items)
    | idopt, classopt, propopt,
      (Object_object _ | Object_binding _ |
       Object_array _ | Object_custom_property _ |
       Object_type _ | Object_name _ as item) :: items ->
      classify (item :: accu) (idopt, classopt, propopt, items) in
  match classify [] (None, None, None, items) with
  | Some object_id, class_name_option, prop_name_option, items ->
    let deep_items = deep_object_items items in
    {
      Mdl_deep_ast.
      deep_object_id = object_id;
      deep_object_class_name = class_name_option;
      deep_object_prop_name = prop_name_option;
      deep_object_items = deep_items;
    }
  | None, _, _, _items ->
    Say.fatal_error "object without object id"
;;

let mask_object_ident = ident_from_string "MaskObject";;

let is_deep_object_a_mask = function
  | {
      Mdl_deep_ast.
      deep_object_id = _object_id;
      deep_object_class_name = _object_class_name_opt;
      deep_object_prop_name = Some prop_name_ident;
      deep_object_items = _object_items;
    } ->  prop_name_ident = mask_object_ident
  | _ -> false
;;

let rec is_mask_items_a_mask = function
  | [] -> false
  | (Mask_display _
    | Mask_value_string _) :: _ -> true
  | (Mask_callback_string _
    | Mask_capabilities _
    | Mask_description _
    | Mask_self_modifiable _
    | Mask_display_string _
    | Mask_enable_string _
    | Mask_help _
    | Mask_style_string _
    | Mask_hide_contents _
    | Mask_icon_frame _
    | Mask_icon_opaque _
    | Mask_icon_rotate _
    | Mask_tab_name_string _
    | Mask_icon_units _
    | Mask_initialization _
    | Mask_port_rotate _
    | Mask_tooltip_string _
    | Mask_prompt_string _
    | Mask_run_init_for_icon_redraw _
    | Mask_tunable_value_string _
    | Mask_type _
    | Mask_var_alias_string _
    | Mask_variables _
    | Mask_visibility_string _) :: block_mask_items ->
    is_mask_items_a_mask block_mask_items
;;

let deep_block_body_elements deep_system blis =

  match classify_block_array_system blis with
  | {
      array_description_option = ar_opt;
      system_option = sys_opt;
      object_description_option = obj_opt;
      block_mask_item_list = mask_items;
      bindings_option = inst_data_opt;
      block_sf_block_type_option = sf_block_type_opt;
      block_item_list = blis;
    } ->

    let deep_ar_opt =
      match ar_opt with
      | None -> None
      | Some bl_array ->
        Some (
          Mdl_deep_ast.
          Deep_block_array (deep_array_description bl_array)) in

    let deep_sys_opt =
      match sys_opt with
      | None -> None
      | Some bl_system ->
        Some (deep_system bl_system) in

    let deep_obj_opt =
      match obj_opt with
      | None -> None
      | Some object_items ->
        Some (deep_object_description object_items) in

    let mask_by_items = is_mask_items_a_mask mask_items in

    let mask_by_deep_object =
      match deep_obj_opt with
      | None -> false
      | Some deep_object ->
        is_deep_object_a_mask deep_object in

    let deep_mask_opt =
      if mask_by_items then
        Some
          (Mdl_deep_ast.
           Deep_block_mask
            (Mdl_deep_mask.deep_mask_of_mask_items mask_items)) else
      if mask_by_deep_object then
        let bl_deep_object = from_option "block mask object" deep_obj_opt in
        Some
          (Mdl_deep_ast.
           Deep_block_mask
            (Mdl_deep_mask.deep_mask_of_mask_object bl_deep_object)) else
      None in

    let deep_obj_opt =
      if mask_by_deep_object then None else
      match deep_obj_opt with
      | None -> None
      | Some deep_bl_object ->
        Some
          (Mdl_deep_ast.
           Deep_block_object (deep_bl_object)) in

    let deep_inst_data_opt =
      match inst_data_opt with
      | None -> None
      | Some instance_data ->
        Some
          (Mdl_deep_ast.
           Deep_block_instance_data (deep_bindings instance_data)) in

    let deep_sf_block_type_opt =
      match sf_block_type_opt with
      | None -> None
      | Some sf_block_type ->
        Some
          (Mdl_deep_ast.
           Deep_block_sf_block_type (deep_sf_block_type sf_block_type)) in

    let deep_block_bindings = function
      | [] -> []
      | blis ->
        [ Mdl_deep_ast.
          Deep_block_bindings blis ] in

    let bl_bindings =
      deep_block_bindings (deep_block_items blis) in

    let add_option l = function
      | None -> l
      | Some x -> x :: l in

    deep_sys_opt,
    List.fold_left add_option bl_bindings [
      deep_ar_opt;
      deep_obj_opt;
      deep_mask_opt;
      deep_inst_data_opt;
      deep_sf_block_type_opt;
    ]
;;

(* Computes a deep block from the block item list description of a block *)
let deep_block deep_system blis =
  let (bl_type, bl_name, bl_sid_opt, bl_ref_ty_opt, bl_tag_opt, bl_msk_opt,
       blis) =

    (* Each get_\(.*\) function here extracts $1 from a kind of $1 option *)
    let get_block_type_name block_fields_classifier =
      match
        block_fields_classifier.block_type_option,
        block_fields_classifier.block_name_option
      with
      | (Some bl_type, Some bl_name) -> bl_type, bl_name
      | (None, _) ->
        deep_error "missing block type"
      | (_, None) ->
        deep_error "missing block name"
    and get_block_reference_type_opt bl_name block_fields_classifier =
      match
        block_fields_classifier.block_source_type_option,
        block_fields_classifier.block_source_path_option
      with
      | (None, None) -> None
      | (Some src_ty, Some src_path) ->
        Some (src_ty, src_path)
      | (None, _) ->
        deep_error (Printf.sprintf
          "spurious block source type in regular block \'%s\'"
          (string_of_block_name bl_name))
      | (_, None) ->
        deep_error (Printf.sprintf
          "spurious block source path in regular block \'%s\'"
          (string_of_block_name bl_name)) in

    let block_fields_classifier, blis = classify_block_fields blis in

    let bl_type, bl_name = get_block_type_name block_fields_classifier in
    let bl_ref_type_opt =
      get_block_reference_type_opt bl_name block_fields_classifier in
    (bl_type,
     bl_name,
     block_fields_classifier.block_sid_option,
     bl_ref_type_opt,
     block_fields_classifier.block_tag_option,
     block_fields_classifier.block_mask_option,
     blis) in

  let bl_deep_sys_opt, bl_body_elements =
    deep_block_body_elements deep_system blis in

  let bl_body_elements =
    match bl_msk_opt with
    | None -> bl_body_elements
    | Some bl_msk ->
      let deep_mask =
        Mdl_deep_mask.deep_mask_of_mask_decription bl_msk in
      Mdl_deep_ast.Deep_block_mask deep_mask :: bl_body_elements in

  let make_deep_block_regular bl_type =
    let bl_desc =
      Mdl_deep_ast.
      Deep_block_regular bl_type in
    {
      Mdl_deep_ast.
      deep_block_name = bl_name;
      deep_block_sid = bl_sid_opt;
      deep_block_body = bl_body_elements;
      deep_block_desc = bl_desc;
    } in

  let make_deep_block_reference bl_reference_type =
    let bl_desc =
      Mdl_deep_ast.
      Deep_block_reference bl_reference_type in
    {
      Mdl_deep_ast.
      deep_block_name = bl_name;
      deep_block_sid = bl_sid_opt;
      deep_block_body = bl_body_elements;
      deep_block_desc = bl_desc;
    } in

  let make_deep_block_sub_system bl_deep_sys =
    let bl_desc =
      Mdl_deep_ast.
      Deep_block_sub_system bl_deep_sys in
    {
      Mdl_deep_ast.
      deep_block_name = bl_name;
      deep_block_sid = bl_sid_opt;
      deep_block_body = bl_body_elements;
      deep_block_desc = bl_desc;
    } in

  let make_deep_block_s_function bl_s_fun =
    let bl_desc =
      Mdl_deep_ast.
      Deep_block_s_function bl_s_fun in
    {
      Mdl_deep_ast.
      deep_block_name = bl_name;
      deep_block_sid = bl_sid_opt;
      deep_block_body = bl_body_elements;
      deep_block_desc = bl_desc;
    } in

  let get_deep_block_reference_type bl_name bl_ref_ty_opt =
    match bl_ref_ty_opt with
    | None ->
      deep_error (Printf.sprintf
        "missing block source type and block source path in \
         block reference %s"
        (string_of_block_name bl_name))
    | Some (bl_src_type, bl_src_path) ->
      make_deep_block_reference_type bl_src_type bl_src_path in

  let get_deep_block_sub_system bl_name bl_deep_sys_opt =
    match bl_deep_sys_opt with
    | None ->
      deep_error (Printf.sprintf
        "the system is missing in sub-system block %s"
        (string_of_block_name bl_name))
    | Some bl_deep_sys -> bl_deep_sys in

  let get_deep_block_s_function bl_name bl_tag_opt =
    match  bl_tag_opt with
    | None -> Mdl_deep_ast.Deep_s_function
    | Some tag_text ->
      let tag_string = string_of_mdl_text tag_text in
      try
        let mdl_name, deep_chart_file_number = deep_tag_text tag_text in
        Mdl_deep_ast.
        Deep_s_function_stateflow (mdl_name, deep_chart_file_number)
      with
      | Scanf.Scan_failure _ | End_of_file | Failure _ ->
        deep_warning
          (Printf.sprintf
             "unexpected tag format \'%s\' dropped for block \'%s\'"
             tag_string (string_of_block_name bl_name));
        Mdl_deep_ast.
        Deep_s_function in

  if bl_type = Mdl_block_types.Reference then
    let deep_block_reference_type =
      get_deep_block_reference_type bl_name bl_ref_ty_opt in
    make_deep_block_reference deep_block_reference_type else

  if bl_ref_ty_opt <> None then
    deep_error (Printf.sprintf
      "spurious block source specification in regular block %s"
      (string_of_block_name bl_name)) else

  if bl_type = Mdl_block_types.SubSystem then
    let bl_deep_sys = get_deep_block_sub_system bl_name bl_deep_sys_opt in
    make_deep_block_sub_system bl_deep_sys else

  if bl_type = Mdl_block_types.S_Function then
    let bl_s_function_desc = get_deep_block_s_function bl_name bl_tag_opt in
    make_deep_block_s_function bl_s_function_desc else

  make_deep_block_regular bl_type
;;

let annotation_bind_warning = function
  | { Binding.bound_ident = bound_ident; bound_value = _; } ->
    deep_warning (Format.sprintf
      "Unknown annotation property \'%s\'."
      (string_of_ident bound_ident))
;;

let deep_annotation_fields ais =

  let rec loop name_opt sid_opt pos_opt asize_opt wrap_opt z_opt
      fis facs ads = function
  | [] ->
    (* We skip here fields asize_opt, wrap_opt, and z_opt since none
       is used in Scicos/Scicos-pro.
       We also skip bindings which represent fields unknown to simport,
       after emitting a warning *)

    (name_opt, sid_opt, pos_opt,
     List.rev fis, List.rev facs, List.rev ads)
  | ai :: ais ->
    match ai with
    | Annotation_name name ->
      if name_opt = None then
        loop (Some name) sid_opt pos_opt asize_opt wrap_opt z_opt
          fis facs ads ais
      else deep_error "multiple annotation name definition"
    | Annotation_sid sid ->
      if sid_opt = None then
        loop name_opt (Some sid) pos_opt asize_opt wrap_opt z_opt
          fis facs ads ais
      else deep_error "multiple annotation sid definition"
(*
    | Annotation_tag _tag ->
        loop name_opt sid_opt pos_opt asize_opt wrap_opt z_opt
          fis facs ads ais
*)
    | Annotation_position pos ->
      if pos_opt = None then
        loop name_opt sid_opt
          (Some (deep_annotation_position pos)) asize_opt wrap_opt z_opt
          fis facs ads ais
      else deep_error "multiple annotation position definition"
    | Annotation_auto_size b ->
      if asize_opt = None then
        loop name_opt sid_opt pos_opt
          (Some b) wrap_opt z_opt fis facs ads ais
      else deep_error "multiple annotation auto size definition"
    | Annotation_word_wrap b ->
      if wrap_opt = None then
        loop name_opt sid_opt pos_opt
          asize_opt (Some b) z_opt fis facs ads ais
      else deep_error "multiple annotation word wrap definition"
    | Annotation_binding binding ->
      annotation_bind_warning binding;
      loop name_opt sid_opt pos_opt asize_opt wrap_opt z_opt
        fis facs ads ais
    | Annotation_zorder z ->
      if z_opt = None then
        loop name_opt sid_opt pos_opt
          asize_opt wrap_opt (Some (deep_zorder z)) fis facs ads ais
      else deep_error "multiple annotation zorder definition"
    | Annotation_graphical_default annotation_graphical_default ->
      match annotation_graphical_default with
      | Horizontal_alignment _
      | Interpreter _
      | Text_as_click_cb _
      | Tex_mode _
      | Vertical_alignment _
      | Click_fcn _ ->
        loop name_opt sid_opt pos_opt asize_opt wrap_opt z_opt
          fis facs (annotation_graphical_default :: ads) ais
      | Font_and_color fac ->
        begin match fac with
        | Font_item fit ->
          loop name_opt sid_opt pos_opt asize_opt wrap_opt z_opt
            (fit :: fis) facs ads ais
        | Background _ | Drop_shadow _ | Foreground _ ->
          loop name_opt sid_opt pos_opt asize_opt wrap_opt z_opt
            fis (fac :: facs) ads ais
        end
      | Annotation_graphical_binding binding ->
        annotation_bind_warning binding;
        loop name_opt sid_opt pos_opt asize_opt wrap_opt z_opt
          fis facs ads ais in

  loop None None None None None None [] [] [] ais
;;

let deep_annotation_graphical_default annotation_graphical_default =
  match annotation_graphical_default with
  | Horizontal_alignment _
  | Interpreter _
  | Text_as_click_cb _
  | Tex_mode _
  | Vertical_alignment _
  | Click_fcn _ as ad -> ad
  | Annotation_graphical_binding _
  | Font_and_color _ ->
    deep_error "spurious font and color annotation default"
;;

let deep_annotations annotations =
  let rec loop accu = function
    | [] -> List.rev accu
    | (Annotation ais) :: annotations ->
      match deep_annotation_fields ais with
      | (None, _, _, _, _, _) ->
        Say.warning "an annotation without text was dropped";
        loop accu annotations
      | (Some name, _, _, _, _, _) when name = mdl_text_of_string "" ->
        Say.warning "an annotation without text was dropped";
        loop accu annotations
      | (_, _, None, _, _, _) ->
        deep_error "annotation position was missing"
      | (Some name, None, Some position, fis, facs, agds)
      | (Some name, Some _, Some position, fis, facs, agds) ->
        let deep_annotation =
          {
            Mdl_basics_deep_ast.
            deep_annotation_name = name;
            deep_annotation_position = position;
            deep_annotation_font = deep_font_of_font_items fis;
            deep_annotation_graphical =
              List.map deep_font_and_color facs;
            deep_annotation_graphical_defaults =
              List.map deep_annotation_graphical_default agds;
          } in
        loop (deep_annotation :: accu) annotations in
  loop [] annotations
;;

let rec deep_system_connectors system_connectors =
  List.map deep_system_connector system_connectors
and deep_system_connector = deep_bindings
;;

(* FIXME should dissociate target and input type for functions like
   deep_branch_body
type implem_branch_body =
  {
    implem_branch_source = connector option;
    implem_branch_destination = connector option;
    implem_branch_branches = deep_branch list;
    implem_branch_labels = labels option;
    implem_branch_points = deep_points option;
    implem_branch_zorder = deep_zorder option;
  }
;;
*)

let deep_connection_type connection_type =
  let parent, self =
    match connection_type with
    | Dest_src ->
      Mdl_basics_deep_ast.Deep_side_dst, Mdl_basics_deep_ast.Deep_side_src
    | Dest_dest ->
      Mdl_basics_deep_ast.Deep_side_dst, Mdl_basics_deep_ast.Deep_side_dst
    | Src_dest ->
      Mdl_basics_deep_ast.Deep_side_src, Mdl_basics_deep_ast.Deep_side_dst
    | Src_src ->
      Mdl_basics_deep_ast.Deep_side_src, Mdl_basics_deep_ast.Deep_side_src in
  {
    Mdl_basics_deep_ast.
    deep_connection_type_parent = parent;
    deep_connection_type_self = self;
  }
;;

let rec deep_branch_body lis =

  let rec loop connection_type_opt accu = function
  | [] -> (connection_type_opt, accu)
  | li :: lis ->
    match li with
    | Line_target_branch_connection_type connection_type ->
      if connection_type_opt <> None then
        deep_error "multiple connection types in branch" else
      loop (Some connection_type) accu lis
    | Line_target_branch_labels labels ->
      loop connection_type_opt {
        accu with
        Mdl_basics_deep_ast.
        deep_branch_labels = Some labels; } lis
    | Line_target_branch_points points ->
      loop connection_type_opt {
        accu with
        Mdl_basics_deep_ast.
        deep_branch_points = Some (deep_points points);
      } lis
    | Line_target_branch_source (Line_source_connector line_connector) ->
      if accu.
         Mdl_basics_deep_ast.
         deep_branch_source <> None then
        deep_error "multiple source in branch" else
      loop connection_type_opt {
        accu with
        Mdl_basics_deep_ast.
        deep_branch_source = Some line_connector; } lis
    | Line_target_branch_target line_target ->
      begin match line_target with
      | Line_target_connector line_connector ->
        if accu.
           Mdl_basics_deep_ast.
           deep_branch_target <> None then
          deep_error "multiple destination in branch" else
        loop connection_type_opt {
          accu with
          Mdl_basics_deep_ast.
          deep_branch_target = Some line_connector;
        } lis
      | Line_target_branch branch_items ->
        loop connection_type_opt {
          accu with
          Mdl_basics_deep_ast.
          deep_branch_branches =
            deep_branch branch_items ::
            accu.Mdl_basics_deep_ast.deep_branch_branches;
        } lis
      end
    | Line_target_branch_zorder zorder ->
      if accu.
         Mdl_basics_deep_ast.
         deep_branch_zorder <> None then
        deep_error "multiple zorder in branch" else
      loop connection_type_opt {
        accu with
        Mdl_basics_deep_ast.
        deep_branch_zorder = Some (deep_zorder zorder);
      } lis in

  loop None
    {
      Mdl_basics_deep_ast.
      deep_branch_source = None;
      deep_branch_target = None;
      deep_branch_branches = [];
      deep_branch_labels = None;
      deep_branch_points = None;
      deep_branch_zorder = None;
    } lis

and deep_branch lis =
  let (connection_type_opt, branch_body) = deep_branch_body lis in

  match connection_type_opt with
  | Some connection_type ->
    Mdl_basics_deep_ast.
    Deep_branch_connection
      (deep_connection_type connection_type, branch_body)
  | None ->
    Mdl_basics_deep_ast.
    Deep_branch_directed branch_body
;;

(* Separate line fields into line font items and other line items *)
let deep_line_body lis =

  let rec loop connection_type_opt fis accu = function
  | [] -> (connection_type_opt, fis, accu)
  | li :: lis ->
    match li with
    | Line_name name ->
      loop connection_type_opt fis {
        accu with
        Mdl_basics_deep_ast.
        deep_line_name = Some name;
      } lis
    | Line_source (Line_source_connector line_connector) ->
      if accu.
         Mdl_basics_deep_ast.
         deep_line_source <> None then
        deep_error "multiple source in line" else
      loop connection_type_opt fis {
        accu with
        Mdl_basics_deep_ast.
        deep_line_source = Some line_connector;
      } lis
    | Line_target line_target ->
      begin match line_target with
      | Line_target_connector line_connector ->
        if accu.
           Mdl_basics_deep_ast.
           deep_line_target <> None then
          deep_error "multiple destination in line" else
        loop connection_type_opt fis {
          accu with
          Mdl_basics_deep_ast.
          deep_line_target = Some line_connector;
        } lis
      | Line_target_branch branch_items ->
        loop connection_type_opt fis {
          accu with
          Mdl_basics_deep_ast.
          deep_line_branches =
            deep_branch branch_items ::
            accu.Mdl_basics_deep_ast.deep_line_branches;
        } lis
      end
    | Line_type Connection ->
      loop (Some Connection) fis accu lis
    | Line_labels labels ->
      loop connection_type_opt fis {
        accu with
        Mdl_basics_deep_ast.
        deep_line_labels = Some labels;
      } lis
    | Line_points points ->
      loop connection_type_opt fis {
        accu with
        Mdl_basics_deep_ast.
        deep_line_points = Some (deep_points points);
      } lis
    | Line_graphical_default (Line_font_item font_item) ->
      loop connection_type_opt (font_item :: fis) accu lis
    | Line_graphical_default (Line_zorder_item zorder_item) ->
      if accu.
         Mdl_basics_deep_ast.
         deep_line_zorder <> None then
        deep_error "multiple zorder in line" else
      let deep_zorder = deep_zorder zorder_item in
      loop connection_type_opt fis {
        accu with
          Mdl_basics_deep_ast.
          deep_line_zorder = Some deep_zorder;
      } lis in

  loop None [] {
    Mdl_basics_deep_ast.
    deep_line_font = None;
    deep_line_zorder = None;
    deep_line_source = None;
    deep_line_target = None;
    deep_line_branches = [];
    deep_line_name = None;
    deep_line_labels = None;
    deep_line_points = None;
  } lis
;;

let deep_line (Line lis) =
  let (connection_type_opt, fis, line_body) = deep_line_body lis in
  let font_opt =
    if fis = [] then None else Some (deep_font_of_font_items fis) in
  let deep_line_body =
    {
      line_body with
      Mdl_basics_deep_ast.
      deep_line_font = font_opt; } in

  match connection_type_opt with
  | Some Connection ->
    Mdl_basics_deep_ast.
    Deep_line_connection deep_line_body
  | None ->
    Mdl_basics_deep_ast.
    Deep_line_directed deep_line_body
;;

let deep_lines lines = List.map deep_line lines;;

(** {6 Systems} *)

let rec deep_system system =
  match system with
  | {
      system_name = system_name;
      system_position = system_position;
      system_bindings = bindings;
      system_blocks = blocks;
      system_lines = lines;
      system_annotations = annotations;
      system_connectors = connectors;
    } ->
    let rec loop accu = function
      | [] -> List.rev accu
      | Block blis :: blocks ->
        loop (deep_block deep_system blis :: accu) blocks
      | Block_vendor _block_vendor :: blocks ->
        Say.warning "discarding a block vendor in system part of the model";
        loop accu blocks in
    let deep_system_blocks = loop [] blocks in
    {
      Mdl_deep_ast.
      deep_system_name = system_name;
      deep_system_position = deep_position system_position;
      deep_system_bindings = deep_bindings bindings;
      deep_system_blocks = deep_system_blocks;
      deep_system_lines = deep_lines lines;
      deep_system_annotations = deep_annotations annotations;
      deep_system_connectors = deep_system_connectors connectors;
    }
;;

(** {6 Models} *)

let make_deep_block_parameter_defaults
    { block_type = bl_type;
      parameter_defaults = bindings; } =
  {
    Mdl_basics_deep_ast.
    deep_parameter_block_type = bl_type;
    deep_parameter_defaults = List.map deep_binding bindings;
  }
;;

let make_default_deep_block_parameter_defaults bl_type =
  make_deep_block_parameter_defaults
    {
      block_type = bl_type;
      parameter_defaults = [];
    }
;;

let deep_line_graphical_defaults line_graphical_defaults =

  let rec loop font_items z_opt = function
    | [] -> List.rev font_items, z_opt
    | Line_font_item fit :: lgds ->
      loop (fit :: font_items) z_opt lgds
    | Line_zorder_item zit :: lgds ->
      if z_opt <> None then
        deep_error "multiple zorder in line" else
        loop font_items (Some zit) lgds in
  let font_items, z_opt = loop [] None line_graphical_defaults in
  let deep_font = deep_font_of_font_items font_items in
  let deep_zorder_opt =
    match z_opt with
    | None -> None
    | Some zit -> Some (deep_zorder zit) in
  {
    Mdl_basics_deep_ast.
    deep_line_default_font = deep_font;
    deep_line_default_zorder = deep_zorder_opt;
  }
;;

let deep_mask_graphical_defaults = List.map deep_binding;;
let deep_mask_parameter_defaults = List.map deep_binding;;

let deep_annotation_graphical_defaults agds =

  let rec loop fis accu agds =
    match agds with
    | [] ->
      let font = deep_font_of_font_items fis in
       Font_and_color (Mdl_basics_deep_ast.Deep_font font) :: accu
    | agd :: agds ->
      match agd with
      | Horizontal_alignment _
      | Interpreter _
      | Text_as_click_cb _
      | Tex_mode _
      | Vertical_alignment _
      | Click_fcn _ as annotation_graphical_default ->
        loop fis (annotation_graphical_default :: accu) agds
      | Font_and_color fc ->
        begin match fc with
        | Font_item fitem ->
          loop (fitem :: fis) accu agds
        | Background _ | Drop_shadow _ | Foreground _ ->
          loop fis (Font_and_color (deep_font_and_color fc) :: accu) agds
        end
      | Annotation_graphical_binding _binding ->
        loop fis accu agds in
  loop [] [] agds
;;

let deep_workspace_data_source_code wsv =
  match deep_simple_model_value wsv with
  | Mdl_basics_deep_ast.Deep_string s ->
    Mdl_basics_deep_ast.
    Deep_workspace_data_source_code (Ast_types.mdl_code_of_string s)
  | Mdl_basics_deep_ast.Deep_code (Mdl_code _mdl_text as code) ->
    Mdl_basics_deep_ast.
    Deep_workspace_data_source_code code
  | Mdl_basics_deep_ast.Deep_wstring _
  | Mdl_basics_deep_ast.Deep_expression _
  | Mdl_basics_deep_ast.Deep_bool _
  | Mdl_basics_deep_ast.Deep_float _
  | Mdl_basics_deep_ast.Deep_int _
  | Mdl_basics_deep_ast.Deep_matrix _
  | Mdl_basics_deep_ast.Deep_vector _
  | Mdl_basics_deep_ast.Deep_bindings _
  | Mdl_basics_deep_ast.Deep_ident _ ->
    deep_error (Printf.sprintf
      "wrong workspace data specification, code item should be a string")
;;

let set_option in_section_name item_name v = function
  | None -> Some v
  | Some _v0 ->
    deep_error (Printf.sprintf
      "in \'%s\', %s is defined twice" in_section_name item_name)
;;

let set_uniq_name_in_config_set item_name =
  set_option "config set" item_name
;;

let set_uniq_name_in_workspace_data item_name =
  set_option "workspace data definition" item_name
;;

let classify_model_workspace_datas wss_items =
  let rec loop
      data_source_type_opt mat_file_name_opt source_file_name_opt
      mcode_opt mdl_file_data_opt = function
    | [] ->
      data_source_type_opt, mat_file_name_opt, source_file_name_opt,
      mcode_opt, mdl_file_data_opt

    | Workspace_data_source_type data_source_type :: wss_items ->
      let data_source_type_opt =
        set_uniq_name_in_workspace_data "data source type"
          data_source_type data_source_type_opt in
      loop data_source_type_opt mat_file_name_opt source_file_name_opt
           mcode_opt mdl_file_data_opt wss_items

    | Workspace_data_mat_file_name mat_file_name :: wss_items ->
      let mat_file_name_opt =
        set_uniq_name_in_workspace_data "mat file name"
          mat_file_name mat_file_name_opt in
      loop data_source_type_opt mat_file_name_opt source_file_name_opt
           mcode_opt mdl_file_data_opt wss_items

    | Workspace_data_source_file_name source_file_name :: wss_items ->
      let source_file_name_opt =
        set_uniq_name_in_workspace_data "source file name"
          source_file_name source_file_name_opt in
      loop data_source_type_opt mat_file_name_opt source_file_name_opt
           mcode_opt mdl_file_data_opt wss_items

    | Workspace_data_mcode simple_value :: wss_items ->
      let mcode_opt =
        set_uniq_name_in_workspace_data "m code"
          simple_value mcode_opt in
      loop data_source_type_opt mat_file_name_opt source_file_name_opt
           mcode_opt mdl_file_data_opt wss_items

    | Workspace_data_mdl_file tag_ident :: wss_items ->
      let mdl_file_data_opt =
        set_uniq_name_in_workspace_data "mdl file data"
          tag_ident source_file_name_opt in
      loop data_source_type_opt mat_file_name_opt source_file_name_opt
           mcode_opt mdl_file_data_opt wss_items in

  loop None None None None None wss_items
;;

let deep_model_workspace_data_of_items
      (data_source_type_opt, mat_file_name_opt, source_file_name_opt,
       mcode_opt, mdl_file_data_opt) =
  match data_source_type_opt with

  (* With data source (old scheme) *)
  | Some Mdl_ast.M_Code ->
    if mat_file_name_opt <> None ||
       source_file_name_opt <> None ||
       mdl_file_data_opt <> None then
      deep_error (Printf.sprintf
        "wrong workspace data specification, \
         too many items or invalid item combination")
    else
      begin
        match mcode_opt with
        | None ->
          Say.warning
            "Truncated workspace data specification without code was ignored";
          []
        | Some wsv ->
          let dwsc = deep_workspace_data_source_code wsv in
          let dcw =  {
            Mdl_deep_ast.
            workspace_data_source_type =
              Some Mdl_deep_ast.Deep_M_Code;
            workspace_data_source_code = dwsc;
          } in
          [ Mdl_deep_ast.Deep_code_workspace_data dcw; ]
      end

  | Some Mdl_ast.MATLAB_Code ->
    if mat_file_name_opt <> None ||
       source_file_name_opt <> None ||
       mdl_file_data_opt <> None then
      deep_error (Printf.sprintf
        "wrong workspace data specification, \
         too many items or invalid item combination")
    else
      begin
        match mcode_opt with
        | None ->
          Say.warning
            "Truncated workspace data specification without code was ignored";
          []
        | Some wsv ->
          let dwsc = deep_workspace_data_source_code wsv in
          let dcw =  {
            Mdl_deep_ast.
            workspace_data_source_type =
              Some Mdl_deep_ast.Deep_MATLAB_Code;
            workspace_data_source_code = dwsc;
          } in
          [ Mdl_deep_ast.Deep_code_workspace_data dcw; ]
      end

  | Some Mdl_ast.MAT_File ->
    if mat_file_name_opt <> None ||
       mcode_opt <> None ||
       mdl_file_data_opt <> None then
      deep_error (Printf.sprintf
        "wrong workspace data specification, \
         too many items or invalid item combination")
    else
      begin
        match source_file_name_opt with
        | None ->
          Say.warning
            "Truncated workspace data specification without code was ignored";
          []
        | Some wssfn ->
          [ Mdl_deep_ast.Deep_mat_file_workspace_data wssfn; ]
      end
  | Some Mdl_ast.MATLAB_File ->
    if mat_file_name_opt <> None ||
       mdl_file_data_opt <> None then
      deep_error (Printf.sprintf
        "wrong workspace data specification, \
         too many items or invalid item combination")
    else
      begin
        match source_file_name_opt, mcode_opt with
        | Some wssfn, Some wsv ->
          let wssc = deep_workspace_data_source_code wsv in
          let dcws = {
            Mdl_deep_ast.workspace_data_source_type =
              Some (Mdl_deep_ast.Deep_MATLAB_File wssfn);
            workspace_data_source_code = wssc;
          } in
          [ Mdl_deep_ast.Deep_code_workspace_data dcws ]
        | None, None ->
          Say.warning
            "Truncated workspace data specification without code was ignored";
          []
        | _, _ ->
          deep_error (Printf.sprintf
            "wrong workspace data specification, \
             too many items or invalid item combination")
      end
  | None ->
    begin
      match mat_file_name_opt, source_file_name_opt,
            mcode_opt, mdl_file_data_opt with
      (* Without data source (old scheme) *)
      | _wsfn_opt, None, None, Some tag_ident ->
        let tag_name = Ast_types.string_of_ident tag_ident in
        Say.warning (Printf.sprintf
          "reference to datatag is not supported.\
         \n  Data %s was ignored."
          tag_name);
        []
      | None, None, Some wsv, tag_ident_opt ->
        begin
          match tag_ident_opt with
          | None -> ()
          | Some tag_ident ->
            let tag_name = Ast_types.string_of_ident tag_ident in
            Say.warning (Printf.sprintf
              "reference to datatag is not supported.\
             \n  Data %s was ignored."
              tag_name);
        end;
        let dwsc = deep_workspace_data_source_code wsv in
        let dcw = {
          Mdl_deep_ast.
          workspace_data_source_type = None;
          workspace_data_source_code = dwsc;
        } in
        [ Mdl_deep_ast.Deep_code_workspace_data dcw; ]
      | Some _, Some _, _, _
      | Some _, None, Some _, _
      | Some _, None, None, None
      | None, Some _, _, _
      | None, None, None, None ->
        deep_error (Printf.sprintf
          "wrong workspace data specification, \
           too many items or invalid item combination")
    end
;;

let deep_model_workspace wss =
  match wss with
  | None -> []
  | Some [] ->
    deep_error (Printf.sprintf
      "wrong workspace data specification with \
       %d items (instead of 1 or 2 or 3)"
      0)
  | Some wss_items ->
    let sorted_wss_items =
      classify_model_workspace_datas wss_items in
    deep_model_workspace_data_of_items sorted_wss_items
;;

let deep_ident vendor_ident vendor_configuration_longident =
  let buff = Buffer.create 15 in
  let add_ident ident =
    Buffer.add_string buff (Ast_types.string_of_ident ident) in
  let vendor_name = Mdl_ast_functions.mdl_name_of_vendor vendor_ident in
  Buffer.add_string buff (Ast_types.string_of_mdl_name vendor_name);
  Buffer.add_char buff '.';
  let rec loop = function
    | Configuration_lident configuration_ident ->
      add_ident configuration_ident
    | Configuration_ldot (configuration_longident, configuration_ident) ->
      loop configuration_longident;
      Buffer.add_char buff '.';
      add_ident configuration_ident in
  loop vendor_configuration_longident;
  Buffer.contents buff
;;

let rec deep_vendor_items items =
  let rec loop acc = function
    | [] -> acc
    | Vendor_dollar_binding (
        Dollar_object_id _
      | Dollar_backup_class _
      | Dollar_class_name _
      | Dollar_prop_name _) :: t ->
      loop acc t
    | Vendor_binding b :: t ->
      begin match b with
      | Object_type _ -> assert false
      | Object_name _ -> loop acc t
      | Object_dollar_binding _ -> assert false
      | Object_array _ -> loop acc t
      | Object_custom_property _ -> assert false
      | Object_object _ -> assert false
      | Object_binding b -> loop (deep_binding b :: acc) t
      end
    | Vendor_vendor_binding vb :: t ->
      let deep_binding = deep_binding_of_vendor_binding vb in
      loop (deep_binding :: acc) t in
  loop [] items
and deep_binding_of_vendor_binding = function
  | {
      vendor_ident = vendor_ident;
      vendor_configuration_longident = vendor_configuration_longident;
      vendor_items = vendor_items;
    } ->
    let vendor_bindings = deep_vendor_items vendor_items in
    let bound_ident = deep_ident vendor_ident vendor_configuration_longident in
    let bound_value = Mdl_basics_deep_ast.Deep_bindings vendor_bindings in
    {
      Binding.
      bound_ident = Ast_types.ident_from_string bound_ident;
      bound_value = bound_value;
    }
;;

let deep_vendor_binding = function
  | {
      vendor_ident = _vendor_ident;
      vendor_configuration_longident = configuration_longident;
      vendor_items = vendor_items;
    } ->
    configuration_longident, deep_vendor_items vendor_items
;;

let deep_component = function
  | Array_vendor_binding vendor_binding -> deep_vendor_binding vendor_binding
  | Array_name _ | Array_binding _ | Array_vendor_configset (_, _)
  | Array_sub_array _ | Array_custom_property _
  | Array_matstruct _ | Array_object _ -> assert false
;;

let deep_config_set l =
  let rec loop (name_opt, object_id_opt, accu_vendor_items, comps_opt) =
    function
    | [] ->
      {
        Mdl_deep_ast.
        deep_configuration_object_id =
          from_option "object id" object_id_opt;
        deep_configuration_vendor_items =
          accu_vendor_items;
        deep_configuration_vendor_components =
          from_option "components array" comps_opt;
        deep_configuration_name =
          from_option "name of configset" name_opt;
      }
    | Vendor_dollar_binding (
        Dollar_prop_name _
      | Dollar_class_name _
      | Dollar_backup_class _) as vendor_item :: vendor_items ->
      let accu_vendor_items = vendor_item :: accu_vendor_items in
      loop
        (name_opt, object_id_opt, accu_vendor_items, comps_opt) vendor_items
    | Vendor_dollar_binding (Dollar_object_id i) :: vendor_items ->
      let object_id_opt =
        set_uniq_name_in_config_set "object id" i object_id_opt in
      loop
        (name_opt, object_id_opt, accu_vendor_items, comps_opt) vendor_items
    | Vendor_binding object_item as vendor_item :: vendor_items ->
      begin match object_item with
      | Object_type _ -> assert false
      | Object_name n ->
        let name_opt =
          set_uniq_name_in_config_set "object name" n name_opt in
        loop
          (name_opt, object_id_opt, accu_vendor_items, comps_opt) vendor_items
      | Object_dollar_binding _ -> assert false
      | Object_array {
          array_type = _;
          array_dimension = _;
          array_prop_name = _;
          array_items = bdgs;
        } ->
        let comps = List.map deep_component bdgs in
        let comps_opt =
          set_uniq_name_in_config_set "components array" comps comps_opt in
        loop
          (name_opt, object_id_opt, accu_vendor_items, comps_opt) vendor_items
      | Object_binding _
      | Object_custom_property _
      | Object_object _ ->
        let accu_vendor_items = vendor_item :: accu_vendor_items in
        loop
          (name_opt, object_id_opt, accu_vendor_items, comps_opt) vendor_items
      end
    | Vendor_vendor_binding _vb :: vendor_items ->
      deep_warning (Printf.sprintf
        "Dropping a nested vendor binding in model config set");
      loop
        (name_opt, object_id_opt, accu_vendor_items, comps_opt) vendor_items in
  loop (None, None, [], None) l
;;

let deep_model_configurations confs =
  let deep_model_configuration conf =
    let {
          array_type = _;
          array_dimension = _;
          array_items = bdgs;
          array_prop_name = _;
        } = conf in
    let fold_cf_set l = function
      | Array_vendor_configset (_vn, vi) ->
        deep_config_set vi :: l
      | Array_name _ | Array_binding _
      | Array_vendor_binding _ | Array_sub_array _
      | Array_custom_property _
      | Array_matstruct _ | Array_object _ -> l in
    List.fold_left fold_cf_set [] bdgs in

  match confs with
  | [] -> []
  | [ conf ] -> deep_model_configuration conf
  | _conf :: _confs ->
    deep_error "more than one configuration set_section in model"
;;

let separate_configuration_parameters csp =
  let rec loop gr_params vendor_bindings obindings bindings = function
  | [] -> (gr_params, vendor_bindings, obindings, bindings)
  | Configuration_graphical_interface graphical_interface_items :: t ->
    loop (graphical_interface_items @ gr_params)
         vendor_bindings obindings bindings t
  | Configuration_binding {
      Binding.bound_ident = id; bound_value = v;
    } :: t ->
    loop gr_params vendor_bindings obindings
      ((id, deep_simple_model_value v) :: bindings) t
  | Configuration_vendor_binding bdg :: t ->
    loop gr_params (bdg :: vendor_bindings) obindings bindings t
  | Configuration_object bdg :: t ->
    loop gr_params vendor_bindings (bdg :: obindings) bindings t in

  loop [] [] [] [] csp
;;

(* Computing version and sub-version of the model version. *)
let deep_version model_version =
  try
    Scanf.sscanf model_version "%d.%d"
      (fun mv sv ->
       {
         Mdl_deep_ast.
         deep_major_version = Unatural.of_int mv;
         deep_minor_version = Unatural.of_int sv;
       }) with
  | Scanf.Scan_failure _
  | Failure _
  | End_of_file as exn ->
    deep_error (
      Printf.sprintf "%s is not a valid version number, %s"
        model_version (Printexc.to_string exn)
    )
;;

let is_supported_mdl_version model_version =
  model_version.Mdl_deep_ast.deep_major_version >= Unatural._6_
;;

let deep_model_version = function
  | Mdl_model_version model_version ->
    let mdl_version = deep_version model_version in
    if not (is_supported_mdl_version mdl_version) then
      Say.fatal_error (
        Printf.sprintf
          "unsupported MDL version %d: only versions >= 6.0 supported"
          (Unatural.to_int mdl_version.Mdl_deep_ast.deep_major_version)
      ) else
    Mdl_deep_ast.Deep_mdl_model_version mdl_version
  | Slx_model_version model_version ->
    Mdl_deep_ast.Deep_slx_model_version (deep_version model_version)
;;

let deep_model_body efname = function
  {
    model_ident = model_ident;
    model_version = model_version;
    model_configuration_parameters = mcps;
    model_configurations = mcs;
    model_set_configuration_option = mcso;
    model_workspace = mws;
    block_graphical_defaults_section = bgds;
    block_parameter_defaults_section = bpd;
    annotation_graphical_defaults_section = agd;
    line_graphical_defaults_section = lgds;
    mask_graphical_defaults_section = mgds;
    mask_parameter_defaults_section = mpd;
    model_system = system;
  } ->

  let dmcs = deep_model_configurations mcs in
  let gr_params, vendor_bindings, object_bindings, bindings =
    separate_configuration_parameters mcps in
  let dmc_section = {
    Mdl_deep_ast.
    deep_graphical_parameters = gr_params;
    deep_configuration_bindings = Hash_table.of_bindings bindings;
    deep_vendor_bindings = vendor_bindings;
    deep_object_bindings = object_bindings;
    deep_model_configurations = dmcs;
    deep_model_set_configuration_option = mcso;
  } in
  let dmd = {
    Mdl_basics_deep_ast.
    deep_model_block_graphical_defaults =
      deep_model_block_graphical_defaults bgds;
    deep_model_block_parameter_defaults =
      List.map make_deep_block_parameter_defaults bpd;
    deep_model_annotation_graphical_defaults =
      deep_annotation_graphical_defaults agd;
    deep_model_line_graphical_defaults =
      deep_line_graphical_defaults lgds;
    deep_model_mask_graphical_defaults =
      deep_mask_graphical_defaults mgds;
    deep_model_mask_parameter_defaults =
      deep_mask_parameter_defaults mpd;
  } in
  let ds = deep_system system
  and dmws = deep_model_workspace mws
  and dmv = deep_model_version model_version in
  {
    Mdl_deep_ast.
    deep_model_ident = model_ident;
    deep_model_version = dmv;
    deep_model_explicit_file_name = efname;
    deep_model_configuration_section = dmc_section;
    deep_model_workspace = dmws;
    deep_model_defaults = dmd;
    deep_model_system = ds;
  }
;;

(** Interfaces *)

(** {6 Block interfaces} *)

let deep_interface_block_parameter_type ty =
  match string_of_ident ty with
  | "boolean" -> Mdl_deep_ast.Parameter_type_bool
  | "string" -> Mdl_deep_ast.Parameter_type_string
  | "enum" -> Mdl_deep_ast.Parameter_type_enum
  | "matrix" -> Mdl_deep_ast.Parameter_type_matrix
  | "list" -> Mdl_deep_ast.Parameter_type_list
  | "rectangle" -> Mdl_deep_ast.Parameter_type_rectangle
  | ty ->
    deep_error
      (Printf.sprintf "unknown block interface parameter type \'%s\'" ty)
;;

let deep_interface_block_parameter bl_ip = {
   bl_ip with
   Binding.bound_value =
     deep_interface_block_parameter_type bl_ip.Binding.bound_value;
}
;;

let deep_block_reference_type bl_type bl_src_type_opt bl_src_path_opt =
  match bl_src_type_opt, bl_src_path_opt with
  | None, _ ->
    let ty = string_of_block_type bl_type in
    deep_error (Printf.sprintf "no BlockSourceType for \'%s\'" ty)
  | _, None ->
    let ty = string_of_block_type bl_type in
    deep_error (Printf.sprintf "no BlockSourceBlock for \'%s\'" ty)
  | Some bl_src_type, Some bl_src_path ->
    make_deep_block_reference_type bl_src_type bl_src_path
;;

let deep_interface_block {
    interface_block_type = bl_type;
    interface_block_source_type = bl_src_type_opt;
    interface_block_source_path = bl_src_path_opt;
    interface_block_ports_description = bl_ports_desc;
    interface_block_parameters = bl_params;
  } =

  (* To recognize the S-function and Sub-System simulink primitive block
     definition: their bl_src_type_opt and bl_src_path_opt are not None. *)
  let deep_bl_interface_desc =
    if bl_type = Mdl_block_types.SubSystem && bl_src_type_opt <> None then
      Mdl_deep_ast.
      Deep_interface_block_sub_system (
        deep_block_reference_type
          bl_type bl_src_type_opt bl_src_path_opt) else

    if bl_type = Mdl_block_types.S_Function && bl_src_type_opt <> None then
      Mdl_deep_ast.
      Deep_interface_block_s_function (
        deep_block_reference_type
          bl_type bl_src_type_opt bl_src_path_opt) else
    Mdl_deep_ast.
    Deep_interface_block_simulink bl_type in

  let deep_bl_ports_desc = deep_ports_description bl_ports_desc in
  let deep_bl_params = List.map deep_interface_block_parameter bl_params in

  {
    Mdl_deep_ast.
    deep_interface_block_desc = deep_bl_interface_desc;
    deep_interface_block_ports_description = deep_bl_ports_desc;
    deep_interface_block_parameters = deep_bl_params;
  }
;;

(** {6 Interface files} *)

let deep_interface_section efname = function
  | {
      interface_ident = iident;
      interface_blocks = bis;
    } ->
    {
      Mdl_deep_ast.
      deep_interface_ident = iident;
      deep_interface_explicit_file_name = efname;
      deep_interface_blocks = List.map deep_interface_block bis;
    }
;;

(* Trailor *)
let deep_matdata_section = function
  | Matdata_section { matdata_num_records = n; matdata_records = l; } ->
    let effective_len = List.length l in
    let records =
      if n = Unatural_types.to_natural effective_len then l else
      deep_error (Printf.sprintf
        "inconsistent Matdata section:\n\
         declared number of records is %d\n\
         but effective number is %d"
        (Unatural.to_int n) effective_len) in
    Mdl_deep_ast.
    Deep_matdata_section records
;;

let deep_implementation_model
    efname model_section implementation_section_items =
  let deep_model_body = deep_model_body efname model_section in
  let deep_implementation_model =
    {
      Mdl_deep_ast.
      deep_model_body = deep_model_body;
      deep_model_matdatas_option = None;
      deep_model_stateflow_option = None;
    } in

  let rec loop deep_implementation_model accu_matdata = function
    | [] ->
      let matdata_option =
        match accu_matdata with
        | [] -> None
        | _ -> Some accu_matdata in
      {
        deep_implementation_model with
        Mdl_deep_ast.
        deep_model_matdatas_option = matdata_option;
      }
    | Implementation_model _model_section :: _ ->
      deep_error "forbidden extraneous Model section"
    | Implementation_library _model_section :: _ ->
      deep_error "forbidden extraneous Library section"
    | Implementation_stateflow _sf :: _
        when
        deep_implementation_model.
        Mdl_deep_ast.deep_model_stateflow_option <> None ->
      deep_error "forbidden extraneous Stateflow section"
    | Implementation_stateflow sf :: implementation_section_items ->
      let stateflow =
        Mdl_stateflow_to_deep_ast.deep_stateflow_section sf in
      let deep_implementation_model =
        {
          deep_implementation_model with
          Mdl_deep_ast.
          deep_model_stateflow_option = Some stateflow;
        } in
      loop deep_implementation_model accu_matdata
        implementation_section_items
    | Implementation_matdata md :: implementation_section_items ->
      let accu_matdata = deep_matdata_section md :: accu_matdata in
      loop deep_implementation_model accu_matdata
        implementation_section_items in
  loop deep_implementation_model [] implementation_section_items
;;

let deep_implementation_library = deep_implementation_model;;

let deep_implementation_sections efname = function
  | [] -> deep_error "empty model"
  | Implementation_model model_section :: l_s ->
    Mdl_deep_ast.
    Deep_implementation_model
      (deep_implementation_model efname model_section l_s)
  | Implementation_library model_section :: l_s ->
    Mdl_deep_ast.
    Deep_implementation_library
      (deep_implementation_library efname model_section l_s)
  | Implementation_matdata _m :: _ ->
    deep_error "forbidden toplevel MatData section"
  | Implementation_stateflow _s :: _ ->
    deep_error "forbidden toplevel Stateflow section"
;;

let simulink_file_to_deep_ast
    (efname, simulink_file_contents as simulink_file) =
  Say.fdebug (fun ppf ->
    Mdl_ast_print.print_simulink_file ppf simulink_file
  );
  match simulink_file_contents with
  | Mdl_ast.Implementation sections ->
    Mdl_deep_ast.
    Deep_implementation (deep_implementation_sections efname sections)
  |  Mdl_ast.Interface interface_section ->
    Mdl_deep_ast.
    Deep_interface (deep_interface_section efname interface_section)
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
