(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_basics_deep_ast.mli,v 1.1 2019-05-10 14:07:14 jpc Exp $ *)

(** From shadow parse tree to deep syntax:
   - we perform various validity checks on the tree obtained from the parser,
   - we build explicit record representations of data instead of simple
     values with associated so-called property lists.
*)

(** Shorthands borrowed from basic modules *)
type ident = Ident.ident
;;

(** Borrowed from Ast_types *)

type natural = Ast_types.natural
and deep_text = Ast_types.mdl_text
and deep_code = Ast_types.mdl_code
;;

(** Borrowed from Mdl_ast *)
type deep_port_number = Mdl_ast.port_number
and deep_port_name = Mdl_ast.port_name
and deep_line_name = Mdl_ast.line_name

and deep_array_name = Mdl_ast.array_name
and deep_prop_name = Mdl_ast.prop_name
and deep_class_name = Mdl_ast.class_name
and deep_line_connector = Mdl_ast.line_connector
and deep_labels = Mdl_ast.labels
and deep_block_name = Mdl_ast.block_name
and deep_system_name = Mdl_ast.system_name
and deep_array_type = Mdl_ast.array_type
and deep_sf_block_type = Mdl_ast.block_sf_block_type
and deep_name = Ast_types.mdl_name
;;

(** {3 Deep values} *)

type deep_binding = deep_value Binding.ident_binding

and deep_value =
  | Deep_bool of bool
  | Deep_float of string
  | Deep_int of string
  | Deep_string of string
  | Deep_wstring of Wstring.wstring
  | Deep_ident of ident
  | Deep_code of deep_code
  | Deep_expression of deep_text
  | Deep_vector of deep_values
  | Deep_matrix of deep_values
  | Deep_bindings of deep_bindings

and deep_bindings = deep_binding list
and deep_binding_table = (ident, deep_binding) Hash_table.t
and deep_values = deep_value list
;;

type deep_connection_type_side =
  | Deep_side_src
  | Deep_side_dst
;;

type deep_connection_type = {
  deep_connection_type_parent : deep_connection_type_side;
  deep_connection_type_self : deep_connection_type_side;
}
;;

type deep_color =
  | Deep_red
  | Deep_green
  | Deep_blue
  | Deep_cyan
  | Deep_magenta
  | Deep_yellow
  | Deep_white
  | Deep_black
  | Deep_grey
  | Deep_orange
  | Deep_lightred
  | Deep_lightgreen
  | Deep_lightblue
  | Deep_lightgrey
  | Deep_darkred
  | Deep_darkgreen
  | Deep_darkblue
  | Deep_darkgrey
  | Deep_purple
  | Deep_rgb of rgb_color

and rgb_color = { red : float; green : float; blue : float; }
;;

type deep_position = {
  xmin : int;
  ymin : int;
  xmax : int;
  ymax : int;
}
;;

type deep_font = {
  deep_font_name : Ast_types.mdl_font_name option;
  deep_font_angle : deep_font_angle option;
  deep_font_size : deep_font_size option;
  deep_font_weight : deep_font_weight option;
}

and deep_font_angle =
  | Angle_italic
  | Angle_normal
  | Angle_oblique

and deep_font_size = natural

and deep_font_weight =
  | Weight_normal
  | Weight_bold
  | Weight_light
;;

type deep_block_port = {
  deep_block_port_number : deep_port_number;
  deep_block_port_name : deep_port_name option;
  deep_block_port_bindings : deep_block_port_bindings;
}

and deep_block_port_bindings = deep_bindings

and deep_block_ports = deep_block_port list
;;

type deep_chart_file_number = natural;;
type deep_machine_name = deep_name;;

type deep_state = deep_value
and deep_tag = deep_machine_name * deep_chart_file_number
and deep_rotation = natural
and deep_mirror = bool
and deep_zorder = int
;;

type deep_annotation_position = {
  xpos : int;
  ypos : int;
  maxpos : deep_annotation_max_position option;
}
and deep_annotation_max_position = {
  xmaxpos : int;
  ymaxpos : int;
}
;;

type deep_orientation =
  | Orientation_right
  | Orientation_left
  | Orientation_up
  | Orientation_down
;;

type deep_placement =
  | Deep_placement_normal
  | Deep_placement_alternate
;;

type deep_font_and_colors = deep_font_and_color list
and deep_font_and_color =
  | Deep_background of deep_color
  | Deep_drop_shadow of deep_drop_shadow
  | Deep_font of deep_font
  | Deep_foreground of deep_color
and deep_drop_shadow = bool
;;

(** {6 Implementations} *)

(** TODO: regular -> directed *)
type deep_number_of_regular_ports = {
  port_nin : natural;
  port_nout : natural;
}
;;

type deep_number_of_implicit_ports = {
  port_niin : natural;
  port_niout : natural;
}
;;

type deep_ports_description = {
  number_of_regular_ports : deep_number_of_regular_ports;
  (* Number of regular ports, in and out. *)
  port_enable : bool;
  port_trigger : bool;
  port_ifaction : bool;
  (* Ports of action: enable, trigger, if_action. *)
  port_state : bool;
  (* Is there a special state out port ? *)
  number_of_implicit_ports : deep_number_of_implicit_ports;
  (* Number of implicit ports, in and out. *)
}
;;

type deep_workspace_data_source_code =
  | Deep_workspace_data_source_code of deep_code
;;

type deep_source_code =
(* FIXME: should be
  | Deep_source_code of deep_code
*)
  | Deep_source_code of deep_text
;;

type deep_block_graphical_default =
  | Deep_block_mirror of deep_mirror
  | Deep_block_name_placement of deep_placement
  | Deep_block_orientation of deep_orientation
  | Deep_block_rotation of deep_rotation
  | Deep_block_show_name of bool
  | Deep_block_hide_automatic_name of bool
  | Deep_block_zorder of deep_zorder
  | Deep_block_ports of deep_block_ports
  | Deep_block_ports_description of deep_ports_description
  | Deep_block_font_and_color of deep_font_and_color

and deep_block_graphical_defaults = deep_block_graphical_default list
(** Essentially graphical defaults for all the blocks of a model. *)
;;

type deep_block_parameter_defaults = {
  deep_parameter_block_type : Mdl_block_types.block_type;
  deep_parameter_defaults : deep_bindings;
}
(** Deep block parameter defaults: the model gives, for each block type, the
  non graphical defaults for the block parameters. *)

and deep_annotation_graphical_defaults = deep_annotation_graphical_default list

and deep_annotation_graphical_default =
  deep_font_and_color Mdl_ast.poly_annotation_graphical_default

and deep_line_graphical_defaults = {
  deep_line_default_font : deep_font;
  deep_line_default_zorder : deep_zorder option;
}

and deep_mask_graphical_defaults = deep_bindings
and deep_mask_parameter_defaults = deep_bindings
;;

type deep_model_defaults = {
  deep_model_block_graphical_defaults : deep_block_graphical_defaults;
  deep_model_block_parameter_defaults : deep_block_parameter_defaults list;
  deep_model_annotation_graphical_defaults : deep_annotation_graphical_defaults;
  deep_model_line_graphical_defaults : deep_line_graphical_defaults;
  deep_model_mask_graphical_defaults : deep_mask_graphical_defaults;
  deep_model_mask_parameter_defaults : deep_mask_parameter_defaults;
}
(** The defaults that the model gives:
    - the graphical defaults standing for all the blocks,
    - the parameter defaults standing for blocks of a given type,
    - the annotation and line defaults for all lines and annotations.
    - the graphical defaults standing for all the masked blocks,
    - the graphical defaults standing for masked blocks of a given type.
*)
;;

(** {6 Deep (block) mask parameter description} *)

type deep_mask_parameter_ident = ident;;

type mask_parameter_style_range =
  | Mask_parameter_style_range of string list
;;

type deep_mask_parameter_style =
  | Deep_mask_parameter_edit
  | Deep_mask_parameter_checkbox
  | Deep_mask_parameter_popup of mask_parameter_style_items
  | Deep_mask_parameter_unidt of deep_text (* FIXME: what is it? *)
  | Deep_mask_parameter_min
  | Deep_mask_parameter_max
  | Deep_mask_parameter_promote
  | Deep_mask_parameter_radiobutton of mask_parameter_style_items
  | Deep_mask_parameter_slider of mask_parameter_style_range option
  | Deep_mask_parameter_spinbox of mask_parameter_style_range option

and mask_parameter_style_item =
  | Item_int of int
  | Item_text of deep_text

and mask_parameter_style_items = mask_parameter_style_item list
;;

type deep_mask_parameter_prompt = deep_text;;

type deep_mask_parameter_callback = deep_source_code;;

type deep_mask_parameter_tunable = bool

type deep_mask_parameter_value =
  | Deep_mask_parameter_value of deep_text
;;

type deep_mask_parameter_evaluation =
  | Quote
  | Eval
;;

type deep_mask_parameter_enable =
  | Enable
  | Disable
;;

type deep_mask_parameter_visible =
  | Visible
  | Hidden
;;

type deep_mask_parameter_internal =
  | Internal
  | External
;;

type deep_mask_parameter_never_save =
  | Never_save
  | Save
;;

type deep_mask_parameter_read_only =
  | Read_only
  | Read_write
;;

type deep_mask_parameter = {
  deep_mask_parameter_ident : deep_mask_parameter_ident;
  deep_mask_parameter_style : deep_mask_parameter_style;
  deep_mask_parameter_prompt : deep_mask_parameter_prompt;
  deep_mask_parameter_callback : deep_mask_parameter_callback option;
  deep_mask_parameter_tunable : deep_mask_parameter_tunable option;
  deep_mask_parameter_value : deep_mask_parameter_value;
  deep_mask_parameter_evaluation : deep_mask_parameter_evaluation option;
  deep_mask_parameter_enable : deep_mask_parameter_enable option;
  deep_mask_parameter_visible : deep_mask_parameter_visible option;
  deep_mask_parameter_internal : deep_mask_parameter_internal option;
  deep_mask_parameter_never_save : deep_mask_parameter_never_save option;
  deep_mask_parameter_read_only : deep_mask_parameter_read_only option;
}

and deep_mask_parameters = deep_mask_parameter list
;;

(** {6 Deep (block) mask description} *)

type deep_mask_type = deep_text
and deep_mask_description = deep_text
;;

type deep_mask = {
  deep_mask_type : deep_mask_type option;
  deep_mask_description : deep_mask_description option;
  deep_mask_initialisation : deep_source_code option;
  deep_mask_display : deep_source_code option;
  deep_mask_parameters : deep_mask_parameters;
  deep_mask_bindings : deep_bindings;
}
;;

(*
 Example of mask found in a .mdl file

 MaskType "Stateflow"
 MaskDescription "Embedded MATLAB block"

 MaskDisplay
   "bgColor = Simulink.Root.ColorString2Rgb(get_param(gcbh, 'BackgroundColor'));
    image(imread('private/eml_membrane_16.png';"

 MaskDisplayString "AeroPendulum Controller \\n-Phase lead\\n-Phase lag"
*)

(** {6 Deep block item and line description} *)

type deep_sid = Mdl_ast.block_sid
;;

type deep_block_item =
  | Deep_block_position of deep_position
  | Deep_block_sid of deep_sid
  | Deep_block_state of deep_state
  | Deep_block_tag of deep_tag
  | Deep_block_graphical_default of deep_block_graphical_default
  | Deep_block_binding of deep_binding

and deep_block_items = deep_block_item list
;;

type deep_points = deep_point list
and deep_point = {
  deep_x : int;
  deep_y : int;
}
;;

type deep_branches = deep_branch list
and deep_branch =
  | Deep_branch_connection of deep_connection_type * deep_branch_body
  | Deep_branch_directed of deep_branch_body

and deep_branch_body = {
  deep_branch_source : deep_line_connector option;
  deep_branch_target : deep_line_connector option;
  deep_branch_branches : deep_branches;
  deep_branch_labels : deep_labels option;
  deep_branch_points : deep_points option;
  deep_branch_zorder : deep_zorder option;
}
;;

type deep_line =
  | Deep_line_connection of deep_line_body
  | Deep_line_directed of deep_line_body

and deep_line_body = {
  deep_line_name : deep_line_name option;
  deep_line_source : deep_line_connector option;
  deep_line_target : deep_line_connector option;
  deep_line_branches : deep_branches;
  deep_line_labels : deep_labels option;
  deep_line_points : deep_points option;
  deep_line_font : deep_font option;
  deep_line_zorder : deep_zorder option;
}
;;

type deep_annotation = {
  deep_annotation_name : deep_text;
  deep_annotation_position : deep_annotation_position;
  (* Those fields are graphical only and not used in scicos we ignore them
     in the deep syntax
     deep_annotation_auto_size : bool;
     deep_annotation_word_wrap : bool;
     deep_annotation_bindings : deep_bindings; *)
  deep_annotation_font : deep_font;
  deep_annotation_graphical : deep_font_and_colors;
  deep_annotation_graphical_defaults : deep_annotation_graphical_defaults;
}
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
