(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_deep_ast.mli,v 1.1 2019-05-10 14:07:14 jpc Exp $ *)

(** From shadow parse tree to deep syntax:
   - we perform various validity checks on the tree obtained from the parser,
   - we build explicit record representations of data instead of simple
     values with associated so-called property lists.
*)

open Ast_types;;
open Mdl_ast;;

type deep_file_name = Path.explicit_file_name
;;

(* Borrowed from Mdl_basics_deep_ast *)

type natural = Mdl_basics_deep_ast.natural
and ident = Mdl_basics_deep_ast.ident
;;

type deep_name = Mdl_basics_deep_ast.deep_name
and deep_value = Mdl_basics_deep_ast.deep_value

and deep_binding = Mdl_basics_deep_ast.deep_binding
and deep_bindings = Mdl_basics_deep_ast.deep_bindings

and deep_mask = Mdl_basics_deep_ast.deep_mask
and deep_sf_block_type = Mdl_basics_deep_ast.deep_sf_block_type
and deep_ports_description = Mdl_basics_deep_ast.deep_ports_description
;;

type deep_block_name = Mdl_basics_deep_ast.deep_block_name
and deep_sid = Mdl_basics_deep_ast.deep_sid
and deep_block_items = Mdl_basics_deep_ast.deep_block_items
;;

type deep_line = Mdl_basics_deep_ast.deep_line
and deep_annotation = Mdl_basics_deep_ast.deep_annotation
;;

type deep_system_name = Mdl_basics_deep_ast.deep_system_name
and deep_prop_name = Mdl_basics_deep_ast.deep_prop_name
and deep_array_name = Mdl_basics_deep_ast.deep_array_name
and deep_class_name = Mdl_basics_deep_ast.deep_class_name
and deep_object_name = Mdl_ast.object_name
and deep_object_type = Mdl_ast.object_type
and deep_object_id = Mdl_ast.object_id
and object_description = Mdl_ast.object_description

and deep_position = Mdl_basics_deep_ast.deep_position
and deep_array_type = Mdl_basics_deep_ast.deep_array_type
and deep_model_defaults = Mdl_basics_deep_ast.deep_model_defaults
;;

type deep_chart_file_number =
  Mdl_stateflow_deep_ast.deep_chart_file_number
;;

(** {6 Deep block types} *)

type deep_block_regular_type = Mdl_block_types.block_type
and deep_block_source_type = Mdl_ast.block_source_type
and deep_block_source_path = Mdl_ast.block_source_path
;;

type deep_block_reference_type = {
  deep_block_source_type : deep_block_source_type;
  deep_block_source_path : deep_block_source_path;
}
;;

(** {6 Deep model interface description} *)

type deep_interface_block_parameter_type =
  | Parameter_type_bool
  | Parameter_type_string
  | Parameter_type_enum
  | Parameter_type_matrix
  | Parameter_type_list
  | Parameter_type_rectangle
;;

type deep_interface_block_binding =
  deep_interface_block_parameter_type Binding.ident_binding

and deep_interface_block_bindings = deep_interface_block_binding list
;;

type deep_interface_block = {
  deep_interface_block_desc : deep_interface_block_desc;
  deep_interface_block_ports_description : deep_ports_description;
  deep_interface_block_parameters : deep_interface_block_bindings;
}

and deep_interface_block_desc =
  | Deep_interface_block_simulink of deep_block_regular_type
  | Deep_interface_block_sub_system of deep_block_reference_type
  | Deep_interface_block_s_function of deep_block_reference_type
;;

(** {6 Deep block and deep system description} *)

type deep_s_function_desc =
  | Deep_s_function
  | Deep_s_function_stateflow of deep_name * deep_chart_file_number
;;

type deep_block = {
  deep_block_name : deep_block_name;
  deep_block_sid : deep_sid option;
  deep_block_body : deep_block_body_elements;
  deep_block_desc : deep_block_desc;
}

and deep_block_desc =
  | Deep_block_regular of deep_block_regular_type
  | Deep_block_reference of deep_block_reference_type
  | Deep_block_sub_system of deep_system
  | Deep_block_s_function of deep_s_function_desc

and deep_block_body_element =
  | Deep_block_mask of deep_mask
  | Deep_block_sf_block_type of deep_sf_block_type
  | Deep_block_bindings of deep_block_items
  | Deep_block_array of deep_array_description
  | Deep_block_object of deep_object_description
  | Deep_block_instance_data of deep_bindings

and deep_block_body_elements = deep_block_body_element list

and deep_system = {
  deep_system_name : deep_system_name;
  deep_system_position : deep_position;
  deep_system_bindings : deep_bindings;
  deep_system_blocks : deep_system_blocks;
  deep_system_lines : deep_lines;
  deep_system_annotations : deep_annotations;
  deep_system_connectors : deep_system_connectors;
}

and deep_array_description = {
  deep_array_type : deep_array_type;
  deep_array_items : deep_array_items;
  deep_array_prop_name : deep_prop_name;
}

and deep_array_items = deep_array_item array
and deep_array_item =
  | Deep_array_name of deep_array_name
  | Deep_array_sub_array of deep_array_description
  | Deep_array_custom_property of Mdl_ast.custom_property
  | Deep_array_matstruct of Mdl_ast.matstruct_items
  | Deep_array_object of deep_object_description
  | Deep_array_vendor_binding of Mdl_ast.vendor_binding
  | Deep_array_vendor_configset of Mdl_ast.vendor_ident * Mdl_ast.vendor_items
  | Deep_array_binding of deep_binding

and deep_object_description = {
  deep_object_id : deep_object_id;
  deep_object_class_name : deep_class_name option;
  deep_object_prop_name : deep_prop_name option;
  deep_object_items : deep_object_items;
}
and deep_object_items = deep_object_item list
and deep_object_item =
  | Deep_object_type of deep_object_type
  | Deep_object_name of deep_object_name
  | Deep_object_dollar_binding of dollar_binding
  | Deep_object_array of deep_array_description
  | Deep_object_custom_property of custom_property
  | Deep_object_object of deep_object_description
  | Deep_object_binding of deep_binding

and deep_lines = deep_line list
and deep_annotations = deep_annotation list

and deep_system_connectors = deep_system_connector list
and deep_system_connector = deep_bindings

and deep_system_blocks = deep_block list
;;

(** {6 Deep model description} *)

type deep_model_configuration_section = {
  deep_graphical_parameters : graphical_interface_item list;
  deep_configuration_bindings : deep_environment;
  deep_vendor_bindings : vendor_bindings;
  deep_object_bindings : object_descriptions;
  deep_model_configurations : deep_model_configurations;
  deep_model_set_configuration_option : model_set_configuration_option;
}

and deep_environment = (ident, deep_value) Hash_table.t

and vendor_bindings = vendor_binding list

and object_descriptions = object_description list

and deep_model_configurations = deep_model_configuration list

and deep_model_configuration = {
  deep_configuration_name : configuration_name;
  deep_configuration_object_id : object_id;
  deep_configuration_vendor_items : vendor_items;
  deep_configuration_vendor_components : vendor_components;
}

and vendor_components = vendor_component list

and vendor_component = component_name * deep_bindings

and component_name = configuration_longident
and configuration_name = mdl_name
;;

type deep_model_body = {
  deep_model_ident : model_ident;
  deep_model_version : deep_model_version;
  deep_model_explicit_file_name : deep_file_name;
  deep_model_configuration_section : deep_model_configuration_section;
  deep_model_defaults : deep_model_defaults;
  deep_model_system : deep_system;
  deep_model_workspace : deep_model_workspace;
}

and deep_model_version =
  | Deep_mdl_model_version of deep_version
  | Deep_slx_model_version of deep_version

and deep_version = {
  deep_major_version : natural;
  deep_minor_version : natural;
}

and deep_model_workspace = deep_model_workspace_data list

and deep_model_workspace_data =
  | Deep_code_workspace_data of deep_model_code_workspace_data
  | Deep_mat_file_workspace_data of Path.file_name

and deep_model_code_workspace_data = {
  workspace_data_source_type : deep_workspace_data_source_type option;
  workspace_data_source_code :
    Mdl_basics_deep_ast.deep_workspace_data_source_code;
}

and deep_workspace_data_source_type =
  | Deep_M_Code
  | Deep_MATLAB_Code
  | Deep_MATLAB_File of Path.file_name
;;

type deep_model_stateflow =
  Mdl_stateflow_deep_ast.deep_stateflow_section
;;

(* End of stateflow section *)

type deep_model_matdatas = deep_model_matdata list
and deep_model_matdata =
  | Deep_matdata_section of deep_matdata_section_item
and deep_matdata_section_item = Mdl_ast.matdata_records
;;

type deep_implementation_model =
  {
    deep_model_body : deep_model_body;
    deep_model_matdatas_option : deep_model_matdatas option;
    deep_model_stateflow_option : deep_model_stateflow option;
  }
;;

(** {6 Interface, implementation, and files} *)

type deep_interface =
  {
    deep_interface_ident : model_ident;
    deep_interface_explicit_file_name : deep_file_name;
    deep_interface_blocks : deep_interface_blocks;
  }

and deep_interface_blocks = deep_interface_block list
;;

type deep_implementation =
  | Deep_implementation_model of deep_implementation_model
  | Deep_implementation_library of deep_implementation_model
;;

type deep_simulink_file =
  | Deep_implementation of deep_implementation
  | Deep_interface of deep_interface
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
