(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_deep_basics.ml,v 1.1 2019-05-10 14:07:14 jpc Exp $ *)

(* From parse tree to deep syntax: the basics. *)

open Mdl_ast;;
open Mdl_basics_deep_ast;;

let deep_error msg =
  Say.fatal_error (
    Printf.sprintf "Deep phase: in Simulink file %s, %s"
      (Configuration.get_source_file ())
      msg
  )
;;

let deep_warning msg =
  Say.warning (
    Printf.sprintf "Deep phase: in Simulink file %s, %s"
      (Configuration.get_source_file ())
      msg
  )
;;

let list_mapi f init =
  let rec loop accu i = function
    | [] -> List.rev accu
    | x :: l -> loop (f i x :: accu) (Unatural.succ i) l in
  loop [] init
;;

let deep_mdl_bool (Ast_types.Mdl_bool _ as mdl_bool) =
  Deep_bool (Ast_types.bool_of_mdl_bool mdl_bool)
;;

let deep_simple_model_value = function
  | Bool s -> Mdl_basics_deep_ast.Deep_bool (Ast_types.bool_of_mdl_string s)
  | Float string -> Mdl_basics_deep_ast.Deep_float string
  | Int string -> Mdl_basics_deep_ast.Deep_int string
  | String string -> Mdl_basics_deep_ast.Deep_string string
  | Ident id -> Mdl_basics_deep_ast.Deep_ident id
;;

let rec deep_model_value = function
  | Value_simple smv -> deep_simple_model_value smv
  | Value_bindings bindings ->
    Mdl_basics_deep_ast.Deep_bindings (deep_bindings bindings)
  | Value_vector svs ->
    Mdl_basics_deep_ast.Deep_vector (List.map deep_simple_model_value svs)
  | Value_matrix vs ->
    Mdl_basics_deep_ast.Deep_matrix (List.map deep_model_value vs)

and deep_bindings bds = List.map deep_binding bds

and deep_binding = function
  | { Binding.bound_ident = _; bound_value = v; } as bdg ->
    { bdg with Binding.bound_value = deep_model_value v; }
;;

let bool_of_simple_value message = function
  | Bool s -> Ast_types.bool_of_mdl_string s
  | String "on" -> true
  | String "off" -> false
  | Float s
  | Int s
  | String s
  | Ident (Ident.Ident s) ->
    deep_error
      (Printf.sprintf "``%s'' is %s, which is not a Simulink boolean"
         message s)
;;

let string_of_simple_value = function
  | Bool s
  | Float s
  | Int s
  | String s -> s
  | Ident (Ident.Ident s) -> s
;;

let simple_value_of_model_value message = function
  | Value_simple smv -> smv
  | Value_bindings _ | Value_vector _ | Value_matrix _ ->
    deep_error (Printf.sprintf "``%s'' is not a simple model value" message)
;;

let bool_of_model_value message mv =
  bool_of_simple_value message (simple_value_of_model_value message mv)
;;

let string_of_model_value message mv =
  string_of_simple_value (simple_value_of_model_value message mv)
;;

let mdl_text_of_model_value message mv =
  Ast_types.mdl_text_of_string (string_of_model_value message mv)
;;

let bool_of_deep_value message = function
  | Deep_bool bool -> bool
  | Deep_string "on" -> true
  | Deep_string "off" -> false
  | Deep_float s
  | Deep_int s
  | Deep_string s ->
    deep_error (Printf.sprintf
      "Wrong boolean for %s: ``%s'' is not a boolean"
      message s)
  | Deep_ident _
  | Deep_wstring _
  | Deep_expression _
  | Deep_code _
  | Deep_bindings _
  | Deep_vector _
  | Deep_matrix _ ->
    deep_error (Printf.sprintf "Wrong boolean for %s" message)
;;

let mdl_string_of_bool = function
  | true -> "on"
  | false -> "off"
;;

let string_of_deep_value message = function
  | Deep_bool b -> mdl_string_of_bool b
  | Deep_float s
  | Deep_int s
  | Deep_string s -> s
  | Deep_wstring _
  | Deep_ident _
  | Deep_expression _
  | Deep_code _
  | Deep_bindings _
  | Deep_vector _
  | Deep_matrix _ ->
    deep_error (Printf.sprintf "Wrong deep value for %s" message)
;;

let matrix_of_deep_value message = function
  | Deep_matrix m -> m
  | Deep_bool _
  | Deep_float _
  | Deep_int _
  | Deep_string _
  | Deep_wstring _
  | Deep_ident _
  | Deep_expression _
  | Deep_code _
  | Deep_bindings _
  | Deep_vector _ ->
    deep_error (Printf.sprintf "Wrong deep value for %s" message)
;;

let vector_of_deep_value message = function
  | Deep_vector v -> v
  | Deep_bool _
  | Deep_float _
  | Deep_int _
  | Deep_string _
  | Deep_wstring _
  | Deep_ident _
  | Deep_expression _
  | Deep_code _
  | Deep_bindings _
  | Deep_matrix _ ->
    deep_error (Printf.sprintf "Wrong deep value for %s" message)
;;

let bindings_of_deep_value message = function
  | Deep_bool _
  | Deep_float _
  | Deep_int _
  | Deep_string _
  | Deep_wstring _
  | Deep_ident _
  | Deep_expression _
  | Deep_code _
  | Deep_vector _
  | Deep_matrix _ ->
    deep_error (Printf.sprintf "Wrong deep value for %s" message)
  | Deep_bindings bindings -> bindings
;;

let make_deep_block_reference_type bl_src_ty bl_src_path =
  { Mdl_deep_ast.
    deep_block_source_type = bl_src_ty;
    deep_block_source_path = bl_src_path;
  }
;;

let deep_block_reference_type_of_strings src path =
  let bl_src_ty =
    Ast_types.mdl_name_of_string src
  and bl_src_path =
    Ast_types.mdl_path_of_string path in
  make_deep_block_reference_type bl_src_ty bl_src_path
;;

(** Model description helpers *)

let deep_bool_of_natural n =
  match Unatural.to_int n with
  | 0 -> false
  | _ -> true
;;

let natural_of_deep_value message = function
  | Mdl_basics_deep_ast.
    Deep_int s -> Unatural.of_string s
  | Deep_string s ->
    deep_warning
      (Printf.sprintf
         "turning string \"%s\" into a natural for %s"
         s message);
    Unatural.of_string s
  | Deep_bool _
  | Deep_float _
  | Deep_wstring _
  | Deep_ident _
  | Deep_expression _
  | Deep_code _
  | Deep_bindings _
  | Deep_vector _
  | Deep_matrix _ ->
    deep_error
      (Printf.sprintf "Unknown natural for %s" message)
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
