(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_deep_mask.ml,v 1.1 2019-05-10 14:07:14 jpc Exp $ *)

(* From parse tree to deep syntax: we handle masking of blocks. *)

(*
open Mdl_block_types;;
See test_data/mdl/explicit/simport/legonxtlib.mdl for
a good example of mask object.
*)

open Ast_types;;
open Mdl_ast;;
open Mdl_basics_deep_ast;;
open Mdl_deep_ast;;
open Mdl_deep_basics;;

(* Mask functions and constants *)

type loop_mask_parameter_ident =
  Mdl_basics_deep_ast.deep_mask_parameter_ident
and loop_mask_parameter_style =
  Mdl_basics_deep_ast.deep_mask_parameter_style option
and loop_mask_parameter_prompt =
  Mdl_basics_deep_ast.deep_mask_parameter_prompt
and loop_mask_parameter_value =
  Mdl_basics_deep_ast.deep_mask_parameter_value option
and loop_mask_parameter_evaluation =
  Mdl_basics_deep_ast.deep_mask_parameter_evaluation option
and loop_mask_parameter_enable =
  Mdl_basics_deep_ast.deep_mask_parameter_enable option
and loop_mask_parameter_visible =
  Mdl_basics_deep_ast.deep_mask_parameter_visible option
and loop_mask_parameter_internal =
  Mdl_basics_deep_ast.deep_mask_parameter_internal option
and loop_mask_parameter_never_save =
  Mdl_basics_deep_ast.deep_mask_parameter_never_save option
and loop_mask_parameter_read_only =
  Mdl_basics_deep_ast.deep_mask_parameter_read_only option
;;

type loop_mask_parameter =
  {
    loop_mask_parameter_var_num : Unatural.t;
    loop_mask_parameter_ident : loop_mask_parameter_ident;
    loop_mask_parameter_style : loop_mask_parameter_style;
    loop_mask_parameter_prompt : loop_mask_parameter_prompt;
    loop_mask_parameter_value : loop_mask_parameter_value;
    loop_mask_parameter_evaluation : loop_mask_parameter_evaluation;
    loop_mask_parameter_enable : loop_mask_parameter_enable;
    loop_mask_parameter_visible : loop_mask_parameter_visible;
    loop_mask_parameter_internal : loop_mask_parameter_internal;
    loop_mask_parameter_never_save : loop_mask_parameter_never_save;
    loop_mask_parameter_read_only : loop_mask_parameter_read_only;
  }
;;

type loop_mask_parameter_classifier = {
  loop_mask_parameter_ident_opt :
    Ident.ident option;
  loop_mask_parameter_style_opt :
    Mdl_basics_deep_ast.deep_mask_parameter_style option;
  loop_mask_parameter_style_options_opt :
    Mdl_basics_deep_ast.mask_parameter_style_item list option;
  loop_mask_parameter_prompt_opt :
    Ast_types.mdl_text option;
  loop_mask_parameter_value_opt :
    Mdl_basics_deep_ast.deep_mask_parameter_value option;
  loop_mask_parameter_eval_opt :
    Mdl_basics_deep_ast.deep_mask_parameter_evaluation option;
  loop_mask_parameter_visible_opt :
    Mdl_basics_deep_ast.deep_mask_parameter_visible option;
  loop_mask_parameter_enable_opt :
    Mdl_basics_deep_ast.deep_mask_parameter_enable option;
  loop_mask_parameter_range_opt :
    Mdl_basics_deep_ast.mask_parameter_style_range option;
}
;;

type loop_mask_parameters_description =
  {
    (* From MDL MaskPromptString "Base Address|Sample Time" *)
    loop_mask_prompts : mdl_pipe_texts option;
    (* From MDL MaskVariables "BaseAddress=@1;T0=@2;" *)
    loop_mask_variables : mdl_semi_texts option;
    (* From Mdl MaskEnableString "on,on" *)
    loop_mask_enables : mdl_comma_bools option;
    (* From MDL MaskVisibilityString "on,on" *)
    loop_mask_visibilities : mdl_comma_bools option;
    (* From MDL MaskValueString "sqrt(8/5)|-1/(Rb*Spb)|8 Bit|on|T0" *)
    loop_mask_values : mdl_pipe_texts option;
    (* From MDL MaskStyleString "edit,popup(0|1|2|3),edit" *)
    loop_mask_styles : mdl_comma_texts option;
  }
;;

let fdebug f =
  Say.fdebug (fun epf -> Format.fprintf epf "Mdl_deep_mask, %t" f)
;;

let fatal_error s =
  Mdl_deep_basics.deep_error (Printf.sprintf "Mdl_deep_mask, %s" s)
;;

let empty_loop_mask_parameters_description =
  {
    loop_mask_prompts = None;
    loop_mask_variables = None;
    loop_mask_enables = None;
    loop_mask_visibilities = None;
    loop_mask_values = None;
    loop_mask_styles = None;
  }
;;

let empty_loop_mask_parameter_classifier = {
  loop_mask_parameter_ident_opt = None;
  loop_mask_parameter_style_opt = None;
  loop_mask_parameter_style_options_opt = None;
  loop_mask_parameter_prompt_opt = None;
  loop_mask_parameter_value_opt = None;
  loop_mask_parameter_eval_opt = None;
  loop_mask_parameter_visible_opt = None;
  loop_mask_parameter_enable_opt = None;
  loop_mask_parameter_range_opt = None;
}
;;

let deep_source_code_of_mdl_code = function
  | Mdl_code code -> Deep_source_code code
;;

let split_string_at_pipe = fun s ->
  let ib = Scanf.Scanning.from_string s in
  let rec loop accu =
    if Scanf.Scanning.end_of_input ib then List.rev accu else
    Scanf.bscanf ib "|%[^|]"
      (fun mask_part -> loop (mask_part :: accu)) in
  try
    Scanf.bscanf ib "%[^|]"
      (fun mask_part -> loop [ mask_part ])
  with
  | Scanf.Scan_failure error
  | Failure error ->
    fatal_error
      (Printf.sprintf
         "split_string_at_pipe, \
          while scanning ``%s'', encountered error %s"
         s
         error)
;;

let find_mask_prompts (Pipe_texts txt) =
  let prompt_strings = split_string_at_pipe (string_of_mdl_text txt) in
  List.map (fun s -> mdl_text_of_string s) prompt_strings
;;

let find_mask_values s =
  let value_strings = split_string_at_pipe s in
  value_strings
;;

let split_string_at_comma = Lib_strings.split_string_at_comma;;

let find_mask_enables (Comma_bools enables_string) =
  split_string_at_comma (string_of_mdl_text enables_string)
;;

let find_mask_visibilities = find_mask_enables
and find_mask_styles (Comma_texts mask_style_string) =
  split_string_at_comma (string_of_mdl_text mask_style_string)
;;

(** {6 Legacy masks defined as a set of block parameter bindings } *)

let read_well_parenthesed_expression ib =

  let well_parenthesed_failwith m =
    failwith (Printf.sprintf
      "read_well_parenthesed_expression: can not find %s"
      m) in

  let check_end_of_file message ib =
    if Scanf.Scanning.end_of_input ib
    then well_parenthesed_failwith message in

  let buffer = Buffer.create 64 in
  let add_current_char ib =
    Scanf.bscanf ib "%c" (fun c -> Buffer.add_char buffer c) in
  let add_current_string s =
    Buffer.add_string buffer s in

  let rec state_0 ib =
    check_end_of_file "an openning parenthesis" ib;
    Scanf.bscanf ib "%0c"
      (function
        | '(' ->
          add_current_char ib;
          state_1 ib
        | _ ->
          well_parenthesed_failwith "an openning parenthesis")

  (* looking for a closing paren ')' *)
  and state_1 ib =
    check_end_of_file "a closing parenthesis" ib;
    Scanf.bscanf ib "%[^()]" (fun s ->
      add_current_string s;
      check_end_of_file "a closing parenthesis" ib;
      Scanf.bscanf ib "%0c" (function
        | '(' ->
          add_current_char ib;
          state_1 ib;
          state_1 ib
        | ')' ->
          add_current_char ib
        | _ -> assert false)) in

  state_0 ib;
  Buffer.contents buffer
;;

let mask_parameter_style_items ss =
  fdebug (fun epf ->
    Format.fprintf epf "mask_parameter_style_items <- %a"
      (Og_printer.printer_List
         Og_printer.printer_String Og_precedence.Ck_record_field)
      ss);
  let style_items =
    List.map
      (fun s ->
       try Mdl_basics_deep_ast.Item_int (int_of_string s) with
       | Failure _ -> Mdl_basics_deep_ast.Item_text (mdl_text_of_string s))
      ss in
  fdebug (fun epf ->
    Format.fprintf epf "mask_parameter_style_items -> %a"
      (Og_printer.printer_List
         Mdl_basics_deep_ast_print.printer_mask_parameter_style_item
         Og_precedence.Ck_record_field)
      style_items);
  style_items
;;

let remove_surrounding_parentheses str =
  fdebug (fun epf ->
    Format.fprintf epf "remove_surrounding_parentheses <- %S" str);
  assert (String.length str >= 2);
  assert (str.[0] = '(');
  let len = String.length str in
  assert (str.[pred len] = ')');
  let str = String.sub str 1 (len - 2) in
  fdebug (fun epf ->
    Format.fprintf epf "remove_surrounding_parentheses -> %S" str);
  str
;;

let deep_mask_parameter_style_of_string = function
  | "edit" -> Mdl_basics_deep_ast.Deep_mask_parameter_edit
  | "checkbox" -> Deep_mask_parameter_checkbox
  | "min" -> Deep_mask_parameter_min
  | "max" -> Deep_mask_parameter_max
  | "promote" -> Deep_mask_parameter_promote
  | "popup" -> Deep_mask_parameter_popup []
  | "radiobutton" -> Deep_mask_parameter_radiobutton []
  | "unidt" -> Deep_mask_parameter_unidt (mdl_text_of_string "")
  | "slider" -> Deep_mask_parameter_slider None
  | "spinbox" -> Deep_mask_parameter_spinbox None
  | style_string ->
    fdebug
      (fun epf ->
       Format.fprintf
         epf "deep_mask_parameter_style_of_string style_string = %s"
         style_string);
    try
       let ib = Scanf.Scanning.from_string style_string in
       Scanf.kscanf ib
         (fun ib _exn ->
          Scanf.bscanf ib "unidt %r %!"
            read_well_parenthesed_expression
            (fun unidt_arg ->
             Deep_mask_parameter_unidt (mdl_text_of_string unidt_arg)))
         "popup %r %!"
         read_well_parenthesed_expression
         (fun choices ->
          let choices = remove_surrounding_parentheses choices in
          fdebug (fun epf ->
            Format.fprintf epf
              "deep_mask_parameter_style_of_string: choices %S" choices);
          Deep_mask_parameter_popup
            (mask_parameter_style_items (split_string_at_pipe choices)))
     with
     | End_of_file ->
       fatal_error
         (Printf.sprintf
            "deep_mask_parameter_style_of_string, \
             while scanning mask style ``%s'', encountered EOF"
             style_string)
     | Scanf.Scan_failure s
     | Failure s ->
       fatal_error
         (Printf.sprintf
            "deep_mask_parameter_style_of_string, \
             while scanning mask style ``%s'', encountered %s"
             style_string
             s)
;;

let get_with_fatal_error field_opt err_msg =
  match field_opt with
  | None -> fatal_error err_msg
  | Some field -> field
;;

let mask_parameters_of_loop_mask_parameters loop_mask_parameters =

  let rec loop accu = function
    | [] -> List.rev accu
    | {
        loop_mask_parameter_var_num = var_num;
        loop_mask_parameter_ident = loop_mask_parameter_ident;
        loop_mask_parameter_style = loop_mask_parameter_style;
        loop_mask_parameter_prompt = loop_mask_parameter_prompt;
        loop_mask_parameter_value = loop_mask_parameter_value;
        loop_mask_parameter_evaluation = loop_mask_parameter_evaluation;
        loop_mask_parameter_enable = loop_mask_parameter_enable;
        loop_mask_parameter_visible = loop_mask_parameter_visible;
        loop_mask_parameter_internal = loop_mask_parameter_internal;
        loop_mask_parameter_never_save = loop_mask_parameter_never_save;
        loop_mask_parameter_read_only = loop_mask_parameter_read_only;
      } :: loop_mask_parameters ->
      let mask_parameter_style =
        get_with_fatal_error
          loop_mask_parameter_style
          "no style for mask parameter" in
      let mask_parameter_value =
        get_with_fatal_error
          loop_mask_parameter_value
          "no value for mask parameter" in
      loop (
          (var_num,
           {
             deep_mask_parameter_ident = loop_mask_parameter_ident;
             deep_mask_parameter_style = mask_parameter_style;
             deep_mask_parameter_prompt = loop_mask_parameter_prompt;
             deep_mask_parameter_callback = None;
             deep_mask_parameter_tunable = None;
             deep_mask_parameter_value = mask_parameter_value;
             deep_mask_parameter_evaluation =
               loop_mask_parameter_evaluation;
             deep_mask_parameter_enable = loop_mask_parameter_enable;
             deep_mask_parameter_visible = loop_mask_parameter_visible;
             deep_mask_parameter_internal = loop_mask_parameter_internal;
             deep_mask_parameter_never_save = loop_mask_parameter_never_save;
             deep_mask_parameter_read_only = loop_mask_parameter_read_only;
           }) :: accu
        )
        loop_mask_parameters in

  loop [] loop_mask_parameters
;;

let make_loop_mask_parameter var var_num mask_prompt evaluation =
  {
    loop_mask_parameter_var_num = var_num;
    loop_mask_parameter_ident = var;
    loop_mask_parameter_style = None;
    loop_mask_parameter_prompt = mask_prompt;
    loop_mask_parameter_value = None;
    loop_mask_parameter_evaluation = evaluation;
    loop_mask_parameter_enable = None;
    loop_mask_parameter_visible = None;
    loop_mask_parameter_internal = None;
    loop_mask_parameter_never_save = None;
    loop_mask_parameter_read_only = None;
  }
;;

let deep_mask_parameter_enable_of_string = function
  | "on" -> Enable
  | "off" -> Disable
  | s ->
    fatal_error (Printf.sprintf
      "invalid enable mode specification ``%s''" s)
;;

let deep_mask_parameter_visible_of_string = function
  | "on" -> Visible
  | "off" -> Hidden
  | s ->
    fatal_error (Printf.sprintf
      "invalid visibility specification ``%s''" s)
;;

let make_mask_ident var_num =
  Ident.Ident (Printf.sprintf "%s%d" "m_a_s_k_" (Unatural.to_int var_num))
;;

let make_mask_variable var_num prompt =
  let var = make_mask_ident var_num in
  make_loop_mask_parameter var var_num prompt None
;;

let make_mask_variables mask_prompts =
  list_mapi make_mask_variable Unatural._1_ mask_prompts
;;

let get_mask_variables (Semi_texts mask_variables) mask_prompts =

  let mask_variables = string_of_mdl_text mask_variables in
  fdebug (fun epf ->
    Format.fprintf epf "mask_variables is %S"
      mask_variables);
  let ib = Scanf.Scanning.from_string mask_variables in

  let rec loop i accu = function
    | [] -> List.rev accu
    | mask_prompt :: mask_prompts ->
      if Scanf.Scanning.end_of_input ib then
        let mask_variable = make_mask_variable i mask_prompt in
        loop (Unatural.succ i) (mask_variable :: accu) mask_prompts
      else
        Scanf.bscanf ib "%s@=%c%d%[;]"
          (fun var var_code var_num semicolons ->
           if String.length semicolons = 0 &&
              not (Scanf.Scanning.end_of_input ib) then
             fatal_error (Printf.sprintf
               "get_mask_variables, \
                while scanning ``%s'', missing ';' separator"
               mask_variables) else
           let var_num = Unatural_types.to_natural var_num in
           let evaluate =
             match var_code with
             | '&' -> Quote
             | '@' -> Eval
             | c ->
               fatal_error
                 (Printf.sprintf
                    "get_mask_variables, \
                     while scanning ``%s'', encountered %C char separator"
                    mask_variables
                    c) in
           let mask_variable =
             if var_num = i then
               let ident = Ident.Ident var in
               make_loop_mask_parameter
                 ident var_num mask_prompt (Some evaluate)
             else make_mask_variable i mask_prompt in
           loop (Unatural.succ i) (mask_variable :: accu) mask_prompts) in

  try
    loop Unatural._1_ [] mask_prompts
  with
  | End_of_file ->
    fatal_error
      (Printf.sprintf
         "get_mask_variables, while scanning ``%s'', encountered EOF"
         mask_variables)
  | Scanf.Scan_failure s
  | Failure s ->
    fatal_error
      (Printf.sprintf
         "get_mask_variables, while scanning ``%s'', encountered %s"
         mask_variables
         s)
;;

let get_mask_items_enables loop_mask_enables loop_mask_parameters =

  let enables = find_mask_enables loop_mask_enables in

  let add_mask_enable loop_mask_parameter enable_string =
    let enable = deep_mask_parameter_enable_of_string enable_string in
    {
      loop_mask_parameter with
      loop_mask_parameter_enable = Some enable;
    } in

  try
    List.rev
      (List.rev_map2
         add_mask_enable loop_mask_parameters enables)
  with
  | _ ->
    fatal_error
      "get_mask_items_enables, \
       enable modes are not on par with mask parameters"
;;

let get_mask_items_visibles loop_mask_visibilities loop_mask_parameters =

  let visibilities = find_mask_visibilities loop_mask_visibilities in

  let add_mask_visible loop_mask_parameter visible_string =
    let visible = deep_mask_parameter_visible_of_string visible_string in
    {
      loop_mask_parameter with
      loop_mask_parameter_visible = Some visible;
    } in

  try
    List.rev
      (List.rev_map2
         add_mask_visible loop_mask_parameters visibilities)
  with
  | _ ->
    fatal_error
      "get_mask_items_visibles, \
       visibility modes are not on par with mask parameters"
;;

let combine_comma_separated_strings ppf = function
  | [] -> ()
  | mask_value :: mask_values ->
    Format.fprintf ppf "``%s''%a"
      mask_value
      (fun ppf ->
       List.iter (Format.fprintf ppf ",@ ``%s''")) mask_values
;;

let print_comma_separated_strings strings =
  combine_comma_separated_strings Format.str_formatter strings;
  Format.flush_str_formatter ()
;;

let print_mask_values = print_comma_separated_strings
and print_mask_styles = print_comma_separated_strings
;;

let print_loop_mask_parameters loop_mask_parameters =
  let mask_parameter_strings =
    List.map
      (fun loop_mask_parameter ->
       string_of_ident (loop_mask_parameter.loop_mask_parameter_ident))
      loop_mask_parameters in
  combine_comma_separated_strings Format.str_formatter mask_parameter_strings;
  Format.flush_str_formatter ()
;;

let get_mask_items_values (Pipe_texts mask_value_string) loop_mask_parameters =

  let mask_value_string = string_of_mdl_text mask_value_string in
  let mask_values = find_mask_values mask_value_string in

  let add_mask_value loop_mask_parameter s =
    {
      loop_mask_parameter with
      loop_mask_parameter_value =
        Some (Deep_mask_parameter_value (mdl_text_of_string s));
    } in

  try
    List.rev
      (List.rev_map2 add_mask_value
         loop_mask_parameters mask_values)
  with
  | _ ->
    let num_values = List.length mask_values in
    let num_params = List.length loop_mask_parameters in
    let mask_values_string =
      print_mask_values mask_values in
    let loop_mask_parameters_string =
      print_loop_mask_parameters loop_mask_parameters in
    fatal_error (Format.sprintf
      "get_mask_items_values, \
       there are %d mask parameter values \
       (%s), for the following %d mask parameters:@ \
       %s."
      num_values
      mask_values_string
      num_params
      loop_mask_parameters_string)
;;

let get_mask_items_styles mask_style_string loop_mask_parameters =

  let mask_styles = find_mask_styles mask_style_string in

  let add_mask_style loop_mask_parameter s =
    let style = deep_mask_parameter_style_of_string s in
    {
      loop_mask_parameter with
      loop_mask_parameter_style = Some style;
    } in

  try
    List.rev
      (List.rev_map2
         add_mask_style loop_mask_parameters mask_styles)
  with
  | _ ->
    let num_styles = List.length mask_styles in
    let num_params = List.length loop_mask_parameters in
    let mask_styles_string =
      print_mask_styles mask_styles in
    let loop_mask_parameters_string =
      print_loop_mask_parameters loop_mask_parameters in
    fatal_error (Format.sprintf
      "get_mask_items_styles, \
       there are %d mask parameter styles \
       (%s), for the following %d mask parameters:@ \
       %s."
      num_styles
      mask_styles_string
      num_params
      loop_mask_parameters_string)
;;

let substitute_variables_in_script loop_mask_parameters script =

  let script_contents = Ast_types.string_of_mdl_shallow_code script in

  fdebug
    (fun ppf ->
     Format.fprintf ppf "Substituting in script |%s|" script_contents);

  let loop_mask_parameters =
    Array.of_list
      (List.sort
        (fun lmp1 lmp2 ->
         Pervasives.compare
           lmp1.loop_mask_parameter_var_num
           lmp2.loop_mask_parameter_var_num)
        loop_mask_parameters) in

  (* Mask variable number i is normally associated to some matlab ident id
     and to be replaced by id in the script;
     if associated ident id is empty, then variable number i should be
     substituted by its associated value field. *)
  let get_mask_variable i =
    let loop_mask_parameter = loop_mask_parameters.(i) in
    string_of_ident loop_mask_parameter.loop_mask_parameter_ident in

  let ib = Scanf.Scanning.from_string script_contents in
  let b = Buffer.create (6 * String.length script_contents) in

  let rec loop () =
    if Scanf.Scanning.end_of_input ib then Buffer.contents b else
    Scanf.bscanf ib "%c"
      (function
       | ('&' | '@') as c ->
         if Scanf.Scanning.end_of_input ib then go_on c else
         Scanf.bscanf ib "%0c"
           (function
            | '1' .. '9' ->
              Scanf.bscanf ib "%d"
                (function i ->
                 let s = get_mask_variable (i - 1) in
                 fdebug
                   (fun ppf ->
                     Format.fprintf ppf "Substituting variable %d by |%s|"
                       i s);
                 Buffer.add_string b s;
                 loop ())
            | _ -> go_on c)
       | c -> go_on c)

  and go_on c = Buffer.add_char b c; loop () in

  try Deep_source_code (mdl_text_of_string (loop ())) with
  | Scanf.Scan_failure s
  | Failure s ->
    fatal_error
      (Printf.sprintf
         "get_mask_and_context, \
          while scanning ``%s'', encountered error %s"
         script_contents
         s)
;;

let get_loop_mask_parameters mask_parameters_description =

  match
    mask_parameters_description.loop_mask_prompts,
    mask_parameters_description.loop_mask_values,
    mask_parameters_description.loop_mask_styles
  with
  | None, None, None -> []
  | Some mask_prompt_string,
    Some mask_value_string,
    Some mask_style_string ->

    let mask_prompts = find_mask_prompts mask_prompt_string in

    let loop_mask_parameters =
      match mask_parameters_description.loop_mask_variables with
      | None -> make_mask_variables mask_prompts
      | Some mask_variables ->
        get_mask_variables mask_variables mask_prompts in

    let loop_mask_parameters =
      get_mask_items_values mask_value_string loop_mask_parameters in

    let loop_mask_parameters =
      get_mask_items_styles mask_style_string loop_mask_parameters in

    let loop_mask_parameters =
      match mask_parameters_description.loop_mask_enables with
      | None -> loop_mask_parameters
      | Some mask_enables ->
        get_mask_items_enables mask_enables loop_mask_parameters in

    let loop_mask_parameters =
      match mask_parameters_description.loop_mask_visibilities with
      | None -> loop_mask_parameters
      | Some mask_visibilities ->
        get_mask_items_visibles mask_visibilities loop_mask_parameters in

    loop_mask_parameters

  | mask_prompt_string_option,
    mask_value_string_option,
    mask_style_string_option ->

    let prompt_string_info =
      match mask_prompt_string_option with
      | None -> ""
      | Some (Pipe_texts mask_prompt_string) ->
        Printf.sprintf
          " mask prompt string is ``%s''"
          (Ast_types.string_of_mdl_text mask_prompt_string) in
    let value_string_info =
      match mask_value_string_option with
      | None -> ""
      | Some (Pipe_texts mask_value_string) ->
        Printf.sprintf
          " mask value string is ``%s''"
          (Ast_types.string_of_mdl_text mask_value_string) in
    let style_string_info =
      match mask_style_string_option with
      | None -> ""
      | Some (Comma_texts mask_style_string) ->
        Printf.sprintf
          " mask style string is ``%s''"
          (Ast_types.string_of_mdl_text mask_style_string) in

    let error_message =
      Printf.sprintf
        "missing string values in mask:%s%s%s"
        prompt_string_info value_string_info style_string_info in
    fatal_error error_message
;;

let get_deep_script loop_mask_parameters script_opt =
  match script_opt with
  | None -> None
  | Some shallow_script ->
    let script =
      substitute_variables_in_script loop_mask_parameters shallow_script in
    Some script
;;

let make_binding = Mdl_ast_functions.make_binding;;

let make_deep_binding_str_string id ws =
  make_binding (ident_from_string id) (Deep_string ws)
;;

let make_deep_binding_str_expression id e =
  make_binding (ident_from_string id) (Deep_expression e)
;;
(***
let make_deep_binding_str_code id s =
  make_binding (ident_from_string id) (Deep_code s)
;;
***)
let make_deep_binding_str_ident id s =
  make_binding (ident_from_string id) (Deep_ident s)
;;

let make_deep_binding_str_bool id mdl_bool =
  make_binding (ident_from_string id) (Mdl_deep_basics.deep_mdl_bool mdl_bool)
;;

let check_opt error opt v =
  match opt with
  | None -> Some v
  | Some _ -> fatal_error (Printf.sprintf "Multiple definition of %s" error)
;;

let check_type_opt = check_opt "mask type"
and check_descr_opt = check_opt "mask description"
and check_init_opt = check_opt "mask initialisation"
and check_display_opt = check_opt "mask display"
;;

let check_loop_mask_parameters_prompts parameters s =
  {
    parameters with
    loop_mask_prompts =
      check_opt "mask parameters prompts" parameters.loop_mask_prompts s;
  }

and check_loop_mask_parameters_variables parameters s =
  {
    parameters with
    loop_mask_variables =
      check_opt "mask parameters variables" parameters.loop_mask_variables s;
  }

and check_loop_mask_parameters_enables parameters s =
  {
    parameters with
    loop_mask_enables =
      check_opt "mask parameters enable modes"
        parameters.loop_mask_enables s;
  }

and check_loop_mask_parameters_visibilities parameters s =
  {
    parameters with
    loop_mask_visibilities =
      check_opt "mask parameters visibility modes"
        parameters.loop_mask_visibilities s;
  }

and check_loop_mask_parameters_values parameters s =
  {
    parameters with
    loop_mask_values =
      check_opt "mask parameters values string"
        parameters.loop_mask_values s;
  }

and check_loop_mask_parameters_styles parameters s =
  {
    parameters with
    loop_mask_styles =
      check_opt "mask parameters styles string"
        parameters.loop_mask_styles s;
  }
;;

(* The main function for legacy masks. *)
let classify_mask_items bindings =

  let rec loop
    type_opt descr_opt init_opt display_opt
    loop_mask_parameters_description bindings = function
  | [] ->
    type_opt, descr_opt, init_opt, display_opt,
    loop_mask_parameters_description, List.rev bindings
  | bmi :: bmis ->
    match bmi with
    | Mask_callback_string (Pipe_expressions mdl_pipe_expressions) ->
      let bdg =
        make_deep_binding_str_string "Mask_callback_string"
          (string_of_mdl_text mdl_pipe_expressions) in
      let bindings = bdg :: bindings in
      loop type_opt descr_opt init_opt display_opt
        loop_mask_parameters_description bindings bmis
    | Mask_capabilities (Expression mdl_expression) ->
      let bdg =
        make_deep_binding_str_expression "Mask_capabilities" mdl_expression in
      let bindings = bdg :: bindings in
      loop type_opt descr_opt init_opt display_opt
        loop_mask_parameters_description bindings bmis
    | Mask_description text ->
      let descr_opt = check_descr_opt descr_opt text in
      loop type_opt descr_opt init_opt display_opt
        loop_mask_parameters_description bindings bmis
    | Mask_display code ->
      let display_opt = check_display_opt display_opt code in
      loop type_opt descr_opt init_opt display_opt
        loop_mask_parameters_description bindings bmis
    | Mask_display_string mdl_text ->
      let bdg =
        make_deep_binding_str_string "Mask_display_string"
          (string_of_mdl_text mdl_text) in
      let bindings = bdg :: bindings in
      loop type_opt descr_opt init_opt display_opt
        loop_mask_parameters_description bindings bmis
    | Mask_enable_string mdl_comma_bools ->
      let loop_mask_parameters_description =
        check_loop_mask_parameters_enables
          loop_mask_parameters_description mdl_comma_bools in
      loop type_opt descr_opt init_opt display_opt
        loop_mask_parameters_description bindings bmis
    | Mask_help text ->
      let bdg =
        make_deep_binding_str_string "Mask_help" (string_of_mdl_text text) in
      let bindings = bdg :: bindings in
      loop type_opt descr_opt init_opt display_opt
        loop_mask_parameters_description bindings bmis
    | Mask_hide_contents mdl_bool ->
      let bdg = make_deep_binding_str_bool "Mask_hide_contents" mdl_bool in
      let bindings = bdg :: bindings in
      loop type_opt descr_opt init_opt display_opt
        loop_mask_parameters_description bindings bmis
    | Mask_icon_frame mdl_bool ->
      let bdg = make_deep_binding_str_bool "Mask_icon_frame" mdl_bool in
      let bindings = bdg :: bindings in
      loop type_opt descr_opt init_opt display_opt
        loop_mask_parameters_description bindings bmis
    | Mask_icon_opaque mdl_bool ->
      let bdg = make_deep_binding_str_bool "Mask_icon_opaque" mdl_bool in
      let bindings = bdg :: bindings in
      loop type_opt descr_opt init_opt display_opt
        loop_mask_parameters_description bindings bmis
    | Mask_icon_rotate id ->
      let bdg = make_deep_binding_str_ident "Mask_icon_rotate" id in
      let bindings = bdg :: bindings in
      loop type_opt descr_opt init_opt display_opt
        loop_mask_parameters_description bindings bmis
    | Mask_icon_units id ->
      let bdg = make_deep_binding_str_ident "Mask_icon_units" id in
      let bindings = bdg :: bindings in
      loop type_opt descr_opt init_opt display_opt
        loop_mask_parameters_description bindings bmis
    | Mask_initialization shallow_code ->
      (* From MDL
         MaskInitialization "ax = [0, Tend,ylimits(1),ylimits(2)];dt=-1;" *)
      let init_opt = check_init_opt init_opt shallow_code in
      loop type_opt descr_opt init_opt display_opt
        loop_mask_parameters_description bindings bmis
    | Mask_port_rotate id ->
      let bdg = make_deep_binding_str_ident "Mask_port_rotate" id in
      let bindings = bdg :: bindings in
      loop type_opt descr_opt init_opt display_opt
        loop_mask_parameters_description bindings bmis
    | Mask_prompt_string mdl_comma_texts ->
      let loop_mask_parameters_description =
        check_loop_mask_parameters_prompts
          loop_mask_parameters_description mdl_comma_texts in
      loop type_opt descr_opt init_opt display_opt
        loop_mask_parameters_description bindings bmis
    | Mask_run_init_for_icon_redraw mdl_bool ->
      let bdg = make_deep_binding_str_bool "Mask_run_init_for_icon_redraw" mdl_bool in
      let bindings = bdg :: bindings in
      loop type_opt descr_opt init_opt display_opt
        loop_mask_parameters_description bindings bmis
    | Mask_self_modifiable mdl_bool ->
      let bdg = make_deep_binding_str_bool "Mask_self_modifiable" mdl_bool in
      let bindings = bdg :: bindings in
      loop type_opt descr_opt init_opt display_opt
        loop_mask_parameters_description bindings bmis
    | Mask_style_string mdl_comma_texts ->
      let loop_mask_parameters_description =
        check_loop_mask_parameters_styles
          loop_mask_parameters_description mdl_comma_texts in
      loop type_opt descr_opt init_opt display_opt
        loop_mask_parameters_description bindings bmis
    | Mask_tab_name_string text ->
      let bdg =
        make_deep_binding_str_string "Mask_tab_name_string" (string_of_mdl_text text) in
      let bindings = bdg :: bindings in
      loop type_opt descr_opt init_opt display_opt
        loop_mask_parameters_description bindings bmis
    | Mask_tooltip_string (Comma_bools mdl_comma_bools) ->
      let bdg =
        make_deep_binding_str_string "Mask_tooltip_string"
          (string_of_mdl_text mdl_comma_bools) in
      let bindings = bdg :: bindings in
      loop type_opt descr_opt init_opt display_opt
        loop_mask_parameters_description bindings bmis
    | Mask_tunable_value_string (Comma_bools mdl_comma_bools) ->
      let bdg =
        make_deep_binding_str_string "Mask_tunable_value_string"
          (string_of_mdl_text mdl_comma_bools) in
      let bindings = bdg :: bindings in
      loop type_opt descr_opt init_opt display_opt
        loop_mask_parameters_description bindings bmis
    | Mask_type text ->
      let type_opt = check_type_opt type_opt text in
      loop type_opt descr_opt init_opt display_opt
        loop_mask_parameters_description bindings bmis
    | Mask_value_string mdl_pipe_texts ->
      let loop_mask_parameters_description =
        check_loop_mask_parameters_values
          loop_mask_parameters_description mdl_pipe_texts in
      loop type_opt descr_opt init_opt display_opt
        loop_mask_parameters_description bindings bmis
    | Mask_var_alias_string text ->
      let bdg =
        make_deep_binding_str_string "Mask_var_alias_string"
          (string_of_mdl_text text) in
      let bindings = bdg :: bindings in
      loop type_opt descr_opt init_opt display_opt
        loop_mask_parameters_description bindings bmis
    | Mask_variables mdl_semi_texts ->
      let loop_mask_parameters_description =
        check_loop_mask_parameters_variables
          loop_mask_parameters_description mdl_semi_texts in
      loop type_opt descr_opt init_opt display_opt
        loop_mask_parameters_description bindings bmis
    | Mask_visibility_string s ->
      let loop_mask_parameters_description =
        check_loop_mask_parameters_visibilities
          loop_mask_parameters_description s in
      loop type_opt descr_opt init_opt display_opt
        loop_mask_parameters_description bindings bmis in

  loop None None None None empty_loop_mask_parameters_description [] bindings
;;

type shallow_mask = {
  shallow_mask_type : deep_mask_type option;
  shallow_mask_description : deep_mask_description option;
  shallow_mask_initialisation : mdl_shallow_code option;
  shallow_mask_display : deep_source_code option;
  shallow_mask_parameters : loop_mask_parameters;
  shallow_mask_bindings : Mdl_deep_ast.deep_bindings;
}

and loop_mask_parameters = loop_mask_parameter list
;;

let deep_mask_of_shallow_mask shallow_mask =
  match shallow_mask with
  | {
      shallow_mask_type = shallow_mask_type;
      shallow_mask_description = shallow_mask_description;
      shallow_mask_initialisation = shallow_mask_initialisation;
      shallow_mask_display = shallow_mask_display;
      shallow_mask_parameters = loop_mask_parameters;
      shallow_mask_bindings = shallow_mask_bindings;
    } ->
    let deep_mask_initialisation =
      get_deep_script loop_mask_parameters shallow_mask_initialisation in
    let shallow_mask_parameters =
      List.map snd
        (mask_parameters_of_loop_mask_parameters loop_mask_parameters) in

    {
      deep_mask_type = shallow_mask_type;
      deep_mask_description = shallow_mask_description;
      deep_mask_initialisation = deep_mask_initialisation;
      deep_mask_display = shallow_mask_display;
      deep_mask_parameters = shallow_mask_parameters;
      deep_mask_bindings = shallow_mask_bindings;
    }
;;

let deep_mask_of_mask_items block_mask_items =
  let shallow_mask_type,
      shallow_mask_description,
      shallow_mask_initialisation,
      shallow_mask_display,
      loop_mask_parameters_description,
      shallow_mask_bindings =
    classify_mask_items block_mask_items in

  let loop_mask_parameters =
    get_loop_mask_parameters
      loop_mask_parameters_description in

  let shallow_mask_display =
    Uoption.map deep_source_code_of_mdl_code shallow_mask_display in

  let shallow_mask =
    {
      shallow_mask_type = shallow_mask_type;
      shallow_mask_description = shallow_mask_description;
      shallow_mask_initialisation = shallow_mask_initialisation;
      shallow_mask_display = shallow_mask_display;
      shallow_mask_parameters = loop_mask_parameters;
      shallow_mask_bindings = shallow_mask_bindings;
    } in
  deep_mask_of_shallow_mask shallow_mask
;;

(* End of mask by mask_items functions *)

(** {6 Masks defined as objects within the block body} *)

let deep_mask_parameter_evaluation_of_bool = function
  | true -> Mdl_basics_deep_ast.Eval
  | false -> Mdl_basics_deep_ast.Quote
;;

let deep_mask_parameter_evaluation_of_deep_value v =

  let mode =
    Mdl_deep_basics.bool_of_deep_value
      "Property evaluate of a mask parameter" v in
  deep_mask_parameter_evaluation_of_bool mode
;;

let parameters_ident = Ast_types.ident_from_string "Parameters";;

let deep_mask_of_object deep_object_items =

  let find_value_in_deep_object_items ident =

    let rec find_value = function
    | [] -> None
    | (Deep_object_type _
      | Deep_object_name _
      | Deep_object_dollar_binding _
      | Deep_object_custom_property _
      | Deep_object_array _
      | Deep_object_object _) :: items -> find_value items
    | Deep_object_binding binding :: items ->
      if binding.Binding.bound_ident = ident
      then Some (binding.Binding.bound_value) else
      find_value items in

    find_value deep_object_items in

  let find_string_value_in_object_items ident =
    match find_value_in_deep_object_items ident with
    | None -> None
    | Some v ->
      let id = string_of_ident ident in
      Some (
        mdl_text_of_string (
          Mdl_deep_basics.string_of_deep_value
            (Printf.sprintf "a ``%s'' of a mask was not a string" id)
            v
          )
      ) in

  let find_shallow_code_value_in_object_items ident =
    match find_value_in_deep_object_items ident with
    | None -> None
    | Some v ->
      let id = string_of_ident ident in
      let shallow_source_code =
        mdl_shallow_code_of_string (
          Mdl_deep_basics.string_of_deep_value
            (Printf.sprintf "a ``%s'' of a mask was not a string" id)
            v) in
      Some shallow_source_code in

  let find_code_value_in_object_items ident =
    match find_value_in_deep_object_items ident with
    | None -> None
    | Some v ->
      let id = string_of_ident ident in
      let source_code =
        Deep_source_code (
          mdl_text_of_string (
            Mdl_deep_basics.string_of_deep_value
              (Printf.sprintf "a ``%s'' of a mask was not a string" id)
              v
          )
        ) in
      Some source_code in

  let shallow_mask_initialisation =
    find_shallow_code_value_in_object_items (Ident.Ident "Initialization") in

  let shallow_mask_description =
    find_string_value_in_object_items (Ident.Ident "Description") in

  let shallow_mask_display =
    find_code_value_in_object_items (Ident.Ident "Display") in

  let shallow_mask_type = find_string_value_in_object_items (Ident.Ident "Type") in
  fdebug (fun epf ->
    Format.fprintf epf "deep_mask_of_object: deep_mask_type %a"
      (Og_printer.printer_Option
         Ast_types_print.printer_mdl_text Og_precedence.Ck_argument)
      shallow_mask_type);

  let find_loop_mask_parameter_in_deep_object_items idx deep_object_items =

    let set_opt e x = function
      | None -> Some x
      | _ -> fatal_error e in

    let set_ident_opt =
      set_opt (Printf.sprintf
        "in mask parameter: ``%s'' set twice" "Name") in

    let set_style_opt (Mdl_name style) =
      set_opt (Printf.sprintf
        "in mask parameter: ``%s'' set twice" "Style")
        (deep_mask_parameter_style_of_string style) in

    let mask_parameter_style_item_of_deep_array_item = function
      | Deep_array_binding binding ->
        begin
          match Ast_types.string_of_ident binding.Binding.bound_ident with
          | "Cell" ->
            begin
              match binding.Binding.bound_value with
              | Deep_string string ->
                Item_text (Ast_types.mdl_text_of_string string)
              | Deep_bool _
              | Deep_float _
              | Deep_int _
              | Deep_wstring _
              | Deep_ident _
              | Deep_code _
              | Deep_expression _
              | Deep_vector _
              | Deep_matrix _
              | Deep_bindings _ ->
                fatal_error
                  "invalid deep value when reading popup style option"
            end
          | ident ->
            fatal_error (Printf.sprintf
              "invalid binding ident when reading popup style option %S" ident)
        end
      | Deep_array_name _
      | Deep_array_sub_array _
      | Deep_array_custom_property _
      | Deep_array_matstruct _
      | Deep_array_object _
      | Deep_array_vendor_binding _
      | Deep_array_vendor_configset _ ->
        fatal_error
          "invalid deep_array_item when reading popup style option" in
    let mask_parameter_style_items_of_deep_array_items deep_array_items =
      let rec loop accu = function
        | [] -> List.rev accu
        | array_item :: array_items ->
          let mask_parameter_style_item =
            mask_parameter_style_item_of_deep_array_item array_item in
          let accu = mask_parameter_style_item :: accu in
          loop accu array_items in
      loop [] (Array.to_list deep_array_items) in
    let set_style_options_opt deep_array_items =
      set_opt (Printf.sprintf
        "in mask parameter: ``%s'' set twice" "StyleOptions")
        (mask_parameter_style_items_of_deep_array_items deep_array_items) in

    let set_range_opt range =
      let range =
        fdebug
          (fun epf ->
           Format.fprintf
             epf "set_range_opt %a"
             Mdl_basics_deep_ast_print.print_deep_value range);
        let mat = vector_of_deep_value "Range" range in
        let rec loop accu = function
          | [] -> List.rev accu
          | deep_value :: deep_values ->
            let value =
              Mdl_deep_basics.string_of_deep_value
                "a value of Range was not a string" deep_value in
            loop (value :: accu) deep_values in
        Mdl_basics_deep_ast.Mask_parameter_style_range (loop [] mat) in
      set_opt
        (Printf.sprintf "in mask parameter: ``%s'' set twice" "Range")
        range in

    let set_prompt_opt v =
      set_opt (Printf.sprintf
        "in mask parameter: ``%s'' set twice" "Prompt")
        (mdl_text_of_string (Mdl_deep_basics.string_of_deep_value
           (Printf.sprintf "a ``%s'' of a parameter was not a string" "Prompt")
           v)) in

    let set_value_opt v =
      set_opt (Printf.sprintf
        "in mask parameter: ``%s'' set twice" "Value")
        (Deep_mask_parameter_value
          (mdl_text_of_string
            (Mdl_deep_basics.string_of_deep_value
              (Printf.sprintf "a ``%s'' of a parameter was not a string" "Value")
              v))) in

    let set_evaluation_opt v =
      set_opt (Printf.sprintf
        "in mask parameter: ``%s'' set twice" "Evaluate")
        (deep_mask_parameter_evaluation_of_deep_value v) in

    let set_enable_opt v =
      set_opt (Printf.sprintf
        "in mask parameter: ``%s'' set twice" "Enable")
        (deep_mask_parameter_enable_of_string
          (Mdl_deep_basics.string_of_deep_value
             (Printf.sprintf "a ``%s'' of a parameter was not a string" "Enable")
             v)) in

    let set_visible_opt v =
      set_opt (Printf.sprintf
        "in mask parameter: ``%s'' set twice" "Visibility")
        (deep_mask_parameter_visible_of_string
          (Mdl_deep_basics.string_of_deep_value
             (Printf.sprintf "a ``%s'' of a parameter was not a string" "Visible")
             v)) in

    let rec loop classifier accu = function
      | [] -> classifier, List.rev accu
      | Deep_object_type object_style :: items ->
        fdebug (fun epf ->
          Format.fprintf epf "deep_mask_of_object: object_style: %a"
            Ast_types_print.print_mdl_name object_style);
        let classifier =
          { classifier with
            loop_mask_parameter_style_opt =
              set_style_opt
                object_style classifier.loop_mask_parameter_style_opt;
          } in
        loop classifier accu items
      | Deep_object_name object_name :: items ->
        let classifier =
          { classifier with
            loop_mask_parameter_ident_opt =
              set_ident_opt
                (ident_from_mdl_name object_name)
                classifier.loop_mask_parameter_ident_opt;
          } in
        loop classifier accu items
      | Deep_object_binding {
          Binding.bound_ident = Ident.Ident "Prompt";
          bound_value = prompt_value;
        } :: items ->
        let classifier =
          { classifier with
            loop_mask_parameter_prompt_opt =
              set_prompt_opt
                prompt_value classifier.loop_mask_parameter_prompt_opt;
          } in
        loop classifier accu items
      | Deep_object_binding {
          Binding.bound_ident = Ident.Ident "Value";
          bound_value = value_value;
        } :: items ->
        let classifier =
          { classifier with
            loop_mask_parameter_value_opt =
              set_value_opt
                value_value classifier.loop_mask_parameter_value_opt;
          } in
        loop classifier accu items
      | Deep_object_binding {
          Binding.bound_ident = Ident.Ident "Evaluate";
          bound_value = evaluation_value;
        } :: items ->
        let classifier =
          { classifier with
            loop_mask_parameter_eval_opt =
              set_evaluation_opt
                evaluation_value
                classifier.loop_mask_parameter_eval_opt;
          } in
        loop classifier accu items
      | Deep_object_binding {
          Binding.bound_ident = Ident.Ident "Visible";
          bound_value = visible_value;
        } :: items ->
        let classifier =
          { classifier with
            loop_mask_parameter_visible_opt =
              set_visible_opt
                visible_value
                classifier.loop_mask_parameter_visible_opt;
          } in
        loop classifier accu items
      | Deep_object_binding {
          Binding.bound_ident = Ident.Ident "Enabled";
          bound_value = enable_value;
        } :: items ->
        let classifier =
          { classifier with
            loop_mask_parameter_enable_opt =
              set_enable_opt
                enable_value classifier.loop_mask_parameter_enable_opt;
          } in
        loop classifier accu items
      | Deep_object_binding {
          Binding.bound_ident = Ident.Ident "Range";
          bound_value = range_value;
        } :: items ->
        let classifier =
          { classifier with
            loop_mask_parameter_range_opt =
              set_range_opt
                range_value classifier.loop_mask_parameter_range_opt;
          } in
        loop classifier accu items
      | Deep_object_array {
            deep_array_type = Cell;
            deep_array_items = deep_array_items;
            deep_array_prop_name = Ident.Ident "TypeOptions";
          } :: items ->
        let classifier =
          { classifier with
            loop_mask_parameter_style_options_opt =
              set_style_options_opt
                deep_array_items
                classifier.loop_mask_parameter_style_options_opt;
          } in
        loop classifier accu items
      | ( Deep_object_binding {
            Binding.bound_ident = Ident.Ident _;
            bound_value = _;
          }
        | Deep_object_array {
            deep_array_type = (Struct | Handle | Array_type _);
            deep_array_items = _;
            deep_array_prop_name = _;
          }
        | Deep_object_array {
            deep_array_type = Cell;
            deep_array_items = _;
            deep_array_prop_name = _
          }
        | Deep_object_dollar_binding _
        | Deep_object_custom_property _
        | Deep_object_object _
        ) as item :: items ->
        loop classifier (item :: accu) items in

    let ident_opt, style_opt, style_options_opt, prompt_opt, value_opt,
        eval_opt, visible_opt, enable_opt, range_opt =
      (* FIXME: are residual deep mask bindings useless
         or should them be transmitted along ? *)
      let classifier, _deep_mask_bindings =
        loop
          empty_loop_mask_parameter_classifier [] deep_object_items in
      match classifier with
      | {
        loop_mask_parameter_ident_opt = ident_opt;
        loop_mask_parameter_style_opt = style_opt;
        loop_mask_parameter_style_options_opt = style_options_opt;
        loop_mask_parameter_prompt_opt = prompt_opt;
        loop_mask_parameter_value_opt = value_opt;
        loop_mask_parameter_eval_opt = eval_opt;
        loop_mask_parameter_visible_opt = visible_opt;
        loop_mask_parameter_enable_opt = enable_opt;
        loop_mask_parameter_range_opt = range_opt;
      } ->
        ident_opt, style_opt, style_options_opt, prompt_opt, value_opt,
        eval_opt, visible_opt, enable_opt, range_opt in

    let style_opt =
      (* todo: why are we matching empty list for popup and
         radiobutton? *)
      match style_opt with
      | None -> None
      | Some (Deep_mask_parameter_popup []) ->
        let mask_parameter_popup_items =
          match style_options_opt with
          | None -> []
          | Some mask_parameter_popup_items -> mask_parameter_popup_items in
        Some (Deep_mask_parameter_popup mask_parameter_popup_items)
      | Some (Deep_mask_parameter_radiobutton []) ->
        let mask_parameter_radiobutton_items =
          match style_options_opt with
          | None -> []
          | Some mask_parameter_radiobutton_items ->
            mask_parameter_radiobutton_items in
        Some
          (Deep_mask_parameter_radiobutton
             mask_parameter_radiobutton_items)
      | Some (Deep_mask_parameter_slider None) ->
        Some (Deep_mask_parameter_slider range_opt)
      | Some (Deep_mask_parameter_spinbox None) ->
        Some (Deep_mask_parameter_spinbox range_opt)
      | Some (
            Deep_mask_parameter_edit
          | Deep_mask_parameter_checkbox
          | Deep_mask_parameter_popup _
          | Deep_mask_parameter_unidt _
          | Deep_mask_parameter_min
          | Deep_mask_parameter_max
          | Deep_mask_parameter_promote
          | Deep_mask_parameter_radiobutton _
          | Deep_mask_parameter_slider _
          | Deep_mask_parameter_spinbox _) as style_opt -> style_opt in

    match ident_opt, style_opt, prompt_opt, value_opt with
    | Some ident, Some _style, Some prompt, Some _value ->

      fdebug (fun ppf ->
        Format.fprintf ppf "Loop parameter number %d is |%s|"
          (Unatural.to_int idx)
          (Ast_types.string_of_ident ident)
      );

      let ident =
        (* Some mask parameters have an empty name, so we create a fake one. *)
        match ident with
        | Ident.Ident "" -> make_mask_ident idx
        | Ident.Ident _ -> ident in

      {
        loop_mask_parameter_var_num = idx;
        loop_mask_parameter_ident = ident;
        loop_mask_parameter_style = style_opt;
        loop_mask_parameter_prompt = prompt;
        loop_mask_parameter_value = value_opt;
        loop_mask_parameter_evaluation = eval_opt;
        loop_mask_parameter_enable = enable_opt;
        loop_mask_parameter_visible = visible_opt;
        (* TODO URGENT: fill those 3 fields. *)
        loop_mask_parameter_internal = None;
        loop_mask_parameter_never_save = None;
        loop_mask_parameter_read_only = None;
      }

    | _, _, _, _ -> fatal_error "an incomplete mask parameter was found" in

  let find_mask_in_deep_array_items deep_items =

    let rec loop accu i =
      if i < Unatural_types.to_natural (Array.length deep_items) then
        let accu =
          match deep_items.(Unatural.to_int i) with
          | Deep_array_name _
          | Deep_array_sub_array _
          | Deep_array_custom_property _
          | Deep_array_matstruct _
          | Deep_array_vendor_binding _
          | Deep_array_vendor_configset _
          | Deep_array_binding _ -> accu
          | Deep_array_object { deep_object_items; _ } ->
            let idx = Unatural.succ i in
            find_loop_mask_parameter_in_deep_object_items idx deep_object_items :: accu in
        loop accu (Unatural.succ i) else
      match accu with
      | [] -> []
      | _ -> List.rev accu in
    loop [] Unatural._0_ in

  let find_loop_mask_parameters_in_deep_object_items prop_name_ident =

    let rec find_loop_mask_parameters = function
      | [] -> []
      | (Deep_object_type _
        | Deep_object_name _
        | Deep_object_dollar_binding _
        | Deep_object_binding _) :: items -> find_loop_mask_parameters items
        (* There is only one parameter that is not in an array *)
      | Deep_object_object {
          deep_object_id = _deep_object_id;
          deep_object_class_name = _class_name_option;
          deep_object_prop_name = Some prop_name;
          deep_object_items = deep_object_items;
        } :: _items when prop_name = prop_name_ident ->
        [ find_loop_mask_parameter_in_deep_object_items Unatural._1_ deep_object_items ]
        (* There many parameter(s) that are inside an array *)
      | Deep_object_array {
          deep_array_type = Array_type "Simulink.MaskParameter";
          deep_array_prop_name = prop_name;
          deep_array_items = deep_array_items;
        } :: _items when prop_name = prop_name_ident ->
        find_mask_in_deep_array_items deep_array_items
      | (Deep_object_array {
           deep_array_type =
             (Struct | Cell | Handle | Array_type _);
           deep_array_prop_name = _;
           deep_array_items = _;
         }
        | Deep_object_custom_property _
        | Deep_object_object _
        ) :: items ->
        find_loop_mask_parameters items in

    find_loop_mask_parameters in

  let loop_mask_parameters =
    find_loop_mask_parameters_in_deep_object_items parameters_ident
      deep_object_items in

  (* FIXME:
     if we keep residual deep mask bindings when finding mask parameters, we
     can pass along some more meaningful list of mask bindings.
  *)
  let shallow_mask_bindings = [] in

  let shallow_mask =
    {
      shallow_mask_type = shallow_mask_type;
      shallow_mask_description = shallow_mask_description;
      shallow_mask_initialisation = shallow_mask_initialisation;
      shallow_mask_display = shallow_mask_display;
      shallow_mask_parameters = loop_mask_parameters;
      shallow_mask_bindings = shallow_mask_bindings;
    } in
  deep_mask_of_shallow_mask shallow_mask
;;

let deep_mask_of_mask_object deep_object_description =
  match deep_object_description with
  | { deep_object_class_name = Some (Mdl_name "Simulink.Mask");
      deep_object_items = deep_object_items; _ } ->
    deep_mask_of_object deep_object_items
  | {
      deep_object_id = _object_id;
      deep_object_class_name = (None | Some (Mdl_name _));
      deep_object_prop_name = _prop_name_opt;
      deep_object_items = _deep_object_items;
    } ->
    fatal_error "unknown mask object class name"
;;

let deep_mask_parameter_of_mask_parameter = function
  | {
      mask_parameter_ident = parameter_ident;
      mask_parameter_style = parameter_style;
      mask_parameter_prompt = parameter_prompt;
      mask_parameter_callback = parameter_callback;
      mask_parameter_tunable = parameter_tunable;
      mask_parameter_value = parameter_value;
      mask_parameter_evaluation = parameter_evaluation;
      mask_parameter_enable = parameter_enable;
      mask_parameter_visible = parameter_visible;
      mask_parameter_internal = parameter_internal;
      mask_parameter_never_save = parameter_never_save;
      mask_parameter_read_only = parameter_read_only;
    } ->
    let deep_mask_parameter_style_of_mdl_text style =
      deep_mask_parameter_style_of_string
        (Ast_types.string_of_mdl_text style) in
    let deep_mask_parameter_ident = parameter_ident in
    let deep_mask_parameter_style =
      deep_mask_parameter_style_of_mdl_text parameter_style in
    let deep_mask_parameter_prompt = parameter_prompt in
    let deep_mask_parameter_callback =
      Uoption.map
        (fun parameter_callback ->
         Deep_source_code
           (Ast_types.mdl_text_of_mdl_code parameter_callback))
        parameter_callback in
    let deep_mask_parameter_tunable_opt =
      Uoption.map
        (fun mdl_bool -> Ast_types.bool_of_mdl_bool mdl_bool)
        parameter_tunable in
    let deep_mask_parameter_value =
      Deep_mask_parameter_value parameter_value in
    let deep_mask_parameter_evaluation_opt =
      Uoption.map
        (fun mdl_bool ->
         deep_mask_parameter_evaluation_of_bool
           (Ast_types.bool_of_mdl_bool mdl_bool))
        parameter_evaluation in
    let deep_mask_parameter_enable_opt =
      Uoption.map
        (fun enable ->
         if Ast_types.bool_of_mdl_bool enable then
           Enable
         else
           Disable)
        parameter_enable in
    let deep_mask_parameter_visible_opt =
      Uoption.map
        (fun visible ->
         if Ast_types.bool_of_mdl_bool visible then
           Visible
         else
           Hidden)
        parameter_visible in
    let deep_mask_parameter_internal_opt =
      Uoption.map
        (fun internal ->
         if Ast_types.bool_of_mdl_bool internal then
           Internal
         else
           External)
        parameter_internal in
    let deep_mask_parameter_never_save_opt =
      Uoption.map
        (fun never_save ->
         if Ast_types.bool_of_mdl_bool never_save then
           Never_save
         else
           Save)
        parameter_never_save in
    let deep_mask_parameter_read_only_opt =
      Uoption.map
        (fun read_only ->
         if Ast_types.bool_of_mdl_bool read_only then
           Read_only
         else
           Read_write)
        parameter_read_only in
    {
      deep_mask_parameter_ident = deep_mask_parameter_ident;
      deep_mask_parameter_style = deep_mask_parameter_style;
      deep_mask_parameter_prompt = deep_mask_parameter_prompt;
      deep_mask_parameter_callback = deep_mask_parameter_callback;
      deep_mask_parameter_tunable = deep_mask_parameter_tunable_opt;
      deep_mask_parameter_value = deep_mask_parameter_value;
      deep_mask_parameter_evaluation = deep_mask_parameter_evaluation_opt;
      deep_mask_parameter_enable = deep_mask_parameter_enable_opt;
      deep_mask_parameter_visible = deep_mask_parameter_visible_opt;
      deep_mask_parameter_internal = deep_mask_parameter_internal_opt;
      deep_mask_parameter_never_save = deep_mask_parameter_never_save_opt;
      deep_mask_parameter_read_only = deep_mask_parameter_read_only_opt;
    }
;;

let deep_mask_of_mask_decription = function
  | {
      mask_type = mtype;
      mask_display = display;
      mask_description = description;
      mask_initialisation = initialisation;
      mask_parameters = parameters;
      mask_bindings = bindings;
    } ->
    let deep_mask_type = mtype in
    let deep_mask_display, deep_mask_display_bindings =
      match display with
      | {
          mask_display_source = display_source_opt;
          mask_display_bindings = bindings;
        } ->
        Uoption.map deep_source_code_of_mdl_code display_source_opt,
        Mdl_deep_basics.deep_bindings bindings in
    let deep_mask_description = description in
    let deep_mask_initialisation =
      Uoption.map deep_source_code_of_mdl_code initialisation in
    let deep_mask_parameters =
      List.map deep_mask_parameter_of_mask_parameter parameters in
    let deep_mask_bindings =
      List.rev_append
        deep_mask_display_bindings
        (Mdl_deep_basics.deep_bindings bindings) in
    {
      deep_mask_type = deep_mask_type;
      deep_mask_description = deep_mask_description;
      deep_mask_initialisation = deep_mask_initialisation;
      deep_mask_display = deep_mask_display;
      deep_mask_parameters = deep_mask_parameters;
      deep_mask_bindings = deep_mask_bindings;
    }
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
