(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_stateflow_deep_ast.mli,v 1.1 2019-05-10 14:07:14 jpc Exp $ *)

(**
  Similar to [Mdl_deep_ast] syntax definition but specialized to stateflow.

  From shadow parse tree to deep syntax:
  - we perform various validity checks on the tree obtained from the parser,
  - we build explicit record representations of data instead of simple
    values with associated so-called property lists.
*)

type ident = Ident.ident
and natural = Mdl_basics_deep_ast.natural
;;

(* From module Mdl_basics_deep_ast *)

type deep_text = Mdl_basics_deep_ast.deep_text
and deep_name = Mdl_basics_deep_ast.deep_name
and deep_code = Mdl_basics_deep_ast.deep_code
and deep_binding_table = Mdl_basics_deep_ast.deep_binding_table
;;

(** Stateflow basic field and argument type definitions *)

type matlab_code = deep_code;;

type deep_stateflow_id = natural
and deep_stateflow_name = deep_name
and deep_stateflow_reference = natural
and deep_stateflow_number = Mdl_ast.stateflow_number
and deep_chart_file_number =
  Mdl_basics_deep_ast.deep_chart_file_number
and deep_stateflow_chart_file_number = deep_chart_file_number
;;

(** May be optional field and argument type definitions *)
type deep_stateflow_name_option = deep_stateflow_name option;;
type deep_stateflow_reference_option = deep_stateflow_reference option;;

(** Always optional field and argument type definitions *)
type deep_stateflow_execution_order_option = natural option
and deep_stateflow_label_string_option = deep_text option
and deep_stateflow_id_option = deep_stateflow_id option
and deep_stateflow_ssid_number_option = natural option
;;

(** Stateflow deep section items *)

type deep_stateflow_state_type =
  | Deep_state_type_and (* "AND_STATE" *)
  | Deep_state_type_or (* "OR_STATE" *)
  | Deep_state_type_func (* "FUNC_STATE" *)
  | Deep_state_type_group (* "GROUP_STATE" *)
  | Deep_state_type_sub (* "SUB" *)
  | Deep_state_type_super (* "SUPER" *)
  | Deep_state_type_eml_chart (* "EML_CHART" *)
  | Deep_state_type_string of deep_text
;;

type deep_stateflow_chart_type = deep_stateflow_state_type;;

type deep_stateflow_junction_type =
  | Deep_junction_type_connective_junction (* "CONNECTIVE_JUNCTION" *)
  | Deep_junction_type_history_junction (* "HISTORY_JUNCTION" *)
  | Deep_junction_type_string of deep_text
;;

type deep_stateflow_transition_type =
  | Deep_transition_type_sub (* "SUB" *)
  | Deep_transition_type_string of deep_text
;;

type deep_stateflow_transition_type_option =
  deep_stateflow_transition_type option
;;

type deep_stateflow_data_scope =
  | Deep_scope_constant_data (* "CONSTANT_DATA" *)
  | Deep_scope_data_store_memory_data (* "DATA_STORE_MEMORY_DATA" *)
  | Deep_scope_exported_data (* "EXPORTED_DATA" *)
  | Deep_scope_function_input_data (* "FUNCTION_INPUT_DATA" *)
  | Deep_scope_function_output_data (* "FUNCTION_OUTPUT_DATA" *)
  | Deep_scope_input_data (* "INPUT_DATA" *)
  | Deep_scope_input_event (* "INPUT_EVENT" *)
  | Deep_scope_local_data (* "LOCAL_DATA" *)
  | Deep_scope_local_event (* "LOCAL_EVENT" *)
  | Deep_scope_output_data (* "OUTPUT_DATA" *)
  | Deep_scope_output_event (* "OUTPUT_EVENT" *)
  | Deep_scope_parameter_data (* "PARAMETER_DATA" *)
  | Deep_scope_temporary_data (* "TEMPORARY_DATA" *)
  | Deep_scope_data_string of string
;;

type deep_stateflow_props_array =
  {
    deep_props_array_item_size : deep_text;
    deep_props_array_item_first_index : deep_text option;
    deep_props_array_item_binding_table : deep_binding_table;
  }
;;

type deep_stateflow_data_props =
  {
    deep_props_array : deep_stateflow_props_array option;
    deep_props_type : deep_stateflow_props_type option;
    deep_props_binding_table : deep_binding_table;
  }

and deep_stateflow_props_type = deep_binding_table
;;

type deep_data_reference =
  | Deep_data_reference_chart of deep_stateflow_reference
  | Deep_data_reference_machine of deep_stateflow_reference
;;

type deep_stateflow_event_trigger =
  | Deep_event_trigger_either_edge_event
  | Deep_event_trigger_falling_edge_event
  | Deep_event_trigger_function_call_event
  | Deep_event_trigger_rising_edge_event
  | Deep_event_trigger_ident of ident
;;

type deep_intersection =
  {
    i0 : deep_stateflow_number;
    i1 : deep_stateflow_number;
    i2 : deep_stateflow_number;
    i3 : deep_stateflow_number;
    i4 : deep_stateflow_number;
    i5 : deep_stateflow_number;
    i6 : deep_stateflow_number;
    i7 : deep_stateflow_number;
  }
;;

type deep_transition_src =
  {
    src_id : deep_stateflow_id_option;
    src_intersection : deep_intersection;
  }
;;

type deep_transition_dst =
  {
    dst_id : deep_stateflow_id_option;
    dst_intersection : deep_intersection;
  }
;;

type deep_stateflow_state_eml =
  {
    state_eml_script : deep_code option;
    state_eml_binding_table : deep_binding_table;
  }
;;

type deep_stateflow_chart_eml =
  {
    chart_eml_name : deep_name;
    chart_eml_binding_table : deep_binding_table;
  }
;;

type deep_stateflow_junctions = deep_stateflow_junction list
and deep_stateflow_junction =
  {
    deep_junction_id : deep_stateflow_id;
    deep_junction_ssid_number : deep_stateflow_ssid_number_option;
    deep_junction_type : deep_stateflow_junction_type;
    deep_junction_label_string : deep_stateflow_label_string_option;
    deep_junction_binding_table : deep_binding_table;
  }
;;

type deep_stateflow_transitions = deep_stateflow_transition list
and deep_stateflow_transition =
  {
    deep_transition_id : deep_stateflow_id;
    deep_transition_ssid_number : deep_stateflow_ssid_number_option;
    deep_transition_type : deep_stateflow_transition_type_option;
    deep_transition_label_string : deep_stateflow_label_string_option;
    deep_transition_src : deep_transition_src; (* src *)
    deep_transition_dst : deep_transition_dst; (* dst *)
    deep_transition_execution_order :
    deep_stateflow_execution_order_option; (* executionOrder *)
    deep_transition_binding_table : deep_binding_table;
  }
;;

type deep_stateflow_events = deep_stateflow_event list
and deep_stateflow_event =
  {
    deep_event_id : deep_stateflow_id;
    deep_event_name : deep_stateflow_name;
    deep_event_scope : deep_stateflow_data_scope;
    deep_event_trigger : deep_stateflow_event_trigger; (* trigger *)
    deep_event_binding_table : deep_binding_table;
  }
;;

type deep_stateflow_datas = deep_stateflow_data list
and deep_stateflow_data =
  {
    deep_data_id : deep_stateflow_id;
    deep_data_name : deep_stateflow_name;
    deep_data_ssid_number : deep_stateflow_ssid_number_option;
    deep_data_scope : deep_stateflow_data_scope; (* scope *)
    deep_data_props : deep_stateflow_data_props; (* props *)
    deep_data_binding_table : deep_binding_table;
  }
;;

type deep_stateflow_states = deep_stateflow_state list
and deep_stateflow_state =
  {
    deep_state_id : deep_stateflow_id;
    deep_state_ssid_number :
    deep_stateflow_ssid_number_option; (* "ssIdNumber" *)
    deep_state_type : deep_stateflow_state_type; (* type *)
    deep_state_label_string : deep_text; (* labelString *)
    deep_state_execution_order : deep_stateflow_execution_order_option;
    deep_stateflow_state_eml : deep_stateflow_state_eml option;
    deep_stateflow_state_datas : deep_stateflow_datas;
    deep_stateflow_state_events : deep_stateflow_events;
    deep_stateflow_state_junctions : deep_stateflow_junctions;
    deep_stateflow_state_states : deep_stateflow_states;
    deep_stateflow_state_transitions : deep_stateflow_transitions;
    deep_state_binding_table : deep_binding_table;
  }
;;

type deep_stateflow_targets = deep_stateflow_target list
and deep_stateflow_target =
  {
    deep_target_id : deep_stateflow_id;
    deep_target_name : deep_stateflow_name_option;
    deep_target_binding_table : deep_binding_table;
  }
;;

type deep_stateflow_charts = deep_stateflow_chart list
and deep_stateflow_chart =
  {
    deep_chart_id : deep_stateflow_id;
    deep_chart_name : deep_stateflow_name;
    deep_chart_file_number : deep_stateflow_chart_file_number;
    deep_chart_type : deep_stateflow_chart_type option;
    deep_stateflow_chart_eml : deep_stateflow_chart_eml option;
    deep_stateflow_chart_datas : deep_stateflow_datas;
    deep_stateflow_chart_events : deep_stateflow_events;
    deep_stateflow_chart_junctions : deep_stateflow_junctions;
    deep_stateflow_chart_states : deep_stateflow_states;
    deep_stateflow_chart_transitions : deep_stateflow_transitions;
    deep_chart_binding_table : deep_binding_table;
  }
;;

(** Machine section item *)
type deep_stateflow_machine =
  {
    deep_machine_id : deep_stateflow_id; (* "id" *)
    deep_machine_name : deep_stateflow_name_option; (* "name" *)
    deep_stateflow_machine_charts : deep_stateflow_charts;
    deep_stateflow_machine_datas : deep_stateflow_datas;
    deep_stateflow_machine_events : deep_stateflow_events;
    deep_stateflow_machine_targets : deep_stateflow_targets;
    deep_machine_binding_table : deep_binding_table;
  }
;;

(** Instance section item *)
type deep_stateflow_instance =
  {
    deep_instance_id : deep_stateflow_id;
    deep_instance_name : deep_stateflow_name;
    deep_instance_machine_ref : deep_stateflow_reference;
    deep_instance_chart_ref : deep_stateflow_reference;
  }
;;

type deep_stateflow_section_items = deep_stateflow_section_item list
and deep_stateflow_section_item =
  | Deep_stateflow_section_machine of deep_stateflow_machine
  | Deep_stateflow_section_instance of deep_stateflow_instance
;;

type deep_stateflow_section =
  | Deep_stateflow_section of deep_stateflow_section_items
;;

(*
 Local Variables:
  compile-command: "cd ../../.. && make"
  End:
*)
