(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_stateflow_to_deep_ast.ml,v 1.1 2019-05-10 14:07:14 jpc Exp $ *)

(* From parse tree to deep syntax:
   - we perform various validity checks on the tree obtained from the parser,
   - we build explicit record representations of data instead of simple
     values with associated so-called property lists.
*)

open Mdl_ast;;
(*open Mdl_basics_deep_ast;;*)
open Mdl_stateflow_deep_ast;;

let module_name = "Mdl_stateflow_to_deep_ast";;

let warning msg =
  Say.warning (
    Printf.sprintf "%s: in Simulink file %s, %s"
      module_name
      (Configuration.get_source_file ())
      msg
  )
;;
let fatal_error s =
  Say.fatal_error (Printf.sprintf
    "%s: %s"
    module_name s)
;;

(* Convert binding [mdl_binding] to a [deep-binding] and add it to the
   [deep_binding] hash table tbl. *)
let add_deep_binding tbl mdl_binding =
  let ident = mdl_binding.Binding.bound_ident in
  let deep_binding =
    Mdl_deep_basics.deep_binding mdl_binding in
    Hashtbl.add tbl ident deep_binding;
;;

let string_of_stateflow_ref stateflow_ref =
  Unatural.to_string stateflow_ref
;;

type classified_stateflow_section_items =
  {
    stateflow_name : Ast_types.mdl_name option;
    stateflow_trigger : Ident.ident option;
    stateflow_chart_file_number : Ast_types.natural option;
    stateflow_type : Ident.ident option;
    stateflow_eml : stateflow_eml_items option;
    stateflow_execution_order : Ast_types.natural option;
    stateflow_label_string : Ast_types.mdl_text option;
    stateflow_scope : Ident.ident option;
    stateflow_props : Mdl_ast.stateflow_props_items option;
    stateflow_ssid_number : Ast_types.natural option;
    stateflow_src : Mdl_ast.stateflow_src_dst_items option;
    stateflow_dst : stateflow_src_dst_items option;
    stateflow_plant_modeling_info : unit option;
    stateflow_binding_table : deep_binding_table;
  }
;;

let mk_empty_classified_stateflow_section_items () =
  {
    stateflow_name = None;
    stateflow_trigger = None;
    stateflow_chart_file_number = None;
    stateflow_type = None;
    stateflow_eml = None;
    stateflow_execution_order = None;
    stateflow_label_string = None;
    stateflow_scope = None;
    stateflow_props = None;
    stateflow_ssid_number = None;
    stateflow_src = None;
    stateflow_dst = None;
    stateflow_plant_modeling_info = None;
    stateflow_binding_table = Hash_table.of_bindings [];
  }
;;

let check_unset field_name = function
  | None -> ()
  | Some _ ->
    fatal_error (Printf.sprintf
      "multiply defined item \'%s\' in stateflow" field_name)
;;

let classify_stateflow_items items =
  let rec loop classification = function
    | [] -> classification
    | stateflow_section_item :: items ->
      let classification =
        match stateflow_section_item with
        | Stateflow_trigger ident ->
          check_unset "trigger" classification.stateflow_trigger;
          {
            classification with
            stateflow_trigger = Some ident;
          }
        | Stateflow_chart_file_number natural ->
          check_unset "chart file number"
            classification.stateflow_chart_file_number;
          {
            classification with
            stateflow_chart_file_number = Some natural;
          }
        | Stateflow_type ident ->
          check_unset "type" classification.stateflow_type;
          {
            classification with
            stateflow_type = Some ident;
          }
        | Stateflow_name mdl_name ->
          check_unset "name" classification.stateflow_name;
          {
            classification with
            stateflow_name = Some mdl_name;
          }
        | Stateflow_eml stateflow_eml_items ->
          check_unset "eml" classification.stateflow_eml;
          {
            classification with
            stateflow_eml = Some stateflow_eml_items;
          }
        | Stateflow_execution_order natural ->
          check_unset "executionOrder"
            classification.stateflow_execution_order;
          {
            classification with
            stateflow_execution_order = Some natural;
          }
        | Stateflow_label_string mdl_text ->
          check_unset "labelString" classification.stateflow_label_string;
          {
            classification with
            stateflow_label_string = Some mdl_text;
          }
        | Stateflow_scope ident ->
          check_unset "scope" classification.stateflow_scope;
          {
            classification with
            stateflow_scope = Some ident;
          }
        | Stateflow_props stateflow_props_items ->
          check_unset "props" classification.stateflow_props;
          {
            classification with
            stateflow_props = Some stateflow_props_items;
          }


        | Stateflow_ssid_number natural ->
          check_unset "ssIdNumber" classification.stateflow_ssid_number;
          {
            classification with
            stateflow_ssid_number = Some natural;
          }

        | Stateflow_src stateflow_items ->
          check_unset "src" classification.stateflow_src;
          {
            classification with
            stateflow_src = Some stateflow_items;
          }
        | Stateflow_dst stateflow_items ->
          check_unset "dst" classification.stateflow_dst;
          {
            classification with
            stateflow_dst = Some stateflow_items;
          }
        | Stateflow_plant_modeling_info ->
          check_unset
            "plant_modeling_info"
            classification.stateflow_plant_modeling_info;
          {
            classification with
            stateflow_plant_modeling_info = Some ();
          }
        | Stateflow_binding binding ->
          add_deep_binding classification.stateflow_binding_table binding;
          classification in
      loop classification items in

  loop (mk_empty_classified_stateflow_section_items ()) items
;;

let name_ident = Mdl_ast_functions.lname_ident
and type_ident = Mdl_ast_functions.ltype_ident
and label_string_ident = Mdl_ast_functions.llabel_string_ident
and trigger_ident = Mdl_ast_functions.ltrigger_ident
and src_ident = Mdl_ast_functions.lsrc_ident
and dst_ident = Mdl_ast_functions.ldst_ident
and scope_ident = Mdl_ast_functions.lscope_ident
and props_ident = Mdl_ast_functions.lprops_ident
and execution_order_ident = Mdl_ast_functions.lexecution_order_ident
and ssid_number_ident = Mdl_ast_functions.lss_id_number_ident
and eml_ident = Mdl_ast_functions.leml_ident
and chart_file_number_ident = Mdl_ast_functions.lchart_file_number_ident
and script_ident = Mdl_ast_functions.lscript_ident
;;

(* Find a field which is not option:
  verify that the field has been found in the classification and return its
  proper value. *)
let find_field field_ident item_name = function
  | None ->
    fatal_error (Printf.sprintf
      "cannot find field \'%s\' for stateflow item \'%s\'"
      (Ast_types.string_of_ident field_ident)
      item_name)
  | Some f -> f
;;

(* Find a field which is an option: no verification to do, simply emit a
  debug message if the debug mode is fired. *)
let find_field_option field_ident item_name field_option =
  if field_option = None then
    Say.fdebug
      (fun ppf -> Format.fprintf ppf
         "@[<v 0>stateflow_to_deep_ast:@ \
          cannot find field \'%s\' for stateflow item \'%s\'\
          @]"
         (Ast_types.string_of_ident field_ident)
         item_name);
  field_option
;;

let check_field_empty field_ident item_name = function
  | None -> ()
  | Some _ ->
    fatal_error (Printf.sprintf
      "field \'%s\' should be empty for stateflow item \'%s\'"
      (Ast_types.string_of_ident field_ident)
      item_name)
;;

let find_name_field = find_field name_ident;;
let find_name_field_option = find_field_option name_ident;;
let check_name_field_empty = check_field_empty name_ident;;

(* ssid number is an option anyway: no need to test *)
let find_ssid_number_field = find_field_option ssid_number_ident;;
let check_ssid_number_field_empty = check_field_empty ssid_number_ident;;

let find_type_field = find_field type_ident;;
let find_type_field_option = find_field_option type_ident;;
let check_type_field_empty = check_field_empty type_ident;;

let find_label_string_field = find_field label_string_ident;;
let find_label_string_field_option = find_field_option label_string_ident;;
let check_label_string_field_empty = check_field_empty label_string_ident;;

let deep_stateflow_event_trigger trigger_ident =
  match Ast_types.string_of_ident trigger_ident with
  | "EITHER_EDGE_EVENT" -> Deep_event_trigger_either_edge_event
  | "FALLING_EDGE_EVENT" -> Deep_event_trigger_falling_edge_event
  | "FUNCTION_CALL_EVENT" -> Deep_event_trigger_function_call_event
  | "RISING_EDGE_EVENT" -> Deep_event_trigger_rising_edge_event
  | _ ->
    warning
      (Printf.sprintf
         "unknown stateflow event trigger field \'%s\'"
         (Ast_types.string_of_ident trigger_ident));
    Deep_event_trigger_ident trigger_ident
;;

let find_trigger_field item_name field_option =
  let trigger = find_field trigger_ident item_name field_option in
  deep_stateflow_event_trigger trigger
;;
let check_trigger_field_empty = check_field_empty trigger_ident;;

let find_execution_order_field_option =
  find_field_option execution_order_ident
;;
let check_execution_order_field_empty =
  check_field_empty execution_order_ident
;;

let check_eml_field_empty = check_field_empty eml_ident;;

let find_chart_file_number_field = find_field chart_file_number_ident;;
let check_chart_file_number_field_empty =
  check_field_empty chart_file_number_ident
;;

let string_of_stateflow_number = function
  | Stateflow_int s -> s
  | Stateflow_float s -> s
;;

let deep_intersection intersection =
  let intersection_array = Array.of_list intersection in
  if Array.length intersection_array = 8 then
    {
      i0 = intersection_array.(0);
      i1 = intersection_array.(1);
      i2 = intersection_array.(2);
      i3 = intersection_array.(3);
      i4 = intersection_array.(4);
      i5 = intersection_array.(5);
      i6 = intersection_array.(6);
      i7 = intersection_array.(7);
    } else
  let intersection_string =
    String.concat " " (List.map string_of_stateflow_number intersection) in
  fatal_error (Printf.sprintf
    "intersection [ %s ] must have 8 numbers"
    intersection_string)
;;

type classified_stateflow_src_dst_item =
  {
    stateflow_src_dst_ref : stateflow_ref option;
    stateflow_intersection : stateflow_numbers option;
  }
;;

type classified_stateflow_eml_item =
  {
    stateflow_eml_name : Ast_types.mdl_name option;
    stateflow_eml_script : Ast_types.mdl_code option;
    stateflow_eml_binding_table : deep_binding_table;
  }
;;

let empty_classified_stateflow_src_dst_item =
  {
    stateflow_src_dst_ref = None;
    stateflow_intersection = None;
  }
;;

let mk_empty_classified_stateflow_eml_item () =
  {
    stateflow_eml_name = None;
    stateflow_eml_script = None;
    stateflow_eml_binding_table = Hash_table.of_bindings [];
  }
;;

let classify_stateflow_src_dst_items item_name stateflow_src_dst_items =
  let rec loop accu = function
    | [] -> accu
    | stateflow_src_dst_item :: stateflow_src_dst_items ->
      let accu =
        match stateflow_src_dst_item with
        | Stateflow_src_dst_ref stateflow_ref ->
          if accu.stateflow_src_dst_ref = None then
          {
            accu with
            stateflow_src_dst_ref = Some stateflow_ref;
          } else
          fatal_error (Printf.sprintf
            "Duplicated reference %s in src/dst %s"
            (string_of_stateflow_ref stateflow_ref) item_name)
        | Stateflow_intersection stateflow_numbers ->
          if accu.stateflow_intersection = None then
          {
            accu with
            stateflow_intersection = Some stateflow_numbers;
          } else
          fatal_error (Printf.sprintf
            "Duplicated intersection in src/dst in field %s" item_name) in
      loop accu stateflow_src_dst_items in
  loop empty_classified_stateflow_src_dst_item stateflow_src_dst_items
;;

let classify_stateflow_eml_items container_description stateflow_eml_items =
  let rec loop accu = function
    | [] -> accu
    | stateflow_eml_item :: stateflow_eml_items ->
      let accu =
        match stateflow_eml_item with
        | Stateflow_eml_name mdl_name ->
          check_unset
            (Printf.sprintf "eml name in %s" container_description)
            accu.stateflow_eml_name;
          {
            accu with
            stateflow_eml_name = Some mdl_name;
          }
        | Stateflow_eml_script mdl_script ->
          check_unset
            (Printf.sprintf "eml script in %s" container_description)
            accu.stateflow_eml_script;
          {
            accu with
            stateflow_eml_script = Some mdl_script;
          }

        | Stateflow_eml_binding binding ->
          match binding with
          | { Binding.
              bound_ident = bound_ident;
              bound_value = _bound_value;
            } ->
            let deep_binding = Mdl_deep_basics.deep_binding binding in
            Hashtbl.add
              accu.stateflow_eml_binding_table bound_ident deep_binding;
            accu in
      loop accu stateflow_eml_items in
  loop (mk_empty_classified_stateflow_eml_item ()) stateflow_eml_items
;;

let find_ref_field_option _item_name classification =
  classification.stateflow_src_dst_ref
;;

let find_intersection_field item_name classification =
  match classification.stateflow_intersection with
  | None ->
    fatal_error (Printf.sprintf
      "cannot find field \'%s\' of src/dst sub-item \'%s\'"
      "intersection"
      item_name)
  | Some intersection -> intersection
;;

let deep_src_dst_fields item_name stateflow_src_dst_items =
  let classification =
    classify_stateflow_src_dst_items item_name stateflow_src_dst_items in
  let id_option =
    find_ref_field_option item_name classification in
  let intersection =
    find_intersection_field item_name classification in

  id_option, intersection
;;

let deep_stateflow_src item_name stateflow_items =
  let id_option, intersection =
    deep_src_dst_fields item_name stateflow_items in
  {
    src_id = id_option;
    src_intersection = deep_intersection intersection;
  }
;;

let deep_stateflow_dst item_name stateflow_items =
  let id_option, intersection =
    deep_src_dst_fields item_name stateflow_items in
  {
    dst_id = id_option;
    dst_intersection = deep_intersection intersection;
  }
;;

let find_src_field item_name field_option =
  let stateflow_items = find_field src_ident item_name field_option in
  let item_name = "src" in
  deep_stateflow_src item_name stateflow_items
;;
let check_src_field_empty = check_field_empty src_ident;;

let find_dst_field item_name field_option =
  let stateflow_items = find_field src_ident item_name field_option in
  let item_name = "dst" in
  deep_stateflow_dst item_name stateflow_items
;;
let check_dst_field_empty = check_field_empty dst_ident;;

let deep_stateflow_data_scope = function
  | Ident.Ident "CONSTANT_DATA" -> Deep_scope_constant_data
  | Ident.Ident "DATA_STORE_MEMORY_DATA" -> Deep_scope_data_store_memory_data
  | Ident.Ident "EXPORTED_DATA" -> Deep_scope_exported_data
  | Ident.Ident "FUNCTION_INPUT_DATA" -> Deep_scope_function_input_data
  | Ident.Ident "FUNCTION_OUTPUT_DATA" -> Deep_scope_function_output_data
  | Ident.Ident "INPUT_DATA" -> Deep_scope_input_data
  | Ident.Ident "INPUT_EVENT" -> Deep_scope_input_event
  | Ident.Ident "LOCAL_DATA" -> Deep_scope_local_data
  | Ident.Ident "LOCAL_EVENT" -> Deep_scope_local_event
  | Ident.Ident "OUTPUT_DATA" -> Deep_scope_output_data
  | Ident.Ident "OUTPUT_EVENT" -> Deep_scope_output_event
  | Ident.Ident "PARAMETER_DATA" -> Deep_scope_parameter_data
  | Ident.Ident "TEMPORARY_DATA" -> Deep_scope_temporary_data
  | Ident.Ident s -> Deep_scope_data_string s
;;

let find_scope_field item_name field_option =
  let scope = find_field scope_ident item_name field_option in
  deep_stateflow_data_scope scope
;;
let check_scope_field_empty = check_field_empty scope_ident;;

let default_props_array_size = Ast_types.mdl_text_of_string "-1";;

let deep_stateflow_data_props props =

  let deep_stateflow_props_array stateflow_props_items =

    let deep_props_array_item_binding_table = Hash_table.of_bindings [] in

    let rec classify (size_opt, first_index_opt as accu) =
      function
      | [] -> accu
      | Stateflow_props_array_item_size mdl_text :: rest ->
        check_unset "size props array" size_opt;
        classify (Some mdl_text, first_index_opt) rest
      | Stateflow_props_array_item_first_index mdl_text :: rest ->
        check_unset "firstIndex props array" first_index_opt;
        classify (size_opt, Some mdl_text) rest
      | Stateflow_props_array_item_binding binding :: rest ->
        add_deep_binding deep_props_array_item_binding_table binding;
        classify (size_opt, first_index_opt) rest in
    let size_opt, first_index_opt =
      classify (None, None) stateflow_props_items in
    let size =
      match size_opt with
      | None -> default_props_array_size
      | Some size -> size in
    {
      deep_props_array_item_size = size;
      deep_props_array_item_first_index = first_index_opt;
      deep_props_array_item_binding_table =
        deep_props_array_item_binding_table;
    } in

  let deep_stateflow_props_type stateflow_props_type_items =
     let props_type_deep_binding_table = Hash_table.of_bindings [] in
     List.iter
       (add_deep_binding props_type_deep_binding_table)
       stateflow_props_type_items;
     props_type_deep_binding_table in

  let deep_props_binding_table = Hash_table.of_bindings [] in

  let rec classify_stateflow_props_items array_opt type_opt = function
    | [] -> array_opt, type_opt
    | Stateflow_props_array stateflow_props_array_items :: rest ->
      check_unset "props array" array_opt;
      let deep_props_array =
        deep_stateflow_props_array stateflow_props_array_items in
      classify_stateflow_props_items (Some deep_props_array) type_opt rest
    | Stateflow_props_type stateflow_props_type_items :: rest ->
      check_unset "props type" type_opt;
      let deep_props_type =
        deep_stateflow_props_type stateflow_props_type_items in
      classify_stateflow_props_items array_opt (Some deep_props_type) rest
    | Stateflow_props_binding binding :: rest ->
      add_deep_binding deep_props_binding_table binding;
      classify_stateflow_props_items array_opt type_opt rest in

  let deep_props_array_opt, deep_props_type_opt =
    classify_stateflow_props_items None None props in
  {
    deep_props_array = deep_props_array_opt;
    deep_props_type = deep_props_type_opt;
    deep_props_binding_table = deep_props_binding_table;
  }
;;

let find_props_field item_name field_option =
  let props = find_field props_ident item_name field_option in
  deep_stateflow_data_props props
;;

let check_props_field_empty = check_field_empty props_ident;;

let deep_stateflow_state_type = function
  | "AND_STATE" -> Mdl_stateflow_deep_ast.Deep_state_type_and
  | "OR_STATE" -> Deep_state_type_or
  | "FUNC_STATE" -> Deep_state_type_func
  | "GROUP_STATE" -> Deep_state_type_group
  | "SUB" -> Deep_state_type_sub
  | "SUPER" -> Deep_state_type_super
  | "EML_CHART" -> Deep_state_type_eml_chart
  | s -> Deep_state_type_string (Ast_types.mdl_text_of_string s)
;;

let deep_stateflow_chart_type = deep_stateflow_state_type;;

let deep_stateflow_junction_type = function
  | "CONNECTIVE_JUNCTION" -> Deep_junction_type_connective_junction
  | "HISTORY_JUNCTION" -> Deep_junction_type_history_junction
  | s -> Deep_junction_type_string (Ast_types.mdl_text_of_string s)
;;

let deep_stateflow_transition_type = function
  | "SUB" -> Deep_transition_type_sub
  | s -> Deep_transition_type_string (Ast_types.mdl_text_of_string s)
;;

(* FIXME: to be removed
let find_eml_script_field = find_field script_ident;;
*)
let find_eml_script_field_option = find_field_option script_ident;;
let check_eml_script_empty = check_field_empty script_ident;;

let find_eml_name_field = find_field name_ident;;
let check_eml_name_empty = check_field_empty name_ident;;

let check_eml_binding_table_empty container_description binding_table =
  if Hashtbl.length binding_table <> 0 then
    warning (Printf.sprintf
      "spurious bindings for stateflow item \'eml\' of \'%s\'"
      container_description)
;;

let deep_stateflow_data = function
  | {
      stateflow_data_id = id;
      stateflow_data_items = stateflow_items;
    } ->
    let item_name = "data" in
    let classification =
      classify_stateflow_items stateflow_items in
    match classification with
    | {
        stateflow_eml = stateflow_eml_opt;
        stateflow_execution_order = stateflow_execution_order_opt;
        stateflow_label_string = stateflow_label_string_opt;
        stateflow_chart_file_number = stateflow_chart_file_number_opt;
        stateflow_name = stateflow_name_opt;
        stateflow_scope = stateflow_scope_opt;
        stateflow_props = stateflow_props_opt;
        stateflow_ssid_number = stateflow_ssid_number_opt;
        stateflow_trigger = stateflow_trigger_opt;
        stateflow_type = stateflow_type_opt;
        stateflow_src = stateflow_src_opt;
        stateflow_dst = stateflow_dst_opt;
        stateflow_plant_modeling_info = _stateflow_plant_modeling_info_opt;
        stateflow_binding_table = stateflow_binding_table;
      } ->

    let name = find_name_field item_name stateflow_name_opt in
    let ssid_number =
      find_ssid_number_field item_name stateflow_ssid_number_opt in
    let scope = find_scope_field item_name stateflow_scope_opt in
    let props = find_props_field item_name stateflow_props_opt in
    let table = stateflow_binding_table in

    check_eml_field_empty item_name stateflow_eml_opt;
    check_execution_order_field_empty item_name stateflow_execution_order_opt;
    check_label_string_field_empty item_name stateflow_label_string_opt;
    check_chart_file_number_field_empty
      item_name stateflow_chart_file_number_opt;
    check_trigger_field_empty item_name stateflow_trigger_opt;
    check_type_field_empty item_name stateflow_type_opt;
    check_src_field_empty item_name stateflow_src_opt;
    check_dst_field_empty item_name stateflow_dst_opt;

    {
      deep_data_id = id;
      deep_data_name = name;
      deep_data_ssid_number = ssid_number;
      deep_data_scope = scope; (* scope *)
      deep_data_props = props; (* props *)
      deep_data_binding_table = table;
    }
;;

let deep_stateflow_junction = function
  | {
      stateflow_junction_id = id;
      stateflow_junction_items = stateflow_items;
    } ->

    let item_name = "junction" in
    let classification =
      classify_stateflow_items stateflow_items in
    match classification with
    | {
        stateflow_eml = stateflow_eml_opt;
        stateflow_execution_order = stateflow_execution_order_opt;
        stateflow_label_string = stateflow_label_string_opt;
        stateflow_chart_file_number = stateflow_chart_file_number_opt;
        stateflow_name = stateflow_name_opt;
        stateflow_scope = stateflow_scope_opt;
        stateflow_props = stateflow_props_opt;
        stateflow_ssid_number = stateflow_ssid_number_opt;
        stateflow_trigger = stateflow_trigger_opt;
        stateflow_type = stateflow_type_opt;
        stateflow_src = stateflow_src_opt;
        stateflow_dst = stateflow_dst_opt;
        stateflow_plant_modeling_info = _stateflow_plant_modeling_info_opt;
        stateflow_binding_table = stateflow_binding_table;
      } ->

    let ssid_number =
      find_ssid_number_field item_name stateflow_ssid_number_opt in
    let ty_text = find_type_field item_name stateflow_type_opt in
    let ty =
      deep_stateflow_junction_type (Ast_types.string_of_ident ty_text) in
    let label_string_option =
      find_label_string_field_option item_name stateflow_label_string_opt in
    let table = stateflow_binding_table in

    check_eml_field_empty item_name stateflow_eml_opt;
    check_execution_order_field_empty item_name stateflow_execution_order_opt;
    check_chart_file_number_field_empty
      item_name stateflow_chart_file_number_opt;
    check_name_field_empty item_name stateflow_name_opt;
    check_scope_field_empty item_name stateflow_scope_opt;
    check_props_field_empty item_name stateflow_props_opt;
    check_trigger_field_empty item_name stateflow_trigger_opt;
    check_src_field_empty item_name stateflow_src_opt;
    check_dst_field_empty item_name stateflow_dst_opt;

    {
      deep_junction_id = id;
      deep_junction_ssid_number = ssid_number;
      deep_junction_type = ty;
      deep_junction_label_string = label_string_option;
      deep_junction_binding_table = table;
    }
;;

let deep_stateflow_transition = function
  | {
      stateflow_transition_id = id;
      stateflow_transition_items = stateflow_items;
    } ->

  let item_name = "transition" in
  let classification =
    classify_stateflow_items stateflow_items in
    match classification with
    | {
        stateflow_eml = stateflow_eml_opt;
        stateflow_execution_order = stateflow_execution_order_opt;
        stateflow_label_string = stateflow_label_string_opt;
        stateflow_chart_file_number = stateflow_chart_file_number_opt;
        stateflow_name = stateflow_name_opt;
        stateflow_scope = stateflow_scope_opt;
        stateflow_props = stateflow_props_opt;
        stateflow_ssid_number = stateflow_ssid_number_opt;
        stateflow_trigger = stateflow_trigger_opt;
        stateflow_type = stateflow_type_opt;
        stateflow_src = stateflow_src_opt;
        stateflow_dst = stateflow_dst_opt;
        stateflow_plant_modeling_info = _stateflow_plant_modeling_info_opt;
        stateflow_binding_table = stateflow_binding_table;
      } ->

  let ssid_number =
    find_ssid_number_field item_name stateflow_ssid_number_opt in
  let ty_text_opt = find_type_field_option item_name stateflow_type_opt in
  let ty_opt =
    match ty_text_opt with
    | None -> None
    | Some ty_text ->
      Some
        (deep_stateflow_transition_type (Ast_types.string_of_ident ty_text)) in
  let label_string_option =
    find_label_string_field_option item_name stateflow_label_string_opt in
  let src = find_src_field item_name stateflow_src_opt in
  let dst = find_dst_field item_name stateflow_dst_opt in
  let execution_order =
    find_execution_order_field_option
      item_name stateflow_execution_order_opt in

  check_eml_field_empty item_name stateflow_eml_opt;
  check_chart_file_number_field_empty
    item_name stateflow_chart_file_number_opt;
  check_name_field_empty item_name stateflow_name_opt;
  check_scope_field_empty item_name stateflow_scope_opt;
  check_props_field_empty item_name stateflow_props_opt;
  check_trigger_field_empty item_name stateflow_trigger_opt;

  let table = stateflow_binding_table in

  {
    deep_transition_id = id;
    deep_transition_type = ty_opt;
    deep_transition_ssid_number = ssid_number;
    deep_transition_label_string = label_string_option;
    deep_transition_src = src;
    deep_transition_dst = dst;
    deep_transition_execution_order = execution_order;
    deep_transition_binding_table = table;
  }
;;

let deep_stateflow_event = function
  | {
      stateflow_event_id = id;
      stateflow_event_items = stateflow_items;
    } ->

    let item_name = "event" in
    let classification =
      classify_stateflow_items stateflow_items in
    match classification with
    | {
        stateflow_eml = stateflow_eml_opt;
        stateflow_execution_order = stateflow_execution_order_opt;
        stateflow_label_string = stateflow_label_string_opt;
        stateflow_chart_file_number = stateflow_chart_file_number_opt;
        stateflow_name = stateflow_name_opt;
        stateflow_scope = stateflow_scope_opt;
        stateflow_props = stateflow_props_opt;
        stateflow_ssid_number = stateflow_ssid_number_opt;
        stateflow_trigger = stateflow_trigger_opt;
        stateflow_type = stateflow_type_opt;
        stateflow_src = stateflow_src_opt;
        stateflow_dst = stateflow_dst_opt;
        stateflow_plant_modeling_info = _stateflow_plant_modeling_info_opt;
        stateflow_binding_table = stateflow_binding_table;
      } ->
    let name = find_name_field item_name stateflow_name_opt in
    let scope = find_scope_field item_name stateflow_scope_opt in
    let trigger =
      find_trigger_field item_name stateflow_trigger_opt in

    check_eml_field_empty item_name stateflow_eml_opt;
    check_execution_order_field_empty item_name stateflow_execution_order_opt;
    check_label_string_field_empty item_name stateflow_label_string_opt;
    check_chart_file_number_field_empty
      item_name stateflow_chart_file_number_opt;
    check_props_field_empty item_name stateflow_props_opt;
    check_ssid_number_field_empty item_name stateflow_ssid_number_opt;
    check_type_field_empty item_name stateflow_type_opt;
    check_src_field_empty item_name stateflow_src_opt;
    check_dst_field_empty item_name stateflow_dst_opt;

    {
      deep_event_id = id;
      deep_event_name = name;
      deep_event_scope = scope; (* scope *)
      deep_event_trigger = trigger; (* trigger *)
      deep_event_binding_table = stateflow_binding_table;
    }
;;

let deep_stateflow_state_eml_opt container_description = function
  | None -> None
  | Some stateflow_eml_items ->
    let classification =
      classify_stateflow_eml_items container_description stateflow_eml_items in
    match classification with
    | {
        stateflow_eml_name = stateflow_eml_name_opt;
        stateflow_eml_script = stateflow_eml_script_opt;
        stateflow_eml_binding_table = stateflow_eml_binding_table;
      } ->
      let eml_script_opt =
        find_eml_script_field_option
          container_description stateflow_eml_script_opt in
      check_eml_name_empty container_description stateflow_eml_name_opt;
      Some {
        state_eml_script = eml_script_opt;
        state_eml_binding_table = stateflow_eml_binding_table;
      }
;;

let deep_stateflow_chart_eml_opt container_description = function
  | None -> None
  | Some stateflow_eml_items ->
    let classification =
      classify_stateflow_eml_items container_description stateflow_eml_items in
    match classification with
    | {
        stateflow_eml_name = stateflow_eml_name_opt;
        stateflow_eml_script = stateflow_eml_script_opt;
        stateflow_eml_binding_table = stateflow_eml_binding_table;
      } ->
      Say.fdebug (fun epf ->
        Format.fprintf epf "container %s: %a"
          container_description
          Mdl_basics_deep_ast_print.print_deep_binding_table
            stateflow_eml_binding_table);
      let eml_name =
        find_eml_name_field container_description stateflow_eml_name_opt in
      check_eml_script_empty container_description stateflow_eml_script_opt;
      check_eml_binding_table_empty
        container_description stateflow_eml_binding_table;
      Some {
        chart_eml_name = eml_name;
        chart_eml_binding_table = stateflow_eml_binding_table;
      }
;;

let rec deep_stateflow_state = function
  | {
      stateflow_state_id = id;
      stateflow_state_datas = stateflow_datas;
      stateflow_state_events = stateflow_events;
      stateflow_state_junctions = stateflow_junctions;
      stateflow_state_states = stateflow_states;
      stateflow_state_transitions = stateflow_transitions;
      stateflow_state_items = stateflow_items;
    } ->
    let item_name = "state" in
    let container_description =
      Printf.sprintf "%s %d" item_name (Unatural.to_int id) in
    let classification =
      classify_stateflow_items stateflow_items in
    match classification with
    | {
        stateflow_eml = stateflow_eml_opt;
        stateflow_execution_order = stateflow_execution_order_opt;
        stateflow_label_string = stateflow_label_string_opt;
        stateflow_chart_file_number = stateflow_chart_file_number_opt;
        stateflow_name = stateflow_name_opt;
        stateflow_scope = stateflow_scope_opt;
        stateflow_props = stateflow_props_opt;
        stateflow_ssid_number = stateflow_ssid_number_opt;
        stateflow_trigger = stateflow_trigger_opt;
        stateflow_type = stateflow_type_opt;
        stateflow_src = stateflow_src_opt;
        stateflow_dst = stateflow_dst_opt;
        stateflow_plant_modeling_info = _stateflow_plant_modeling_info_opt;
        stateflow_binding_table = stateflow_binding_table;
      } ->
    let ssid_number =
      find_ssid_number_field item_name stateflow_ssid_number_opt in
    let ty_ident = find_type_field item_name stateflow_type_opt in
    let ty =
      deep_stateflow_state_type (Ast_types.string_of_ident ty_ident) in
    let label_string =
      find_label_string_field item_name stateflow_label_string_opt in
    let execution_order =
      find_execution_order_field_option
        item_name stateflow_execution_order_opt in
    let deep_stateflow_state_eml_opt =
      deep_stateflow_state_eml_opt container_description stateflow_eml_opt in

    check_chart_file_number_field_empty
      item_name stateflow_chart_file_number_opt;
    check_name_field_empty item_name stateflow_name_opt;
    check_scope_field_empty item_name stateflow_scope_opt;
    check_props_field_empty item_name stateflow_props_opt;
    check_trigger_field_empty item_name stateflow_trigger_opt;
    check_src_field_empty item_name stateflow_src_opt;
    check_dst_field_empty item_name stateflow_dst_opt;

    let deep_datas =
      List.map deep_stateflow_data stateflow_datas
    and deep_events =
      List.map deep_stateflow_event stateflow_events
    and deep_junctions =
      List.map deep_stateflow_junction stateflow_junctions
    and deep_states =
      List.map deep_stateflow_state stateflow_states
    and deep_transitions =
      List.map deep_stateflow_transition stateflow_transitions in

    {
      deep_state_id = id; (* "id" *)
      deep_state_ssid_number = ssid_number; (* "ssIdNumber" *)
      deep_state_type = ty; (* type *)
      deep_state_label_string = label_string; (* labelString *)
      deep_state_execution_order = execution_order;
      deep_stateflow_state_eml = deep_stateflow_state_eml_opt;
      deep_stateflow_state_datas = deep_datas;
      deep_stateflow_state_events = deep_events;
      deep_stateflow_state_junctions = deep_junctions;
      deep_stateflow_state_states = deep_states;
      deep_stateflow_state_transitions = deep_transitions;
      deep_state_binding_table = stateflow_binding_table;
    }
;;

let deep_stateflow_chart = function
  | {
      stateflow_chart_id = id;
      stateflow_chart_datas = stateflow_datas;
      stateflow_chart_events = stateflow_events;
      stateflow_chart_junctions = stateflow_junctions;
      stateflow_chart_states = stateflow_states;
      stateflow_chart_transitions = stateflow_transitions;
      stateflow_chart_items = stateflow_items;
    } ->
    let item_name = "chart" in
    let container_description =
      Printf.sprintf "%s %d" item_name (Unatural.to_int id) in
    let classification =
      classify_stateflow_items stateflow_items in
    match classification with
    | {
        stateflow_eml = stateflow_eml_opt;
        stateflow_execution_order = stateflow_execution_order_opt;
        stateflow_label_string = stateflow_label_string_opt;
        stateflow_chart_file_number = stateflow_chart_file_number_opt;
        stateflow_name = stateflow_name_opt;
        stateflow_scope = stateflow_scope_opt;
        stateflow_props = stateflow_props_opt;
        stateflow_ssid_number = stateflow_ssid_number_opt;
        stateflow_trigger = stateflow_trigger_opt;
        stateflow_type = stateflow_type_opt;
        stateflow_src = stateflow_src_opt;
        stateflow_dst = stateflow_dst_opt;
        stateflow_plant_modeling_info = _stateflow_plant_modeling_info_opt;
        stateflow_binding_table = stateflow_binding_table;
      } ->
    let name = find_name_field item_name stateflow_name_opt in
    let deep_stateflow_chart_eml_opt =
      deep_stateflow_chart_eml_opt container_description stateflow_eml_opt in
    let deep_chart_file_number =
      find_chart_file_number_field item_name stateflow_chart_file_number_opt in
    let ty_ident_opt = find_type_field_option item_name stateflow_type_opt in
    let ty_opt =
      match ty_ident_opt with
      | None -> None
      | Some ty_ident ->
        Some
          (deep_stateflow_chart_type (Ast_types.string_of_ident ty_ident)) in

    check_execution_order_field_empty item_name stateflow_execution_order_opt;
    check_label_string_field_empty item_name stateflow_label_string_opt;
    check_scope_field_empty item_name stateflow_scope_opt;
    check_props_field_empty item_name stateflow_props_opt;
    check_ssid_number_field_empty item_name stateflow_ssid_number_opt;
    check_trigger_field_empty item_name stateflow_trigger_opt;
    check_src_field_empty item_name stateflow_src_opt;
    check_dst_field_empty item_name stateflow_dst_opt;

    let deep_datas =
      List.map deep_stateflow_data stateflow_datas
    and deep_events =
      List.map deep_stateflow_event stateflow_events
    and deep_junctions =
      List.map deep_stateflow_junction stateflow_junctions
    and deep_states =
      List.map deep_stateflow_state stateflow_states
    and deep_transitions =
      List.map deep_stateflow_transition stateflow_transitions in

    {
      Mdl_stateflow_deep_ast.
      deep_chart_id = id;
      deep_chart_name = name;
      deep_chart_file_number = deep_chart_file_number;
      deep_chart_type = ty_opt;
      deep_stateflow_chart_eml = deep_stateflow_chart_eml_opt;
      deep_stateflow_chart_datas = deep_datas;
      deep_stateflow_chart_events = deep_events;
      deep_stateflow_chart_junctions = deep_junctions;
      deep_stateflow_chart_states = deep_states;
      deep_stateflow_chart_transitions = deep_transitions;
      deep_chart_binding_table = stateflow_binding_table;
    }
;;

let deep_stateflow_target = function
  | {
      stateflow_target_id = id;
      stateflow_target_items = stateflow_items;
    } ->
    let item_name = "target" in
    let classification =
      classify_stateflow_items stateflow_items in
    match classification with
    | {
        stateflow_eml = stateflow_eml_opt;
        stateflow_execution_order = stateflow_execution_order_opt;
        stateflow_label_string = stateflow_label_string_opt;
        stateflow_chart_file_number = stateflow_chart_file_number_opt;
        stateflow_name = stateflow_name_opt;
        stateflow_scope = stateflow_scope_opt;
        stateflow_props = stateflow_props_opt;
        stateflow_ssid_number = stateflow_ssid_number_opt;
        stateflow_trigger = stateflow_trigger_opt;
        stateflow_type = stateflow_type_opt;
        stateflow_src = stateflow_src_opt;
        stateflow_dst = stateflow_dst_opt;
        stateflow_plant_modeling_info = _stateflow_plant_modeling_info_opt;
        stateflow_binding_table = stateflow_binding_table;
      } ->
    let name_option = find_name_field_option item_name stateflow_name_opt in

    check_eml_field_empty item_name stateflow_eml_opt;
    check_execution_order_field_empty item_name stateflow_execution_order_opt;
    check_label_string_field_empty item_name stateflow_label_string_opt;
    check_chart_file_number_field_empty
      item_name stateflow_chart_file_number_opt;
    check_scope_field_empty item_name stateflow_scope_opt;
    check_props_field_empty item_name stateflow_props_opt;
    check_ssid_number_field_empty item_name stateflow_ssid_number_opt;
    check_trigger_field_empty item_name stateflow_trigger_opt;
    check_type_field_empty item_name stateflow_type_opt;
    check_src_field_empty item_name stateflow_src_opt;
    check_dst_field_empty item_name stateflow_dst_opt;

    {
      deep_target_id = id;
      deep_target_name = name_option;
      deep_target_binding_table = stateflow_binding_table;
    }
;;

let deep_stateflow_machine = function
  | {
      stateflow_machine_id = id;
      stateflow_machine_charts = stateflow_charts;
      stateflow_machine_datas = stateflow_datas;
      stateflow_machine_events = stateflow_events;
      stateflow_machine_targets = stateflow_targets;
      stateflow_machine_items = stateflow_items;
    } ->
    let item_name = "machine" in
    let classification =
      classify_stateflow_items stateflow_items in
    match classification with
    | {
        stateflow_eml = stateflow_eml_opt;
        stateflow_execution_order = stateflow_execution_order_opt;
        stateflow_label_string = stateflow_label_string_opt;
        stateflow_chart_file_number = stateflow_chart_file_number_opt;
        stateflow_name = stateflow_name_opt;
        stateflow_scope = stateflow_scope_opt;
        stateflow_props = stateflow_props_opt;
        stateflow_ssid_number = stateflow_ssid_number_opt;
        stateflow_trigger = stateflow_trigger_opt;
        stateflow_type = stateflow_type_opt;
        stateflow_src = stateflow_src_opt;
        stateflow_dst = stateflow_dst_opt;
        stateflow_plant_modeling_info = _stateflow_plant_modeling_info_opt;
        stateflow_binding_table = stateflow_binding_table;
      } ->
    let name_option = find_name_field_option item_name stateflow_name_opt in
    let deep_charts =
      List.map deep_stateflow_chart stateflow_charts
    and deep_datas =
      List.map deep_stateflow_data stateflow_datas
    and deep_events =
      List.map deep_stateflow_event stateflow_events
    and deep_targets =
      List.map deep_stateflow_target stateflow_targets in

    check_eml_field_empty item_name stateflow_eml_opt;
    check_execution_order_field_empty item_name stateflow_execution_order_opt;
    check_label_string_field_empty item_name stateflow_label_string_opt;
    check_chart_file_number_field_empty
      item_name stateflow_chart_file_number_opt;
    check_scope_field_empty item_name stateflow_scope_opt;
    check_props_field_empty item_name stateflow_props_opt;
    check_ssid_number_field_empty item_name stateflow_ssid_number_opt;
    check_trigger_field_empty item_name stateflow_trigger_opt;
    check_type_field_empty item_name stateflow_type_opt;
    check_src_field_empty item_name stateflow_src_opt;
    check_dst_field_empty item_name stateflow_dst_opt;

    {
      Mdl_stateflow_deep_ast.
      deep_machine_id = id;
      deep_machine_name = name_option;
      deep_stateflow_machine_charts = deep_charts;
      deep_stateflow_machine_datas = deep_datas;
      deep_stateflow_machine_events = deep_events;
      deep_stateflow_machine_targets = deep_targets;
      deep_machine_binding_table = stateflow_binding_table;
    }
;;

let deep_stateflow_section_instance = function
  | {
      stateflow_instance_id = id;
      stateflow_instance_name = mdl_name;
      stateflow_instance_machine_ref = machine_ref;
      stateflow_instance_chart_ref = chart_ref;
    } ->
    {
      deep_instance_id = id;
      deep_instance_name = mdl_name;
      deep_instance_machine_ref = machine_ref;
      deep_instance_chart_ref = chart_ref;
    }
;;

let deep_stateflow_section_item = function
  | Mdl_ast.
    Stateflow_section_machine stateflow_machine ->
    Deep_stateflow_section_machine
      (deep_stateflow_machine stateflow_machine)
  | Stateflow_section_instance stateflow_instance ->
    Deep_stateflow_section_instance
      (deep_stateflow_section_instance stateflow_instance)
;;

let deep_stateflow_section stateflow_section =
  match stateflow_section with
  | Stateflow_section stateflow_section_items ->
(*
    Say.fdebug
      (fun epf ->
       Mdl_ast_print.print_stateflow_section_items
         epf stateflow_section_items);
*)
    Deep_stateflow_section
      (List.rev_map deep_stateflow_section_item stateflow_section_items)
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
