(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_basics_to_explicit_ast.ml,v 1.1 2019-05-10 14:04:56 jpc Exp $ *)

(* From deep syntax to completely explicit inherited tree:
   - defaults sections are incorporated into values,
   - no more default sections in the resulting trees: all fields are explicit
     and checked with respect to various type constraints or range constraints.
*)
open Mdl_basics_deep_ast;;

let explicit_error s =
  Say.fatal_error (
    Printf.sprintf "Inheritance: %s" s)
;;

let natural_of_deep_value = Mdl_deep_basics.natural_of_deep_value;;

let bool_of_deep_value = Mdl_deep_basics.bool_of_deep_value;;

let string_of_deep_value message = function
  | Deep_string s -> s
  | Deep_wstring _
  | Deep_int _
  | Deep_bool _
  | Deep_float _
  | Deep_ident _
  | Deep_expression _
  | Deep_code _
  | Deep_bindings _
  | Deep_vector _
  | Deep_matrix _ ->
    explicit_error (Printf.sprintf "Unknown string for %s" message)
;;

let natural_list_of_deep_value message = function
  | Deep_vector deep_value_list ->
    List.map (natural_of_deep_value message) deep_value_list
  | Deep_wstring _
  | Deep_bool _
  | Deep_float _
  | Deep_int _
  | Deep_string _
  | Deep_ident _
  | Deep_expression _
  | Deep_code _
  | Deep_bindings _
  | Deep_matrix _ ->
    explicit_error (Printf.sprintf "Unknown value list for %s" message)
;;

let ports_of_deep_value = natural_list_of_deep_value "ports";;

let name_of_explicit_block_regular_type =
  Mdl_block_types_tables.string_of_block_type

and name_of_explicit_block_reference_type = function
  | { Mdl_deep_ast.
      deep_block_source_type = Ast_types.Mdl_name ty;
      deep_block_source_path = Ast_types.Mdl_path path;
    } ->
    Printf.sprintf "%s, %s" ty path
;;

let name_of_explicit_block_type = function
  | Mdl_basics_explicit_ast.Explicit_block_regular_type bst ->
    name_of_explicit_block_regular_type bst
  | Mdl_basics_explicit_ast.Explicit_block_reference_type brt ->
    name_of_explicit_block_reference_type brt
;;

let name_of_explicit_block_source_type = function
  | Mdl_basics_explicit_ast.Explicit_block_regular_type bst ->
    name_of_explicit_block_regular_type bst
  | Mdl_basics_explicit_ast.Explicit_block_reference_type brt ->
    Ast_types.string_of_mdl_name
      brt.Mdl_deep_ast.deep_block_source_type
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
