(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_implementation_explicit_ast.mli,v 1.1 2019-05-10 14:04:56 jpc Exp $ *)

open Mdl_ast;;
open Mdl_basics_explicit_ast;;

type explicit_block_name =
  Mdl_basics_explicit_ast.explicit_block_name
and explicit_model_version =
  Mdl_basics_explicit_ast.explicit_model_version
and explicit_matlab_function_script =
  Mdl_basics_explicit_ast.explicit_source_code
;;

type explicit_block_graphical = {
  explicit_block_background_color : Mdl_basics_explicit_ast.explicit_color;
  explicit_block_foreground_color : explicit_color;
  explicit_block_mirror : Mdl_basics_explicit_ast.explicit_mirror;
  explicit_block_ports : Mdl_basics_explicit_ast.explicit_block_ports;
  explicit_block_ports_description :
    Mdl_basics_explicit_ast.explicit_ports_description;
  explicit_block_position : Mdl_basics_explicit_ast.explicit_position;
  explicit_block_rotation : Mdl_basics_explicit_ast.explicit_rotation;
  explicit_block_show_name : Mdl_basics_explicit_ast.explicit_show_name;
}
(** The graphical properties of a block.

  All those properties must be specified for each block, either explicitly
  in a binding of the block, or inherited from the MDL file (system or
  library) in the block defaults section or in the block parameter defaults
  section. *)
;;

type explicit_block_optional_graphical = {
  explicit_block_drop_shadow : explicit_drop_shadow option;
  explicit_block_font : explicit_font;
  explicit_block_name_placement : explicit_placement option;
  explicit_block_hide_automatic_name : explicit_hide_automatic_name option;
  explicit_block_state : explicit_state option;
  explicit_block_tag : explicit_tag option;
}
(** Optional properties that may not be defined at all.
  Potentially useful for translation to Scicos host language. *)
;;

type explicit_binding_table = explicit_value Binding.ident_binding_table
(** A table that defines bindings (i.e. name to value associations) for a
  block: a binding table can specifies the value of proper block parameters
  as well as other bindings of a block. *)
;;

type explicit_block_parameters = explicit_binding_table
(** Just for easy human memorization. *)
;;

type explicit_block_all_bindings = {
  explicit_block_parameters : explicit_block_parameters;
  explicit_block_bindings : explicit_binding_table;
  explicit_block_graphical : explicit_block_graphical;
  explicit_block_optional_graphical : explicit_block_optional_graphical;
}
(** Block explicit bindings are made of all the properties defined for a
  block instance:
    - field [explicit_block_parameters]: the block actual parameter values as
      specified in the block or inherited, stored in a table of bindings,
    - field [explicit_block_bindings]: other block specific bindings
      defined in the block, properly stored in a table of bindings,
    - field [explicit_block_graphical]: the graphical properties of the
      block, as specified or inherited,
    - field [explicit_block_optional_graphical]: other (optional)
      graphical properties.
*)
;;

(** {6 Blocks and systems} *)

type explicit_annotation = {
  explicit_annotation_name : Ast_types.mdl_text;
  explicit_annotation_position : explicit_annotation_position;
  explicit_annotation_font : explicit_font;
  explicit_annotation_graphical : explicit_font_and_colors;
  explicit_annotation_graphical_defaults :
    explicit_annotation_graphical_defaults;
}
and explicit_annotations = explicit_annotation list
(** An annotation is a piece of text written somewhere on the model. *)
;;

(** The blocks *)

type explicit_simple_block_activation =
  | Simple_block_plain
  | Simple_block_activated of simple_block_activation

and simple_block_activation =
  | Simple_block_enabled
  | Simple_block_if_action
  | Simple_block_triggered
  | Simple_block_triggered_and_enabled
(** Activation mode for Simulink simple blocks of action.
  A simple block is a block that is not a sub-system
  (it is either a regular block or a reference block). *)
;;

type 'ty explicit_simple_block_desc = {
  explicit_simple_block_type : 'ty;
  explicit_simple_block_activation : explicit_simple_block_activation;
}
(** An explicit simple block descriptor completely describes all Simulink
  blocks that are not sub-systems.

  The block activation field is similar to the one for sub-system, except
  that the Sub_system_plain_atomic_unit kind of activation does not apply,
  and there is no *Port block to fire the activation; hence the simpler
  summand of type [explicit_simple_block_activation].
*)
;;

type explicit_action_type =
  | Action_case
  | Action_default
  | Action_else
  | Action_elseif
  | Action_then
  | Action_unset
(** From Simulink property ActionType of an ActionPort block *)

and explicit_trigger_type =
  | Trigger_either
  | Trigger_falling
  | Trigger_function_call
  | Trigger_rising
  | Trigger_rising_edge
  | Trigger_unset
(** From Simulink property TriggerType of a TriggerPort block *)
;;

type explicit_matlab_function_data_desc = {
  explicit_matlab_function_data_name : Ast_types.mdl_name;
  explicit_matlab_function_data_size : Ast_types.mdl_text;
}
;;

type explicit_matlab_function_parameter_data_desc = {
  explicit_matlab_function_parameter_data_name : Ast_types.mdl_name;
  explicit_matlab_function_parameter_data_size : Ast_types.mdl_text;
  (* Information coming from binding paramIndexForInitFromWorkspace of
     stateflow data with PARAMETER_DATA scope. *)
  explicit_matlab_function_parameter_data_index : Unatural.t;
}
;;

type explicit_matlab_function_arguments =
  explicit_matlab_function_argument array
and explicit_matlab_function_argument =
  | Explicit_matlab_function_input of explicit_matlab_function_data_desc
  | Explicit_matlab_function_parameter of
    explicit_matlab_function_parameter_data_desc
;;

type explicit_matlab_function_returns = explicit_matlab_function_return array
and explicit_matlab_function_return =
  | Explicit_matlab_function_output of explicit_matlab_function_data_desc
;;

type explicit_s_function_type =
  | Explicit_s_function
  | Explicit_s_function_stateflow of
      explicit_stateflow_name * explicit_stateflow_chart_file_number
  | Explicit_matlab_function of
      explicit_matlab_function_desc
and explicit_matlab_function_desc = {
  explicit_matlab_function_name : Ast_types.mdl_name;
  explicit_matlab_function_script : explicit_matlab_function_script;
  explicit_matlab_function_arguments : explicit_matlab_function_arguments;
  explicit_matlab_function_returns : explicit_matlab_function_returns;
}
;;

type 'line explicit_block = {
  explicit_block_name : explicit_block_name;
  explicit_block_sid : explicit_block_sid;
  explicit_block_body : explicit_block_body;
  explicit_block_known : bool;
  explicit_block_desc : 'line explicit_block_desc;
}

and explicit_block_sid = explicit_sid option

and 'line explicit_block_desc =
  | Explicit_block_regular of
      explicit_block_regular_type explicit_simple_block_desc
  | Explicit_block_reference of
      explicit_block_reference_type explicit_simple_block_desc
  | Explicit_block_sub_system of
      'line explicit_block_sub_system_desc
  | Explicit_block_s_function of
      explicit_s_function_type explicit_simple_block_desc
(** The block descriptor completely describes all Simulink blocks.
  The descriptor distinguishes all kinds of Simulink blocks, namely
  - regular block (basic blocks),
  - reference blocks (blocks that are a reference to another one),
  - sub-system blocks (blocks that made of a combination of other blocks). *)

and 'line explicit_block_sub_system_desc = {
  explicit_block_sub_system_activation : 'line explicit_sub_system_activation;
  (** The activation classification of the sub-system block. *)
  explicit_block_sub_system_body : 'line explicit_system;
  (** The system associated to the sub-system block (the system contains in
     particular the list of blocks embedded in the sub-system block). *)
  explicit_block_sub_system_treat_as_atomic_unit :
    Mdl_basics_explicit_ast.explicit_treat_as_atomic_unit;
  (** The treat_as_atomic property of the sub-system block (set or not). *)
  explicit_block_sub_system_sample_time :
    Mdl_basics_explicit_ast.explicit_system_sample_time;
  (** The explicit sample time of the sub-system block
     (the value -1 means inherited). *)
}

(** The activation classification of a sub-system block. *)
and 'line explicit_sub_system_activation =
  | Sub_system_plain
    (** A Sub_system_plain sub-system block needs no activation.
      This is the regular case: the sub-system has no embedded *Port block.

      According to its treat_as_atomic binding value, a Sub_system_plain
      sub-system block may be treated as an atomic unit or not.
      Note that, if a Sub_system_plain sub-system must be treated as an
      atomic unit, its sample time must be inherited (hence, its value is
      -1), since otherwise the sub-system would need an activation and thus
      it would have been classified as Sub_system_plain_atomic_unit.
    *)
  | Sub_system_activated of 'line sub_system_activation

and 'line sub_system_activation =

  | Sub_system_plain_atomic_unit of Mdl_deep_ast.deep_value
    (** A Sub_system_plain_atomic_unit sub-system block is a regular
      sub-system block (it has no embedded *Port block), but it has the
      treat_as_atomic_unit property set and a specific sample time that is
      not inherited (meaning, the sample time is not -1).

      A Sub_system_plain_atomic_unit sub-system block needs a clock
      activation with the given sample time.
    *)

  | Sub_system_enabled of 'line explicit_block
    (** The sub-system block needs an event generator.
      A block EnablePort has been found in the blocks of the sub-system
      block. *)

  | Sub_system_if_action of explicit_action_type * 'line explicit_block
    (** The sub-system block needs an event generator.
      An ActionPort block has been found in the blocks of the sub-system
      block. *)

  | Sub_system_triggered of (explicit_trigger_type * 'line explicit_block)
    (** The sub-system block needs an event generator.
      A TriggerPort block has been found in the blocks of the sub-system
      block. *)

  | Sub_system_triggered_and_enabled of
      (explicit_trigger_type * 'line explicit_block) * 'line explicit_block
    (** The sub-system block needs an event generator.
      Both a TriggerPort block && an EnablePort block have been found
      in the blocks of the sub-system block. *)

and explicit_block_body = {
  explicit_block_all_bindings : explicit_block_all_bindings;
  explicit_block_array : explicit_array option;
  explicit_block_object : explicit_object option;
  explicit_block_mask : Mdl_basics_explicit_ast.explicit_mask option;
  explicit_block_sf_block_type :
    Mdl_basics_explicit_ast.explicit_sf_block_type option;
  explicit_block_instance_data : explicit_binding_table option;
}
(**
  Note: the field [explicit_block_instance_data] contains instance data
  properties (the meaning is unknown as of 2015/11/30).
*)

and 'line explicit_system = {
  explicit_system_name : system_name;
  explicit_system_position : Mdl_basics_explicit_ast.explicit_position;
  explicit_system_bindings : explicit_binding_table;
  explicit_system_blocks : 'line explicit_blocks;
  explicit_system_lines : 'line list;
  explicit_system_annotations : explicit_annotations;
  explicit_system_connectors : explicit_system_connectors;
  explicit_system_zoom_factor : explicit_zoom_factor;
}

and 'line explicit_blocks = 'line explicit_block list

and explicit_system_connectors = explicit_system_connector list

and explicit_system_connector = explicit_binding_table
;;

(** {6 The polymorphic model definition} *)

(** The line kind of the model is left abstract here: the kind of
  line of the model will change when processing goes on.

  Each line kind hence defines an explicit model kind that features
  this kind of lines.
 *)

type 'line explicit_model = {
  explicit_model_body : 'line explicit_model_body;
  explicit_model_matdatas_option : explicit_model_matdatas option;
  explicit_stateflow_option :
    Mdl_stateflow_explicit_ast.explicit_stateflow_section option;
}

and explicit_model_matdata =
  | Explicit_matdata_section of explicit_matdata_section_item
and explicit_matdata_section_item = Mdl_deep_ast.deep_matdata_section_item
and explicit_model_matdatas = explicit_model_matdata list

and 'line explicit_model_body = {
  explicit_model_ident : model_ident;
  explicit_model_version : explicit_model_version;
  explicit_model_explicit_file_name : Path.explicit_file_name;
  explicit_model_configuration : explicit_model_configuration;
  explicit_model_workspace : explicit_model_workspace;
  explicit_model_system : 'line explicit_system;
}

and explicit_model_workspace =
  Mdl_basics_explicit_ast.explicit_model_workspace_data list

and explicit_model_configuration = {
  explicit_graphical_parameters : graphical_interface_item list;
  explicit_configuration_bindings :
    (Ident.ident, explicit_value) Hash_table.t;
  explicit_vendor_bindings : explicit_vendor_bindings;
  explicit_active_configuration : explicit_active_configuration option;
(* TORM
  Everything is explicit: no need to keep configurations nor
  eplicit_set_configuration_option ?
  explicit_configurations : explicit_configurations;
  explicit_set_configuration_option : model_set_configuration_option;
*)
}
;;

(** {6 The 'explicit line' connection kind} *)

(**
  Explicit lines represent an explicit representation of the original
  Simulink lines of the model (including dandling lines with no
  target).

  A model with explicit line connections is the result of the complete
  explicitation of inheritance and activation block distinction.
*)

type 'port_type explicit_connector_port =
  {
    explicit_connector_port_block : block_name;
    explicit_connector_port_desc : 'port_type;
  }
(** An [explicit_connector_port] is a connector for line to a block port: it
  defines a block and a connection point, that could be a port number
  (input or output) of the named block or a special port (a port of a
  Simulink action block (a block [Enable], [Trigger], [Enable_and_Trigger],
  or [IfAction]).
*)
;;

(** {7 Explicit connector port for directed lines} *)
type explicit_directed_connector_port =
  Mdl_ast.port_value explicit_connector_port
;;

(** {7 Explicit connector port for connection lines} *)

type explicit_connection_connector_port =
  explicit_connection_connector_desc explicit_connector_port

and explicit_connection_connector_desc =
  | Explicit_left_connector of connection_port_value
  | Explicit_right_connector of connection_port_value
(** Connection line connectors have a left/right annotation to tell where
  they should appear on the block they are connected to (being it on the
  left, the side of regular directed input connectors, or on the right,
  the side of regular directed output connectors). *)

and connection_port_value =
  | Connection_port_value of Ast_types.mdl_port_number
;;

(** {6 Defining connectors} *)

(** {7 Defining points for lines} *)

type explicit_relative_points = explicit_points;;

type explicit_absolute_point =
  {
    explicit_x : int;
    explicit_y : int;
  }
;;

type explicit_directed_points = explicit_relative_points
and explicit_connection_points =
  | Explicit_relative_points of explicit_relative_points
  | Explicit_absolute_points of
      explicit_absolute_point * explicit_relative_points
;;

(** {7 Connectors for connection lines} *)

type explicit_connection_connector =
  | Explicit_connection_connector_port of explicit_connection_connector_port
  | Explicit_connection_connector_branches of explicit_connection_branches

and explicit_connection_branches = explicit_connection_branch list
and explicit_connection_branch = {
  explicit_connection_branch_points : explicit_connection_points option;
  explicit_connection_branch_child : explicit_connection_branch_child;
}

and explicit_connection_branch_child =
  | Explicit_connection_branch_source of explicit_connection_connector
  | Explicit_connection_branch_target of explicit_connection_connector
;;

type explicit_connection_connectors = {
  explicit_source_connection_connector : explicit_connection_connector;
  explicit_target_connection_connector : explicit_connection_connector;
}
;;

(** {7 Connectors for directed lines} *)

type explicit_directed_connectors =
  {
    explicit_directed_source : explicit_directed_connector_port;
    explicit_directed_target : explicit_directed_connector_port option;
    (* Branches are still there: they will disappear after the split pass
       when branches become split boxes and lines become links. *)
    explicit_directed_branches : Mdl_basics_explicit_ast.explicit_branches;
  }
;;

(** {6 Defining explicit lines} *)

(** {7 Defining explicit line body} *)

type ('connectors, 'points) explicit_line_body =
  {
    explicit_line_name : line_name option;
    explicit_line_connectors : 'connectors;
    explicit_line_font : explicit_font;
    explicit_line_labels : labels option;
    explicit_line_points : 'points option;
  }
(**
  For directed lines, ['points] is a list of displacements (relative
  positions).
  For connection lines, ['points] should be more sophisticated since in
  some case the first point of the list is an absolute point that specifies
  the location of the connection split to be inserted in the line; other
  points are all relative positions.
*)
;;

(** {7 Defining explicit line variant} *)
type explicit_lines = explicit_line list
and explicit_line =
  | Explicit_line_connection of
      (explicit_connection_connectors, explicit_connection_points)
        explicit_line_body
  | Explicit_line_directed of
      (explicit_directed_connectors, explicit_directed_points)
        explicit_line_body
;;

(** {6 Explicit line model} *)
type explicit_line_model = explicit_line explicit_model
(**
  An [explicit_model] is an explicit model featuring explicit lines.
*)
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
