(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_implementation_to_explicit_ast.ml,v 1.1 2019-05-10 14:04:56 jpc Exp $ *)

(**
  From deep syntax to completely explicit inherited tree:
  - defaults sections are incorporated into values,
  - no more default sections in the resulting trees: all fields are explicit
    and checked with respect to various type or range constraints.
*)

open Ast_types;;
open Mdl_deep_ast;;
open Mdl_basics_explicit_ast;;
open Mdl_basics_to_explicit_ast;;
open Mdl_interface_explicit_ast;;
open Mdl_implementation_explicit_ast;;
open Mdl_stateflow_deep_ast;;

(* Borrowed from deep phase *)
type deep_font = Mdl_basics_deep_ast.deep_font;;

let current_phase = "explicit phase";;

let explicit_message s =
  Format.sprintf "@[<hv 2>during %s,@ %s@]" current_phase s
;;

let explicit_error s = Say.fatal_error (explicit_message s)
and explicit_warning s = Say.warning (explicit_message s)
and explicit_debug s = Say.debug (explicit_message s)
;;

let duplicated_block_specification_error s =
  explicit_error (Printf.sprintf "Duplicated block %s specification" s)
;;

let duplicated_error s =
  explicit_error (Printf.sprintf "Duplicated %s specification" s)

let explicit_deep_zorder_dropped context = function
  | None -> ()
  | Some z ->
    explicit_debug (Format.sprintf
      "zorder option (%d) for %s was dropped" z context)
;;

let explicit_sf_block_type sf_block_type = sf_block_type;;

let explicit_orientation = function
  | Mdl_basics_deep_ast.
    Orientation_right -> false, Unatural._0_
  | Mdl_basics_deep_ast.
    Orientation_left -> true, Unatural._0_
  | Mdl_basics_deep_ast.
    Orientation_up -> false, Unatural_types.to_natural 270
  | Mdl_basics_deep_ast.
    Orientation_down -> true, Unatural_types.to_natural 270
;;

module Full_deep_font :
  sig

    type t = private
      | Full_deep_font of deep_font;;

    val of_deep_font : deep_font -> t;;

  end =

  struct

    type t =
      | Full_deep_font of deep_font
    ;;

    (* Build of full deep font out of a deep font that is indeed full,
       i.e. there is no None option in this deep font. *)
    let of_deep_font = function
      | {
          Mdl_basics_deep_ast.
          deep_font_name = Some _;
          deep_font_angle = Some _;
          deep_font_size = Some _;
          deep_font_weight = Some _;
        } as deep_font -> Full_deep_font deep_font
      | _ ->
        explicit_error
          "Full_deep_font.of_deep_font: deep font argument is not full"
    ;;

  end
;;

(* To gain pattern marching unqualified access to private
   constructor [Full_deep_font.Full_deep_font]. *)
open Full_deep_font;;

(* Building the full deep font corresponding to
   the configuration default font. *)
let configuration_default_full_deep_font =
  let deep_font =
    Mdl_ast_to_deep_ast.deep_font (Configuration.get_default_font ()) in
   Full_deep_font.of_deep_font deep_font
;;

(* Given a (MDL specified) deep font that may have missing fields
   we inherit those fields from a full default deep font. *)
let full_deep_font_of_deep_font (Full_deep_font default_font) deep_font =

  let deep_font_name =
    match deep_font.Mdl_basics_deep_ast.deep_font_name with
    | None ->
      default_font.Mdl_basics_deep_ast.deep_font_name
    | Some _ as option -> option

  and deep_font_angle =
    match deep_font.Mdl_basics_deep_ast.deep_font_angle with
    | None ->
      default_font.Mdl_basics_deep_ast.deep_font_angle
    | Some _ as option -> option

  and deep_font_size =
    match deep_font.Mdl_basics_deep_ast.deep_font_size with
    | None ->
      default_font.Mdl_basics_deep_ast.deep_font_size
    | Some _ as option -> option

  and deep_font_weight =
    match deep_font.Mdl_basics_deep_ast.deep_font_weight with
    | None ->
      default_font.Mdl_basics_deep_ast.deep_font_weight
    | Some _ as option -> option in

   Full_deep_font.of_deep_font
   {
      Mdl_basics_deep_ast.
      deep_font_name = deep_font_name;
      deep_font_angle = deep_font_angle;
      deep_font_size = deep_font_size;
      deep_font_weight = deep_font_weight;
    }
;;

(* Given a default deep font (that may still lack some field(s),
  turn it to a [full_deep_font] by inheriting from the
  configuration default full deep font ([configuration_full_deep_font]). *)
let full_deep_font_of_default_deep_font =
  full_deep_font_of_deep_font configuration_default_full_deep_font
;;

(* Given a deep font option and a default full deep font, return
   a [full_deep_font] *)
let full_deep_font_of_deep_font_option default_full_deep_font = function
  | None -> default_full_deep_font
  | Some deep_font ->
    full_deep_font_of_deep_font default_full_deep_font deep_font
;;

(* Return a [full_deep_font] from a default deep font option:
  fill missing fields with the configuration default full deep font
  ([configuration_default_full_deep_font]).
*)
let full_deep_font_of_default_deep_font_option =
  full_deep_font_of_deep_font_option configuration_default_full_deep_font
;;

(* Build an explicit font out of a full deep font. *)
let explicit_font_of_full_deep_font = function
  | Full_deep_font
    {
      Mdl_basics_deep_ast.
      deep_font_name = Some fn;
      deep_font_angle = Some fa;
      deep_font_size = Some fs;
      deep_font_weight = Some fw;
    } ->
    {
      explicit_font_name = fn;
      explicit_font_angle = fa;
      explicit_font_size = fs;
      explicit_font_weight = fw;
    }
  | Full_deep_font _ -> assert false
;;

let explicit_font_of_deep_font default_full_deep_font deep_font =
  let full_deep_font =
    full_deep_font_of_deep_font
      default_full_deep_font deep_font in
  explicit_font_of_full_deep_font full_deep_font
;;

let explicit_font_of_deep_font_option default_full_deep_font deep_font_opt =
  let full_deep_font =
    full_deep_font_of_deep_font_option
      default_full_deep_font deep_font_opt in
  explicit_font_of_full_deep_font full_deep_font
;;

(*
  The properties that a block must have.

  Those properties are given in the block description as a list of
  block items or even simple bindings. We must sort them out to find the set
  of real bindings of the block (afterwards we sort this list of bindings to
  separate block parameters from block ordinary bindings).

  The type [intrinsic_properties] is not exported outside this implementation
  file.
*)
type 'a intrinsic_properties = {
  ip_background_color : explicit_color option;
  ip_drop_shadow : bool option;
  ip_foreground_color : explicit_color option;
  ip_font : 'a option;
  ip_mirror : explicit_mirror option;
  ip_name_placement : explicit_placement option;
  ip_orientation : explicit_orientation option;
  ip_ports : explicit_block_ports option;
  ip_ports_description : explicit_ports_description option;
  ip_position : explicit_position option;
  ip_rotation : explicit_rotation option;
  ip_show_name : bool option;
  ip_hide_automatic_name : bool option;
  ip_sid : explicit_sid option;
  ip_state : explicit_state option;
  ip_tag : explicit_tag option;
  ip_zorder : explicit_zorder option;
  ip_instance_data : explicit_binding_table option;
}

and explicit_orientation = explicit_mirror * explicit_rotation
;;

let binding_table_of_list bds =
  let tbl = Hashtbl.create 73 in
  let enter_binding bd =
    let { Binding.bound_ident = id; bound_value = _; } = bd in
    if Hashtbl.mem tbl id then
      explicit_warning
        (Printf.sprintf
           "binding_table_of_list: duplicated binding for identifier %s"
           (string_of_ident id));
    Hashtbl.add tbl id bd in

  List.iter enter_binding bds;
  tbl
;;

(**

  [separate_deep_block_properties] sorts out deep block items of a block
  to compute the record of intrinsic properties for this block; default
  values for intrinsic properties are given in the list
  [deep_block_properties] (a deep binding list).

  The rest of deep bindings found in the block items which are non intrinsic
  bindings flagged as [Deep_block_binding] elements in the list
  [deep_block_properties]) are returned as a list of bindings.

  Also takes care of special case of orientation property which may
  be given with a simple binding with special bound ident ``Orientation''.

*)
let separate_deep_block_properties deep_block_items =

  let mark_ip, is_ip_marked =
    let ids = Hash_table.create (Configuration.get_mean_block_bindings ()) in
    (fun id -> Hashtbl.add ids id ()), Hashtbl.mem ids in

  let rec loop got bds = function
    | [] -> got, bds
    | deep_block_item :: deep_block_items ->
      match deep_block_item with
      | Mdl_basics_deep_ast.
        Deep_block_position deep_position ->
        if got.ip_position = None then
        loop {
          got with
          ip_position = Some deep_position;
        } bds deep_block_items else
        duplicated_block_specification_error "position"

      | Mdl_basics_deep_ast.
        Deep_block_sid sid ->
        if got.ip_sid = None then
        loop {
          got with
          ip_sid = Some sid;
        } bds deep_block_items else
        duplicated_block_specification_error "sid"

      | Mdl_basics_deep_ast.
        Deep_block_state deep_value ->
        if got.ip_state = None then
        loop {
          got with
          ip_state = Some deep_value;
        } bds deep_block_items else
        duplicated_block_specification_error "state"

      | Mdl_basics_deep_ast.
        Deep_block_tag deep_value ->
        if got.ip_tag = None then
        loop {
          got with
          ip_tag = Some deep_value;
        } bds deep_block_items else
        duplicated_block_specification_error "tag"

      | Mdl_basics_deep_ast.
        Deep_block_binding ({
          Binding.bound_ident = id;
          _;
        } as deep_binding) ->
        if is_ip_marked id then
          duplicated_block_specification_error (Printf.sprintf
            "binding (%S)"
            (Ast_types.string_of_ident id)) else
          mark_ip id;
        loop got (deep_binding :: bds) deep_block_items

      | Mdl_basics_deep_ast.
        Deep_block_graphical_default deep_block_graphical_default ->

        (match deep_block_graphical_default with
         | Mdl_basics_deep_ast.
           Deep_block_mirror bool ->
           if got.ip_mirror = None then
           loop {
             got with
             ip_mirror = Some bool;
           } bds deep_block_items else
           duplicated_block_specification_error "mirror"
         | Mdl_basics_deep_ast.
           Deep_block_name_placement deep_placement ->
           if got.ip_name_placement = None then
           loop {
             got with
             ip_name_placement = Some deep_placement;
           } bds deep_block_items else
           duplicated_block_specification_error "name placement"
         | Mdl_basics_deep_ast.
           Deep_block_orientation deep_orientation ->
           if got.ip_orientation = None then
           loop {
             got with
             ip_orientation =
               Some (explicit_orientation deep_orientation);
           } bds deep_block_items else
           duplicated_block_specification_error "orientation"
         | Mdl_basics_deep_ast.
           Deep_block_rotation natural ->
           if got.ip_rotation = None then
           loop {
             got with
             ip_rotation = Some natural;
           } bds deep_block_items else
           duplicated_block_specification_error "rotation"
         | Mdl_basics_deep_ast.
           Deep_block_show_name b ->
           if got.ip_show_name = None then
           loop {
             got with
             ip_show_name = Some b;
           } bds deep_block_items else
           duplicated_block_specification_error "show name"
         | Mdl_basics_deep_ast.
           Deep_block_hide_automatic_name b ->
           if got.ip_hide_automatic_name = None then
           loop {
             got with
             ip_hide_automatic_name = Some b;
           } bds deep_block_items else
           duplicated_block_specification_error "hide automatic name"
         | Mdl_basics_deep_ast.
           Deep_block_font_and_color deep_font_and_color ->
           (match deep_font_and_color with
            | Mdl_basics_deep_ast.
              Deep_background deep_color ->
              if got.ip_background_color = None then
              loop {
                got with
                ip_background_color = Some deep_color;
              } bds deep_block_items else
              duplicated_block_specification_error "background color"
            | Mdl_basics_deep_ast.
              Deep_drop_shadow b ->
              if got.ip_drop_shadow = None then
              loop {
                got with
                ip_drop_shadow = Some b;
              } bds deep_block_items else
              duplicated_block_specification_error "drop shadow"
            | Mdl_basics_deep_ast.
              Deep_foreground deep_color ->
              if got.ip_foreground_color = None then
              loop {
                got with
                ip_foreground_color = Some deep_color;
              } bds deep_block_items else
              duplicated_block_specification_error "foreground color"
            | Mdl_basics_deep_ast.
              Deep_font deep_font ->
              if got.ip_font = None then
              loop {
                got with
                ip_font = Some deep_font;
              } bds deep_block_items else
              duplicated_block_specification_error "font"
           )
         | Mdl_basics_deep_ast.
           Deep_block_zorder z ->
           if got.ip_zorder = None then
           loop {
             got with
             ip_zorder = Some z;
           } bds deep_block_items else
           duplicated_block_specification_error "zorder"
         | Mdl_basics_deep_ast.
           Deep_block_ports deep_port_bodies ->
           if got.ip_ports = None then
           loop {
             got with
             ip_ports = Some deep_port_bodies;
           } bds deep_block_items else
           duplicated_block_specification_error "ports"
         | Mdl_basics_deep_ast.
           Deep_block_ports_description ports_desc ->
           if got.ip_ports_description = None then
           loop {
             got with
             ip_ports_description = Some ports_desc;
           } bds deep_block_items else
           duplicated_block_specification_error "ports description"
        ) in

  loop {
    ip_background_color = None;
    ip_foreground_color = None;
    ip_drop_shadow = None;
    ip_name_placement = None;
    ip_show_name = None;
    ip_hide_automatic_name = None;
    ip_position = None;
    ip_rotation = None;
    ip_sid = None;
    ip_mirror = None;
    ip_orientation = None;
    ip_ports = None;
    ip_ports_description = None;
    ip_font = None;
    ip_state = None;
    ip_tag = None;
    ip_zorder = None;
    ip_instance_data = None;
  } [] deep_block_items
(**

  Separate deep block items of a block into items that are known to be
  block intrinsic specifications (all those items are returned as
  explicit fields of an [intrinsic_properties] record) from other items
  that may be parameters or ordinary bindings (all those bindings are
  returned in a list of bindings). Also verify uniqueness of intrinsic
  specifications.

*)
;;

let explicit_block_type deep_block =
  match deep_block.Mdl_deep_ast.deep_block_desc with
  | Deep_block_reference deep_block_reference_type ->
    Explicit_block_reference_type deep_block_reference_type
  | Deep_block_regular deep_block_regular_type ->
    Explicit_block_regular_type deep_block_regular_type
  | Deep_block_sub_system _ ->
    Explicit_block_regular_type Mdl_block_types.SubSystem
  | Deep_block_s_function _ ->
    Explicit_block_regular_type Mdl_block_types.S_Function
;;

let inherit_deep_bindings
    deep_model_defaults deep_block_type deep_bindings =

  let model_block_parameter_defaults =
    deep_model_defaults.
    Mdl_basics_deep_ast.deep_model_block_parameter_defaults in
  let block_parameter_defaults =
    try
      List.find
        (fun bpd ->
         bpd.Mdl_basics_deep_ast.deep_parameter_block_type = deep_block_type)
        model_block_parameter_defaults
    with
    | Not_found ->
      explicit_warning (
        Printf.sprintf "defaults for block type %s was not found in model"
          (Mdl_block_types_tables.string_of_block_type deep_block_type));
      Mdl_ast_to_deep_ast.make_default_deep_block_parameter_defaults
        deep_block_type in

  let deep_parameter_defaults_table =
    binding_table_of_list
      block_parameter_defaults.Mdl_basics_deep_ast.deep_parameter_defaults in

  (*
  Say.fdebug (fun ppf ->
    Format.fprintf ppf "\
      @[<v 2>@ \
      Deep_defaults_table:@ \
      %a@ \
      @]";
      (fun ppf ->
       Hashtbl.iter
         (fun id _bd ->
          Printf.fprintf ppf
            "%s found in deep_parameter_defaults_table."
            id))
      deep_parameter_defaults_table);
  *)
  let deep_bindings_table =
    binding_table_of_list deep_bindings in

  (*
  Say.fdebug (fun ppf ->
    Format.fprintf ppf "\
      @[<v 2>@ \
      Deep_bindings_table:@ \
      %a@ \
      @]";
      (fun ppf ->
       Hashtbl.iter
         (fun id _bd ->
          Printf.fprintf ppf
            "%s found in deep_bindings_table."
            id))
      deep_bindings_table);
  *)
  Hashtbl.iter
    (fun id bd ->
     if not (Hashtbl.mem deep_bindings_table id)
     then begin
       (*
       Say.fdebug (fun ppf ->
         Format.fprintf ppf "\
           %s added in deep_bindings_table."
           id);
       *)
       Hashtbl.add deep_bindings_table id bd
     end)
    deep_parameter_defaults_table;

  (*
  Say.fdebug (fun ppf ->
    Format.fprintf ppf "\
      @[<v 2>@ \
      AFTER inheritance, deep_bindings_table is:@ \
      %a@ \
      @]";
      (fun ppf ->
       Hashtbl.iter
         (fun id _bd ->
          Printf.fprintf ppf
            "%s found in deep_bindings_table."
            id))
      deep_bindings_table);
  *)

  deep_bindings_table
;;

let explicit_block_type_unknown =
  let bl_src_type =
    Ast_types.mdl_name_of_string "SimportUnknown" in
  let bl_src_path =
    Ast_types.mdl_path_of_string
      "simport/genr_block/genr_not_yet_translated" in
  Explicit_block_reference_type
    (Mdl_deep_basics.make_deep_block_reference_type bl_src_type bl_src_path)
;;

let find_explicit_interface_block
    (explicit_interfaces :
       Mdl_interface_explicit_ast.explicit_interfaces)
    explicit_block_type =

  let find_block_in_interface intf block_type =
    let tbl = intf.ei_table in
    if Hashtbl.mem tbl block_type then
      Hashtbl.find tbl block_type else raise Not_found in

  let rec loop block_type = function
  | [] -> raise Not_found
  | intf :: intfs ->
    try find_block_in_interface intf block_type with
    | Not_found -> loop block_type intfs in

  try loop explicit_block_type explicit_interfaces with
  | Not_found ->
    let ty_name =
      Mdl_basics_to_explicit_ast.
      name_of_explicit_block_type explicit_block_type in
    explicit_warning
      (Format.sprintf
         "@[<v 0>\
          the type specification for block %S was not found,@,\
          hence the block was replaced by an empty super block;@,\
          consider adding a -I option to access %S block interface.\
          @]"
          ty_name ty_name);
    loop explicit_block_type_unknown explicit_interfaces
;;

let mem_explicit_interface_block explicit_interfaces explicit_block_type =

  let mem_block_in_interface intf block_type =
    let tbl = intf.ei_table in
    Hashtbl.mem tbl block_type in

  let rec loop block_type = function
  | [] -> false
  | intf :: intfs ->
    mem_block_in_interface intf block_type || loop block_type intfs in

  loop explicit_block_type explicit_interfaces
;;

let deep_block_type_of_deep_block deep_block =
  match deep_block.deep_block_desc with
  | Deep_block_reference _deep_block_reference_type ->
    Mdl_block_types.Reference
  | Deep_block_regular deep_block_regular_type -> deep_block_regular_type
  | Deep_block_sub_system _deep_system -> Mdl_block_types.SubSystem
  | Deep_block_s_function _ -> Mdl_block_types.S_Function
;;

let separate_deep_bindings
    (explicit_interfaces :
       Mdl_interface_explicit_ast.explicit_interfaces)
    deep_model_defaults
    deep_block deep_bindings =

  let explicit_block_type = explicit_block_type deep_block in
  let deep_block_type = deep_block_type_of_deep_block deep_block in

  let inherited_deep_bindings_table =
    (* No inheritance for references, just use the block bindings as table. *)
    if deep_block_type = Mdl_block_types.Reference then
      binding_table_of_list deep_bindings
    else
    (* The true thing: inherit from the model defaults. *)
      inherit_deep_bindings deep_model_defaults
        deep_block_type deep_bindings in

  (* Find the block interface as declared in the list of (library) interfaces
  found on the command line (for instance simport_pervasives.mdli). *)
  let explicit_interface_block =
    find_explicit_interface_block explicit_interfaces explicit_block_type in
  (* Then find the declared parameters of this block. *)
  let ebi_parameters_table = explicit_interface_block.ebi_parameters in

  (* Separate all bindings for the block into parameters or bindings. *)
  let block_parameters, block_bindings =
    Hashtbl.fold
      (fun id bd (bp, bb) ->
       (*let (Ast_types.Mdl_name block_name) =
           deep_block.deep_block_name in*)
       if Hashtbl.mem ebi_parameters_table id
       then
         begin
           (* prerr_endline (Printf.sprintf "%s is a parameter for block %s"
             (Ast_types.string_of_ident id) block_name); *)
           (bd :: bp, bb)
         end
       else
         begin
           (* prerr_endline (Printf.sprintf "%s is a binding for block %s"
             (Ast_types.string_of_ident id) block_name); *)
           (bp, bd :: bb)
         end)
      inherited_deep_bindings_table
      ([], []) in

  let deep_block_parameters_table =
    binding_table_of_list block_parameters
  and deep_bindings_table =
    binding_table_of_list block_bindings in

  deep_block_parameters_table, deep_bindings_table
;;

let inherit_intrinsic_properties deep_model_defaults
    (ip as _intrinsic_properties) =

  let deep_block_graphical_defaults =
    deep_model_defaults.
      Mdl_basics_deep_ast.deep_model_block_graphical_defaults in
  let defaults_intrinsic_properties, _ =
    separate_deep_block_properties
      (List.map
         (fun d -> Mdl_basics_deep_ast.Deep_block_graphical_default d)
         deep_block_graphical_defaults) in
  {
    ip_background_color =
      (let opt = ip.ip_background_color in
       if opt = None
       then defaults_intrinsic_properties.ip_background_color
       else opt);
    ip_drop_shadow =
      (let opt = ip.ip_drop_shadow in
       if opt = None
       then defaults_intrinsic_properties.ip_drop_shadow
       else opt);
    ip_foreground_color =
      (let opt = ip.ip_foreground_color in
       if opt = None
       then defaults_intrinsic_properties.ip_foreground_color
       else opt);
    ip_font =
      (let default_full_deep_font =
         let ip_default_font_opt = defaults_intrinsic_properties.ip_font in
         full_deep_font_of_default_deep_font_option ip_default_font_opt in
       let font_opt = ip.ip_font in
       let full_deep_font =
         full_deep_font_of_deep_font_option default_full_deep_font font_opt in
       Some full_deep_font);
    ip_mirror =
      (let opt = ip.ip_mirror in
       if opt = None
       then defaults_intrinsic_properties.ip_mirror
       else opt);
    ip_name_placement =
      (let opt = ip.ip_name_placement in
       if opt = None
       then defaults_intrinsic_properties.ip_name_placement
       else opt);
    ip_orientation =
      (let opt = ip.ip_orientation in
       if opt = None
       then defaults_intrinsic_properties.ip_orientation
       else opt);
    ip_ports =
      (let opt = ip.ip_ports in
       if opt = None
       then defaults_intrinsic_properties.ip_ports
       else opt);
    ip_ports_description =
      (let opt = ip.ip_ports_description in
       if opt = None
       then defaults_intrinsic_properties.ip_ports_description
       else opt);
    ip_position =
      (let opt = ip.ip_position in
       if opt = None
       then defaults_intrinsic_properties.ip_position
       else opt);
    ip_rotation =
      (let opt = ip.ip_rotation in
       if opt = None
       then defaults_intrinsic_properties.ip_rotation
       else opt);
    ip_show_name =
      (let opt = ip.ip_show_name in
       if opt = None
       then defaults_intrinsic_properties.ip_show_name
       else opt);
    ip_hide_automatic_name =
      (let opt = ip.ip_hide_automatic_name in
       if opt = None
       then defaults_intrinsic_properties.ip_hide_automatic_name
       else opt);
    ip_sid =
      (let opt = ip.ip_sid in
       if opt = None
       then defaults_intrinsic_properties.ip_sid
       else opt);
    ip_state =
      (let opt = ip.ip_state in
       if opt = None
       then defaults_intrinsic_properties.ip_state
       else opt);
    ip_tag =
      (let opt = ip.ip_tag in
       if opt = None
       then defaults_intrinsic_properties.ip_tag
       else opt);
   ip_zorder =
      (let opt = ip.ip_zorder in
       if opt = None
       then defaults_intrinsic_properties.ip_zorder
       else opt);
   ip_instance_data =
      (let opt = ip.ip_instance_data in
       if opt = None
       then defaults_intrinsic_properties.ip_instance_data
       else opt);
  }
;;

(* Computes the definitions of a block, as specified in the block definition.
   No inheritance has been applied yet. *)
let create_explicit_block_all_bindings
    (explicit_interfaces :
       Mdl_interface_explicit_ast.explicit_interfaces)
    deep_model_defaults
    deep_block
    deep_block_properties =

  let explicit_block_type = explicit_block_type deep_block in
  let (Ast_types.Mdl_name block_name) = deep_block.deep_block_name in

  let correct_intrinsic_properties intrinsic_properties =
    match intrinsic_properties.ip_mirror,
          intrinsic_properties.ip_rotation,
          intrinsic_properties.ip_orientation with
    | Some _, Some _, None -> intrinsic_properties
    | None, None, Some (mirror, rotation) ->
      {
        intrinsic_properties with
          ip_mirror = Some mirror;
          ip_rotation = Some rotation;
          ip_orientation = None;
      }
    | None, None, None -> intrinsic_properties
    | _, _, _ ->
      explicit_error
        (Printf.sprintf
           "Incorrectly defined orientation of block %S (with block type %S)."
           block_name
           (name_of_explicit_block_type explicit_block_type)) in

  (* Properties as defined in the block *)
  let intrinsic_properties, deep_bindings =
    separate_deep_block_properties deep_block_properties in

  let inherited_intrinsic_properties =
    inherit_intrinsic_properties deep_model_defaults intrinsic_properties in

  let inherited_intrinsic_properties =
    correct_intrinsic_properties inherited_intrinsic_properties in

  let explicit_block_parameters_table, explicit_block_bindings =
    separate_deep_bindings
      explicit_interfaces deep_model_defaults
      deep_block deep_bindings in

  let explicit_block_graphical, explicit_block_optional_graphical =
    match inherited_intrinsic_properties with
    | {
        ip_background_color = Some bg;
        ip_foreground_color = Some fg;
        ip_drop_shadow = ds;
        ip_name_placement = np;
        ip_show_name = Some sn;
        ip_hide_automatic_name = han;
        ip_position = Some pos;
        ip_rotation = Some rot;
        ip_mirror = Some mir;
        ip_sid = _sid;
        ip_orientation = _ori;
        ip_ports = (ports : Mdl_basics_deep_ast.deep_block_ports option);
        ip_font = Some full_deep_font;
        ip_state = state;
        ip_tag = tag;
        ip_zorder = zorder_opt;
        ip_ports_description = ports_desc;
        ip_instance_data = _instance_data;
      } ->
      explicit_deep_zorder_dropped "block" zorder_opt;

      let explicit_ports_description =
        match ports_desc with
        | None ->
          let explicit_interface_block =
            find_explicit_interface_block
              explicit_interfaces explicit_block_type in
          explicit_interface_block.ebi_ports_description
        | Some ports_desc -> ports_desc in
      let ports =
        match ports with
        | None -> []
        | Some ports -> ports in

      {
        explicit_block_background_color = bg;
        explicit_block_foreground_color = fg;
        explicit_block_mirror = mir;
        explicit_block_ports = ports;
        explicit_block_ports_description = explicit_ports_description;
        explicit_block_position = pos;
        explicit_block_rotation = rot;
        explicit_block_show_name = sn;
      },
      {
        explicit_block_font =
          explicit_font_of_full_deep_font full_deep_font;
        explicit_block_state = state;
        explicit_block_tag = tag;
        explicit_block_drop_shadow = ds;
        explicit_block_name_placement = np;
        explicit_block_hide_automatic_name = han;
(*
        explicit_block_sid = sid;
*)
      }
    | {
        ip_background_color = bg;
        ip_foreground_color = fg;
        ip_drop_shadow = _ds;
        ip_name_placement = _np;
        ip_show_name = sn;
        ip_hide_automatic_name = _han;
        ip_position = pos;
        ip_rotation = rot;
        ip_mirror = mir;
        ip_sid = _sid;
        ip_orientation = _ori;
        ip_ports = ports;
        ip_font = font;
        ip_state = _state;
        ip_tag = _tag;
        ip_zorder = _z;
        ip_ports_description = ports_desc;
        ip_instance_data = _instance_data;
      } ->
      let s = if bg = None then "bg " else "" in
      let s = if fg = None then s ^ "fg " else s in
      let s = if sn = None then s ^ "sn " else s in
(*
      let s = if han = None then s ^ "han " else s in
*)
      let s = if pos = None then s ^ "pos " else s in
      let s = if rot = None then s ^ "rot " else s in
      let s = if mir = None then s ^ "mir " else s in
      let s = if ports = None then s ^ "ports " else s in
      let s = if ports_desc = None then s ^ "ports description " else s in
      let s = if font = None then s ^ "font " else s in
      prerr_endline (
        Printf.sprintf "In file %S, block %s, property \'%s\' is missing"
          (Configuration.get_source_file ())
          block_name
          s);
      assert false in

  {
    explicit_block_parameters = explicit_block_parameters_table;
    explicit_block_bindings = explicit_block_bindings;
    explicit_block_graphical = explicit_block_graphical;
    explicit_block_optional_graphical = explicit_block_optional_graphical;
  }
;;

let explicit_array array_description = array_description;;

let explicit_object object_description = object_description;;

let explicit_line_body
    default_full_deep_font explicit_line_connectors deep_line_body =

  match deep_line_body with
  | {
      Mdl_basics_deep_ast.
      deep_line_font = deep_font_opt;
      deep_line_zorder = deep_zorder_opt;
      deep_line_source = _;
      deep_line_target = _;
      deep_line_branches = _;
      deep_line_name = deep_line_name;
      deep_line_labels = deep_line_labels;
      deep_line_points = deep_line_points;
    } ->
    explicit_deep_zorder_dropped "line" deep_zorder_opt;
    let explicit_font =
      explicit_font_of_deep_font_option
        default_full_deep_font deep_font_opt in
    {
      explicit_line_name = deep_line_name;
      explicit_line_font = explicit_font;
      explicit_line_connectors = explicit_line_connectors;
      explicit_line_labels = deep_line_labels;
      explicit_line_points = deep_line_points;
    }
;;

let make_explicit_connector_port block_name port_type =
  {
    explicit_connector_port_block = block_name;
    explicit_connector_port_desc = port_type;
  }
;;

let explicit_connector_port_of_deep_port = function
  | {
      Mdl_ast.
      line_connector_block = block_name;
      line_connector_port = port_value;
    } ->
    make_explicit_connector_port block_name port_value
;;

let explicit_line_directed_body default_full_deep_font accu = function
  | {
      Mdl_basics_deep_ast.
      deep_line_source = None; _
    } ->
    explicit_warning "Line directed without source was dropped";
    accu
  | {
      Mdl_basics_deep_ast.
      deep_line_target = None;
      deep_line_branches = [];
      _
    } ->
    explicit_warning
      "Line directed without target or branches was dropped";
    accu
  | ({
      Mdl_basics_deep_ast.
      deep_line_source = Some csrc;
      deep_line_target = target;
      deep_line_branches = branches;
      _
    } as deep_line_body) ->
    let eline_body =
      let explicit_directed_source =
        explicit_connector_port_of_deep_port csrc in
      let explicit_directed_target =
        match target with
        | None -> None
        | Some dst -> Some (explicit_connector_port_of_deep_port dst) in
      let directed_connectors =
        {
          Mdl_implementation_explicit_ast.
          explicit_directed_source = explicit_directed_source;
          explicit_directed_target = explicit_directed_target;
          explicit_directed_branches = branches;
        } in
      explicit_line_body
        default_full_deep_font directed_connectors deep_line_body in
    Explicit_line_directed eline_body :: accu
;;

(* Expliciting connection lines *)

let partition_line_connection_branches branches =
  let part (source_branches, target_branches) = function
    | Mdl_basics_deep_ast.
      Deep_branch_connection (deep_connection_type, deep_branch_body) ->
      let (_deep_connection_side, _deep_branch_body as branch) =
        deep_connection_type.
          Mdl_basics_deep_ast.
          deep_connection_type_self,
        deep_branch_body in
      begin
        match
           deep_connection_type.
             Mdl_basics_deep_ast.deep_connection_type_parent with
        | Mdl_basics_deep_ast.
          Deep_side_src -> branch :: source_branches, target_branches
        | Mdl_basics_deep_ast.
          Deep_side_dst -> source_branches, branch :: target_branches
      end
    | Mdl_basics_deep_ast.
      Deep_branch_directed _ ->
      explicit_error "Directed branch in connection line" in
  List.fold_left part ([], []) branches
;;

let make_explicit_line_connection_connector_port deep_connector =
  match deep_connector with
  | {
      Mdl_ast.
      line_connector_block = block_name;
      line_connector_port = connector_port;
    } ->
    let explicit_connection_connector_desc =
      match connector_port with
      | Mdl_ast.Port_trigger
      | Mdl_ast.Port_enable
      | Mdl_ast.Port_state
      | Mdl_ast.Port_ifaction
      | Mdl_ast.Port_number _ ->
        explicit_error "invalid connection line connector port"
      | Mdl_ast.Port_lconn port_number ->
        Explicit_left_connector (Connection_port_value port_number)
      | Mdl_ast.Port_rconn port_number ->
        Explicit_right_connector (Connection_port_value port_number) in

    {
      explicit_connector_port_block = block_name;
      explicit_connector_port_desc = explicit_connection_connector_desc;
    }
;;

let explicit_absolute_points_of_deep_points deep_points =
  match deep_points with
  | [] -> explicit_error "Empty list of points in connection line"
  | deep_point :: deep_points ->
    let explicit_absolute_point =
      match deep_point with
      | {
          Mdl_basics_deep_ast.
          deep_x = x;
          deep_y = y;
        } ->
        {
          explicit_x = x;
          explicit_y = y;
        } in
    let explicit_relative_points = deep_points in
    Explicit_absolute_points
      (explicit_absolute_point, explicit_relative_points)
;;

let explicit_absolute_connection_points_option deep_points_option =
  match deep_points_option with
  | None ->
    explicit_error "Missing absolute position of branch"
  | Some deep_points ->
    Some (explicit_absolute_points_of_deep_points deep_points)
;;

let explicit_relative_connection_points_option deep_points_option =
  match deep_points_option with
  | None -> None
  | Some deep_points -> Some (Explicit_relative_points deep_points)
;;

let make_explicit_connection_points_option
    explicit_connection_connector deep_points_option =
  match explicit_connection_connector with
  | Explicit_connection_connector_port _ ->
    explicit_relative_connection_points_option deep_points_option
  | Explicit_connection_connector_branches _ ->
    explicit_absolute_connection_points_option deep_points_option
;;

let rec make_explicit_line_connection_connector
    deep_connector_port_option
    deep_line_branches =
  match deep_connector_port_option with
  | None ->
    begin
      match deep_line_branches with
      | [] -> explicit_warning "Dangling connection line"
      | _ -> ()
    end;
    let explicit_branches =
      List.map
        make_explicit_line_connection_branch deep_line_branches in
    Explicit_connection_connector_branches explicit_branches
  | Some deep_connector_port ->
    begin
      match deep_line_branches with
      | [] ->
        let explicit_connector_port =
          make_explicit_line_connection_connector_port deep_connector_port in
        Explicit_connection_connector_port explicit_connector_port
      | _branch :: _branches ->
        explicit_error
          "Connector port cannot have branches in connection line"
    end

and make_explicit_line_connection_branch
    (deep_connection_side, deep_branch_body) =
  let check_no_child message deep_line_connector_option branches =
    match deep_line_connector_option, branches with
    | None, [] -> ()
    | _, _ ->
      explicit_error (Printf.sprintf
        "a generation skipping connection is not allowed:\n\
         a connection line branch with %s side cannot connect \
         to parent node and child node"
        message) in

  match deep_branch_body with
  | {
      Mdl_basics_deep_ast.
      deep_branch_source = deep_line_connector_source_option;
      deep_branch_target = deep_line_connector_target_option;
      deep_branch_branches = deep_branches;
      deep_branch_labels = _deep_labels_option;
      deep_branch_points = deep_points_option;
      deep_branch_zorder = deep_zorder_option;
    } ->

    let source_branches, target_branches =
      partition_line_connection_branches deep_branches in

    let explicit_connection_branch_child,
        explicit_connection_points_option =
      match deep_connection_side with
      | Mdl_basics_deep_ast.Deep_side_src ->
        check_no_child "source"
          deep_line_connector_source_option source_branches;
        let explicit_line_connection_connector =
          make_explicit_line_connection_connector
            deep_line_connector_target_option target_branches in
        let explicit_connection_points_option =
          explicit_relative_connection_points_option deep_points_option in

        Explicit_connection_branch_target explicit_line_connection_connector,
        explicit_connection_points_option

      | Mdl_basics_deep_ast.Deep_side_dst ->
        check_no_child "target"
          deep_line_connector_target_option target_branches;
        let explicit_line_connection_connector =
          make_explicit_line_connection_connector
            deep_line_connector_source_option source_branches in

        let explicit_line_connection_points_option =
          make_explicit_connection_points_option
            explicit_line_connection_connector deep_points_option in
        Explicit_connection_branch_source explicit_line_connection_connector,
        explicit_line_connection_points_option in

    explicit_deep_zorder_dropped "connection line" deep_zorder_option;

    {
      explicit_connection_branch_points = explicit_connection_points_option;
      explicit_connection_branch_child = explicit_connection_branch_child;
    }
;;

let explicit_line_connection_body default_full_deep_font accu = function
  | {
      Mdl_basics_deep_ast.
      deep_line_name = deep_line_name_option;
      deep_line_font = deep_font_option;
      deep_line_source = deep_connector_src_option;
      deep_line_target = deep_connector_dst_option;
      deep_line_branches = deep_line_branches;
      deep_line_labels = deep_labels_option;
      deep_line_points = deep_points_option;
      deep_line_zorder = deep_zorder_option;
      _
    } as _deep_line_body ->

    let source_branches, target_branches =
      partition_line_connection_branches deep_line_branches in

    let explicit_source_connection_connector =
      make_explicit_line_connection_connector
        deep_connector_src_option source_branches in

    let explicit_target_connection_connector =
      make_explicit_line_connection_connector
        deep_connector_dst_option target_branches in

    let explicit_connection_connectors =
      {
        explicit_source_connection_connector =
          explicit_source_connection_connector;
        explicit_target_connection_connector =
          explicit_target_connection_connector;
      } in

    let explicit_font =
      explicit_font_of_deep_font_option
        default_full_deep_font deep_font_option in

    let explicit_connection_points_option =
      make_explicit_connection_points_option
        explicit_source_connection_connector deep_points_option in

    explicit_deep_zorder_dropped "connection line" deep_zorder_option;

    let eline_body =
      {
        Mdl_implementation_explicit_ast.
        explicit_line_name = deep_line_name_option;
        explicit_line_connectors = explicit_connection_connectors;
        explicit_line_font = explicit_font;
        explicit_line_labels = deep_labels_option;
        explicit_line_points = explicit_connection_points_option;
      } in

    Explicit_line_connection eline_body :: accu
;;

let explicit_lines deep_model_defaults deep_lines =

  let line_graphical_defaults =
    deep_model_defaults.
      Mdl_basics_deep_ast.deep_model_line_graphical_defaults in
  let default_full_deep_font =
    let default_deep_font =
      line_graphical_defaults.
      Mdl_basics_deep_ast.deep_line_default_font in
    full_deep_font_of_default_deep_font default_deep_font in
  let zorder_opt =
    line_graphical_defaults.
      Mdl_basics_deep_ast.deep_line_default_zorder in
  explicit_deep_zorder_dropped "line default" zorder_opt;

  let rec loop accu = function
    | [] -> accu
    | Mdl_basics_deep_ast.
      Deep_line_directed deep_directed_body :: deep_lines ->
      let accu =
        explicit_line_directed_body
          default_full_deep_font accu deep_directed_body in
      loop accu deep_lines
    | Mdl_basics_deep_ast.
      Deep_line_connection deep_connection_body :: deep_lines ->
      let accu =
        explicit_line_connection_body
          default_full_deep_font accu deep_connection_body in
      loop accu deep_lines in
  loop [] deep_lines
;;

let fields_of_model_annotation_graphical_defaults
    model_annotation_graphical_defaults =

  let rec loop font_opt fcs accu = function
    | [] -> font_opt, List.rev fcs, List.rev accu
    | mad :: mads ->
      match mad with
      | Mdl_ast.Horizontal_alignment _
      | Mdl_ast.Interpreter _
      | Mdl_ast.Text_as_click_cb _
      | Mdl_ast.Tex_mode _
      | Mdl_ast.Vertical_alignment _
      | Mdl_ast.Click_fcn _ -> loop font_opt fcs (mad :: accu) mads
      | Mdl_ast.Font_and_color fc ->
        begin match fc with
        | Mdl_basics_deep_ast.
          Deep_font dfn ->
          if font_opt = None then loop (Some dfn) fcs accu mads else
          duplicated_error "default font for annotations"
        | ( Mdl_basics_deep_ast.Deep_background _
        |   Mdl_basics_deep_ast.Deep_drop_shadow _
        |   Mdl_basics_deep_ast.Deep_foreground _ ) ->
          loop font_opt (fc :: fcs) accu mads
        end
      | Mdl_ast.Annotation_graphical_binding _binding ->
        loop font_opt fcs accu mads in

  loop None [] [] model_annotation_graphical_defaults
;;

let inherit_annotation_graphical_defaults
    dads deep_annotation_graphical_defaults =
  let same_dad dad1 dad2 =
    match dad1, dad2 with
    | Mdl_ast.Horizontal_alignment _, Mdl_ast.Horizontal_alignment _
    | Mdl_ast.Interpreter _, Mdl_ast.Interpreter _
    | Mdl_ast.Text_as_click_cb _, Mdl_ast.Text_as_click_cb _
    | Mdl_ast.Tex_mode _, Mdl_ast.Tex_mode _
    | Mdl_ast.Vertical_alignment _, Mdl_ast.Vertical_alignment _
    | Mdl_ast.Click_fcn _, Mdl_ast.Click_fcn _ -> true
    | Mdl_ast.Font_and_color _, Mdl_ast.Font_and_color _
    | Mdl_ast.Annotation_graphical_binding _,
      Mdl_ast.Annotation_graphical_binding _ -> assert false
    | (Mdl_ast.Horizontal_alignment _ |
       Mdl_ast.Interpreter _ |
       Mdl_ast.Text_as_click_cb _ | Mdl_ast.Tex_mode _ |
       Mdl_ast.Vertical_alignment _ | Mdl_ast.Font_and_color _ |
       Mdl_ast.Click_fcn _ |
       Mdl_ast.Annotation_graphical_binding _),
      (Mdl_ast.Horizontal_alignment _ |
       Mdl_ast.Interpreter _ |
       Mdl_ast.Text_as_click_cb _ | Mdl_ast.Tex_mode _ |
       Mdl_ast.Vertical_alignment _ | Mdl_ast.Font_and_color _ |
       Mdl_ast.Click_fcn _ |
       Mdl_ast.Annotation_graphical_binding _) -> false in

  let rec loop accu = function
    | [] -> accu
    | dad :: dads ->
      loop (dad :: List.filter (same_dad dad) accu) dads in

  loop dads deep_annotation_graphical_defaults
;;

let inherit_annotation_graphical mdfcs deep_font_and_colors =

  let same_dfc dfc1 dfc2 =
    match dfc1, dfc2 with
    | Mdl_basics_deep_ast.Deep_font _,
      Mdl_basics_deep_ast.Deep_font _
    | Mdl_basics_deep_ast.Deep_background _,
       Mdl_basics_deep_ast.Deep_background _
    | Mdl_basics_deep_ast.Deep_drop_shadow _,
       Mdl_basics_deep_ast.Deep_drop_shadow _
    | Mdl_basics_deep_ast.Deep_foreground _,
       Mdl_basics_deep_ast.Deep_foreground _ -> true
    | (Mdl_basics_deep_ast.Deep_font _ |
       Mdl_basics_deep_ast.Deep_background _ |
       Mdl_basics_deep_ast.Deep_drop_shadow _ |
       Mdl_basics_deep_ast.Deep_foreground _),
      (Mdl_basics_deep_ast.Deep_font _ |
       Mdl_basics_deep_ast.Deep_background _ |
       Mdl_basics_deep_ast.Deep_drop_shadow _ |
       Mdl_basics_deep_ast.Deep_foreground _) -> false in

  let rec loop accu = function
    | [] -> accu
    | dfc :: dfcs ->
      loop (dfc :: List.filter (same_dfc dfc) accu) dfcs in

  loop mdfcs deep_font_and_colors
;;

let explicit_annotation deep_model_defaults deep_annotation =

  let model_annotation_graphical_defaults =
    deep_model_defaults.
      Mdl_basics_deep_ast.deep_model_annotation_graphical_defaults in

  let (default_deep_font_option, mdfcs, dads) =
    fields_of_model_annotation_graphical_defaults
      model_annotation_graphical_defaults in

  let default_full_deep_font =
    full_deep_font_of_default_deep_font_option default_deep_font_option in

  match deep_annotation with
  | {
      Mdl_basics_deep_ast.
      deep_annotation_name = annotation_name;
      deep_annotation_position = deep_annotation_position;
      deep_annotation_font = deep_font;
      deep_annotation_graphical = deep_font_and_colors;
      deep_annotation_graphical_defaults = deep_annotation_graphical_defaults;
    } ->

    {
      explicit_annotation_name = annotation_name;
      explicit_annotation_position = deep_annotation_position;
      explicit_annotation_font =
        explicit_font_of_deep_font default_full_deep_font deep_font;
      explicit_annotation_graphical =
        inherit_annotation_graphical mdfcs deep_font_and_colors;
      explicit_annotation_graphical_defaults =
        inherit_annotation_graphical_defaults
          dads deep_annotation_graphical_defaults;
    }
;;

(* FIXME: should inherit using block mask defaults and so on. *)
let explicit_mask_parameters deep_mask_parameters =
  let explicit_mask_parameter rank deep_mask_parameter =
    match deep_mask_parameter with
    | {
        Mdl_basics_deep_ast.
        deep_mask_parameter_ident = deep_mask_parameter_ident;
        deep_mask_parameter_style = deep_mask_parameter_style;
        deep_mask_parameter_prompt = deep_mask_parameter_prompt;
        deep_mask_parameter_callback = deep_mask_parameter_callback;
        deep_mask_parameter_tunable = deep_mask_parameter_tunable;
        deep_mask_parameter_value =
           Mdl_basics_deep_ast.
           Deep_mask_parameter_value mdl_text;
        deep_mask_parameter_evaluation =
          deep_mask_parameter_evaluation_option;
        deep_mask_parameter_enable =
          deep_mask_parameter_enable_option;
        deep_mask_parameter_visible =
          deep_mask_parameter_visible_option;
        deep_mask_parameter_internal =
          deep_mask_parameter_internal_option;
        deep_mask_parameter_never_save =
          deep_mask_parameter_never_save_option;
        deep_mask_parameter_read_only =
          deep_mask_parameter_read_only_option;
      } ->
      let () =
        ignore deep_mask_parameter_callback;
        explicit_warning "mask parameter callback dropped" in
      let () =
        ignore deep_mask_parameter_tunable;
        explicit_warning "mask parameter tunable dropped" in
      let () =
        match deep_mask_parameter_internal_option with
        | None -> ()
        | Some _ -> explicit_warning "mask parameter internal dropped" in
      let () =
        match deep_mask_parameter_never_save_option with
        | None -> ()
        | Some _ -> explicit_warning "mask parameter never_save dropped" in
      let () =
        match deep_mask_parameter_read_only_option with
        | None -> ()
        | Some _ -> explicit_warning "mask parameter read_only dropped" in
      let explicit_mask_parameter_evaluation =
        match deep_mask_parameter_evaluation_option with
        | None -> Mdl_basics_deep_ast.Eval
        | Some (
            Mdl_basics_deep_ast.Eval |
            Mdl_basics_deep_ast.Quote as eval) -> eval in

      let explicit_mask_parameter_value =
        match deep_mask_parameter_style with
        |  (
           Mdl_basics_deep_ast.Deep_mask_parameter_edit |
           Mdl_basics_deep_ast.Deep_mask_parameter_min |
           Mdl_basics_deep_ast.Deep_mask_parameter_max |
           Mdl_basics_deep_ast.Deep_mask_parameter_slider _)
          when
            explicit_mask_parameter_evaluation = Mdl_basics_deep_ast.Eval ->
          Explicit_mask_parameter_eval mdl_text
        | (Mdl_basics_deep_ast.Deep_mask_parameter_edit |
           Mdl_basics_deep_ast.Deep_mask_parameter_min |
           Mdl_basics_deep_ast.Deep_mask_parameter_max)
        | Mdl_basics_deep_ast.Deep_mask_parameter_checkbox
        | Mdl_basics_deep_ast.Deep_mask_parameter_popup _
        | Mdl_basics_deep_ast.Deep_mask_parameter_unidt _
        | Mdl_basics_deep_ast.Deep_mask_parameter_promote
        | Mdl_basics_deep_ast.Deep_mask_parameter_radiobutton _
        | Mdl_basics_deep_ast.Deep_mask_parameter_slider _
        | Mdl_basics_deep_ast.Deep_mask_parameter_spinbox _ ->
          Explicit_mask_parameter_quote mdl_text in
      {
        explicit_mask_parameter_ident = deep_mask_parameter_ident;
        explicit_mask_parameter_rank = rank;
        explicit_mask_parameter_style = deep_mask_parameter_style;
        explicit_mask_parameter_prompt = deep_mask_parameter_prompt;
        explicit_mask_parameter_value = explicit_mask_parameter_value;
        explicit_mask_parameter_evaluation = explicit_mask_parameter_evaluation;
        explicit_mask_parameter_enable = deep_mask_parameter_enable_option;
        explicit_mask_parameter_visible = deep_mask_parameter_visible_option;
      } in
  let explicit_mask_parameters, _ =
    List.fold_left
      (fun (explicit_mask_parameters, rank) deep_mask_parameter ->
       explicit_mask_parameter rank deep_mask_parameter ::
         explicit_mask_parameters,
       Unatural.succ rank)
      ([], Unatural._0_)
      deep_mask_parameters in
  List.rev explicit_mask_parameters
;;

let explicit_mask deep_mask =
  match deep_mask with
  | {
      Mdl_basics_deep_ast.
      deep_mask_type = deep_mask_type_option;
      deep_mask_description = deep_mask_description_option;
      deep_mask_initialisation = deep_mask_initialisation_option;
      deep_mask_display = deep_mask_display_option;
      deep_mask_parameters = deep_mask_parameters;
      deep_mask_bindings = deep_bindings;
    } ->
    let explicit_mask_parameters =
      explicit_mask_parameters deep_mask_parameters in
    {
      explicit_mask_type = deep_mask_type_option;
      explicit_mask_description = deep_mask_description_option;
      explicit_mask_initialisation = deep_mask_initialisation_option;
      explicit_mask_display = deep_mask_display_option;
      explicit_mask_parameters = explicit_mask_parameters;
      explicit_mask_bindings = deep_bindings;
    }
;;

let paramIndexForInitFromWorkspace_ident =
  Mdl_ast_functions.lparamIndexForInitFromWorkspace_ident
;;

let explicit_treat_as_atomic_unit explicit_block_bindings =
  let binding_table =
    explicit_block_bindings.explicit_block_bindings in
  if Hashtbl.mem binding_table
       Mdl_ast_functions.treat_as_atomic_unit_ident
  then
    let { Binding.bound_ident = _; bound_value = v; } =
      Hashtbl.find binding_table
        Mdl_ast_functions.treat_as_atomic_unit_ident in
    Some (
      Mdl_basics_to_explicit_ast.
      bool_of_deep_value "treat as atomic unit" v)
  else None
;;

let explicit_system_sample_time explicit_block_bindings =
  let binding_table =
    explicit_block_bindings.explicit_block_bindings in
  if Hashtbl.mem binding_table
       Mdl_ast_functions.system_sample_time_ident
  then
    let { Binding.bound_ident = _; bound_value = v; } =
      Hashtbl.find binding_table
        Mdl_ast_functions.system_sample_time_ident in
    Some v
(*
to find out the string contents just write
      Mdl_basics_to_explicit_ast.
      string_of_deep_value "system sample time" v
*)
  else None
;;

(* FIXME: should inherit from defaults *)
let get_explicit_zoom_factor explicit_binding_table =
  if Hashtbl.mem explicit_binding_table
       Mdl_ast_functions.zoom_factor_ident
  then
    let { Binding.bound_ident = _; bound_value = v; } =
      Hashtbl.find explicit_binding_table
        Mdl_ast_functions.zoom_factor_ident in
    Some (
      Mdl_basics_to_explicit_ast.
      natural_of_deep_value "zoom factor" v)
  (* Must inherit from defaults. *)
  else
    None
;;

let explicit_trigger_type_of_deep_value v =
  match string_of_deep_value "explicit_trigger_type_of" v with
  | "falling" -> Trigger_falling
  | "rising" -> Trigger_rising
  | "either" -> Trigger_either
  | "function-call" -> Trigger_function_call
  | "unset" -> Trigger_unset
  | tt ->
    explicit_error (Printf.sprintf
      "Unknown TriggerType value %s in a TriggerPort block definition" tt)
;;

let find_trigger_type explicit_block_parameters =
  let v =
    if Hashtbl.mem explicit_block_parameters
         Mdl_ast_functions.trigger_type_ident
    then
      let { Binding.bound_ident = _; bound_value = v; } =
        Hashtbl.find explicit_block_parameters
          Mdl_ast_functions.trigger_type_ident in
      v
    else Mdl_basics_deep_ast.Deep_string "unset" in
  explicit_trigger_type_of_deep_value v
;;

let explicit_action_type_of_deep_value v =
  match string_of_deep_value "explicit_action_type_of" v with
  | "case" -> Action_case
  | "default" -> Action_default
  | "else" -> Action_else
  | "elseif" -> Action_elseif
  | "then" -> Action_then
  | "unset" -> Action_unset
  | tt ->
    explicit_error (Printf.sprintf
      "Unknown ActionType value %s \
       in parameters of an ActionPort block definition" tt)
;;

let find_action_type explicit_block_parameters =
  let v =
    if Hashtbl.mem explicit_block_parameters
         Mdl_ast_functions.action_type_ident
    then
      let { Binding.bound_ident = _; bound_value = v; } =
        Hashtbl.find explicit_block_parameters
          Mdl_ast_functions.action_type_ident in
      v
    else Mdl_basics_deep_ast.Deep_string "unset" in
  explicit_action_type_of_deep_value v
;;

let explicit_block_graphical explicit_block_body =
  explicit_block_body.
  explicit_block_all_bindings.
  explicit_block_graphical
;;

let get_ports_description explicit_block_body =
  let graphical = explicit_block_graphical explicit_block_body in
  graphical.explicit_block_ports_description
;;

let get_explicit_block_graphical blk =
  let explicit_block_body = blk.explicit_block_body in
  explicit_block_graphical explicit_block_body
;;

let get_explicit_block_ports_description blk =
  let explicit_block_body = blk.explicit_block_body in
  get_ports_description explicit_block_body
;;

let get_explicit_block_parameters blk =
  let body = blk.explicit_block_body in
  body.explicit_block_all_bindings.explicit_block_parameters
;;

let find_sub_system_explicit_activation initial_guess explicit_system =
  let find_desc desc blk =
    match blk.explicit_block_desc with
    | Explicit_block_reference _
    | Explicit_block_s_function _ -> desc
    | Explicit_block_regular explicit_simple_block_desc ->
      let ty = explicit_simple_block_desc.explicit_simple_block_type in
        if ty = Mdl_block_types.EnablePort then
          let sub_system_activation =
            match desc with
            | Sub_system_plain -> Sub_system_enabled blk
            | Sub_system_activated sub_system_activation ->
              match sub_system_activation with
              | Sub_system_plain_atomic_unit _ -> Sub_system_enabled blk
              | Sub_system_enabled _blk0 ->
                explicit_error "Doubly enabled subsystem"
              | Sub_system_if_action _ ->
                explicit_error "Enable cannot occur in if action subsystem"
              | Sub_system_triggered tblk ->
                Sub_system_triggered_and_enabled (tblk, blk)
              | Sub_system_triggered_and_enabled (_, _) ->
                explicit_error "Doubly enabled & triggered subsystem" in
          Sub_system_activated sub_system_activation else
        if ty = Mdl_block_types.TriggerPort then
          let explicit_block_parameters = get_explicit_block_parameters blk in
          let trig_type = find_trigger_type explicit_block_parameters in
          let sub_system_activation =
            match desc with
            | Sub_system_plain -> Sub_system_triggered (trig_type, blk)
            | Sub_system_activated sub_system_activation ->
              match sub_system_activation with
              | Sub_system_plain_atomic_unit _ ->
                Sub_system_triggered (trig_type, blk)
              | Sub_system_enabled blk0 ->
                Sub_system_triggered_and_enabled ((trig_type, blk), blk0)
              | Sub_system_if_action _ ->
                explicit_error "Trigger cannot occur in if action subsystem"
              | Sub_system_triggered _blk0 ->
                explicit_error "Doubly triggered subsystem"
              | Sub_system_triggered_and_enabled (_, _) ->
                explicit_error "Doubly triggered & enabled subsystem" in
          Sub_system_activated sub_system_activation else
        if ty = Mdl_block_types.ActionPort (*Then || Else || Default*) then
          let explicit_block_parameters = get_explicit_block_parameters blk in
          let action_type = find_action_type explicit_block_parameters in
          let sub_system_activation =
            match desc with
            | Sub_system_plain -> Sub_system_if_action (action_type, blk)
            | Sub_system_activated sub_system_activation ->
              match sub_system_activation with
              | Sub_system_plain_atomic_unit _ ->
                Sub_system_if_action (action_type, blk)
              | Sub_system_enabled _blk0 ->
                explicit_error
                  "If action cannot occur in an enabled subsystem"
              | Sub_system_if_action _ ->
                explicit_error
                  "If action cannot occur in an if action subsystem"
              | Sub_system_triggered _blk0 ->
                explicit_error
                  "If action cannot occur in a triggered subsystem"
              | Sub_system_triggered_and_enabled (_, _) ->
                explicit_error
                  "If action cannot occur in a triggered & enabled \
                   subsystem" in
          Sub_system_activated sub_system_activation
        else desc
    | Explicit_block_sub_system _explicit_block_sub_system -> desc in

  List.fold_left
    find_desc
    initial_guess
    explicit_system.explicit_system_blocks
;;

let find_block_explicit_activation explicit_block_body =
  let ports_description = get_ports_description explicit_block_body in
  match ports_description with
  | {
      Mdl_basics_deep_ast.
      port_enable = pe;
      port_trigger = pt;
      port_ifaction = pi;
      _;
    } ->
    if pe then
      if pt
        then Simple_block_activated Simple_block_triggered_and_enabled
        else Simple_block_activated Simple_block_enabled else
    if pt then Simple_block_activated Simple_block_triggered else
    if pi then Simple_block_activated Simple_block_if_action else
    Simple_block_plain
;;

let correct_treat_as_atomic_unit_option
    treat_as_atomic_unit_option explicit_activation =
  match explicit_activation with
  | Sub_system_plain -> treat_as_atomic_unit_option
  | Sub_system_activated (
      Sub_system_plain_atomic_unit _
    | Sub_system_enabled _
    | Sub_system_if_action _
    | Sub_system_triggered _
    | Sub_system_triggered_and_enabled _) -> Some true
;;

type explicit_block_body_elements_classifier =
  {
    deep_block_items_option : deep_block_items option;
    explicit_array_option : explicit_array option;
    explicit_object_option : explicit_object option;
    explicit_mask_option : explicit_mask option;
    explicit_sf_block_type_option : explicit_sf_block_type option;
    explicit_binding_table_option :
      (Ident.ident, Mdl_basics_explicit_ast.explicit_binding)
      Hashtbl.t option;
  }
;;

let empty_explicit_block_body_elements_classifier =
  {
    deep_block_items_option = None;
    explicit_array_option = None;
    explicit_object_option = None;
    explicit_mask_option = None;
    explicit_sf_block_type_option = None;
    explicit_binding_table_option = None;
  }
;;

let set_option error_message v = function
  | None -> Some v
  | Some _ -> explicit_error error_message
;;

let find_stateflow_machine explicit_stateflow_section =
  match explicit_stateflow_section with
  | Mdl_stateflow_explicit_ast.
    Explicit_stateflow_section explicit_section_items ->
    let rec loop machine_opt = function
      | [] ->
        begin
          match machine_opt with
          | None -> explicit_error "No machines in stateflow section"
          | Some machine -> machine
        end
      | Deep_stateflow_section_instance _ :: items -> loop machine_opt items
      | Mdl_stateflow_deep_ast.
        Deep_stateflow_section_machine deep_stateflow_machine :: items ->
        let machine_opt =
          match machine_opt with
          | None -> Some deep_stateflow_machine
          | Some _ ->
            explicit_error "Too many machines in stateflow section" in
        loop machine_opt items in
    loop None explicit_section_items
;;

let find_chart_in_stateflow_machine stateflow_machine chart_file_number =
  let machine_charts =
    stateflow_machine.Mdl_stateflow_deep_ast.deep_stateflow_machine_charts in
  List.find
    (fun chart ->
     chart.
     Mdl_stateflow_deep_ast.deep_chart_file_number = chart_file_number)
    machine_charts
;;

let unknown_data_size = Ast_types.mdl_text_of_string "-1";;

let get_size_of_data_props = function
  | {
      deep_props_array = None;
      deep_props_type = _;
      deep_props_binding_table = _;
    } -> unknown_data_size
  | {
      deep_props_array = Some deep_stateflow_props_array;
      deep_props_type = _;
      deep_props_binding_table = _;
    } ->
    match deep_stateflow_props_array with
    | {
        deep_props_array_item_size = deep_props_array_item_size;
        deep_props_array_item_first_index = _;
        deep_props_array_item_binding_table = _;
      } -> deep_props_array_item_size
;;

let find_script_profile_in_datas datas =
  let classify_data (arguments, returns as accu) data =
    match data with
    | {
        Mdl_stateflow_deep_ast.
        deep_data_id = _deep_stateflow_id;
        deep_data_name = name;
        deep_data_ssid_number = _deep_stateflow_ssid_number_option;
        deep_data_scope = scope;
        deep_data_props = props;
        deep_data_binding_table = binding_table;
      } ->
      let size = get_size_of_data_props props in
      match scope with
      | Deep_scope_input_data ->
        let input_desc =
          {
            explicit_matlab_function_data_name = name;
            explicit_matlab_function_data_size = size;
          } in
        (Explicit_matlab_function_input input_desc :: arguments,
         returns)
      | Deep_scope_parameter_data ->
        let idx =
          try
            let binding =
              Hashtbl.find
                binding_table paramIndexForInitFromWorkspace_ident in
            let v = binding.Binding.bound_value in
            natural_of_deep_value "scope parameter data" v
          with
          | Not_found -> Unatural._0_ in
        let parameter_desc =
          {
            explicit_matlab_function_parameter_data_name = name;
            explicit_matlab_function_parameter_data_size = size;
            explicit_matlab_function_parameter_data_index = idx;
          } in
        (Explicit_matlab_function_parameter parameter_desc :: arguments,
         returns)
      | Mdl_stateflow_deep_ast.Deep_scope_output_data ->
        let output_desc =
          {
            explicit_matlab_function_data_name = name;
            explicit_matlab_function_data_size = size;
          } in
        (arguments, Explicit_matlab_function_output output_desc :: returns)

      | Deep_scope_constant_data
      | Deep_scope_data_store_memory_data
      | Deep_scope_exported_data
      | Deep_scope_function_input_data
      | Deep_scope_function_output_data
      | Deep_scope_input_event
      | Deep_scope_local_data
      | Deep_scope_local_event
      | Deep_scope_output_event
      | Deep_scope_temporary_data
      | Deep_scope_data_string _ -> accu in
  List.fold_left classify_data ([], []) datas
;;

exception Not_a_matlab_function_stateflow_s_function;;

(* FIXME: should avoid so much cascading type conversions! *)
let clean_matlab_script script =
  match script with
  | Mdl_code mdl_text ->
    let matlab_target_language =
      Simport_configuration.
      Target_mtlb Simport_configuration.Matlab in
    let cleanse_matlab_script =
      M_to_host.
      translate_script_to matlab_target_language mdl_text in
    match cleanse_matlab_script with
    | Mdl_text s ->
      Ast_types.mdl_code_of_string s
;;

let get_matlab_function_script explicit_stateflow_section chart_file_number =

   let stateflow_machine = find_stateflow_machine explicit_stateflow_section in

   let chart =
     let chart =
       find_chart_in_stateflow_machine stateflow_machine chart_file_number in
     (* Verify that chart type (field is deep_chart_type) is EML_CHART *)
     let chart_type = chart.Mdl_stateflow_deep_ast.deep_chart_type in
     match chart_type with
     | Some Deep_state_type_eml_chart -> chart
     | None
     | Some (
         Deep_state_type_and |
         Deep_state_type_or |
         Deep_state_type_func |
         Deep_state_type_group |
         Deep_state_type_sub |
         Deep_state_type_super |
         Deep_state_type_string _) ->
       raise Not_a_matlab_function_stateflow_s_function in

   let datas =
     chart.Mdl_stateflow_deep_ast.deep_stateflow_chart_datas in

   let arguments, returns =
     find_script_profile_in_datas datas in

   (*
      Chart must have a field deep_stateflow_chart_eml containing
      a binding for the name of the matlab function used in the script.
   *)
   let function_name =
     match chart.deep_stateflow_chart_eml with
     | None -> explicit_error "No eml field found in the eml chart"
     | Some deep_stateflow_chart_eml ->
       deep_stateflow_chart_eml.chart_eml_name in

   let state =
     match chart.deep_stateflow_chart_states with
     | [ state ] -> state
     | [] | _ ->
       explicit_error "One and only one state expected in eml chart" in

   let deep_stateflow_state_eml =
     match state with
     | {
         deep_stateflow_state_eml = Some state_eml;
         _;
       } -> state_eml
     | _ ->
       explicit_error "Cannot find eml part of state in chart" in

   let function_script =
     match deep_stateflow_state_eml with
     | {
         state_eml_script = Some script (*(Mdl_code script)*);
         state_eml_binding_table = _;
       } ->
       (* Surprisingly, we must translate the matlab script we found for the
          function to another matlab script! This may seem useless, since
          translating a matlab script to matlab should give the same script.
          In fact, this operation is not the identity: it normalizes the
          script to a more regular form that can be embedded into another
          matlab script (in particular, omitted [end] keywords are now
          explicit, so that embedded sub-functions are clearly delimited).
       *)
       clean_matlab_script script
     | {
         state_eml_script = None;
         state_eml_binding_table = _;
       } ->
       explicit_error "Cannot find script in eml part of state in chart" in

   function_name, function_script, arguments, returns
;;

(* FIXME: should disappear *)
(* FIXME: should not call Mdl_basics_deep_ast here *)
let explicit_source_code explicit_matlab_function_script =
  match explicit_matlab_function_script with
  | Mdl_code mdl_text ->
    Mdl_basics_deep_ast.
    Deep_source_code mdl_text
;;

let explicit_s_function_type explicit_stateflow_option = function
  | Deep_s_function -> Explicit_s_function
  | Deep_s_function_stateflow (deep_name, chart_file_number) ->
    let explicit_stateflow_section =
      match explicit_stateflow_option with
      | None ->
        explicit_error
          "no stateflow section when encountering a stateflow S-Function"
      | Some explicit_stateflow_section -> explicit_stateflow_section in
    try
      let explicit_matlab_function_name,
          explicit_matlab_function_script,
          arguments, returns =
        get_matlab_function_script
          explicit_stateflow_section chart_file_number in
      let explicit_matlab_function_arguments =
        Array.of_list (List.rev arguments) in
      let explicit_matlab_function_returns =
        Array.of_list (List.rev returns) in
      Explicit_matlab_function {
        explicit_matlab_function_name = explicit_matlab_function_name;
        explicit_matlab_function_script =
          explicit_source_code explicit_matlab_function_script;
        explicit_matlab_function_arguments =
          explicit_matlab_function_arguments;
        explicit_matlab_function_returns = explicit_matlab_function_returns;
      } with
    | Not_a_matlab_function_stateflow_s_function ->
      Explicit_s_function_stateflow (deep_name, chart_file_number)
;;

let rec explicit_block
    (explicit_interfaces :
       Mdl_interface_explicit_ast.explicit_interfaces)
    explicit_stateflow_option
    deep_model_defaults deep_block =

  let explicit_block_type = explicit_block_type deep_block in
  let explicit_block_known =
    mem_explicit_interface_block explicit_interfaces explicit_block_type in

  match deep_block with
  | {
      deep_block_name = block_name;
      deep_block_sid = deep_sid_opt;
      deep_block_body = deep_block_body_elements;
      deep_block_desc = deep_block_desc;
    } ->

    let explicit_block_body =
      explicit_block_body_elements
        explicit_interfaces deep_model_defaults
        deep_block deep_block_body_elements in

    let explicit_block_desc =

      match deep_block_desc with
      | Deep_block_reference deep_block_reference_type ->
        let explicit_block_activation =
          find_block_explicit_activation explicit_block_body in
        let explicit_simple_block_desc = {
          explicit_simple_block_type = deep_block_reference_type;
          explicit_simple_block_activation = explicit_block_activation;
        } in
        Explicit_block_reference explicit_simple_block_desc
      | Deep_block_regular deep_block_regular_type ->
        let explicit_block_activation =
          find_block_explicit_activation explicit_block_body in
        let explicit_simple_block_desc = {
          explicit_simple_block_type = deep_block_regular_type;
          explicit_simple_block_activation = explicit_block_activation;
        } in
        Explicit_block_regular explicit_simple_block_desc

      | Deep_block_s_function deep_s_function_desc ->
        let explicit_block_activation =
          find_block_explicit_activation explicit_block_body in
        let explicit_block_s_function_type =
          explicit_s_function_type
             explicit_stateflow_option deep_s_function_desc in
        let explicit_simple_block_desc = {
          explicit_simple_block_type = explicit_block_s_function_type;
          explicit_simple_block_activation = explicit_block_activation;
        } in
        Explicit_block_s_function explicit_simple_block_desc

      | Deep_block_sub_system deep_system ->

        let explicit_block_sub_system_desc =
          let explicit_block_all_bindings =
            explicit_block_body.explicit_block_all_bindings in
          let explicit_system =
            explicit_system
              explicit_interfaces explicit_stateflow_option
              deep_model_defaults deep_system in
          let treat_as_atomic_unit_option =
            explicit_treat_as_atomic_unit explicit_block_all_bindings in
          let system_sample_time_option =
            explicit_system_sample_time explicit_block_all_bindings in
          (* Find in the explicit_system if there are activated blocks *)
          let initial_guess =
            match treat_as_atomic_unit_option, system_sample_time_option with
            | (None | Some false), _ -> Sub_system_plain
            | Some true, None -> Sub_system_plain
            | Some true, Some v ->
              let st = string_of_deep_value "system sample time" v in
              if st = "-1"
              then Sub_system_plain
              else Sub_system_activated (Sub_system_plain_atomic_unit v) in

          let explicit_activation =
            find_sub_system_explicit_activation
              initial_guess explicit_system in

          let treat_as_atomic_unit_option =
            correct_treat_as_atomic_unit_option
              treat_as_atomic_unit_option
              explicit_activation in

          {
            explicit_block_sub_system_activation = explicit_activation;
            explicit_block_sub_system_body = explicit_system;
            explicit_block_sub_system_treat_as_atomic_unit =
              treat_as_atomic_unit_option;
            explicit_block_sub_system_sample_time =
              system_sample_time_option;
          } in

        Explicit_block_sub_system explicit_block_sub_system_desc in

    {
      explicit_block_name = block_name;
      explicit_block_sid = deep_sid_opt;
      explicit_block_body = explicit_block_body;
      explicit_block_known = explicit_block_known;
      explicit_block_desc = explicit_block_desc;
    }

and explicit_block_body_elements
    (explicit_interfaces :
       Mdl_interface_explicit_ast.explicit_interfaces)
    deep_model_defaults
    deep_block
    deep_block_body_elements =

  let
    {
      deep_block_items_option = deep_block_properties_option;
      explicit_array_option = explicit_array_option;
      explicit_object_option = explicit_object_option;
      explicit_mask_option = explicit_mask_option;
      explicit_sf_block_type_option = explicit_sf_block_type_option;
      explicit_binding_table_option = explicit_instance_data_option;
    } =

    let set_deep_block_properties =
      set_option "defaults body_element"
    and set_explicit_array =
      set_option "array body_element"
    and set_explicit_object =
      set_option "object body_element"
    and set_explicit_mask =
      set_option "mask body_element"
    and set_explicit_instance_data =
      set_option "instance data body_element"
    and set_explicit_sf_block_type =
      set_option "sf block type body_element" in

    let rec loop classifier = function
      | [] -> classifier
      | body_element :: deep_block_body_elements ->
        let classifier =
          match body_element with
          | Deep_block_bindings deep_block_properties ->
            {
              classifier with
              deep_block_items_option =
                set_deep_block_properties
                  deep_block_properties
                  classifier.deep_block_items_option;
            }
          | Deep_block_array deep_array_description ->
            let explicit_array =
              explicit_array deep_array_description in
            {
              classifier with
              explicit_array_option =
                set_explicit_array
                  explicit_array
                  classifier.explicit_array_option;
            }
          | Deep_block_object deep_object_description ->
            let explicit_object =
              explicit_object deep_object_description in
            {
              classifier with
              explicit_object_option =
                set_explicit_object
                  explicit_object
                  classifier.explicit_object_option;
            }
          | Deep_block_mask deep_mask ->
            let explicit_mask = explicit_mask deep_mask in
            {
              classifier with
              explicit_mask_option =
                set_explicit_mask
                  explicit_mask
                  classifier.explicit_mask_option;
            }
          | Deep_block_sf_block_type deep_sf_block_type ->
            let explicit_sf_block_type =
              explicit_sf_block_type deep_sf_block_type in
            {
              classifier with
              explicit_sf_block_type_option =
                set_explicit_sf_block_type
                  explicit_sf_block_type
                  classifier.explicit_sf_block_type_option;
            }
          | Deep_block_instance_data deep_instance_data ->
            let explicit_binding_table =
              binding_table_of_list deep_instance_data in
            {
              classifier with
              explicit_binding_table_option =
                set_explicit_instance_data
                  explicit_binding_table
                  classifier.explicit_binding_table_option;
            } in
        loop classifier deep_block_body_elements in

    loop
      empty_explicit_block_body_elements_classifier
      deep_block_body_elements in

  let deep_block_properties =
    match deep_block_properties_option with
    | None -> []
    | Some deep_block_properties -> deep_block_properties in

  let explicit_block_all_bindings =
    create_explicit_block_all_bindings
      explicit_interfaces deep_model_defaults
      deep_block
      deep_block_properties in

  {
    explicit_block_all_bindings = explicit_block_all_bindings;
    explicit_block_array = explicit_array_option;
    explicit_block_object = explicit_object_option;
    explicit_block_mask = explicit_mask_option;
    explicit_block_sf_block_type = explicit_sf_block_type_option;
    explicit_block_instance_data = explicit_instance_data_option;
  }

and explicit_system
    (explicit_interfaces :
       Mdl_interface_explicit_ast.explicit_interfaces)
    explicit_stateflow_option
    deep_model_defaults deep_system =
  match deep_system with
  | {
      deep_system_name = system_name;
      deep_system_position = system_position;
      deep_system_bindings = deep_bindings;
      deep_system_blocks = deep_system_blocks;
      deep_system_lines = deep_lines;
      deep_system_annotations = deep_annotations;
      deep_system_connectors = deep_system_connectors;
    } ->
    let explicit_binding_table =
      binding_table_of_list deep_bindings in
(* FIXME : should also inherit from deep_model_defaults *)
    let explicit_zoom_factor =
      get_explicit_zoom_factor explicit_binding_table in
    {
      explicit_system_name = system_name;
      explicit_system_position = system_position;
      explicit_system_bindings = explicit_binding_table;
      explicit_system_blocks =
        List.map
          (explicit_block
             explicit_interfaces explicit_stateflow_option deep_model_defaults)
          deep_system_blocks;
      explicit_system_lines =
        explicit_lines deep_model_defaults deep_lines;
      explicit_system_annotations =
        List.map (explicit_annotation deep_model_defaults)
          deep_annotations;
      explicit_system_connectors =
        List.map binding_table_of_list deep_system_connectors;
      explicit_system_zoom_factor = explicit_zoom_factor;
    }
;;

let explicit_model_configuration
    body_kind deep_model_configuration_section =

  let find_active_configuration sections = function
    | Some { Mdl_ast.configuration_object_id = (path, nat) as id; _ } ->
      Some (
        try
          List.find
            (fun conf ->
             let conf_id = conf.deep_configuration_object_id in
             (*prerr_endline (Printf.sprintf "Found config %d" conf_id);*)
             conf_id = id)
            sections with
        | Not_found ->
          explicit_error (Printf.sprintf
            "configuration set section %s with number %d not found."
            path
            (Unatural_types.of_natural nat)))
    | None ->
      match sections with
      | c :: _ -> Some c
      | [] ->
        begin match body_kind with
        | Mdl_basics_explicit_ast.Explicit_library ->
          explicit_warning "configuration set section not found";
          None
        | Mdl_basics_explicit_ast.Explicit_model ->
          explicit_warning
            "configuration set section not found, using default \
             simulation parameters";
          None
        end in

  let active_configuration =
    find_active_configuration
      deep_model_configuration_section.deep_model_configurations
      deep_model_configuration_section.deep_model_set_configuration_option in

  {
    explicit_graphical_parameters =
      deep_model_configuration_section.deep_graphical_parameters;
    explicit_configuration_bindings =
      deep_model_configuration_section.deep_configuration_bindings;
    explicit_vendor_bindings =
      deep_model_configuration_section.deep_vendor_bindings;
    explicit_active_configuration = active_configuration;
  }
;;

let explicit_model_body
    body_kind
    (explicit_interfaces :
     Mdl_interface_explicit_ast.explicit_interfaces)
    explicit_stateflow_option
    deep_model_body =

  match deep_model_body with
  | {
      deep_model_ident = model_ident;
      deep_model_version = deep_model_version;
      deep_model_explicit_file_name = efname;
      deep_model_configuration_section = deep_model_configuration;
      deep_model_defaults = deep_model_defaults;
      deep_model_system = deep_system;
      deep_model_workspace = deep_model_workspace;
    } ->
    {
      explicit_model_ident = model_ident;
      explicit_model_version = deep_model_version;
      explicit_model_explicit_file_name = efname;
      explicit_model_configuration =
        explicit_model_configuration body_kind deep_model_configuration;
      explicit_model_system =
        explicit_system
          explicit_interfaces explicit_stateflow_option
          deep_model_defaults deep_system;
      explicit_model_workspace = deep_model_workspace;
    }
;;

let explicit_stateflow_option = function
  | None -> None
  | Some deep_stateflow_section ->
    let explicit_stateflow_section =
      Mdl_stateflow_to_explicit_ast.explicit_stateflow_section
        deep_stateflow_section in
    Some explicit_stateflow_section
;;

let explicit_model_matdata = function
  | Deep_matdata_section deep_matdata_section_item ->
    (* Just do nothing here *)
    Explicit_matdata_section deep_matdata_section_item
;;

let explicit_model_matdatas deep_model_matdatas =
  List.rev_map explicit_model_matdata deep_model_matdatas
;;

let explicit_model_matdatas_option deep_model_matdatas_option =
  match deep_model_matdatas_option with
  | None | Some [] -> None
  | Some deep_model_matdatas ->
    Some (explicit_model_matdatas deep_model_matdatas)
;;

let make_explicit_model_body
    body_kind
    (explicit_interfaces :
     Mdl_interface_explicit_ast.explicit_interfaces) dml =

  match dml with
  | {
      deep_model_body = deep_model_body;
      deep_model_matdatas_option = deep_matdatas_option;
      deep_model_stateflow_option = deep_stateflow_option;
    } ->
    let explicit_stateflow_option =
      explicit_stateflow_option deep_stateflow_option in
    let explicit_model_body =
      explicit_model_body
        body_kind explicit_interfaces explicit_stateflow_option
        deep_model_body in
    let explicit_model_matdatas_option =
      explicit_model_matdatas_option deep_matdatas_option in
    {
      explicit_model_body = explicit_model_body;
      explicit_model_matdatas_option = explicit_model_matdatas_option;
      explicit_stateflow_option = explicit_stateflow_option;
    }
;;

let make_explicit_model = make_explicit_model_body Explicit_model;;
let make_explicit_library = make_explicit_model_body Explicit_library;;

let implementation_to_explicit_line_ast explicit_interfaces = function
  | Deep_implementation_model dml ->
    Mdl_explicit_ast.Explicit_model
      (make_explicit_model explicit_interfaces dml)
  | Deep_implementation_library dml ->
    Mdl_explicit_ast.Explicit_library
      (make_explicit_library explicit_interfaces dml)
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
