(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_interface_to_explicit_ast.ml,v 1.1 2019-05-10 14:04:56 jpc Exp $ *)

(*
  From deep syntax to completely explicit inherited tree:
  We treat here our interface semantics for Simulink files;
  in particular the predefined pervasives list of blocks,
  as defined in simport file "pervasives.mdli".
*)

open Mdl_deep_ast;;
open Mdl_basics_to_explicit_ast;;
open Mdl_basics_explicit_ast;;
open Mdl_interface_explicit_ast;;

let make_ebip_tbl block_parameters =
  let (ebip_tbl : explicit_interface_block_parameters_table) =
    Hash_table.create (Configuration.get_mean_block_parameters ()) in
  let enter_parameter { Binding.bound_ident = id; bound_value = v; } =
    if Hashtbl.mem ebip_tbl id then
      Say.fatal_error (Printf.sprintf
        "Identifier %s encountered twice in block parameters"
        (Ast_types.string_of_ident id)) else
    Hashtbl.add ebip_tbl id v in
  List.iter enter_parameter block_parameters;
  ebip_tbl
;;

(*
  The function enters the list of blocks found in an interface block
  specification file, in the explicit_interface_table association table
  argument.
*)
let enter_explicit_interface_block eibt_tbl ebt ebi =
  if not (Hashtbl.mem eibt_tbl ebt) then Hashtbl.add eibt_tbl ebt ebi else
  begin
    Hashtbl.replace eibt_tbl ebt ebi;
    if Configuration.get_debug () then
      let sebt = name_of_explicit_block_type ebt in
      let fname = Configuration.get_source_file () in
      Say.warning (Printf.sprintf
        "block type %S occurs multiple times in interface %s" sebt fname);
  end
;;

let enter_deep_interface_block eibt_tbl
    {
      deep_interface_block_desc = bl_desc;
      deep_interface_block_ports_description = bl_ports_desc;
      deep_interface_block_parameters = bl_params;
    } =
  let desc, ty =
    match bl_desc with
    | Deep_interface_block_simulink bl_type ->
      Ebi_primitive, Explicit_block_regular_type bl_type
    | Deep_interface_block_sub_system bl_ref_type ->
      Ebi_sub_system, Explicit_block_reference_type bl_ref_type
    | Deep_interface_block_s_function bl_ref_type ->
      Ebi_s_function, Explicit_block_reference_type bl_ref_type in
  let ebip_tbl = make_ebip_tbl bl_params in
  let ebi = {
    ebi_desc = desc;
    ebi_ports_description = bl_ports_desc;
    ebi_parameters = ebip_tbl;
  } in
  enter_explicit_interface_block eibt_tbl ty ebi
;;

let enter_explicit_interface_block eibt_tbl dbis =
  List.iter (enter_deep_interface_block eibt_tbl) dbis
;;

let create_explicit_interface_block_table () =
  let max_block_types = Configuration.get_max_block_types () in
  Hash_table.create max_block_types
;;

let create_explicit_interface iname explicit_fname =
  {
    ei_version = Configuration.get_interface_version ();
    ei_ident = iname;
    ei_source = Path.of_explicit_file_name explicit_fname;
    ei_table = create_explicit_interface_block_table ();
  }
;;

let check_interface_version intf =
  let v = intf.ei_version in
  if not (v = Configuration.get_interface_version ()) then
    explicit_error (Printf.sprintf
      "enter_explicit_interface: interface in %s has version %s, \
       which is not compatible with the current %s interface version (%s)"
      intf.ei_source v
      (Configuration.get_software_name ())
      (Configuration.get_interface_version ()))
;;

let check_interface interface_ident intf =
  if interface_ident <> intf.ei_ident then
    explicit_error (Printf.sprintf
      "enter_explicit_interface: interface in %s has name %s, \
       which should be %s"
      intf.ei_source
      (Ast_types.string_of_ident intf.ei_ident)
      (Ast_types.string_of_ident interface_ident)) else
    check_interface_version intf
;;

let read_simulink_interface_file explicit_fname =
  let dast = Mdl_to_deep_ast.read_mdl_file explicit_fname in
  match dast with
  | Deep_interface dint -> dint
  | Deep_implementation _ds ->
    Mdl_basics_to_explicit_ast.explicit_error (Printf.sprintf
      "%s is not a valid Simulink interface file"
      (Path.of_explicit_file_name explicit_fname))
;;

let enter_simulink_interface dintf =
  match dintf with
  | {
      deep_interface_ident = iname;
      deep_interface_explicit_file_name = efname;
      deep_interface_blocks = dis;
    } ->
    let dintf = create_explicit_interface iname efname in
    let eibt_tbl = dintf.ei_table in
    enter_explicit_interface_block eibt_tbl dis;
    dintf
;;

let read_simulink_interface_source_file explicit_fname =
  Say.debug (Printf.sprintf
    "read_interface_source_file %s"
    (Path.of_explicit_file_name explicit_fname));
  let dintf = read_simulink_interface_file explicit_fname in
  enter_simulink_interface dintf
;;

(** From Simulink interface file to compiled interface file. *)

let read_simulink_interface_marshaled_file,
    write_simulink_interface_marshaled_file =
  Umarshal.read_write_file ()
;;

let save_marshaled_interface explicit_file_name intf =
  let explicit_cfname =
    Configuration.simulink_interface_marshaled_file_name explicit_file_name in
  write_simulink_interface_marshaled_file explicit_cfname intf
;;

let load_marshaled_interface_file explicit_fname =
  let intf = read_simulink_interface_marshaled_file explicit_fname in
  check_interface_version intf;
  intf
;;

let load_simulink_interface_source_file explicit_file_name =
  Say.debug (Printf.sprintf
    "load_simulink_interface_source_file %s"
    (Path.of_explicit_file_name explicit_file_name));
  read_simulink_interface_source_file explicit_file_name
;;

let load_simulink_interface interface_ident =
  let iname = Ast_types.string_of_ident interface_ident in
  Say.debug (Printf.sprintf "load_simulink_interface %s" iname);
  let cfname =
    Path.add_extension
      iname Configuration.simulink_interface_marshaled_file_extension in
  if Path.exists cfname then
    let explicit_cfname = Path.find cfname in
    let intf = load_marshaled_interface_file explicit_cfname in
    check_interface interface_ident intf;
    intf else
  let fname =
    Path.add_extension
      iname Configuration.simulink_interface_source_file_extension in
  if Path.exists fname then
    let explicit_file_name = Path.find fname in
    let intf = load_simulink_interface_source_file explicit_file_name in
    check_interface interface_ident intf;
    save_marshaled_interface explicit_file_name intf;
    intf
  else
    explicit_error (Printf.sprintf
      "load_simulink_interface: interface %s was not found." iname)
;;

let load_simulink_interfaces () =
  let interfaces = Configuration.get_interfaces () in
  List.map load_simulink_interface interfaces
;;

let interface_to_explicit_ast dintf =
  let eintf = enter_simulink_interface dintf in
  Mdl_explicit_ast.Explicit_interface eintf
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
