(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_utilities_common.ml,v 1.1 2019-05-10 14:07:14 jpc Exp $ *)

open Ast_types;;
open Mdl_basics_deep_ast;;
open Genr_types;;
open Instructions;;

type natural = Genr_types.natural;;

type connection_port_desc_generator =
  natural -> natural ->
  Mdl_rename_ast.rename_connection_connector_desc
;;

(** {6 General utilities to translate Simulink blocks} *)

let fdebug f =
  Say.fdebug (fun epf ->
    Format.fprintf epf
    "@[<v 4>during block translation,@ \
     %t@]" f)
;;

let warning w =
  Say.warning (Format.sprintf
    "@[<v 4>during block translation,@ \
     %s@]" w)
;;

let warning_coercion ident w =
  warning (Format.sprintf "%s for ident %S"
    w (Ast_types.string_of_ident ident))
;;

let fatal_error err =
  Say.fatal_error (Format.sprintf
    "@[<v 4>during block translation,@ \
     %s@]" err)
;;

let wrong_type_for ident deep_value ty =
  let deep_value_text =
    Mdl_deep_ast_print.print_deep_value Format.str_formatter deep_value;
    Format.flush_str_formatter () in
  fatal_error (Format.sprintf
    "@[<v 0>identifier ``%s'' is bound to ``%s''@ \
     which is not of type ``%s''@]"
    (Ast_types.string_of_ident ident) deep_value_text ty)
;;

let wrong_value_for_type message deep_value ty =
  let deep_value_text =
    Mdl_deep_ast_print.print_deep_value Format.str_formatter deep_value;
    Format.flush_str_formatter () in
  fatal_error (Format.sprintf
    "@[<v 0>%s: value ``%s'' @ \
     is not of type ``%s''@]"
     message deep_value_text ty)
;;

let debug_with_info genr_info mess =
  fdebug (fun epf ->
      Format.fprintf epf "%s@ %a"
        mess
        Genr_block_types_print.print_genr_info genr_info
    )
;;

let warning_with_info genr_info w =
  warning (Format.sprintf
    "@[<v 4>%S,@ \
     %s@]"
    (Genr_block.block_identity genr_info) w)
;;

let fatal_error_with_info genr_info err =
  fatal_error (Format.sprintf
    "@[<v 4>%S,@ \
     %s@]"
    (Genr_block.block_identity genr_info) err)
;;

(** {6 General iterators for list using natural} *)

let rev_map_i_start start f l =
  let rec loop accu i = function
    | [] -> accu
    | x :: l -> loop (f i x :: accu) (Unatural.succ i) l in
  loop [] start l
;;

let rev_map_i_start2 start f l1 l2 =
  let rec loop accu i l1 l2 =
    match l1, l2 with
    | [], [] -> accu
    | x :: l1, y :: l2 -> loop (f i x y :: accu) (Unatural.succ i) l1 l2
    | _, _ ->
     failwith
       "rev_map_i_start2: \
        the two list arguments should have the same length" in
  loop [] start l1 l2
;;

(** {6 Injection from Caml values to scicos block parameter binding values} *)

let code_out_int i = Mdl_basics_deep_ast.Deep_int (string_of_int i);;
let code_out_natural n = code_out_int (Unatural.to_int n);;
let code_out_float f = Deep_float (string_of_float f);;
let code_out_string s = Deep_string s;;
let code_out_bool =
  let scicos_int_of_bool = function
    | true -> 1
    | false -> 0 in
  (fun b ->
   code_out_int (scicos_int_of_bool b))
;;

let code_out_expression txt =
  Deep_expression (Ast_types.mdl_text_of_string txt)
;;

let code_out_port_number (Mdl_port_number n) = code_out_natural n;;

(** {6 Auxiliaries for Model_to_code} *)

let block_parameter_binding = Genr_utils.block_parameter_binding;;
let struct_parameter_binding = Genr_utils.struct_parameter_binding;;

(** {6 Retrieving Simulink block argument and parameter binding values} *)

let get_nin_nout genr_info =
  let ports_desc = genr_info.Genr_block_types.block_ports_description in
  Mdl_ast_to_deep_ast.get_regular_port_nin ports_desc,
  Mdl_ast_to_deep_ast.get_regular_port_nout ports_desc
;;

let exists_argument genr_info ident =
  Hashtbl.mem genr_info.Genr_block_types.block_arguments ident ||
  match genr_info.Genr_block_types.block_explicit_instance_data with
  | None-> false
  | Some tbl ->
    Hashtbl.mem tbl ident
;;

let exists_parameter genr_info ident =
  Hashtbl.mem genr_info.Genr_block_types.block_bindings ident
;;

let find_argument genr_info ident =
  try Hashtbl.find genr_info.Genr_block_types.block_arguments ident with
  | Not_found ->
    match genr_info.Genr_block_types.block_explicit_instance_data with
    | None-> raise Not_found
    | Some tbl ->
      Hashtbl.find tbl ident
;;

let raw_get_argument_value genr_info ident =
  let bdg = find_argument genr_info ident in
  bdg.Binding.bound_value
;;

let raw_get_parameter_value genr_info ident =
  let bdg = Hashtbl.find genr_info.Genr_block_types.block_bindings ident in
  bdg.Binding.bound_value
;;

(*
   Redefining module Hashtbl to prevent direct access to genr_info tables,
   without going through authorized channels raw_get* functions.
*)
module Hashtbl = struct end;;

(*
let print_explicit_block_type ppf = function
  | Mdl_basics_explicit_ast.Explicit_block_type b ->
    Format.fprintf ppf "@[<1>block type %a@]%!"
      Mdl_block_types_print.print_block_type b
  | Mdl_basics_explicit_ast.Explicit_block_source_type b ->
    Format.fprintf ppf "@[<1>block source type %a@]%!"
      Mdl_ast_print.print_block_source_type b
;;

let string_of_explicit_block_type ebt =
  let b = Buffer.create 64 in
  let ppf = Format.formatter_of_buffer b in
  print_explicit_block_type ppf ebt;
  Buffer.contents b
;;
*)

(* [get_argument_value genr_info ident] return the explicit value bound to
  argument [ident] in a Simulink block instance arguments. *)
let get_argument_value genr_info ident =
  try raw_get_argument_value genr_info ident with
  | Not_found ->
    fatal_error_with_info genr_info
      (Printf.sprintf
         "block must have argument %s"
         (Ast_types.string_of_ident ident))
;;

(* [get_parameter_value genr_info ident] return the explicit value bound to
  parameter [ident] in a Simulink block instance parameters. *)
let get_parameter_value genr_info ident =
  try raw_get_parameter_value genr_info ident with
  | Not_found ->
    fatal_error_with_info genr_info
      (Printf.sprintf
         "block must have parameter binding for %s"
         (string_of_ident ident))
;;

(* As a general hint, function [get_t] returns a value of Caml type [t]
  from the explicit value bolund to a Simulink block parameter. *)

(* [get_string ident v] return the Caml [string] from the explicit value
  [v], bound to [ident] in a Simulink block *)
let rec get_string ident = function
  | Deep_string s -> s
  | Deep_int s ->
    warning_coercion ident
      (Printf.sprintf "Implicit coercion of int ``%s'' to a string" s);
    s
  | Deep_float s ->
    warning_coercion ident
      (Printf.sprintf "Implicit coercion of float ``%s'' to a string" s);
    s
  | Deep_bool b ->
    warning_coercion ident
      (Printf.sprintf "Implicit coercion of bool %B to a string" b);
    if b then "on" else "off"
  | Deep_ident id ->
    let s = Ast_types.string_of_ident id in
    warning_coercion ident
      (Printf.sprintf "Implicit coercion of ident ``%s'' to a string" s);
    s
  | Deep_vector v ->
    let vs = List.map (get_string ident) v in
    let s = Printf.sprintf "[%s]" (String.concat ", " vs) in
    warning_coercion ident
      (Printf.sprintf "Implicit coercion of vector ``%s'' to a string" s);
    s
  | Deep_matrix m ->
    let ms = List.map (get_string ident) m in
    let s = Printf.sprintf "[%s]" (String.concat "; " ms) in
    warning_coercion ident
      (Printf.sprintf "Implicit coercion of matrix ``%s'' to a string" s);
    s
  |(Deep_bindings _ | Deep_expression _ | Deep_code _ | Deep_wstring _)
    as deep_value ->
    wrong_type_for ident deep_value "string"
;;

let get_bool_internal simulink_type ident = function
  | Deep_bool b -> b
  | Deep_string ("on" | "off" as s) ->
    warning_coercion ident (Printf.sprintf
      "Implicit coercion of string ``%s'' to a %s value" s simulink_type);
    s = "on"
  |(Deep_string _ | Deep_wstring _ | Deep_matrix _ | Deep_vector _
  | Deep_bindings _
  | Deep_ident _ | Deep_expression _ | Deep_code _
  | Deep_int _ | Deep_float _) as deep_value ->
    wrong_type_for ident deep_value simulink_type
;;

(* [get_bool ident v] return the Caml [bool] from the explicit value
  [v] bound to [ident] in a Simulink block *)
let get_bool = get_bool_internal "boolean";;

let floor_round_method = code_out_string "Floor";;

let external_activation_mode_on = code_out_bool true
and external_activation_mode_off = code_out_bool false
;;

(* A dummy external activation mode to allow a single profile for functions of
  Scicos that do not handle a external activation parameter. *)
let external_activation_mode_dummy = external_activation_mode_off;;

let saturation_mode_off = Genr_types.Saturation_off;;
let saturation_mode_on = Genr_types.Saturation_on;;

(* [get_simulink_saturation_mode ident v] return the Caml
  [Genr_types.simulink_saturation] value found in the explicit
  value [v] bound to [ident] in a Simulink block. *)
let get_simulink_saturation_mode ident v =
  match get_bool_internal "saturation" ident v with
  | false -> saturation_mode_off
  | true -> saturation_mode_on
;;

let get_simulink_code ident = function
  | (Deep_int _
  | Deep_string _
  | Deep_wstring _ | Deep_matrix _ | Deep_vector _ | Deep_bindings _
  | Deep_ident _ | Deep_expression _
  | Deep_float _ | Deep_bool _) as deep_value ->
    wrong_type_for ident deep_value "code"
  | Deep_code _ as v -> v
;;

let get_text_expression ident = function
  | Deep_expression txt -> txt
  | (Deep_int _
  | Deep_string _
  | Deep_wstring _ | Deep_matrix _ | Deep_vector _ | Deep_bindings _
  | Deep_ident _
  | Deep_float _ | Deep_bool _ | Deep_code _) as deep_value ->
    wrong_type_for ident deep_value "expression"
;;

let get_expression ident v =
  Deep_expression (get_text_expression ident v)
;;

(* [get_int ident v] return the Caml int value from the explicit
  value bound to [ident] in a Simulink block. *)
let get_int ident = function
  | Deep_int i -> int_of_string i
  | Deep_string s as deep_value ->
    begin
    try
      let i = int_of_string s in
      warning_coercion ident
        (Printf.sprintf "Implicit coercion of string ``%s'' to an integer" s);
      i
    with
    | Failure _ -> wrong_type_for ident deep_value "int"
    end
  |(Deep_wstring _ | Deep_matrix _ | Deep_vector _ | Deep_bindings _
  | Deep_ident _ | Deep_expression _ | Deep_code _
  | Deep_float _ | Deep_bool _) as deep_value ->
    wrong_type_for ident deep_value "int"
;;

(* [get_float ident v] return the Caml [float] value from the explicit
  value bound to [ident] in a Simulink block. *)
let get_float ident = function
  | Deep_float i -> float_of_string i
  | Deep_string s as deep_value ->
    begin
    try
      let i = float_of_string s in
      warning_coercion ident
        (Printf.sprintf "Implicit coercion of string ``%s'' to a float" s);
      i
    with
    | Failure _ -> wrong_type_for ident deep_value "float"
    end
  |(Deep_wstring _ | Deep_matrix _ | Deep_vector _ | Deep_bindings _
  | Deep_ident _ | Deep_expression _ | Deep_code _
  | Deep_int _ | Deep_bool _) as deep_value ->
    wrong_type_for ident deep_value "float"
;;

let get_vector_items ident = function
  | Deep_vector v -> v
  |(Deep_matrix _ | Deep_bindings _
  | Deep_ident _ | Deep_expression _ | Deep_code _
  | Deep_int _ | Deep_string _ | Deep_wstring _ | Deep_float _
  | Deep_bool _) as deep_value ->
    wrong_type_for ident deep_value "vector"
;;

(* [get_vector ident v] return the Caml [Explicit_value array] value from the
  explicit value bound to [ident] in a Simulink block. *)
let get_vector ident v =
  Array.of_list (get_vector_items ident v)
;;

(* [get_int_vector ident v] return the Caml [int array] value from the
  explicit value bound to [ident] in a Simulink block. *)
let get_int_vector ident dv =
  let v = get_vector ident dv in
  Array.map (get_int ident) v
;;

(* [get_expression_vector ident v] return the Caml [expression array] value
   from the explicit value bound to [ident] in a Simulink block. *)
let get_expression_vector ident dv =
  let v = get_vector ident dv in
  Array.map (get_expression ident) v
;;

let rewrite_expression f ident v =
  let expression = get_text_expression ident v in
  match expression with
  | Mdl_text txt ->
    Deep_expression (Ast_types.mdl_text_of_string (f txt))
;;

let rewrite_expression_vector f ident v =
  let v = get_vector_items ident v in
  let txts =
    List.map (fun item ->
      Ast_types.string_of_mdl_text (get_text_expression ident item)) v in
  Deep_expression (Ast_types.mdl_text_of_string (f txts))
;;

let rewrite_script f _ident = function
(* FIXME: too many conversions! *)
  | Deep_code (Mdl_code (Mdl_text txt)) ->
    Deep_code (Ast_types.mdl_code_of_string (f txt))
  | (Deep_int _
  | Deep_string _
  | Deep_wstring _ | Deep_matrix _ | Deep_vector _ | Deep_bindings _
  | Deep_ident _
  | Deep_float _ | Deep_bool _ | Deep_expression _) as deep_value ->
    wrong_value_for_type
      "In function rewrite_script" deep_value "script"
;;

let raw_get_argument_value_string genr_info ident =
  get_string ident (raw_get_argument_value genr_info ident)
;;

let raw_get_argument_value_bool genr_info ident =
  get_bool ident (raw_get_argument_value genr_info ident)
;;

let raw_get_argument_value_simulink_saturation genr_info ident =
  get_simulink_saturation_mode ident (raw_get_argument_value genr_info ident)
;;

let get_argument_value_string genr_info ident =
  get_string ident (get_argument_value genr_info ident)
;;

let get_parameter_value_string genr_info ident =
  get_string ident (get_parameter_value genr_info ident)
;;

let get_argument_value_bool genr_info ident =
  get_bool ident (get_argument_value genr_info ident)
;;

let get_argument_value_int genr_info ident =
  get_int ident (get_argument_value genr_info ident)
;;

(*
let get_argument_value_float genr_info ident =
  get_float ident (get_argument_value genr_info ident)
;;
*)
let get_argument_value_int_vector genr_info ident =
  get_int_vector ident (get_argument_value genr_info ident)
;;

let get_argument_value_expression_vector genr_info ident =
  get_expression_vector ident (get_argument_value genr_info ident)
;;


let get_argument_value_simulink_saturation genr_info ident =
  get_simulink_saturation_mode ident (get_argument_value genr_info ident)
;;

let get_argument_value_code genr_info ident =
  get_simulink_code ident (get_argument_value genr_info ident)
;;

let get_argument_value_expression genr_info ident =
  let txt =
    Ast_types.mdl_text_of_string
      (get_argument_value_string genr_info ident) in
  Deep_expression txt
;;

(** {6 Simulink to Scicos type correspondance} *)

let out_data_type_str_ident = ident_from_string "OutDataTypeStr";;

let get_out_data_type_string genr_info =
  try raw_get_argument_value_string genr_info out_data_type_str_ident with
  | Not_found ->
    warning_with_info genr_info
      "OutDataTypeStr was not found in the Simulink block arguments,\
     \n    using default (\"Inherit\" mode)";
    "Inherit"
;;

let simulink_out_data_type_of_string = function
  | "double" | "float(\'double\')" -> Genr_types.Simulink_double
  | "single" | "float(\'single\')" -> Simulink_single
  | "int8" | "int(8)" -> Simulink_int8
  | "uint8" | "uint(8)" -> Simulink_uint8
  | "int16" | "int(16)" -> Simulink_int16
  | "uint16" | "uint(16)" -> Simulink_uint16
  | "int32" | "int(32)" -> Simulink_int32
  | "uint32" | "uint(32)" -> Simulink_uint32
  | "boolean" -> Simulink_boolean
  | s when
      String.length s >= 7 &&
      String.sub s 0 7 = "Inherit" -> Simulink_inherit
  | s ->
    warning
      (Printf.sprintf
         "out data type ``%s'' is not supported, using double instead." s);
    Simulink_double
;;

let get_simulink_out_data_type genr_info =
  let out_data_type = get_out_data_type_string genr_info in
  simulink_out_data_type_of_string out_data_type
;;

let string_of_simulink_out_data_type = function
  | Simulink_single -> "single"
  | Simulink_double -> "double"
  | Simulink_int8 -> "int8"
  | Simulink_uint8 -> "uint8"
  | Simulink_int16 -> "int16"
  | Simulink_uint16 -> "uint16"
  | Simulink_int32 -> "int32"
  | Simulink_uint32 -> "uint32"
  | Simulink_boolean -> "boolean"
  | Simulink_inherit -> "inherit"
(* Used to print out Simulink out data type in case of error *)
;;

let concat_instruction_lists = Concat_list.clist_of_lists;;

(** {6 Block generators} *)

let mdl_port_number_of_natural = Ast_types.mdl_port_number_of_natural;;

let port_1_ = mdl_port_number_of_natural Unatural._1_;;
let port_2_ = mdl_port_number_of_natural Unatural._2_;;
let port_3_ = mdl_port_number_of_natural Unatural._3_;;
let port_4_ = mdl_port_number_of_natural Unatural._4_;;

(* Building connection ports *)
let make_connection_port_info port_number nin nout =
  let mdl_port_number =
    mdl_port_number_of_natural port_number in
  let connection_port_value =
    Mdl_implementation_explicit_ast.Connection_port_value mdl_port_number in
  {
    Mdl_rename_ast.
    connector_block_nin = nin;
    connector_block_nout = nout;
    connection_port_value = connection_port_value;
  }
;;

let mdl_left_connection_port_number_of_natural port_number nin nout =
  let connection_port_info =
    make_connection_port_info port_number nin nout in
  Mdl_rename_ast.
  Rename_left_connector connection_port_info
;;

let mdl_right_connection_port_number_of_natural port_number nin nout =
  let connection_port_info =
    make_connection_port_info port_number nin nout in
  Mdl_rename_ast.
  Rename_right_connector connection_port_info
;;

let port_l1_ = mdl_left_connection_port_number_of_natural Unatural._1_;;
let port_l2_ = mdl_left_connection_port_number_of_natural Unatural._2_;;
let port_l3_ = mdl_left_connection_port_number_of_natural Unatural._3_;;
let port_l4_ = mdl_left_connection_port_number_of_natural Unatural._4_;;

let port_r1_ = mdl_right_connection_port_number_of_natural Unatural._1_;;
let port_r2_ = mdl_right_connection_port_number_of_natural Unatural._2_;;
let port_r3_ = mdl_right_connection_port_number_of_natural Unatural._3_;;
let port_r4_ = mdl_right_connection_port_number_of_natural Unatural._4_;;

let default_block_port_size = {
  ss_w =
    Unatural.to_float (Configuration.get_block_port_width ());
  ss_h =
    Unatural.to_float (Configuration.get_block_port_height ());
}
;;

let set_block_geometry block_id (nin, nout) (x, y) =
  let m = Configuration.get_block_geometry_magnification () in
  [
    Instructions.Set_block_nin (block_id, nin);
    Instructions.Set_block_nout (block_id, nout);
    Instructions.Set_block_flip (block_id, false);
    Instructions.Set_block_origin (
      block_id,
      {
        Instructions.
        sp_x = Unatural.to_float (Unatural.mul m x);
        sp_y = Unatural.to_float (Unatural.mul m y);
      }
    );
  ]
;;

let genr_add_block =
  Mdl_translation.add_block Mdl_translation.genr_table
;;

let block_name block_string = Ast_types.mdl_name_of_string block_string;;

let blk0 = Genr_utils.blk0
and scsm0 = Genr_utils.scsm0
;;

let add_block_body scsm_id block_id block_string block_body nports origin =
  let add_instruction =
    genr_add_block scsm_id block_id (block_name block_string) in
  block_body @
  set_block_geometry block_id nports origin @
  [ add_instruction; ]
;;

let add_block_body_in_scsm0 = add_block_body scsm0 blk0;;

let get_block_tag block_name =
  Mdl_translation.get_block_tag
    Mdl_translation.genr_table block_name
;;

let add_directed_link
    scsm_id src_string src_port_desc dst_string dst_port_desc points =
  let src_name = block_name src_string
  and dst_name = block_name dst_string in
  let src_tag = get_block_tag src_name
  and dst_tag = get_block_tag dst_name in

  let src_port =
    {
      Mdl_implementation_explicit_ast.
      explicit_connector_port_block = src_name;
      explicit_connector_port_desc = src_port_desc;
    }
  and dst_port =
    {
      Mdl_implementation_explicit_ast.
      explicit_connector_port_block = dst_name;
      explicit_connector_port_desc = dst_port_desc;
    } in

  Instructions.Add_directed_link (
    scsm_id,
    Dcs (src_tag, src_port),
    Dcd (dst_tag, dst_port),
    points
  )
;;

let add_directed_link_in_scsm0 = add_directed_link scsm0;;

let add_connection_link
    scsm_id src_string
    (src_port_desc :
     Mdl_rename_ast.rename_connection_connector_desc)
    dst_string dst_port_desc
    points =
  let src_name = block_name src_string
  and dst_name = block_name dst_string in
  let src_tag = get_block_tag src_name
  and dst_tag = get_block_tag dst_name in

  let src_port =
    {
      Mdl_implementation_explicit_ast.
      explicit_connector_port_block = src_name;
      explicit_connector_port_desc = src_port_desc;
    }
  and dst_port =
    {
      Mdl_implementation_explicit_ast.
      explicit_connector_port_block = dst_name;
      explicit_connector_port_desc = dst_port_desc;
    } in

  Instructions.Add_connection_link (
    scsm_id,
    Ccs (src_tag, src_port),
    Ccd (dst_tag, dst_port),
    points
  )
;;

let add_connection_link_in_scsm0 = add_connection_link scsm0;;

let add_event_link scsm_id src_string src_port dst_string dst_port points =
  let src_name = block_name src_string
  and dst_name = block_name dst_string in
  let src_tag = get_block_tag src_name
  and dst_tag = get_block_tag dst_name in
  Add_event_link (scsm_id,
    Dcs (src_tag, {
          Mdl_implementation_explicit_ast.
          explicit_connector_port_block = src_name;
          explicit_connector_port_desc = src_port;
        }
    ),
    Dcd (dst_tag, {
          Mdl_implementation_explicit_ast.
          explicit_connector_port_block = dst_name;
          explicit_connector_port_desc = dst_port;
        }
    ),
    points
  )
;;

let add_event_link_in_scsm0 = add_event_link scsm0;;

(** {6 Ubiquitous identificators value retrieving functions} *)

(* Sample time retrieving. *)
let sampletime_ident = ident_from_string "SampleTime";;
let tsamp_ident = ident_from_string "tsamp";;
let samptime_ident = ident_from_string "samptime";;
let sample_time_ident = ident_from_string "sample_time";;

let get_sample_time genr_info =
  try raw_get_argument_value_string genr_info sampletime_ident with
  | Not_found ->
    try raw_get_argument_value_string genr_info tsamp_ident with
    | Not_found ->
      try raw_get_argument_value_string genr_info sample_time_ident with
        | Not_found ->
          try raw_get_argument_value_string genr_info samptime_ident with
          | Not_found ->
            warning_with_info genr_info
            "Neither SampleTime, nor tsamp, nor sample_time was found\
           \n    in the block arguments, using default (\"Inherit\" mode)";
           "-1"
;;

let system_sample_time_ident = Mdl_ast_functions.system_sample_time_ident;;
let get_system_sample_time genr_info =
  try raw_get_argument_value_string genr_info system_sample_time_ident with
  | Not_found ->
    fatal_error_with_info genr_info
      (Printf.sprintf
         "SystemSampleTime was not found in the block arguments.");
;;

let string_to_float_value s =
  try
    let _ = float_of_string s in
    Deep_float s with
  | Failure _ (*"float_of_string"*) ->
    Deep_expression (mdl_text_of_string s)
;;

let get_sample_time_value genr_info =
  string_to_float_value (get_sample_time genr_info)
;;

let get_system_sample_time_value genr_info =
  string_to_float_value (get_system_sample_time genr_info)
;;

(* Zero crossing retrieving. *)
let zero_cross_ident = ident_from_string "ZeroCross";;

let get_zero_crossing genr_info =
  let zero_cross =
    try raw_get_argument_value_bool genr_info zero_cross_ident with
    | Not_found ->
      warning_with_info genr_info
        "ZeroCross was not found in the block arguments,\
       \n    using default (no zero crossing)";
      false in
  zero_cross
;;

let get_zero_crossing_value genr_info =
  code_out_bool (get_zero_crossing genr_info)
;;

(* Saturation retrieving. *)
let saturate_on_integer_overflow_ident =
  ident_from_string "SaturateOnIntegerOverflow"
;;

let do_satur_ident = ident_from_string "DoSatur";;

let get_simulink_saturation genr_info =
  try
    raw_get_argument_value_simulink_saturation
      genr_info saturate_on_integer_overflow_ident
  with
  | Not_found ->
    try
      raw_get_argument_value_simulink_saturation
        genr_info do_satur_ident
    with
    | Not_found ->
      warning_with_info genr_info
        "SaturateOnIntegerOverflow nor DoSatur was found in the block \
         arguments,\
       \n    using default (no saturation)";
      saturation_mode_off
;;

(* Show enable port retrieving. *)
let show_enable_port_ident = ident_from_string "ShowEnablePort";;

let get_show_enable_port genr_info =
  try raw_get_argument_value_bool genr_info show_enable_port_ident with
  | Not_found ->
    warning_with_info genr_info
      "ShowEnablePort was not found in the integrator block arguments,\
     \n    using default (false)";
    false
;;

let get_show_enable_port_value genr_info =
  code_out_bool (get_show_enable_port genr_info)
;;

(* Show saturation port retrieving. *)
let show_saturation_port_ident = ident_from_string "ShowSaturationPort";;

let get_show_saturation_port genr_info =
  try raw_get_argument_value_bool genr_info show_saturation_port_ident with
  | Not_found ->
    warning_with_info genr_info
      "ShowSaturationPort was not found in the integrator block arguments,\
     \n    using default (false)";
    false
;;

let get_show_saturation_port_value genr_info =
  code_out_bool (get_show_saturation_port genr_info)
;;

(* Argument binding value retrieving. *)
let get_ident_string = get_argument_value_string;;
let get_ident_bool = get_argument_value_bool;;
let get_ident_int = get_argument_value_int;;
let get_ident_int_vector = get_argument_value_int_vector;;
let get_ident_simulink_saturation = get_argument_value_simulink_saturation;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
