(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_utilities_common.mli,v 1.1 2019-05-10 14:07:14 jpc Exp $ *)

open Instructions;;
open Genr_types;;

(** {3 General utilities to translate Simulink blocks} *)

type natural = Genr_types.natural;;

type connection_port_desc_generator =
  natural -> natural ->
  Mdl_rename_ast.rename_connection_connector_desc
;;

(** {6 Failing and debugging} *)

val fdebug : (Format.formatter -> unit) -> unit;;
val warning : string -> unit;;
val fatal_error : string -> 'a;;
val wrong_type_for :
  Genr_types.ident -> Instructions.code_value -> string -> 'a
;;
val debug_with_info : Genr_block_types.genr_info -> string -> unit;;
val warning_with_info : Genr_block_types.genr_info -> string -> unit;;
val fatal_error_with_info : genr_info -> string -> 'a;;

(** {6 General iterators for lists using natural indexes} *)

val rev_map_i_start :
  natural -> (natural -> 'a -> 'b) -> 'a list -> 'b list
;;
val rev_map_i_start2 :
  natural ->
  (natural -> 'a -> 'b -> 'c) -> 'a list -> 'b list -> 'c list
;;

(** {6 Injection/projection from Caml to Scicos binding values} *)

(** From Caml values created in block generators
  to Scicos block parameter binding values. *)

val code_out_int : int -> code_value;;
val code_out_natural : natural -> code_value;;
val code_out_float : float -> code_value;;
val code_out_string : string -> code_value;;
val code_out_bool : bool -> code_value;;
val code_out_expression : string -> code_value;;
val code_out_port_number : Ast_types.mdl_port_number -> code_value;;

(** Building scicos block parameter bindings. *)

val block_parameter_binding :
  Instructions.modeler_ident * code_value ->
  Instructions.modeler_block_parameter_binding
(** Simple block parameter binding: an ident bound to a value. *)
;;

val struct_parameter_binding :
  Instructions.modeler_ident *
    (Instructions.modeler_ident * code_value list) Ulist1.t ->
  Instructions.modeler_block_parameter_binding
(** Structured block parameter bindings: a list of idents bound to a record,
  i.e. a non empty list of pairs of a label and a list of values. *)
;;

(** {6 Retrieving Simulink block argument and parameter binding values} *)

(** Retrieving Simulink block argument and parameters binding values from
  [genr_info] description of a source block. *)
val get_nin_nout : genr_info -> Unatural.t * Unatural.t;;
val exists_argument : genr_info -> ident -> bool;;
val exists_parameter : genr_info -> ident -> bool
(** Test if a given identifier is bound as a parameter or as an argument. *)
;;

(** The [get_*] family abort compilation if the argument has no binding in
  the [genr_info] description, or if the ident is bound to a value that
  has not the expected type. *)

val get_argument_value : genr_info -> ident -> code_value
(** [get_argument_value genr_info ident] return the explicit value bound to
  argument [ident] in a Simulink block instance arguments. *)
;;

val get_parameter_value : genr_info -> ident -> code_value
(** [get_parameter_value genr_info ident] return the explicit value bound to
  parameter [ident] in a Simulink block instance parameters. *)

(** As a general hint, function [get_t ident v] returns a Caml value of type
  [t] from the value [v] bound to a Simulink block parameter [ident]. *)

val get_string : ident -> code_value -> string
(** [get_string ident v] return the Caml [string] found in the explicit value
  [v], bound to [ident] in a Simulink block *)
;;

val get_bool : ident -> code_value -> bool
(** [get_bool ident v] return the Caml [bool] found in the explicit value
  [v] bound to [ident] in a Simulink block *)
;;

val get_int : ident -> code_value -> int
(** [get_int ident v] return the Caml [int] value found in the explicit
  value [v] bound to [ident] in a Simulink block. *)
;;
val get_float : ident -> code_value -> float
(** [get_float ident v] return the Caml [float] value found in the explicit
  value [v] bound to [ident] in a Simulink block. *)
;;

(** Defining usual constants for Scicos block definitions *)

(** Rounding methods. *)

val floor_round_method : code_value;;

(** External activation values. *)

val external_activation_mode_on : code_value;;
val external_activation_mode_off : code_value;;
val external_activation_mode_dummy : code_value
(** A dummy external activation mode to allow a single profile for Scicos
  functions that do not handle an external activation parameter. *)
;;

(** Saturation mode values. *)

val saturation_mode_off : Genr_types.simulink_saturation;;
val saturation_mode_on : Genr_types.simulink_saturation;;

(** {6 Retrieving Simulink block parameter binding values} *)

val get_argument_value_string : genr_info -> ident -> string;;

val get_parameter_value_string : genr_info -> ident -> string;;

val get_argument_value_expression : genr_info -> ident -> code_value;;

val get_argument_value_expression_vector : genr_info -> ident -> code_value array;;

val get_argument_value_code : genr_info -> ident -> code_value;;

val rewrite_expression : (string -> string) -> ident -> code_value -> code_value;;
val rewrite_expression_vector :
    (string list -> string) -> ident -> code_value -> code_value
;;

val rewrite_script : (string -> string) -> ident -> code_value -> code_value;;

(** {6 Simulink to Scicos type correspondance} *)

val get_simulink_out_data_type :
  genr_info -> Genr_types.simulink_out_data_type
;;

val string_of_simulink_out_data_type :
  Genr_types.simulink_out_data_type -> string
;;

(** {6 Helpers for block generation} *)

val concat_instruction_lists :
  Instructions.t list list -> Instructions.code
(** Catenate a list of lists of instructions into a piece of code.
  Used typically to build the entire code of a super block body, out of the
  list of its internal components (each component being implemented by a
  list of instructions).
*)
;;

val mdl_port_number_of_natural : natural -> Ast_types.mdl_port_number;;

val port_1_ : Ast_types.mdl_port_number;;
val port_2_ : Ast_types.mdl_port_number;;
val port_3_ : Ast_types.mdl_port_number;;
val port_4_ : Ast_types.mdl_port_number
(** Some useful predefined Simulink block port. *)
;;

val port_l1_ : connection_port_desc_generator;;
val port_l2_ : connection_port_desc_generator;;
val port_l3_ : connection_port_desc_generator;;
val port_l4_ : connection_port_desc_generator

val port_r1_ : connection_port_desc_generator;;
val port_r2_ : connection_port_desc_generator;;
val port_r3_ : connection_port_desc_generator;;
val port_r4_ : connection_port_desc_generator
(** Some useful predefined Simscape block port. *)
;;

val default_block_port_size : Instructions.modeler_size;;

val block_name : string -> Ast_types.mdl_name;;

val blk0 : Instructions.host_block_ident;;
val scsm0 : Instructions.host_diagram_ident;;

val set_block_geometry :
  host_block_ident -> natural * natural ->
  origin -> Instructions.instruction list
;;

val genr_add_block :
  Instructions.host_diagram_ident -> host_block_ident ->
  Mdl_translation.block_name -> instruction
;;

val add_block_body :
  host_diagram_ident ->
  host_block_ident ->
  Genr_types.block_string -> instruction list -> natural * natural ->
  origin -> instruction list
;;

val add_block_body_in_scsm0 :
  Genr_types.block_string -> instruction list -> natural * natural ->
  origin -> instruction list
;;

val add_directed_link_in_scsm0 :
  block_string -> explicit_port_number ->
  block_string -> explicit_port_number ->
  Instructions.modeler_point list -> instruction
;;

 val add_connection_link_in_scsm0 :
  block_string -> Mdl_rename_ast.rename_connection_connector_desc ->
  block_string -> Mdl_rename_ast.rename_connection_connector_desc ->
  Instructions.modeler_point list -> instruction
;;

val add_event_link_in_scsm0 :
  block_string -> explicit_port_number ->
  block_string -> explicit_port_number ->
  Instructions.modeler_point list -> instruction
;;

(** {6 Ubiquitous Simulink block property value retrieving functions} *)

val get_sample_time : genr_info -> string;;
val get_sample_time_value : genr_info -> code_value
(** Sample time retrieving. *)
;;
val get_system_sample_time : genr_info -> string;;
val get_system_sample_time_value : genr_info -> code_value
(** System Sample time corresponds to the sample time of the Atomic
  sub-system. *)
;;

val get_zero_crossing : genr_info -> bool;;
val get_zero_crossing_value : genr_info -> code_value
(** Zero crossing retrieving. *)
;;

val get_simulink_saturation :
  genr_info -> Genr_types.simulink_saturation
(** Saturation retrieving. *)
;;

val get_show_enable_port : genr_info -> bool;;
val get_show_enable_port_value : genr_info -> code_value
(** Show enable port retrieving. *)
;;

val get_show_saturation_port : genr_info -> bool;;
val get_show_saturation_port_value : genr_info -> code_value
(** Show saturation port retrieving. *)
;;

(** {6 Argument binding value retrieving functions} *)

val get_ident_string : genr_info -> ident -> string;;
val get_ident_bool : genr_info -> ident -> bool;;
val get_ident_int : genr_info -> ident -> int;;
val get_ident_int_vector : genr_info -> ident -> int array;;
val get_ident_simulink_saturation :
  genr_info -> ident -> Genr_types.simulink_saturation
;;


(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
