(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: make_genr_utilities.ml,v 1.1 2019-05-10 14:07:14 jpc Exp $ *)

open Concat_list
open Ast_types
open Instructions
open Unatural
;;

module
  Make (Target_modeler : Make_genr_utilities_spec.MODELER) =
  struct

  include Genr_utilities_common;;
  include Target_modeler;;

  Genr_utils.set_modeler_input_port_type
    Target_modeler.target_modeler
    Target_modeler.modeler_input_port_type;
  Genr_utils.set_modeler_output_port_type
    target_modeler Target_modeler.modeler_output_port_type;
  Genr_utils.set_modeler_annotation_type
    target_modeler Target_modeler.modeler_annotation_type
  ;;

  let get_simulink_out_data_type_value genr_info =
    let simulink_out_data_type =
      Genr_utilities_common.get_simulink_out_data_type genr_info in
    Target_modeler.
    code_out_simulink_out_data_type simulink_out_data_type
  ;;

  let get_simulink_saturation_value genr_info =
    Target_modeler.code_out_simulink_saturation
      (Genr_utilities_common.get_simulink_saturation genr_info)
  ;;

  (** Translation of mdl blocks that are just aliases for modeler blocks *)

  let simulink_block_reference_type =
    Genr_utils.simulink_block_reference_type
  ;;

  let simulink_block_type =
    Genr_utils.simulink_block_type
  ;;

  let find_alias genr_info aliases =
    let rec loop argument_flag = function
      | [] ->
        (* If we were looking for an argument,
               now we try to find a parameter. *)
        if argument_flag then loop false aliases else
          (* Otherwise, we already looked everywhere
               and this is fatal. *)
          let aliases =
            String.concat ", " (List.map string_of_ident aliases) in
          Genr_utilities_common.fatal_error_with_info
            genr_info
            (Printf.sprintf
               "none of %s arguments where found in block."
               aliases)
      | aliase :: aliases ->
        if argument_flag then
          if
            Genr_utilities_common.exists_argument genr_info aliase
          then
            Genr_utilities_common.get_argument_value genr_info aliase
          else
            loop argument_flag aliases
        else
          if
            Genr_utilities_common.exists_parameter genr_info aliase
          then
            Genr_utilities_common.get_parameter_value genr_info aliase
          else
            loop argument_flag aliases in
    loop true aliases
  ;;

  let get_modeler_block_value genr_info =

    let rec get_block_value = function
      | Genr_types.Simulink_arg_sample_time ->
        Genr_utilities_common.get_sample_time_value genr_info
      | Genr_types.Simulink_arg_out_data_type ->
        get_simulink_out_data_type_value genr_info
      | Genr_types.Simulink_arg_zero_crossing ->
        Genr_utilities_common.get_zero_crossing_value genr_info
      | Genr_types.Simulink_arg_saturation ->
        get_simulink_saturation_value genr_info
      | Genr_types.Simulink_arg_saturation_set sat ->
        Target_modeler.code_out_simulink_saturation sat
      | Genr_types.Simulink_arg_expression arg ->
        let v =
          Genr_utilities_common.
          get_argument_value_expression genr_info arg in
        Genr_utilities_common.fdebug
          (fun epf ->
           Format.fprintf epf
             "get_block_value Simulink_arg_expression %s %a"
             (Ast_types.string_of_ident arg)
             Mdl_deep_ast_print.print_deep_value v);
        v
      | Genr_types.Simulink_arg_value arg ->
        Genr_utilities_common.get_argument_value genr_info arg
      | Genr_types.Simulink_arg_value_aliases args ->
        find_alias genr_info args
      | Genr_types.Simulink_arg_expression_aliases args ->
        let main_alias =
          match args with
          | [] -> failwith "empty alias list"
          | main_alias :: _ -> main_alias in
        let value = find_alias genr_info args in
        Mdl_basics_deep_ast.Deep_expression
          (Ast_types.mdl_text_of_string
             (Genr_utilities_common.get_string main_alias value))
      | Genr_types.Simulink_arg_with_default (v, arg) ->
        begin
          try
            get_block_value arg
          with
          | Not_found -> v
        end
      | Genr_types.Simulink_arg_fun1 (get_fun, arg1) ->
        get_fun
          (get_block_value arg1)
      | Genr_types.Simulink_arg_fun2 (get_fun, arg1, arg2) ->
        get_fun
          (get_block_value arg1)
          (get_block_value arg2)
      | Genr_types.Simulink_arg_fun3 (get_fun, arg1, arg2, arg3) ->
        get_fun
          (get_block_value arg1)
          (get_block_value arg2)
          (get_block_value arg3)
      | Genr_types.Code_value v -> v in

    get_block_value
  ;;

  let add_set_block_parameters
      (args : Genr_types.modeler_block_parameters)
      code block_id genr_info =

    let get_parameter_value sbvs =
      Ulist1.map
        (fun (sid, sbv) ->
         sid, List.map (get_modeler_block_value genr_info) sbv)
        sbvs in

    match args with
    | [] -> code
    | arg :: args ->
      let modeler_block_parameter_bindings =
        Ulist1.map
          (function
           | Genr_types.Value_parameter (sid, sbv) ->
             let v = get_modeler_block_value genr_info sbv in
             Genr_utilities_common.block_parameter_binding (sid, v)
           | Genr_types.Struct_parameter (sid, sbv) ->
             let v = get_parameter_value sbv in
             Genr_utilities_common.struct_parameter_binding (sid, v))
          (arg, args) in

      Set_block_parameters (block_id, modeler_block_parameter_bindings) :: code
  ;;

  let genr_instructions_for_modeler_block_type modeler_block_type =
    fun args code block_id genr_info ->
      Instantiate_block (block_id, modeler_block_type) ::
      add_set_block_parameters args code block_id genr_info
  ;;

  let genr_instructions_for simulink_block_type =
    let modeler_block_type =
      Target_modeler.modeler_block_type_for simulink_block_type in
    genr_instructions_for_modeler_block_type modeler_block_type
  ;;

  let genr_code_for_modeler_block_type modeler_block_type =
    fun args code block_id genr_info ->
      list (
        Instantiate_block (block_id, modeler_block_type) ::
        add_set_block_parameters args code block_id genr_info)
  ;;

  let genr_code_for simulink_block_type =
    let modeler_block_type =
      Target_modeler.modeler_block_type_for simulink_block_type in
    genr_code_for_modeler_block_type modeler_block_type
  ;;

  let add_translator =
    Genr_block.add_translator Target_modeler.target_modeler
  ;;

  (* Translation registering utilities *)
  let add_modeler_translation simulink_block_type args code =
    (* Warning: in the definition of genr, the eta expansion is mandatory,
       since we must delay the generator registration until
       Driver.main is running. *)
    let genr block_id =
      genr_code_for simulink_block_type args code block_id in
    add_translator simulink_block_type genr
  ;;

  let add_modeler_equivalent simulink_block_type =
    add_modeler_translation simulink_block_type [] []
  ;;

  let add_modeler_conditional_translation_gen applyer
      simulink_block_type
      get_block_type_and_block_args
      args code =

    let genr block_id genr_info =
      let modeler_block_type, modeler_args =
        applyer get_block_type_and_block_args genr_info args in
      genr_code_for_modeler_block_type
        modeler_block_type modeler_args code block_id genr_info in

    add_translator simulink_block_type genr
  ;;

  let add_modeler_conditional_translation1 =
    let applyer get_block_type_and_block_args genr_info arg =
      get_block_type_and_block_args
        (get_modeler_block_value genr_info arg) in
    add_modeler_conditional_translation_gen applyer
  ;;

  let add_modeler_conditional_translation2 =
    let applyer
        get_modeler_block_type_and_block_args genr_info (arg1, arg2) =
      get_modeler_block_type_and_block_args
        (get_modeler_block_value genr_info arg1)
        (get_modeler_block_value genr_info arg2) in
    add_modeler_conditional_translation_gen applyer
  ;;

  let add_modeler_conditional_translation3 =
    let applyer
      get_modeler_block_type_and_block_args genr_info (arg1, arg2, arg3) =
      get_modeler_block_type_and_block_args
        (get_modeler_block_value genr_info arg1)
        (get_modeler_block_value genr_info arg2)
        (get_modeler_block_value genr_info arg3) in
    add_modeler_conditional_translation_gen applyer
  ;;

  let add_modeler_conditional_translation =
    let applyer get_modeler_block_type_and_block_args genr_info args =
      get_modeler_block_type_and_block_args
        (List.map (get_modeler_block_value genr_info) args) in
    add_modeler_conditional_translation_gen applyer
  ;;

  let inherit_mode_type =
    code_out_simulink_out_data_type Genr_types.Simulink_inherit
  ;;
  let double_type =
    code_out_simulink_out_data_type Genr_types.Simulink_double
  ;;

  (* Regular ports *)

  (* Regular input ports *)
  let in_port scsm_id block_id block_string i origin =
    let nports = (Unatural._0_, Unatural._1_) in
    let body = Target_modeler.in_port_body block_id i in
    Genr_utilities_common.add_block_body
      scsm_id block_id block_string body nports origin
  ;;

  let in_port_in_scsm0 =
    in_port Genr_utilities_common.scsm0 Genr_utilities_common.blk0
  ;;

  let in_ports scsm_id blk_ident nin =

    let rec loop in_strings accu i =
      if i > nin then in_strings, accu else
        let block_string = Printf.sprintf "in%d" (Unatural.to_int i) in
        let origin = (_0_, Unatural.sub nin i) in
        let lst =
          in_port scsm_id blk_ident block_string
            (Genr_utilities_common.mdl_port_number_of_natural i) origin in
        loop
          (block_string :: in_strings)
          (concat accu (list lst))
          (Unatural.succ i) in

    let in_strings, code = loop [] (list []) _1_ in
    List.rev in_strings, code
  ;;

  let in_ports_in_scsm0 = in_ports scsm0 blk0;;

  (* Regular output ports *)
  let out_port scsm_id block_id block_string i origin =
    let body = Target_modeler.out_port_body block_id i in
    let nports = (_1_, _0_) in
    Genr_utilities_common.add_block_body
      scsm_id block_id block_string body nports origin
  ;;

  let out_port_in_scsm0 = out_port scsm0 blk0;;

  let out_ports width scsm_id blk_ident nout =
    let rec loop out_strings accu i =
      if i < _1_ then out_strings, accu else
        let block_string = Printf.sprintf "out%d" (Unatural.to_int i) in
        let origin = (width, Unatural.sub nout i) in
        let lst =
          out_port scsm_id blk_ident block_string
            (Genr_utilities_common.mdl_port_number_of_natural i)
            origin in
        loop
          (block_string :: out_strings) (cons_list lst accu)
          (Unatural.pred i) in
    let out_strings, code = loop [] (list []) nout in
    List.rev out_strings, code
  ;;

  let out_ports_in_scsm0 width = out_ports width scsm0 blk0;;
  (* End of regular input/output ports *)

  (* Implicit ports *)

  (* Implicit input ports *)
  let implicit_in_port scsm_id block_id block_string i origin =
    let nports = (_0_, _1_) in
    let body = Target_modeler.implicit_in_port_body block_id i in
    Genr_utilities_common.add_block_body
      scsm_id block_id block_string body nports origin
  ;;

  let implicit_in_port_in_scsm0 = implicit_in_port scsm0 blk0;;

  let implicit_in_ports scsm_id blk_ident nin niin =
    let rec loop in_strings accu i =
      let n = add nin niin in
      if i > niin then in_strings, accu else
        let nin_i = Unatural.add i  nin in
        let block_string = Printf.sprintf "in%d" (Unatural.to_int nin_i) in
        let origin = (_0_, Unatural.sub n i) in
        let lst =
          implicit_in_port scsm_id blk_ident block_string
            (Genr_utilities_common.mdl_port_number_of_natural nin_i) origin in
        loop
          (block_string :: in_strings)
          (concat accu (list lst))
          (Unatural.succ i) in

    let in_strings, code = loop [] (list []) _1_ in
    List.rev in_strings, code
  ;;

  let implicit_in_ports_in_scsm0 = implicit_in_ports scsm0 blk0;;

  (* Implicit output ports *)
  let implicit_out_port scsm_id block_id block_string i origin =
    let body = Target_modeler.implicit_out_port_body block_id i in
    let nports = (_1_, _0_) in
    Genr_utilities_common.add_block_body
      scsm_id block_id block_string body nports origin
  ;;

  let implicit_out_port_in_scsm0 = implicit_out_port scsm0 blk0;;

  let implicit_out_ports width scsm_id blk_ident nout niout =
    let rec loop out_strings accu i =
      let n = Unatural.add nout niout in
      if i < _1_ then out_strings, accu else
        let nout_i = Unatural.add i nout in
        let block_string =
          Printf.sprintf "out%d" (Unatural.to_int nout_i) in
        let origin = (width, Unatural.sub n i) in
        let lst =
          implicit_out_port scsm_id blk_ident block_string
            (Genr_utilities_common.mdl_port_number_of_natural nout_i)
            origin in
        loop
          (block_string :: out_strings) (cons_list lst accu)
          (Unatural.pred i) in
    let out_strings, code = loop [] (list []) niout in
    List.rev out_strings, code
  ;;

  let implicit_out_ports_in_scsm0 width =
    implicit_out_ports width scsm0 blk0
  ;;
  (* End of implicit ports *)

  (* Event ports. *)

  (* Event input ports. *)
  let in_port_event scsm_id block_id block_string i origin =
    let body = Target_modeler.in_activation_port_body block_id i in
    let nports = (_0_, _0_) in
    Genr_utilities_common.add_block_body
      scsm_id block_id block_string body nports origin
  ;;

  let in_port_event_in_scsm0 = in_port_event scsm0 blk0;;

  let in_port_events width scsm_id blk_ident nin =
    let rec loop accu i =
      if i < _1_ then accu else
        let block_string = Printf.sprintf "in%d" (Unatural.to_int i) in
        let origin = (width, Unatural.sub nin i) in
        let lst =
          in_port_event scsm_id blk_ident block_string
            (Genr_utilities_common.mdl_port_number_of_natural i)
            origin in
        loop (cons_list lst accu) (Unatural.pred i) in
    loop (list []) nin
  ;;

  let in_port_events_in_scsm0 width = in_port_events width scsm0 blk0;;

  (* Event output ports. *)
  let out_port_event scsm_id block_id block_string i origin =
    let body = Target_modeler.out_activation_port_body block_id i in
    let nports = (_0_, _0_) in
    Genr_utilities_common.add_block_body
      scsm_id block_id block_string body nports origin
  ;;

  let out_port_event_in_scsm0 = out_port_event scsm0 blk0;;

  let out_port_events width scsm_id blk_ident nout =
    let rec loop accu i =
      if i < _1_ then accu else
        let block_string = Printf.sprintf "out%d" (Unatural.to_int i) in
        let origin = (width, Unatural.sub nout i) in
        let lst =
          out_port_event scsm_id blk_ident block_string
            (Genr_utilities_common.mdl_port_number_of_natural i)
            origin in
        loop (cons_list lst accu) (Unatural.pred i) in
    loop (list []) nout
  ;;

  let out_port_events_in_scsm0 width = out_port_events width scsm0 blk0;;
  (* End of event ports. *)

  let split scsm_id block_id block_string origin =
    let block_name = Genr_utilities_common.block_name block_string in
    let nports = (_1_, Unatural._3_) in
    Instantiate_block (block_id, modeler_split_type) ::
    set_block_geometry block_id nports origin @ [
      Genr_utilities_common.
      genr_add_block scsm_id block_id block_name;
    ]
  ;;

  let split_in_scsm0 = split scsm0 blk0;;

  let implicit_split scsm_id block_id block_string origin =
    let block_name = Genr_utilities_common.block_name block_string in
    let nports = (_1_, Unatural._3_) in
    Instantiate_block (block_id, modeler_implicit_split_type) ::
    set_block_geometry block_id nports origin @ [
      Genr_utilities_common.
      genr_add_block scsm_id block_id block_name;
    ]
  ;;

  let implicit_split_in_scsm0 = implicit_split scsm0 blk0;;

  let split_event scsm_id block_id block_string origin =
    let block_name = block_name block_string in
    let nports = (_0_, _0_) in
    Instantiate_block (block_id, modeler_clk_split_type) ::
    set_block_geometry block_id nports origin @ [
      genr_add_block scsm_id block_id block_name;
    ]
  ;;

  let split_event_in_scsm0 = split_event scsm0 blk0;;

  let sample_clock scsm_id block_id block_string sample_time origin =
    let block_name = block_name block_string in
    let modeler_parameter_bindings =
      block_parameter_binding (Si "frequ", sample_time), [
        block_parameter_binding
          (Si "offset", Genr_utilities_common.code_out_int 0);
      ] in
    let nports = (_0_, _0_) in
    Instantiate_block (block_id, modeler_sample_clk_type) ::
    set_block_geometry block_id nports origin @ [
      Set_block_parameters (block_id, modeler_parameter_bindings);
      genr_add_block scsm_id block_id block_name;
    ]
  ;;

  let sample_clock_in_scsm0 = sample_clock scsm0 blk0;;

  let virtual_clk0 scsm_id block_id block_string origin =
    let block_name = block_name block_string in
    let nports = (_0_, _0_) in
    Instantiate_block (block_id, modeler_virtual_clk0_type) ::
    set_block_geometry block_id nports origin @ [
      genr_add_block scsm_id block_id block_name;
    ]
  ;;

  let virtual_clk0_in_scsm0 = virtual_clk0 scsm0 blk0;;

  let samphold_in_scsm0 = samphold scsm0 blk0;;
  let sampholds_in_scsm0 = sampholds scsm0 blk0;;
  let time_f_in_scsm0 = time_f scsm0 blk0;;
  let relational_op_in_scsm0 = relational_op scsm0 blk0;;

  let convert_in_scsm0 block_string out_data_type genr_info =
    let saturation =
      Genr_utilities_common.get_simulink_saturation genr_info in
    Target_modeler.convert_in_scsm0_with_saturation
      block_string out_data_type saturation
  ;;

  let super_block_in scsm_id block_id body =
    let prelude = [
      Instantiate_super_block block_id;
      Instantiate_diagram scsm_id;
    ] in

    let postlude = [
      Fill_super_block (block_id, scsm_id);
    ] in

    concat
      (Genr_utilities_common.concat_instruction_lists (prelude :: body))
      (list postlude)
  ;;

  let super_block_in_scsm0 = super_block_in scsm0;;

  let set_diagram_context_in_scsm0 ctx = Set_diagram_context (scsm0, ctx);;

end
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
