(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: scicos_pro_variant.ml,v 1.1 2019-05-10 14:07:14 jpc Exp $ *)

open Mdl_basics_explicit_ast;;
open Mdl_block_types;;
open Instructions;;
open Genr_types;;
open Genr_utils;;
open Unatural;;

module M =
  struct

    let target_modeler = Simport_configuration.Modeler_scicos_pro;;

    let code_out_simulink_saturation =
      let scicos_string_of_saturation = function
        | Genr_types.Saturation_off -> "Nothing"
        | Genr_types.Saturation_on -> "Saturate" in
      (fun sat ->
       Genr_utilities_common.code_out_string (scicos_string_of_saturation sat))
    ;;

    let scicos_pro_input_port_type =
      Instructions.Sbt "system/Ports/Input"
    and scicos_pro_output_port_type =
      Sbt "system/Ports/Output"
    ;;

    let scicos_pro_implicit_input_port_type =
      Sbt "system/Ports/ImplicitInput"
    and scicos_pro_implicit_output_port_type =
      Sbt "system/Ports/ImplicitOutput"
    ;;

    let scicos_pro_input_activation_port_type =
      Sbt "system/Ports/EventInput"
    and scicos_pro_output_activation_port_type =
      Sbt "system/Ports/EventOutput"
    ;;

    let scicos_pro_annotation_type =
      Sbt "TEXT_f"
    ;;
    let scicos_pro_samphold_type =
      Sbt "system/Hybrid/SampleHold"
    ;;
    let scicos_pro_split_type =
      Sbt "system/Links/Split"
    ;;
    let scicos_pro_implicit_split_type =
      Sbt "system/Links/ImplicitSplit"
    ;;
    let scicos_pro_clk_split_type =
      Sbt "system/Links/ActivationSplit"
    ;;
    let scicos_pro_atomic_port_type =
      scicos_pro_input_activation_port_type
    ;;
    let scicos_pro_sample_clk_type =
      Sbt "system/ActivationOperations/SampleClock"
    ;;
    let scicos_pro_time_type =
      Sbt "system/SignalGenerators/Time"
    ;;
    let scicos_pro_virtual_clk0_type =
      Sbt "system/ActivationOperations/EventPortZero"
    ;;

    let scicos_pro_relational_op_type =
      Sbt "system/LogicalOperations/RelationalOp"
    ;;

    let scicos_pro_compare_to_constant_type =
      Sbt "system/LogicalOperations/CompareToConstant"
    ;;

    let scicos_pro_math_expression_type =
      Sbt "system/MathOperations/MathExpression"
    ;;

    let scicos_pro_unit_conversion_type =
      Sbt "system/SignalConversions/UnitConversion"
    ;;

    let scicos_pro_edge_type =
      Sbt "system/ActivationOperations/EdgeTrigger"
    ;;

    let scicos_pro_extract_activation_type =
      Sbt "system/ActivationOperations/ExtractActivation"
    ;;

    let scicos_pro_if_then_else_type =
      Sbt "system/ActivationOperations/IfThenElse"
    ;;

    let scicos_pro_interpreted_function_type =
      Sbt "system/CustomBlocks/OmlCustomBlock"
    ;;

    let get_relop_value genr_info relop =
      let relop =
        match
          Genr_utilities_common.get_argument_value_string genr_info relop
        with
        | "=="
        | "~="
        | "<"
        | "<="
        | ">"
        | ">=" as relop -> relop
        | "isNaN" ->
          (* isNaN is translated into x -> x <> x, have a look into
             genr_scicos_pro_relational_operator.ml *)
          "~="
        | s ->
          Genr_utilities_common.fatal_error_with_info genr_info
            (Printf.sprintf
               "unknown comparison operator ``%s''"
               s) in
      Genr_utilities_common.code_out_string relop
    ;;

    let code_out_simulink_out_data_type simulink_out_data_type =
      let scicos_pro_out_data_type =
        match simulink_out_data_type with
        | Genr_types.Simulink_single ->
          Genr_utilities_common.warning
            "out data type single is not supported, double is used instead.";
          "double"
        | Simulink_double -> "double"
        | Simulink_int8 -> "int8"
        | Simulink_uint8 -> "uint8"
        | Simulink_int16 -> "int16"
        | Simulink_uint16 -> "uint16"
        | Simulink_int32 -> "int32"
        | Simulink_uint32 -> "uint32"
        | Simulink_boolean -> "boolean"
        | Simulink_inherit -> "inherit" in
      Genr_utilities_common.code_out_string scicos_pro_out_data_type
    ;;

    let inherit_mode_type =
      code_out_simulink_out_data_type Simulink_inherit
    ;;

    let simport_block_translation_bug s =
      (* This is a bug: the compiler should normally compile this block
         specially *)
      ignore (Genr_utilities_common.fatal_error (Printf.sprintf
        "Fatal simport bug: code for block %s should have been generated by\
       \nthe compiler, but block %s was erroneously proposed to translation."
        s s));
      assert false
    ;;

    let scicos_pro_block_type_for_regular_block_type = function
      | Mdl_block_types.Abs ->
        Sbt "system/MathOperations/Abs"
      | ActionPort ->
        Say.fatal_error "Unexpected ActionPort encountered, please report."

      | Backlash ->
        Sbt "system/Hybrid/Backlash"
      | Bias ->
        Sbt "system/MathOperations/Bias"
      | BusAssignment ->
        Sbt "system/BusOperations/BusAssignment"
      | BusCreator ->
        Sbt "system/BusOperations/BusCreator"
      | BusSelector ->
        Sbt "system/BusOperations/BusExtractor"

      | Clock ->
        scicos_pro_time_type
      | CombinatorialLogic ->
        Sbt "system/LogicalOperations/TruthTable"
      | ComplexToRealImag ->
        Sbt "system/SignalConversions/ComplexDecomposition"
      | ComplexToMagnitudeAngle ->
        Sbt "system/SignalConversions/ComplexToPolar"
      | Concatenate ->
        Sbt "system/MatrixOperations/MatrixExpression"
      | Constant ->
        Sbt "system/SignalGenerators/Constant"
      | DataTypeConversion ->
        Sbt "system/SignalConversions/DatatypeConversion"

      | DeadZone ->
        Sbt "system/Hybrid/Deadzone"
      | Delay ->
        Sbt "system/Dynamical/ShiftRegister"
      | Demux ->
        Sbt "system/Routing/Demux"
      | Derivative ->
        Sbt "system/Dynamical/Derivative"
      | DigitalClock ->
        scicos_pro_time_type
      | DiscreteFilter ->
        Sbt "system/Dynamical/DiscrTransFunc"
      | DiscreteIntegrator ->
        Sbt "system/Dynamical/DiscreteIntegral"
      | DiscretePulseGenerator ->
        scicos_pro_math_expression_type
      | DiscreteStateSpace ->
        Sbt "system/Dynamical/DiscrStateSpace"
      | DiscreteTransferFcn ->
        Sbt "system/Dynamical/DiscrTransFunc"
      | DiscreteZeroPole ->
        Sbt "system/Dynamical/DiscrPoleZero"
      | Display ->
        Sbt "system/SignalViewers/Display"
      | DotProduct ->
        Sbt "system/MathOperations/Product"

      | ElementaryMath ->
        Sbt "system/MathOperations/TrigFunc"

      | Fcn ->
        scicos_pro_math_expression_type
      | From ->
        Sbt "system/Routing/GetSignal"
      | FromWorkspace ->
        Sbt "system/SignalImporters/SignalIn"

      | Gain ->
        Sbt "system/MathOperations/Gain"
      | Goto ->
        Sbt "system/Routing/SetSignal"
      | GotoTagVisibility ->
        Sbt "system/Routing/SignalScoping"
      | Ground ->
        Sbt "system/SignalGenerators/DefaultSource"

      | HitCross ->
        scicos_pro_edge_type

      | If ->
        Sbt "system/ActivationOperations/IfExpressions"
      | InitialCondition ->
        Sbt "system/Routing/SelectInput"
      | Inport -> scicos_pro_input_port_type
      | InportShadow ->
        simport_block_translation_bug "InportShadow"

      | Integrator ->
        Sbt "system/Dynamical/Integral"

      | Logic ->
        Sbt "system/LogicalOperations/LogicalOp"
      | Lookup ->
        Sbt "system/LookupTables/LookupTable"
      | Lookup2D
      | Lookup_n_D ->
        Sbt "system/LookupTables/LookupTable2D"

        (* Interpreted Matlab Fcn:
           Same as M_S_Function, but the function is a native Matlab
           function *)
      | MATLABFcn ->
        Sbt "system/MathOperations/OmlExpression"
      | MagnitudeAngleToComplex ->
        Sbt "system/SignalConversions/ComplexToPolar"
      | ManualSwitch ->
        Sbt "system/Routing/ManualSelect"
      | Math ->
        Sbt "system/MathOperations/MathFunc"
      | Memory ->
        simport_block_translation_bug "Memory"
      | Merge ->
        Sbt "system/Routing/SelectInput"
      | MinMax ->
        Sbt "system/MathOperations/MaxMin"
      | MultiPortSwitch ->
        Sbt "system/Routing/ConditionalNSelect"
      | Mux ->
        Sbt "system/Routing/Mux"

      | Outport -> scicos_pro_output_port_type

      | PMIOPort ->
        Sbt "system/Ports/ImplicitInput"

      | Polyval ->
        Sbt "system/MathOperations/Horner"
      | Product ->
        Sbt "system/MathOperations/Product"

      | Quantizer ->
        Sbt "system/Hybrid/Quantization"

      | RandomNumber ->
        Sbt "system/SignalGenerators/Random"
      | RateLimiter ->
        Sbt "system/Hybrid/RateLimiter"
      | RateTransition ->
        Sbt "system/Dynamical/DiscreteDelay"
      | RealImagToComplex ->
        Sbt "system/SignalConversions/ComplexDecomposition"
      | RelationalOperator ->
        scicos_pro_relational_op_type
      | Relay ->
        Sbt "system/Hybrid/Hysteresis"
      | Reshape ->
        Sbt "system/MatrixOperations/MatrixReshape"
      | Rounding ->
        Sbt "system/MathOperations/MathFunc"

      | Saturate ->
        Sbt "system/Hybrid/Saturation"
      | Scope ->
        Sbt "system/SignalViewers/Scope"
      | Selector ->
        Sbt "system/MatrixOperations/MatrixExtractor"
      | SignalGenerator ->
        Sbt "system/SignalGenerators/SineWaveGenerator"
      | Signum ->
        Sbt "system/MathOperations/Sign"
      | Sin ->
        Sbt "system/SignalGenerators/SineWaveGenerator"
      | Sqrt ->
        Sbt "system/MathOperations/Power"
      | StateSpace ->
        simport_block_translation_bug "SateSpace"
      | Step ->
        Sbt "system/SignalGenerators/StepGenerator"
      | Stop ->
        Sbt "system/ActivationOperations/End"
      | SubSystem ->
        simport_block_translation_bug "Subsystem"
      | Sum ->
        Sbt "system/MathOperations/Sum"
      | Switch ->
        Sbt "system/Routing/ConditionalSelect"
      | SwitchCase ->
        Sbt "system/ActivationOperations/IfExpressions"

      | Terminator ->
        Sbt "system/SignalProperties/Sink"
      | TimeScope ->
        Sbt "system/SignalViewers/Scope"
      | ToWorkspace ->
        Sbt "system/SignalExporters/SignalOut"
      | TransferFcn ->
        Sbt "system/Dynamical/ContTransFunc"
      | TransportDelay ->
        Sbt "system/Dynamical/FixedDelay"
      | TriggerPort ->
        scicos_pro_input_activation_port_type
      | Trigonometry ->
        Sbt "system/MathOperations/TrigFunc"

      | UnaryMinus ->
        Sbt "system/MathOperations/Negate"
      | UniformRandomNumber ->
        Sbt "system/SignalGenerators/Random"
      | UnitDelay ->
        Sbt "system/Dynamical/DiscreteDelay"

      | VariableTransportDelay ->
        Sbt "system/Dynamical/VariableDelay"

      | ZeroOrderHold ->
        scicos_pro_samphold_type
      | ZeroPole ->
        Sbt "system/Dynamical/ContPoleZero"

        (* Native blocks we do not translate *)
      | (Actuator |
         ArithShift |
         Assertion | Assignment | AsynchronousTaskSpecification |

        Buffer |
        BusToVector |

        ConstellationDiagram |

        DataStoreMemory | DataStoreRead | DataStoreWrite |
        DataTypeDuplicate |
        DiscreteFir |
        DSPFlip |

        EnablePort |

        Find |
        ForEach | ForIterator |
        FrameConversion | FromFile |
        FunctionCallFeedbackLatch | FunctionCallSplit |

        Interpolation_n_D |

        LookupNDDirect |

        (* Level 2 Matlab S-Function:
           A call to an external function defined in Matlab (the function
           name is given in the block) *)
        M_S_Function |
        MATLABSystem |
        ModelReference |

        PMComponent |
        PermuteDimensions | PreLookup | Probe |

        Reference |

        (* S-Function block (in the interface):
           A call to an external function defined in another langage
           and probably compiled (the function name is given in the block)

           Note: the MATLAB Function block (in the interface):
           A special case of S_Function, the code is given in the block but
           the MATLAB Function block generates a S_Function with a
           stateflow section that embbeds the function definition.

        *)
        S_Function |
        SampleTimeMath |
        SecondOrderIntegrator | Sensor |
        SignalConversion |
        SignalFromWorkspace | SignalSpecification |
        SignalToWorkspace |
        SignalViewerScope | SpectrumAnalyzer |
        Squeeze |

        ToFile | TwoWayConnection |

        Unbuffer | UnitConversion |

        VideoViewer |

        Width | WhileIterator as ty) ->
      Genr_utilities_common.fatal_error (Printf.sprintf
        "Mdl block %s has no Activate counterpart yet"
        (Mdl_block_types_tables.string_of_block_type ty))
    ;;

    let reference_block_type_bindings = [
      Mdl_fake_block_types.
      split_block_reference_type, scicos_pro_split_type;
      Mdl_fake_block_types.
      split_connection_block_reference_type, scicos_pro_implicit_split_type;
      Mdl_fake_block_types.
      atomic_port_block_reference_type, scicos_pro_atomic_port_type;
      Mdl_fake_block_types.
      atomic_event_generator_block_reference_type, scicos_pro_atomic_port_type;
      Mdl_fake_block_types.
      if_action_event_generator_block_reference_type,
        scicos_pro_if_then_else_type;
      Mdl_fake_block_types.
      enable_event_generator_block_reference_type,
        scicos_pro_if_then_else_type;
      Mdl_fake_block_types.
      trigger_rising_event_generator_block_reference_type,
        scicos_pro_edge_type;
      Mdl_fake_block_types.
      trigger_falling_event_generator_block_reference_type,
        scicos_pro_edge_type;
      Mdl_fake_block_types.
      trigger_either_event_generator_block_reference_type,
        scicos_pro_edge_type;
      Mdl_fake_block_types.
      trigger_function_call_generator_block_reference_type,
        scicos_pro_extract_activation_type;
      Mdl_fake_block_types.
      trigger_and_enable_rising_event_generator_block_reference_type,
        scicos_pro_edge_type;
      Mdl_fake_block_types.
      trigger_and_enable_falling_event_generator_block_reference_type,
        scicos_pro_edge_type;
      Mdl_fake_block_types.
      trigger_and_enable_either_event_generator_block_reference_type,
        scicos_pro_edge_type;
      Mdl_fake_block_types.
      trigger_and_enable_function_call_generator_block_reference_type,
        scicos_pro_extract_activation_type;
      Mdl_fake_block_types.
      matlab_function_block_reference_type,
        scicos_pro_interpreted_function_type;

      {
        Mdl_deep_ast.
        deep_block_source_type = Ast_types.mdl_name_of_string
          "Algebraic Constraint";
        deep_block_source_path = Ast_types.mdl_path_of_string
          "simulink/Math\nOperations/Algebraic Constraint";
      },
      Sbt "system/MathOperations/Constraint";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Band-Limited White Noise.";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Sources/Band-Limited\nWhite Noise";
      },
      Sbt "system/SignalGenerators/Random";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Cart2Polar";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink_extras/Transformations/Cartesian to\nPolar";
      },
      Sbt "system/SignalConversions/CartesianToPolar";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "CelsiusToFahrenheit";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink_extras/Transformations/Celsius to\nFahrenheit";
      },
      scicos_pro_unit_conversion_type;

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Checks_DGap";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Model\nVerification/Check \nDynamic Gap";
      },
      Sbt "Checks_DGap";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Checks_DMin";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Model\nVerification/Check Dynamic \nLower Bound";
      },
      Sbt "Checks_DMin";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Checks_DMax";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Model\nVerification/Check Dynamic \nUpper Bound";
      },
      Sbt "Checks_DMax";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Checks_DRange";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Model\nVerification/Check \nDynamic Range";
      },
      Sbt "Checks_DRange";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Checks_Gradient";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Model\nVerification/Check\nDiscrete Gradient";
      },
      Sbt "Checks_Gradient";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Checks_Resolution";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Model\nVerification/Check Input \nResolution";
      },
      Sbt "Checks_Resolution";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Checks_SGap";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Model\nVerification/Check \nStatic Gap";
      },
      Sbt "Checks_SGap";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Checks_SMin";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Model\nVerification/Check Static \nLower Bound";
      },
      Sbt "Checks_SMin";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Checks_SMax";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Model\nVerification/Check Static \nUpper Bound";
      },
      Sbt "Checks_SMax";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Checks_SRange";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Model\nVerification/Check \nStatic Range";
      },
      Sbt "Checks_SRange";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Compare To Constant";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Logic and Bit\nOperations/Compare\nTo Constant";
      },
      scicos_pro_relational_op_type;

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Compare To Zero";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Logic and Bit\nOperations/Compare\nTo Zero";
      },
      scicos_pro_relational_op_type;

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Matrix Concatenation";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Math\nOperations/Matrix\nConcatenation";
      },
      Sbt "system/MatrixOperations/MatrixExpression";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Matrix Concatenation";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink3/Signals\n& Systems/Matrix\nConcatenation"
      },
      Sbt "system/MatrixOperations/MatrixExpression";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Coulombic and Viscous Friction";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Discontinuities/Coulomb &\nViscous Friction";
      },
      Sbt "system/MathOperations/Gain";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Counter Limited";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Sources/Counter\nLimited";
      },
      Sbt "system/SignalGenerators/ModuloCounter";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Cross Product";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "aerospace/Axes\nTransformations/3x3 cross product";
      },
      Sbt "system/MatrixOperations/MatrixExpression";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Data Type Duplicate";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Signal\nAttributes/Data Type\nDuplicate";
      },
      Sbt "system/SignalProperties/TypeConstraint";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Dead Zone Dynamic";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Discontinuities/Dead Zone\nDynamic";
      },
      Sbt "system/Hybrid/ControlledDeadzone";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Detect Change";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Logic and Bit\nOperations/Detect\nChange";
      },
      scicos_pro_relational_op_type;

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Detect Decrease";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Logic and Bit\nOperations/Detect\nDecrease";
      },
      scicos_pro_relational_op_type;

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Detect Increase";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Logic and Bit\nOperations/Detect\nIncrease";
      },
      scicos_pro_relational_op_type;

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Detect Fall Negative";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Logic and Bit\nOperations/Detect Fall\nNegative";
      },
      scicos_pro_compare_to_constant_type;

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Detect Fall Nonpositive";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Logic and Bit\nOperations/Detect Fall\nNonpositive";
      },
      scicos_pro_compare_to_constant_type;

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Detect Rise Positive";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Logic and Bit\nOperations/Detect Rise\nPositive";
      },
      scicos_pro_compare_to_constant_type;

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Detect Rise Nonnegative";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Logic and Bit\nOperations/Detect Rise\nNonnegative";
      },
      scicos_pro_compare_to_constant_type;

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "DegreesToRadians";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink_extras/Transformations/Degrees to\nRadians";
      },
      scicos_pro_unit_conversion_type;

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "DFlipFlop";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink_extras/Flip Flops/D Flip-Flop";
      },
      Sbt "system/FlipFlops/DFlipFlop";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Difference";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Discrete/Difference";
      },
      Sbt "system/Dynamical/DiscreteDelay";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Discrete Derivative";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Discrete/Discrete Derivative";
      },
      Sbt "system/Dynamical/DiscreteDerivative";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "DLatch";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink_extras/Flip Flops/D Latch";
      },
      Sbt "system/FlipFlops/DLatch";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Dot Product";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Math\nOperations/Dot Product";
      },
      Sbt "system/MathOperations/Product";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "FahrenheitToCelsius";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink_extras/Transformations/Fahrenheit\nto Celsius";
      },
      scicos_pro_unit_conversion_type;

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Function-Call Generator";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Ports &\nSubsystems/Function-Call\nGenerator";
      },
      Sbt "system/SignalGenerators/Constant";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Integer Delay";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Discrete/Integer Delay";
      },
      Sbt "system/Dynamical/ShiftRegister";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Interval Test";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Logic and Bit\nOperations/Interval Test";
      },
      scicos_pro_compare_to_constant_type;

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "JKFlipFlop";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink_extras/Flip Flops/J-K\nFlip-Flop";
      },
      Sbt "system/FlipFlops/JKFlipFlop";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "LTI Block";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "cstblocks/LTI System";
      },
      Sbt "system/Dynamical/ContStateSpace";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Manual Switch";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Signal\nRouting/Manual Switch";
      },
      Sbt "system/Routing/ManualSelect";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Manual Switch";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink3/Nonlinear/Manual Switch";
      },
      Sbt "system/Routing/ManualSelect";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Signal\nRouting/Manual Switch";
      },
      Sbt "system/Routing/ManualSelect";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Unknown";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Signal\nRouting/Manual Switch";
      },
      Sbt "system/Routing/ManualSelect";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "MultiPlot Graph";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "RCTblocks/MultiPlot Graph";
      },
      Sbt "system/SignalViewers/Scope";

       {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "PID Controller";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink_extras/Additional\nLinear/PID Controller";
      },
      Sbt "system/Dynamical/PID";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "PID Controller";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink_need_slupdate/PID Controller";
      },
      Sbt "system/Dynamical/PID";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "RadiansToDegrees";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink_extras/Transformations/Radians\nto Degrees";
      },
      scicos_pro_unit_conversion_type;

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Ramp";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Sources/Ramp";
      },
      Sbt "system/SignalGenerators/Ramp";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Repeating table";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Sources/Repeating\nSequence";
      },
      Sbt "system/SignalGenerators/SignalGenerator";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Repeating Sequence Interpolated";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Sources/Repeating\nSequence\nInterpolated";
      },
      Sbt "system/SignalGenerators/SignalGenerator";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Repeating Sequence Stair";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Sources/Repeating\nSequence\nStair";
      },
      Sbt "system/Dynamical/ShiftRegister";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "PS-Simulink\nConverter";
        deep_block_source_path = Ast_types.mdl_path_of_string
          "nesl_utility/PS-Simulink\nConverter";
      },
      Sbt "system/Interfaces/FromModelica";

     {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Simulink-PS\nConverter";
        deep_block_source_path = Ast_types.mdl_path_of_string
          "nesl_utility/Simulink-PS\nConverter";
      },
      Sbt "system/Interfaces/ToModelica";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Reshape";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Math\nOperations/Reshape";
      },
      Sbt "system/MatrixOperations/MatrixReshape";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Reshape";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink3/Signals\n& Systems/Reshape";
      },
      Sbt "system/MatrixOperations/MatrixReshape";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Saturation Dynamic";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Discontinuities/Saturation\nDynamic";
      },
      Sbt "system/Hybrid/ControlledSaturation";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Slider Gain";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Math\nOperations/Slider\nGain";
      },
      Sbt "system/MathOperations/Gain";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "SRFlipFlop";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink_extras/Flip Flops/S-R\nFlip-Flop";
      },
      Sbt "system/FlipFlops/SRFlipFlop";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Tapped Delay Line";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Discrete/Tapped Delay";
      },
      Sbt "system/Buffers/Buffer";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Transfer Function with Initial Outputs";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink_extras/Additional\nLinear/Transfer Fcn\n\
                    (with initial outputs)";
      },
      Sbt "system/AdvancedDynamical/ContTransFunc_Y";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Transfer Function with Initial States";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink_extras/Additional\nLinear/Transfer Fcn\n\
                    (with initial states)";
      },
      Sbt "system/AdvancedDynamical/ContTransFunc_X";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Transpose";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "dspmtrx3/Transpose";
      },
      Sbt "system/MatrixOperations/Transpose";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Unary Minus";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Math\nOperations/Unary Minus";
      },
      Sbt "system/MathOperations/Negate";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "XY scope.";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "simulink/Sinks/XY Graph";
      },
      Sbt "system/SignalViewers/ScopeXY";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Electrical Reference";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "fl_lib/Electrical/Electrical Elements/Electrical Reference";
      },
      Sbt "Modelica 3.2.1/Electrical/Analog/Basic/Ground";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Resistor";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "fl_lib/Electrical/Electrical Elements/Resistor";
      },
      Sbt "Modelica 3.2.1/Electrical/Analog/Basic/Resistor";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Inductor";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "fl_lib/Electrical/Electrical Elements/Inductor";
      },
      Sbt "Modelica 3.2.1/Electrical/Analog/Basic/Inductor";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Voltage Sensor";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "fl_lib/Electrical/Electrical Sensors/Voltage Sensor";
      },
      Sbt "Modelica 3.2.1/Electrical/Analog/Sensors/VoltageSensor";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Current Sensor";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "fl_lib/Electrical/Electrical Sensors/Current Sensor";
      },
      Sbt "Modelica 3.2.1/Electrical/Analog/Sensors/CurrentSensor";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Controlled Voltage\nSource";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "fl_lib/Electrical/Electrical Sources/Controlled Voltage\nSource";
      },
      Sbt "Modelica 3.2.1/Electrical/Analog/Sources/SignalVoltage";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Controlled Current\nSource";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "fl_lib/Electrical/Electrical Sources/Controlled Current\nSource";
      },
      Sbt "Modelica 3.2.1/Electrical/Analog/Sources/SignalCurrent";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Rotational Damper";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "fl_lib/Mechanical/Rotational Elements/Rotational Damper";
      },
      Sbt "Modelica 3.2.1/Mechanics/Rotational/Components/Damper";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Mechanical\nRotational Reference";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "fl_lib/Mechanical/Rotational Elements/Mechanical\nRotational Reference";
      },
      Sbt "Modelica 3.2.1/Mechanics/Rotational/Components/Fixed";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Ideal Torque Source";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "fl_lib/Mechanical/Mechanical Sources/Ideal Torque Source";
      },
      Sbt "Modelica 3.2.1/Mechanics/Rotational/Sources/Torque2";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Inertia";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "fl_lib/Mechanical/Rotational Elements/Inertia";
      },
      Sbt "Modelica 3.2.1/Mechanics/Rotational/Components/Inertia";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Ideal Rotational\nMotion Sensor";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "fl_lib/Mechanical/Mechanical Sensors/Ideal Rotational\nMotion Sensor";
      },
      Sbt "Modelica 3.2.1/Mechanics/Rotational/Sensors/RelSpeedSensor";

      {
        Mdl_deep_ast.
          deep_block_source_type = Ast_types.mdl_name_of_string
            "Solver\nConfiguration";
        deep_block_source_path = Ast_types.mdl_path_of_string
            "nesl_utility/Solver\nConfiguration";
      },
      scicos_pro_implicit_split_type

    ]
    ;;

    let reference_block_type_table =
      Hash_table.of_bindings reference_block_type_bindings
    ;;

    let scicos_pro_block_type_for_reference_block_type = function
      | {
          Mdl_deep_ast.
          deep_block_source_type = Ast_types.Mdl_name ty;
          deep_block_source_path = Ast_types.Mdl_path path;
        } as ref_ty ->
        if Hashtbl.mem reference_block_type_table ref_ty
        then Hashtbl.find reference_block_type_table ref_ty
        else
          Genr_utilities_common.fatal_error (Printf.sprintf
            "Mdl block reference %s, with block path %s\
           \nhas no Activate counterpart yet"
            ty path)
    ;;

    let scicos_pro_block_type_for = function
      | Mdl_basics_explicit_ast.
        Explicit_block_regular_type ty ->
        scicos_pro_block_type_for_regular_block_type ty
      | Explicit_block_reference_type ref_ty ->
        scicos_pro_block_type_for_reference_block_type ref_ty
    ;;

    let scicos_pro_port_body scicos_pro_port_type other_bindings block_id i =
      let scicos_pro_parameter_bindings =
        Genr_utils.block_parameter_binding
          (Si "portNumber",
           Genr_utilities_common.code_out_port_number i),
        other_bindings in
      [
        Instantiate_block (block_id, scicos_pro_port_type);
        Set_block_parameters (block_id, scicos_pro_parameter_bindings);
        Set_block_size
          (block_id, Genr_utilities_common.default_block_port_size);
      ]
    ;;

    let in_port_body =
      let other_bindings =
        [
          block_parameter_binding
            (Si "outsize",
             Genr_utilities_common.code_out_expression "[-1, -2]");
          block_parameter_binding
            (Si "outtyp", inherit_mode_type);
          block_parameter_binding
            (Si "dept", Genr_utilities_common.code_out_int 0);
        ] in
      scicos_pro_port_body scicos_pro_input_port_type other_bindings
    ;;

    let out_port_body =
      let other_bindings =
        [
          block_parameter_binding
            (Si "insize",
             Genr_utilities_common.code_out_expression "[-1, -2]");
          block_parameter_binding
            (Si "intyp", inherit_mode_type);
        ] in
      scicos_pro_port_body scicos_pro_output_port_type other_bindings
    ;;

    (* Implicit ports *)
    let implicit_in_port_body =
      scicos_pro_port_body scicos_pro_implicit_input_port_type []
    ;;

    let implicit_out_port_body =
      scicos_pro_port_body scicos_pro_implicit_output_port_type []
    ;;

    let in_activation_port_body =
      scicos_pro_port_body scicos_pro_input_activation_port_type []
    and out_activation_port_body =
      scicos_pro_port_body scicos_pro_output_activation_port_type []
    ;;

    let samphold scsm_id block_id block_string out_data_type ext origin =
      let block_name = Genr_utilities_common.block_name block_string in
      let scicos_pro_parameter_bindings =
        block_parameter_binding (Si "typ", out_data_type),
        [
          block_parameter_binding (Si "externalActivation", ext);
        ] in
      Instantiate_block (block_id, scicos_pro_samphold_type) ::
      let nports = (Unatural._1_, _1_) in
      Genr_utilities_common.set_block_geometry block_id nports origin @ [
        Set_block_parameters (block_id, scicos_pro_parameter_bindings);
        Genr_utilities_common.genr_add_block scsm_id block_id block_name;
      ]
    ;;

    let sampholds scsm_id block_id out_data_type ext nhold =

      let rec loop samphold_strings accu i =
        if i > nhold then samphold_strings, accu else
        let block_string = Printf.sprintf "samphold%d" (Unatural.to_int i) in
        let origin = (_1_, Unatural.sub nhold i) in
        let lst =
          samphold scsm_id block_id block_string out_data_type ext origin in
        loop
          (block_string :: samphold_strings)
          (Concat_list.concat accu (Concat_list.list lst))
          (Unatural.succ i) in

      let samphold_strings, code =
        loop [] (Concat_list.list []) _1_ in
      List.rev samphold_strings, code
    ;;

    let time_f scsm_id block_id block_string ext origin =
      let block_name = Genr_utilities_common.block_name block_string in
      let nports = (Unatural._0_, _1_) in
      let scicos_pro_parameter_bindings =
        block_parameter_binding (Si "externalActivation", ext), [] in
      Instantiate_block (block_id, scicos_pro_time_type) ::
      Genr_utilities_common.set_block_geometry block_id nports origin @ [
        Set_block_parameters (block_id, scicos_pro_parameter_bindings);
        Genr_utilities_common.genr_add_block scsm_id block_id block_name;
      ]
    ;;

    let relational_op
        scsm_id block_id block_string op zcr out_data_type ext origin =
      let block_name = Genr_utilities_common.block_name block_string in
      let nports = (Unatural._2_, _1_) in
      let scicos_pro_parameter_bindings =
        block_parameter_binding (Si "rule", op),
        [
          block_parameter_binding (Si "zcr", zcr);
          block_parameter_binding (Si "typei", inherit_mode_type);
          block_parameter_binding (Si "typeo", out_data_type);
          block_parameter_binding (Si "externalActivation", ext);
        ] in
      Instantiate_block (block_id, scicos_pro_relational_op_type) ::
      Genr_utilities_common.set_block_geometry block_id nports origin @ [
        Set_block_parameters (block_id, scicos_pro_parameter_bindings);
        Genr_utilities_common.genr_add_block scsm_id block_id block_name;
      ]
    ;;

    let convert
        scsm_id block_id block_string out_data_type saturation origin =
      let block_name = Genr_utilities_common.block_name block_string in
      let nports = (_1_, _1_) in
      let sbt =
        scicos_pro_block_type_for
          (Explicit_block_regular_type DataTypeConversion) in
      let scicos_pro_parameter_bindings =
        Genr_utilities_common.block_parameter_binding
          (Si "it", inherit_mode_type),
        [
          block_parameter_binding
            (Si "ot", out_data_type);
          block_parameter_binding
            (Si "satur", code_out_simulink_saturation saturation);
          block_parameter_binding
            (Si "rmethod", Genr_utilities_common.floor_round_method);
          block_parameter_binding
            (Si "externalActivation",
             Genr_utilities_common.external_activation_mode_off);
        ] in
      Instantiate_block (block_id, sbt) ::
      Genr_utilities_common.set_block_geometry block_id nports origin @ [
        Set_block_parameters (block_id, scicos_pro_parameter_bindings);
        Genr_utilities_common.genr_add_block scsm_id block_id block_name;
      ]
    ;;

    let convert_in_scsm0_with_saturation = convert scsm0 blk0;;

    let modeler_block_type_for = scicos_pro_block_type_for
    and modeler_math_expression_type = scicos_pro_math_expression_type
    and modeler_relational_op_type = scicos_pro_relational_op_type
    and modeler_virtual_clk0_type = scicos_pro_virtual_clk0_type
    and modeler_time_type = scicos_pro_time_type
    and modeler_sample_clk_type = scicos_pro_sample_clk_type
    and modeler_clk_split_type = scicos_pro_clk_split_type
    and modeler_implicit_split_type = scicos_pro_implicit_split_type
    and modeler_samphold_type = scicos_pro_samphold_type
    and modeler_split_type = scicos_pro_split_type
    and modeler_annotation_type = scicos_pro_annotation_type
    and modeler_output_activation_port_type =
      scicos_pro_output_activation_port_type
    and modeler_input_activation_port_type =
      scicos_pro_input_activation_port_type
    and modeler_output_port_type =
      scicos_pro_output_port_type
    and modeler_input_port_type =
      scicos_pro_input_port_type
   ;;

  end
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
