(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: scicos_variant.ml,v 1.1 2019-05-10 14:07:14 jpc Exp $ *)

open Mdl_basics_explicit_ast;;
open Mdl_block_types;;
open Instructions;;
open Genr_types;;
open Genr_utils;;
open Unatural;;

module M =
  struct

    let target_modeler = Simport_configuration.Modeler_scicos;;

    let code_out_simulink_saturation =
      let scicos_int_of_saturation = function
        | Genr_types.Saturation_off -> 0
        | Genr_types.Saturation_on -> 1 in
      (fun sat ->
       Genr_utilities_common.code_out_int (scicos_int_of_saturation sat))
    ;;

    let scicos_input_port_type = Instructions.Sbt "IN_f"
    and scicos_output_port_type = Sbt "OUT_f"
    ;;

    let scicos_implicit_input_port_type = Instructions.Sbt "INIMPL_f"
    and scicos_implicit_output_port_type = Sbt "OUTIMPL_f"
    ;;

    let scicos_input_activation_port_type = Sbt "CLKINV_f"
    and scicos_output_activation_port_type = Sbt "CLKOUTV_f"
    ;;
    let scicos_annotation_type =
      Sbt "TEXT_f"
    ;;
    let scicos_samphold_type =
      Sbt "SAMPHOLD_m"
    ;;
    let scicos_split_type =
      Sbt "SPLIT_f"
    ;;
    let scicos_implicit_split_type =
      Sbt "IMPSPLIT_f"
    ;;
    let scicos_atomic_port_type =
      scicos_input_activation_port_type
    ;;
    let scicos_clk_split_type =
      Sbt "CLKSPLIT_f"
    ;;
    let scicos_sample_clk_type =
      Sbt "SampleCLK"
    ;;
    let scicos_time_type =
      Sbt "TIME_f"
    ;;
    let scicos_virtual_clk0_type =
      Sbt "VirtualCLK0"
    ;;
    let scicos_relational_op_type =
      Sbt "RELATIONALOP"
    ;;
    let scicos_math_expression_type =
      Sbt "EXPRESSION"
    ;;
    let scicos_neg_to_pos_type =
      Sbt "NEGTOPOS_f"
    ;;
    let scicos_pos_to_neg_type =
      Sbt "POSTONEG_f"
    ;;
    let scicos_zcross_type =
      Sbt "ZCROSS_f"
    ;;
(* To be revisited when Edge_Trigger will exist in Scicos.
    let scicos_edge_type =
      Sbt "Edge_Trigger"
    ;;
*)
    let scicos_extract_activation_type =
      Sbt "Extract_Activation"
    ;;

    let scicos_if_then_else_type =
      Sbt "IFTHEL_f"
    ;;

    let scicos_interpreted_function_type =
      Sbt "scifunc_block5"
    ;;

    let get_relop_value genr_info relop =
      let relop =
        match
          Genr_utilities_common.get_argument_value_string genr_info relop
        with
        | "==" -> 0
        | "~=" -> 1
        | "isNaN" ->
          (* isNaN is translated into x -> x <> x, have a look into
             genr_scicos_relational_operator.ml *)
          1
        | "<" -> 2
        | "<=" -> 3
        | ">" -> 4
        | ">=" -> 5
        | s ->
          Genr_utilities_common.fatal_error_with_info genr_info
            (Printf.sprintf
               "unknown comparison operator ``%s''"
               s) in
      Genr_utilities_common.code_out_int relop
    ;;

    let code_out_simulink_out_data_type simulink_out_data_type =
      let scicos_out_data_type =
        match simulink_out_data_type with
        | Genr_types.Simulink_single ->
          Genr_utilities_common.warning (
            "out data type single is not supported, double is used instead.");
          1
        | Simulink_double -> 1
        | Simulink_int8 -> 5
        | Simulink_uint8 -> 8
        | Simulink_int16 -> 4
        | Simulink_uint16 -> 7
        | Simulink_int32 -> 3
        | Simulink_uint32 -> 6
        | Simulink_boolean -> 9
        | Simulink_inherit -> -1 in
      Genr_utilities_common.code_out_int scicos_out_data_type
    ;;

    let inherit_mode_type = code_out_simulink_out_data_type Simulink_inherit;;

    let simport_block_translation_bug s =
      (* This is a bug: the compiler should normally compile this block
           specially *)
      ignore (Genr_utilities_common.fatal_error (Printf.sprintf
        "Fatal simport bug: code for block %s should have been generated by\
       \nthe compiler, but block %s was erroneously proposed to translation."
        s s));
      assert false
    ;;

    let scicos_block_type_for = function
      | Mdl_basics_explicit_ast.
        Explicit_block_regular_type Mdl_block_types.Abs ->
        Sbt "ABS_VALUEi"
      | Explicit_block_regular_type Mdl_block_types.ActionPort ->
        Say.fatal_error "Unexpected ActionPort encountered, please report."

      | Explicit_block_regular_type Backlash -> Sbt "BACKLASH"
      | Explicit_block_regular_type Bias -> Sbt "SUMMATION"
      | Explicit_block_regular_type BusCreator -> Sbt "MUX"

      | Explicit_block_regular_type Clock -> scicos_time_type

      | Explicit_block_regular_type ComplexToMagnitudeAngle -> Sbt "MATMAGPHI"
      | Explicit_block_regular_type ComplexToRealImag -> Sbt "MATZREIM"

      | Explicit_block_regular_type Constant -> Sbt "CONST_m"

      | Explicit_block_regular_type DataTypeConversion -> Sbt "CONVERT"
      | Explicit_block_regular_type Delay -> Sbt "DELAY_f"
      | Explicit_block_regular_type Demux -> Sbt "DEMUX"
      | Explicit_block_regular_type Derivative -> Sbt "DERIV"
      | Explicit_block_regular_type DigitalClock -> Sbt "TIME_f"
      | Explicit_block_regular_type DiscretePulseGenerator -> Sbt "PULSE_SD"
      | Explicit_block_regular_type DiscreteFilter
      | Explicit_block_regular_type DiscreteTransferFcn
      | Explicit_block_regular_type DiscreteZeroPole -> Sbt "DLR"
      | Explicit_block_regular_type DiscreteStateSpace -> Sbt "DLSS"
      | Explicit_block_regular_type Display -> Sbt "AFFICH_m"
      | Explicit_block_regular_type DotProduct -> Sbt "MATMUL"

      | Explicit_block_regular_type ElementaryMath -> Sbt "TrigFun"

        (* Fcn: empbedded simulink expression *)
      | Explicit_block_regular_type Fcn -> Sbt "EXPRESSION"
      | Explicit_block_regular_type From -> Sbt "FROM"
      | Explicit_block_regular_type FromWorkspace -> Sbt "FROMWS_c"

      | Explicit_block_regular_type Gain -> Sbt "GAINBLK"
      | Explicit_block_regular_type Goto -> Sbt "GOTO"
      | Explicit_block_regular_type GotoTagVisibility -> Sbt "GotoTagVisibility"
      | Explicit_block_regular_type Ground -> Sbt "Ground_g"

      | Explicit_block_regular_type Inport -> scicos_input_port_type
      | Explicit_block_regular_type InportShadow ->
        simport_block_translation_bug "InportShadow"
      | Explicit_block_regular_type Integrator -> Sbt "SIM_INTEGRAL"

      | Explicit_block_regular_type Logic -> Sbt "LOGICAL_OP"
      | Explicit_block_regular_type Lookup -> Sbt "LOOKUP_c"
      | Explicit_block_regular_type Lookup2D -> Sbt "LOOKUP2D"

      | Explicit_block_regular_type MagnitudeAngleToComplex -> Sbt "MATMAGPHI"
      | Explicit_block_regular_type ManualSwitch -> Sbt "TKSWITCH"
      | Explicit_block_regular_type Math -> Sbt "EXPBLK_m"
      | Explicit_block_regular_type Merge -> Sbt "NRMSOM_f"
      | Explicit_block_regular_type MinMax -> Sbt "MAXMIN"
      | Explicit_block_regular_type MultiPortSwitch -> Sbt "M_SWITCH"
      | Explicit_block_regular_type Mux -> Sbt "MUX"

      | Explicit_block_regular_type Outport -> scicos_output_port_type

      | Explicit_block_regular_type PMIOPort -> Sbt "INIMPL_f"

      | Explicit_block_regular_type Product -> Sbt "PRODUCT"

      | Explicit_block_regular_type Quantizer -> Sbt "QUANT_f"

      | Explicit_block_regular_type RandomNumber
      | Explicit_block_regular_type UniformRandomNumber -> Sbt "RAND_m"
      | Explicit_block_regular_type RateLimiter -> Sbt "RATELIMITER"

      | Explicit_block_regular_type RealImagToComplex -> Sbt "MATZREIM"

      | Explicit_block_regular_type RelationalOperator ->
        scicos_relational_op_type
      | Explicit_block_regular_type Relay -> Sbt "HYSTHERESIS"

      | Explicit_block_regular_type Saturate -> Sbt "SATURATION"
      | Explicit_block_regular_type Scope -> Sbt "CSCOPE"
      | Explicit_block_regular_type Selector -> Sbt "EXTRACTOR"
      | Explicit_block_regular_type SignalGenerator -> Sbt "GENSIN_f"
      | Explicit_block_regular_type Signum -> Sbt "SIGNUM"
      | Explicit_block_regular_type Sin -> Sbt "GENSIN_f"
      | Explicit_block_regular_type StateSpace ->
        simport_block_translation_bug "StateSpace"
      | Explicit_block_regular_type Step -> Sbt "STEP_FUNC"
      | Explicit_block_regular_type Stop -> Sbt "END_c"
      | Explicit_block_regular_type SubSystem ->
        simport_block_translation_bug "SubSystem"
      | Explicit_block_regular_type Sum -> Sbt "SUMMATION"
      | Explicit_block_regular_type Switch -> Sbt "SWITCH2_s"

      | Explicit_block_regular_type Terminator -> Sbt "MUX"
      | Explicit_block_regular_type TimeScope -> Sbt "CSCOPE"
      | Explicit_block_regular_type ToWorkspace -> Sbt "TOWS_c"
      | Explicit_block_regular_type TransferFcn -> Sbt "CLR"
      | Explicit_block_regular_type TransportDelay -> Sbt "TIME_DELAY"

      | Explicit_block_regular_type Mdl_block_types.TriggerPort ->
        scicos_input_activation_port_type

      | Explicit_block_regular_type Trigonometry -> Sbt "TrigFun"

      | Explicit_block_regular_type UnitDelay -> Sbt "DOLLAR_m"

      | Explicit_block_regular_type ZeroOrderHold -> scicos_samphold_type
      | Explicit_block_regular_type ZeroPole -> Sbt "CLR"

      | Explicit_block_regular_type (
          Actuator |
          Assertion | Assignment | AsynchronousTaskSpecification |

          Buffer |
          BusAssignment | BusSelector | BusToVector |

          CombinatorialLogic |
          Concatenate | ConstellationDiagram |

          DataStoreMemory | DataStoreRead | DataStoreWrite |

          DeadZone | DiscreteFir | DiscreteIntegrator |

          DSPFlip |

          EnablePort |

          ForIterator | FrameConversion | FromFile |

          HitCross |

          If | InitialCondition | Interpolation_n_D |

          LookupNDDirect | Lookup_n_D |

          (* Level 2 Matlab S-Function:
             A call to an external function defined in Matlab (the function
             name is given in the block) *)
          M_S_Function |
          (* Interpreted Matlab Fcn:
             Same as M_S_Function, but the function is a native Matlab
             function *)
          MATLABFcn |

          MATLABSystem |

          Memory | ModelReference |

          PMComponent |
          PermuteDimensions | Polyval | PreLookup | Probe |

          RateTransition |
          Reference |
          Reshape | Rounding |

          (* S-Function block (in the interface):
             A call to an external function defined in another langage
             and probably compiled (the function name is given in the block)

             Note: the MATLAB Function block (in the interface):
             A special case of S_Function, the code is given in the block but
             the MATLAB Function block generates a S_Function with a
             stateflow section that embbeds the function definition.

          *)
          S_Function |
          SampleTimeMath | Sensor |

          SignalConversion |
          SignalFromWorkspace | SignalSpecification |
          SignalToWorkspace |
          SignalViewerScope | SpectrumAnalyzer |

          Squeeze | Sqrt | SwitchCase |

          ToFile | TwoWayConnection |

          UnitConversion | UnaryMinus |

          VariableTransportDelay |
          VideoViewer |

          Width | WhileIterator |

          Unbuffer | SecondOrderIntegrator | FunctionCallSplit |
          ForEach | DataTypeDuplicate | ArithShift | Find |
          FunctionCallFeedbackLatch

          as ty) ->
        Genr_utilities_common.fatal_error (Printf.sprintf
          "Mdl block %s has no Scicos counterpart yet"
          (Mdl_block_types_tables.string_of_block_type ty))
      | Explicit_block_reference_type {
          Mdl_deep_ast.
          deep_block_source_type = ty;
          deep_block_source_path = path;
        } as ref_ty ->
        if ref_ty = Mdl_fake_block_types.
           split_block_type
          then scicos_split_type else
        if ref_ty = Mdl_fake_block_types.
           split_connection_block_type
          then scicos_implicit_split_type else
        if ref_ty = Mdl_fake_block_types.
           atomic_port_block_type
          then scicos_atomic_port_type else
        if ref_ty = Mdl_fake_block_types.
           atomic_event_generator_block_type
          then scicos_atomic_port_type else
        if ref_ty = Mdl_fake_block_types.
           if_action_event_generator_block_type
          then scicos_if_then_else_type else
        if ref_ty = Mdl_fake_block_types.
           enable_event_generator_block_type
          then scicos_if_then_else_type else
        if ref_ty = Mdl_fake_block_types.
           trigger_rising_event_generator_block_type
          then scicos_neg_to_pos_type else
        if ref_ty = Mdl_fake_block_types.
           trigger_falling_event_generator_block_type
          then scicos_pos_to_neg_type else
        if ref_ty = Mdl_fake_block_types.
           trigger_either_event_generator_block_type
          then scicos_zcross_type else
        if ref_ty = Mdl_fake_block_types.
           trigger_function_call_generator_block_type
          then scicos_extract_activation_type else
(* To be revisited when scicos_edge_type will exist in Scicos
        if ref_ty = Mdl_fake_block_types.
           trigger_and_enable_rising_event_generator_block_type
          then scicos_edge_type else
        if ref_ty = Mdl_fake_block_types.
           trigger_and_enable_falling_event_generator_block_type
          then scicos_edge_type else
        if ref_ty = Mdl_fake_block_types.
           trigger_and_enable_either_event_generator_block_type
        then scicos_edge_type else
*)
        if ref_ty = Mdl_fake_block_types.
           trigger_and_enable_function_call_generator_block_type
          then scicos_extract_activation_type else
        if ref_ty = Mdl_fake_block_types.
           matlab_function_block_type
          then scicos_interpreted_function_type else

        begin match ty, path with
        | Ast_types.Mdl_name "Algebraic Constraint",
          Ast_types.Mdl_path
            "simulink/Math\nOperations/Algebraic Constraint" ->
          Sbt "CONSTRAINT_c"

        | Ast_types.Mdl_name "Band-Limited White Noise.",
          Ast_types.Mdl_path
            "simulink/Sources/Band-Limited\nWhite Noise" ->
          Sbt "RAND_m"

        | Ast_types.Mdl_name "CelsiusToFahrenheit",
          Ast_types.Mdl_path
            "simulink_extras/Transformations/Celsius to\nFahrenheit" ->
          Sbt "EXPRESSION"

        | Ast_types.Mdl_name "Checks_DGap",
          Ast_types.Mdl_path
            "simulink/Model\nVerification/Check \nDynamic Gap" ->
          Sbt "Checks_DGap"
        | Ast_types.Mdl_name "Checks_DMin",
          Ast_types.Mdl_path
            "simulink/Model\nVerification/Check Dynamic \nLower Bound" ->
          Sbt "Checks_DMin"
        | Ast_types.Mdl_name "Checks_DMax",
          Ast_types.Mdl_path
            "simulink/Model\nVerification/Check Dynamic \nUpper Bound" ->
          Sbt "Checks_DMax"
        | Ast_types.Mdl_name "Checks_DRange",
          Ast_types.Mdl_path "simulink/Model\nVerification/Check \nDynamic Range" ->
          Sbt "Checks_DRange"
        | Ast_types.Mdl_name "Checks_Gradient",
          Ast_types.Mdl_path
            "simulink/Model\nVerification/Check\nDiscrete Gradient" ->
          Sbt "Checks_Gradient"
        | Ast_types.Mdl_name "Checks_Resolution",
          Ast_types.Mdl_path
            "simulink/Model\nVerification/Check Input \nResolution" ->
          Sbt "Checks_Resolution"
        | Ast_types.Mdl_name "Checks_SGap",
          Ast_types.Mdl_path
            "simulink/Model\nVerification/Check \nStatic Gap" ->
          Sbt "Checks_SGap"
        | Ast_types.Mdl_name "Checks_SMin",
          Ast_types.Mdl_path
            "simulink/Model\nVerification/Check Static \nLower Bound" ->
          Sbt "Checks_SMin"
        | Ast_types.Mdl_name "Checks_SMax",
          Ast_types.Mdl_path
            "simulink/Model\nVerification/Check Static \nUpper Bound" ->
          Sbt "Checks_SMax"
        | Ast_types.Mdl_name "Checks_SRange",
          Ast_types.Mdl_path
            "simulink/Model\nVerification/Check \nStatic Range" ->
          Sbt "Checks_SRange"

        | Ast_types.Mdl_name "Compare To Constant",
          Ast_types.Mdl_path
            "simulink/Logic and Bit\nOperations/Compare\nTo Constant" ->
          scicos_relational_op_type
        | Ast_types.Mdl_name "Compare To Zero",
          Ast_types.Mdl_path
            "simulink/Logic and Bit\nOperations/Compare\nTo Zero" ->
          scicos_relational_op_type

        | Ast_types.Mdl_name "DFlipFlop",
          Ast_types.Mdl_path
            "simulink_extras/Flip Flops/D Flip-Flop" ->
          Sbt "DFLIPFLOP"

        | Ast_types.Mdl_name "Difference",
          Ast_types.Mdl_path
            "simulink/Discrete/Difference" ->
          Sbt "DOLLAR_m"

(*
        | Ast_types.Mdl_name "Discrete Derivative",
          Ast_types.Mdl_path
            "simulink/Discrete/Discrete Derivative" ->
          Sbt "SUPER_f"
*)

        | Ast_types.Mdl_name "DLatch",
          Ast_types.Mdl_path
            "simulink_extras/Flip Flops/D Latch" ->
          Sbt "DLATCH"
        | Ast_types.Mdl_name "Dot Product",
          Ast_types.Mdl_path
            "simulink/Math\nOperations/Dot Product" ->
          Sbt "MATMUL"

        | Ast_types.Mdl_name "FahrenheitToCelsius",
          Ast_types.Mdl_path
            "simulink_extras/Transformations/Fahrenheit\nto Celsius" ->
          Sbt "EXPRESSION"

        | Ast_types.Mdl_name "Function-Call Generator",
          Ast_types.Mdl_path
            "simulink/Ports &\nSubsystems/Function-Call\nGenerator" ->
          Sbt "CONST_m"

        | Ast_types.Mdl_name "LTI Block",
          Ast_types.Mdl_path
            "cstblocks/LTI System" ->
          Sbt "CLSS"

        | Ast_types.Mdl_name "JKFlipFlop",
          Ast_types.Mdl_path
            "simulink_extras/Flip Flops/J-K\nFlip-Flop" ->
          Sbt "JKFLIPFLOP"

        | (Ast_types.Mdl_name "Manual Switch" |
           Ast_types.Mdl_name "" |
           Ast_types.Mdl_name "Unknown"),
          Ast_types.Mdl_path
            "simulink/Signal\nRouting/Manual Switch" ->
          Sbt "TKSWITCH"

        | Ast_types.Mdl_name "Manual Switch",
          Ast_types.Mdl_path
            "simulink3/Nonlinear/Manual Switch" ->
          Sbt "TKSWITCH"

        | Ast_types.Mdl_name "PID Controller",
          Ast_types.Mdl_path
            "simulink_extras/Additional\nLinear/PID Controller" ->
          Sbt "PID2"

        | Ast_types.Mdl_name "PID Controller",
          Ast_types.Mdl_path
            "simulink_need_slupdate/PID Controller" ->
          Sbt "PID2"

        | Ast_types.Mdl_name "Ramp",
          Ast_types.Mdl_path
            "simulink/Sources/Ramp" ->
          Sbt "RAMP"
        | Ast_types.Mdl_name "Repeating table",
          Ast_types.Mdl_path
            "simulink/Sources/Repeating\nSequence"
        | Ast_types.Mdl_name "Repeating Sequence Interpolated",
          Ast_types.Mdl_path
            "simulink/Sources/Repeating\nSequence\nInterpolated" ->
          Sbt "CURVE_c"
        | Ast_types.Mdl_name "Repeating Sequence Stair",
          Ast_types.Mdl_path
            "simulink/Sources/Repeating\nSequence\nStair" ->
          Sbt "REGISTER"

        | Ast_types.Mdl_name "Slider Gain",
          Ast_types.Mdl_path
            "simulink/Math\nOperations/Slider\nGain" ->
          Sbt "GAINBLK"

        | Ast_types.Mdl_name "SRFlipFlop",
          Ast_types.Mdl_path
            "simulink_extras/Flip Flops/S-R\nFlip-Flop" ->
          Sbt "SRFLIPFLOP"

        | Ast_types.Mdl_name "Transpose",
          Ast_types.Mdl_path
            "dspmtrx3/Transpose" ->
          Sbt "MATTRAN"

        | Ast_types.Mdl_name "XY scope.",
          Ast_types.Mdl_path
            "simulink/Sinks/XY Graph" ->
          Sbt "CSCOPXY"

        | Ast_types.Mdl_name ty, Ast_types.Mdl_path p ->
          Genr_utilities_common.fatal_error (Printf.sprintf
            "Mdl block reference %s, with block path %s\
           \nhas no Scicos counterpart yet"
            ty p)
        end
    ;;

    let scicos_port_body scicos_port_type other_bindings block_id i =
      let scicos_parameter_bindings =
        Genr_utils.block_parameter_binding
          (Si "prt",
           Genr_utilities_common.code_out_port_number i),
        other_bindings in
      [
        Instantiate_block (block_id, scicos_port_type);
        Set_block_parameters (block_id, scicos_parameter_bindings);
        Set_block_size
          (block_id, Genr_utilities_common.default_block_port_size);
      ]
    ;;

    let in_port_body =
      let other_bindings =
        [
          block_parameter_binding
           (Si "otsz", Genr_utilities_common.code_out_int (-1));
          block_parameter_binding
           (Si "ot", inherit_mode_type);
        ] in
      scicos_port_body scicos_input_port_type other_bindings
    ;;

    let out_port_body =
      scicos_port_body scicos_output_port_type []
    ;;

    let implicit_in_port_body =
      scicos_port_body scicos_implicit_input_port_type []
    ;;

    let implicit_out_port_body =
      scicos_port_body scicos_implicit_output_port_type []
    ;;

    let in_activation_port_body =
      scicos_port_body scicos_input_activation_port_type []
    and out_activation_port_body =
      scicos_port_body scicos_output_activation_port_type []
    ;;

    let samphold scsm_id block_id block_string out_data_type _ext origin =
      let block_name = Genr_utilities_common.block_name block_string in
      let scicos_parameter_bindings =
        block_parameter_binding (Si "it", out_data_type),
        [] in
      Instantiate_block (block_id, scicos_samphold_type) ::
        let nports = (Unatural._1_, _1_) in
        Genr_utilities_common.set_block_geometry block_id nports origin @ [
          Set_block_parameters (block_id, scicos_parameter_bindings);
          Genr_utilities_common.genr_add_block scsm_id block_id block_name;
        ]
    ;;

    let sampholds scsm_id block_id out_data_type ext nhold =

      let rec loop samphold_strings accu i =
        if i > nhold then samphold_strings, accu else
        let block_string = Printf.sprintf "samphold%d" (Unatural.to_int i) in
        let origin = (_1_, Unatural.sub nhold i) in
        let lst =
          samphold scsm_id block_id block_string out_data_type ext origin in
        loop
          (block_string :: samphold_strings)
          (Concat_list.concat accu (Concat_list.list lst))
          (Unatural.succ i) in

      let samphold_strings, code =
        loop [] (Concat_list.list []) _1_ in
      List.rev samphold_strings, code
    ;;

    let time_f scsm_id block_id block_string _ext origin =
      let block_name = Genr_utilities_common.block_name block_string in
      let nports = (Unatural._0_, _1_) in
      Instantiate_block (block_id, scicos_time_type) ::
        Genr_utilities_common.set_block_geometry block_id nports origin @ [
        Genr_utilities_common.genr_add_block scsm_id block_id block_name;
      ]
    ;;

    let relational_op
        scsm_id block_id block_string op zcr out_data_type _ext origin =
      let block_name = Genr_utilities_common.block_name block_string in
      let nports = (Unatural._2_, _1_) in
      let scicos_parameter_bindings =
        block_parameter_binding (Si "rule", op),
        [
          block_parameter_binding (Si "zcr", zcr);
          block_parameter_binding (Si "Datatype", out_data_type);
        ] in
      Instantiate_block (block_id, scicos_relational_op_type) ::
        Genr_utilities_common.set_block_geometry block_id nports origin @ [
        Set_block_parameters (block_id, scicos_parameter_bindings);
        Genr_utilities_common.genr_add_block scsm_id block_id block_name;
      ]
    ;;

    let convert
        scsm_id block_id block_string out_data_type saturation origin =
      let block_name = Genr_utilities_common.block_name block_string in
      let nports = (_1_, _1_) in
      let sbt =
        scicos_block_type_for
          (Explicit_block_regular_type DataTypeConversion) in
      let scicos_parameter_bindings =
        Genr_utilities_common.block_parameter_binding
          (Si "it", inherit_mode_type),
        [
          block_parameter_binding
            (Si "ot", out_data_type);
          block_parameter_binding
            (Si "np", code_out_simulink_saturation saturation);
        ] in
      Instantiate_block (block_id, sbt) ::
        Genr_utilities_common.set_block_geometry block_id nports origin @ [
        Set_block_parameters (block_id, scicos_parameter_bindings);
        Genr_utilities_common.genr_add_block scsm_id block_id block_name;
      ]
    ;;

    let convert_in_scsm0_with_saturation = convert scsm0 blk0;;

    let modeler_block_type_for = scicos_block_type_for
    and modeler_math_expression_type = scicos_math_expression_type
    and modeler_relational_op_type = scicos_relational_op_type
    and modeler_virtual_clk0_type = scicos_virtual_clk0_type
    and modeler_time_type = scicos_time_type
    and modeler_sample_clk_type = scicos_sample_clk_type
    and modeler_clk_split_type = scicos_clk_split_type
    and modeler_implicit_split_type = scicos_implicit_split_type
    and modeler_samphold_type = scicos_samphold_type
    and modeler_split_type = scicos_split_type
    and modeler_annotation_type = scicos_annotation_type
    and modeler_output_activation_port_type =
      scicos_output_activation_port_type
    and modeler_input_activation_port_type =
      scicos_input_activation_port_type
    and modeler_output_port_type =
      scicos_output_port_type
    and modeler_input_port_type =
      scicos_input_port_type
   ;;

  end
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
