(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_modelicos_gain.ml,v 1.1 2019-05-10 14:07:14 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_modelicos_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;
module Genr_utilities_scicos_pro = struct end;;

open Instructions
open Genr_types
open Genr_utilities_modelicos
open Unatural
;;

(* Simulink notion of gain multiplication method. *)
type mdl_multiplication =
  | Element_wise
  | Left_matrix
  | Right_matrix
;;

let mdl_block_type =
  Mdl_basics_explicit_ast.Explicit_block_regular_type
    Mdl_block_types.Gain
;;

let default_multiplication = Element_wise;;

let get_rule_multiplication multiplication =
  match multiplication with
  | "Element-wise(K.*u)" | "Element-wise(.*)" -> Element_wise
  | "Matrix(K*u)" | "Matrix(*)" | "Matrix(K*u) (u vector)" -> Left_matrix
  | "Matrix(u*K)" -> Right_matrix
  |  c ->
     Say.warning (Printf.sprintf
       "Gain block translation: unexpected rule %s" c);
     default_multiplication
;;

(* VSS notion of gain multiplication method of simulink multiplication. *)
let modelicos_of_multiplication = function
  | Element_wise -> "Gain.*u1"
  | Left_matrix -> "Gain*u1"
  | Right_matrix -> "u1*Gain"
;;

let gain_ident = Ast_types.ident_from_string "Gain"
and multiplication_ident = Ast_types.ident_from_string "Multiplication"
;;

let genr block_id genr_info =

  let multiplication = get_ident_string genr_info multiplication_ident
  and sample_time =
    Genr_utilities_modelicos.get_sample_time genr_info in

  let mulmethod = get_rule_multiplication multiplication in

  let modelicos_block_parameters ext = [
    Value_parameter (Si "gain", Simulink_arg_expression gain_ident);
    Value_parameter (Si "overflow", Simulink_arg_saturation);
    Value_parameter (Si "externalActivation", Code_value ext);
  ] in

  let modelicos_basic_block_instructions blk_id ext =
    genr_instructions_for
      mdl_block_type (modelicos_block_parameters ext)
      [] blk_id genr_info in

  let modelicos_block_parameters2 ext = [
    Value_parameter (Si "gain",  Simulink_arg_expression gain_ident);
    Value_parameter (Si "overflow", Simulink_arg_saturation);
    Value_parameter (Si "mulmethod", Code_value (
      code_out_string (modelicos_of_multiplication mulmethod)));
    Value_parameter (Si "externalActivation", ext);
  ] in

  let modelicos_matrix_block_instructions blk_id ext =
    let sci_block_type = Sbt "system/MatrixOperations/MatrixGain" in
    genr_instructions_for_modeler_block_type
      sci_block_type (modelicos_block_parameters2 ext)
      [] blk_id genr_info in

  match sample_time with
  | "-1" ->
    if mulmethod = Element_wise then
      Concat_list.list (
        modelicos_basic_block_instructions
          block_id external_activation_mode_off)
    else
      Concat_list.list (
        modelicos_matrix_block_instructions
          block_id (Code_value external_activation_mode_off))

  | _ ->
    let sample_time = code_out_expression sample_time in

    let in_port_name = "in"
    and out_port_name = "out"
    and matrix_gain_name = "matrix_gain"
    and sample_clock_name = "sample_clock" in

    let in_port =
      Genr_utilities_modelicos.in_port_in_scsm0
        in_port_name port_1_
        (_0_, _0_) in
    let out_port =
      Genr_utilities_modelicos.
      out_port_in_scsm0
        out_port_name port_1_
        (_2_, _0_) in
    let matrix_gain =
      if mulmethod = Element_wise then (
        let nports = (_1_, _1_) in
        Genr_utilities_modelicos.
        add_block_body_in_scsm0 matrix_gain_name
          (modelicos_basic_block_instructions blk0 external_activation_mode_on)
          nports
          (_1_, _0_)
      ) else (
        let nports = (_1_, _1_) in
        Genr_utilities_modelicos.
        add_block_body_in_scsm0 matrix_gain_name
          (modelicos_basic_block_instructions blk0 external_activation_mode_on)
          nports
          (_1_, _0_)
      ) in
    let sample_clock =
      Genr_utilities_modelicos.
      sample_clock_in_scsm0
        sample_clock_name sample_time
        (_1_, _1_) in

    let links = [
      Genr_utilities_modelicos.
      add_directed_link_in_scsm0
        in_port_name port_1_ matrix_gain_name port_1_ [];
      add_directed_link_in_scsm0
        matrix_gain_name port_1_ out_port_name port_1_ [];
      Genr_utilities_modelicos.
      add_event_link_in_scsm0
        sample_clock_name port_1_ matrix_gain_name port_1_ [];
    ] in

    let super_body = [
      in_port;
      out_port;
      matrix_gain;
      sample_clock;
      links;
    ] in

  super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
