#!/bin/sh -e

#######################################################################
#                                                                     #
#                               Simport                               #
#                                                                     #
#                   Pierre Weis, INRIA Rocquencourt                   #
#                                                                     #
#  Copyright 2010-2015,                                               #
#  Institut National de Recherche en Informatique et en Automatique.  #
#  All rights reserved.                                               #
#                                                                     #
#  This file is distributed under the terms of the BSD License.       #
#                                                                     #
#######################################################################

# $Id: make_generator_interfaces.sh,v 1.1 2019-05-10 14:07:14 jpc Exp $

# The shell command that (re-) creates interfaces for all block generator
# modules.

# All block generator modules should share the same empty interface, since
# block generator modules have nothing to export.
#
# This command properly copies one common source interface
# (namely genr_*_compare_to_constant.mli) to be the interface of all block
# generator modules.

# Notes:
# - Module Genr_*_not_yet_translated is treated especially, since it must
#  have its proper non empty interface in order to export its
#  generating function to the simport compiler.
# - Module Genr_*_compare_to_constant is also treated especially since
#  its interface is the shared interface source file for all the generator
#  modules.
# - Since svn supports link usage, we should use links instead of file copies
#  here; unfortunately Windows does not support links, so let's stay to the
#  safe side by using the ubiquitous file copy basic facility.

BLOCK_SYSTEM_VARIANT=modelicos

GENERATOR_PREFIX=genr_${BLOCK_SYSTEM_VARIANT}
COMMON_INTERFACE=${GENERATOR_PREFIX}_compare_to_constant.mli

# Prevent spurious executable status for our module source files.
chmod -x $COMMON_INTERFACE

cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_abs.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_aliases.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_atomic_event_gen.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_backlash.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_band_limited_white_noise.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_bus_assignment.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_bus_creator.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_bus_selector.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_c_m_block.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_combinatorial_logic.mli
#${GENERATOR_PREFIX}_compare_to_constant.mli is the source file for all
#module interfaces in this directory: no need to copy it to itself!
#cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_compare_to_constant.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_complex_to_coordinates.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_concatenate.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_conversion.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_coordinates_to_complex.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_coulomb_viscous_friction.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_counter_limited.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_cross_product.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_data_type_conversion.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_dead_zone.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_dead_zone_dynamic.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_delay.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_demux.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_detect.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_detect_rise_and_fall.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_difference.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_digital_clock.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_discrete_derivative.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_discrete_integral.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_discrete_pulse_generator.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_discrete_state_space.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_discrete_transfer_fcn.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_discrete_zero_pole.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_dot_product.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_elementary_math.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_enable_event_gen.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_fcn.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_from_workspace.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_function_call_generator.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_gain.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_hit_cross.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_if.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_if_action_event_gen.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_initial_condition.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_integer_delay.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_integrator.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_interval_test.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_logic.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_lookup.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_lookup2d.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_lookup_nd.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_math.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_matlab_function.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_matlab_fcn.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_matrix_concatenation.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_merge.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_min_max.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_mport_switch.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_multi_plot_graph.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_mux.mli
# ${GENERATOR_PREFIX}_not_yet_translated indeed exports its genr function.
# It could not share the common interface file!
#cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_not_yet_translated.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_pmio_port.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_quantizer.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_random_number.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_rate_transition.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_relational_operator.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_repeating_sequence_interpolated.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_repeating_sequence_stair.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_repeating_table.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_rounding.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_saturate.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_saturation_dynamic.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_scope.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_selector.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_signal_generator.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_signum.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_sin.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_sqrt.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_stop.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_switch.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_tapped_delay_line.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_time_scope.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_to_workspace.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_trigger_and_enable_event_gen.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_trigger_event_gen.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_trigonometry.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_unary_minus.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_uniform_random_number.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_unit_delay.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_xy_scope.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_zero_order_hold.mli

# Simscape block generators
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_electrical_inductor.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_electrical_sensors.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_electrical_mechanical_sources.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_mechanical_rotational_inertia.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_mechanical_rotational_reference.mli
cp -p $COMMON_INTERFACE ${GENERATOR_PREFIX}_mechanical_ideal_rotational_sensor.mli
