(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_aliases.ml,v 1.1 2019-05-10 14:05:17 jpc Exp $ *)

(* In scicos we must use module Genr_utilities_scicos and not
   Genr_utilities_scicos_pro. *)
module Genr_utilities_scicos_pro = struct end;;

open Ast_types;;
open Genr_utilities_scicos;;
open Genr_types;;
open Instructions;;

let list_replicate n x =

  let rec loop accu = function
    | 0 -> accu
    | n -> loop (x :: accu) (n - 1) in

  if n < 0 then invalid_arg (
    Printf.sprintf "list_replicate: negative argument %d" n) else
  loop [] n
(** A List.replicate function. *)
;;

(* Idents commonly used in this file. *)

let inputs_ident = Ast_types.ident_from_string "Inputs"
and outputs_ident = ident_from_string "Outputs"
and value_ident = ident_from_string "Value"
;;

let gain_ident = ident_from_string "Gain"
and multiplication_ident = ident_from_string "Multiplication"
;;

(** Definition of block translators *)

(** Registering translators *)

open Mdl_block_types;;

let add_scicos_translation =
  Genr_utilities_scicos.add_modeler_translation
;;

let add_scicos_equivalent =
  Genr_utilities_scicos.add_modeler_equivalent
;;

let add_scicos_conditional_translation1 =
  Genr_utilities_scicos.add_modeler_conditional_translation1
;;

let add_scicos_conditional_translation2 =
  Genr_utilities_scicos.add_modeler_conditional_translation2
;;

let add_scicos_conditional_translation3 =
  Genr_utilities_scicos.add_modeler_conditional_translation3
;;

let scicos_block_type_for =
  Genr_utilities_scicos.modeler_block_type_for
(*
and scicos_math_expression_type =
  Genr_utilities_scicos.modeler_math_expression_type
and scicos_relational_op_type =
  Genr_utilities_scicos.modeler_relational_op_type
and scicos_virtual_clk0_type =
  Genr_utilities_scicos.modeler_virtual_clk0_type
and scicos_time_type =
  Genr_utilities_scicos.modeler_time_type
and scicos_sample_clk_type =
  Genr_utilities_scicos.modeler_sample_clk_type
and scicos_clk_split_type =
  Genr_utilities_scicos.modeler_clk_split_type
and scicos_implicit_split_type =
  Genr_utilities_scicos.modeler_implicit_split_type
and scicos_samphold_type =
  Genr_utilities_scicos.modeler_samphold_type
and scicos_split_type =
  Genr_utilities_scicos.modeler_split_type
and scicos_annotation_type =
  Genr_utilities_scicos.modeler_annotation_type
and scicos_output_activation_port_type =
  Genr_utilities_scicos.modeler_output_activation_port_type
and scicos_input_activation_port_type =
  Genr_utilities_scicos.modeler_input_activation_port_type
and scicos_output_port_type =
  Genr_utilities_scicos.modeler_output_port_type
and scicos_input_port_type =
  Genr_utilities_scicos.modeler_input_port_type
*)
;;

(** Simulink block [ActionPort]
  This block is a simple marker for an IfAction sub-system.
  It is suppressed when translating the corresponding sub-system block of
  action. *)

(** Simulink block [Algebraic Constraint] *)

add_scicos_translation
  (Genr_utilities_scicos.simulink_block_reference_type
     (Ast_types.mdl_name_of_string "Algebraic Constraint")
     (Ast_types.mdl_path_of_string
       "simulink/Math\nOperations/Algebraic Constraint"))
  [
    Value_parameter (Si "x0",
      Simulink_arg_value (Ast_types.ident_from_string "z0"));
  ]
  []
;;

(** Simulink block [Slider Gain] *)

(* FIXME:
  Slider is translated only to a gain block ignoring the bounds for the moment
  let low_ident = ident_from_string "low"
  and hi_ident = ident_from_string "hi" *)
add_scicos_translation
  (simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Slider Gain")
    (Ast_types.mdl_path_of_string
      "simulink/Math\nOperations/Slider\nGain"))
  [
    Value_parameter (Si "gain",
      Simulink_arg_expression (Ast_types.ident_from_string "gain"));
    Value_parameter (Si "over",
      Simulink_arg_saturation_set Genr_types.Saturation_off);
  ]
  []
;;

(** Simulink block [BusCreator] *)

let get_scicos_block_type_and_block_args nin =
  Say.warning (Printf.sprintf
    "BusCreator not supported yet: using Mux block instead.");
  let scicos_block_type =
    scicos_block_type_for (simulink_block_type BusCreator) in
  let scicos_block_args = [
    Value_parameter (Si "in",
      Code_value nin);
  ] in

  scicos_block_type, scicos_block_args in

add_scicos_conditional_translation1
  (simulink_block_type Mdl_block_types.BusCreator)
  get_scicos_block_type_and_block_args
    (Simulink_arg_expression inputs_ident)
  []
;;

(* TODO: those blocks are partially describe (missing source block)

(** Simulink blocks for checking assertions, blocks [Checks_*]. *)

List.iter
  (fun ty ->
   add_scicos_equivalent
     (simulink_block_source_type (Ast_types.mdl_name_of_string ty)))
  [
    "Checks_DGap";
    "Checks_DMin";
    "Checks_DMax";
    "Checks_DRange";
    "Checks_Gradient";
    "Checks_Resolution";
    "Checks_SGap";
    "Checks_SMin";
    "Checks_SMax";
    "Checks_SRange";
  ]
;;
*)

(** Simulink block [Clock] *)

add_scicos_equivalent (simulink_block_type Mdl_block_types.Clock)
;;

(** Simulink block [Constant] *)

add_scicos_translation
  (simulink_block_type Mdl_block_types.Constant)
  [
    Value_parameter (Si "C",
      Simulink_arg_expression value_ident);
  ]
  []
;;

(** Simulink block [Demux] *)

add_scicos_translation
  (simulink_block_type Mdl_block_types.Demux)
  [
    Value_parameter (Si "out",
      Simulink_arg_expression outputs_ident);
  ]
  []
;;

(** Simulink block [Derivative] *)

add_scicos_equivalent (simulink_block_type Derivative)
;;

(** Simulink block [DFlipFlop] *)

add_scicos_equivalent
  (simulink_block_reference_type
     (mdl_name_of_string "DFlipFlop")
     (mdl_path_of_string "simulink_extras/Flip Flops/D Flip-Flop"))
;;

(** Simulink block [Display] *)

let format_ident = ident_from_string "Format"
and floating_ident = ident_from_string "Floating"
and decimation_ident = ident_from_string "Decimation" in

let get_scicos_block_type_and_block_args format floating decimation =

  let scicos_block_type =
    scicos_block_type_for (simulink_block_type Display) in

  begin match get_bool floating_ident floating with
  | false -> ()
  | true ->
    Say.warning (Printf.sprintf
      "Display block translation: unsupported floating ``on''")
  end;

  begin match get_string decimation_ident decimation with
  | "1" -> ()
  | decimation ->
    Say.warning (Printf.sprintf
      "Display block translation: unsupported decimation ``%s''" decimation)
  end;

  let add_nt_nd args =
    match get_string format_ident format with
    | "long" ->
      Value_parameter (Si "nt",
        Code_value (code_out_int 12)) ::
      Value_parameter (Si "nd",
        Code_value (code_out_int 5)) :: args
   | "short" ->
     Value_parameter (Si "nt",
       Code_value (code_out_int 6)) ::
     Value_parameter (Si "nd",
       Code_value (code_out_int 3)) :: args
   | format ->
     Say.warning (Printf.sprintf
       "Display block translation: unsupported format ``%s'', using long"
       format);
     Value_parameter (Si "nt",
       Code_value (code_out_int 12)) ::
     Value_parameter (Si "nd",
       Code_value (code_out_int 5)) :: args in

  let scicos_block_args =
    Value_parameter (Si "in",
      Code_value (code_out_expression "[ -1, -2 ]")) ::
    Value_parameter (Si "font",
      Code_value (code_out_int 1)) ::
    Value_parameter (Si "fontsize",
      Code_value (code_out_int 1)) ::
    Value_parameter (Si "colr",
      Code_value (code_out_int 1)) ::
    add_nt_nd [
      Value_parameter (Si "herit",
        Code_value (code_out_int 1)); ] in

  scicos_block_type, scicos_block_args in

add_scicos_conditional_translation3
  (simulink_block_type Display)
  get_scicos_block_type_and_block_args
  (Simulink_arg_value format_ident,
   Simulink_arg_value floating_ident,
   Simulink_arg_value decimation_ident)
  []
;;

(** Simulink block [DLatch] *)

add_scicos_equivalent
  (simulink_block_reference_type
    (mdl_name_of_string "DLatch")
    (mdl_path_of_string "simulink_extras/Flip Flops/D Latch"))
;;

(** Simulink block [From] *)

add_scicos_translation
  (simulink_block_type From)
  [
    Value_parameter (Si "tag",
      Simulink_arg_value (ident_from_string "GotoTag"));
    Value_parameter (Si "BS",
      Code_value (code_out_int 1));
  ]
  []
;;

(** Simulink block [Gain] *)

let get_gain =
  let diag_flag = "Element-wise" in
  let diag_flag_len = String.length diag_flag in
  fun gainv multv ->
  match get_string gain_ident gainv,
        get_string multiplication_ident multv with
  | g, mult ->
  (*Say.debug "*** mult is %S\n" mult;*)
    let remove_trailing_comments expr =
      let ib = Scanf.Scanning.from_string expr in
      Scanf.bscanf ib "%[^%%]"
        (fun s ->
         if Scanf.Scanning.end_of_input ib then
           s
         else
           begin
             Scanf.bscanf ib "%0c" (fun c -> assert (c = '%'));
             s
           end) in
    let g =
      if String.length mult >= diag_flag_len &&
         String.sub mult 0 diag_flag_len = diag_flag
      then Printf.sprintf "%s" (remove_trailing_comments g)
      else g in
    code_out_expression g in

let get_mult_method =
  fun gainv multv ->
  match get_string gain_ident gainv,
        get_string multiplication_ident multv with
  | _g, mult ->
    let scicos_mult_method =
      match mult with
      | "Element-wise(K.*u)" | "Element-wise(.*)" -> 1
      | "Matrix(K*u)" | "Matrix(*)" | "Matrix(K*u) (u vector)" -> 0
      | "Matrix(u*K)" -> 0
      |  c ->
         Say.warning (Printf.sprintf
           "Gain block translation: unexpected rule %s" c);
         0 in
    (* Say.debug "*** mult is %S\n" mult; *)
    code_out_int scicos_mult_method  in

add_scicos_translation
  (simulink_block_type Gain)
  [
    Value_parameter (Si "gain",
      Simulink_arg_fun2 (
        get_gain,
        Simulink_arg_value gain_ident,
        Simulink_arg_value multiplication_ident));
    Value_parameter (Si "over",
      Simulink_arg_saturation);
    Value_parameter (Si "mulmethod",
      Simulink_arg_fun2 (
        get_mult_method,
        Simulink_arg_value gain_ident,
        Simulink_arg_value multiplication_ident));
  ]
  []
;;

(** Simulink block [Goto] *)

let tag_vis_ident = ident_from_string "TagVisibility"
and goto_tag_ident = ident_from_string "GotoTag" in

let get_visibility v =
  let s = get_string tag_vis_ident v in
  let vty =
    match s with
    | "local" -> 1
    | "scoped" -> 2
    | "global" -> 3
    | s ->
      Say.warning (Printf.sprintf
        "Goto block translation: unexpected visibility ``%s''\
       \nChoosing local as visibility tag" s);
      1 in
  code_out_int vty in

add_scicos_translation
  (simulink_block_type Goto)
  [
    Value_parameter (Si "tag",
      Simulink_arg_value goto_tag_ident);
    Value_parameter (Si "tagvis",
      Simulink_arg_fun1 (
        get_visibility, Simulink_arg_value tag_vis_ident));
    Value_parameter (Si "BS",
      Code_value (code_out_int 1));
  ]
  []
;;

(** Simulink block [GotoTagVisibility] *)

add_scicos_translation
  (simulink_block_type GotoTagVisibility)
  [
    Value_parameter (Si "tag",
      Simulink_arg_value (ident_from_string "GotoTag"));
  ]
  []
;;

(** Simulink block [Ground] *)

add_scicos_equivalent (simulink_block_type Ground)
;;

(** Simulink block [Inport] *)

let port_ident = ident_from_string "Port";;
let port_dimensions_ident = ident_from_string "PortDimensions";;

let get_port_dims d =
  let s = get_string port_dimensions_ident d in
  try
    let _nd = int_of_string s in
     code_out_expression (Printf.sprintf "[ %s, 1]" s)
  with
  | Failure _ (*"int_of_string"*) -> code_out_expression s in

add_scicos_translation
  (simulink_block_type Inport)
  [
    Value_parameter (Si "prt",
      Simulink_arg_expression port_ident);
    Value_parameter (Si "otsz",
      Simulink_arg_fun1 (
        get_port_dims,
        Simulink_arg_value port_dimensions_ident));
    Value_parameter (Si "ot",
      Simulink_arg_out_data_type);
  ]
  []
;;

(** Simulink block [JKFlipFlop] *)

let initial_condition_ident = ident_from_string "initial_condition";;

add_scicos_translation
  (simulink_block_reference_type
    (mdl_name_of_string "JKFlipFlop")
    (mdl_path_of_string "simulink_extras/Flip Flops/J-K\nFlip-Flop"))
  [
    Value_parameter (Si "init",
      Simulink_arg_expression initial_condition_ident);
  ]
  []
;;

(** Simulink block [LTI Block] *)

let sys_ident = ident_from_string "sys"
and ic_ident = ident_from_string "IC" in
let getA s = code_out_expression ((get_string sys_ident s)^".A")
and getB s = code_out_expression ((get_string sys_ident s)^".B")
and getC s = code_out_expression ((get_string sys_ident s)^".C")
and getD s = code_out_expression ((get_string sys_ident s)^".D")
in

add_scicos_translation
  (simulink_block_reference_type
    (mdl_name_of_string "LTI Block")
    (mdl_path_of_string "cstblocks/LTI System"))
    [
      Value_parameter (Si "A",
        Simulink_arg_fun1 (getA,Simulink_arg_value sys_ident));
      Value_parameter (Si "B",
        Simulink_arg_fun1 (getB,Simulink_arg_value sys_ident));
      Value_parameter (Si "C",
        Simulink_arg_fun1 (getC,Simulink_arg_value sys_ident));
      Value_parameter (Si "D",
        Simulink_arg_fun1 (getD,Simulink_arg_value sys_ident));
      Value_parameter (Si "x0",
        Simulink_arg_expression ic_ident);
    ]
  []
;;

(** Simulink reference block [Manual switch] *)

let mdl_manual_switch_block_type,
    mdl_manual_switch_block_type_empty,
    mdl_manual_switch_block_type_unknown =
  let path =
    Ast_types.mdl_path_of_string
      "simulink/Signal\nRouting/Manual Switch" in
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Manual Switch")
    path,
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "")
    path,
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Unknown")
    path in

let sw_ident = ident_from_string "sw" in
(*
Unused in Scicos
and action_ident = ident_from_string "action" in
*)
let get_sw sw =
  let sw = get_int sw_ident sw in
  let sw =
    match sw with
    | 0 -> 2
    | 1 -> 1
    | sw ->
      let default = 1 in
      Say.warning (Printf.sprintf
        "ManualSwitch block translation: \
         unexpected connected input %d, using %d"
        sw default);
      default in
  code_out_int sw in

let block_parameters =
  [
    Value_parameter (Si "ch",
      Simulink_arg_fun1 (get_sw, Simulink_arg_value sw_ident));
    Value_parameter (Si "inh",
      Code_value (code_out_int 1));
(* Fixme:
   Field varsize is not there on all mdl models.
   Value_parameter
     (Si "varsize", Simulink_arg_bool "varsize");

   Should define a default when varsize is not there.
*)
  ] in

add_scicos_translation
  mdl_manual_switch_block_type
  block_parameters
  [];

add_scicos_translation
  mdl_manual_switch_block_type_empty
  block_parameters
  [];

add_scicos_translation
  mdl_manual_switch_block_type_unknown
  block_parameters
  [];

let mdl_manual_switch_block_type3 =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Manual Switch")
    (Ast_types.mdl_path_of_string "simulink3/Nonlinear/Manual Switch") in

add_scicos_translation
  mdl_manual_switch_block_type3
  [
    Value_parameter (Si "ch",
      Simulink_arg_fun1 (get_sw, Simulink_arg_value sw_ident));
    Value_parameter (Si "inh",
      Code_value (code_out_int 1));
  ]
  []
;;

(** Native Simulink block [ManualSwitch] *)

let current_setting_ident = Ast_types.ident_from_string "CurrentSetting" in

let get_cs cs =
  let sw = get_int current_setting_ident cs in
  let sw =
    match sw with
    | 0 -> 2
    | 1 -> 1
    | sw ->
      let default = 1 in
      Say.warning (Printf.sprintf
        "ManualSwitch block translation: \
         unexpected connected input %d, using %d"
        sw default);
      default in
  code_out_int sw in

let mdl_manual_switch_native_block_type =
  simulink_block_type Mdl_block_types.ManualSwitch in

add_scicos_translation
  mdl_manual_switch_native_block_type
  [
    Value_parameter (Si "ch",
      Simulink_arg_fun1 (get_cs, Simulink_arg_value current_setting_ident));
    Value_parameter (Si "inh",
      Code_value (code_out_int 1));
  (* FIXME:
     Fields varsize and sample are not yet treated.
  *)
  ]
  []

;;

(** Simulink block [Mux] *)

add_scicos_translation
  (simulink_block_type Mdl_block_types.Mux)
  [
    Value_parameter
      (Si "in", Simulink_arg_expression inputs_ident);
  ]
  []
;;

(** Simulink block [Outport] *)

add_scicos_translation
  (simulink_block_type Outport)
  [
    Value_parameter (Si "prt",
      Simulink_arg_expression port_ident);
  ]
  []
;;

(** Simulink block [PID Controller] *)

let i_ident = ident_from_string "I"
and p_ident = ident_from_string "P"
and d_ident = ident_from_string "D" in

add_scicos_translation
  (simulink_block_reference_type
    (mdl_name_of_string "PID Controller")
    (mdl_path_of_string "simulink_extras/Additional\nLinear/PID Controller"))
  [
    Value_parameter (Instructions.Si "p",
      Simulink_arg_expression p_ident);
    Value_parameter (Instructions.Si "i",
      Simulink_arg_expression i_ident);
    Value_parameter (Instructions.Si "d",
      Simulink_arg_expression d_ident);
  ]
  []
;;

(** Simulink block [PID Controller] *)

let i_ident = ident_from_string "I"
and p_ident = ident_from_string "P"
and d_ident = ident_from_string "D" in

add_scicos_translation
  (simulink_block_reference_type
    (mdl_name_of_string "PID Controller")
    (mdl_path_of_string "simulink_need_slupdate/PID Controller"))
  [
    Value_parameter (Instructions.Si "p",
      Simulink_arg_expression p_ident);
    Value_parameter (Instructions.Si "i",
      Simulink_arg_expression i_ident);
    Value_parameter (Instructions.Si "d",
      Simulink_arg_expression d_ident);
  ]
  []
;;

(** Simulink block [Product] *)

let get_sign_string accu =
  Printf.sprintf "[ %s ]" (String.concat ", " accu)
;;

let out_data_type_ident = ident_from_string "OutDataType" in

let get_sign inputs =

  let ib = Scanf.Scanning.from_string inputs in

  if Scanf.Scanning.end_of_input ib then
    Say.fatal_error "Block Product with empty Inputs argument" else

  let rec loop accu =
    if Scanf.Scanning.end_of_input ib then List.rev accu else
    Scanf.bscanf ib "%c" (function
    | '*' -> loop ("+1" :: accu)
    | '/' -> loop ("-1" :: accu)
    | '|' -> loop accu
    | c ->
      Say.warning (Printf.sprintf
        "Product block translation: unexpected character \
         %C in Inputs argument"
        c);
      loop accu) in

  Scanf.bscanf ib "%0c" (function
  | '*' | '/' | '|' -> get_sign_string (loop [])
  | '0' .. '9' ->
    Scanf.bscanf ib "%d" (function n ->
      get_sign_string (list_replicate n "+1")
    )
  | _ -> get_sign_string [ inputs; ]) in

let get_scicos_block_type_and_block_args inputs ty =

  let inputs = get_string inputs_ident inputs
  and ty = get_int out_data_type_ident ty in

  begin match ty with
  | 1 -> ()
  | _ ->
    Say.warning "Block Product only works for doubles, using doubles"
  end;

  match inputs with
  | "/" ->
    Say.warning
      "Block Product with only one \"/\" argument, using scicos \
       block INVBLK";
    Sbt "INVBLK", [ ]
  | inputs ->
    scicos_block_type_for (simulink_block_type Product),
    [
      Value_parameter
        (Si "sgn",
         Code_value (code_out_expression (get_sign inputs)));
    ] in

add_scicos_conditional_translation2
  (simulink_block_type Product)
  get_scicos_block_type_and_block_args (
    Simulink_arg_value inputs_ident,
    Simulink_arg_out_data_type
  )
  []
;;

(** Simulink block [Ramp] *)

add_scicos_translation
  (simulink_block_reference_type
    (mdl_name_of_string "Ramp")
    (mdl_path_of_string "simulink/Sources/Ramp"))
  [
    Value_parameter (Si "slope",
      Simulink_arg_expression (ident_from_string "slope"));
    Value_parameter (Si "stt",
      Simulink_arg_expression (ident_from_string "start"));
    Value_parameter (Si "iout",
      Simulink_arg_expression (ident_from_string "X0"));
  ]
  []
;;

(** Simulink block [RateLimiter] *)

add_scicos_translation
  (simulink_block_type Mdl_block_types.RateLimiter)
  [
    Value_parameter (Si "maxp",
      Simulink_arg_expression (ident_from_string "RisingSlewLimit"));
    Value_parameter (Si "minp",
      Simulink_arg_expression (ident_from_string "FallingSlewLimit"));
  ]
  []
;;

(** Simulink block [Relay] *)

add_scicos_translation
  (simulink_block_type Mdl_block_types.Relay)
  [
    Value_parameter (Si "high_lim",
      Simulink_arg_value (ident_from_string "OnSwitchValue"));
    Value_parameter (Si "low_lim",
      Simulink_arg_value (ident_from_string "OffSwitchValue"));
    Value_parameter (Si "out_high",
      Simulink_arg_value (ident_from_string "OnOutputValue"));
    Value_parameter (Si "out_low",
      Simulink_arg_value (ident_from_string "OffOutputValue"));
    Value_parameter (Si "nzz",
      Simulink_arg_zero_crossing);
  ]
  []
;;

(** Fake Simulink block [Split] *)

add_scicos_equivalent
  Mdl_fake_block_types.split_block_type
;;

(** Fake Simulink block [Split_connection] *)

add_scicos_equivalent
  Mdl_fake_block_types.split_connection_block_type
;;

(* Fake Simulink block [AtomicPort] *)
add_scicos_equivalent
  Mdl_fake_block_types.atomic_port_block_type
;;

(** Simulink block [SRFlipFlop] *)

add_scicos_translation
  (simulink_block_reference_type
     (mdl_name_of_string "SRFlipFlop")
     (mdl_path_of_string "simulink_extras/Flip Flops/S-R\nFlip-Flop"))
  [
    Value_parameter (Si "init",
      Simulink_arg_expression initial_condition_ident);
  ]
  []
;;

let is_empty s =
  let ib = Scanf.Scanning.from_string s in
  Scanf.Scanning.end_of_input ib ||
  let k _ _ = false in
  Scanf.kscanf ib k " [ ] " true
;;

(** Simulink block [StateSpace] *)

let a_ident = ident_from_string "A" in

let get_scicos_block_type_and_block_args a =
  match get_string a_ident a with
  | a when is_empty a ->
    Sbt "GAINBLK",
    [
      Value_parameter (Si "gain",
        Simulink_arg_expression (ident_from_string "D"));
      Value_parameter (Si "over",
        Simulink_arg_saturation_set Genr_types.Saturation_off);
    ]
  | _ ->
    Sbt "CLSS",
    [
      Value_parameter (Si "A",
        Simulink_arg_expression a_ident);
      Value_parameter (Si "B",
        Simulink_arg_expression (ident_from_string "B"));
      Value_parameter (Si "C",
        Simulink_arg_expression (ident_from_string "C"));
      Value_parameter (Si "D",
        Simulink_arg_expression (ident_from_string "D"));
      Value_parameter (Si "x0",
        Simulink_arg_expression_aliases [
          ident_from_string "X0";
          ident_from_string "InitialCondition";
          ]);
    ] in

add_scicos_conditional_translation1
  (simulink_block_type StateSpace)
  get_scicos_block_type_and_block_args
  (Simulink_arg_value a_ident)
  []
;;

(** Simulink block [Step] *)

add_scicos_translation
  (simulink_block_type Step)
  [
    Value_parameter (Si "tim",
      Simulink_arg_expression (ident_from_string "Time"));
    Value_parameter (Si "be",
      Simulink_arg_expression (ident_from_string "Before"));
    Value_parameter (Si "af",
      Simulink_arg_expression (ident_from_string "After"));
  ]
  []
;;

(** Simulink block [Sum] *)

let get_sign_value inputs =

  let inputs = get_string inputs_ident inputs in
  let ib = Scanf.Scanning.from_string inputs in
  let rec loop accu =
    if Scanf.Scanning.end_of_input ib then List.rev accu else
    Scanf.bscanf ib "%c" (function
    | '+' -> loop ("+1" :: accu)
    | '-' -> loop ("-1" :: accu)
    | '|' -> loop accu
    | _ -> loop accu) in
  let txt = get_sign_string (loop []) in
  code_out_expression txt in

add_scicos_translation
  (simulink_block_type Sum)
  [
    Value_parameter (Si "Datatype",
      Simulink_arg_out_data_type);
    Value_parameter (Si "sgn",
      Simulink_arg_fun1 (
         get_sign_value,
         Simulink_arg_value inputs_ident)
      );
    Value_parameter (Si "satur",
      Simulink_arg_saturation);
  ]
  []
;;

(** Simulink block [Terminator] *)

add_scicos_translation
  (simulink_block_type Mdl_block_types.Terminator)
  [
    Value_parameter (Si "in",
      Code_value (code_out_int 1));
  ]
  []
;;

(** Simulink block [TransferFcn] *)

let numerator_ident = ident_from_string "Numerator"
and denominator_ident = ident_from_string "Denominator" in

let get_component n =
  let n = get_string numerator_ident n in
  let s =
    Printf.sprintf "poly (fliplr (%s), \'s\', \'coeff\')" n in
  code_out_expression s in

add_scicos_translation
  (simulink_block_type Mdl_block_types.TransferFcn)
  [
    Value_parameter (Si "num",
      Simulink_arg_fun1 (
        get_component,
        Simulink_arg_value numerator_ident));
    Value_parameter (Si "den",
      Simulink_arg_fun1 (
        get_component,
        Simulink_arg_value denominator_ident));
  ]
  []
;;

(** Simulink block [TransportDelay] *)

let initial_input_ident = ident_from_string "InitialInput"
and initial_output_ident = ident_from_string "InitialOutput"
;;

let get_init v =
  let txt = get_string initial_input_ident v in
  code_out_expression txt in
add_scicos_translation
  (simulink_block_type Mdl_block_types.TransportDelay)
  [
    Value_parameter (Si "T",
      Simulink_arg_expression (ident_from_string "DelayTime"));
    Value_parameter (Si "init",
      Simulink_arg_fun1 (
        get_init,
        Simulink_arg_value_aliases [
          initial_output_ident; initial_input_ident;
        ]));
    Value_parameter (Si "N",
      Simulink_arg_expression (ident_from_string "BufferSize"));
  ]
  []
;;

(** Simulink block [Transpose] *)

add_scicos_translation
  (simulink_block_reference_type
    (mdl_name_of_string "Transpose")
    (mdl_path_of_string "dspmtrx3/Transpose"))
  [
    Value_parameter (Si "typ",
      Code_value (code_out_int 1));
    Value_parameter (Si "rule",
      Code_value (code_out_int 1));
  ]
  []
;;

(** Simulink block [TriggerPort] *)

add_scicos_translation
  (Genr_utilities_scicos.simulink_block_type TriggerPort)
  [
    Value_parameter (Si "prt",
      Code_value (code_out_int 1));
  ]
  []
;;

(** Simulink block [ZeroPole]
    num = gain * poly (zeros, z, root)
    poly (num, s, coeff).
*)

let zeros_ident = ident_from_string "Zeros"
and poles_ident = ident_from_string "Poles" in

let get_num gain zeros =

  let gain = get_string gain_ident gain in
  let zeros = get_string zeros_ident zeros in

  let txt = Printf.sprintf "%s * poly (%s, \"s\", \"root\")" gain zeros in
  code_out_expression txt in

let get_den poles =

  let poles = get_string poles_ident poles in

  let txt = Printf.sprintf "poly (%s, \"s\", \"root\")" poles in
  code_out_expression txt in

add_scicos_translation
  (simulink_block_type Mdl_block_types.ZeroPole)
  [
    Value_parameter (Si "num",
      Simulink_arg_fun2 (
        get_num,
        Simulink_arg_value gain_ident,
        Simulink_arg_value zeros_ident));
    Value_parameter (Si "den",
      Simulink_arg_fun1 (
        get_den,
        Simulink_arg_value poles_ident));
  ]
  []
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
