(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_compare_to_constant.ml,v 1.1 2019-05-10 14:05:17 jpc Exp $ *)

(* In scicos we must use module Genr_utilities_scicos and not
   Genr_utilities_scicos_pro. *)
module Genr_utilities_scicos_pro = struct end;;

(** Translating Simulink blocs "Compare To Constant" and "Compare To Zero" *)

open Instructions
open Genr_types
open Genr_utilities_scicos
open Unatural
;;

let mdl_block_type_compare_to_constant =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Compare To Constant")
    (Ast_types.mdl_path_of_string
       "simulink/Logic and Bit\nOperations/Compare\nTo Constant")
;;

let mdl_block_type_compare_to_zero =
  Genr_utilities_scicos.simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Compare To Zero")
    (Ast_types.mdl_path_of_string
       "simulink/Logic and Bit\nOperations/Compare\nTo Zero")
;;

let relop_ident = Ast_types.ident_from_string "relop"
and const_ident = Ast_types.ident_from_string "const"
;;

let genr_comparison const block_id genr_info =

  let zcr =
    Genr_utilities_scicos.get_zero_crossing_value genr_info
  and out_data_type =
    Genr_utilities_scicos.get_simulink_out_data_type_value genr_info
  and op =
    Genr_utilities_scicos.get_relop_value genr_info relop_ident in

  let in_port_name = "in"
  and out_port_name = "out"
  and constant_name = "constant"
  and relational_name = "relational" in

  let in_port =
    Genr_utilities_scicos.
    in_port_in_scsm0 in_port_name port_1_ (_0_, _0_) in

  let out_port =
    Genr_utilities_scicos.
    out_port_in_scsm0 out_port_name port_1_ (_2_, _0_) in

  let constant =
    let nports = (_0_, _1_) in
    let scicos_block_type = Sbt "CONST_m" in
    Genr_utilities_scicos.
    add_block_body_in_scsm0 constant_name
      (genr_instructions_for_modeler_block_type scicos_block_type
         [
           Value_parameter (Si "C", Code_value const);
         ]
         []
         Genr_utils.blk0 genr_info)
      nports
      (_0_, _1_) in

  let scicos_block_parameters = [
    Genr_types.Value_parameter
      (Instructions.Si "rule", Genr_types.Code_value op);
    Value_parameter (Si "zcr", Code_value zcr);
    Value_parameter (Si "Datatype", Code_value out_data_type);
    ] in

  let scicos_basic_block_instructions blk_id =
    genr_instructions_for
      mdl_block_type_compare_to_constant scicos_block_parameters
      [] blk_id genr_info in

  let relational =
    let nports = (_2_, _1_) in
    let relational_body = scicos_basic_block_instructions blk0 in
    add_block_body_in_scsm0
      relational_name relational_body nports (_1_, _0_) in

  let links = [
    Genr_utilities_scicos.
    add_directed_link_in_scsm0
      constant_name port_1_ relational_name port_2_ [];
    add_directed_link_in_scsm0
      in_port_name port_1_ relational_name port_1_ [];
    add_directed_link_in_scsm0
      relational_name port_1_ out_port_name port_1_ [];
  ] in

  let super_body = [
    in_port;
    out_port;
    constant;
    relational;
    links;
  ] in

  super_block_in_scsm0 block_id super_body
;;

let genr_compare_to_constant block_id genr_info =
  let const =
    code_out_expression (get_ident_string genr_info const_ident) in
  genr_comparison const block_id genr_info
;;

add_translator mdl_block_type_compare_to_constant genr_compare_to_constant
;;

let genr_compare_to_zero =
  genr_comparison (Genr_utilities_scicos.code_out_int 0)
;;

add_translator mdl_block_type_compare_to_zero genr_compare_to_zero
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
