(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_complex_to_coordinates.ml,v 1.1 2019-05-10 14:05:17 jpc Exp $ *)

(* In scicos we must use module Genr_utilities_scicos and not
   Genr_utilities_scicos_pro. *)
module Genr_utilities_scicos_pro = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos
open Unatural
;;

(* Computing representation of complex numbers *)

(* Simulink bloc types ComplexToMagnitudeAngle and ComplexToRealImag share
   a lot of code.
   Hence the sharing of the generator.
*)

let magnitude_angle_options = [| "Magnitude and angle";
                                 "Magnitude"; "Angle" |];;
let real_imag_options = [| "Real and imag";
                           "Real"; "Imag" |];;

let mdl_block_type_complex_to_magnitude_angle =
  Mdl_basics_explicit_ast.Explicit_block_regular_type
    Mdl_block_types.ComplexToMagnitudeAngle
;;

let complex_to_magnitude_angle_formula = code_out_int 1;;

let complex_to_magnitude_angle_options = magnitude_angle_options;;

let mdl_block_type_complex_to_real_imag =
  Mdl_basics_explicit_ast.Explicit_block_regular_type
    Mdl_block_types.ComplexToRealImag
;;

let complex_to_real_imag_formula = code_out_int 1;;

let complex_to_real_imag_options = real_imag_options;;

(* The generator for all the above blocks. *)

let output_ident = Ast_types.ident_from_string "Output";;

let genr mdl_block_type conversion_formula output_options block_id genr_info =

  let get_ident_string =
    Genr_utilities_scicos.get_ident_string genr_info in

  let output = get_ident_string output_ident in

  let scicos_block_parameters = [
    Value_parameter (Si "decomptyp", Code_value conversion_formula);
  ] in

  let option =
    if output_options.(0) = output then 0
    else if output_options.(1) = output then 1
    else if output_options.(2) = output then 2
    else (
      Say.warning (Printf.sprintf
        "Complex to %s Simulink block translation: \
         Option %s not supported." output_options.(0) output );
      0) in

  let scicos_basic_block_instructions blk_id =
    genr_instructions_for
      mdl_block_type
      scicos_block_parameters
      [] blk_id genr_info in

  match option with
  | 0 ->
    Concat_list.list
      (scicos_basic_block_instructions block_id)

  | option ->
    let in_port_name = "in"
    and out_port_name1 = "out1"
    and ctri_name = "ctri" in

    let in_port =
      Genr_utilities_scicos.
      in_port_in_scsm0
        in_port_name port_1_
          (_0_, _0_) in
    let out_port1 =
      Genr_utilities_scicos.
      out_port_in_scsm0
        out_port_name1 port_1_
        (_2_, _0_) in
    let ctri =
      let nports = (_1_, _2_) in
      Genr_utilities_scicos.
      add_block_body_in_scsm0
        ctri_name
        (scicos_basic_block_instructions blk0)
        nports
        (_1_, _0_) in

    let links = [
      Genr_utilities_scicos.
      add_directed_link_in_scsm0
        in_port_name port_1_ ctri_name port_1_ [];
      add_directed_link_in_scsm0
        ctri_name
        (Ast_types.mdl_port_number_of_int option)
        out_port_name1 port_1_ [];
    ] in

    let super_body = [
      in_port;
      out_port1;
      ctri;
      links;
    ] in

    super_block_in_scsm0 block_id super_body
;;

let genr_magnitude_angle =
  genr
    mdl_block_type_complex_to_magnitude_angle
    complex_to_magnitude_angle_formula
    complex_to_magnitude_angle_options
;;

add_translator
  mdl_block_type_complex_to_magnitude_angle genr_magnitude_angle
;;

let genr_real_imag =
  genr
    mdl_block_type_complex_to_real_imag
    complex_to_real_imag_formula
    complex_to_real_imag_options
;;

add_translator
  mdl_block_type_complex_to_real_imag genr_real_imag
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
