(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_discrete_pulse_generator.ml,v 1.1 2019-05-10 14:05:18 jpc Exp $ *)

(* In scicos we must use module Genr_utilities_scicos and not
   Genr_utilities_scicos_pro. *)
module Genr_utilities_scicos_pro = struct end;;

open Concat_list
open Instructions
open Genr_types
open Genr_utilities_scicos
open Unatural
;;

let mdl_block_type =
  Mdl_basics_explicit_ast.Explicit_block_regular_type
    Mdl_block_types.DiscretePulseGenerator
;;

let pulse_type_ident = Ast_types.ident_from_string "PulseType"
and time_source_ident = Ast_types.ident_from_string "TimeSource"
and amplitude_ident = Ast_types.ident_from_string "Amplitude"
and phase_delay_ident = Ast_types.ident_from_string "PhaseDelay"
and pulse_width_ident = Ast_types.ident_from_string "PulseWidth"
and period_ident = Ast_types.ident_from_string "Period"
;;

let genr block_id genr_info =

  let get_ident_string =
    Genr_utilities_scicos.get_ident_string genr_info in

  let get_ident_expression ident =
    Genr_types.Code_value (
      Genr_utilities_scicos.
      code_out_expression (get_ident_string ident)) in

  let sample_time = Genr_utilities_scicos.get_sample_time genr_info in

  let pulse_type = get_ident_string pulse_type_ident in
  let time_source = get_ident_string time_source_ident in
  let amplitude = get_ident_expression amplitude_ident in
  let period = get_ident_expression period_ident in
  let pulse_width = get_ident_expression pulse_width_ident in
  let phase_delay = get_ident_expression phase_delay_ident in

  let scicos_block_parameters = [
    Genr_types.
    Value_parameter (Instructions.Si "E", phase_delay);
    Value_parameter (Si "W", pulse_width);
    Value_parameter (Si "F", period);
    Value_parameter (Si "A", amplitude);
  ] in

  let time_source_generator blk_id =
    genr_instructions_for
      mdl_block_type scicos_block_parameters
      [] blk_id genr_info in

  let time_based_generator () =

    let sample_time = code_out_expression sample_time in

    let out_port_name = "out"
    and pulse_sd_name = "pulse_sd"
    and virtual_clk0_name = "virtual_clk0"
    and sample_clock_name = "sample_clock" in

    let out_port =
      Genr_utilities_scicos.out_port_in_scsm0
        out_port_name port_1_ (_1_, Unatural._0_) in

    let pulse_sd =
      let nports = (_0_, _1_) in
      let body = time_source_generator Genr_utils.blk0 in
      Genr_utilities_scicos.add_block_body_in_scsm0
        pulse_sd_name body nports (_0_, _0_) in

    let virtual_clk0 =
      Genr_utilities_scicos.
      virtual_clk0_in_scsm0 virtual_clk0_name (_1_, _1_) in
    let sample_clock =
      Genr_utilities_scicos.
      sample_clock_in_scsm0
        sample_clock_name sample_time (_1_, _2_) in

    let links = [
      Genr_utilities_scicos.
      add_directed_link_in_scsm0
        pulse_sd_name port_1_ out_port_name port_1_ [];
      Genr_utilities_scicos.
      add_event_link_in_scsm0
        sample_clock_name port_1_ virtual_clk0_name port_1_ [];
    ] in

    let super_body = [
      out_port;
      pulse_sd;
      virtual_clk0;
      sample_clock;
      links;
    ] in

    Genr_utilities_scicos.super_block_in_scsm0 block_id super_body in

  let zero_time_source_generator () =
    list (time_source_generator block_id) in

  let non_time_based_generator () =
    Say.warning "DiscretePulseGenerator: \
      non time based pulse type is not supported";
    Genr_scicos_not_yet_translated.
    genr block_id genr_info in

  let code =
    match pulse_type, time_source with
    | "Time based", _ -> time_based_generator ()
    | _, "0" -> zero_time_source_generator ()
    | _, _ -> non_time_based_generator () in

  code
;;

add_translator mdl_block_type genr;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
