(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_dot_product.ml,v 1.1 2019-05-10 14:05:18 jpc Exp $ *)

(* In scicos we must use module Genr_utilities_scicos and not
   Genr_utilities_scicos_pro. *)
module Genr_utilities_scicos_pro = struct end;;

open Mdl_block_types
open Instructions
open Genr_types
open Genr_utilities_scicos
open Unatural
;;

(* The dot product block has two names in Simulink:
   a reference type and a native type.
   This translator evidently works for both.
*)

let mdl_block_type_native =
  Mdl_basics_explicit_ast.Explicit_block_regular_type DotProduct
;;

let mdl_block_type_reference =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Dot Product")
    (Ast_types.mdl_path_of_string "simulink/Math\nOperations/Dot Product")
;;

let genr block_id genr_info =

  (* let SaturateOnIntegerOverflow =
     // Not supported because MATTRAN does not support integers *)

  let sample_time = get_sample_time genr_info in

  let out_data_type = get_simulink_out_data_type genr_info in

  let get_dtype () =
    (* MATTRAN only supports real and complex *)
    let dtype =
      match out_data_type with
      | Genr_types.Simulink_single
      | Simulink_double
      | Simulink_inherit -> out_data_type
      | Simulink_int8
      | Simulink_uint8
      | Simulink_int16
      | Simulink_uint16
      | Simulink_int32
      | Simulink_uint32
      | Simulink_boolean ->
        let out_data_type_string =
          Genr_utilities_scicos.
          string_of_simulink_out_data_type out_data_type in
        Say.warning (Printf.sprintf
          "Dot Product Simulink block translation: \
           double is used in place of Simulink type %s"
          out_data_type_string);
        Simulink_double in
    code_out_simulink_out_data_type dtype in

  let scicos_block_parameters = [
    Genr_types.
    Value_parameter (Instructions.Si "dtype",
      Genr_types.Code_value (get_dtype ()));
    Value_parameter (Si "rule",
      Code_value (code_out_int 1));
    Value_parameter (Si "np",
      Code_value (code_out_int 2));
  ] in

  let scicos_basic_block_instructions blk_id =
    genr_instructions_for
      mdl_block_type_native scicos_block_parameters
      [] blk_id genr_info in

  let in_port1_name = "in1"
  and in_port2_name = "in2"
  and out_port_name = "out"
  and mat_mul_name = "mat_mul"
  and transpose_name = "tranpose"  in

  let in_port1 =
    in_port_in_scsm0 in_port1_name port_1_ (_0_, _1_) in

  let in_port2 =
    in_port_in_scsm0 in_port2_name port_2_ (_0_, _0_) in

  let out_port =
    out_port_in_scsm0 out_port_name port_1_ (_4_, _0_) in

  let mat_mul =
    let nports = (_2_, _1_) in
    add_block_body_in_scsm0 mat_mul_name
      (scicos_basic_block_instructions blk0)
      nports
      (_3_, _0_) in

  let transpose =
    let nports = (_1_, _1_) in
    add_block_body_in_scsm0 transpose_name
      (genr_instructions_for_modeler_block_type
         (Sbt "MATTRAN")
         [
           Value_parameter (Si "type", Code_value (get_dtype ()));
           Value_parameter (Si "rule", Code_value (code_out_int 1));
         ]
         [] blk0 genr_info)
      nports
      (_2_, _1_) in

  let super_body =
    match sample_time with
    | "-1" ->
      let links = [
        Genr_utilities_scicos.
        add_directed_link_in_scsm0
          in_port1_name port_1_ transpose_name port_1_ [];
        add_directed_link_in_scsm0
          in_port2_name port_1_ mat_mul_name port_2_ [];
        add_directed_link_in_scsm0
          transpose_name port_1_ mat_mul_name port_1_ [];
        add_directed_link_in_scsm0
          mat_mul_name port_1_ out_port_name port_1_ [];
      ] in

      [
        in_port1;
        in_port2;
        transpose;
        mat_mul;
        out_port;
        links;
      ]

    | _ ->

      let sample_time = code_out_expression sample_time in
      let out_data_type = code_out_simulink_out_data_type out_data_type in

      let samphold1_name = "samphold1"
      and samphold2_name = "samphold2"
      and split_event_name = "split_event"
      and sample_clock_name = "sample_clock" in

      let split_event =
        Genr_utilities_scicos.
        split_event_in_scsm0 split_event_name (_1_, _2_) in

      let samphold1 =
        Genr_utilities_scicos.
        samphold_in_scsm0
          samphold1_name
          out_data_type
          external_activation_mode_dummy
          (_1_, _0_) in

      let samphold2 =
        samphold_in_scsm0
          samphold2_name
          out_data_type
          external_activation_mode_dummy
          (_1_, _1_) in

      let sample_clock =
        Genr_utilities_scicos.
        sample_clock_in_scsm0
          sample_clock_name sample_time (_1_, _3_) in

      let links = [
        add_directed_link_in_scsm0
          in_port1_name port_1_ samphold1_name port_1_ [];
        add_directed_link_in_scsm0
          in_port2_name port_1_ samphold2_name port_1_ [];
        add_directed_link_in_scsm0 samphold1_name port_1_ transpose_name port_1_ [];
        add_directed_link_in_scsm0 transpose_name port_1_ mat_mul_name port_1_ [];
        add_directed_link_in_scsm0 samphold2_name port_1_ mat_mul_name port_2_ [];
        add_directed_link_in_scsm0 mat_mul_name port_1_ out_port_name port_1_ [];
        Genr_utilities_scicos.
        add_event_link_in_scsm0
          sample_clock_name port_1_ split_event_name port_1_ [];
        add_event_link_in_scsm0
          split_event_name port_1_ samphold1_name port_1_ [];
        add_event_link_in_scsm0
          split_event_name port_2_ samphold2_name port_1_ [];
      ] in

      [
        in_port1;
        in_port2;
        split_event;
        samphold1;
        samphold2;
        mat_mul;
        transpose;
        sample_clock;
        out_port;
        links;
      ] in

  super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type_native genr
;;

add_translator mdl_block_type_reference genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
