(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_from_workspace.ml,v 1.1 2019-05-10 14:05:18 jpc Exp $ *)

(* In scicos we must use module Genr_utilities_scicos and not
   Genr_utilities_scicos_pro. *)
module Genr_utilities_scicos_pro = struct end;;

open Mdl_block_types
open Instructions
open Genr_types
open Genr_utilities_scicos
open Unatural
;;

let mdl_block_type = simulink_block_type FromWorkspace;;

let method_of_output_after_value = function
  | "Setting to zero" -> 0
  | "Holding final value" -> 1
  | "Cyclic repetition" -> 2
  | "Extrapolation" -> 3
  | s ->
    Say.warning (
      Printf.sprintf
        "genr_from_workspace: output method %s is not supported,\
         holding final value"
        s);
    1
;;

(* In scicos there are four interpolation methods (0, 1, 2 or 3), whereas
  simulink only stipulates a boolean (interpolation or not). *)
let interpolation_method_of_simulink_interpolate = function
  | false -> 0
  | true -> 1
;;

let variable_name_ident =
  Ast_types.ident_from_string "VariableName"
and interpolate_ident =
  Ast_types.ident_from_string "Interpolate"
and output_after_final_value_ident =
  Ast_types.ident_from_string "OutputAfterFinalValue"
;;

let genr block_id genr_info =

  let sample_time = Genr_utilities_scicos.get_sample_time genr_info in

  let variable_name =
    code_out_expression
      (Genr_utilities_scicos.
       get_ident_string genr_info variable_name_ident)
  and interpolate =
    Genr_utilities_scicos.
      code_out_int
      (interpolation_method_of_simulink_interpolate
         (Genr_utilities_scicos.
          get_ident_bool genr_info interpolate_ident))
  and zcr = Genr_utilities_scicos.get_zero_crossing_value genr_info
  and output_after_final_value =
    Genr_utilities_scicos.
      get_ident_string genr_info output_after_final_value_ident in

  let scicos_block_parameters = [
    Genr_types.Value_parameter
      (Instructions.Si "varnam", Genr_types.Code_value variable_name);
    Value_parameter (Si "Method", Code_value interpolate);
    Value_parameter (Si "ZC", Code_value zcr);
    Value_parameter (Si "OutEnd",
      Code_value (code_out_int (
        method_of_output_after_value output_after_final_value)));
  ] in

  let scicos_basic_block_instructions blk_id =
    genr_instructions_for
      mdl_block_type scicos_block_parameters
      [] blk_id genr_info in

  let sample_time = code_out_expression sample_time in

  let out_port_name = "out"
  and from_workspace_name = "from_workspace"
  and sample_clock_name = "sample_clock" in

  let out_port =
    Genr_utilities_scicos.
      out_port_in_scsm0 out_port_name port_1_ (_1_, _0_) in
  let from_workspace =
    let nports = (_0_, _1_) in
    Genr_utilities_scicos.
    add_block_body_in_scsm0 from_workspace_name
      (scicos_basic_block_instructions blk0)
      nports
      (_0_, _0_) in
  let sample_clock =
    Genr_utilities_scicos.
    sample_clock_in_scsm0
      sample_clock_name sample_time (_0_, _1_) in

  let links = [
    Genr_utilities_scicos.
    add_directed_link_in_scsm0
      from_workspace_name port_1_ out_port_name port_1_ [];
    Genr_utilities_scicos.
    add_event_link_in_scsm0
      sample_clock_name port_1_ from_workspace_name port_1_ [];
  ] in

  let super_body = [
    out_port;
    from_workspace;
    sample_clock;
    links;
  ] in

  super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
