(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_integrator.ml,v 1.1 2019-05-10 14:05:18 jpc Exp $ *)

(* In scicos we must use module Genr_utilities_scicos and not
   Genr_utilities_scicos_pro. *)
module Genr_utilities_scicos_pro = struct end;;

open Mdl_block_types
open Concat_list
open Instructions
open Genr_types
open Genr_utilities_scicos
(* open Unatural *)
;;

let mdl_block_type = simulink_block_type Integrator;;

let external_reset_ident =
  Ast_types.ident_from_string "ExternalReset"
and initial_condition_source_ident =
  Ast_types.ident_from_string "InitialConditionSource"
and initial_condition_ident =
  Ast_types.ident_from_string "InitialCondition"
  (* Limit_output_ident is a boolean (on or off) in Simulink.
     In scios this boolean should be translated as a small int.
     To obtain this effect, we use encoding into type Genr_types.saturation,
     since this boolean is written as a small int. *)
and limit_output_ident =
  Ast_types.ident_from_string "LimitOutput"
and upper_saturation_limit_ident =
  Ast_types.ident_from_string "UpperSaturationLimit"
and lower_saturation_limit_ident =
  Ast_types.ident_from_string "LowerSaturationLimit"
;;

let genr block_id genr_info =

  let get_ident_string = Genr_utilities_scicos.get_ident_string genr_info in

  let external_reset =
    get_ident_string external_reset_ident
  and initial_condition_source =
    get_ident_string initial_condition_source_ident in

  let saturation =
    get_ident_simulink_saturation genr_info limit_output_ident in

  let scicos_basic_block_instructions reinit blk_id external_init =
    let scicos_block_parameters = [
      Genr_types.Value_parameter (Instructions.Si "external init",
	Code_value (code_out_bool external_init));
      Value_parameter (Si "x0",
        Simulink_arg_expression initial_condition_ident);
      Value_parameter (Si "reinit",
        Code_value (code_out_bool reinit));
      Value_parameter (Si "satur",
        Simulink_arg_saturation_set saturation);
      Value_parameter (Si "maxp",
        Simulink_arg_expression upper_saturation_limit_ident);
      Value_parameter (Si "lowp",
	Simulink_arg_expression lower_saturation_limit_ident);
    ] in

    genr_instructions_for
      mdl_block_type scicos_block_parameters
      [] blk_id genr_info in

  match external_reset, initial_condition_source with
  | "none", "internal" ->
    list (scicos_basic_block_instructions false block_id false)

  | "none", "external" ->
      list (scicos_basic_block_instructions false block_id true)
(*	
    let in_port1_name = "in1"
    and in_port2_name = "in2"
    and out_port_name = "out"
    and integrator_name = "integrator"
    and event_gen_name = "event_gen" in

    let in_port1 =
      Genr_utilities_scicos.
      in_port_in_scsm0 in_port1_name port_1_ (_0_, _1_) in
    let in_port2 =
      Genr_utilities_scicos.
      in_port_in_scsm0 in_port2_name port_2_ (_0_, _0_) in
    let out_port =
      Genr_utilities_scicos.
      out_port_in_scsm0 out_port_name port_1_ (_3_, _0_) in
    let integrator =
      let nports = (_2_, _1_) in
      Genr_utilities_scicos.
      add_block_body_in_scsm0 integrator_name
        (scicos_basic_block_instructions true blk0)
        nports
        (_1_, _0_) in

    let event_gen =
      let nports = (_0_, _0_) in
      Genr_utilities_scicos.
      add_block_body_in_scsm0 event_gen_name
        (genr_instructions_for_modeler_block_type (Sbt "EVTGEN_f")
           [
            Value_parameter (Si "tt", Code_value (code_out_int 0));
           ]
           [] blk0 genr_info)
        nports
        (_1_, _1_) in

    let links = [
      Genr_utilities_scicos.
      add_directed_link_in_scsm0
        in_port1_name port_1_ integrator_name port_1_ [];
      add_directed_link_in_scsm0 in_port2_name port_1_ integrator_name port_2_ [];
      add_directed_link_in_scsm0 integrator_name port_1_ out_port_name port_1_ [];
      Genr_utilities_scicos.
      add_event_link_in_scsm0
        event_gen_name port_1_ integrator_name port_1_ [];
    ] in

    let super_body = [
      in_port1;
      in_port2;
      out_port;
      integrator;
      event_gen;
      links;
    ] in

    super_block_in_scsm0 block_id super_body
*)
  | ("rising" | "falling" | "either" | "level" | "level hold" as s), _ ->
    Say.warning
      (Printf.sprintf
         "Integrator Simulink block translation: \
          %S configuration is not yet implemented" s);
    Genr_scicos_not_yet_translated.genr block_id genr_info
  | s, _ ->
    Say.warning
      (Printf.sprintf
         "Integrator Simulink block translation: \
          unknown configuration %S" s);
    Genr_scicos_not_yet_translated.genr block_id genr_info
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
