(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_logic.ml,v 1.1 2019-05-10 14:05:18 jpc Exp $ *)

(* In scicos we must use module Genr_utilities_scicos and not
   Genr_utilities_scicos_pro. *)
module Genr_utilities_scicos_pro = struct end;;

open Mdl_block_types
open Concat_list
open Instructions
open Genr_types
open Genr_utilities_scicos
open Unatural
;;

(** Simulink block [Logic] *)

let mdl_block_type = simulink_block_type Logic;;

let operator_ident = Ast_types.ident_from_string "Operator"
and inputs_ident = Ast_types.ident_from_string "Inputs"
;;

let get_scicos_logical_operator = function
  | "AND" -> 0
  | "OR" -> 1
  | "NAND" -> 2
  | "NOR" -> 3
  | "XOR" -> 4
  | "NOT" -> 5
  | "NXOR" -> 6
  | s ->
    Genr_utilities_scicos.
    fatal_error
      (Printf.sprintf "Logic block: unsupported logical operator %S" s)
;;

let genr block_id genr_info =

  let get_ident_string =
    Genr_utilities_scicos.get_ident_string genr_info in

  let sample_time =
    Genr_utilities_scicos.get_sample_time genr_info in

  let out_data_type =
    Genr_utilities_scicos.get_simulink_out_data_type_value genr_info in

  let operator =
    get_scicos_logical_operator (get_ident_string operator_ident) in

  let nin =
    if operator = 5 then "1" else get_ident_string inputs_ident in

  let nin_nat = Unatural.of_string nin in

  let scicos_block_parameters = [
    Genr_types.
    Value_parameter (Instructions.Si "nin",
      Genr_types.Code_value (code_out_expression nin));
    Value_parameter (Si "rule", Code_value (code_out_int operator));
    Value_parameter (Si "Datatype", Code_value out_data_type);
    Value_parameter (Si "tp", Code_value (code_out_int 0));
  ] in

  let scicos_basic_block_instructions blk_id =
    genr_instructions_for
      mdl_block_type scicos_block_parameters
      [] blk_id genr_info in

  match sample_time with
  | "-1" ->
    (* This simply is a Scicos LOGICAL_OP block, with
       nin, rule, Datatype, tp. *)
    list (scicos_basic_block_instructions block_id)

  | _ ->

    let sample_time = code_out_expression sample_time in

    let in_names, in_ports =
      Genr_utilities_scicos.
      in_ports_in_scsm0 nin_nat in

    let out_port_name = "out_port"
    and samphold_name = "samphold"
    and logic_name = "logic"
    and sample_clock_name = "sample_clock" in

    let samphold =
      Genr_utilities_scicos.
      samphold_in_scsm0
        samphold_name
        out_data_type
        external_activation_mode_dummy
        (_1_, _0_) in
    let sample_clock =
      Genr_utilities_scicos.
      sample_clock_in_scsm0
        sample_clock_name sample_time (_1_, _1_) in
    let logic =
      let nports = (_1_, _1_) in
      Genr_utilities_scicos.
      add_block_body_in_scsm0 logic_name
        (scicos_basic_block_instructions blk0)
        nports
        (_2_, _0_) in
    let out_port =
      Genr_utilities_scicos.
      out_port_in_scsm0 out_port_name port_1_ (_3_, _0_) in

    let in_links =
      Genr_utilities_scicos.
      rev_map_i_start Unatural._1_
        (fun i in_port_name ->
         add_directed_link_in_scsm0
           in_port_name port_1_ logic_name
           (Genr_utilities_scicos.mdl_port_number_of_natural i) [])
        in_names in

    let links = [
      Genr_utilities_scicos.
      add_directed_link_in_scsm0 samphold_name port_1_ logic_name port_1_ [];
      add_directed_link_in_scsm0 logic_name port_1_ out_port_name port_1_ [];
      Genr_utilities_scicos.
      add_event_link_in_scsm0
        sample_clock_name port_1_ samphold_name port_1_ [];
    ] in

    let in_ports = list_of_clist in_ports in

    let super_body = [
      in_ports;
      samphold;
      sample_clock;
      logic;
      out_port;
      in_links;
      links;
    ] in

  super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
