(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_lookup2d.ml,v 1.1 2019-05-10 14:05:18 jpc Exp $ *)

(* In scicos we must use module Genr_utilities_scicos and not
   Genr_utilities_scicos_pro. *)
module Genr_utilities_scicos_pro = struct end;;

open Mdl_block_types
open Concat_list
open Instructions
open Genr_types
open Genr_utilities_scicos
open Unatural
;;

let mdl_block_type = simulink_block_type Lookup2D;;

let output_values_ident = Ast_types.ident_from_string "OutputValues"
and look_up_meth_ident = Ast_types.ident_from_string "LookUpMeth"
and table_ident = Ast_types.ident_from_string "Table"
and row_index_ident = Ast_types.ident_from_string "RowIndex"
and column_index_ident = Ast_types.ident_from_string "ColumnIndex"
;;

let rec get_method_extrapo = function
  | "Interpolation-Extrapolation" -> (1, 1)
  | "Interpolation-Use End Values" -> (1, 0)
  | "Use Input Above" -> (8, 0)
  | "Use Input Below" -> (0, 0)
  | "Use Input Nearest" -> (9, 0)
  | s ->
    Say.warning (
      Printf.sprintf "genr_lookup: extrapolation method %s is not\
         supported, using Interpolation-Extrapolation" s);
    get_method_extrapo "Interpolation-Extrapolation"
;;

let genr block_id genr_info =

  let get_ident_string = Genr_utilities_scicos.get_ident_string genr_info in

  let sample_time = Genr_utilities_scicos.get_sample_time genr_info in

  let row_index = get_ident_string row_index_ident
  and column_index = get_ident_string column_index_ident
  and out_data_type =
    Genr_utilities_scicos.get_simulink_out_data_type_value genr_info
  and table =
    if Genr_utilities_scicos.exists_argument genr_info table_ident then
      get_ident_string table_ident
    else
      get_parameter_value_string genr_info output_values_ident
  and look_up_meth = get_ident_string look_up_meth_ident in

  let sci_method = fst (get_method_extrapo look_up_meth) in

  let scicos_block_parameters = [
    Genr_types.Value_parameter
      (Instructions.Si "Method",
       Genr_types.Code_value (code_out_int sci_method));
    Value_parameter (Si "xx", Code_value (code_out_expression row_index));
    Value_parameter (Si "yy", Code_value (code_out_expression column_index));
    Value_parameter (Si "zz", Code_value (code_out_expression table));
    Value_parameter (Si "graf", Code_value (code_out_expression "n"));
  ] in

  let scicos_basic_block_instructions blk_id =
    genr_instructions_for
      mdl_block_type scicos_block_parameters
      [] blk_id genr_info in

  match sample_time with
  | "-1" -> list (scicos_basic_block_instructions block_id)

  | _ ->

    let sample_time = code_out_expression sample_time in

    let in_port1_name = "in_port1"
    and in_port2_name = "in_port2"
    and out_port_name = "out_port"
    and samphold1_name = "samphold1"
    and samphold2_name = "samphold2"
    and split_event_name = "split_event"
    and lookup_name = "lookup"
    and sample_clock_name = "sample_clock" in

    let in_port1 =
      Genr_utilities_scicos.
      in_port_in_scsm0 in_port1_name port_1_ (_0_, _0_) in
    let in_port2 =
      in_port_in_scsm0 in_port2_name port_2_ (_0_, _1_) in
    let out_port =
      Genr_utilities_scicos.
      out_port_in_scsm0 out_port_name port_1_ (_3_, _0_) in
    let split_event =
      Genr_utilities_scicos.
      split_event_in_scsm0 split_event_name (_1_, _2_) in
    let samphold1 =
      Genr_utilities_scicos.
      samphold_in_scsm0
        samphold1_name
        out_data_type
        external_activation_mode_dummy
        (_1_, _0_) in
    let samphold2 =
      samphold_in_scsm0
        samphold2_name
        out_data_type
        external_activation_mode_dummy
        (_1_, _1_) in
    let lookup =
      let nports = (_2_, _1_) in
      Genr_utilities_scicos.
      add_block_body_in_scsm0 lookup_name
        (scicos_basic_block_instructions blk0)
        nports
        (_2_, _0_) in
    let sample_clock =
      Genr_utilities_scicos.
      sample_clock_in_scsm0
        sample_clock_name sample_time (_1_, _3_) in

    let links = [
      Genr_utilities_scicos.
      add_directed_link_in_scsm0
        in_port1_name port_1_ samphold1_name port_1_ [];
      add_directed_link_in_scsm0
        in_port2_name port_1_ samphold2_name port_1_ [];
      add_directed_link_in_scsm0 samphold1_name port_1_ lookup_name port_1_ [];
      add_directed_link_in_scsm0 samphold2_name port_1_ lookup_name port_2_ [];
      add_directed_link_in_scsm0 lookup_name port_1_ out_port_name port_1_ [];
      Genr_utilities_scicos.
      add_event_link_in_scsm0
        sample_clock_name port_1_ split_event_name port_1_ [];
      add_event_link_in_scsm0
        split_event_name port_1_ samphold1_name port_1_ [];
      add_event_link_in_scsm0
        split_event_name port_2_ samphold2_name port_1_ [];
    ] in

    let super_body = [
      in_port1;
      in_port2;
      out_port;
      split_event;
      samphold1;
      samphold2;
      lookup;
      sample_clock;
      links;
    ] in

  super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
