(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_math.ml,v 1.1 2019-05-10 14:05:18 jpc Exp $ *)

(* In scicos we must use module Genr_utilities_scicos and not
   Genr_utilities_scicos_pro. *)
module Genr_utilities_scicos_pro = struct end;;

open Mdl_block_types
open Concat_list
open Instructions
open Genr_types
open Genr_utilities_scicos
open Unatural
;;

let mdl_block_type = simulink_block_type Math;;

let scicos_block_type =
  Genr_utilities_scicos.modeler_block_type_for mdl_block_type
;;

let operator_ident = Ast_types.ident_from_string "Operator";;

let exp_block = Instructions.Sbt "EXPBLK_m"
and log_block = Sbt "LOGBLK_f"
and inv_block = Sbt "INVBLK"
and pow_block = Sbt "POWBLK_f"
;;

let euler_number = Genr_utilities_scicos.code_out_expression "exp(1)"
and constant_10 = code_out_int 10
and constant_2 = code_out_int 2
;;

let si_base = Si "a";;

let get_block_type_params operator =
  match operator with
  | "exp" -> exp_block, [ (si_base, euler_number); ]
  | "10^u" -> scicos_block_type, [ (si_base, constant_10); ]
  | "log" -> log_block, [ (si_base, euler_number); ]
  | "log10" -> log_block, [ (si_base, constant_10); ]
  | "reciprocal" -> inv_block, []
  | "square" -> pow_block, [ (si_base, constant_2); ]
  | t ->
    Say.warning
      (Printf.sprintf
         "Math Simulink block translation: \
          Operator %s not yet implemented" t);
    raise Not_found
;;

let genr block_id genr_info =

  let sample_time = Genr_utilities_scicos.get_sample_time genr_info in

  let get_ident_string = Genr_utilities_scicos.get_ident_string genr_info in

  let operator = get_ident_string operator_ident in

  match sample_time with
  | "-1" ->
    begin
      try
        let scicos_block_type, params = get_block_type_params operator in

        let instructions =
          if scicos_block_type = inv_block then [
            Instantiate_block (block_id, inv_block);
          ] else

          let scicos_parameters =
            match params with
            | [] -> assert false
            | param :: params ->
              Ulist1.map
                Genr_utilities_scicos.
                block_parameter_binding (param, params) in
          [
            Instantiate_block (block_id, scicos_block_type);
            Set_block_parameters (block_id, scicos_parameters);
          ] in

        list instructions

      with
      | Not_found -> Genr_scicos_not_yet_translated.genr block_id genr_info
    end

  | _ when not (
    operator = "exp" || operator = "10^u" || operator = "log" ||
    operator="log10" || operator="reciprocal" || operator = "square"
    ) ->
    Say.warning
      (Printf.sprintf
          "Math Simulink block translation: \
           Operator %s not yet implemented" operator);
    Genr_scicos_not_yet_translated.genr block_id genr_info
  | _ ->

    let sample_time = code_out_expression sample_time in

    let in_port_name = "in"
    and out_port_name = "out"
    and samphold_name = "samphold"
    and math_name = "math"
    and sample_clock_name = "sample_clock" in

    let inherit_data_type = code_out_int (-1) in

    let in_port =
      Genr_utilities_scicos.
      in_port_in_scsm0 in_port_name port_1_ (_0_, _0_) in
    let out_port =
      Genr_utilities_scicos.
      out_port_in_scsm0 out_port_name port_1_ (_3_, _0_) in
    let samphold =
      Genr_utilities_scicos.
      samphold_in_scsm0
        samphold_name
        inherit_data_type
        external_activation_mode_dummy
        (_1_, _0_) in

    let nports = (_1_, _1_) in

    let math =
      match operator with
      | "exp" ->
         Genr_utilities_scicos.
         add_block_body_in_scsm0 math_name
           (genr_instructions_for_modeler_block_type exp_block
              [ Genr_types.Value_parameter
                  (si_base, Genr_types.Code_value euler_number); ]
              [] blk0 genr_info)
           nports
           (_2_, _0_)
      | "10^u" ->
         Genr_utilities_scicos.
         add_block_body_in_scsm0 math_name
           (genr_instructions_for mdl_block_type
              [ Value_parameter (si_base, Code_value constant_10); ]
              [] blk0 genr_info)
           nports
           (_2_, _0_)
      | "log" ->
         add_block_body_in_scsm0 math_name
           (genr_instructions_for_modeler_block_type log_block
              [ Value_parameter (si_base, Code_value euler_number); ]
              [] blk0 genr_info)
           nports
           (_2_, _0_)
      | "log10" ->
         add_block_body_in_scsm0 math_name
           (genr_instructions_for_modeler_block_type log_block
              [ Value_parameter (si_base, Code_value constant_10); ]
              [] blk0 genr_info)
           nports
           (_2_, _0_)
      | "reciprocal" ->
         add_block_body_in_scsm0 math_name
           (genr_instructions_for_modeler_block_type inv_block
              []
              [] blk0 genr_info)
           nports
           (_2_, _0_)
      | "square" ->
         add_block_body_in_scsm0 math_name
           (genr_instructions_for_modeler_block_type pow_block
              [ Value_parameter (si_base, Code_value constant_2); ]
              [] blk0 genr_info)
           nports
           (_2_, _0_)
      | _ -> assert false in

    let sample_clock =
      Genr_utilities_scicos.
      sample_clock_in_scsm0
        sample_clock_name sample_time (_1_, _1_) in

    let links = [
      Genr_utilities_scicos.
      add_directed_link_in_scsm0
        in_port_name port_1_ samphold_name port_1_ [];
      add_directed_link_in_scsm0 samphold_name port_1_ math_name port_1_ [];
      add_directed_link_in_scsm0 math_name port_1_ out_port_name port_1_ [];
      Genr_utilities_scicos.
      add_event_link_in_scsm0
        sample_clock_name port_1_ samphold_name port_1_ [];
    ] in

    let super_body = [
      in_port;
      samphold;
      math;
      sample_clock;
      out_port;
      links;
    ] in

  super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
