(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_matlab_function.ml,v 1.1 2019-05-10 14:05:18 jpc Exp $ *)

(* In scicos we must use module Genr_utilities_scicos and not
   Genr_utilities_scicos_pro. *)
module Genr_utilities_scicos_pro = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos
;;

let mdl_block_type =
  Mdl_fake_block_types.matlab_function_block_type
;;

let matlab_function_name_ident = Mdl_rename.matlab_function_name_ident
and matlab_function_script_ident = Mdl_rename.matlab_function_script_ident
and block_input_offsets_ident = Mdl_rename.block_input_offsets_ident
and block_parameter_offsets_ident = Mdl_rename.block_parameter_offsets_ident
and block_argument_sizes_ident = Mdl_rename.block_argument_sizes_ident
and block_return_sizes_ident = Mdl_rename.block_return_sizes_ident
and parameters_ident =
  Ast_types.ident_from_string "Parameters"
;;

let genr block_id genr_info =
  let get_ident_string =
    Genr_utilities_scicos.get_ident_string genr_info
  and get_ident_int_vector =
    Genr_utilities_scicos.get_ident_int_vector genr_info
  and get_ident_code =
    Genr_utilities_scicos.get_argument_value_code genr_info in

  let function_name = get_ident_string matlab_function_name_ident in

  let function_script_value = get_ident_code matlab_function_script_ident in

  let input_offsets = get_ident_int_vector block_input_offsets_ident
  and parameter_offsets =
    get_ident_int_vector block_parameter_offsets_ident
  and argument_sizes =
    get_argument_value genr_info block_argument_sizes_ident
  and return_sizes =
    get_argument_value genr_info block_return_sizes_ident in

  let opar_expression =
    if exists_argument genr_info parameters_ident then
      let parameters = get_ident_string parameters_ident in
      Printf.sprintf "list(%s)" parameters
    else
      "list()" in

  let nin, nout = Genr_utilities_common.get_nin_nout genr_info in
  let nparams = Array.length parameter_offsets in
  let ninputs = Array.length input_offsets in

  let body_in =
    let b = Buffer.create 256 in
    let rec loop i =
      if i <= Unatural.to_int nin then
        let line =
          Printf.sprintf "\
            u%d = blk.inptr(%d);\n"
            i i in
        Buffer.add_string b line;
        loop (succ i) in
    loop 1;
    Buffer.contents b in

  let body_out =
    let b = Buffer.create 256 in
    let rec loop i =
      if i <= Unatural.to_int nout then
        let line =
          Printf.sprintf "\
            blk.outptr(%d) = y%d;\n"
            i i in
        Buffer.add_string b line;
        loop (succ i) in
    loop 2;
    Buffer.contents b in

  let body_params =
    let b = Buffer.create 256 in
    let rec loop i =
      if i <= nparams then
      let line =
        Printf.sprintf "\
          p%d = blk.opar(%d);\n"
          i i in
      Buffer.add_string b line;
      loop (succ i) in
    loop 1;
    Buffer.contents b in

  let body_call =
    let arg_out_opt =
      if Unatural.ge Unatural._1_ nout then None else
      let b = Buffer.create 256 in
      let rec loop i =
        let line = Printf.sprintf "y%d" i in
        Buffer.add_string b line;
        if i < Unatural.to_int nout then
          begin
            Buffer.add_char b ',';
            loop (succ i)
          end in
      loop 2;
      Some (Buffer.contents b) in

    let arg_in =
      let v = Array.make (ninputs + nparams) "" in
      Array.iteri (fun i offset ->
        v.(offset) <- Printf.sprintf "u%d" (succ i))
        input_offsets;
      Array.iteri (fun i offset ->
        v.(offset) <- Printf.sprintf "p%d" (succ i))
        parameter_offsets;
      String.concat ", " (Array.to_list v) in
    match arg_out_opt with
    | None ->
      Printf.sprintf "%s(%s)" function_name arg_in
    | Some arg_out ->
      Printf.sprintf "[%s]=%s(%s)"
        arg_out function_name arg_in in

  let translate_script_value function_script_string =
    Printf.sprintf "\
      function blk = sciblk(blk, flag)\
    \n  %s\
    \n  %s\
    \n  %s\
    \n  if flag == 1\
    \n    %s\
    \n    %s\
    \n  end\
    \nend"
      function_script_string
      body_in body_params body_call body_out in

  let function_script_value =
    rewrite_script
      translate_script_value matlab_function_script_ident
      function_script_value in

  let it_or_ot_expression n =
    if n = 0 then
      "[]"
    else
      let arg_in =
        let b = Buffer.create 256 in
        let rec loop i =
          Buffer.add_char b '1';
          if i < n - 1 then
            begin Buffer.add_char b ';'; loop (succ i) end in
        loop 0;
        Buffer.contents b in
      Printf.sprintf "[%s]" arg_in in

  let it_expression = it_or_ot_expression (Unatural.to_int nin) in
  let ot_expression = it_or_ot_expression (Unatural.to_int nout) in

  let rewrite_output_sizes sizes =
    let dummy_output_sizes = "[1,1]" in
    let dims = String.concat "," sizes in
    if dims = "" then
      Printf.sprintf "scicos_get_out_indices(%s)"
        dummy_output_sizes else
      Printf.sprintf "scicos_get_out_indices(%s,%s)"
        dummy_output_sizes dims in

  let out_sizes_expression =
    rewrite_expression_vector
      rewrite_output_sizes block_return_sizes_ident return_sizes in

  let rewrite_input_sizes sizes =
    let sizes = Array.of_list sizes in
    let b = Buffer.create 128 in
    begin
      if Array.length input_offsets = 0 then
        let append_gen_size () = Buffer.add_string b "-1" in
        let rec mk_generic_sizes = function
          | 0 -> ()
          | 1 -> append_gen_size ()
          | k ->
            append_gen_size ();
            Buffer.add_char b ',';
            mk_generic_sizes (pred k) in
        mk_generic_sizes (Unatural.to_int nin)
      else
        Array.iteri
          (fun i offset ->
           if i > 0 then
             Buffer.add_string b ", ";
           Buffer.add_string b sizes.(offset))
          input_offsets
    end;
    Printf.sprintf "scicos_get_in_indices(%s)"
      (Buffer.contents b) in

  let in_sizes_expression =
    rewrite_expression_vector
      rewrite_input_sizes block_argument_sizes_ident argument_sizes in

  let scicos_block_parameters = [

    Genr_types.Value_parameter
      (Instructions.Si "junction_name",
       Genr_types.Code_value (code_out_expression "sciblk"));
    Value_parameter
      (Si "impli", Code_value (code_out_expression "n"));
    Value_parameter
      (Si "in", Code_value in_sizes_expression);
    Value_parameter
      (Si "it",  Code_value (code_out_expression it_expression));
    Value_parameter
      (Si "out", Code_value out_sizes_expression);
    Value_parameter
      (Si "ot", Code_value (code_out_expression ot_expression));
    Value_parameter
      (Si "ci", Code_value (code_out_expression "[]"));
    Value_parameter
      (Si "co", Code_value (code_out_expression "[]"));
    Value_parameter
      (Si "xx", Code_value (code_out_expression "[]"));
    Value_parameter
      (Si "z", Code_value (code_out_expression "[]"));
    Value_parameter
      (Si "oz", Code_value (code_out_expression "list()"));
    Value_parameter
      (Si "rpar", Code_value (code_out_expression "[]"));
    Value_parameter
      (Si "ipar", Code_value (code_out_expression "[]"));
    Value_parameter
      (Si "opar", Code_value (code_out_expression opar_expression));
    Value_parameter
      (Si "nmode", Code_value (code_out_int 0));
    Value_parameter
      (Si "nzcr", Code_value (code_out_int 0));
    Value_parameter
      (Si "auto0", Code_value (code_out_expression "[]"));
    Value_parameter
      (Si "depu", Code_value (code_out_expression "y"));
    Value_parameter
      (Si "dept", Code_value (code_out_expression "n"));
    Value_parameter
      (Si "block_function_script",
       Code_value function_script_value);
  ] in

  let scicos_basic_block_instructions blk_id =
    genr_instructions_for
      mdl_block_type scicos_block_parameters
      [] blk_id genr_info in

  Concat_list.list (scicos_basic_block_instructions block_id)
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
