(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_mport_switch.ml,v 1.1 2019-05-10 14:05:18 jpc Exp $ *)

(* In scicos we must use module Genr_utilities_scicos and not
   Genr_utilities_scicos_pro. *)
module Genr_utilities_scicos_pro = struct end;;

open Mdl_block_types
open Concat_list
open Instructions
open Genr_types
open Genr_utilities_scicos
open Unatural
;;

let mdl_block_type = simulink_block_type MultiPortSwitch;;

let get_rule = function
  | "Floor" -> 3
  | "Round" -> 1
  | "Zero" -> 0
  | "Ceiling" -> 2
  | s ->
    Genr_utilities_scicos.warning (Printf.sprintf
      "Rounding rule %s not supported; Round used instead."
       s);
    1

and get_base = function
  | true -> 0
  | false -> 1
;;

let inputs_ident =
  Ast_types.ident_from_string "Inputs"
and rnd_meth_ident =
  Ast_types.ident_from_string "RndMeth"
and zero_idx_ident =
  Ast_types.ident_from_string "zeroidx"
;;

let genr block_id genr_info =

  let sample_time = Genr_utilities_scicos.get_sample_time genr_info in

  let nin = Genr_utilities_scicos.get_ident_int genr_info inputs_ident in
  let nin_nat = Unatural.succ (Unatural.of_int nin) in

  let rule =
    get_rule
      (Genr_utilities_scicos.get_ident_string genr_info rnd_meth_ident)

  and base =
    get_base
     (Genr_utilities_scicos.get_ident_bool genr_info zero_idx_ident) in

  let scicos_block_parameters = [
    Genr_types.Value_parameter
      (Instructions.Si "nin", Genr_types.Code_value (code_out_int nin));
    Value_parameter (Si "base", Code_value (code_out_int base));
    Value_parameter (Si "rule", Code_value (code_out_int rule));
  ] in

  let scicos_basic_block_instructions blk_id =
    genr_instructions_for
      mdl_block_type scicos_block_parameters
      [] blk_id genr_info in

  match sample_time with
  | "-1" -> list (scicos_basic_block_instructions block_id)

  | _ ->

    let sample_time = code_out_expression sample_time in

    let out_data_type =
      Genr_utilities_scicos.
      get_simulink_out_data_type_value genr_info in

    let in_names, in_ports =
      Genr_utilities_scicos.
      in_ports_in_scsm0 nin_nat in

    let samphold_names, sampholds =
      Genr_utilities_scicos.
      sampholds_in_scsm0
        out_data_type external_activation_mode_dummy nin_nat in

    let split_event_name = "split_event"
    and sample_clock_name = "sample_clock" in

    let out_port_name = "out"
    and mport_switch_name = "mport_switch" in

    let split_event =
      Genr_utilities_scicos.
      split_event_in_scsm0
        split_event_name (Unatural._1_, nin_nat) in
    let sample_clock =
      Genr_utilities_scicos.
      sample_clock_in_scsm0
        sample_clock_name sample_time (_1_, Unatural.succ nin_nat) in

    let out_port =
      Genr_utilities_scicos.
      out_port_in_scsm0 out_port_name port_1_ (_3_, _0_) in

    let mport_switch =
      let nports = (nin_nat, _1_) in
      Genr_utilities_scicos.
      add_block_body_in_scsm0 mport_switch_name
        (scicos_basic_block_instructions blk0)
        nports
        (_2_, _0_) in

    let in_links =
      Genr_utilities_scicos.
      rev_map_i_start2 _1_
        (fun _i in_port_name samphold_name ->
         add_directed_link_in_scsm0 in_port_name port_1_ samphold_name port_1_ [])
        in_names
        samphold_names in

    let samphold_mport_switch_links =
      Genr_utilities_scicos.
      rev_map_i_start _1_
        (fun i samphold_port_name ->
         add_directed_link_in_scsm0
           samphold_port_name port_1_
           mport_switch_name
           (Genr_utilities_scicos.mdl_port_number_of_natural i)
           [])
        samphold_names in

    let samphold_split_links =
      Genr_utilities_scicos.
      rev_map_i_start _1_
        (fun i samphold_port_name ->
         add_event_link_in_scsm0
           split_event_name
           (Genr_utilities_scicos.mdl_port_number_of_natural i)
           samphold_port_name port_1_
           [])
        samphold_names in

    let links = [
      Genr_utilities_scicos.
      add_event_link_in_scsm0
        sample_clock_name port_1_ split_event_name port_1_ [];
      add_directed_link_in_scsm0 mport_switch_name port_1_ out_port_name port_1_ [];
    ] in

    let in_ports = list_of_clist in_ports in
    let sampholds = list_of_clist sampholds in

    let super_body =
    [
      in_ports;
      sampholds;
      split_event;
      sample_clock;
      out_port;
      mport_switch;
      in_links;
      samphold_mport_switch_links;
      samphold_split_links;
      links;
    ] in

    super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
