(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_relational_operator.ml,v 1.1 2019-05-10 14:05:18 jpc Exp $ *)

(* In scicos we must use module Genr_utilities_scicos and not
   Genr_utilities_scicos_pro. *)
module Genr_utilities_scicos_pro = struct end;;

open Mdl_block_types
open Genr_utilities_scicos
open Unatural
;;

let mdl_block_type = simulink_block_type RelationalOperator;;

let operator_ident = Ast_types.ident_from_string "Operator";;

let genr block_id genr_info =

  let sample_time = Genr_utilities_scicos.get_sample_time genr_info in

  let op = Genr_utilities_scicos.get_relop_value genr_info operator_ident
(*  and input_same_data_type = get_value genr_info "InputSameDT" *)
  and zcr = Genr_utilities_scicos.get_zero_crossing_value genr_info
  and out_data_type =
    Genr_utilities_scicos.get_simulink_out_data_type_value genr_info in

  let in1_name = "in1"
  and in2_name = "in2"
  and out_name = "out"
  and relational_op_name = "relational_op"
  and split_name = "split"
  and convert_name = "convert" in

  let is_nan_op =
    let op_as_string =
      Genr_utilities_common.get_argument_value_string
        genr_info operator_ident in
    op_as_string = "isNaN" in

  let in_port1 =
    Genr_utilities_scicos.
    in_port_in_scsm0 in1_name port_1_ (_0_, _1_) in
  let in_port2 =
    Genr_utilities_scicos.
    in_port_in_scsm0 in2_name port_2_ (_0_, _0_) in
  let out_port =
    Genr_utilities_scicos.
    out_port_in_scsm0 out_name port_1_ (_4_, _1_) in

  let inherit_data_type = code_out_int (-1) in

  let relational_op =
    Genr_utilities_scicos.relational_op_in_scsm0
      relational_op_name
      op
      zcr
      out_data_type
      external_activation_mode_dummy
      (_2_, _1_) in

  let convert =
    let saturation = Genr_utilities_scicos.saturation_mode_off in
    Genr_utilities_scicos.
    convert_in_scsm0_with_saturation
      convert_name inherit_data_type saturation (_3_, _1_) in

  let in_port2_opt, inside_super_body =
    match sample_time, is_nan_op with
    | "-1", false ->

      let links = [
        Genr_utilities_scicos.
        add_directed_link_in_scsm0 in1_name port_1_ relational_op_name port_1_ [];
        add_directed_link_in_scsm0 in2_name port_1_ relational_op_name port_2_ [];
        add_directed_link_in_scsm0 relational_op_name port_1_ convert_name port_1_ [];
        add_directed_link_in_scsm0 convert_name port_1_ out_name port_1_ [];
      ] in
      Some in_port2,
      [
        links;
      ]

    | _, false ->

      let sample_time = code_out_expression sample_time in

      let sample_clock_name = "sample_clock"
      and samphold1_name = "samphold1"
      and samphold2_name = "samphold2"
      and clock_split_name = "clock_split" in

      let sample_clock =
        sample_clock_in_scsm0 sample_clock_name sample_time (_2_, _3_) in
      let samphold1 =
        samphold_in_scsm0
          samphold1_name
          inherit_data_type
          external_activation_mode_dummy
          (_1_, _1_) in
      let samphold2 =
        samphold_in_scsm0
          samphold2_name
          inherit_data_type
          external_activation_mode_dummy
          (_1_, _0_) in
      let clock_split = split_in_scsm0 clock_split_name (_2_, _2_) in

      let links = [
        Genr_utilities_scicos.
        add_directed_link_in_scsm0 in1_name port_1_ samphold1_name port_1_ [];
        add_directed_link_in_scsm0 in2_name port_1_ samphold2_name port_1_ [];
        add_directed_link_in_scsm0 samphold1_name port_1_ relational_op_name port_1_ [];
        add_directed_link_in_scsm0 samphold2_name port_1_ relational_op_name port_2_ [];
        add_directed_link_in_scsm0 relational_op_name port_1_ convert_name port_1_ [];
        add_directed_link_in_scsm0 convert_name port_1_ out_name port_1_ [];

        add_event_link_in_scsm0
          sample_clock_name port_1_ clock_split_name port_1_ [];
        add_event_link_in_scsm0
          clock_split_name port_1_ samphold1_name port_1_ [];
        add_event_link_in_scsm0
          clock_split_name port_2_ samphold2_name port_1_ [];
      ] in

      Some in_port2,
      [
        sample_clock;
        samphold1;
        samphold2;
        clock_split;
        links;
      ]

    | "-1", true ->

      let split =
        Genr_utilities_scicos.
          split_in_scsm0
            split_name (_1_, _1_) in

      let links = [
        Genr_utilities_scicos.
        add_directed_link_in_scsm0 in1_name port_1_ split_name port_1_ [];
        add_directed_link_in_scsm0 split_name port_1_ relational_op_name port_1_ [];
        add_directed_link_in_scsm0 split_name port_2_ relational_op_name port_2_ [];
        add_directed_link_in_scsm0 relational_op_name port_1_ convert_name port_1_ [];
        add_directed_link_in_scsm0 convert_name port_1_ out_name port_1_ [];
      ] in
      None,
      [
        split;
        links;
      ]

    | _, true ->

      let split =
        Genr_utilities_scicos.
          split_in_scsm0
            split_name (_1_, _1_) in

      let sample_time = code_out_expression sample_time in

      let sample_clock_name = "sample_clock"
      and samphold1_name = "samphold1"
      and samphold2_name = "samphold2"
      and clock_split_name = "clock_split" in

      let sample_clock =
        sample_clock_in_scsm0 sample_clock_name sample_time (_2_, _3_) in
      let samphold1 =
        samphold_in_scsm0
          samphold1_name
          inherit_data_type
          external_activation_mode_dummy
          (_1_, _1_) in
      let samphold2 =
        samphold_in_scsm0
          samphold2_name
          inherit_data_type
          external_activation_mode_dummy
          (_1_, _0_) in
      let clock_split = split_in_scsm0 clock_split_name (_2_, _2_) in

      let links = [
        Genr_utilities_scicos.
        add_directed_link_in_scsm0 in1_name port_1_ split_name port_1_ [];
        add_directed_link_in_scsm0 split_name port_1_ samphold1_name port_1_ [];
        add_directed_link_in_scsm0 split_name port_2_ samphold2_name port_1_ [];
        add_directed_link_in_scsm0 samphold1_name port_1_ relational_op_name port_1_ [];
        add_directed_link_in_scsm0 samphold2_name port_1_ relational_op_name port_2_ [];
        add_directed_link_in_scsm0 relational_op_name port_1_ convert_name port_1_ [];
        add_directed_link_in_scsm0 convert_name port_1_ out_name port_1_ [];

        add_event_link_in_scsm0
          sample_clock_name port_1_ clock_split_name port_1_ [];
        add_event_link_in_scsm0
          clock_split_name port_1_ samphold1_name port_1_ [];
        add_event_link_in_scsm0
          clock_split_name port_2_ samphold2_name port_1_ [];
      ] in

      None,
      [
        split;
        sample_clock;
        samphold1;
        samphold2;
        clock_split;
        links;
      ] in

  let super_body =
    let super_body =
      out_port
      :: relational_op
      :: convert
      :: inside_super_body in
    let super_body =
      match in_port2_opt with
      | None -> super_body
      | Some in_port2 -> in_port2 :: super_body in
    in_port1 :: super_body in

  super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
