(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_scope.ml,v 1.1 2019-05-10 14:05:19 jpc Exp $ *)

(* In scicos we must use module Genr_utilities_scicos and not
   Genr_utilities_scicos_pro. *)
module Genr_utilities_scicos_pro = struct end;;

open Concat_list
open Instructions
open Genr_types
open Genr_utilities_scicos
;;

let mdl_block_type =
  simulink_block_type Mdl_block_types.Scope
;;

let list_init n f =
  let rec loop accu = function
  | 0 -> accu
  | n -> loop (f n :: accu) (n - 1) in
  loop [] n
;;

let num_in_ports_ident =
  Ast_types.ident_from_string "NumInputPorts"
and y_min_ident =
  Ast_types.ident_from_string "YMin"
and y_max_ident =
  Ast_types.ident_from_string "YMax"
and time_range_ident =
  Ast_types.ident_from_string "TimeRange"
and location_ident =
  Ast_types.ident_from_string "Location"
and scope_specification_ident =
  Ast_types.ident_from_string "ScopeSpecificationString"
(*
FIXME: could also define axes titles ?
and axes_titles_ident =
  Ast_types.ident_from_string "AxesTitles"
*)
;;

let string_of_int_pair (x, y) = Printf.sprintf "[ %d; %d ]" x y;;

let scicos_vector_of_strings strs =
  Printf.sprintf "[%s]" (String.concat "; " strs)
;;

let get_min_max_bounds bounds_min bounds_max =
  let get_bounds bounds_string =
    let ib = Scanf.Scanning.from_string bounds_string in
    let rec loop accu =
      if Scanf.Scanning.end_of_input ib then List.rev accu else
      try
        Scanf.bscanf ib "%s@~"
          (fun s -> loop (s :: accu)) with
      | Scanf.Scan_failure s
      | Failure s ->
        Say.fatal_error
          (Printf.sprintf
             "Genr_scicos_scope: \
              while scanning ``%s'', encountered ``%s''"
             bounds_string s) in
    loop [] in

  (get_bounds bounds_min, get_bounds bounds_max)
;;

let genr block_id genr_info =

  let get_ident_string =
    Genr_utilities_scicos.get_ident_string genr_info
  and get_ident_int =
    Genr_utilities_scicos.get_ident_int genr_info
  and exists_argument =
    Genr_utilities_scicos.exists_argument genr_info
  and exists_parameter =
    Genr_utilities_scicos.exists_parameter genr_info
(* FIXME: used for axes titles
  and get_argument_value =
    Genr_utilities_scicos.get_argument_value genr_info
*)
  and get_ident_int_vector =
    Genr_utilities_scicos.get_ident_int_vector genr_info in

  let get_num_in_ports num_in_ports_ident =
    if exists_argument num_in_ports_ident then
      get_ident_int num_in_ports_ident else
    let ports_desc = genr_info.Genr_block_types.block_ports_description in
    let nin = Mdl_ast_to_deep_ast.get_regular_port_nin ports_desc in
    Unatural.to_int nin in

  (* For debugging purpose *)
  Genr_utilities_common.debug_with_info genr_info "Genr_scicos_scope";

  let num_in_ports = get_num_in_ports num_in_ports_ident in

  let y_min, y_max =
    let y_mins, y_maxs =
      if exists_parameter scope_specification_ident
      then get_min_max_bounds "[-5]" "[5]" else
      let bound_min = get_ident_string y_min_ident
      and bound_max = get_ident_string y_max_ident in
      get_min_max_bounds bound_min bound_max in
    scicos_vector_of_strings y_mins,
    scicos_vector_of_strings y_maxs in

(*
FIXME: could also define axes titles ?
  let axes_titles =
    let axes_titles_bindings =
      if not (exists_argument axes_titles_ident) then [] else
      let titles = get_argument_value axes_titles_ident in
      Mdl_deep_basics.bindings_of_deep_value "AxesTitles" titles in
    let rec loop accu = function
      | 0 -> accu
      | i -> loop (get_axis_title axes_titles_bindings i :: accu) (pred i) in
    loop [] num_in_ports in
*)

  let time_range =
    if exists_argument time_range_ident
    then get_ident_string time_range_ident
    else "20"

  and location =
    if exists_argument location_ident
    then get_ident_int_vector location_ident
    else [| 0; 0; 200; 200; |]

  and block_name =
    genr_info.Genr_block_types.block_explicit_name in

  let per =
    if time_range = "auto" then "_final_simulation_time" else time_range
  and sz =
    string_of_int_pair
      (location.(2) - location.(0), location.(3) - location.(1))
  and og =
    string_of_int_pair (location.(0), location.(1)) in
    (* loc(1:2)  // this is probably wrong- must be height-loc(2)+-sz(2) *)

  let scicos_block_parameters = [
    Genr_types.Value_parameter
      (Instructions.Si "clrs",
       Genr_types.Code_value
         (code_out_expression "[ 1, 3, 5, 7, 9, 11, 13, 15 ]"));
    Value_parameter (Si "win", Code_value (code_out_int (-1)));
    Value_parameter (Si "wpos", Code_value (code_out_expression og));
    Value_parameter (Si "wdim", Code_value (code_out_expression sz));
    Value_parameter (Si "ymin", Code_value (code_out_expression y_min));
    Value_parameter (Si "ymax", Code_value (code_out_expression y_max));
    Value_parameter (Si "per", Code_value (code_out_expression per));
    Value_parameter (Si "N", Code_value (code_out_int 15));
    Value_parameter (Si "heritance", Code_value (code_out_int 1));
    Value_parameter (Si "nom",
      Code_value (code_out_string (
        Ast_types.string_of_mdl_name block_name)));
  ] in

  let scicos_basic_block_instructions blk_id =
    genr_instructions_for
      mdl_block_type scicos_block_parameters
      [] blk_id genr_info in

  match num_in_ports with
  | 1 -> list (scicos_basic_block_instructions block_id)
  | _ ->
    let in_sizes =
      list_init num_in_ports (fun i -> string_of_int (- i)) in
    let sizes =
      let sz = String.concat ", " in_sizes in
      Printf.sprintf "[ %s ]" sz in
    let scicos_block_parameters =
      (Value_parameter (Si "in", Code_value (code_out_expression sizes))) ::
      scicos_block_parameters in

    list (
      genr_instructions_for_modeler_block_type
        (Sbt "CMSCOPE") scicos_block_parameters
        [] block_id genr_info)
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
