(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_selector.ml,v 1.1 2019-05-10 14:05:19 jpc Exp $ *)

(* In scicos we must use module Genr_utilities_scicos and not
   Genr_utilities_scicos_pro. *)
module Genr_utilities_scicos_pro = struct end;;

open Mdl_block_types
open Concat_list
open Genr_utilities_scicos
open Unatural
;;

let mdl_block_type = simulink_block_type Selector;;

let number_of_dimensions_ident =
  Ast_types.ident_from_string "NumberOfDimensions"
and index_mode_ident =
  Ast_types.ident_from_string "IndexMode"
and index_options_ident =
  Ast_types.ident_from_string "IndexOptions"
and indices_ident =
  Ast_types.ident_from_string "Indices"
and elements_ident =
  Ast_types.ident_from_string "Elements"
;;

let genr block_id genr_info =

  let sample_time = Genr_utilities_scicos.get_sample_time genr_info in

  let get_ident_string = Genr_utilities_scicos.get_ident_string genr_info in

  let default_number_of_dimensions = "1"
  and default_index_options = "Index vector (dialog)" in

  let number_of_dimensions =
    if Genr_utilities_scicos.
       exists_argument genr_info number_of_dimensions_ident
    then get_ident_string number_of_dimensions_ident
    else default_number_of_dimensions

  and index_mode = get_ident_string index_mode_ident

  and index_options =
    if Genr_utilities_scicos.exists_argument genr_info index_options_ident
    then get_ident_string index_options_ident
    else default_index_options

  and indices_string =
    if Genr_utilities_scicos.exists_argument genr_info indices_ident
    then get_ident_string indices_ident
    else get_parameter_value_string genr_info elements_ident in

  let ind =
    match index_mode with
    | "One-based" -> indices_string
    | _ -> Printf.sprintf "(%s) + 1" indices_string in

  let scicos_block_parameters = [
    Genr_types.Value_parameter
      (Instructions.Si "ind",
       Genr_types.Code_value (code_out_expression ind));
  ] in

  let scicos_basic_block_instructions blk_id =
    genr_instructions_for
      mdl_block_type scicos_block_parameters
      [] blk_id genr_info in

  match number_of_dimensions, index_options with
  | ("1", index_options) when index_options = default_index_options ->
    begin match sample_time with
    | "-1" -> list (scicos_basic_block_instructions block_id)

    | _ ->

      let sample_time = code_out_expression sample_time in

      let in_port_name = "in"
      and out_port_name = "out"
      and samphold_name = "samphold"
      and extractor_name = "extractor"
      and sample_clock_name = "sample_clock" in

      let inherit_data_type = code_out_int (-1) in

      let in_port =
        Genr_utilities_scicos.
        in_port_in_scsm0 in_port_name port_1_ (_0_, _0_) in
      let out_port =
        Genr_utilities_scicos.
        out_port_in_scsm0 out_port_name port_1_ (_3_, _0_) in
      let samphold =
        Genr_utilities_scicos.
        samphold_in_scsm0
          samphold_name
          inherit_data_type
          external_activation_mode_dummy
          (_1_, _0_) in
      let extractor =
        let nports = (_1_, _1_) in
        Genr_utilities_scicos.
        add_block_body_in_scsm0 extractor_name
          (scicos_basic_block_instructions blk0)
          nports
          (_2_, _0_) in
      let sample_clock =
        Genr_utilities_scicos.
        sample_clock_in_scsm0
          sample_clock_name sample_time (_1_, _1_) in

      let links = [
        Genr_utilities_scicos.
        add_directed_link_in_scsm0
          in_port_name port_1_ samphold_name port_1_ [];
        add_directed_link_in_scsm0 samphold_name port_1_ extractor_name port_1_ [];
        add_directed_link_in_scsm0 extractor_name port_1_ out_port_name port_1_ [];
        Genr_utilities_scicos.
        add_event_link_in_scsm0
          sample_clock_name port_1_ samphold_name port_1_ [];
      ] in

      let super_body = [
        in_port;
        out_port;
        samphold;
        extractor;
        sample_clock;
        links;
      ] in

      super_block_in_scsm0 block_id super_body
    end

  | t1, t2 ->
    Say.warning
      (Printf.sprintf "Selector Simulink block translation: \
         Case of number of dimensions %s and index options %s \
         is not yet implemented" t1 t2);
    Genr_scicos_not_yet_translated.genr block_id genr_info
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
