(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_signal_generator.ml,v 1.1 2019-05-10 14:05:19 jpc Exp $ *)

(* In scicos we must use module Genr_utilities_scicos and not
   Genr_utilities_scicos_pro. *)
module Genr_utilities_scicos_pro = struct end;;

open Mdl_block_types
open Concat_list
open Instructions
open Genr_types
open Genr_utilities_scicos
;;

let mdl_block_type = simulink_block_type SignalGenerator;;

let wave_form_ident =
  Ast_types.ident_from_string "WaveForm"
and time_soource_ident =
  Ast_types.ident_from_string "TimeSource"
and amplitude_ident =
  Ast_types.ident_from_string "Amplitude"
and frequency_ident =
  Ast_types.ident_from_string "Frequency"
and units_ident =
  Ast_types.ident_from_string "Units"
;;

let genr block_id genr_info =

  let get_ident_string = Genr_utilities_scicos.get_ident_string genr_info in

  let wave_form = get_ident_string wave_form_ident
  and time_source = get_ident_string time_soource_ident
  and amplitude = get_ident_string amplitude_ident
  and frequency = get_ident_string frequency_ident
  and units = get_ident_string units_ident in

  let get_F = function
    | "Hertz" -> Printf.sprintf "2 * pi * (%s)" frequency
    | _ -> frequency in

  match time_source with
  | "Use simulation time" ->
    begin match wave_form with
    | "sine" ->

      let scicos_block_parameters = [
        Genr_types.Value_parameter (Instructions.Si "M",
          Genr_types.Code_value (code_out_expression amplitude));
        Value_parameter (Si "F",
          Code_value (code_out_expression (get_F units)));
        Value_parameter (Si "P",
          Code_value (code_out_int 0));
      ] in

      list (
        genr_instructions_for
          mdl_block_type scicos_block_parameters
          [] block_id genr_info)

    | "square" ->

      let get_Amin () = Printf.sprintf "-abs(%s)" amplitude
      and get_Amax () = Printf.sprintf "abs(%s)" amplitude
      and get_rule () =
        Printf.sprintf "(%s >= 0) + 2 * (%s < 0)" amplitude amplitude in

      let scicos_block_parameters = [
        Value_parameter (Si "Amin",
          Code_value (code_out_expression (get_Amin ())));
        Value_parameter (Si "Amax",
          Code_value (code_out_expression (get_Amax ())));
        Value_parameter (Si "rule",
          Code_value (code_out_expression (get_rule ())));
        Value_parameter (Si "F",
          Code_value (code_out_expression (get_F units)));
      ] in

      let scicos_block_type = Sbt "GEN_SQR" in

      list (
        genr_instructions_for_modeler_block_type
          scicos_block_type scicos_block_parameters
          [] block_id genr_info)

    | w ->
      Say.warning
        (Printf.sprintf
           "Signal generator Simulink block translation: \
          \nWaveform %s not yet implemented" w);
      Genr_scicos_not_yet_translated.genr block_id genr_info
    end
  | ts ->
    Say.warning
      (Printf.sprintf
         "Signal generator Simulink block translation: \
        \nTime source %s not yet implemented" ts);
    Genr_scicos_not_yet_translated.genr block_id genr_info
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
