(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_sin.ml,v 1.1 2019-05-10 14:05:19 jpc Exp $ *)

(* In scicos we must use module Genr_utilities_scicos and not
   Genr_utilities_scicos_pro. *)
module Genr_utilities_scicos_pro = struct end;;

open Mdl_block_types
open Concat_list
open Instructions
open Genr_types
open Genr_utilities_scicos
open Unatural
;;

let mdl_block_type = simulink_block_type Sin;;

let phase_ident =
  Ast_types.ident_from_string "Phase"
and amplitude_ident =
  Ast_types.ident_from_string "Amplitude"
and frequency_ident =
  Ast_types.ident_from_string "Frequency"
and bias_ident =
  Ast_types.ident_from_string "Bias"
and time_source_ident =
  Ast_types.ident_from_string "TimeSource"
and sine_type_ident =
  Ast_types.ident_from_string "SineType"
;;

let genr block_id genr_info =

  let sample_time = Genr_utilities_scicos.get_sample_time genr_info in

  let get_ident_string = Genr_utilities_scicos.get_ident_string genr_info in

  let phase = get_ident_string phase_ident
  and amplitude = get_ident_string amplitude_ident
  and frequency = get_ident_string frequency_ident
  and bias = get_ident_string bias_ident in

  let sine_type = get_ident_string sine_type_ident
  and time_source = get_ident_string time_source_ident in

  let scicos_block_parameters = [
    Genr_types.Value_parameter (Instructions.Si "M",
      Genr_types.Code_value (code_out_expression amplitude));
    Value_parameter (Si "F",
      Code_value (code_out_expression frequency));
    Value_parameter (Si "P",
      Code_value (code_out_expression phase));
  ] in

  let scicos_basic_block_instructions blk_id =
    genr_instructions_for
      mdl_block_type scicos_block_parameters
      [] blk_id genr_info in

  let scicos_sample_time_block_instructions sample_time blk_id =

    let sample_time = code_out_expression sample_time in

    let out_port_name = "out"
    and sin_name = "sin"
    and virtual_clk0_name = "virtual_clk0"
    and sample_clock_name = "sample_clock" in

    let out_port =
      Genr_utilities_scicos.
      out_port_in_scsm0 out_port_name
        port_1_
        (_1_, Unatural._0_) in

    let sin =
      let nports = (_0_, _1_) in
      Genr_utilities_scicos.
      add_block_body_in_scsm0 sin_name
        (scicos_basic_block_instructions blk0)
        nports
        (_0_, _0_) in

    let virtual_clk0 =
      Genr_utilities_scicos.
      virtual_clk0_in_scsm0 virtual_clk0_name (_1_, _1_) in

    let sample_clock =
      Genr_utilities_scicos.
      sample_clock_in_scsm0 sample_clock_name sample_time (_1_, _2_) in

    let links = [
      Genr_utilities_scicos.add_directed_link_in_scsm0
        sin_name port_1_ out_port_name port_1_ [];
      Genr_utilities_scicos.add_event_link_in_scsm0
        sample_clock_name port_1_
        virtual_clk0_name port_1_
        [];
    ] in

    let super_body = [
      out_port;
      sin;
      virtual_clk0;
      sample_clock;
      links;
    ] in
    Genr_utilities_scicos.
    super_block_in_scsm0 blk_id super_body in

  match time_source, sine_type with
  | "Use simulation time", "Time based" when bias = "0" ->
    begin
      match sample_time with
      | "0" -> list (scicos_basic_block_instructions block_id)
      | sample_time ->
        scicos_sample_time_block_instructions sample_time block_id
    end
  | "Use simulation time", "Time based" ->
    Say.warning
      "Sin Simulink block translation: \
       non zero bias not yet supported";
    Genr_scicos_not_yet_translated.genr block_id genr_info
  | "Use simulation time", "Sample based" ->
    Say.warning
      "Sin Simulink block translation: \
       \"Frame Based\" type not yet implemented";
    Genr_scicos_not_yet_translated.genr block_id genr_info
  | "Use external signal", "Time based" ->
    Say.warning
      "Sin Simulink block translation: \
       \"Use external signal\" option not yet implemented";
    Genr_scicos_not_yet_translated.genr block_id genr_info
  | "Use external signal", "Sample based" ->
    Say.warning
      "Sin Simulink block translation: \
       \"Frame Based\" type not yet implemented";
    Genr_scicos_not_yet_translated.genr block_id genr_info
  | t1, t2 ->
    Say.warning
       (Printf.sprintf
          "Sin Simulink block translation: \
          time source %s or sine type %s not yet implemented" t1 t2);
    Genr_scicos_not_yet_translated.genr block_id genr_info
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
