(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_time_scope.ml,v 1.1 2019-05-10 14:05:19 jpc Exp $ *)

(* In scicos we must use module Genr_utilities_scicos and not
   Genr_utilities_scicos_pro. *)
module Genr_utilities_scicos_pro = struct end;;

open Mdl_block_types
open Concat_list
open Instructions
open Genr_types
open Genr_utilities_scicos
;;

let mdl_block_type = simulink_block_type TimeScope;;

let list_init n f =
  let rec loop accu = function
  | 0 -> accu
  | n -> loop (f n :: accu) (n - 1) in
  loop [] n
;;

let num_in_ports_ident =
  Ast_types.ident_from_string "NumInputPorts"
and _y_min_ident =
  Ast_types.ident_from_string "YMin"
and _y_max_ident =
  Ast_types.ident_from_string "YMax"
and _time_range_ident =
  Ast_types.ident_from_string "TimeRange"
and _location_ident =
  Ast_types.ident_from_string "Location"
;;

let y_min_default = "-5"
and y_max_default = "5"
and time_range_default = "30"
and location_default = [| 100; 100; 400; 400; |]
;;

let genr block_id genr_info =

  let get_bound bound_string =
    let ib = Scanf.Scanning.from_string bound_string in
    let rec loop accu =
      if Scanf.Scanning.end_of_input ib then List.rev accu else
      try
        Scanf.bscanf ib "%s@~"
          (fun s -> loop (s :: accu)) with
      | Scanf.Scan_failure s
      | Failure s ->
        Say.fatal_error
          (Printf.sprintf
             "Genr_time_scope.genr: while scanning %S, encountered %s"
             bound_string s) in
    Printf.sprintf "[%s]" (String.concat "; " (loop [])) in

  let num_in_ports =
    Genr_utilities_scicos.get_ident_int genr_info num_in_ports_ident
  and y_min = get_bound y_min_default
  and y_max = get_bound y_max_default
  and time_range = time_range_default
  and location = location_default
  and block_name =
    genr_info.Genr_block_types.block_explicit_name in

  let string_of_int_pair (x, y) =
    Printf.sprintf "[ %d; %d ]" x y in

  let per =
    if time_range = "auto" then "_final_simulation_time" else time_range
  and sz =
    string_of_int_pair
      (location.(2) - location.(0), location.(3) - location.(1))
  and og =
    string_of_int_pair (location.(0), location.(1)) in
      (* loc(1:2)  // this is probably wrong- must be height-loc(2)+-sz(2) *)

  let scicos_block_parameters = [
    Genr_types.Value_parameter (Instructions.Si "clrs",
      Genr_types.Code_value
        (code_out_expression "[ 1, 3, 5, 7, 9, 11, 13, 15 ]"));
    Value_parameter (Si "win", Code_value (code_out_int (-1)));
    Value_parameter (Si "wpos", Code_value (code_out_expression og));
    Value_parameter (Si "wdim", Code_value (code_out_expression sz));
    Value_parameter (Si "ymin", Code_value (code_out_expression y_min));
    Value_parameter (Si "ymax", Code_value (code_out_expression y_max));
    Value_parameter (Si "per", Code_value (code_out_expression per));
    Value_parameter (Si "N", Code_value (code_out_int 15));
    Value_parameter (Si "heritance", Code_value (code_out_int 1));
    Value_parameter (Si "nom",
      Code_value (code_out_string (
        Ast_types.string_of_mdl_name block_name)));
  ] in

  let scicos_basic_block_instructions blk_id =
    genr_instructions_for
      mdl_block_type scicos_block_parameters
      [] blk_id genr_info in

  match num_in_ports with
  | 1 -> list (scicos_basic_block_instructions block_id)
  | _ ->
    let in_sizes =
      list_init num_in_ports (fun i -> string_of_int (- i)) in
    let sizes =
      let sz = String.concat ", " in_sizes in
      Printf.sprintf "[ %s ]" sz in
    let scicos_block_parameters =
      (Value_parameter (Si "in", Code_value (code_out_expression sizes))) ::
      scicos_block_parameters in

    list (
      genr_instructions_for_modeler_block_type
        (Sbt "CMSCOPE") scicos_block_parameters
        [] block_id genr_info)
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
