(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_trigger_event_gen.ml,v 1.1 2019-05-10 14:05:19 jpc Exp $ *)

(* In scicos we must use module Genr_utilities_scicos and not
   Genr_utilities_scicos_pro. *)
module Genr_utilities_scicos_pro = struct end;;

(** Generate a super block for a Simulink sub system block that is triggered,
  meaning that is has an TriggerPort block in its sub system. *)

open Instructions
open Genr_types
open Genr_utilities_scicos
open Unatural
;;

type event_kind =
  | Rising
  | Falling
  | Either
  | FunctionCall
;;

let mdl_block_type_rising =
  Mdl_fake_block_types.trigger_rising_event_generator_block_type
and mdl_block_type_falling =
  Mdl_fake_block_types.trigger_falling_event_generator_block_type
and mdl_block_type_either =
  Mdl_fake_block_types.trigger_either_event_generator_block_type
and mdl_block_type_function_call =
  Mdl_fake_block_types.trigger_function_call_generator_block_type
;;

let mdl_block_type_and_scicos_arguments = function
  | Rising ->
    mdl_block_type_rising, []
  | Falling ->
    mdl_block_type_falling, []
  | Either ->
    mdl_block_type_either,
    [
      Genr_types.
      Value_parameter (Si "in",
      Code_value (code_out_expression  "1"));
    ]
  | FunctionCall ->
    mdl_block_type_function_call,
    [
      Genr_types.
      Value_parameter (Si "inh",
        Code_value (code_out_expression  "1"));
      Value_parameter (Si "nmod",
        Code_value (code_out_expression  "0"));
    ]
;;

let genr event_kind block_id genr_info =

  let trigger_name = "Trigger" in

  let in_port_name = "in"
  and out_port_event_name = "out" in

  let in_port =
    Genr_utilities_scicos.
    in_port_in_scsm0 in_port_name port_1_ (Unatural._0_, Unatural._1_) in
  let out_port_event =
    Genr_utilities_scicos.
    out_port_event_in_scsm0 out_port_event_name port_1_ (_1_, _0_) in

  let mdl_block_type, scicos_block_parameters =
    mdl_block_type_and_scicos_arguments event_kind in

  let scicos_block_instructions block_id =
    genr_instructions_for
      mdl_block_type scicos_block_parameters
      [] block_id genr_info in

  let trigger =
    let nports = (_1_, _0_) in
    Genr_utilities_scicos.add_block_body_in_scsm0
      trigger_name
      (scicos_block_instructions blk0)
      nports
      (_1_, _1_) in

  let links = [
    Genr_utilities_scicos.
    add_directed_link_in_scsm0
      in_port_name port_1_ trigger_name port_1_ [];
    Genr_utilities_scicos.
    add_event_link_in_scsm0
      trigger_name port_1_ out_port_event_name port_1_ [];
  ] in

  let super_body = [
    in_port;
    out_port_event;
    trigger;
    links;
  ] in

  super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type_rising (genr Rising);;
add_translator mdl_block_type_falling (genr Falling);;
add_translator mdl_block_type_either (genr Either);;
add_translator mdl_block_type_function_call (genr FunctionCall);;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
