(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_aliases.ml,v 1.1 2019-05-10 14:04:56 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Genr_utilities_scicos_pro;;
open Genr_types;;
open Instructions;;

let list_replicate n x =

  let rec loop accu = function
    | 0 -> accu
    | n -> loop (x :: accu) (n - 1) in

  if n < 0 then invalid_arg (
    Printf.sprintf "list_replicate: negative argument %d" n) else
  loop [] n
(** A List.replicate function. *)
;;

(* Idents commonly used in this file. *)

let inputs_ident = Ast_types.ident_from_string "Inputs"
and value_ident = Ast_types.ident_from_string "Value"
;;

let gain_ident = Ast_types.ident_from_string "Gain"
and multiplication_ident = Ast_types.ident_from_string "Multiplication"
;;

(** Definition of block translators *)

(** Registering translators *)

let add_scicos_pro_translation =
  Genr_utilities_scicos_pro.add_modeler_translation
;;

let add_scicos_pro_equivalent =
  Genr_utilities_scicos_pro.add_modeler_equivalent
;;

let add_scicos_pro_conditional_translation1 =
  Genr_utilities_scicos_pro.add_modeler_conditional_translation1
;;

let add_scicos_pro_conditional_translation2 =
  Genr_utilities_scicos_pro.add_modeler_conditional_translation2
;;

let add_scicos_pro_conditional_translation3 =
  Genr_utilities_scicos_pro.add_modeler_conditional_translation3
;;

let scicos_pro_block_type_for =
  Genr_utilities_scicos_pro.modeler_block_type_for
;;

(** Simulink block [ActionPort]
  This block is a simple marker for an IfAction sub-system.
  It is suppressed when translating the corresponding sub-system block of
  action. *)

(** Simulink block [Algebraic Constraint] *)

let mdl_algebraic_constraint_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Algebraic Constraint")
    (Ast_types.mdl_path_of_string
       "simulink/Math\nOperations/Algebraic Constraint") in

add_scicos_pro_translation
  mdl_algebraic_constraint_type
  [
    Value_parameter (Si "dae",
      Code_value (code_out_int 0));
    Value_parameter (Si "x0",
      Simulink_arg_expression (Ast_types.ident_from_string "z0"));
  ]
  []
;;

(** Simulink block [Slider Gain] *)
(* FIXME. Slider is translated only to a gain block,
   ignoring the bounds for the moment...
let low_ident = ident_from_string "low"
and hi_ident = ident_from_string "hi" *)

let mdl_slider_gain_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Slider Gain")
    (Ast_types.mdl_path_of_string
       "simulink/Math\nOperations/Slider\nGain") in

add_scicos_pro_translation
  mdl_slider_gain_type
  [
    Value_parameter (Si "gain",
      Simulink_arg_expression (Ast_types.ident_from_string "gain"));
    Value_parameter (Si "overflow",
      Code_value (code_out_string "Nothing"));
    Value_parameter (Si "externalActivation",
      Code_value (Genr_utilities_scicos_pro.external_activation_mode_off));
  ]
  []
;;

(** Simulink block [Bias] *)

let mdl_bias_type = simulink_block_type Mdl_block_types.Bias in

add_scicos_pro_translation
  mdl_bias_type
  [
    Value_parameter (Si "b",
      Simulink_arg_expression (Ast_types.ident_from_string "Bias"));
    Value_parameter (Si "over",
      Simulink_arg_saturation);
  ]
  []
;;

(* TODO: those blocks are partially describe (missing source block)

(** Simulink blocks for checking assertions, blocks [Checks_*]. *)

List.iter
  (fun ty ->
   add_scicos_pro_equivalent
     (simulink_block_source_type (Ast_types.mdl_name_of_string ty)))
  [
    "Checks_DGap";
    "Checks_DMin";
    "Checks_DMax";
    "Checks_DRange";
    "Checks_Gradient";
    "Checks_Resolution";
    "Checks_SGap";
    "Checks_SMin";
    "Checks_SMax";
    "Checks_SRange";
  ]
;;
*)

(** Simulink block [PS-Simulink\nConverter] *)

let mdl_converter_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "PS-Simulink\nConverter")
    (Ast_types.mdl_path_of_string
       "nesl_utility/PS-Simulink\nConverter") in

add_scicos_pro_translation
  mdl_converter_type
  []
  []
;;

(** Simulink block [Simulink-PS\nConverter] *)

let mdl_ps_converter_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Simulink-PS\nConverter")
    (Ast_types.mdl_path_of_string
       "nesl_utility/Simulink-PS\nConverter") in

add_scicos_pro_translation
  mdl_ps_converter_type
  []
  []
;;

(** Simulink block [Clock] *)

add_scicos_pro_equivalent (simulink_block_type Mdl_block_types.Clock)
;;

(** Simulink block [Constant] *)

let mdl_constant_block_type =
  simulink_block_type Mdl_block_types.Constant in

add_scicos_pro_translation
  mdl_constant_block_type
  [
    Value_parameter (Instructions.Si "C",
      Simulink_arg_expression value_ident);
    Value_parameter (Si "typ",
      Code_value Genr_utilities_scicos_pro.inherit_mode_type);
    Value_parameter (Si "externalActivation",
      Code_value (code_out_int 0));
  ]
  []
;;

(** Simulink block [Derivative] *)

add_scicos_pro_equivalent (simulink_block_type Mdl_block_types.Derivative)
;;

(** Simulink block [DFlipFlop] *)

let mdl_dflipflop_block_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "DFlipFlop")
    (Ast_types.mdl_path_of_string
       "simulink_extras/Flip Flops/D Flip-Flop") in

add_scicos_pro_equivalent mdl_dflipflop_block_type
;;

(** Simulink block [Display] *)

let format_ident = Ast_types.ident_from_string "Format"
and floating_ident = Ast_types.ident_from_string "Floating"
and decimation_ident = Ast_types.ident_from_string "Decimation" in

let get_scicos_pro_block_type_and_block_args format floating decimation =

  let scicos_pro_block_type =
    scicos_pro_block_type_for
      (simulink_block_type Mdl_block_types.Display) in

  begin match
    Genr_utilities_scicos_pro.
    get_bool floating_ident floating with
  | false -> ()
  | true ->
    Say.warning (Printf.sprintf
      "Display block translation: unsupported floating ``on''")
  end;

  begin
    match
      Genr_utilities_scicos_pro.
        get_string decimation_ident decimation with
    | decimation ->
      Say.warning (Printf.sprintf
        "Display block translation: unsupported decimation ``%s''"
        decimation)
  end;

  let fmt = get_string format_ident format in

  let get_fmt = function
    | "short" -> "%5.5f"
    | "long" -> "%15.15f"
    | "short_e" -> "%5.5e"
    | "long_e" -> "%15.15e"
    | "bank" -> "%2.2f"
    |  s -> Say.warning (Printf.sprintf
      "Display block translation: format argument ``%s'' ignored" s);
    "%5.5f"  in

  let scicos_pro_block_args = [
    Value_parameter (Si "fmt",
      Code_value (code_out_string (get_fmt fmt)));
    Value_parameter (Si "dt",
      Code_value (code_out_int 1));
  ] in

  scicos_pro_block_type, scicos_pro_block_args in

add_scicos_pro_conditional_translation3
  (simulink_block_type Mdl_block_types.Display)
  get_scicos_pro_block_type_and_block_args
  (Simulink_arg_value format_ident,
   Simulink_arg_value floating_ident,
   Simulink_arg_value decimation_ident)
  []
;;

(** Simulink block [DLatch] *)

let mdl_dlatch_block_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "DLatch")
    (Ast_types.mdl_path_of_string "simulink_extras/Flip Flops/D Latch") in
add_scicos_pro_equivalent mdl_dlatch_block_type
;;

(** Simulink block [From] *)

add_scicos_pro_translation
  (simulink_block_type Mdl_block_types.From)
  [
    Value_parameter (Si "tag",
      Simulink_arg_value (Ast_types.ident_from_string "GotoTag"));
  ]
  []
;;

(** Simulink block [Goto] *)

let tag_visibility_ident = Ast_types.ident_from_string "TagVisibility"
and goto_tag_ident = Ast_types.ident_from_string "GotoTag" in

let get_visibility v =
  let s = get_string tag_visibility_ident v in
  let vty =
    match s with
    | "local" -> "Local"
    | "scoped" -> "Scoped"
    | "global" -> "Global"
    | s ->
      Say.warning (Printf.sprintf
        "Goto block translation: unexpected visibility ``%s''\
       \nChoosing local as visibility tag" s);
      "Local" in
  code_out_string vty in

add_scicos_pro_translation
  (simulink_block_type Mdl_block_types.Goto)
  [
    Value_parameter (Si "tag",
      Simulink_arg_value goto_tag_ident);
    Value_parameter (Si "tagvis",
      Simulink_arg_fun1 (
        get_visibility, Simulink_arg_value tag_visibility_ident));
  ]
  []
;;

(** Simulink block [GotoTagVisibility] *)

add_scicos_pro_translation
  (simulink_block_type Mdl_block_types.GotoTagVisibility)
  [
    Value_parameter (Si "tag",
      Simulink_arg_value (Ast_types.ident_from_string "GotoTag"));
  ]
  []
;;

(** Simulink block [Ground] *)

add_scicos_pro_equivalent (simulink_block_type Mdl_block_types.Ground)
;;

(** Simulink block [Inport] *)

let port_ident = Ast_types.ident_from_string "Port";;
let port_dimensions_ident = Ast_types.ident_from_string "PortDimensions";;

let get_port_dims d =
  let s = get_string port_dimensions_ident d in
  try
    let _nd = int_of_string s in
     code_out_expression (Printf.sprintf "[ %s, 1]" s)
  with
  | Failure _ (*"int_of_string"*) -> code_out_expression s in

add_scicos_pro_translation
  (simulink_block_type Mdl_block_types.Inport)
  [
    Value_parameter (Si "portNumber",
      Simulink_arg_expression port_ident);
    Value_parameter (Si "outsize",
      Simulink_arg_fun1 (
        get_port_dims,
        Simulink_arg_value port_dimensions_ident));
    Value_parameter (Si "outtyp",
      Simulink_arg_out_data_type);
  ]
  []
;;

(** Simulink block [JKFlipFlop] *)

let initial_condition_ident =
  Ast_types.ident_from_string "initial_condition"
;;

let mdl_jkflipflop_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "JKFlipFlop")
    (Ast_types.mdl_path_of_string
       "simulink_extras/Flip Flops/J-K\nFlip-Flop") in

add_scicos_pro_translation
  mdl_jkflipflop_type
  [
    Value_parameter (Si "init",
      Simulink_arg_expression initial_condition_ident);
    Value_parameter (Si "edge",
      Code_value (code_out_string "Falling"));
  ]
  []
;;

(** Simulink block [LTI Block] *)

(* TO BE FIXED
   sys may be an expression so the translation should create
   a super block and define sys in the context and extract
   parameters A, B, C and D from it in the context
*)
let mdl_lti_block_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "LTI Block")
    (Ast_types.mdl_path_of_string "cstblocks/LTI System") in

let sys_ident = Ast_types.ident_from_string "sys"
and ic_ident = Ast_types.ident_from_string "IC" in
let getA s = code_out_expression (get_string sys_ident s ^ ".A")
and getB s = code_out_expression (get_string sys_ident s ^ ".B")
and getC s = code_out_expression (get_string sys_ident s ^ ".C")
and getD s = code_out_expression (get_string sys_ident s ^ ".D") in

add_scicos_pro_translation
  mdl_lti_block_type
  [
    Value_parameter (Si "A",
      Simulink_arg_fun1 (getA,Simulink_arg_value sys_ident));
    Value_parameter (Si "B",
      Simulink_arg_fun1 (getB,Simulink_arg_value sys_ident));
    Value_parameter (Si "C",
      Simulink_arg_fun1 (getC,Simulink_arg_value sys_ident));
    Value_parameter (Si "D",
      Simulink_arg_fun1 (getD,Simulink_arg_value sys_ident));
    Value_parameter
        (Si "x0", Simulink_arg_expression ic_ident);
    ]
  []
;;

(** Simulink reference block [Manual switch] *)

let mdl_manual_switch_block_type,
    mdl_manual_switch_block_type_empty,
    mdl_manual_switch_block_type_unknown =
  let path =
    Ast_types.mdl_path_of_string
      "simulink/Signal\nRouting/Manual Switch" in
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Manual Switch")
    path,
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "")
    path,
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Unknown")
    path in

let sw_ident = Ast_types.ident_from_string "sw" in
(*
Unused in Scicos pro
and action_ident = Ast_types.ident_from_string "action" in
*)
let get_sw sw =
  let sw = get_int sw_ident sw in
  let sw =
    match sw with
    | 0 -> 2
    | 1 -> 1
    | sw ->
      let default = 1 in
      Say.warning (Printf.sprintf
        "ManualSwitch block translation: \
         unexpected connected input %d, using %d"
        sw default);
      default in
  code_out_int sw in

let block_parameters =
  [
    Value_parameter (Si "nin",
      Code_value (code_out_int 2));
    Value_parameter (Si "z0",
      Simulink_arg_fun1 (get_sw, Simulink_arg_value sw_ident));
(* FIXME:
   Field varsize is not there on all mdl models.
   Value_parameter
     (Si "varsize", Simulink_arg_bool "varsize");
   Should define a default when varsize is not there.
*)
  ] in

add_scicos_pro_translation
  mdl_manual_switch_block_type
  block_parameters
  [];

add_scicos_pro_translation
  mdl_manual_switch_block_type_empty
  block_parameters
  [];

add_scicos_pro_translation
  mdl_manual_switch_block_type_unknown
  block_parameters
  [];

let mdl_manual_switch_block_type3 =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Manual Switch")
    (Ast_types.mdl_path_of_string "simulink3/Nonlinear/Manual Switch") in

add_scicos_pro_translation
  mdl_manual_switch_block_type3
  [
    Value_parameter (Si "nin",
      Code_value (code_out_int 2));
    Value_parameter (Si "z0",
      Simulink_arg_fun1 (get_sw, Simulink_arg_value sw_ident));
  ]
  []
;;

(** Native Simulink block [ManualSwitch] *)

let current_setting_ident = Ast_types.ident_from_string "CurrentSetting" in

let get_cs cs =
  let sw = get_int current_setting_ident cs in
  let sw =
    match sw with
    | 0 -> 2
    | 1 -> 1
    | sw ->
      let default = 1 in
      Say.warning (Printf.sprintf
        "ManualSwitch block translation: \
         unexpected connected input %d, using %d"
        sw default);
      default in
  code_out_int sw in

let mdl_manual_switch_native_block_type =
  simulink_block_type Mdl_block_types.ManualSwitch in

add_scicos_pro_translation
  mdl_manual_switch_native_block_type
  [
    Value_parameter (Si "nin",
      Code_value (code_out_int 2));
    Value_parameter (Si "z0",
      Simulink_arg_fun1 (
        get_cs,
        Simulink_arg_value current_setting_ident));
(* FIXME:
   Fields varsize and sample are not yet treated.
 *)
  ]
  []
;;

(** Simulink block [Outport] *)

add_scicos_pro_translation
  (simulink_block_type Mdl_block_types.Outport)
  [
    Value_parameter (Si "portNumber",
      Simulink_arg_expression port_ident);
  ]
  []
;;

(** Simulink block [PID Controller] *)

let mdl_pid_controller_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "PID Controller")
    (Ast_types.mdl_path_of_string
       "simulink_extras/Additional\nLinear/PID Controller") in

let i_ident = Ast_types.ident_from_string "I"
and p_ident = Ast_types.ident_from_string "P"
and d_ident = Ast_types.ident_from_string "D" in

add_scicos_pro_translation
  mdl_pid_controller_type
  [
    Value_parameter (Instructions.Si "Kp",
      Simulink_arg_expression p_ident);
    Value_parameter (Instructions.Si "Ki",
      Simulink_arg_expression i_ident);
    Value_parameter (Instructions.Si "Kd",
      Simulink_arg_expression d_ident);
    Value_parameter (Si "N", Code_value (code_out_int 100));
  ]
  []
;;

(** Simulink block [PID Controller] *)

let mdl_pid_controller_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "PID Controller")
    (Ast_types.mdl_path_of_string
       "simulink_need_slupdate/PID Controller") in

let i_ident = Ast_types.ident_from_string "I"
and p_ident = Ast_types.ident_from_string "P"
and d_ident = Ast_types.ident_from_string "D" in

add_scicos_pro_translation
  mdl_pid_controller_type
  [
    Value_parameter (Instructions.Si "Kp",
      Simulink_arg_expression p_ident);
    Value_parameter (Instructions.Si "Ki",
      Simulink_arg_expression i_ident);
    Value_parameter (Instructions.Si "Kd",
      Simulink_arg_expression d_ident);
    Value_parameter (Si "N", Code_value (code_out_int 100));
  ]
  []
;;


(** Simulink block [Polyval] *)

let coefs_ident = Ast_types.ident_from_string "Coefs" in

add_scicos_pro_translation
  (simulink_block_type Mdl_block_types.Polyval)
  [
    Value_parameter (Instructions.Si "mode",
      Code_value (code_out_int 1));
    Value_parameter (Si "coeffs",
      Simulink_arg_expression coefs_ident);
  ]
  []
;;

(** Simulink block [Product] *)

let mul = Code_value (code_out_string "*")
and div = Code_value (code_out_string "/")
;;

let get_mul_op = function
  | '*' -> mul
  | '/' -> div
  | _ -> raise Not_found
;;

let add = Code_value (code_out_string "+")
and sub = Code_value (code_out_string "-")
;;

let get_add_op = function
  | '+' -> add
  | '-' -> sub
  | _ -> raise Not_found
;;

let get_ops (block_string, default_op, get_block_op) inputs =

  let ib = Scanf.Scanning.from_string inputs in

  if Scanf.Scanning.end_of_input ib then
    Say.fatal_error
      (Printf.sprintf "Block %s has empty Inputs argument"
         block_string) else

  let rec loop (n, ops as accu) =
    if Scanf.Scanning.end_of_input ib then (n, List.rev ops) else
    Scanf.bscanf ib "%c" (function
    | '|' -> loop accu
    | c ->
      try loop (succ n, get_block_op c :: ops) with
      | Not_found ->
        Say.warning (Printf.sprintf
          "%s block translation: unexpected character \
           %C was ignored in Inputs argument"
          block_string c);
        loop accu) in

  Scanf.bscanf ib "%0c" (function
  | '0' .. '9' ->
    Scanf.bscanf ib "%d"
      (fun n -> n, list_replicate n default_op)
  | _ -> loop (0, []))
;;

let rounding_method_ident = Ast_types.ident_from_string "RndMeth";;

let mdl_product_block_type =
  simulink_block_type Mdl_block_types.Product in

let get_scicos_pro_block_type_and_block_args inputs multiplication =

  let inputs = get_string inputs_ident inputs in
  let multiplication = get_string multiplication_ident multiplication in

  let number_of_ops, ops =
    get_ops ("Product", mul, get_mul_op) inputs in

  match multiplication with
  | "Element-wise(.*)" ->
    let scicos_pro_block_type =
      scicos_pro_block_type_for mdl_product_block_type in
    let get_rounding_method r =
      let r = get_string rounding_method_ident r in
      let rounding_method =
        match r with
        | "Ceiling" -> "Ceil"
        | "Floor" -> "Floor"
        | "Round" -> "Round"
        | "Zero" -> "Int"
        | c ->
          Say.warning (Printf.sprintf
            "Round method ``%s'' not supported in Product;\
           \n using method ``Int'' instead." c);
          "Int" in
      code_out_string rounding_method in

    scicos_pro_block_type,
    [
      Value_parameter (Si "nin",
        Code_value (code_out_int number_of_ops));
      Struct_parameter (Si "in_ports",
        ((Si "sgn", ops), []));
      Value_parameter (Si "overflow",
        Simulink_arg_saturation);
      Value_parameter (Si "rmethod",
        Simulink_arg_fun1 (
          get_rounding_method,
          Simulink_arg_value rounding_method_ident));
      Value_parameter (Si "externalActivation",
        Code_value (code_out_int 0));
    ]
  | "Matrix(*)" ->
    let scicos_pro_block_type =
      Sbt "system/MatrixOperations/MatrixMultiplication" in
    scicos_pro_block_type,
    [
      Value_parameter (Si "nin",
        Code_value (code_out_int number_of_ops));
      Struct_parameter (Si "in_ports",
        ((Si "sgn", ops), []));
      Value_parameter (Si "overflow",
        Simulink_arg_saturation);
      Value_parameter (Si "externalActivation",
        Code_value (code_out_int 0));
    ]
  | s ->
(* FIXME: fatal error or elementwise ??? *)
    Say.fatal_error (Printf.sprintf
      "Product block translation: unexpected rule ``%s''; using elementwise"
      s) in

add_scicos_pro_conditional_translation2
  mdl_product_block_type
  get_scicos_pro_block_type_and_block_args (
    Simulink_arg_value inputs_ident,
    Simulink_arg_value multiplication_ident
  )
  []
;;

(** Simulink block [Sum] *)

let mdl_sum_block_type =
  simulink_block_type Mdl_block_types.Sum in

let get_scicos_pro_block_type_and_block_args inputs =

  let inputs = get_string inputs_ident inputs in

  let number_of_ops, ops =
    get_ops ("Sum", add, get_add_op) inputs in

  let scicos_pro_block_type =
    scicos_pro_block_type_for mdl_sum_block_type in

  scicos_pro_block_type,
  [
    Value_parameter (Si "nin",
      Code_value (code_out_int number_of_ops));
    Struct_parameter (Si "in_ports",
      ((Si "sgn", ops), []));
    Value_parameter (Si "satur",
      Simulink_arg_saturation);
    Value_parameter (Si "externalActivation",
      Code_value (external_activation_mode_off));
  ] in

add_scicos_pro_conditional_translation1
  mdl_sum_block_type
  get_scicos_pro_block_type_and_block_args
  (Simulink_arg_value inputs_ident)
  []
;;

(** Simulink block [Data Type Duplicate] *)

let num_inputs_ident = Ast_types.ident_from_string "NumInputPorts";;

let list_of_minus_ones n =
  let rec loop accu = function
    | 0 -> accu
    | n ->
      let v = Code_value (code_out_int (-1)) in
      loop (v :: accu) (succ n) in
  loop [] (-n)
;;

let mdl_duplicate_block_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Data Type Duplicate")
    (Ast_types.mdl_path_of_string
        "simulink/Signal\nAttributes/Data Type\nDuplicate") in

let get_scicos_pro_block_type_and_block_args inputs =

  let inputs = get_int num_inputs_ident inputs in

  let ops = list_of_minus_ones inputs in

  let scicos_pro_block_type =
    scicos_pro_block_type_for mdl_duplicate_block_type in

  scicos_pro_block_type,
  [
    Value_parameter (Si "nin",
      Code_value (code_out_int inputs));
    Struct_parameter (Si "in_ports",
      ((Si "tp", ops), []));
  ] in

add_scicos_pro_conditional_translation1
  mdl_duplicate_block_type
  get_scicos_pro_block_type_and_block_args
  (Simulink_arg_value num_inputs_ident)
  []
;;

(** Simulink block [Ramp] *)

let mdl_ramp_block_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Ramp")
    (Ast_types.mdl_path_of_string "simulink/Sources/Ramp") in

add_scicos_pro_translation
  mdl_ramp_block_type
  [
    Value_parameter (Si "slope",
      Simulink_arg_expression (Ast_types.ident_from_string "slope"));
    Value_parameter (Si "startt",
      Simulink_arg_expression (Ast_types.ident_from_string "start"));
    Value_parameter (Si "initout",
      Simulink_arg_expression  (Ast_types.ident_from_string "X0"));
    Value_parameter (Si "zc",
      Code_value (code_out_int 0));
  ]
  []
;;

(** Simulink block [Cart2Polar] *)

let mdl_cart2polar_block_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Cart2Polar")
    (Ast_types.mdl_path_of_string
       "simulink_extras/Transformations/Cartesian to\nPolar") in

add_scicos_pro_translation
  mdl_cart2polar_block_type
  []
  []
;;

(** Simulink block [Memory] *)

let mdl_memory_block_type =
  simulink_block_type Mdl_block_types.Memory in

let x0_ident = Ast_types.ident_from_string "X0"
and initial_condition_ident =
  Ast_types.ident_from_string "InitialCondition" in

let get_scicos_pro_block_type_and_block_args x0 =
  match get_string x0_ident x0 with
  | "0" ->
    Sbt "system/Hybrid/LastInput",
    [
      Value_parameter (Si "externalActivation",
        Code_value (code_out_int 0));
    ]
  | _ ->
    Sbt "system/Dynamical/DiscreteDelay",
    [
      Value_parameter (Si "init_cond",
        Simulink_arg_expression_aliases
          [ x0_ident; initial_condition_ident; ]);
      Value_parameter (Si "typ",
        Code_value (inherit_mode_type));
      Value_parameter (Si "externalActivation",
        Code_value (code_out_int 0));
    ] in

add_scicos_pro_conditional_translation1
  mdl_memory_block_type
  get_scicos_pro_block_type_and_block_args
  (Simulink_arg_value_aliases [ x0_ident; initial_condition_ident; ])
  []
;;

(** Simulink block [RateLimiter] *)

let mdl_rate_limiter_blcok_type =
  simulink_block_type Mdl_block_types.RateLimiter in

add_scicos_pro_translation
  mdl_rate_limiter_blcok_type
  [
    Value_parameter (Si "upper",
      Simulink_arg_expression
        (Ast_types.ident_from_string "RisingSlewLimit"));
    Value_parameter (Si "lower",
      Simulink_arg_expression
        (Ast_types.ident_from_string "FallingSlewLimit"));
    Value_parameter (Si "init",
      Simulink_arg_expression
        (Ast_types.ident_from_string "InitialCondition"));
  ]
  []
;;

(** Simulink block [Relay] *)

let mdl_relay_block_type =
  simulink_block_type Mdl_block_types.Relay in

add_scicos_pro_translation
  mdl_relay_block_type
  [
    Value_parameter (Si "high_lim",
      Simulink_arg_expression
        (Ast_types.ident_from_string  "OnSwitchValue"));
    Value_parameter (Si "low_lim",
      Simulink_arg_expression
        (Ast_types.ident_from_string  "OffSwitchValue"));
    Value_parameter (Si "out_high",
      Simulink_arg_expression
        (Ast_types.ident_from_string  "OnOutputValue"));
    Value_parameter (Si "out_low",
      Simulink_arg_expression
        (Ast_types.ident_from_string  "OffOutputValue"));
    Value_parameter (Si "initout",
      Code_value (code_out_int 0));
    Value_parameter (Si "zc",
      Simulink_arg_zero_crossing);
  ]
  []
;;

(** Simulink block [Reshape] *)

let mdl_reshape_block_type =
  simulink_block_type Mdl_block_types.Reshape in

let output_dimensionality_ident =
  Ast_types.ident_from_string "OutputDimensionality"
and output_dimensions_ident =
  Ast_types.ident_from_string "OutputDimensions" in

let get_out_size dimensionality dims =
  (* tick "get_out_size"; *)
  let dimensionality =
    get_string output_dimensionality_ident dimensionality
  and dims =
    get_string output_dimensions_ident dims in
  match dimensionality with
  | "1-D array"
  | "Column vector (2-D)" -> code_out_expression "[-1, 1]"
  | "Column vector" -> code_out_expression "[-1, 1]"
  | "Row vector (2-D)" -> code_out_expression "[1, -1]"
  | "Row vector" -> code_out_expression "[1, -1]"
  | "Customize (N-D)"
  | "Customize" -> code_out_expression dims
  | c ->
    Say.warning (Printf.sprintf "Reshape option ``%s'' not supported" c);
    code_out_expression dims in

let scicos_pro_block_args =
  [
    Value_parameter (Si "outsize",
      Simulink_arg_fun2 (
        get_out_size,
        Simulink_arg_value output_dimensionality_ident,
        Simulink_arg_value output_dimensions_ident));
    Value_parameter (Si "externalActivation",
      Code_value external_activation_mode_off);
  ] in

(* Simulink builtin block Reshape *)
add_scicos_pro_translation
  mdl_reshape_block_type
  scicos_pro_block_args
  [];

(* Simulink block reference Reshape *)

let mdl_reshape_block_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Reshape")
    (Ast_types.mdl_path_of_string "simulink/Math\nOperations/Reshape") in

add_scicos_pro_translation
  mdl_reshape_block_type
  scicos_pro_block_args
  [];

(* Simulink block reference Reshape *)

let mdl_reshape_block_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Reshape")
    (Ast_types.mdl_path_of_string "simulink3/Signals\n& Systems/Reshape") in

add_scicos_pro_translation
  mdl_reshape_block_type
  scicos_pro_block_args
  []
;;

(** Fake Simulink block [Split] *)

add_scicos_pro_equivalent Mdl_fake_block_types.split_block_type
;;

(** Fake Simulink block [Split_connection] *)

add_scicos_pro_equivalent Mdl_fake_block_types.split_connection_block_type
;;

(** Fake Simulink block [AtomicPort] *)

add_scicos_pro_translation
  Mdl_fake_block_types.atomic_port_block_type
  [
    Value_parameter (Si "portNumber",
      Code_value (code_out_int 1));
  ]
  []
;;

(** Simulink block [SRFlipFlop] *)

let mdl_srflipflop_block_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "SRFlipFlop")
    (Ast_types.mdl_path_of_string
       "simulink_extras/Flip Flops/S-R\nFlip-Flop") in

add_scicos_pro_translation
  mdl_srflipflop_block_type
  [
    Value_parameter (Si "init",
      Simulink_arg_expression initial_condition_ident);
  ]
  []
;;

(** Simulink block [StateSpace] *)

let mdl_state_space_block_type =
  simulink_block_type Mdl_block_types.StateSpace in

let a_ident = Ast_types.ident_from_string "A"
and b_ident = Ast_types.ident_from_string "B"
and c_ident = Ast_types.ident_from_string "C"
and d_ident = Ast_types.ident_from_string "D"
and x0_ident = Ast_types.ident_from_string "X0"
and initial_condition_ident = (* XXX *)
  Ast_types.ident_from_string "InitialCondition" in

let is_empty s =
  let ib = Scanf.Scanning.from_string s in
  Scanf.Scanning.end_of_input ib ||
  let k _ _ = false in
  Scanf.kscanf ib k " [ ] " true in

let get_scicos_pro_block_type_and_block_args a =
  match get_string a_ident a with
  | a when is_empty a ->
    Sbt "system/MatrixOperations/MatrixGain",
    [
      Value_parameter (Si "gain",
        Simulink_arg_expression d_ident);
      Value_parameter (Si "overflow",
        Simulink_arg_saturation_set Genr_types.Saturation_off);
      Value_parameter (Si "mulmethod",
        Code_value (code_out_string "Gain*u1"));
      Value_parameter (Si "externalActivation",
        Code_value (code_out_int 0));
    ]
  | _ ->
    Sbt "system/Dynamical/ContStateSpace",
    [
      Value_parameter
        (Si "A", Simulink_arg_expression a_ident);
      Value_parameter
        (Si "B", Simulink_arg_expression b_ident);
      Value_parameter
        (Si "C", Simulink_arg_expression c_ident);
      Value_parameter
        (Si "D", Simulink_arg_expression d_ident);
      Value_parameter
        (Si "x0",
         Simulink_arg_expression_aliases
           [ x0_ident; initial_condition_ident; ] );
    ] in

add_scicos_pro_conditional_translation1
  mdl_state_space_block_type
  get_scicos_pro_block_type_and_block_args
  (Simulink_arg_value a_ident)
  []
;;

(** Simulink block [Step] *)

add_scicos_pro_translation
  (simulink_block_type Mdl_block_types.Step)
  [
    Value_parameter (Si "steptime",
      Simulink_arg_expression (Ast_types.ident_from_string "Time"));
    Value_parameter (Si "iniv",
      Simulink_arg_expression (Ast_types.ident_from_string "Before"));
    Value_parameter (Si "finv",
      Simulink_arg_expression (Ast_types.ident_from_string "After"));
  ]
  []
;;

(** Simulink block [Terminator] *)

add_scicos_pro_translation
  (simulink_block_type Mdl_block_types.Terminator)
  [
    Value_parameter (Si "dim",
      Code_value (code_out_expression "[-1,-2]"));
    Value_parameter (Si "intyp",
      Code_value inherit_mode_type);
    Value_parameter (Si "externalActivation",
      Code_value (code_out_int 0));
  ]
  []
;;

(** Simulink block [TransferFcn] *)

let mdl_transfer_fcn_block_type =
  simulink_block_type Mdl_block_types.TransferFcn in

let numerator_ident = Ast_types.ident_from_string "Numerator"
and denominator_ident = Ast_types.ident_from_string "Denominator" in

let get_component n =
  let n = get_string numerator_ident n in
  code_out_expression (Printf.sprintf "%s" n) in

add_scicos_pro_translation
  mdl_transfer_fcn_block_type
  [
    Value_parameter (Si "num",
      Simulink_arg_fun1 (
        get_component,
        Simulink_arg_value numerator_ident));
    Value_parameter (Si "den",
      Simulink_arg_fun1 (
        get_component,
        Simulink_arg_value denominator_ident));
  ]
  []
;;

(** Simulink block [Transfer Function with Initial Outputs] *)

let mdl_transfer_function_block_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Transfer Function with Initial Outputs")
    (Ast_types.mdl_path_of_string
       "simulink_extras/Additional\nLinear/Transfer Fcn\n\
        (with initial outputs)") in

let numerator_ident = Ast_types.ident_from_string "N"
and denominator_ident = Ast_types.ident_from_string "D"
and y0_ident = Ast_types.ident_from_string "Y0" in

let get_component n =
  let n = get_string numerator_ident n in
  code_out_expression (Printf.sprintf "%s" n) in

add_scicos_pro_translation
  mdl_transfer_function_block_type
  [
    Value_parameter (Si "num",
      Simulink_arg_fun1 (
        get_component,
        Simulink_arg_value numerator_ident));
    Value_parameter (Si "den",
      Simulink_arg_fun1 (
        get_component,
        Simulink_arg_value denominator_ident));
    Value_parameter (Si "y0",
      Simulink_arg_fun1 (
        get_component,
        Simulink_arg_value y0_ident));  ]
  []
;;

(** Simulink block [Transfer Function with Initial States] *)

let mdl_transfer_function_block_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Transfer Function with Initial States")
    (Ast_types.mdl_path_of_string
       "simulink_extras/Additional\nLinear/Transfer Fcn\
      \n(with initial states)") in

let numerator_ident = Ast_types.ident_from_string "N"
and denominator_ident = Ast_types.ident_from_string "D"
and x0_ident = Ast_types.ident_from_string "X0" in

let get_component n =
  let n = get_string numerator_ident n in
  code_out_expression (Printf.sprintf "%s" n) in

add_scicos_pro_translation
  mdl_transfer_function_block_type
  [
    Value_parameter (Si "num",
      Simulink_arg_fun1 (
        get_component,
        Simulink_arg_value numerator_ident));
    Value_parameter (Si "den",
      Simulink_arg_fun1 (
        get_component,
        Simulink_arg_value denominator_ident));
    Value_parameter (Si "x0",
      Simulink_arg_fun1 (
        get_component,
        Simulink_arg_value x0_ident));  ]
  []
;;

(** Simulink block [TransportDelay] *)

let initial_input_ident = Ast_types.ident_from_string "InitialInput"
and initial_output_ident = Ast_types.ident_from_string "InitialOutput"
;;

let mdl_transport_delay_block_type =
  simulink_block_type Mdl_block_types.TransportDelay in

let get_init v =
  let txt = get_string initial_input_ident v in
  code_out_expression txt in
add_scicos_pro_translation
  mdl_transport_delay_block_type
  [
    Value_parameter (Si "T",
      Simulink_arg_expression (Ast_types.ident_from_string "DelayTime"));
    Value_parameter (Si "init",
      Simulink_arg_fun1 (
        get_init,
        Simulink_arg_value_aliases [
          initial_output_ident; initial_input_ident;
        ]));
    Value_parameter (Si "N",
      Simulink_arg_expression (Ast_types.ident_from_string "BufferSize"));
  ]
  []
;;

(** Simulink block [Transpose] *)

let mdl_transpose_block_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Transpose")
    (Ast_types.mdl_path_of_string "dspmtrx3/Transpose") in

add_scicos_pro_translation
  mdl_transpose_block_type
  [
    Value_parameter (Si "rule",
      Code_value (code_out_string "Transpose"));
    Value_parameter (Si "externalActivation",
      Code_value (code_out_int 0));
  ]
  []
;;

(** Simulink block [TriggerPort] *)

add_scicos_pro_translation
  (simulink_block_type Mdl_block_types.TriggerPort)
  [
    Value_parameter (Si "portNumber",
      Code_value (code_out_int 1));
  ]
  []
;;

(** Simulink block [VariableTransportDelay] *)

let mdl_variable_transport_delay_block_type =
  simulink_block_type Mdl_block_types.VariableTransportDelay in

let variable_delay_type_ident =
  Ast_types.ident_from_string "VariableDelayType" in

let get_delay_type d =
  let d = get_string variable_delay_type_ident d in
  let ty =
    match d with
    | "Variable time delay" -> "Time"
    | "Variable transport delay" -> "Transport"
    | d ->
      Say.fatal_error (Printf.sprintf
        "Variable delay block: Unsupported delay type ``%s''"
        d) in
  code_out_string ty in

add_scicos_pro_translation
  mdl_variable_transport_delay_block_type
  [
    Value_parameter (Si "T",
      Simulink_arg_expression
        (Ast_types.ident_from_string "MaximumDelay"));
    Value_parameter (Si "init",
      Simulink_arg_expression initial_output_ident);
    Value_parameter (Si "N",
      Simulink_arg_expression
        (Ast_types.ident_from_string "MaximumPoints"));
    Value_parameter (Si "dtype",
      Simulink_arg_fun1 (
        get_delay_type,
        Simulink_arg_value variable_delay_type_ident));
  ]
  []
;;

(** Simulink block [ZeroPole] *)

(*
    num = gain * poly (zeros, z, root)
    poly (num, s, coeff).
*)

add_scicos_pro_translation
  (simulink_block_type Mdl_block_types.ZeroPole)
  [
    Value_parameter (Si "zer",
      Simulink_arg_expression (Ast_types.ident_from_string "Zeros"));
    Value_parameter (Si "pol",
      Simulink_arg_expression (Ast_types.ident_from_string "Poles"));
    Value_parameter (Si "k_gain",
      Simulink_arg_expression gain_ident);
  ]
  []
;;

(** Aliases for Simscape Reference blocks supported by simport. *)

(** Simscape block [Electrical Reference] *)

let mdl_electrical_reference_block_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Electrical Reference")
    (Ast_types.mdl_path_of_string
       "fl_lib/Electrical/Electrical Elements/Electrical Reference") in

add_scicos_pro_equivalent
  mdl_electrical_reference_block_type
;;


(** Simscape block [Resistor] *)

let mdl_resistor_block_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Resistor")
    (Ast_types.mdl_path_of_string
       "fl_lib/Electrical/Electrical Elements/Resistor") in

add_scicos_pro_translation
  mdl_resistor_block_type
  [
    Value_parameter (Si "R",
      Simulink_arg_expression (Ast_types.ident_from_string "R"));
  ]
  []
;;

(** Simscape block [Mechanical Rotational Damper] *)

let mdl_rotational_damper_block_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Rotational Damper")
    (Ast_types.mdl_path_of_string
       "fl_lib/Mechanical/Rotational Elements/Rotational Damper") in

add_scicos_pro_translation
  mdl_rotational_damper_block_type
  [
    Value_parameter (Si "d",
      Simulink_arg_expression (Ast_types.ident_from_string "D"));
  ]
  []
;;

(** Simscape block [Solver\nConfiguration]  *)

let mdl_solver_configuration_block_type =
  simulink_block_reference_type
   (Ast_types.mdl_name_of_string "Solver\nConfiguration")
   (Ast_types.mdl_path_of_string
      "nesl_utility/Solver\nConfiguration") in

add_scicos_pro_equivalent mdl_solver_configuration_block_type
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
