(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_bus_creator.ml,v 1.1 2019-05-10 14:04:56 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
;;

let mdl_block_type = simulink_block_type Mdl_block_types.BusCreator;;

let propagated_signals_ident = Ast_types.ident_from_string "PropagatedSignals"
and use_bus_object_ident = Ast_types.ident_from_string "UseBusObject"
;;

let inputs_ident = Ast_types.ident_from_string "Inputs"
;;

let genr block_id genr_info =

  let ports_desc = genr_info.Genr_block_types.block_ports_description in
  let nin =
    Mdl_ast_to_deep_ast.get_regular_port_nin ports_desc in

  let inputs =
    Genr_utilities_scicos_pro.get_ident_string genr_info inputs_ident in

  let get_propagated_signals bindings =
    try
      let value =
        let rec loop = function
          | [] -> raise Not_found
          | {
              Binding.
              bound_ident = ident;
              bound_value = value;
            } :: _ when ident = propagated_signals_ident -> value
          | _ :: bindings -> loop bindings in
        loop bindings in
      let signals =
        Mdl_deep_basics.string_of_deep_value
          "PropagatedSignals is not a string" value in
      let signals = Lib_strings.split_string_at_comma signals in
      let signals = List.map String.trim signals in
      Some signals
    with
    | Not_found -> None in

  let bus_name, propagated_signals =
    match genr_info.Genr_block_types.block_ports with
    | [] ->
      "", None
    | [ {
          Mdl_basics_deep_ast.
          deep_block_port_number = _;
          deep_block_port_name = None;
          deep_block_port_bindings = bindings;
        }; ] ->
      "", get_propagated_signals bindings
    | [ {
          Mdl_basics_deep_ast.
          deep_block_port_number = _;
          deep_block_port_name = Some port_name;
          deep_block_port_bindings = bindings;
        }; ] ->
      Ast_types.
      string_of_mdl_name port_name, get_propagated_signals bindings
    | _ ->
      Say.warning (Printf.sprintf
        "Multiple output ports for BusCreator are not supported");
      "", None in

  let get_signal_names_from_lines () =
    let rec loop accu port_number =
      if port_number = Unatural._0_ then accu else
        let signal_name =
          try
            let signal_name_as_string =
              Ast_types.string_of_mdl_name
                (Hashtbl.find
                   genr_info.Genr_block_types.block_inport_names
                   (Genr_utilities_scicos_pro.
                    mdl_port_number_of_natural port_number)) in
            signal_name_as_string
          with
          | Not_found -> "" in
        loop (signal_name :: accu) (Unatural.pred port_number) in
    loop [] nin in

  let is_quoted name =
    let len = String.length name in
    len >= 2 && name.[0] = '\'' && name.[pred len] = '\'' in

  let get_signal_names_from_inputs () =
    try
      ignore (int_of_string inputs);
      None
    with
    | Failure _ (*"int_of_string"*) ->
      let names = Lib_strings.split_string_at_comma inputs in
      match names with
      | [ name; ] when not (is_quoted name) -> None
      | quoted_names ->
        let unquote_name name =
          let len = String.length name in
          if not (is_quoted name) then
            fatal_error (Printf.sprintf
              "signal name %S from Inputs parameter is not quoted" name)
          else
            String.sub name 1 (len - 2) in
        Some (List.map unquote_name quoted_names) in

  let use_bus_object =
    if Genr_utilities_scicos_pro.
       exists_argument genr_info use_bus_object_ident then
      Genr_utilities_scicos_pro.
      get_ident_bool genr_info use_bus_object_ident
    else
      false in

  let insignals =
    match get_signal_names_from_inputs () with
    | Some insignals -> insignals
    | None ->
      if use_bus_object then
        get_signal_names_from_lines ()
      else
        match propagated_signals with
        | Some insignals -> insignals
        | None -> get_signal_names_from_lines () in

  let insignals =
    List.map
      (fun insignal -> Code_value (code_out_string insignal)) insignals in

  let scicos_block_parameters = [
    Value_parameter
      (Si "busname", Code_value (code_out_string bus_name));
    Value_parameter
      (Si "nin", Code_value (code_out_int (Unatural.to_int nin)));
    Struct_parameter
      (Si "reg_ports", ((Si "sign", insignals), []));
  ] in

  let scicos_basic_block_instructions blk_id =
    genr_instructions_for
      mdl_block_type scicos_block_parameters
      [] blk_id genr_info in

  Concat_list.list (scicos_basic_block_instructions block_id)
;;

add_translator mdl_block_type genr;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
