(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_bus_selector.ml,v 1.1 2019-05-10 14:04:56 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
;;

let mdl_block_type = simulink_block_type Mdl_block_types.BusSelector;;

let output_signals_ident =
  Ast_types.ident_from_string "OutputSignals"
and output_as_bus_ident =
  Ast_types.ident_from_string "OutputAsBus"
;;

let genr block_id genr_info =

  let output_as_bus =
    Genr_utilities_scicos_pro.get_ident_bool genr_info output_as_bus_ident
  and output_signals =
    List.map
      (fun output_signal -> Code_value (code_out_string output_signal))
      (Lib_strings.split_string_at_comma
         (Genr_utilities_scicos_pro.
          get_ident_string genr_info output_signals_ident)) in

  let nout = Unatural.of_int (List.length output_signals)
  and busname =
    let input_signal_name =
      try
        Ast_types.string_of_mdl_name
          (Hashtbl.find
             genr_info.Genr_block_types.block_inport_names
             port_1_)
      with
      | Not_found -> "" in
    Code_value (code_out_string input_signal_name) in

  let bus_extractor_block_instructions blk_id =
    let scicos_block_parameters = [
      Value_parameter (Si "busname", busname);
      Value_parameter
        (Si "nout", Code_value (code_out_int (Unatural.to_int nout)));
      Struct_parameter
        (Si "outports", ((Si "Signals", output_signals), []));
    ] in

    genr_instructions_for
      mdl_block_type scicos_block_parameters
      []
      blk_id genr_info in

  match output_as_bus with
  | false -> Concat_list.list (bus_extractor_block_instructions block_id)
  | true ->
    let bus_name =
      match genr_info.Genr_block_types.block_ports with
      | [] -> ""
      | [ {
            Mdl_basics_deep_ast.
            deep_block_port_number = _;
            deep_block_port_name = None;
            deep_block_port_bindings = _;
          }; ] -> ""
      | [ {
            Mdl_basics_deep_ast.
            deep_block_port_number = _;
            deep_block_port_name = Some port_name;
            deep_block_port_bindings = _;
          }; ] -> Ast_types.string_of_mdl_name port_name
      | _ ->
        Say.warning (Printf.sprintf
          "Multiple output ports for BusCreator not supported");
        "" in
    let in_block_port_name = "in" in
    let bus_extractor_name = "busextractor" in
    let bus_creator_name = "buscreator" in
    let out_block_port_name = "out" in

    let in_port =
      Genr_utilities_scicos_pro.in_port_in_scsm0
        in_block_port_name port_1_ (Unatural._0_, Unatural._0_) in

    let bus_extractor =
      let nports = (Unatural._1_, nout) in
      Genr_utilities_scicos_pro.add_block_body_in_scsm0
        bus_extractor_name
        (bus_extractor_block_instructions blk0)
        nports
        (Unatural._1_, Unatural._0_) in

    let bus_creator =
      let nports = (nout, Unatural._1_) in
      let scicos_block_type = Sbt "system/BusOperations/BusCreator" in
      Genr_utilities_scicos_pro.add_block_body_in_scsm0
        bus_creator_name
        (genr_instructions_for_modeler_block_type
           scicos_block_type
           [
             Value_parameter
               (Si "busname", Code_value (code_out_string bus_name));
             Value_parameter
               (Si "nin", Code_value (code_out_int (Unatural.to_int nout)));
             Struct_parameter
               (Si "reg_ports", ((Si "sign", output_signals), []));
           ] [] Genr_utils.blk0 genr_info)
        nports
        (Unatural._2_, Unatural._0_) in

    let out_port =
      Genr_utilities_scicos_pro.out_port_in_scsm0
        out_block_port_name port_1_ (Unatural._3_, Unatural._0_) in

    let ports_links = [
      Genr_utilities_scicos_pro.
      add_directed_link_in_scsm0
        in_block_port_name port_1_ bus_extractor_name port_1_ [];
      Genr_utilities_scicos_pro.
      add_directed_link_in_scsm0
        bus_creator_name port_1_ out_block_port_name port_1_ [];
    ] in

    let internal_links =
      let rec loop links port_nat =
        if port_nat = Unatural._0_ then links else
        let port_number = mdl_port_number_of_natural port_nat in
        let link =
          Genr_utilities_scicos_pro.
          add_directed_link_in_scsm0
            bus_extractor_name port_number
            bus_creator_name port_number [] in
          loop (link :: links) (Unatural.pred port_nat) in
      loop [] nout in

    let super_body = [
      in_port;
      bus_extractor;
      bus_creator;
      out_port;
      ports_links;
      internal_links;
    ] in

    super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
