(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_complex_to_coordinates.ml,v 1.1 2019-05-10 14:04:56 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
open Unatural
;;

(* Computing representation of complex numbers *)

(* Simulink bloc types ComplexToMagnitudeAngle and ComplexToRealImag share
   a lot of code.
   Hence the sharing of the generator.
*)

let magnitude_angle_options = [| "Magnitude and angle";
                                 "Magnitude"; "Angle" |];;
let real_imag_options = [| "Real and imag";
                           "Real"; "Imag" |];;

let mdl_block_type_complex_to_magnitude_angle =
  Mdl_basics_explicit_ast.Explicit_block_regular_type
    Mdl_block_types.ComplexToMagnitudeAngle
;;

let complex_to_magnitude_angle_formula = code_out_expression "CtoP";;

let complex_to_magnitude_angle_options = magnitude_angle_options;;

let mdl_block_type_complex_to_real_imag =
  Mdl_basics_explicit_ast.Explicit_block_regular_type
    Mdl_block_types.ComplexToRealImag
;;

let complex_to_real_imag_formula = code_out_expression "CtoR";;

let complex_to_real_imag_options = real_imag_options;;

(* The generator for all the above blocks. *)

let output_ident = Ast_types.ident_from_string "Output";;

let genr mdl_block_type conversion_formula output_options block_id genr_info =

  let get_ident_string =
    Genr_utilities_scicos_pro.get_ident_string genr_info in

  let output = get_ident_string output_ident
  and sample_time = Genr_utilities_scicos_pro.get_sample_time genr_info in

  let scicos_block_parameters activation = [
    Value_parameter (Si "decomptyp", Code_value conversion_formula);
    Value_parameter (Si "externalActivation", Code_value activation);
  ] in

  let option =
    if output_options.(0) = output then 0
    else if output_options.(1) = output then 1
    else if output_options.(2) = output then 2
    else (
      Say.warning (Printf.sprintf
        "Complex to %s Simulink block translation: \
         Option %s not supported." output_options.(0) output );
      0) in

  let scicos_basic_block_instructions blk_id activation =
    genr_instructions_for
      mdl_block_type
      (scicos_block_parameters activation)
      [] blk_id genr_info in

  match sample_time, option with
  | "-1", 0 ->
    Concat_list.list
      (scicos_basic_block_instructions block_id
         Genr_utilities_scicos_pro.external_activation_mode_off)

  | "-1", option ->
    let in_port_name = "in"
    and out_port_name1 = "out1"
    and ctri_name = "ctri" in

    let in_port =
      Genr_utilities_scicos_pro.
      in_port_in_scsm0
        in_port_name port_1_
          (_0_, _0_) in
    let out_port1 =
      Genr_utilities_scicos_pro.
      out_port_in_scsm0
        out_port_name1 port_1_
        (_2_, _0_) in
    let ctri =
      let nports = (_1_, _2_) in
      Genr_utilities_scicos_pro.
      add_block_body_in_scsm0
        ctri_name
        (scicos_basic_block_instructions blk0
         Genr_utilities_scicos_pro.external_activation_mode_off)
        nports
        (_1_, _0_) in

    let links = [
      Genr_utilities_scicos_pro.
      add_directed_link_in_scsm0
        in_port_name port_1_ ctri_name port_1_ [];
      add_directed_link_in_scsm0
        ctri_name
        (Ast_types.mdl_port_number_of_int option)
        out_port_name1 port_1_ [];
    ] in

    let super_body = [
      in_port;
      out_port1;
      ctri;
      links;
    ] in

    super_block_in_scsm0 block_id super_body

  | sample_time, 0 ->

    let sample_time = code_out_expression sample_time in

    let in_port_name = "in"
    and out_port_name1 = "out1"
    and out_port_name2 = "out2"
    and ctri_name = "ctri"
    and sample_clock_name = "sample_clock" in

    let in_port =
      Genr_utilities_scicos_pro.
      in_port_in_scsm0
        in_port_name port_1_
          (_0_, _0_) in
    let out_port1 =
      Genr_utilities_scicos_pro.
      out_port_in_scsm0
        out_port_name1 port_1_
        (_2_, _0_) in
    let out_port2 =
      out_port_in_scsm0
        out_port_name2 port_2_
        (_2_, _1_) in
    let ctri =
      let nports = (_1_, _2_) in
      Genr_utilities_scicos_pro.
      add_block_body_in_scsm0
        ctri_name
        (scicos_basic_block_instructions blk0
         Genr_utilities_scicos_pro.external_activation_mode_on)
        nports
        (_1_, _0_) in
    let sample_clock =
      Genr_utilities_scicos_pro.
      sample_clock_in_scsm0
        sample_clock_name sample_time
        (_1_, _1_) in

    let links = [
      Genr_utilities_scicos_pro.
      add_directed_link_in_scsm0
        in_port_name port_1_ ctri_name port_1_ [];
      add_directed_link_in_scsm0
        ctri_name port_1_ out_port_name1 port_1_ [];
      add_directed_link_in_scsm0
        ctri_name port_2_ out_port_name2 port_1_ [];
      Genr_utilities_scicos_pro.
      add_event_link_in_scsm0
        sample_clock_name port_1_ ctri_name port_1_ [];
    ] in

    let super_body = [
      in_port;
      out_port1;
      out_port2;
      ctri;
      sample_clock;
      links;
    ] in

    super_block_in_scsm0 block_id super_body

  | sample_time, option ->

    let sample_time = code_out_expression sample_time in

    let in_port_name = "in"
    and out_port_name1 = "out1"
    and ctri_name = "ctri"
    and sample_clock_name = "sample_clock" in

    let in_port =
      Genr_utilities_scicos_pro.
      in_port_in_scsm0
        in_port_name port_1_
          (_0_, _0_) in
    let out_port1 =
      Genr_utilities_scicos_pro.
      out_port_in_scsm0
        out_port_name1 port_1_
        (_2_, _0_) in
    let ctri =
      let nports = (_1_, _2_) in
      Genr_utilities_scicos_pro.
      add_block_body_in_scsm0
        ctri_name
        (scicos_basic_block_instructions blk0
         Genr_utilities_scicos_pro.external_activation_mode_on)
        nports
        (_1_, _0_) in
    let sample_clock =
      Genr_utilities_scicos_pro.
      sample_clock_in_scsm0
        sample_clock_name sample_time
        (_1_, _1_) in

    let links = [
      Genr_utilities_scicos_pro.
      add_directed_link_in_scsm0
        in_port_name port_1_ ctri_name port_1_ [];
      add_directed_link_in_scsm0
        ctri_name
        (Ast_types.mdl_port_number_of_int option)
        out_port_name1 port_1_ [];
      Genr_utilities_scicos_pro.
      add_event_link_in_scsm0
        sample_clock_name port_1_ ctri_name port_1_ [];
    ] in

    let super_body = [
      in_port;
      out_port1;
      ctri;
      sample_clock;
      links;
    ] in

    super_block_in_scsm0 block_id super_body
;;

let genr_magnitude_angle =
  genr
    mdl_block_type_complex_to_magnitude_angle
    complex_to_magnitude_angle_formula
    complex_to_magnitude_angle_options
;;

add_translator
  mdl_block_type_complex_to_magnitude_angle genr_magnitude_angle
;;

let genr_real_imag =
  genr
    mdl_block_type_complex_to_real_imag
    complex_to_real_imag_formula
    complex_to_real_imag_options
;;

add_translator mdl_block_type_complex_to_real_imag genr_real_imag
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
