(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_concatenate.ml,v 1.1 2019-05-10 14:04:56 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
;;

let fatal_error s =
  Say.fatal_error (Printf.sprintf "Genr_concatenate: %s" s)
;;

let mdl_block_type =
  Mdl_basics_explicit_ast.
  Explicit_block_regular_type Mdl_block_types.Concatenate
;;

let nin_ident = Ast_types.ident_from_string "NumInputs"
and mode_ident = Ast_types.ident_from_string "Mode"
and concatenate_dimension_ident =
  Ast_types.ident_from_string "ConcatenateDimension"
;;

let si_external_activation = Si "externalActivation";;
let no_activation =
  Value_parameter
    (si_external_activation, Code_value external_activation_mode_off)
;;

let get_nin genr_info nin_ident =
  let nin = get_ident_string genr_info nin_ident in
  try
    int_of_string nin
  with
  | Failure _ (*"int_of_string"*) ->
    Say.warning (Printf.sprintf
      "Concatenate block translation: number of inputs not specifed as \
      an integer.");
    let ports_desc = genr_info.Genr_block_types.block_ports_description in
    Unatural.to_int
      (Mdl_ast_to_deep_ast.get_regular_port_nin ports_desc)
;;

let genr block_id genr_info =

  let nin = get_nin genr_info nin_ident
  and mode = get_ident_string genr_info mode_ident
  and work = code_out_expression "getcurrentenv" in

  match mode with
  | "Multidimensional array"
    when
      let dim = get_ident_int genr_info concatenate_dimension_ident in
      dim <> 1 && dim <> 2 ->
    Genr_utilities_scicos_pro.warning (Printf.sprintf
      "Higher than 2 dimensional matrices not supported");
    Genr_scicos_pro_not_yet_translated.genr block_id genr_info
  | _ ->

    let separator =
      match mode with
      | "Multidimensional array" ->
        let concatenate_dimension =
          get_ident_int genr_info concatenate_dimension_ident in
        begin match concatenate_dimension with
        | 1 -> ";"
        | 2 -> ","
        | _ -> assert false
        end
      | "Vertical matrix concatenation"
      | "Vector"
      | "Vector concatenation" -> ";"
      | "Horizontal matrix concatenation" -> ","
      | mode ->
        fatal_error (Printf.sprintf
          "Unsupported concatenation mode %s" mode) in

    let rec loop i expr =
      if i == nin then Printf.sprintf "%su%d" expr i else
      loop (i + 1) (Printf.sprintf "%su%d%s" expr i separator) in

    let expr = loop 1 "[" ^ "]" in
    let exx = code_out_string expr in

    let scicos_block_parameters = [
      Value_parameter (Si "nin", Code_value (code_out_int nin));
      Value_parameter (Si "work", Code_value work);
      Value_parameter (Si "exx", Code_value exx);
      no_activation;
    ] in

    let scicos_basic_block_instructions blk_id =
      genr_instructions_for
        mdl_block_type scicos_block_parameters
        [] blk_id genr_info in

    Concat_list.list (scicos_basic_block_instructions block_id)
;;

add_translator mdl_block_type genr;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
