(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_conversion.ml,v 1.1 2019-05-10 14:04:56 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
open Unatural
;;

(* Simulink bloc types FahrenheitToCelsius/CelsiusToFahrenheit
   and DegreesToRadians/RadiansToDegrees are all translated to the same
   UnitConversion Scicos block.
   Hence the sharing of their generators.
*)

(* Defining the Simulink conversions we can handle. *)
type conversion_block = {
  block_source_type_name : string;
  block_source_block : string;
  from_name : string;
  to_name : string;
  category_name : string;
}
;;

let simulink_block_of_conversion conv =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string conv.block_source_type_name)
    (Ast_types.mdl_path_of_string conv.block_source_block)
;;

(* Temperature conversions: celsius <-> fahrenheit *)

let celsius_to_fahrenheit_conversion = {
  block_source_type_name = "CelsiusToFahrenheit";
  block_source_block =
    "simulink_extras/Transformations/Celsius to\nFahrenheit";
  from_name = "celsius";
  to_name = "fahrenheit";
  category_name = "Temperature";
}
;;

let fahrenheit_to_celsius_conversion = {
  block_source_type_name = "FahrenheitToCelsius";
  block_source_block =
    "simulink_extras/Transformations/Fahrenheit\nto Celsius";
  from_name = "fahrenheit";
  to_name = "celsius";
  category_name = "Temperature";
}
;;

let mdl_block_type_celsius_to_fahrenheit =
  simulink_block_of_conversion celsius_to_fahrenheit_conversion
;;

let mdl_block_type_fahrenheit_to_celsius =
  simulink_block_of_conversion fahrenheit_to_celsius_conversion
;;

(* Rotation conversions: degrees <-> radians *)

let degrees_to_radians_conversion = {
  block_source_type_name = "DegreesToRadians";
  block_source_block =
    "simulink_extras/Transformations/Degrees to\nRadians";
  from_name = "degrees";
  to_name = "radians";
  category_name = "Rotation";
}
;;

let radians_to_degrees_conversion = {
  block_source_type_name = "RadiansToDegrees";
  block_source_block =
    "simulink_extras/Transformations/Radians\nto Degrees";
  from_name = "radians";
  to_name = "degrees";
  category_name = "Rotation";
}
;;

let mdl_block_type_degrees_to_radians =
  simulink_block_of_conversion degrees_to_radians_conversion
;;

let mdl_block_type_radians_to_degrees =
  simulink_block_of_conversion radians_to_degrees_conversion
;;

(** The conversion generator *)

let st_ident =
  Ast_types.ident_from_string "SystemSampleTime"
;;

let genr_conversion conversion_block block_id genr_info =

  let get_ident = Genr_utilities_scicos_pro.get_ident_string genr_info in

  let st =
    if Genr_utilities_scicos_pro.exists_argument genr_info st_ident
    then get_ident st_ident
    else "-1" in

  let category_parameter =
    Value_parameter (
      Si "ucategory",
      Code_value (code_out_expression conversion_block.category_name))
  and pi_parameter =
    Value_parameter (
      Si (Printf.sprintf "PI_%s" conversion_block.category_name),
      Code_value (code_out_string conversion_block.from_name))
  and po_parameter =
    Value_parameter (
      Si (Printf.sprintf "PO_%s" conversion_block.category_name),
      Code_value (code_out_string conversion_block.to_name)) in

  let scicos_block_parameters ext =
    let ext_parameter =
      Value_parameter (
        Si "externalActivation",
        Code_value ext) in
    [
      category_parameter;
      pi_parameter;
      po_parameter;
      ext_parameter;
    ] in

  let scicos_basic_block_instructions blk_id ext =
    genr_instructions_for
      mdl_block_type_celsius_to_fahrenheit
      (scicos_block_parameters ext)
      [] blk_id genr_info in

  match st with
  | "-1" ->
    Concat_list.list
      (scicos_basic_block_instructions
         block_id external_activation_mode_off)

  | st ->

    let in_port_name = "in"
    and out_port_name = "out"
    and sample_clock_name = "sample_clock"
    and conv_name = "conv" in

    let in_port =
      Genr_utilities_scicos_pro.
      in_port_in_scsm0
        in_port_name port_1_
        (_0_, _0_) in
    let out_port =
      Genr_utilities_scicos_pro.
      out_port_in_scsm0
        out_port_name port_1_
        (_2_, _0_) in
    let conv =
      let nports = (_1_, _1_) in
      Genr_utilities_scicos_pro.
      add_block_body_in_scsm0
        conv_name
        (scicos_basic_block_instructions blk0 external_activation_mode_on)
        nports
        (_1_, _0_) in
    let sample_clock =
      Genr_utilities_scicos_pro.
      sample_clock_in_scsm0
        sample_clock_name (code_out_expression st)
        (_1_, _1_) in

    let links = [
      add_directed_link_in_scsm0
        in_port_name port_1_ conv_name port_1_ [];
      add_directed_link_in_scsm0
        conv_name port_1_ out_port_name port_1_ [];
      Genr_utilities_scicos_pro.
      add_event_link_in_scsm0
        sample_clock_name port_1_ conv_name port_1_ [];
    ] in

    let super_body = [
      in_port;
      out_port;
      conv;
      sample_clock;
      links;
    ] in

    super_block_in_scsm0 block_id super_body
;;

(* Celsius to Fahrenheit_conversions *)

let genr_celsius_to_fahrenheit_conversion =
  genr_conversion celsius_to_fahrenheit_conversion
;;

add_translator
  mdl_block_type_celsius_to_fahrenheit
  genr_celsius_to_fahrenheit_conversion
;;

let genr_fahrenheit_to_celsius_conversion =
  genr_conversion fahrenheit_to_celsius_conversion
;;

add_translator
  mdl_block_type_fahrenheit_to_celsius
  genr_fahrenheit_to_celsius_conversion
;;

(* Degrees to radians conversions *)

let genr_degrees_to_radians_conversion =
  genr_conversion degrees_to_radians_conversion
;;

add_translator
  mdl_block_type_degrees_to_radians
  genr_degrees_to_radians_conversion
;;

let genr_radians_to_degrees_conversion =
  genr_conversion radians_to_degrees_conversion
;;

add_translator
  mdl_block_type_radians_to_degrees
  genr_radians_to_degrees_conversion
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
