(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_coordinates_to_complex.ml,v 1.1 2019-05-10 14:04:56 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
open Unatural
;;

(* Computing complex numbers from their representation *)

(* Simulink bloc types MagnitudeAngleToComplex and RealImagToComplex share
   a lot of code.
   Hence the sharing of the generator.
*)

let magnitude_angle_option = "Magnitude and angle";;
let real_imag_option = "Real and imag";;

let mdl_block_type_real_imag_to_complex =
  Mdl_basics_explicit_ast.Explicit_block_regular_type
    Mdl_block_types.RealImagToComplex
;;

let real_imag_to_complex_formula = code_out_expression "RtoC";;
let real_imag_to_complex_option = real_imag_option;;

let mdl_block_type_magnitude_angle_to_complex =
  Mdl_basics_explicit_ast.Explicit_block_regular_type
    Mdl_block_types.MagnitudeAngleToComplex
;;

let magnitude_angle_to_complex_formula = code_out_expression "PtoC";;
let magnitude_angle_to_complex_option = magnitude_angle_option;;

(* The generator for all the above blocks. *)

let input_ident = Ast_types.ident_from_string "Input";;

let genr mdl_block_type conversion_formula input_option block_id genr_info =

  let input = get_ident_string genr_info input_ident
  and sample_time = Genr_utilities_scicos_pro.get_sample_time genr_info in

  let scicos_block_parameters activation = [
    Value_parameter (Si "decomptyp", Code_value conversion_formula);
    Value_parameter (Si "externalActivation", Code_value activation);
  ] in

  let scicos_basic_block_instructions blk_id activation =
    genr_instructions_for
      mdl_block_type
      (scicos_block_parameters activation)
      [] blk_id genr_info in

  if input <> input_option then begin
    Say.warning
      (Printf.sprintf
         "%S to complex Simulink block translation: \
          unsupported option %S"
         input_option
         input);
    Genr_scicos_pro_not_yet_translated.genr block_id genr_info
  end else

  match sample_time with
  | "-1" ->
    Concat_list.list
      (scicos_basic_block_instructions block_id
         Genr_utilities_scicos_pro.external_activation_mode_off)

  | _ ->

    let sample_time = code_out_expression sample_time in

    let in_port_name1 = "in1"
    and in_port_name2 = "in2"
    and out_port_name = "out"
    and ritc_name = "ritc"
    and sample_clock_name = "sample_clock" in

    let in_port1 =
      Genr_utilities_scicos_pro.
      in_port_in_scsm0
        in_port_name1 port_1_
        (_0_, _0_) in
    let in_port2 =
      Genr_utilities_scicos_pro.
      in_port_in_scsm0
        in_port_name2 port_1_
        (_0_, _1_) in
    let out_port =
      Genr_utilities_scicos_pro.
      out_port_in_scsm0
        out_port_name port_2_
        (_2_, _1_) in
    let ritc =
      let nports = (_2_, _1_) in
      Genr_utilities_scicos_pro.
      add_block_body_in_scsm0
        ritc_name
        (scicos_basic_block_instructions blk0
         Genr_utilities_scicos_pro.external_activation_mode_on)
        nports
        (_1_, _0_) in
    let sample_clock =
      Genr_utilities_scicos_pro.
      sample_clock_in_scsm0
        sample_clock_name sample_time
        (_1_, _1_) in

    let links = [
      Genr_utilities_scicos_pro.
      add_directed_link_in_scsm0
        in_port_name1 port_1_ ritc_name port_1_ [];
      add_directed_link_in_scsm0
        in_port_name2 port_1_ ritc_name port_2_ [];
      add_directed_link_in_scsm0
        ritc_name port_1_ out_port_name port_1_ [];
      Genr_utilities_scicos_pro.
      add_event_link_in_scsm0
        sample_clock_name port_1_ ritc_name port_1_ [];
    ] in

    let super_body = [
      in_port1;
      in_port2;
      out_port;
      ritc;
      sample_clock;
      links;
    ] in

    super_block_in_scsm0 block_id super_body
;;

let genr_magnitude_angle =
  genr
    mdl_block_type_magnitude_angle_to_complex
    magnitude_angle_to_complex_formula
    magnitude_angle_to_complex_option
;;

add_translator
  mdl_block_type_magnitude_angle_to_complex genr_magnitude_angle
;;

let genr_real_imag =
  genr
    mdl_block_type_real_imag_to_complex
    real_imag_to_complex_formula
    real_imag_to_complex_option
;;

add_translator mdl_block_type_real_imag_to_complex genr_real_imag
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
