(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_coulomb_viscous_friction.ml,v 1.1 2019-05-10 14:04:56 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
open Unatural
;;

let mdl_block_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Coulombic and Viscous Friction")
    (Ast_types.mdl_path_of_string
       "simulink/Discontinuities/Coulomb &\nViscous Friction")
;;

let offset_ident =
  Ast_types.ident_from_string "offset"
and gain_ident =
  Ast_types.ident_from_string "gain"
;;

let add_block_gen
    genr_info block_name block_type
    block_parameters nports position =

  let block_body =
    genr_instructions_for_modeler_block_type block_type
      block_parameters
      [] Genr_utils.blk0 genr_info in
  Genr_utilities_scicos_pro.
  add_block_body_in_scsm0 block_name
    block_body
    nports
    position
;;

let genr block_id genr_info =

  let add_block = add_block_gen genr_info in

  let scicos_block_parameters = [
    Value_parameter (Si "gain",
      Simulink_arg_expression gain_ident);
    Value_parameter (Si "overflow",
      Simulink_arg_saturation_set saturation_mode_off);
    Value_parameter (Si "externalActivation",
      Code_value
        Genr_utilities_scicos_pro.external_activation_mode_off);
  ] in

  let scicos_basic_block_instructions blk_id =
    genr_instructions_for
      mdl_block_type scicos_block_parameters
      [] blk_id genr_info in

  let in_port_name = "in"
  and out_port_name = "out"
  and abs_name = "abs"
  and signum_name = "signum"
  and gain_name = "gain"
  and bias_name = "bias"
  and product_name = "product"
  and split_name = "split" in

  let in_port =
    Genr_utilities_scicos_pro.
    in_port_in_scsm0 in_port_name port_1_

  and split_block =
    Genr_utilities_scicos_pro.
    split_in_scsm0 split_name

  and out_port =
    Genr_utilities_scicos_pro.
    out_port_in_scsm0 out_port_name port_1_

  and gain_block =
    let nports = (_1_, _1_) in
    Genr_utilities_scicos_pro.
    add_block_body_in_scsm0 gain_name
      (scicos_basic_block_instructions blk0)
      nports

  and abs_block =
    let abs_type = Sbt "system/MathOperations/Abs"
    and abs_parameters = [
      Value_parameter (Si "zcross", Code_value (code_out_bool true));
      Value_parameter (Si "overflow",
        Simulink_arg_saturation_set
          Genr_utilities_scicos_pro.saturation_mode_off);
      Value_parameter (Si "externalActivation",
        Code_value
          Genr_utilities_scicos_pro.external_activation_mode_off);
    ]
    and abs_ports = (_1_, _1_) in
    add_block
      abs_name abs_type abs_parameters abs_ports

  and signum_block =
    let signum_type = Sbt "system/MathOperations/Sign"
    and signum_parameters = [
      Value_parameter (Si "nzz",
        Code_value (code_out_bool true));
      Value_parameter (Si "externalActivation",
        Code_value Genr_utilities_scicos_pro.external_activation_mode_off);
    ]
    and signum_ports = (_1_, _1_) in
    add_block
      signum_name signum_type signum_parameters signum_ports

  and bias_block =
    let bias_type = Sbt "system/MathOperations/Bias"
    and bias_parameters = [
      Value_parameter (Si "b",
        Simulink_arg_expression offset_ident);
      Value_parameter (Si "over",
        Simulink_arg_saturation_set
          Genr_utilities_scicos_pro.saturation_mode_off);
    ]
    and bias_ports = (_1_, _1_) in
    add_block
      bias_name bias_type bias_parameters bias_ports

  and product_block =
    let product_type = Sbt "system/MathOperations/Product"
    and product_parameters =
      let inports_prod =
        (Si "sgn", [
           Code_value (code_out_string "*");
           Code_value (code_out_string "*");
         ]),
        [] in
      [
        Value_parameter (Si "nin",
          Code_value (code_out_int 2));
        Struct_parameter (Si "in_ports", inports_prod);
        Value_parameter (Si "overflow",
          Simulink_arg_saturation_set
            Genr_utilities_scicos_pro.saturation_mode_off);
        Value_parameter (Si "rmethod",
          Code_value (code_out_string "Int"));
        Value_parameter (Si "externalActivation",
          Code_value (code_out_int 0));
      ]
    and product_ports = (_2_, _1_) in
    add_block
      product_name product_type product_parameters product_ports in

  let blocks = [
    in_port (_0_, _0_);
    split_block (_1_, _0_);
    out_port (_6_, _1_);
    abs_block (_2_, _0_);
    signum_block (_2_, _1_);
    product_block (_5_, _1_);
    gain_block (_3_, _0_);
    bias_block (_4_, _0_);
  ] in

  let links = [
    Genr_utilities_scicos_pro.
    add_directed_link_in_scsm0
      in_port_name port_1_ split_name port_1_ [];
    add_directed_link_in_scsm0
      split_name port_1_ abs_name port_1_ [];
    add_directed_link_in_scsm0
      split_name port_2_ signum_name port_1_ [];
    add_directed_link_in_scsm0
      abs_name port_1_ gain_name port_1_ [];
    add_directed_link_in_scsm0
      signum_name port_1_ product_name port_1_ [];
    add_directed_link_in_scsm0
      gain_name port_1_ bias_name port_1_ [];
    add_directed_link_in_scsm0
      bias_name port_1_ product_name port_2_ [];
    add_directed_link_in_scsm0
      product_name port_1_ out_port_name port_1_ [];
  ] in

  let super_body =
    blocks @ [ links; ] in

  super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
