(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_delay.ml,v 1.1 2019-05-10 14:04:57 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

(** {3 Translating Simulink block Delay} *)

(**
  This block should be translated into two distinct scicos blocks,
  according to the delay length parameter: if length is 1 we use
  the Scicos block for unit delay (DiscreteDelay), otherwise we use
  the Scicos block for multiple delay (ShiftRegister).
*)

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
open Unatural
;;

(* The general case *)
let mdl_block_type_multiple_delay =
  Mdl_basics_explicit_ast.Explicit_block_regular_type
    Mdl_block_types.Delay
;;

(* The particular case for unit delay. *)
let mdl_block_type_unit_delay =
  Mdl_basics_explicit_ast.Explicit_block_regular_type
    Mdl_block_types.UnitDelay
;;

let mdl_block_type = mdl_block_type_multiple_delay;;

let external_reset_ident =
  Ast_types.ident_from_string "ExternalReset"
and initial_condition_source_ident =
  Ast_types.ident_from_string "InitialConditionSource"
and delay_length_ident =
  Ast_types.ident_from_string "DelayLength"
and initial_condition_ident =
  Ast_types.ident_from_string "InitialCondition"
and use_circular_buffer_ident =
  Ast_types.ident_from_string "UseCircularBuffer"
and delay_length_source_ident =
  Ast_types.ident_from_string "DelayLengthSource"
;;

let genr block_id genr_info =

  let get_ident_string =
    Genr_utilities_scicos_pro.
    get_ident_string genr_info in

  let initial_condition_source =
    get_ident_string initial_condition_source_ident in

  let delay_length_source =
    get_ident_string delay_length_source_ident in

  let sample_time =
    Genr_utilities_scicos_pro.
    get_sample_time genr_info in

  let initial_condition =
    get_ident_string initial_condition_ident in

  let show_enable_port =
    Genr_utilities_scicos_pro.
    get_show_enable_port genr_info in

  let use_circular_buffer =
    get_ident_bool genr_info use_circular_buffer_ident in

  let external_reset =
    get_ident_string external_reset_ident in

  let delay_length =
    get_ident_string delay_length_ident in

  match
    external_reset, initial_condition_source,
    delay_length_source, show_enable_port,
    use_circular_buffer with
  | "None", "Dialog", "Dialog", false, false ->

    let mdl_block_type_and_parameters delay_length ext =
      if delay_length = "1"
      then
        mdl_block_type_unit_delay,
        [
          Value_parameter (Si "init_cond",
            Code_value (code_out_expression initial_condition));
          Value_parameter (Si "typ",
            Code_value (inherit_mode_type));
          Value_parameter (Si "externalActivation",
            Code_value ext)
        ]
      else
        mdl_block_type_multiple_delay,
        [
          Value_parameter (Si "K",
            Code_value (code_out_expression delay_length));
          Value_parameter (Si "z0",
            Code_value (code_out_expression initial_condition));
          Value_parameter (Si "type",
            Code_value (inherit_mode_type));
          Value_parameter (Si "externalActivation",
            Code_value ext)
        ] in

    let scicos_basic_block_instructions
        mdl_block_type scicos_block_parameters block_id =
      genr_instructions_for
        mdl_block_type scicos_block_parameters
        [] block_id genr_info in

    if sample_time = "-1" then
      let mdl_block_type, scicos_block_parameters =
        mdl_block_type_and_parameters
          delay_length
          external_activation_mode_off in
      Concat_list.list
        (scicos_basic_block_instructions
           mdl_block_type
           scicos_block_parameters block_id) else

    let in_port_name = "in"
    and out_port_name = "out"
    and dollar_name = "dollar"
    and sample_clock_name = "sample_clock" in

    let in_port position =
      Genr_utilities_scicos_pro.
      in_port_in_scsm0
        in_port_name port_1_ position

    and out_port position =
      Genr_utilities_scicos_pro.
      out_port_in_scsm0
        out_port_name port_1_ position in

    let dollar_block_type, dollar_block_parameters =
      mdl_block_type_and_parameters delay_length
      external_activation_mode_on in

    let dollar =
      let dollar_nports = (_1_, _1_) in
      Genr_utilities_scicos_pro.
      add_block_body_in_scsm0
        dollar_name
        (scicos_basic_block_instructions
           dollar_block_type
           dollar_block_parameters blk0)
        dollar_nports in

    let sample_clock =
      Genr_utilities_scicos_pro.
      sample_clock_in_scsm0
        sample_clock_name
        (code_out_expression sample_time) in

    (* Creating blocks before adding links. *)
    let in_port_block = in_port (_0_, _0_)
    and dollar_block = dollar (_1_, _0_)
    and sample_clock_block = sample_clock (_1_, _1_)
    and out_port_block = out_port (_2_, _0_) in

    let links = [
      Genr_utilities_scicos_pro.
      add_directed_link_in_scsm0
        in_port_name port_1_ dollar_name port_1_ [];
      add_directed_link_in_scsm0
        dollar_name port_1_ out_port_name port_1_ [];
      Genr_utilities_scicos_pro.
      add_event_link_in_scsm0
        sample_clock_name port_1_ dollar_name port_1_ [];
    ] in

    let super_body = [
      in_port_block;
      out_port_block;
      dollar_block;
      sample_clock_block;
      links;
    ] in

    super_block_in_scsm0 block_id super_body

  | _, _, _, _, _ ->
    Say.warning (Printf.sprintf
      "Delay Simulink block translation: \
       configuration not yet supported.");
    Genr_scicos_pro_not_yet_translated.genr block_id genr_info
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
