(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_detect_rise_and_fall.ml,v 1.1 2019-05-10 14:04:57 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
open Unatural
;;

(* We translate 4 (reference) blocks here:
   Detect Fall Negative
   Detect Fall Non positive
   Detect Rise Positive
   Detect Rise Non negative
*)

let mdl_block_type_detect_fall_negative =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Detect Fall Negative")
    (Ast_types.mdl_path_of_string
       "simulink/Logic and Bit\nOperations/Detect Fall\nNegative")
;;

let mdl_block_type_detect_fall_nonpositive =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Detect Fall Nonpositive")
    (Ast_types.mdl_path_of_string
       "simulink/Logic and Bit\nOperations/Detect Fall\nNonpositive")
;;

let mdl_block_type_detect_rise_positive =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Detect Rise Positive")
    (Ast_types.mdl_path_of_string
       "simulink/Logic and Bit\nOperations/Detect Rise\nPositive")
;;

let mdl_block_type_detect_rise_nonnegative =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Detect Rise Nonnegative")
    (Ast_types.mdl_path_of_string
       "simulink/Logic and Bit\nOperations/Detect Rise\nNonnegative")
;;

let vinit_ident = Ast_types.ident_from_string "vinit"
;;

let genr_detect_comparison mdl_block_type op1 op2 block_id genr_info =

  let get_ident_string =
    Genr_utilities_scicos_pro.
    get_ident_string genr_info in

  let out_data_type =
    Genr_utilities_scicos_pro.
    get_simulink_out_data_type_value genr_info in

  let vinit = get_ident_string vinit_ident in

  let scicos_block_parameters op = [
    Value_parameter (Si "op", Code_value (code_out_string op));
    Value_parameter (Si "C", Code_value (code_out_int 0));
    Value_parameter (Si "zcr", Code_value (code_out_bool false));
  ] in

  let scicos_basic_block_instructions blk_id op =
    genr_instructions_for
      mdl_block_type (scicos_block_parameters op)
      [] blk_id genr_info in

  let in_port_name = "in"
  and out_port_name = "out"
  and dollar_name = "dollar"
  and logical_op_name = "and"
  and comp1_name = "comp1"
  and comp2_name = "comp2"
  and split_name = "split" in

  let split_block =
    Genr_utilities_scicos_pro.
    split_in_scsm0 split_name (_1_, _1_)
  and in_port =
    Genr_utilities_scicos_pro.
    in_port_in_scsm0 in_port_name port_1_ (_0_, _0_)
  and out_port =
    Genr_utilities_scicos_pro.
    out_port_in_scsm0 out_port_name port_1_ (_5_, _0_)

  and dollar_block =
    let scicos_block_type = Sbt "system/Dynamical/DiscreteDelay"
    and dollar_parameters = [
      Value_parameter (Si "init_cond",
        Code_value (code_out_expression vinit));
      Value_parameter (Si "typ",
        Code_value
          Genr_utilities_scicos_pro.inherit_mode_type);
      Value_parameter (Si "externalActivation",
        Code_value
          Genr_utilities_scicos_pro.external_activation_mode_off);
    ]
    and dollar_nports = (_1_, _1_) in

    Genr_utilities_scicos_pro.
    add_block_body_in_scsm0 dollar_name
      (genr_instructions_for_modeler_block_type scicos_block_type
         dollar_parameters
         [] Genr_utils.blk0 genr_info)
      dollar_nports
      (_2_, _0_)

  and comp1_block =
    let comp1_nports = (_1_, _1_) in
    Genr_utilities_scicos_pro.
    add_block_body_in_scsm0 comp1_name
      (scicos_basic_block_instructions blk0 op1)
      comp1_nports
      (_3_, _1_)

  and comp2_block =
    let comp2_nports = (_1_, _1_) in
    Genr_utilities_scicos_pro.
    add_block_body_in_scsm0 comp2_name
      (scicos_basic_block_instructions blk0 op2)
      comp2_nports
      (_3_, _0_)

  and logical_op_block =
    let scicos_block_type = Sbt "system/LogicalOperations/LogicalOp"
    and logical_op_parameters = [
      Value_parameter (Si "nin", Code_value (code_out_int 2));
      Value_parameter (Si "rule", Code_value (code_out_string "AND"));
      Value_parameter (Si "Datatype", Code_value out_data_type);
      Value_parameter (Si "tp", Code_value (code_out_int 0));
      Value_parameter (Si "externalActivation", Code_value
           Genr_utilities_scicos_pro.external_activation_mode_off);
    ]
    and logical_op_nports = (_2_, _1_) in

    Genr_utilities_scicos_pro.
    add_block_body_in_scsm0 logical_op_name
      (genr_instructions_for_modeler_block_type scicos_block_type
         logical_op_parameters
         [] Genr_utils.blk0 genr_info)
      logical_op_nports
      (_4_, _0_) in

  let links = [
    Genr_utilities_scicos_pro.
    add_directed_link_in_scsm0
      in_port_name port_1_ split_name port_1_ [];
    add_directed_link_in_scsm0
      split_name port_1_ comp1_name port_1_ [];
    add_directed_link_in_scsm0
      split_name port_2_ dollar_name port_1_ [];
    add_directed_link_in_scsm0
      dollar_name port_1_ comp2_name port_1_ [];
    add_directed_link_in_scsm0
      comp1_name port_1_ logical_op_name port_1_ [];
    add_directed_link_in_scsm0
      comp2_name port_1_ logical_op_name port_2_ [];
    add_directed_link_in_scsm0
      logical_op_name port_1_ out_port_name port_1_ [];
  ] in

  let super_body = [
    in_port;
    split_block;
    out_port;
    dollar_block;
    comp1_block;
    comp2_block;
    logical_op_block;
    links;
  ] in

  super_block_in_scsm0 block_id super_body
;;

let genr_detect_fall_negative =
  genr_detect_comparison
    mdl_block_type_detect_fall_negative "<" ">="
and genr_detect_fall_nonpositive =
  genr_detect_comparison
    mdl_block_type_detect_fall_nonpositive "<=" ">"
and genr_detect_rise_positive =
  genr_detect_comparison
    mdl_block_type_detect_rise_positive ">" "<="
and genr_detect_rise_nonnegative =
  genr_detect_comparison
    mdl_block_type_detect_rise_nonnegative ">=" "<"
;;

add_translator
  mdl_block_type_detect_fall_negative genr_detect_fall_negative
;;

add_translator
  mdl_block_type_detect_fall_nonpositive genr_detect_fall_nonpositive
;;

add_translator
  mdl_block_type_detect_rise_positive genr_detect_rise_positive
;;

add_translator
  mdl_block_type_detect_rise_nonnegative genr_detect_rise_nonnegative
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
