(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_discrete_integral.ml,v 1.1 2019-05-10 14:04:57 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
open Unatural
;;

let mdl_block_type =
  Mdl_basics_explicit_ast.
  Explicit_block_regular_type
    Mdl_block_types.
    DiscreteIntegrator
;;

let external_reset_ident =
  Ast_types.ident_from_string "ExternalReset"
and initial_condition_source_ident =
  Ast_types.ident_from_string "InitialConditionSource"
and initial_condition_ident =
  Ast_types.ident_from_string "InitialCondition"
and limit_output_ident =
  (* Limit_output_ident is a boolean (on or off) in Simulink.
     In scios_pro this boolean should be translated as a boolean as well. *)
  Ast_types.ident_from_string "LimitOutput"
and upper_saturation_limit_ident =
  Ast_types.ident_from_string "UpperSaturationLimit"
and lower_saturation_limit_ident =
  Ast_types.ident_from_string "LowerSaturationLimit"
and gainval_ident =
  Ast_types.ident_from_string "gainval"
and int_method_ident =
  Ast_types.ident_from_string "IntegratorMethod"
;;

let genr block_id genr_info =

  let get_ident_string =
    Genr_utilities_scicos_pro.get_ident_string genr_info in

  let sample_time =
    Genr_utilities_scicos_pro.get_sample_time genr_info
  and external_reset =
    get_ident_string external_reset_ident
  and show_saturation_port =
    Genr_utilities_scicos_pro.
    get_show_saturation_port genr_info
  and int_method =
    get_ident_string int_method_ident in

  if show_saturation_port ||
     external_reset <> "none" then
    begin
      Say.warning
        (Printf.sprintf
           "This configuration of discrete integrator Simulink block\n \
            is not supported yet:\n \
            show_saturation=%b, external_reset=%S\n"
            show_saturation_port external_reset);
      Genr_scicos_pro_not_yet_translated.genr block_id genr_info
    end else

  let st = code_out_expression sample_time
  and initial_condition_source =
    get_ident_string initial_condition_source_ident in

  let unsupported_method simulink_method scicos_method =
    Say.warning (Printf.sprintf
      "Discrete Integrator Simulink block translation: \
       Unsupported method '%s'; using method '%s'"
      simulink_method scicos_method);
      scicos_method in

  let forward_euler_method = "Forward Euler"
  and backward_euler_method = "Backward Euler"
  and trapezoidal_method = "Trapezoidal" in

  let get_method simulink_method =
    match simulink_method with
    | "Integration: Forward Euler" -> forward_euler_method
    | "Integration: Backward Euler" -> backward_euler_method
    | "Integration: Trapezoidal" -> trapezoidal_method
    | "Accumulation: Forward Euler" ->
      unsupported_method simulink_method forward_euler_method
    | "Accumulation: Backward Euler" ->
      unsupported_method simulink_method backward_euler_method
    | "Accumulation: Trapezoidal" ->
      unsupported_method simulink_method trapezoidal_method
    | simulink_method ->
      unsupported_method simulink_method forward_euler_method in

  let scicos_basic_block_instructions reinit blk_id =
    let scicos_block_parameters = [
      Value_parameter (Instructions.Si "x0",
        Simulink_arg_expression initial_condition_ident);
      Value_parameter (Si "method",
        Code_value (code_out_string (get_method int_method)));
      Value_parameter (Si "typ",
        Code_value (inherit_mode_type));
      Value_parameter (Si "reinit",
        Code_value (code_out_bool reinit));
      Value_parameter (Si "satur",
        Simulink_arg_value limit_output_ident);
      Struct_parameter (Si "limit", (
        (Si "max",
           [Simulink_arg_expression upper_saturation_limit_ident]),
        [Si "min",
           [Simulink_arg_expression lower_saturation_limit_ident]]
        )
      );
    ] in

    genr_instructions_for
      mdl_block_type scicos_block_parameters
      [] blk_id genr_info in

  let in_port1_name = "in1"
  and out_port_name = "out"
  and gain_name = "gain"
  and split_name = "split"
  and sample_clock_name = "sample_clock"
  and extract_activation_name = "extract_activation"
  and integrator_name = "integrator" in

  let in_port1 =
    Genr_utilities_scicos_pro.
    in_port_in_scsm0 in_port1_name port_1_
      (Unatural._0_, Unatural._1_) in

  let split =
    Genr_utilities_scicos_pro.
    split_in_scsm0 split_name
      (_1_, _0_) in

  let extract_activation =
    let nports = (_1_, _0_) in
    let extraction_body =
      genr_instructions_for_modeler_block_type
        (Sbt "system/ActivationOperations/ExtractActivation")
        [] [] blk0 genr_info in
    Genr_utilities_scicos_pro.
    add_block_body_in_scsm0
      extract_activation_name
      extraction_body
      nports
      (_2_, _1_) in

  let gain =
    let nports = (_1_, _1_) in
    let gain_body =
      genr_instructions_for_modeler_block_type
        (Sbt "system/MathOperations/Gain")
        [
          Value_parameter (Si "gain",
            Simulink_arg_expression gainval_ident);
          Value_parameter (Si "overflow",
            Code_value (code_out_string "Nothing"));
          Value_parameter (Si "externalActivation",
            Code_value external_activation_mode_off);
        ]
        [] blk0 genr_info in
    Genr_utilities_scicos_pro.
    add_block_body_in_scsm0
      gain_name
      gain_body
      nports
      (_2_, _0_) in

  let out_port =
    Genr_utilities_scicos_pro.
    out_port_in_scsm0
      out_port_name port_1_
      (_4_, _0_) in

  let sample_clock =
    Genr_utilities_scicos_pro.
    sample_clock_in_scsm0
      sample_clock_name st
      (_2_, _1_) in

  match initial_condition_source with
  | "internal" ->

    let integrator =
      let nports = (_1_, _1_) in
      let integrator_body =
        scicos_basic_block_instructions false blk0 in
      Genr_utilities_scicos_pro.
      add_block_body_in_scsm0
        integrator_name
        integrator_body
        nports
        (_3_, _0_) in

    let links =
      match sample_time with
      | "-1" ->
        [
          Genr_utilities_scicos_pro.
          add_directed_link_in_scsm0
            in_port1_name port_1_ split_name port_1_ [];
          add_directed_link_in_scsm0
            split_name port_1_ gain_name port_1_ [];
          add_directed_link_in_scsm0
            split_name port_2_ extract_activation_name port_1_ [];
          add_directed_link_in_scsm0
            gain_name port_1_ integrator_name port_1_ [];
          add_directed_link_in_scsm0
            integrator_name port_1_ out_port_name port_1_ [];
          Genr_utilities_scicos_pro.
          add_event_link_in_scsm0
             extract_activation_name port_1_ integrator_name port_1_ [];
        ]
      | _ ->
        [
          add_directed_link_in_scsm0
            in_port1_name port_1_ gain_name port_1_ [];
          add_directed_link_in_scsm0
            gain_name port_1_ integrator_name port_1_ [];
          add_directed_link_in_scsm0
            integrator_name port_1_ out_port_name port_1_ [];
          add_event_link_in_scsm0
             sample_clock_name port_1_ integrator_name port_1_ [];
        ] in

    let super_body =
      match sample_time with
      | "-1" ->
        [
          in_port1;
          extract_activation;
          split;
          out_port;
          gain;
          integrator;
          links;
        ]
      | _ ->
        [
          in_port1;
          out_port;
          gain;
          integrator;
          sample_clock;
          links;
        ] in
    super_block_in_scsm0 block_id super_body

  | "external" ->

    let integrator =
      let nports = (_2_, _1_) in
      let integrator_body =
        scicos_basic_block_instructions true blk0 in
      add_block_body_in_scsm0
        integrator_name
        integrator_body
        nports
        (_3_, _0_) in

    let in_port2_name = "in2"
    and sample_clock2_name = "sample_clock2" in

    let in_port2 =
      in_port_in_scsm0
        in_port2_name port_2_
        (_0_, _0_) in

    let sample_clock2 =
      let nports = (_0_, _0_) in
      let sample_clock_body =
        let sci_block_type =
          Sbt "system/ActivationOperations/InitialEvent" in
        genr_instructions_for_modeler_block_type sci_block_type []
          [] blk0 genr_info in
      add_block_body_in_scsm0
        sample_clock2_name
        sample_clock_body
        nports
        (_3_, _1_) in

    let links =
      match sample_time with
      | "-1" ->
        [
          add_directed_link_in_scsm0
            in_port1_name port_1_ split_name port_1_ [];
          add_directed_link_in_scsm0
            split_name port_1_ gain_name port_1_ [];
          add_directed_link_in_scsm0
            split_name port_2_ extract_activation_name port_1_ [];
          add_directed_link_in_scsm0
            gain_name port_1_ integrator_name port_1_ [];
          add_directed_link_in_scsm0
            in_port2_name port_1_ integrator_name port_2_ [];
          add_directed_link_in_scsm0
            integrator_name port_1_ out_port_name port_1_ [];
          add_event_link_in_scsm0
             extract_activation_name port_1_ integrator_name port_1_ [];
          add_event_link_in_scsm0
             sample_clock2_name port_1_ integrator_name port_2_ [];
        ]
      | _ ->
        [
          add_directed_link_in_scsm0
            in_port1_name port_1_ gain_name port_1_ [];
          add_directed_link_in_scsm0
            gain_name port_1_ integrator_name port_1_ [];
          add_directed_link_in_scsm0
            in_port2_name port_1_ integrator_name port_2_ [];
          add_directed_link_in_scsm0
            integrator_name port_1_ out_port_name port_1_ [];
          add_event_link_in_scsm0
             sample_clock_name port_1_ integrator_name port_1_ [];
          add_event_link_in_scsm0
             sample_clock2_name port_1_ integrator_name port_2_ [];
        ] in

    let super_body =
      match sample_time with
      | "-1" ->
        [
          in_port1;
          in_port2;
          out_port;
          split;
          gain;
          integrator;
          extract_activation;
          sample_clock2;
          links;
        ]
      | _ ->
        [
          in_port1;
          in_port2;
          out_port;
          gain;
          integrator;
          sample_clock;
          sample_clock2;
          links;
        ] in
    super_block_in_scsm0 block_id super_body

  | initial_condition_source ->
    Say.warning (Printf.sprintf
      "Discrete Integrator Simulink block translation: \
       unknown initial condition source %S"
      initial_condition_source);
    Genr_scicos_pro_not_yet_translated.genr block_id genr_info
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
