(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_discrete_pulse_generator.ml,v 1.1 2019-05-10 14:04:57 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
open Unatural
;;

let mdl_block_type =
  Mdl_basics_explicit_ast.Explicit_block_regular_type
    Mdl_block_types.DiscretePulseGenerator
;;

let pulse_type_ident = Ast_types.ident_from_string "PulseType"
and time_source_ident = Ast_types.ident_from_string "TimeSource"
and amplitude_ident = Ast_types.ident_from_string "Amplitude"
and phase_delay_ident = Ast_types.ident_from_string "PhaseDelay"
and pulse_width_ident = Ast_types.ident_from_string "PulseWidth"
and period_ident = Ast_types.ident_from_string "Period"
;;

let genr block_id genr_info =

  let get_ident_string =
    Genr_utilities_scicos_pro.
    get_ident_string genr_info in

  let pulse_type = get_ident_string pulse_type_ident in
  let time_source = get_ident_string time_source_ident in
  let amplitude = get_ident_string amplitude_ident in
  let pulse_width = get_ident_string pulse_width_ident in
  let phase_delay = get_ident_string phase_delay_ident in
  let period =
    code_out_expression (get_ident_string period_ident) in

  let get_exx tmin tmax =
     code_out_string
       (Printf.sprintf "((u1>=%s)&&(u1<%s))*(%s)" tmin tmax amplitude) in

  let get_expression position expression_name tmin tmax =
    let nports = (_1_, _1_) in
    Genr_utilities_scicos_pro.
    add_block_body_in_scsm0 expression_name
      (genr_instructions_for mdl_block_type
         [
           Value_parameter (Si "nin",
             Code_value (code_out_int 1));
           Value_parameter (Si "work",
             Code_value (code_out_expression "getcurrentenv"));
           Value_parameter (Si "exx",
             Code_value (get_exx tmin tmax));
           Value_parameter (Si "zc",
             Code_value (code_out_bool true));
           Value_parameter (Si "externalActivation",
             Code_value external_activation_mode_off);
         ]
         [] Genr_utils.blk0 genr_info)
      nports
      position in

  let time_based (tmin,tmax) =
    let out_port_name = "out"
    and expression_name = "expression"
    and saw_tooth_name = "saw_tooth_name" in

    let out_port =
      Genr_utilities_scicos_pro.out_port_in_scsm0
        out_port_name port_1_
        (_2_, _0_) in

    let expression = get_expression (_1_, _0_) expression_name tmin tmax in

    let saw_tooth =
      let nports = (_0_, _1_) in
      let scicos_block_type = Sbt "system/SignalGenerators/Sawtooth" in
      Genr_utilities_scicos_pro.
      add_block_body_in_scsm0 saw_tooth_name
        (genr_instructions_for_modeler_block_type scicos_block_type
           [
             Value_parameter (Si "period", Code_value period);
             Value_parameter (Si "down", Code_value (code_out_float 0.0));
             Value_parameter (Si "up", Code_value period);
             Value_parameter (Si "tinit", Code_value (code_out_float 0.0));
           ] [] Genr_utils.blk0 genr_info)
        nports
        (_0_, _0_) in

    let links = [
      Genr_utilities_scicos_pro.
      add_directed_link_in_scsm0
        expression_name port_1_ out_port_name port_1_ [];
      add_directed_link_in_scsm0
        saw_tooth_name port_1_ expression_name port_1_ [];
    ] in

    let super_body = [
      out_port;
      expression;
      saw_tooth;
      links;
    ] in

    Genr_utilities_scicos_pro.super_block_in_scsm0 block_id super_body in

  let extern (tmin,tmax) =
    let out_port_name = "out"
    and expression_name = "expression"
    and modulo_name =" modulo"
    and constant_name = "constant"
    and in_port_name = "in" in

    let out_port =
      Genr_utilities_scicos_pro.out_port_in_scsm0
        out_port_name port_1_
        (_3_, _1_) in

    let in_port =
      Genr_utilities_scicos_pro.in_port_in_scsm0
        in_port_name port_1_
        (_0_, _1_) in

    let constant =
      let nports = (_0_, _1_) in
      let scicos_block_type = Sbt "system/SignalGenerators/Constant" in
      Genr_utilities_scicos_pro.
      add_block_body_in_scsm0 constant_name
        (genr_instructions_for_modeler_block_type scicos_block_type
         [
           Value_parameter (Si "C", Code_value period);
           Value_parameter (Si "typ",
             Code_value (Genr_utilities_scicos_pro.inherit_mode_type));
           Value_parameter (Si "externalActivation",
             Code_value external_activation_mode_off);
         ] [] Genr_utils.blk0 genr_info)
        nports
        (_0_, _0_) in

    let modulo =
      let nports = (_2_, _1_) in
      let scicos_block_type = Sbt "system/MathOperations/Modulo" in
      Genr_utilities_scicos_pro.
      add_block_body_in_scsm0 modulo_name
        (genr_instructions_for_modeler_block_type scicos_block_type
           [
             Value_parameter (Si "op",
               Code_value (code_out_string "Modulo"));
             Value_parameter (Si "externalActivation",
               Code_value external_activation_mode_off);
           ] [] Genr_utils.blk0 genr_info)
        nports
        (_1_, _1_) in

    let expression = get_expression (_2_, _1_) expression_name tmin tmax in

    let links = [
      Genr_utilities_scicos_pro.
      add_directed_link_in_scsm0
        expression_name port_1_ out_port_name port_1_ [];
      add_directed_link_in_scsm0
        modulo_name port_1_ expression_name port_1_ [];
      add_directed_link_in_scsm0
        in_port_name port_1_ modulo_name port_1_ [];
      add_directed_link_in_scsm0
        constant_name port_1_ modulo_name port_2_ [];
    ] in

    let super_body = [
      out_port;
      expression;
      in_port;
      modulo;
      constant;
      links;
    ] in

    Genr_utilities_scicos_pro.
    super_block_in_scsm0 block_id super_body in

  let sample_based (tmin, tmax) =
    let out_port_name = "out"
    and expression_name = "expression"
    and sample_clock_name = "sample_clock"
    and modulo_counter_name = "modulo_counter" in

    let out_port =
      Genr_utilities_scicos_pro.
      out_port_in_scsm0
        out_port_name port_1_
        (_2_, _0_) in

    let expression = get_expression (_1_, _0_) expression_name tmin tmax in

    let sample_time =
      Genr_utilities_scicos_pro.get_sample_time_value genr_info in

    let sample_clock =
      Genr_utilities_scicos_pro.
      sample_clock_in_scsm0
        sample_clock_name sample_time
        (_1_, _1_) in

    let modulo_counter =
      let nports = (_0_, _1_) in
      let scicos_block_type = Sbt "system/SignalGenerators/ModuloCounter" in
      let ini_state = Printf.sprintf "(%s)-1" phase_delay in
      Genr_utilities_scicos_pro.
      add_block_body_in_scsm0 modulo_counter_name
        (genr_instructions_for_modeler_block_type scicos_block_type
          [
            Value_parameter
              (Si "ini_state", Code_value (code_out_expression ini_state));
            Value_parameter (Si "base", Code_value period);
            Value_parameter
              (Si "typ",
               Code_value
                 (Genr_utilities_scicos_pro.code_out_simulink_out_data_type
                    Simulink_double));
          ] [] Genr_utils.blk0 genr_info)
        nports
        (_0_, _0_) in

    let links = [
      add_directed_link_in_scsm0
        expression_name port_1_ out_port_name port_1_ [];
      add_directed_link_in_scsm0
        modulo_counter_name port_1_ expression_name port_1_ [];
      Genr_utilities_scicos_pro.
      add_event_link_in_scsm0
        sample_clock_name port_1_ modulo_counter_name port_1_ [];
    ] in

    let super_body = [
      out_port;
      expression;
      modulo_counter;
      sample_clock;
      links;
    ] in

    super_block_in_scsm0 block_id super_body in


  let period = get_ident_string period_ident in

  let get_discrete_pulse_timing () =
    (phase_delay,
     Printf.sprintf "(%s) + (%s)" pulse_width phase_delay)
  and get_continuous_pulse_timing () =
    (phase_delay,
     Printf.sprintf "(%s)*(%s)/100+%s" period pulse_width phase_delay) in

  let code =
    match pulse_type, time_source with
    | "Time based", "Use simulation time" ->
       time_based (get_continuous_pulse_timing ())
    | "Time based", "Use external signal" ->
       extern (get_continuous_pulse_timing ())
    | "Sample based", "Use simulation time" ->
       sample_based (get_discrete_pulse_timing ())
    | "Sample based", "Use external signal" ->
       extern (get_discrete_pulse_timing ())
    | pulse_type, time_source ->
      Say.fatal_error (Printf.sprintf
        "DiscretePuseGenerator translation: unknown pulse-type '%s'\
       \nor time-source '%s'"
        pulse_type time_source
      ) in

  code
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
