(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_dot_product.ml,v 1.1 2019-05-10 14:04:57 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
open Unatural
;;

(* The dot product block has two names in Simulink:
   a reference type and a native type.
   This translator evidently works for both.
*)

let mdl_block_type_native =
  Mdl_basics_explicit_ast.Explicit_block_regular_type
    Mdl_block_types.DotProduct
;;

let mdl_block_type_reference =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Dot Product")
    (Ast_types.mdl_path_of_string "simulink/Math\nOperations/Dot Product")
;;

let genr block_id genr_info =

  let sample_time = get_sample_time genr_info in

  let inports_prod =
    let star = Code_value (code_out_string "*") in
    [star; star] in
  let inports_prod = (Si "sgn", inports_prod), [] in

  let scicos_block_parameters ext = [
    Genr_types.
    Value_parameter (Si "nin", Code_value (code_out_int 2));
    Struct_parameter (Si "in_ports", inports_prod);
    Value_parameter (Si "overflow", Simulink_arg_saturation);
    Value_parameter (Si "rmethod", Code_value floor_round_method);
    Value_parameter (Si "externalActivation", Code_value ext);
  ] in

  let scicos_basic_block_instructions blk_id ext =
    genr_instructions_for
      mdl_block_type_native (scicos_block_parameters ext)
      [] blk_id genr_info in

  let in_port1_name = "in1"
  and in_port2_name = "in2"
  and out_port_name = "out"
  and product0_name = "product0"
  and product1_name = "product1"
  and sum_name = "sum"
  and sample_clock_name = "sample_clock" in

  let in_port1 =
    in_port_in_scsm0 in_port1_name port_1_
    (_0_, _1_) in

  let in_port2 =
    in_port_in_scsm0 in_port2_name port_2_
    (_0_, _0_) in

  let out_port =
    out_port_in_scsm0 out_port_name port_1_
    (_3_, _0_) in

  let product0 =
    let nports = (_2_, _1_) in
    add_block_body_in_scsm0 product0_name
      (scicos_basic_block_instructions blk0 external_activation_mode_off)
      nports
      (_1_, _0_) in

  let sum =
    let inports_sum =
      (Si "sgn",
       [ Code_value (code_out_string "+") ]),
      [] in
    let nports = (_1_, _1_) in

    add_block_body_in_scsm0 sum_name
      (genr_instructions_for_modeler_block_type
         (Sbt "system/MathOperations/Sum")
         [
           Value_parameter (Si "nin",
             Code_value (code_out_int 1));
           Struct_parameter (Si "in_ports",
             inports_sum);
           Value_parameter (Si "satur",
             Simulink_arg_saturation);
           Value_parameter (Si "externalActivation",
             Code_value (external_activation_mode_off));
         ]
         [] Genr_utils.blk0 genr_info)
      nports
      (_2_, _0_) in

  let super_body =

    match sample_time with
    | "-1" ->

      let links = [
        Genr_utilities_scicos_pro.
        add_directed_link_in_scsm0
          in_port1_name port_1_ product0_name port_1_ [];
        add_directed_link_in_scsm0
          in_port2_name port_1_ product0_name port_2_ [];
        add_directed_link_in_scsm0
          product0_name port_1_ sum_name port_1_ [];
        add_directed_link_in_scsm0
          sum_name port_1_ out_port_name port_1_ [];
      ] in

      [
        in_port1;
        in_port2;
        out_port;
        sum;
        product0;
        links;
      ]

    | sample_time ->

      let sample_clock =
        let sample_time = code_out_expression sample_time in
        Genr_utilities_scicos_pro.
        sample_clock_in_scsm0
          sample_clock_name sample_time
          (_1_, _1_) in

      let product1 =
        let nports = (_2_, _1_) in
        Genr_utilities_scicos_pro.
        add_block_body_in_scsm0 product1_name
          (scicos_basic_block_instructions blk0 external_activation_mode_on)
        nports
        (_1_, _0_) in

      let links = [
        add_directed_link_in_scsm0
          in_port1_name port_1_ product1_name port_1_ [];
        add_directed_link_in_scsm0
          in_port2_name port_1_ product1_name port_2_ [];
        add_directed_link_in_scsm0
          product1_name port_1_ sum_name port_1_ [];
        add_directed_link_in_scsm0
          sum_name port_1_ out_port_name port_1_ [];
        Genr_utilities_scicos_pro.add_event_link_in_scsm0
          sample_clock_name port_1_ product1_name port_1_ [];
      ] in

      [
        in_port1;
        in_port2;
        out_port;
        sum;
        product1;
        sample_clock;
        links;
      ] in

  super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type_native genr
;;

add_translator mdl_block_type_reference genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
