(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_hit_cross.ml,v 1.1 2019-05-10 14:04:57 jpc Exp $ *)

(*open Unatural*)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Ast_types
open Genr_utilities_scicos_pro
open Unatural
;;

let mdl_block_type =
  Mdl_basics_explicit_ast.Explicit_block_regular_type
    Mdl_block_types.HitCross
;;

let hit_cross_offset_ident = ident_from_string "HitCrossingOffset"
and hit_cross_dir_ident = ident_from_string "HitCrossingDirection"
and show_port_ident = ident_from_string "ShowOutputPort"
;;

let add_block_gen
    genr_info block_name block_type
    block_parameters nports position =
  let block_body =
    genr_instructions_for_modeler_block_type
      block_type
      block_parameters
      [] Genr_utils.blk0 genr_info in

  Genr_utilities_scicos_pro.
  add_block_body_in_scsm0
    block_name
    block_body
    nports
    position
;;

let genr block_id genr_info =

  let get_ident_string =
    Genr_utilities_scicos_pro.get_ident_string genr_info in

  let add_block = add_block_gen genr_info in

  let edge_type = match (get_ident_string hit_cross_dir_ident) with
    | "rising" -> "- to +"
    | "falling" -> "+ to -"
    | "either" -> "Zcross"
    | s ->
      Genr_utilities_scicos_pro.
      fatal_error (Printf.sprintf
        "HitCross block: unsupported direction %S" s) in

  let bias_expression = get_ident_string hit_cross_offset_ident in
  let bias_expression = Printf.sprintf "-(%s)" bias_expression in

  let show_port = get_ident_bool genr_info show_port_ident in

  let scicos_block_parameters  = [
    Value_parameter (Si "edge",
      Code_value (code_out_string edge_type));
    Value_parameter (Si "showport",
      Code_value (code_out_bool show_port));
  ] in

  let scicos_basic_block_instructions blk_id =
    genr_instructions_for
      mdl_block_type scicos_block_parameters
      [] blk_id genr_info in

  let in_port_name = "in"
  and out_port_name = "out"
  and edge_trigger_name = "trigger"
  and abs_name = "abs"
  and bias_name = "bias" in

  let in_port =
    Genr_utilities_scicos_pro.
    in_port_in_scsm0 in_port_name port_1_

  and out_port =
    Genr_utilities_scicos_pro.
    out_port_in_scsm0 out_port_name port_1_ in

  let bias =
    let bias_type = Sbt "system/MathOperations/Bias"
    and bias_parameters = [
      Value_parameter (Si "b",
        Code_value (code_out_expression bias_expression));
      Value_parameter (Si "over",
        Simulink_arg_saturation_set
          Genr_utilities_scicos_pro.saturation_mode_off);
    ]
    and bias_ports = (_1_, _1_) in

    add_block
      bias_name bias_type bias_parameters bias_ports in

  let edge_trigger =
    let nports = (_1_, _1_) in
    Genr_utilities_scicos_pro.
    add_block_body_in_scsm0
      edge_trigger_name
      (scicos_basic_block_instructions blk0)
      nports in

  let abs =
    let abs_type = Sbt "system/MathOperations/Abs"
    and abs_parameters = [
      Value_parameter (Si "zcross", Code_value (code_out_bool false));
      Value_parameter (Si "overflow",
        Simulink_arg_saturation_set
          Genr_utilities_scicos_pro.saturation_mode_off);
      Value_parameter (Si "externalActivation",
        Code_value
          Genr_utilities_scicos_pro.external_activation_mode_off);
    ]
    and abs_ports = (_1_, _1_) in

    add_block
      abs_name abs_type abs_parameters abs_ports in

  let blocks = if show_port then
    [
      in_port (_0_, _0_);
      out_port (_4_, _1_);
      abs (_3_, _0_);
      edge_trigger (_2_, _0_);
      bias (_1_, _0_);
    ] else
    [
      in_port (_0_, _0_);
      edge_trigger (_2_, _0_);
      bias (_1_, _0_);
    ] in

  let links = if show_port then
    [
      Genr_utilities_scicos_pro.
      add_directed_link_in_scsm0
        in_port_name port_1_ bias_name port_1_ [];
      add_directed_link_in_scsm0
        bias_name port_1_ edge_trigger_name port_1_ [];
      add_directed_link_in_scsm0
        edge_trigger_name port_1_ abs_name port_1_ [];
      add_directed_link_in_scsm0
        abs_name port_1_ out_port_name port_1_ [];
    ] else
    [
      Genr_utilities_scicos_pro.
      add_directed_link_in_scsm0
        in_port_name port_1_ bias_name port_1_ [];
      add_directed_link_in_scsm0
        bias_name port_1_ edge_trigger_name port_1_ [];
    ] in

  let super_body =
    blocks @ [ links; ] in

  super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
