(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_if.ml,v 1.1 2019-05-10 14:04:57 jpc Exp $ *)

(*open Unatural*)
module Genr_utilities_scicos = struct end;;

(* Pure Caml auxiliaries. *)

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
;;

(** The block translator code for Simulink blocks [If]
  and [SwitchCase].. *)

let mdl_block_type_if =
  Mdl_basics_explicit_ast.Explicit_block_regular_type
  Mdl_block_types.If
;;

let mdl_block_type_switch_case =
  Mdl_basics_explicit_ast.Explicit_block_regular_type
  Mdl_block_types.SwitchCase
;;

(* [If] block parameters:

    NumInputs : string
    IfExpression : string
    ElseIfExpressions : string
    ShowElse : boolean

    ZeroCross : boolean
    SampleTime : string
*)

(* [SwitchCase] block parameters:

    CaseConditions : string
    ShowDefaultCase : boolean  OR     CaseShowDefault : boolean
    ZeroCross : boolean
    SampleTime : string
*)


let get_items s =
  let lim = String.length s in
  let accu = Buffer.create lim in
  Buffer.add_string accu "u1==";
  let rec loop paren_count bracket_count i items =
    if i >= lim then
      let s = Buffer.contents accu in
      Buffer.clear accu;
      List.rev (Code_value (code_out_string s) :: items)
    else
      let c = s.[i] in
      let i = succ i in
      match c with
      | '(' ->
        Buffer.add_char accu c;
        loop (succ paren_count) bracket_count i items
      | ')' ->
        Buffer.add_char accu c;
        loop (pred paren_count) bracket_count i items
      | '[' ->
(*        Buffer.add_char accu c; *)
        loop paren_count (succ bracket_count) i items
      | ']' ->
(*        Buffer.add_char accu c; *)
        loop paren_count (pred bracket_count) i items
      | ',' when paren_count = 0 ->
        if bracket_count = 0 then
          let s = Buffer.contents accu in
            Buffer.clear accu;
            Buffer.add_string accu "u1==";
            loop paren_count bracket_count i (Code_value (code_out_string s) :: items)
        else
          let s = Buffer.contents accu in
            Buffer.add_string accu " || u1==";
            loop paren_count bracket_count i (Code_value (code_out_string s) :: items)
      | '}' | '{' ->  (* skip curly braces *)
        loop paren_count bracket_count i items
      | _ ->
        Buffer.add_char accu c;
        loop paren_count bracket_count i items in
  loop 0 0 0 []
;;

(* Computes the list of code value 1 corresponding port sizes. *)
let list_of_ones_of_int n =
  let rec loop accu = function
    | 0 -> accu
    | n ->
      let v = Code_value (code_out_int 1) in
      loop (v :: accu) (succ n) in
  loop [] (-n)
;;

(** [char_is_blank c] tests if character [c] is a blank, i.e. a character
  that prints as a spacing character. *)
let char_is_blank = function
  | ' ' | '\t' | '\r' | '\n' -> true
  | _c -> false
;;

(** [chomp s] Removes spurious blanks from both ends of string [s].
  Simply returns [s], if no end of [s] is a blank characters, *)
let chomp s =
  let len = String.length s in

  let rec loop_left i =
    if i < len
    then if char_is_blank s.[i] then loop_left (i + 1) else i
    else i in

  let min = loop_left 0 in

  let rec loop_right i =
    if min < i
    then if char_is_blank s.[i] then loop_right (i - 1) else i
    else i in

  let max = loop_right (len - 1) in

  let len_sub = max - min + 1 in
  if len_sub = len then s else String.sub s min len_sub
;;

(** [get_tuple_items] returns the comma separated parts of a string. *)
let get_tuple_items s =
  let ib = Scanf.Scanning.from_string s in
  let rec loop accu =
    if Scanf.Scanning.end_of_input ib then List.rev accu else
    Scanf.bscanf ib " %s@, " (fun item ->
      loop (chomp item :: accu)) in
  loop []
;;

let case_conditions_ident =
  Ast_types.ident_from_string "CaseConditions"
and show_default_case_ident =
  Ast_types.ident_from_string "ShowDefaultCase"
and case_show_default_ident =
  Ast_types.ident_from_string "CaseShowDefault"
;;

let num_inputs_ident =
  Ast_types.ident_from_string "NumInputs"
and if_expression_ident =
  Ast_types.ident_from_string "IfExpression"
and else_if_expressions_ident =
  Ast_types.ident_from_string "ElseIfExpressions"
and show_else_ident =
  Ast_types.ident_from_string "ShowElse"
;;

let get_show_else_ident genr_info =
  get_ident_bool genr_info show_else_ident
;;

let get_show_default_case genr_info =
  if exists_argument genr_info show_default_case_ident
  then get_ident_bool genr_info show_default_case_ident
  else get_ident_bool genr_info case_show_default_ident
;;

let get_else_if_expressions genr_info =
  if exists_argument genr_info else_if_expressions_ident
  then get_ident_string genr_info else_if_expressions_ident
  else ""
;;

let get_nin genr_info =
  get_ident_int genr_info num_inputs_ident
;;

let _0_ = Unatural._0_
and _1_ = Unatural._1_
and _2_ = Unatural._2_
and _3_ = Unatural._3_
and _4_ = Unatural._4_
;;

let genr mdl_block_type block_id genr_info =

  let get_ident_string =
    Genr_utilities_scicos_pro.
    get_ident_string genr_info
  and sample_time =
    Genr_utilities_scicos_pro.
    get_sample_time genr_info
  and zcr =
    Genr_utilities_scicos_pro.
    get_zero_crossing_value genr_info in

  let (all_expressions, show_else, nin) =
    if mdl_block_type = mdl_block_type_if
    then
      let if_expression =
        get_ident_string if_expression_ident
      and else_if_expressions =
        get_else_if_expressions genr_info in

      let else_ifs = get_tuple_items else_if_expressions in
      let all_expressions = if_expression :: else_ifs in
      (List.map (fun ex -> Code_value (code_out_string ex)) all_expressions,
       get_show_else_ident genr_info,
       get_nin genr_info)
    else
      (* mdl_block_type_switch_case ->*)
      (get_items (get_ident_string case_conditions_ident),
       get_show_default_case genr_info,
       1) in

  let nexx = List.length all_expressions in

  let n_evout = nexx + 1
  and nout = if show_else then nexx + 1 else nexx in

  let expressions = (Si "exx", all_expressions), [] in

  let scicos_block_parameters ext = [
    Value_parameter (Si "nexx", Code_value (code_out_int nexx));
    Struct_parameter (Si "Exp", expressions);
    Value_parameter (Si "zc", Code_value zcr );
    Value_parameter (Si "externalActivation", Code_value ext);
  ] in

  let scicos_basic_block_instructions blk_id activation =
    genr_instructions_for
      mdl_block_type (scicos_block_parameters activation)
      [] blk_id genr_info in

  let nin_nat = Unatural.of_int nin
  and nout_nat = Unatural.of_int nout in

  (* Blocks and port names *)
  let mux_name = "Mux"
  and if_expressions_name = "if_expressions"
  and select_out_name = "select_out"
  and constant_name = "constant"
  and data_type_conversion_name = "type conversion"
  and sample_clock_name = "sample_clock" in

  let in_names, in_ports =
    Genr_utilities_scicos_pro.
    in_ports_in_scsm0 nin_nat
  and out_names, out_ports =
    out_ports_in_scsm0 _4_ nout_nat in

  let in_ports = Concat_list.list_of_clist in_ports
  and out_ports = Concat_list.list_of_clist out_ports in

  let mux =
    let sizes = (Si "size", list_of_ones_of_int nin), [] in
    let nports = (nin_nat, _1_) in
    Genr_utilities_scicos_pro.
    add_block_body_in_scsm0
      mux_name
      (genr_instructions_for_modeler_block_type
         (Sbt "system/Routing/Mux")
         [
           Value_parameter (Si "nin",
             Code_value (code_out_int nin));
           Struct_parameter (Si "inports", sizes);
         ]
         [] Genr_utils.blk0 genr_info)
      nports

  and select_out =
    let nports = (_1_, Unatural.of_int nexx) in
    Genr_utilities_scicos_pro.
    add_block_body_in_scsm0
      select_out_name
      (genr_instructions_for_modeler_block_type
         (Sbt "system/Routing/SelectOutput")
         [
           Value_parameter (Si "typ",
             Code_value double_type);
           Value_parameter (Si "nout",
             Code_value (code_out_int n_evout));
         ]
         [] Genr_utils.blk0 genr_info)
      nports

  and constant =
    let scicos_block_type = Sbt "system/SignalGenerators/Constant" in
    let nports = (_0_, _1_) in
    Genr_utilities_scicos_pro.
    add_block_body_in_scsm0 constant_name
      (genr_instructions_for_modeler_block_type scicos_block_type
         [
           Value_parameter (Si "C",
             Code_value (code_out_int 1));
           Value_parameter (Si "typ",
             Code_value (
               Genr_utilities_scicos_pro.double_type));
           Value_parameter (Si "externalActivation",
             Code_value
               Genr_utilities_scicos_pro.external_activation_mode_off);
         ] [] Genr_utils.blk0 genr_info)
      nports

  and data_type_conversion =
    let nports = (_1_, _1_) in
    Genr_utilities_scicos_pro.add_block_body_in_scsm0
      data_type_conversion_name
      (genr_instructions_for_modeler_block_type
        (Sbt "system/SignalConversions/DatatypeConversion")
        [
          Value_parameter (Si "it",
            Code_value inherit_mode_type);
          Value_parameter (Si "ot",
            Code_value double_type);
          Value_parameter (Si "satur",
            Simulink_arg_saturation_set saturation_mode_off);
          Value_parameter (Si "rmethod",
            Code_value floor_round_method);
          Value_parameter (Si "externalActivation",
            Code_value
              Genr_utilities_scicos_pro.external_activation_mode_off);
         ]
        [] Genr_utils.blk0 genr_info)
      nports

  and if_expressions =
    let ext =
      if sample_time = "-1"
      then external_activation_mode_off
      else external_activation_mode_on in
    let nports = (_1_, _0_) in
    Genr_utilities_scicos_pro.
    add_block_body_in_scsm0 if_expressions_name
      (scicos_basic_block_instructions blk0 ext)
      nports

  (* Sample clock is mandatory when sample time it not -1 *)
  and sample_clock =
    let sample_time = code_out_expression sample_time in
    Genr_utilities_scicos_pro.
    sample_clock_in_scsm0
      sample_clock_name sample_time in

  (* Really create required blocks before creating links;
     in particular a sample clock block is created if sample_time
     is not inherited (-1) *)
  let blocks =
    let blocks = [
      in_ports;
      out_ports;
      mux (_1_, _3_);
      data_type_conversion (_2_, _3_);
      select_out (_3_, _1_);
      constant (_1_, _1_);
      if_expressions (_3_, _3_);
    ] in

    if sample_time = "-1" then blocks else
    sample_clock (_2_, _4_) :: blocks in

  (* Links from input ports to Mux *)
  let in_links =
    Genr_utilities_scicos_pro.
    rev_map_i_start _1_
      (fun i in_port_name ->
       add_directed_link_in_scsm0
         in_port_name port_1_ mux_name
         (Genr_utilities_scicos_pro.mdl_port_number_of_natural i) [])
      in_names in

  (* Links from selector block ports to output ports *)
  let out_links =
    Genr_utilities_scicos_pro.
    rev_map_i_start _1_
      (fun i out_port_name ->
       add_directed_link_in_scsm0
         select_out_name
         (Genr_utilities_scicos_pro.mdl_port_number_of_natural i)
         out_port_name port_1_ [])
      (List.rev out_names) in

  (* Add a link from sample clock to If block when sample time it not -1 *)
  let add_sample_time_link code =
    if sample_time = "-1" then code else
    let sample_time_link =
      Genr_utilities_scicos_pro.
      add_event_link_in_scsm0
         sample_clock_name port_1_ if_expressions_name port_1_ [] in
    sample_time_link :: code in

  let links =
    Genr_utilities_scicos_pro.
    add_directed_link_in_scsm0
      data_type_conversion_name port_1_
      if_expressions_name port_1_ [] ::
    add_directed_link_in_scsm0
      mux_name port_1_ data_type_conversion_name port_1_ [] ::
    add_directed_link_in_scsm0
      constant_name port_1_ select_out_name port_1_ [] ::
    add_sample_time_link (in_links @ out_links) in

  (* Add event links *)
  let rec add_event_links links i =
    if i > 0
    then
      add_event_links (
        Genr_utilities_scicos_pro.
        add_event_link_in_scsm0
          if_expressions_name
          (Genr_utilities_scicos_pro.
           mdl_port_number_of_natural (Unatural.of_int i))
          select_out_name
          (Genr_utilities_scicos_pro.
           mdl_port_number_of_natural (Unatural.of_int i)) [] ::
        links)
        (i - 1)
    else links in

  let links = add_event_links links n_evout in

  let super_body = blocks @ [links]  in

  super_block_in_scsm0 block_id super_body
;;

let genr_if = genr mdl_block_type_if
;;

let genr_switch_case = genr mdl_block_type_switch_case
;;

add_translator mdl_block_type_if genr_if
;;

add_translator mdl_block_type_switch_case genr_switch_case
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
