(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_integer_delay.ml,v 1.1 2019-05-10 14:04:57 jpc Exp $ *)

(*open Unatural*)
module Genr_utilities_scicos = struct end;;

(** {3 Translating Simulink block Delay} *)

(**
  TODO
  This block should be translated into two distinct scicos blocks,
  according to the delay length parameter: if length is 1 we use
  the Scicos block for unit delay (DiscreteDelay), otherwise we use
  the Scicos block for multiple delay (ShiftRegister).
*)

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
open Unatural
;;

(* The Simulink block we translate. *)
let mdl_block_type_multiple_delay =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Integer Delay")
    (Ast_types.mdl_path_of_string "simulink/Discrete/Integer Delay")
;;

(* The particular case for unit delay, used in the translation. *)
let mdl_block_type_unit_delay =
  Mdl_basics_explicit_ast.
  Explicit_block_regular_type Mdl_block_types.UnitDelay
;;

(* The Simulink translator we should register. *)
let mdl_block_type = mdl_block_type_multiple_delay;;

let delay_length_ident =
  Ast_types.ident_from_string "NumDelays"
and initial_condition_ident =
  Ast_types.ident_from_string "vinit"
;;

let genr block_id genr_info =

  let get_ident_string =
    Genr_utilities_scicos_pro.get_ident_string genr_info in

  let sample_time =
    Genr_utilities_scicos_pro.
    get_sample_time genr_info in

  let initial_condition =
    get_ident_string initial_condition_ident in

  let delay_length =
    get_ident_string delay_length_ident in

  let block_type_and_parameters delay_length ext =
    if delay_length = "1"
    then
      mdl_block_type_unit_delay,
      [
        Value_parameter (Si "init_cond",
          Code_value (code_out_expression initial_condition));
        Value_parameter (Si "typ",
          Code_value (inherit_mode_type));
        Value_parameter (Si "externalActivation",
          Code_value ext)
      ]
    else
      mdl_block_type_multiple_delay,
      [
        Value_parameter (Si "K",
          Code_value (code_out_expression delay_length));
        Value_parameter (Si "z0",
          Code_value (code_out_expression initial_condition));
        Value_parameter (Si "type",
          Code_value (inherit_mode_type));
        Value_parameter (Si "externalActivation",
          Code_value ext)
      ] in

  let scicos_basic_block_instructions
      simulink_block_type scicos_block_parameters block_id =
    genr_instructions_for
      simulink_block_type scicos_block_parameters
      [] block_id genr_info in

  match sample_time with
  | "-1" ->
    let simulink_block_type, scicos_block_parameters =
      block_type_and_parameters
        delay_length external_activation_mode_off in
    Concat_list.
    list
      (scicos_basic_block_instructions
         simulink_block_type
         scicos_block_parameters block_id)

  | _ ->
    let in_port_name = "in"
    and out_port_name = "out"
    and dollar_name = "dollar"
    and sample_clock_name = "sample_clock" in

    let in_port =
      Genr_utilities_scicos_pro.
      in_port_in_scsm0
        in_port_name port_1_
        (_0_, _0_) in

    let out_port =
      Genr_utilities_scicos_pro.
      out_port_in_scsm0
        out_port_name port_1_
        (_2_, _0_) in

    let dollar_block =
      let simulink_block_type, scicos_block_parameters =
        block_type_and_parameters
          delay_length external_activation_mode_on in

      let dollar_ports = (_1_, _1_) in
      Genr_utilities_scicos_pro.
      add_block_body_in_scsm0
        dollar_name
        (scicos_basic_block_instructions
           simulink_block_type
           scicos_block_parameters blk0)
        dollar_ports
        (_1_, _0_) in

    let sample_clock =
      Genr_utilities_scicos_pro.
      sample_clock_in_scsm0
        sample_clock_name
        (code_out_expression sample_time)
        (_1_, _1_) in

    let links = [
      Genr_utilities_scicos_pro.
      add_directed_link_in_scsm0
        in_port_name port_1_ dollar_name port_1_ [];
      add_directed_link_in_scsm0
        dollar_name port_1_ out_port_name port_1_ [];
      Genr_utilities_scicos_pro.
      add_event_link_in_scsm0
        sample_clock_name port_1_ dollar_name port_1_ [];
    ] in

    let super_body = [
      in_port;
      out_port;
      dollar_block;
      sample_clock;
      links;
    ] in

    super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
