(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_integrator.ml,v 1.1 2019-05-10 14:04:57 jpc Exp $ *)

(*open Unatural*)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
open Unatural
;;

let mdl_block_type =
  Mdl_basics_explicit_ast.Explicit_block_regular_type
    Mdl_block_types.Integrator
;;

let external_reset_ident =
  Ast_types.ident_from_string "ExternalReset"
and initial_condition_source_ident =
  Ast_types.ident_from_string "InitialConditionSource"
and initial_condition_ident =
  Ast_types.ident_from_string "InitialCondition"
  (* Limit_output_ident is a boolean (on or off) in Simulink.
     In scios_pro this boolean should be translated as a boolean as well.
  *)
and limit_output_ident =
  Ast_types.ident_from_string "LimitOutput"
and upper_saturation_limit_ident =
  Ast_types.ident_from_string "UpperSaturationLimit"
and lower_saturation_limit_ident =
  Ast_types.ident_from_string "LowerSaturationLimit"
and show_state_port_ident =
  Ast_types.ident_from_string "ShowStatePort"
;;

let genr block_id genr_info =

  let external_reset =
    get_ident_string genr_info external_reset_ident
  and initial_condition_source =
    get_ident_string genr_info initial_condition_source_ident in

  let scicos_basic_block_instructions reinit blk_id =
    let scicos_block_parameters = [
      Value_parameter (Si "x0",
        Simulink_arg_expression initial_condition_ident);
      Value_parameter (Si "reinit",
        Code_value (code_out_bool reinit));
      Value_parameter (Si "satur",
        Simulink_arg_value limit_output_ident);
      Struct_parameter (Si "limit", (
        (Si "max",
           [Simulink_arg_expression upper_saturation_limit_ident]),
        [
         Si "min",
           [Simulink_arg_expression lower_saturation_limit_ident];
        ])
      );
    ] in

    genr_instructions_for
      mdl_block_type
      scicos_block_parameters
      [] blk_id genr_info in

  let last_input_block_instructions blk_id =
    let sci_block_type = Sbt "system/Hybrid/LastInput" in
    genr_instructions_for_modeler_block_type
      sci_block_type
      [
        Value_parameter (Si "externalActivation",
          Code_value external_activation_mode_off)
      ]
      [] blk_id genr_info in

  let show_saturation_port =
    Genr_utilities_scicos_pro.
    get_show_saturation_port genr_info in

  let show_state_port =
    get_ident_bool genr_info show_state_port_ident in

  match external_reset, initial_condition_source,
    show_saturation_port, show_state_port with
  | "none", "internal", false, false ->
    Concat_list.list (scicos_basic_block_instructions false block_id)

  | "none", "internal", false, true ->
    let in_port_name = "in"
    and out_port1_name = "out1"
    and out_port2_name = "out2"
    and split_name = "split"
    and integrator_name = "integrator"
    and last_input_name = "lastinput" in

    let in_port =
      Genr_utilities_scicos_pro.
      in_port_in_scsm0
        in_port_name port_1_
        (_0_, _1_) in
    let out_port1 =
      Genr_utilities_scicos_pro.
      out_port_in_scsm0
        out_port1_name port_1_
        (_4_, _1_) in
    let out_port2 =
      out_port_in_scsm0
        out_port2_name port_2_
        (_4_, _0_) in
    let integrator =
      let nports = (_1_, _1_) in
      Genr_utilities_scicos_pro.
      add_block_body_in_scsm0
        integrator_name
          (scicos_basic_block_instructions true blk0)
        nports
        (_1_, _0_) in

    let last_input =
      let nports = (_1_, _1_) in
      add_block_body_in_scsm0
        last_input_name
        (last_input_block_instructions blk0)
        nports
        (_3_, _0_) in

    let split =
      Genr_utilities_scicos_pro.
      split_in_scsm0
        split_name (_2_, _0_) in

    let links = [
      Genr_utilities_scicos_pro.
      add_directed_link_in_scsm0
        in_port_name port_1_ integrator_name port_1_ [];
      add_directed_link_in_scsm0
        integrator_name port_1_ split_name port_1_ [];
      add_directed_link_in_scsm0
        split_name port_1_ out_port1_name port_1_ [];
      add_directed_link_in_scsm0
        split_name port_2_ last_input_name port_1_ [];
      add_directed_link_in_scsm0
        last_input_name port_1_ out_port2_name port_1_ [];
    ] in

    let super_body = [
      in_port;
      out_port1;
      out_port2;
      integrator;
      split;
      last_input;
      links;
    ] in

    super_block_in_scsm0 block_id super_body

  | "none", "external", false, false ->
    let in_port1_name = "in1"
    and in_port2_name = "in2"
    and out_port_name = "out"
    and integrator_name = "integrator"
    and event_gen_name = "event_gen" in

    let in_port1 =
      in_port_in_scsm0
        in_port1_name port_1_
        (_0_, _1_) in
    let in_port2 =
      in_port_in_scsm0
        in_port2_name port_2_
        (_0_, _0_) in
    let out_port =
      out_port_in_scsm0
        out_port_name port_1_
        (_3_, _0_) in
    let integrator =
      let nports = (_2_, _1_) in
      add_block_body_in_scsm0
        integrator_name
          (scicos_basic_block_instructions true blk0)
        nports
        (_1_, _0_) in

    let event_gen =
      let nports = (_0_, _0_) in
      add_block_body_in_scsm0
        event_gen_name
        (let sci_block_type =
           Sbt "system/ActivationOperations/InitialEvent" in
         genr_instructions_for_modeler_block_type
           sci_block_type []
           [] blk0 genr_info)
        nports
        (_1_, _1_) in

    let links = [
      add_directed_link_in_scsm0
        in_port1_name port_1_ integrator_name port_1_ [];
      add_directed_link_in_scsm0
        in_port2_name port_1_ integrator_name port_2_ [];
      add_directed_link_in_scsm0
        integrator_name port_1_ out_port_name port_1_ [];
      add_event_link_in_scsm0
        event_gen_name port_1_ integrator_name port_1_ [];
    ] in

    let super_body = [
      in_port1;
      in_port2;
      out_port;
      integrator;
      event_gen;
      links;
    ] in

    super_block_in_scsm0 block_id super_body

  | "none", "external", false, true ->
    let in_port1_name = "in1"
    and in_port2_name = "in2"
    and out_port1_name = "out1"
    and out_port2_name = "out2"
    and split_name = "split"
    and integrator_name = "integrator"
    and event_gen_name = "event_gen"
    and last_input_name = "lastinput"  in

    let in_port1 =
      in_port_in_scsm0
        in_port1_name port_1_
        (_0_, _1_) in
    let in_port2 =
      in_port_in_scsm0
        in_port2_name port_2_
        (_0_, _0_) in
    let out_port1 =
      out_port_in_scsm0
        out_port1_name port_1_
        (_4_, _1_) in
    let out_port2 =
      out_port_in_scsm0
        out_port2_name port_2_
        (_4_, _0_) in
    let integrator =
      let nports = (_2_, _1_) in
      add_block_body_in_scsm0
        integrator_name
          (scicos_basic_block_instructions true blk0)
        nports
        (_1_, _0_) in
    let last_input =
      let nports = (_1_, _1_) in
      add_block_body_in_scsm0
        last_input_name
        (last_input_block_instructions blk0)
        nports
        (_3_, _0_) in

    let split =
      split_in_scsm0
        split_name (_2_, _0_) in

    let event_gen =
      let nports = (_0_, _0_) in
      Genr_utilities_scicos_pro.
      add_block_body_in_scsm0
        event_gen_name
        (let sci_block_type =
           Sbt "system/ActivationOperations/InitialEvent" in
         genr_instructions_for_modeler_block_type sci_block_type []
           [] blk0 genr_info)
        nports
        (_1_, _1_) in

    let links = [
      add_directed_link_in_scsm0
        in_port1_name port_1_ integrator_name port_1_ [];
      add_directed_link_in_scsm0
        in_port2_name port_1_ integrator_name port_2_ [];
      add_directed_link_in_scsm0
        integrator_name port_1_ split_name port_1_ [];
      add_directed_link_in_scsm0
        split_name port_1_ out_port1_name port_1_ [];
      add_directed_link_in_scsm0
        split_name port_2_ last_input_name port_1_ [];
      add_directed_link_in_scsm0
        last_input_name port_1_ out_port2_name port_1_ [];
      add_event_link_in_scsm0
        event_gen_name port_1_ integrator_name port_1_ [];
    ] in

    let super_body = [
      in_port1;
      in_port2;
      out_port1;
      out_port2;
      integrator;
      split;
      last_input;
      event_gen;
      links;
    ] in

    super_block_in_scsm0 block_id super_body

  | ("rising" | "falling" | "either" | "level" | "level hold"
     as external_reset), _,
    false, false ->
    Say.warning
      (Printf.sprintf
         "Integrator Simulink block translation: \
          %S configuration is not yet implemented"
         external_reset);
    Genr_scicos_pro_not_yet_translated.genr block_id genr_info
  | external_reset, _, _, _ ->
    Say.warning
      (Printf.sprintf
         "Integrator Simulink block translation: \
          unknown configuration %S"
         external_reset);
    Genr_scicos_pro_not_yet_translated.genr block_id genr_info
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
