(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_lookup2d.ml,v 1.1 2019-05-10 14:04:57 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
open Unatural
;;

let mdl_block_type =
  Mdl_basics_explicit_ast.Explicit_block_regular_type
    Mdl_block_types.Lookup2D
;;

let output_values_ident =
  Ast_types.ident_from_string "OutputValues"
and look_up_meth_ident =
  Ast_types.ident_from_string "LookUpMeth"
and table_ident =
  Ast_types.ident_from_string "Table"
and row_index_ident =
  Ast_types.ident_from_string "RowIndex"
and column_index_ident =
  Ast_types.ident_from_string "ColumnIndex"
;;

let rec get_method_extrapo = function
  | "Interpolation-Extrapolation" ->
    ("Bilinear interpolation", "Bilinear extrapolation")
  | "Interpolation-Use End Values" ->
    ("Bilinear interpolation", "Hold")
  | "Use Input Above" -> ("Point_just_above", "Hold")
  | "Use Input Below" -> ("Point_just_below", "Hold")
  | "Use Input Nearest" -> ("Nearest_point", "Hold")
  | s ->
    let extrapolation = "Interpolation-Extrapolation" in
    Say.warning (Printf.sprintf
      "Lookup2d translation: extrapolation method %s is not supported,\
     \nusing %s extrapolation method"
      s extrapolation);
    get_method_extrapo extrapolation
;;

let genr block_id genr_info =

  let get_ident_string =
    Genr_utilities_scicos_pro.get_ident_string genr_info in

  let sample_time =
    Genr_utilities_scicos_pro.get_sample_time genr_info in

  let table =
    if Genr_utilities_scicos_pro.exists_argument genr_info table_ident
    then get_ident_string table_ident
    else get_parameter_value_string genr_info output_values_ident
  and look_up_meth = get_ident_string look_up_meth_ident in

  let sci_method = fst (get_method_extrapo look_up_meth) in
  let sci_extrapo = code_out_string "Bilinear extrapolation" in

  let scicos_block_parameters ext = [
    Value_parameter (Si "xx",
      Simulink_arg_expression row_index_ident);
    Value_parameter (Si "yy",
      Simulink_arg_expression column_index_ident);
    Value_parameter (Si "zz",
      Code_value (code_out_expression table));
    Value_parameter (Si "Method",
      Code_value (code_out_string sci_method));
    Value_parameter (Si "DoOutside",
      Code_value sci_extrapo);
    Value_parameter (Si "EventGen",
      Code_value (code_out_int 0));
    Value_parameter (Si "externalActivation",
      Code_value ext);
  ] in

  let scicos_basic_block_instructions blk_id ext =
    genr_instructions_for
      mdl_block_type (scicos_block_parameters ext)
      [] blk_id genr_info in

  match sample_time with
  | "-1" ->
    Concat_list.list
      (scicos_basic_block_instructions
         block_id external_activation_mode_off)

  | _ ->

    let sample_time = code_out_expression sample_time in

    let in_port1_name = "in_port1"
    and in_port2_name = "in_port2"
    and out_port_name = "out_port"
    and lookup_name = "lookup"
    and sample_clock_name = "sample_clock" in

    let in_port1 =
      Genr_utilities_scicos_pro.
      in_port_in_scsm0
        in_port1_name port_1_
        (_0_, _0_) in
    let in_port2 =
      Genr_utilities_scicos_pro.
      in_port_in_scsm0
        in_port2_name port_2_
        (_0_, _1_) in
    let out_port =
      Genr_utilities_scicos_pro.
      out_port_in_scsm0
        out_port_name port_1_
        (_2_, _0_) in
    let lookup =
      let nports = (_2_, _1_) in
      Genr_utilities_scicos_pro.
      add_block_body_in_scsm0
        lookup_name
        (scicos_basic_block_instructions blk0 external_activation_mode_on)
        nports
        (_2_, _0_) in
    let sample_clock =
      Genr_utilities_scicos_pro.
      sample_clock_in_scsm0
        sample_clock_name sample_time
        (_1_, _3_) in

    let links = [
      Genr_utilities_scicos_pro.
      add_directed_link_in_scsm0
        in_port1_name port_1_ lookup_name port_1_ [];
      add_directed_link_in_scsm0
        in_port2_name port_1_ lookup_name port_2_ [];
      add_directed_link_in_scsm0
        lookup_name port_1_ out_port_name port_1_ [];
      Genr_utilities_scicos_pro.
      add_event_link_in_scsm0
        sample_clock_name port_1_ lookup_name port_1_ [];
    ] in

    let super_body = [
      in_port1;
      in_port2;
      out_port;
      lookup;
      sample_clock;
      links;
    ] in

  super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type genr;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
