(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_lookup_nd.ml,v 1.1 2019-05-10 14:04:57 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
open Unatural
;;

let mdl_block_type =
  Mdl_basics_explicit_ast.Explicit_block_regular_type
    Mdl_block_types.Lookup_n_D
;;

let number_of_dimensions_ident =
  Ast_types.ident_from_string "NumberOfTableDimensions"
and output_values_ident =
  Ast_types.ident_from_string "OutputValues"
and interp_method_ident =
  Ast_types.ident_from_string "InterpMethod"
and extrap_method_ident =
  Ast_types.ident_from_string "ExtrapMethod"
and table_ident =
  Ast_types.ident_from_string "Table"
and row_index_ident =
  Ast_types.ident_from_string "BreakpointsForDimension1"
and column_index_ident =
 Ast_types.ident_from_string "BreakpointsForDimension2"
;;

let warning_replaced_by_bilinear opt =
  Say.warning (Printf.sprintf
    "2D lookup cubic spline option %S replaced by bilinear." opt);
;;

let warning_only_1D_or_2D n =
  Say.warning (Printf.sprintf
    "Lookup table with dimension %d not supported yet" n)
;;

let rec get_intrapo dims opt =
  match opt with
  | "Flat" -> if dims = 2 then "Point_just_below" else "ZeroOrder(floor)"
  | "Nearest" -> if dims = 2 then "Nearest_point" else "ZeroOrder(nearest)"
  | "Linear" -> if dims = 2 then "Bilinear interpolation" else "Linear"
  | "Cubic spline"
  | _ when dims = 2 ->
    warning_replaced_by_bilinear opt;
    get_intrapo dims "Linear"
  | _ -> "Not_a_knot"
;;

let rec get_extrapo dims opt =
  match opt with
  | "Clip" -> "Hold"
  | "Nearest" -> "Hold"
  | "Linear" ->
    if dims = 2 then "Bilinear extrapolation" else "Extrapolation"
  | "Cubic spline"
  | _ when dims = 2 ->
    warning_replaced_by_bilinear opt;
    get_extrapo dims "Linear"
  | _ -> "extrapolation"
;;

let genr block_id genr_info =

  let get_ident_string =
    Genr_utilities_scicos_pro.get_ident_string genr_info in
  let get_ident_int =
    Genr_utilities_scicos_pro.get_ident_int genr_info in

  let sample_time =
    Genr_utilities_scicos_pro.get_sample_time genr_info in

  let dims = get_ident_int number_of_dimensions_ident in
  let table =
    if Genr_utilities_scicos_pro.exists_argument genr_info table_ident
    then get_ident_string table_ident
    else get_parameter_value_string genr_info output_values_ident
  and interp_method = get_ident_string interp_method_ident
  and extrap_method = get_ident_string extrap_method_ident
  and out_data_type =
    Genr_utilities_scicos_pro.
    get_simulink_out_data_type_value genr_info in

  let scicos_block_parameters ext = [
    Value_parameter (Si "xx",
      Simulink_arg_expression row_index_ident);
    Value_parameter (Si "yy",
      Simulink_arg_expression column_index_ident);
    Value_parameter (Si "zz",
      Code_value (code_out_expression table));
    Value_parameter (Si "Method",
      Code_value (code_out_string (get_intrapo 2 interp_method)));
    Value_parameter (Si "DoOutside",
      Code_value (code_out_string (get_extrapo 2 extrap_method)));
    Value_parameter (Si "EventGen",
      Code_value (code_out_int 0));
    Value_parameter (Si "externalActivation",
      Code_value ext);
  ] in

  let scicos_block_instructions blk_id ext =
    genr_instructions_for
      mdl_block_type (scicos_block_parameters ext)
      [] blk_id genr_info in

  let lookup_block_parameters ext = [
    Value_parameter (Si "xx",
      Simulink_arg_expression row_index_ident);
    Value_parameter (Si "yy",
      Code_value (code_out_expression table));
    Value_parameter (Si "Method",
      Code_value (code_out_string (get_intrapo 1 interp_method)));
    Value_parameter (Si "typ",
      Code_value out_data_type);
    Value_parameter (Si "EventGen",
      Code_value (code_out_int 0));
    Value_parameter (Si "externalActivation",
      Code_value ext);
    Value_parameter (Si "Leftside",
      Code_value (code_out_string (get_extrapo 1 extrap_method)));
    Value_parameter (Si "Rightside",
      Code_value (code_out_string (get_extrapo 1 extrap_method)));
  ] in

  let lookup_block_instructions blk_id ext =
    let sci_block_type = Sbt "system/LookupTables/LookupTable" in
      genr_instructions_for_modeler_block_type
        sci_block_type (lookup_block_parameters ext)
        [] blk_id genr_info in

  match sample_time with
  | "-1" ->
    begin match dims with
    | 1 ->
      Concat_list.list
        (lookup_block_instructions block_id
           external_activation_mode_off)
    | 2 ->
      Concat_list.list
        (scicos_block_instructions block_id
           external_activation_mode_off)
    | n ->
      warning_only_1D_or_2D n;
      Genr_scicos_pro_not_yet_translated.genr block_id genr_info
    end

  | _ ->

    let sample_time = code_out_expression sample_time in

    let in_port1_name = "in_port1"
    and in_port2_name = "in_port2"
    and out_port_name = "out_port"
    and lookup_name = "lookup"
    and sample_clock_name = "sample_clock" in

    let in_port1 =
      Genr_utilities_scicos_pro.
      in_port_in_scsm0
        in_port1_name port_1_
        (_0_, _0_) in
    let out_port =
      Genr_utilities_scicos_pro.
      out_port_in_scsm0
        out_port_name port_1_
        (_2_, _0_) in
    let sample_clock =
      Genr_utilities_scicos_pro.
      sample_clock_in_scsm0
        sample_clock_name sample_time
        (_1_, _3_) in

    match dims with
    | 2 ->
      let in_port2 =
        Genr_utilities_scicos_pro.
        in_port_in_scsm0
          in_port2_name port_2_
          (_0_, _1_) in
      let lookup =
        let nports = (_2_, _1_) in
        Genr_utilities_scicos_pro.
        add_block_body_in_scsm0
          lookup_name
          (scicos_block_instructions blk0 external_activation_mode_on)
          nports
          (_2_, _0_) in
      let links = [
        Genr_utilities_scicos_pro.
        add_directed_link_in_scsm0
          in_port1_name port_1_ lookup_name port_1_ [];
        add_directed_link_in_scsm0
          in_port2_name port_1_ lookup_name port_2_ [];
        add_directed_link_in_scsm0
          lookup_name port_1_ out_port_name port_1_ [];
        Genr_utilities_scicos_pro.
        add_event_link_in_scsm0
          sample_clock_name port_1_ lookup_name port_1_ [];
      ] in

      let super_body = [
        in_port1;
        in_port2;
        out_port;
        lookup;
        sample_clock;
        links;
      ] in

      super_block_in_scsm0 block_id super_body

    | 1 ->
      let lookup =
        let nports = (_2_, _1_) in
        Genr_utilities_scicos_pro.
        add_block_body_in_scsm0
          lookup_name
          (lookup_block_instructions blk0 external_activation_mode_on)
          nports
          (_2_, _0_) in
      let links = [
        add_directed_link_in_scsm0
          in_port1_name port_1_ lookup_name port_1_ [];
        add_directed_link_in_scsm0
          lookup_name port_1_ out_port_name port_1_ [];
        add_event_link_in_scsm0
          sample_clock_name port_1_ lookup_name port_1_ [];
      ] in

      let super_body = [
        in_port1;
        out_port;
        lookup;
        sample_clock;
        links;
      ] in

      super_block_in_scsm0 block_id super_body

    | dims ->
      warning_only_1D_or_2D dims;
      Genr_scicos_pro_not_yet_translated.genr block_id genr_info
;;

add_translator mdl_block_type genr;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
