(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_math.ml,v 1.1 2019-05-10 14:04:57 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
open Unatural
;;

let mdl_block_type =
  Mdl_basics_explicit_ast.
  Explicit_block_regular_type Mdl_block_types.Math
;;

let scicos_block_type =
  Genr_utilities_scicos_pro.modeler_block_type_for mdl_block_type
;;

let operator_ident = Ast_types.ident_from_string "Operator"
;;

let mod_block_type = Sbt "system/MathOperations/Modulo"
and transpose_block_type = Sbt "system/MatrixOperations/Transpose"
and pow_block_type = Sbt "system/MathOperations/Pow"
and hypot_block_type = Sbt "system/MathOperations/Hypot"
;;

let get_block_type_params operator =
  match operator with
  | "exp" -> scicos_block_type, "func", "exp"
  | "log" -> scicos_block_type, "func", "log"
  | "10^u" -> scicos_block_type, "func", "10^u"
  | "log10" -> scicos_block_type, "func", "log10"
  | "magnitude^2" -> scicos_block_type, "func", "|u|^2"
  | "square" -> scicos_block_type, "func", "u^2"
  | "sqrt" -> scicos_block_type, "func", "sqrt"
  | "reciprocal" -> scicos_block_type, "func", "1/u"
  | "conj" ->  scicos_block_type, "func", "conj"
  | "pow" -> pow_block_type, "none", "none"
  | "hypot" -> hypot_block_type, "none", "none"
  | "transpose" -> transpose_block_type, "rule", "Transpose"
  | "hermitian" ->
    transpose_block_type, "rule", "Hermitian(Transpose-Conjugate)"
  | "rem" -> mod_block_type, "op", "Modulo"
  | "mod" -> mod_block_type, "op", "Remainder"
  | _ ->
    Say.fatal_error (Printf.sprintf
      "Math operator %s is not supported" operator)
;;

let genr block_id genr_info =

  let sample_time =
    Genr_utilities_scicos_pro.get_sample_time genr_info in

  let get_ident_string =
    Genr_utilities_scicos_pro.get_ident_string genr_info in

  let operator = get_ident_string operator_ident in

  let scicos_block_type, param_name, param_value =
    get_block_type_params operator in

  let scicos_block_parameters ext =
    if param_name = "none" && param_value = "none"
    then
      [
        Value_parameter
          (Si "externalActivation",
           Code_value ext);
      ]
    else
      [
        Value_parameter
          (Si param_name,
           Code_value (code_out_string param_value));
        Value_parameter
          (Si "externalActivation",
           Code_value ext);
      ] in

  let scicos_basic_block_instructions blk_id ext =
    genr_instructions_for_modeler_block_type scicos_block_type
      (scicos_block_parameters ext)
      [] blk_id genr_info in


  match sample_time with
  | "-1"->
    Concat_list.list
      (scicos_basic_block_instructions
         block_id external_activation_mode_off)
  | _ ->

    let sample_time = code_out_expression sample_time in

    let block_name = "math" in
    let in_port_name = "in"
    and out_port_name = "out"
    and sample_clock_name = "sample_clock" in

    let in_port =
      Genr_utilities_scicos_pro.
      in_port_in_scsm0
        in_port_name port_1_ (_0_, _0_) in
    let out_port =
      Genr_utilities_scicos_pro.
      out_port_in_scsm0
        out_port_name port_1_ (_2_, _0_) in
    let sample_clock =
      Genr_utilities_scicos_pro.
      sample_clock_in_scsm0
        sample_clock_name sample_time (_1_, _1_) in

    let super_body =

      match operator with
      | "mod" | "rem" | "pow" | "hypot" ->

        let in_port2_name = "in" in
        let in_port2 =
          Genr_utilities_scicos_pro.
          in_port_in_scsm0
            in_port2_name port_2_ (_0_, _1_) in
        let block =
          let nports = (_2_, _1_) in
          Genr_utilities_scicos_pro.
          add_block_body_in_scsm0
            block_name
            (scicos_basic_block_instructions
               blk0 external_activation_mode_on)
            nports
            (_1_, _0_) in

        let links = [
          Genr_utilities_scicos_pro.
          add_directed_link_in_scsm0
            in_port_name port_1_ block_name port_1_ [];
          Genr_utilities_scicos_pro.
          add_directed_link_in_scsm0
            in_port2_name port_1_ block_name port_2_ [];
          add_directed_link_in_scsm0
            block_name port_1_ out_port_name port_1_ [];
          Genr_utilities_scicos_pro.
          add_event_link_in_scsm0
            sample_clock_name port_1_ block_name port_1_ [];
        ] in

        [
          in_port;
          in_port2;
          out_port;
          block;
          sample_clock;
          links;
        ]

      | _ ->

        let block =
          let nports = (_1_, _1_) in
          add_block_body_in_scsm0
            block_name
            (scicos_basic_block_instructions
               blk0 external_activation_mode_on)
            nports
            (_1_, _0_) in

        let links = [
          add_directed_link_in_scsm0
            in_port_name port_1_ block_name port_1_ [];
          add_directed_link_in_scsm0
            block_name port_1_ out_port_name port_1_ [];
          add_event_link_in_scsm0
            sample_clock_name port_1_ block_name port_1_ [];
        ] in

        [
          in_port;
          out_port;
          block;
          sample_clock;
          links;
        ] in

    super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type genr;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
