(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_matlab_fcn.ml,v 1.1 2019-05-10 14:04:57 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
open Unatural
;;

let mdl_block_type =
  Mdl_basics_explicit_ast.
  Explicit_block_regular_type Mdl_block_types.MATLABFcn
;;

let matlab_fcn_ident =
  Ast_types.ident_from_string "MATLABFcn"
;;

let output_dimensions_type_ident =
  Ast_types.ident_from_string "OutputDimensions"
;;

let add_arg_if_needed exx =
  Scanf.sscanf exx "%_1[A-Za-z]%_[A-Za-z_0-9]%[\000-\255]"
    (fun ending ->
     match ending with
     | "" -> exx ^ "(u)"
     | _ -> exx)
;;

let genr block_id genr_info =

  let get_ident_string =
    Genr_utilities_scicos_pro.
    get_ident_string genr_info in

  let out_size =
    get_ident_string output_dimensions_type_ident in
  let out_size =
    code_out_expression (Printf.sprintf "[ %s, 1]" out_size) in

  let exx =
    get_ident_string matlab_fcn_ident in

  let exx = add_arg_if_needed (String.trim exx) in

  let txt = Ast_types.mdl_text_of_string exx in

  let string_token =
    let txt =
       Ast_types.string_of_mdl_text (
         M_to_host.translate_expression txt) in
    Lib_strings.suppress_spaces txt in

  let sample_time =
    Genr_utilities_scicos_pro.
    get_sample_time genr_info in

  let scicos_block_parameters activation = [
    Value_parameter (Si "nin",
      Code_value (code_out_int 1));
    Value_parameter (Si "exx",
      Code_value (code_out_string string_token));
    Value_parameter (Si "outsz",
      Code_value out_size);
    Value_parameter (Si "outtp",
      Code_value (code_out_bool false));
    Value_parameter (Si "externalActivation",
      Code_value activation);
  ] in

  let scicos_basic_block_instructions blk_id activation =
    genr_instructions_for
      mdl_block_type (scicos_block_parameters activation)
      [] blk_id genr_info in

  match sample_time with
  | "-1" ->
    Concat_list.list
      (scicos_basic_block_instructions block_id
         Genr_utilities_scicos_pro.external_activation_mode_off)

  | _ ->

    let sample_time = code_out_expression sample_time in

    let in_port_name = "in"
    and out_port_name = "out"
    and hml_name = "hml"
    and sample_clock_name = "sample_clock" in

    let in_port =
      Genr_utilities_scicos_pro.
      in_port_in_scsm0 in_port_name port_1_
        (_0_, _0_) in
    let out_port =
      Genr_utilities_scicos_pro.
      out_port_in_scsm0 out_port_name port_1_
        (_2_, _0_) in
    let hml =
      let nports = (_1_, _1_) in
      Genr_utilities_scicos_pro.
      add_block_body_in_scsm0 hml_name
        (scicos_basic_block_instructions blk0
           Genr_utilities_scicos_pro.external_activation_mode_on)
        nports
        (_1_, _0_) in
    let sample_clock =
      Genr_utilities_scicos_pro.
      sample_clock_in_scsm0
        sample_clock_name sample_time
        (_1_, _1_) in

    let links = [
      Genr_utilities_scicos_pro.
      add_directed_link_in_scsm0
        in_port_name port_1_ hml_name port_1_ [];
      add_directed_link_in_scsm0 hml_name port_1_ out_port_name port_1_ [];
      Genr_utilities_scicos_pro.
      add_event_link_in_scsm0
        sample_clock_name port_1_ hml_name port_1_ [];
    ] in

    let super_body = [
      in_port;
      sample_clock;
      hml;
      out_port;
      links;
    ] in

    super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
