(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_matlab_function.ml,v 1.1 2019-05-10 14:04:57 jpc Exp $ *)

(*
   Compiling the ambedded Matlab block to Activate custom block.
*)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
;;

let mdl_block_type =
  Mdl_fake_block_types.matlab_function_block_type
;;

(* Name of Api functions are different for Activate/Oml and Activate/Hml. *)
type host_activate_api =
  {
    host_language_prefix : string;
    output_flag : string;
    get_input_data : string;
    set_output_data : string;
    get_output_data_type : string;
    get_opar_data : string;
  }
;;

let oml_activate_api =
  {
    host_language_prefix = "Oml";
    output_flag = "vssBlockOutputUpdateFlag";
    get_input_data = "vssGetInputData";
    set_output_data = "vssSetOutputData";
    get_output_data_type = "vssGetOutputDataType";
    get_opar_data = "vssGetOparData";
  };;

let hml_activate_api =
  {
    host_language_prefix = "Hml";
    output_flag = "hwscpEngineBlock.OUTPUT_UPDATE";
    get_input_data = "vssGetInPortData";
    set_output_data = "vssSetOutPortData";
    get_output_data_type = "vssGetOutputDataType";
    get_opar_data = "vssGetOparData";
  }
;;

let get_activate_api () =
  match Configuration.get_simulink_target_language () with
  | Simport_configuration.Oml -> oml_activate_api
  | Simport_configuration.Hypermath -> hml_activate_api
  | Simport_configuration.ScicosLab
  | Simport_configuration.Nsp
  | Simport_configuration.Modelica
  | Simport_configuration.Octave ->
    Say.fatal_error
      "No Nsp, Octave, Modelica, nor Scilab \
       translation for Matlab function blocks"
;;

let matlab_function_name_ident = Mdl_rename.matlab_function_name_ident
and matlab_function_script_ident = Mdl_rename.matlab_function_script_ident
and block_input_offsets_ident = Mdl_rename.block_input_offsets_ident
and block_parameter_offsets_ident = Mdl_rename.block_parameter_offsets_ident
and block_argument_sizes_ident = Mdl_rename.block_argument_sizes_ident
and block_return_sizes_ident = Mdl_rename.block_return_sizes_ident
and parameters_ident =
  Ast_types.ident_from_string "Parameters"
;;

let genr block_id genr_info =
  let get_ident_string =
    Genr_utilities_scicos_pro.get_ident_string genr_info
  and get_ident_int_vector =
    Genr_utilities_scicos_pro.get_ident_int_vector genr_info
  and get_ident_code =
    Genr_utilities_scicos_pro.get_argument_value_code genr_info
  and get_ident_expression_vector =
    Genr_utilities_scicos_pro.get_argument_value_expression_vector
      genr_info in

  let activate_api = get_activate_api () in

  let function_name = get_ident_string matlab_function_name_ident in

  let function_script_value = get_ident_code matlab_function_script_ident in

  let input_offsets = get_ident_int_vector block_input_offsets_ident
  and parameter_offsets =
    get_ident_int_vector block_parameter_offsets_ident
  and argument_sizes =
    get_ident_expression_vector block_argument_sizes_ident
  and return_sizes =
    get_ident_expression_vector block_return_sizes_ident in

  let opar_expression =
    if exists_argument genr_info parameters_ident then
      let parameters = get_ident_string parameters_ident in
      Printf.sprintf "{%s}" parameters
    else
      "{}" in

  let nin, nout = Genr_utilities_common.get_nin_nout genr_info in
  let nparams = Array.length parameter_offsets in
  let ninputs = Array.length input_offsets in

  let body_in =
    let b = Buffer.create 256 in
    let rec loop i =
      if i <= Unatural.to_int nin then
        let line =
          Printf.sprintf "\
            u%d = %s(block, %d);\n"
            i activate_api.get_input_data i in
        Buffer.add_string b line;
        loop (succ i) in
    loop 1;
    Buffer.contents b in

  let body_out =
    let b = Buffer.create 256 in
    let rec loop i =
      if i <= Unatural.to_int nout then
        let line =
          Printf.sprintf "\
            %s(block,%d,y%d,%s(block,%d));\n"
            activate_api.set_output_data i
            i activate_api.get_output_data_type i in
        Buffer.add_string b line;
        loop (succ i) in
    loop 2;
    Buffer.contents b in

  let body_params =
    let b = Buffer.create 256 in
    let rec loop i =
      if i <= nparams then
      let line =
        Printf.sprintf "\
          p%d = %s(block, %d);\n"
          i activate_api.get_opar_data i in
      Buffer.add_string b line;
      loop (succ i) in
    loop 1;
    Buffer.contents b in

  let body_call =
    let arg_out_opt =
      if Unatural.ge Unatural._1_ nout then None else
      let b = Buffer.create 256 in
      let rec loop i =
        let line = Printf.sprintf "y%d" i in
        Buffer.add_string b line;
        if i < Unatural.to_int nout then
          begin
            Buffer.add_char b ',';
            loop (succ i)
          end in
      loop 2;
      Some (Buffer.contents b) in

    let arg_in =
      let v = Array.make (ninputs + nparams) "" in
      Array.iteri (fun i offset ->
        v.(offset) <- Printf.sprintf "u%d" (succ i))
        input_offsets;
      Array.iteri (fun i offset ->
        v.(offset) <- Printf.sprintf "p%d" (succ i))
        parameter_offsets;
      String.concat ", " (Array.to_list v) in
    match arg_out_opt with
    | None ->
      Printf.sprintf "%s(%s)" function_name arg_in
    | Some arg_out ->
      Printf.sprintf "[%s]=%s(%s)"
        arg_out function_name arg_in in

  let translate_script_value function_script_string =
    Printf.sprintf "\
      function %sBlockFunction (block, flag)\
    \n  %s\
    \n  %s\
    \n  %s\
    \n  if flag == %s\
    \n    %s\
    \n    %s\
    \n  end\
    \nend"
      activate_api.host_language_prefix
      function_script_string
      body_in body_params
      activate_api.output_flag
      body_call body_out in

  let function_script_value =
    rewrite_script
      translate_script_value matlab_function_script_ident
      function_script_value in

  let opar_values =
    Array.to_list (
      Array.init nparams
        (fun i ->
         Code_value (code_out_expression (
           Printf.sprintf "get_index(%s,%d)"
             opar_expression (i + 1))))) in

  let translate_port_size =
    let mk_var =
      let indx = ref 0 in
      fun () -> decr indx; string_of_int !indx in
    fun idx size_string ->
      match size_string with
      | "-1" -> mk_var ()
      | expr_string ->
        Printf.sprintf "get_index(%s,%d)" expr_string idx in

  let compute_port_size ident sz =
    let rewrite idx =
      Code_value (
        Genr_utilities_common.
        rewrite_expression (translate_port_size idx) ident sz
      ) in
    let x = rewrite 1 in
    let y = rewrite 2 in
    x, y in

  let dummy_out_port_size = Code_value (code_out_int 1) in

  let get_output_sizes ident (return_sizes: code_value array) =
    let outrows, outcols =
      Array.fold_left
        (fun (rows, cols) sz ->
         let outrow, outcol = compute_port_size ident sz in
         outrow :: rows, outcol :: cols)
        ([dummy_out_port_size], [dummy_out_port_size]) return_sizes in
    List.rev outrows, List.rev outcols

  and get_input_sizes ident argument_sizes input_offsets =
    if Array.length input_offsets = 0 then
      let mk_generic_size () =
        compute_port_size ident (code_out_expression "-1") in
      let rec mk_generic_sizes rows cols = function
        | 0 -> rows, cols
        | k ->
          let inrow, incol = mk_generic_size () in
          mk_generic_sizes (inrow :: rows) (incol :: cols) (pred k) in
      mk_generic_sizes [] [] (Unatural_types.of_natural nin)
    else
      let inrows, incols =
        Array.fold_left
          (fun (rows, cols) offset ->
           let inrow, incol =
             compute_port_size ident argument_sizes.(offset) in
           inrow :: rows, incol :: cols)
          ([], []) input_offsets in
      List.rev inrows, List.rev incols in

  let inrow, incol =
    get_input_sizes block_argument_sizes_ident argument_sizes input_offsets in
  let outrow, outcol =
    get_output_sizes block_return_sizes_ident return_sizes in

  let scicos_block_parameters = [

    Genr_types.
    Value_parameter (Instructions.Si "nin",
      Genr_types.Code_value (code_out_natural nin)
    );
    Struct_parameter (Si "inports",
      ((Si "inrow", inrow),
       [ Si "incol", incol ])
    );
    Value_parameter (Si "nout",
      Code_value (code_out_natural nout)
    );
    Struct_parameter (Si "outports",
      ((Si "outrow", outrow),
       [ Si "outcol", outcol ])
    );
    Value_parameter (Si "nopar",
      Code_value (code_out_int nparams)
    );
    Struct_parameter (Si "oparparams",
      ((Si "oparvalue", opar_values),
       [ ])
    );
    Value_parameter (Si "code",
      Code_value function_script_value
    );
  ] in

  let scicos_basic_block_instructions blk_id =
    genr_instructions_for
      mdl_block_type scicos_block_parameters
      [] blk_id genr_info in

  Concat_list.list (scicos_basic_block_instructions block_id)
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
