(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_mechanical_ideal_rotational_sensor.ml,v 1.1 2019-05-10 14:04:58 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
open Unatural
;;

let mdl_block_path =
  "fl_lib/Mechanical/Mechanical Sensors/Ideal Rotational\nMotion Sensor"
;;

let mdl_block_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string
       "Ideal Rotational\nMotion Sensor")
    (Ast_types.mdl_path_of_string mdl_block_path)
;;

let offset_ident = Ast_types.ident_from_string "offset"
and port_l1_ = Genr_utilities_scicos_pro.port_l1_ Unatural._0_ _0_
and port_r1_ = Genr_utilities_scicos_pro.port_r1_ _0_ _0_
and port_r2_ = Genr_utilities_scicos_pro.port_r2_ _0_ _0_
;;

let genr block_id genr_info =

  let fixity_value, offset_value =
    if Genr_utilities_scicos_pro.
       exists_parameter genr_info offset_ident
    then
      Code_value (code_out_int 1),
      Simulink_arg_expression offset_ident
    else
      Code_value (code_out_int 0),
      Code_value (code_out_int 0) in

  let scicos_block_parameters = [] in

  let scicos_basic_block_instructions blk_id =
    genr_instructions_for
      mdl_block_type scicos_block_parameters
      [] blk_id genr_info in

  let in_port_name = "in"
  and rel_speed_sensor_name = "RelSpeedSensor"
  and rel_angle_sensor_name = "RelAngleSensor"
  and out_port1_name = "out1"
  and out_port2_name = "out2"
  and out_port3_name = "out3"
  and split1_name = "split1"
  and split2_name = "split2" in

  let in_port =
    Genr_utilities_scicos_pro.
    implicit_in_port_in_scsm0 in_port_name port_1_
    (_0_, _0_) in

  let split1 =
    Genr_utilities_scicos_pro.
    implicit_split_in_scsm0 split1_name (_1_, _0_) in

  let split2 =
    Genr_utilities_scicos_pro.
    implicit_split_in_scsm0 split2_name (_4_, _2_) in

  let rel_speed_sensor =
   let nports = (_1_, _2_) in
   Genr_utilities_scicos_pro.
   add_block_body_in_scsm0
     rel_speed_sensor_name
     (scicos_basic_block_instructions blk0)
     nports
     (_2_, _1_) in

  let rel_angle_sensor =
    let nports = (_1_, _2_) in
    let angle_sensor_body =
      genr_instructions_for_modeler_block_type
        (Sbt "Modelica 3.2.1/Mechanics/Rotational/Sensors/RelAngleSensor")
        [
          Struct_parameter (Si "phi_rel",
          ((Si "fixed", [ fixity_value; ]),
           [ Si "start", [ offset_value; ]]));
            ] [] blk0 genr_info in
    add_block_body_in_scsm0
      rel_angle_sensor_name
      angle_sensor_body
      nports
      (_2_, _0_) in

  let out_port1 =
    Genr_utilities_scicos_pro.
    implicit_out_port_in_scsm0 out_port1_name port_1_
    (_5_, _2_) in

  let out_port2 =
    Genr_utilities_scicos_pro.
    implicit_out_port_in_scsm0 out_port2_name port_2_
    (_5_, _1_) in

  let out_port3 =
    Genr_utilities_scicos_pro.
    implicit_out_port_in_scsm0 out_port3_name port_3_
    (_5_, _0_) in

  let links = [
    Genr_utilities_scicos_pro.
    add_connection_link_in_scsm0
      in_port_name port_r1_ split1_name port_l1_ [];
    Genr_utilities_scicos_pro.
    add_connection_link_in_scsm0
      split1_name port_r1_ rel_speed_sensor_name port_l1_ [];
    Genr_utilities_scicos_pro.
    add_connection_link_in_scsm0
      split1_name port_r2_  rel_angle_sensor_name port_l1_ [];
    Genr_utilities_scicos_pro.
    add_connection_link_in_scsm0
      rel_speed_sensor_name port_r1_ split2_name port_r1_ [];
    Genr_utilities_scicos_pro.
    add_connection_link_in_scsm0
      rel_angle_sensor_name port_r1_ split2_name port_r2_ [];
    Genr_utilities_scicos_pro.
    add_connection_link_in_scsm0
      rel_speed_sensor_name port_r2_ out_port2_name port_l1_ [];
    Genr_utilities_scicos_pro.
    add_connection_link_in_scsm0
      rel_angle_sensor_name port_r2_ out_port3_name port_l1_ [];
    Genr_utilities_scicos_pro.
    add_connection_link_in_scsm0
      split2_name port_l1_ out_port1_name port_l1_ [];
  ] in

  let super_body = [
    in_port;
    out_port1;
    out_port2;
    out_port3;
    split1;
    split2;
    rel_angle_sensor;
    rel_speed_sensor;
    links;
  ] in

  super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
