(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_min_max.ml,v 1.1 2019-05-10 14:04:58 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
open Unatural
;;

let mdl_block_type =
  Mdl_basics_explicit_ast.Explicit_block_regular_type
    Mdl_block_types.MinMax
;;

let inputs_ident =
  Ast_types.ident_from_string "Inputs"
and function_ident =
  Ast_types.ident_from_string "Function"
;;

let get_mm_option = function
  | "min" -> Some "Min"
  | "max" -> Some "Max"
  | s ->
    Say.warning (
      Printf.sprintf
        "MinMax block with Function %s not yet supported" s);
    None
;;

let genr block_id genr_info =

  let get_ident_string =
    Genr_utilities_scicos_pro.get_ident_string genr_info in

  let sample_time =
    Genr_utilities_scicos_pro.get_sample_time genr_info in

  let zcr =
    Genr_utilities_scicos_pro.get_zero_crossing_value genr_info
  and nin =
    Genr_utilities_scicos_pro.get_ident_int genr_info inputs_ident in

  let fun_name = get_ident_string function_ident in

  match get_mm_option fun_name with
  | None ->
    Genr_scicos_pro_not_yet_translated.genr block_id genr_info
  | Some mm ->

  let scicos_block_parameters ext = [
    Value_parameter (Si "mm", Code_value (code_out_string mm));
    Value_parameter (Si "nin", Code_value (code_out_int nin));
    Value_parameter (Si "zc", Code_value zcr);
    Value_parameter (Si "externalActivation", Code_value ext);
  ] in

  let scicos_basic_block_instructions blk_id ext =
    genr_instructions_for
      mdl_block_type (scicos_block_parameters ext)
      [] blk_id genr_info in

  let nin_nat = Unatural.of_int nin in

  match sample_time with
  | "-1" ->
    Concat_list.list
      (scicos_basic_block_instructions block_id
         external_activation_mode_off)

  | _ ->
    let sample_time = code_out_expression sample_time in

    let sample_clock_name = "sample_clock" in
    let out_port_name = "out"
    and max_min_name = "max_min" in

    let sample_clock =
      Genr_utilities_scicos_pro.
      sample_clock_in_scsm0
        sample_clock_name sample_time
        (Unatural._1_, _1_) in

    let out_port =
      Genr_utilities_scicos_pro.
      out_port_in_scsm0
        out_port_name port_1_
        (Unatural._2_, Unatural._0_) in

    let max_min =
      let nports = (nin_nat, _1_) in
      Genr_utilities_scicos_pro.
      add_block_body_in_scsm0 max_min_name
        (scicos_basic_block_instructions blk0 external_activation_mode_on)
        nports
        (_1_, _0_) in

    let in_names, in_ports =
      Genr_utilities_scicos_pro.in_ports_in_scsm0 nin_nat in

    let in_links =
      Genr_utilities_scicos_pro.
      rev_map_i_start _1_
        (fun i in_port_name ->
         add_directed_link_in_scsm0
           in_port_name port_1_ max_min_name
           (Genr_utilities_scicos_pro.mdl_port_number_of_natural i) [])
        in_names in

    let links = [
      Genr_utilities_scicos_pro.
      add_directed_link_in_scsm0
        max_min_name port_1_ out_port_name port_1_ [];
      Genr_utilities_scicos_pro.add_event_link_in_scsm0
        sample_clock_name port_1_  max_min_name port_1_ [];
    ] in

    let in_ports = Concat_list.list_of_clist in_ports in

    let super_body = [
      in_ports;
      sample_clock;
      out_port;
      max_min;
      in_links;
      links;
    ] in

    super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
