(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_multi_plot_graph.ml,v 1.1 2019-05-10 14:04:58 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
;;

let mdl_block_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string
       "MultiPlot Graph")
    (Ast_types.mdl_path_of_string
       "RCTblocks/MultiPlot Graph")
;;

let y_min_ident =
  Ast_types.ident_from_string "ymin"
and y_max_ident =
  Ast_types.ident_from_string "ymax"
and time_range_ident =
  Ast_types.ident_from_string "tmax"
and title_ident =
  Ast_types.ident_from_string "titleString"
;;

let get_min_max_bounds bounds_min bounds_max =
  let get_bounds bounds_string =
    let ib = Scanf.Scanning.from_string bounds_string in
    let rec loop accu =
      if Scanf.Scanning.end_of_input ib then List.rev accu else
      try
        Scanf.bscanf ib "%s@~"
          (fun s -> loop (s :: accu)) with
      | Scanf.Scan_failure s
      | Failure s ->
        Say.fatal_error
          (Printf.sprintf
             "Genr_scicos_pro_scope: \
              while scanning ``%s'', encountered ``%s''"
             bounds_string s) in
    loop [] in

  (get_bounds bounds_min, get_bounds bounds_max)
;;

let row_concat a b =
  Code_value (code_out_expression (Printf.sprintf "[ %s; %s ]" a b))
;;

let genr block_id genr_info =

  let get_ident_string =
    Genr_utilities_scicos_pro.get_ident_string genr_info
  and exists_argument =
    Genr_utilities_scicos_pro.exists_argument genr_info in

  let y_bounds =
    let y_mins, y_maxs =
      let bounds_min = get_ident_string y_min_ident
      and bounds_max = get_ident_string y_max_ident in
      get_min_max_bounds bounds_min bounds_max in

    List.map2 row_concat y_mins y_maxs in

  let axes_titles =
    Code_value (code_out_string (get_ident_string title_ident)) in

  let time_range =
    if exists_argument time_range_ident
    then get_ident_string time_range_ident
    else "20"


  and block_name =
    genr_info.Genr_block_types.block_explicit_name in

  let sample_time =
    Genr_utilities_scicos_pro.
    get_sample_time genr_info in

  let refresh =
    [ Code_value (code_out_expression time_range) ] in

  let in_names, in_ports =
    Genr_utilities_scicos_pro.
    in_ports_in_scsm0 Unatural._1_ in

  let scicos_block_parameters activation = [
    Value_parameter (Si "nin",
      Code_value (code_out_int 1));
    Value_parameter (Si "externalActivation",
      Code_value activation);
    Value_parameter (Si "useSubplot",
      Code_value (code_out_bool false));
    Value_parameter (Si "winNam",
      Code_value (code_out_string (
        Ast_types.string_of_mdl_name block_name)));
    Struct_parameter (Si "win_s",
       ((Si "xRefresh", refresh),
        [ Si "yInterval", y_bounds;
          Si "title", [axes_titles] ]));
  ] in

  let scicos_basic_block_instructions blk_id activation =
    genr_instructions_for
      mdl_block_type (scicos_block_parameters activation)
      [] blk_id genr_info in

  match sample_time with
  | "0" | "-1" ->
    Concat_list.list
      (scicos_basic_block_instructions block_id
         external_activation_mode_off)
  | _ ->

    let sample_clock_name = "sample_clock"
    and scope_name = "scope" in

    let sample_clock =
      Genr_utilities_scicos_pro.
      sample_clock_in_scsm0
        sample_clock_name
        (code_out_expression sample_time)
        (Unatural._1_, Unatural._1_) in

    let scope =
      let nports = (Unatural._1_, Unatural._0_) in
      let scope_body =
        scicos_basic_block_instructions blk0 external_activation_mode_on in
      Genr_utilities_scicos_pro.
      add_block_body_in_scsm0
        scope_name
        scope_body
        nports
        (Unatural._1_, Unatural._0_) in

    let in_links =
      Genr_utilities_scicos_pro.
      rev_map_i_start Unatural._1_
        (fun i in_port_name ->
         add_directed_link_in_scsm0
           in_port_name port_1_ scope_name
           (Genr_utilities_scicos_pro.
            mdl_port_number_of_natural i) [])
        in_names in

    let links = [
      Genr_utilities_scicos_pro.
      add_event_link_in_scsm0
        sample_clock_name port_1_ scope_name port_1_ [];
    ] in

    let in_ports = Concat_list.list_of_clist in_ports in

    let super_body = [
      in_ports;
      sample_clock;
      scope;
      in_links;
      links;
    ] in

    super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
