(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_pmio_port.ml,v 1.1 2019-05-10 14:04:58 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

(* Generator for connection block ports, PMIOPort in Simulink.

  This block has two variants, one for input and the other for output,
  specified by its 'Side' parameter: Left is input and Right is output.

  Numbering of these ports is done via a complex algorithm, since it depends
  on all port configuration of the surrounding system.

  In Simulink, PMIOPort number is unique (starting from 1 for each diagram),
  and the numbering is unrelated to the Side parameter value. For instance,
  in a diagram with 3 PMIOPort blocks with Sides Left, Right, and Left,
  successivelyh numbered 1 to 3, the sub-system contains two LConn ports,
  LConn1 and LConn2 and one RConn port, RConn1.
  - LConn1 corresponds to PMIOPort with port number 1
  - LConn2 corresponds to PMIOPort with port number 3
  - RConn1 corresponds to PMIOPort with port number 2

  Connectioon ports are placed on the left or right of the sub-system, after
  the regaular (directed) input and output ports respectively.

  In Scicos, each variant of PMIOPort translates into a different block:
  - system/Ports/ImplicitInput for Side=Left
  - system/Ports/ImplicitOutput for Side=Right

  Scicos block numberings of ImplicitInput and ImplicitOutput blocks are
  independent of each other. Regular input and implicit input port blocks are
  numbered together. Same for output port blocks.

  Since Simulink places connection ports after regular ports, implicit input
  ports in Scicos are numbered starting from nin + 1 (nin is the number of
  regular input ports). Same for output port starting from nout + 1.

  So this translator should get the following information:
  - nin: the number of input ports (Inport blocks) of the surrounding system,
  - nout: the number of output ports (Outport blocks) of the surrounding
    system,
  - an array of Sides, side_ports:
    side_ports.(i - 1) should provide the Side parameter value of the
    PMIOPort with parameter 'Port' (number) i in the surrounding system.

  Using this scheme, PMIOPort with Side s and Port (number) p should be
  numbered in Scicos as:
  - if s is Left: nin + index, where index is the number of occurrences of
    Left in array side_ports from 0 to p - 1 (included),
  - if s is Right: nout +  index, where index is the number of occurrences of
    Left in array side_ports from 0 to p - 1 (included).

  Note: in practice array side_ports elements are not Side value but options
  on Side values, since the Simulink numbering may contain holes. that would
  be represented as absent values (None in Caml parlance).

*)

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
open Genr_block_types
;;

let mdl_block_type =
  Mdl_basics_explicit_ast.
  Explicit_block_regular_type Mdl_block_types.PMIOPort
;;

let port_ident = Ast_types.ident_from_string "Port"
and side_ident = Ast_types.ident_from_string "Side"
;;

let genr block_id genr_info =

  let get_ident_string =
    Genr_utilities_scicos_pro.get_ident_string genr_info
  and get_ident_int =
    Genr_utilities_scicos_pro.get_ident_int genr_info in

  let port_int = get_ident_int port_ident
  and side_string = get_ident_string side_ident in

  let
    { Genr_block_types.
      system_pmio_port_sides = pmio_port_sides;
      system_inport_number = parent_nin;
      system_outport_number = parent_nout;
    } = genr_info.Genr_block_types.block_parent_system_io_descriptor in

  let scicos_block_parameters index = [
    Value_parameter (Si "portNumber",
      Code_value (code_out_int index));
  ] in

  let lim = Array.length pmio_port_sides in

  let scicos_block_type, index =
    match side_string with
    | "Left" ->
      Sbt "system/Ports/ImplicitInput",
      if port_int > lim then
        Say.fatal_error "incorrect numbering of PMIOPort" else
      let rec loop i accu =
        if i = port_int then accu else
        begin
          match pmio_port_sides.(i) with
          | Some (
              Genr_block_types.
              Pmio_port_left) -> loop (i + 1) (accu + 1)
          | None | Some Pmio_port_right ->
            loop (i + 1) accu
        end in
      loop 0 (Unatural.to_int parent_nin)
    | "Right" ->
      Sbt "system/Ports/ImplicitOutput",
      if port_int > lim then
        Say.fatal_error "incorrect numbering of PMIOPort" else
      let rec loop i accu =
        if i = port_int then accu else
        begin
          match pmio_port_sides.(i) with
          | None | Some Pmio_port_left ->
            loop (i + 1) accu
          | Some Pmio_port_right ->
            loop (i + 1) (accu + 1)
        end in
      loop 0 (Unatural.to_int parent_nout)
    | s ->
      Say.fatal_error
        (Printf.sprintf "Unknown port connection side '%s'" s) in

  let scicos_basic_block_instructions =
    genr_instructions_for_modeler_block_type
      scicos_block_type (scicos_block_parameters index)
      [] block_id genr_info in
  Concat_list.list scicos_basic_block_instructions
;;

add_translator mdl_block_type genr;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
