(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_random_number.ml,v 1.1 2019-05-10 14:04:58 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
open Unatural
;;

let mdl_block_type =
  Mdl_basics_explicit_ast.
  Explicit_block_regular_type Mdl_block_types.RandomNumber
;;

let mean_ident =
  Ast_types.ident_from_string "Mean"
and variance_ident =
  Ast_types.ident_from_string "Variance"
and seed_ident =
  Ast_types.ident_from_string "Seed"
;;

let genr block_id genr_info =

  let get_ident_string =
    Genr_utilities_scicos_pro.get_ident_string genr_info in

  let sample_time =
    Genr_utilities_scicos_pro.get_sample_time genr_info in

  let variance =
    let var_string = get_ident_string variance_ident in
    Printf.sprintf "sqrt(%s)" var_string in

  let scicos_block_parameters = [
    Value_parameter (Si "distribution",
      Code_value (code_out_string "Normal"));
    Value_parameter (Si "A",
      Simulink_arg_expression mean_ident);
    Value_parameter (Si "B",
      Code_value (code_out_expression variance));
    Value_parameter (Si "seed",
      Simulink_arg_expression seed_ident);
  ] in

  let scicos_basic_block_instructions blk_id =
    genr_instructions_for
      mdl_block_type scicos_block_parameters
      [] blk_id genr_info in

  let sample_time =
    match sample_time with
    | "-1" -> "0"
      (* above is in case of atomic SB *)
    | _ -> sample_time in

  let out_port_name = "out"
  and rand_name = "rand"
  and sample_clock_name = "sample_clock" in

  let out_port =
    Genr_utilities_scicos_pro.
    out_port_in_scsm0
      out_port_name port_1_ (_1_, _0_) in

  let rand =
    let nports = (_0_, _1_) in
    Genr_utilities_scicos_pro.
    add_block_body_in_scsm0
      rand_name
      (scicos_basic_block_instructions blk0)
      nports
      (_0_, _0_) in

  let sample_clock =
    Genr_utilities_scicos_pro.
    sample_clock_in_scsm0
      sample_clock_name
      (code_out_expression sample_time) (_0_, _1_) in

  let links = [
    add_directed_link_in_scsm0
      rand_name port_1_ out_port_name port_1_ [];
    Genr_utilities_scicos_pro.
    add_event_link_in_scsm0
      sample_clock_name port_1_ rand_name port_1_ [];
  ] in

  let super_body = [
    out_port;
    rand;
    sample_clock;
    links;
  ] in

  super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
