(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_rate_transition.ml,v 1.1 2019-05-10 14:04:58 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_utilities_scicos_pro
open Unatural
open Genr_types
;;

let mdl_block_type =
  Mdl_basics_explicit_ast.Explicit_block_regular_type
    Mdl_block_types.RateTransition
;;

let x0_ident =
  Ast_types.ident_from_string "X0"
and outport_sampletime_multiple_ident =
  Ast_types.ident_from_string "OutPortSampleTimeMultiple"
and outport_sampletime_ident =
  Ast_types.ident_from_string "OutPortSampleTime"
and outport_sampletime_opt_ident =
  Ast_types.ident_from_string "OutPortSampleTimeOpt"
;;

let genr block_id genr_info =

  let get_ident_string =
    Genr_utilities_scicos_pro.
    get_ident_string genr_info in

  let outport_sampletime_opt =
    if exists_argument genr_info outport_sampletime_opt_ident then
    get_ident_string outport_sampletime_opt_ident else "Specify" in

  let  outport_sampletime =
    get_ident_string outport_sampletime_ident in

  let out_port_name = "out"
  and in_port_name = "in" in

  let in_port =
    Genr_utilities_scicos_pro.
    in_port_in_scsm0 in_port_name port_1_
    (_0_, _0_)

  and out_port =
    Genr_utilities_scicos_pro.
    out_port_in_scsm0 out_port_name port_1_
    (_3_, _0_) in

  match outport_sampletime_opt, outport_sampletime with

  | "Specify", "-1"
  | "Inherit", _ ->

    let links = [
      Genr_utilities_scicos_pro.
      add_directed_link_in_scsm0
        in_port_name port_1_ out_port_name port_1_ [];
    ] in

    let super_body = [
      in_port;
      out_port;
      links;
    ] in
    super_block_in_scsm0 block_id super_body

  | "Specify", sample_time ->

    let sample_time = code_out_expression sample_time in

    let sample_clock_name = "sample_clock"
    and samphold_name = "samphold" in

    let sample_clock =
      Genr_utilities_scicos_pro.
      sample_clock_in_scsm0
        sample_clock_name sample_time
        (_1_, _1_)

    and samphold =
      Genr_utilities_scicos_pro.
      samphold_in_scsm0 samphold_name
        Genr_utilities_scicos_pro.
        inherit_mode_type external_activation_mode_on
        (_1_, _0_) in

    let links = [
      Genr_utilities_scicos_pro.
      add_directed_link_in_scsm0
        in_port_name port_1_ samphold_name port_1_ [];
      Genr_utilities_scicos_pro.
      add_directed_link_in_scsm0
        samphold_name port_1_ out_port_name port_1_ [];
      Genr_utilities_scicos_pro.
      add_event_link_in_scsm0
        sample_clock_name port_1_ samphold_name port_1_ [];
    ] in

    let super_body = [
      in_port;
      sample_clock;
      samphold;
      out_port;
      links;
    ] in
    super_block_in_scsm0 block_id super_body

  | "Multiple of input port sample time", _ ->

    let split_name = "split"
    and resample_name = "resample"
    and extract_name = "extract_activation"
    and dollar_name = "dollar" in

    let split =
      Genr_utilities_scicos_pro.
      split_in_scsm0 split_name
      (_1_, _0_) in

    let scicos_block_parameters = [
      Value_parameter (Si "init_cond",
        Simulink_arg_expression x0_ident);
      Value_parameter (Si "typ",
        Code_value Genr_utilities_scicos_pro.inherit_mode_type);
      Value_parameter (Si "externalActivation",
        Code_value Genr_utilities_scicos_pro.external_activation_mode_off);
    ] in

    let dollar =
      let nports = (_1_, _1_) in
      Genr_utilities_scicos_pro.
      add_block_body_in_scsm0 dollar_name
        (genr_instructions_for mdl_block_type
           scicos_block_parameters
           [] blk0 genr_info) nports
           (_2_, _0_) in

    let resample =
      let scicos_block_type =
        Sbt "system/ActivationOperations/ReSampleClock" in
      let nports = (_1_, _0_) in
      Genr_utilities_scicos_pro.
      add_block_body_in_scsm0 resample_name
        (genr_instructions_for_modeler_block_type scicos_block_type
          [
            Value_parameter (Si "up",
              Simulink_arg_expression outport_sampletime_multiple_ident);
            Value_parameter (Si "down",
              Code_value (code_out_int 1));
          ]
          [] Genr_utils.blk0 genr_info)
       nports
       (_2_, _1_) in

    let extract_activation =
      let nports = (_1_, _0_) in
      add_block_body_in_scsm0 extract_name
        (genr_instructions_for_modeler_block_type
           (Sbt "system/ActivationOperations/ExtractActivation")
           [] [] blk0 genr_info)
        nports
        (_2_, _2_) in

    let links = [
      Genr_utilities_scicos_pro.
      add_directed_link_in_scsm0
        in_port_name port_1_ split_name port_1_ [];
      add_directed_link_in_scsm0
        split_name port_1_ dollar_name port_1_ [];
      add_directed_link_in_scsm0
        split_name port_2_ extract_name port_1_ [];
      add_directed_link_in_scsm0
        dollar_name port_1_ out_port_name port_1_ [];
      Genr_utilities_scicos_pro.
      add_event_link_in_scsm0
        extract_name port_1_ resample_name port_1_ [];
      add_event_link_in_scsm0
        resample_name port_1_ dollar_name port_1_ [];
    ] in

    let super_body = [
      in_port;
      extract_activation;
      dollar;
      split;
      resample;
      out_port;
      links;
    ] in
    super_block_in_scsm0 block_id super_body

  | outport_sampletime_opt, _ ->
    Say.warning (Printf.sprintf
      "RateTransition Simulink block translation: \
       unknown configuration '%s'"
      outport_sampletime_opt);
    Genr_scicos_pro_not_yet_translated.genr block_id genr_info
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
