(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_relational_operator.ml,v 1.1 2019-05-10 14:04:58 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Genr_utilities_scicos_pro
open Unatural
;;

let mdl_block_type =
  Mdl_basics_explicit_ast.Explicit_block_regular_type
    Mdl_block_types.RelationalOperator
;;

let operator_ident =
  Ast_types.ident_from_string "Operator"
and input_same_data_type_ident =
  Ast_types.ident_from_string "InputSameDT"
;;

let genr block_id genr_info =

  let op =
    Genr_utilities_scicos_pro.get_relop_value genr_info operator_ident
  and input_same_data_type =
    get_ident_bool genr_info input_same_data_type_ident
  and zcr =
    Genr_utilities_scicos_pro.get_zero_crossing_value genr_info
  and out_data_type =
    Genr_utilities_scicos_pro.get_simulink_out_data_type_value genr_info
  and sample_time =
    Genr_utilities_scicos_pro.get_sample_time genr_info in

  let in1_name = "in1"
  and in2_name = "in2"
  and out_name = "out"
  and relational_op_name = "relational_op"
  and split_name = "split"
  and convert_name = "convert"
  and sample_clock_name = "sample_clock" in

  let is_nan_op =
    let op_as_string =
      Genr_utilities_common.get_argument_value_string
        genr_info operator_ident in
    op_as_string = "isNaN" in

  let in_port1 =
    Genr_utilities_scicos_pro.
    in_port_in_scsm0 in1_name port_1_
    (_0_, _1_) in
  let in_port2 =
    Genr_utilities_scicos_pro.
    in_port_in_scsm0 in2_name port_2_
    (_0_, _0_) in
  let out_port =
    Genr_utilities_scicos_pro.
    out_port_in_scsm0 out_name port_1_
    (_3_, _1_) in

  let inherit_data_type =
    Genr_utilities_scicos_pro.inherit_mode_type in

  let convert =
    let saturation = saturation_mode_off in
    Genr_utilities_scicos_pro.
    convert_in_scsm0_with_saturation
      convert_name inherit_data_type saturation
      (_1_, _1_) in

  let super_body =

    let relational_op =
      Genr_utilities_scicos_pro.
      relational_op_in_scsm0
        relational_op_name op zcr out_data_type
        external_activation_mode_off
        (_2_, _1_) in

    match (sample_time, input_same_data_type, is_nan_op) with
    | "-1", true, false ->

      let links = [
        Genr_utilities_scicos_pro.
        add_directed_link_in_scsm0 in1_name port_1_ relational_op_name port_1_ [];
        add_directed_link_in_scsm0 in2_name port_1_ relational_op_name port_2_ [];
        add_directed_link_in_scsm0 relational_op_name port_1_ out_name port_1_ [];
      ] in

      [
        in_port1; in_port2; out_port;
        relational_op;
        links;
      ]

    | "-1", false, false ->

      let links = [
        Genr_utilities_scicos_pro.
        add_directed_link_in_scsm0 in1_name port_1_ convert_name port_1_ [];
        add_directed_link_in_scsm0 convert_name port_1_ relational_op_name port_1_ [];
        add_directed_link_in_scsm0 in2_name port_1_ relational_op_name port_2_ [];
        add_directed_link_in_scsm0 relational_op_name port_1_ out_name port_1_ [];
      ] in

      [
        in_port1; in_port2; out_port;
        relational_op; convert;
        links;
      ]

    | _, true, false ->

      let sample_clock =
        sample_clock_in_scsm0 sample_clock_name
        (code_out_expression sample_time)
        (_2_, _2_) in

      let links = [
        Genr_utilities_scicos_pro.
        add_directed_link_in_scsm0 in1_name port_1_ relational_op_name port_1_ [];
        add_directed_link_in_scsm0 in2_name port_1_ relational_op_name port_2_ [];
        add_directed_link_in_scsm0 relational_op_name port_1_ out_name port_1_ [];

        Genr_utilities_scicos_pro.
        add_event_link_in_scsm0
          sample_clock_name port_1_ relational_op_name port_1_ [];
      ] in

      [
        in_port1; in_port2; out_port;
        relational_op;
        sample_clock;
        links;
      ]

    | _, false, false ->

      let sample_clock =
        sample_clock_in_scsm0 sample_clock_name
        (code_out_expression sample_time)
        (_2_, _2_) in

      let links = [
        Genr_utilities_scicos_pro.
        add_directed_link_in_scsm0 in1_name port_1_ convert_name port_1_ [];
        add_directed_link_in_scsm0 convert_name port_1_ relational_op_name port_1_ [];
        add_directed_link_in_scsm0 in2_name port_1_ relational_op_name port_2_ [];
        add_directed_link_in_scsm0 relational_op_name port_1_ out_name port_1_ [];

        add_event_link_in_scsm0 sample_clock_name port_1_ relational_op_name port_1_ [];
      ] in

      [
        in_port1; in_port2; out_port;
        relational_op; convert;
        sample_clock;
        links;
      ]

    | "-1", _, true ->

      let split =
        Genr_utilities_scicos_pro.
          split_in_scsm0
            split_name (_1_, _1_) in

      let links = [
        Genr_utilities_scicos_pro.
        add_directed_link_in_scsm0 in1_name port_1_ split_name port_1_ [];
        add_directed_link_in_scsm0 split_name port_1_ relational_op_name port_1_ [];
        add_directed_link_in_scsm0 split_name port_2_ relational_op_name port_2_ [];
        add_directed_link_in_scsm0 relational_op_name port_1_ out_name port_1_ [];
      ] in

      [
        in_port1; out_port;
        split; relational_op;
        links;
      ]

    | _, _, true ->

      let split =
        Genr_utilities_scicos_pro.
          split_in_scsm0
            split_name (_1_, _1_) in

      let sample_clock =
        sample_clock_in_scsm0 sample_clock_name
        (code_out_expression sample_time)
        (_2_, _2_) in

      let links = [
        Genr_utilities_scicos_pro.
        add_directed_link_in_scsm0 in1_name port_1_ split_name port_1_ [];
        add_directed_link_in_scsm0 split_name port_1_ relational_op_name port_1_ [];
        add_directed_link_in_scsm0 split_name port_2_ relational_op_name port_2_ [];
        add_directed_link_in_scsm0 relational_op_name port_1_ out_name port_1_ [];

        Genr_utilities_scicos_pro.
        add_event_link_in_scsm0
          sample_clock_name port_1_ relational_op_name port_1_ [];
      ] in

      [
        in_port1; out_port;
        split; relational_op;
        sample_clock;
        links;
      ] in

  super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
