(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_scope.ml,v 1.1 2019-05-10 14:04:58 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
;;

let mdl_block_type =
  simulink_block_type Mdl_block_types.Scope
;;

let num_in_ports_ident =
  Ast_types.ident_from_string "NumInputPorts"
and y_min_ident =
  Ast_types.ident_from_string "YMin"
and y_max_ident =
  Ast_types.ident_from_string "YMax"
and time_range_ident =
  Ast_types.ident_from_string "TimeRange"
and location_ident =
  Ast_types.ident_from_string "Location"
and scope_specification_ident =
  Ast_types.ident_from_string "ScopeSpecificationString"
and axes_titles_ident =
  Ast_types.ident_from_string "AxesTitles"
;;

let find_bound_value ident bindings =
  let same_ident ident binding =
    binding.Binding.bound_ident = ident in
  let binding =
    List.find (same_ident ident) bindings in
  binding.Binding.bound_value
;;

let string_of_int_pair (x, y) = Printf.sprintf "[ %d; %d ]" x y;;

let get_min_max_bounds bounds_min bounds_max =
  let get_bounds bounds_string =
    let ib = Scanf.Scanning.from_string bounds_string in
    let rec loop accu =
      if Scanf.Scanning.end_of_input ib then List.rev accu else
      try
        Scanf.bscanf ib "%s@~"
          (fun s -> loop (s :: accu)) with
      | Scanf.Scan_failure s
      | Failure s ->
        Say.fatal_error
          (Printf.sprintf
             "Genr_scicos_pro_scope: \
              while scanning ``%s'', encountered ``%s''"
             bounds_string s) in
    loop [] in

  (get_bounds bounds_min, get_bounds bounds_max)
;;

let get_axis_title axes_titles_bindings index =

  let sought_ident =
    let sought_name = "axes" ^ string_of_int index in
    Ident.Ident sought_name in

  let title =
    try
      let sought_value =
        find_bound_value sought_ident axes_titles_bindings in
      let axis_title =
        Mdl_deep_basics.string_of_deep_value "axis title" sought_value in
      if axis_title <> "%<SignalLabel>" then axis_title
      (* todo: signal name is dropped. We may obtain it using a too
         complex algorithm. *)
      else "" with
    | Not_found -> "" in

  Code_value (code_out_string title)
;;

let row_concat a b =
  Code_value (code_out_expression (Printf.sprintf "[ %s; %s ]" a b))
;;

let genr block_id genr_info =

  let get_ident_string =
    Genr_utilities_scicos_pro.get_ident_string genr_info
  and get_ident_int =
    Genr_utilities_scicos_pro.get_ident_int genr_info
  and exists_argument =
    Genr_utilities_scicos_pro.exists_argument genr_info
  and exists_parameter =
    Genr_utilities_scicos_pro.exists_parameter genr_info
  and get_argument_value =
    Genr_utilities_scicos_pro.get_argument_value genr_info
  and get_ident_int_vector =
    Genr_utilities_scicos_pro.get_ident_int_vector genr_info in

  let get_num_in_ports num_in_ports_ident =
    if exists_argument num_in_ports_ident then
      get_ident_int num_in_ports_ident else
    let ports_desc = genr_info.Genr_block_types.block_ports_description in
    let nin = Mdl_ast_to_deep_ast.get_regular_port_nin ports_desc in
    Unatural.to_int nin in

  (* For debugging purpose *)
  Genr_utilities_common.debug_with_info genr_info "Genr_scicos_pro_scope";

  let num_in_ports = get_num_in_ports num_in_ports_ident in

  let y_bounds =
    let y_mins, y_maxs =
      if exists_parameter scope_specification_ident
      then get_min_max_bounds "-5" "5" else
      let bounds_min = get_ident_string y_min_ident
      and bounds_max = get_ident_string y_max_ident in
      get_min_max_bounds bounds_min bounds_max in

    List.map2 row_concat y_mins y_maxs in

  let axes_titles =
    let axes_titles_bindings =
      if not (exists_argument axes_titles_ident) then [] else
      let titles = get_argument_value axes_titles_ident in
      Mdl_deep_basics.bindings_of_deep_value "AxesTitles" titles in
    let rec loop accu = function
      | 0 -> accu
      | i -> loop (get_axis_title axes_titles_bindings i :: accu) (pred i) in
    loop [] num_in_ports in

  let time_range =
    if exists_argument time_range_ident
    then get_ident_string time_range_ident
    else "20"

  and location =
    if exists_argument location_ident
    then get_ident_int_vector location_ident
    else [| 0; 0; 200; 200; |]

  and block_name =
    genr_info.Genr_block_types.block_explicit_name in

  let nin_string = string_of_int num_in_ports in
  let nin_nat = Unatural.of_string nin_string
  and sample_time =
    Genr_utilities_scicos_pro.
    get_sample_time genr_info in

  let refresh =
    [ Code_value (code_out_expression time_range) ]
  and sz =
    string_of_int_pair
      (location.(2) - location.(0), location.(3) - location.(1))
  and og =
    string_of_int_pair (location.(0), location.(1)) in
    (* loc(1:2) // this is probably wrong- must be height-loc(2)+-sz(2) *)

  let in_names, in_ports =
    Genr_utilities_scicos_pro.
    in_ports_in_scsm0 nin_nat in

  let scicos_block_parameters activation = [
    Value_parameter (Si "nin",
      Code_value (code_out_expression nin_string));
    Value_parameter (Si "externalActivation",
      Code_value activation);
    Value_parameter (Si "winSize",
      Code_value (code_out_expression sz));
    Value_parameter (Si "winPos",
      Code_value (code_out_expression og));
    Value_parameter (Si "useSubplot",
      Code_value (code_out_bool true));
    Value_parameter (Si "winNam",
      Code_value (code_out_string (
        Ast_types.string_of_mdl_name block_name)));
    Struct_parameter (Si "win_s",
       ((Si "xRefresh", refresh),
        [ Si "yInterval", y_bounds;
          Si "title", axes_titles ]));
  ] in

  let scicos_basic_block_instructions blk_id activation =
    genr_instructions_for
      mdl_block_type (scicos_block_parameters activation)
      [] blk_id genr_info in

  match sample_time with
  | "0" | "-1" ->
    Concat_list.list
      (scicos_basic_block_instructions block_id
         external_activation_mode_off)
  | _ ->

    let sample_clock_name = "sample_clock"
    and scope_name = "scope" in

    let sample_clock =
      Genr_utilities_scicos_pro.
      sample_clock_in_scsm0
        sample_clock_name
        (code_out_expression sample_time)
        (Unatural._1_, Unatural._1_) in

    let scope =
      let nports = (nin_nat, Unatural._0_) in
      Genr_utilities_scicos_pro.
      add_block_body_in_scsm0 scope_name
        (scicos_basic_block_instructions blk0 external_activation_mode_on)
        nports
        (Unatural._1_, Unatural._0_) in

    let in_links =
      Genr_utilities_scicos_pro.
      rev_map_i_start Unatural._1_
        (fun i in_port_name ->
         add_directed_link_in_scsm0
           in_port_name port_1_ scope_name
           (Genr_utilities_scicos_pro.mdl_port_number_of_natural i) [])
        in_names in

    let links = [
      Genr_utilities_scicos_pro.
      add_event_link_in_scsm0
        sample_clock_name port_1_ scope_name port_1_ [];
    ] in

    let in_ports = Concat_list.list_of_clist in_ports in

    let super_body = [
      in_ports;
      sample_clock;
      scope;
      in_links;
      links;
    ] in

    super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
