(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_selector.ml,v 1.1 2019-05-10 14:04:58 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_utilities_scicos_pro
;;

let mdl_block_type =
  Mdl_basics_explicit_ast.
  Explicit_block_regular_type Mdl_block_types.Selector
;;
let scicos_block_type =
  Genr_utilities_scicos_pro.modeler_block_type_for mdl_block_type
;;

let number_of_dimensions_ident =
  Ast_types.ident_from_string "NumberOfDimensions"
and index_mode_ident =
  Ast_types.ident_from_string "IndexMode"
and index_options_ident =
  Ast_types.ident_from_string "IndexOptions"
and indices_ident =
  Ast_types.ident_from_string "Indices"
and elements_ident =
  Ast_types.ident_from_string "Elements"
and output_sizes_ident =
  Ast_types.ident_from_string "OutputSizes"
;;

let genr block_id genr_info =

  let get_ident_string =
    Genr_utilities_scicos_pro.get_ident_string genr_info in

  let default_number_of_dimensions = 1
  and default_index_options = "Index vector (dialog)"
  and default_output_sizes = "1" in

  let indices_string =
    if Genr_utilities_scicos_pro.exists_argument genr_info indices_ident
    then get_ident_string indices_ident
    else get_parameter_value_string genr_info elements_ident in

  let output_sizes =
    if exists_argument genr_info output_sizes_ident
    then get_ident_string output_sizes_ident
    else default_output_sizes in

  let number_of_dimensions =
    if exists_argument genr_info number_of_dimensions_ident
    then get_ident_int genr_info number_of_dimensions_ident
    else default_number_of_dimensions

  and index_options =
    if exists_argument genr_info index_options_ident
    then get_ident_string index_options_ident
    else default_index_options in

  let index_mode = match
    get_ident_string index_mode_ident with
    | "One-based" -> 1
    | "Zero-based" -> 0
    | s ->
      Say.fatal_error (Printf.sprintf
        "Unsupported base indexing %s in selector" s) in

  let translate_index_option dir t =
    let index_option =
      match t with
      | "Index vector (dialog)" -> "vectorindex"
      | "Index vector (port)" -> "vectorindexport"
      | "Starting index (port)" -> "startingindexport"
      | "Starting index (dialog)" -> "startingindex"
      | "Select all" -> "selectall"
      | t ->
        Say.warning (Printf.sprintf
          "Unsupported index option in selector: %s" t);
        raise Not_found in
    Printf.sprintf "%s_%s" dir index_option in

  let get_items s =
    let lim = String.length s in
    let accu = Buffer.create lim in
    let rec loop paren_count bracket_count i items =
      if i >= lim then
        let s = Buffer.contents accu in
        Buffer.clear accu;
        Ulist1.of_list (List.rev (s :: items))
      else
        let c = s.[i] in
        let i = succ i in
        match c with
        | '(' ->
          Buffer.add_char accu c;
          loop (succ paren_count) bracket_count i items
        | ')' ->
          Buffer.add_char accu c;
          loop (pred paren_count) bracket_count i items
        | '[' ->
          Buffer.add_char accu c;
          loop paren_count (succ bracket_count) i items
        | ']' ->
          Buffer.add_char accu c;
          loop paren_count (pred bracket_count) i items
        | ',' when paren_count = 0 && bracket_count = 0 ->
          let s = Buffer.contents accu in
          Buffer.clear accu;
          loop paren_count bracket_count i (s :: items)
        | _ ->
          Buffer.add_char accu c;
          loop paren_count bracket_count i items in
    loop 0 0 0 []  in

  let si, _indices = get_items indices_string in
  let osz, _o_sizes = get_items output_sizes in

  let get_x_params index_option =
    match index_option with
    | "x_vectorindex" ->
      struct_parameter_binding
        (Si "x_vector_index",
         ((Si "x_vi", [ code_out_expression si; ]), []))

    | "x_startingindex" ->
      struct_parameter_binding
        (Si "x_starting_index",
         ((Si "x_si", [ code_out_expression si; ]),
          [
            Si "x_osz", [ code_out_expression osz; ];
          ]))

    | "x_startingindexport" ->
      struct_parameter_binding
        (Si "x_starting_index_port",
         ((Si "x_osz", [ code_out_expression osz; ]), []))

    | _ -> assert false

  and get_y_params index_option =
    match index_option with
    | "y_vectorindex" ->
      struct_parameter_binding
        (Si "y_vector_index",
         ((Si "y_vi", [ code_out_expression si; ]), []))

    | "y_startingindex" ->
      struct_parameter_binding
        (Si "y_starting_index",
         ((Si "y_si", [ code_out_expression si; ]),
          [
            Si "y_osz", [ code_out_expression osz; ]
          ]))

    | "y_startingindexport" ->
      struct_parameter_binding
        (Si "y_starting_index_port",
         ((Si "y_osz", [ code_out_expression osz; ]), []))

    | _ -> assert false in

  match number_of_dimensions with
  | 1 ->
    begin match get_items index_options with
    | index_option, [] ->
      let blk =
        try
          let index_option = translate_index_option "x" index_option in
          let scicos_parameters =
            let params = [
              block_parameter_binding
                (Si "extracty", code_out_expression "y_vectorindex");
              struct_parameter_binding
                (Si "y_vector_index",
                 ((Si "y_vi", [code_out_int 1]), []));
              block_parameter_binding
                (Si "onebased", code_out_int index_mode);
              block_parameter_binding
                (Si "externalActivation", external_activation_mode_off);
            ] in
            let params =
              match index_option with
              | "x_vectorindex" | "x_startingindex" | "x_startingindexport" ->
                get_x_params index_option :: params
              | "x_vectorindexport" | "x_selectall" -> params
              | c ->
                Say.fatal_error (Printf.sprintf
                  "Unsupported index option %s" c) in

            (block_parameter_binding
               (Si "extractx", code_out_expression index_option),
             params) in
            Concat_list.list [
              Instantiate_block (block_id, scicos_block_type);
              Set_block_parameters (block_id, scicos_parameters);
            ]
        with
        | Not_found ->
          Genr_scicos_pro_not_yet_translated.genr block_id genr_info in
      blk
    | _ ->
      Say.fatal_error "incompatible index option in Selector"
    end

  | 2 ->
    begin match get_items index_options with
    | index_x_option, [ index_y_option; ] ->
      let blk =
        try
          let index_x_option = translate_index_option "x" index_x_option
          and index_y_option = translate_index_option "y" index_y_option in
          let scicos_parameters =
            let params = [
              block_parameter_binding
                (Si "extracty", (code_out_expression index_y_option));
              struct_parameter_binding
                (Si "y_vector_index",
                   ((Si "y_vi", [ code_out_int 1; ]), []));
              block_parameter_binding
                (Si "onebased", code_out_int index_mode);
              block_parameter_binding
                (Si "externalActivation", external_activation_mode_off);
            ] in

            let params =
              match index_y_option with
              | "y_vectorindex" | "y_startingindex" | "y_startingindexport" ->
                get_y_params index_y_option :: params
              | "y_vectorindexport" | "y_selectall" -> params
              | c ->
                Say.fatal_error (Printf.sprintf
                  "Unsupported index option %s" c) in

            let params =
              match index_x_option with
              | "x_vectorindex" | "x_startingindex" | "x_startingindexport" ->
                get_x_params index_x_option :: params
              | "x_vectorindexport" | "x_selectall" -> params
              | c ->
                Say.fatal_error (Printf.sprintf
                  "Unsupported index option %s" c) in

            (block_parameter_binding
               (Si "extractx", code_out_expression index_x_option),
             params) in

          Concat_list.list [
            Instantiate_block (block_id, scicos_block_type);
            Set_block_parameters (block_id, scicos_parameters);
          ]
        with
        | Not_found ->
          Genr_scicos_pro_not_yet_translated.genr block_id genr_info in
      blk
    | _ ->
      Say.fatal_error "incompatible index option in Selector"
    end

  | d ->
    Say.warning (Printf.sprintf
      "Selector of dimension ``%d'' not supported" d);
    Genr_scicos_pro_not_yet_translated.genr block_id genr_info
;;

add_translator mdl_block_type genr;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
