(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_signal_generator.ml,v 1.1 2019-05-10 14:04:58 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
;;

let mdl_block_type =
  Mdl_basics_explicit_ast.
  Explicit_block_regular_type Mdl_block_types.SignalGenerator
;;

let wave_form_ident =
  Ast_types.ident_from_string "WaveForm"
and time_source_ident =
  Ast_types.ident_from_string "TimeSource"
and amplitude_ident =
  Ast_types.ident_from_string "Amplitude"
and frequency_ident =
  Ast_types.ident_from_string "Frequency"
and units_ident =
  Ast_types.ident_from_string "Units"
;;

let genr block_id genr_info =

  let get_ident = Genr_utilities_scicos_pro.get_ident_string genr_info in

  let wave_form = get_ident wave_form_ident
  and time_source = get_ident time_source_ident
  and amplitude = get_ident amplitude_ident
  and frequency = get_ident frequency_ident
  and units = get_ident units_ident in

  let get_Amin () = Printf.sprintf "-abs(%s)" amplitude
  and get_Amax () = Printf.sprintf "abs(%s)" amplitude
  and get_minus_amplitude () = Printf.sprintf "-(%s)" amplitude in

  let get_F () =
    if units = "Hertz"
    then Printf.sprintf "2 * pi * (%s)" frequency
    else frequency in

  let get_P () =
    if units = "Hertz"
    then Printf.sprintf "1 / (%s)" frequency
    else Printf.sprintf "2 * pi / (%s)" frequency in

  match time_source with
  | "Use simulation time" ->
    begin match wave_form with

    | "sine" ->

      let scicos_block_parameters = [
        Value_parameter (Si "M", Code_value (code_out_expression amplitude));
        Value_parameter (Si "F", Code_value (code_out_expression (get_F ())));
        Value_parameter (Si "P", Code_value (code_out_float 0.0));
        Value_parameter (Si "offs", Code_value (code_out_float 0.0));
      ] in
      Concat_list.list (
        genr_instructions_for mdl_block_type
          scicos_block_parameters
          [] block_id genr_info)

    | "square" ->
      let scicos_block_parameters = [
         Value_parameter
           (Si "FV",
            Code_value (code_out_expression (get_minus_amplitude ())));
         Value_parameter
           (Si "SV",
            Code_value (code_out_expression amplitude));
         Value_parameter
           (Si "period", Code_value (code_out_expression (get_P ())));
         Value_parameter
           (Si "dutyCycle", Code_value (code_out_float 50.0));
         Value_parameter
           (Si "offset", Code_value (code_out_float 0.0));
         Value_parameter (Si "datatype",
           Code_value Genr_utilities_scicos_pro.inherit_mode_type);
      ] in

      let scicos_block_type =
        Sbt "system/SignalGenerators/SquareWaveGenerator" in

      Concat_list.list (
        genr_instructions_for_modeler_block_type scicos_block_type
          scicos_block_parameters
          [] block_id genr_info)

    | "sawtooth" ->

      let scicos_block_parameters = [
        Value_parameter
          (Si "period", Code_value (code_out_expression (get_P ())));
        Value_parameter
          (Si "down", Code_value (code_out_expression amplitude));
        Value_parameter (Si "up",
          Code_value (code_out_expression (get_minus_amplitude ())));
        Value_parameter
          (Si "tinit", Code_value (code_out_float 0.0));
      ] in

      let scicos_block_type =
        Sbt "system/SignalGenerators/Sawtooth" in

      Concat_list.list (
        genr_instructions_for_modeler_block_type scicos_block_type
          scicos_block_parameters
          [] block_id genr_info)

    | "random" ->

      let get_B () = Printf.sprintf "2 * %s" (get_Amax ()) in

      let scicos_block_parameters = [
        Value_parameter
          (Si "distribution", Code_value (code_out_string "Uniform"));
         Value_parameter
          (Si "A", Code_value (code_out_expression (get_Amin ())));
         Value_parameter
          (Si "B", Code_value (code_out_expression (get_B ())));
         Value_parameter
          (Si "seed", Code_value (code_out_int 0));
      ] in

      let scicos_block_type = Sbt "system/SignalGenerators/Random" in

      Concat_list.list (
        genr_instructions_for_modeler_block_type scicos_block_type
          scicos_block_parameters
          [] block_id genr_info)

    | wave_form ->
      Say.warning (Printf.sprintf
        "Signal generator Simulink block translation: \
         Waveform ``%s'' not yet implemented"
        wave_form);
      Genr_scicos_pro_not_yet_translated.genr block_id genr_info
    end

  | time_source ->
    Say.warning (Printf.sprintf
      "Signal generator Simulink block translation: \
       Time source ``%s'' not yet implemented"
      time_source);
    Genr_scicos_pro_not_yet_translated.genr block_id genr_info
;;

add_translator mdl_block_type genr;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
