(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_sin.ml,v 1.1 2019-05-10 14:04:58 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
open Unatural
;;

let mdl_block_type =
  Mdl_basics_explicit_ast.Explicit_block_regular_type
    Mdl_block_types.Sin
;;

let phase_ident =
  Ast_types.ident_from_string "Phase"
and amplitude_ident =
  Ast_types.ident_from_string "Amplitude"
and frequency_ident =
  Ast_types.ident_from_string "Frequency"
and bias_ident =
  Ast_types.ident_from_string "Bias"
and time_source_ident =
  Ast_types.ident_from_string "TimeSource"
and sine_type_ident =
  Ast_types.ident_from_string "SineType"
;;

let genr block_id genr_info =

  let get_ident_string =
    Genr_utilities_scicos_pro.get_ident_string genr_info in

  let sample_time =
    Genr_utilities_scicos_pro.get_sample_time genr_info in

  let sine_type = get_ident_string sine_type_ident
  and time_source = get_ident_string time_source_ident in

  let scicos_block_parameters ext = [
    Value_parameter (Si "M",
      Simulink_arg_expression amplitude_ident);
    Value_parameter (Si "F",
      Simulink_arg_expression frequency_ident);
    Value_parameter (Si "P",
      Simulink_arg_expression phase_ident);
    Value_parameter (Si "offs",
      Simulink_arg_expression bias_ident);
    Value_parameter (Si "externalActivation",
      Code_value ext);
  ] in

  let scicos_basic_block_instructions blk_id ext =
    genr_instructions_for mdl_block_type
      (scicos_block_parameters ext)
      [] blk_id genr_info in

  let scicos_sample_time_block_instructions sample_time blk_id =

    let out_port_name = "out"
    and sin_name = "sin"
    and sample_clock_name = "sample_clock" in

    let out_port =
      Genr_utilities_scicos_pro.out_port_in_scsm0
        out_port_name port_1_
        (_1_, _0_) in

    let sin =
      let nports = (_0_, _1_) in
      Genr_utilities_scicos_pro.
      add_block_body_in_scsm0
        sin_name
        (scicos_basic_block_instructions blk0 external_activation_mode_on)
        nports
        (_0_, _0_) in

    let sample_clock =
      Genr_utilities_scicos_pro.
      sample_clock_in_scsm0
        sample_clock_name
        (code_out_expression sample_time)
        (_0_, _1_) in

    let links = [
      Genr_utilities_scicos_pro.
      add_directed_link_in_scsm0
        sin_name port_1_ out_port_name port_1_ [];
      Genr_utilities_scicos_pro.
      add_event_link_in_scsm0
        sample_clock_name port_1_ sin_name port_1_ [];
    ] in

    let super_body = [
      out_port;
      sin;
      sample_clock;
      links;
    ] in
    Genr_utilities_scicos_pro.
    super_block_in_scsm0 blk_id super_body  in

  match time_source, sine_type with
  | "Use simulation time", "Time based" ->
    begin
      match sample_time with
      | "0" ->
        Concat_list.list
          (scicos_basic_block_instructions block_id
             external_activation_mode_off)
      | sample_time ->
        scicos_sample_time_block_instructions sample_time block_id
    end
  | "Use simulation time", "Sample based" ->
    Say.warning
      "Sin Simulink block translation: \
       ``Frame Based'' type not yet implemented";
    Genr_scicos_pro_not_yet_translated.genr block_id genr_info
  | "Use external signal", "Time based" ->
    Say.warning
      "Sin Simulink block translation: \
       ``Use external signal'' option not yet implemented";
    Genr_scicos_pro_not_yet_translated.genr block_id genr_info
  | "Use external signal", "Sample based" ->
    Say.warning
      "Sin Simulink block translation: \
       ``Frame Based'' type not yet implemented";
    Genr_scicos_pro_not_yet_translated.genr block_id genr_info
  | t1, t2 ->
    Say.warning (Printf.sprintf
      "Sin Simulink block translation: \
       time source ``%s'' or sine type ``%s'' not yet implemented" t1 t2);
    Genr_scicos_pro_not_yet_translated.genr block_id genr_info
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
