(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_switch.ml,v 1.1 2019-05-10 14:04:58 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
open Unatural
;;

let mdl_block_type =
  Mdl_basics_explicit_ast.Explicit_block_regular_type
    Mdl_block_types.Switch
;;

let criteria_ident =
  Ast_types.ident_from_string "Criteria"
and threshold_ident =
  Ast_types.ident_from_string "Threshold"
;;

let genr block_id genr_info =

  let get_ident_string =
    Genr_utilities_scicos_pro.get_ident_string genr_info in

  let sample_time =
    Genr_utilities_scicos_pro.
    get_sample_time genr_info
  and zcr =
    Genr_utilities_scicos_pro.
    get_zero_crossing_value genr_info
  and criteria = get_ident_string criteria_ident
  and default_criteria = "u2 >= Threshold" in

  let rec get_op criteria =
    match criteria with
    | c when c = default_criteria -> "u2>=a"
    | "u2 > Threshold" -> "u2>a"
    | "u2 ~= 0" -> "u2~=a"
    | c ->
      Genr_utilities_scicos_pro.warning (Printf.sprintf
        "Unsupported criterion in Switch block: \
         using ``%s'' instead of ``%s''."
         default_criteria c);
      get_op default_criteria in

  let op = get_op criteria
  and thresh =
    if criteria = "u2 ~= 0"
    then code_out_float 0.0
    else code_out_expression (get_ident_string threshold_ident) in

  let scicos_block_parameters activation = [
    Value_parameter (Si "u2typ",
      Code_value Genr_utilities_scicos_pro.inherit_mode_type);
    Value_parameter (Si "rule",
      Code_value (code_out_string op));
    Value_parameter (Si "thra",
      Code_value thresh);
    Value_parameter (Si "nzz",
      Code_value zcr);
    Value_parameter (Si "externalActivation",
      Code_value activation);
  ] in

  let scicos_basic_block_instructions blk_id activation =
    genr_instructions_for
      mdl_block_type (scicos_block_parameters activation)
      [] blk_id genr_info in

  match sample_time with
  | "-1" ->
    Concat_list.list
      (scicos_basic_block_instructions block_id
       Genr_utilities_scicos_pro.external_activation_mode_off)

  | _ ->

    let sample_time = code_out_expression sample_time in

    let in_port1_name = "in1"
    and in_port2_name = "in2"
    and in_port3_name = "in3"
    and out_port_name = "out"
    and sw_name = "sw"
    and sample_clock_name = "sample_clock" in

    let in_port1 =
      Genr_utilities_scicos_pro.
      in_port_in_scsm0 in_port1_name port_1_
    and in_port2 =
      Genr_utilities_scicos_pro.
      in_port_in_scsm0 in_port2_name port_2_
    and in_port3 =
      Genr_utilities_scicos_pro.
      in_port_in_scsm0 in_port3_name port_3_
    and out_port =
      Genr_utilities_scicos_pro.
      out_port_in_scsm0 out_port_name port_1_
    and sw =
      let nports = (_3_, _1_) in
      Genr_utilities_scicos_pro.
      add_block_body_in_scsm0 sw_name
        (scicos_basic_block_instructions blk0
           Genr_utilities_scicos_pro.external_activation_mode_on)
        nports
    and sample_clock =
      Genr_utilities_scicos_pro.
      sample_clock_in_scsm0
        sample_clock_name sample_time in

    (* Creating required blocks before adding links. *)
    let in_port1_block = in_port1 (_0_, _2_)
    and in_port2_block = in_port2 (_0_, _1_)
    and in_port3_block = in_port3 (_0_, _0_)
    and sample_clock_block = sample_clock (_1_, _1_)
    and sw_block = sw (_1_, _0_)
    and out_port_block = out_port (_2_, _0_) in

    let links = [
      Genr_utilities_scicos_pro.
      add_directed_link_in_scsm0 in_port1_name port_1_ sw_name port_1_ [];
      add_directed_link_in_scsm0 in_port2_name port_1_ sw_name port_2_ [];
      add_directed_link_in_scsm0 in_port3_name port_1_ sw_name port_3_ [];
      add_directed_link_in_scsm0 sw_name port_1_ out_port_name port_1_ [];
      Genr_utilities_scicos_pro.
      add_event_link_in_scsm0
        sample_clock_name port_1_ sw_name port_1_ [];
    ] in

    let super_body = [
      in_port1_block;
      in_port2_block;
      in_port3_block;
      sample_clock_block;
      sw_block;
      out_port_block;
      links;
    ] in

    super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
