(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_tapped_delay_line.ml,v 1.1 2019-05-10 14:04:58 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
open Unatural
;;

let mdl_block_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "Tapped Delay Line")
    (Ast_types.mdl_path_of_string
       "simulink/Discrete/Tapped Delay")
;;

let delay_order_ident =
  Ast_types.ident_from_string "DelayOrder"
and num_delays_ident =
  Ast_types.ident_from_string "NumDelays"
and vinit_ident =
  Ast_types.ident_from_string "vinit"
and include_current_ident =
  Ast_types.ident_from_string "includeCurrent"
;;

let genr block_id genr_info =

  let get_ident_string =
    Genr_utilities_scicos_pro.get_ident_string genr_info in

  let sample_time =
    Genr_utilities_scicos_pro.get_sample_time genr_info in

  let s = get_ident_string delay_order_ident in
  let reverse =
    if s = "Oldest" then code_out_bool true else
    if s = "Newest" then code_out_bool false else
    Say.fatal_error (Printf.sprintf
      "Unsupported delay order %s in tapped delay line" s) in

  let include_current = get_ident_bool genr_info include_current_ident in

  let scicos_block_parameters = [
    Value_parameter (Si "bufferlength",
      Simulink_arg_expression num_delays_ident);
    Value_parameter (Si "initialoutput",
      Simulink_arg_expression vinit_ident);
    Value_parameter (Si "reverseBuffer",
      Code_value reverse);
    Value_parameter (Si "circularBuffer",
      Code_value (code_out_bool false));
    Value_parameter (Si "feedthrough",
      Code_value (code_out_bool include_current));
  ] in

  let scicos_basic_block_instructions blk_id =
    genr_instructions_for
      mdl_block_type scicos_block_parameters
      [] blk_id genr_info in

  let in_port_name = "in"
  and out_port_name = "out"
  and buffer_name = "buffer"
  and extract_activation_name = "extract_activation"
  and reshape_name = "Reshape"
  and event_split_name = "event_split"
  and split_name = "split"
  and sample_clock_name = "sample_clock" in

  let in_port =
    in_port_in_scsm0 in_port_name port_1_ (_0_, _0_) in

  let split =
    Genr_utilities_scicos_pro.
    split_in_scsm0 split_name (_1_, _0_) in

  let event_split =
    Genr_utilities_scicos_pro.
    split_event_in_scsm0 event_split_name (_2_, _1_) in

  let out_port =
    Genr_utilities_scicos_pro.
    out_port_in_scsm0 out_port_name port_1_
    (_4_, _0_) in

  let buffer =
    let nports = (_1_, _1_) in
    Genr_utilities_scicos_pro.
    add_block_body_in_scsm0 buffer_name
      (scicos_basic_block_instructions blk0)
      nports
      (_2_, _0_) in

  let genr_activation sample_time =
    let activation_location = (_2_, _2_) in
    let activation =
      if sample_time = "-1"
      then
        let extract_activation =
          let nports = (_1_, _0_) in
          Genr_utilities_scicos_pro.
          add_block_body_in_scsm0 extract_activation_name
            (genr_instructions_for_modeler_block_type
               (Sbt "system/ActivationOperations/ExtractActivation")
               [] [] blk0 genr_info)
            nports in
         extract_activation
      else
        let sample_clock =
          Genr_utilities_scicos_pro.sample_clock_in_scsm0
            sample_clock_name
            (code_out_expression sample_time) in
        sample_clock in
    activation activation_location in

  let reshape =
    let scicos_block_type =
      Sbt "system/MatrixOperations/MatrixReshape" in
    let reshape_parameters = [
      Value_parameter (Si "outsize",
        Code_value (code_out_expression "[-1, 1]"));
      Value_parameter (Si "externalActivation",
        Code_value
          Genr_utilities_scicos_pro.external_activation_mode_off);
    ] in

    let nports = (_1_, _1_) in
    Genr_utilities_scicos_pro.
    add_block_body_in_scsm0 reshape_name
      (genr_instructions_for_modeler_block_type scicos_block_type
         reshape_parameters
         [] Genr_utils.blk0 genr_info)
      nports
      (_3_, _0_) in

  let activation = genr_activation sample_time in

  let links = [
    Genr_utilities_scicos_pro.
    add_directed_link_in_scsm0
      buffer_name port_1_ reshape_name port_1_ [];
    add_directed_link_in_scsm0
      reshape_name port_1_ out_port_name port_1_ [];
    add_event_link_in_scsm0
      event_split_name port_1_ buffer_name port_1_ [];
    add_event_link_in_scsm0
      event_split_name port_2_ buffer_name port_2_ [];
  ] in

  let conditional_links =
    if sample_time = "-1"
    then [
      Genr_utilities_scicos_pro.
      add_directed_link_in_scsm0
        in_port_name port_1_ split_name port_1_ [];
      add_event_link_in_scsm0
        extract_activation_name port_1_ event_split_name port_1_ [];
      add_directed_link_in_scsm0
        split_name port_1_ buffer_name port_1_ [];
      add_directed_link_in_scsm0
        split_name port_2_ extract_activation_name port_1_ [];
    ]
    else [
      Genr_utilities_scicos_pro.
      add_directed_link_in_scsm0
        in_port_name port_1_ buffer_name port_1_ [];
      add_event_link_in_scsm0
        sample_clock_name port_1_ event_split_name port_1_ [];
    ] in

  let super_body = [
    in_port;
    split;
    out_port;
    reshape;
    buffer;
    activation;
    event_split;
    conditional_links;
    links;
  ] in

  super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
