(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_time_scope.ml,v 1.1 2019-05-10 14:04:58 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
;;

let mdl_block_type =
  simulink_block_type Mdl_block_types.TimeScope
;;

let num_in_ports_ident =
  Ast_types.ident_from_string "NumInputPorts"
and _y_min_ident =
  Ast_types.ident_from_string "YMin"
and _y_max_ident =
  Ast_types.ident_from_string "YMax"
and _time_range_ident =
  Ast_types.ident_from_string "TimeRange"
and _location_ident =
  Ast_types.ident_from_string "Location"
;;

let y_min_default = "-5"
and y_max_default = "5"
and time_range_default = "30"
and location_default = [| 100; 100; 400; 400; |]
;;

let genr block_id genr_info =

  let get_bound bound_string =
    let ib = Scanf.Scanning.from_string bound_string in
    let rec loop accu =
      if Scanf.Scanning.end_of_input ib then List.rev accu else
      try
        Scanf.bscanf ib "%s@~"
          (fun s -> loop (s :: accu)) with
      | Scanf.Scan_failure s
      | Failure s ->
        Say.fatal_error
          (Printf.sprintf
             "Genr_scope.genr: while scanning ``%s'', encountered ``%s''"
             bound_string s) in
    loop [] in

  let num_in_ports = get_ident_int genr_info num_in_ports_ident
  and y_min = get_bound y_min_default
  and y_max = get_bound y_max_default
  and time_range = time_range_default
  and location = location_default in

  let row_concat a b =
    Code_value (code_out_expression (Printf.sprintf "[ %s; %s ]" a b)) in

  let ybound = List.map2 row_concat y_min y_max in

  let block_name =
    genr_info.Genr_block_types.block_explicit_name in

  let nin_string = string_of_int num_in_ports in
  let nin_nat = Unatural.of_string nin_string
  and sample_time =
    Genr_utilities_scicos_pro.get_sample_time genr_info in

  let string_of_int_pair (x, y) =
    Printf.sprintf "[ %d; %d ]" x y in

  let per =
    code_out_expression
      (if time_range = "auto" then "-1" else time_range)
  and sz =
    string_of_int_pair
      (location.(2) - location.(0), location.(3) - location.(1))
  and og =
    string_of_int_pair (location.(0), location.(1)) in

  let in_names, in_ports =
    Genr_utilities_scicos_pro.
    in_ports_in_scsm0 nin_nat in

  let scicos_block_parameters ext = [
    Value_parameter (Si "nin",
      Code_value (code_out_expression nin_string));
    Value_parameter (Si "externalActivation",
      Code_value ext);
    Value_parameter (Si "winSize",
      Code_value (code_out_expression sz));
    Value_parameter (Si "winPos",
      Code_value (code_out_expression og));
    Value_parameter (Si "useSubplot",
      Code_value (code_out_bool true));
    Value_parameter (Si "winNam",
      Code_value
        (code_out_string (Ast_types.string_of_mdl_name  block_name)));
    Struct_parameter (Si "win_s",
       ((Si "xRefresh", [Code_value per]),
        [ Si "yInterval", ybound ]));
  ] in

  let scicos_basic_block_instructions blk_id ext =
    genr_instructions_for
      mdl_block_type (scicos_block_parameters ext)
      [] blk_id genr_info in

  match sample_time with
  | "0" | "-1" ->
    Concat_list.list
      (scicos_basic_block_instructions
         block_id
         external_activation_mode_off)

  | sample_time ->

    let sample_clock_name = "sample_clock"
    and scope_name = "timescope" in

    let sample_clock =
      Genr_utilities_scicos_pro.
      sample_clock_in_scsm0
        sample_clock_name
        (code_out_expression sample_time)
        (Unatural._1_, Unatural._1_) in

    let scope =
      let nports = (nin_nat, Unatural._0_) in
      Genr_utilities_scicos_pro.
      add_block_body_in_scsm0 scope_name
        (scicos_basic_block_instructions blk0 external_activation_mode_on)
        nports
        (Unatural._1_, Unatural._0_) in

    let in_links =
      Genr_utilities_scicos_pro.
      rev_map_i_start Unatural._1_
        (fun i in_port_name ->
         add_directed_link_in_scsm0
           in_port_name port_1_ scope_name
           (Genr_utilities_scicos_pro.mdl_port_number_of_natural i) [])
        in_names in

    let links = [
      Genr_utilities_scicos_pro.
      add_event_link_in_scsm0
        sample_clock_name port_1_ scope_name port_1_ [];
    ] in

    let in_ports = Concat_list.list_of_clist in_ports in

    let super_body = [
      in_ports;
      sample_clock;
      scope;
      in_links;
      links;
    ] in

    super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
