(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_trigger_and_enable_event_gen.ml,v 1.1 2019-05-10 14:04:58 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

(** Generate a super block for a Simulink sub system block that is actioned,
  meaning that is has an ActionPort block in its sub system. *)

open Instructions
open Genr_utilities_scicos_pro
open Unatural
open Genr_types
;;

type event_kind =
  | Rising
  | Falling
  | Either
  | FunctionCall
;;

let mdl_block_type_rising =
  Mdl_fake_block_types.trigger_and_enable_rising_event_generator_block_type
and mdl_block_type_falling =
  Mdl_fake_block_types.trigger_and_enable_falling_event_generator_block_type
and mdl_block_type_either =
  Mdl_fake_block_types.trigger_and_enable_either_event_generator_block_type
and mdl_block_type_function_call =
  Mdl_fake_block_types.trigger_and_enable_function_call_generator_block_type
;;

let mdl_block_type_and_scicos_argument = function
  | Rising -> mdl_block_type_rising,
    [
      Genr_types.
      Value_parameter (Si "edge",
        Code_value (code_out_string "- to +"));
      Value_parameter (Si "showport",
      Code_value (code_out_bool true));
    ]
  | Falling -> mdl_block_type_falling,
    [
      Genr_types.
      Value_parameter (Si "edge",
        Code_value (code_out_string "+ to -"));
      Value_parameter (Si "showport",
      Code_value (code_out_bool true));
    ]
   | Either -> mdl_block_type_either,
    [
      Genr_types.
      Value_parameter (Si "edge",
        Code_value (code_out_string  "Zcross"));
      Value_parameter (Si "showport",
      Code_value (code_out_bool true));
    ]
   | FunctionCall -> mdl_block_type_function_call, [ ]
;;

let genr event_kind block_id genr_info =
  let if_then_else_name = "If"
  and trigger_name = "Trigger"
  and split2_name = "split2"
  and in_port1_name = "in1"
  and in_port2_name = "in2"
  and out_port1_name = "out1"
  and out_port2_name = "out2"
  and out_port_event_name = "out_event" in

  let in_port1 =
    Genr_utilities_scicos_pro.
    in_port_in_scsm0 in_port1_name port_1_ (Unatural._0_, Unatural._2_)
  and in_port2 =
    in_port_in_scsm0 in_port2_name port_1_ (Unatural._0_, Unatural._1_)
  and out_port1 =
    Genr_utilities_scicos_pro.
    out_port_in_scsm0 out_port1_name port_1_ (Unatural._3_, Unatural._3_)
  and out_port2 =
    out_port_in_scsm0 out_port2_name port_1_ (Unatural._3_, Unatural._2_)
  and out_port_event =
    Genr_utilities_scicos_pro.
    out_port_event_in_scsm0 out_port_event_name port_1_ (_2_, _0_)
  and split2 =
    split_in_scsm0
      split2_name (_1_, _1_) in

  let mdl_block_type, scicos_block_parameters =
    mdl_block_type_and_scicos_argument event_kind in

  let if_then_else_block_parameters = [
    Genr_types.
    Value_parameter (Si "externalActivation",
      Genr_types.Code_value
        Genr_utilities_scicos_pro.external_activation_mode_on);
    Value_parameter (Si "nmod",
      Genr_types.Code_value (code_out_expression "0"));
  ] in

  let if_then_else_block_instructions blk_id =
    let sci_block_type = Sbt "system/ActivationOperations/IfThenElse" in
    genr_instructions_for_modeler_block_type
      sci_block_type if_then_else_block_parameters
      [] blk_id genr_info in

  let if_then_else =
    let nports = (_1_, _0_) in
    Genr_utilities_scicos_pro.
    add_block_body_in_scsm0
      if_then_else_name
      (if_then_else_block_instructions blk0)
      nports
      (_2_, _1_) in

  let scicos_block_instructions block_id =
    genr_instructions_for
      mdl_block_type scicos_block_parameters
      [] block_id genr_info in

  let trigger =
    let nports = (_1_, _0_) in
    add_block_body_in_scsm0
      trigger_name
      (scicos_block_instructions blk0)
      nports
      (_2_, _2_) in

  let links = [
    Genr_utilities_scicos_pro.
    add_directed_link_in_scsm0 in_port1_name port_1_ trigger_name port_1_ [];
    add_directed_link_in_scsm0 in_port2_name port_1_ split2_name port_1_ [];
    add_directed_link_in_scsm0 trigger_name port_1_ out_port1_name port_1_ [];
    add_directed_link_in_scsm0 split2_name port_1_ if_then_else_name port_1_ [];
    add_directed_link_in_scsm0 split2_name port_2_ out_port2_name port_1_ [];
    Genr_utilities_scicos_pro.
    add_event_link_in_scsm0
      trigger_name port_1_ if_then_else_name port_1_ [];
    add_event_link_in_scsm0
      if_then_else_name port_1_ out_port_event_name port_1_ [];
  ] in

  let super_body = [
    in_port1;
    out_port1;
    in_port2;
    out_port2;
    out_port_event;
    split2;
    trigger;
    if_then_else;
    links;
  ] in

  super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type_rising (genr Rising);;
add_translator mdl_block_type_falling (genr Falling);;
add_translator mdl_block_type_either (genr Either);;
add_translator mdl_block_type_function_call (genr FunctionCall);;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
