(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_trigger_event_gen.ml,v 1.1 2019-05-10 14:04:58 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

(** Generate a super block for a Simulink sub system block that is triggered,
  meaning that is has an TriggerPort block in its sub system. *)

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
open Unatural
;;

type event_kind =
  | Rising
  | Falling
  | Either
  | FunctionCall
;;

let mdl_block_type_rising =
  Mdl_fake_block_types.trigger_rising_event_generator_block_type
and mdl_block_type_falling =
  Mdl_fake_block_types.trigger_falling_event_generator_block_type
and mdl_block_type_either =
  Mdl_fake_block_types.trigger_either_event_generator_block_type
and mdl_block_type_function_call =
  Mdl_fake_block_types.trigger_function_call_generator_block_type
;;

let mdl_block_type_and_scicos_argument = function
  | Rising -> mdl_block_type_rising,
    [
      Genr_types.
      Value_parameter (Si "edge",
      Code_value (code_out_string "- to +"));
      Value_parameter (Si "showport",
      Code_value (code_out_bool true));
    ]
  | Falling -> mdl_block_type_falling,
    [
      Genr_types.
      Value_parameter (Si "edge",
      Code_value (code_out_string "+ to -"));
      Value_parameter (Si "showport",
      Code_value (code_out_bool true));
    ]
   | Either -> mdl_block_type_either,
    [
      Genr_types.
      Value_parameter (Si "edge",
      Code_value (code_out_string  "Zcross"));
      Value_parameter (Si "showport",
      Code_value (code_out_bool true));
    ]
   | FunctionCall -> mdl_block_type_function_call, [ ]
;;

let genr event_kind block_id genr_info =

  let trigger_name = "Trigger"
  and in_port_name = "in"
  and out_port_name = "out"
  and out_port_event_name = "out_event" in

  let in_port =
    Genr_utilities_scicos_pro.
    in_port_in_scsm0 in_port_name port_1_ (Unatural._0_, Unatural._1_) in
  let out_port =
    Genr_utilities_scicos_pro.
    out_port_in_scsm0 out_port_name port_1_ (Unatural._3_, Unatural._2_) in
  let out_port_event =
    Genr_utilities_scicos_pro.
    out_port_event_in_scsm0 out_port_event_name port_1_ (_2_, _0_) in

  let mdl_block_type, scicos_block_parameters =
    mdl_block_type_and_scicos_argument event_kind in

  let scicos_block_instructions block_id =
    genr_instructions_for
      mdl_block_type scicos_block_parameters
      [] block_id genr_info in

  let trigger =
    let nports = (_1_, _0_) in
    Genr_utilities_scicos_pro.add_block_body_in_scsm0
      trigger_name
      (scicos_block_instructions blk0)
      nports
      (_2_, _1_) in

  let links = [
    Genr_utilities_scicos_pro.
    add_directed_link_in_scsm0 in_port_name port_1_ trigger_name port_1_ [];
    Genr_utilities_scicos_pro.
    add_event_link_in_scsm0
      trigger_name port_1_ out_port_event_name port_1_ [];
  ] in

  let links =
    match event_kind with
      (* The trigger block for FunCall has no output ports *)
    | FunctionCall -> links
    | Rising | Falling | Either ->
      add_directed_link_in_scsm0
        trigger_name port_1_ out_port_name port_1_ [] ::
      links in

  let super_body = [
    in_port;
    out_port;
    out_port_event;
    trigger;
    links;
  ] in

  super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type_rising (genr Rising);;
add_translator mdl_block_type_falling (genr Falling);;
add_translator mdl_block_type_either (genr Either);;
add_translator mdl_block_type_function_call (genr FunctionCall);;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
