(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_trigonometry.ml,v 1.1 2019-05-10 14:04:58 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
open Unatural
;;

let mdl_block_type =
  Mdl_basics_explicit_ast.
  Explicit_block_regular_type Mdl_block_types.Trigonometry
;;

let operator_ident =
  Ast_types.ident_from_string "Operator"
;;

let genr block_id genr_info =

  let get_ident_string =
    Genr_utilities_scicos_pro.get_ident_string genr_info in

  let sample_time =
    Genr_utilities_scicos_pro.get_sample_time genr_info
  and operator = get_ident_string operator_ident in

  let scicos_block_parameters operator ext = [
    Value_parameter (Si "func",
      Code_value (code_out_string operator));
    Value_parameter (Si "externalActivation",
      Code_value ext)
  ] in

  let scicos_basic_block_instructions blk_id operator ext =
    genr_instructions_for
      mdl_block_type (scicos_block_parameters operator ext)
      [] blk_id genr_info in

  let atan2_block_parameters ext = [
    Value_parameter (Si "externalActivation",
      Code_value ext)
  ] in

  let atan2_block_instructions blk_id ext =
    let sci_block_type = Sbt "system/MathOperations/Atan2" in
    genr_instructions_for_modeler_block_type
      sci_block_type (atan2_block_parameters ext)
      [] blk_id genr_info in

  let in_port1_name = "in"
  and in_port2_name = "in2"
  and out_port1_name = "out"
  and out_port2_name = "out2"
  and split_name = "split"
  and event_split_name = "event_split"
  and sinop_name = "sinop"
  and cosop_name = "cosop"
  and atan2_name = "atan2"
  and sample_clock_name = "sample_clock" in

  let out_port_name = out_port1_name
  and in_port_name = in_port1_name in

  match sample_time with
  | "-1" ->
    begin match operator with
    | "atan2" ->
      Concat_list.list
        (atan2_block_instructions block_id
           external_activation_mode_off)

    | "sincos" ->

      let in_port =
        Genr_utilities_scicos_pro.
        in_port_in_scsm0
          in_port_name port_1_ (_0_, _0_) in

      let out_port1 =
        Genr_utilities_scicos_pro.
        out_port_in_scsm0
          out_port1_name port_1_ (_3_, _0_) in

      let out_port2 =
        Genr_utilities_scicos_pro.
        out_port_in_scsm0
          out_port2_name port_2_ (_3_, _1_) in

      let sinop =
        let nports = (_1_, _1_) in
        Genr_utilities_scicos_pro.
        add_block_body_in_scsm0
          sinop_name
          (scicos_basic_block_instructions blk0 "sin"
            external_activation_mode_off)
          nports
          (_2_, _0_) in

      let cosop =
        let nports = (_1_, _1_) in
        add_block_body_in_scsm0
          cosop_name
          (scicos_basic_block_instructions blk0 "cos"
            external_activation_mode_off)
          nports
          (_2_, _1_) in

      let split =
        Genr_utilities_scicos_pro.
        split_in_scsm0 split_name (_1_, _0_) in

      let links = [
        Genr_utilities_scicos_pro.
        add_directed_link_in_scsm0 in_port_name port_1_ split_name port_1_ [];
        add_directed_link_in_scsm0 split_name port_1_ sinop_name port_1_ [];
        add_directed_link_in_scsm0 split_name port_2_ cosop_name port_1_ [];
        add_directed_link_in_scsm0 sinop_name port_1_ out_port1_name port_1_ [];
        add_directed_link_in_scsm0 cosop_name port_1_ out_port2_name port_1_ [];
      ] in

      let super_body = [
        in_port;
        out_port1;
        out_port2;
        sinop;
        cosop;
        split;
        links;
      ] in

      super_block_in_scsm0 block_id super_body

    | operator ->
      Concat_list.list
        (scicos_basic_block_instructions block_id
           operator external_activation_mode_off)
    end

  | sample_time ->

    let in_port =
      Genr_utilities_scicos_pro.
      in_port_in_scsm0 in_port_name port_1_ (_0_, _0_) in
    let out_port =
      Genr_utilities_scicos_pro.
      out_port_in_scsm0 out_port_name port_1_ (_2_, _0_) in
    let sample_clock =
      Genr_utilities_scicos_pro.
      sample_clock_in_scsm0
        sample_clock_name
        (code_out_expression sample_time) (_1_, _2_) in

    begin match operator with
    | "atan2" ->

      let in_port2 =
        Genr_utilities_scicos_pro.
        in_port_in_scsm0 in_port2_name port_2_ (_0_, _1_) in

      let atan2 =
        let nports = (_2_, _1_) in
        add_block_body_in_scsm0
          atan2_name
          (atan2_block_instructions blk0 external_activation_mode_on)
          nports
          (_1_, _0_) in

      let links = [
        add_directed_link_in_scsm0 in_port_name port_1_ atan2_name port_1_ [];
        add_directed_link_in_scsm0 in_port2_name port_1_ atan2_name port_2_ [];
        add_directed_link_in_scsm0 atan2_name port_1_ out_port_name port_1_ [];
        Genr_utilities_scicos_pro.
        add_event_link_in_scsm0
          sample_clock_name port_1_ atan2_name port_1_ [];
      ] in

      let super_body = [
        in_port;
        in_port2;
        out_port;
        atan2;
        sample_clock;
        links;
      ] in

      super_block_in_scsm0 block_id super_body

    | "sincos" ->

      let out_port2 =
        Genr_utilities_scicos_pro.
        out_port_in_scsm0
          out_port2_name port_2_ (_3_, _1_) in

      let sinop =
        let nports = (_1_, _1_) in
        add_block_body_in_scsm0
          sinop_name
          (scicos_basic_block_instructions blk0 "sin"
            external_activation_mode_on)
          nports
          (_2_, _0_) in

      let cosop =
        let nports = (_1_, _1_) in
        add_block_body_in_scsm0
          cosop_name
          (scicos_basic_block_instructions blk0 "cos"
            external_activation_mode_off)
          nports
          (_2_, _1_) in

      let split =
        Genr_utilities_scicos_pro.
        split_in_scsm0
          split_name (_1_, _0_) in

      let event_split =
        Genr_utilities_scicos_pro.split_event_in_scsm0
          event_split_name (_1_, _1_) in

      let links = [
        add_directed_link_in_scsm0 in_port_name port_1_ split_name port_1_ [];
        add_directed_link_in_scsm0 split_name port_1_ sinop_name port_1_ [];
        add_directed_link_in_scsm0 split_name port_2_ cosop_name port_1_ [];
        add_directed_link_in_scsm0 sinop_name port_1_ out_port_name port_1_ [];
        add_directed_link_in_scsm0 cosop_name port_1_ out_port2_name port_1_ [];
        add_event_link_in_scsm0
          sample_clock_name port_1_ event_split_name port_1_ [];
        add_event_link_in_scsm0
          event_split_name port_1_ sinop_name port_1_ [];
        add_event_link_in_scsm0
          event_split_name port_2_ cosop_name port_1_ [];
      ] in

      let super_body = [
        in_port;
        out_port;
        out_port2;
        sinop;
        cosop;
        split;
        event_split;
        sample_clock;
        links;
      ] in

      super_block_in_scsm0 block_id super_body

    | operator ->

      let trigonometry_name = "trigonometry" in

      let trigonometry =
        let nports = (_1_, _1_) in
        add_block_body_in_scsm0
          trigonometry_name
          (scicos_basic_block_instructions blk0 operator
            external_activation_mode_on)
          nports
          (_1_, _0_) in

      let links = [
        add_directed_link_in_scsm0
          in_port_name port_1_ trigonometry_name port_1_ [];
        add_directed_link_in_scsm0
          trigonometry_name port_1_ out_port_name port_1_ [];
        add_event_link_in_scsm0
          sample_clock_name port_1_ trigonometry_name port_1_ [];
      ] in

      let super_body = [
        in_port;
        out_port;
        trigonometry;
        sample_clock;
        links;
      ] in

      super_block_in_scsm0 block_id super_body
    end
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
