(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_scicos_pro_xy_scope.ml,v 1.1 2019-05-10 14:04:58 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Instructions
open Genr_types
open Genr_utilities_scicos_pro
open Unatural
;;

let mdl_block_type =
  simulink_block_reference_type
    (Ast_types.mdl_name_of_string "XY scope.")
    (Ast_types.mdl_path_of_string "simulink/Sinks/XY Graph")
;;

let xmin_ident =
  Ast_types.ident_from_string "xmin"
and xmax_ident =
  Ast_types.ident_from_string "xmax"
and ymin_ident =
  Ast_types.ident_from_string "ymin"
and ymax_ident =
  Ast_types.ident_from_string "ymax"
and st_ident =
  Ast_types.ident_from_string "st"
;;

let genr block_id genr_info =

  let get_ident = Genr_utilities_scicos_pro.get_ident_string genr_info in

  let xmin = get_ident xmin_ident
  and xmax = get_ident xmax_ident
  and ymin = get_ident ymin_ident
  and ymax = get_ident ymax_ident in
  let col_concat a b =
    Code_value (code_out_expression (Printf.sprintf "[ %s, %s ]" a b)) in
  let xbound = col_concat xmin xmax in
  let ybound = col_concat ymin ymax in
  let st = get_ident st_ident
  and block_name = genr_info.Genr_block_types.block_explicit_name in

  let scicos_block_parameters ext = [
    Value_parameter
      (Si "externalActivation", Code_value ext);
    Value_parameter
      (Si "winNam",
       Code_value
         (code_out_string (Ast_types.string_of_mdl_name  block_name)));
    Value_parameter
      (Si "xInterval", xbound);
    Value_parameter
      (Si "yInterval", ybound);
  ] in

  let scicos_basic_block_instructions blk_id ext =
    genr_instructions_for
      mdl_block_type (scicos_block_parameters ext)
      [] blk_id genr_info in

  match st with
  | "-1" ->
    Concat_list.list
      (scicos_basic_block_instructions block_id external_activation_mode_off)

  | st ->

    let in_port1_name = "in1"
    and in_port2_name = "in2"
    and sample_clock_name = "sample_clock"
    and cscopxy_name = "cscopxy" in

    let in_port1 =
      Genr_utilities_scicos_pro.
      in_port_in_scsm0
        in_port1_name port_1_ (_0_, _1_) in
    let in_port2 =
      Genr_utilities_scicos_pro.
      in_port_in_scsm0
        in_port2_name port_2_ (_0_, _0_) in
    let cscopxy =
      let nports = (_2_, _0_) in
      Genr_utilities_scicos_pro.
      add_block_body_in_scsm0
        cscopxy_name
        (scicos_basic_block_instructions blk0 external_activation_mode_on)
        nports
        (_1_, _0_) in
    let sample_clock =
      Genr_utilities_scicos_pro.
      sample_clock_in_scsm0
        sample_clock_name (code_out_expression st) (_1_, _1_) in

    let links = [
      Genr_utilities_scicos_pro.
      add_directed_link_in_scsm0
        in_port1_name port_1_ cscopxy_name port_1_ [];
      add_directed_link_in_scsm0
        in_port2_name port_1_ cscopxy_name port_2_ [];
      Genr_utilities_scicos_pro.
      add_event_link_in_scsm0
        sample_clock_name port_1_ cscopxy_name port_1_ [];
    ] in

    let super_body = [
      in_port1;
      in_port2;
      cscopxy;
      sample_clock;
      links;
    ] in

    super_block_in_scsm0 block_id super_body
;;

add_translator mdl_block_type genr
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
