(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_utilities_pro.ml,v 1.1 2019-05-10 14:04:59 jpc Exp $ *)

(* Defensive programming: defining an empty Genr_utilities_scicos module.
  (See genr_scicos_pro_compare_to_constant.ml for more information.) *)
module Genr_utilities_scicos = struct end;;

open Genr_utilities_scicos_pro;;
open Genr_types;;

type size_array =
  | Size_array_vector of modeler_block_value list
  | Size_array_expression
;;

(* Computes the list of code value corresponding to the range of negative
   integers [-1 .. -n]. *)
let size_array_of_int n =
  let rec loop accu = function
    | 0 -> accu
    | n ->
      let v = Code_value (code_out_int n) in
      loop (v :: accu) (succ n) in
  loop [] (-n)
;;

(* Computes the number of entries in a vector expression. *)
let get_size_array block_type_name input_string =
  let context =
    Printf.sprintf "port size specifications in %s block" block_type_name in

  let scicos_value_of_string s = Code_value (code_out_expression s) in

  let expr_buff = Buffer.create 16 in

  let get_contents buff =
    let contents = Buffer.contents buff in
    Buffer.clear buff;
    contents in

  let enter_non_empty_expr accu =
    let expr = get_contents expr_buff in
    if String.length expr > 0 then expr :: accu else accu in

  let enter_expr accu =
    if Buffer.length expr_buff = 0 then
      Say.fatal_error (Printf.sprintf
        "Empty size specification in %s (%S)" context input_string) else
    enter_non_empty_expr accu in

  let rec enter_square_bracket accu ib =
    if Scanf.Scanning.end_of_input ib then
      Say.fatal_error (Printf.sprintf "Empty set of %s" context) else
    Scanf.bscanf ib "%c" (function
    | ' ' | ',' | ';' ->
      let accu = enter_expr accu in
      Scanf.bscanf ib "%_[ ]" enter_square_bracket accu ib
    | '[' (* Helping emacs ']' *) ->
      Say.fatal_error (Printf.sprintf "nested %C %s" '[' context)
    | ']' -> List.rev (enter_non_empty_expr accu)
    | c ->
      Buffer.add_char expr_buff c;
      enter_square_bracket accu ib) in

  let fallback_result = Size_array_expression in

  let return_result ib result =
    match Scanf.Scanning.end_of_input ib with
    | true -> Size_array_vector result
    | false -> fallback_result in

  let scan_port_size ib =
    Scanf.bscanf ib "%0c" (function
    | '0' .. '9' ->
      Scanf.bscanf ib "%d" (fun n ->
        match Scanf.Scanning.end_of_input ib with
        | true ->
          let sizes = size_array_of_int n in
          begin
            match sizes with
            | [] -> Say.fatal_error (Printf.sprintf "Empty set of %s" context)
            | sizes -> return_result ib sizes
          end
        | false -> fallback_result)
    | '[' (* Helping emacs ']' *) ->
      let exprs =
        Scanf.bscanf ib "%c " (fun _ -> enter_square_bracket [] ib) in
      return_result ib (List.map scicos_value_of_string exprs)
    | '{' (* Helping emacs ']' *) ->
      Genr_utilities_scicos_pro.warning (Printf.sprintf
        "get_size_array: unsupported cell array in expression %S in %s"
        input_string context);
      fallback_result
    | _ ->
      Say.fdebug (fun epf ->
        Format.fprintf epf "get_size_array: expression %S in %s"
          input_string context);
      fallback_result) in

  if input_string = "" then
    Say.fatal_error (Printf.sprintf "Empty set of %s" context) else

  scan_port_size (Scanf.Scanning.from_string input_string)
;;

(*
 Local Variables:
  compile-command: "cd ../../../..; make"
  End:
*)
