(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_rename.ml,v 1.1 2019-05-10 14:04:59 jpc Exp $ *)

(** {3 Renaming of links and bocks} *)

(** From [action_link] blocks to [rename_link] blocks.

  - Some block types should be renamed to classify original Simulink
  blocks to translate them into the target block model system (Scicos).
  This concerns in particular embedded matlab blocks (specifically
  S-Function defined via a stateflow embedded script).

  - In addition, this phase supplements the connection link information to
  compute its port numbers in the target block system.
*)

open Mdl_implementation_explicit_ast
open Mdl_rename_ast;;

let current_phase = "rename phase";;

let rename_message s =
  Format.sprintf "@[<hv 2>during %s,@ %s@]" current_phase s
;;

let rename_error s = Say.fatal_error (rename_message s)
and _rename_warning s = Say.warning (rename_message s)
and _rename_debug s = Say.debug (rename_message s)
;;

(* Building a hash table to retrieve blocks by their name
   (the block name is unique un any given system) *)
let make_block_by_names blocks =
  let block_by_names = Hashtbl.create 100 in
  List.iter
    (fun blk ->
     let block_name = blk.explicit_block_name in
     if not (Hashtbl.mem block_by_names block_name) then
       Hashtbl.add block_by_names block_name blk else
       rename_error (
         let Ast_types.Mdl_name blk_name = block_name in
         Printf.sprintf
         "duplicated block %s in sub-system"
         blk_name)
    )
    blocks;
   block_by_names
;;

(* Computing offsets of parameters of MATLAB Function blocks. *)
let get_arguments_offsets function_arguments =

  let deep_int i =
    Mdl_basics_deep_ast.Deep_int (string_of_int i) in

  let _, rev_inputs, tbl =
    let tbl = Hash_table.create Unatural._5_ in
    let classify_arguments (i, inputs, parameters) =
      fun matlab_function_argument ->
      match matlab_function_argument with
      | Explicit_matlab_function_input
          _explicit_matlab_function_data_desc ->
        (succ i, deep_int i :: inputs, parameters)
      | Explicit_matlab_function_parameter
          explicit_matlab_function_parameter_data_desc ->
        let index =
          explicit_matlab_function_parameter_data_desc.
          explicit_matlab_function_parameter_data_index in
        Hashtbl.add parameters index i;
        (succ i, inputs, parameters) in
    Array.fold_left classify_arguments
      (0, [], tbl)
      function_arguments in
  let parameters =
    let len = Hashtbl.length tbl in
    let rec loop i accu =
      if i < 0 then accu else
      let param =
        let idx = Unatural.of_int i in
        if Hashtbl.mem tbl idx then
           deep_int (Hashtbl.find tbl idx) else
        rename_error
          (Printf.sprintf "Cannot find parameter index %d" i) in
      loop (i - 1) (param :: accu) in
    loop (pred len) [] in
  List.rev rev_inputs, parameters
;;

let matlab_function_name_ident =
  Ast_types.ident_from_string "matlab_function_name"
and matlab_function_script_ident =
  Ast_types.ident_from_string "matlab_function_script"
and block_input_offsets_ident =
  Ast_types.ident_from_string "block_input_offsets"
and block_parameter_offsets_ident =
  Ast_types.ident_from_string "block_parameter_offsets"
and block_argument_sizes_ident =
  Ast_types.ident_from_string "block_argument_sizes"
and block_return_sizes_ident =
  Ast_types.ident_from_string "block_return_sizes"
(* FIXME: to be removed ?
and block_parameter_values_ident =
  Ast_types.ident_from_string "block_parameter_values"
*)
;;

let get_argument_size matlab_function_argument =
  match matlab_function_argument with
  | Explicit_matlab_function_input
      explicit_matlab_function_data_desc ->
    Mdl_basics_deep_ast.Deep_expression
      explicit_matlab_function_data_desc.
      explicit_matlab_function_data_size
  | Explicit_matlab_function_parameter
      explicit_matlab_function_parameter_data_desc ->
    Mdl_basics_deep_ast.Deep_expression
      explicit_matlab_function_parameter_data_desc.
      explicit_matlab_function_parameter_data_size
;;
let get_return_size matlab_function_return =
  match matlab_function_return with
  | Explicit_matlab_function_output explicit_matlab_function_data_desc ->
    Mdl_basics_deep_ast.Deep_expression
      explicit_matlab_function_data_desc.
      explicit_matlab_function_data_size
;;

let get_port_sizes function_arguments function_returns =
  Say.fdebug
    (fun epf ->
     Format.fprintf epf "get_port_sizes [|%a|] [|%a|]"
       (fun epf ->
        Array.iter (fun function_argument ->
          (Format.fprintf epf "%a; "
            Mdl_implementation_explicit_ast_print.
              print_explicit_matlab_function_argument
            function_argument)))
       function_arguments
       (fun epf ->
        Array.iter (fun function_argument ->
          (Format.fprintf epf "%a; "
            Mdl_implementation_explicit_ast_print.
         print_explicit_matlab_function_return
            function_argument)))
       function_returns);
  Array.to_list (Array.map get_argument_size function_arguments),
  Array.to_list (Array.map get_return_size function_returns)
;;

let get_block_name_nin block_by_names block_name =
  let eb = Hashtbl.find block_by_names block_name in
  Mdl_explicit_block_utils.get_block_regular_port_nin eb
;;

let get_block_name_nout block_by_names block_name =
  let eb = Hashtbl.find block_by_names block_name in
  Mdl_explicit_block_utils.get_block_regular_port_nout eb
;;

let make_rename_connection_port_info
    block_by_names block_name connection_port_value =

  let port_nin = get_block_name_nin block_by_names block_name in
  let port_nout = get_block_name_nout block_by_names block_name in
  {
    Mdl_rename_ast.
    connector_block_nin = port_nin;
    connector_block_nout = port_nout;
    connection_port_value = connection_port_value;
  }
;;

(* Simulink PMIO_port blocks have Rconn connection, when
   Scicos needs Lconn connections *)
let make_lconn1_connector block_by_names connector_port_desc block_name =
   match connector_port_desc with
   | Mdl_implementation_explicit_ast.
     Explicit_right_connector port_value ->
     Mdl_rename_ast.
     Rename_left_connector
       (make_rename_connection_port_info
          block_by_names block_name port_value)
   | Mdl_implementation_explicit_ast.
     Explicit_left_connector _port_value ->
     rename_error (Printf.sprintf
       "PMIOPort block %s has an Lconn connection"
       (Mdl_explicit_block_utils.string_of_block_name block_name))
;;

let flip_pmioport_block blk =
  (* Should change the implicit input/output port numbers
     when side is Left *)
  let blk =
    Mdl_explicit_block_utils.
    set_block_implicit_port_nin blk Unatural._1_ in
  let blk =
    Mdl_explicit_block_utils.
    set_block_implicit_port_nout blk Unatural._0_ in
  (* Flip the pmioport block *)
  Mdl_explicit_block_utils.map_block_mirror not blk
;;

let should_flip_pmioport_block action_block =
  let side =
    let deep_side_string_value =
      Mdl_explicit_block_utils.
      get_block_parameter_value action_block Mdl_ast_functions.side_ident in
    Mdl_basics_to_explicit_ast.string_of_deep_value
      "Port parameter Side should contain a string value"
      deep_side_string_value in
  (* PMIOPort to the right should be flipped change their mirror state *)
  match side with
  | "Left" -> false
  | "Right" -> true
  | s ->
    Say.fatal_error (Printf.sprintf "Unexpected port side '%s'" s)
;;

let is_pmioport_block explicit_simple_block_desc =
  let explicit_simple_block_type =
    explicit_simple_block_desc.explicit_simple_block_type in

  explicit_simple_block_type = Mdl_block_types.PMIOPort
;;

let is_pmioport_block_to_flip = function
  | {
      explicit_block_desc =
        Explicit_block_regular explicit_simple_block_desc;
      _;
    } as blk ->
    is_pmioport_block explicit_simple_block_desc &&
    should_flip_pmioport_block blk
  | {
      explicit_block_desc =
        Explicit_block_reference _ |
        Explicit_block_s_function _ |
        Explicit_block_sub_system _ ;
      _;
    } -> false
;;

let make_rename_link_desc block_by_names action_connection_connectors =
  match action_connection_connectors with
  | {
      Mdl_split_ast.
      split_link_src = split_connector_port_src;
      split_link_dst = split_connector_port_dst;
    } ->

   let rename_connection_connector split_connector_port =
     match split_connector_port with
     | {
         explicit_connector_port_block = block_name;
         explicit_connector_port_desc = connector_port_desc;
       } ->
       let block = Hashtbl.find block_by_names block_name in
       if is_pmioport_block_to_flip block then
         (* Should rename the connection link to connect to the Left *)
         make_lconn1_connector block_by_names connector_port_desc block_name
       else
       match connector_port_desc with
       | Mdl_implementation_explicit_ast.
         Explicit_right_connector port_value ->
         Mdl_rename_ast.
         Rename_right_connector
           (make_rename_connection_port_info
              block_by_names block_name port_value)
       | Mdl_implementation_explicit_ast.
         Explicit_left_connector port_value ->
         Rename_left_connector
           (make_rename_connection_port_info
              block_by_names block_name port_value) in

   let rename_connection_connector_src =
     {
       split_connector_port_src with
       explicit_connector_port_desc =
         rename_connection_connector split_connector_port_src;
     } in
   let rename_connection_connector_dst =
     {
       split_connector_port_dst with
       explicit_connector_port_desc =
         rename_connection_connector split_connector_port_dst;
     } in

   let rename_connection_connectors =
     {
        Mdl_split_ast.
        split_link_src = rename_connection_connector_src;
        split_link_dst = rename_connection_connector_dst;
      } in

   Mdl_rename_ast.
   Rename_connection_link rename_connection_connectors
;;

let make_rename_link block_by_names action_link =
  match action_link with
  | {
      Mdl_split_ast.
      split_link_name = _;
      split_link_font = _;
      split_link_labels = _;
      split_link_points = _;
      split_link_desc = action_link_desc;
    } ->
    let rename_link_desc =
      match action_link_desc with
      | Mdl_action_ast.
        Action_directed_link connectors ->
        Mdl_rename_ast.Rename_directed_link connectors
      | Mdl_action_ast.
        Action_event_link connectors ->
        Mdl_rename_ast.
        Rename_event_link connectors
      | Mdl_action_ast.
        Action_connection_link action_connection_connectors ->
        make_rename_link_desc block_by_names action_connection_connectors in

    {
      action_link with
      Mdl_split_ast.
      split_link_desc = rename_link_desc;
    }
;;

let rename_action_links block_by_names action_links =
  List.map (make_rename_link block_by_names) action_links
;;

let rec make_rename_blocks action_links action_blocks =

  (* Recurse in sub-systems to turn out all action blocks to rename blocks. *)
  let rename_blocks =
(*
    List.rev_map make_rename_block action_blocks in
*)
    List.map make_rename_block action_blocks in

  let block_by_names = make_block_by_names action_blocks in

  let rename_links =
    rename_action_links block_by_names action_links in

  rename_links, rename_blocks

and make_rename_block
  (action_block : Mdl_action_ast.action_block) :
  Mdl_rename_ast.rename_block =
  match action_block with
  | {
      explicit_block_desc =
        Explicit_block_regular explicit_simple_block_desc;
      _;
    } as blk ->
    if is_pmioport_block explicit_simple_block_desc &&
       should_flip_pmioport_block action_block
    then flip_pmioport_block blk
    else blk
  | {
      explicit_block_desc =
        Explicit_block_reference _;
      _;
    } as blk -> blk
  | {
      explicit_block_desc =
         Explicit_block_s_function explicit_s_function_desc;
      _;
    } as blk ->
    let explicit_s_function_type =
      explicit_s_function_desc.explicit_simple_block_type in
    begin match explicit_s_function_type with
    | Explicit_s_function -> blk
    | Explicit_s_function_stateflow
        (_explicit_stateflow_name, _explicit_stateflow_chart_file_number) ->
      let simulink_block_reference_type =
        Mdl_fake_block_types.
        s_function_stateflow_block_reference_type in
      let explicit_simple_block_desc =
        {
          explicit_s_function_desc with
          explicit_simple_block_type = simulink_block_reference_type;
        } in
      {
        blk with
        explicit_block_desc =
          Explicit_block_reference explicit_simple_block_desc;
      }
    | Explicit_matlab_function explicit_matlab_function_desc ->
      let simulink_block_reference_type =
        Mdl_fake_block_types.
        matlab_function_block_reference_type in

      let explicit_simple_block_desc =
        {
          explicit_s_function_desc with
          explicit_simple_block_type = simulink_block_reference_type;
        } in

      let parameters =
        match explicit_matlab_function_desc with
        | {
            explicit_matlab_function_name = function_name;
            explicit_matlab_function_script = function_script;
            explicit_matlab_function_arguments = function_arguments;
            explicit_matlab_function_returns = function_returns;
          } ->
          let input_offsets, parameter_offsets =
            get_arguments_offsets function_arguments
          and argument_sizes, return_sizes =
            get_port_sizes function_arguments function_returns in
          [
            matlab_function_name_ident,
              Mdl_basics_deep_ast.Deep_ident (
                Ast_types.ident_from_mdl_name function_name);
            matlab_function_script_ident,
(* FIXME: remove spurious conversions *)
             (match function_script with
              | Mdl_basics_deep_ast.
                Deep_source_code (Ast_types.Mdl_text s (*as deep_text*)) ->
                Mdl_basics_deep_ast.
                Deep_code (Ast_types.mdl_code_of_string s));
            block_input_offsets_ident,
              Mdl_basics_deep_ast.Deep_vector input_offsets;
            block_parameter_offsets_ident,
              Mdl_basics_deep_ast.Deep_vector parameter_offsets;
            block_argument_sizes_ident,
              Mdl_basics_deep_ast.Deep_vector argument_sizes;
            block_return_sizes_ident,
              Mdl_basics_deep_ast.Deep_vector return_sizes;

(*          block_parameter_values_ident,
              Mdl_basics_deep_ast.Deep_vector ; *)
          ] in

      Mdl_explicit_block_utils.add_block_parameters blk parameters;

      {
        blk with
        explicit_block_desc =
          Explicit_block_reference explicit_simple_block_desc;
      }
    end
  | {
      explicit_block_desc =
        Explicit_block_sub_system explicit_block_sub_system;
      _;
    } as blk ->
    let rename_block_desc =
      let rename_block_sub_system =
        make_rename_block_sub_system explicit_block_sub_system in
      Explicit_block_sub_system rename_block_sub_system in
    {
      blk with
      explicit_block_desc = rename_block_desc;
    }

and make_rename_block_sub_system ebss_desc =
  let ebss_system = ebss_desc.explicit_block_sub_system_body in
  let rename_system = make_rename_system ebss_system in
  let rename_activation =
    make_rename_activation
      ebss_desc.explicit_block_sub_system_activation in

  {
    ebss_desc with
    explicit_block_sub_system_activation = rename_activation;
    explicit_block_sub_system_body = rename_system;
  }

and make_rename_activation ea =
(* FIXME: should compute the relevant explicit_activation ??? *)
  match ea with
  | _ -> Sub_system_plain (* or Sub_system_activated some_activation *)

and make_rename_system
  (action_system : Mdl_action_ast.action_system) :
  Mdl_rename_ast.rename_system =
  let action_links =
    action_system.explicit_system_lines
  and action_blocks =
    action_system.explicit_system_blocks in

  let rename_links, rename_blocks =
    make_rename_blocks action_links action_blocks in

  {
    action_system with
    explicit_system_blocks = rename_blocks;
    explicit_system_lines = rename_links;
  }
;;

let rename_model_body action_model_body =
  {
    action_model_body with
    explicit_model_system =
      make_rename_system
        action_model_body.explicit_model_system;
  }
;;

let rename_model action_model =
  {
    action_model with
    explicit_model_body =
      rename_model_body action_model.explicit_model_body;
  }
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
