(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_rename_ast.mli,v 1.1 2019-05-10 14:04:59 jpc Exp $ *)

(** {3 Rename phase} *)

(** {6 Rename link kind models} *)

(*
  A model with rename link connections is a [rename_model] (as defined
  below).

  A model with rename link connections is the result of the activation
  handling treatment of an explicit link Simulink model (a
  [explicit_link_model] value).

  Rename links are made of directed connectors (as defined below as a
  [directed_connector poly_explicit_link]), either regular port to port
  connector or event connector.

  There are no more special port blocks in a [rename_model]:
  all port blocks (EnablePort, TriggerPort, ...) have been replaced by
  regular input ports. Activation of blocks is translated via extra external
  blocks and links. External blocks are event generators activating blocks
  through event links (event generators are introduced in the model as fake
  Simulink blocks).

  WIP: Rename should renumber the connection ports according
  to the following rule:

  LConni => nin + i
    where nin is the number of directed input ports of the destination block.
  RConni => nout + i
    where nout is the number of directed output ports of the source block.
*)

(** {6 Rename type definitions} *)

type natural = Unatural.t;;

(** {7 Borrowed from explicit phase} *)

type 'a explicit_block = 'a Mdl_implementation_explicit_ast.explicit_block
and 'a explicit_system = 'a Mdl_implementation_explicit_ast.explicit_system
and 'a explicit_model = 'a Mdl_implementation_explicit_ast.explicit_model
and 'a explicit_simulink_file = 'a Mdl_explicit_ast.explicit_simulink_file
;;

(** {7 Borrowed from action phase} *)

type 'a action_link_connectors = 'a Mdl_action_ast.action_link_connectors
and action_directed_connectors = Mdl_action_ast.action_directed_connectors
and action_event_connectors = Mdl_action_ast.action_event_connectors
;;

(** {7 Defining rename links *)

type 'link poly_rename_link = 'link Mdl_action_ast.poly_action_link;;

(* TORM
(** For the time being,
  the rename phase does not modify the type of links.
  However, to ensure safety (rename phase must be run on all input Simulink
  models), we should make this type private. *)
module Private_link_desc : sig
  type rename_link_desc = private Mdl_action_ast.action_link_desc;;
  val rename_link_desc :
    Mdl_action_ast.action_link_desc -> rename_link_desc
  ;;
end;;
*)

type connection_port_info = {
  connector_block_nin : natural;
  connector_block_nout : natural;
  connection_port_value :
    Mdl_implementation_explicit_ast.connection_port_value;
}
;;

type rename_connection_connector_desc =
  | Rename_left_connector of connection_port_info
  | Rename_right_connector of connection_port_info
(** Provide information to renumber connection line connector ports *)
;;

type rename_directed_connector = Mdl_action_ast.action_directed_connector
(* = action_directed_connector_desc
       Mdl_action.action_connector_port
   = action_directed_connector_desc
       Mdl_implementation_explicit_ast.explicit_connector_port
   =
     {
       explicit_connector_port_block : block_name;
       explicit_connector_port_desc : mdl_port_number;
     }
 *)
(*and rename_event_connector = Mdl_action_ast.action_event_connector*)
and rename_connection_connector =
  rename_connection_connector_desc Mdl_action_ast.action_connector_port
(*
   =
     {
       explicit_connector_port_block : block_name;
       explicit_connector_port_desc : rename_connection_connector_desc;
     }

rename_connection_connector_desc =
  | Rename_left_connector of connection_port_info
  | Rename_right_connector of connection_port_info

connection_port_info = {
  connector_block_nin : natural;
  connector_block_nout : natural;
  connection_port_value :
    Mdl_implementation_explicit_ast.connection_port_value;
}
*)
;;

type rename_connection_connectors =
  rename_connection_connector_desc action_link_connectors
;;

type rename_directed_connectors = Mdl_action_ast.action_directed_connectors
and rename_event_connectors = Mdl_action_ast.action_event_connectors
;;

type rename_link_desc =
  | Rename_directed_link of rename_directed_connectors
    (* The same as an Action_directed_link *)
  | Rename_event_link of rename_event_connectors
    (* The same as an Action_event_link *)
  | Rename_connection_link of rename_connection_connectors
    (* We must change the descriptor of connection connectors to provide
       information for renumbering connection ports *)
;;

type rename_links = rename_link list
and rename_link = rename_link_desc poly_rename_link
;;

(** {7 Defining rename models *)

type rename_block = rename_link explicit_block
and rename_system = rename_link explicit_system
and rename_model = rename_link explicit_model
and rename_simulink_file = rename_link explicit_simulink_file
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
