(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_fake_block_types.mli,v 1.1 2019-05-10 14:04:01 jpc Exp $ *)

(**
  We sometimes need to insert new blocks in a Simulink model in order to
  faithfully translate the model into Scicos. Those blocks are more or less
  Scicos blocks maskeraded into Simulink blocks (e.g. we add Scicos
  split blocks to translate Simulink line branches, or add Scicos event
  producer blocks to translate a Simulink block of action).

  This module defines the reference types of these additional blocks.
*)

(** {3 Building blocks in explicit syntax} *)

(** Some additional block explicit types *)

(** We introduce explicit block types for blocks we need to insert in the
  Simulink model in order to faithfully translate them to Scicos.

  All those types are fake Simulink block types that uniquely identify those
  auxiliary blocks. *)

(** Some Simulink blocks may not have any translation scheme written yet.
  In this case, we generate an empty Scicos super-block in place of the
  original Simulink block. This generated super-block has the same name as
  the original Simulink block (plus a suffix mark as given by
  [not_yet_translated_block_mark]) and the list of original block arguments
  is given as comments in the super-block context.

  The generator for this super-block is defined as any other translator;
  hence, it should be associated with a Simulink block type.
  We define here this new Simulink reference block type. *)

val not_yet_translated_block_type :
  Mdl_basics_explicit_ast.explicit_block_type
(** The explicit type of a block that has no translation *)
;;

val not_yet_translated_block_mark : string
(** The suffix written in the name of the super-block place holder that
  replaces a Simulink block that has no translation yet. *)
;;

val split_block_reference_type :
  Mdl_basics_explicit_ast.explicit_block_reference_type
(** The reference type of the fake Simulink split block that implements
  Simulink Branches. *)
;;

val split_block_type :
  Mdl_basics_explicit_ast.explicit_block_type
(** The explicit type of the Scicos Split block that implements Simulink
  Branches. *)
;;

val split_connection_block_reference_type :
  Mdl_basics_explicit_ast.explicit_block_reference_type
(** The reference type of the fake Simulink split block that implements
  Simulink Branches of connection lines. *)
;;

val split_connection_block_type :
  Mdl_basics_explicit_ast.explicit_block_type
(** The explicit type of the Scicos Split block that implements Simulink
  Branches of connection lines. *)
;;

val s_function_stateflow_block_reference_type :
  Mdl_basics_explicit_ast.explicit_block_reference_type
;;
val s_function_stateflow_block_type :
  Mdl_basics_explicit_ast.explicit_block_type
(** The explicit type of a Simulink S-Function block obtained via the
  Level-2 MATLAB S-Function user-defined functions palette entry.
  The function to call at runtime is not given in the model. The eml part of
  the chart only contains the name of the function to call.
  The Scicos translation should suppose hat the user has already translated,
  compiled, and imported (via some foreign interface) the correspondiong code
  in the Scicos embedding language. *)
;;

val matlab_function_block_reference_type :
  Mdl_basics_explicit_ast.explicit_block_reference_type
;;
val matlab_function_block_type :
  Mdl_basics_explicit_ast.explicit_block_type
(** The explicit type of a Simulink S-Function block obtained via the MATLAB
  Function block in the Simulink user-defined functions palette.
  This block defines a script embedded in the stateflow part of the model.
  This script should be translated into the Scicos embedding language, and
  the Scicos equivalent block should be the Scicos block that interprets
  scripts of the embedding language. *)
;;

val atomic_port_block_reference_type :
  Mdl_basics_explicit_ast.explicit_block_reference_type
;;

val atomic_port_block_type :
  Mdl_basics_explicit_ast.explicit_block_type
(** The explicit type of the Scicos event port block that implements Simulink
  atomic sub systems with sample time <> -1. *)
;;

(** Event block generator explicit block types *)

val atomic_event_generator_block_reference_type :
  Mdl_basics_explicit_ast.explicit_block_reference_type
;;

val atomic_event_generator_block_type :
  Mdl_basics_explicit_ast.explicit_block_type
(** Event generator block type for atomic sub-system blocks. *)
;;

val enable_event_generator_block_reference_type :
  Mdl_basics_explicit_ast.explicit_block_reference_type
;;

val enable_event_generator_block_type :
  Mdl_basics_explicit_ast.explicit_block_type
(** Event generator block type for enabled blocks. *)
;;

val if_action_event_generator_block_reference_type :
  Mdl_basics_explicit_ast.explicit_block_reference_type
;;

val if_action_event_generator_block_type :
  Mdl_basics_explicit_ast.explicit_block_type
(** Event generator block type for if_action blocks. *)
;;

val trigger_either_event_generator_block_reference_type :
  Mdl_basics_explicit_ast.explicit_block_reference_type
val trigger_falling_event_generator_block_reference_type :
  Mdl_basics_explicit_ast.explicit_block_reference_type
val trigger_rising_event_generator_block_reference_type :
  Mdl_basics_explicit_ast.explicit_block_reference_type
val trigger_function_call_generator_block_reference_type :
  Mdl_basics_explicit_ast.explicit_block_reference_type
;;

val trigger_either_event_generator_block_type :
  Mdl_basics_explicit_ast.explicit_block_type
val trigger_falling_event_generator_block_type :
  Mdl_basics_explicit_ast.explicit_block_type
val trigger_rising_event_generator_block_type :
  Mdl_basics_explicit_ast.explicit_block_type
val trigger_function_call_generator_block_type :
  Mdl_basics_explicit_ast.explicit_block_type
(** Event generator block types for triggered blocks of action. *)
;;

val trigger_and_enable_either_event_generator_block_reference_type :
  Mdl_basics_explicit_ast.explicit_block_reference_type
val trigger_and_enable_falling_event_generator_block_reference_type :
  Mdl_basics_explicit_ast.explicit_block_reference_type
val trigger_and_enable_rising_event_generator_block_reference_type :
  Mdl_basics_explicit_ast.explicit_block_reference_type
val trigger_and_enable_function_call_generator_block_reference_type :
  Mdl_basics_explicit_ast.explicit_block_reference_type
;;

val trigger_and_enable_either_event_generator_block_type :
  Mdl_basics_explicit_ast.explicit_block_type
val trigger_and_enable_falling_event_generator_block_type :
  Mdl_basics_explicit_ast.explicit_block_type
val trigger_and_enable_rising_event_generator_block_type :
  Mdl_basics_explicit_ast.explicit_block_type
val trigger_and_enable_function_call_generator_block_type :
  Mdl_basics_explicit_ast.explicit_block_type
(** Event generator block types for triggered and enabled blocks of action. *)
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
