(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_make_block.ml,v 1.1 2019-05-10 14:04:01 jpc Exp $ *)

(* From explicit syntax with Simulink-style lines to
   splitting blocks witnh Scicos-style lines. *)

open Ast_types
open Mdl_basics_deep_ast
open Mdl_implementation_explicit_ast
;;

(** {3 Simple block generation: splits, connection splits, and
       event generator blocks} *)

let dummy_explicit_font = {
  Mdl_basics_explicit_ast.
  explicit_font_name = "Helvetica";
  explicit_font_angle = Angle_oblique;
  explicit_font_size = Unatural_types.to_natural 12;
  explicit_font_weight = Weight_bold;
}
;;

let dummy_position = {
  xmin = 0;
  ymin = 0;
  xmax = 0;
  ymax = 0;
}
;;

let make_explicit_simple_block_desc explicit_simple_block_type =
  {
    explicit_simple_block_type = explicit_simple_block_type;
    explicit_simple_block_activation = Simple_block_plain;
  }
;;

let make_simple_block
  block_bg_color block_fg_color
  explicit_simple_block_type
  block_parameters
  block_mirror block_position
  line_name_opt
  block_name
  block_number_of_regular_ports
  block_number_of_implicit_ports =

  let block_ports_desc =
    Mdl_ast_to_deep_ast.
    make_default_deep_ports_description
      block_number_of_regular_ports
      block_number_of_implicit_ports in

  let block_ports =
    match line_name_opt with
    | None -> []
    | Some _ ->
      let block_port i = {
        Mdl_basics_deep_ast.
        deep_block_port_number = mdl_port_number_of_natural i;
        deep_block_port_name = line_name_opt;
        deep_block_port_bindings = [];
      } in
      let rec loop accu = function
        | i when i = Unatural._0_ -> accu
        | i ->
          loop (block_port i :: accu) (Unatural.pred i) in
      loop [] block_number_of_regular_ports.port_nout in

  let block_graphical = {
    Mdl_implementation_explicit_ast.
    explicit_block_background_color = block_bg_color;
    explicit_block_foreground_color = block_fg_color;
    explicit_block_mirror = block_mirror;
    explicit_block_ports = block_ports;
    explicit_block_ports_description = block_ports_desc;
    explicit_block_position = block_position;
    explicit_block_rotation = Unatural._0_;
    explicit_block_show_name = false;
  } in

  let block_optional_graphical = {
    Mdl_implementation_explicit_ast.
    explicit_block_drop_shadow = None;
    explicit_block_font = dummy_explicit_font;
    explicit_block_name_placement = None;
    explicit_block_state = None;
    explicit_block_tag = None;
    explicit_block_hide_automatic_name = None;
  } in

  let block_binding_table = Hashtbl.create 1 in

  let block_bindings = {
    Mdl_implementation_explicit_ast.
    explicit_block_parameters = block_parameters;
    explicit_block_bindings = block_binding_table;
    explicit_block_graphical = block_graphical;
    explicit_block_optional_graphical = block_optional_graphical;
  } in

  let block_body = {
    explicit_block_all_bindings = block_bindings;
    explicit_block_array = None;
    explicit_block_object = None;
    explicit_block_mask = None;
    explicit_block_sf_block_type = None;
    explicit_block_instance_data = None;
  } in

  let block_desc =
    let explicit_simple_block_desc =
      make_explicit_simple_block_desc explicit_simple_block_type in
    Explicit_block_reference explicit_simple_block_desc in

  {
    Mdl_implementation_explicit_ast.
    explicit_block_name = block_name;
    explicit_block_sid = None;
    explicit_block_body = block_body;
    explicit_block_known = true;
    explicit_block_desc = block_desc;
  }
;;

let empty_block_parameters = Hashtbl.create 1;;

let make_simple_block_without_parameter bg fg block_type =
  make_simple_block bg fg block_type empty_block_parameters
;;

(* Generator for split blocks to replace branches of directed lines. *)
let null_block_number_of_regular_ports =
  {
    port_nin = Unatural._0_;
    port_nout = Unatural._0_;
  }
;;

let null_block_number_of_connection_ports =
  {
    port_niin = Unatural._0_;
    port_niout = Unatural._0_;
  }
;;

(* Generator for event-generator blocks. *)
let mk_event_generator_block =
  let mk =
    make_simple_block Deep_white Deep_red in
  (fun explicit_simple_block_type
       block_parameters
       block_mirror block_position
       line_name_opt
       block_name
       block_number_of_regular_ports ->
   mk
     explicit_simple_block_type
     block_parameters
     block_mirror block_position
     line_name_opt
     block_name
     block_number_of_regular_ports
     null_block_number_of_connection_ports)
;;

let make_split_directed_block =
  let mk =
    make_simple_block_without_parameter Deep_black Deep_black
      Mdl_fake_block_types.split_block_reference_type
      false dummy_position in
  (fun line_name_opt
       block_name
       block_number_of_regular_ports ->
    mk line_name_opt block_name
      block_number_of_regular_ports null_block_number_of_connection_ports)
;;

(* Generator for connection split blocks to replace branches of connection
   lines. *)
let make_split_connection_block =
  let mk =
    make_simple_block_without_parameter Deep_blue Deep_blue
      Mdl_fake_block_types.split_connection_block_reference_type
      false in
  (fun block_position
     line_name_opt
     block_name ->
     mk
       block_position
       line_name_opt
       block_name
       null_block_number_of_regular_ports)
;;

(* Generating port blocks to be placed inside original sub-system blocks. *)
(* An [Atomic_port_block] has not ports: no regular not connection ports. *)
let make_atomic_port_block =
  let mk =
    make_simple_block_without_parameter Deep_white Deep_red
      Mdl_fake_block_types.atomic_port_block_reference_type
      false in
   (fun block_position
      line_name_opt
      block_name ->
    mk
      block_position
      line_name_opt
      block_name
      null_block_number_of_regular_ports
      null_block_number_of_connection_ports)
;;

(* Generating event-generator blocks. *)

let make_atomic_event_generator_block parameters =
  mk_event_generator_block
    Mdl_fake_block_types.atomic_event_generator_block_reference_type
    parameters
;;

let make_trigger_rising_event_generator_block =
  mk_event_generator_block
    Mdl_fake_block_types.trigger_rising_event_generator_block_reference_type
and make_trigger_falling_event_generator_block =
  mk_event_generator_block
    Mdl_fake_block_types.trigger_falling_event_generator_block_reference_type
and make_trigger_either_event_generator_block =
  mk_event_generator_block
    Mdl_fake_block_types.trigger_either_event_generator_block_reference_type
and make_trigger_function_call_generator_block =
  mk_event_generator_block
    Mdl_fake_block_types.trigger_function_call_generator_block_reference_type
;;

let make_enable_event_generator_block =
  mk_event_generator_block
    Mdl_fake_block_types.enable_event_generator_block_reference_type
;;

let make_if_action_event_generator_block =
  mk_event_generator_block
    Mdl_fake_block_types.if_action_event_generator_block_reference_type
;;

let make_trigger_and_enable_rising_event_generator_block =
  mk_event_generator_block
    Mdl_fake_block_types.
    trigger_and_enable_rising_event_generator_block_reference_type
and make_trigger_and_enable_falling_event_generator_block =
  mk_event_generator_block
    Mdl_fake_block_types.
    trigger_and_enable_falling_event_generator_block_reference_type
and make_trigger_and_enable_either_event_generator_block =
  mk_event_generator_block
    Mdl_fake_block_types.
    trigger_and_enable_either_event_generator_block_reference_type
and make_trigger_and_enable_function_call_generator_block =
  mk_event_generator_block
    Mdl_fake_block_types.
    trigger_and_enable_function_call_generator_block_reference_type
;;

let make_event_generator_block
    egb_ty egb_parameters
    egb_mirror egb_position
    egb_name egb_number_of_regular_ports =
  mk_event_generator_block
    egb_ty egb_parameters
    egb_mirror egb_position
    None
    egb_name egb_number_of_regular_ports
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
