(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_make_block.mli,v 1.1 2019-05-10 14:04:01 jpc Exp $ *)

(** {3 Building blocks in explicit syntax} *)

(**
  We sometimes need to insert new blocks in a Simulink model in order to
  faithfully translate the model into Scicos. Those blocks are more or less
  Scicos blocks maskeraded into Simulink blocks (e.g. we add Scicos
  split blocks to translate Simulink line branches, or add Scicos event
  producer blocks to translate a Simulink block of action).

  This module defines helper functions to build these additional block
  instances.
*)

open Mdl_make_block_types;;

(** {6 Building mdl blocks} *)

(** We provide here functions to build blocks that are mandatory to translate
  a Simulink model into a Scicos model.
  We need to insert Scicos split blocks to translate Simulink lines with
  branches into proper Scicos links.

  In the same vein, we need to add event generator blocks to mimmick Simulink
  blocks of action with so-called action lines into Scicos super blocks with
  event links.

  The new blocks here are directly explicit blocks, ready to be compiled:
  there is no treatment nor analysis to perform on them since all the data
  of these new blocks are already completely explicit.
*)

val dummy_explicit_font : Mdl_basics_explicit_ast.explicit_font
(** A fake font for our generated blocks. *)
;;
val dummy_position : Mdl_basics_deep_ast.deep_position
(** A fake position for our generated blocks. *)
;;

val make_split_directed_block : 'a regular_block_generator;;
val make_split_connection_block :
  Mdl_basics_explicit_ast.explicit_position ->
  'a implicit_block_generator
;;

val make_atomic_port_block :
  Mdl_basics_explicit_ast.explicit_position ->
  Mdl_basics_deep_ast.deep_port_name option ->
  Mdl_implementation_explicit_ast.explicit_block_name ->
  'a Mdl_implementation_explicit_ast.explicit_block
;;

val make_atomic_event_generator_block :
  'a event_generator_block_generator
;;

val make_trigger_rising_event_generator_block :
  'a event_generator_block_generator
;;
val make_trigger_falling_event_generator_block :
  'a event_generator_block_generator
;;
val make_trigger_either_event_generator_block :
  'a event_generator_block_generator
;;
val make_trigger_function_call_generator_block :
  'a event_generator_block_generator
;;

val make_trigger_and_enable_rising_event_generator_block :
  'a event_generator_block_generator
;;
val make_trigger_and_enable_falling_event_generator_block :
  'a event_generator_block_generator
;;
val make_trigger_and_enable_either_event_generator_block :
  'a event_generator_block_generator
;;
val make_trigger_and_enable_function_call_generator_block :
  'a event_generator_block_generator
;;

val make_enable_event_generator_block :
  'a event_generator_block_generator
;;

val make_if_action_event_generator_block :
  'a event_generator_block_generator
;;

val empty_block_parameters :
  (Ident.ident,
   Mdl_basics_explicit_ast.explicit_value Binding.ident_binding) Hashtbl.t
;;

val make_event_generator_block :
  Mdl_basics_explicit_ast.explicit_block_reference_type ->
  Mdl_implementation_explicit_ast.explicit_binding_table ->
  Mdl_basics_explicit_ast.explicit_mirror ->
  Mdl_basics_explicit_ast.explicit_position ->
  Mdl_ast.block_name ->
  Mdl_basics_explicit_ast.explicit_number_of_regular_ports ->
  'a Mdl_implementation_explicit_ast.explicit_block
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
