(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_split.ml,v 1.1 2019-05-10 14:04:01 jpc Exp $ *)

(* From explicit syntax with Simulink-style lines to
   Scicos-style links with split blocks. *)

open Mdl_basics_deep_ast
open Mdl_implementation_explicit_ast
open Mdl_split_ast
;;

type split_connection_side =
  | Destination
  | Source of
      Mdl_implementation_explicit_ast.explicit_absolute_point option
;;

type template_line_directed_body =
  {
    template_line_directed_name : line_name option;
    template_line_directed_font : explicit_font;
  }
;;

type template_line_connection_body =
  {
    template_line_connection_name : line_name option;
    template_line_connection_font : explicit_font;
    template_line_connection_labels : labels_option;
  }
;;

let make_template_line_directed_body = function
  | {
      explicit_line_name = line_name_option;
      explicit_line_connectors = connectors;
      explicit_line_font = explicit_font;
      explicit_line_labels = labels_option;
      explicit_line_points = points_option;
    } ->
    {
      template_line_directed_name = line_name_option;
      template_line_directed_font = explicit_font;
    },
    connectors,
    labels_option,
    points_option
;;

let make_template_line_connection_body = function
  | {
      explicit_line_name = line_name_option;
      explicit_line_connectors = connectors;
      explicit_line_font = explicit_font;
      explicit_line_labels = labels_option;
      explicit_line_points = points_option;
    } ->
    {
      template_line_connection_name = line_name_option;
      template_line_connection_font = explicit_font;
      template_line_connection_labels = labels_option;
    },
    connectors,
    points_option
;;

(* Builds an explicit link from the body of an explicit line. *)
let make_split_link_directed
    template_line_body link_desc labels_option points_option =
  match template_line_body with
  | {
      template_line_directed_name = line_name_option;
      template_line_directed_font = explicit_font;
    } ->
    {
      Mdl_split_ast.
      split_link_name = line_name_option;
      split_link_font = explicit_font;
      split_link_labels = labels_option;
      split_link_points = points_option;
      split_link_desc = link_desc;
    }
;;

let make_split_link_connection template_line_body link_desc points_option =
  match template_line_body with
  | {
      template_line_connection_name = line_name_option;
      template_line_connection_font = explicit_font;
      template_line_connection_labels = labels_option;
    } ->
    {
      Mdl_split_ast.
      split_link_name = line_name_option;
      split_link_font = explicit_font;
      split_link_labels = labels_option;
      split_link_points = points_option;
      split_link_desc = link_desc;
    }
;;

(* Borrowed from Ast_types module. *)
let mdl_port_number_of_natural = Ast_types.mdl_port_number_of_natural;;
let string_of_mdl_name = Ast_types.string_of_mdl_name;;
let ident_from_string = Ast_types.ident_from_string;;
let mdl_name_of_string = Ast_types.mdl_name_of_string;;

(* Borrowed from Mdl_make_block module. *)
let dummy_explicit_font = Mdl_make_block.dummy_explicit_font;;
let dummy_position = Mdl_make_block.dummy_position;;

(* Messages and errors for this module *)

let split_message s = Printf.sprintf "Mdl_split: %s" s;;

let split_error s = Say.fatal_error (split_message s)
and split_warning s = Say.warning (split_message s)
and split_debug s = Say.debug (split_message s)
;;

let split_deep_zorder_dropped = function
  | None -> ()
  | Some z ->
    split_debug (Format.sprintf
      "branch zorder option dropped (%d)" z)
;;

(* Functions in this file use an accumulator made of three compoonents:
   - an integer accumulator for the SID of split blocks introduced during
     branch elimination,
   - an accumulator for split blocks introduced during branch elimination,
   - an accumulator for links computed during the branch elimination process.

   We provide simple functions to manipulate the accumulator:
   adding a new link, adding a new split, or incrementing the SID integer.
*)
let add_link (i, accu_links, accu_splits) link =
  i, link :: accu_links, accu_splits
(* Accumulator manipulation: adding a new split block to the accumulator *)
and add_split (i, accu_links, accu_splits) split =
  i, accu_links, split :: accu_splits
(* Accumulator manipulation: adding one to the block SID register *)
and next_block_id (i, accu_links, accu_splits) =
  i + 1, accu_links, accu_splits
;;

let get_split_directed (i, _, _ as accu) line_name_opt nout =
  let split_directed_block_name =
    mdl_name_of_string (Printf.sprintf "_S_p_l_i_t_%d" i) in
  let explicit_number_of_regular_ports =
    Mdl_ast_to_deep_ast.
    make_number_of_regular_ports Unatural._1_ nout in
  next_block_id accu,
  split_directed_block_name,
  Mdl_make_block.
  make_split_directed_block
    line_name_opt split_directed_block_name
    explicit_number_of_regular_ports
;;

(* TORM
(* A connection line split has one left and one right ports *)
let default_connection_split_number_of_implicit_ports =
  Mdl_ast_to_deep_ast.
  make_number_of_implicit_ports Unatural._1_ Unatural._1_
;;
*)

let get_connection_split_source, get_connection_split_target =

  let get_connection_split (i, _, _ as accu)
      line_name_opt niin niout split_position =
    let split_connection_block_name =
      mdl_name_of_string (Printf.sprintf "_C_S_p_l_i_t_%d" i) in
    let explicit_number_of_implicit_ports =
      Mdl_ast_to_deep_ast.
      make_number_of_implicit_ports niin niout in
    next_block_id accu,
    split_connection_block_name,
    Mdl_make_block.
    make_split_connection_block
      split_position
      line_name_opt split_connection_block_name
      explicit_number_of_implicit_ports in

  let default_port_count = Unatural._1_ in
  (fun accu line_name_opt niin split_position ->
   get_connection_split accu
     line_name_opt niin default_port_count split_position),
  (fun accu line_name_opt niout split_position ->
   get_connection_split accu
     line_name_opt default_port_count niout split_position)
;;

let get_split_shadow block_name =
  let explicit_number_of_regular_ports =
    Mdl_ast_to_deep_ast.
    make_number_of_regular_ports Unatural._1_ Unatural._2_ in
  Mdl_make_block.make_split_directed_block
    None block_name
    explicit_number_of_regular_ports
;;

(** {3 Explicitation of all lines/branches into links/split_blocks} *)

let make_connector_port block_name descriptor =
  {
    Mdl_implementation_explicit_ast.
    explicit_connector_port_block = block_name;
    explicit_connector_port_desc = descriptor;
  }
;;

let make_explicit_connector_port block_name port_number =
  let descriptor =
    Mdl_ast.Port_number (mdl_port_number_of_natural port_number) in
  make_connector_port block_name descriptor
;;

let explicit_directed_connector_of_deep_connector = function
  | {
      Mdl_ast.
      line_connector_block = block_name;
      line_connector_port = line_connector_port;
    } ->
    make_connector_port block_name line_connector_port
;;

let string_of_activation_port = function
  | Mdl_split_ast.Enable_port -> "Enable_port"
  | Trigger_port -> "Trigger_port"
  | Enable_trigger_port -> "Enable_trigger_port"
  | Trigger_enable_port -> "Trigger_enable_port"
  | If_action_port -> "If_action_port"
;;

(* Defining a polymorphic activation_port finding function for
   block_block_descriptors: this way we can share the code for regular and
   reference blocks that need activation lines, when their respective
   descriptors do not have the same type. *)
let activation_port_of_block_block_descriptor
    activation_port explicit_simple_block_desc =
  let explicit_block_activation =
    explicit_simple_block_desc.explicit_simple_block_activation in
  match explicit_block_activation with
  | Simple_block_plain ->
    split_error (Printf.sprintf
      "attempt to get activation port %s \
       from a block that is a Plain reference"
      (string_of_activation_port activation_port))
  | Simple_block_activated simple_block_activation ->
    match simple_block_activation with
    | Simple_block_enabled ->
      if activation_port = Enable_port then activation_port
      else assert false
    | Simple_block_if_action ->
      if activation_port = If_action_port then activation_port
      else assert false
    | Simple_block_triggered ->
      if activation_port = Trigger_port
      then activation_port else assert false
    | Simple_block_triggered_and_enabled ->
      if activation_port = Trigger_port then Trigger_enable_port else
      if activation_port = Enable_port then Enable_trigger_port
      else assert false
;;

let activation_port block_by_names block_name activation_port =
  let blk = Hashtbl.find block_by_names block_name in
  let desc = blk.explicit_block_desc in
  match desc with
  | Explicit_block_regular explicit_simple_block_desc ->
    activation_port_of_block_block_descriptor
      activation_port explicit_simple_block_desc
  | Explicit_block_reference explicit_simple_block_desc ->
    activation_port_of_block_block_descriptor
      activation_port explicit_simple_block_desc
  | Explicit_block_sub_system explicit_block_sub_system_desc ->
    let explicit_activation =
      explicit_block_sub_system_desc.explicit_block_sub_system_activation in
    begin match explicit_activation with
    | Sub_system_plain
    | Sub_system_activated (Sub_system_plain_atomic_unit _) ->
      split_error (Printf.sprintf
        "attempt to get activation port %s \
         from a block that is a Plain sub-system"
        (string_of_activation_port activation_port))
    | Sub_system_activated (Sub_system_enabled _) ->
      if activation_port = Enable_port then activation_port
      else assert false
    | Sub_system_activated (Sub_system_if_action (_, _)) ->
      if activation_port = If_action_port then activation_port
      else assert false
    | Sub_system_activated (Sub_system_triggered (_, _)) ->
      if activation_port = Trigger_port then activation_port
      else assert false
    | Sub_system_activated (Sub_system_triggered_and_enabled ((_, _), _)) ->
      if activation_port = Trigger_port then Trigger_enable_port else
      if activation_port = Enable_port then Enable_trigger_port else
      assert false
    end
  | Explicit_block_s_function _ -> assert false
;;

let activation_port_connector_desc block_by_names block_name sport =
  let port = activation_port block_by_names block_name sport in
  Mdl_split_ast.
  Split_activation_connector_desc port
;;

let make_split_dst_connector block_name n =
  let descriptor =
    let port_number = mdl_port_number_of_natural n in
    Mdl_split_ast.
    Split_directed_connector_desc port_number in
  make_connector_port block_name descriptor
;;

let make_directed_dst_connector block_by_names dst_connector =
  match dst_connector with
  | {
      explicit_connector_port_block = dst_block_name;
      explicit_connector_port_desc = dst_connector_port;
    } ->
    let explicit_connector_desc =
      match dst_connector_port with
      | Mdl_ast.Port_number port_number ->
        Split_directed_connector_desc port_number
      | Mdl_ast.Port_trigger ->
        activation_port_connector_desc
          block_by_names dst_block_name Trigger_port
      | Mdl_ast.Port_enable ->
        activation_port_connector_desc
          block_by_names dst_block_name Enable_port
      | Mdl_ast.Port_ifaction ->
        activation_port_connector_desc
          block_by_names dst_block_name If_action_port
      | Mdl_ast.Port_state ->
        split_error "Line destination can not be a State port"
      | Mdl_ast.Port_lconn _port_number
      | Mdl_ast.Port_rconn _port_number ->
        split_error "Line connection as link connector" in

    make_connector_port dst_block_name explicit_connector_desc
;;

let get_directed_src_connector block_by_names = function
  | {
      explicit_connector_port_block = source_block_name;
      explicit_connector_port_desc = connector_port;
    } ->
    let source_port_number =
      match connector_port with
      | Mdl_ast.Port_number source_port_number -> source_port_number
      | Mdl_ast.Port_trigger
      | Mdl_ast.Port_enable | Mdl_ast.Port_ifaction ->
        split_error "Event activation port in source port."
      | Mdl_ast.Port_state ->
        let blk = Hashtbl.find block_by_names source_block_name in
        (* The State port number is the last port of the block,
           since this port has been added during the deep phase when
           the port_state number of the block in the MDL file was 1.
        *)
        let source_port_number =
          let ports_desc =
            blk.
              explicit_block_body.
                explicit_block_all_bindings.
                  explicit_block_graphical.
                    explicit_block_ports_description in
          Mdl_ast_to_deep_ast.get_regular_port_nout ports_desc in
        mdl_port_number_of_natural source_port_number
      | Mdl_ast.Port_lconn _ | Mdl_ast.Port_rconn _ ->
        (* Should call code for connection lines *)
        split_error "Connection line port in source port." in

    make_connector_port source_block_name source_port_number
;;

(* Builds an explicit link from the body (template) of an explicit line. *)
let rec flatten_line_template
    block_by_names accu
    template_line_body connectors labels_option points_option =
  match connectors with
  | {
      explicit_directed_branches = [];
      explicit_directed_target = None;
      _;
    } ->
    (* Skip the line and go on *)
    Say.warning "Skipping dangling directed line without target";
    accu

  | {
      explicit_directed_branches = [];
      explicit_directed_source = src_connector;
      explicit_directed_target = Some dst_connector;
      _;
    } ->
    (* Just add the line to the resulting links *)
    let directed_dst_connector =
      make_directed_dst_connector block_by_names dst_connector in
    let directed_src_connector =
      get_directed_src_connector block_by_names src_connector in

    let split_directed_link_connectors =
      {
        Mdl_split_ast.
        split_link_src = directed_src_connector;
        split_link_dst = directed_dst_connector;
      } in

    let split_desc =
      Mdl_split_ast.
      Split_directed_link_desc split_directed_link_connectors in
    let split_link =
      make_split_link_directed template_line_body split_desc
        labels_option points_option in
    add_link accu split_link

    (* Case branches *)
  | {
      explicit_directed_branches = branches;
      explicit_directed_source = src_connector;
      _;
    } ->
    let nb_branches = Unatural_types.to_natural (List.length branches) in
    (* Create a new Split block with
       - i as name suffix
       - nb_branches output ports *)
    let accu, split_name, split =
      get_split_directed
        accu template_line_body.template_line_directed_name nb_branches in

    (* Target of the link is the new split *)
    let directed_dst_connector =
      make_split_dst_connector split_name Unatural._1_ in

    let directed_src_connector =
      get_directed_src_connector block_by_names src_connector in

    let split_directed_link_connectors =
      {
        Mdl_split_ast.
        split_link_src = directed_src_connector;
        split_link_dst = directed_dst_connector;
      } in

    let split_link =
      let split_desc =
        Mdl_split_ast.
        Split_directed_link_desc split_directed_link_connectors in
      make_split_link_directed
        template_line_body split_desc
        labels_option points_option in

    (* Add the new link to the resulting split link list *)
    let accu = add_link accu split_link in
    (* Add the split block to the resulting split block list *)
    let accu = add_split accu split in
    (* Cumulate new links and blocks for all the branches *)
    flatten_branches_directed
      block_by_names accu template_line_body split_name branches

and flatten_branches_directed
    block_by_names accu template_line_body split_name branches =

  let rec loop (_i, _accu_links, _accu_splits as accu) port_number branches =
    match branches with
    | [] -> accu
    | Deep_branch_connection (_connect_type, _deep_branch_body) :: _ ->
      split_error "Directed lines cannot lead to connection branches"
    | Deep_branch_directed deep_branch_body :: branches ->
      begin match deep_branch_body with
      | {
          deep_branch_source = None;
          deep_branch_target = Some dst_conn;
          deep_branch_branches = [];
          deep_branch_labels = labels_opt;
          deep_branch_points = points_opt;
          deep_branch_zorder = zorder_opt;
        } ->
        split_deep_zorder_dropped zorder_opt;

        let explicit_line_connectors =
          {
            explicit_directed_source =
              make_explicit_connector_port split_name port_number;
            explicit_directed_target =
              Some (
                explicit_directed_connector_of_deep_connector dst_conn);
            explicit_directed_branches = [];
          } in

        let accu =
          flatten_line_template block_by_names accu
            template_line_body explicit_line_connectors
            labels_opt points_opt in
        loop accu (Unatural.succ port_number) branches

      | {
          deep_branch_source = None;
          deep_branch_target = Some dst_conn;
          deep_branch_branches = _branch_branches;
          deep_branch_labels = _labels_opt;
          deep_branch_points = _points_opt;
          deep_branch_zorder = _zorder_opt;
        } ->
        split_error (
          Printf.sprintf
            "explicit branch with both target (%S) and branches"
            (Ast_types.string_of_mdl_name
               dst_conn.Mdl_ast.line_connector_block))

      | {
          deep_branch_source = Some src_conn;
          deep_branch_target = _dst_conn;
          deep_branch_branches = _branch_branches;
          deep_branch_labels = _labels_opt;
          deep_branch_points = _points_opt;
          deep_branch_zorder = _zorder_opt;
        } ->
        split_error (
          Printf.sprintf "explicit branch with a source connector (%S)"
            (string_of_mdl_name
               src_conn.Mdl_ast.line_connector_block))
      | {
          deep_branch_source = None;
          deep_branch_target = None;
          deep_branch_branches = [];
          deep_branch_labels = _labels_opt;
          deep_branch_points = _points_opt;
          deep_branch_zorder = _zorder_opt;
        } ->
        split_warning "explicit branch without target";
        loop accu port_number branches
      | {
          deep_branch_source = None;
          deep_branch_target = None;
          deep_branch_branches = branch_branches;
          deep_branch_labels = labels_opt;
          deep_branch_points = points_opt;
          deep_branch_zorder = zorder_opt;
        } ->
        split_deep_zorder_dropped zorder_opt;
        let explicit_line_connectors =
          {
            explicit_directed_source =
              make_explicit_connector_port split_name port_number;
            explicit_directed_target = None;
            explicit_directed_branches = branch_branches;
          } in

        let accu =
          flatten_line_template block_by_names accu
            template_line_body explicit_line_connectors
            labels_opt points_opt in
        loop accu (Unatural.succ port_number) branches
      end in

  let port_number = Unatural._1_ in
  loop accu port_number branches
;;

let flatten_line_directed
    block_by_names accu explicit_line_body =
  let template, connectors, labels_option, points_option =
    make_template_line_directed_body explicit_line_body in
  flatten_line_template
    block_by_names accu template connectors labels_option points_option
;;

(* Constant connection port number 1. *)
let connection_port_number_1 =
  Connection_port_value (mdl_port_number_of_natural Unatural._1_)
;;

(* Build a connection right connector port from a split block name *)
let make_explicit_connection_right_connector_port
    block_name connection_port_number =
  let explicit_connector_port_desc =
    Mdl_implementation_explicit_ast.
    Explicit_right_connector connection_port_number in
  make_connector_port block_name explicit_connector_port_desc
;;

(* Build a connection left connector port from a split block name *)
let make_explicit_connection_left_connector_port
    block_name connection_port_number =
  let explicit_connector_port_desc =
    Mdl_implementation_explicit_ast.
    Explicit_left_connector connection_port_number in
  make_connector_port block_name explicit_connector_port_desc
;;

let get_make_explicit_connection_side_connector_ports block_name = function
  | Source _ ->
    make_explicit_connection_right_connector_port
      block_name connection_port_number_1,
    make_explicit_connection_left_connector_port block_name
  | Destination ->
    make_explicit_connection_left_connector_port
      block_name connection_port_number_1,
    make_explicit_connection_right_connector_port block_name
;;

let get_explicit_connection_connector connection_side = function
  | {
      explicit_source_connection_connector =
        explicit_source_connection_connector;
      explicit_target_connection_connector =
        explicit_target_connection_connector;
    } ->
    match connection_side with
    | Source _ ->
      explicit_source_connection_connector
    | Destination ->
      explicit_target_connection_connector
;;

let split_explicit_points_option explicit_points_option =
  match explicit_points_option with
  | None -> None, None
  | Some (Explicit_relative_points explicit_relative_points) ->
    None, Some explicit_relative_points
  | Some (
      Explicit_absolute_points (
        explicit_absolute_point, explicit_relative_points)) ->
    Some explicit_absolute_point, Some explicit_relative_points
;;

let rec make_link_connection_connector
    connection_side accu
    template_line_body
    explicit_connection_connector =

  let make_explicit_connection_side_connector_ports block_name =
    get_make_explicit_connection_side_connector_ports
      block_name connection_side in

  let explicit_line_name_option =
    template_line_body.template_line_connection_name in

  match explicit_connection_connector with
  | Explicit_connection_connector_port
      explicit_connection_connector_port ->
    accu,
    explicit_connection_connector_port

  | Explicit_connection_connector_branches
      explicit_connection_branches ->

    let nb_branches =
      Unatural_types.to_natural (List.length explicit_connection_branches) in
    (* Create a new Split block with
       - i as name suffix
       - nb_branches output ports *)

    let accu, split_block_name, split_connection_block =
      let get_connection_split, connection_split_position =
        match connection_side with
        | Destination -> get_connection_split_target, dummy_position
        | Source None -> get_connection_split_source, dummy_position
        | Source (Some explicit_absolute_point) ->
          let make_split_position = function
            | {
                Mdl_implementation_explicit_ast.
                explicit_x = x;
                explicit_y = y;
              } ->
              {
                xmin = x;
                ymin = y;
                xmax = x;
                ymax = y;
              } in
          get_connection_split_source,
          make_split_position explicit_absolute_point in

      get_connection_split
        accu explicit_line_name_option nb_branches
        connection_split_position in

    let explicit_connection_connector_port_for_self,
        make_explicit_connection_connector_port_for_child =
      make_explicit_connection_side_connector_ports split_block_name in

    let accu = add_split accu split_connection_block in

    let accu =
      flatten_line_connection_branches
        accu
        template_line_body
        make_explicit_connection_connector_port_for_child
        explicit_connection_branches in

    accu,
    explicit_connection_connector_port_for_self

and flatten_line_connection_branches
    accu
    explicit_line_body
    make_explicit_connection_connector_port_for_child
    explicit_connection_branches =

  let accu, _ =
    List.fold_left
      (fun (accu, raw_port_number) explicit_connection_branch ->
       let port_number =
         Connection_port_value (mdl_port_number_of_natural raw_port_number) in
       let accu =
         flatten_line_connection_branch
           explicit_line_body
           (make_explicit_connection_connector_port_for_child port_number)
           accu
           explicit_connection_branch in
       accu, Unatural.succ raw_port_number)
      (accu, Unatural._1_)
      explicit_connection_branches in
  accu

and flatten_line_connection_branch
    template_line_body
    explicit_connection_connector_port_for_child
    accu = function
  | {
      explicit_connection_branch_points = explicit_points_option;
      explicit_connection_branch_child = explicit_connection_branch_child;
    } ->

    let absolute_point_option, explicit_relative_points_option =
      split_explicit_points_option explicit_points_option in

    let accu, explicit_link_src, explicit_link_tgt =
      match explicit_connection_branch_child  with
      | Explicit_connection_branch_source explicit_connection_connector ->
        let accu, explicit_link_src =
          let connection_side = Source absolute_point_option in
          make_link_connection_connector
            connection_side accu
            template_line_body
            explicit_connection_connector in
        let explicit_link_tgt =
          explicit_connection_connector_port_for_child in
        accu, explicit_link_src, explicit_link_tgt
      | Explicit_connection_branch_target explicit_connection_connector ->
        let explicit_link_src =
          explicit_connection_connector_port_for_child in
        let accu, explicit_link_tgt =
          let connection_side =
            match absolute_point_option with
            | None -> Destination
            | Some _ -> assert false in
          make_link_connection_connector
            connection_side accu
            template_line_body
            explicit_connection_connector in
        accu, explicit_link_src, explicit_link_tgt in

    let split_connection_link_connectors =
      {
        Mdl_split_ast.
        split_link_src = explicit_link_src;
        split_link_dst = explicit_link_tgt;
      } in
    let split_desc =
      Mdl_split_ast.
      Split_connection_link_desc split_connection_link_connectors in

    let split_link =
      make_split_link_connection template_line_body split_desc
        explicit_relative_points_option in

    add_link accu split_link
;;

let flatten_line_connection accu explicit_line_body =

  match explicit_line_body with
(* TODO? use explicit_line_body as a template *)
  | {
      Mdl_implementation_explicit_ast.
      explicit_line_name = _line_name_option;
      explicit_line_connectors = {
        explicit_source_connection_connector =
          _explicit_source_connection_connector;
        explicit_target_connection_connector =
          _explicit_target_connection_connector;
      };
      explicit_line_font = _explicit_font;
      (* Will be transmited with the line *)
      explicit_line_labels = _labels_option;
      (* Will be transmited with the line *)
      explicit_line_points = _explicit_points_option;
    } ->

    let get_link_connection_connector
        connection_side
        accu
        template_line_body
        connectors =
      let explicit_connection_connector =
        get_explicit_connection_connector
          connection_side connectors in
      make_link_connection_connector
        connection_side
        accu
        template_line_body
        explicit_connection_connector in

    let template_line_body, connectors, points_option =
      make_template_line_connection_body explicit_line_body in

    let absolute_point_option, explicit_relative_points_option =
      split_explicit_points_option points_option in

    let accu, explicit_link_src =
      get_link_connection_connector
        (Source absolute_point_option)
        accu
        template_line_body connectors in

    let accu, explicit_link_tgt =
      get_link_connection_connector
        Destination
        accu
        template_line_body connectors in

    (* Simply add the line to the resulting links *)
    let split_connection_link_connectors =
      {
        Mdl_split_ast.
        split_link_src = explicit_link_src;
        split_link_dst = explicit_link_tgt;
      } in

    let split_desc =
      Mdl_split_ast.
      Split_connection_link_desc split_connection_link_connectors in

    let split_link =
      make_split_link_connection template_line_body split_desc
        explicit_relative_points_option in

    let accu = add_link accu split_link in

    accu
;;

(* Remove branches by splitting Simulink lines *)
let split_lines block_by_names lines accu_splits =

  let split_line (i, accu_links, accu_splits as accu) = function
    | Explicit_line_directed explicit_line_body ->
      flatten_line_directed block_by_names accu explicit_line_body
    | Explicit_line_connection explicit_line_body ->
      let i, line_accu_links, accu_splits =
        flatten_line_connection (i, [], accu_splits) explicit_line_body in
      i, List.rev_append line_accu_links accu_links, accu_splits in

  let (_i, accu_links, accu_splits) =
    List.fold_left split_line (0, [], accu_splits) lines in

  accu_links, accu_splits
;;

let port_ident = ident_from_string "Port";;

let get_inport_port_value block_name explicit_block_all_bindings =
  try
    let bdg =
      Hashtbl.find
        explicit_block_all_bindings.explicit_block_parameters port_ident in
    bdg.Binding.bound_value with
  | Not_found ->
    split_error (Printf.sprintf
      "No Port value in Inport (shadow) %s block bindings"
      (string_of_mdl_name block_name))
;;

let has_input_port block_num blk =
  match blk.explicit_block_desc with
  | Explicit_block_regular explicit_simple_block_desc ->
    let ty = explicit_simple_block_desc.explicit_simple_block_type in
    if ty <> Mdl_block_types.Inport then false else
    let inport_num =
      get_inport_port_value
        blk.explicit_block_name
        blk.explicit_block_body.explicit_block_all_bindings in
    inport_num = block_num
  | Explicit_block_reference _explicit_simple_block_desc -> false
  | Explicit_block_sub_system _explicit_block_sub_system -> false
  | Explicit_block_s_function _ -> false
;;

let route_to_inport_split
    inport_block_name inport_shadow_block_name all_lines =

  let lines_from_inport_block =
    List.map
      (fun line ->
       match line with
       | Explicit_line_directed ({
           explicit_line_connectors = {
             explicit_directed_source = {
               explicit_connector_port_block = block_name;
               explicit_connector_port_desc = _;
             _;
             };
           _;
           } as explicit_line_connectors;
          _;
         } as explicit_line_body) ->
         if block_name = inport_block_name then
           Explicit_line_directed {
             explicit_line_body with
             explicit_line_connectors = {
               explicit_line_connectors with
                 explicit_directed_source =
                   make_explicit_connector_port
                     inport_shadow_block_name Unatural._2_;
             }
          }
         else line
       | Explicit_line_connection _ -> line)
      all_lines in

  let line =
    let explicit_line_connectors = {
      explicit_directed_source =
        make_explicit_connector_port inport_block_name Unatural._1_;
      explicit_directed_target =
        Some (
          make_explicit_connector_port
            inport_shadow_block_name Unatural._1_);
      explicit_directed_branches = [];
    } in
    Explicit_line_directed {
      explicit_line_font = dummy_explicit_font;
      explicit_line_connectors = explicit_line_connectors;
      explicit_line_name = None;
      explicit_line_labels = None;
      explicit_line_points = None;
    } in

  line :: lines_from_inport_block
;;

let split_inport_shadow_block
    inport_shadow_block_name inport_shadow_block_bindings
    all_lines blocks =

  let inport_block =
    let block_num =
      get_inport_port_value
        inport_shadow_block_name inport_shadow_block_bindings in
    try
      List.find (fun blk -> has_input_port block_num blk) blocks with
    | Not_found ->
      split_error (Printf.sprintf
        "Inportshadow block %s contains reference to \
         non existent InputPort block with port number %s"
        (string_of_mdl_name inport_shadow_block_name)
        (Mdl_deep_basics.string_of_deep_value "port number" block_num)) in

  let inport_block_name = inport_block.explicit_block_name in

  let all_lines =
    route_to_inport_split
      inport_block_name inport_shadow_block_name all_lines in

  let inport_split_block = get_split_shadow inport_shadow_block_name in
  all_lines, inport_split_block
;;

let split_inport_shadow lines blocks =

  let rec loop accu_blks lines blks =
    match blks with
    | [] -> lines, accu_blks
    | blk :: blks ->
      match blk.explicit_block_desc with
      | Explicit_block_regular explicit_simple_block_desc
          when
           let ty =
             explicit_simple_block_desc.explicit_simple_block_type in
           ty = Mdl_block_types.InportShadow ->
        let inport_shadow_block_name = blk.explicit_block_name in
        let inport_shadow_block_bindings =
          blk.explicit_block_body.explicit_block_all_bindings in
        Say.debug (Printf.sprintf
          "Split: Block %s is an inport shadow block"
          (string_of_mdl_name inport_shadow_block_name));
        let lines, blk =
          split_inport_shadow_block
            inport_shadow_block_name inport_shadow_block_bindings
            lines blocks in
        loop (blk :: accu_blks) lines blks
      | Explicit_block_regular _
      | Explicit_block_reference _
      | Explicit_block_sub_system _
      | Explicit_block_s_function _ ->
        loop (blk :: accu_blks) lines blks in

  loop [] lines blocks
;;

let make_block_by_names blocks =
  let block_by_names = Hashtbl.create 100 in
  List.iter
    (fun blk ->
     let block_name = blk.explicit_block_name in
     if not (Hashtbl.mem block_by_names block_name) then
       Hashtbl.add block_by_names block_name blk else
       split_error (
         let Ast_types.Mdl_name blk_name = block_name in
         Printf.sprintf
         "duplicated block %s in sub-system"
         blk_name)
    )
    blocks;
   block_by_names
;;

(* From Simulink lines to Scicos links. *)
let rec make_split_block explicit_line_block :
  Mdl_split_ast.split_link Mdl_split_ast.explicit_block =
  match explicit_line_block with
    (* A regular, S-Function, and reference block has no internal lines:
       there is nothing to do inside. *)
  | {
      explicit_block_desc =
        (Explicit_block_regular _ |
         Explicit_block_reference _ |
         Explicit_block_s_function _);
      _;
    } as blk -> blk
    (* A sub-system block has internal lines in its system:
       change them recursively. *)
  | {
      explicit_block_desc =
        Explicit_block_sub_system explicit_block_sub_system_desc;
      _;
    } as blk ->
    let split_block_desc =
      let split_block_sub_system_desc =
        make_split_block_sub_system_desc
          explicit_block_sub_system_desc in
      Explicit_block_sub_system split_block_sub_system_desc in
    {
      blk with
      explicit_block_desc = split_block_desc;
    }

and make_split_block_sub_system_desc = function
  | {
      explicit_block_sub_system_activation = explicit_activation;
      explicit_block_sub_system_body = explicit_system;
      explicit_block_sub_system_treat_as_atomic_unit =
        explicit_treat_as_atomic_unit;
      explicit_block_sub_system_sample_time = explicit_system_sample_time;
    } ->
    let split_activation =
      make_split_activation explicit_activation in
    let split_system =
      make_split_system
        (explicit_system :
           Mdl_implementation_explicit_ast.explicit_line
           Mdl_implementation_explicit_ast.explicit_system) in
    {
      explicit_block_sub_system_activation = split_activation;
      explicit_block_sub_system_body = split_system;
      explicit_block_sub_system_treat_as_atomic_unit =
        explicit_treat_as_atomic_unit;
      explicit_block_sub_system_sample_time = explicit_system_sample_time;
    }

and make_split_activation explicit_activation =
  match explicit_activation with
  | Sub_system_plain -> Sub_system_plain
  | Sub_system_activated sub_system_activation ->

    let activation =
      match sub_system_activation with
      | Sub_system_plain_atomic_unit _ as act -> act
      | Sub_system_enabled explicit_block ->
        let split_block =
          make_split_block explicit_block in
        Sub_system_enabled split_block
      | Sub_system_if_action (explicit_activation_type, explicit_block) ->
        let split_block =
          make_split_block explicit_block in
        Sub_system_if_action (explicit_activation_type, split_block)
      | Sub_system_triggered (explicit_trigger_type, explicit_block) ->
        let split_block =
          make_split_block explicit_block in
        Sub_system_triggered (explicit_trigger_type, split_block)
      | Sub_system_triggered_and_enabled
          ((explicit_trigger_type, explicit_block1), explicit_block2) ->
        let split_block1 =
          make_split_block explicit_block1 in
        let split_block2 =
          make_split_block explicit_block2 in
        Sub_system_triggered_and_enabled
          ((explicit_trigger_type, split_block1),
           split_block2) in

    Sub_system_activated activation

and make_split_system
  (explicit_system :
     Mdl_implementation_explicit_ast.explicit_line
     Mdl_implementation_explicit_ast.explicit_system) :
     Mdl_split_ast.split_system =
  let explicit_lines =
    explicit_system.explicit_system_lines
  and explicit_line_blocks =
    explicit_system.explicit_system_blocks in

  (* Adding inport shadow blocks.
     We still return lines and line blocks.
     We thus compute a new Simulink model with inlined shadow blocks with
     the addition of proper Simulink lines.
     We then turn lines to split links on the new model. *)
  let explicit_lines, explicit_line_blocks =
    split_inport_shadow explicit_lines explicit_line_blocks in

  let split_blocks =
    List.rev_map make_split_block explicit_line_blocks in

  let block_by_names = make_block_by_names split_blocks in

  (* We now turn the Simulink lines into split links (almost Scicos links),
     generating Split blocks on the way when eliminating branches. *)
  let accu_splits, new_split_blocks =
    split_lines block_by_names explicit_lines [] in

  let split_blocks =
    List.rev_append new_split_blocks split_blocks in

  {
    explicit_system with
    explicit_system_blocks = split_blocks;
    explicit_system_lines = accu_splits;
  }
;;

let split_model_body explicit_model_body =
  {
    explicit_model_body with
    explicit_model_system =
      make_split_system
        explicit_model_body.explicit_model_system;
  }
;;

let split_model model =
  {
    model with
    explicit_model_body =
      split_model_body model.explicit_model_body;
  }
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
