(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_split_ast.mli,v 1.1 2019-05-10 14:04:01 jpc Exp $ *)

(** {3 Split link kind models} *)

(**
  A model with split link connections is a [split_model] (as
  defined below).

  A model with split link connections is the result of the line splitting
  treatment of an explicit Simulink model (an explicit Simulink model is
  an [Mdl_inplementation_explicit_ast.explicit_line_model] value).

  Split link connections does not have branches: those branches have been
  explicitly translated into devoted split blocks that are connected with
  other link connections to other blocks in the model.
*)

(** {7 Stuff borrowed from external modules} *)

(** {8 Borrowed from the parsing phase definitions} *)

type block_name = Mdl_ast.block_name
and line_name = Mdl_ast.line_name
;;

type port_number = Mdl_ast.port_number
(** A port number is simply a natural number. *)
;;

type labels_option = Mdl_ast.labels option;;

(** {8 Borrowed from the explicit phase definitions} *)
type explicit_font = Mdl_basics_explicit_ast.explicit_font
and explicit_relative_points =
  Mdl_implementation_explicit_ast.explicit_relative_points
;;

type 'a explicit_block = 'a Mdl_implementation_explicit_ast.explicit_block
and 'a explicit_system = 'a Mdl_implementation_explicit_ast.explicit_system
and 'a explicit_model = 'a Mdl_implementation_explicit_ast.explicit_model
and 'a explicit_simulink_file = 'a Mdl_explicit_ast.explicit_simulink_file
;;

(** {7 Defining split links} *)

(** A split link is either a directed or connection link. *)

(** The connectors of a split link: there is a source connector and a
  destination connector. Those two connectors may have the same type or not.
  For connection lines the two connectors are both [connection_connector]
  connectors. For standard directed lines the source port is a
  [directed_connector] while the target connector is either a
  [directed_connector] or a [activation_connector] (see above for the definition
  of activation port).
*)

type 'a split_connector_port =
  'a Mdl_implementation_explicit_ast.explicit_connector_port
;;

type split_connection_connector_desc =
  Mdl_implementation_explicit_ast.explicit_connection_connector_desc
;;

type directed_connector_desc = port_number;;

type activation_connector_desc =
  | Enable_port
  | Trigger_port
  | Enable_trigger_port
  | Trigger_enable_port
  | If_action_port
(** The target of a Simulink so-called activation lines. *)
;;

type activation_connector = activation_connector_desc split_connector_port;;

type ('src, 'dst) split_link_connectors =
  {
    split_link_src : 'src split_connector_port;
    split_link_dst : 'dst split_connector_port;
  }
;;

type split_directed_link_connectors =
  (directed_connector_desc, dst_directed_connector_desc)
  split_link_connectors

and dst_directed_connector_desc =
  | Split_directed_connector_desc of directed_connector_desc
    (** A destination connector for a directed link *)
  | Split_activation_connector_desc of activation_connector_desc
    (** A destination connector for a activation directed link
      (a link that maps to a Scicos event link). *)
;;

type split_connection_link_connectors =
  (split_connection_connector_desc, split_connection_connector_desc)
    split_link_connectors
;;

type 'link_desc poly_split_link = {
  split_link_name : line_name option;
  split_link_font : explicit_font;
  split_link_labels : labels_option;
  split_link_points : explicit_relative_points option;
  split_link_desc : 'link_desc;
}

and split_link_desc =
  | Split_directed_link_desc of split_directed_link_connectors
  | Split_connection_link_desc of split_connection_link_connectors
    (* Simulink line connection (Scicos counterpart are implicit links) *)
;;

type split_links = split_link list
and split_link = split_link_desc poly_split_link
;;

type split_block = split_link explicit_block
and split_system = split_link explicit_system
and split_model = split_link explicit_model
and split_simulink_file = split_link explicit_simulink_file
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
