(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_block.ml,v 1.1 2019-05-10 14:04:02 jpc Exp $ *)

open Mdl_basics_explicit_ast
open Genr_block_types
;;

let translation_error msg =
  Say.fatal_error (Printf.sprintf "Translation: %s" msg)
;;

(* The hash table containing the translator for (known) Simulink blocks. *)
let translator_table =
  (Hashtbl.create 273 :
   (explicit_block_type, block_translator) Hash_table.t)
;;

let register_translator explicit_block_type block_translator =
  if Hashtbl.mem translator_table explicit_block_type then
    translation_error
      (Printf.sprintf
         "Multiple definition of translator for block %S"
         (Mdl_basics_to_explicit_ast.
          name_of_explicit_block_type explicit_block_type)) else
  Hashtbl.add translator_table explicit_block_type block_translator
;;

let add_initialization, init =
  let scicos_initializers = ref [] in
  let scicos_pro_initializers = ref [] in
  let modelica_initializers = ref [] in
  let get_initializers = function
    | Simport_configuration.Modeler_scicos ->
      scicos_initializers
    | Simport_configuration.Modeler_scicos_pro ->
      scicos_pro_initializers
    | Simport_configuration.Modeler_modelicos ->
      modelica_initializers in
  (fun target_model_system bt tr ->
   let initializers = get_initializers target_model_system in
   initializers := (bt, tr) :: !initializers),
  (fun target_model_system ->
   let initializers = get_initializers target_model_system in
   List.iter
     (fun (bt, tr) -> register_translator bt tr)
     (List.rev !initializers))
;;

let add_translator = add_initialization;;

(* Find available translator for a Simulink block type. *)
let find_translator explicit_block_type =
  Hashtbl.find translator_table explicit_block_type
;;

let block_identity genr_info =
  let block_type = genr_info.block_explicit_type in
  let block_name = genr_info.block_explicit_name in
  Printf.sprintf "block %s, with type %s"
    (Ast_types.string_of_mdl_name block_name)
    (Mdl_basics_to_explicit_ast.name_of_explicit_block_type block_type)
;;

let print_block_identity ppf genr_info =
  Format.fprintf ppf "%s" (block_identity genr_info)
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
