(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: genr_utils.ml,v 1.1 2019-05-10 14:04:02 jpc Exp $ *)

let host_id id = id;;

let host_ident id = Ident.Ident (host_id id);;
let modeler_ident id = Instructions.Si id;;
let host_block_ident id = Instructions.Hbi (host_id id);;
let host_diagram_ident id = Instructions.Hdi (host_id id);;
let host_block_tag_ident id = Instructions.Hbti (host_id id);;
let host_function_ident id = Instructions.Hfi (host_id id);;

let host_block level =
  host_block_ident (Printf.sprintf "blk%d" level)
;;

let host_diagram level =
  host_diagram_ident (Printf.sprintf "scsm%d" level)
;;

let next_block_level level =
  let level = level + 1 in
  level, host_block level
;;

let blk = host_block_ident "blk";;
let blk0 = host_block 0;;
(*let blk1 = host_block 1;;*)
let scsm = host_diagram_ident "scsm";;
let scsm0 = host_diagram 0;;
(*let scsm1 = host_diagram 1;;*)

let modeler_variant = ref None;;

let init sv = modeler_variant := Some sv;;

let no_init_call_error () =
  Say.fatal_error
    "You must call Genr_utils.init during initialization\
   \nto set the target modeler variant"
;;

let no_port_kind_error kind =
  Say.fatal_error
    (Printf.sprintf "modeler %s port type was not set" kind)

let no_input_port_error () = no_port_kind_error "input";;
let no_output_port_error () = no_port_kind_error "output";;

let port_kind_set_twice_error kind =
 Say.fatal_error
   (Printf.sprintf "modeler %s port type was set twice" kind)
;;

let input_port_set_twice_error () = port_kind_set_twice_error "input";;
let output_port_set_twice_error () = port_kind_set_twice_error "output";;

let get_modeler_input_port_type, set_modeler_input_port_type =
  let scicos_input_port_type = ref None
  and scicos_pro_input_port_type = ref None
  and modelicos_input_port_type = ref None in

  let get_input_port_type = function
   | Simport_configuration.Modeler_scicos ->
     scicos_input_port_type
   | Simport_configuration.Modeler_scicos_pro ->
     scicos_pro_input_port_type
   | Simport_configuration.Modeler_modelicos ->
     modelicos_input_port_type in

  (fun () ->
   match !modeler_variant with
   | None -> no_init_call_error ()
   | Some sv ->
     let input_port_type = get_input_port_type sv in
     match !input_port_type with
     | None -> no_input_port_error ()
     | Some t -> t),
  (fun sv t ->
   let input_port_type = get_input_port_type sv in
   match !input_port_type with
   | None -> input_port_type := Some t
   | Some _ -> input_port_set_twice_error ())
;;

let get_modeler_output_port_type, set_modeler_output_port_type =
  let scicos_output_port_type = ref None
  and scicos_pro_output_port_type = ref None
  and modelicos_output_port_type = ref None in

  let get_output_port_type = function
   | Simport_configuration.Modeler_scicos ->
     scicos_output_port_type
   | Simport_configuration.Modeler_scicos_pro ->
     scicos_pro_output_port_type
   | Simport_configuration.Modeler_modelicos ->
     modelicos_output_port_type in

  (fun () ->
   match !modeler_variant with
   | None -> no_init_call_error ()
   | Some sv ->
     let output_port_type = get_output_port_type sv in
     match !output_port_type with
     | None -> no_input_port_error ()
     | Some t -> t),
  (fun sv t ->
   let output_port_type = get_output_port_type sv in
   match !output_port_type with
   | None -> output_port_type := Some t
   | Some _ -> input_port_set_twice_error ())
;;

let get_modeler_annotation_type, set_modeler_annotation_type =
  let scicos_annotation_type = ref None
  and scicos_pro_annotation_type = ref None
  and modelicos_annotation_type = ref None in

  let get_annotation_type = function
   | Simport_configuration.Modeler_scicos ->
     scicos_annotation_type
   | Simport_configuration.Modeler_scicos_pro ->
     scicos_pro_annotation_type
   | Simport_configuration.Modeler_modelicos ->
     modelicos_annotation_type in

  (fun () ->
   match !modeler_variant with
   | None -> no_init_call_error ()
   | Some sv ->
     let annotation_type = get_annotation_type sv in
     match !annotation_type with
     | None -> no_output_port_error ()
     | Some t -> t),
  (fun sv t ->
   let annotation_type = get_annotation_type sv in
   match !annotation_type with
   | None -> annotation_type := Some t
   | Some _ -> output_port_set_twice_error ())
;;

let make_binding = Mdl_ast_functions.make_binding;;

let block_parameter_binding (Instructions.Si id, code_value) =
  make_binding (host_ident id) (Instructions.Parameter_value code_value)
;;

let struct_parameter_binding (Instructions.Si id, struct_values) =
  let struct_value =
    Ulist1.map
      (fun (Instructions.Si id, values) ->
       make_binding (host_ident id) values)
      struct_values in
  make_binding (host_ident id) (Instructions.Parameter_struct struct_value)
;;

(* Turns a Caml string contents (as given by the Mdl parser from a mdl
   file) to a ``valid'' Matlab string.
   It escapes all \' characters in the text as two \'
   characters and enclose this result into \' characters to get a valid
   matlab (string) lexem. *)
let matlab_string_of_mdl_string text =
  let lim = String.length text - 1 in
  let accu = Buffer.create (lim + 2) in

  let rec loop i =
    if i <= lim then
      begin
        (match text.[i] with
         | '\'' -> Buffer.add_string accu "''"
         | '\n' ->
           Say.fatal_error (Printf.sprintf
             "can not turn text %S into a Matlab string token,\
            \na newline character (ASCII 10) was encountered"
             text)
         | c -> Buffer.add_char accu c
        );
        loop (i + 1);
      end in

  Buffer.add_char accu '\'';
  loop 0;
  Buffer.add_char accu '\'';
  Buffer.contents accu
;;

let matlab_string_of_file_name fname =
  let text = fname in
  matlab_string_of_mdl_string text
;;

let simulink_block_reference_type bl_src_ty bl_src_path =
  Mdl_basics_explicit_ast.Explicit_block_reference_type
    (Mdl_deep_basics.make_deep_block_reference_type  bl_src_ty bl_src_path)
;;

let simulink_block_type ty =
  Mdl_basics_explicit_ast.Explicit_block_regular_type ty
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
