(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: instructions.mli,v 1.1 2019-05-10 14:04:02 jpc Exp $ *)

(** {3 The set of instructions to generate target block system models}

Instructions are specific to a given target block modeling system (either
scicos or scicos-pro or modelica block system. In addition, instructions
are bound to the block modeling system host language (for instance Nsp or
Scilab for Scicos, Oml, Hml, or Nsp for scicos-pro, or modelica for the
modelicos block modeling system).

In the following type names, string 'host_' denotes any target host language;
similarly, string 'modeler_' denotes any target block modeling system. For
instance, 'host_ident' is the type of identifiers of the (target block
modeling system) host language.

*)

(** Borrowed from module Ast_types. *)

type natural = Ast_types.natural;;
type mdl_text = Ast_types.mdl_text;;

type diagram_name = Ast_types.mdl_name;;

(* New type definitions *)

(* FIXME hostid hould be an ident ? *)
type host_id = string;;

type host_ident =
  | Hi of host_id
;;

type host_block_ident =
  | Hbi of host_id
;;

type host_block_tag_ident =
  | Hbti of string
;;

type host_diagram_ident =
  | Hdi of host_id
;;

type host_model_ident =
  | Hmi of host_id
;;

type host_function_ident =
  | Hfi of host_id
;;

type host_string =
  | Host_string of string
;;

type modeler_font_size =
  | Sfs of mdl_text
;;

type modeler_font_value =
  | Sfv of natural
;;

type modeler_color =
  | Sc of Mdl_basics_deep_ast.rgb_color
;;

type modeler_block_type =
  | Sbt of string
;;

type modeler_ident =
  | Si of string
;;

type matlab_script =
  | Matlab_script of mdl_text
;;

type code_value = Mdl_basics_explicit_ast.explicit_value;;

type modeler_block_parameter_bindings =
  modeler_block_parameter_binding Ulist1.t
and modeler_block_parameter_binding =
  modeler_block_parameter_value Binding.ident_binding

and modeler_block_parameter_value =
  | Parameter_value of code_value
  | Parameter_struct of (code_value list Binding.ident_binding) Ulist1.t
;;

type evaluation_mode = Mdl_basics_deep_ast.deep_mask_parameter_evaluation;;
type enable_mode = Mdl_basics_deep_ast.deep_mask_parameter_enable;;
type visible_mode = Mdl_basics_deep_ast.deep_mask_parameter_visible;;
type mask = Mdl_basics_explicit_ast.explicit_mask;;
type mask_parameter =
  Mdl_basics_explicit_ast.explicit_mask_parameter
;;
type mask_parameter_value =
  Mdl_basics_explicit_ast.explicit_mask_parameter_value
;;

type context =
  | Context_code of matlab_script
  | Context_bindings of modeler_block_parameter_bindings
  | Context_commented of modeler_block_parameter_bindings
;;

type workspace = Mdl_implementation_explicit_ast.explicit_model_workspace
;;

type directed_connector_src =
  | Dcs of host_block_tag_ident * Mdl_rename_ast.rename_directed_connector
and directed_connector_dst =
  | Dcd of host_block_tag_ident * Mdl_rename_ast.rename_directed_connector
;;

type connection_connector_src =
  | Ccs of host_block_tag_ident * Mdl_rename_ast.rename_connection_connector
and connection_connector_dst =
  | Ccd of host_block_tag_ident * Mdl_rename_ast.rename_connection_connector
;;

type modeler_point = {
  sp_x : float;
  sp_y : float;
}
;;

type modeler_size = {
  ss_w : float;
  ss_h : float;
}
;;

type modeler_position = {
  sp_xmin : float;
  sp_xmax : float;
  sp_ymin : float;
  sp_ymax : float;
}
;;

type modeler_simulation_parameters = {
  ssp_abs_tol : code_value;
  ssp_rel_tol : code_value;
  ssp_tim_tol : code_value;
  ssp_delta_t : code_value;
  ssp_rt_scale : code_value;
  ssp_solver : code_value;
  ssp_max_step : code_value;
}
;;

type code = instruction Concat_list.clist

and instruction =
  | Add_block of
      host_diagram_ident * host_block_tag_ident *
      host_block_ident * Mdl_ast.block_name
  | Add_connection_link of
      host_diagram_ident *
      connection_connector_src * connection_connector_dst *
      modeler_point list
  | Add_directed_link of
      host_diagram_ident *
      directed_connector_src * directed_connector_dst *
      modeler_point list
  | Add_event_link of
      host_diagram_ident *
      directed_connector_src * directed_connector_dst *
      modeler_point list
  | Define_function of
      host_diagram_ident * host_function_ident *
      host_ident list * code
  | Evaluate_model of host_model_ident * host_diagram_ident
  | Fill_super_block of host_block_ident * host_diagram_ident
  | Instantiate_block of host_block_ident * modeler_block_type
  | Instantiate_super_block of host_block_ident
  | Instantiate_diagram of host_diagram_ident
  | Mask_block of host_block_ident * context * mask
  | Set_block_bg_color of host_block_ident * modeler_color
  | Set_block_fg_color of host_block_ident * modeler_color
  | Set_block_evtnin of host_block_ident * natural
  | Set_block_evtnout of host_block_ident * natural
  | Set_block_flip of host_block_ident * bool
  | Set_block_icon of host_block_ident * matlab_script
  | Set_block_icon_text of host_block_ident * mdl_text
  | Set_block_ident of host_block_ident * Mdl_ast.block_name
  | Set_block_mask of host_block_ident * mask
  | Set_block_nin of host_block_ident * natural
  | Set_block_nout of host_block_ident * natural
  | Set_block_origin of host_block_ident * modeler_point
  | Set_block_parameters of host_block_ident * modeler_block_parameter_bindings
  | Set_block_size of host_block_ident * modeler_size
  | Set_block_theta of host_block_ident * float
  | Set_diagram_3d of host_diagram_ident * bool
  | Set_diagram_bg_color of host_diagram_ident * modeler_color
  | Set_diagram_context of host_diagram_ident * context
  | Set_diagram_position of host_diagram_ident * modeler_position
  | Set_diagram_name of host_diagram_ident * diagram_name
  | Set_diagram_zoom of host_diagram_ident * float
  | Set_function_call of
      host_diagram_ident * host_function_ident * host_ident list

  | Set_model_close_fcn of host_model_ident * matlab_script
  | Set_model_final_simulation_time of host_model_ident * code_value
  | Set_model_init_fcn of host_model_ident * matlab_script
  | Set_model_initial_simulation_time of host_model_ident * code_value
  | Set_model_postload_fcn of host_model_ident * matlab_script
  | Set_model_postlude of host_model_ident * matlab_script
  | Set_model_postsave_fcn of host_model_ident * matlab_script
  | Set_model_preload_fcn of host_model_ident * matlab_script
  | Set_model_prelude of host_model_ident * matlab_script
  | Set_model_presave_fcn of host_model_ident * matlab_script
  | Set_model_start_fcn of host_model_ident * matlab_script
  | Set_model_stop_fcn of host_model_ident * matlab_script
  | Set_model_workspace of host_model_ident * workspace

  | Set_model_solver_parameters of
      host_model_ident * modeler_simulation_parameters

  | Set_super_block_atomic_property of host_block_ident * bool

  | Set_target_modeler of Simport_configuration.target_modeler

  | Comment of string
;;

type t = instruction;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
