(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_translation.ml,v 1.1 2019-05-10 14:04:02 jpc Exp $ *)

(** From explicit syntax with splitting blocks and Scicos-style lines,
   to real target host language code.
*)

open Ast_types;;
open Mdl_basics_deep_ast;;
open Instructions;;
open Genr_utils;;

(* FIXME: should be suppressed!

let float = Unatural.to_float;;

let diagram_left_margin =
  float (Configuration.get_diagram_left_margin ())
and diagram_right_margin =
  float (Configuration.get_diagram_right_margin ())
and diagram_bottom_margin =
  float (Configuration.get_diagram_bottom_margin ())
;;

let block_vertical_spacing =
  float (Configuration.get_block_vertical_spacing ())
;;

let block_port_size = {
  ss_w = float (Configuration.get_block_port_width ());
  ss_h = float (Configuration.get_block_port_height ());
}
;;

let window_size = {
   ss_w = float (Configuration.get_window_width ());
   ss_h = float (Configuration.get_window_height ());
}
;;
 FIXME: end of should be suppressed *)

(* FIXME:
  font value and font size should be command line settable and incorporated
  in the simport configuration module.

TODO: scicos font value is just a number specific to the scios/sicos_pro
modelers identified fonts.
Should be pull out of simport in the API of the modeler...
*)
let scicos_font_value = Sfv Unatural._2_;;

let modeler_font_value _ = scicos_font_value;;

(*
let modeler_font_size explicit_font =
  (* Return the natural size of the font. *)
  explicit_font.explicit_font_size
;;
*)
let scicos_font_size = Sfs (mdl_text_of_string "4");;

let modeler_font_size _ = scicos_font_size;;

let modeler_rgb r g b =

  let r = float_of_int r /. 255.0
  and g = float_of_int g /. 255.0
  and b = float_of_int b /. 255.0 in

  Sc { Mdl_basics_deep_ast.red = r; green = g; blue = b; }
;;

let modeler_color = function
  | Mdl_basics_deep_ast.Deep_black -> (* 1 *)
    modeler_rgb 0 0 0
  | Mdl_basics_deep_ast.Deep_blue -> (* 2 *)
    modeler_rgb 0 0 255
  | Mdl_basics_deep_ast.Deep_green -> (* 3 *)
    modeler_rgb 0 255 0
  | Mdl_basics_deep_ast.Deep_cyan -> (* 4 *)
    modeler_rgb 0 255 255
  | Mdl_basics_deep_ast.Deep_red -> (* 5 *)
    modeler_rgb 255 0 0
  | Mdl_basics_deep_ast.Deep_yellow -> (* 7 *)
    modeler_rgb 255 255 0
  | Mdl_basics_deep_ast.Deep_white -> (* 8 *)
    modeler_rgb 255 255 255
  | Mdl_basics_deep_ast.Deep_darkblue -> (* 9 *)
    modeler_rgb 0 0 0x8b

  | Mdl_basics_deep_ast.Deep_lightblue -> (* 12 *)
    modeler_rgb 0xad 0xd8 0xe6
  | Mdl_basics_deep_ast.Deep_darkgreen -> (* 13 *)
    modeler_rgb 0 0x64 0

  | Mdl_basics_deep_ast.Deep_lightgreen -> (* 15 *)
    modeler_rgb 0x90 0xee 0x90

  | Mdl_basics_deep_ast.Deep_magenta -> (* 19 *)
    modeler_rgb 255 0 255
  | Mdl_basics_deep_ast.Deep_darkred -> (* 20 *)
    modeler_rgb 0x8b 0 0
  | Mdl_basics_deep_ast.Deep_lightred -> (* 21 *)
    modeler_rgb 0xff 0x47 0x4c

  | Mdl_basics_deep_ast.Deep_orange -> (* 27 *)
    modeler_rgb 0xFF 0xA5 0

  | Mdl_basics_deep_ast.Deep_grey -> (* 33 *)
    modeler_rgb 0xBE 0xBE 0xBE
  | Mdl_basics_deep_ast.Deep_lightgrey ->
    modeler_rgb 0xD3 0xD3 0xD3
  | Mdl_basics_deep_ast.Deep_darkgrey ->
    modeler_rgb 0xA9 0xA9 0xA9
  | Mdl_basics_deep_ast.Deep_purple ->
    modeler_rgb 0x80 0x00 0x80
  | Mdl_basics_deep_ast.Deep_rgb rgb_color -> Sc rgb_color
;;

let modeler_evaluation_commented_context ebp_table =
  let evaluation_context = Hash_table.values ebp_table in
  List.map
    (fun bdg ->
     Genr_utils.block_parameter_binding
       (Si (string_of_ident bdg.Binding.bound_ident),
        bdg.Binding.bound_value))
    evaluation_context
;;

type block_name = mdl_name
and block_tag = host_block_tag_ident
and translation_table = {
  mutable tag_count : int;
  mutable diagram_count : int;
  mutable block_count : int;
  mutable diagram_ident : host_diagram_ident;
  mutable block_ident : host_block_ident;
          table : (block_name, block_tag) Hashtbl.t;
}
;;

let new_translation_table () = {
  tag_count = 0;
  diagram_count = 0;
  block_count = 0;
  diagram_ident = host_diagram_ident "scsm";
  block_ident = host_block_ident "blk";
  table = Hashtbl.create 17;
}
;;

let get_diagram_ident table = table.diagram_ident;;
let get_block_ident table = table.block_ident;;
let get_block_tag translation_table block_name =
  let tbl = translation_table.table in
  if Hashtbl.mem tbl block_name
  then Hashtbl.find tbl block_name else
  Genr_block.translation_error
    (Printf.sprintf
       "block tag table: Simulink block %S has no associated \
        tag identifier."
       (string_of_mdl_name block_name))
;;

let set_diagram_ident table =
  let count = table.diagram_count in
  let name =
    if count = 0 then "scsm" else
    Printf.sprintf "scsm%d" (count - 1) in
  let diagram_ident = host_diagram_ident name in
  table.diagram_ident <- diagram_ident;
  diagram_ident;
;;

let set_block_ident table =
  let count = table.block_count in
  let name =
    if count = 0 then "blk" else
    Printf.sprintf "blk%d" (count - 1) in
  let block_ident = host_block_ident name in
  table.block_ident <- block_ident;
  block_ident
;;

let set_block_tag translation_table block_name block_tag =
(*prerr_endline (Printf.sprintf "Adding %S in the tag table" block_name);*)
  let tbl = translation_table.table in
  if Hashtbl.mem tbl block_name then
    Say.fatal_error (
      Printf.sprintf
        "block tag table: attempt to store %S twice"
        (string_of_mdl_name block_name));
  Hashtbl.add tbl block_name block_tag
;;

let new_diagram_ident table =
  let count = table.diagram_count in
  table.diagram_count <- count + 1;
  set_diagram_ident table
;;

let new_block_ident table =
  let count = table.block_count in
  table.block_count <- count + 1;
  set_block_ident table
;;

let new_block_tag table =
  let count = table.tag_count in
  table.tag_count <- count + 1;
  let tag =
    if count = 0 then "block_tag" else
    Printf.sprintf "block_tag%d" (count - 1) in
  host_block_tag_ident tag
;;

let genr_table = new_translation_table ();;
let reset_genr_table =
  let reset_translation_table translation_table table =
    translation_table.tag_count <- table.tag_count;
    translation_table.diagram_count <- table.diagram_count;
    translation_table.block_count <- table.block_count;
    ignore (set_diagram_ident translation_table);
    ignore (set_block_ident translation_table);
    Hashtbl.clear translation_table.table in
  reset_translation_table genr_table
;;

let new_genr_diagram_ident () = new_diagram_ident genr_table;;
let new_genr_block_ident () = new_block_ident genr_table;;

let add_block translation_table diagram_id block_id block_name =
  let block_tag = new_block_tag translation_table in
  set_block_tag translation_table block_name block_tag;
  Add_block (diagram_id, block_tag, block_id, block_name)
;;

let annotation_block_name = mdl_name_of_string "Annotation Block";;

let add_annotation_block translation_table diagram_id block_id =
  let block_tag = new_block_tag translation_table in
  Add_block (diagram_id, block_tag, block_id, annotation_block_name)
;;

(* FIXME: should be removed!
let evaluation_mode_of_deep_value dv =
  let mode =
    Mdl_basics_to_explicit_ast.bool_of_deep_value
      "Property evaluate of Simulink.MaskParameter is not a boolean" dv in
  match mode with
  | true -> Mdl_basics_deep_ast.Eval
  | false -> Mdl_basics_deep_ast.Quote
;;
*)

let bool_of_evaluation_mode = function
  | Mdl_basics_deep_ast.Eval -> true
  | Mdl_basics_deep_ast.Quote -> false
;;

let default_visible_mode = true;;

let bool_of_visible_mode_opt = function
  | None -> default_visible_mode
  | Some Mdl_basics_deep_ast.Visible -> true
  | Some Mdl_basics_deep_ast.Hidden -> false
;;

let default_enable_mode = true;;

let bool_of_enable_mode_opt = function
  | None -> default_enable_mode
  | Some Enable -> true
  | Some Disable -> false
;;

let target_string_of_parameter_style = function
  | Deep_mask_parameter_edit -> "edit"
  | Deep_mask_parameter_checkbox -> "checkbox"
  | Deep_mask_parameter_popup _ -> "popup"
  | Deep_mask_parameter_unidt _ -> "unidt"
  | Deep_mask_parameter_min -> "min"
  | Deep_mask_parameter_max -> "max"
  | Deep_mask_parameter_promote -> "promote"
  | Deep_mask_parameter_radiobutton _ -> "radiobutton"
  | Deep_mask_parameter_slider _ -> "slider"
  | Deep_mask_parameter_spinbox _ -> "spinbox"
;;

let target_style_of_parameter_style ps =
  mdl_text_of_string (target_string_of_parameter_style ps)
;;

let transcode_text txt =
  let src_encoding = Configuration.get_source_encoding () in
  let dst_encoding = Configuration.get_target_encoding () in
  let str_txt = Ast_types.string_of_mdl_text txt in
  Ast_types.mdl_text_of_string
    (Wstring.transcode src_encoding str_txt dst_encoding)
;;

let print_text_as_raw_string_token ppf txt =
  M_to_host.print_quoted_text ppf
    (M_to_host.quote_text txt)
;;

let print_text_as_encoded_string_token ppf txt =
  let txt =
    if Configuration.get_ignore_encoding () then txt else
      transcode_text txt in
  print_text_as_raw_string_token ppf txt
;;

(* FIXME
let print_text ppf txt =
  Format.fprintf ppf "%s" (Ast_types.string_of_mdl_text txt)
;;
*)

let print_mask_parameter_style_item ppf = function
  | Item_int int ->
    print_text_as_raw_string_token ppf
      (mdl_text_of_string (string_of_int int))
  | Item_text mdl_text ->
    print_text_as_encoded_string_token ppf mdl_text
;;

let print_mask_range_item ppf range_item =
  print_text_as_raw_string_token ppf (mdl_text_of_string range_item)
;;

let print_comma_separated pp_x ppf = function
  | [] -> ()
  | x :: xs ->
    Format.fprintf
      ppf "%a%a"
      pp_x x
      (fun ppf ->
       List.iter
         (fun x ->
          Format.fprintf
            ppf ", %a"
            pp_x x))
      xs
;;

let print_mask_parameter_style_items style ppf style_items =
  Say.fdebug
    (fun epf ->
     Format.fprintf
       epf "print_parameter_style_items: printing %s_items: %a"
       style
       Mdl_basics_deep_ast_print.print_mask_parameter_style_items
       style_items);
  print_comma_separated print_mask_parameter_style_item ppf style_items
;;

let print_mask_parameter_range style ppf range =
  Say.fdebug
    (fun epf ->
     Format.fprintf
       epf "print_mask_parameter_range: printing %s_items: %a"
       style
       Mdl_basics_deep_ast_print.print_mask_parameter_style_range
       range);
  let range =
    match range with
    | Mdl_basics_deep_ast.Mask_parameter_style_range range -> range in
  print_comma_separated print_mask_range_item ppf range
;;

let print_mask_parameter_range_opt style ppf = function
  | None -> ()
  | Some range -> print_mask_parameter_range style ppf range
;;

let print_parameter_style_items ppf = function
  | Deep_mask_parameter_edit
  | Deep_mask_parameter_checkbox
  | Deep_mask_parameter_min
  | Deep_mask_parameter_max
  | Deep_mask_parameter_promote -> ()
  | Deep_mask_parameter_radiobutton radiobutton_items ->
    print_mask_parameter_style_items "radiobutton" ppf radiobutton_items
  | Deep_mask_parameter_popup popup_items ->
    print_mask_parameter_style_items "popup" ppf popup_items
  | Deep_mask_parameter_unidt mdl_text ->
    Format.fprintf ppf "%a"
      print_text_as_encoded_string_token mdl_text
  | Deep_mask_parameter_slider range_opt ->
    print_mask_parameter_range_opt "slider" ppf range_opt
  | Deep_mask_parameter_spinbox range_opt ->
    print_mask_parameter_range_opt "spinbox" ppf range_opt
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
