(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: model_to_code.ml,v 1.1 2019-05-10 14:04:02 jpc Exp $ *)

(** From split form tree to symbolic code for the
  target block modeling system API. *)
open Ast_types;;
open Mdl_ast;;
open Mdl_basics_deep_ast;;
open Mdl_basics_explicit_ast;;
open Mdl_implementation_explicit_ast;;
open Mdl_split_ast;;
open Mdl_translation;;
open Instructions;;
open Genr_utils;;
open Concat_list;;

let string_of_host_block_ident (Hbi s) = s;;

let find_deep_binding id bds =
  List.find (fun bd -> bd.Binding.bound_ident = id) bds
;;

let find_deep_value id bds =
  (find_deep_binding id bds).Binding.bound_value
;;

let scicos_point position { deep_x = x; deep_y = y; } =
  {
    sp_x = float_of_int x;
    sp_y =
      float_of_int (position.ymax - position.ymin - y);
  }
;;

(* FIXME: should properly define a function point specific to the target
   block system *)
let modeler_point = scicos_point;;

let scicos_geometry height theta
    {
      xmin = xmin; ymin = ymin;
      xmax = xmax; ymax = ymax;
    } =
  let xmin = float_of_int xmin and ymin = float_of_int ymin
  and xmax = float_of_int xmax and ymax = float_of_int ymax in
  let w = xmax -. xmin in
  let h = ymax -. ymin in

  if Unatural.rem theta (Unatural_types.to_natural 180) <> Unatural._0_
  then
    let size = {
      ss_w = h;
      ss_h = w;
    } in
    let origin = {
      sp_x = xmin +. size.ss_h /. 2. -. size.ss_w /. 2.;
      sp_y = height -. ymin -. size.ss_h /. 2. -. size.ss_w /. 2.;
    } in
    (origin, size)
  else
    let size = {
      ss_w = w;
      ss_h = h;
    } in
    let origin = {
      sp_x = xmin;
      sp_y = height -. ymin -. size.ss_h;
    } in
    (origin, size)
;;

(* FIXME: should properly define a geometry function specific to the target
   block modeling system *)
let modeler_geometry = scicos_geometry;;

let (_dummy_origin, _dummy_size as scicos_dummy_geometry) =
  {
    sp_x = -1.0;
    sp_y = -1.0;
  },
  {
    ss_w = 0.0;
    ss_h = 0.0;
  }
;;

let modeler_dummy_geometry = scicos_dummy_geometry;;

let get_modeler_geometry height theta position =
  if Mdl_make_block.dummy_position = position
  then modeler_dummy_geometry
  else modeler_geometry height theta position
;;

(* FIXME: should properly define a function modeler_origin specific to the
   target block modeling system *)
let modeler_origin position { xpos = x; ypos = y; maxpos = maxpos; } =
  match maxpos with
  | None -> modeler_point position { deep_x = x; deep_y = y; }, None
  | Some { xmaxpos = xmaxpos; ymaxpos = ymaxpos; } ->
    let origin, size =
      modeler_geometry
        (float_of_int (position.ymax - position.ymin))
        Unatural._0_
        {
          xmin = x; ymin = y;
          xmax = xmaxpos; ymax = ymaxpos;
        } in
    origin, Some size
;;

let scicos_link_point dpoint =
  {
    sp_x = float_of_int dpoint.deep_x;
    sp_y = float_of_int (- dpoint.deep_y);
  }
;;

(* FIXME: should properly define a function link_point specific to the target
   block modeling system *)
let modeler_link_point = scicos_link_point;;

let modeler_link_points = List.map modeler_link_point;;

let scicos_position position =
  let xmax = float_of_int position.xmax
  and ymax = float_of_int position.ymax
  and xmin = float_of_int position.xmin
  and ymin = float_of_int position.ymin in
  {
    sp_xmin = xmin;
    sp_xmax = xmax;
    sp_ymin = ymin;
    sp_ymax = ymax;
  }
;;

(* FIXME: should properly define a function position specific to the target
   block system *)
let modeler_position = scicos_position;;

let make_genr_info
    explicit_block_type explicit_block_name known
    explicit_block_all_bindings
    explicit_instance_data_opt
    system_io_descriptor
    links_by_dst =
  let block_ports =
    explicit_block_all_bindings.
      explicit_block_graphical.explicit_block_ports in
  let ports_description =
    explicit_block_all_bindings.
      explicit_block_graphical.explicit_block_ports_description in
  let block_parameters =
    explicit_block_all_bindings.explicit_block_parameters
  and block_bindings =
    explicit_block_all_bindings.explicit_block_bindings in
  let ebp_table =
    if known then block_parameters else block_bindings in
  let activation_needed =
    match ports_description with
    | {
        Mdl_basics_deep_ast.
        port_enable = pe;
        port_trigger = pt;
        port_ifaction = pi;
        _;
      } ->
      pe || pt || pi in

  {
    Genr_block_types.
    block_arguments = ebp_table;
    block_bindings = block_bindings;
    block_inport_names = links_by_dst;
    block_ports = block_ports;
    block_ports_description = ports_description;
    block_parent_system_io_descriptor = system_io_descriptor;
    block_port_activation_needed = activation_needed;
    block_explicit_type = explicit_block_type;
    block_explicit_name = explicit_block_name;
    block_explicit_instance_data = explicit_instance_data_opt;
  }
;;

(* Get a translator for a block type.
   This function never fails: it finds the proper registered translator for
   the block type, or return the translator associated to blocks with no
   translator yet! *)
let get_translator explicit_block_type =
  try Genr_block.find_translator explicit_block_type with
  | Not_found ->
    let ty_name =
      Mdl_basics_to_explicit_ast.
      name_of_explicit_block_type explicit_block_type in
    Say.warning (Format.sprintf
      "%S Simulink block translation is not yet available."
      ty_name);
    Genr_block.find_translator
      Mdl_fake_block_types.not_yet_translated_block_type
;;

(* Compiling links *)

let get_ports_of_rename_directed_connectors
    translation_table
    rename_directed_connectors =
  let get_directed_src_port = function
    | {
        explicit_connector_port_block = src_block_name;
        explicit_connector_port_desc = _port_number;
      } as connector_src ->
      let src_block_tag =
        get_block_tag translation_table src_block_name in
      Dcs (src_block_tag, connector_src) in

  let get_directed_dst_port = function
    | {
        explicit_connector_port_block = dst_block_name;
        explicit_connector_port_desc = _port_number;
      } as connector_dst ->
      let dst_block_tag =
        get_block_tag translation_table dst_block_name in
      Dcd (dst_block_tag, connector_dst) in

  let split_link_src =
    rename_directed_connectors.Mdl_split_ast.split_link_src
  and split_link_dst =
    rename_directed_connectors.Mdl_split_ast.split_link_dst in

  get_directed_src_port split_link_src,
  get_directed_dst_port split_link_dst
;;

let get_ports_of_rename_connection_connectors
    translation_table
    rename_connection_connectors =
(* Waiting the right instruction for connection links... *)

  let get_connection_src_port = function
    | {
        explicit_connector_port_block = src_block_name;
        explicit_connector_port_desc =
         (_rename_connection_connector_desc :
          Mdl_rename_ast.rename_connection_connector_desc);
      } as connector_src ->
(* RQ
 Should
  match rename_connection_connector_desc with
  | Explicit_left_connector port_number ->
     Csl (src_block_tag, connector_src)
  | Explicit_right_connector port_number ->
     Csr (src_block_tag, connector_src)

Since type explicit_connection_connector_desc is defined as:
and explicit_connection_connector_desc =
  | Explicit_left_connector of port_number
  | Explicit_right_connector of port_number
*)
      let src_block_tag = get_block_tag translation_table src_block_name in
      Ccs (src_block_tag, connector_src) in

  let get_connection_dst_port = function
    | {
        explicit_connector_port_block = dst_block_name;
        explicit_connector_port_desc =
          (_rename_connection_connector_desc :
           Mdl_rename_ast.rename_connection_connector_desc);
      } as connector_dst ->
      let dst_block_tag =
        get_block_tag translation_table dst_block_name in
      Ccd (dst_block_tag, connector_dst) in

  let split_link_src =
    rename_connection_connectors.Mdl_split_ast.split_link_src
  and split_link_dst =
    rename_connection_connectors.Mdl_split_ast.split_link_dst in

  get_connection_src_port split_link_src,
  get_connection_dst_port split_link_dst
;;

let compile_link translation_table scsm_id
    ({
      Mdl_split_ast.
      split_link_name = _;
      split_link_font = _;
      split_link_labels = _;
      split_link_points = link_points;
      split_link_desc = link_desc;
    } : Mdl_rename_ast.rename_link) =

  let points =
    match link_points with
    | None -> []
    | Some points -> points in

  let points = modeler_link_points points in

  match link_desc with
  | Mdl_rename_ast.Rename_directed_link rename_directed_connectors ->
    let src_port, dst_port =
      get_ports_of_rename_directed_connectors
        translation_table
        rename_directed_connectors in
    Add_directed_link (scsm_id, src_port, dst_port, points)

  | Mdl_rename_ast.Rename_event_link rename_event_connectors ->
    let src_port, dst_port =
      get_ports_of_rename_directed_connectors
        translation_table rename_event_connectors in
    Add_event_link (scsm_id, src_port, dst_port, points)

  | Mdl_rename_ast.Rename_connection_link rename_connection_connectors ->
    let src_port, dst_port =
      get_ports_of_rename_connection_connectors
        translation_table
        rename_connection_connectors in
    Add_connection_link (scsm_id, src_port, dst_port, points)
;;

let compile_links translation_table scsm_id explicit_links =
  Concat_list.list
  (*
     If you want the links to be listed in reverse order in the
     generated files, call rev_map instead of map
  *)
    (List.rev_map (compile_link translation_table scsm_id) explicit_links)
  (*
    (List.map (compile_link translation_table scsm_id) explicit_links)
  *)
;;

(* Compiling annotations. *)
let compile_annotations
    table scsm_id block_id system_position explicit_annotations =

  let rec loop accu = function
    | [] -> accu
    | {
        explicit_annotation_name = mdl_text;
        explicit_annotation_position = explicit_annotation_position;
        explicit_annotation_font = explicit_font;
        explicit_annotation_graphical = _explicit_font_and_colors;
        explicit_annotation_graphical_defaults =
         _explicit_annotation_graphical_defaults;
      } :: annotations ->
      assert (String.length (string_of_mdl_text mdl_text) <> 0);
      let parameter_bindings =
        Genr_utils.block_parameter_binding
          (Si "txt",
           Deep_string (string_of_mdl_text mdl_text)),
        [
          block_parameter_binding
            (Si "font",
             let (Sfv v) = Mdl_translation.modeler_font_value explicit_font in
             Deep_int (Unatural.to_string v));
          block_parameter_binding
            (Si "siz",
             let (Sfs v) = Mdl_translation.modeler_font_size explicit_font in
             Deep_expression v);
        ] in

      let add_annotation_instruct =
        add_annotation_block table scsm_id block_id in

      let block_body =

        let orig, size_opt =
          modeler_origin system_position explicit_annotation_position in

        let l = [
          Set_block_parameters (block_id, parameter_bindings);
          add_annotation_instruct;
        ] in
        let l =
          match size_opt with
          | None -> l
          | Some size -> Set_block_size (block_id, size) :: l in
        Instantiate_block
          (block_id, Genr_utils.get_modeler_annotation_type ()) ::
        Set_block_origin (block_id, orig) ::
        l in

      loop (Concat_list.concat accu (list block_body)) annotations in

  loop (list []) explicit_annotations
;;

let compile_graphics height ebg block_id genr_info =

  let bg, fg =
    modeler_color (ebg.explicit_block_background_color),
    modeler_color (ebg.explicit_block_foreground_color) in

  let block_mirror = ebg.explicit_block_mirror
  and theta = ebg.explicit_block_rotation in

  let block_name = genr_info.Genr_block_types.block_explicit_name in

  let add_show_name_instruction code =
    if ebg.explicit_block_show_name
    then Set_block_ident (block_id, block_name) :: code
    else code in

  let origin, size =
    get_modeler_geometry height theta ebg.explicit_block_position in

  let ports_desc = genr_info.Genr_block_types.block_ports_description in

  let nin, nout, niin, niout, evtnin, evtnout =
    Mdl_ast_to_deep_ast.get_regular_port_nin ports_desc,
    Mdl_ast_to_deep_ast.get_regular_port_nout ports_desc,
    Mdl_ast_to_deep_ast.get_implicit_port_nin ports_desc,
    Mdl_ast_to_deep_ast.get_implicit_port_nout ports_desc,
Unatural._0_,
Unatural._0_
(* TODO
    Mdl_ast_to_deep_ast.get_event_port_nin ports_desc,
    Mdl_ast_to_deep_ast.get_event_port_nout ports_desc
*) in

  Set_block_bg_color (block_id, bg) ::
  Set_block_fg_color (block_id, fg) ::
  Set_block_nin (block_id, Unatural.add nin niin) ::
  Set_block_nout (block_id, Unatural.add nout niout) ::
  Set_block_evtnin (block_id, evtnin) ::
  Set_block_evtnout (block_id, evtnout) ::
  Set_block_flip (block_id, block_mirror) ::
  add_show_name_instruction [
    Set_block_origin (block_id, origin);
    Set_block_size (block_id, size);
    Set_block_theta (block_id, Unatural.to_float theta);
  ]
;;

(* The name of the host language function that generates the sub-system. *)
let make_sub_system_function_ident index =
  let fun_name = Printf.sprintf "subsystem_%d" index in
  index + 1,
  Genr_utils.host_function_ident fun_name
;;

let translate_block table translator explicit_block_type block_id genr_info =
  Mdl_translation.reset_genr_table table;
  Say.fdebug (fun epf ->
    Format.fprintf epf "Translating block %S@ (with type %a)"
      (string_of_mdl_name genr_info.Genr_block_types.block_explicit_name)
      Mdl_basics_explicit_ast_print.print_explicit_block_type
        explicit_block_type);
  translator block_id genr_info
;;

let explicit_block_type blk =
  match blk.explicit_block_desc with
  | Explicit_block_regular explicit_simple_block_desc ->
    let explicit_simple_block_type =
      explicit_simple_block_desc.explicit_simple_block_type in
    Explicit_block_regular_type explicit_simple_block_type
  | Explicit_block_reference explicit_simple_block_desc ->
    let explicit_simple_block_type =
      explicit_simple_block_desc.explicit_simple_block_type in
    Explicit_block_reference_type explicit_simple_block_type
  | Explicit_block_s_function _explicit_s_function_desc ->
    Explicit_block_regular_type Mdl_block_types.S_Function
  | Explicit_block_sub_system _ ->
    Explicit_block_regular_type Mdl_block_types.SubSystem
;;

(* Link names normalization:
   sometimes when link names are generated, they are
   enclosed by angle brackets, we remove them in this case. *)
let normalize_link_name link_name =
  let link_string = Ast_types.string_of_mdl_name link_name in
  let len = String.length link_string in
  if len >= 2 &&
     link_string.[0] = '<' && link_string.[pred len] = '>'
  then Ast_types.mdl_name_of_string (String.sub link_string 1 (len - 2))
  else link_name
;;

let group_links_by_dst links =
  let hashtbl = Hashtbl.create 10 in
  let collect_input = function
    | ({
        Mdl_split_ast.
        split_link_name = link_name_option;
        split_link_desc = link_desc;
        _;
      } : Mdl_rename_ast.rename_link) ->
      match link_name_option with
      | None -> ()
      | Some link_name ->
        let link_name = normalize_link_name link_name in
        match link_desc with
        | Mdl_rename_ast.Rename_directed_link directed_connectors
        | Mdl_rename_ast.Rename_event_link directed_connectors ->
          let { split_link_dst = link_destination; _ } =
            directed_connectors in
          begin match link_destination with
          | {
              explicit_connector_port_block = connector_block;
              explicit_connector_port_desc =
                Ast_types.Mdl_port_number _ as port_number;
            } ->
            let block_inputs =
              try
                Hashtbl.find hashtbl connector_block
              with
              | Not_found -> [] in
            Hashtbl.replace hashtbl connector_block
              ((link_name, port_number) :: block_inputs)
          end
        | Mdl_rename_ast.Rename_connection_link
            _rename_connection_connectors ->
          begin
          (* TODO: complete this clause ? *)
          assert false
          end in
  List.iter collect_input links;

  let info_hashtbl = Hashtbl.create (Hashtbl.length hashtbl) in
  Hashtbl.iter
    (fun connector_block block_inputs ->
     let local_hashtbl = Hashtbl.create 3 in
     let add_input (name, number) =
       Hashtbl.add local_hashtbl number name in
     List.iter add_input block_inputs;
     Hashtbl.add info_hashtbl connector_block local_hashtbl)
    hashtbl;
  info_hashtbl
;;

let pmio_port_side_of_string = function
  | "Left" -> Genr_block_types.Pmio_port_left
  | "Right" -> Genr_block_types.Pmio_port_right
  | s -> Say.fatal_error (Printf.sprintf "Unexpected port side '%s'" s)
;;

let caml_index idx = Unatural.pred idx;;

let get_pmio_port_index_and_side block =

  let port_nat =
    let deep_port_nat_value =
      Mdl_explicit_block_utils.
      get_block_parameter_value block Mdl_ast_functions.port_ident in
    Mdl_basics_to_explicit_ast.
    natural_of_deep_value
      "Port parameter of PMIOPort block should contain a natural value"
      deep_port_nat_value in
  let side_string =
    let deep_side_string_value =
      Mdl_explicit_block_utils.
      get_block_parameter_value block Mdl_ast_functions.side_ident in
    Mdl_basics_to_explicit_ast.string_of_deep_value
      "Port parameter Side should contain a string value"
      deep_side_string_value in

  caml_index port_nat, pmio_port_side_of_string side_string
;;

let make_system_io_descriptor blocks =
  let system_nin, system_nout, system_pmio_port_index_sides =
    let rec loop (nin, nout, index_sides as accu) = function
      | [] -> accu
      | block :: blocks ->
        let explicit_block_type = explicit_block_type block in
        match explicit_block_type with
        | Explicit_block_regular_type explicit_simple_block_type ->
          if explicit_simple_block_type = Mdl_block_types.Inport then
            loop (Unatural.succ nin, nout, index_sides) blocks else
          if explicit_simple_block_type = Mdl_block_types.Outport then
            loop (nin, Unatural.succ nout, index_sides) blocks else
          if explicit_simple_block_type = Mdl_block_types.PMIOPort then
            let index_sides =
              get_pmio_port_index_and_side block :: index_sides in
            loop (nin, nout, index_sides) blocks else
          loop (nin, nout, index_sides) blocks
        | Explicit_block_reference_type _ ->
          loop (nin, nout, index_sides) blocks in
    loop (Unatural._0_, Unatural._0_, []) blocks in
  (* Find the maximum index of pmioports to get the number of such ports. *)
  let max_index =
    List.fold_left
      (fun accu (index, _) -> Unatural.max accu index)
      Unatural._0_ system_pmio_port_index_sides in
  let length = Unatural.succ max_index in
  let system_pmio_port_sides =
    Array.make (Unatural.to_int length) None in

  List.iter
    (fun (nat, side) ->
     let idx = Unatural.to_int nat in
     system_pmio_port_sides.(idx) <- Some side)
    system_pmio_port_index_sides;

  {
    Genr_block_types.
    system_pmio_port_sides = system_pmio_port_sides;
    system_inport_number = system_nin;
    system_outport_number = system_nout;
  }
;;

let rec compile_blocks
        table system_position index system_funs blocks links_by_dst =

  let height =
    float_of_int (system_position.ymax - system_position.ymin) in

  let system_io_descriptor = make_system_io_descriptor blocks in

  let rec loop index funs code_accu = function
    | [] -> index, funs, code_accu
    | blk :: blks ->

      let block_id = Mdl_translation.get_block_ident table
      and scsm_id = Mdl_translation.get_diagram_ident table in

      let index, funs, code_block =

        let explicit_block_type = explicit_block_type blk in

        match blk with
        | {
            explicit_block_name = explicit_block_name;
            explicit_block_sid = _explicit_sid_opt;
            explicit_block_body = ({
              explicit_block_all_bindings = explicit_block_all_bindings;
              explicit_block_array = _explicit_array_opt;
              explicit_block_object = _explicit_object_opt;
              explicit_block_mask = explicit_mask_opt;
              (* FIXME: we should compile subsystems with embedded matlab
                 explicit_block_sf_block_type. *)
              explicit_block_sf_block_type = _explicit_sf_block_type_opt;
              explicit_block_instance_data = explicit_instance_data_opt;
            } as _explicit_block_reg_body);
            explicit_block_known = known;
            explicit_block_desc = desc;
          } ->

          let link_names =
            try Hashtbl.find links_by_dst explicit_block_name with
            | Not_found -> Hashtbl.create 3 in

          let genr_info =
            make_genr_info
              explicit_block_type explicit_block_name known
              explicit_block_all_bindings explicit_instance_data_opt
              system_io_descriptor
              link_names in

          let code_graphics =
            let ebg = explicit_block_all_bindings.explicit_block_graphical in
            compile_graphics height ebg block_id genr_info in

          let add_mask_block_code code =
            match explicit_mask_opt with
            | None -> code
            | Some
              ({
                explicit_mask_type = _;
                explicit_mask_description = _;
                explicit_mask_initialisation = explicit_mask_initialisation;
                explicit_mask_display = _;
                explicit_mask_parameters = _;
                explicit_mask_bindings = _;
              } as explicit_mask) ->
              let context =
                let context_opt =
                  context_opt_of_explicit_mask_initialisation_opt
                    block_id explicit_mask_initialisation in
                match context_opt with
                | None ->
                  Context_code
                    (Matlab_script (Ast_types.mdl_text_of_string ""))
                | Some context -> context in
              let mask_block =
                Mask_block (block_id, context, explicit_mask) in
              Concat_list.cons_list (mask_block :: code_graphics) code in

          match desc with
          | Explicit_block_regular _
          | Explicit_block_reference _
          | Explicit_block_s_function _ ->
            let translator = get_translator explicit_block_type in
            let block_code =
              translate_block
                table translator explicit_block_type block_id genr_info in

            index,
            funs,
            Concat_list.concat
              block_code
              (Concat_list.cons_list
                 code_graphics
                 (add_mask_block_code
                    (list [
                       add_block table scsm_id block_id explicit_block_name;
                     ])
                 )
              )

          | Explicit_block_sub_system {
              explicit_block_sub_system_activation = explicit_activation;
              explicit_block_sub_system_body = sub_system;
              explicit_block_sub_system_treat_as_atomic_unit = atomic_option;
              explicit_block_sub_system_sample_time = _;
            } ->

            let add_treat_as_atomic_unit_instruction code =
              let atomic_option =
                match explicit_activation with
                | Mdl_implementation_explicit_ast.
                  Sub_system_plain -> atomic_option
                | Mdl_implementation_explicit_ast.
                  Sub_system_activated activation ->
                  match activation with
                  | Sub_system_plain_atomic_unit _
                  | Sub_system_enabled _
                  | Sub_system_if_action _
                  | Sub_system_triggered _
                  | Sub_system_triggered_and_enabled (_, _) -> Some true in
              match atomic_option with
              | None -> code
              | Some b ->
                Set_super_block_atomic_property (block_id, b) :: code in

            let index, sub_system_function_ident, funs =
              compile_sub_system index funs sub_system in

            let super_block_diagram_ident = Genr_utils.scsm0 in

            let add_super_block_mask code =
              match explicit_mask_opt with
              | None -> code
              | Some explicit_mask ->
                mask_super_block block_id super_block_diagram_ident
                  explicit_mask code in

            let code_system =
              Concat_list.cons
                (Instantiate_super_block block_id)
                (Concat_list.cons_list
                   (add_treat_as_atomic_unit_instruction code_graphics)
                   (list (
                      Set_function_call (
                        scsm0, sub_system_function_ident, []) ::
                      add_super_block_mask [
                        Fill_super_block (
                          block_id, super_block_diagram_ident);
                        add_block table scsm_id block_id explicit_block_name;
                      ]
                    )
                   )
                ) in

            index,
            funs,
            code_system in

      loop index funs (concat code_accu code_block) blks in

  loop index system_funs (list []) blocks

and context_opt_of_explicit_mask_initialisation_opt block_id = function
  | None ->
    Say.debug (Printf.sprintf
      "No initialisation defined in mask for block %s"
      (string_of_host_block_ident block_id));
    None
  | Some (Deep_source_code initialisation) ->
    Say.debug (Printf.sprintf
      "Initialisation in mask %S added to context for block %s"
      (string_of_mdl_text initialisation)
      (string_of_host_block_ident block_id));
    Some (Context_code (Matlab_script initialisation))

and mask_super_block block_id super_scsm_id mask code =

  let add_mask_type explicit_mask_type code =
    match explicit_mask_type with
    | None -> code
    | Some _mask_type ->
      code in

  let add_mask_description explicit_mask_description_opt code =
    match explicit_mask_description_opt with
    | None -> code
    | Some _explicit_mask_description ->
      code in

  let add_mask_initialisation explicit_mask_initialisation_opt code =
    let context_opt =
      context_opt_of_explicit_mask_initialisation_opt
        block_id explicit_mask_initialisation_opt in
    match context_opt with
    | None -> code
    | Some context ->
      Set_diagram_context (super_scsm_id, context) :: code in

  let add_mask_display explicit_mask_display_opt code =
    match explicit_mask_display_opt with
    | None -> code
    | Some (Deep_source_code mask_display) ->
      Set_block_icon (block_id, Matlab_script mask_display) :: code in

  add_mask_type mask.Mdl_basics_explicit_ast.explicit_mask_type
    (add_mask_description mask.explicit_mask_description
      (add_mask_initialisation mask.explicit_mask_initialisation
        (add_mask_display mask.explicit_mask_display
          (Set_block_mask (block_id, mask) ::
           code)
        )
      )
    )

and compile_sub_system index funs = function
  | {
      explicit_system_name = system_name;
      explicit_system_position = system_position;
      explicit_system_bindings = _explicit_binding_table;
      explicit_system_blocks = explicit_blocks;
      explicit_system_lines = explicit_links;
      explicit_system_annotations = explicit_annotations;
      explicit_system_connectors = explicit_system_connectors;
      explicit_system_zoom_factor = explicit_zoom_factor;
    } ->

    begin
      match explicit_system_connectors with
      | [] -> ()
      | _ :: _ ->
        Say.warning
          "when translating model to code: dropping system connectors"
    end;

    let index, sub_system_function_ident =
      make_sub_system_function_ident index in

    let translation_table = Mdl_translation.new_translation_table () in

    let scsm_id =
      Mdl_translation.get_diagram_ident translation_table
    and block_id =
      Mdl_translation.get_block_ident translation_table in

    let links_by_dst = group_links_by_dst explicit_links in

    let index, funs, code_blocks =
      compile_blocks
        translation_table system_position index funs
        explicit_blocks links_by_dst in

    let code_links =
      compile_links translation_table scsm_id explicit_links in

    let code_annotations =
      compile_annotations
        translation_table scsm_id block_id system_position
        explicit_annotations in

    let get_modeler_zoom_factor = function
    | None -> 1.0
    | Some z ->
      let f = Unatural.to_float z in
      f /. 100.0 in

    let sci_zoom_factor = get_modeler_zoom_factor explicit_zoom_factor in
    let sci_bg_color = modeler_color Deep_white in
    let sci_position = modeler_position system_position in

    let code_graphics = [
      Instantiate_diagram (scsm_id);
      Set_diagram_bg_color (scsm_id, sci_bg_color);
      Set_diagram_3d (scsm_id, true);
      Set_diagram_zoom (scsm_id, sci_zoom_factor);
      Set_diagram_position (scsm_id, sci_position);
      Set_diagram_name (scsm_id, system_name);
    ] in

    let function_code =
      cons_list code_graphics
        (concat code_blocks
           (concat code_links
              code_annotations)) in

    index,
    sub_system_function_ident,
    Define_function (
      scsm_id, sub_system_function_ident, [], function_code) ::
    funs
;;

let compile_explicit_system _model_name explicit_system =
  let _, sub_system_function_ident, system_funs =
    compile_sub_system 0 [] explicit_system in
  sub_system_function_ident, system_funs
;;

(* FIXME: Nasty hack to use the pre 4.02 definition of function
   Pervasives.input using Bytes.t instead of string:
   we need code compatible with version 4.01-4.03. *)
external compatible_input_primitive :
  in_channel -> string -> int -> int -> int =
  "caml_ml_input"
;;

(** Returns a string out of an input channel. *)
let string_of_ic ic =
  let l = in_channel_length ic in
  (* Must use a string here to be compatible with version >= 4.01. *)
  let buf = String.make 1024 ' ' in
  let res = Buffer.create l in
  try
    while true do
      (* Must use a string here to be compatible with 4.02 version. *)
      let n = compatible_input_primitive ic buf 0 1024 in
      if n = 0 then raise Exit else
      Buffer.add_substring res buf 0 n
    done;
    Buffer.contents res
  with Exit -> Buffer.contents res
;;

let file_contents fname =
  if not (Path.exists fname) then
    Say.fatal_error (Printf.sprintf
      "unable to find companion file %S" fname) else
  let efname = Path.find fname in
  Path.with_explicit_in_file efname string_of_ic
;;

let get_matlab_script = function
(* FIXME: Matlab_script should take a script argument
   (may be a renaming of constructor Mdl_code to Mdl_script)
   or type Mdl_code could be divided into Mdl_expression and Mdl_script ? *)
  | Deep_code
     (Mdl_code mdl_text) ->
    Matlab_script mdl_text
  | Deep_string s -> Matlab_script (Ast_types.mdl_text_of_string s)
  | Deep_wstring _ ->
    Say.fatal_error (Printf.sprintf "found a%s, %s" " text"
      "unable to find a matlab script without a code value")
  | Deep_expression _ ->
    Say.fatal_error (Printf.sprintf "found a%s, %s" "n expression"
      "unable to find a matlab script without a code value")
  | Deep_bool b ->
    Say.fatal_error (Printf.sprintf "found %s,\n %s"
      (Printf.sprintf "bool ``%B''" b)
      "unable to find a matlab script without a code value")
  | Deep_float s ->
    Say.fatal_error (Printf.sprintf "found %s,\n %s"
      (Printf.sprintf "float ``%s''" s)
      "unable to find a matlab script without a code value")
  | Deep_int s ->
    Say.fatal_error (Printf.sprintf "found %s,\n %s"
      (Printf.sprintf "int ``%s''" s)
      "unable to find a matlab script without a code value")
  | Deep_ident id ->
    Say.fatal_error (Printf.sprintf "found %s,\n %s"
      (Printf.sprintf "ident ``%s''" (Ast_types.string_of_ident id))
      "unable to find a matlab script without a code value")
  | Deep_vector _ ->
    Say.fatal_error (Printf.sprintf "found %s,\n %s"
      (Printf.sprintf "%s" "a vector")
      "unable to find a matlab script without a code value")
  | Deep_matrix _ ->
    Say.fatal_error (Printf.sprintf "found %s, %s"
      (Printf.sprintf "%s" "a matrix")
      "unable to find a matlab script without a code value")
  | Deep_bindings _ ->
    Say.fatal_error (Printf.sprintf "found %s, %s"
      (Printf.sprintf "%s" "some bindings")
      "unable to find a matlab script without a code value")
;;

let get_matlab_scripts fnames =

  let catenate_scripts =
    let rec loop accu = function
      | [] ->
       Matlab_script
         (mdl_text_of_string
            (String.concat "\n" accu))
      | Matlab_script text :: scripts ->
        loop (string_of_mdl_text text :: accu) scripts in
    loop [] in

(* FIXME: should call a new function
   get_matlab_script_in_file that should take the appropriate argument,
   and return the proper value:
   get_matlab_script_in_file : explicit_file_name -> deep_value
   or better
   get_matlab_script_in_file : explicit_file_name -> deep_script
   with deep_script the argument of the Deep_code summand of deep_value.

   However, the compiler should not fail because the file does not exist
   during compilation ? *)
  let get_fname_script fname =
    get_matlab_script
      (Deep_code (mdl_code_of_string (file_contents fname))) in

  let scripts =
    List.map get_fname_script fnames in

  catenate_scripts scripts
;;

let add_companion_files model_ident code =
  match Configuration.get_companion_source_files () with
  | [] -> code
  | fnames ->
    let instrs = get_matlab_scripts fnames in
    Set_model_prelude (model_ident, instrs) :: code
;;

let init_fcn_ident = ident_from_string "InitFcn";;
let close_fcn_ident = ident_from_string "CloseFcn";;
let pre_load_fcn_ident = ident_from_string "PreLoadFcn";;
let post_load_fcn_ident = ident_from_string "PostLoadFcn";;
let pre_save_fcn_ident = ident_from_string "PreSaveFcn";;
let post_save_fcn_ident = ident_from_string "PostSaveFcn";;
let start_fcn_ident = ident_from_string "StartFcn";;
let stop_fcn_ident = ident_from_string "StopFcn";;

let get_instrs ident bindings code seq =
  if not (Hashtbl.mem bindings ident) then code else
  let instrs = Hashtbl.find bindings ident in
  seq (get_matlab_script instrs) :: code
;;

let add_model_prelude = add_companion_files

and add_init_fcn model_ident bindings code =
  get_instrs init_fcn_ident bindings code
    (fun instrs -> Set_model_init_fcn (model_ident, instrs))

and add_close_fcn model_ident bindings code =
  get_instrs close_fcn_ident bindings code
    (fun instrs -> Set_model_close_fcn (model_ident, instrs))

and add_preload_fcn model_ident bindings code =
  get_instrs pre_load_fcn_ident bindings code
    (fun instrs -> Set_model_preload_fcn (model_ident, instrs))

and add_postload_fcn model_ident bindings code =
  get_instrs post_load_fcn_ident bindings code
    (fun instrs -> Set_model_postload_fcn (model_ident, instrs))

and add_presave_fcn model_ident bindings code =
  get_instrs pre_save_fcn_ident bindings code
    (fun instrs -> Set_model_presave_fcn (model_ident, instrs))

and add_postsave_fcn model_ident bindings code =
  get_instrs post_save_fcn_ident bindings code
    (fun instrs -> Set_model_postsave_fcn (model_ident, instrs))

and add_start_fcn model_ident bindings code =
  get_instrs start_fcn_ident bindings code
    (fun instrs -> Set_model_start_fcn (model_ident, instrs))

and add_stop_fcn model_ident bindings code =
  get_instrs stop_fcn_ident bindings code
    (fun instrs -> Set_model_stop_fcn (model_ident, instrs))
;;

let add_workspace_instruction model_ident explicit_model_workspace code =
  Set_model_workspace (model_ident, explicit_model_workspace)
  :: code
;;

let solver_cc_longident =
  Configuration_lident (ident_from_string "SolverCC")
;;

let stop_time_ident = ident_from_string "StopTime";;
let start_time_ident = ident_from_string "StartTime";;
let rel_tol_ident = ident_from_string "RelTol";;
let abs_tol_ident = ident_from_string "AbsTol";;
let solver_ident = ident_from_string "Solver";;
let max_step_ident = ident_from_string "MaxStep";;

let add_solver_parameters model_ident active_configuration code =
  match active_configuration with
  | None -> code
  | Some active_configuration ->
    let cvcs =
      active_configuration.
      Mdl_deep_ast.deep_configuration_vendor_components in
    let solver_cfg = List.assoc solver_cc_longident cvcs in

    let find_value_in_solver_cfg ident =
      let dv = find_deep_value ident solver_cfg in
      match dv with
      | Deep_bool _ -> dv
      | Deep_int _s -> dv
      | Deep_float _s -> dv
      | Deep_string s ->
        Deep_expression (Ast_types.mdl_text_of_string s)
      | Deep_ident id -> Deep_expression (Ast_types.mdl_text_of_ident id)
      | Deep_expression _ | Deep_code _ -> dv
      |(Deep_wstring _ | Deep_matrix _ | Deep_vector _ | Deep_bindings _) ->
        Say.fatal_error (Printf.sprintf
          "wrong type for ident %s" (Ast_types.string_of_ident ident)) in

    let final_simulation_time =
      find_value_in_solver_cfg stop_time_ident in

    let initial_simulation_time =
      find_value_in_solver_cfg start_time_ident in

    let modeler_simulation_parameters = {
      ssp_abs_tol =
        find_value_in_solver_cfg abs_tol_ident;
      ssp_rel_tol =
        find_value_in_solver_cfg rel_tol_ident;
      ssp_tim_tol = Deep_expression (mdl_text_of_string "auto");
      ssp_delta_t = final_simulation_time;
      ssp_rt_scale = Deep_int "0";
      ssp_solver =
       find_value_in_solver_cfg solver_ident;
      ssp_max_step =
       find_value_in_solver_cfg max_step_ident;
    } in
    Set_model_initial_simulation_time
      (model_ident, initial_simulation_time) ::
    Set_model_final_simulation_time
      (model_ident, final_simulation_time) ::
    Set_model_solver_parameters
      (model_ident, modeler_simulation_parameters) ::
    code
;;

let rec compile_model = function
  | {
      explicit_model_body = explicit_model_body;
      explicit_model_matdatas_option = _matdata_records;
      explicit_stateflow_option = _stateflow_section_items;
    } ->
    compile_model_body explicit_model_body

and compile_model_body = function
  | {
      explicit_model_ident = model_ident;
      explicit_model_version = _model_version;
      explicit_model_explicit_file_name = _efname;
      explicit_model_configuration = {
        explicit_graphical_parameters = _;
        explicit_configuration_bindings = bindings;
        explicit_vendor_bindings = _;
        explicit_active_configuration = active_configuration;
      };
      explicit_model_workspace = explicit_model_workspace;
      explicit_model_system = explicit_system;
    } ->
    let sub_system_function_ident, system_funs =
      compile_explicit_system model_ident explicit_system in

    let host_model_ident = Hmi (string_of_ident model_ident) in

    let context_code =
      List.fold_left
        (fun instructions g -> g instructions)
        []
        [
          add_close_fcn host_model_ident bindings;
          add_postsave_fcn host_model_ident bindings;
          add_presave_fcn host_model_ident bindings;
          add_stop_fcn host_model_ident bindings;
          add_start_fcn host_model_ident bindings;

          add_solver_parameters host_model_ident active_configuration;
          add_init_fcn host_model_ident bindings;

          add_workspace_instruction
            host_model_ident explicit_model_workspace;

          add_postload_fcn host_model_ident bindings;
          add_preload_fcn host_model_ident bindings;

          add_model_prelude host_model_ident;
        ] in

    let postlude_code =
      Concat_list.cons
        (Set_function_call (
           Hdi (string_of_ident model_ident),
           sub_system_function_ident, []
           )
        )
        (Concat_list.cons_list
           context_code
           (Concat_list.list [
              Evaluate_model (
                host_model_ident,
                Hdi (
                  string_of_ident (
                    Configuration.get_target_model_ident ())
                )
              );
            ]
           )
        ) in

    Concat_list.cons_list
      (List.rev system_funs)
      postlude_code
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
