# 17 "src/mdl_file_format/parsing/mdl_lexer.mll"
 

(* Prelude part: this is pure Caml. *)

open Lexing;;
open Mdl_parser;;

(** {6 Lexing errors} *)

type error =
  | Illegal_character of char
  | Illegal_escape of string
  | Unterminated_string
(** The various errors when lexing. *)
;;

(** {7 Explaining lexing errors} *)

let report_error ppf = function
  | Illegal_character c ->
    Format.fprintf ppf "Illegal character (%C)" c
  | Illegal_escape s ->
    Format.fprintf ppf "Illegal escape (%S)" s
  | Unterminated_string ->
    Format.fprintf ppf "Unterminated string"
;;

(** {7 Raising lexing errors} *)
exception Error of error * Lexing.position * Lexing.position;;

let lexical_error error start_pos end_pos =
  raise (Error (error, start_pos, end_pos))
;;

(** {6 Lexer table definitions} *)

(** {7 The keyword table} *)

let keyword_table, x_keyword_table =
  let keyword_bindings = [
    (*
      "on", BOOL true;
      "off", BOOL false;
    *)
    "$BackupClass", DOLLAR_BACKUPCLASS;
    "$ClassName", DOLLAR_CLASSNAME;
    "$ObjectID", DOLLAR_OBJECTID;
    "$PropName", DOLLAR_PROPNAME;
    "Annotation", ANNOTATION;
    "AnnotationDefaults", ANNOTATIONDEFAULTS;
    "Array", ARRAY;
    "AutoSize", AUTOSIZE;
    "BackgroundColor", BACKGROUNDCOLOR;
    "Block", BLOCK;
    "BlockDefaults", BLOCKDEFAULTS;
    "BlockInterfaceParameters", BLOCKINTERFACEPARAMETERS;
    "BlockInterfaces", BLOCKINTERFACES;
    "BlockMirror", BLOCKMIRROR;
    "BlockParameterDefaults", BLOCKPARAMETERDEFAULTS;
    "BlockRotation", BLOCKROTATION;
    "BlockType", BLOCKTYPE;
    "Branch", BRANCH;
    "ClickFcn", CLICKFCN;
    "ConfigSet", CONFIGSET;
    "ConnectType", CONNECTTYPE;
    "CustomProperty", CUSTOMPROPERTY;
    "Data", DATA;
    "DataRecord", DATARECORD;
    "DataType", DATATYPE;
    "Dimension", DIMENSION;
    "DropShadow", DROPSHADOW;
    "DstBlock", DSTBLOCK;
    "DstPort", DSTPORT;
    "Enableport", ENABLEPORT;
    "EnumStrPairs", ENUMSTRPAIRS;
    "FontAngle", FONTANGLE;
    "FontName", FONTNAME;
    "FontSize", FONTSIZE;
    "FontWeight", FONTWEIGHT;
    "ForegroundColor", FOREGROUNDCOLOR;
    "GraphicalInterface", GRAPHICALINTERFACE;
    "HideAutomaticName", HIDEAUTOMATICNAME;
    "HorizontalAlignment", HORIZONTALALIGNMENT;
    "Inport", INPORT;
    "InterfaceName", INTERFACENAME;
    "Interpreter", INTERPRETER;
    "Labels", LABELS;
    "Library", LIBRARY;
    "Line", LINE;
    "LineDefaults", LINEDEFAULTS;
    "LineType", LINETYPE;
    "List", LIST;
    "ListType", LISTTYPE;
    "Location", LOCATION;
    "MATStruct", MATSTRUCT;
    "MaskCallbackString", MASKCALLBACKSTRING;
    "MaskCapabilities", MASKCAPABILITIES;
    "MaskDefaults", MASKDEFAULTS;
    "MaskDescription", MASKDESCRIPTION;
    "MaskDisplay", MASKDISPLAY;
    "MaskDisplayString", MASKDISPLAYSTRING;
    "MaskEnableString", MASKENABLESTRING;
    "MaskHelp", MASKHELP;
    "MaskIconFrame", MASKICONFRAME;
    "MaskIconOpaque", MASKICONOPAQUE;
    "MaskIconRotate", MASKICONROTATE;
    "MaskIconUnits", MASKICONUNITS;
    "MaskInitialization", MASKINITIALIZATION;
    "MaskParameterDefaults", MASKPARAMETERDEFAULTS;
    "MaskPortRotate", MASKPORTROTATE;
    "MaskPromptString", MASKPROMPTSTRING;
    "MaskRunInitForIconRedraw", MASKRUNINITFORICONREDRAW;
    "MaskSelfModifiable", MASKSELFMODIFIABLE;
    "MaskStyleString", MASKSTYLESTRING;
    "MaskTabNameString", MASKTABNAMESTRING;
    "MaskToolTipString", MASKTOOLTIPSTRING;
    "MaskTunableValueString", MASKTUNABLEVALUESTRING;
    "MaskType", MASKTYPE;
    "MaskValueString", MASKVALUESTRING;
    "MaskVarAliasString", MASKVARALIASSTRING;
    "MaskVariables", MASKVARIABLES;
    "MaskVisibilityString", MASKVISIBILITYSTRING;
    "MatData", MATDATA;
    "Model", MODEL;
    "Name", NAME;
    "NamePlacement", NAMEPLACEMENT;
    "NumRecords", NUMRECORDS;
    "NumRootInports", NUMROOTINPORTS;
    "NumRootOutports", NUMROOTOUTPORTS;
    "NumRootTrigports", NUMROOTTRIGPORTS;
    "Object", OBJECT;
    "Orientation", ORIENTATION;
    "Outport", OUTPORT;
    "Points", POINTS;
    "Port", PORT;
    "PortNumber", PORTNUMBER;
    "Ports", PORTS;
    "Position", POSITION;
    "PropName", PROPNAME;
    "RTW", VENDOR Mdl_ast.Rtw;
    "SFBlockType", SFBLOCKTYPE;
    "SID", SID;
    "SavedCharacterEncoding", SAVEDCHARACTERENCODING;
    "Scicos", VENDOR Mdl_ast.Scicos;
    "ShowName", SHOWNAME;
    "Simulink", VENDOR Mdl_ast.Simulink;
    "SourceBlock", SOURCEBLOCK;
    "SourceType", SOURCETYPE;
    "SrcBlock", SRCBLOCK;
    "SrcPort", SRCPORT;
    "Stateflow", STATEFLOW;
    "String", STR;
    "System", SYSTEM;
    "Tag", TAG;
    "TeXMode", TEXMODE;
    "Text", TEXT;
    "TreatAsAtomicUnit", TREATASATOMICUNIT;
    "Trigport", TRIGPORT;
    "Type", TYPE;
    "UseDisplayTextAsClickCallback", USEDISPLAYTEXTASCLICKCALLBACK;
    "Value", VALUE;
    "Version", VERSION;
    "VerticalAlignment", VERTICALALIGNMENT;
    "WSDataSource", WSDATASOURCE;
    "WSMATFileName", WSMATFILENAME;
    "WSMATLABCode", WSMATLABCODE;
    "WSMCode", WSMCODE;
    "WSMdlFileData", WSMDLFILEDATA;
    "WSSourceFileName", WSSOURCEFILENAME;
    "WordWrap", WORDWRAP;
    "ZOrder", ZORDER;
    "array", LARRAY;
    "chart", LCHART;
    "chartFileNumber", LCHARTFILENUMBER;
    "data", LDATA;
    "dst", LDST;
    "eml", LEML;
    "enable", LENABLE;
    "event", LEVENT;
    "executionOrder", LEXECUTIONORDER;
    "firstData", LFIRSTDATA;
    "firstEvent", LFIRSTEVENT;
    "firstJunction", LFIRSTJUNCTION;
    "firstIndex", LFIRSTINDEX;
    "firstTransition", LFIRSTTRANSITION;
    "firstTarget", LFIRSTTARGET;
    "id", LID;
    "ifaction", LIFACTION;
    "instance", LINSTANCE;
    "intersection", LINTERSECTION;
    "junction", LJUNCTION;
    "labelString", LLABELSTRING;
    "linkNode", LLINKNODE;
    "machine", LMACHINE;
    "name", LNAME;
    "plantModelingInfo", LPLANTMODELINGINFO;
    "props", LPROPS;
    "scope", LSCOPE;
    "script", LSCRIPT;
    "size", LSIZE;
    "src", LSRC;
    "state", LSTATE;
    "target", LTARGET;
    "transition", LTRANSITION;
    "treeNode", LTREENODE;
    "trigger", LTRIGGER;
    "type", LTYPE;
    (* "MaskHideContents", MASKHIDECONTENTS; *)
  ] in
  Hash_table.t_xt_of_bindings keyword_bindings
;;

(** {7 Returning tokens from lexical entities} *)

let token_of_ident s =
  assert (String.length s > 0);
  try Hashtbl.find keyword_table s with
  | Not_found ->
    try BLOCKTYPENAME (Mdl_block_types_tables.block_type_of_string s) with
    | Not_found -> IDENT s
;;

let token_of_number fractional_part exponent_part lexem =
  let lf = String.length fractional_part
  and le = String.length exponent_part in
  match lf, le with
  | 0, 0 -> INT lexem
  (* We could also admit:
     | 0, _ -> INT lexem
     meaning that we acknowledge a decimal exponent for integer numbers,
     provided we verify that the resulting number is indeed an integer!
     We must forbid for instance 20e-2 !
  *)
  | _, _ -> FLOAT lexem
;;

let token_of_string s = STRING s;;

(** {6 Keeping the internal buffer locations up to date} *)

let update_loc lexbuf fname line absolute chars =
  let pos = lexbuf.lex_curr_p in
  lexbuf.lex_curr_p <- {
    pos with
    pos_fname = fname;
    pos_lnum = if absolute then line else pos.pos_lnum + line;
    pos_bol = pos.pos_cnum - chars;
  }
;;

(** Add one to the current line counter of the file being lexed. *)
let incr_line_num lexbuf =
  update_loc lexbuf (Configuration.get_source_file ()) 1 false 0
;;

(** {6 Lexing the string tokens} *)
let string_buffer = Buffer.create 256;;

let string_start_pos = ref None;;

let store_string_char c =
  Buffer.add_char string_buffer c
;;

let get_stored_string () =
  let s = Buffer.contents string_buffer in
  Buffer.reset string_buffer;
  s
;;

let char_for_character = function
  | '\\' -> '\\'
  | '\"' -> '\"'
  | 'n' -> '\n'
  | 't' -> '\t'
  | 'b' -> '\b'
  | 'r' -> '\r'
  | c ->
    Say.fatal_error
      (Printf.sprintf "Unknown escaped character %C" c)
;;


# 286 "src/mdl_file_format/parsing/mdl_lexer.ml"
let __ocaml_lex_tables = {
  Lexing.lex_base = 
   "\000\000\231\255\232\255\001\000\235\255\237\255\238\255\239\255\
    \240\255\241\255\242\255\243\255\244\255\245\255\062\000\163\000\
    \248\255\241\000\137\000\063\001\141\001\149\000\219\001\040\002\
    \006\000\006\000\255\255\062\002\094\002\253\255\104\002\126\002\
    \136\002\158\002\236\002\057\003\014\000\043\000\052\000\251\255\
    \053\000\094\003\172\003\250\003\069\004\106\004\184\004\006\005\
    \007\000\234\255\222\001\003\000\250\255\251\255\063\005\010\000\
    \089\002\030\000\015\000\255\255\252\255\253\255";
  Lexing.lex_backtrk = 
   "\255\255\255\255\255\255\022\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\024\000\008\000\
    \255\255\008\000\019\000\008\000\008\000\024\000\008\000\005\000\
    \001\000\000\000\255\255\255\255\005\000\255\255\005\000\005\000\
    \255\255\008\000\003\000\005\000\255\255\255\255\255\255\255\255\
    \255\255\008\000\004\000\008\000\005\000\006\000\008\000\009\000\
    \021\000\255\255\255\255\255\255\255\255\255\255\005\000\001\000\
    \255\255\255\255\255\255\255\255\255\255\255\255";
  Lexing.lex_default = 
   "\001\000\000\000\000\000\050\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\050\000\052\000\000\000\000\000\060\000\255\255\
    \255\255\255\255\255\255\000\000\000\000\000\000";
  Lexing.lex_trans = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\024\000\026\000\049\000\024\000\025\000\048\000\024\000\
    \026\000\049\000\024\000\058\000\057\000\000\000\000\000\056\000\
    \058\000\057\000\000\000\000\000\056\000\000\000\000\000\000\000\
    \024\000\000\000\016\000\003\000\014\000\055\000\024\000\057\000\
    \013\000\012\000\058\000\021\000\007\000\021\000\018\000\058\000\
    \023\000\023\000\023\000\023\000\023\000\023\000\023\000\023\000\
    \023\000\023\000\006\000\005\000\000\000\000\000\057\000\000\000\
    \059\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\019\000\015\000\015\000\015\000\015\000\022\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\009\000\000\000\008\000\000\000\054\000\
    \000\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\020\000\015\000\015\000\015\000\015\000\015\000\017\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\011\000\040\000\010\000\004\000\047\000\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\038\000\039\000\039\000\000\000\000\000\000\000\047\000\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\044\000\044\000\044\000\044\000\044\000\044\000\044\000\
    \044\000\044\000\044\000\000\000\000\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\000\000\
    \015\000\000\000\000\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\000\000\036\000\000\000\
    \000\000\000\000\000\000\000\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\037\000\000\000\
    \002\000\255\255\015\000\053\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\000\000\
    \000\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\000\000\000\000\000\000\000\000\
    \015\000\000\000\015\000\015\000\015\000\015\000\015\000\046\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\045\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\000\000\000\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\000\000\000\000\000\000\000\000\015\000\000\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\043\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\000\000\000\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \049\000\000\000\000\000\048\000\015\000\000\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\041\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\000\000\000\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\000\000\000\000\
    \000\000\000\000\015\000\000\000\033\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\028\000\000\000\
    \023\000\023\000\023\000\023\000\023\000\023\000\023\000\023\000\
    \023\000\023\000\057\000\057\000\000\000\000\000\000\000\000\000\
    \000\000\032\000\000\000\032\000\027\000\027\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \000\000\057\000\000\000\059\000\000\000\029\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\023\000\
    \000\000\000\000\000\000\000\000\027\000\027\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\027\000\027\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\027\000\027\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\027\000\027\000\000\000\000\000\000\000\030\000\
    \000\000\000\000\000\000\015\000\027\000\027\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \000\000\000\000\000\000\000\000\000\000\031\000\255\255\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\034\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\000\000\000\000\000\000\000\000\015\000\000\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\000\000\000\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\000\000\
    \000\000\000\000\000\000\015\000\000\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\028\000\
    \000\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\027\000\027\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\015\000\000\000\000\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \035\000\000\000\000\000\000\000\000\000\027\000\027\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\000\000\000\000\000\000\000\000\015\000\000\000\015\000\
    \015\000\015\000\015\000\015\000\042\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\000\000\000\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\000\000\
    \000\000\000\000\000\000\015\000\000\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \000\000\000\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\000\000\000\000\000\000\
    \000\000\015\000\000\000\015\000\015\000\015\000\015\000\015\000\
    \042\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\044\000\044\000\044\000\
    \044\000\044\000\044\000\044\000\044\000\044\000\044\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\027\000\027\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\015\000\
    \000\000\000\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\044\000\000\000\000\000\000\000\
    \000\000\027\000\027\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\000\000\000\000\000\000\
    \000\000\015\000\000\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\000\000\000\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\000\000\000\000\000\000\000\000\015\000\
    \000\000\015\000\015\000\015\000\015\000\015\000\045\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\047\000\000\000\000\000\047\000\047\000\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\047\000\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\061\000\000\000\000\000\000\000\047\000\000\000\047\000\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\061\000\000\000\000\000\000\000\000\000\
    \000\000\061\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\061\000\000\000\000\000\
    \000\000\061\000\000\000\061\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\255\255\
    ";
  Lexing.lex_check = 
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\003\000\000\000\000\000\003\000\024\000\
    \025\000\048\000\024\000\055\000\055\000\255\255\255\255\055\000\
    \058\000\058\000\255\255\255\255\058\000\255\255\255\255\255\255\
    \000\000\255\255\000\000\000\000\000\000\051\000\024\000\057\000\
    \000\000\000\000\055\000\000\000\000\000\000\000\000\000\058\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\255\255\255\255\057\000\255\255\
    \057\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\255\255\000\000\255\255\051\000\
    \255\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\036\000\000\000\000\000\014\000\
    \014\000\014\000\014\000\014\000\014\000\014\000\014\000\014\000\
    \014\000\014\000\014\000\014\000\014\000\014\000\014\000\014\000\
    \014\000\014\000\014\000\014\000\014\000\014\000\014\000\014\000\
    \014\000\037\000\038\000\040\000\255\255\255\255\255\255\014\000\
    \014\000\014\000\014\000\014\000\014\000\014\000\014\000\014\000\
    \014\000\014\000\014\000\014\000\014\000\014\000\014\000\014\000\
    \014\000\014\000\014\000\014\000\014\000\014\000\014\000\014\000\
    \014\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\255\255\255\255\021\000\021\000\021\000\
    \021\000\021\000\021\000\021\000\021\000\021\000\021\000\255\255\
    \015\000\255\255\255\255\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\255\255\021\000\255\255\
    \255\255\255\255\255\255\255\255\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\021\000\255\255\
    \000\000\003\000\015\000\051\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\017\000\255\255\
    \255\255\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\255\255\255\255\255\255\255\255\
    \017\000\255\255\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\019\000\255\255\255\255\019\000\
    \019\000\019\000\019\000\019\000\019\000\019\000\019\000\019\000\
    \019\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \019\000\019\000\019\000\019\000\019\000\019\000\019\000\019\000\
    \019\000\019\000\019\000\019\000\019\000\019\000\019\000\019\000\
    \019\000\019\000\019\000\019\000\019\000\019\000\019\000\019\000\
    \019\000\019\000\255\255\255\255\255\255\255\255\019\000\255\255\
    \019\000\019\000\019\000\019\000\019\000\019\000\019\000\019\000\
    \019\000\019\000\019\000\019\000\019\000\019\000\019\000\019\000\
    \019\000\019\000\019\000\019\000\019\000\019\000\019\000\019\000\
    \019\000\019\000\020\000\255\255\255\255\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \050\000\255\255\255\255\050\000\020\000\255\255\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \022\000\255\255\255\255\022\000\022\000\022\000\022\000\022\000\
    \022\000\022\000\022\000\022\000\022\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\022\000\022\000\022\000\022\000\
    \022\000\022\000\022\000\022\000\022\000\022\000\022\000\022\000\
    \022\000\022\000\022\000\022\000\022\000\022\000\022\000\022\000\
    \022\000\022\000\022\000\022\000\022\000\022\000\255\255\255\255\
    \255\255\255\255\022\000\255\255\022\000\022\000\022\000\022\000\
    \022\000\022\000\022\000\022\000\022\000\022\000\022\000\022\000\
    \022\000\022\000\022\000\022\000\022\000\022\000\022\000\022\000\
    \022\000\022\000\022\000\022\000\022\000\022\000\023\000\255\255\
    \023\000\023\000\023\000\023\000\023\000\023\000\023\000\023\000\
    \023\000\023\000\056\000\056\000\255\255\255\255\255\255\255\255\
    \255\255\027\000\255\255\027\000\023\000\023\000\027\000\027\000\
    \027\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \255\255\056\000\255\255\056\000\255\255\023\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\023\000\
    \255\255\255\255\255\255\255\255\023\000\023\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\028\000\028\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\030\000\030\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\028\000\028\000\255\255\255\255\255\255\030\000\
    \255\255\255\255\255\255\033\000\030\000\030\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \255\255\255\255\255\255\255\255\255\255\031\000\050\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\255\255\255\255\255\255\255\255\033\000\255\255\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\034\000\255\255\255\255\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\255\255\
    \255\255\255\255\255\255\034\000\255\255\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\035\000\
    \255\255\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\035\000\035\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\041\000\255\255\255\255\041\000\041\000\
    \041\000\041\000\041\000\041\000\041\000\041\000\041\000\041\000\
    \035\000\255\255\255\255\255\255\255\255\035\000\035\000\041\000\
    \041\000\041\000\041\000\041\000\041\000\041\000\041\000\041\000\
    \041\000\041\000\041\000\041\000\041\000\041\000\041\000\041\000\
    \041\000\041\000\041\000\041\000\041\000\041\000\041\000\041\000\
    \041\000\255\255\255\255\255\255\255\255\041\000\255\255\041\000\
    \041\000\041\000\041\000\041\000\041\000\041\000\041\000\041\000\
    \041\000\041\000\041\000\041\000\041\000\041\000\041\000\041\000\
    \041\000\041\000\041\000\041\000\041\000\041\000\041\000\041\000\
    \041\000\042\000\255\255\255\255\042\000\042\000\042\000\042\000\
    \042\000\042\000\042\000\042\000\042\000\042\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\042\000\042\000\042\000\
    \042\000\042\000\042\000\042\000\042\000\042\000\042\000\042\000\
    \042\000\042\000\042\000\042\000\042\000\042\000\042\000\042\000\
    \042\000\042\000\042\000\042\000\042\000\042\000\042\000\255\255\
    \255\255\255\255\255\255\042\000\255\255\042\000\042\000\042\000\
    \042\000\042\000\042\000\042\000\042\000\042\000\042\000\042\000\
    \042\000\042\000\042\000\042\000\042\000\042\000\042\000\042\000\
    \042\000\042\000\042\000\042\000\042\000\042\000\042\000\043\000\
    \255\255\255\255\043\000\043\000\043\000\043\000\043\000\043\000\
    \043\000\043\000\043\000\043\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\043\000\043\000\043\000\043\000\043\000\
    \043\000\043\000\043\000\043\000\043\000\043\000\043\000\043\000\
    \043\000\043\000\043\000\043\000\043\000\043\000\043\000\043\000\
    \043\000\043\000\043\000\043\000\043\000\255\255\255\255\255\255\
    \255\255\043\000\255\255\043\000\043\000\043\000\043\000\043\000\
    \043\000\043\000\043\000\043\000\043\000\043\000\043\000\043\000\
    \043\000\043\000\043\000\043\000\043\000\043\000\043\000\043\000\
    \043\000\043\000\043\000\043\000\043\000\044\000\044\000\044\000\
    \044\000\044\000\044\000\044\000\044\000\044\000\044\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\044\000\044\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\045\000\
    \255\255\255\255\045\000\045\000\045\000\045\000\045\000\045\000\
    \045\000\045\000\045\000\045\000\044\000\255\255\255\255\255\255\
    \255\255\044\000\044\000\045\000\045\000\045\000\045\000\045\000\
    \045\000\045\000\045\000\045\000\045\000\045\000\045\000\045\000\
    \045\000\045\000\045\000\045\000\045\000\045\000\045\000\045\000\
    \045\000\045\000\045\000\045\000\045\000\255\255\255\255\255\255\
    \255\255\045\000\255\255\045\000\045\000\045\000\045\000\045\000\
    \045\000\045\000\045\000\045\000\045\000\045\000\045\000\045\000\
    \045\000\045\000\045\000\045\000\045\000\045\000\045\000\045\000\
    \045\000\045\000\045\000\045\000\045\000\046\000\255\255\255\255\
    \046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\
    \046\000\046\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\046\000\046\000\046\000\046\000\046\000\046\000\046\000\
    \046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\
    \046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\
    \046\000\046\000\046\000\255\255\255\255\255\255\255\255\046\000\
    \255\255\046\000\046\000\046\000\046\000\046\000\046\000\046\000\
    \046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\
    \046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\
    \046\000\046\000\046\000\047\000\255\255\255\255\047\000\047\000\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\047\000\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\054\000\255\255\255\255\255\255\047\000\255\255\047\000\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\054\000\255\255\255\255\255\255\255\255\
    \255\255\054\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\054\000\255\255\255\255\
    \255\255\054\000\255\255\054\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\054\000\
    ";
  Lexing.lex_base_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\010\000\000\000\000\000\020\000\000\000\032\000\
    \000\000\000\000\000\000\000\000\042\000\000\000\052\000\000\000\
    \000\000\000\000\000\000\064\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\080\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_backtrk_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\013\000\
    \000\000\000\000\000\000\000\000\013\000\000\000\013\000\013\000\
    \000\000\000\000\000\000\013\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\013\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_default_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_trans_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \001\000\001\000\001\000\001\000\001\000\001\000\001\000\001\000\
    \001\000\001\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\001\000\001\000\001\000\001\000\
    \001\000\001\000\001\000\001\000\001\000\001\000\008\000\000\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\000\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\000\000\000\000\000\000\000\000\000\000\020\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\008\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\020\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\008\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000";
  Lexing.lex_check_code = 
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\021\000\021\000\021\000\021\000\
    \021\000\021\000\021\000\021\000\021\000\021\000\023\000\255\255\
    \023\000\023\000\023\000\023\000\023\000\023\000\023\000\023\000\
    \023\000\023\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\035\000\255\255\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\255\255\255\255\255\255\255\255\255\255\023\000\
    \044\000\044\000\044\000\044\000\044\000\044\000\044\000\044\000\
    \044\000\044\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\030\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\035\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\044\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255";
  Lexing.lex_code = 
   "\255\005\255\004\255\003\255\255\005\255\004\255\255\002\003\001\
    \005\000\004\255\003\255\005\255\004\255\255";
}

let rec token lexbuf =
  lexbuf.Lexing.lex_mem <- Array.make 6 (-1) ; (* L=1 [3] <- p ;  *)
  lexbuf.Lexing.lex_mem.(3) <- lexbuf.Lexing.lex_curr_pos ;
  __ocaml_lex_token_rec lexbuf 0
and __ocaml_lex_token_rec lexbuf __ocaml_lex_state =
  match Lexing.new_engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 344 "src/mdl_file_format/parsing/mdl_lexer.mll"
    ( incr_line_num lexbuf;
      token lexbuf )
# 849 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 1 ->
# 347 "src/mdl_file_format/parsing/mdl_lexer.mll"
    ( token lexbuf )
# 854 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 2 ->
# 351 "src/mdl_file_format/parsing/mdl_lexer.mll"
    ( INT (Lexing.lexeme lexbuf) )
# 859 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 3 ->
# 354 "src/mdl_file_format/parsing/mdl_lexer.mll"
    ( FLOAT (Lexing.lexeme lexbuf) )
# 864 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 4 ->
# 356 "src/mdl_file_format/parsing/mdl_lexer.mll"
    ( FLOAT (Lexing.lexeme lexbuf) )
# 869 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 5 ->
let
# 359 "src/mdl_file_format/parsing/mdl_lexer.mll"
                                                   fractional_part
# 875 "src/mdl_file_format/parsing/mdl_lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(2) lexbuf.Lexing.lex_mem.(1)
and
# 362 "src/mdl_file_format/parsing/mdl_lexer.mll"
                                 exponent_part
# 880 "src/mdl_file_format/parsing/mdl_lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_curr_pos in
# 363 "src/mdl_file_format/parsing/mdl_lexer.mll"
    ( token_of_number fractional_part exponent_part (Lexing.lexeme lexbuf) )
# 884 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 6 ->
# 367 "src/mdl_file_format/parsing/mdl_lexer.mll"
    ( BOOL (Lexing.lexeme lexbuf) )
# 889 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 7 ->
# 371 "src/mdl_file_format/parsing/mdl_lexer.mll"
    (
      string_start_pos :=
        Some (lexbuf.lex_start_p, lexbuf.lex_curr_p);
      string lexbuf;
      begin match !string_start_pos with
      | Some (start_pos, _) -> lexbuf.lex_start_p <- start_pos
      | _ -> assert false end;
      let s = get_stored_string () in
      token_of_string s
    )
# 903 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 8 ->
# 384 "src/mdl_file_format/parsing/mdl_lexer.mll"
    ( token_of_ident (Lexing.lexeme lexbuf) )
# 908 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 9 ->
# 386 "src/mdl_file_format/parsing/mdl_lexer.mll"
    ( token_of_ident (Lexing.lexeme lexbuf) )
# 913 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 10 ->
# 389 "src/mdl_file_format/parsing/mdl_lexer.mll"
        ( OPEN_PAREN )
# 918 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 11 ->
# 390 "src/mdl_file_format/parsing/mdl_lexer.mll"
        ( CLOSE_PAREN )
# 923 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 12 ->
# 391 "src/mdl_file_format/parsing/mdl_lexer.mll"
        ( OPEN_BRACE )
# 928 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 13 ->
# 392 "src/mdl_file_format/parsing/mdl_lexer.mll"
        ( CLOSE_BRACE )
# 933 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 14 ->
# 393 "src/mdl_file_format/parsing/mdl_lexer.mll"
        ( OPEN_BRACKET )
# 938 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 15 ->
# 394 "src/mdl_file_format/parsing/mdl_lexer.mll"
        ( CLOSE_BRACKET )
# 943 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 16 ->
# 395 "src/mdl_file_format/parsing/mdl_lexer.mll"
        ( COMMA )
# 948 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 17 ->
# 396 "src/mdl_file_format/parsing/mdl_lexer.mll"
        ( COLON )
# 953 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 18 ->
# 397 "src/mdl_file_format/parsing/mdl_lexer.mll"
        ( SEMICOLON )
# 958 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 19 ->
# 398 "src/mdl_file_format/parsing/mdl_lexer.mll"
        ( DOT )
# 963 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 20 ->
# 399 "src/mdl_file_format/parsing/mdl_lexer.mll"
        ( TILDA )
# 968 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 21 ->
# 402 "src/mdl_file_format/parsing/mdl_lexer.mll"
    ( incr_line_num lexbuf;
      token lexbuf )
# 974 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 22 ->
# 404 "src/mdl_file_format/parsing/mdl_lexer.mll"
        ( SHARP )
# 979 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 23 ->
# 406 "src/mdl_file_format/parsing/mdl_lexer.mll"
        ( EOF )
# 984 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 24 ->
# 408 "src/mdl_file_format/parsing/mdl_lexer.mll"
    ( lexical_error
        (Illegal_character (Lexing.lexeme_char lexbuf 0))
        lexbuf.lex_start_p lexbuf.lex_curr_p )
# 991 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_token_rec lexbuf __ocaml_lex_state

and string lexbuf =
    __ocaml_lex_string_rec lexbuf 51
and __ocaml_lex_string_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 414 "src/mdl_file_format/parsing/mdl_lexer.mll"
    ( incr_line_num lexbuf;
      string lexbuf )
# 1004 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 1 ->
# 417 "src/mdl_file_format/parsing/mdl_lexer.mll"
    ( () )
# 1009 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 2 ->
# 420 "src/mdl_file_format/parsing/mdl_lexer.mll"
    ( store_string_char (char_for_character (Lexing.lexeme_char lexbuf 1));
      string lexbuf )
# 1015 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 3 ->
# 424 "src/mdl_file_format/parsing/mdl_lexer.mll"
    ( lexical_error
        (Illegal_escape (Lexing.lexeme lexbuf))
        lexbuf.lex_start_p lexbuf.lex_curr_p )
# 1022 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 4 ->
# 429 "src/mdl_file_format/parsing/mdl_lexer.mll"
    ( match !string_start_pos with
      | Some (start_pos, end_pos) ->
        lexical_error Unterminated_string start_pos end_pos
      | _ -> assert false )
# 1030 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | 5 ->
# 434 "src/mdl_file_format/parsing/mdl_lexer.mll"
    ( store_string_char (Lexing.lexeme_char lexbuf 0);
      string lexbuf )
# 1036 "src/mdl_file_format/parsing/mdl_lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_string_rec lexbuf __ocaml_lex_state

;;

