(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_lexer.mll,v 1.1 2019-05-10 14:04:02 jpc Exp $ *)

{

(* Prelude part: this is pure Caml. *)

open Lexing;;
open Mdl_parser;;

(** {6 Lexing errors} *)

type error =
  | Illegal_character of char
  | Illegal_escape of string
  | Unterminated_string
(** The various errors when lexing. *)
;;

(** {7 Explaining lexing errors} *)

let report_error ppf = function
  | Illegal_character c ->
    Format.fprintf ppf "Illegal character (%C)" c
  | Illegal_escape s ->
    Format.fprintf ppf "Illegal escape (%S)" s
  | Unterminated_string ->
    Format.fprintf ppf "Unterminated string"
;;

(** {7 Raising lexing errors} *)
exception Error of error * Lexing.position * Lexing.position;;

let lexical_error error start_pos end_pos =
  raise (Error (error, start_pos, end_pos))
;;

(** {6 Lexer table definitions} *)

(** {7 The keyword table} *)

let keyword_table, x_keyword_table =
  let keyword_bindings = [
    (*
      "on", BOOL true;
      "off", BOOL false;
    *)
    "$BackupClass", DOLLAR_BACKUPCLASS;
    "$ClassName", DOLLAR_CLASSNAME;
    "$ObjectID", DOLLAR_OBJECTID;
    "$PropName", DOLLAR_PROPNAME;
    "Annotation", ANNOTATION;
    "AnnotationDefaults", ANNOTATIONDEFAULTS;
    "Array", ARRAY;
    "AutoSize", AUTOSIZE;
    "BackgroundColor", BACKGROUNDCOLOR;
    "Block", BLOCK;
    "BlockDefaults", BLOCKDEFAULTS;
    "BlockInterfaceParameters", BLOCKINTERFACEPARAMETERS;
    "BlockInterfaces", BLOCKINTERFACES;
    "BlockMirror", BLOCKMIRROR;
    "BlockParameterDefaults", BLOCKPARAMETERDEFAULTS;
    "BlockRotation", BLOCKROTATION;
    "BlockType", BLOCKTYPE;
    "Branch", BRANCH;
    "ClickFcn", CLICKFCN;
    "ConfigSet", CONFIGSET;
    "ConnectType", CONNECTTYPE;
    "CustomProperty", CUSTOMPROPERTY;
    "Data", DATA;
    "DataRecord", DATARECORD;
    "DataType", DATATYPE;
    "Dimension", DIMENSION;
    "DropShadow", DROPSHADOW;
    "DstBlock", DSTBLOCK;
    "DstPort", DSTPORT;
    "Enableport", ENABLEPORT;
    "EnumStrPairs", ENUMSTRPAIRS;
    "FontAngle", FONTANGLE;
    "FontName", FONTNAME;
    "FontSize", FONTSIZE;
    "FontWeight", FONTWEIGHT;
    "ForegroundColor", FOREGROUNDCOLOR;
    "GraphicalInterface", GRAPHICALINTERFACE;
    "HideAutomaticName", HIDEAUTOMATICNAME;
    "HorizontalAlignment", HORIZONTALALIGNMENT;
    "Inport", INPORT;
    "InterfaceName", INTERFACENAME;
    "Interpreter", INTERPRETER;
    "Labels", LABELS;
    "Library", LIBRARY;
    "Line", LINE;
    "LineDefaults", LINEDEFAULTS;
    "LineType", LINETYPE;
    "List", LIST;
    "ListType", LISTTYPE;
    "Location", LOCATION;
    "MATStruct", MATSTRUCT;
    "MaskCallbackString", MASKCALLBACKSTRING;
    "MaskCapabilities", MASKCAPABILITIES;
    "MaskDefaults", MASKDEFAULTS;
    "MaskDescription", MASKDESCRIPTION;
    "MaskDisplay", MASKDISPLAY;
    "MaskDisplayString", MASKDISPLAYSTRING;
    "MaskEnableString", MASKENABLESTRING;
    "MaskHelp", MASKHELP;
    "MaskIconFrame", MASKICONFRAME;
    "MaskIconOpaque", MASKICONOPAQUE;
    "MaskIconRotate", MASKICONROTATE;
    "MaskIconUnits", MASKICONUNITS;
    "MaskInitialization", MASKINITIALIZATION;
    "MaskParameterDefaults", MASKPARAMETERDEFAULTS;
    "MaskPortRotate", MASKPORTROTATE;
    "MaskPromptString", MASKPROMPTSTRING;
    "MaskRunInitForIconRedraw", MASKRUNINITFORICONREDRAW;
    "MaskSelfModifiable", MASKSELFMODIFIABLE;
    "MaskStyleString", MASKSTYLESTRING;
    "MaskTabNameString", MASKTABNAMESTRING;
    "MaskToolTipString", MASKTOOLTIPSTRING;
    "MaskTunableValueString", MASKTUNABLEVALUESTRING;
    "MaskType", MASKTYPE;
    "MaskValueString", MASKVALUESTRING;
    "MaskVarAliasString", MASKVARALIASSTRING;
    "MaskVariables", MASKVARIABLES;
    "MaskVisibilityString", MASKVISIBILITYSTRING;
    "MatData", MATDATA;
    "Model", MODEL;
    "Name", NAME;
    "NamePlacement", NAMEPLACEMENT;
    "NumRecords", NUMRECORDS;
    "NumRootInports", NUMROOTINPORTS;
    "NumRootOutports", NUMROOTOUTPORTS;
    "NumRootTrigports", NUMROOTTRIGPORTS;
    "Object", OBJECT;
    "Orientation", ORIENTATION;
    "Outport", OUTPORT;
    "Points", POINTS;
    "Port", PORT;
    "PortNumber", PORTNUMBER;
    "Ports", PORTS;
    "Position", POSITION;
    "PropName", PROPNAME;
    "RTW", VENDOR Mdl_ast.Rtw;
    "SFBlockType", SFBLOCKTYPE;
    "SID", SID;
    "SavedCharacterEncoding", SAVEDCHARACTERENCODING;
    "Scicos", VENDOR Mdl_ast.Scicos;
    "ShowName", SHOWNAME;
    "Simulink", VENDOR Mdl_ast.Simulink;
    "SourceBlock", SOURCEBLOCK;
    "SourceType", SOURCETYPE;
    "SrcBlock", SRCBLOCK;
    "SrcPort", SRCPORT;
    "Stateflow", STATEFLOW;
    "String", STR;
    "System", SYSTEM;
    "Tag", TAG;
    "TeXMode", TEXMODE;
    "Text", TEXT;
    "TreatAsAtomicUnit", TREATASATOMICUNIT;
    "Trigport", TRIGPORT;
    "Type", TYPE;
    "UseDisplayTextAsClickCallback", USEDISPLAYTEXTASCLICKCALLBACK;
    "Value", VALUE;
    "Version", VERSION;
    "VerticalAlignment", VERTICALALIGNMENT;
    "WSDataSource", WSDATASOURCE;
    "WSMATFileName", WSMATFILENAME;
    "WSMATLABCode", WSMATLABCODE;
    "WSMCode", WSMCODE;
    "WSMdlFileData", WSMDLFILEDATA;
    "WSSourceFileName", WSSOURCEFILENAME;
    "WordWrap", WORDWRAP;
    "ZOrder", ZORDER;
    "array", LARRAY;
    "chart", LCHART;
    "chartFileNumber", LCHARTFILENUMBER;
    "data", LDATA;
    "dst", LDST;
    "eml", LEML;
    "enable", LENABLE;
    "event", LEVENT;
    "executionOrder", LEXECUTIONORDER;
    "firstData", LFIRSTDATA;
    "firstEvent", LFIRSTEVENT;
    "firstJunction", LFIRSTJUNCTION;
    "firstIndex", LFIRSTINDEX;
    "firstTransition", LFIRSTTRANSITION;
    "firstTarget", LFIRSTTARGET;
    "id", LID;
    "ifaction", LIFACTION;
    "instance", LINSTANCE;
    "intersection", LINTERSECTION;
    "junction", LJUNCTION;
    "labelString", LLABELSTRING;
    "linkNode", LLINKNODE;
    "machine", LMACHINE;
    "name", LNAME;
    "plantModelingInfo", LPLANTMODELINGINFO;
    "props", LPROPS;
    "scope", LSCOPE;
    "script", LSCRIPT;
    "size", LSIZE;
    "src", LSRC;
    "state", LSTATE;
    "target", LTARGET;
    "transition", LTRANSITION;
    "treeNode", LTREENODE;
    "trigger", LTRIGGER;
    "type", LTYPE;
    (* "MaskHideContents", MASKHIDECONTENTS; *)
  ] in
  Hash_table.t_xt_of_bindings keyword_bindings
;;

(** {7 Returning tokens from lexical entities} *)

let token_of_ident s =
  assert (String.length s > 0);
  try Hashtbl.find keyword_table s with
  | Not_found ->
    try BLOCKTYPENAME (Mdl_block_types_tables.block_type_of_string s) with
    | Not_found -> IDENT s
;;

let token_of_number fractional_part exponent_part lexem =
  let lf = String.length fractional_part
  and le = String.length exponent_part in
  match lf, le with
  | 0, 0 -> INT lexem
  (* We could also admit:
     | 0, _ -> INT lexem
     meaning that we acknowledge a decimal exponent for integer numbers,
     provided we verify that the resulting number is indeed an integer!
     We must forbid for instance 20e-2 !
  *)
  | _, _ -> FLOAT lexem
;;

let token_of_string s = STRING s;;

(** {6 Keeping the internal buffer locations up to date} *)

let update_loc lexbuf fname line absolute chars =
  let pos = lexbuf.lex_curr_p in
  lexbuf.lex_curr_p <- {
    pos with
    pos_fname = fname;
    pos_lnum = if absolute then line else pos.pos_lnum + line;
    pos_bol = pos.pos_cnum - chars;
  }
;;

(** Add one to the current line counter of the file being lexed. *)
let incr_line_num lexbuf =
  update_loc lexbuf (Configuration.get_source_file ()) 1 false 0
;;

(** {6 Lexing the string tokens} *)
let string_buffer = Buffer.create 256;;

let string_start_pos = ref None;;

let store_string_char c =
  Buffer.add_char string_buffer c
;;

let get_stored_string () =
  let s = Buffer.contents string_buffer in
  Buffer.reset string_buffer;
  s
;;

let char_for_character = function
  | '\\' -> '\\'
  | '\"' -> '\"'
  | 'n' -> '\n'
  | 't' -> '\t'
  | 'b' -> '\b'
  | 'r' -> '\r'
  | c ->
    Say.fatal_error
      (Printf.sprintf "Unknown escaped character %C" c)
;;

}

(** The lexer: this is the pure ocamllex part. *)

let newline = ( '\n' | '\r' | "\r\n" )
let blank = [ ' ' '\t' '\012' ]
let whites = [ ' ' '\t' ]*

(** {6 Decimal integer numbers} *)

let decimal_digit = [ '0' - '9' ]
let unsigned_decimal_literal = decimal_digit ( decimal_digit | '_' )*

let sign = [ '+' '-' ]

let decimal_literal = sign? unsigned_decimal_literal

(** {6 Decimal float numbers} *)

let scientific_notation = [ 'e' 'E' 'd' 'D' ]

let float_scientific_suffix = (scientific_notation decimal_literal)

let float_fractional_part = ('.' unsigned_decimal_literal* )

(** {3 Identifiers} *)
let lowercase_alphabetic = [ 'a' - 'z' ]
let uppercase_alphabetic = [ 'A' - 'Z' ]
let alphabetic = lowercase_alphabetic | uppercase_alphabetic

let inside_ident =
    alphabetic
  | decimal_digit
  | '_' | '-'

let ident = alphabetic inside_ident*

let escaped_character =
  '\\' [ '\\' '\"' 'n' 't' 'b' 'r' ]
  (* ` Helping emacs }]) *)

(** {3 The main lexer. *)

rule token = parse
  | newline
    { incr_line_num lexbuf;
      token lexbuf }
  | blank +
    { token lexbuf }

  (* Numbers *)
  | unsigned_decimal_literal 'U'
    { INT (Lexing.lexeme lexbuf) }

  | "NaN"
    { FLOAT (Lexing.lexeme lexbuf) }
  | (sign ?) ("inf" | "Inf")
    { FLOAT (Lexing.lexeme lexbuf) }

  | (
        decimal_literal (float_fractional_part? as fractional_part)
      | ('.' unsigned_decimal_literal as fractional_part)
    )
    (float_scientific_suffix? as exponent_part)
    { token_of_number fractional_part exponent_part (Lexing.lexeme lexbuf) }

  (* Booleans *)
  | ("off" | "on")
    { BOOL (Lexing.lexeme lexbuf) }

  (* Strings *)
  | '\"'
    {
      string_start_pos :=
        Some (lexbuf.lex_start_p, lexbuf.lex_curr_p);
      string lexbuf;
      begin match !string_start_pos with
      | Some (start_pos, _) -> lexbuf.lex_start_p <- start_pos
      | _ -> assert false end;
      let s = get_stored_string () in
      token_of_string s
    }

  (* Identifiers *)
  | ident
    { token_of_ident (Lexing.lexeme lexbuf) }
  | '$' ident
    { token_of_ident (Lexing.lexeme lexbuf) }

  (* Usual simple tokens *)
  | '(' { OPEN_PAREN }
  | ')' { CLOSE_PAREN }
  | '{' { OPEN_BRACE }
  | '}' { CLOSE_BRACE }
  | '[' { OPEN_BRACKET }
  | ']' { CLOSE_BRACKET }
  | ',' { COMMA }
  | ':' { COLON }
  | ';' { SEMICOLON }
  | '.' { DOT }
  | '~' { TILDA }

  | '#' [^ '\n' '\r']* newline
    { incr_line_num lexbuf;
      token lexbuf }
  | '#' { SHARP }

  | eof { EOF }
  | _
    { lexical_error
        (Illegal_character (Lexing.lexeme_char lexbuf 0))
        lexbuf.lex_start_p lexbuf.lex_curr_p }

and string = parse
  | '\"' whites newline whites '\"'
    { incr_line_num lexbuf;
      string lexbuf }
  | '\"'
    { () }

  | escaped_character
    { store_string_char (char_for_character (Lexing.lexeme_char lexbuf 1));
      string lexbuf }

  | '\\' _
    { lexical_error
        (Illegal_escape (Lexing.lexeme lexbuf))
        lexbuf.lex_start_p lexbuf.lex_curr_p }

  | eof
    { match !string_start_pos with
      | Some (start_pos, end_pos) ->
        lexical_error Unterminated_string start_pos end_pos
      | _ -> assert false }
  | _
    { store_string_char (Lexing.lexeme_char lexbuf 0);
      string lexbuf }

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
