/***********************************************************************/
/*                                                                     */
/*                               Simport                               */
/*                                                                     */
/*                   Pierre Weis, INRIA Rocquencourt                   */
/*                                                                     */
/*  Copyright 2010-2019,                                               */
/*  Institut National de Recherche en Informatique et en Automatique.  */
/*  All rights reserved.                                               */
/*                                                                     */
/*  This file is distributed under the terms of the BSD License.       */
/*                                                                     */
/***********************************************************************/
%{

(* $Id: mdl_parser.mly,v 1.1 2019-05-10 14:04:02 jpc Exp $ *)

open Ast_types;;
open Mdl_ast;;
open Mdl_ast_functions;;
open Mdl_parse_error;;

(* Parsing errors. *)

let while_parsing construct =
  Printf.sprintf "while parsing %s" construct
;;

(* Error notification helpers *)
let not_an_integer loc s reason =
  Mdl_parse_error.parsing_error loc (Not_an_integer (s, reason))
;;

let wrong_integer loc s reason =
  parsing_error loc (Wrong_integer (s, reason))
;;

let not_a_boolean loc s reason =
  parsing_error loc (Not_a_boolean (s, reason))
;;

let duplicated_defaults_section loc s =
  parsing_error loc (Duplicated_defaults_section s)
;;

let invalid_connection_port loc s =
  parsing_error loc (Wrong_connection_port s)
;;

(* Basic construction functions *)

let make_int construct loc s =
  try
    int_of_string s
  with
  | Failure _ ->
    let reason = while_parsing construct in
    not_an_integer loc s reason
;;

let natural_of_mdl_string construct loc s =
  try
    Unatural.of_string s
  with
  | Failure reason ->
    let reason = while_parsing (construct ^ reason) in
    wrong_integer loc s reason
;;

let mdl_bool_of_mdl_string loc s =
  try mdl_bool_of_string s with
  | Failure _ ->
    let reason = "should be \"on\" or \"off\"" in
    not_a_boolean loc s reason
;;

let make_natural = natural_of_mdl_string;;

let make_float_tilda_list2 list2 =
  String (String.concat "~" list2)
;;

let make_object_id loc s =
  try
    let id =
      let id = Unatural.of_string s in
      if id >= Unatural._1_ then id else
      let reason = "an object identifier is strictly positive" in
      wrong_integer loc s reason in
    Configuration.default_slx_blockdiagram_file_name, id
  with
  | Failure _ ->
    let reason = while_parsing "an object identifier" in
    wrong_integer loc s reason
;;

let make_vendor = function
  | Ident.Ident "Simulink" -> Mdl_ast.Simulink
  | Ident.Ident "Scicos" -> Scicos
  | Ident.Ident "RTW" -> Rtw
  | Ident.Ident _ as ident -> Ui ident
;;

let make_vendor_binding vendor_ident vendor_config_longident vendor_items =
  {
    vendor_ident = vendor_ident;
    vendor_configuration_longident = vendor_config_longident;
    vendor_items = vendor_items;
  }
;;

let make_configuration_longident idents =
  Say.fdebug
    (fun epf ->
     Format.fprintf epf "make_configuration_longident <-@ %a"
       (Og_printer.printer_List
          Ident_print.printer_ident Og_precedence.Ck_toplevel)
       idents);
  let rec loop = function
    | [] -> assert false
    | [ ident; ] -> Configuration_lident ident
    | ident :: idents -> Configuration_ldot (loop idents, ident) in
  let configuration_longident = loop (List.rev idents) in
  Say.fdebug
    (fun epf ->
     Format.fprintf epf "make_configuration_longident ->@ %a"
       Mdl_ast_print.print_configuration_longident configuration_longident);
  configuration_longident
;;

let make_block_vendor = make_vendor_binding;;

let make_mdl_location loc ints =
  let construct = "a location" in
  List.map (make_int construct loc) ints
;;

let make_integer_matrix loc =
(* FIXME: matrices are rectangular! Should check the fact! *)

  let construct = "an integer matrix" in

  let make_integer s = make_int construct loc s in

  let make_integer = function
    | Int s -> make_integer s
    | Float s ->
      let reason = while_parsing construct in
      not_an_integer loc s reason
    | Bool _ | String _ | Ident _ ->
      let reason = while_parsing construct in
      not_an_integer loc "not an immediate integer" reason in

  let make_integer_vector = function
    | Value_simple _ | Value_bindings _ | Value_matrix _ ->
      assert false
    | Value_vector v ->
      Array.of_list (List.map make_integer v) in

  function
  | Value_simple _ | Value_bindings _ ->
    assert false
  | Value_vector _ as vect ->
    Array.of_list [make_integer_vector vect]
  | Value_matrix a ->
    Array.of_list (List.map make_integer_vector a)
;;

let make_integer_matrix_from_string loc =

  let construct = "an integer matrix" in

  let scan_line ib =
    let rec loop accu =
      Scanf.bscanf ib " %0c" (function
      | ',' -> Scanf.bscanf ib "%c " (fun _ -> loop accu)
      | '0'..'9' | '-' | '+' ->
        Scanf.bscanf ib " %i " (fun i -> loop (i :: accu))
      | _ -> Array.of_list (List.rev accu)) in
    loop [] in

  let scan_lines ib =
    let rec loop accu =
      Scanf.bscanf ib " %0c" (function
      | ';' -> Scanf.bscanf ib "%c " (fun _ -> loop accu)
      | '0'..'9' | '-' | '+' ->
        Scanf.bscanf ib " %r " scan_line (fun l -> loop (l :: accu))
      | _ -> Array.of_list (List.rev accu)) in
    loop [] in

  fun s ->
  let ib = Scanf.Scanning.from_string s in
  try Scanf.bscanf ib "[ %r ]" scan_lines (fun lines -> lines) with
  | _ ->
    let because = Invalid_string_specification (s, construct) in
    parsing_error loc because
;;

let make_matdata_section_item loc nr rs =
  let construct = "a matdata section" in
  let nr = make_natural construct loc nr in
  { matdata_num_records = nr; matdata_records = rs; }
;;

let make_data_record tag data =
  { matdata_record_tag = tag; matdata_record_data = data; }
;;

let make_interface_block btype bstype bsblock bports biparams = {
  interface_block_type = btype;
  interface_block_source_type = bstype;
  interface_block_source_path = bsblock;
  interface_block_ports_description = bports;
  interface_block_parameters = biparams;
}
;;

type defaults_section =
  | Block_graphical_defaults_section of block_graphical_defaults
  | Block_parameter_defaults_section of block_parameter_defaults list
  | Annotation_graphical_defaults_section of annotation_graphical_defaults
  | Line_graphical_defaults_section of line_graphical_defaults
  | Mask_graphical_defaults_section of mask_graphical_defaults
  | Mask_parameter_defaults_section of mask_parameter_defaults

and block_graphical_defaults = block_graphical_default list
and annotation_graphical_defaults = annotation_graphical_default list
and line_graphical_defaults = line_graphical_default list
and mask_graphical_defaults = mask_graphical_default list
and mask_parameter_defaults = mask_parameter_default list
;;

type configuration_set_parameter =
  | Configuration_parameter of Mdl_ast.model_configuration_parameter
  | Model_configuration of Mdl_ast.model_configuration
;;

let make_model loc mid mvers csp mcso _cgs wsl defaults_sections ss =

  let separate_parameters csp =
    let rec loop accu_params accu_confs = function
      | Configuration_parameter bdg :: params ->
        loop (bdg :: accu_params) accu_confs params
      | Model_configuration conf :: params ->
        loop accu_params (conf :: accu_confs) params
      | [] -> accu_params, accu_confs in
    loop [] [] csp in

  let rec separate_defaults accu = function
    | Block_graphical_defaults_section l :: defaults ->
      if accu.block_graphical_defaults_section = [] then
        separate_defaults
          {accu with block_graphical_defaults_section = l} defaults else
      duplicated_defaults_section loc "block_graphical_defaults_section"
    | Block_parameter_defaults_section l :: defaults ->
      if accu.block_parameter_defaults_section = [] then
        separate_defaults
          {accu with block_parameter_defaults_section = l} defaults else
      duplicated_defaults_section loc "block_parameter_defaults_section"
    | Annotation_graphical_defaults_section l :: defaults ->
      if accu.annotation_graphical_defaults_section = [] then
        separate_defaults
          {accu with annotation_graphical_defaults_section = l}
          defaults else
      duplicated_defaults_section
        loc "annotation_graphical_defaults_section"
    | Line_graphical_defaults_section l :: defaults ->
      if accu.line_graphical_defaults_section = [] then
        separate_defaults
          {accu with line_graphical_defaults_section = l}
          defaults else
      duplicated_defaults_section loc "line_graphical_defaults_section"
    | Mask_graphical_defaults_section l :: defaults ->
      if accu.mask_graphical_defaults_section = [] then
        separate_defaults
          {accu with mask_graphical_defaults_section = l}
          defaults else
      duplicated_defaults_section
        loc "mask_graphical_defaults_section"
    | Mask_parameter_defaults_section l :: defaults ->
      if accu.mask_parameter_defaults_section = [] then
        separate_defaults
          {accu with mask_parameter_defaults_section = l}
          defaults else
      duplicated_defaults_section
        loc "mask_parameter_defaults_section"
    | [] -> accu in

  let model_configuration_parameters, model_configurations =
    let rev_model_configuration_parameters, rev_model_configurations =
      separate_parameters csp in
    List.rev rev_model_configuration_parameters,
    List.rev rev_model_configurations in

  let model =
    let init_model = {
      model_ident = mid;
      model_version = mvers;
      model_configuration_parameters = model_configuration_parameters;
      model_configurations = model_configurations;
      model_set_configuration_option = mcso;
      model_workspace = wsl;
  (* Should disappear *)
      block_graphical_defaults_section = [];
      block_parameter_defaults_section = [];
      annotation_graphical_defaults_section = [];
      line_graphical_defaults_section = [];
      mask_graphical_defaults_section = [];
      mask_parameter_defaults_section = [];
  (* Should disappear in favor of
      code_generation_section = cgs;
      model_defaults_sections: defaults_sections;
  *)
      model_system = ss;
    } in

    separate_defaults init_model defaults_sections in

  model
;;
(* FIXME: remove or implement!
  {
    rev_model with
    model_ident = mid;
    model_version = mvers;
    model_configuration_parameters_section =
      List.rev rev_model.model_configuration_parameters;
    model_configurations_section =
      List.rev rev_model.model_configurations;
    model_configuration_set_section = mcso;
    model_workspace = wsl;
(* Should disappear *)
    block_graphical_defaults_section =
      List.rev rev_model.block_graphical_defaults_section;
    block_parameter_defaults_section =
      List.rev rev_model.block_parameter_defaults_section;
    annotation_graphical_defaults_section =
      List.rev rev_model.annotation_graphical_defaults_section;
    line_graphical_defaults_section =
      List.rev rev_model.line_graphical_defaults_section;
    mask_graphical_defaults_section =
      List.rev rev_model.mask_graphical_defaults_section;
    mask_parameter_defaults_section =
      List.rev rev_model.mask_parameter_defaults_section;
(* should disappear in favor of
    code_generation_section = cgs;
    model_defaults_sections: defaults_sections;
*)
    model_system = ss;
  }
;;
*)

let make_interface id bls = {
  interface_ident = id;
  interface_blocks = bls;
}
;;

let make_line_connector block_name block_port =
  {
    line_connector_block = block_name;
    line_connector_port = block_port;
  }
;;

let make_connection_type loc = function
  | "DEST_SRC" -> Dest_src
  | "DEST_DEST" -> Dest_dest
  | "SRC_DEST" -> Src_dest
  | "SRC_SRC" -> Src_src
  | s -> parsing_error loc (Unknown_connection_type s)
;;

let make_line_type loc = function
  | "Connection" -> Connection
  | s -> parsing_error loc (Unknown_line_type s)
;;

let make_port_number loc id port_number =
  let Ident.Ident s = id in
  match s with
  | "in"
  | "out" -> Port_number port_number
  | "lconn" -> Port_lconn port_number
  | "rconn" -> Port_rconn port_number
  | s -> parsing_error loc (Unknown_port_kind s)
;;

let make_horizontal_alignment loc = function
  | "center" -> Alignment_Center
  | "left" -> Alignment_Left
  | "right" -> Alignment_Right
  | s -> parsing_error loc (Unknown_alignment s)
;;

let make_vertical_alignment loc = function
  | "top" -> Alignment_Top
  | "middle" -> Alignment_Middle
  | "bottom" -> Alignment_Bottom
  | s -> parsing_error loc (Unknown_alignment s)
;;

let make_array_type _loc = function
  | "Cell" -> Cell
  | "Handle" -> Handle
  | "Struct" -> Struct
  | s -> Array_type s
;;

let make_workspace_data_source_type loc = function
  | "M-Code" -> M_Code
  | "MATLAB Code" -> MATLAB_Code
  | "MATLAB File" -> MATLAB_File
  | "MAT-File" -> MAT_File
  | s ->
    parsing_error loc (Unknown_workspace_data_source_type s)
;;

let make_custom_property_type = function
  | "uint8" -> Custom_property_type_uint8
  | "uint16" -> Custom_property_type_uint16
  | "uint32" -> Custom_property_type_uint32
  | "uint64" -> Custom_property_type_uint64
  | "int8" -> Custom_property_type_int8
  | "int16" -> Custom_property_type_int16
  | "int32" -> Custom_property_type_int32
  | "int64" -> Custom_property_type_int64
  | "slbool" -> Custom_property_type_slbool
  | "string" -> Custom_property_type_string
  | s -> Custom_property_type_custom (Ast_types.mdl_name_of_string s)
;;

let make_tag_ident loc s =
  try
    Ast_types.ident_from_string s
  with
  | Failure _ -> parsing_error loc (Invalid_tag_ident s)
;;

(* Some block types are not valid Matlab identifiers; so they are given as
   strings in the MDL files. We find their associated block types here.
   We call Mdl_block_types_tables.block_type_of_string and catch
   the Not_found in case of error.
   From the time being, this problem arises for blocks specified as
   "Interpolation_n-D" | "Lookup_n-D" | "M-S-Function" | "S-Function".
 *)
let make_block_type_from_string_specification loc s =
  let quoted_s = Printf.sprintf "\"%s\"" s in
  try Mdl_block_types_tables.string_specified_block_type_of_string quoted_s with
  | Not_found ->
    let message =
      Printf.sprintf "%s (a string specified block type)" quoted_s in
    parsing_error loc (Unknown_block_type message)
;;

let make_block_sf_block_type_from_string_specification loc = function
  | "NONE" -> Block_sf_block_type_none
  | "MATLAB Function" -> Block_sf_block_type_matlab_function
  | "Chart" -> Block_sf_block_type_chart
  | s ->
    let quoted_s = Printf.sprintf "\"%s\"" s in
    let message =
      Printf.sprintf "%s (a string specified sf block type)" quoted_s in
    parsing_error loc (Unknown_block_sf_block_type message)
;;

let connection_of_ident loc id =
  let Ident.Ident s = id in
  let ib = Scanf.Scanning.from_string s in
  try
    Scanf.bscanf ib "%cConn%d%!" (fun c n ->
      match c with
      | 'L' -> Port_lconn (mdl_port_number_of_int n)
      | 'R' -> Port_rconn (mdl_port_number_of_int n)
      | _ -> invalid_connection_port loc s)
  with
  | Scanf.Scan_failure _ | Failure _ | End_of_file ->
    invalid_connection_port loc s
;;

let simple_value_of_ident id = Ident id;;

%}

/* Identifiers */
%token <string> IDENT /* Lower/Upper case ident (e.g. x, cos, X, or xY) */
%token <Mdl_block_types.block_type> BLOCKTYPENAME
%token <Mdl_ast.vendor_ident> VENDOR

/* Basic constants */
%token <string> FLOAT
%token <string> INT
%token <string> STRING
%token <string> BOOL

/* Nested symbols */
%token OPEN_PAREN
%token CLOSE_PAREN
%token OPEN_BRACKET
%token CLOSE_BRACKET
%token OPEN_BRACE
%token CLOSE_BRACE

/* Standard punctuation */
%token COMMA
%token COLON
%token DOT
%token SEMICOLON
%token TILDA

/* End of file */
%token EOF

/* Keywords */
%token ANNOTATION
%token ANNOTATIONDEFAULTS
%token ARRAY
%token AUTOSIZE
%token BACKGROUNDCOLOR
%token BLOCK
%token BLOCKDEFAULTS
%token BLOCKINTERFACEPARAMETERS
%token BLOCKINTERFACES
%token BLOCKMIRROR
%token BLOCKPARAMETERDEFAULTS
%token BLOCKROTATION
%token BLOCKTYPE
%token BRANCH
%token CLICKFCN
%token CONFIGSET
%token CONNECTTYPE
%token CUSTOMPROPERTY
%token DATARECORD
%token DATATYPE
%token DIMENSION
%token DOLLAR_BACKUPCLASS
%token DOLLAR_CLASSNAME
%token DOLLAR_OBJECTID
%token DOLLAR_PROPNAME
%token DROPSHADOW
%token DSTBLOCK
%token DSTPORT
%token ENABLEPORT
%token ENUMSTRPAIRS
%token FONTANGLE
%token FONTNAME
%token FONTSIZE
%token FONTWEIGHT
%token FOREGROUNDCOLOR
%token GRAPHICALINTERFACE
%token HIDEAUTOMATICNAME
%token HORIZONTALALIGNMENT
%token INPORT
%token INTERFACENAME
%token INTERPRETER
%token LABELS
%token LENABLE
%token LIBRARY
%token LIFACTION
%token LINE
%token LINEDEFAULTS
%token LINETYPE
%token LIST
%token LISTTYPE
%token LOCATION
%token MASKCALLBACKSTRING
%token MASKCAPABILITIES
%token MASKDEFAULTS
%token MASKDESCRIPTION
%token MASKDISPLAY
%token MASKDISPLAYSTRING
%token MASKENABLESTRING
%token MASKHELP
%token MASKICONFRAME
%token MASKICONOPAQUE
%token MASKICONROTATE
%token MASKICONUNITS
%token MASKINITIALIZATION
%token MASKPARAMETERDEFAULTS
%token MASKPORTROTATE
%token MASKPROMPTSTRING
%token MASKRUNINITFORICONREDRAW
%token MASKSELFMODIFIABLE
%token MASKSTYLESTRING
%token MASKTABNAMESTRING
%token MASKTOOLTIPSTRING
%token MASKTUNABLEVALUESTRING
%token MASKTYPE
%token MASKVALUESTRING
%token MASKVARALIASSTRING
%token MASKVARIABLES
%token MASKVISIBILITYSTRING
%token MATDATA
%token MATSTRUCT
%token MODEL
%token NAME
%token NAMEPLACEMENT
%token NUMRECORDS
%token NUMROOTINPORTS
%token NUMROOTOUTPORTS
%token NUMROOTTRIGPORTS
%token OBJECT
%token ORIENTATION
%token OUTPORT
%token POINTS
%token PORT
%token PORTNUMBER
%token PORTS
%token POSITION
%token PROPNAME
%token SAVEDCHARACTERENCODING
%token SFBLOCKTYPE
%token SHARP
%token SHOWNAME
%token SID
%token SOURCEBLOCK
%token SOURCETYPE
%token SRCBLOCK
%token SRCPORT
%token STATEFLOW
%token STR
%token SYSTEM
%token TAG
%token TEXMODE
%token TEXT
%token TREATASATOMICUNIT
%token TRIGPORT
%token TYPE
%token DATA
%token USEDISPLAYTEXTASCLICKCALLBACK
%token VALUE
%token VERSION
%token VERTICALALIGNMENT
%token WORDWRAP
%token WSDATASOURCE
%token WSMATFILENAME
%token WSMATLABCODE
%token WSMCODE
%token WSMDLFILEDATA
%token WSSOURCEFILENAME
%token ZORDER
/* FIXME: remove or explain why this is use;ess
%token MASKHIDECONTENTS
*/

/* Stateflow specific (lowercase) keywords */
%token LARRAY
%token LCHART
%token LDATA
%token LDST
%token LEML
%token LEVENT
%token LEXECUTIONORDER
%token LFIRSTDATA
%token LFIRSTEVENT
%token LFIRSTINDEX
%token LFIRSTJUNCTION
%token LFIRSTTRANSITION
%token LFIRSTTARGET
%token LID
%token LINTERSECTION
%token LINSTANCE
%token LJUNCTION
%token LLABELSTRING
%token LLINKNODE
%token LTREENODE
%token LMACHINE
%token LCHARTFILENUMBER
%token LNAME
%token LPROPS
%token LPLANTMODELINGINFO
%token LSCOPE
%token LSIZE
%token LSCRIPT
%token LSRC
%token LSSIDNUMBER
%token LSTATE
%token LTARGET
%token LTRANSITION
%token LTRIGGER
%token LTYPE

/* Precedences and associativities. */

/* Predefined precedences to resolve conflicts. */

%nonassoc TILDA
%nonassoc DOT

%nonassoc CLOSE_PAREN
%nonassoc OPEN_PAREN
%nonassoc CLOSE_BRACKET
%nonassoc OPEN_BRACKET
%nonassoc CLOSE_BRACE
%nonassoc OPEN_BRACE

/* Entry points */

%start mdl_file
%type <Mdl_parsing_ast.simulink_file_contents> mdl_file

/* Parser entries used by convertor from XML to mdl AST */

%start position_entry
%type <Mdl_ast.position> position_entry

%start location_entry
%type <Mdl_ast.position> location_entry

%start simple_value_entry
%type <Mdl_ast.simple_value> simple_value_entry

%start slx_model_value_entry
%type <Mdl_ast.model_value> slx_model_value_entry

%start block_type_value_entry
%type <Mdl_block_types.block_type> block_type_value_entry

%start array_type_entry
%type <Mdl_ast.array_type> array_type_entry

%start line_type_entry
%type <Mdl_ast.line_type> line_type_entry

%start connection_type_entry
%type <Mdl_ast.connection_type> connection_type_entry

%start naturals_entry
%type <Unatural.t list> naturals_entry

%start natural_vector_entry
%type <Unatural.t list> natural_vector_entry

%start horizontal_alignment_entry
%type <Mdl_ast.horizontal_alignment> horizontal_alignment_entry

%start vertical_alignment_entry
%type <Mdl_ast.vertical_alignment> vertical_alignment_entry

%start integer_matrix_entry
%type <Mdl_ast.integer_matrix> integer_matrix_entry

%start block_sid_value_entry
%type <Mdl_ast.block_sid> block_sid_value_entry

%start slx_port_connector_entry
%type <Mdl_ast.block_sid * Mdl_ast.port_value> slx_port_connector_entry

%start vendor_configuration_longident_entry
%type <Mdl_ast.vendor_ident * Mdl_ast.configuration_longident>
      vendor_configuration_longident_entry

%start workspace_data_source_type_entry
%type <Mdl_ast.workspace_data_source_type> workspace_data_source_type_entry

%start custom_property_type_entry
%type <Mdl_ast.custom_property_type> custom_property_type_entry

%start intersection_entry
%type <Mdl_ast.stateflow_numbers> intersection_entry

%%

/** Parser entries used by the converter from XML to mdl AST **/

position_entry:
  | position_value EOF { $1 }
;

location_entry:
  | location_value EOF { $1 }
;

simple_value_entry:
  | simple_binding_value EOF { $1 }
;

slx_model_value_entry:
  | slx_binding_value EOF { $1 }
;

block_type_value_entry:
  | block_type_value EOF { $1 }
;

array_type_entry:
  | array_type EOF { $1 }
;

line_type_entry:
  | line_type_value EOF { $1 }
;

connection_type_entry:
  | connection_type_value EOF { $1 }
;

naturals_entry:
  | natural_commas EOF { $1 }
;

natural_vector_entry:
  | natural_vector EOF { $1 }

horizontal_alignment_entry:
  | horizontal_alignment_value EOF { $1 }
;

vertical_alignment_entry:
  | vertical_alignment_value EOF { $1 }
;

integer_matrix_entry:
  | integer_matrix EOF { $1 }
;

block_sid_value_entry:
  | block_sid_value EOF { $1 }
;

slx_port_connector_entry:
  | block_sid_value SHARP slx_port_value EOF { $1, $3 }
;

vendor_configuration_longident_entry:
  | vendor_configuration_longident EOF { $1 }
;

workspace_data_source_type_entry:
  | STRING EOF
    { let loc = Location.get_symbol_loc () in
      make_workspace_data_source_type loc $1 }
;

custom_property_type_entry:
  | STRING EOF
    { make_custom_property_type $1 }
;

intersection_entry:
  | intersection EOF
    { $1 }
;

/* Slx specific rules */

slx_port_value:
  | ident COLON port_number
    { let loc = Location.get_symbol_loc () in
      make_port_number loc $1 $3 }
  | common_port_value { $1 }
;

/** Main parser entry **/

mdl_file:
  | INTERFACENAME interface_section EOF
    { Interface $2 }
  | MODEL model_section model_trailor_section EOF
    { Implementation (Implementation_model $2 :: $3)}
  | LIBRARY model_section model_trailor_section EOF
    { Implementation (Implementation_library $2 :: $3)}
;

/** Parsing model and library interface **/

interface_section:
  | ident BLOCKINTERFACES OPEN_BRACE interface_blocks CLOSE_BRACE
    { make_interface $1 $4 }
;

interface_blocks:
  | { [] }
  | interface_block interface_blocks { $1 :: $2 }
;

/** Parsing model and library implementation **/

/* Model and library section contents */

model_section:
  | OPEN_BRACE model_section_items CLOSE_BRACE { $2 }
;

model_section_items:
  model_ident
  model_version
  model_configuration_parameters_section
  model_configuration_set_section
  code_generation_section
  ws_data_section
  defaults_sections
  system
  { let loc = Location.get_symbol_loc () in
    make_model loc $1 $2 $3 $4 $5 $6 $7 $8}
;

model_ident:
  | NAME ident_in_string { $2 }
  | NAME ident { $2 }
;

ident_in_string:
 | STRING { ident_from_string $1 }
;

model_version:
  | VERSION FLOAT { Mdl_ast.Mdl_model_version $2 }
;

/** Sub-section 1 of model and library sections:
  configuration set parameters section **/

model_configuration_parameters_section:
  | { [] }
  | model_configuration_parameter_section_item
    model_configuration_parameters_section
    { $1 :: $2 }
;

model_configuration_parameter_section_item:
  | array_description
    { Model_configuration $1 }
  | GRAPHICALINTERFACE OPEN_BRACE graphical_interface_items CLOSE_BRACE
    { Configuration_parameter (Configuration_graphical_interface $3) }
  | vendor_configuration_longident vendor_items
    { let vendor_ident, configuration_longident = $1 in
      let bdg =
        Configuration_vendor_binding {
          vendor_ident = vendor_ident;
          vendor_configuration_longident = configuration_longident;
          vendor_items = $2;
        } in
      Configuration_parameter bdg
    }
  | configuration_binding { Configuration_parameter $1 }
;

graphical_interface_items:
  | { [] }
  | graphical_interface_item graphical_interface_items
    { $1 :: $2 }
;

graphical_interface_item:
  | NUMROOTINPORTS natural { Graphical_inports_number $2 }
  | NUMROOTOUTPORTS natural { Graphical_outports_number $2 }
  | NUMROOTTRIGPORTS natural { Graphical_trigports_number $2 }
  | INPORT OPEN_BRACE io_port_bindings CLOSE_BRACE { Graphical_inport $3 }
  | OUTPORT OPEN_BRACE io_port_bindings CLOSE_BRACE { Graphical_outport $3 }
  | TRIGPORT OPEN_BRACE io_port_bindings CLOSE_BRACE { Graphical_trigport $3 }
  | ENABLEPORT OPEN_BRACE io_port_bindings CLOSE_BRACE
    { Graphical_enableport $3 }
  | binding { Graphical_binding $1 }
;

vendor_configuration_longident:
  | VENDOR DOT CONFIGSET { $1, config_set_longident }
  | VENDOR DOT configuration_longident { $1, $3 }
  | ident DOT configuration_longident { make_vendor $1, $3 }
;

configuration_longident:
  | dot_idents { make_configuration_longident $1 }
;

dot_idents:
  | binding_ident { [ $1; ] }
  | binding_ident DOT dot_idents { $1 :: $3 }
;

io_port_bindings:
  | { [] }
  | io_port_binding_item io_port_bindings
    { $1 :: $2 }
;

io_port_binding_item:
  | NAME name { Io_port_name $2 }
  | binding { Io_port_binding $1 }
;

configuration_binding:
  | SAVEDCHARACTERENCODING STRING
    { Configuration.fix_source_encoding $2;
      Configuration_binding
        (make_binding saved_character_encoding_ident (String $2)) }
  | configuration_binding_ident simple_binding_value
    { Configuration_binding (Mdl_ast_functions.make_binding $1 $2) }
  | object_description
    { Configuration_object $1 }
;

configuration_binding_ident:
  | TAG { tag_ident }
  | ident { $1 }
;

object_description:
  | OBJECT OPEN_BRACE object_items CLOSE_BRACE
    { $3 }
;

object_items:
  | { [] }
  | object_item object_items
    { $1 :: $2 }
;

object_item:
  | dollar_binding { Object_dollar_binding $1 }
  | simple_object_item { $1 }
;

simple_object_item:
  | CUSTOMPROPERTY OPEN_BRACE custom_property_bindings CLOSE_BRACE
    { Object_custom_property $3 }
  | binding { Object_binding $1 }
  | array_description { Object_array $1 }
  | object_description { Object_object $1 }
  | NAME name { Object_name $2 }
  | TYPE name { Object_type $2 }
;

vendor_items:
  | OPEN_BRACE vendor_item_list CLOSE_BRACE { $2 }
;

vendor_item_list:
  | { [] }
  | vendor_item vendor_item_list { $1 :: $2 }
;

vendor_item:
  | dollar_binding { Vendor_dollar_binding $1 }
  | simple_object_item { Vendor_binding $1 }
  | vendor_configuration_longident vendor_items
    { let vendor_ident, configuration_longident = $1 in
      let vendor_binding =
        make_vendor_binding vendor_ident configuration_longident $2 in
      Vendor_vendor_binding vendor_binding
    }
;

dollar_binding:
  | dollar_object_id { Dollar_object_id $1 }
  | dollar_class_name { Dollar_class_name $1 }
  | dollar_backup_class { Dollar_backup_class $1 }
  | dollar_prop_name { Dollar_prop_name $1 }
;

dollar_prop_name:
  | DOLLAR_PROPNAME ident_in_string { $2 }
;

dollar_class_name:
  | DOLLAR_CLASSNAME name { $2 }
;

dollar_backup_class:
  DOLLAR_BACKUPCLASS name { $2 }
;
/** End of Sub-section 1 of model and library sections **/

/** Sub-section 2 of model and library sections:
  model configuration section list **/
/* FIXME: remove or implement!
model_configurations_section:
  | { [] }
  | array_description model_configurations_section { $1 :: $2}
;
*/

array_description:
  | ARRAY OPEN_BRACE array_contents CLOSE_BRACE { $3 }
;

array_contents:
  TYPE array_type
  dimension_simple_value
  array_items
  PROPNAME ident_in_string
  {
    {
      array_type = $2;
      array_dimension = $3;
      array_items = $4;
      array_prop_name = $6;
    }
  }
;

array_type:
  | STRING
    {
      let loc = Location.get_symbol_loc () in
      make_array_type loc $1
    }
;

dimension_simple_value:
  | DIMENSION natural_or_string { $2 }
;

array_items:
  | { [] }
  | array_item array_items { $1 :: $2 }
;

array_item:
  | binding
    { Array_binding $1 }
  | block_name
    { Array_name $1 }

/*
  | vendor_configuration_longident vendor_items
    { let vendor, configuration_longident = $1 in
      make_array_vendor_item vendor configuration_longident $2}

TODO: Can we improve that and use general rule
We use make_array_vendor_item instead of
    Array_vendor_binding
      (make_vendor_binding vendor configuration_longident $3)
since we must distinguish between Array_vendor_binding and
    Array_vendor_configset cases.

For the time being we expand the rule vendor_configuration_longident to
distinguish those cases via parsing rules.

*/

  | VENDOR DOT CONFIGSET vendor_items
    { Array_vendor_configset ($1, $4) }
  | ident DOT CONFIGSET vendor_items
    { Array_vendor_configset (make_vendor $1, $4) }

  | VENDOR DOT configuration_longident vendor_items
    { Array_vendor_binding (make_vendor_binding $1 $3 $4) }
  | ident DOT configuration_longident vendor_items
    { let vendor = make_vendor $1 in
      Array_vendor_binding (make_vendor_binding vendor $3 $4) }

  | array_description
    { Array_sub_array $1 }
  | matstruct_description
    { Array_matstruct $1 }
  | object_description
    { Array_object $1 }
;

matstruct_description:
  | MATSTRUCT OPEN_BRACE matstruct_items CLOSE_BRACE
    { $3 }
;

matstruct_items:
  | { [] }
  | matstruct_item matstruct_items { $1 :: $2 }
;

matstruct_item:
  | TYPE name { Matstruct_type $2 }
  | NAME name { Matstruct_name $2 }
  | PORT simple_binding_value
    { Matstruct_binding (make_binding port_ident (Value_simple $2)) }
  | binding { Matstruct_binding $1 }
  | array_description { Matstruct_array $1 }
;

custom_property_bindings:
  | DATATYPE STRING
    enum_str_pairs
    ident_value_pair
    {
      let ident, v = $4 in
      {
        custom_property_type = make_custom_property_type $2;
        custom_property_enum_str_pairs = $3;
        custom_property_ident = ident;
        custom_property_value = v;
      }
    }
;

enum_str_pairs:
  | { [] }
  | ENUMSTRPAIRS OPEN_BRACE enum_str_pair_list CLOSE_BRACE
    { $3 }
;

enum_str_pair_list:
  | { [] }
  | enum_str_pair enum_str_pair_list { $1 :: $2 }
;

enum_str_pair:
  VALUE natural
  STR text
  { $2, $4 }
;

ident_value_pair:
 |  NAME ident_in_string
    VALUE simple_binding_value
    { $2, $4 }
;
/** End of Sub-section 2 of model and library sections **/

/** Sub-section 3 of model and library sections:
   model configuration set **/

model_configuration_set_section:
  | vendor_configuration_longident
    OPEN_BRACE dollar_prop_name dollar_object_id CLOSE_BRACE
    { let vendor_ident, configuration_longident = $1 in
      Some {
        configuration_vendor = vendor_ident;
        configuration_longident = configuration_longident;
        configuration_prop_name = $3;
        configuration_object_id = $4;
      }
    }
  | { None }
;

/* FIXME: remove or say why this does not work
configset:
  | CONFIGSET { config_set_ident }
;
*/

dollar_object_id:
  DOLLAR_OBJECTID INT
  { let loc = Location.get_symbol_loc () in
    make_object_id loc $2 }
;
/* End of Sub-section 3 of model and library sections */

/** Sub-section 4 of model and library sections:
  code generation bindings **/

code_generation_section:
  | { [] }
  | configuration_binding code_generation_section { $1 :: $2 }
;
/** End of Sub-section 4 of model and library sections **/

/** Sub-section 5 of model and library sections:
  workspace data section **/

ws_data_section:
  | { None }
  | ws_data_section_item ws_data_section_items
    { Some ($1 :: $2) }
;

ws_data_section_items:
  | { [ ] }
  | ws_data_section_item ws_data_section_items
    { $1 :: $2 }
;

ws_data_section_item:
  | WSDATASOURCE STRING
    { let loc = Location.get_symbol_loc () in
      let source_type =
        make_workspace_data_source_type loc $2 in
      Workspace_data_source_type source_type }
  | WSMATFILENAME STRING
    { Workspace_data_mat_file_name $2 }
  | WSSOURCEFILENAME STRING
    { Workspace_data_source_file_name $2 }
  | WSMATLABCODE simple_binding_value
    { Workspace_data_mcode $2 }
  | WSMCODE simple_binding_value
    { Workspace_data_mcode $2 }
  | WSMDLFILEDATA STRING
    { let loc = Location.get_symbol_loc () in
      Workspace_data_mdl_file (make_tag_ident loc $2) }
;
/** End of Sub-section 5 of model and library sections **/

/** Sub-section 6 of model and library sections:
  defaults sections **/

defaults_sections:
  | { [] }
  | defaults_section defaults_sections { $1 :: $2 }
;

defaults_section:
  | block_graphical_defaults_section { Block_graphical_defaults_section $1 }
  | block_parameter_defaults_section { Block_parameter_defaults_section $1 }
  | annotation_graphical_defaults_section
    { Annotation_graphical_defaults_section $1 }
  | line_graphical_defaults_section { Line_graphical_defaults_section $1 }
  | mask_graphical_defaults_section { Mask_graphical_defaults_section $1 }
  | mask_parameter_defaults_section { Mask_parameter_defaults_section $1 }
;

/* Sub-section block graphical defaults */
block_graphical_defaults_section:
  BLOCKDEFAULTS OPEN_BRACE block_graphical_defaults CLOSE_BRACE { $3 }
;

block_graphical_defaults:
  | { [] }
  | block_graphical_default block_graphical_defaults { $1 :: $2 }
;

font_and_color:
  | background { Background $1 }
  | drop_shadow { Drop_shadow $1 }
  | font_item { Font_item $1 }
  | foreground { Foreground $1 }
;

block_graphical_default:
  | mirror { Block_mirror $1 }
  | name_placement { Block_name_placement $1 }
  | orientation { Block_orientation $1 }
  | rotation { Block_rotation $1 }
  | show_name { Block_show_name $1 }
  | hide_automatic_name { Block_hide_automatic_name $1 }
  | font_and_color { Block_font_and_color $1 }
  | zorder { Block_zorder $1 }
;
/* End of Sub-section block graphical defaults */

/* Sub-section block parameter defaults */
block_parameter_defaults_section:
  | BLOCKPARAMETERDEFAULTS OPEN_BRACE
      block_parameter_defaults_list
    CLOSE_BRACE
    { $3 }
;

block_parameter_defaults_list:
  | { [] }
  | block_parameter_defaults block_parameter_defaults_list { $1 :: $2 }
;

block_parameter_defaults:
  BLOCK OPEN_BRACE block_type parameter_defaults CLOSE_BRACE
  { { block_type = $3; parameter_defaults = $4; } }
;

parameter_defaults:
  | { [] }
  | parameter_default parameter_defaults { $1 :: $2 }
;

parameter_default:
  | parameter_default_binding { $1 }
  | block_sf_block_type_binding { $1 }
;

parameter_default_binding:
  | PORT simple_binding_value
    { make_port_binding (Value_simple $2) }
  | SYSTEM simple_binding_value
    { make_system_binding (Value_simple $2) }
  | block_treat_as_atomic_unit
    { $1 }
  | simple_binding
    { $1 }
;

block_treat_as_atomic_unit:
  | TREATASATOMICUNIT bool_or_string_value
    { Mdl_ast_functions.make_treat_as_atomic_unit_binding $2 }
;

/* End of Sub-section block parameter defaults */

/* Sub-section annotation defaults */
/* TODO: is there no tags in annotation?
annotation_tag:
  | TAG STRING
    { Configuration.convert_source_file_string $2 }
;
*/

annotation_graphical_defaults_section:
  ANNOTATIONDEFAULTS OPEN_BRACE annotation_graphical_defaults CLOSE_BRACE { $3 }
;

annotation_graphical_defaults:
  | { [] }
    /* To handle all unknown new features such that: we simply parse a
       binding that would be ignores afterwards...
       MarkupType string
       AnnotationType string
       FixedHeight of bool
       FixedWidth of bool
    */
  | binding annotation_graphical_defaults
    { Annotation_graphical_binding $1 :: $2}
  | annotation_graphical_default annotation_graphical_defaults { $1 :: $2 }
;

annotation_graphical_default:
  | font_and_color { Font_and_color $1 }
  | horizontal_alignment { Horizontal_alignment $1 }
  | vertical_alignment { Vertical_alignment $1 }
  | text_as_click_cb { Text_as_click_cb $1 }
  | text_click_fcn { Click_fcn $1 }
  | tex_mode_option { Tex_mode $1 }
  | interpreter { Interpreter $1 }
;

mdl_bool:
  | BOOL { Ast_types.mdl_bool_of_string $1 }
;

annotation_auto_size:
  | AUTOSIZE mdl_bool { $2 }
;

annotation_word_wrap:
  | WORDWRAP mdl_bool { $2 }
;

annotation_binding:
  | binding { $1 }
;
/* End of Sub-section annotation defaults */

/* Sub-section line defaults */
line_graphical_defaults_section:
  LINEDEFAULTS OPEN_BRACE line_graphical_defaults CLOSE_BRACE { $3 }
;

line_graphical_defaults:
  | { [] }
  | line_graphical_default line_graphical_defaults { $1 :: $2 }
;

line_graphical_default:
  | font_item { Line_font_item $1 }
  | zorder { Line_zorder_item $1 }
;
/* End of Sub-section line defaults */

/* Sub-section mask graphical defaults */
mask_graphical_defaults_section:
  | MASKDEFAULTS OPEN_BRACE mask_graphical_defaults CLOSE_BRACE { $3 }
;

mask_graphical_defaults:
  | { [] }
  | mask_graphical_default mask_graphical_defaults { $1 :: $2 }
;

mask_graphical_default:
  | binding { $1 }
;
/* End of Sub-section mask graphical defaults */

/* Sub-section mask parameter defaults */
mask_parameter_defaults_section:
  | MASKPARAMETERDEFAULTS OPEN_BRACE mask_parameter_defaults CLOSE_BRACE
    { $3 }
;

mask_parameter_defaults:
  | { [] }
  | binding mask_parameter_defaults { $1 :: $2 }
;

bindings:
  | { [] }
  | binding bindings
    { $1 :: $2 }
;

binding:
  | binding_ident binding_value
    { make_binding $1 $2 }
  | LIST OPEN_BRACE
      LISTTYPE binding_ident
      bindings
    CLOSE_BRACE
    { make_binding $4 (Value_bindings $5) }
;

sf_binding_ident:
  | ident { $1 }
  | BLOCKTYPENAME
    { let s = Mdl_block_types_tables.string_of_block_type $1 in
      make_ident s }
  | LOCATION { location_ident }
  | VALUE { value_ident }
  | DATATYPE { data_type_ident }
  | STR { string_ident }
  | INTERFACENAME { interface_name_ident }
  | TEXT { text_ident }
  | VERSION { version_ident }
  | SID { sid_ident }
  | TYPE { type_ident }
/* FIXME: remove or explain why this is no more useful
  | SYSTEM { system_ident }
*/
;

ident:
  | IDENT { make_ident $1 }
;

/* Adding lowercase stateflow keywords to regular binding idents */
binding_ident:
  | sf_binding_ident { $1 }
  | LCHART { Mdl_ast_functions.lchart_ident }
  | LDATA { Mdl_ast_functions.ldata_ident }
  | LDST { ldst_ident }
  | LEML { leml_ident }
  | LEVENT { levent_ident }
  | LENABLE { lenable_ident }
  | LEXECUTIONORDER { lexecution_order_ident }
  | LFIRSTDATA { lfirst_data_ident }
  | LFIRSTEVENT { lfirst_event_ident }
  | LFIRSTJUNCTION { lfirst_junction_ident }
  | LFIRSTTRANSITION { lfirst_transition_ident }
  | LFIRSTTARGET { lfirst_target_ident }
  | LID { lid_ident }
  | LINSTANCE { linstance_ident }
  | LINTERSECTION { lintersection_ident }
  | LJUNCTION { ljunction_ident }
  | LLABELSTRING { llabel_string_ident }
  | LLINKNODE { llink_node_ident }
  | LTREENODE { ltree_node_ident }
  | LMACHINE { lmachine_ident }
  | LCHARTFILENUMBER { lchart_file_number_ident }
  | LNAME { lname_ident }
  | LPROPS { lprops_ident }
  | LSCOPE { lscope_ident }
  | LSCRIPT { lscript_ident }
  | LSIZE { lsize_ident }
  | LSRC { lsrc_ident }
  | LSSIDNUMBER { lss_id_number_ident }
  /* LSTATE cannot be added here, since it could not give a binding,
     because this keyword introduces block item Block_state.
  | LSTATE { lstate_ident }
  */
  | LTARGET { ltarget_ident }
  | LTRANSITION { ltransition_ident }
  | LTRIGGER { ltrigger_ident }
  | LTYPE { ltype_ident }
;

binding_value:
  | simple_binding_value
    { Value_simple $1 }
  | OPEN_BRACKET simple_binding_value_array CLOSE_BRACKET
    { $2 }
  | OPEN_BRACE bindings CLOSE_BRACE
    { Value_bindings $2 }
;

slx_binding_value:
  | constant_binding_value
    { Value_simple $1 }
  | OPEN_BRACKET simple_binding_value_array CLOSE_BRACKET
    { $2 }
/* FIXME: explain why this is useless
  | OPEN_BRACE bindings CLOSE_BRACE
    { Value_bindings $2 }
*/
;

constant_binding_value:
  | BOOL
    { Bool $1 }
  | FLOAT
    { Float $1 }
  | INT
    { Int $1 }
  | STRING
    { String $1 }
    /* Dealing with compound tokens made of tilda separated integers or
       floats (for instance -1~0.2).
       We simply return a String constant simple value */
  | float_or_int_tilda_list2
    { make_float_tilda_list2 $1 }
;

float_or_int_string:
  | INT
    { $1 }
  | FLOAT
    { $1 }
;

float_or_int_tilda_list2:
  | float_or_int_string TILDA float_or_int_string { [ $1; $3; ] }
  | float_or_int_string TILDA float_or_int_tilda_list2 { $1 :: $3 }
;

ident_or_keyword:
  | ident
    { simple_value_of_ident $1 }
  | ARRAY
    { simple_value_of_ident Mdl_ast_functions.array_ident }
;

simple_binding_value:
  | constant_binding_value
    { $1 }
  | ident_or_keyword
    { $1 }
;

simple_binding:
  | LIST OPEN_BRACE
      LISTTYPE binding_ident
      simple_bindings
    CLOSE_BRACE
    { make_binding $4 (Value_bindings $5) }
  | list_binding_ident simple_binding_value
    { make_binding $1 (Value_simple $2) }
  | list_binding_ident OPEN_BRACKET simple_binding_value_array CLOSE_BRACKET
    { make_binding $1 $3 }
  | list_binding_ident OPEN_BRACE simple_bindings CLOSE_BRACE
    { make_binding $1 (Value_bindings $3) }
;

list_binding_ident:
  | binding_ident { $1 }
  | TYPE { type_ident }
;

simple_bindings:
  | { [] }
  | simple_binding simple_bindings
    { $1 :: $2 }
;

simple_binding_value_array:
  | simple_binding_value_vector
    { $1 }
  | simple_binding_value_vector
    SEMICOLON
    simple_binding_value_vector_semicolons
    { Value_matrix ($1 :: $3) }
;

simple_binding_value_vector:
  | simple_binding_value_commas
    { Value_vector $1 }
;

simple_binding_value_commas:
  | { [] }
  | simple_binding_value
    COMMA simple_binding_value_commas
    { $1 :: $3 }
  | simple_binding_value
    simple_binding_value_commas
    { $1 :: $2 }
;

simple_binding_value_vector_semicolons:
  | simple_binding_value_vector
    { [ $1 ] }
  | simple_binding_value_vector
    SEMICOLON simple_binding_value_vector_semicolons
    { $1 :: $3 }
;

natural_commas:
  | { [ ] }
  | natural natural_commas
    { $1 :: $2 }
  | natural COMMA natural_commas
    { $1 :: $3 }
;

integer_commas:
  | { [ ] }
  | integer integer_commas
    { $1 :: $2 }
  | integer COMMA integer_commas
    { $1 :: $3 }
;

natural_vector:
  | OPEN_BRACKET natural_commas CLOSE_BRACKET
    { $2 }
;
integer_vector:
  | OPEN_BRACKET integer_commas CLOSE_BRACKET
    { $2 }
;

int_commas:
  | { [ ] }
  | INT int_commas
    { $1 :: $2 }
  | INT COMMA int_commas
    { $1 :: $3 }
;

int_vector:
  | OPEN_BRACKET int_commas CLOSE_BRACKET
    { $2 }
;
/* End of Sub-section mask parameter defaults */

/** End of Sub-section 6 of model and library sections **/

/** Sub-section 7 of model and library sections:
  the system section **/

system:
  | SYSTEM system_body { $2 }
;

system_body:
  | OPEN_BRACE
      block_name location bindings blocks lines annotations
    CLOSE_BRACE
    {
      {
        system_name = $2;
        system_position = $3;
        system_bindings = $4;
        system_blocks = $5;
        system_lines = $6;
        system_annotations = $7;
        system_connectors = [];
      }
    }
;

block_name:
  | NAME name { $2 }
  | NAME IDENT { mdl_name_of_string $2 }
;

location:
  | LOCATION location_value { $2 }
;

location_value:
/* A vector of 2 points */
  | int_vector {
    let loc = Location.get_symbol_loc () in
    make_mdl_location loc $1 }
;

interface_block:
  BLOCK OPEN_BRACE interface_block_body CLOSE_BRACE
  { $3 }
;

interface_block_body:
  | block_type
    block_source_type_opt
    block_source_path_opt
    block_ports_description
    interface_block_parameters
    { make_interface_block $1 $2 $3 $4 $5 }
;

block_source_type_opt:
  | { None }
  | block_source_type { Some $1 }
;
block_source_path_opt:
  | { None }
  | block_source_path { Some $1 }
;

interface_block_parameters:
  | BLOCKINTERFACEPARAMETERS OPEN_BRACE interface_block_params CLOSE_BRACE
    { $3 }
;

interface_block_params:
  | { [] }
  | interface_block_param interface_block_params { $1 :: $2 }
;

interface_block_param:
  | interface_binding_ident COLON ident
    { make_interface_block_parameter $1 $3 }
;

interface_binding_ident:
  | binding_ident { $1 }
  | PORT { port_ident }
;

blocks:
  | { [] }
  | block blocks { $1 :: $2 }
;

block:
  | BLOCK OPEN_BRACE block_items CLOSE_BRACE
    { Block $3 }
  | VENDOR DOT configuration_longident vendor_items
    { Block_vendor (make_block_vendor $1 $3 $4) }
;

block_items:
  | { [] }
  | block_item block_items { $1 :: $2 }
;

block_item:
  | block_type { Block_type $1 }
  | block_name { Block_name $1 }
  | block_source_path { Block_source_path $1 }
  | block_source_type { Block_source_type $1 }
  | block_port { $1 }
  | block_ports_description { Block_ports_description $1 }
  | block_position { Block_position $1 }
  | block_tag { Block_tag $1 }
  | block_sid { Block_sid $1 }
  | block_state { Block_state $1 }
  | block_sf_block_type { Block_sf_block_type $1 }
 /* FIXME: Is this one a bug?
     Found in only one example as an array of size 0! */
  | array_description { Block_array $1 }
  | object_description { Block_object $1 }
/* FIXME: is this useless ?
 | SYSTEM simple_binding_value
    { Block_binding (make_system_binding (Value_simple $2)) }
*/
  | system { Block_system $1 }
  | dimension_simple_value
    { Block_binding (make_dimension_binding (Value_simple $1)) }
  | port_number_value
    { Block_binding (make_port_number_binding $1) }
  | parameter_default_binding { Block_binding $1 }
  | block_graphical_default { Block_graphical_default $1 }
  | block_mask_item { Block_mask_item $1 }
/* TODO:
  replaced by next rule. Can we remove that comment?
  | vendor_configuration_longident vendor_items */
  | VENDOR DOT configuration_longident vendor_items
    { Block_vendor_binding (make_vendor_binding $1 $3 $4) }
;

block_mask_item:
  | MASKCALLBACKSTRING text
    { Mask_callback_string (Pipe_expressions $2) }
  | MASKCAPABILITIES text
    { Mask_capabilities (Expression $2) }
  | MASKDESCRIPTION text
    { Mask_description $2 }
  | MASKDISPLAY code
    { Mask_display $2 }
  | MASKDISPLAYSTRING text
    { Mask_display_string $2 }
  | MASKENABLESTRING text
    { Mask_enable_string (Comma_bools $2) }
  | MASKHELP text
    { Mask_help $2 }
/* FIME: remove or explain why this is useless
  | MASKHIDECONTENTS mdl_bool { Mask_hide_contents $2 }
 */
  | MASKICONFRAME mdl_bool { Mask_icon_frame $2 }
  | MASKICONOPAQUE mdl_bool { Mask_icon_opaque $2 }
  | MASKICONROTATE ident_in_string { Mask_icon_rotate $2 }
  | MASKICONUNITS ident_in_string { Mask_icon_units $2 }
  | MASKINITIALIZATION STRING
    { Mask_initialization (mdl_shallow_code_of_string $2) }
  | MASKPORTROTATE ident_in_string { Mask_port_rotate $2 }
  | MASKPROMPTSTRING text
    { Mask_prompt_string (Pipe_texts $2) }
  | MASKRUNINITFORICONREDRAW mdl_bool
    { Mask_run_init_for_icon_redraw $2 }
  | MASKSELFMODIFIABLE mdl_bool
    { Mask_self_modifiable $2 }
  | MASKSTYLESTRING text
    { Mask_style_string (Comma_texts $2) }
  | MASKTABNAMESTRING text
    { Mask_tab_name_string $2 }
  | MASKTOOLTIPSTRING text
    { Mask_tooltip_string (Comma_bools $2) }
  | MASKTUNABLEVALUESTRING text
    { Mask_tunable_value_string (Comma_bools $2) }
  | MASKTYPE text
    { Mask_type $2 }
  | MASKVALUESTRING text
    { Mask_value_string (Pipe_texts $2) }
  | MASKVARALIASSTRING text
    { Mask_var_alias_string $2 }
  | MASKVARIABLES text
    { Mask_variables (Semi_texts $2) }
  | MASKVISIBILITYSTRING text
    { Mask_visibility_string (Comma_bools $2) }
;

block_type:
  | BLOCKTYPE block_type_value { $2 }
;

block_type_value:
  | OUTPORT { Mdl_block_types.Outport }
  | INPORT { Mdl_block_types.Inport }
  | BLOCKTYPENAME { $1 }
  | STRING
    { let loc = Location.get_symbol_loc () in
      make_block_type_from_string_specification loc $1 }
;

block_sf_block_type:
  | block_sf_block_type_string
    { let loc, type_string = $1 in
      make_block_sf_block_type_from_string_specification loc type_string }
;

block_sf_block_type_string:
  | SFBLOCKTYPE STRING { Location.get_rhs_loc 2, $2 }
;

block_sf_block_type_binding:
  | block_sf_block_type_string
    { let loc, type_string = $1 in
      ignore
        (make_block_sf_block_type_from_string_specification loc type_string);
      let model_value = Value_simple (String type_string) in
      Mdl_ast_functions.make_sf_block_type_binding model_value }
;

block_source_path:
  | SOURCEBLOCK STRING { mdl_path_of_string $2 }
;

block_source_type:
   | SOURCETYPE name { $2 }
/* FIXME: should be suppressed ? */
   | SOURCETYPE VENDOR
     { failwith "VENDOR source type!" (*mdl_name_of_vendor $2*) }
/* FIXME: end of should be suppressed */
;

block_port:
  | PORT OPEN_BRACE block_port_body CLOSE_BRACE
    { Block_port $3 }
/* FIXME: explain why this is now useless
  | PORT simple_binding_value
     { Block_binding (make_port_binding (Value_simple $2)) }
*/
;

block_ports_description:
  | PORTS natural_vector
    { $2 }
;

block_position:
  | POSITION position_value
    { $2 }
;

position_value:
  | integer_vector
    { $1 }
;

block_tag:
  | TAG text
    { $2 }
;

block_sid:
  | SID block_sid_value { $2 }
;

block_sid_value:
  | block_sid_compound
    { Sid_compound $1 }
  | STRING
    { Sid_string $1 }
;

block_sid_compound:
  | natural
    { Sid_natural $1 }
  | natural COLON block_sid_compound
    { Sid_ref ($1, $3) }
  | natural COLON COLON block_sid_compound
    { Sid_child ($1, $4) }
;

block_state:
  | LSTATE simple_binding_value
    { $2 }
;

background:
  | BACKGROUNDCOLOR STRING
    { $2 }
;

drop_shadow:
  | DROPSHADOW mdl_bool
    { $2 }
;

foreground:
  | FOREGROUNDCOLOR STRING
    { $2 }
;

mirror:
  | BLOCKMIRROR mdl_bool
    { $2 }
;

rotation:
  | BLOCKROTATION natural
    { $2 }
;

orientation:
  | ORIENTATION STRING
    { $2 }
;

name_placement:
  | NAMEPLACEMENT STRING
    { $2 }
;

show_name:
  | SHOWNAME mdl_bool
    { $2 }
;

hide_automatic_name:
  | HIDEAUTOMATICNAME mdl_bool
    { $2 }
;

horizontal_alignment:
  | HORIZONTALALIGNMENT horizontal_alignment_value
    { $2 }
;

horizontal_alignment_value:
  | STRING
    { let loc = Location.get_symbol_loc () in
      make_horizontal_alignment loc $1 }
;

vertical_alignment:
  | VERTICALALIGNMENT vertical_alignment_value
    { $2 }
;

vertical_alignment_value:
  | STRING
    { let loc = Location.get_symbol_loc () in
      make_vertical_alignment loc $1 }
;

text_as_click_cb:
  | USEDISPLAYTEXTASCLICKCALLBACK mdl_bool { $2 }
;

text_click_fcn:
  | CLICKFCN code { $2 }
;

tex_mode_option:
  | TEXMODE mdl_bool { $2 }
  | TEXMODE STRING {
    let loc = Location.get_symbol_loc () in
    mdl_bool_of_mdl_string loc $2 }
;

interpreter:
  | INTERPRETER STRING { $2 }
;

zorder:
  | ZORDER INT { $2 }
;

port_number_value:
  | PORTNUMBER simple_binding_value { Value_simple $2 }
;

port_name:
  | { None }
  | block_name { Some $1 }
;

block_port_body:
  | PORTNUMBER port_number port_name block_port_bindings
    {
      { block_port_number = $2;
        block_port_name = $3;
        block_port_bindings = $4; }
    }
;

block_port_bindings:
  | { [] }
  | block_port_binding block_port_bindings
    { $1 :: $2 }
;

block_port_binding:
  | binding_ident simple_binding_value
    { Block_port_binding (make_binding $1 (Value_simple $2)) }
  | vendor_configuration_longident vendor_items
    {
      let vendor_ident, configuration_longident = $1 in
      let vendor_binding =
        make_vendor_binding vendor_ident configuration_longident $2 in
      Block_port_binding_vendor vendor_binding
    }
  | array_description { Block_port_binding_array $1 }
;

natural:
  | INT
    { let loc = Location.get_symbol_loc () in
      make_natural "a natural" loc $1 }
;
integer:
  | INT
    { let loc = Location.get_symbol_loc () in
      make_int "an integer" loc $1 }
;
port_number:
  | integer { Ast_types.mdl_port_number_of_int $1 }
;

text:
  | STRING
    { Ast_types.mdl_text_of_string $1 }
;

code:
  | STRING
    { Ast_types.mdl_code_of_string $1 }
;

name:
  | STRING
    { Ast_types.mdl_name_of_string $1 }
;

natural_or_string:
  | INT
    { let loc = Location.get_symbol_loc () in
      ignore (make_natural "a natural" loc $1);
      Int $1 }
  | STRING
    { let loc = Location.get_symbol_loc () in
      ignore
        (natural_of_mdl_string "a string natural" loc $1);
      String $1 }
;

bool_or_string_value:
  | BOOL
    { Value_simple (Bool $1) }
  | STRING
    { let loc = Location.get_symbol_loc () in
      ignore (mdl_bool_of_mdl_string loc $1);
      Value_simple (String $1) }
;

lines:
  | { [] }
  | line lines { $1 :: $2 }
;

line:
  | LINE OPEN_BRACE line_items CLOSE_BRACE { Line $3 }
;

line_items:
  | line_item { [ $1 ] }
  | line_item line_items { $1 :: $2 }
;

line_item:
  | block_name { Line_name $1 }
  | line_type { Line_type $1 }
  | labels { Line_labels $1 }
  | line_source { Line_source $1 }
  | line_target { Line_target $1 }
  | points { Line_points $1 }
  | line_graphical_default { Line_graphical_default $1 }
;

line_type:
  | LINETYPE line_type_value
    { $2 }
;

line_type_value:
  | STRING
    { let loc = Location.get_symbol_loc () in
      make_line_type loc $1 }
;

font_item:
  | FONTANGLE STRING { Font_angle $2 }
    /* Default is "normal" */
  | FONTNAME STRING { Font_name $2 }
    /* Default is "Helvetica" */
  | FONTSIZE natural { Font_size $2 }
    /* Default could be 10 */
  | FONTWEIGHT STRING { Font_weight $2 }
    /* Default is "normal" */
;

labels:
  | LABELS integer_matrix { $2 }
;

integer_matrix:
  | OPEN_BRACKET simple_binding_value_array CLOSE_BRACKET
    { let loc = Location.get_symbol_loc () in
      make_integer_matrix loc $2 }
  | STRING
    { let loc = Location.get_symbol_loc () in
      make_integer_matrix_from_string loc $1 }
;

line_source:
  | SRCBLOCK name SRCPORT mdl_port_value
    { Line_source_connector (make_line_connector $2 $4) }
;

common_port_value:
  | LTRIGGER { Port_trigger }
  | LIFACTION { Port_ifaction }
  | LENABLE { Port_enable }
  | LSTATE { Port_state }
;

mdl_port_value:
  | port_number { Port_number $1 }
  | ident
    { let loc = Location.get_symbol_loc () in
      connection_of_ident loc $1 }
  | common_port_value { $1 }
;

line_target:
  | line_target_connector { $1 }
  | line_target_branch { $1 }
;

line_target_connector:
  | DSTBLOCK name DSTPORT mdl_port_value
    { Line_target_connector (make_line_connector $2 $4) }
;

line_target_branch:
  | BRANCH OPEN_BRACE line_target_branch_items CLOSE_BRACE
    { Line_target_branch $3 }
;

line_target_branch_items:
  | { [] }
  | line_target_branch_item line_target_branch_items { $1 :: $2 }
;

line_target_branch_item:
  | connection_type { Line_target_branch_connection_type $1 }
  | labels { Line_target_branch_labels $1 }
  | points { Line_target_branch_points $1 }
  | line_source { Line_target_branch_source $1 }
  | line_target { Line_target_branch_target $1 }
  | zorder { Line_target_branch_zorder $1 }
;

connection_type:
  | CONNECTTYPE connection_type_value
    { $2 }
;

connection_type_value:
  | STRING
    { let loc = Location.get_symbol_loc () in
      make_connection_type loc $1 }
;

points:
  | POINTS integer_matrix
    { $2 }
;

annotations:
  | { [] }
  | annotation annotations { $1 :: $2 }
;

annotation:
  | ANNOTATION OPEN_BRACE annotation_items CLOSE_BRACE
    { Annotation $3 }
;

annotation_items:
  | { [] }
  | annotation_item annotation_items { $1 :: $2 }
;

annotation_name:
  | NAME IDENT { mdl_text_of_string $2 }
  | NAME text { $2 }
  | TEXT text { $2 }
;

annotation_item:
  | annotation_name { Annotation_name $1 }
  | block_position { Annotation_position $1 }
  | block_sid { Annotation_sid $1 }
/*
  | annotation_tag { Annotation_tag $1 }
*/
  | zorder { Annotation_zorder $1 }
  | annotation_graphical_default { Annotation_graphical_default $1 }
  | annotation_auto_size { Annotation_auto_size $1 }
  | annotation_word_wrap { Annotation_word_wrap $1 }
  | annotation_binding { Annotation_binding $1 }
;
/* End of Sub-section 7 of model and library sections */

/** The trailor of model and library descriptions:
  the matdata or stateflow section list. **/

model_trailor_section:
  | EOF { [] }
  | model_trailor_section_item model_trailor_section
    { ($1 :: $2) }
;

model_trailor_section_item:
  | matdata_trailor_section_item { $1 }
  | stateflow_trailor_section_item { $1 }
;

/* The matdata sub-section */
matdata_trailor_section_item:
  | MATDATA OPEN_BRACE matdata_section_item CLOSE_BRACE
    { Implementation_matdata (Matdata_section $3) }
;

matdata_section_item:
  NUMRECORDS INT data_records
  { let loc = Location.get_symbol_loc () in
    make_matdata_section_item loc $2 $3 }
;

data_records:
  | { [] }
  | data_record data_records { $1 :: $2 }
;

data_record:
  DATARECORD OPEN_BRACE TAG ident DATA text CLOSE_BRACE
  { make_data_record $4 $6 }
;
/* End of The matdata sub-section */

/* The stateflow sub-section */
stateflow_trailor_section_item:
  | STATEFLOW OPEN_BRACE stateflow_section_items CLOSE_BRACE
    { Implementation_stateflow (Mdl_parsing_ast.Mdl_stateflow_section $3) }
;

stateflow_section_items:
  | { [] }
  | stateflow_section_item stateflow_section_items { $1 :: $2 }
;

stateflow_section_item:
  | LCHART OPEN_BRACE stateflow_items CLOSE_BRACE
    { Mdl_parsing_ast.Mdl_stateflow_chart $3 }
  | LDATA OPEN_BRACE stateflow_items CLOSE_BRACE
    { Mdl_parsing_ast.Mdl_stateflow_data $3 }
  | LEVENT OPEN_BRACE stateflow_items CLOSE_BRACE
    { Mdl_parsing_ast.Mdl_stateflow_event $3 }
  | LINSTANCE OPEN_BRACE stateflow_items CLOSE_BRACE
    { Mdl_parsing_ast.Mdl_stateflow_instance $3 }
  | LJUNCTION OPEN_BRACE stateflow_items CLOSE_BRACE
    { Mdl_parsing_ast.Mdl_stateflow_junction $3 }
  | LMACHINE OPEN_BRACE stateflow_items CLOSE_BRACE
    { Mdl_parsing_ast.Mdl_stateflow_machine $3 }
  | LSTATE OPEN_BRACE stateflow_items CLOSE_BRACE
    { Mdl_parsing_ast.Mdl_stateflow_state $3 }
  | LTARGET OPEN_BRACE stateflow_items CLOSE_BRACE
    { Mdl_parsing_ast.Mdl_stateflow_target $3 }
  | LTRANSITION OPEN_BRACE stateflow_items CLOSE_BRACE
    { Mdl_parsing_ast.Mdl_stateflow_transition $3 }
;

stateflow_items:
  | { [] }
  | stateflow_item stateflow_items { $1 :: $2 }
;

sf_binding_value:
  | binding_value { $1 }
;

sf_binding:
 | sf_binding_ident sf_binding_value
   { make_binding $1 $2 }
;

link_node:
  | OPEN_BRACKET natural natural natural CLOSE_BRACKET
    {
      {
        Mdl_parsing_ast.
        link_node_parent_ref = $2;
        link_node_prev_ref = $3;
        link_node_next_ref = $4;
      }
    }
;

tree_node:
  | OPEN_BRACKET natural natural natural natural CLOSE_BRACKET
    {
      {
        Mdl_parsing_ast.
        tree_node_parent_ref = $2;
        tree_node_first_child_ref = $3;
        tree_node_prev_ref = $4;
        tree_node_next_ref = $5;
      }
    }
;

stateflow_item:
  | sf_binding
    { Mdl_parsing_ast.Mdl_stateflow_binding $1 }
  | LCHART natural
    { Mdl_parsing_ast.Mdl_stateflow_chart_ref $2 }
  | LEML OPEN_BRACE stateflow_eml_items CLOSE_BRACE
    { Mdl_parsing_ast.Mdl_stateflow_eml $3 }
  | LEXECUTIONORDER natural
    { Mdl_parsing_ast.Mdl_stateflow_execution_order $2 }
  | LFIRSTDATA natural
    { Mdl_parsing_ast.Mdl_stateflow_first_data $2 }
  | LFIRSTEVENT natural
    { Mdl_parsing_ast.Mdl_stateflow_first_event $2 }
  | LFIRSTJUNCTION natural
    { Mdl_parsing_ast.Mdl_stateflow_first_junction $2 }
  | LFIRSTTRANSITION natural
    { Mdl_parsing_ast.Mdl_stateflow_first_transition $2 }
  | LFIRSTTARGET natural
    { Mdl_parsing_ast.Mdl_stateflow_first_target $2 }
  | LID natural
    { Mdl_parsing_ast.Mdl_stateflow_id $2 }
  | LLABELSTRING text
    { Mdl_parsing_ast.Mdl_stateflow_label_string $2 }
  | LLINKNODE link_node
    { Mdl_parsing_ast.Mdl_stateflow_link_node $2 }
  | LTREENODE tree_node
    { Mdl_parsing_ast.Mdl_stateflow_tree_node $2 }
  | LMACHINE natural
    { Mdl_parsing_ast.Mdl_stateflow_machine_ref $2 }
  | LCHARTFILENUMBER natural
    { Mdl_parsing_ast.Mdl_stateflow_chart_file_number $2 }
  | LNAME name
    { Mdl_parsing_ast.Mdl_stateflow_name $2 }
  | LPROPS OPEN_BRACE stateflow_props_items CLOSE_BRACE
    { Mdl_parsing_ast.Mdl_stateflow_props $3 }
  | LSCOPE ident
    { Mdl_parsing_ast.Mdl_stateflow_scope $2 }
  | LSSIDNUMBER natural
    { Mdl_parsing_ast.Mdl_stateflow_ssid_number $2 }
  | LTRIGGER ident
    { Mdl_parsing_ast.Mdl_stateflow_trigger $2 }
  | LTYPE ident
    { Mdl_parsing_ast.Mdl_stateflow_type $2 }
  | LSRC OPEN_BRACE stateflow_src_dst_items CLOSE_BRACE
    { Mdl_parsing_ast.Mdl_stateflow_src $3 }
  | LDST OPEN_BRACE stateflow_src_dst_items CLOSE_BRACE
    { Mdl_parsing_ast.Mdl_stateflow_dst $3 }
  | LPLANTMODELINGINFO OPEN_BRACE CLOSE_BRACE
    { Mdl_parsing_ast.Mdl_stateflow_plant_modeling_info }
;

/* Items for src/dst stateflow link/tree nodes */
stateflow_src_dst_items:
  | stateflow_src_dst_item { [ $1 ] }
  | stateflow_src_dst_item stateflow_src_dst_items { $1 :: $2 }
;

stateflow_src_dst_item:
  | LID natural
    { Mdl_ast.Stateflow_src_dst_ref $2 }
  | LINTERSECTION intersection
    { Mdl_ast.Stateflow_intersection $2 }
;

intersection:
  | OPEN_BRACKET stateflow_numbers CLOSE_BRACKET
    { $2 }
;

stateflow_number:
  | INT { Mdl_ast.Stateflow_int $1 }
  | FLOAT { Mdl_ast.Stateflow_float $1 }
;

stateflow_numbers:
  | { [ ] }
  | stateflow_number stateflow_numbers
    { $1 :: $2 }
;

/* Items for src/dst stateflow eml */
stateflow_eml_items:
  | stateflow_eml_item { [ $1 ] }
  | stateflow_eml_item stateflow_eml_items { $1 :: $2 }
;

stateflow_eml_item:
  | LNAME name
    { Mdl_ast.Stateflow_eml_name $2 }
  | LSCRIPT code
    { Mdl_ast.Stateflow_eml_script $2 }
  | sf_binding
    { Mdl_ast.Stateflow_eml_binding $1 }
;

/* Items for src/dst stateflow props */
stateflow_props_items:
  | stateflow_props_item { [ $1 ] }
  | stateflow_props_item stateflow_props_items { $1 :: $2 }
;

stateflow_props_item:
  | LARRAY OPEN_BRACE stateflow_props_array_items CLOSE_BRACE
    { Mdl_ast.Stateflow_props_array $3 }
  | LTYPE OPEN_BRACE stateflow_props_type_items CLOSE_BRACE
    { Mdl_ast.Stateflow_props_type $3 }
  | sf_binding
    { Mdl_ast.Stateflow_props_binding $1 }
;

stateflow_props_array_items:
  | stateflow_props_array_item { [ $1 ] }
  | stateflow_props_array_item stateflow_props_array_items { $1 :: $2 }
;

stateflow_props_array_item:
  | LSIZE text
    { Mdl_ast.Stateflow_props_array_item_size $2}
  | LFIRSTINDEX text
    { Mdl_ast.Stateflow_props_array_item_first_index $2}
  | sf_binding
    { Mdl_ast.Stateflow_props_array_item_binding $1 }
;

stateflow_props_type_items:
  | stateflow_props_type_item { [ $1 ] }
  | stateflow_props_type_item stateflow_props_type_items { $1 :: $2 }
;

stateflow_props_type_item:
/* | LMETHOD ident */
  | sf_binding { $1 }
;

/* End of The stateflow sub-section */

/** End of The trailor of model and library descriptions **/

/*

 Local Variables:
  compile-command: "cd ../../..; make"
  End:

*/
