(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mdl_parsing_ast_to_mdl_ast.ml,v 1.1 2019-05-10 14:04:02 jpc Exp $ *)

open Mdl_ast;;
open Mdl_parsing_ast;;

let cautious_hashtbl_find tbl key message =
  if Hashtbl.mem tbl key then
    Hashtbl.find tbl key else
  Say.fatal_error (Printf.sprintf
    "Key was not found in table: %s"
    message)
;;

type 'b stateflow_table =
  (Mdl_ast.stateflow_ref, 'b) Hash_table.t
(** A hash table to record stateflow_items for a given stateflow reference
  (an id natural number). *)
;;

type classified_machine =
  {
    machine_id : stateflow_id;
    machine_first_data_ref_opt : stateflow_ref option;
    machine_first_event_ref_opt : stateflow_ref option;
    machine_first_target_ref_opt : stateflow_ref option;
    machine_stateflow_items : Mdl_ast.stateflow_items;
  }
;;

type classified_instance =
  {
    instance_id : stateflow_id;
    instance_machine_ref : stateflow_ref;
    instance_chart_ref : stateflow_ref;
    instance_stateflow_items : stateflow_items;
  }
;;

type classified_chart =
  {
    chart_id : stateflow_id;
    chart_machine_ref : stateflow_ref;
    chart_tree_node_opt : stateflow_tree_node option;
    chart_first_data_ref_opt : stateflow_ref option;
    chart_first_event_ref_opt : stateflow_ref option;
    chart_first_junction_ref_opt : stateflow_ref option;
    chart_first_transition_ref_opt : stateflow_ref option;
    chart_stateflow_items : stateflow_items;
  }

and classified_charts = classified_chart stateflow_table
;;

type classified_state =
  {
    state_id : stateflow_id;
    state_chart_ref : stateflow_ref;
    state_tree_node : stateflow_tree_node;
    state_first_data_ref_opt : stateflow_ref option;
    state_first_event_ref_opt : stateflow_ref option;
    state_first_junction_ref_opt : stateflow_ref option;
    state_first_transition_ref_opt : stateflow_ref option;
    state_stateflow_items : stateflow_items;
  }
;;

type classified_data =
  {
    data_id : stateflow_id;
    data_machine_ref : stateflow_ref;
    data_link_node : stateflow_link_node;
    data_stateflow_items : stateflow_items;
  }

and classified_event =
  {
    event_id : stateflow_id;
    event_machine_ref : stateflow_ref;
    event_link_node : stateflow_link_node;
    event_stateflow_items : stateflow_items;
  }

and classified_junction =
  {
    junction_id : stateflow_id;
    junction_chart_ref : stateflow_ref;
    junction_link_node : stateflow_link_node;
    junction_stateflow_items : stateflow_items;
  }

and classified_transition =
  {
    transition_id : stateflow_id;
    transition_chart_ref : stateflow_ref;
    transition_link_node : stateflow_link_node;
    transition_stateflow_items : stateflow_items;
  }

and classified_target =
  {
    target_id : stateflow_id;
    target_machine_ref : stateflow_ref;
    target_link_node : stateflow_link_node;
    target_stateflow_items : stateflow_items;
  }
;;

type stateflow_tables =
  {
    machines : classified_machine stateflow_table;
    instances : classified_instance stateflow_table;
    charts_by_machine_ref : classified_charts stateflow_table;
    datas : classified_data stateflow_table;
    events : classified_event stateflow_table;
    states : classified_state stateflow_table;
    junctions : classified_junction stateflow_table;
    transitions : classified_transition stateflow_table;
    targets : classified_target stateflow_table;
  }
(** Tables to record stateflow items for stateflow entities. *)
;;

type container_kind =
  | Container_machine
  | Container_instance
  | Container_chart
  | Container_data
  | Container_event
  | Container_state
  | Container_junction
  | Container_transition
  | Container_target
;;

type stateflow_item_key =
  | Key_id
  | Key_machine_ref
  | Key_chart_ref
  | Key_link_node
  | Key_tree_node
  | Key_first_data_ref
  | Key_first_event_ref
  | Key_first_junction_ref
  | Key_first_transition_ref
  | Key_first_target_ref
(** The key to retrieve information of interest in hash tables. *)
;;

type classified_mdl_stateflow_items =
  {
    key_id_value : stateflow_id option;
    key_machine_ref_value : stateflow_machine_ref option;
    key_chart_ref_value : stateflow_chart_ref option;
    key_link_node_value : Mdl_parsing_ast.stateflow_link_node option;
    key_tree_node_value : Mdl_parsing_ast.stateflow_tree_node option;
    key_first_data_ref_value : stateflow_data_ref option;
    key_first_event_ref_value : stateflow_event_ref option;
    key_first_junction_ref_value : stateflow_junction_ref option;
    key_first_transition_ref_value : stateflow_transition_ref option;
    key_first_target_ref_value : stateflow_target_ref option;
  }
(** The stateflow items classified in a record, to handle their various
  types and easily retrieve their value using a label. *)
;;

let fatal_error s =
  Say.fatal_error (Printf.sprintf "stateflow tree building: %s" s)
;;

let sprint_natural () natural =
  Printf.sprintf "%d" (Unatural.to_int natural)
;;

let string_of_container_kind = function
  | Container_machine -> "machine"
  | Container_instance -> "instance"
  | Container_chart -> "chart"
  | Container_data -> "data"
  | Container_event -> "event"
  | Container_state -> "state"
  | Container_junction -> "junction"
  | Container_transition -> "transition"
  | Container_target -> "target"
;;

let string_of_stateflow_item_key = function
  | Key_id -> "id"
  | Key_machine_ref -> "machine"
  | Key_chart_ref -> "chart"
  | Key_link_node -> "linkNode"
  | Key_tree_node -> "treeNode"
  | Key_first_data_ref -> "firstData"
  | Key_first_event_ref -> "firstEvent"
  | Key_first_junction_ref -> "firstJunction"
  | Key_first_transition_ref -> "firstTransition"
  | Key_first_target_ref -> "firstTarget"
;;

let set_option container_kind stateflow_item_key v = function
  | None -> Some v
  | Some _ ->
    fatal_error (Printf.sprintf "%s was set twice in %s"
      (string_of_stateflow_item_key stateflow_item_key)
      (string_of_container_kind container_kind))
;;

let empty_classified_stateflow_items =
  {
    key_id_value = None;
    key_machine_ref_value = None;
    key_chart_ref_value = None;
    key_link_node_value = None;
    key_tree_node_value = None;
    key_first_data_ref_value = None;
    key_first_event_ref_value = None;
    key_first_junction_ref_value = None;
    key_first_transition_ref_value = None;
    key_first_target_ref_value = None;
  }
(** The empty record of classified items. *)
;;

let create_stateflow_tables () =
  {
    machines = Hash_table.create Unatural._1_;
    instances = Hash_table.create Unatural._3_;
    charts_by_machine_ref = Hash_table.create (Unatural.of_int 10);
    datas = Hash_table.create (Unatural.of_int 71);
    events = Hash_table.create (Unatural.of_int 10);
    states = Hash_table.create (Unatural.of_int 123);
    junctions = Hash_table.create (Unatural.of_int 10);
    transitions = Hash_table.create (Unatural.of_int 10);
    targets = Hash_table.create (Unatural.of_int 10);
  }
(** A record of hash tables for each stateflow node kind
  (a stateflow section in MDL). *)
;;

let spurious_key_error container_kind container_id_opt key =
  let container_id_string =
    match container_id_opt with
    | None -> "unknown"
    | Some container_id -> sprint_natural () container_id in
  fatal_error (Printf.sprintf "spurious key %s in %s (id: %s)"
    (string_of_stateflow_item_key key)
    (string_of_container_kind container_kind)
    container_id_string)
;;

let classify_mdl_stateflow_items container_kind mdl_stateflow_items =

  let set_option x = set_option container_kind x in

  let rec loop classified_stateflow_items accu = function
    | [] -> classified_stateflow_items, accu
    | mdl_stateflow_item :: mdl_stateflow_items ->
      let classified_stateflow_items, accu =
        loop classified_stateflow_items accu mdl_stateflow_items in
      match mdl_stateflow_item with

      (* Stateflow items that disappear from the resulting MDL AST tree. *)
      | Mdl_stateflow_chart_ref stateflow_chart_ref ->
        let opt =
          set_option Key_chart_ref stateflow_chart_ref
            classified_stateflow_items.key_chart_ref_value in
        {
          classified_stateflow_items with
          key_chart_ref_value = opt;
        },
        accu
      | Mdl_stateflow_id stateflow_id ->
        let opt =
          set_option Key_id stateflow_id
            classified_stateflow_items.key_id_value in
        {
          classified_stateflow_items with
          key_id_value = opt;
        },
        accu
      | Mdl_stateflow_machine_ref stateflow_machine_ref ->
        let opt =
          set_option Key_machine_ref stateflow_machine_ref
            classified_stateflow_items.key_machine_ref_value in
        {
          classified_stateflow_items with
          key_machine_ref_value = opt;
        },
        accu
      | Mdl_stateflow_link_node stateflow_link_node ->
        let opt =
          set_option Key_link_node stateflow_link_node
            classified_stateflow_items.key_link_node_value in
        {
          classified_stateflow_items with
          key_link_node_value = opt;
        },
        accu
      | Mdl_stateflow_tree_node stateflow_tree_node ->
        let opt =
          set_option Key_tree_node stateflow_tree_node
            classified_stateflow_items.key_tree_node_value in
        {
          classified_stateflow_items with
          key_tree_node_value = opt;
        },
        accu
      | Mdl_stateflow_first_data stateflow_first_data_ref ->
        let opt =
          set_option Key_first_data_ref stateflow_first_data_ref
            classified_stateflow_items.key_first_data_ref_value in
        {
          classified_stateflow_items with
          key_first_data_ref_value = opt;
        },
        accu
      | Mdl_stateflow_first_event stateflow_first_event_ref ->
        let opt =
          set_option Key_first_event_ref stateflow_first_event_ref
            classified_stateflow_items.key_first_event_ref_value in
        {
          classified_stateflow_items with
          key_first_event_ref_value = opt;
        },
        accu
      | Mdl_stateflow_first_junction stateflow_first_junction_ref ->
        let opt =
          set_option Key_first_junction_ref stateflow_first_junction_ref
            classified_stateflow_items.key_first_junction_ref_value in
        {
          classified_stateflow_items with
          key_first_junction_ref_value = opt;
        },
        accu
      | Mdl_stateflow_first_transition stateflow_first_transition_ref ->
        let opt =
          set_option Key_first_transition_ref stateflow_first_transition_ref
            classified_stateflow_items.key_first_transition_ref_value in
        {
          classified_stateflow_items with
          key_first_transition_ref_value = opt;
        },
        accu
      | Mdl_stateflow_first_target stateflow_first_target_ref ->
        let opt =
          set_option Key_first_target_ref stateflow_first_target_ref
            classified_stateflow_items.key_first_target_ref_value in
        {
          classified_stateflow_items with
          key_first_target_ref_value = opt;
        },
        accu

      (* Stateflow items that persist in the MDL AST tree. *)
      | Mdl_stateflow_src stateflow_src ->
        classified_stateflow_items,
        Stateflow_src stateflow_src :: accu
      | Mdl_stateflow_dst stateflow_dst ->
        classified_stateflow_items,
        Stateflow_dst stateflow_dst :: accu
      | Mdl_stateflow_eml stateflow_eml ->
        classified_stateflow_items,
        Stateflow_eml stateflow_eml :: accu
      | Mdl_stateflow_binding binding ->
        classified_stateflow_items,
        Stateflow_binding binding :: accu
      | Mdl_stateflow_execution_order execution_order ->
        classified_stateflow_items,
        Stateflow_execution_order execution_order :: accu
      | Mdl_stateflow_label_string label_string ->
        classified_stateflow_items,
        Stateflow_label_string label_string :: accu
      | Mdl_stateflow_chart_file_number chart_file_number ->
        classified_stateflow_items,
        Stateflow_chart_file_number chart_file_number :: accu
      | Mdl_stateflow_name name ->
        classified_stateflow_items,
        Stateflow_name name :: accu
      | Mdl_stateflow_scope scope ->
        classified_stateflow_items,
        Stateflow_scope scope :: accu
      | Mdl_parsing_ast.Mdl_stateflow_props stateflow_props_items ->
        classified_stateflow_items,
        Stateflow_props stateflow_props_items :: accu
      | Mdl_stateflow_ssid_number ssid_number ->
        classified_stateflow_items,
        Stateflow_ssid_number ssid_number :: accu
      | Mdl_stateflow_trigger trigger ->
        classified_stateflow_items,
        Stateflow_trigger trigger :: accu
      | Mdl_stateflow_type ty ->
        classified_stateflow_items,
        Stateflow_type ty :: accu
      | Mdl_stateflow_plant_modeling_info ->
        classified_stateflow_items,
        Stateflow_plant_modeling_info :: accu in
  loop empty_classified_stateflow_items [] mdl_stateflow_items
;;

let machine_specs,
    instance_specs,
    chart_specs,
    data_specs,
    state_specs,
    event_specs,
    junction_specs,
    transition_specs,
    target_specs =
  (Container_machine,
   [ Key_id ],
   [
     Key_first_data_ref;
     Key_first_event_ref;
     Key_first_target_ref;
   ]),
  (Container_instance,
   [ Key_id; Key_machine_ref; Key_chart_ref; ],
   [ ]),
  (Container_chart,
   [ Key_id; Key_machine_ref; ],
   [
     Key_tree_node;
     Key_first_data_ref;
     Key_first_event_ref;
     Key_first_junction_ref;
     Key_first_transition_ref;
   ]),
  (Container_data,
   [ Key_id; Key_machine_ref; Key_link_node; ],
   [ ]),
  (Container_state,
   [ Key_id; Key_chart_ref; Key_tree_node; ],
   [
     Key_first_data_ref;
     Key_first_event_ref;
     Key_first_junction_ref;
     Key_first_transition_ref;
   ]),
  (Container_event,
   [ Key_id; Key_machine_ref; Key_link_node; ],
   []),
  (Container_junction,
   [ Key_id; Key_chart_ref; Key_link_node; ],
   [ ]),
  (Container_transition,
   [ Key_id; Key_chart_ref; Key_link_node; ],
   [ ]),
  (Container_target,
   [ Key_id; Key_machine_ref; Key_link_node; ],
   [ ])
;;

let check_classified_mdl_stateflow_items container_specs = function
  | {
      key_id_value = id_opt;
      key_machine_ref_value = machine_ref_opt;
      key_chart_ref_value = chart_ref_opt;
      key_link_node_value = link_node_opt;
      key_tree_node_value = tree_node_opt;
      key_first_data_ref_value = first_data_ref_opt;
      key_first_event_ref_value = first_event_ref_opt;
      key_first_junction_ref_value = first_junction_ref_opt;
      key_first_transition_ref_value = first_transition_ref_opt;
      key_first_target_ref_value = first_target_ref_opt;
    } ->
    match container_specs with
    | container_kind, mandatory_keys, optional_keys ->
      let value_set = function
        | None -> false
        | Some _ -> true in
      let htbl = Hashtbl.create 17 in
      let add_key stateflow_item_key key_value =
        Hashtbl.add htbl stateflow_item_key (value_set key_value) in

      let add_keys () =
        add_key Key_id id_opt;
        add_key Key_machine_ref machine_ref_opt;
        add_key Key_chart_ref chart_ref_opt;
        add_key Key_link_node link_node_opt;
        add_key Key_tree_node tree_node_opt;
        add_key Key_first_data_ref first_data_ref_opt;
        add_key Key_first_event_ref first_event_ref_opt;
        add_key Key_first_junction_ref first_junction_ref_opt;
        add_key Key_first_transition_ref first_transition_ref_opt;
        add_key Key_first_target_ref first_target_ref_opt; in

      let remove_key key = Hashtbl.remove htbl key in

      let retrieve_key_set key =
        let key_set =
          let message = string_of_stateflow_item_key key in
          cautious_hashtbl_find htbl key message in
        remove_key key;
        key_set in

      let check_mandatory_key mandatory_key =
        if not (retrieve_key_set mandatory_key) then
          fatal_error (Printf.sprintf
            "key %s not found in %s"
            (string_of_stateflow_item_key mandatory_key)
            (string_of_container_kind container_kind)) in

      let check_unset_keys () =
        Hashtbl.iter
          (fun key v ->
           if v then spurious_key_error container_kind id_opt key)
          htbl in

      add_keys ();
      List.iter check_mandatory_key mandatory_keys;
      List.iter remove_key optional_keys;
      check_unset_keys ();
;;

let classify_machine_items mdl_stateflow_items =
  let classified_mdl_stateflow_items, stateflow_items =
    classify_mdl_stateflow_items Container_machine mdl_stateflow_items in
  check_classified_mdl_stateflow_items
    machine_specs classified_mdl_stateflow_items;
  match classified_mdl_stateflow_items with
  | {
      key_id_value = Some id;
      key_machine_ref_value = _machine_ref_opt;
      key_chart_ref_value = _chart_ref_opt;
      key_link_node_value = _link_node_opt;
      key_tree_node_value = _tree_node_opt;
      key_first_data_ref_value = first_data_ref_opt;
      key_first_event_ref_value = first_event_ref_opt;
      key_first_junction_ref_value = _first_junction_ref_opt;
      key_first_transition_ref_value = _first_transition_ref_opt;
      key_first_target_ref_value = first_target_ref_opt;
    } ->
    {
      machine_id = id;
      machine_first_data_ref_opt = first_data_ref_opt;
      machine_first_event_ref_opt = first_event_ref_opt;
      machine_first_target_ref_opt = first_target_ref_opt;
      machine_stateflow_items = stateflow_items;
    }
  | _ -> assert false
;;

let classify_instance_items mdl_stateflow_items =
  let classified_mdl_stateflow_items, stateflow_items =
    classify_mdl_stateflow_items Container_instance mdl_stateflow_items in
  check_classified_mdl_stateflow_items
    instance_specs classified_mdl_stateflow_items;
  match classified_mdl_stateflow_items with
  | {
      key_id_value = Some id;
      key_machine_ref_value = Some machine_ref;
      key_chart_ref_value = Some chart_ref;
      key_link_node_value = _link_node_opt;
      key_tree_node_value = _tree_node_opt;
      key_first_data_ref_value = _first_data_ref_opt;
      key_first_event_ref_value = _first_event_ref_opt;
      key_first_junction_ref_value = _first_junction_ref_opt;
      key_first_transition_ref_value = _first_transition_ref_opt;
      key_first_target_ref_value = _first_target_ref_opt;
    } ->
    {
      instance_id = id;
      instance_machine_ref = machine_ref;
      instance_chart_ref = chart_ref;
      instance_stateflow_items = stateflow_items;
    }
  | _ -> assert false
;;

let classify_chart_items mdl_stateflow_items =
  let classified_mdl_stateflow_items, stateflow_items =
    classify_mdl_stateflow_items Container_chart mdl_stateflow_items in
  check_classified_mdl_stateflow_items
    chart_specs classified_mdl_stateflow_items;
  match classified_mdl_stateflow_items with
  | {
      key_id_value = Some id;
      key_machine_ref_value = Some machine_ref;
      key_chart_ref_value = _chart_ref_opt;
      key_link_node_value = _link_node_opt;
      key_tree_node_value = tree_node_opt;
      key_first_data_ref_value = first_data_ref_opt;
      key_first_event_ref_value = first_event_ref_opt;
      key_first_junction_ref_value = first_junction_ref_opt;
      key_first_transition_ref_value = first_transition_ref_opt;
      key_first_target_ref_value = _first_target_ref_opt;
    } ->
    {
      chart_id = id;
      chart_machine_ref = machine_ref;
      chart_tree_node_opt = tree_node_opt;
      chart_first_data_ref_opt = first_data_ref_opt;
      chart_first_event_ref_opt = first_event_ref_opt;
      chart_first_junction_ref_opt = first_junction_ref_opt;
      chart_first_transition_ref_opt = first_transition_ref_opt;
      chart_stateflow_items = stateflow_items;
    }
  | _ -> assert false
;;

let classify_data_items mdl_stateflow_items =
  let classified_mdl_stateflow_items, stateflow_items =
    classify_mdl_stateflow_items Container_data mdl_stateflow_items in
  check_classified_mdl_stateflow_items
    data_specs classified_mdl_stateflow_items;
  match classified_mdl_stateflow_items with
  | {
      key_id_value = Some id;
      key_machine_ref_value = Some machine_ref;
      key_chart_ref_value = _chart_ref_opt;
      key_link_node_value = Some link_node;
      key_tree_node_value = _tree_node_opt;
      key_first_data_ref_value = _first_data_ref_opt;
      key_first_event_ref_value = _first_event_ref_opt;
      key_first_junction_ref_value = _first_junction_ref_opt;
      key_first_transition_ref_value = _first_transition_ref_opt;
      key_first_target_ref_value = _first_target_ref_opt;
    } ->
    {
      data_id = id;
      data_machine_ref = machine_ref;
      data_link_node = link_node;
      data_stateflow_items = stateflow_items;
    }
  | _ -> assert false
;;

let classify_event_items mdl_stateflow_items =
  let classified_mdl_stateflow_items, stateflow_items =
    classify_mdl_stateflow_items Container_chart mdl_stateflow_items in
  check_classified_mdl_stateflow_items
    event_specs classified_mdl_stateflow_items;
  match classified_mdl_stateflow_items with
  | {
      key_id_value = Some id;
      key_machine_ref_value = Some machine_ref;
      key_chart_ref_value = _chart_ref_opt;
      key_link_node_value = Some link_node;
      key_tree_node_value = _tree_node_opt;
      key_first_data_ref_value = _first_data_ref_opt;
      key_first_event_ref_value = _first_event_ref_opt;
      key_first_junction_ref_value = _first_junction_ref_opt;
      key_first_transition_ref_value = _first_transition_ref_opt;
      key_first_target_ref_value = _first_target_ref_opt;
    } ->
    {
      event_id = id;
      event_machine_ref = machine_ref;
      event_link_node = link_node;
      event_stateflow_items = stateflow_items;
    }
  | _ -> assert false
;;

let ref_option_of_stateflow_ref stateflow_ref =
  if stateflow_ref = Unatural._0_
  then None
  else Some stateflow_ref
;;

let classify_state_items mdl_stateflow_items =
  let classified_mdl_stateflow_items, stateflow_items =
    classify_mdl_stateflow_items Container_state mdl_stateflow_items in
  check_classified_mdl_stateflow_items
    state_specs classified_mdl_stateflow_items;
  match classified_mdl_stateflow_items with
  | {
      key_id_value = Some id;
      key_machine_ref_value = _machine_ref_opt;
      key_chart_ref_value = Some chart_ref;
      key_link_node_value = _link_node_opt;
      key_tree_node_value = Some tree_node;
      key_first_data_ref_value = first_data_ref_opt;
      key_first_event_ref_value = first_event_ref_opt;
      key_first_junction_ref_value = first_junction_ref_opt;
      key_first_transition_ref_value = first_transition_ref_opt;
      key_first_target_ref_value = _first_target_ref_opt;
    } ->
    {
      state_id = id;
      state_chart_ref = chart_ref;
      state_tree_node = tree_node;
      state_first_data_ref_opt = first_data_ref_opt;
      state_first_event_ref_opt = first_event_ref_opt;
      state_first_junction_ref_opt = first_junction_ref_opt;
      state_first_transition_ref_opt = first_transition_ref_opt;
      state_stateflow_items = stateflow_items;
    }
  | _ -> assert false
;;

let classify_junction_items mdl_stateflow_items =
  let classified_mdl_stateflow_items, stateflow_items =
    classify_mdl_stateflow_items Container_junction mdl_stateflow_items in
  check_classified_mdl_stateflow_items
    junction_specs classified_mdl_stateflow_items;
  match classified_mdl_stateflow_items with
  | {
      key_id_value = Some id;
      key_machine_ref_value = _machine_ref_opt;
      key_chart_ref_value = Some chart_ref;
      key_link_node_value = Some link_node;
      key_tree_node_value = _tree_node_opt;
      key_first_data_ref_value = _first_data_ref_opt;
      key_first_event_ref_value = _first_event_ref_opt;
      key_first_junction_ref_value = _first_junction_ref_opt;
      key_first_transition_ref_value = _first_transition_ref_opt;
      key_first_target_ref_value = _first_target_ref_opt;
    } ->
      {
        junction_id = id;
        junction_chart_ref = chart_ref;
        junction_link_node = link_node;
        junction_stateflow_items = stateflow_items;
      }
  | _ -> assert false
;;

let classify_transition_items mdl_stateflow_items =
  let classified_mdl_stateflow_items, stateflow_items =
    classify_mdl_stateflow_items Container_transition mdl_stateflow_items in
  check_classified_mdl_stateflow_items
    transition_specs classified_mdl_stateflow_items;
  match classified_mdl_stateflow_items with
  | {
      key_id_value = Some id;
      key_machine_ref_value = _machine_ref_opt;
      key_chart_ref_value = Some chart_ref;
      key_link_node_value = Some link_node;
      key_tree_node_value = _tree_node_opt;
      key_first_data_ref_value = _first_data_ref_opt;
      key_first_event_ref_value = _first_event_ref_opt;
      key_first_junction_ref_value = _first_junction_ref_opt;
      key_first_transition_ref_value = _first_transition_ref_opt;
      key_first_target_ref_value = _first_target_ref_opt;
    } ->
    {
      transition_id = id;
      transition_chart_ref = chart_ref;
      transition_link_node = link_node;
      transition_stateflow_items = stateflow_items;
    }
  | _ -> assert false
;;

let classify_target_items mdl_stateflow_items =
  let classified_mdl_stateflow_items, stateflow_items =
    classify_mdl_stateflow_items Container_target mdl_stateflow_items in
  check_classified_mdl_stateflow_items
    target_specs classified_mdl_stateflow_items;
  match classified_mdl_stateflow_items with
  | {
      key_id_value = Some id;
      key_machine_ref_value = Some machine_ref;
      key_chart_ref_value = _chart_ref_opt;
      key_link_node_value = Some link_node;
      key_tree_node_value = _tree_node_opt;
      key_first_data_ref_value = _first_data_ref_opt;
      key_first_event_ref_value = _first_event_ref_opt;
      key_first_junction_ref_value = _first_junction_ref_opt;
      key_first_transition_ref_value = _first_transition_ref_opt;
      key_first_target_ref_value = _first_target_ref_opt;
    } ->
    {
      target_id = id;
      target_machine_ref = machine_ref;
      target_link_node = link_node;
      target_stateflow_items = stateflow_items;
    }
  | _ -> assert false
;;

let register table container_kind id content =
  if Hashtbl.mem table id then
    fatal_error (Printf.sprintf
      "%s with id %a is declared twice in stateflow"
      (string_of_container_kind container_kind)
      sprint_natural id)
  else
    Hashtbl.add table id content
;;

let register_machine stateflow_tables classified_machine =
  register stateflow_tables.machines Container_machine
    classified_machine.machine_id classified_machine
;;

let register_instance stateflow_tables classified_instance =
  register stateflow_tables.instances Container_instance
    classified_instance.instance_id classified_instance
;;

let register_in_sub_table
    tbl parent_ref (item_type, item_id) classified_item =
  let classified_items =
    if Hashtbl.mem tbl parent_ref then
      Hashtbl.find tbl parent_ref
    else
      let classified_items = Hash_table.create Unatural._7_ in
      Hashtbl.add tbl parent_ref classified_items;
      classified_items in
  register classified_items item_type item_id classified_item
;;

let register_chart stateflow_tables classified_chart =
  let tbl = stateflow_tables.charts_by_machine_ref in
  let machine_ref = classified_chart.chart_machine_ref in
  let chart_id = classified_chart.chart_id in
  register_in_sub_table
    tbl machine_ref (Container_chart, chart_id) classified_chart
;;

let register_data stateflow_tables classified_data =
  register stateflow_tables.datas Container_data
    classified_data.data_id classified_data
;;

let register_event stateflow_tables classified_event =
  register stateflow_tables.events Container_event
    classified_event.event_id classified_event
;;

let register_junction stateflow_tables classified_junction =
  register stateflow_tables.junctions Container_junction
    classified_junction.junction_id classified_junction
;;

let register_state stateflow_tables classified_state =
  register stateflow_tables.states Container_state
    classified_state.state_id classified_state
;;

let register_transition stateflow_tables classified_transition =
  register stateflow_tables.transitions Container_transition
    classified_transition.transition_id classified_transition
;;

let register_target stateflow_tables classified_target =
  register stateflow_tables.targets Container_target
    classified_target.target_id classified_target
;;

let make_stateflow_tables stateflow_section_items =

  let stateflow_tables = create_stateflow_tables () in

  let rec loop = function
    | [] -> ()
    | stateflow_section_item :: stateflow_section_items ->
      begin
        match stateflow_section_item with
        | Mdl_stateflow_machine mdl_stateflow_items ->
          let classified_machine =
            classify_machine_items mdl_stateflow_items in
          register_machine stateflow_tables classified_machine
        | Mdl_stateflow_instance mdl_stateflow_items ->
          let classified_instance =
            classify_instance_items mdl_stateflow_items in
          register_instance stateflow_tables classified_instance
        | Mdl_stateflow_chart mdl_stateflow_items ->
          let classified_chart =
            classify_chart_items mdl_stateflow_items in
          register_chart stateflow_tables classified_chart
        | Mdl_stateflow_data mdl_stateflow_items ->
          let classified_data =
            classify_data_items mdl_stateflow_items in
          register_data stateflow_tables classified_data
        | Mdl_stateflow_event mdl_stateflow_items ->
          let classified_event =
            classify_event_items mdl_stateflow_items in
          register_event stateflow_tables classified_event
        | Mdl_stateflow_state mdl_stateflow_items ->
          let classified_state =
            classify_state_items mdl_stateflow_items in
          register_state stateflow_tables classified_state
        | Mdl_stateflow_junction mdl_stateflow_items ->
          let classified_junction =
            classify_junction_items mdl_stateflow_items in
          register_junction stateflow_tables classified_junction
        | Mdl_stateflow_transition mdl_stateflow_items ->
          let classified_transition =
            classify_transition_items mdl_stateflow_items in
          register_transition stateflow_tables classified_transition;
        | Mdl_stateflow_target mdl_stateflow_items ->
          let classified_target =
            classify_target_items mdl_stateflow_items in
          register_target stateflow_tables classified_target
      end;
      loop stateflow_section_items in

  loop stateflow_section_items;
  stateflow_tables
;;

let check_machine_ref stateflow_tables instance_id instance_machine_ref =
  if not (Hashtbl.mem stateflow_tables.machines instance_machine_ref) then
    fatal_error
      (Printf.sprintf "unkown machine %a referenced by instance %a"
      sprint_natural instance_machine_ref
      sprint_natural instance_id)
;;

let check_chart_ref stateflow_tables instance_id
      instance_machine_ref instance_chart_ref =
  let charts_table = stateflow_tables.charts_by_machine_ref in
  if not (Hashtbl.mem charts_table instance_machine_ref) then
    fatal_error (Printf.sprintf
      "missing chart referencing machine %a in instance %a"
      sprint_natural instance_machine_ref
      sprint_natural instance_id);
  let charts = Hashtbl.find charts_table instance_machine_ref in
  if not (Hashtbl.mem charts instance_chart_ref) then
    fatal_error
      (Printf.sprintf "unkown chart %a referenced by instance %a"
      sprint_natural instance_chart_ref
      sprint_natural instance_id)
;;

(* Compute a Caml list from a stateflow encoded list
   (encoding may use stateflow link_node bindings or
   and tree_node bindings). *)
let caml_list_of_stateflow_list
    tbl tree_name get_next_ref item_of_classified_item_node
    first_item_ref_opt =

  match first_item_ref_opt with
  | None -> []
  | Some first_item_ref ->

    let rec classify_item_nodes accu item_ref =
      if item_ref = Unatural._0_ then accu else

      let classified_item =
        let item_message =
          Printf.sprintf "%s reference %s"
            tree_name
            (Unatural.to_string item_ref) in
        cautious_hashtbl_find tbl item_ref item_message in
      classify_item_nodes
        (classified_item :: accu) (get_next_ref classified_item) in

    let classified_item_nodes =

      (* Weird behavior in MDL files:
         in some case, the first node reference found in stateflow section
         is plain wrong: there is no corresponding node in the stateflow
         specification.
         It could presumably be a bug in the graphical interface:
         some nodes of this kind may have been once defined by the user,
         and the first_**_reference been set to the proper node; but then,
         the user removed all the nodes of this kind, including the node
         corresponding to the first_**_reference; then the reference become a
         dandling reference still stored in the simulink file...
      *)
      if not (Hashtbl.mem tbl first_item_ref) then begin
        Say.warning (Printf.sprintf
          "Dandling reference %s to %s node"
          (Unatural.to_string first_item_ref) tree_name);
        [] end else
      classify_item_nodes [] first_item_ref in

    List.rev_map item_of_classified_item_node classified_item_nodes
;;

let caml_list_of_stateflow_link_node tbl list_name get_link_node =
  let get_next_ref classified_item =
    let link_node = get_link_node classified_item in
    link_node.Mdl_parsing_ast.link_node_next_ref in
  caml_list_of_stateflow_list tbl list_name get_next_ref
;;

let caml_list_of_stateflow_tree_node tbl tree_name get_tree_node =
  let get_next_ref classified_item =
    let tree_node = get_tree_node classified_item in
    tree_node.Mdl_parsing_ast.tree_node_first_child_ref in
  caml_list_of_stateflow_list tbl tree_name get_next_ref
;;

(* Find targets *)
let stateflow_target_of_classified_target_node machine_id = function
  | {
      target_id = target_id;
      target_machine_ref = machine_ref;
      target_link_node = _link_node;
      target_stateflow_items = stateflow_items;
    } ->
    if machine_ref <> machine_id then
      Say.fatal_error (Printf.sprintf
         "incompatible reference between machine %a and target %a"
         sprint_natural machine_id
         sprint_natural target_id
      ) else
    {
      stateflow_target_id = target_id;
      stateflow_target_items = stateflow_items;
    }
;;

let find_stateflow_targets
    stateflow_tables machine_id machine_first_target_ref_opt =
  let tbl = stateflow_tables.targets in
  let item_of_classified_item_node =
    stateflow_target_of_classified_target_node machine_id in
  let get_link_node classified_target =
    classified_target.target_link_node in
  caml_list_of_stateflow_link_node
    tbl "target list" get_link_node
    item_of_classified_item_node
    machine_first_target_ref_opt
;;

(* Find datas *)
let stateflow_data_of_classified_data_node machine_id = function
  | {
      data_id = data_id;
      data_machine_ref = machine_ref;
      data_link_node = _link_node;
      data_stateflow_items = stateflow_items;
    } ->
    if machine_ref <> machine_id then
      Say.fatal_error (Printf.sprintf
         "incompatible reference between machine %a and data %a"
         sprint_natural machine_id
         sprint_natural data_id
      ) else
    {
      stateflow_data_id = data_id;
      stateflow_data_items = stateflow_items;
    }
;;

let find_stateflow_datas
    stateflow_tables machine_id machine_first_data_ref_opt =
  let tbl = stateflow_tables.datas in
  let item_of_classified_item_node =
    stateflow_data_of_classified_data_node machine_id in
  let get_link_node classified_data =
    classified_data.data_link_node in
  caml_list_of_stateflow_link_node
    tbl "data list" get_link_node item_of_classified_item_node
    machine_first_data_ref_opt
;;

(* Find events *)
let stateflow_event_of_classified_event_node machine_id = function
  | {
      event_id = event_id;
      event_machine_ref = machine_ref;
      event_link_node = _link_node;
      event_stateflow_items = stateflow_items;
    } ->
    if machine_ref <> machine_id then
      Say.fatal_error (Printf.sprintf
         "incompatible reference between machine %a and event %a"
         sprint_natural machine_id
         sprint_natural event_id
      ) else
    {
      stateflow_event_id = event_id;
      stateflow_event_items = stateflow_items;
    }
;;

let find_stateflow_events
    stateflow_tables machine_id first_event_ref_opt =
  let tbl = stateflow_tables.events in
  let item_of_classified_item_node =
    stateflow_event_of_classified_event_node machine_id in
  let get_link_node classified_event =
    classified_event.event_link_node in
  caml_list_of_stateflow_link_node
    tbl "event list" get_link_node item_of_classified_item_node
    first_event_ref_opt
;;

(* Find junctions *)
let stateflow_junction_of_classified_junction_node chart_id = function
  | {
      junction_id = junction_id;
      junction_chart_ref = chart_ref;
      junction_link_node = _link_node;
      junction_stateflow_items = stateflow_items;
    } ->
    if chart_ref <> chart_id then
      Say.fatal_error (Printf.sprintf
         "incompatible reference between chart %a and junction %a"
         sprint_natural chart_id
         sprint_natural junction_id
      ) else
    {
      stateflow_junction_id = junction_id;
      stateflow_junction_items = stateflow_items;
    }
;;

let find_stateflow_junctions
    stateflow_tables chart_id first_junction_ref_opt =
  let tbl = stateflow_tables.junctions in
  let item_of_classified_item_node =
    stateflow_junction_of_classified_junction_node chart_id in
  let get_link_node classified_junction =
    classified_junction.junction_link_node in
  caml_list_of_stateflow_link_node
    tbl "junction list" get_link_node item_of_classified_item_node
    first_junction_ref_opt
;;

(* Find transitions *)
let stateflow_transition_of_classified_transition_node chart_id = function
  | {
      transition_id = transition_id;
      transition_chart_ref = chart_ref;
      transition_link_node = _link_node;
      transition_stateflow_items = stateflow_items;
    } ->
    if chart_ref <> chart_id then
      Say.fatal_error (Printf.sprintf
         "incompatible reference between chart %a and transition %a"
         sprint_natural chart_id
         sprint_natural transition_id
      ) else
    {
      stateflow_transition_id = transition_id;
      stateflow_transition_items = stateflow_items;
    }
;;

let find_stateflow_transitions
    stateflow_tables chart_id first_transition_ref_opt =
  let tbl = stateflow_tables.transitions in
  let item_of_classified_item_node =
    stateflow_transition_of_classified_transition_node chart_id in
  let get_link_node classified_transition =
    classified_transition.transition_link_node in
  caml_list_of_stateflow_link_node
    tbl "transition list" get_link_node item_of_classified_item_node
    first_transition_ref_opt
;;

(* Find states *)
let first_state_ref_of_tree_node tree_node =
  tree_node.Mdl_parsing_ast.tree_node_first_child_ref
;;

let rec stateflow_state_of_classified_state_node
    stateflow_tables machine_ref chart_id classified_state_node =
  match classified_state_node with
  | {
      state_id = state_id;
      state_chart_ref = chart_ref;
      state_tree_node = tree_node;
      state_first_data_ref_opt = state_first_data_ref_opt;
      state_first_event_ref_opt = state_first_event_ref_opt;
      state_first_junction_ref_opt = state_first_junction_ref_opt;
      state_first_transition_ref_opt = state_first_transition_ref_opt;
      state_stateflow_items = stateflow_items;
    } ->
    if chart_ref <> chart_id then
      Say.fatal_error (Printf.sprintf
         "incompatible reference between chart %a and state %a"
         sprint_natural chart_id
         sprint_natural state_id
      ) else
    let stateflow_datas =
      find_stateflow_datas
        stateflow_tables machine_ref state_first_data_ref_opt in
    let stateflow_events =
      find_stateflow_events
        stateflow_tables machine_ref state_first_event_ref_opt in
    let stateflow_junctions =
      find_stateflow_junctions
        stateflow_tables chart_id state_first_junction_ref_opt in
    let stateflow_states =
      let state_first_state_ref_opt =
        first_state_ref_of_tree_node tree_node in
      find_stateflow_states
        stateflow_tables machine_ref chart_id state_first_state_ref_opt in
    let stateflow_transitions =
      find_stateflow_transitions
        stateflow_tables chart_id state_first_transition_ref_opt in
    {
      stateflow_state_id = state_id;
      stateflow_state_datas = stateflow_datas;
      stateflow_state_events = stateflow_events;
      stateflow_state_junctions = stateflow_junctions;
      stateflow_state_states = stateflow_states;
      stateflow_state_transitions = stateflow_transitions;
      stateflow_state_items = stateflow_items;
    }

and find_stateflow_states
    stateflow_tables machine_ref chart_id state_first_child_ref =
  let tbl = stateflow_tables.states in
  let item_of_classified_item_node =
    stateflow_state_of_classified_state_node
      stateflow_tables machine_ref chart_id in
  let get_tree_node classified_state =
    classified_state.state_tree_node in
  caml_list_of_stateflow_tree_node
    tbl "state tree" get_tree_node
    item_of_classified_item_node
    (ref_option_of_stateflow_ref state_first_child_ref)
;;

(* Find stateflow items according to a given
   stateflow_item_of_classified_item retrieving function. *)
let find_stateflow_items stateflow_item_of_classified_item tbl parent_id =
  if Hashtbl.mem tbl parent_id then
    let items =
      Hashtbl.find tbl parent_id in
    Hashtbl.fold
      (fun _item_id classified_item stateflow_items ->
       let stateflow_item =
         stateflow_item_of_classified_item classified_item in
       stateflow_item :: stateflow_items)
      items []
  else
    []
;;

(* Find charts *)
let stateflow_chart_of_classified_chart stateflow_tables classified_chart =
  match classified_chart with
  | {
      chart_id = chart_id;
      chart_machine_ref = machine_ref;
      chart_tree_node_opt = tree_node_opt;
      chart_first_data_ref_opt = chart_first_data_ref_opt;
      chart_first_event_ref_opt = chart_first_event_ref_opt;
      chart_first_junction_ref_opt = chart_first_junction_ref_opt;
      chart_first_transition_ref_opt = chart_first_transition_ref_opt;
      chart_stateflow_items = stateflow_items;
    } ->
    let stateflow_datas =
      find_stateflow_datas
        stateflow_tables machine_ref chart_first_data_ref_opt in
    let stateflow_events =
      find_stateflow_events
        stateflow_tables machine_ref chart_first_event_ref_opt in
    let stateflow_junctions =
      find_stateflow_junctions
        stateflow_tables chart_id chart_first_junction_ref_opt in
    let stateflow_states =
      match tree_node_opt with
      | None -> []
      | Some tree_node ->
        let chart_first_state_ref_opt =
          first_state_ref_of_tree_node tree_node in
        find_stateflow_states
          stateflow_tables machine_ref chart_id chart_first_state_ref_opt in
    let stateflow_transitions =
      find_stateflow_transitions
        stateflow_tables chart_id chart_first_transition_ref_opt in
    {
      stateflow_chart_id = chart_id;
      stateflow_chart_datas = stateflow_datas;
      stateflow_chart_events = stateflow_events;
      stateflow_chart_junctions = stateflow_junctions;
      stateflow_chart_states = stateflow_states;
      stateflow_chart_transitions = stateflow_transitions;
      stateflow_chart_items = stateflow_items;
    }
;;

let find_stateflow_charts stateflow_tables machine_id =
  find_stateflow_items
    (stateflow_chart_of_classified_chart stateflow_tables)
    stateflow_tables.charts_by_machine_ref
    machine_id
;;

(* Find machines *)
let stateflow_machine stateflow_tables = function
  | {
      machine_id = machine_id;
      machine_first_data_ref_opt = machine_first_data_ref_opt;
      machine_first_event_ref_opt = machine_first_event_ref_opt;
      machine_first_target_ref_opt = machine_first_target_ref_opt;
      machine_stateflow_items = machine_stateflow_items;
    } ->
    let stateflow_charts =
      find_stateflow_charts stateflow_tables machine_id in
    let stateflow_datas =
      find_stateflow_datas
        stateflow_tables machine_id machine_first_data_ref_opt in
    let stateflow_events =
      find_stateflow_events
        stateflow_tables machine_id machine_first_event_ref_opt in
    let stateflow_targets =
      find_stateflow_targets
        stateflow_tables machine_id machine_first_target_ref_opt in
    {
      stateflow_machine_id = machine_id;
      stateflow_machine_charts = stateflow_charts;
      stateflow_machine_datas = stateflow_datas;
      stateflow_machine_events = stateflow_events;
      stateflow_machine_targets = stateflow_targets;
      stateflow_machine_items = machine_stateflow_items;
    }
;;

let name_option_of_stateflow_item = function
  | Stateflow_name name -> Some name
  | Stateflow_binding _
  | Stateflow_eml _
  | Stateflow_execution_order _
  | Stateflow_label_string _
  | Stateflow_chart_file_number _
  | Stateflow_scope _
  | Stateflow_props _
  | Stateflow_ssid_number _
  | Stateflow_trigger _
  | Stateflow_type _
  | Stateflow_src _
  | Stateflow_dst _
  | Stateflow_plant_modeling_info -> None
(* Extracting a name from the proper stateflow item. *)
;;

let find_single_item f items =
  let found_item_opt =
    List.fold_left
      (fun found_item_opt item ->
       match f item with
       | None -> found_item_opt
       | Some found_item ->
         match found_item_opt with
           | None -> Some found_item
           | Some _ -> fatal_error "find_single_item: duplicate item")
      None items in
  match found_item_opt with
  | None -> raise Not_found
  | Some found_item -> found_item
(* Find an item, checking it is indeed unique. *)
;;

(* Finding names in stateflow items. *)
let find_name item_type item_id stateflow_items =
  try
    find_single_item name_option_of_stateflow_item stateflow_items
  with
  | Not_found ->
    fatal_error (Printf.sprintf "missing name for %s %a"
      item_type
      sprint_natural item_id)
;;

(* Building Stateflow instances *)
let stateflow_instance stateflow_tables = function
| {
    instance_id = instance_id;
    instance_machine_ref = instance_machine_ref;
    instance_chart_ref = instance_chart_ref;
    instance_stateflow_items = instance_stateflow_items;
  } ->
  check_machine_ref stateflow_tables instance_id instance_machine_ref;
  check_chart_ref stateflow_tables instance_id
    instance_machine_ref instance_chart_ref;
  let stateflow_instance_name =
    find_name "instance" instance_id instance_stateflow_items in
  {
    stateflow_instance_id = instance_id;
    stateflow_instance_name = stateflow_instance_name;
    stateflow_instance_machine_ref = instance_machine_ref;
    stateflow_instance_chart_ref = instance_chart_ref;
  }
;;

(* Building the Stateflow tree AST of MDL files *)

let build_stateflow_tree stateflow_tables =
  let stateflow_section_items =
    Hashtbl.fold
      (fun _machine_id classified_machine stateflow_section_items ->
       let stateflow_machine =
         stateflow_machine stateflow_tables classified_machine in
       Stateflow_section_machine stateflow_machine :: stateflow_section_items
      )
      stateflow_tables.machines [] in
  let stateflow_section_items =
    Hashtbl.fold
      (fun _instance_id classified_instance stateflow_section_items ->
       let stateflow_instance =
         stateflow_instance stateflow_tables classified_instance in
       Stateflow_section_instance stateflow_instance :: stateflow_section_items
      )
      stateflow_tables.instances stateflow_section_items in
  Stateflow_section (List.rev stateflow_section_items)
;;

let tree_section_items stateflow_section_items =
  let stateflow_tables = make_stateflow_tables stateflow_section_items in
  build_stateflow_tree stateflow_tables
;;

let tree_stateflow_section = function
  | Mdl_stateflow_section mdl_stateflow_section_items ->
    tree_section_items mdl_stateflow_section_items
;;

let tree_implementation_section_item = function
  | Mdl_ast.Implementation_library _
  | Implementation_model _
  | Implementation_matdata _ as implementation_section_item ->
    implementation_section_item
  | Implementation_stateflow parsing_stateflow_section ->
    Implementation_stateflow (tree_stateflow_section parsing_stateflow_section)
;;

let tree_implementation_section implementation_section =
  List.map tree_implementation_section_item implementation_section
;;

let tree_simulink_file_contents = function
  | Mdl_ast.Implementation implementation_section ->
    Mdl_ast.Implementation (tree_implementation_section implementation_section)
  | Interface _ as interface -> interface
;;

(* Parsing MDL files *)
let parsing_simulink_file_to_ast_simulink_file
      (path, parsing_simulink_file_contents) =
  let ast_simulink_file_contents =
    tree_simulink_file_contents parsing_simulink_file_contents in
  path, ast_simulink_file_contents
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
  End:
*)
