(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mbd_deep_ast_pprint.ml,v 1.1 2019-05-10 13:58:20 jpc Exp $ *)

open Mbd_predef;;
open Mbd_predefined;;
open Mbd_deep_ast;;
open Format;;

let is_prefix_binary_prim = function
  | Bpow
  | Bpower -> false

  | Batan2
  | Bhypot

  | Bmin
  | Bmax
  | Brem -> true
;;

let print_unary_op ppf op =
  let op =
    match Mbd_predefined.string_of_unary_op op with
    | "!" -> "~"
    | op -> op in
  fprintf ppf "%s" op
;;

let print_binary_op ppf op =
  let op =
    match Mbd_predefined.string_of_binary_op op with
    | "!=" -> "~="
    | "||" -> "|"
    | "&&" -> "&"
    | op -> op in
  fprintf ppf "%s" op
;;

let print_unary_prim ppf prim =
  let prim =
    match Mbd_predefined.string_of_unary_prim prim with
    | "fabs" -> "abs"
    | "sgn" -> "sign"
    | "ln" -> "log"
    | op -> op in
  fprintf ppf "%s" prim
;;

let print_binary_prim ppf prim =
  let prim =
    match Mbd_predefined.string_of_binary_prim prim with
    | "pow" -> "^"
    | "power" -> "^"
    | prim -> prim in
  fprintf ppf "%s" prim
;;

let print_variable ppf v =
  let v_name =
    match v with
    | Mbd_predef.Gop1 op -> string_of_unary_op op
    | Gop2 op -> string_of_binary_op op
    | Gprim1 prim -> string_of_unary_prim prim
    | Gprim2 prim -> string_of_binary_prim prim
    | Gident v_name -> v_name in
  fprintf ppf "%s" v_name
;;

let is_closed_expression = function
  | D_number _
  | D_string _
  | D_variable _
  | D_uparameter_paren _
  | D_uparameter_bracket _
  | D_parameter _ -> true
  | D_apply _
  | D_unary_op _
  | D_binary_op _
  | D_unary_prim _
  | D_binary_prim _ -> false
  | D_record_access _
  | D_paren _ -> true
;;

let rec pprint ppf = function
  | D_number n -> fprintf ppf "%s" n
  | D_string s -> fprintf ppf "%S" s
  | D_variable v -> print_variable ppf v
  | D_uparameter_paren i -> fprintf ppf "u (%a)" pprint i
  | D_uparameter_bracket i -> fprintf ppf "u [%a]" pprint i
  | D_parameter i -> fprintf ppf "u%d" i
  | D_apply (e_f, e_args) ->
    fprintf ppf "%a (%a)"
      pprint e_f print_args e_args
  | D_unary_op (op, x) ->
    fprintf ppf "%a %a"
      print_unary_op op
      pprint x
  | D_binary_op (op, x, y) ->
    fprintf ppf "%a %a %a"
      pprint x
      print_binary_op op
      pprint y
  | D_unary_prim (prim, x) ->
    fprintf ppf "%a (%a)"
      print_unary_prim prim
      pprint x
  | D_binary_prim (prim, x, y)
      when is_prefix_binary_prim prim ->
    fprintf ppf "%a (%a, %a)"
      print_binary_prim prim
      pprint x
      pprint y
  | D_binary_prim (prim, x, y) ->
    fprintf ppf "%a %a %a"
      pprint_closed_expression x
      print_binary_prim prim
      pprint_closed_expression y
  | D_record_access (e1, label) ->
    fprintf ppf "%a.%s"
      pprint e1
      label
  | D_paren e ->
    fprintf ppf "(%a)" pprint e

and print_args ppf = function
  | [] -> ()
  | arg :: args ->
    pprint ppf arg;
    List.iter (function arg -> fprintf ppf ", %a" pprint arg) args

and pprint_closed_expression ppf expr =
  if is_closed_expression expr
  then pprint ppf expr
  else Format.fprintf ppf "(%a)" pprint expr
;;

(*
 Local Variables:
  compile-command: "cd ../..; make"
  End:
*)
