(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mbd_lexer.mll,v 1.1 2019-05-10 13:58:20 jpc Exp $ *)

{

(* Prelude part: this is pure Caml. *)

open Lexing;;
open Mbd_parser;;

(** {6 Lexing errors} *)

type error =
   | Illegal_character of char
   | Illegal_escape of string
   | Unterminated_string
(** The various errors when lexing. *)
;;

exception Error of error * Lexing.position * Lexing.position;;

let error lexbuf err =
  raise (Error (err, lexbuf.lex_start_p, lexbuf.lex_curr_p))
;;

(** {6 Explaining lexing errors} *)

let report_error ppf = function
  | Illegal_character c ->
      Format.fprintf ppf "Illegal character (%C)" c
  | Illegal_escape s ->
      Format.fprintf ppf "Illegal escape (%S)" s
  | Unterminated_string ->
      Format.fprintf ppf "Unterminated string"
;;

let token_of_ident s =
  assert (String.length s > 0);
  IDENT s
;;

(** {6 Keeping the internal buffer locations up to date} *)

let update_loc lexbuf fname line absolute chars =
  let pos = lexbuf.lex_curr_p in
  let new_fname =
    match fname with
    | None -> pos.pos_fname
    | Some s -> s in
  lexbuf.lex_curr_p <- {
    pos with
    pos_fname = new_fname;
    pos_lnum = if absolute then line else pos.pos_lnum + line;
    pos_bol = pos.pos_cnum - chars;
  }
;;

(** Add one to the current line counter of the file being lexed. *)
let incr_line_num lexbuf =
  update_loc lexbuf None 1 false 0
;;

(** {6 Lexing the string tokens} *)
let string_buffer = Buffer.create 256;;

let string_start_pos = ref None;;

let store_string_char c =
  Buffer.add_char string_buffer c
;;

let get_stored_string () =
  let s = Buffer.contents string_buffer in
  Buffer.reset string_buffer;
  s
;;

let char_for_character = function
  | '\\' -> '\\'
  | '\"' -> '\"'
  | 'n' -> '\n'
  | 't' -> '\t'
  | 'b' -> '\b'
  | 'r' -> '\r'
  | c ->
    failwith
      (Printf.sprintf "mbd: lexer error, unknown escaped character %C" c)
;;

}

(** The lexer: this is the pure ocamllex part. *)

let newline = ( '\n' | '\r' | "\r\n" )
let blank = [ ' ' '\t' '\012' ]

(** {6 Floating point numbers} *)

let decimal_digit = [ '0' - '9' ]
let unsigned_decimal_literal = decimal_digit+

let sign = [ '+' '-' ]

let decimal_literal = (sign ?) unsigned_decimal_literal

let scientific_notation = [ 'e' 'E' 'd' 'D' ]
let exponent = scientific_notation decimal_literal

let unsigned_decimal_float_literal_body =
    unsigned_decimal_literal
    ('.' (unsigned_decimal_literal ?)) ?
  | ('.' unsigned_decimal_literal)

let float_literal = unsigned_decimal_float_literal_body (exponent ?)

(** {6 Strings} *)
let escaped_character =
    "\\" [ '\\' '\"' 'n' 't' 'b' 'r' ]

(** {6 Identifiers} *)
let lowercase_alphabetic = [ 'a' - 'z' ]
let uppercase_alphabetic = [ 'A' - 'Z' ]

let inside_ident =
    lowercase_alphabetic
  | uppercase_alphabetic
  | decimal_digit
  | '_'

let ident = ( lowercase_alphabetic | uppercase_alphabetic) inside_ident*

(** {6 Comments} *)

let comment = "//" [^ '\n' '\r' ]* newline

(** {3 The main lexer. *)

rule token = parse
  | comment
    { incr_line_num lexbuf;
      token lexbuf }
  | newline
    { incr_line_num lexbuf;
      token lexbuf }
  | blank +
    { token lexbuf }

  (* Numbers *)
  | float_literal
    { FLOAT (Lexing.lexeme lexbuf) }

  (* Strings *)
  | '\"'
    {
      string_start_pos :=
        Some (lexbuf.lex_start_p, lexbuf.lex_curr_p);
      string lexbuf;
      STRING (get_stored_string ())
    }

  (* Parameters *)
  | 'u'
    { UPARAM }
  | 'u' ([ '1' - '9' ] [ '0' - '9' ]*)
    { PARAM (Lexing.lexeme lexbuf) }

  (* Identifiers *)
  | ident
    { token_of_ident (Lexing.lexeme lexbuf) }

  (* Symbols (or symbolic idents) *)
  | '^' { HAT_OP "^" }
  | '+' { PLUS_OP "+" }
  | '-' { DASH_OP "-" }
  | '*' { STAR_OP "*" }
  | '/' { SLASH_OP "/" }
  | "&&" { AMPER_AMPER_OP "&&" }
  | "||" { BAR_BAR_OP "||" }
  | '<' { LT_OP "<" }
  | "<=" { LT_OP "<=" }
  | "==" { EQ_OP "==" }
  | '>' { GT_OP ">" }
  | ">=" { GT_OP ">=" }
  | '!' { BANG_OP "!" }
  | "!=" { NE_OP "!=" }

  (* Usual simple tokens *)
  | '(' { LPAREN }
  | ')' { RPAREN }
  | '[' { LBRACKET }
  | ']' { RBRACKET }
  | ',' { COMMA }
  | ';' { SEMI }
  | '.' { DOT_OP "." }
  | eof { EOF }
  | _
    { error lexbuf (Illegal_character (Lexing.lexeme_char lexbuf 0)) }

and string = parse
  | '\"'
    { () }
  | escaped_character
    { store_string_char (char_for_character (Lexing.lexeme_char lexbuf 1));
      string lexbuf }
  | '\\' _
    { error lexbuf (Illegal_escape (Lexing.lexeme lexbuf)) }
  | eof
    { match !string_start_pos with
      | Some (start_pos, end_pos) ->
        raise (Error (Unterminated_string, start_pos, end_pos))
      | _ -> assert false }
  | _
    { store_string_char (Lexing.lexeme_char lexbuf 0);
      string lexbuf }

(*
 Local Variables:
  compile-command: "cd ../..; make"
  End:
*)
