/***********************************************************************/
/*                                                                     */
/*                               Simport                               */
/*                                                                     */
/*                   Pierre Weis, INRIA Rocquencourt                   */
/*                                                                     */
/*  Copyright 2010-2019,                                               */
/*  Institut National de Recherche en Informatique et en Automatique.  */
/*  All rights reserved.                                               */
/*                                                                     */
/*  This file is distributed under the terms of the BSD License.       */
/*                                                                     */
/***********************************************************************/
%{

(* $Id: mbd_parser.mly,v 1.1 2019-05-10 13:58:20 jpc Exp $ *)

open Mbd_ast;;

let mk_infix_application e1 op e2 = E_binary_op (op, e1, e2);;

let mk_unary_application op e1 = E_unary_op (op, e1);;

%}

%token EOF
%token END

%token <string> IDENT /* Lower case ident (e.g. x, cos, _1, or _xY) */
%token <string> PARAM /* Named block parameter (e.g. u1, or u20) */
%token UPARAM /* Anonymous block parameter: u */

%token <string> DOT_OP

/* Basic constants */
%token <string> FLOAT
%token <string> STRING

/* Arithmetic operators */
%token <string> BANG_OP
%token <string> HAT_OP

%token <string> STAR_OP
%token <string> SLASH_OP

%token <string> PLUS_OP
%token <string> DASH_OP

%token <string> LT_OP
%token <string> GT_OP
%token <string> EQ_OP
%token <string> NE_OP

%token <string> BAR_BAR_OP
%token <string> AMPER_AMPER_OP

%token COMMA

/* Nested symbols */
%token LPAREN
%token RPAREN
%token LBRACKET
%token RBRACKET

/* Keywords */
%token SEMI

/* Precedences and associativities. */

/* Binary operators. */

%right    BAR_BAR_OP               /* expr (e || e || e) */
%right    AMPER_AMPER_OP           /* expr (e && e && e) */

%left     EQ_OP NE_OP              /* expr (e = e, != e) */
%left     LT_OP GT_OP              /* expr (e < e, e <= e, e > e, e >= e) */

%left     COMMA

%left     PLUS_OP DASH_OP
          /* expr (e OP e OP OP e) with OP starting with '+', or '-' */

%left     STAR_OP SLASH_OP
          /* expr (e OP e OP e) with OP starting with '*', '/', or '%' */

%right    BANG_OP
          /* expr (OP e) with OP starting with '!' */

%left     DOT_OP
          /* expr (e OP e) with OP starting with '.' */

%nonassoc prec_unary_bang

%nonassoc prec_unary_minus

%right    HAT_OP
          /* expr (e OP e OP e) with OP starting with '^' */

/* Predefined precedences to resolve conflicts. */
/* TO CHECK */

%nonassoc below_RPAREN
%nonassoc RPAREN
%nonassoc LPAREN
%nonassoc RBRACKET
%nonassoc LBRACKET

%start expression
%type <Mbd_ast.expression> expression
%%

expression:
  | expr { $1 }
  | expr SEMI { $1 }
;

expr:
  | FLOAT
    { E_number $1 }
  | STRING
    { E_string $1 }
  | IDENT
    { E_variable $1 }

  | UPARAM LPAREN expr RPAREN
    { E_uparameter_paren ($3) }
  | UPARAM LBRACKET expr RBRACKET
    { E_uparameter_bracket ($3) }
  | UPARAM %prec below_RPAREN
    { E_uparameter }
  | PARAM
    { E_parameter $1 }

  /* Application */
  | expr LPAREN expr_comma_list RPAREN
    { E_apply ($1, $3) }

  /* Binary operators */

  | expr BAR_BAR_OP expr
    { mk_infix_application $1 $2 $3 }
  | expr AMPER_AMPER_OP expr
    { mk_infix_application $1 $2 $3 }

  | expr EQ_OP expr
    { mk_infix_application $1 $2 $3 }
  | expr LT_OP expr
    { mk_infix_application $1 $2 $3 }
  | expr GT_OP expr
    { mk_infix_application $1 $2 $3 }
  | expr NE_OP expr
    { mk_infix_application $1 $2 $3 }

  | expr PLUS_OP expr
    { mk_infix_application $1 $2 $3 }
  | expr DASH_OP expr
    { mk_infix_application $1 $2 $3 }

  | expr STAR_OP expr
    { mk_infix_application $1 $2 $3 }
  | expr SLASH_OP expr
    { mk_infix_application $1 $2 $3 }

  | expr HAT_OP expr
    { mk_infix_application $1 $2 $3 }

  | expr DOT_OP label
    { E_record_access ($1, $3) }

  /* Unary operators. */

  | DASH_OP expr %prec prec_unary_minus
    { mk_unary_application $1 $2 }
  | PLUS_OP expr %prec prec_unary_minus
    { mk_unary_application $1 $2 }
  | BANG_OP expr %prec prec_unary_bang
    { mk_unary_application $1 $2 }

  /* Parens expressions */
  | LPAREN expr RPAREN
    { E_paren $2 }
;

label:
  | IDENT { $1 }
;

expr_comma_list:
  | { [ ] }
  | expr_comma_list1 { $1 }
;

expr_comma_list1:
  | expr { [ $1 ] }
  | expr COMMA expr_comma_list1 { $1 :: $3 }
;

/*

 Local Variables:
  compile-command: "cd ../..; make"
  End:

*/
