(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mbd_transl.ml,v 1.1 2019-05-10 13:58:20 jpc Exp $ *)

open Mbd_predef;;
open Mbd_deep_ast;;

let transl_error s =
  Printf.eprintf "transl: fatal error %s\n" s;
  failwith "transl"
;;

(*let transl_unary_prim = Mbd_predefined.unary_prim_of_string;;

let transl_binary_prim = Mbd_predefined.binary_prim_of_string;;

let transl_unary_op = Mbd_predefined.unary_op_of_string;;

let transl_binary_op = Mbd_predefined.binary_op_of_string;;
*)
let mk_uminus n =
  let len = String.length n in
  assert (len > 0);
  match n.[0] with
  | '-' -> Printf.sprintf "%s" (String.sub n 1 (len - 1))
  | '+' -> Printf.sprintf "-%s" (String.sub n 1 (len - 1))
  | _ -> Printf.sprintf "-%s" n
;;

(*
let make_paren exp =
  match exp with
  | D_number s ->
    if String.length s = 0 then assert false else
    if s.[0] = '-' then D_paren exp else exp
  | D_string _ -> exp
  | D_variable _ -> exp
  | D_uparameter_paren _ -> exp
  | D_uparameter_bracket _ -> exp
  | D_parameter _ -> exp
  | D_unary (_, _)
  | D_binary (_, _, _)
  | D_apply (_, _) -> D_paren exp
  | D_record_access (_, _) -> exp
  | D_paren _ -> exp
;;
*)

let transl_parameter v = int_of_string v;;

let rec transl e =
  match e with
  | D_number _
  | D_string _ -> e
    (* Should restrict variables to real identifiers (not op nor prim) ? *)
  | D_variable _
  | D_parameter _ -> e
  | D_uparameter_paren e ->
    transl_uparameter_expr (fun e -> D_uparameter_paren e) (transl e)
  | D_uparameter_bracket e ->
    transl_uparameter_expr (fun e -> D_uparameter_bracket e) (transl e)

  | D_unary_op (Mbd_predef.Uplus, e) -> transl e
  | D_unary_op (Mbd_predef.Uminus, D_number n) -> D_number (mk_uminus n)
  | D_unary_op (Mbd_predef.Uminus as op, e) -> D_unary_op (op, transl e)
  | D_unary_op (Mbd_predef.Ubang as op, e) -> D_unary_op (op, transl e)

  | D_binary_op (op, x, y) -> D_binary_op (op, transl x, transl y)

  | D_unary_prim (prim, x) -> D_unary_prim (prim, transl x)
  | D_binary_prim (prim, x, y) -> D_binary_prim (prim, transl x, transl y)

  | D_apply (D_paren (D_record_access (e, label)), args)
  | D_apply (D_record_access (e, label), args) ->
    let _v =
      match e with
      | D_variable _ -> ()
      | D_number _
      | D_string _
      | D_uparameter_paren _
      | D_uparameter_bracket _
      | D_parameter _
      | D_unary_op (_, _)
      | D_binary_op (_, _, _)
      | D_unary_prim (_, _)
      | D_binary_prim (_, _, _)
      | D_apply (_, _)
      | D_record_access (_, _)
      | D_paren _ ->
        transl_error
          (Printf.sprintf
             "illegal record access to field %s of a non \
              variable expression"
             label) in
    let args = List.map transl args in
    D_apply (D_record_access (transl e, label), args)

  | D_apply ((D_variable v as f), [ x ]) ->
    begin match v with
    | Mbd_predef.Gop1 op -> D_unary_op (op, transl x)
    | Mbd_predef.Gprim1 prim -> D_unary_prim (prim, transl x)
      (* Binary operator/prim arity error ? *)
    | Mbd_predef.Gop2 _
    | Mbd_predef.Gprim2 _
    | Mbd_predef.Gident _ -> D_apply (f, [ transl x ])
    end

  | D_apply ((D_variable v as f), [ x; y ]) ->
    begin match v with
      (* Unary operator/prim arity error ? *)
    | Gop1 _
    | Gprim1 _ -> D_apply (f, [ transl x; transl y ])
    | Gop2 op -> D_binary_op (op, transl x, transl y)
    | Gprim2 prim -> D_binary_prim (prim, transl x, transl y)
    | Gident _ -> D_apply (f, [ transl x; transl y ])
    end

  | D_apply (f, args) ->
    D_apply (transl f, List.map transl args)

  | D_record_access (e1, label) ->
    D_record_access (transl e1, label)
  | D_paren exp -> D_paren (transl exp)

and transl_uparameter_expr construct e =
  match e with
  | D_number n -> D_parameter (transl_parameter n)
  | D_paren (D_number _ as e) -> transl_uparameter_expr construct e
  | D_string _
  | D_variable _
  | D_uparameter_paren _
  | D_uparameter_bracket _
  | D_parameter _
  | D_unary_op (_, _)
  | D_binary_op (_, _, _)
  | D_unary_prim (_, _)
  | D_binary_prim (_, _, _)
  | D_apply (_, _)
  | D_record_access (_, _)
  | D_paren _ ->
    construct (e)
;;

let translate = transl;;

(*
 Local Variables:
  compile-command: "cd ../..; make"
  End:
*)
