(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2010-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mbd_treat.ml,v 1.1 2019-05-10 13:58:20 jpc Exp $ *)

open Format;;

type config = {
  print_ast : bool;
  print_deep_ast : bool;
  print_transl_ast : bool;
  ppf : Format.formatter;
  eppf : Format.formatter;
}
;;

let mbd_config = {
  print_ast = false;
  print_deep_ast = false;
  print_transl_ast = false;
  ppf = std_formatter;
  eppf = err_formatter;
}
;;

let debug_config = {
  print_ast = true;
  print_deep_ast = true;
  print_transl_ast = true;
  ppf = std_formatter;
  eppf = err_formatter;
}
;;

let get_config, set_config =
  let config = ref mbd_config in
  (fun () -> !config),
  (fun c -> config := c)
;;
let set_debug () = set_config debug_config;;

let parse config efname_opt lexbuf =
  try
    Mbd_parser.expression Mbd_lexer.token lexbuf
  with
  | Mbd_lexer.Error (e, pos_begin, pos_end) ->
    let eppf = config.eppf in
    let file_info =
      match efname_opt with
      | None -> "in string input"
      | Some ef_name ->
        Printf.sprintf "file %s" (Path.of_explicit_file_name ef_name) in
    Format.fprintf eppf
      "Lexical error: %s, %a, %a@."
      file_info
      Location.print_token_location (pos_begin, pos_end)
      Mbd_lexer.report_error e;
    Say.fatal_error "sime parsing error"
;;

let parse_file config efname =
  begin
    if Configuration.get_verbose () then
      let eppf = config.eppf in
      Format.fprintf eppf
        "parsing file %S@."
        (Path.of_explicit_file_name efname);
  end;
  Path.with_explicit_in_file efname (fun ic ->
    let lexbuf = Lexing.from_channel ic in
    parse config (Some efname) lexbuf)
;;

let parse_string config s =
  begin
    if Configuration.get_verbose () then
      let eppf = config.eppf in
      Format.fprintf eppf "parsing string %S@." s;
  end;
  let lexbuf = Lexing.from_string s in
  parse config None lexbuf
;;

let treat_ast config parse =

  let eppf = config.eppf in

  if config.print_ast then Format.fprintf eppf "@.Parsing:@.";
  let ast = parse () in
  if config.print_ast then
    Format.fprintf eppf "%a@." Mbd_ast_pprint.pprint ast;

  if config.print_deep_ast then Format.printf "@.Scoping:@.";
  let ast = Mbd_ast_to_deep_ast.ast_to_deep_ast ast in
  if config.print_deep_ast then
    Format.fprintf eppf "%a@." Mbd_deep_ast_pprint.pprint ast;

  if config.print_transl_ast then Format.printf "@.Translating:@.";
  let ast = Mbd_transl.translate ast in
  if config.print_transl_ast then
    Format.fprintf eppf "%a@." Mbd_deep_ast_pprint.pprint ast;

  ast
;;

let treat_file f =
  let config = get_config () in
  treat_ast config (fun () -> parse_file config f)
;;

let treat_string s =
  let config = get_config () in
  treat_ast config (fun () -> parse_string config s)
;;

(*
 Local Variables:
  compile-command: "cd ../..; make"
  End:
*)
