(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, Clément Franchini                    *)
(*                   INRIA Rocquencourt                                *)
(*                                                                     *)
(*  Copyright 2013-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: xml_ast_to_xml_deep_ast.ml,v 1.1 2019-05-10 14:00:16 jpc Exp $ *)

open Xml_ast;;
open Xml_deep_ast;;
(* open Slx_xml_check;; *)

let to_xml_deep_ast_error s =
  Say.fatal_error ("to xml deep ast " ^ s)
;;

let deep_attribute_value = function
  | [ Xml_basics_ast.Att_value_text att_value_text; ] -> att_value_text
  | [  Xml_basics_ast.Att_value_reference _]
  | []
  | _ :: _ :: _ ->
    (* Xml_expand should have translated the list of
       att_value_elem into a single att_value_text *)
    assert false
;;

let deep_attribute = function
  | {
    tag_attribute_name = tag_attribute_name;
    tag_attribute_value = att_value;
    } ->
    let da_name = Xml_basics_ast.of_xml_name tag_attribute_name
    and da_value = deep_attribute_value att_value in
    {
      da_name = da_name;
      da_value = da_value;
    }
;;

let deep_attributes tag_attributes =
  List.rev (List.rev_map deep_attribute tag_attributes)
;;

let deep_char_data (Char_data char_data_text) = char_data_text

let rec deep_child accu = function
  | Child_element element ->
    Intercal.In_right (deep_element element) :: accu
  | Child_cd_sect _cdata ->
    to_xml_deep_ast_error "not supported CData"
  | Child_pi _pi ->
    to_xml_deep_ast_error "not supported PI (Processing Instructions)"
  | Child_comment _comment ->
    accu
  | Child_reference _reference ->
    (* Xml_expand should have translated the Child_reference to Character
       Data *)
    assert false

and deep_children children =
  let children =
    Intercal.fold_left
      (fun accu char_data ->
        Intercal.In_left (deep_char_data char_data) :: accu)
      deep_child
      []
      children in
  List.rev children

and deep_element element =
  let tag, element_content =
    match element with
    | Xml_ast.Empty_element_tag t ->
      t, Intercal.empty
    | Xml_ast.Content_element_tag (tag, Element_content element_content) ->
      tag, element_content in
  let de_name = Xml_basics_ast.of_xml_name tag.tag_name
  and de_attributes = deep_attributes tag.tag_attributes
  and de_children = deep_children element_content in
  {
    de_name = de_name;
    de_attributes = de_attributes;
    de_children = de_children;
  }
;;

let deep_document document =
  match Xml_expand.of_document document with
  | {
      prolog = prolog;
      element = element;
      misc_star = misc_star;
    } ->
    Slx_xml_check.check_misc_star misc_star;
    Slx_xml_check.check_prolog prolog;
    let encoding = Slx_xml_check.get_encoding prolog in
    {
      deep_prolog = encoding;
      deep_element = deep_element element;
    }
;;

let deep_annex_name = function
  | Slx_to_xml_ast.Block_diagram_defaults -> Deep_block_diagram_defaults
  | Slx_to_xml_ast.Config_set_info -> Deep_config_set_info
  | Slx_to_xml_ast.Graphical_interface -> Deep_graphical_interface
  | Slx_to_xml_ast.Stateflow -> Deep_stateflow
  | Slx_to_xml_ast.Annex_file_name fname -> Deep_annex_file_name fname
;;

let deep_annexes annexes =
  List.map
    (fun (annex_name, annexe) ->
     deep_annex_name annex_name, deep_document annexe)
    annexes
;;

(* Slx_to_xml_ast.xml_model -> Xml_deep_ast.deep_xml_model *)
let xml_ast_to_xml_deep_ast = function
  {
    Slx_to_xml_ast.
    model_ident = model_ident;
    blockdiagram = blockdiagram;
    annexes = annexes;
  } ->
  let deep_blockdiagram = deep_document blockdiagram in
  let deep_annexes = deep_annexes annexes in
  {
    deep_model_ident = model_ident;
    deep_blockdiagram = deep_blockdiagram;
    deep_annexes = deep_annexes;
  }
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
 End:
*)
