(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, Clément Franchini                    *)
(*                   INRIA Rocquencourt                                *)
(*                                                                     *)
(*  Copyright 2013-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: xml_deep_ast_to_xml_explicit_ast.ml,v 1.1 2019-05-10 14:00:17 jpc Exp $ *)

(* TODO: gather all the tag names *)
let dialog_parameters_tag_name = "DialogParameters"
and mask_tag_name = "Mask"
and display_tag_name = "Display"
;;

open Xml_deep_ast;;
open Xml_explicit_ast;;
open Ast_types;;

type object_attributes_classifier = {
  explicit_object_id_opt : explicit_object_id option;
  explicit_object_class_name_opt : Mdl_ast.class_name option;
  explicit_object_prop_name_opt : Mdl_ast.prop_name option;
  explicit_object_backup_class_opt : Mdl_ast.backup_class_name option;
  explicit_object_version_opt : explicit_version option;
  explicit_object_reference_opt : explicit_reference option;
}
;;

type explicit_block_diagram_defaults_classifier = {
  explicit_system_defaults_opt : explicit_system_defaults option;
  explicit_block_defaults_opt : explicit_block_defaults option;
  explicit_annotation_defaults_opt : explicit_annotation_defaults option;
  explicit_line_defaults_opt : explicit_line_defaults option;
  explicit_mask_defaults_opt : explicit_mask_defaults option;
  explicit_mask_parameter_defaults_opt :
    explicit_mask_parameter_defaults option;
  explicit_block_parameter_defaults_opt :
    explicit_block_parameter_defaults option;
}
;;

type implementation_section_items_classifier = {
  explicit_model_workspace_opt : explicit_model_workspace;
  explicit_model_configuration_opt : explicit_model_configuration option;
  explicit_configuration_set_object_opt : explicit_object option;
  explicit_configuration_set_binding_list : explicit_bindings;
  explicit_block_diagram_defaults_classifier :
    explicit_block_diagram_defaults_classifier;
  explicit_system_opt : explicit_system option;
}
;;

type display_attribute_classifier = {
  explicit_display_icon_frame : Xml_deep_ast.attribute_value option;
  explicit_display_icon_units : Xml_deep_ast.attribute_value option;
  explicit_display_icon_opaque : Xml_deep_ast.attribute_value option;
  explicit_display_icon_rotate : Xml_deep_ast.attribute_value option;
  explicit_display_run_init_for_icon_redraw : Xml_deep_ast.attribute_value option;
}
;;

let empty_display_attribute_classifier = {
  explicit_display_icon_frame = None;
  explicit_display_icon_units = None;
  explicit_display_icon_opaque = None;
  explicit_display_icon_rotate = None;
  explicit_display_run_init_for_icon_redraw = None;
}
;;

type explicit_mask_classifier = {
  explicit_mask_type_opt : explicit_mask_type option;
  explicit_mask_display_opt : explicit_mask_display option;
  explicit_mask_description_opt : explicit_mask_description option;
  explicit_mask_initialisation_opt :
    explicit_mask_initialisation option;
  explicit_mask_help_opt : explicit_mask_help option;
  explicit_mask_parameters_opt : explicit_mask_parameter list;
}
;;

type mask_parameter_classifier = {
  explicit_mask_parameter_name_opt :
    Xml_deep_ast.attribute_value option;
  explicit_mask_parameter_type_opt :
    Xml_deep_ast.attribute_value option;
  explicit_mask_parameter_tunable_opt :
    Xml_deep_ast.attribute_value option;
  explicit_mask_parameter_evaluate_opt :
    Xml_deep_ast.attribute_value option;
  explicit_mask_parameter_enabled_opt :
    Xml_deep_ast.attribute_value option;
  explicit_mask_parameter_internal_opt :
    Xml_deep_ast.attribute_value option;
  explicit_mask_parameter_visible_opt :
    Xml_deep_ast.attribute_value option;
  explicit_mask_parameter_never_save_opt :
    Xml_deep_ast.attribute_value option;
  explicit_mask_parameter_read_only_opt :
    Xml_deep_ast.attribute_value option;
}
;;

let empty_explicit_mask_classifier = {
  explicit_mask_type_opt = None;
  explicit_mask_display_opt = None;
  explicit_mask_description_opt = None;
  explicit_mask_initialisation_opt = None;
  explicit_mask_help_opt = None;
  explicit_mask_parameters_opt = [];
}
and empty_mask_parameter_classifier = {
  explicit_mask_parameter_name_opt = None;
  explicit_mask_parameter_type_opt = None;
  explicit_mask_parameter_tunable_opt = None;
  explicit_mask_parameter_evaluate_opt = None;
  explicit_mask_parameter_enabled_opt = None;
  explicit_mask_parameter_visible_opt = None;
  explicit_mask_parameter_internal_opt = None;
  explicit_mask_parameter_never_save_opt = None;
  explicit_mask_parameter_read_only_opt = None;
}
;;

let default_config_set_object_id = Unatural._1_;;

let here s =
  Printf.sprintf "xml deep ast to xml explicit ast: %s" s
;;

(*
let _not_yet s _ =
  failwith (Printf.sprintf "xml deep ast to xml explicit ast: not yet: %s" s)
;;

let _debug s = fdebug (fun ppf -> Format.fprintf ppf "%s" s);;
*)

let fdebug f =
  Say.fdebug
    (fun ppf -> Format.fprintf ppf "xml deep ast to xml explicit ast: %t" f)
;;

let fatal_error s = Say.fatal_error (here s);;

let unknown_attribute_list_for_tag tag_name =
  fatal_error (Printf.sprintf "unknown attribute list for tag %S" tag_name)
;;

let unknown_attribute_list_for_element element_name =
  fatal_error
    (Printf.sprintf "unknown attribute list for element name %S" element_name)
;;

let warning s = Say.warning (here s);;

let get_option_value message item_name = function
  | None -> fatal_error (Printf.sprintf "%s: missing %s" message item_name)
  | Some x -> x
;;

let get_list_option_value = function
  | None -> []
  | Some values -> values
;;

let list_of_options options =
  List.rev
    (List.fold_left
       (fun accu opt ->
        match opt with
        | None -> accu
        | Some item -> item :: accu)
       []
       options)
;;

let unatural_of_string s =
  try
    Unatural.of_string s
  with
  | _ ->
    fatal_error (Printf.sprintf
      "%S is not a natural" s)
;;

let dimension_of_string s =
  try
    Scanf.sscanf s " %d * %d %!"
      (fun i j -> Unatural.of_int i, Unatural.of_int j)
  with
  | _ ->
    (* TODO: implement a better error management *)
    fatal_error (Printf.sprintf
      "invalid dimension specification %S" s)
;;

let cell_class_of_string = Mdl_entry_parsers.parameter_class_of_string
and field_class_of_string = Mdl_entry_parsers.parameter_class_of_string
;;

let prop_data_type_of_string = Mdl_entry_parsers.custom_property_type_parser;;

let check_opt_is_empty m = function
  | None -> ()
  | Some _ -> fatal_error m
;;

let set_opt m opt x =
  check_opt_is_empty m opt;
  Some x
;;

let check_deep_element element_name = function
  | {
      de_name = de_name;
      de_attributes = de_attributes;
      de_children = de_children;
    } ->
    if de_name = element_name then de_attributes, de_children
    else
      fatal_error (Printf.sprintf
        "%S tag read, but %S tag was expected" de_name element_name)
;;

let check_no_attributes element_name = function
  | [] -> ()
  | _ ->
    unknown_attribute_list_for_element element_name
;;

let check_deep_attribute element_name attribute_name = function
  | {
      da_name = da_name;
      da_value = da_value;
    } ->
    if da_name = attribute_name then da_value
    else
      fatal_error (Printf.sprintf
        "%S attribute read, but %S attribute was expected for tag %S"
        da_name attribute_name element_name)
;;

let is_whitespaces s = Scanf.sscanf s " %s" (fun p -> String.length p = 0);;

let block_sid_value s = Mdl_entry_parsers.block_sid_value_parser s;;

let ssid_value s = unatural_of_string s;;

let with_elements f deep_children =
  let rec loop accu = function
    | [] -> List.rev accu
    | Intercal.In_left char_data :: deep_elements
        when is_whitespaces char_data -> loop accu deep_elements
    | Intercal.In_left _char_data :: _ ->
      fatal_error "non whitespaces char_data"
    | Intercal.In_right deep_element :: deep_elements ->
      loop (deep_element :: accu) deep_elements
  in
  let deep_elements = loop [] deep_children in
  f deep_elements
;;

let map_elements f = with_elements (List.map f);;

let fail_unknown_tag parent_element_name deep_element =
  fatal_error
    (Printf.sprintf
       "``%s'' tag inside ``%s'' tag: not supported"
       deep_element.de_name
       parent_element_name)
;;

let rec classify_element parent_element_name deep_element = function
  | [] -> fail_unknown_tag parent_element_name deep_element
  | (de_name, f) :: _ when deep_element.de_name = de_name -> f deep_element
  | _ :: classifiers ->
    classify_element parent_element_name deep_element classifiers
;;

let simplify_attribute = function
  | {
      da_name = da_name;
      da_value = da_value;
    } ->
    da_name, da_value
;;

let simplify_attributes = List.map simplify_attribute;;

let string_of_children element_name (children : deep_children) =
  match children with
  | [ Intercal.In_left char_data; ] -> char_data
  | [] -> ""
  | [ Intercal.In_right _; ]
  | _ :: _ :: _ ->
    fatal_error (Printf.sprintf
      "an element was read inside a %S tag" element_name)
;;

let string_of_element element_name deep_element =
  let de_attributes, de_children =
    check_deep_element element_name deep_element in
  check_no_attributes element_name de_attributes;
  string_of_children element_name de_children
;;

let explicit_binding_of_element debug deep_element =
  fdebug
    (fun epf -> Format.fprintf epf "explicit_binding_of_element: %S" debug);
  let de_attributes, de_children = check_deep_element "P" deep_element in
  let bound_name, property_class_opt, string_value =
    match simplify_attributes de_attributes with
    | [] -> fatal_error (Printf.sprintf "%S tag without attribute" "P")
    | [ "Name", name; ] ->
      let string_value = string_of_children "P" de_children in
      name, None, string_value
    | [ "Name", name; "Class", parameter_class; ] ->
      name,
      Some (Mdl_entry_parsers.parameter_class_of_string parameter_class),
      string_of_children "P" de_children
    | [ "Name", name; "Ref", mref; ] ->
      begin
        match de_children with
        | [] -> ()
        | _ ->
          fatal_error (Printf.sprintf
            "found reference P tag named as %S with children" name)
      end;
      name, None, mref
    | _ -> fatal_error "P tag does not have a correct attributes list" in
  let explicit_document_value =
    match property_class_opt with
    | None -> Ddv string_value
    | Some property_class ->
      Ddv_classified (property_class, string_value) in
  {
    Binding.bound_ident = Ast_types.ident_from_string bound_name;
    bound_value = explicit_document_value;
  }
;;

let explicit_classifier element_name local_function f =
  element_name, fun deep_element -> f (local_function deep_element)
;;

let binding_classifier debug f =
  explicit_classifier
    "P" (explicit_binding_of_element ("binding_classifier: " ^ debug)) f
;;

let explicit_bindings debug =
  map_elements (explicit_binding_of_element ("explicit_bindings: " ^ debug))
;;

let explicit_block_parameter_default deep_element =
  let de_attributes, de_children =
    check_deep_element "Block" deep_element in
    let explicit_block_type =
      match de_attributes with
      | [block_type_attribute] ->
        let block_type_string =
          check_deep_attribute "Block" "BlockType" block_type_attribute in
        Mdl_entry_parsers.block_type_value_parser block_type_string
      | _ ->
        fatal_error (Printf.sprintf
          "%S tag inside %S block do not have a correct attributes list"
          "Block" "BlockParameterDefaults")
    and explicit_block_default_bindings =
      explicit_bindings "explicit_block_parameter_default" de_children in
    {
      explicit_block_parameter_default_type = explicit_block_type;
      explicit_block_parameter_default_bindings =
        explicit_block_default_bindings;
    }
;;

let explicit_system_defaults = explicit_bindings "explicit_system_defaults"
and explicit_block_defaults = explicit_bindings "explicit_block_defaults"
and explicit_annotation_defaults =
  explicit_bindings "explicit_annotation_defaults"
and explicit_line_defaults  = explicit_bindings "explicit_line_defaults "
and explicit_mask_defaults = explicit_bindings "explicit_mask_defaults"
and explicit_mask_parameter_defaults =
  explicit_bindings "explicit_mask_parameter_defaults"

and explicit_block_parameter_defaults =
  map_elements explicit_block_parameter_default
;;

let explicit_binding_of_attribute = function
  | {
    da_name = da_name;
    da_value = da_value;
  } ->
    {
      Binding.bound_ident = Ast_types.ident_from_string da_name;
      bound_value = Ddv da_value;
    }
;;

let explicit_mask_defaults_of_annex deep_element =
  let rec loop_attributes bindings = function
    | [] -> bindings
    | attribute :: attributes ->
      let binding = explicit_binding_of_attribute attribute in
      let bindings = binding :: bindings in
      loop_attributes bindings attributes in
  let rec loop_element bindings element =
    let bindings = loop_attributes bindings element.de_attributes in
    let bindings =
      with_elements (loop_elements bindings) element.de_children in
    bindings
  and loop_elements bindings = function
    | [] -> bindings
    | element :: elements ->
      let bindings = loop_element bindings element in
      let bindings = loop_elements bindings elements in
      bindings in
  let bindings = loop_element [] deep_element in
  List.rev bindings
;;

let rec explicit_branch_item deep_element =
  classify_element "Branch" deep_element
    [
      binding_classifier
        "explicit_branch_item"
        (fun binding -> Explicit_branch_binding binding);
      branch_classifier (fun branch -> Explicit_branch_branch branch);
    ]

and explicit_branch deep_element =
  let de_attributes, de_children =
    check_deep_element "Branch" deep_element in
  let connection_type =
    match de_attributes with
    | [] -> None
    | [ connection_type_attribute; ] ->
      Some
        (Mdl_entry_parsers.connection_type_parser
           (check_deep_attribute
              "Branch" "ConnectType" connection_type_attribute))
    | _ ->
      unknown_attribute_list_for_tag "Branch" in
  let branch_items = map_elements explicit_branch_item de_children in
  {
    explicit_branch_connection_type = connection_type;
    explicit_branch_items = branch_items;
  }

and branch_classifier :
    'a. (explicit_branch -> 'a) -> string * (deep_element -> 'a) =
  fun f -> explicit_classifier "Branch" explicit_branch f
;;

let explicit_line_item deep_element =
  classify_element "Line" deep_element
    [
      binding_classifier
        "explicit_line_item"
        (fun binding -> Explicit_line_binding binding);
      branch_classifier (fun branch -> Explicit_line_branch branch);
    ]
;;

let explicit_cell deep_element =
  let de_attributes, de_children =
    check_deep_element "Cell" deep_element in
  let explicit_cell_class_opt =
    match de_attributes with
      | [] -> None
      | [ class_attribute; ] ->
        let class_attribute =
          check_deep_attribute "Cell" "Class" class_attribute in
        let explicit_cell_class = cell_class_of_string class_attribute in
        Some explicit_cell_class
      | _ ->
        unknown_attribute_list_for_tag "Cell" in
  let explicit_cell_value = string_of_children "Cell" de_children in
  {
    explicit_cell_class = explicit_cell_class_opt;
    explicit_cell_value = explicit_cell_value;
  }
;;

let cell_classifier f = explicit_classifier "Cell" explicit_cell f;;

let empty_object_attributes_classifier = {
  explicit_object_id_opt = None;
  explicit_object_class_name_opt = None;
  explicit_object_prop_name_opt = None;
  explicit_object_backup_class_opt = None;
  explicit_object_version_opt = None;
  explicit_object_reference_opt = None;
}
;;

let explicit_field deep_element =
  let de_attributes, de_children =
    check_deep_element "Field" deep_element in
  match de_attributes with
  | [
      field_name_attribute;
      field_class_attribute;
    ] ->
    let field_name =
      check_deep_attribute "Field" "Name" field_name_attribute in
    let field_class =
      field_class_of_string
        (check_deep_attribute "Field" "Class" field_class_attribute) in
    let field_value = string_of_children "Field" de_children in
    {
      Binding.
      bound_ident = Ast_types.ident_from_string field_name;
      bound_value = Ddv_classified (field_class, field_value);
    }
  | _ ->
    unknown_attribute_list_for_tag "Field"
;;

let field_classifier f =
  explicit_classifier "Field" explicit_field f
;;

let explicit_enum_str_pairs deep_element =
  let _de_attributes, de_children =
    check_deep_element "EnumStrPairs" deep_element in
  explicit_bindings "explicit_enum_str_pairs" de_children
;;

let enum_str_pairs_classifier f =
  explicit_classifier "EnumStrPairs" explicit_enum_str_pairs f
;;

let explicit_custom_property_item deep_element =
  classify_element "CustomProperty" deep_element
    [
      binding_classifier
        "explicit_custom_property_item"
        (fun binding -> Explicit_custom_property_item_binding binding);
      enum_str_pairs_classifier
        (fun enum_str_pairs ->
         Explicit_custom_property_item_enum_str_pairs enum_str_pairs);
    ]
;;

let explicit_custom_property_items deep_children =
  map_elements explicit_custom_property_item deep_children
;;

let explicit_custom_property deep_element =
  let de_attributes, de_children =
    check_deep_element "CustomProperty" deep_element in
  match de_attributes with
  | [
      name_attribute;
      data_type_attribute;
    ] ->
    let name =
      Ast_types.ident_from_string
        (check_deep_attribute "CustomProperty" "Name" name_attribute) in
    let data_type =
      prop_data_type_of_string
        (check_deep_attribute
           "CustomProperty" "DataType" data_type_attribute) in
    let items = explicit_custom_property_items de_children in
    {
      explicit_custom_property_name = name;
      explicit_custom_property_data_type = data_type;
      explicit_custom_property_items = items;
    }
  | _ ->
    unknown_attribute_list_for_tag "CustomProperty"
;;

let custom_property_classifier f =
  explicit_classifier "CustomProperty" explicit_custom_property f
;;

let rec explicit_array_item fname deep_element =
  classify_element "Array" deep_element
    [
      object_classifier fname
        (fun e_object -> Explicit_array_item_object e_object);
      cell_classifier
        (fun cell -> Explicit_array_item_cell cell);
      matstruct_fields_classifier fname
        (fun matsruct_fields -> Explicit_array_matstruct matsruct_fields);
      array_classifier fname
        (fun array -> Explicit_array_array array);
    ]

and explicit_array fname deep_element =
  let de_attributes, de_children =
    check_deep_element "Array" deep_element in
  let explicit_array_prop_name, explicit_array_type, explicit_array_dimension =
    match de_attributes with
    | [
        array_prop_name_attribute;
        array_type_attribute;
        array_dimension_attribute;
      ] ->
      check_deep_attribute "Array" "PropName" array_prop_name_attribute,
      check_deep_attribute "Array" "Type" array_type_attribute,
      check_deep_attribute "Array" "Dimension" array_dimension_attribute
    | _ ->
      unknown_attribute_list_for_tag "Array" in
  let explicit_array_type =
    let explicit_array_type =
      Printf.sprintf "%S" explicit_array_type in
    Mdl_entry_parsers.array_type_parser explicit_array_type in
  let explicit_array_dimension =
    dimension_of_string explicit_array_dimension in

  let explicit_array_items =
    map_elements (explicit_array_item fname) de_children in

  {
    explicit_array_type = explicit_array_type;
    explicit_array_dimension = explicit_array_dimension;
    explicit_array_prop_name =
      Ast_types.ident_from_string explicit_array_prop_name;
    explicit_array_items = explicit_array_items;
  }

and explicit_matstruct_item fname deep_element =
  classify_element "MATStruct" deep_element
    [
      field_classifier (fun field -> Explicit_matstruct_field field);
      array_classifier fname (fun array -> Explicit_matstruct_array array);
    ]

and explicit_matstruct_items fname deep_element =
  let de_attributes, de_children =
    check_deep_element "MATStruct" deep_element in
  match de_attributes with
  | [] -> map_elements (explicit_matstruct_item fname) de_children
  | _ ->
    unknown_attribute_list_for_tag "MATStruct"

and matstruct_fields_classifier fname f =
  explicit_classifier "MATStruct" (explicit_matstruct_items fname) f

and explicit_object_item fname deep_element =
  classify_element "Object" deep_element
    [
      binding_classifier
        "explicit_object_item"
        (fun binding -> Explicit_object_binding binding);
      array_classifier fname
        (fun array -> Explicit_object_array array);
      object_classifier fname
        (fun explicit_object -> Explicit_object_object explicit_object);
      custom_property_classifier
        (fun custom_property ->
         Explicit_object_custom_property custom_property);
    ]

and explicit_object fname deep_element =
  let de_attributes, de_children =
    check_deep_element "Object" deep_element in
  let classify_attribute classifier deep_attribute =
    let check_opt_is_empty opt =
      check_opt_is_empty
        (Printf.sprintf "attribute %S of %S tag set twice"
           deep_attribute.da_name "Object") opt in
    match deep_attribute.da_name with
    | "ObjectID" ->
      check_opt_is_empty classifier.explicit_object_id_opt;
      let explicit_object_nat = unatural_of_string deep_attribute.da_value in
      let explicit_object_fname = fname in
      let explicit_object_id = explicit_object_fname, explicit_object_nat in
      {
        classifier with
          explicit_object_id_opt = Some explicit_object_id;
      }
    | "ClassName" ->
      check_opt_is_empty classifier.explicit_object_class_name_opt;
      let explicit_object_class_name = deep_attribute.da_value in
      {
        classifier with
          explicit_object_class_name_opt =
            Some (Ast_types.mdl_name_of_string explicit_object_class_name);
      }
    | "PropName" ->
      check_opt_is_empty classifier.explicit_object_prop_name_opt;
      let explicit_object_prop_name = deep_attribute.da_value in
      {
        classifier with
          explicit_object_prop_name_opt =
            Some (Ast_types.ident_from_string explicit_object_prop_name);
      }
    | "BackupClass" ->
      check_opt_is_empty classifier.explicit_object_backup_class_opt;
      let explicit_object_backup_class = deep_attribute.da_value in
      {
        classifier with
          explicit_object_backup_class_opt =
          Some (Ast_types.mdl_name_of_string explicit_object_backup_class);
      }
    | "Version" ->
      check_opt_is_empty classifier.explicit_object_version_opt;
      let explicit_object_version = deep_attribute.da_value in
      {
        classifier with
          explicit_object_version_opt = Some explicit_object_version;
      }
    | "Reference" ->
      check_opt_is_empty classifier.explicit_object_reference_opt;
      let explicit_object_reference = deep_attribute.da_value in
      {
        classifier with
          explicit_object_reference_opt = Some explicit_object_reference;
      }
    | da_name ->
      fatal_error (Printf.sprintf
        "unknown %S attribute of %S tag" da_name "Object") in

  let classifier =
    List.fold_left
      classify_attribute empty_object_attributes_classifier de_attributes in

  let explicit_object_id =
    match classifier.explicit_object_id_opt with
    | None -> fatal_error "object without id detected"
    | Some explicit_object_id -> explicit_object_id in

  let explicit_object_items =
    map_elements (explicit_object_item fname) de_children in

  {
    explicit_object_id = explicit_object_id;
    explicit_object_class_name = classifier.explicit_object_class_name_opt;
    explicit_object_prop_name = classifier.explicit_object_prop_name_opt;
    explicit_object_backup_class =
      classifier.explicit_object_backup_class_opt;
    explicit_object_version = classifier.explicit_object_version_opt;
    explicit_object_reference = classifier.explicit_object_reference_opt;
    explicit_object_items = explicit_object_items;
  }

and array_classifier :
    'a. Path.file_name -> (explicit_array -> 'a) -> string * (deep_element -> 'a) =
  fun fname f -> explicit_classifier "Array" (explicit_array fname) f
and object_classifier :
    'b. Path.file_name -> (explicit_object -> 'b) -> string * (deep_element -> 'b) =
  fun fname f -> explicit_classifier "Object" (explicit_object fname) f
;;

(* Turn an XML tag attribute into an equivalent XML element:
   from attribute_name="3.14" to the deep child obtained from treating XML
   element:
   <P Name="attribute_name">3.14</P>
   In short, we treat XML source
   <Tag attribute="pcdata"> ...
   <Tag>
     <P>Name="attribute">pcdata</P>
*)
let deep_element_of_attribute = function
  | {
      da_name = da_name;
      da_value = da_value;
    } ->
    {
      de_name = "P";
      de_attributes = [
        {
          da_name = "Name";
          da_value = da_name;
        };
      ];
      de_children = [
        Intercal.In_left da_value;
      ];
    }
;;

let child_of_attribute attribute =
  Intercal.In_right (deep_element_of_attribute attribute)
;;

let explicit_block_list deep_element =
  let de_attributes, de_children =
    check_deep_element "List" deep_element in
  let explicit_list_type, de_attributes =
    match de_attributes with
    | list_type_attribute :: de_attributes ->
      check_deep_attribute "List" "ListType" list_type_attribute,
      de_attributes
    | _ ->
      unknown_attribute_list_for_tag "List" in
  let attributes_children = List.rev_map child_of_attribute de_attributes in
  let de_children =
    List.rev_append attributes_children de_children in
  let explicit_bindings =
    explicit_bindings "explicit_block_list" de_children in
  {
    explicit_list_type = explicit_list_type;
    explicit_list_bindings = explicit_bindings;
  }
;;

let list_classifier f = explicit_classifier "List" explicit_block_list f;;

let explicit_port_item fname deep_element =
  classify_element "Port" deep_element
    [
      binding_classifier
        "explicit_port_item"
        (fun binding -> Explicit_port_binding binding);
      array_classifier fname (fun array -> Explicit_port_array array);
    ]
;;

let explicit_port_items fname deep_children =
  map_elements (explicit_port_item fname) deep_children
;;

let explicit_block_port fname deep_element =
  let _de_attributes, de_children =
    check_deep_element "Port" deep_element in
  explicit_port_items fname de_children
;;

let block_port_classifier fname f =
  explicit_classifier "Port" (explicit_block_port fname) f
;;

let explicit_instance_data fname deep_element =
  let element_name = "InstanceData" in
  let de_attributes, de_children =
    check_deep_element element_name deep_element in
  check_no_attributes element_name de_attributes;
  explicit_bindings fname de_children
;;

let instance_data_classifier fname f =
  explicit_classifier "InstanceData" (explicit_instance_data fname) f
;;

let explicit_system_line deep_element =
  let de_attributes, de_children =
    check_deep_element "Line" deep_element in
  let line_type =
    match de_attributes with
    | [] -> None
    | [ line_type_attribute; ] ->
      Some
        (Mdl_entry_parsers.line_type_parser
           (check_deep_attribute "Line" "LineType" line_type_attribute))
    | _ ->
      fatal_error (Printf.sprintf
        "%S tag with an unknown attributes list" "Line") in
  let items = map_elements explicit_line_item de_children in
  {
    explicit_line_type = line_type;
    explicit_line_items = items;
  }
;;

let line_classifier f =
  explicit_classifier "Line" explicit_system_line f
;;

let explicit_system_annotation deep_element =
  let de_attributes, de_children =
    check_deep_element "Annotation" deep_element in
  let explicit_annotation_sid =
    match de_attributes with
    | [sid_attribute;] ->
      let sid_value = check_deep_attribute "Annotation" "SID" sid_attribute in
      block_sid_value sid_value
    | _ ->
      fatal_error (Printf.sprintf
        "%S tag with an unknown attributes list" "Annotation")
  and explicit_annotation_bindings =
    explicit_bindings "explicit_system_annotation" de_children in
  {
    explicit_annotation_sid = explicit_annotation_sid;
    explicit_annotation_bindings = explicit_annotation_bindings;
  }
;;

let annotation_classifier f =
  explicit_classifier "Annotation" explicit_system_annotation f
;;

let explicit_connector deep_element =
  let de_attributes, de_children =
    check_deep_element "Connector" deep_element in
  match de_attributes with
  | [] -> explicit_bindings "Connector" de_children
  | _ ->
    unknown_attribute_list_for_tag "Connector"


let connector_classifier f =
  explicit_classifier "Connector" explicit_connector f
;;

let explicit_dialog_parameters deep_element =
  let de_attributes, de_children =
    check_deep_element dialog_parameters_tag_name deep_element in
  let block_name =
    match de_attributes with
    | [ name_attribute; ] ->
      mdl_name_of_string
        (check_deep_attribute
           dialog_parameters_tag_name "BlockName" name_attribute)
    | _ ->
    unknown_attribute_list_for_tag dialog_parameters_tag_name in
  let bindings =
    explicit_bindings dialog_parameters_tag_name de_children in
  {
    explicit_dialog_parameters_block_name = block_name;
    explicit_dialog_parameters_bindings = bindings;
  }
;;

let explicit_link_data deep_element =
  let _de_attributes, de_children =
    check_deep_element "LinkData" deep_element in
  map_elements explicit_dialog_parameters de_children
;;

let link_data_classifier f =
  explicit_classifier "LinkData" explicit_link_data f
;;

let explicit_description = string_of_element "Description"
and explicit_initialization = string_of_element "Initialization"
and explicit_help = string_of_element "Help"
and explicit_type = string_of_element "Type"
;;

let classify_mask_parameter_attribute classifier attribute =
  match attribute.da_name with
  | "Name" ->
    { classifier with
      explicit_mask_parameter_name_opt =
        set_opt
          "Name found twice"
          classifier.explicit_mask_parameter_name_opt
          attribute.da_value;
    }
  | "Type" ->
    { classifier with
      explicit_mask_parameter_type_opt =
        set_opt
          "Type found twice"
          classifier.explicit_mask_parameter_type_opt
          attribute.da_value;
    }
  | "Tunable" ->
    { classifier with
      explicit_mask_parameter_tunable_opt =
        set_opt
          "Tunable found twice"
          classifier.explicit_mask_parameter_tunable_opt
          attribute.da_value;
    }
  | "Evaluate" ->
    { classifier with
      explicit_mask_parameter_evaluate_opt =
        set_opt
          "Evaluate found twice"
          classifier.explicit_mask_parameter_evaluate_opt
          attribute.da_value;
    }
  | "Enabled" ->
    { classifier with
      explicit_mask_parameter_enabled_opt =
        set_opt
          "Enabled found twice"
          classifier.explicit_mask_parameter_enabled_opt
          attribute.da_value;
    }
  | "Visible" ->
    { classifier with
      explicit_mask_parameter_visible_opt =
        set_opt
          "Visible found twice"
          classifier.explicit_mask_parameter_visible_opt
          attribute.da_value;
    }
  | "Internal" ->
    { classifier with
      explicit_mask_parameter_internal_opt =
        set_opt
          "Internal found twice"
          classifier.explicit_mask_parameter_internal_opt
          attribute.da_value;
    }
  | "NeverSave" ->
    { classifier with
      explicit_mask_parameter_never_save_opt =
        set_opt
          "NeverSave found twice"
          classifier.explicit_mask_parameter_never_save_opt
          attribute.da_value;
    }
  | "ReadOnly" ->
    { classifier with
      explicit_mask_parameter_read_only_opt =
        set_opt
          "ReadOnly found twice"
          classifier.explicit_mask_parameter_read_only_opt
          attribute.da_value;
    }
  | da_name ->
    fatal_error
      (Printf.sprintf
         "unknown %S attribute of %S tag"
         da_name "MaskParameter")
;;

let explicit_mask_parameter deep_element =
  let de_attributes, de_children =
    check_deep_element "MaskParameter" deep_element in
  let classifier =
    List.fold_left
      classify_mask_parameter_attribute
      empty_mask_parameter_classifier
      de_attributes in
  let rec classify_children p_opt v_opt c_opt = function
    | [] -> p_opt, v_opt, c_opt
    | child :: children ->
      let p_opt, v_opt, c_opt =
        match child.de_name with
        | "Prompt" ->
          let prompt = string_of_element "Prompt" child in
          set_opt "Prompt found twice" p_opt prompt,
          v_opt,
          c_opt
        | "Value" ->
          let value = string_of_element "Value" child in
          p_opt,
          set_opt "Value found twice" v_opt value,
          c_opt
        | "Callback" ->
          let callback = string_of_element "Callback" child in
          p_opt,
          v_opt,
          set_opt "Callback found twice" c_opt callback
        | "TypeOptions" ->
          warning (Printf.sprintf
            "dropping %S from MaskParameter" "TypeOptions");
          p_opt, v_opt, c_opt
        | _ -> fail_unknown_tag "MaskParameter" child in
      classify_children p_opt v_opt c_opt children in
  let p_opt, v_opt, c_opt =
    with_elements
      (fun elements ->
       classify_children None None None elements)
      de_children in
  match classifier with
  | {
    explicit_mask_parameter_name_opt = name_opt;
    explicit_mask_parameter_type_opt = type_opt;
    explicit_mask_parameter_tunable_opt = tunable_opt;
    explicit_mask_parameter_evaluate_opt = evaluate_opt;
    explicit_mask_parameter_enabled_opt = enabled_opt;
    explicit_mask_parameter_internal_opt = internal_opt;
    explicit_mask_parameter_visible_opt = visible_opt;
    explicit_mask_parameter_never_save_opt = never_save_opt;
    explicit_mask_parameter_read_only_opt = read_only_opt;
  } ->
    let explicit_mask_parameter_name =
      Ast_types.mdl_name_of_string
        (get_option_value
           "incomplete mask parameter" "Name" name_opt) in
    let get_option_value missing_field opt =
      let message =
        Printf.sprintf
          "incomplete mask parameter %S"
          (Ast_types.string_of_mdl_name explicit_mask_parameter_name) in
      (get_option_value message missing_field opt) in
    let explicit_mask_parameter_type =
      get_option_value "Type" type_opt in
    let explicit_mask_parameter_prompt =
      match p_opt with
      | None -> "" (* yes they are mask parameters without a prompt *)
      | Some p -> p in
    let explicit_mask_parameter_callback = c_opt in
    let explicit_mask_parameter_tunable = tunable_opt in
    let explicit_mask_parameter_value =
      match v_opt with
      | None -> "" (* yes they are mask parameters without a value *)
      | Some v -> v in
    let explicit_mask_parameter_enabled = enabled_opt in
    let explicit_mask_parameter_visible = visible_opt in
    let explicit_mask_parameter_evaluate = evaluate_opt in
    let explicit_mask_parameter_internal = internal_opt in
    let explicit_mask_parameter_never_save = never_save_opt in
    let explicit_mask_parameter_read_only = read_only_opt in
    {
      explicit_mask_parameter_name =
        explicit_mask_parameter_name;
      explicit_mask_parameter_type =
        explicit_mask_parameter_type;
      explicit_mask_parameter_prompt =
        explicit_mask_parameter_prompt;
      explicit_mask_parameter_value =
        explicit_mask_parameter_value;
      explicit_mask_parameter_callback =
        explicit_mask_parameter_callback;
      explicit_mask_parameter_tunable =
        explicit_mask_parameter_tunable;
      explicit_mask_parameter_evaluate =
        explicit_mask_parameter_evaluate;
      explicit_mask_parameter_enabled =
        explicit_mask_parameter_enabled;
      explicit_mask_parameter_visible =
        explicit_mask_parameter_visible;
      explicit_mask_parameter_internal =
        explicit_mask_parameter_internal;
      explicit_mask_parameter_never_save =
        explicit_mask_parameter_never_save;
      explicit_mask_parameter_read_only =
        explicit_mask_parameter_read_only;
    }
;;

let explicit_display deep_element =
  let de_attributes, de_children =
    check_deep_element display_tag_name deep_element in
  let classify_attribute classifier attribute =
    match attribute.da_name with
    | "IconFrame" ->
      { classifier with
        explicit_display_icon_frame =
          set_opt
            "IconFrame found twice"
            classifier.explicit_display_icon_frame
            attribute.da_value;
      }
    | "IconUnits" ->
      { classifier with
        explicit_display_icon_units =
          set_opt
            "IconUnits found twice"
            classifier.explicit_display_icon_units
            attribute.da_value;
      }
    | "IconOpaque" ->
      { classifier with
        explicit_display_icon_opaque =
          set_opt
            "IconOpaque found twice"
            classifier.explicit_display_icon_opaque
            attribute.da_value;
      }
    | "IconRotate" ->
      { classifier with
        explicit_display_icon_rotate =
          set_opt
            "IconRotate found twice"
            classifier.explicit_display_icon_rotate
            attribute.da_value;
      }
    | "RunInitForIconRedraw" ->
      { classifier with
        explicit_display_run_init_for_icon_redraw =
          set_opt
            "RunInitForIconRedraw found twice"
            classifier.explicit_display_run_init_for_icon_redraw
            attribute.da_value;
      }
    | da_name ->
      fatal_error
        (Printf.sprintf
           "unknown %S attribute of %S tag"
           da_name display_tag_name) in
  let classifier =
    List.fold_left
      classify_attribute
      empty_display_attribute_classifier
      de_attributes in
  let script = string_of_children display_tag_name de_children in
  match classifier with
  | {
    explicit_display_icon_frame = icon_frame;
    explicit_display_icon_units = icon_units;
    explicit_display_icon_opaque = icon_opaque;
    explicit_display_icon_rotate = icon_rotate;
    explicit_display_run_init_for_icon_redraw = run_init_for_icon_redraw;
  } ->
    {
      explicit_mask_display_icon_frame = icon_frame;
      explicit_mask_display_icon_units = icon_units;
      explicit_mask_display_icon_opaque = icon_opaque;
      explicit_mask_display_icon_rotate = icon_rotate;
      explicit_mask_display_run_init_for_icon_redraw =
        run_init_for_icon_redraw;
      explicit_mask_display_script = Some script;
    }
;;

let classify_mask_element classifier deep_element =
  match deep_element.de_name with
  | de_name when de_name = display_tag_name ->
    let display = explicit_display deep_element in
    { classifier with
      explicit_mask_display_opt =
        set_opt
          (Printf.sprintf "%s found twice" display_tag_name)
          classifier.explicit_mask_display_opt
          display
    }
  | "Description" ->
    let description = explicit_description deep_element in
    { classifier with
      explicit_mask_description_opt =
        set_opt
          "Description found twice"
          classifier.explicit_mask_description_opt
          description
    }
  | "Initialization" ->
    let initialization = explicit_initialization deep_element in
    { classifier with
      explicit_mask_initialisation_opt =
        set_opt
          "Initialization found twice"
          classifier.explicit_mask_initialisation_opt
          initialization
    }
  | "Help" ->
    let help = explicit_help deep_element in
    { classifier with
      explicit_mask_help_opt =
        set_opt
          "Help found twice"
          classifier.explicit_mask_help_opt
          help
    }
  | "Type" ->
    let mask_type = explicit_type deep_element in
    { classifier with
      explicit_mask_type_opt =
        set_opt
          "Mask_Type found twice"
          classifier.explicit_mask_type_opt
          mask_type
    }
  | "MaskParameter" ->
    let mask_parameter = explicit_mask_parameter deep_element in
    { classifier with
      explicit_mask_parameters_opt =
        mask_parameter :: classifier.explicit_mask_parameters_opt;
    }
  | "Capabilities"
  | "DialogControl" ->
    warning
      (Printf.sprintf "dropping %S from Mask" deep_element.de_name);
    classifier
  | _ -> fail_unknown_tag mask_tag_name deep_element
;;

let explicit_mask deep_element =
  let _de_attributes, de_children =
    check_deep_element mask_tag_name deep_element in
  let classifier =
    with_elements
      (fun deep_elements ->
       List.fold_left
         classify_mask_element
         empty_explicit_mask_classifier
         deep_elements)
      de_children in
  let classifier =
    { classifier with
      explicit_mask_parameters_opt =
        List.rev classifier.explicit_mask_parameters_opt;
    } in
  let explicit_mask_display =
    get_option_value
      "incomplete mask" display_tag_name
      classifier.explicit_mask_display_opt in
  {
    explicit_mask_type = classifier.explicit_mask_type_opt;
    explicit_mask_display = explicit_mask_display;
    explicit_mask_description =
      classifier.explicit_mask_description_opt;
    explicit_mask_initialisation =
      classifier.explicit_mask_initialisation_opt;
    explicit_mask_help = classifier.explicit_mask_help_opt;
    explicit_mask_parameters = classifier.explicit_mask_parameters_opt;
  }
;;

let mask_classifier f =
  explicit_classifier mask_tag_name explicit_mask f
;;

let rec explicit_block fname deep_element =
  let de_attributes, de_children =
    check_deep_element "Block" deep_element in
    let explicit_block_type, block_name, explicit_block_sid =
      match de_attributes with
      | [
          block_type_attribute;
          name_attribute;
          sid_attribute;
        ] ->
        Mdl_entry_parsers.block_type_value_parser
          (check_deep_attribute "Block" "BlockType" block_type_attribute),
        mdl_name_of_string
          (check_deep_attribute "Block" "Name" name_attribute),
        block_sid_value (check_deep_attribute "Block" "SID" sid_attribute)
      | _ ->
        fatal_error (Printf.sprintf
          "%S tag with an unknown attributes list" "Block") in
    let explicit_block_items = explicit_block_items fname de_children in
    {
      explicit_block_type = explicit_block_type;
      explicit_block_name = block_name;
      explicit_block_sid = explicit_block_sid;
      explicit_block_items = explicit_block_items;
    }

and explicit_block_item fname deep_element =
  classify_element "Block" deep_element
    [
      binding_classifier
        "explicit_block_item"
        (fun binding -> Explicit_block_binding binding);
      object_classifier fname (fun e_object -> Explicit_block_object e_object);
      system_classifier fname (fun system -> Explicit_block_system system);
      list_classifier (fun list -> Explicit_block_list list);
      block_port_classifier fname (fun port -> Explicit_block_port port);
      array_classifier fname (fun array -> Explicit_block_array array);
      instance_data_classifier fname
        (fun instance_data -> Explicit_block_instance_data instance_data);
      link_data_classifier
        (fun link_data -> Explicit_block_link_data link_data);
      mask_classifier (fun mask -> Explicit_block_mask mask);
    ]

and explicit_block_items fname deep_children =
  map_elements (explicit_block_item fname) deep_children

and explicit_system fname deep_element =

  let _de_attributes, de_children =
    check_deep_element "System" deep_element in

  let rec classify
          position_opt bindings blocks lines annotations connectors =
    function
    | [] ->
      position_opt,
      List.rev bindings,
      List.rev blocks,
      List.rev lines,
      List.rev annotations,
      List.rev connectors
    | deep_element :: deep_elements ->
      classify_element "System" deep_element [
        binding_classifier
          "explicit_system"
          (fun binding ->
           match Ast_types.string_of_ident binding.Binding.bound_ident with
           | "Location" ->
              fdebug
                (fun ppf ->
                 Format.fprintf ppf
                  "explicit_system: classify: Location read %a"
                  (Binding_print.printer_ident_binding
                     Xml_explicit_ast_print.printer_explicit_document_value
                     Og_precedence.Ck_argument)
                  binding);
             check_opt_is_empty (Printf.sprintf
               "%S binding for %S tag is defined twice" "Location" "System")
               position_opt;
             let position_opt =
               match binding.Binding.bound_value with
               | Ddv text -> Some text
               | Ddv_classified _ ->
                 fatal_error "classified location" in
             classify
               position_opt bindings blocks lines
               annotations connectors deep_elements
           | _ ->
             let bindings = binding :: bindings in
             classify
               position_opt bindings blocks lines
               annotations connectors deep_elements);
        block_classifier fname (fun block ->
          let blocks = block :: blocks in
          classify
            position_opt bindings blocks lines
            annotations connectors deep_elements);
        line_classifier (fun line ->
          let lines = line :: lines in
          classify
            position_opt bindings blocks lines
            annotations connectors deep_elements);
        annotation_classifier (fun annotation ->
          let annotations = annotation :: annotations in
          classify
            position_opt bindings blocks lines
            annotations connectors deep_elements);
        connector_classifier (fun connector ->
          let connectors = connector :: connectors in
          classify
            position_opt bindings blocks lines
            annotations connectors deep_elements);
      ] in

  with_elements (fun deep_elements ->
    let position_opt, bindings, blocks, lines, annotations, connectors  =
      classify None [] [] [] [] [] deep_elements in
    match position_opt with
    | None ->
      fatal_error (Printf.sprintf
        "%S tag with no Location" "System")
    | Some position ->
      {
        explicit_system_position = position;
        explicit_system_bindings = bindings;
        explicit_system_blocks = blocks;
        explicit_system_connections = lines;
        explicit_system_annotations = annotations;
        explicit_system_connectors = connectors;
      })
    de_children

and system_classifier :
    'a. Path.file_name -> (explicit_system -> 'a) -> string * (deep_element -> 'a) =
  fun fname f -> explicit_classifier "System" (explicit_system fname) f

and block_classifier :
    'a. Path.file_name -> (explicit_block -> 'a) -> string * (deep_element -> 'a) =
  fun fname f -> explicit_classifier "Block" (explicit_block fname) f
;;

let explicit_configuration_set deep_elements =
  let fname = Configuration.default_slx_blockdiagram_file_name in
  let rec loop cfsets acf_opt = function
  | [] -> List.rev cfsets, acf_opt
  | deep_element :: deep_elements ->
    match deep_element with
    | {
        de_name = "Array";
        de_attributes =
          {
            da_name = "PropName";
            da_value = "ConfigurationSets";
          } :: _deep_attributes;
        de_children = _de_children;
      } ->
      let explicit_array = explicit_array fname deep_element in
      let cfsets = explicit_array :: cfsets in
      loop cfsets acf_opt deep_elements
    | _ ->
      let explicit_object = explicit_object fname deep_element in
      begin
        match explicit_object.explicit_object_prop_name with
        | Some (Ident.Ident "ActiveConfigurationSet") -> ()
        | Some (Ident.Ident prop_name) ->
          fatal_error (Printf.sprintf
            "unknown object %S found in ConfigurationSet"
            prop_name)
        | None ->
          fatal_error (Printf.sprintf
            "object without PropName found in ConfigurationSet")
      end;
      begin
        match explicit_object.explicit_object_class_name with
        | Some (Mdl_name "Simulink.ConfigSet") -> ()
        | Some (Mdl_name class_name) ->
          fatal_error (Printf.sprintf
            "unknown object class %S found in ConfigurationSet"
            class_name)
        | None ->
          fatal_error (Printf.sprintf
            "object without class found in ConfigurationSet")
      end;
      check_opt_is_empty
        "two active configuration found in ConfigurationSet"
        acf_opt;
      let acf_opt = Some explicit_object in
      loop cfsets acf_opt deep_elements in

  loop [] None deep_elements
;;

let explicit_model_workspace = explicit_bindings "explicit_model_workspace";;

let empty_explicit_block_diagram_defaults_classifier = {
  explicit_system_defaults_opt = None;
  explicit_block_defaults_opt = None;
  explicit_annotation_defaults_opt = None;
  explicit_line_defaults_opt = None;
  explicit_mask_defaults_opt = None;
  explicit_mask_parameter_defaults_opt = None;
  explicit_block_parameter_defaults_opt = None;
}
;;

let empty_implementation_section_items_classifier = {
  explicit_model_workspace_opt = None;
  explicit_model_configuration_opt = None;
  explicit_configuration_set_object_opt = None;
  explicit_configuration_set_binding_list = [];
  explicit_block_diagram_defaults_classifier =
    empty_explicit_block_diagram_defaults_classifier;
  explicit_system_opt = None;
}
;;

let with_bdd_classifier classifier f =
  { classifier with
    explicit_block_diagram_defaults_classifier =
      f classifier.explicit_block_diagram_defaults_classifier;
  }
;;

let classify_implementation_section_item classifier deep_element =
  let file_name = Configuration.default_slx_blockdiagram_file_name in
  match deep_element.de_name with
  | "P" ->
    let binding =
      explicit_binding_of_element "implementation_section_item" deep_element in
    {
      classifier with
      explicit_configuration_set_binding_list =
        binding :: classifier.explicit_configuration_set_binding_list;
    }
  | "ConfigurationSet" ->
    check_opt_is_empty
      "two ConfigurationSets elements read"
      classifier.explicit_model_configuration_opt;
    check_opt_is_empty
      "two ActiveConfigurationSet elements read"
      classifier.explicit_configuration_set_object_opt;
    let explicit_model_configuration, explicit_configuration_set_object_opt =
      with_elements
        explicit_configuration_set deep_element.de_children in
    {
      classifier with
      explicit_model_configuration_opt =
        Some explicit_model_configuration;
      explicit_configuration_set_object_opt =
        explicit_configuration_set_object_opt;
    }
  | "SystemDefaults" ->
    check_opt_is_empty
      "two SystemDefaults elements read"
      classifier
      .explicit_block_diagram_defaults_classifier
      .explicit_system_defaults_opt;
    let explicit_system_defaults =
      explicit_system_defaults deep_element.de_children in
    with_bdd_classifier
      classifier
      (fun bdd ->
       { bdd with
         explicit_system_defaults_opt =
           Some explicit_system_defaults;
       })
  | "BlockDefaults" ->
    check_opt_is_empty
      "two BlockDefaults elements read"
      classifier
      .explicit_block_diagram_defaults_classifier
      .explicit_block_defaults_opt;
    let explicit_block_defaults =
      explicit_block_defaults deep_element.de_children in
    with_bdd_classifier
      classifier
      (fun bdd ->
       { bdd with
         explicit_block_defaults_opt =
           Some explicit_block_defaults;
       })
  | "AnnotationDefaults" ->
    check_opt_is_empty
      "two AnnotationDefaults elements read"
      classifier
      .explicit_block_diagram_defaults_classifier
      .explicit_annotation_defaults_opt;
    let explicit_annotation_defaults =
      explicit_annotation_defaults deep_element.de_children in
    with_bdd_classifier
      classifier
      (fun bdd ->
       { bdd with
         explicit_annotation_defaults_opt =
           Some explicit_annotation_defaults;
       })
  | "LineDefaults" ->
    check_opt_is_empty
      "two LineDefaults elements read"
      classifier
      .explicit_block_diagram_defaults_classifier
      .explicit_line_defaults_opt;
    let explicit_line_defaults =
      explicit_line_defaults deep_element.de_children in
    with_bdd_classifier
      classifier
      (fun bdd ->
       { bdd with
         explicit_line_defaults_opt = Some explicit_line_defaults;
       })
  | "MaskDefaults" ->
    check_opt_is_empty
      "two MaskDefaults elements read"
      classifier
      .explicit_block_diagram_defaults_classifier
      .explicit_mask_defaults_opt;
    let explicit_mask_defaults =
      explicit_mask_defaults deep_element.de_children in
    with_bdd_classifier
      classifier
      (fun bdd ->
       { bdd with
         explicit_mask_defaults_opt = Some explicit_mask_defaults;
       })
  | "MaskParameterDefaults" ->
    check_opt_is_empty
      "two MaskParameterDefaults elements read"
      classifier
      .explicit_block_diagram_defaults_classifier
      .explicit_mask_parameter_defaults_opt;
    let explicit_mask_parameter_defaults =
      explicit_mask_parameter_defaults deep_element.de_children in
    with_bdd_classifier
      classifier
      (fun bdd ->
       { bdd with
         explicit_mask_parameter_defaults_opt =
           Some explicit_mask_parameter_defaults;
       })
  | "BlockParameterDefaults" ->
    check_opt_is_empty
      "two BlockParameterDefaults elements read"
      classifier
      .explicit_block_diagram_defaults_classifier
      .explicit_block_parameter_defaults_opt;
    let explicit_block_parameter_defaults =
      explicit_block_parameter_defaults deep_element.de_children in
    with_bdd_classifier
      classifier
      (fun bdd ->
       { bdd with
         explicit_block_parameter_defaults_opt =
           Some explicit_block_parameter_defaults;
       })
  | "System" ->
    check_opt_is_empty
      "two System elements read"
      classifier.explicit_system_opt;
    let explicit_system = explicit_system file_name deep_element in
    {
      classifier with
        explicit_system_opt = Some explicit_system;
    }
  | "ModelWorkspace" ->
    check_opt_is_empty
      "two ModelWorkspace elements read"
      classifier.explicit_model_workspace_opt;
    let explicit_model_workspace =
      explicit_model_workspace deep_element.de_children in
    {
      classifier with
      explicit_model_workspace_opt = Some explicit_model_workspace;
    }
  | "GraphicalInterface"
  | "Object"
  | "ConfigManagerSettings"
  | "EditorSettings"
  | "SimulationSettings"
  | "Verification"
  | "ExternalMode"
  | "EngineSettings"
  | "ModelReferenceSettings"
  | "ConcurrentExecutionSettings"
  | "UserParameters" ->
    warning (Printf.sprintf
      "dropping %S implementation section item"
      deep_element.de_name);
    classifier
  | de_name ->
    fatal_error (Printf.sprintf
      "%S tag inside section item: not supported" de_name)
;;

let is_corresponding_annex_name annex_name part_name =
  fdebug
    (fun epf ->
     Format.fprintf epf
       "is_corresponding_annex_name %a %S"
       Xml_deep_ast_print.print_deep_annex_name annex_name
       part_name);
  match annex_name with
  | Deep_block_diagram_defaults
  | Deep_config_set_info
  | Deep_graphical_interface
  | Deep_stateflow -> false
  | Deep_annex_file_name fname ->
    let part_name =
      let len = String.length part_name in
      if len >= 1 && part_name.[0] = '/' then
        String.sub part_name 1 (pred len)
      else
        part_name in
  fdebug
    (fun epf ->
     Format.fprintf epf
       "is_corresponding_annex_name string_comparison %S %S"
       fname
       part_name);
    Path.eq_file_name fname part_name
;;

let insert_object_id deep_children =
  let rec loop accu deep_element_count = function
    | [] ->
      if deep_element_count = 1 then
        List.rev accu
      else if deep_element_count = 0 then
        fatal_error "trying to insert ObjectId, but no object was found"
      else
        assert false
    | deep_child :: deep_children ->
      let deep_child, deep_element_count =
        match deep_child with
        | Intercal.In_left char_data ->
          Intercal.In_left char_data, deep_element_count
        | Intercal.In_right deep_element ->
          if deep_element_count >= 1 then
            fatal_error "too many ConfigSet when inserting ObjectId"
          else
            let object_id_attribute = {
              da_name = "ObjectID";
              da_value = Unatural.to_string default_config_set_object_id;
            } in
            let deep_element = {
              deep_element with
              de_attributes = object_id_attribute :: deep_element.de_attributes;
            } in
            Intercal.In_right deep_element, 1 in
        loop (deep_child :: accu) deep_element_count deep_children in
  loop [] 0 deep_children
;;

let file_name_of_deep_annex_name = function
  | Xml_deep_ast.Deep_block_diagram_defaults ->
    Configuration.default_slx_block_diagram_defaults_file_name
  | Xml_deep_ast.Deep_config_set_info ->
    Configuration.default_slx_config_set_info_file_name
  | Xml_deep_ast.Deep_graphical_interface ->
    Configuration.default_slx_graphical_interface_file_name
  | Xml_deep_ast.Deep_stateflow ->
    Configuration.default_slx_stateflow_file_name
  | Xml_deep_ast.Deep_annex_file_name fname ->
    fname
;;

let config_set_of_annex_document (deep_annex_name, annex_document) =
  let element_name = "ConfigSet" in
  let de_attributes, de_children =
    check_deep_element element_name annex_document.deep_element in
  begin
    match de_attributes with
    | [] -> ()
    | _ ->
      unknown_attribute_list_for_element element_name
  end;

  let fname = file_name_of_deep_annex_name deep_annex_name in

  fdebug
    (fun epf ->
     Format.fprintf epf "inserting object id into:@ %a"
       Xml_deep_ast_print.print_deep_children de_children);
  let de_children = insert_object_id de_children in

  let explicit_objects =
    map_elements (explicit_object fname) de_children in
  let explicit_array_items =
    List.rev
      (List.rev_map
         (fun explicit_object ->
          Explicit_array_item_object explicit_object)
         explicit_objects) in
  let explicit_array = {
    explicit_array_type = Mdl_entry_parsers.array_type_parser "\"Handle\"";
    explicit_array_dimension =
      dimension_of_string
        (Printf.sprintf "1*%d" (List.length explicit_objects));
    explicit_array_prop_name =
      Ast_types.ident_from_string "ConfigurationSets";
    explicit_array_items = explicit_array_items;
  } in
  explicit_array
;;

let get_config_set_of_annex deep_annexes part_name =
  try
    let annex =
      List.find
        (fun (annex_name, _) ->
         is_corresponding_annex_name annex_name part_name)
        deep_annexes in
    config_set_of_annex_document annex
  with
  | Not_found ->
    fatal_error (Printf.sprintf "config set %s not found" part_name)
;;

let part_name_active deep_element =
  let element_name = "ConfigSet" in
  let de_attributes, de_children =
    check_deep_element element_name deep_element in
  (* We want to be sure that children are just texts elements, so we retrieve
     them as a string. *)
  ignore (string_of_children element_name de_children);
  let part_name, active =
    match de_attributes with
    | [
        part_name_attribute;
        active_attribute;
      ] ->
      let part_name =
        check_deep_attribute element_name "PartName" part_name_attribute in
      let active =
        check_deep_attribute element_name "Active" active_attribute in
      part_name, active
    | [
        part_name_attribute;
      ] ->
      let part_name =
        check_deep_attribute element_name "PartName" part_name_attribute in
      part_name, "false"
    | _ ->
      unknown_attribute_list_for_element element_name in
  part_name, active
;;

let part_name_active_classifier f =
  explicit_classifier "ConfigSet" part_name_active f
;;

let part_name_active_item deep_element =
  classify_element "ConfigSetInfo" deep_element
    [
      part_name_active_classifier
        (fun (part_name, active) -> part_name, active);
    ]
;;

let part_name_active_items deep_children =
  map_elements part_name_active_item deep_children
;;

let get_part_name_active deep_annexes =
  let element_name = "ConfigSetInfo" in
  let rec loop = function
    | [] -> fatal_error "config_set_info not found"
    | (Deep_config_set_info, deep_document) :: _ ->
      let de_attributes, de_children =
        check_deep_element element_name deep_document.deep_element in
      begin
        match de_attributes with
        | [] -> ()
        | _ ->
          unknown_attribute_list_for_element element_name
      end;
      let part_name_active_items = part_name_active_items de_children in
      let part_name_active_items =
        List.filter
          (fun (_, active) -> active = "true") part_name_active_items in
      begin
        match part_name_active_items with
        | [ part_name, _; ] -> part_name
        | _ -> fatal_error "too many active ConfigSetInfo"
      end
    | (Deep_block_diagram_defaults, _) :: deep_annexes
    | (Deep_graphical_interface, _) :: deep_annexes
    | (Deep_stateflow, _) :: deep_annexes
    | (Deep_annex_file_name _, _) :: deep_annexes ->
      loop deep_annexes in
  loop deep_annexes
;;

let get_explicit_configuration_of_annexes deep_annexes =
  let part_name = get_part_name_active deep_annexes in

  let config_set = get_config_set_of_annex deep_annexes part_name in
  let object_id = part_name, default_config_set_object_id in
  let active_configuration_set_explicit_object =
      Some {
        explicit_object_id = object_id;
        explicit_object_class_name =
          Some (Ast_types.mdl_name_of_string "Simulink.ConfigSet");
        explicit_object_prop_name =
          Some (Ast_types.ident_from_string "ActiveConfigurationSet");
        explicit_object_backup_class = None;
        explicit_object_version = None;
        explicit_object_reference = None;
        explicit_object_items = [];
        } in
  [ config_set; ], active_configuration_set_explicit_object
;;

let fill_block_diagram_defaults_of_annex accu de =
  let set_opt name = set_opt (Printf.sprintf "%s found twice" name) in
  let insert get set x = set accu (set_opt de.de_name (get accu) x) in
  match de.de_name with
  | "MaskDefaults" ->
    (* BEWARE: that one is different from the others, it uses a custom
       function  to translate the MaskDefaults node *)
    insert
      (fun accu -> accu.explicit_mask_defaults_opt)
      (fun accu x -> { accu with explicit_mask_defaults_opt = x; })
      (explicit_mask_defaults_of_annex de)
  | "AnnotationDefaults" -> insert
      (fun accu -> accu.explicit_annotation_defaults_opt)
      (fun accu x ->
       { accu with explicit_annotation_defaults_opt = x; })
      (explicit_annotation_defaults de.de_children)
  | "BlockDefaults" ->
    insert
      (fun accu -> accu.explicit_block_defaults_opt)
      (fun accu x -> { accu with explicit_block_defaults_opt = x; })
      (explicit_block_defaults de.de_children)
  | "BlockParameterDefaults" ->
    insert
      (fun accu -> accu.explicit_block_parameter_defaults_opt)
      (fun accu x ->
       { accu with explicit_block_parameter_defaults_opt = x; })
      (explicit_block_parameter_defaults de.de_children)
  | "LineDefaults" ->
    insert
      (fun accu -> accu.explicit_line_defaults_opt)
      (fun accu x -> { accu with explicit_line_defaults_opt = x; })
      (explicit_line_defaults de.de_children)
  | "MaskParameterDefaults" ->
    insert
      (fun accu -> accu.explicit_mask_parameter_defaults_opt)
      (fun accu x ->
       { accu with explicit_mask_parameter_defaults_opt = x; })
      (explicit_mask_parameter_defaults de.de_children)
  | "SystemDefaults" ->
    insert
      (fun accu -> accu.explicit_system_defaults_opt)
      (fun accu x -> { accu with explicit_system_defaults_opt = x; })
      (explicit_system_defaults de.de_children)
  | _ -> fail_unknown_tag "BlockDiagramDefaults" de
;;

let classify_block_diagram_defaults_of_annex _fname deep_element =
  let accu = empty_explicit_block_diagram_defaults_classifier in
  let _de_attributes, de_children =
    check_deep_element "BlockDiagramDefaults" deep_element in
  with_elements
    (List.fold_left fill_block_diagram_defaults_of_annex accu)
    de_children
;;

let get_of_annexes
      is_annex_name section_name of_deep_element deep_annexes =
  let rec check_no_more = function
    | [] -> ()
    | (deep_annex_name, _) :: deep_annexes ->
      if is_annex_name deep_annex_name then
        fatal_error
          (Printf.sprintf "More than one %s section" section_name);
      check_no_more deep_annexes in
  let rec get = function
    | [] -> None
    | (deep_annex_name, deep_document) :: deep_annexes ->
      if is_annex_name deep_annex_name then
        let fname = file_name_of_deep_annex_name deep_annex_name in
        let x = of_deep_element fname deep_document.deep_element in
        check_no_more deep_annexes;
        Some x
      else
        get deep_annexes in
  get deep_annexes
;;

let get_explicit_block_diagram_defaults_of_annexes deep_annexes =
  let is_annex_name = function
    | Deep_config_set_info
    | Deep_stateflow
    | Deep_graphical_interface
    | Deep_annex_file_name _ -> false
    | Deep_block_diagram_defaults -> true in
  let section_name = "block diagram defaults" in
  let of_deep_element = classify_block_diagram_defaults_of_annex in
  get_of_annexes is_annex_name section_name of_deep_element deep_annexes
;;

let explicit_section deep_annexes element_name deep_element =

  let _de_attributes, de_children =
    check_deep_element element_name deep_element in

  let classifier =
    with_elements (fun deep_elements ->
      List.fold_left
        classify_implementation_section_item
        empty_implementation_section_items_classifier
        deep_elements)
      de_children in
  match classifier with
  | {
      explicit_model_workspace_opt = explicit_model_workspace_opt;

      explicit_model_configuration_opt =
        explicit_model_configuration_opt;
      explicit_configuration_set_object_opt =
        explicit_configuration_set_object_opt;
      explicit_configuration_set_binding_list =
        explicit_configuration_set_binding_list;
      explicit_block_diagram_defaults_classifier =
        explicit_block_diagram_defaults_classifier;
      explicit_system_opt =
        explicit_system_opt;
    } ->

    let get_option item_name =
      get_option_value
        "incomplete defaults specifications in model" item_name in

    let explicit_model_configuration, explicit_configuration_set_object_opt =
      match explicit_model_configuration_opt with
      | None ->
        begin
          match explicit_configuration_set_object_opt with
          | None -> get_explicit_configuration_of_annexes deep_annexes
          | Some _ ->
            fatal_error
              ("incorrect defaults specifications in model: \
                ActiveConfigurationSet without ConfigurationSets")
        end
      | Some explicit_model_configuration ->
        explicit_model_configuration, explicit_configuration_set_object_opt in

    let explicit_configuration_set = {
      explicit_configuration_set_object =
        explicit_configuration_set_object_opt;
      explicit_configuration_set_bindings =
        explicit_configuration_set_binding_list;
    } in

    let block_diagram_defaults =
      (* If there is a block_diagram_defaults coming from annex, we use
         it. Otherwise we use the old classified values. The real
         question is: what if both are defined? *)
      match
        get_explicit_block_diagram_defaults_of_annexes deep_annexes
      with
      | None -> explicit_block_diagram_defaults_classifier
      | Some block_diagram_defaults -> block_diagram_defaults in

    match block_diagram_defaults with
    | {
      explicit_system_defaults_opt = explicit_system_defaults_opt;
      explicit_block_defaults_opt = explicit_block_defaults_opt;
      explicit_annotation_defaults_opt = explicit_annotation_defaults_opt;
      explicit_line_defaults_opt = explicit_line_defaults_opt;
      explicit_mask_defaults_opt = explicit_mask_defaults_opt;
      explicit_mask_parameter_defaults_opt =
        explicit_mask_parameter_defaults_opt;
      explicit_block_parameter_defaults_opt =
        explicit_block_parameter_defaults_opt;
    } ->

      {
        explicit_model_workspace = explicit_model_workspace_opt;

        explicit_model_configuration = explicit_model_configuration;
        explicit_configuration_set = explicit_configuration_set;
        explicit_system_defaults =
          get_option "explicit_system_defaults"
            explicit_system_defaults_opt;
        explicit_block_defaults =
          get_option "explicit_block_defaults"
            explicit_block_defaults_opt;
        explicit_annotation_defaults =
          get_option "explicit_annotation_defaults"
            explicit_annotation_defaults_opt;
        explicit_line_defaults =
          get_option "explicit_line_defaults"
            explicit_line_defaults_opt;
        explicit_mask_defaults =
          get_list_option_value explicit_mask_defaults_opt;
        explicit_mask_parameter_defaults =
          get_list_option_value explicit_mask_parameter_defaults_opt;
        explicit_block_parameter_defaults =
          get_option "explicit_block_parameter_defaults"
            explicit_block_parameter_defaults_opt;
        explicit_system =
          get_option "explicit_system" explicit_system_opt;
      }
;;

let section_classifier deep_annexes element_name f =
  explicit_classifier
    element_name (explicit_section deep_annexes element_name) f
;;

let explicit_comment deep_element =
  let _, _ =
    check_deep_element "Comment" deep_element in
  warning "dropping comment";
  ()
;;

let comment_classifier f =
  explicit_classifier "Comment" explicit_comment f
;;

let explicit_debug deep_element =
  let de_attributes, de_children =
    check_deep_element "debug" deep_element in
  match de_attributes with
  | [] ->
    explicit_bindings "explicit_debug" de_children
  | _ ->
    unknown_attribute_list_for_element "debug"
;;

let debug_classifier f =
  explicit_classifier "debug" explicit_debug f
;;

let explicit_target_item =
  explicit_binding_of_element "explicit_target_item"
;;

let explicit_target deep_element =
  let de_attributes, de_children =
    check_deep_element "target" deep_element in
  let id, name =
    match de_attributes with
    | [
        id_attribute;
        name_attribute;
      ] ->
      check_deep_attribute "target" "id" id_attribute,
      Some (check_deep_attribute "target" "name" name_attribute)
    | [
        id_attribute;
      ] ->
      check_deep_attribute "target" "id" id_attribute,
      None
    | _ ->
      unknown_attribute_list_for_element "target" in
  let id = unatural_of_string id in
  let items = map_elements explicit_target_item de_children in
  {
    explicit_target_id = id;
    explicit_target_name = name;
    explicit_target_items = items;
  }
;;

let target_classifier f =
  explicit_classifier "target" explicit_target f
;;

let explicit_active_state_output deep_element =
  let de_attributes, de_children =
    check_deep_element "activeStateOutput" deep_element in
  match de_attributes with
  | [] ->
     explicit_bindings "explicit_active_state_output" de_children
  | _ ->
    unknown_attribute_list_for_element "activeStateOutput"
;;

let active_state_output_classifier f =
  explicit_classifier "activeStateOutput" explicit_active_state_output f
;;

let explicit_src deep_element =
  let de_attributes, de_children =
    check_deep_element "src" deep_element in
  match de_attributes with
  | [] ->
     explicit_bindings "explicit_src" de_children
  | _ ->
    unknown_attribute_list_for_element "src"
;;

let src_classifier f =
  explicit_classifier "src" explicit_src f
;;

let explicit_dst deep_element =
  let de_attributes, de_children =
    check_deep_element "dst" deep_element in
  match de_attributes with
  | [] ->
     explicit_bindings "explicit_dst" de_children
  | _ ->
    unknown_attribute_list_for_element "dst"
;;

let dst_classifier f =
  explicit_classifier "dst" explicit_dst f
;;

let explicit_slide deep_element =
  let de_attributes, de_children =
    check_deep_element "slide" deep_element in
  match de_attributes with
  | [] ->
     explicit_bindings "explicit_slide" de_children
  | _ ->
    unknown_attribute_list_for_element "slide"
;;

let slide_classifier f =
  explicit_classifier "slide" explicit_slide f
;;

let explicit_autogen_item fname deep_element =
  classify_element "autogen" deep_element
    [
      binding_classifier
        "explicit_autogen_item"
        (fun binding -> Explicit_autogen_item_binding binding);
      array_classifier fname (fun array -> Explicit_autogen_item_array array);
    ]
;;

let explicit_autogen fname deep_element =
  let de_attributes, de_children =
    check_deep_element "autogen" deep_element in
  match de_attributes with
  | [] ->
    map_elements (explicit_autogen_item fname) de_children
  | _ ->
    unknown_attribute_list_for_element "autogen"
;;

let autogen_classifier fname f =
  explicit_classifier "autogen" (explicit_autogen fname) f
;;

let explicit_transition_item fname deep_element =
  classify_element "transition" deep_element
    [
      binding_classifier
        "explicit_transition_item"
        (fun binding -> Explicit_transition_item_binding binding);
      src_classifier (fun src -> Explicit_transition_item_src src);
      dst_classifier (fun dst -> Explicit_transition_item_dst dst);
      slide_classifier (fun slide -> Explicit_transition_item_slide slide);
      autogen_classifier fname
        (fun autogen -> Explicit_transition_item_autogen autogen);
    ]
;;

let explicit_transition fname deep_element =
  let de_attributes, de_children =
    check_deep_element "transition" deep_element in
  match de_attributes with
  | [
      ssid_attribute;
    ] ->
    let id =
      ssid_value
        (check_deep_attribute "transition" "SSID" ssid_attribute) in
    let items = map_elements (explicit_transition_item fname) de_children in
    {
      explicit_transition_id = id;
      explicit_transition_items = items;
    }
  | _ ->
    unknown_attribute_list_for_element "transition"
;;

let transition_classifier fname f =
  explicit_classifier "transition" (explicit_transition fname) f
;;

let explicit_sub_view_s deep_element =
  let de_attributes, de_children =
    check_deep_element "subviewS" deep_element in
  match de_attributes with
  | [] ->
     explicit_bindings "explicit_sub_view_s" de_children
  | _ ->
    unknown_attribute_list_for_element "subviewS"
;;

let sub_view_s_classifier f =
  explicit_classifier "subviewS" explicit_sub_view_s f
;;

let explicit_fixpt deep_element =
  let de_attributes, de_children =
    check_deep_element "fixpt" deep_element in
  match de_attributes with
  | [] ->
     explicit_bindings "explicit_fixpt" de_children
  | _ ->
    unknown_attribute_list_for_element "fixpt"
;;

let fixpt_classifier f =
  explicit_classifier "fixpt" explicit_fixpt f
;;

let explicit_type_item deep_element =
  classify_element "type" deep_element
    [
      binding_classifier
        "explicit_type_item"
        (fun binding -> Explicit_type_item_binding binding);
      fixpt_classifier (fun fixpt -> Explicit_type_item_fixpt fixpt);
    ]
;;

let explicit_type deep_element =
  let de_attributes, de_children =
    check_deep_element "type" deep_element in
  match de_attributes with
  | [] ->
    map_elements explicit_type_item de_children
  | _ ->
    unknown_attribute_list_for_element "type"
;;

let type_classifier f =
  explicit_classifier "type" explicit_type f
;;

let explicit_stateflow_array deep_element =
  let de_attributes, de_children =
    check_deep_element "array" deep_element in
  match de_attributes with
  | [] ->
    explicit_bindings "explicit_stateflow_array" de_children
  | _ ->
    unknown_attribute_list_for_element "array"
;;

let stateflow_array_classifier f =
  explicit_classifier "array" explicit_stateflow_array f
;;

let explicit_range_item deep_element =
  classify_element "range" deep_element
    [
      binding_classifier
        "explicit_range_item"
        (fun binding -> binding);
    ]
;;

let explicit_range deep_element =
  let de_attributes, de_children =
    check_deep_element "range" deep_element in
  match de_attributes with
  | [] ->
    map_elements explicit_range_item de_children
  | _ ->
    unknown_attribute_list_for_element "range"
;;

let range_classifier f =
  explicit_classifier "range" explicit_range f
;;

let explicit_unit deep_element =
  let _, _ = check_deep_element "unit" deep_element in
  warning "dropping unit"
;;

let unit_classifier f =
  explicit_classifier "unit" explicit_unit f
;;

let explicit_props_item deep_element =
  classify_element "props" deep_element
    [
      type_classifier
        (fun explicit_type -> Some (Explicit_props_type explicit_type));
      binding_classifier
        "explicit_props_item"
        (fun binding -> Some (Explicit_props_binding binding));
      stateflow_array_classifier
        (fun array -> Some (Explicit_props_stateflow_array array));
      range_classifier (fun range -> Some (Explicit_props_range range));
      unit_classifier (fun () -> None);
    ]
;;

let explicit_props deep_element =
  let de_attributes, de_children =
    check_deep_element "props" deep_element in
  match de_attributes with
  | [] ->
    let props = map_elements explicit_props_item de_children in
    list_of_options props
  | _ ->
    unknown_attribute_list_for_element "props"
;;

let props_classifier f =
  explicit_classifier "props" explicit_props f
;;

let explicit_data_item deep_element =
  classify_element "data" deep_element
    [
      binding_classifier
        "explicit_data_item"
        (fun binding -> Explicit_data_binding binding);
      props_classifier (fun props -> Explicit_data_props props);
    ]
;;

let explicit_data deep_element =
  let de_attributes, de_children =
    check_deep_element "data" deep_element in
  match de_attributes with
  | [
      id_attribute;
      name_attribute;
    ] ->
    let id_opt, ssid_opt =
      match id_attribute with
      | {
          da_name = "id";
          da_value = id;
        } -> Some (Explicit_data_id (unatural_of_string id)), None
      | {
          da_name = "SSID";
          da_value = ssid;
        } -> None, Some (Explicit_data_ssid (unatural_of_string ssid))
      | {
          da_name = da_name;
          da_value = _da_value;
        } ->
        fatal_error (Printf.sprintf
          "%S attribute read, but %S or %S attribute was expected for tag %S"
          da_name "SSID" "id" "data") in
    let name = check_deep_attribute "data" "name" name_attribute in
    let items = map_elements explicit_data_item de_children in
    {
      explicit_data_id = id_opt;
      explicit_data_ssid = ssid_opt;
      explicit_data_name = name;
      explicit_data_items = items;
    }
  | [
      id_attribute;
      ssid_attribute;
      name_attribute;
    ] ->
    let id =
      match id_attribute with
      | {
          da_name = "id";
          da_value = id;
        } -> Explicit_data_id (unatural_of_string id)
      | {
          da_name = da_name;
          da_value = _da_value;
        } ->
        fatal_error (Printf.sprintf
          "%S attribute read, but %S attribute was expected for tag %S"
          da_name "id" "data") in
    let ssid =
      match ssid_attribute with
      | {
          da_name = "SSID";
          da_value = ssid;
        } -> Explicit_data_ssid (unatural_of_string ssid)
      | {
          da_name = da_name;
          da_value = _da_value;
        } ->
        fatal_error (Printf.sprintf
          "%S attribute read, but %S attribute was expected for tag %S"
          da_name "SSID" "data") in
    let name = check_deep_attribute "data" "name" name_attribute in
    let items = map_elements explicit_data_item de_children in
    {
      explicit_data_id = Some id;
      explicit_data_ssid = Some ssid;
      explicit_data_name = name;
      explicit_data_items = items;
    }
  | _ ->
    unknown_attribute_list_for_element "data"
;;

let data_classifier f =
  explicit_classifier "data" explicit_data f
;;

let explicit_junction_items fname deep_element =
  classify_element "junction" deep_element
    [
      binding_classifier
        "explicit_junction_items"
        (fun binding -> Explicit_junction_item_binding binding);
      autogen_classifier fname
        (fun autogen -> Explicit_junction_item_autogen autogen);
    ]
;;

let explicit_junction fname deep_element =
  let de_attributes, de_children =
    check_deep_element "junction" deep_element in
  match de_attributes with
  | [ ssid_attribute; ] ->
    let id =
      ssid_value
        (check_deep_attribute "junction" "SSID" ssid_attribute) in
    let items = map_elements (explicit_junction_items fname) de_children in
    {
      explicit_junction_id = id;
      explicit_junction_items = items;
    }
  | _ ->
    unknown_attribute_list_for_element "junction"
;;

let junction_classifier fname f =
  explicit_classifier "junction" (explicit_junction fname) f
;;

let explicit_event deep_element =
  let de_attributes, de_children =
    check_deep_element "event" deep_element in
  match de_attributes with
  | [ ssid_attribute; ] ->
    let id =
      ssid_value (check_deep_attribute "event" "SSID" ssid_attribute) in
    let items = explicit_bindings "explicit_event" de_children in
    {
      explicit_event_id = id;
      explicit_event_items = items;
    }
  | _ ->
    unknown_attribute_list_for_element "event"
;;

let event_classifier f =
  explicit_classifier "event" explicit_event f
;;

let explicit_eml deep_element =
  let de_attributes, de_children =
    check_deep_element "eml" deep_element in
  match de_attributes with
  | [] ->
    explicit_bindings "explicit_eml" de_children
  | _ ->
    unknown_attribute_list_for_element "eml"
;;

let eml_classifier f =
  explicit_classifier "eml" explicit_eml f
;;

let explicit_simulink deep_element =
  let de_attributes, de_children =
    check_deep_element "simulink" deep_element in
  match de_attributes with
  | [] ->
    explicit_bindings "explicit_simulink" de_children
  | _ ->
    unknown_attribute_list_for_element "simulink"
;;

let simulink_classifier f =
  explicit_classifier "simulink" explicit_simulink f
;;

let explicit_layout_info deep_element =
  let de_attributes, de_children =
    check_deep_element "layoutInfo" deep_element in
  match de_attributes with
  | [] ->
    explicit_bindings "explicit_layout_info" de_children
  | _ ->
    unknown_attribute_list_for_element "layoutInfo"
;;

let layout_info_classifier f =
  explicit_classifier "layoutInfo" explicit_layout_info f
;;

let explicit_table_ui_prefs deep_element =
  let de_attributes, de_children =
    check_deep_element "tableUIPrefs" deep_element in
  match de_attributes with
  | [] ->
    explicit_bindings "explicit_table_ui_prefs" de_children
  | _ ->
    unknown_attribute_list_for_element "tableUIPrefs"
;;

let table_ui_prefs_classifier f =
  explicit_classifier "tableUIPrefs" explicit_table_ui_prefs f
;;

let explicit_state_cell deep_element =
  let de_attributes, de_children =
    check_deep_element "statecell" deep_element in
  match de_attributes with
  | [ id_attribute; ] ->
    let id =
      unatural_of_string
        (check_deep_attribute "statecell" "id" id_attribute) in
    let bindings = explicit_bindings "explicit_state_cell" de_children in
    {
      explicit_state_cell_id = id;
      explicit_state_cell_items = bindings;
    }
  | _ ->
    unknown_attribute_list_for_element "statecell"
;;

let state_cell_classifier f =
  explicit_classifier "statecell" explicit_state_cell f
;;

let explicit_transition_cell deep_element =
  let de_attributes, de_children =
    check_deep_element "transitioncell" deep_element in
  match de_attributes with
  | [ id_attribute; ] ->
    let id =
      unatural_of_string
        (check_deep_attribute "transitioncell" "id" id_attribute) in
    let bindings = explicit_bindings "explicit_transition_cell" de_children in
    {
      explicit_transition_cell_id = id;
      explicit_transition_cell_items = bindings;
    }
  | _ ->
    unknown_attribute_list_for_element "transitioncell"
;;

let transition_cell_classifier f =
  explicit_classifier "transitioncell" explicit_transition_cell f
;;

let explicit_truth_table_item fname deep_element =
  classify_element "truthTable" deep_element
    [
      binding_classifier
        "explicit_truth_table_item"
        (fun binding -> Explicit_truth_table_item_binding binding);
      array_classifier fname
        (fun array -> Explicit_truth_table_item_array array);
    ]
;;

let explicit_truth_table fname deep_element =
  let de_attributes, de_children =
    check_deep_element "truthTable" deep_element in
  match de_attributes with
  | [] -> map_elements (explicit_truth_table_item fname) de_children
  | _ ->
    unknown_attribute_list_for_element "truthTable"
;;

let truth_table_classifier fname f =
  explicit_classifier "truthTable" (explicit_truth_table fname) f
;;

let plant_modeling_info_tag = "plantModelingInfo";;

let plant_modeling_info deep_element =
  let _de_attributes, _de_children =
    check_deep_element plant_modeling_info_tag deep_element in
  (* todo: find a concrete example of content for plantModelingInfo *)
  ()
;;

let plant_modeling_info_classifier f =
  explicit_classifier plant_modeling_info_tag plant_modeling_info f
;;

let rec explicit_chart_item fname deep_element =
  classify_element "chart" deep_element
    [
      binding_classifier
        "explicit_chart_item"
        (fun binding -> Explicit_chart_item_binding binding);
      children_classifier fname
        (fun children -> Explicit_chart_item_children children);
      sub_view_s_classifier
        (fun sub_view_s -> Explicit_chart_item_sub_view_s sub_view_s);
      eml_classifier (fun eml -> Explicit_chart_item_eml eml);
      active_state_output_classifier
        (fun active_state_output ->
         Explicit_chart_item_active_state_output active_state_output);
      plant_modeling_info_classifier
        (fun () -> Explicit_chart_item_plant_modeling_info);

    ]

and explicit_chart fname deep_element =
  let de_attributes, de_children =
    check_deep_element "chart" deep_element in
  match de_attributes with
  | [ id_attribute; ] ->
    let id = check_deep_attribute "chart" "id" id_attribute in
    let id = unatural_of_string id in
    let items = map_elements (explicit_chart_item fname) de_children in
    {
      explicit_chart_id = id;
      explicit_chart_items = items;
    }
  | _ ->
    unknown_attribute_list_for_element "chart"

and chart_classifier fname f =
  explicit_classifier "chart" (explicit_chart fname) f

and explicit_table_item fname deep_element =
  classify_element "table" deep_element
    [
      binding_classifier
        "explicit_table_item"
        (fun binding -> Explicit_table_item_binding binding);
      layout_info_classifier
        (fun layout_info -> Explicit_table_item_layout_info layout_info);
      array_classifier fname (fun array -> Explicit_table_item_array array);
      table_ui_prefs_classifier
        (fun table_ui_prefs ->
         Explicit_table_item_table_ui_prefs table_ui_prefs);
      children_classifier fname
        (fun children -> Explicit_table_item_children children);
    ]

and explicit_table fname deep_element =
  let de_attributes, de_children =
    check_deep_element "table" deep_element in
  match de_attributes with
  | [ id_attribute; ] ->
    let id = check_deep_attribute "table" "id" id_attribute in
    let id = unatural_of_string id in
    let items = map_elements (explicit_table_item fname) de_children in
    {
      explicit_table_id = id;
      explicit_table_items = items;
    }
  | _ ->
    unknown_attribute_list_for_element "table"

and table_classifier fname f =
  explicit_classifier "table" (explicit_table fname) f

and explicit_child fname deep_element =
  classify_element "Children" deep_element
    [
      target_classifier (fun target -> Explicit_child_target target);
      chart_classifier fname (fun chart -> Explicit_child_chart chart);
      state_classifier fname (fun state -> Explicit_child_state state);
      transition_classifier fname
        (fun transition -> Explicit_child_transition transition);
      data_classifier (fun data -> Explicit_child_data data);
      junction_classifier fname
        (fun junction -> Explicit_child_junction junction);
      event_classifier (fun event -> Explicit_child_event event);
      table_classifier fname (fun table -> Explicit_child_table table);
      state_cell_classifier
        (fun state_cell -> Explicit_child_state_cell state_cell);
      transition_cell_classifier
        (fun transition_cell ->
         Explicit_child_transition_cell transition_cell);
    ]

and explicit_children fname deep_element =
  let de_attributes, de_children =
    check_deep_element "Children" deep_element in
  match de_attributes with
  | [] ->
    map_elements (explicit_child fname) de_children
  | _ ->
    unknown_attribute_list_for_element "Children"

and explicit_state_item fname deep_element =
  classify_element "state" deep_element
    [
      binding_classifier
        "explicit_state_item"
        (fun binding -> Explicit_state_item_binding binding);
      children_classifier fname
        (fun children -> Explicit_state_item_children children);
      active_state_output_classifier
        (fun active_state_output ->
         Explicit_state_item_active_state_output active_state_output);
      sub_view_s_classifier
        (fun sub_view_s -> Explicit_state_item_sub_view_s sub_view_s);
      eml_classifier (fun eml -> Explicit_state_item_eml eml);
      simulink_classifier
        (fun simulink -> Explicit_state_item_simulink simulink);
      autogen_classifier fname
        (fun autogen -> Explicit_state_item_autogen autogen);
      truth_table_classifier fname
        (fun truth_table -> Explicit_state_item_truth_table truth_table);
    ]

and explicit_state fname deep_element =
  let de_attributes, de_children =
    check_deep_element "state" deep_element in
  match de_attributes with
  | [
      ssid_attribute;
    ] ->
    let id =
      ssid_value (check_deep_attribute "state" "SSID" ssid_attribute) in
    let items = map_elements (explicit_state_item fname) de_children in
    {
      explicit_state_id = id;
      explicit_state_items = items;
    }
  | _ ->
    unknown_attribute_list_for_element "state"

and state_classifier fname f =
  explicit_classifier "state" (explicit_state fname) f

and children_classifier :
      'a. Path.file_name -> (explicit_children -> 'a) -> string * (deep_element -> 'a) =
  fun fname f -> explicit_classifier "Children" (explicit_children fname) f
;;

let explicit_machine_item fname deep_element =
  classify_element "machine" deep_element
    [
      binding_classifier
        "explicit_machine_item"
        (fun binding -> Explicit_machine_binding binding);
      debug_classifier (fun debug -> Explicit_machine_debug debug);
      children_classifier fname
        (fun children -> Explicit_machine_children children);
    ]
;;

let explicit_machine fname deep_element =
  let de_attributes, de_children =
    check_deep_element "machine" deep_element in
  match de_attributes with
  | [
      id_attribute;
    ] ->
    let id = check_deep_attribute "machine" "id" id_attribute in
    let id = unatural_of_string id in
    let items = map_elements (explicit_machine_item fname) de_children in
    {
      explicit_machine_id = id;
      explicit_machine_items = items;
    }
  | _ ->
    unknown_attribute_list_for_element "machine"
;;

let machine_classifier fname f =
  explicit_classifier "machine" (explicit_machine fname) f
;;

let explicit_instance deep_element =
  let de_attributes, de_children =
    check_deep_element "instance" deep_element in
  match de_attributes with
  | [
      id_attribute;
    ] ->
    let id =
      unatural_of_string (check_deep_attribute "instance" "id" id_attribute) in
    let items = explicit_bindings "explicit_instance" de_children in
    {
      explicit_instance_id = id;
      explicit_instance_items = items;
    }
  | _ ->
    unknown_attribute_list_for_element "instance"
;;

let instance_classifier f =
  explicit_classifier "instance" explicit_instance f
;;

let explicit_stateflow_section_item fname deep_element =
  classify_element "Stateflow" deep_element
    [
      machine_classifier fname
        (fun machine -> Explicit_stateflow_section_machine machine);
      instance_classifier
        (fun instance -> Explicit_stateflow_section_instance instance);
    ]
;;

let explicit_stateflow_section fname deep_element =
  let de_attributes, de_children =
    check_deep_element "Stateflow" deep_element in
  match de_attributes with
  | [] ->
    let explicit_stateflow_section_items =
      map_elements
        (explicit_stateflow_section_item fname) de_children in
    Explicit_stateflow_section explicit_stateflow_section_items
  | _ ->
    unknown_attribute_list_for_element "Stateflow"
;;

let stateflow_classifier fname f =
  explicit_classifier "Stateflow" (explicit_stateflow_section fname) f
;;

let stateflow_of_annex_document deep_annexes =
  let is_annex_name = function
    | Deep_block_diagram_defaults
    | Deep_config_set_info
    | Deep_graphical_interface
    | Deep_annex_file_name _ -> false
    | Deep_stateflow -> true in
  let section_name = "stateflow" in
  let of_deep_element fname de =
    Explicit_implementation_stateflow
      (explicit_stateflow_section fname de) in
  get_of_annexes is_annex_name section_name of_deep_element deep_annexes
;;

let explicit_implementation_section deep_annexes deep_element =
  fdebug (fun epf -> Format.fprintf epf
    "explicit_implementation_section: %S" deep_element.de_name);
  let file_name = Configuration.default_slx_blockdiagram_file_name in
  classify_element "ModelInformation" deep_element
    [
      section_classifier deep_annexes
        "Library"
        (fun section -> Some (Explicit_implementation_library section));
      section_classifier deep_annexes
        "Model"
        (fun section -> Some (Explicit_implementation_model section));
      stateflow_classifier file_name
        (fun stateflow -> Some (Explicit_implementation_stateflow stateflow));
      comment_classifier
        (fun () -> None)
    ]
;;

let explicit_model_version deep_attributes =
  let rec find_version version_opt = function
    | [] -> version_opt
    | deep_attribute :: deep_attributes ->
      match deep_attribute with
      | {
          da_name = "Version";
          da_value = model_version;
        } ->
        check_opt_is_empty
          "two ``Version'' attributes in ``ModelInformation''" version_opt;
        find_version (Some model_version) deep_attributes
      | _ -> find_version version_opt deep_attributes in
  match find_version None deep_attributes with
  | None -> fatal_error "no ``Version'' attributes in ``ModelInformation''"
  | Some model_version -> model_version
;;

let explicit_model_sections deep_annexes = function
  | {
      de_name = "ModelInformation";
      de_attributes = deep_attributes;
      de_children = deep_children;
    } ->
    fdebug (fun epf -> Format.fprintf epf
      "explicit_model_sections: %S" "ModelInformation");
    let model_version = explicit_model_version deep_attributes in
    let explicit_implementation_section_items =
      with_elements
        (List.map (explicit_implementation_section deep_annexes))
        deep_children in
    let stateflow_of_annex_document =
      stateflow_of_annex_document deep_annexes in
    let explicit_implementation_section_items =
      (* StateFlow must not be first, it must be after model. Maybe we can
         insert it just after the header to avoid explorate the full list. But
         it's not a huge list, so not urgent. *)
      explicit_implementation_section_items @ [stateflow_of_annex_document] in
    list_of_options explicit_implementation_section_items, model_version
  | {
      de_name = element_name;
      de_attributes = _deep_attributes;
      de_children = _deep_children;
    } ->
    fatal_error (Printf.sprintf
      "Unknown root element name %S" element_name)
;;

let explicit_encoding = function
  | None -> ASCII
  | Some encoding ->
    match Lib_strings.lowercase encoding with
    | "ascii" | "us-ascii" | "us" -> ASCII
    | "iso-8859-1" | "iso_8859-1" | "latin1" -> Latin1
    | "iso-8859-15" | "iso_8859-15" | "latin-9" -> Latin9
    | "utf-16" -> UTF16
    | "utf-8" -> UTF8
    | encoding -> Encoding encoding
;;

let explicit_blockdiagram deep_annexes = function
  | {
      deep_prolog = encoding_opt;
      deep_element = deep_element;
    } ->
    let explicit_encoding = explicit_encoding encoding_opt
    and explicit_model_sections, explicit_model_version =
      explicit_model_sections deep_annexes deep_element in
    {
      explicit_prolog = explicit_encoding;
      explicit_model_sections = explicit_model_sections;
      explicit_model_version = explicit_model_version;
    }
;;

let xml_deep_ast_to_xml_explicit_ast = function
  | {
      deep_model_ident = model_ident;
      deep_blockdiagram = deep_blockdiagram;
      deep_annexes = deep_annexes;
    } ->
    let explicit_blockdiagram =
      explicit_blockdiagram deep_annexes deep_blockdiagram in
    fdebug
      (fun ppf ->
       Xml_explicit_ast_print.print_explicit_blockdiagram
         ppf explicit_blockdiagram);
    {
      explicit_model_ident = model_ident;
      explicit_blockdiagram = explicit_blockdiagram;
    }
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
 End:
*)
