(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, Clément Franchini                    *)
(*                   INRIA Rocquencourt                                *)
(*                                                                     *)
(*  Copyright 2013-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: xml_explicit_ast.mli,v 1.1 2019-05-10 14:00:17 jpc Exp $ *)

(** Explicit_xml_ast is a specialized version of XML trees specialized to MDL
    specification and translation.

    To facilitate pattern matching and safety, all (or as much as possible)
    string values in the XML tree should be translated to constructors
    similar to MDL constructors and values.

    The Explicit_xml_ast trees will facilitate translation to MDL files (or
    preferably translation to MDL AST values).
*)

open Ast_types;;

type explicit_version = string
and explicit_model_version = string
and explicit_reference = string
and explicit_list_type = string
;;

type explicit_dimension = Ast_types.natural * natural;;

type sid = Mdl_ast.block_sid;;

type explicit_object_id = Path.file_name * natural;;

type explicit_encoding =
  | ASCII
  | Latin1
  | Latin9
  | UTF16
  | UTF8
  | Encoding of string
;;

(*
type explicit_zoom_mode =
  | DZM_on
  | DZM_off
  | DZM_xonly
  | DZM_yonly
;;
*)

type source_port_description =
  {
    src_port : natural;
    src_block_name : Mdl_ast.block_name;
  }
and destination_port_description =
  {
    dst_port : natural;
    dst_block_name : Mdl_ast.block_name;
  }
;;

type parameter_class =
  | Char
  | Double
  | Int32
  | Uint32
  | Logical
;;

type explicit_cell_class = parameter_class
and explicit_field_class = parameter_class
;;

type text = string;;

type explicit_position = text;;

type explicit_document_value =
  (* A simple value. *)
  | Ddv of text
  (* A simple value associated with its class parameter. *)
  | Ddv_classified of parameter_class * text
;;

type explicit_binding = explicit_document_value Binding.ident_binding
and explicit_bindings = explicit_binding list;;

type explicit_block_parameter_default = {
  explicit_block_parameter_default_type : Mdl_block_types.block_type;
  explicit_block_parameter_default_bindings : explicit_bindings;
}
and explicit_block_parameter_defaults = explicit_block_parameter_default list;;

type explicit_annotation = {
  explicit_annotation_sid : sid;
  explicit_annotation_bindings : explicit_bindings;
}
and explicit_annotations = explicit_annotation list
;;

type explicit_connectors = explicit_connector list
and explicit_connector = explicit_bindings
;;

type explicit_custom_property_item =
  | Explicit_custom_property_item_binding of explicit_binding
  | Explicit_custom_property_item_enum_str_pairs of explicit_bindings
;;

type explicit_custom_property_items =
  explicit_custom_property_item list
;;

type explicit_custom_property = {
  explicit_custom_property_name : Mdl_ast.prop_name;
  explicit_custom_property_data_type : Mdl_ast.custom_property_type;
  explicit_custom_property_items : explicit_custom_property_items;
}
;;

type explicit_dialog_parameters = {
  explicit_dialog_parameters_block_name : Mdl_ast.block_name;
  explicit_dialog_parameters_bindings : explicit_bindings;
}
;;

type explicit_link_data = explicit_dialog_parameters list;;

type explicit_mask_display_icon_frame = text
and explicit_mask_display_icon_units = text
and explicit_mask_display_icon_opaque = text
and explicit_mask_display_icon_rotate = text
and explicit_mask_display_run_init_for_icon_redraw = text
and explicit_mask_display_script = text
and explicit_mask_initialisation = text
and explicit_mask_help = text
and explicit_mask_parameter_name = Ast_types.mdl_name
and explicit_mask_parameter_type = text
and explicit_mask_parameter_prompt = text
and explicit_mask_parameter_value = text
and explicit_mask_type = text
and explicit_mask_description = text
and explicit_mask_parameter_evaluate = text
and explicit_mask_parameter_enabled = text
and explicit_mask_parameter_visible = text
and explicit_mask_parameter_callback = text
and explicit_mask_parameter_tunable = text
and explicit_mask_parameter_internal = text
and explicit_mask_parameter_never_save = text
and explicit_mask_parameter_read_only = text
;;

type explicit_mask_parameter = {
  explicit_mask_parameter_name : explicit_mask_parameter_name;
  explicit_mask_parameter_type : explicit_mask_parameter_type;
  explicit_mask_parameter_prompt : explicit_mask_parameter_prompt;
  explicit_mask_parameter_value : explicit_mask_parameter_value;
  explicit_mask_parameter_callback :
    explicit_mask_parameter_callback option;
  explicit_mask_parameter_tunable :
    explicit_mask_parameter_tunable option;
  explicit_mask_parameter_evaluate :
    explicit_mask_parameter_evaluate option;
  explicit_mask_parameter_enabled :
    explicit_mask_parameter_enabled option;
  explicit_mask_parameter_visible :
    explicit_mask_parameter_visible option;
  explicit_mask_parameter_internal :
    explicit_mask_parameter_internal option;
  explicit_mask_parameter_never_save :
    explicit_mask_parameter_never_save option;
  explicit_mask_parameter_read_only :
    explicit_mask_parameter_read_only option;
}
;;

type explicit_mask_display = {
  explicit_mask_display_icon_frame :
    explicit_mask_display_icon_frame option;
  explicit_mask_display_icon_units :
    explicit_mask_display_icon_units option;
  explicit_mask_display_icon_opaque :
    explicit_mask_display_icon_opaque option;
  explicit_mask_display_run_init_for_icon_redraw :
    explicit_mask_display_run_init_for_icon_redraw option;
  explicit_mask_display_icon_rotate :
    explicit_mask_display_icon_rotate option;
  explicit_mask_display_script :
    explicit_mask_display_script option;
}
;;

type explicit_mask = {
  explicit_mask_type : explicit_mask_type option;
  explicit_mask_display : explicit_mask_display;
  explicit_mask_description : explicit_mask_description option;
  explicit_mask_initialisation : explicit_mask_initialisation option;
  explicit_mask_help : explicit_mask_help option;
  explicit_mask_parameters : explicit_mask_parameter list;
}
;;

type explicit_system = {
  explicit_system_position : explicit_position;
  explicit_system_bindings : explicit_bindings;
  explicit_system_blocks : explicit_blocks;
  explicit_system_connections : explicit_lines;
  explicit_system_annotations : explicit_annotations;
  explicit_system_connectors : explicit_connectors;
}

and explicit_block = {
  explicit_block_type : Mdl_block_types.block_type;
  explicit_block_name : Mdl_ast.block_name;
  explicit_block_sid : sid;
  explicit_block_items : explicit_block_items;
}
and explicit_blocks = explicit_block list

and explicit_block_item =
| Explicit_block_binding of explicit_binding
| Explicit_block_object of explicit_object
| Explicit_block_system of explicit_system
| Explicit_block_list of explicit_list
| Explicit_block_port of explicit_port
| Explicit_block_array of explicit_array
| Explicit_block_instance_data of explicit_bindings
| Explicit_block_link_data of explicit_link_data
| Explicit_block_mask of explicit_mask
and explicit_block_items = explicit_block_item list

and explicit_port = explicit_port_items

and explicit_port_item =
| Explicit_port_array of explicit_array
| Explicit_port_binding of explicit_binding
and explicit_port_items = explicit_port_item list

and explicit_list = {
  explicit_list_type : explicit_list_type;
  explicit_list_bindings : explicit_bindings;
}

and explicit_object = {
  explicit_object_id : explicit_object_id;
  explicit_object_class_name : Mdl_ast.class_name option;
  explicit_object_prop_name : Mdl_ast.prop_name option;
  explicit_object_backup_class : Mdl_ast.backup_class_name option;
  explicit_object_version : explicit_version option;
  explicit_object_reference : explicit_reference option;
  explicit_object_items : explicit_object_items;
}

and explicit_object_item =
| Explicit_object_binding of explicit_binding
| Explicit_object_array of explicit_array
| Explicit_object_object of explicit_object
| Explicit_object_custom_property of explicit_custom_property
and explicit_object_items = explicit_object_item list

and explicit_array = {
  explicit_array_type : Mdl_ast.array_type;
  explicit_array_dimension : explicit_dimension;
  explicit_array_prop_name : Mdl_ast.prop_name;
  explicit_array_items : explicit_array_items;
}

and explicit_array_item =
  | Explicit_array_item_object of explicit_object
  | Explicit_array_item_cell of explicit_cell
  | Explicit_array_matstruct of explicit_matstruct_items
  | Explicit_array_array of explicit_array
and explicit_array_items = explicit_array_item list

and explicit_matstruct_item =
  | Explicit_matstruct_field of explicit_binding
  | Explicit_matstruct_array of explicit_array
and explicit_matstruct_items = explicit_matstruct_item list

and explicit_cell = {
  explicit_cell_class : explicit_cell_class option;
  explicit_cell_value : string;
}

and explicit_line = {
  explicit_line_type : Mdl_ast.line_type option;
  explicit_line_items : explicit_line_items;
}
and explicit_lines = explicit_line list

and explicit_line_item =
| Explicit_line_binding of explicit_binding
| Explicit_line_branch of explicit_branch
and explicit_line_items = explicit_line_item list

and explicit_branch = {
  explicit_branch_connection_type : Mdl_ast.connection_type option;
  explicit_branch_items : explicit_branch_items;
}
and explicit_branches = explicit_branch list

and explicit_branch_item =
| Explicit_branch_binding of explicit_binding
| Explicit_branch_branch of explicit_branch
and explicit_branch_items = explicit_branch_item list
;;

type config_manager_settings = string
and editor_settings = string
and simulation_settings = string
and verification = string
and external_mode = string
and model_reference_settings = string
and concurrent_execution_settings = string

and explicit_model_workspace = explicit_bindings option

and explicit_configuration_set = {
  explicit_configuration_set_object : explicit_object option;
  explicit_configuration_set_bindings : explicit_bindings;
}

and explicit_model_configuration = explicit_array list
and explicit_system_defaults = explicit_bindings
and explicit_block_defaults = explicit_bindings
and explicit_annotation_defaults = explicit_bindings
and explicit_line_defaults = explicit_bindings
and explicit_mask_defaults = explicit_bindings
and explicit_mask_parameter_defaults = explicit_bindings
;;

type explicit_section = {
(* We assume we can discard those data.
  explicit_simulation_settings : simulation_settings;
  explicit_model_windows_info : Mdl_ast.object_description;
  explicit_config_manager_settings : config_manager_settings;
  explicit_editor_settings : editor_settings;
  explicit_verification : verification;
  explicit_external_mode : external_mode;
  explicit_model_reference_settings : model_reference_settings;
  explicit_concurrent_execution_settings : concurrent_execution_settings;
*)
  explicit_model_workspace : explicit_model_workspace;
  explicit_model_configuration : explicit_model_configuration;
  explicit_configuration_set : explicit_configuration_set;
  explicit_system_defaults : explicit_system_defaults;
  explicit_block_defaults : explicit_block_defaults;
  explicit_annotation_defaults : explicit_annotation_defaults;
  explicit_line_defaults : explicit_line_defaults;
  explicit_mask_defaults : explicit_mask_defaults;
  explicit_mask_parameter_defaults : explicit_mask_parameter_defaults;
  explicit_block_parameter_defaults : explicit_block_parameter_defaults;
  explicit_system : explicit_system;
}
;;

(** {6 State Flow types section} *)

type explicit_id = Unatural.t

type explicit_ssid_number = Unatural.t;;

type explicit_debug = explicit_bindings;;

type explicit_target_items = explicit_bindings;;

type explicit_target = {
  explicit_target_id : explicit_id;
  explicit_target_name : string option;
  explicit_target_items : explicit_target_items;
}
;;

type explicit_active_state_output = explicit_bindings;;

type explicit_src = explicit_bindings;;

type explicit_dst = explicit_bindings;;

type explicit_slide = explicit_bindings;;

type explicit_autogen_item =
  | Explicit_autogen_item_binding of explicit_binding
  | Explicit_autogen_item_array of explicit_array
;;

type explicit_autogen_items = explicit_autogen_item list;;

type explicit_autogen = explicit_autogen_items;;

type explicit_transition_item =
  | Explicit_transition_item_binding of explicit_binding
  | Explicit_transition_item_src of explicit_src
  | Explicit_transition_item_dst of explicit_dst
  | Explicit_transition_item_slide of explicit_slide
  | Explicit_transition_item_autogen of explicit_autogen
;;

type explicit_transition_items = explicit_transition_item list;;

type explicit_transition = {
  explicit_transition_id : explicit_id;
  explicit_transition_items : explicit_transition_items;
}
;;

type explicit_sub_view_s = explicit_bindings;;

type explicit_fixpt = explicit_bindings;;

type explicit_type_item =
  | Explicit_type_item_binding of explicit_binding
  | Explicit_type_item_fixpt of explicit_fixpt
;;

type explicit_type_items = explicit_type_item list;;

type explicit_type = explicit_type_items;;

type explicit_stateflow_array = explicit_bindings;;

type explicit_range = explicit_bindings;;

type explicit_props_item =
  | Explicit_props_type of explicit_type
  | Explicit_props_binding of explicit_binding
  | Explicit_props_stateflow_array of explicit_stateflow_array
  | Explicit_props_range of explicit_range
;;

type explicit_props_items = explicit_props_item list;;

type explicit_props = explicit_props_items;;

type explicit_data_item =
  | Explicit_data_binding of explicit_binding
  | Explicit_data_props of explicit_props
;;

type explicit_data_items = explicit_data_item list;;

type explicit_data_id =
  | Explicit_data_id of explicit_id
;;

type explicit_data_ssid =
  | Explicit_data_ssid of explicit_ssid_number
;;

type explicit_data = {
  explicit_data_id : explicit_data_id option;
  explicit_data_ssid : explicit_data_ssid option;
  explicit_data_name : string;
  explicit_data_items : explicit_data_item list;
}
;;

type explicit_junction_item =
  | Explicit_junction_item_binding of explicit_binding
  | Explicit_junction_item_autogen of explicit_autogen
;;

type explicit_junction_items = explicit_junction_item list;;

type explicit_junction = {
  explicit_junction_id : explicit_id;
  explicit_junction_items : explicit_junction_items;
}
;;

type explicit_event_item = explicit_binding;;

type explicit_event_items = explicit_event_item list;;

type explicit_event = {
  explicit_event_id : explicit_id;
  explicit_event_items : explicit_event_items;
}
;;

type explicit_eml = explicit_bindings;;

type explicit_simulink = explicit_bindings;;

type explicit_layout_info = explicit_bindings;;

type explicit_table_ui_prefs = explicit_bindings;;

type explicit_state_cell = {
  explicit_state_cell_id : explicit_id;
  explicit_state_cell_items : explicit_bindings;
}
;;

type explicit_transition_cell_items = explicit_bindings;;

type explicit_transition_cell = {
  explicit_transition_cell_id : explicit_id;
  explicit_transition_cell_items : explicit_transition_cell_items;
}
;;

type explicit_truth_table_item =
  | Explicit_truth_table_item_binding of explicit_binding
  | Explicit_truth_table_item_array of explicit_array
;;

type explicit_truth_table_items = explicit_truth_table_item list;;

type explicit_truth_table = explicit_truth_table_items;;

type explicit_chart_item =
  | Explicit_chart_item_binding of explicit_binding
  | Explicit_chart_item_children of explicit_children
  | Explicit_chart_item_sub_view_s of explicit_sub_view_s
  | Explicit_chart_item_eml of explicit_eml
  | Explicit_chart_item_active_state_output of explicit_active_state_output
  | Explicit_chart_item_plant_modeling_info

and explicit_chart_items = explicit_chart_item list

and explicit_chart = {
  explicit_chart_id : explicit_id;
  explicit_chart_items : explicit_chart_items;
}

and explicit_table_item =
  | Explicit_table_item_binding of explicit_binding
  | Explicit_table_item_layout_info of explicit_layout_info
  | Explicit_table_item_array of explicit_array
  | Explicit_table_item_table_ui_prefs of explicit_table_ui_prefs
  | Explicit_table_item_children of explicit_children

and explicit_table_items = explicit_table_item list

and explicit_table = {
  explicit_table_id : explicit_id;
  explicit_table_items : explicit_table_items;
}

and explicit_child =
  | Explicit_child_target of explicit_target
  | Explicit_child_chart of explicit_chart
  | Explicit_child_state of explicit_state
  | Explicit_child_transition of explicit_transition
  | Explicit_child_data of explicit_data
  | Explicit_child_junction of explicit_junction
  | Explicit_child_event of explicit_event
  | Explicit_child_table of explicit_table
  | Explicit_child_state_cell of explicit_state_cell
  | Explicit_child_transition_cell of explicit_transition_cell

and explicit_children = explicit_child list

and explicit_state_item =
  | Explicit_state_item_binding of explicit_binding
  | Explicit_state_item_children of explicit_children
  | Explicit_state_item_active_state_output of explicit_active_state_output
  | Explicit_state_item_sub_view_s of explicit_sub_view_s
  | Explicit_state_item_eml of explicit_eml
  | Explicit_state_item_simulink of explicit_simulink
  | Explicit_state_item_autogen of explicit_autogen
  | Explicit_state_item_truth_table of explicit_truth_table

and explicit_state_items = explicit_state_item list

and explicit_state = {
  explicit_state_id : explicit_id;
  explicit_state_items : explicit_state_items;
}
;;

type explicit_machine_item =
  | Explicit_machine_binding of explicit_binding
  | Explicit_machine_debug of explicit_debug
  | Explicit_machine_children of explicit_children
;;

type explicit_machine_items = explicit_machine_item list;;

type explicit_machine = {
  explicit_machine_id : explicit_id;
  explicit_machine_items : explicit_machine_items;
}
;;

type explicit_instance_items = explicit_bindings;;

type explicit_instance = {
  explicit_instance_id : explicit_id;
  explicit_instance_items : explicit_instance_items;
}
;;

type explicit_stateflow_section_items = explicit_stateflow_section_item list
and explicit_stateflow_section_item =
  | Explicit_stateflow_section_machine of explicit_machine
  | Explicit_stateflow_section_instance of explicit_instance
;;

type explicit_stateflow_section =
  | Explicit_stateflow_section of explicit_stateflow_section_items
;;

(** {6 General section} *)

type explicit_implementation_section_item =
  | Explicit_implementation_library of explicit_section
  (* | Explicit_implementation_matdata of matdata_section *)
  | Explicit_implementation_model of explicit_section
  | Explicit_implementation_stateflow of explicit_stateflow_section
;;

type explicit_model_sections = explicit_implementation_section_item list;;

type explicit_blockdiagram = {
  explicit_prolog : explicit_encoding;
  explicit_model_sections : explicit_model_sections;
  explicit_model_version : explicit_model_version;
}
;;

type explicit_xml_model = {
  explicit_model_ident : Mdl_ast.model_ident;
  explicit_blockdiagram : explicit_blockdiagram;
}
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
 End:
*)
