#######################################################################
#                                                                     #
#                          Ocaml-xml Library                          #
#                                                                     #
#                   Pierre Weis, INRIA Rocquencourt                   #
#                                                                     #
#  Copyright 2013-2018,                                               #
#  Institut National de Recherche en Informatique et en Automatique.  #
#  All rights reserved.                                               #
#                                                                     #
#  This file is distributed under the terms of the BSD License.       #
#                                                                     #
#######################################################################

# $Id: Caml.mk,v 1.1 2019-05-10 14:02:41 jpc Exp $

# This Makefile defines generic rules to compile various kinds of Caml source
# files and to automatically handle their dependencies.

# This Makefile also defines two entries:
# - clean (to clean up the compiled files)
# - depend (to recompute the dependency order).

# This Makefile should be included at the end of the Makefile that handles a
# set of Caml files (to build a library or an application).
# Simpy write at the end of your Makefile:
# include path_to_Caml.mk/Caml.mk

.PHONY: default all generation byt bin
.PHONY: clean clean-all dist-clean clean-spurious clean-compiled
.PHONY: clean-generated clean-caml-generated clean-depend
.PHONY: depend before-depend common-depend local-depend internal-depend
.PHONY: depend-printer before-depend-printer common-depend-printer
.PHONY: local-depend-printer internal-depend-printer
.PHONY: test test-all

# Main targets
all: generation byt bin

generation: $(CONFIG_TEXT_FILES) $(CAML_FILES)

# Generic clean up
COMPILED_EXT_TO_CLEAN=.cm* .o .a .annot .output .obj .lib .so .byt .bin

clean-compiled:
	@for EXT in $(COMPILED_EXT_TO_CLEAN); do \
	echo "Cleaning *$$EXT" && \
	  find . -name "*$$EXT" -exec $(RM) "{}" \; ; \
	done

SPURIOUS_EXT_TO_CLEAN=*~ .*~ a.out .\#*

clean-spurious:
	@for EXT in $(SPURIOUS_EXT_TO_CLEAN); do \
	echo "Cleaning *$$EXT" && \
	  find . -name "$$EXT" -exec $(RM) "{}" \; ; \
	done

clean-generated: clean-caml-generated

clean-caml-generated:
	@$(RM) $(CAML_GENERATED_FILES)

clean: clean-compiled

clean-all: clean-compiled clean-spurious clean-caml-generated
	$(MAKE) depend

dist-clean: clean-spurious clean

# Rebuilding dependencies
clean-depend:
	$(RM) .depend && $(TOUCH) .depend

before-depend: generation

common-depend: clean-depend
	@echo "Computing Caml file dependencies" && \
	$(CAMLC_DEP) $(CAML_INCLUDES) $(CAML_FILES) >> .depend

local-depend:

internal-depend:

depend:\
 before-depend common-depend\
 local-depend internal-depend\
 depend-printer\

# Rebuilding printer rules
before-depend-printer:

common-depend-printer:
	@echo "Computing OCamlgen printer generation rules" && \
	if test -n "$(CAML_PRINTER_SRC_FILES)"; then \
	  (for file in $(CAML_PRINTER_SRC_FILES); do \
	    echo "$$(dirname $$file)/$$(basename $$file .mli)_print.ml : \\" && \
	    echo "    $$file" && \
	    echo "	@echo \"Generating printer for $$file\" && \\" && \
	    echo "	\$$(CAMLC_PPR) $$file"; \
	  done) >> .depend; \
	fi

local-depend-printer:

internal-depend-printer:

depend-printer:\
 before-depend-printer common-depend-printer\
 local-depend-printer internal-depend-printer

# Compilation rules
.SUFFIXES:
.SUFFIXES: .ml .mli .mlin .mliin .mll .mly .mlm .mlms
.SUFFIXES: .cmo .cmi .cma .cmx .cmxa .obj
.SUFFIXES: .c .h .o .a .lib .so
.SUFFIXES: .indata .in

.ml.cmo:
	@echo "Byte compiling $<" && \
	$(CAMLC_BYT) -c $<

.mli.cmi:
	@echo "Compiling interface $<" && \
	$(CAMLC_BYT) -c $<

.ml.cmx:
	@echo "Binary compiling $<" && \
	$(CAMLC_BIN) -c $<

.mlin.ml:
	@echo "Expanding $< into $@" && \
	$(HTMLC) -i $< -o $@

.mliin.mli:
	@echo "Expanding $< into $@" && \
	$(HTMLC) -i $< -o $@

.mll.ml:
	@echo "Generating lexer $<" && \
	$(CAMLC_LEX) $<

.mly.ml:
	@echo "Generating parser $<" && \
	$(CAMLC_YAC) $<

.mly.mli:
	@echo "Generating parser $<" && \
	$(CAMLC_YAC) $<

.mlm.ml:
	@echo "Generating implementation for moca file $<" && \
	$(MOCAC) -c $<

.mlm.mli:
	@echo "Generating interface for moca file $<" && \
	$(MOCAC) -i $<

.mlms.ml:
	@echo "Generating implementation for moca file $<" && \
	$(MOCAC) -c $<

.mlms.mli:
	@echo "Generating interface for moca file $<" && \
	$(MOCAC) -i $<

.c.o:
	@echo "C compiling $<" && \
	$(CAMLC_CC) -c $<

.indata.in:
	@echo "Expanding $< into $@" && \
	$(HTMLC) -i $< -o $@

.mliin.mli_lt_4_02 .mlin.ml_lt_4_02:
	@echo "Expanding $< into $@" && \
	$(HTMLC_LT_4_02) -i $< -o $@

.mliin.mli_ge_4_02 .mlin.ml_ge_4_02:
	@echo "Expanding $< into $@" && \
	$(HTMLC_GE_4_02) -i $< -o $@

banner: banner.in
	@echo "Expanding $< into $@" && \
	$(HTMLC) -i banner.in -o banner

include .depend
