(***********************************************************************)
(*                                                                     *)
(*                          Ocaml-xml Library                          *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2013-2018,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: intercal.mli,v 1.1 2019-05-10 14:00:49 jpc Exp $ *)

(** Intercal operations.
    This module provides basic functions to manipulate intercal values.  *)

type ('a, 'b) intercal =
  'a option * ('b * 'a option) list
(** The type of intercal containing items of type ['a] and ['b]. *)
;;

type ('a, 'b) t = ('a, 'b) intercal
(** An alias for [intercal]. *)
;;

type ('a, 'b) sum =
| In_left of 'a
| In_right of 'b
;;

(** {6 Intercal constructors} *)

val empty : ('a, 'b) intercal
(** The empty intercal. *)
;;

val second : ('a, 'b) intercal -> ('b * 'a option) list
(** [second intercal] returns the second part of given intercal. *)
;;
val first : ('a, 'b) intercal -> 'a option
(** [first intercal] returns the first part of given intercal. *)
;;

val cons_alpha : 'a -> ('a, 'b) intercal -> ('a, 'b) intercal
(** [cons_alpha alpha intercal] returns a new intercal with [alpha] consed. If
    an ['a] was already inserted without a ['b] inserted, a failure will be
    raised. *)
;;
val cons_beta : 'b -> ('a, 'b) intercal -> ('a, 'b) intercal
(** [cons_beta beta intercal] returns a new intercal with [beta] consed. *)
;;

val singleton_alpha : 'a -> ('a, 'b) intercal
(** [singleton_alpha a] returns a new intercal with [a] as unique element. *)
;;
val singleton_beta : 'b -> ('a, 'b) intercal
(** [singleton_alpha b] returns a new intercal with [b] as unique element. *)
;;

(** {6 Test functions} *)

val is_empty : ('a, 'b) intercal -> bool
(** [is_empty intercal] returns true if [intercal] is empty. *)
;;
val is_head_alpha : ('a, 'b) intercal -> bool
(** [is_head_alpha intercal] returns true if the last cons item is of type
    alpha. *)
;;
val is_head_beta : ('a, 'b) intercal -> bool
(** [is_head_beta intercal] returns true if the last cons item is of type
    beta. *)
;;

(** {6 Iterators} *)

val rev : ('a, 'b) intercal -> ('a, 'b) intercal
(** [rev intercal] returns an intercal with the second part of the type
    constructor reversed. *)
;;
val fold_left :
  ('a -> 'b -> 'a) -> ('a -> 'c -> 'a) -> 'a -> ('b, 'c) intercal -> 'a
(** [fold_left alpha_folder beta_folder init intercal] folds [intercal] using
    [alpha_folder] for items of type ['a] and [beta_folder] for items of type
    ['b]. [init] represents the initial value for the folding process. *)
;;

val iter : ('a -> unit) -> ('b -> unit) -> ('a, 'b) intercal -> unit
(** [iter alpha beta intercal] iters through the intercal using [alpha] for
    items of type ['a] and [beta] for items of type ['b]. *)
;;

val map : ('a -> 'b) -> ('c -> 'd) -> ('a, 'c) intercal -> ('b, 'd) intercal
(** [map alpha beta intercal] map [intercal] using [alpha] for
    items of type ['a] and [beta] for items of type ['b]. *)
;;

val rev_append : ('a, 'b) intercal -> ('a, 'b) sum list -> ('a, 'b) sum list
(** [rev_append intercal list] appends in reverse order [intercal] to [list].
*)
;;

(** {6 Convertors} *)

val to_elements : ('a, 'b) intercal -> ('a, 'b) sum list
(** [to_elements intercal] returns the elements of [intercal]. *)
;;

val of_elements : ('a, 'b) sum list -> ('a, 'b) intercal
(** [of_elements elements] returns the intercal containing elements of
    [elements].
    Raises [Failure] if there are two consecutive elements of type ['a]. *)
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
 End:
*)
