(***********************************************************************)
(*                                                                     *)
(*                              OCamlGen                               *)
(*                                                                     *)
(*                       OCaml module Generator                        *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 1997-2018 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Clement Franchini <Clement.Franchini@inria.fr>                     *)
(*                                                                     *)
(***********************************************************************)

(* $Id: og_ident.ml,v 1.1 2019-05-10 14:02:41 jpc Exp $ *)

(* May be revise to_ident to avoid double verifications *)

(** {3 Type definitions for OCaml identifiers} *)

(* Projections are just identity functions: we suppose that projections are
   only done by the lexer, hence verification is just unnecessary. *)

type name = string;;

type capitalization =
  | Upper
  | Lower
  | Symbol
;;

let classify_name name =
  if String.length name = 0 then
    failwith "Og_ident: classify_name, invalid name with no letters" else
  match name.[0] with
  | 'A' .. 'Z' -> Upper
  | '_' | 'a' .. 'z' -> Lower
  | '~' | '!' | '@' | '%' | '^' | '&' | '*' | '-' | '=' | '+' | '|'
  | '<' | '>' | '/' | '?' -> Symbol
  | _ ->
    failwith (Printf.sprintf
      "Og_ident: classify_name, invalid name \'%s\'"
      name)
;;

type lowercase_name = name;;
external of_lowercase_name : lowercase_name -> name = "%identity"
;;

let to_lowercase_name name =
  if classify_name name = Lower then name else
  invalid_arg (Printf.sprintf
    "to_lowercase_name: not a lowercase name \'%s\'" name)
;;

type uppercase_name = name;;
external of_uppercase_name : uppercase_name -> name = "%identity"
;;
let to_uppercase_name name =
  if classify_name name = Upper then name else
  invalid_arg (Printf.sprintf
    "to_uppercase_name: not an uppercase name \'%s\'" name)
;;

type symbolic_name = name;;
external of_symbolic_name : symbolic_name -> name = "%identity"
;;
let to_symbolic_name name =
  if classify_name name = Symbol then name else
  invalid_arg (Printf.sprintf
    "to_symbolic_name: not an symbolic name \'%s\'" name)
;;

type ident =
  | Uppercase of uppercase_name
  | Lowercase of lowercase_name
  | Symbolic of symbolic_name
;;

let of_ident = function
  | Uppercase uid -> of_uppercase_name uid
  | Lowercase lid -> of_lowercase_name lid
  | Symbolic sid -> of_symbolic_name sid
;;

let to_ident name =
  match classify_name name with
  | Upper -> Uppercase (to_uppercase_name name)
  | Lower -> Lowercase (to_lowercase_name name)
  | Symbol -> Symbolic (to_symbolic_name name)
;;

let ident_of_lowercase_name lid = Lowercase lid;;
let ident_of_uppercase_name uid = Uppercase uid;;
let ident_of_symbolic_name sid = Symbolic sid;;

let ident_of_name = to_ident;;
let name_of_ident = of_ident;;

type longident =
  | Lident of ident
  | Ldot of longident * ident
  | Lapply of longident * longident
;;

let longident_of_name name = Lident (ident_of_name name);;
let longident_of_lowercase_name name = Lident (ident_of_name name);;
let longident_of_uppercase_name name = Lident (ident_of_name name);;

let short_name_of_longident = function
  | Lident id -> name_of_ident id
  | Ldot (_longident, id) -> name_of_ident id
  | Lapply (_foncident, _modident) -> assert false
;;

let rec name_of_longident = function
  | Lident id -> name_of_ident id
  | Ldot (lid, id) ->
    Printf.sprintf "%s.%s"
      (name_of_longident lid)
      (name_of_ident id)
  | Lapply (fonc_lid, mod_lid) ->
    Printf.sprintf "%s (%s)"
      (name_of_longident fonc_lid)
      (name_of_longident mod_lid)
;;

let split_longident = function
  | Lident ident -> None, ident
  | Ldot (longident, ident) -> Some longident, ident
  | Lapply (_longident_f, _longident_x) as longident ->
    failwith (Printf.sprintf
      "Og_ident.split_longident: %S can not be split"
      (name_of_longident longident))
;;

let unused_var_name lname =
  let name = of_lowercase_name lname in
  if name.[0] = '_' then lname else
  to_lowercase_name (Printf.sprintf "%c%s" '_' name)
;;
