(***********************************************************************)
(*                                                                     *)
(*                          Ocaml-xml Library                          *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2013-2018,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: ulist1.mli,v 1.1 2019-05-10 14:00:49 jpc Exp $ *)

(** The module defining nonempty lists. *)

type 'a list1 = 'a * 'a list;;

type 'a t = 'a list1;;

val length : 'a list1 -> int
(** Return the length (number of elements) of the given list. *)
;;

val cons : 'a -> 'a list1 -> 'a list1;;

val to_list : 'a list1 -> 'a list
(** [to_list l] returns the (usual) list of all the elements of the
  nonempty list [l]. *)
;;

val of_list : 'a list -> 'a list1
(** [of_list l] returns the nonempty list containing the elements of
   [l].
   Raise [Invalid_argument] if the list is empty. *)
;;

val rev_append : 'a list1 -> 'a list -> 'a list1
(** Concatenate a nonempty list to a list in reverse order.
   Tail-recursive
   (length of the first argument). *)
;;

val rev : 'a list1 -> 'a list1
(** [rev (a1, [a2; ...; an])] is [ (an, [...; a2; a1])] *)
;;

val iter : ('a -> unit) -> 'a list1 -> unit
(** [iter f (a1, [a2; ...; an])] applies function [f] in turn to
   [a1, a2, ..., an]. It is equivalent to
   [begin f a1; f a2; ...; f an; () end]. *)
;;

val map : ('a -> 'b) -> 'a list1 -> 'b list1
(** [map f (a1, [a2; ...; an])] applies function [f] to [a1, a2, ..., an],
   and builds the [list1] [(f a1, [f a2; ...; f an])]
   with the results returned by [f]. Not tail-recursive. *)
;;

val fold_left : ('a -> 'b -> 'a) -> 'a -> 'b list1 -> 'a
(** [fold_left f a (b1, [b2; ...; bn])] is
   [f (... (f (f a b1) b2) ...) bn]. *)
;;

val foldl : ('a -> 'a -> 'a) -> 'a * 'a list1 -> 'a
(** [foldl f (a, (a1, [a2; ...; an]))] is
   [f (... (f (f a a1) a2) a3 ...) an]. *)
;;

val fold_right : ('a -> 'b -> 'b) -> 'a list1 -> 'b -> 'b
(** [fold_right f (a1, [a2; ...; an]) b] is
   [f a1 (f a2 (... (f an b) ...))].
   Not tail-recursive. *)
;;

val foldr : ('a -> 'a -> 'a) -> 'a * 'a list1 -> 'a
(** [foldr f (a, (a1, [a2; ...; an-1; an]))] is
   [f a (f a1 (... (f an-1 an) ...))].
   Not tail-recursive. *)
;;

val fold_map :
  ('a -> 'b -> 'a * 'c) ->
  'a -> 'b list1 -> 'a * 'c list1
;;

val rev_fold_flat_map_end :
  ('a -> 'b -> 'a * 'c list) ->
  'a -> 'b * 'b list -> 'c list -> 'a * 'c list
;;

val fold_flat_map :
  ('a -> 'b -> 'a * 'c list) ->
  'a -> 'b list1 -> 'a * 'c list
;;

val iter2 : ('a -> 'b -> unit) -> 'a list1 -> 'b list1 -> unit
(** [iter2 f (a1, [a2; ...; an]) (b1, [b2; ...; bn])] calls in turn
   [f a1 b1, ..., f an bn].
   Raise [Invalid_argument] if the two lists have different lengths. *)
;;

val map2 : ('a -> 'b -> 'c) -> 'a list1 -> 'b list1 -> 'c list1
(** [map2 f (a1, [a2; ...; an]) (b1, [b2; ...; bn]] is
   [(f a1 b1, [f a2 b2; ...; f an bn]].
   Raise [Invalid_argument] if the two lists have different lengths.
   Not tail-recursive. *)
;;

val find : ('a -> bool) -> 'a list1 -> 'a
(** [find p l] returns the first element of the list [l]
   that satisfies the predicate [p].
   Raise [Not_found] if there is no value that satisfies [p] in the
   list [l]. *)
;;

val find_all : ('a -> bool) -> 'a list1 -> 'a list
(** [find_all p l] returns the list of elements of list [l]
   that satisfies the predicate [p]. *)
;;

val filter : ('a -> bool) -> 'a list1 -> 'a list
(** [filter p l] returns all the elements of the list [l]
   that satisfy the predicate [p].
   The order of the elements in the input list is preserved.  *)
;;

val partition : ('a -> bool) -> 'a list1 -> 'a list * 'a list
(** [partition p l] returns a pair of lists [(l1, l2)], where
   [l1] is the list of all the elements of [l] that satisfy the predicate [p],
   and [l2] is the list of all the elements of [l] that do not satisfy [p].
   The order of the elements in the input list is preserved. *)
;;

(** {6 Association lists} *)

val max_snd_assoc : ('a -> 'a -> int) -> ('b * 'a) list1 -> 'b * 'a
(** [max_snd_assoc l] returns the leftmost element of the list of pairs
   [l] with the maximum second component. *)
;;

val split : ('a * 'b) list1 -> 'a list1 * 'b list1
(** Transform a list1 of pairs into a pair of list1s:
   [split [(a1,b1); ...; (an,bn)]] is [([a1; ...; an], [b1; ...; bn])].
   Not tail-recursive. *)
;;

val combine : 'a list1 -> 'b list1 -> ('a * 'b) list1
(** Transform a pair of list1s into a list1 of pairs:
   [combine [a1; ...; an] [b1; ...; bn]] is
   [[(a1,b1); ...; (an,bn)]].
   Raise [Invalid_argument] if the two list1s
   have different lengths.  Not tail-recursive. *)
;;

(*
 Local Variables:
  compile-command: "cd ../..; make"
  End:
*)
