(***********************************************************************)
(*                                                                     *)
(*                          Ocaml-xml Library                          *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2013-2018,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: uname.mli,v 1.1 2019-05-10 14:00:49 jpc Exp $ *)

(** {3 Unique names} *)

type t = private {
  name : Name.t;
  uid : Uid.t;
}
(** The private type implementing unique names.
  A unique name is a bundle of a name and a uniquely identifying number.

  The name of a [Uname.t] value is a value of type [Name.t].
  The uniquely identifying number of a [Uname.t] value is a value of type [Uid.t].
*)
;;

val fresh : Name.t -> t
(** Given a [Name.t] value returns a unique name with this name and
   a fresh uniquely identifying number. *)
;;

val string_of_name : t -> string
(** Returns the string representation of the [Name.t] component of a unique
  name. *)
;;

val to_string : t -> string
(** Returns the string representation of a unique name: the concatenation of
  its name and its unique id. *)
;;

val to_delimited_string : Name.prefix -> Name.suffix -> t -> string
(** [to_delimited_string prefix suffix uname] returns the concatenation of
  the string representation of the name component of the unique name
  argument, followed by the prefix string, then the string representation
  of the unique id of the unique name, followed by the suffix string. *)
;;

val dummy : Name.t -> t
(** Creates a unique name with a dummy uniquely identifying number. *)
;;
val is_dummy : t -> bool
(** Tests if a unique name has a dummy uniquely identifying number. *)
;;
val make_dummy : string -> t
(** Creates a dummy unique name with name component made out of the string
  argument. *)
;;

val compare : t -> t -> int
(** Compare the unique id of two unique names. *)
;;
val same : t -> t -> bool
(** [same uname1 uname2] is equivalent to [compare uname1 uname2 = 0]. *)
;;
val same_name : t -> t -> bool
(** Tests if the [Name.t] component of two [Uname.t] values are equal. *)
;;

(*
val name : t -> Name.t;;
val uid : t -> Uid.t
(** Accessors for unique names components. *)
;;
*)

(*
 Local Variables:
  compile-command: "cd ../..; make"
  End:
*)
