(***********************************************************************)
(*                                                                     *)
(*                          Ocaml-xml Library                          *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2013-2018,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: lib_strings.mli,v 1.1 2019-05-10 14:00:49 jpc Exp $ *)

(** {3 A collection of functions for string manipulation} *)

(** {6 String searching} *)

val starts_by : string -> int -> string -> bool
(** [starts_by s idx pat] tests if, starting from index [idx],
  the string [s] matches the pattern [pat].

  The empty string never matches a non empty pattern.
  Every string matches the empty pattern.

  Raises [Invalid_argument] if [idx] is not a valid index of [s]. *)
;;

val pos_string : string -> int -> char -> int
(** [pos_string s idx c] finds the position of character [c]
  in string [s] starting from index [idx] and looking forward to the end
  of string [s].

  Raises exception [Not_found] if there is no such position.
  Raises [Invalid_argument] if [idx] is not a valid index of [s]. *)
;;

val pos_string_rev : string -> int -> char -> int
(** [pos_string_rev s idx c] finds the position of the occurrence of
  the character [c] that precedes the index [idx], which should be a
  valid index into [s].

  Raises exception [Not_found] if there is no such position.
  Raises [Invalid_argument] if [idx] is not a valid index of [s]. *)
;;

(** {6 String extraction} *)

val extract_string : string -> int -> int -> string
(** [extract_string s start stop] returns the sub string of [s]
  starting at index [i_start] (inclusively) to index [i_stop] (excluded). *)
;;

val sub_string_from_to : string -> int -> char -> char -> string
(** [sub_string_from_to s idx starter stopper] finds the sub string of
  [s] that is (strictly) comprised between the delimiters [starter]
  and [stopper]: that string starts after the first occurrence of
  character [starter] and ends before the first following
  occurrence of character [stopper]. Search for [starter] begins at
  index [idx].

  Raises exception [Not_found] if there is no such position.
  Raises [Invalid_argument] if [idx] is not a valid index of [s]. *)
;;

val from_char_to_end : string -> char -> string
(** [from_char_to_end s idx c] returns the sub string of string [s] that
  starts from the position following the last occurrence of
  character [c], and ends to the end of the string [s].
  If [c] is not found, the string [s] is returned. *)
;;

val from_char_to : string -> char -> int -> string
(** [from_char_to s c idx] returns the sub string of string [s] that
  starts from the position following the last occurrence of
  character [c] preceding [idx], and ends to [idx - 1].

  If [c] is not found, the sub-string of [s] starting from [0]
  to [idx - 1] is returned. *)
;;

val from_to_char : string -> int -> char -> string
(** [from_to_char s idx c] returns the sub string of string [s] that
  starts from the valid position [idx] to the position preceding the
  first occurrence of character [c] following [idx].

  If [c] is not found, the sub string from [idx] to the end of
  [s] is returned.

  Raises [Invalid_arg] if [idx] is not a valid index into [s]. *)
;;

val split_string_at_comma : string -> string list
(** Splits a comma separated string. *)

(** {6 String alignment} *)

val center_gen : char -> string -> int -> string
(** [center_gen c s len] centers [s] into a line made of [c] characters.

  [center_gen c s len] returns a string of length [len] that contains
  string [s] centered: the result is filled with as many [c]
  characters on the left and on the right as to get [s] centered.
  When [s] cannot fit centered in [len] characters (meaning
  [len - String.length s] is odd), there is one [c] character in excess
  on the right of [s].

  Raises [Failure] if string [s] cannot fit within [len] characters.
*)
;;

val center : string -> int -> string
(** [center] is a shorthand for [center_gen ' ']. *)
;;

val flush_left_gen : char -> string -> int -> string
(** [flush_left_gen c s len] Similar to [center_gen] for flushing a string
  on the left of a line made of [c] characters.

  Raises [Failure] if string [s] cannot fit within [len] characters. *)
;;

val flush_right_gen : char -> string -> int -> string
(** [flush_right_gen c s len] Similar to [center_gen] for flushing a string
  on the left of a line made of [c] characters.

  Raises [Failure] if string [s] cannot fit within [len] characters. *)
;;

val flush_left : string -> int -> string
(** [flush_left] is a shorthand for [flush_left_gen ' ']: the string argument is
  flushed on the left of a blank line of the given length.

  Raises [Failure] if string [s] cannot fit into a line of the given
  length. *)
;;
val flush_right : string -> int -> string
(** [flush_right] is a shorthand for [flush_right_gen ' ']: the string argument is
  flushed on the right of a blank line of the given length.

  Raises [Failure] if string [s] cannot fit into a line of the given
  length. *)
;;

val spaces : int -> string
(** [spaces n] returns a fresh string of [n] space character (ASCII 32). *)
;;

val suppress_spaces : string -> string
(** [suppress_spaces s] returns a fresh string by removing from [s] all space
  (ASCII 32) and tabulation (ASCII 9) characters. *)
;;

(** {6 Case manipulation for ISO-latin characters and character strings} *)

val lowercase_char : char -> char
(** Lowercase mapping for the ISO-8859-1 character set. *)
;;
val uppercase_char : char -> char
(** Uppercase mapping for the ISO-8859-1 character set. *)
;;

val uppercase : string -> string
(** Uppercase mapping for strings in the ISO-8859-1 character
  set encoding. *)
;;
val lowercase : string -> string
(** Lowercase mapping for strings in the ISO-8859-1 character
  set encoding. *)
;;
val capitalize : string -> string
(** Capitalize mapping for strings in the ISO-8859-1 character
  set encoding. *)
;;
val uncapitalize : string -> string
(** Uncapitalize mapping for strings in the ISO-8859-1 character
  set encoding. *)
;;

(** {6 Input to string} *)

val input_string : int -> in_channel -> string
(** [input_string len ic] reads characters from input channel [ic].
 The string result contains [len] characters or all the available characters
 if there are not enough available characters in [ic].

 Raise [End_of_file] if there are no characters to read. *)
;;

val string_of_ic : in_channel -> string
(** [string_of_ic ic] reads all the available characters in input channel [ic].
 Returns the empty string if there are no characters to read. *)
;;

val iter_channel : (string -> unit) -> in_channel -> unit
(** [iter_channel ic f] iterates function [f] on all successive string chunks
 read in input channel [ic]. *)
;;
