(***********************************************************************)
(*                                                                     *)
(*                          Ocaml-xml Library                          *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2013-2018,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: location.ml,v 1.1 2019-05-10 14:00:49 jpc Exp $ *)

open Lexing;;

type file_name = string;;

type position = Lexing.position = {
  pos_fname : file_name;
  pos_lnum : int;
  pos_bol : int;
  pos_cnum : int;
}
;;

(** The location of an AST node,
  beginning and ending position of its corresponding source text. *)
type t = {
  loc_start : position;
  loc_end : position;
}
;;

let mk_loc loc_start loc_end = {
  loc_start = loc_start;
  loc_end = loc_end;
}
;;

let get_lexbuf_loc lexbuf =
  mk_loc lexbuf.Lexing.lex_start_p lexbuf.Lexing.lex_curr_p
;;

let set_lexbuf_filename_positions lexbuf fname =
  lexbuf.Lexing.lex_start_p <-
    { lexbuf.lex_start_p
      with Lexing.pos_fname = fname; };
  lexbuf.Lexing.lex_curr_p <-
    { lexbuf.lex_curr_p
      with pos_fname = fname; }
;;

let init lexbuf fname =
  let init_pos = {
    pos_fname = fname;
    pos_lnum = 1;
    pos_bol = 0;
    pos_cnum = 0;
  } in
  lexbuf.Lexing.lex_start_p <- init_pos;
  lexbuf.Lexing.lex_curr_p <- init_pos
;;

let lexing_from_file fname =
  let ic = Pervasives.open_in fname in
  let lexbuf = Lexing.from_channel ic in
  set_lexbuf_filename_positions lexbuf fname;
  ic, lexbuf
;;

let lexing_from_explicit_file efname =
  let ic = Path.open_explicit_in efname in
  let lexbuf = Lexing.from_channel ic in
  set_lexbuf_filename_positions lexbuf (Path.of_explicit_file_name efname);
  ic, lexbuf
;;

let get_symbol_loc () =
  mk_loc (Parsing.symbol_start_pos ()) (Parsing.symbol_end_pos ())
;;

let get_rhs_loc n =
  mk_loc (Parsing.rhs_start_pos n) (Parsing.rhs_end_pos n)
;;

let dummy = mk_loc Lexing.dummy_pos Lexing.dummy_pos;;

let print_token_location ppf (start_pos, end_pos) =
  if start_pos.pos_lnum = end_pos.pos_lnum then
    (* Starts and ends on a same line. *)
    Format.fprintf ppf
      "line %d, characters %d-%d"
      start_pos.pos_lnum
      (start_pos.pos_cnum - start_pos.pos_bol)
      (end_pos.pos_cnum - end_pos.pos_bol)
  else
    Format.fprintf ppf
      "line %d, character %d, line %d, character %d"
      start_pos.pos_lnum
      (start_pos.pos_cnum - start_pos.pos_bol)
      end_pos.pos_lnum
      (end_pos.pos_cnum - end_pos.pos_bol)
;;

let print_location ppf loc =
  Format.fprintf ppf "File \"%s\", %a:"
    loc.loc_start.pos_fname
    print_token_location
      (loc.loc_start, loc.loc_end)
;;

(*
 Local Variables:
  compile-command: "cd ../..; make"
  End:
*)
