(***********************************************************************)
(*                                                                     *)
(*                          Ocaml-xml Library                          *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2013-2018,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: uencoding_types.mli,v 1.1 2019-05-10 14:00:49 jpc Exp $ *)

type byte_encoding =
  | Us_ascii
    (** American Standard Code for Information Interchange or US-ASCII *)
  | Iso_8859_1
    (** West european encoding, or Latin-1 *)
  | Iso_8859_15
    (** French encoding, or Latin-9 *)
  | Cp_1252
    (** Code page 1252, or Windows-1252 is a variant of Latin-1
      with the range 128 to 159 (0x80 to 0x9F) replaced with additional
      characters (including all the characters missing from encoding Latin-1
      and provided by Latin-9).
      Also confusedly named ANSI (after the American National Standards
      Institute). *)
(** An encoding is a code that pairs each character from a given repertoire
  with a bit pattern.
  A [byte_encoding] represents an encoding using only one octet for each
  character of the repertoire. *)
;;

type character_encoding =
  | US_ASCII
  | ISO_8859_1
  | ISO_8859_15
  | SHIFT_JIS
  | UTF_8
  | CP_1252
(** An encoding is a code that pairs each character from a given repertoire
  with a bit pattern.

  A [character_encoding] represents any encoding from simple byte encodings
  (see [byte_encoding] above), to variable length encodings such as Shift-JIS
  that uses one or two octets or UTF-8 that uses variable-width encoding from
  one to 6 octets for each character.
  Most encodings are preserving the basic US-ASCII encoding repertoire (or a
  large part of it): as a consequence, roman characters have the same
  code in all encodings and most one byte codes in the range [0 .. 127] refer
  to the same characters.

  This library supports the following character encodings:
  - US-ASCII (the American Standard Code for Information Interchange or US-ASCII)
    contains the repertoire of regular latin characters (Upper-case and
    lower-case letters with no dyacritic signs) plus usual punctuation
    characters.
  - ISO-8859-1 (the West european encoding, or Latin-1)
    contains the repertoire of characters used in western european languages
    including (most common) letters with dyacritic signs.
  - ISO-8859-15 (the french encoding, or Latin-9) contains the entire
    repertoire used in french, including all the french characters missing
    from encoding Latin-1.
  - Shift-JIS (Shift Japanese Industrial Standard)
    contains the repertoire of all japanese charaters. It is a variable-width
    encoding using one or two bytes.
  - UTF-8 (UCS transformation format 8 bits) is a variable-width encoding
    for the Universal Character Set (Unicode).
  - Windows-1252 (the Code Page 1252 or CP1252 or ANSI)
 *)
;;

type character_encoding_name = string
(** The string representation of a character encoding: this is the IANA
  preferred mime name of the encoding. *)
;;

(*
 Local Variables:
  compile-command: "cd ../..; make"
  End:
*)
