(***********************************************************************)
(*                                                                     *)
(*                          Ocaml-xml Library                          *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2013-2018,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: wstring.mli,v 1.1 2019-05-10 14:00:49 jpc Exp $ *)

type byte = private char
(** A byte is the unit of digital information; it most commonly consists
  of an octet (a eight bit data).
  A byte is an integer from 0 (0x00) to 255 (0xFF);
  bytes in the inclusive range from 32 (0x20) to 126 (0x7E) are used to code
  usual printable characters for the Western Latin languages (ASCII standard).
  The international standard IEC 80000-13 codified the notion of byte. *)
;;

val of_byte : byte -> char
(** Extracts the [char] internal representation of a [byte] value. *)
;;

val input_byte : in_channel -> byte
(** [input_byte ic] reads a byte from [ic]. Raise End_of_file if an end of file
     was reached. *)
;;

(*
val char_of_byte : Uencoding_types.byte_encoding -> byte -> char;;
val byte_of_char : byte_encoding -> char -> byte
(** Conversion between bytes and chars according to the given encoding. *)
;;
*)

type unicode = int
(** The type of unicode characters codes. *)
;;

type wchar = private int32
(** The type of chararacters in the utf8 encoding. This encoding needs up to
  four octets hence the representation as a 32bits integer. *)
;;

type wstring = private string
(** The type of wide strings containing unicode characters encoded in
  utf-8. *)
;;

val of_wchar : wchar -> int32
(** Extracts the [int32] internal representation of a [wchar] value. *)
;;

val of_wstring : wstring -> string
(** Extracts the [string] internal representation of a [wstring] value. *)
;;

val length : wstring -> int
(** Return the length (number of characters) of the given wide string. *)
;;

val nth : wstring -> int -> wchar
(** Return the [n]-th character of the given wide string.
  The first character is at position 0.
  Raise [Failure "nth"] if the wide string is too short.
  Raise [Invalid_argument "Wstring.nth"] if [n] is negative. *)
;;

val sub : wstring -> int -> int -> wstring
(** [Wstring.sub ws start len] returns a wide string of length [len],
  containing the substring of [ws] that starts at position [start] and
  has length [len].

  Raise [Invalid_argument] if [start] and [len] do not
  designate a valid substring of [s]. *)
;;

val concat : wstring -> wstring list -> wstring
(** [Wstring.concat sep sl] concatenates the list of wide strings [sl],
  inserting the separator wide string [sep] between each. *)

val guess_encoding : string -> Uencoding_types.byte_encoding
(** Try to guess the encoding of the string argument:
  if all characters are lower than 128 then ASCII is returned.
  Otherwise, if C1 characters are found (from 0x80 to 0x9F) then Windows-1252
  is returned else iso-8859-15 (Latin-9) is returned. *)
;;

val wide_encode : unicode -> wchar
(** [wide_encode char] encodes unicode [char] into utf-8. *)
;;

val byte_to_wchar : Uencoding_types.byte_encoding -> byte -> wchar
(** [byte_to_wchar encoding byte] converts [byte] representing a character with
  [encoding] into the corresponding character in utf-8. *)
;;

val string_to_wstring :
  Uencoding_types.character_encoding ->
  string ->
  wstring
(** [string_to_wstring encoding str] converts [str] to a utf-8 string
  assuming that encoding of [str] is [encoding]. *)
;;

val to_utf_8 : string -> wstring
(** [to_utf_8 str] converts [str] to utf-8 string guessing original
 encoding. *)
;;

val fold : ('a -> wchar -> 'a) -> 'a -> wstring -> 'a
(** [Wstring.fold f a wstr] folds function [f] on [wstr] starting from [a].
  If [wstr] has chars [c1; ..; cn;] then [Wstring.fold f a wstr] is
  [f (... (f (f a c1) c2) ...) cn]. *)
;;

val iter : (wchar -> unit) -> wstring -> unit
(** [Wstring.iter f s] applies function [f] in turn to all
  the characters of [s]. It is equivalent to code sequence
  [f s.[0]; f s.[1]; ...; f s.[Wstring.length s - 1]; ()]. *)
;;

val iteri : (int -> wchar -> unit) -> wstring -> unit
(** Same as {!Wstring.iter}, but the
  function is applied to the index of the element as first argument
  (counting from 0), and the character itself as second argument. *)
;;

val map : (wchar -> wchar) -> wstring -> wstring
(** [Wstring.map f s] applies function [f] in turn to all
   the characters of wide string [s] and returns the resulting wide string. *)
;;

val trim : wstring -> wstring
(** Return a copy of the argument, without leading and trailing
  whitespace.  A whitespace is one of [' '], ['\012'], ['\n'], ['\r'], and
  ['\t']. Returns the wide string argument if there is no leading nor trailing
  whitespace character. *)
;;

(*
val wide_scan : wstring -> wchar array
(** [wide_scan wstr] reads the UTF-8 encoded characters in [wstr]. *)
;;
*)

val character_encoding_of_encoding_name :
  Uencoding_types.character_encoding_name -> Uencoding_types.character_encoding
(** [character_encoding_of_encoding_name encoding_name] returns the
  [character_encoding] value that has name [encoding_name].
  Raise [Failure] if [encoding_name] argument is unknown.
*)
;;

val encoding_name_of_character_encoding :
  Uencoding_types.character_encoding -> string
(** [encoding_name_of_character_encoding encoding] returns the (IANA
  preferred) string name of the [encoding] argument.
*)
;;

val wide_decode : wchar -> unicode
(** [wide_decode wchar] returns the unicode value of [wchar]. *)
;;

(*
val unicode_to_ascii : unicode -> byte
(** Raise [Failure] if unicode argument does not belong to the ASCII
    repertoire. *)
;;
val unicode_of_ascii : byte -> unicode
(** Raise [Failure] if byte argument does not belong to the ASCII
    repertoire. *)
;;

val unicode_to_latin_1 : unicode -> byte
(** Raise [Failure] if unicode argument does not belong to the Latin1
    repertoire. *)
;;
val unicode_of_latin_1 : byte -> unicode;;

val unicode_to_latin_9 : unicode -> byte
(** Raise [Failure] if unicode argument does not belong to the Latin9
    repertoire. *)
;;
val unicode_of_latin_9 : byte -> unicode;;

val unicode_to_cp_1252 : unicode -> byte
(** Raise [Failure] if unicode argument does not belong to the Windows1252
    repertoire. *)
;;
val unicode_of_cp_1252 : byte -> unicode;;
*)

val string_of_wstring : Uencoding_types.character_encoding -> wstring -> string
(** [string_of_wstring encoding ws] given a wide string [ws], encode each wide
    char according to [encoding] and returns the resulting string. *)
;;

type t = wstring
(** An alias for the type [wstring]. *)
;;

val compare : t -> t -> int
(** The comparison function for wide strings, with the same specification as
    {!Pervasives.compare}.  Along with the type [t], this function [compare]
    allows the module [Wstring] to be passed as argument to the functors
    {!Set.Make} and {!Map.Make}. *)
;;

val transcode :
  Uencoding_types.character_encoding -> string ->
  Uencoding_types.character_encoding -> string
;;

(*
 Local Variables:
  compile-command: "cd ../..; make"
  End:
*)
