(***********************************************************************)
(*                                                                     *)
(*                              OCamlGen                               *)
(*                                                                     *)
(*                       OCaml module Generator                        *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 1997-2018 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Clement Franchini <Clement.Franchini@inria.fr>                     *)
(*                                                                     *)
(***********************************************************************)

(* $Id: og_printer_xml.ml,v 1.1 2019-05-10 14:02:41 jpc Exp $ *)

(* Runtime library for XML pretty-printers generated with OCamlGen. *)

open Og_printer;;

(** {6 Xml basics functions} *)

let xml_open_tag ppf tag_name = Format.fprintf ppf "<%s>" tag_name;;

let xml_close_tag ppf tag_name = Format.fprintf ppf "</%s>" tag_name;;

let xml_autoclose_tag ppf tag_name = Format.fprintf ppf "<%s />" tag_name;;

let xml_escape_char = function
  | '"' -> "&quot;" (* Helping Emacs '"' *)
  | '&' -> "&amp;"
  | '\'' -> "&apos;"
  | '<' -> "&lt;"
  | '>' -> "&gt;"
  | c -> String.make 1 c
;;

let xml_escape_string s =
  let l = String.length s in
  let buff = Buffer.create l in
  let rec loop i =
    if i = l then
      Buffer.contents buff
    else
      begin
        Buffer.add_string buff (xml_escape_char s.[i]);
        loop (succ i)
      end in
  loop 0
;;

let xml_write_string ppf s = Format.fprintf ppf "%s" (xml_escape_string s);;

let xml_with_tag tag_name pp_item ppf x =
  Format.fprintf
    ppf "@[<hv 2>%a@,%a@;<0 -2>%a@]"
    xml_open_tag tag_name
    pp_item x
    xml_close_tag tag_name
;;

(** {6 OCamlGen xml basics functions} *)

let og_xml_escape_char c =
  let s = xml_escape_char c in
  if String.length s = 1 then
    match c with
    | ' '
    | '\n'
    | '\r'
    | '\t' as c -> Printf.sprintf "&#x%02X;" (int_of_char c)
    | c -> String.make 1 c
  else
    s
;;

let split part_length s =
  let lim = String.length s in
  let rec loop accu i =
    if lim <= i then
      List.rev accu
    else
      let length = min part_length (lim - i) in
      let part = String.sub s i length in
      loop (part :: accu) (i + length) in
  loop [] 0
;;

let og_xml_write_string ppf s =
  let og_xml_write_string_part ppf part =
    let l = String.length part in
    let rec loop ppf i =
      if i = l then
        ()
      else
        Format.fprintf
          ppf "%s%a" (og_xml_escape_char part.[i]) loop (succ i) in
    loop ppf 0 in
  let rec loop ppf = function
    | [] -> ()
    | [ part; ] -> og_xml_write_string_part ppf part
    | part :: parts ->
      Format.fprintf
        ppf "%a@,%a" og_xml_write_string_part part loop parts in
  loop ppf (split 20 s)
;;

(* Printers for basic monomorphic values. *)

let printer_xml_Unit _ctx ppf () =
  xml_with_tag "unit" xml_write_string ppf "()"
and printer_xml_Bool _ctx ppf b =
  xml_with_tag "bool" xml_write_string ppf (if b then "true" else "false")
and printer_xml_Char _ctx ppf c =
  xml_with_tag "char" xml_write_string ppf (xml_escape_char c)
and printer_xml_String _ctx ppf s =
  xml_with_tag "string" og_xml_write_string ppf s
and printer_xml_Bytes = printer_Bytes
and printer_xml_Format = printer_Format
and printer_xml_Format4 = printer_Format4
and printer_xml_Format6 = printer_Format6
;;

let printer_xml_Int _ctx ppf i =
  xml_with_tag "int" (fun ppf -> Format.fprintf ppf "%i") ppf i
and printer_xml_Int32 _ctx ppf i =
  xml_with_tag "int32" (fun ppf -> Format.fprintf ppf "%lil") ppf i
and printer_xml_Int64 _ctx ppf i =
  xml_with_tag "int64" (fun ppf -> Format.fprintf ppf "%LiL") ppf i
and printer_xml_Nativeint _ctx ppf i =
  xml_with_tag "nativeint" (fun ppf -> Format.fprintf ppf "%nin") ppf i
and printer_xml_Float _ctx ppf f =
  xml_with_tag "int" (fun ppf -> Format.fprintf ppf "%f") ppf f
;;

(* Printers for basic polymorphic values. *)

let printer_xml_List pp_a _ctx ppf list =
  let rec loop ppf = function
    | [] -> ()
    | [x] -> pp_a Og_precedence.Ck_toplevel ppf x
    | x :: xs ->
      Format.fprintf ppf "%a@,%a" (pp_a Og_precedence.Ck_toplevel) x loop xs in
  xml_with_tag "list" loop ppf list
and printer_xml_Array pp_a _ctx ppf array =
  let lim = Array.length array in
  let rec loop ppf i =
    if i = lim then
      ()
    else if i = pred lim then
      pp_a Og_precedence.Ck_toplevel ppf array.(i)
    else
      Format.fprintf
        ppf "%a@,%a"
        (pp_a Og_precedence.Ck_toplevel) array.(i) loop (succ i) in
  xml_with_tag "array" loop ppf 0
;;

let printer_xml_Option px_a ctx ppf = function
  | None ->
    ignore ctx; Format.fprintf ppf "%,@[<hv 2><constructor>@,";
    Format.fprintf ppf "%,@[<hv 2><lident>@,"; Format.fprintf ppf "%,None";
    Format.fprintf ppf "%,@;<0 -2></lident>@]";
    Format.fprintf ppf "%,@;<0 -2></constructor>@]"
  | Some a ->
    ignore ctx; Format.fprintf ppf "%,@[<hv 2><constructor>@,";
    Format.fprintf ppf "%,@[<hv 2><lident>@,"; Format.fprintf ppf "%,Some";
    Format.fprintf ppf "%,@;<0 -2></lident>@]"; Format.fprintf ppf "%,@,";
    px_a Og_precedence.Ck_argument ppf a;
    Format.fprintf ppf "%,@;<0 -2></constructor>@]"
and printer_xml_Ref px_a ctx ppf = function
  | { contents = a; } ->
    ignore ctx; Format.fprintf ppf "%,@[<hv 2><constructor>@,";
    Format.fprintf ppf "%,@[<hv 2><lident>@,"; Format.fprintf ppf "%,ref";
    Format.fprintf ppf "%,@;<0 -2></lident>@]"; Format.fprintf ppf "%,@,";
    px_a Og_precedence.Ck_argument ppf a;
    Format.fprintf ppf "%,@;<0 -2></constructor>@]"
;;

(* Printers for functional, abstract and polymorphic values. *)

let printer_xml_Fun _pp_a _pp_b _ctx ppf _function_item =
  xml_autoclose_tag ppf "fun"
and printer_xml_Poly _ctx ppf _value =
  xml_autoclose_tag ppf "poly"
and printer_xml_Abstract _ctx ppf _value =
  xml_autoclose_tag ppf "abstract"
and printer_xml_Lazy _ctx ppf () =
  xml_write_string ppf "*"
and printer_xml_External _ctx ppf _value =
  xml_autoclose_tag ppf "external"
;;
