(***********************************************************************)
(*                                                                     *)
(*                              OCamlGen                               *)
(*                                                                     *)
(*                       OCaml module Generator                        *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 1997-2018 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Clement Franchini <Clement.Franchini@inria.fr>                     *)
(*                                                                     *)
(***********************************************************************)

(* $Id: og_scanner.ml,v 1.1 2019-05-10 14:02:41 jpc Exp $ *)

(** {3 Runtime library for pretty-scanners generated with OCamlGen} *)

type ('a, 'b, 'c, 'd) delimiter_format =
  ('a, 'b, 'c, 'd, 'd, 'a) format6
;;

type ('a, 'b) scanf_delimiter_format =
  ('a, Scanf.Scanning.in_channel, 'b, 'a -> 'a) delimiter_format
;;

type 'a scanner_delimiter_format = (unit, 'a) scanf_delimiter_format;;

let id x = x;;

let scanner_unit ib =
  Scanf.bscanf ib "%, ( )" ()
;;
let scanner_Unit = scanner_unit;;

(* Reading strings with double quotes. *)
let scanner_quoted_string ib = Scanf.bscanf ib "%S" id;;
let scanner_String = scanner_quoted_string;;

(*
let scanner_bytes ib = Scanf.bscanf ib "%s" (fun s -> Bytes.of_string s);;
*)
let scanner_Bytes ib = Scanf.bscanf ib "%S" (fun s -> Bytes.of_string s);;

(* Reading characters with quotes. *)
let scanner_quoted_char ib = Scanf.bscanf ib "%C" id;;
let scanner_Char = scanner_quoted_char;;

(* Reading numbers that may be parenthesized (such as (-1). *)
let scanner_quoted_int ib =
  Scanf.bscanf ib "%0c" (function
  | '(' -> (* ) *)
    Scanf.bscanf ib "( %i )" id
  | _ -> Scanf.bscanf ib "%i" id)
;;
let scanner_Int = scanner_quoted_int;;

let scanner_quoted_nativeint ib =
  Scanf.bscanf ib "%0c" (function
  | '(' -> (* ) *)
    Scanf.bscanf ib "( %ni )" id
  | _ -> Scanf.bscanf ib "%ni" id)
;;
let scanner_Nativeint = scanner_quoted_nativeint;;

let scanner_quoted_int32 ib =
  Scanf.bscanf ib "%0c" (function
  | '(' -> (* ) *)
    Scanf.bscanf ib "( %li )" id
  | _ -> Scanf.bscanf ib "%li" id)
;;
let scanner_Int32 = scanner_quoted_int32;;

let scanner_quoted_int64 ib =
  Scanf.bscanf ib "%0c" (function
  | '(' -> (* ) *)
    Scanf.bscanf ib "( %Li )" id
  | _ -> Scanf.bscanf ib "%Li" id)
;;
let scanner_Int64 = scanner_quoted_int64;;

let scanner_quoted_float ib =
  Scanf.bscanf ib "%0c" (function
  | '(' -> (* ) *)
    Scanf.bscanf ib "( %f )" id
  | _ -> Scanf.bscanf ib "%f" id)
;;
let scanner_Float = scanner_quoted_float;;

(* Scanners for polymorphic basic values *)

(*
type ('a, 'b, 'c, 'd) scan_separator =
  ('a, Scanf.Scanning.in_channel, 'b, unit -> 'c, 'a -> 'd, 'd) format6
;;

  Scanning a list of elements with user specified conventions:
  [scanner_list fmt scanner_a ib] scans input [ib]
  to read a list of elements read with [scanner_a] and separated by
  Format string [fmt].

  For instance:
  - separator [";"] is the Caml convention to parse list of elements;
  - separator [" ; "] allows arbitrary spacing before and after the semicolon.
  - separator [" "] means elements separated by spaces;
  - separator [" %_1[;] "] allows list of elements separated either by
  a semicolon or spaces.

  Note: [scanner_list] reads input as much as possible until scanning cannot
  proceed.
  As a consequence, [scanner_list] always succeeds: if the separator [fmt]
  cannot be read or [scanner_a] fails to scan a new element, [scanner_list]
  returns the list of elements read so far. Furthermore, separator [fmt] may
  appear after the last element of the list: scanning will go on until [fmt]
  has been read.
*)
let scanner_list (fmt : _ scanner_delimiter_format) scanner_a ib =

  let rec loop accu ib =
    try
      let accu = scanner_a ib :: accu in
      try
        Scanf.bscanf ib fmt ();
        loop accu ib
      with _ -> accu
    with _ -> accu in

  List.rev (loop [] ib)
;;

(* Scanning a list with Caml conventions *)
let scanner_List scanner_a ib =

  let rec rl accu ib =
    Scanf.bscanf ib "%0c" (function
    | (* [ *) ']' -> accu
    | _ ->
      Scanf.bscanf ib "%r " scanner_a (fun x ->
        Scanf.bscanf ib "%0c" (function
        | ';' ->
          Scanf.bscanf ib "; " rl (x :: accu) ib
        | _ ->
          rl (x :: accu) ib))) in
  Scanf.bscanf ib "[ %r ]" (rl []) (fun l -> List.rev l)
;;

(*
  Scanning an option with user conventions:
  The option must be either
  - nothing at all, or
  - successful scanning of argument format [fmt],
    followed by a successful scan of option value with scanner argument
    [scanner_a].
*)
(*
let scanner_option =
  let scan_option fmt fmt_start scanner_a ib =
    if Scanning.end_of_input ib then None else
    Scanf.bscanf ib "%0c" (fun c ->
      if c <> fmt_start then None else
      (Scanf.bscanf ib fmt (); Some (scanner_a ib))) in
  (fun fmt ->
    let fmt_start = (string_of_format fmt).[0] in
    scan_option fmt fmt_start)
*)
let scanner_option fmt scanner_a ib =
  let fmt_start = (string_of_format fmt).[0] in
  if Scanf.Scanning.end_of_input ib then None else
  Scanf.bscanf ib "%0c" (fun c ->
    if c <> fmt_start then None else
    (Scanf.bscanf ib fmt (); Some (scanner_a ib)))
;;

(* Scanning an option with Caml conventions:
   The option must be either
   - the word None, or
   - the word Some,
     followed by a successful scan of option value with scanner argument
     [scanner_a] (possibly enclosed in parens).
*)

let scanner_parens_opt scanner_a ib =
  Scanf.bscanf ib "%0c" (function
  | '(' (* ')' *) ->
    Scanf.bscanf ib "( %r )" scanner_a (fun x -> x)
  | _ -> Scanf.bscanf ib "%r" scanner_a (fun x -> x))
;;

let scanner_Option scanner_a ib =
  Scanf.bscanf ib "%0c" (function
  | 'N' ->
    Scanf.bscanf ib "None" None
  | _ ->
    Scanf.bscanf ib "Some %r"
      (scanner_parens_opt scanner_a)
      (fun x -> Some x))
;;

(* Scanning an array with user conventions:
  format string argument [fmt] may appear after each element of the array.
*)
let scanner_array fmt scanner_a ib =
  let rec loop accu ib =
    try
      let accu = scanner_a ib :: accu in
      try
        Scanf.bscanf ib fmt ();
        loop accu ib
      with _ -> accu
    with _ -> accu in

  Array.of_list (List.rev (loop [] ib))
;;

(* Scanning an array with Caml conventions *)
let scanner_Array scanner_a ib =

  let rec rl accu ib =
    Scanf.bscanf ib "%0c" (function
    | '|' -> accu
    | _ ->
      Scanf.bscanf ib "%r " scanner_a  (fun x ->
        Scanf.bscanf ib "%0c" (function
        | ';' ->
          Scanf.bscanf ib "; " rl (x :: accu) ib
        | _ ->
          rl (x :: accu) ib))) in

  Scanf.bscanf ib "[| %r |]" (rl []) (fun l -> Array.of_list (List.rev l))
;;

let scanner_ref fmt scanner_a ib =
  try
    Scanf.bscanf ib fmt ();
    ref (scanner_a ib)
  with _ ->
    ref (scanner_a ib)
;;

let scanner_Ref scanner_a ib =
  Scanf.bscanf ib "%0c"
    (function
     | 'r' ->
       Scanf.bscanf ib "ref %r" scanner_a (fun c -> ref c)
     | _ ->
       Scanf.bscanf ib "{ contents = %r }" scanner_a
         (fun c -> { contents = c }))
;;

let scanner_lazy fmt scanner_a ib =
  Scanf.bscanf ib fmt ();
  let v = scanner_a ib in
  lazy v
;;

let scanner_Lazy scanner_a = scanner_lazy "lazy " scanner_a;;

let scanner_bool ib =
  Scanf.bscanf ib "%0c" (function
  | 't' ->
    Scanf.bscanf ib "true" true
  | _ ->
    Scanf.bscanf ib "false" false)
;;
let scanner_Bool = scanner_bool;;

let scanner_fun
  (fmt : _ scanner_delimiter_format)
  (_scanner_a : 'a Og_types.scanner)
  (_scanner_b : 'b Og_types.scanner) ib =
  Scanf.bscanf ib fmt ();
  (fun _ -> failwith "Cannot read a function")
;;

let scanner_Fun ib = scanner_fun "<fun>" ib;;

let scanner_poly_value message fmt ib =
  Scanf.bscanf ib fmt ();
  failwith message
;;

let scanner_poly fmt =
   scanner_poly_value "Cannot read a polymorphic value" fmt
;;

let scanner_Poly ib = scanner_poly "<poly>" ib;;

let scanner_abstract fmt =
  scanner_poly_value "Cannot read an abstract value" fmt
;;

let scanner_Abstract ib = scanner_abstract "<abstract>" ib;;

let scanner_external fmt =
  scanner_poly_value "Cannot read an external value" fmt
;;

let scanner_External ib = scanner_external "<external>" ib;;
