(***********************************************************************)
(*                                                                     *)
(*                          Ocaml-xml Library                          *)
(*                                                                     *)
(*                   Pierre Weis, INRIA Rocquencourt                   *)
(*                                                                     *)
(*  Copyright 2013-2018,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: xml_lexer.mll,v 1.1 2019-05-10 14:00:49 jpc Exp $ *)

{

(* The lexer *)

(* Prelude part: this is pure Caml. *)

open Lexing;;
open Xml_parser;;
open Xml_lexer_types;;

(** When expanding entities we call the lexer with the replacement text.
  So we need to distinguish the two cases since when called with the
  replacement text, we must find an EOF at the end of the string. *)
type char_data_mode =
| Inside_document (** The normal mode when lexing a whole xml document. *)
| Standalone_char_data (** The mode when calling from the expander. *)
;;

let debug s =
  Format.eprintf "%s@." s
;;

(** {6 Lexical errors} *)

type error =
   | Illegal_character of string * char
   | Illegal_escape of string
   | Invalid_int32 of string
   | Invalid_recursive_cdata
   | Invalid_sequence of string * string
   | Invalid_xml_name of string
   | Invalid_pi_target of string
   | Unknown_entity of string
   | Unterminated of string
(** The various errors when lexing. *)
;;

exception Error of error * Lexing.position * Lexing.position;;

let error err start_pos end_pos =
  raise (Error (err, start_pos, end_pos))
;;

(** {6 Explaining lexing errors} *)

let report_error ppf = function
  | Illegal_character (s, c) ->
    Format.fprintf ppf "Illegal character (%C) during %s lexing." c s
  | Illegal_escape s ->
    Format.fprintf ppf "Illegal escape (%S)" s
  | Invalid_int32 s ->
    Format.fprintf ppf "Invalid int32 (%s)" s
  | Invalid_recursive_cdata ->
    Format.fprintf ppf "Invalid recursive CData."
  | Invalid_sequence (context, sequence) ->
    Format.fprintf ppf "Invalid sequence %S in %s." sequence context
  | Invalid_xml_name s ->
    Format.fprintf ppf "Invalid XML name %S." s
  | Invalid_pi_target s ->
    Format.fprintf ppf "Invalid PI target %S." s
  | Unknown_entity s ->
    Format.fprintf ppf "Unknown entity %S." s
  | Unterminated s ->
    Format.fprintf ppf "Unterminated %s" s
;;

(** {6 Keeping the internal buffer locations up to date} *)

let update_loc lexbuf line absolute chars =
  let pos = lexbuf.lex_curr_p in
  lexbuf.lex_curr_p <- {
    pos with
    pos_lnum = if absolute then line else pos.pos_lnum + line;
    pos_bol = pos.pos_cnum - chars;
  }
;;

(** Add one to the current line counter of the file being lexed. *)
let incr_line_num lexbuf =
  update_loc lexbuf 1 false 0
;;

(** {6 Buffer creation function} *)
let make_buffer buffer_name size =
  let buffer = Buffer.create size in
  let store_char c =
    try
      Buffer.add_char buffer c
    with
    | x ->
       prerr_endline (Printf.sprintf
         "trying to add char %C to buffer %S failed: %s"
         c buffer_name (Printexc.to_string x));
       prerr_endline (Printf.sprintf
         "current buffer length: %d" (Buffer.length buffer));
       prerr_endline (Printf.sprintf
         "current buffer content");
       prerr_endline (Buffer.contents buffer);
       raise x in
  let store_string s =
    try
      Buffer.add_string buffer s
    with
    | x ->
       prerr_endline (Printf.sprintf
         "trying to add string %S to buffer %S failed: %s"
         s buffer_name (Printexc.to_string x));
       prerr_endline (Printf.sprintf
         "current buffer length: %d" (Buffer.length buffer));
       prerr_endline (Printf.sprintf
         "current buffer content");
       prerr_endline (Buffer.contents buffer);
       raise x in
  let get_buffer_contents () =
    let s = Buffer.contents buffer in
    Buffer.reset buffer;
    s in
  store_char, store_string, get_buffer_contents
;;
let make_lexing_buffer () = make_buffer "lexing" 256;;

(** {6 Lexing the att_value tokens} *)
let store_att_value_char, store_att_value_string,
    get_att_value_string =
  make_lexing_buffer ()
;;

let att_value_start_pos = ref None;;

(** {6 Lexing the system_literal tokens (strings or literals)} *)
let store_system_literal_char, store_system_literal_string,
    get_system_literal_string =
  make_lexing_buffer ()
;;

let system_literal_start_pos = ref None;;

(** {6 Lexing the comment tokens} *)
let store_comment_char, store_comment_string,
    get_comment_string =
  make_lexing_buffer ()
;;

let comment_start_pos = ref None;;

(** {6 Lexing the cdata tokens} *)
let store_cdata_char, store_cdata_string,
    get_cdata_string =
  make_lexing_buffer ()
;;

let cdata_start_pos = ref None;;

(** {6 Lexing the CharData tokens} *)
let store_char_data_char, store_char_data_string,
    get_char_data_string =
  make_lexing_buffer ()
;;

let char_data_start_pos = ref None;;

(** {6 Lexing the pi tokens} *)
let store_pi_char, store_pi_string,
    get_pi_string =
  make_lexing_buffer ()
;;

let pi_start_pos = ref None;;

(** {6 Unescaping characters in strings} *)
let store_unescape_char, store_unescape_string,
    get_unescape_string =
  make_lexing_buffer ()
;;

(** {6 Lexing the Ignore token} *)
let store_ignore_char, store_ignore_string,
    get_ignore_string =
  make_lexing_buffer ()
;;

let ignore_start_pos = ref None;;

(* Token construction *)
let token_of_sdd_decl = function
  | "yes" -> SDDecl true
  | "no" -> SDDecl false
  | _ -> assert false
;;

let token_of_nm_token _lexbuf nm = Nmtoken nm;;

let token_of_comment s = Comment s;;
let token_of_ignore s = Ignore s;;

let token_of_cdata s = CData s;;

let make_xml_name lexbuf nm =
  try
    Xml_basics_ast.to_xml_name nm
  with
  | Failure _ ->
    error (Invalid_xml_name nm) lexbuf.lex_start_p lexbuf.lex_curr_p
;;

let tag_of_name = make_xml_name;;

(*
let token_of_name lexbuf nm = Name (make_xml_name lexbuf nm);;
*)

let token_of_pe_reference lexbuf nm = PEReference (make_xml_name lexbuf nm);;

let int32_of_string start_pos end_pos cr =
  try
    Int32.of_string cr
  with
  | Failure _ ->
    error (Invalid_int32 cr) start_pos end_pos
;;

let token_of_dec_char_ref lexbuf cr =
  let i = int32_of_string lexbuf.lex_start_p lexbuf.lex_curr_p cr in
  CharRef i
;;

let token_of_hex_char_ref lexbuf cr =
  let i = int32_of_string lexbuf.lex_start_p lexbuf.lex_curr_p ("0x" ^ cr) in
  CharRef i
;;

let token_of_pe_decl_name lexbuf nm = PEDeclName (make_xml_name lexbuf nm);;
let token_of_ge_decl_name lexbuf nm = GEDeclName (make_xml_name lexbuf nm);;

let token_of_doctype_name lexbuf nm = DocTypeName (make_xml_name lexbuf nm);;

let token_of_element_name lexbuf nm = ElementName (make_xml_name lexbuf nm);;

let token_of_notation_name lexbuf nm = NotationName (make_xml_name lexbuf nm);;

let token_of_etag_name lexbuf nm = ETag (make_xml_name lexbuf nm);;

type first_tag =
   | Empty_first_tag of token_tag
   | Non_empty_first_tag of token_tag
;;

let token_of_first_start_tag = function
  | Empty_first_tag (tag_name, attrs) -> FirstEmptyElemTag (tag_name, attrs)
  | Non_empty_first_tag (tag_name, attrs) -> FirstSTag (tag_name, attrs)
;;

let token_of_start_tag = function
  | Empty_first_tag (tag_name, attrs) -> EmptyElemTag (tag_name, attrs)
  | Non_empty_first_tag (tag_name, attrs) -> STag (tag_name, attrs)
;;

let make_pi_target lexbuf nm =
  try
    let xml_nm = Xml_basics_ast.to_xml_name nm in
    Xml_basics_ast.to_pi_target xml_nm
  with
  | Failure _ ->
    error (Invalid_pi_target nm) lexbuf.lex_start_p lexbuf.lex_curr_p
;;

let token_of_pi lexbuf tgt instrs = PI (make_pi_target lexbuf tgt, instrs);;

let token_of_attlist_name lexbuf nm = AttlistName (make_xml_name lexbuf nm);;

let token_of_notation_type_name lexbuf nm =
  NotationTypeName (make_xml_name lexbuf nm)
;;

let token_of_ndatadecl lexbuf nm =
  let xml_name = make_xml_name lexbuf nm in
  NDataDecl xml_name
;;

let token_of_entityref lexbuf nm =
  let xml_name = make_xml_name lexbuf nm in
  EntityRef xml_name
;;

let unquote_token s =
  let len = String.length s in
  assert (
    len >= 2 &&
    let c0 = s.[0] in
    c0 = s.[len - 1] &&
    (c0 = '\'' || c0 = '\"'));
  String.sub s 1 (len - 2)
;;

(* Decoding entity references '&' name ';' *)
(* <!ENTITY lt     "&#38;#60;"> *)
(* <!ENTITY gt     "&#62;"> *)
(* <!ENTITY amp    "&#38;#38;"> *)
(* <!ENTITY apos   "&#39;"> *)
(* <!ENTITY quot   "&#34;"> *)
(* let char_of_entity_ref s start_pos end_pos = *)
(*   match s with *)
(*   | "&apos;" -> '\'' *)
(*   | "&quot;" -> '\"' *)
(*   | "&amp;" -> '&' *)
(*   | "&lt;" -> '<' *)
(*   | "&gt;" -> '>' *)
(*   | name -> *)
(*     error (Unknown_entity name) start_pos end_pos *)
(*;; *)

(* Decoding character references '&#[x]d+;' *)
(* let char_of_char_ref s = *)
(*   let len = String.length s in *)
(*   assert (len >= 4); *)
(*   let num = "0" ^ String.sub s 2 (len - 3) in *)
(*   (\* We scan a numeric specification in string s *)
(*      - either 0xd+ (with d any hexadecimal digits) *)
(*      - or 0d+ (with d any decimal digits) *)
(*      according to char reference s (&#d+ or &#xd+). *\) *)
(*     (\* FIXME: we should catch char_of_int errors *)
(*        in case of invalid character number *\) *)
(*   Scanf.sscanf num "%i" (fun i -> char_of_int i) *)
(*;; *)

(* DONE WITH A STANDARD LEXER *)
(*
let check_char_data char_data =
  let ib = Scanf.Scanning.from_string (unescape_references char_data) in
  let buffer = Buffer.create (String.length char_data) in
  let store_char c = Buffer.add_char buffer c in
  let rec loop ib =
    if Scanf.Scanning.end_of_input ib then
      Buffer.contents buffer
    else
      Scanf.bscanf ib "%c" (function
      | ']' ->
        store_char ']';
        after_bracket ib
      | c -> store_char c)
  and after_bracket ib =
    if Scanf.Scanning.end_of_input ib then
      Buffer.contents buffer
    else
      Scanf.bscanf ib "%c" (function
      | ']' ->
        store_char ']';
        after_brackets ib
      | c ->
        store_char c;
        loop ib)
  and after_brackets ib =
    if Scanf.Scanning.end_of_input ib then
      Buffer.contents buffer
    else
      Scanf.bscanf ib "%c" (function
      | '>' -> error ("Invalid CharData \"]]>\" found") start_pos end_pos
      | ']' ->
        store_char ']';
        after_brackets ib
      | c ->
        store_char c;
        loop ib) in
  loop ib
;;
*)

let update_lexbuf_newlines lexeme lexbuf =
  let limit = String.length lexeme - 1 in
  for i = 0 to limit do
    match lexeme.[i] with
    | '\n' -> incr_line_num lexbuf
    | _ -> ()
  done
;;

let get_lexeme_with_newlines lexbuf =
  let lexeme = Lexing.lexeme lexbuf in
  update_lexbuf_newlines lexeme lexbuf;
  lexeme
;;

let update_newlines lexbuf =
  ignore (get_lexeme_with_newlines lexbuf)
;;

let unterminated_token tok start_pos_ref =
  match !start_pos_ref with
  | Some (start_pos, end_pos) ->
    error (Unterminated tok) start_pos end_pos
  | _ -> assert false
;;

let unterminated_comment () =
  unterminated_token "Comment" comment_start_pos
;;

let unterminated_ignore () =
  unterminated_token "Ignore" ignore_start_pos
;;

let unterminated_pi () =
  unterminated_token "PI" pi_start_pos
;;

let unterminated_cdata () =
  unterminated_token "CData" cdata_start_pos
;;

let unterminated_char_data () =
  unterminated_token "CharData" char_data_start_pos
;;

let unterminated_system_literal () =
  unterminated_token "SystemLiteral" system_literal_start_pos
;;

let unterminated_att_value () =
  unterminated_token "AttValue" att_value_start_pos
;;
}

(* Regular expression definitions *)

(* [3] Spaces *)
let newline_char = '\n'
let tab_char = '\t'
let space_char = ' '
let return_char = '\r'

let white_space_char_but_newline = space_char | tab_char | return_char
let white_space_char = white_space_char_but_newline | newline_char

let space = white_space_char +
let space_opt = space ?

(* The complete set of characters used in text. *)
let char =
    white_space_char
  | [ '\033' - '\255' ] (* all characters after space *)
    (* In fact we should use:
       Any Unicode character, excluding the surrogate
       blocks, FFFE, and FFFF. *)

(* All characters except '\n' '-' '<' '>' '?' '[' ']' *)
let char_but_specials =
    white_space_char_but_newline
  | [ '\033' - '\044' (* omit '\045' or '-' *)
      '\046' - '\060' (* omit '\061' or '<' and '\062' or '>' *)
      '\064' - '\090' (* omit '\063' or '?' *)
      '\092'          (* omit '\091' or '[' *)
      '\094' - '\255' (* omit '\093' or ']' *)
    ]

(* All characters except '-' '>' (for comments) *)
let char_but_gt_dash =
  char_but_specials | [ '<' '?' '[' ']' ]

(* All characters except '-' *)
let char_but_dash =
  char_but_gt_dash | [ '>' ]

(* All characters except '>' '?' (for PI) *)
let char_but_gt_question =
  char_but_specials | [ '-' '<' '[' ']' ]

(* All characters except '>' ']' (for char_data & cdata) *)
let char_but_gt_right_bracket =
  char_but_specials | [ '-' '<' '?' '[' ]

(* All characters except '<' ']' (for ignore) *)
(* [65] Ignore ::= Char* - (Char* ('<![' | ']]>') Char* ) *)
let char_but_lt_right_bracket =
  char_but_specials | [ '-' '>' '?' '[' ]

(* [4] Names and Tokens *)
let decimal_digit = [ '0' - '9' ]
let hexadecimal_digit = decimal_digit | [ 'a' - 'f' 'A' - 'F' ]

let lowercase_alphabetic = [ 'a' - 'z' ]
let uppercase_alphabetic = [ 'A' - 'Z' ]
let alphabetic = lowercase_alphabetic | uppercase_alphabetic
let alphanumeric = alphabetic | decimal_digit

(* [4] NameStartChar ::=
         ":" | [A-Z] | "_" | [a-z]
       | [#xC0-#xD6] | [#xD8-#xF6] | [#xF8-#x2FF] | [#x370-#x37D]
       | [#x37F-#x1FFF] | [#x200C-#x200D] | [#x2070-#x218F] | [#x2C00-#x2FEF]
       | [#x3001-#xD7FF] | [#xF900-#xFDCF] | [#xFDF0-#xFFFD]
       | [#x10000-#xEFFFF] *)
let name_start_char =
    [ ':' '_' ]
  | alphabetic
  | [ '\192' - '\214' ]
  | [ '\216' - '\246' ]
  | [ '\248' - '\255' ] (* Any character after '\248' *)
    (* In fact we should use:
       | [#xC0-#xD6] | [#xD8-#xF6] | [#xF8-#x2FF] | [#x370-#x37D]
       | [#x37F-#x1FFF] | [#x200C-#x200D] | [#x2070-#x218F]
       | [#x2C00-#x2FEF] | [#x3001-#xD7FF]
       | [#xF900-#xFDCF] | [#xFDF0-#xFFFD] | [#x10000-#xEFFFF]
    *)

(* [4a] NameChar ::=
      NameStartChar | "-" | "." | [0-9] | #xB7
   | [#x0300-#x036F] | [#x203F-#x2040]*)
let name_char =
    name_start_char | [ '-' '.' ] | decimal_digit | '\183'
    (* In fact we should also use:
       | [#x0300-#x036F] | [#x203F-#x2040]
    *)

(* [5] Name *)
let name = name_start_char name_char*
(* Side condition:
   no name could start by xml in any case variant
   (no name may match the regexp
   (('X'|'x') ('M'|'m') ('L'|'l')) name_in_char*
   This is check in function
   make_xml_name
*)

(* [7] Nmtoken ::= (NameChar)+ *)
let nm_token = name_char+

(* Entity Reference *)
(* [66] CharRef ::=
     '&#' [0-9]+ ';'
   | '&#x' [0-9a-fA-F]+ ';'*)
let char_ref =
    "&#" decimal_digit+ ';'
  | "&#x" hexadecimal_digit+ ';'

(* [68] EntityRef ::= '&' Name ';'
   [WFC: Entity Declared] [VC: Entity Declared]
   [WFC: Parsed Entity] [WFC: No Recursion] *)
let entity_ref = '&' name ';'

(* [69] PEReference ::= '%' Name ';'
   [VC: Entity Declared] [WFC: No Recursion] [WFC: In DTD] *)
let pe_reference = '%' name ';'

(* [67] Reference ::= EntityRef | CharRef *)
let reference = entity_ref | char_ref

(* The eq symbol surrounded by optional spaces. *)
let eq = space_opt '=' space_opt

let version_num = "1." decimal_digit+

(* [81] EncName ::= [A-Za-z] ([A-Za-z0-9._] | '-')* *)
let enc_name = alphabetic (alphanumeric | [ '.' '_' '-'])*

let yes_no = "yes" | "no"

let system_literal =
    ('\"' [^ '\"' ]* '\"')
  | ('\'' [^ '\'' ]* '\'')

let att_value =
    '\"' ([^ '<' '&' '\"' ] | reference)* '\"'
  |  '\'' ([^ '<' '&' '\''] | reference)* '\''

let char_data_but_gt_right_bracket =
  [^ '<' '&' '>' ']' ]

let char_data_but_right_bracket =
    char_data_but_gt_right_bracket
  | '>'

let char_data_chunk =
    char_data_but_right_bracket
  | ']' char_data_but_right_bracket
  | ']' (']')+ char_data_but_gt_right_bracket

let char_data_end = (']')*

let char_data = char_data_chunk* char_data_end

(* #x20 | #xD | #xA | [a-zA-Z0-9] | [-'()+,./:=?;!*#@$_%] *)
let pubid_char_but_quote =
    alphanumeric
  | [' ' '\r' '\n'
     '-' '(' ')' '+' ',' '.'
     '/' ':' '=' '?' ';' '!'
     '*' '#' '@' '$' '_' '%']
let pubid_char = '\'' | pubid_char_but_quote

let pubid_literal =
    '\"' pubid_char* '\"'
  | '\'' pubid_char_but_quote* '\''

(* The main entry point lexer (to check if there is a prolog) *)

(* A prolog starts with either
   | "<?" (XMLDecl or Misc/PI)
   | "<!--" (Misc/Comment)
   | space (Misc/S)
   | "<!DOCTYPE" (doctypedecl)
   Hence, a prolog starts either with "<?" or "<!" or space. *)
rule start = parse
  | "<?" | "<!" | space
    { Xml_lexer_types.Start_prolog }
  | '<'
    { Start_element }

(* The entry point for lexing a prolog. *)
and prolog = parse
    (* [22] prolog ::= XMLDecl? Misc* (doctypedecl Misc* )? *)
    (* [23] XMLDecl ::=
       '<?xml' VersionInfo EncodingDecl? SDDecl? S? '?>' *)
  | "<?xml"
    { XmlDeclStart }
    (* [25] Eq ::= S? '=' S? *)
    (* [24] VersionInfo ::=
         S 'version' Eq
         ("'" VersionNum "'" | '"' VersionNum '"') *)
  | space "version" eq (
        '\'' (version_num as vn) '\''
      | '\"' (version_num as vn) '\"' )
    { update_newlines lexbuf;
      VersionInfo vn}

    (* [80] EncodingDecl ::=
         S 'encoding' Eq ('"' EncName '"'
       | "'" EncName "'" ) *)
  | space "encoding" eq (
        '\'' (enc_name as en) '\''
      | '\"' (enc_name as en) '\"')
    { update_newlines lexbuf;
      EncodingDecl en}

    (* [32] SDDecl ::=
       S 'standalone' Eq
       (  ("'" ('yes' | 'no') "'")
        | ('"' ('yes' | 'no') '"') ) *)
  | space "standalone" eq (
        '\'' (yes_no as yn) '\''
      | '\"' (yes_no as yn) '\"')
    { update_newlines lexbuf;
      token_of_sdd_decl yn}

  | space_opt "?>"
    { update_newlines lexbuf;
      XmlDeclEnd }
  | space_opt '>'
    { update_newlines lexbuf;
      GT }

    (* [27] Misc ::= Comment | PI | S *)
    (* Comment ::= '<!--' ... '-->' *)
  | "<!--"
    {
      comment_start_pos :=
        Some (lexbuf.lex_start_p, lexbuf.lex_curr_p);
      in_comment lexbuf;
      begin match !comment_start_pos with
      | Some (start_pos, _) -> lexbuf.lex_start_p <- start_pos
      | _ -> assert false end;
      let s = get_comment_string () in
      token_of_comment s
    }

    (* PI ::= '<?' Name ... *)
  | "<?" (name as nm)
    {
      pi_start_pos :=
        Some (lexbuf.lex_start_p, lexbuf.lex_curr_p);
      in_pi lexbuf;
      begin match !pi_start_pos with
      | Some (start_pos, _) -> lexbuf.lex_start_p <- start_pos
      | _ -> assert false end;
      token_of_pi lexbuf nm (get_pi_string ())
    }

    (* FIXED: Should remove this rule ?
       This is difficult, since we should preserve spacing and allow
       space_opt in element_decl, hence after content_spec, children, and
       mixed; hence after a CloseParen in case of absent regexp_mod_opt.
       Anyway, allowing an optional space after a closing paren would not
       break existing XML code, even if it is a bit laxist. *)
    (* S ::= ... *)
  | space
    { update_newlines lexbuf;
      Space (get_lexeme_with_newlines lexbuf) }

    (* [25] Eq ::= S? '=' S? *)
  | space_opt '=' space_opt
    { update_newlines lexbuf;
      Eq }

    (* Document Type Definition
       [28] doctypedecl ::=
         '<!DOCTYPE' S Name (S ExternalID)? S?
         ('[' intSubset ']' S?)?
         '>'
       [VC: Root Element Type] *)
  | "<!DOCTYPE" space (name as nm)
    { update_newlines lexbuf;
      token_of_doctype_name lexbuf nm }

    (* [75] ExternalID ::=
           'SYSTEM' S SystemLiteral
         | 'PUBLIC' S PubidLiteral S SystemLiteral *)
  | space "SYSTEM" space (system_literal as sl) space_opt
    { update_newlines lexbuf;
      ExternalIDSystem sl}
  | space "PUBLIC" space (pubid_literal as pl) space
     (system_literal as sl) space_opt
    { update_newlines lexbuf;
      ExternalIDPublic (pl, sl)}

  | space_opt '['
    { update_newlines lexbuf;
      OpenBracket }

    (* [28a] DeclSep ::= PEReference | S *)
    (* [69] PEReference ::= '%' Name ';' *)
  | '%' (name as nm) ';'
    { token_of_pe_reference lexbuf nm }

    (* 28b] intSubset ::= (markupdecl | DeclSep)* *)
    (* [29] markupdecl ::=
          elementdecl | AttlistDecl | EntityDecl
        | NotationDecl | PI | Comment *)

    (* [45] elementdecl ::= '<!ELEMENT' S Name S contentspec S? '>' *)
  | "<!ELEMENT" space (name as nm) space
    { update_newlines lexbuf;
      token_of_element_name lexbuf nm }

    (* [46] contentspec ::= 'EMPTY' | 'ANY' | Mixed | children *)

   (* [51] Mixed ::=
        '(' S? '#PCDATA' (S? '|' S? Name)* S? ')*'
      | '(' S? '#PCDATA' S? ')' *)
  | '(' space_opt
    { update_newlines lexbuf;
      OpenParen }
  | "#PCDATA"
    { PCDATA }
  | space_opt '|' space_opt
    { update_newlines lexbuf;
      Bar }

    (* [53] AttDef ::= S Name S AttType S DefaultDecl *)

    (* [54] AttType ::= StringType | TokenizedType | EnumeratedType
       [55] StringType ::= 'CDATA'
       [56] TokenizedType ::=
         'ID'
         [VC: ID] [VC: One ID per Element Type] [VC: ID Attribute Default]
       | 'IDREF' [VC: IDREF]
       | 'IDREFS' [VC: IDREF]
       | 'ENTITY' [VC: Entity Name]
       | 'ENTITIES' [VC: Entity Name]
       | 'NMTOKEN' [VC: Name Token]
       | 'NMTOKENS' *)

  | (nm_token as nm)
    { token_of_nm_token lexbuf nm }
    (* The nm_token rule is mandatory for Enumeration.
       Hence, the rule for names must use the preceding one
       with an additional check in the parser, instead of the natural
       one for name:
       | (name as nm)
         { update_newlines lexbuf;
           Name nm } *)

  | space_opt ")*" space_opt
    { update_newlines lexbuf;
      CloseParenStar }

    (* [47] children ::= (choice | seq) ('?' | '*' | '+')?
       [48] cp       ::= (Name | choice | seq) ('?' | '*' | '+')?
       [49] choice   ::= '(' S? cp ( S? '|' S? cp )+ S? ')'
            [VC: Proper Group/PE Nesting]
       [50] seq      ::= '(' S? cp ( S? ',' S? cp )* S? ')'
            [VC: Proper Group/PE Nesting] *)
  | '?' { Question }
  | '*' { Star }
  | '+' { Plus }

  | space_opt ")?" space_opt
    { update_newlines lexbuf;
      CloseParenQuestion }

  | space_opt ")+" space_opt
    { update_newlines lexbuf;
      CloseParenPlus }

  | space_opt ',' space_opt
    { update_newlines lexbuf;
      Comma }

  | space_opt ')'
    { update_newlines lexbuf;
      CloseParen }

    (* [52] AttlistDecl ::=
        '<!ATTLIST' S Name AttDef* S? '>' *)
  | "<!ATTLIST" space (name as nm)
    { update_newlines lexbuf;
      token_of_attlist_name lexbuf nm }

    (* [57] EnumeratedType ::= NotationType | Enumeration *)
    (* [58] NotationType ::= 'NOTATION' S '(' S? Name (S? '|' S? Name)* S? ')'
         [VC: Notation Attributes]
         [VC: One Notation Per Element Type]
         [VC: No Notation on Empty Element]
         [VC: No Duplicate Tokens] *)
  | "NOTATION" space '(' space_opt (name as nm)
    { update_newlines lexbuf;
      token_of_notation_type_name lexbuf nm }
    (* [59] Enumeration ::= '(' S? Nmtoken (S? '|' S? Nmtoken)* S? ')'
         [VC: Enumeration] [VC: No Duplicate Tokens] *)
    (* [60] DefaultDecl ::=
         '#REQUIRED' | '#IMPLIED'
       | (('#FIXED' S)? AttValue)
         [VC: Required Attribute]
         [VC: Attribute Default Value Syntactically Correct]
         [WFC: No < in Attribute Values]
         [VC: Fixed Attribute Default]
         [WFC: No External Entity References] *)
     (* [10] AttValue ::=
                '"' ([^<&"] | Reference)* '"'
             |  "'" ([^<&'] | Reference)* "'" *)
  | "#REQUIRED" { REQUIRED }
  | "#IMPLIED" { IMPLIED }
    (* Next rule could be:
       | "#FIXED" space (att_value as av)
         { update_newlines lexbuf;
           FIXED av}
       but this would cause discrepancy in the treatment of AttValue,
       since this one would have been checked, whereas the default non #FIXED
       AttValue is parsed as a general system literal which is not. *)
  | "#FIXED" space
    { update_newlines lexbuf;
      FIXED}

    (* [70] EntityDecl ::= GEDecl | PEDecl
       [71] GEDecl ::= '<!ENTITY' S Name S EntityDef S? '>' *)
  | "<!ENTITY" space (name as nm) space
    { update_newlines lexbuf;
      token_of_ge_decl_name lexbuf nm }

    (* [73] EntityDef ::= EntityValue | (ExternalID NDataDecl?)
       [9] EntityValue ::=
              '"' ([^%&"] | PEReference | Reference)* '"'
           |  "'" ([^%&'] | PEReference | Reference)* "'"
       FIXME: EntityValue is treated as a system literal by the lexer.
       A verification in the parser must ensure that references (entities,
       characters and parameter-entities) are indeed well-formed. *)
  | "SYSTEM" space (system_literal as sl)
    { update_newlines lexbuf;
      ExternalIDSystem sl}
  | "PUBLIC" space (pubid_literal as pl) (system_literal as sl)
    { update_newlines lexbuf;
      ExternalIDPublic (pl, sl)}

    (* [76] NDataDecl ::= S 'NDATA' S Name *)
  | space "NDATA" space (name as nm)
    { update_newlines lexbuf;
      token_of_ndatadecl lexbuf nm }

    (* [72] PEDecl ::= '<!ENTITY' S '%' S Name S PEDef S? '>' *)
  | "<!ENTITY" space '%' space (name as nm) space
    { update_newlines lexbuf;
      token_of_pe_decl_name lexbuf nm }
    (* [74] PEDef ::= EntityValue | ExternalID *)

    (* [11] SystemLiteral ::=
            ('"' [^"]* '"') | ("'" [^']* "'") *)
  | ('\"' | '\'') as c
    {
      att_value_start_pos :=
        Some (lexbuf.lex_start_p, lexbuf.lex_curr_p);
      let attribute_value = in_att_value c [] lexbuf in
      begin match !att_value_start_pos with
      | Some (start_pos, _) -> lexbuf.lex_start_p <- start_pos
      | _ -> assert false end;
      AttValue attribute_value
    }

    (* [82] NotationDecl ::=
       '<!NOTATION' S Name S
       (ExternalID | PublicID) S? '>'
       [VC: Unique Notation Name] *)
  | "<!NOTATION" space (name as nm) space
    { update_newlines lexbuf;
      token_of_notation_name lexbuf nm }

    (* [83] PublicID ::= 'PUBLIC' S PubidLiteral *)
  | "PUBLIC" space (pubid_literal as pl)
    { update_newlines lexbuf;
      PublicID pl}

  | ']' space_opt
    { update_newlines lexbuf;
      CloseBracket }

  | "<![" space_opt "INCLUDE" space_opt '['
    { update_newlines lexbuf;
      INCLUDE }
    (* [63] ignoreSect ::= '<![' S? 'IGNORE' S? '['
       ignoreSectContents*
       ']]>'
       [VC: Proper Conditional Section/PE Nesting] *)
  | "<![" space_opt "IGNORE" space_opt '['
    { update_newlines lexbuf;
      ignore_start_pos :=
        Some (lexbuf.lex_start_p, lexbuf.lex_curr_p);
      in_ignore lexbuf;
      begin match !ignore_start_pos with
      | Some (start_pos, _) -> lexbuf.lex_start_p <- start_pos
      | _ -> assert false end;
      let s = get_ignore_string () in
      token_of_ignore s }
  | "<!["
    { OpenConditionalSect }

  | "]]>"
    { update_newlines lexbuf;
      CloseConditionalSect }

  | '<' { Eof }

(* The entry point for lexing an element. *)
and element = parse
  | '<'
    { in_markup lexbuf }
  | '>'
    { in_char_data_or_references Inside_document (None, []) lexbuf }
  | '>' space_opt eof
    { update_newlines lexbuf;
      Eof }

  | '%' (name as nm) ';'
    { token_of_pe_reference lexbuf nm }
  | "&#" (decimal_digit+ as dec_int32) ';'
    { token_of_dec_char_ref lexbuf dec_int32 }
  | "&#x" (hexadecimal_digit+ as hex_int32) ';'
    { token_of_hex_char_ref lexbuf hex_int32 }
  | entity_ref
    { token_of_entityref lexbuf (Lexing.lexeme lexbuf) }

  | space
    { update_newlines lexbuf;
      Space (get_lexeme_with_newlines lexbuf) }

  | space_opt '?' space_opt
    { update_newlines lexbuf;
      Question }
  | space_opt '*' space_opt
    { update_newlines lexbuf;
      Star }
  | space_opt '+' space_opt
    { update_newlines lexbuf;
      Plus }

  | space_opt ")?" space_opt
    { update_newlines lexbuf;
      CloseParenQuestion }
  | space_opt ")+" space_opt
    { update_newlines lexbuf;
      CloseParenPlus }

  | space_opt ',' space_opt
    { update_newlines lexbuf;
      Comma }
  | space_opt '=' space_opt
    { update_newlines lexbuf;
      Eq }

    (* HACK: the '<' is consummed by the prolog lexer/parser;
       so that rule must only be used once, just after the prolog, to lex the
       first start tag of the element.
       In particular, if the document has no prolog, that rule must not be
       used. *)
  | (name as nm)
    { let tag = tag_of_name lexbuf nm in
      let stag = tag_attributes tag [] lexbuf in
      token_of_first_start_tag stag
    }

  | space_opt eof { Eof }

and in_markup = parse
  | "![CDATA["
    { cdata_start_pos :=
        Some (lexbuf.lex_start_p, lexbuf.lex_curr_p);
      in_cdata lexbuf;
      begin match !cdata_start_pos with
      | Some (start_pos, _) -> lexbuf.lex_start_p <- start_pos
      | _ -> assert false end;
      token_of_cdata (get_cdata_string ()) }
  | "!--"
    {
      comment_start_pos :=
        Some (lexbuf.lex_start_p, lexbuf.lex_curr_p);
      in_comment lexbuf;
      begin match !comment_start_pos with
      | Some (start_pos, _) -> lexbuf.lex_start_p <- start_pos
      | _ -> assert false end;
      let s = get_comment_string () in
      token_of_comment s
    }

  | '?' (name as nm)
    {
      pi_start_pos :=
        Some (lexbuf.lex_start_p, lexbuf.lex_curr_p);
      in_pi lexbuf;
      begin match !pi_start_pos with
      | Some (start_pos, _) -> lexbuf.lex_start_p <- start_pos
      | _ -> assert false end;
      token_of_pi lexbuf nm (get_pi_string ())
    }
  | '/' (name as nm) space_opt
    { update_newlines lexbuf;
      token_of_etag_name lexbuf nm }

  | (name as nm)
    (* [40] STag ::= '<' Name (S Attribute)* S? '>' *)
    (* Go on with tag_attributes i.e. (S attribute)* S?
       then
       '/' gives EmptyElemTag (name, attributes)
       '>' gives Stag (name, attributes)
       _ gives illegal character in tag element *)
    { let tag = tag_of_name lexbuf nm in
      token_of_start_tag (tag_attributes tag [] lexbuf) }
  | eof
    { failwith "Unterminated markup"
      (*error (Unterminated "markup") start_pos end_pos*) }

and tag_attribute = parse
  | (name as nm)
    {
      let attribute_name = tag_of_name lexbuf nm in
      tag_attribute_after_name attribute_name lexbuf
    }
  | eof
    {
      failwith "Unterminated tag attribute"
      (*error (Unterminated "attribute") start_pos end_pos*)
    }

and tag_attribute_after_name attribute_name = parse
  | eq ('\'' | '\"' as c)
    {
      update_newlines lexbuf;
      att_value_start_pos :=
        Some (lexbuf.lex_start_p, lexbuf.lex_curr_p);
      let attribute_value = in_att_value c [] lexbuf in
      attribute_name, attribute_value
    }
  | eof
    {
      failwith "Unterminated tag attribute"
      (*error (Unterminated "attribute") start_pos end_pos*)
    }

and in_att_value c accu = parse
  | eof { unterminated_att_value () }
  | newline_char
    { update_newlines lexbuf;
      store_att_value_char '\n';
      in_att_value c accu lexbuf }

  | "&#" (decimal_digit+ as dec_int32) ';'
    { let text_elem = get_att_value_string () in
      let accu =
        if String.length text_elem = 0 then accu else
          Xml_basics_ast.Att_value_text text_elem :: accu in
      let start_pos, end_pos = lexbuf.lex_start_p, lexbuf.lex_curr_p in
      let reference =
        Xml_basics_ast.Char_ref
          (int32_of_string start_pos end_pos dec_int32) in
      let char_ref_elem = Xml_basics_ast.Att_value_reference reference in
      in_att_value c (char_ref_elem :: accu) lexbuf }
  | "&#x" (hexadecimal_digit+ as hex_int32) ';'
    { let text_elem = get_att_value_string () in
      let accu =
        if String.length text_elem = 0 then accu else
          Xml_basics_ast.Att_value_text text_elem :: accu in
      let start_pos, end_pos = lexbuf.lex_start_p, lexbuf.lex_curr_p in
      let reference =
        Xml_basics_ast.Char_ref
          (int32_of_string start_pos end_pos ("0x" ^ hex_int32)) in
      let char_ref_elem = Xml_basics_ast.Att_value_reference reference in
      in_att_value c (char_ref_elem :: accu) lexbuf }

  | "&" (name as nm) ';'
    { let text_elem = get_att_value_string () in
      let accu =
        if String.length text_elem = 0 then accu else
          Xml_basics_ast.Att_value_text text_elem :: accu in
      let reference = Xml_basics_ast.Entity_ref (make_xml_name lexbuf nm) in
      let entity_ref_elem = Xml_basics_ast.Att_value_reference reference in
      in_att_value c (entity_ref_elem :: accu) lexbuf }

  | _
    { let char = Lexing.lexeme_char lexbuf 0 in
      if char <> c then begin
        store_att_value_char char;
        in_att_value c accu lexbuf
      end else
        let text_elem = get_att_value_string () in
        let accu =
          if String.length text_elem = 0 then accu else
            Xml_basics_ast.Att_value_text text_elem :: accu in
        List.rev accu }

and tag_attributes attribute_name accu = parse
  | ' '
    { let attr = tag_attribute lexbuf in
      tag_attributes attribute_name (attr :: accu) lexbuf
    }
  | space_opt '/'
    {
      update_newlines lexbuf;
      Empty_first_tag (attribute_name, List.rev accu)
    }
  | space_opt
    {
      update_newlines lexbuf;
      Non_empty_first_tag (attribute_name, List.rev accu)
    }
  | eof
    { failwith "Unterminated tag"
      (*error (Unterminated "tag") start_pos end_pos*) }

(* [3] S ::= (#x20 | #x9 | #xD | #xA)+ *)
and s_plus = parse
  | [ ' ' '\t' '\r' ]
    { s_star lexbuf }
  | '\n'
    { incr_line_num lexbuf; s_star lexbuf }
  | _ { failwith "Waiting for a white character." }

and s_star = parse
  | [ ' ' '\t' '\r' ]+
    { s_star lexbuf }
  | '\n'
    { incr_line_num lexbuf; s_star lexbuf }
  | _ { () }

(* Comment ::= "<!--" ((Char - '-') | ('-' (Char - '-')))* "-->" *)
and in_comment = parse
  | "--" { in_comment_after_dashes lexbuf }
  | '-'
    { store_comment_char '-';
      in_comment lexbuf }
  | (char_but_dash * as text)
    { store_comment_string text;
      in_comment lexbuf; }
  | newline_char
    { store_comment_char (Lexing.lexeme_char lexbuf 0);
      incr_line_num lexbuf;
      in_comment lexbuf }
  | eof
    { match !comment_start_pos with
      | Some (start_pos, end_pos) ->
        error (Unterminated "comment") start_pos end_pos
      | _ -> assert false }
  | _
    { (* char can not be '\n' (because of newline_char rule) *)
      let char = Lexing.lexeme_char lexbuf 0 in
      match !comment_start_pos with
      | Some (start_pos, end_pos) ->
        error (Illegal_character ("comment", char)) start_pos end_pos
      | _ -> assert false
    }
and in_comment_after_dashes = parse
  | (char_but_gt_dash * as text)
    { if String.length text <> 0 then
        match !comment_start_pos with
        | Some (start_pos, end_pos) ->
          error (Invalid_sequence ("comment", "--")) start_pos end_pos
        | _ -> assert false }
  | eof
    { match !comment_start_pos with
      | Some (start_pos, end_pos) ->
        error (Unterminated "comment") start_pos end_pos
      | _ -> assert false }
  | [^ '>' ]
    { match !comment_start_pos with
      | Some (start_pos, end_pos) ->
        error (Invalid_sequence ("comment", "--")) start_pos end_pos
      | _ -> assert false }

(* [65] Ignore ::= Char* - (Char* ('<![' | ']]>') Char* ) *)
and in_ignore = parse
  | "<!" { in_ignore_after_lt_bang lexbuf }
  | "]]" { in_ignore_after_right_brackets lexbuf }
  | ('<' | ']' as c)
    { store_ignore_char c;
      in_ignore lexbuf }
  | (char_but_lt_right_bracket * as text)
    { store_ignore_string text;
      in_ignore lexbuf; }
  | newline_char
    { store_ignore_char (Lexing.lexeme_char lexbuf 0);
      incr_line_num lexbuf;
      in_ignore lexbuf }
  | eof
    { match !ignore_start_pos with
      | Some (start_pos, end_pos) ->
        error (Unterminated "ignore") start_pos end_pos
      | _ -> assert false }
and in_ignore_after_right_brackets = parse
  | '>'
    { () }
  | newline_char
    { store_ignore_char (Lexing.lexeme_char lexbuf 0);
      incr_line_num lexbuf;
      in_ignore lexbuf }
  | ([^ '>' ] as c)
    {
      store_ignore_string "]]";
      store_ignore_char c;
      in_ignore lexbuf;
     }
  | eof
    { match !ignore_start_pos with
      | Some (start_pos, end_pos) ->
        error (Unterminated "Ignore") start_pos end_pos
      | _ -> assert false }
and in_ignore_after_lt_bang = parse
  | '['
    { store_ignore_string "<![";
      in_ignore lexbuf;
      store_ignore_string "]]>";
      in_ignore lexbuf; }
  | newline_char
    { store_ignore_char (Lexing.lexeme_char lexbuf 0);
      incr_line_num lexbuf;
      in_ignore lexbuf }
  | ([^ '[' ] as c)
    {
      store_ignore_string "<!";
      store_ignore_char c;
      in_ignore lexbuf; }
  | eof
    { match !ignore_start_pos with
      | Some (start_pos, end_pos) ->
        error (Unterminated "Ignore") start_pos end_pos
      | _ -> assert false }

and in_pi = parse
  | space
    { update_newlines lexbuf;
      in_pi_after_space lexbuf }
  | '?' { () }
  | eof { unterminated_pi () }
and in_pi_after_space = parse
  | '?' { in_pi_after_question lexbuf }
  | (char_but_gt_question * as text)
    { store_pi_string text;
      in_pi_after_space lexbuf }
  | eof { unterminated_pi () }
and in_pi_after_question = parse
  | '?'
    { store_pi_char '?';
      in_pi_after_question lexbuf }
  | (char_but_gt_question * as text)
    { if String.length text <> 0 then begin
        store_pi_char '?';
        store_pi_string text;
        in_pi_after_space lexbuf end }
  | eof { unterminated_pi () }

and in_cdata = parse
  | "]]"
    { in_cdata_after_brackets lexbuf }
  | ']'
    { store_cdata_char ']';
      in_cdata lexbuf }
  | newline_char
    { update_newlines lexbuf;
      store_cdata_char '\n';
      in_cdata lexbuf }
  | (char_but_gt_question * as text)
    { store_cdata_string text;
      in_cdata lexbuf }
  | eof { unterminated_cdata () }
and in_cdata_after_brackets = parse
  | ']'
    { store_cdata_char ']';
      in_cdata_after_brackets lexbuf }
  | (char_but_gt_right_bracket * as text)
    { if String.length text <> 0 then begin
        store_cdata_string "]]";
        store_cdata_string text;
        in_cdata lexbuf end
    }
  | newline_char
    { update_newlines lexbuf;
      store_cdata_char '\n';
      in_cdata lexbuf }
  | eof { unterminated_cdata () }

and in_char_data_or_references char_data_mode accu = parse
  | char_data
    { let char_data = Lexing.lexeme lexbuf in
      if String.length char_data = 0 then
        match char_data_mode with
        | Inside_document -> CharDataIntercal (Intercal.rev accu)
        | Standalone_char_data -> unterminated_char_data ()
      else
        let accu = Intercal.cons_alpha char_data accu in
        in_char_data_or_references char_data_mode accu lexbuf }

  | "&#" (decimal_digit+ as dec_int32) ';'
    { let start_pos, end_pos = lexbuf.lex_start_p, lexbuf.lex_curr_p in
      let reference =
        Xml_basics_ast.Char_ref
          (int32_of_string start_pos end_pos dec_int32) in
      let accu = Intercal.cons_beta reference accu in
      in_char_data_or_references char_data_mode accu lexbuf }
  | "&#x" (hexadecimal_digit+ as hex_int32) ';'
    { let start_pos, end_pos = lexbuf.lex_start_p, lexbuf.lex_curr_p in
      let reference =
        Xml_basics_ast.Char_ref
          (int32_of_string start_pos end_pos ("0x" ^ hex_int32)) in
      let accu = Intercal.cons_beta reference accu in
      in_char_data_or_references char_data_mode accu lexbuf }

  | "&" (name as nm) ';'
    { let reference = Xml_basics_ast.Entity_ref (make_xml_name lexbuf nm) in
      let accu = Intercal.cons_beta reference accu in
      in_char_data_or_references char_data_mode accu lexbuf }

  | "]]>"
    { match !char_data_start_pos with
      | Some (start_pos, end_pos) ->
        error (Invalid_sequence ("CharData", "]]>")) start_pos end_pos
      | _ -> assert false }

   (* This case should not occur given the actual patterns in rule element.
      However, we keep it as a safe guard, in case patterns in rule element
      would change. *)
  | eof
    { match char_data_mode with
      | Inside_document -> unterminated_char_data ()
      | Standalone_char_data -> CharDataIntercal (Intercal.rev accu) }

and in_system_literal c = parse
  | eof { unterminated_system_literal () }
  | newline_char
    { update_newlines lexbuf;
      store_system_literal_char '\n';
      in_system_literal c lexbuf }
  | _
    { let char = Lexing.lexeme_char lexbuf 0 in
      if char <> c then begin
        store_system_literal_char char;
        in_system_literal c lexbuf
      end}

(*  Auxiliaries and token validity checkers *)

(* A lexer to check validity of Name read as Nmtoken *)
and name_of_nmtoken_checker start_pos end_pos = parse
  | name as nm
    { try
        make_xml_name lexbuf nm
      with
      | Error (err, _, _) ->
        error err start_pos end_pos }
  | [^'\000']* as nm_token
    { error (Invalid_xml_name nm_token) start_pos end_pos }

(* A lexer to check validity of pubid literals. *)
and pubid_literal_checker start_pos end_pos = parse
  | '\"' (pubid_char | '\'') * '\"' { () }
  | '\'' pubid_char * '\'' { () }
  | _
    { let char = Lexing.lexeme_char lexbuf 0 in
      error (Illegal_character ("PubId", char)) start_pos end_pos }

(* A lexer to check validity of att_value literals. *)
and att_value_literal_checker start_pos end_pos = parse
  | '\"' [^ '<' '&' '\"' ] * '\"' { () }
  | '\'' [^ '<' '&' '\'' ] * '\'' { () }
  | _
    { let char = Lexing.lexeme_char lexbuf 0 in
      error (Illegal_character ("AttValue", char)) start_pos end_pos }

(* A lexer to check the validity of encodings. *)
(* [81] EncName ::= [A-Za-z] ([A-Za-z0-9._] | '-')* *)
and enc_name_literal_checker start_pos end_pos = parse
  | alphabetic (alphanumeric | [ '.' '_' '-']) * { () }
  | _
    { let char = Lexing.lexeme_char lexbuf 0 in
      error (Illegal_character ("Encoding", char)) start_pos end_pos }

(* A lexer to replace character references by true character. *)
(* and unescape_references start_pos end_pos = parse *)
(*   | [^ '&' ] + *)
(*     { store_unescape_string (Lexing.lexeme lexbuf); *)
(*       unescape_references start_pos end_pos lexbuf } *)
(*   | char_ref *)
(*     { let c = char_of_char_ref (Lexing.lexeme lexbuf) in *)
(*       store_unescape_char c; *)
(*       unescape_references start_pos end_pos lexbuf } *)
(*   | entity_ref *)
(*     { let c = *)
(*         char_of_entity_ref *)
(*           (Lexing.lexeme lexbuf) *)
(*           start_pos end_pos in *)
(*       store_unescape_char c; *)
(*       unescape_references start_pos end_pos lexbuf } *)
(*   | eof *)
(*     { get_unescape_string (); } *)

{

(* let unescape_text start_pos end_pos text = *)
(*   let lexbuf = Lexing.from_string text in *)
(*   unescape_references start_pos end_pos lexbuf *)
(*;; *)

let check_enc_name_literal start_pos end_pos system_literal =
  let lexbuf = Lexing.from_string system_literal in
  enc_name_literal_checker start_pos end_pos lexbuf
;;

let make_enc_name start_pos end_pos system_literal =
  check_enc_name_literal start_pos end_pos system_literal;
  system_literal
;;

let check_pubid_literal start_pos end_pos system_literal =
  let lexbuf = Lexing.from_string system_literal in
  pubid_literal_checker start_pos end_pos lexbuf
;;

let make_pubid_literal start_pos end_pos system_literal =
  check_pubid_literal start_pos end_pos system_literal;
  system_literal
;;

let check_att_value_literal start_pos end_pos system_literal =
  let lexbuf = Lexing.from_string system_literal in
  att_value_literal_checker start_pos end_pos lexbuf
;;

(* let make_att_value_literal start_pos end_pos system_literal = *)
(*   let att_value_literal = unescape_text start_pos end_pos system_literal in *)
(*   check_att_value_literal start_pos end_pos att_value_literal; *)
(*   att_value_literal *)
(*;; *)

}

(*
 Local Variables:
  compile-command: "cd ../../..; make"
 End:
*)
