/***********************************************************************/
/*                                                                     */
/*                          Ocaml-xml Library                          */
/*                                                                     */
/*                   Pierre Weis, INRIA Rocquencourt                   */
/*                                                                     */
/*  Copyright 2013-2018,                                               */
/*  Institut National de Recherche en Informatique et en Automatique.  */
/*  All rights reserved.                                               */
/*                                                                     */
/*  This file is distributed under the terms of the BSD License.       */
/*                                                                     */
/***********************************************************************/

/* $Id: xml_parser.mly,v 1.1 2019-05-10 14:00:50 jpc Exp $ */

%{

(* Prelude part: this is pure Caml. *)

(* The parser *)

open Xml_ast;;

(* FIXME: validity checks. *)
let make_enc_name s = s;;

let make_xml_name start_pos end_pos nm_token =
  try
    Xml_basics_ast.to_xml_name nm_token
  with
  | Failure _ ->
    raise (Xml_parser_error.Error
      (Xml_parser_error.Invalid_xml_name nm_token, start_pos, end_pos))
;;

(* FIXME: validity checks. *)
let make_att_value_literal s =
  (* FIXME: Should att_value_text to produce a list of text | Reference. *)
  s
;;

(* FIXME: validity checks. *)
let make_entity_value_of_att_value start_pos end_pos att_value =
  let add_text, add_reference, add_pe_reference, get_result =
    let result = ref [] in
    let add_item i = result := i :: !result in
    let add_text t =
      if t <> "" then
        add_item (Xml_basics_ast.Entity_value_text t)
    and add_reference r = add_item (Xml_basics_ast.Entity_value_reference r)
    and add_pe_reference r =
      let xml_name = Xml_basics_ast.to_xml_name r in
      add_item (Xml_basics_ast.Entity_value_pe_reference xml_name)
    and get_result () = List.rev !result in
    add_text, add_reference, add_pe_reference, get_result in
  let cut s =
    let ib = Scanf.Scanning.from_string s in
    let rec loop () =
      if Scanf.Scanning.end_of_input ib then
        ()
      else
        try
          Scanf.bscanf ib "%[^%%]%%%[^;];%[^%%]" (fun t r s ->
            add_text t;
            add_pe_reference r;
            add_text s);
          loop ()
        with
          | Scanf.Scan_failure message ->
            raise (Xml_parser_error.Error
              (Xml_parser_error.Invalid_entity_value (att_value, message),
               start_pos,
               end_pos))
          | Failure _ -> assert false
          | End_of_file ->
          (* That case happen when there is no parameter entity reference.
             That is why we append all the string. *)
            add_text s
          | Invalid_argument _ -> assert false in
    loop () in
  List.iter
    (function
    | Xml_basics_ast.Att_value_text t ->  cut t
    | Xml_basics_ast.Att_value_reference r -> add_reference r)
    att_value;
  get_result ()
;;

let make_tag (tag_name, tag_attributes) =
  let make_attribute (attr_name, attr_val) =
    let attr_val = make_att_value_literal attr_val in
    {
      tag_attribute_name = attr_name;
      tag_attribute_value = attr_val;
    } in
  let attrs = List.map make_attribute tag_attributes in
  {
    tag_name = tag_name;
    tag_attributes = attrs;
  }
;;

let make_content_element_tag start_pos end_pos tag children etag =
  let tag = make_tag tag in
  if tag.tag_name <> etag then
    raise (Xml_parser_error.Error
      (Xml_parser_error.Invalid_closing_tag (tag.tag_name, etag), start_pos, end_pos))
  else
    Content_element_tag (tag, children)
;;

let make_empty_element_tag (tag_name, attributes) =
  let attributes =
    List.map (fun (att_name, att_value) ->
      {
        tag_attribute_name = att_name;
        tag_attribute_value = att_value;
      }) attributes in
  let tag = {
    tag_name = tag_name;
    tag_attributes = attributes;
  } in
  Xml_ast.Empty_element_tag tag
;;

(*
let check_is_empty_CharData s =
  let ls = String.length s in
  let rec loop i =
    i >= ls ||
    let c = s.[i] in
    (c = ' ' || c = '\t' || c = '\r' || c = '\n') &&
    loop (succ i) in
  if not (loop 0) then
    not_implemented "check_is_empty_CharData error case"
;;
*)

let make_sdd_decl = function
  | true -> Standalone
  | false -> Not_standalone
;;

let make_xml_decl vi enc_opt sd_opt =
  {
    xml_version_info = vi;
    xml_encoding_decl = enc_opt;
    xml_standalone_decl = sd_opt;
  }
;;

let make_att_def start_pos end_pos nm_token att_type att_default =
  let name = make_xml_name start_pos end_pos nm_token in
  {
    att_def_name = name;
    att_def_type = att_type;
    att_def_default = att_default;
  }
;;

let make_att_type start_pos end_pos = function
  | "CDATA" -> String_type
  | "ID" -> Tokenized_type Id_type
  | "IDREF" -> Tokenized_type Idref_type
  | "IDREFS" -> Tokenized_type Idrefs_type
  | "ENTITY" -> Tokenized_type Entity_type
  | "ENTITIES" -> Tokenized_type Entities_type
  | "NMTOKEN" -> Tokenized_type Nmtoken_type
  | "NMTOKENS" -> Tokenized_type Nmtokens_type
  | nm_token ->
    raise (Xml_parser_error.Error
      (Xml_parser_error.Invalid_att_type nm_token, start_pos, end_pos))
;;

let make_pi (tgt, instructions) =
  {
    pi_target = tgt;
    pi_instructions = instructions;
  }
;;

let make_content_spec_mod_choice choice spec_mod =
  Content_spec_choice (spec_mod, choice)
;;

let make_content_spec_mod_seq seq spec_mod =
  Content_spec_seq (spec_mod, seq)
;;

let make_content_spec_particle_mod_choice choice spec_mod =
  Content_particle_choice (spec_mod, choice)
;;

let make_content_spec_particle_mod_seq seq spec_mod =
  Content_particle_seq (spec_mod, seq)
;;

let make_content_spec_particle_mod_name start_pos end_pos nm spec_mod =
  Content_particle_name (spec_mod, make_xml_name start_pos end_pos nm)
;;

let make_content_spec start_pos end_pos = function
  | "EMPTY" -> Content_spec_empty
  | "ANY" -> Content_spec_any
  | nm_token ->
    raise (Xml_parser_error.Error
      (Xml_parser_error.Invalid_content_spec nm_token, start_pos, end_pos))
;;

let make_element_content char_data_intercal_child_intercal =
  let alpha_folder accu alpha =
    let char_data_folder accu char_data_text =
      Intercal.cons_alpha (Char_data char_data_text) accu in
    let reference_folder accu reference =
      Intercal.cons_beta (Child_reference reference) accu in
    Intercal.fold_left char_data_folder reference_folder accu alpha in
  let beta_folder accu beta =
    Intercal.cons_beta beta accu in
  let element_content =
    Intercal.fold_left alpha_folder beta_folder
      Intercal.empty char_data_intercal_child_intercal in
  let element_content = Intercal.rev element_content in
  Element_content element_content
;;

%}

%token Eof
%token OpenConditionalSect
%token CloseConditionalSect
%token OpenBracket
%token CloseBracket
%token OpenParen
%token CloseParen

%token CloseParenQuestion
%token CloseParenStar
%token CloseParenPlus

%token GT

%token Question
%token Star
%token Plus
%token Eq
%token Comma
%token Bar

%token XmlDeclStart
%token XmlDeclEnd

%token <string> ExternalIDSystem
%token <string * string> ExternalIDPublic
%token <string> PublicID
%token PCDATA
%token REQUIRED
%token IMPLIED
%token FIXED

%token INCLUDE
%token <string> Ignore

%token <string> VersionInfo
%token <string> EncodingDecl
%token <bool> SDDecl
%token <Xml_basics_ast.xml_name> NDataDecl

%token <Xml_basics_ast.xml_name> AttlistName
%token <Xml_basics_ast.xml_name> NotationTypeName
%token <string> Space
%token <string> Comment
%token <Xml_basics_ast.xml_name> NotationName
%token <Xml_basics_ast.xml_name> GEDeclName
%token <Xml_basics_ast.xml_name> PEDeclName

%token <Xml_basics_ast.att_value> AttValue
%token <Xml_basics_ast.xml_name> PEReference
%token <int32> CharRef
%token <Xml_basics_ast.xml_name> EntityRef

%token <string> CData
%token <Xml_lexer_types.char_data_intercal> CharDataIntercal

%token <Xml_lexer_types.token_tag> FirstSTag
%token <Xml_lexer_types.token_tag> FirstEmptyElemTag
%token <Xml_lexer_types.token_tag> STag
%token <Xml_lexer_types.token_tag> EmptyElemTag

%token <Xml_basics_ast.tag_name> ETag

%token <Xml_basics_ast.nm_token> Nmtoken
%token <Xml_basics_ast.tag_attribute_name> DocTypeName
%token <Xml_basics_ast.tag_attribute_name> ElementName

%token <Xml_basics_ast.pi_target * Xml_basics_ast.pi_instructions> PI

%start prolog
%type <Xml_ast.prolog> prolog

%start element
%type <Xml_ast.element> element

%start element_after_prolog
%type <Xml_ast.element> element_after_prolog

%start ext_subset
%type <Xml_ast.raw_ext_subset> ext_subset

%%

/* [1] document ::= prolog element Misc* */
/* document :
 | prolog element misc_star
   {
     Xml_document {
       prolog = $1;
       element = $2;
       misc_star = $3;
     }
   } */
/* | prolog element misc_star CharData
   {
     check_is_empty_CharData $4;
     Xml_document {
       prolog = $1;
       element = $2;
       misc_star = $3;
     }
   }
*/
;

/* Document/Prolog */

/* [22] prolog ::= XMLDecl? Misc* (doctypedecl Misc*)? */
prolog : xml_decl_opt misc_star doc_type_decl_opt Eof
  {
    let doc_type_decl, misc_star_trailing =
      match $3 with
      | None -> None, []
      | Some (doc_type_decl, misc_star_trailing) ->
        Some doc_type_decl, misc_star_trailing in
    {
      xml_decl = $1;
      xml_misc_star = $2;
      xml_doc_type_decl = doc_type_decl;
      xml_misc_star_trailing = misc_star_trailing;
    }
  }
;

/* Document/Prolog/xmldecl_opt */

xml_decl_opt :
  | { None }
  | xml_decl { Some $1 }
;

/* [23] XMLDecl ::= '<?xml' VersionInfo EncodingDecl? SDDecl? S? '?>' */
xml_decl :
  | XmlDeclStart VersionInfo encoding_decl_opt sdd_decl_opt XmlDeclEnd
    { make_xml_decl $2 $3 $4 }
;

/* [24] VersionInfo ::=
        S 'version' Eq ("'" VersionNum "'" | '"' VersionNum '"') */
/* [25] Eq ::= S? '=' S? */
/* [26] VersionNum ::= '1.' [0-9]+ */
/* version_info :
  | Version Eq SystemLiteral { make_version $3 }
; */

encoding_decl_opt :
  | { None }
  | EncodingDecl { Some (make_enc_name $1) }
;

/* [80] EncodingDecl ::= S 'encoding' Eq ('"' EncName '"' | "'" EncName "'") */
/* EncName ::= SystemLiteral with some constraints. */
/* EncodingDecl is now a lexem, no more grammar rule. */

sdd_decl_opt :
  | { None }
  | SDDecl
    { Some (make_sdd_decl $1) }
;

/* Document/Prolog/doc_type_decl_opt */

doc_type_decl_opt :
  | { None }
  | doc_type_decl misc_star { Some ($1, $2) }
;

/* [28] doctypedecl ::=
  '<!DOCTYPE' S Name (S ExternalID)? S? ('[' intSubset ']' S?)? '>' */
doc_type_decl :
  | DocTypeName external_id_opt int_subset_outer_opt GT
    {
      {
        doctype_name = $1;
        external_id = $2;
        int_subset = $3;
      }
    }
;

external_id_opt :
  | { None }
  | external_id { Some $1 }
;

external_id :
  | ExternalIDSystem { System $1 }
  | ExternalIDPublic
    { let pl, sl = $1 in
      Public (pl, sl) }
;

/* public_id :
  | NamePublic pubid_literal { $2 }
; */

/* Document/Prolog/doc_type_decl_opt/int_subset_outer_opt */

int_subset_outer_opt :
  | { None }
  | int_subset_outer { Some $1 }
;

int_subset_outer :
  OpenBracket int_subset CloseBracket { $2 }
;

/*
  The rules for intSubset is:

  [28b] intSubset ::= (markupdecl | DeclSep)*
  [28a] DeclSep   ::= PEReference | S

  We compress those two rules into:
  intSubset ::= (markupdecl | PEReference | S)*
*/
int_subset :
  | { [] }
  | int_subset_elem int_subset { $1 :: $2 }
;

/* Document/Prolog/doc_type_decl_opt/int_subset/insubset_elem */
int_subset_elem :
  | markup_decl { Markup_decl $1 }
  | decl_sep    { Sep_decl $1 }
;

decl_sep :
  | PEReference { Pe_reference_decl $1 }
  | Space { Space_decl $1 }
;

/* Document/Prolog/doc_type_decl_opt/int_subset/insubset_elem/markup_decl */
/* [29] markupdecl ::= elementdecl | AttlistDecl | EntityDecl | NotationDecl | PI | Comment */
markup_decl :
  | element_decl { $1 }
  | attlist_decl { $1 }
  | entity_decl { Entity_decl $1 }
  | notation_decl { $1 }
  | PI { Pi_decl (make_pi $1) }
  | Comment { Comment_decl $1 }
;

/* Document/Prolog/doc_type_decl_opt/int_subset/insubset_elem/markup_decl/element_decl */
element_decl :
  | ElementName content_spec GT
    { Element_decl ($1, $2) }
;

/* Document/Prolog/doc_type_decl_opt/int_subset/insubset_elem/markup_decl/element_decl/content_spec */
content_spec :
  | Nmtoken
    {
      let start_pos = Parsing.rhs_start_pos 1 in
      let end_pos = Parsing.rhs_end_pos 1 in
      make_content_spec start_pos end_pos $1
    }
  | children { Content_spec_children $1 }
  | mixed { Content_spec_mixed $1 }
;

/* Document/Prolog/.../element_decl/content_spec/children */
children :
  | OpenParen choice close_paren_spec_mod
    { make_content_spec_mod_choice $2 $3 }
  | OpenParen seq close_paren_spec_mod
    { make_content_spec_mod_seq $2 $3 }
;

choice :
  | cp cp_alternative_plus
    { $1, $2 }
;

seq :
  | cp cp_comma_star
    { $1, $2 }
;

close_paren_spec_mod :
  | CloseParen { None_mod }
  | CloseParenPlus { Plus_mod }
  | CloseParenStar { Star_mod }
  | CloseParenQuestion { Opt_mod }
;

spec_mod :
  | { None_mod }
  | Plus { Plus_mod }
  | Star { Star_mod }
  | Question { Opt_mod }
;

cp :
  | OpenParen choice close_paren_spec_mod
    { make_content_spec_particle_mod_choice $2 $3 }
  | OpenParen seq close_paren_spec_mod
    { make_content_spec_particle_mod_seq $2 $3 }
  | Nmtoken spec_mod
    { let start_pos = Parsing.rhs_start_pos 1 in
      let end_pos = Parsing.rhs_end_pos 1 in
      make_content_spec_particle_mod_name start_pos end_pos $1 $2 }
;

cp_alternative_plus :
  | Bar cp cp_alternative_star { $2, $3 }
;

cp_alternative_star :
  | { [] }
  | Bar cp cp_alternative_star { $2 :: $3 }
;

cp_comma_star :
  | { [] }
  | Comma cp cp_comma_star { $2 :: $3 }
;

/* Document/Prolog/.../element_decl/content_spec/mixed */
/* The rule given in w3 xml recommandation is
   [51] Mixed ::= '(' S? '#PCDATA' (S? '|' S? Name)* S? ')*'
                | '(' S? '#PCDATA' S? ')'

   However,the example given in w3 xml recommandation leads to add a rule for
   PEReferences lists as in:
   [51bis] Mixed ::= '(' S? '#PCDATA' (S? '|' S? Name)* S? ')*'
                   | '(' S? '#PCDATA' (S? '|' S? PEReference)* S? ')*'
                   | '(' S? '#PCDATA' S? ')'
   Now to avoid confusion between empty list of name and pereference, we use:
   [51bis] Mixed ::= '(' S? '#PCDATA' S? ')*'
                  | '(' S? '#PCDATA' (S? '|' S? Name)+ S? ')*'
                  | '(' S? '#PCDATA' (S? '|' S? PEReference)+ S? ')*'
                  | '(' S? '#PCDATA' S? ')'
*/
mixed :
  | OpenParen PCDATA CloseParenStar
    { PCData_empty }
  | OpenParen PCDATA alternative_names1 CloseParenStar
    { PCData_alternatives $3 }
/*
  | OpenParen PCDATA alternative_pereferences1 CloseParenStar
    { PCData_alternatives_PEReferences $3 }
*/
  | OpenParen PCDATA CloseParen
    { PCData }
;

alternative_names1 :
  | alternative_name alternative_names { $1 :: $2 }
;

alternative_names :
  | { [] }
  | alternative_name alternative_names { $1 :: $2 }
;

alternative_name :
  | Bar Nmtoken
    {
      let start_pos = Parsing.rhs_start_pos 2 in
      let end_pos = Parsing.rhs_end_pos 2 in
      make_xml_name start_pos end_pos $2
    }
;

/*
alternative_pereferences1 :
  | alternative_pereference alternative_pereferences { $1 :: $2 }
;

alternative_pereferences :
  | { [] }
  | alternative_pereference alternative_pereferences { $1 :: $2 }
;

alternative_pereference :
  | Bar PEReference { $2 }
;
*/

attlist_decl :
  | AttlistName att_def_list GT { Attlist_decl ($1, $2) }
;

att_def_list :
  | { [] }
  | att_def att_def_list
    { $1 :: $2 }
;

att_def :
  /* Should be token Name, but in prolog the lexer produce a Nmtoken.
     We check that this token is indeed a Name in function make_att_def. */
  | Space Nmtoken Space att_type Space att_default
    {
      let start_pos = Parsing.rhs_start_pos 2 in
      let end_pos = Parsing.rhs_end_pos 2 in
      make_att_def start_pos end_pos $2 $4 $6
    }
;

att_type :
  | Nmtoken
    {
      let start_pos = Parsing.rhs_start_pos 1 in
      let end_pos = Parsing.rhs_end_pos 1 in
      make_att_type start_pos end_pos $1
    }
  | enumerated_type { Enumerated_type $1 }
;

enumerated_type :
  | notation_type { $1}
  | enumeration_type { $1}
;

notation_type :
  | NotationTypeName or_names CloseParen
    { Notation_type ($1, $2) }
;

or_names :
  | { [] }
  | Bar Nmtoken or_names
    { let start_pos = Parsing.rhs_start_pos 2 in
      let end_pos = Parsing.rhs_end_pos 2 in
      make_xml_name start_pos end_pos $2 :: $3 }
;

or_nmtokens :
  | { [] }
  | Bar Nmtoken or_nmtokens
    { $2 :: $3 }
;

/* [59] Enumeration ::= '(' S? Nmtoken (S? '|' S? Nmtoken)* S? ')' */
enumeration_type :
  | OpenParen Nmtoken or_nmtokens CloseParen
    { Enumeration_type ($2, $3) }
;

att_default :
  | REQUIRED { Required }
  | IMPLIED { Implied }
  | FIXED att_value { Fixed_att_value $2 }
  | att_value { Att_value $1 }
;

att_value :
  | AttValue { make_att_value_literal $1 }
;

entity_decl :
  | GEDeclName gentity_def GT { Ge_decl ($1, $2) }
  | PEDeclName pentity_def GT { Pe_decl ($1, $2) }
;

gentity_def :
  | entity_value { Gentity_value $1 }
  | external_id ndata_decl_opt { Gentity_external_id ($1, $2) }
;

ndata_decl_opt :
  | { None }
  | ndata_decl { Some $1 }
;

ndata_decl :
  | NDataDecl { $1 }
;

pentity_def :
  | entity_value { Pentity_value $1 }
  | external_id { Pentity_external_id $1 }
;

entity_value :
  | AttValue
    {
      let start_pos = Parsing.rhs_start_pos 1 in
      let end_pos = Parsing.rhs_end_pos 1 in
      make_entity_value_of_att_value start_pos end_pos $1
    }
;

/* [82] NotationDecl ::= '<!NOTATION' S Name S (ExternalID | PublicID) S? '>'
   [83] PublicID     ::= 'PUBLIC' S PubidLiteral */
notation_decl :
  | NotationName external_id GT { Notation_decl ($1, External_id $2) }
  | NotationName PublicID GT { Notation_decl ($1, Public_id $2) }
;

/* External Subset: the second part of the DTD */

/* [30] extSubset ::= TextDecl? extSubsetDecl */
ext_subset :
  | ext_subset_decl { None, $1 }
  | text_decl ext_subset_decl { Some $1, $2 }
;

/* [77] TextDecl ::= '<?xml' VersionInfo? EncodingDecl S? '?>' */
text_decl :
  | XmlDeclStart version_info_opt EncodingDecl XmlDeclEnd
    {
      {
         text_version_info = $2;
         text_encoding_decl = $3;
      }
    }
;

version_info_opt :
  | { None }
  | VersionInfo { Some $1 }
;

/* [31] extSubsetDecl ::= (markupdecl | conditionalSect | DeclSep)* */
ext_subset_decl :
  | { [] }
  | ext_subset_decl_elem ext_subset_decl { $1 :: $2 }
;

ext_subset_decl_elem :
  | markup_decl { Ext_markup_decl $1 }
  | conditional_sect { Ext_conditional_sect $1 }
  | decl_sep { Ext_sep_decl $1 }
;

/* Conditional Section */
/* [61] conditionalSect ::= includeSect | ignoreSect */
conditional_sect :
  | include_sect { Include_sect $1 }
  | ignore_sect { Ignore_sect $1 }
;

/* [62] includeSect ::= '<![' S? 'INCLUDE' S? '[' extSubsetDecl ']]>'
   [VC: Proper Conditional Section/PE Nesting] */
include_sect :
 | INCLUDE ext_subset_decl CloseConditionalSect { $2 }
;

/* [63] ignoreSect ::= '<![' S? 'IGNORE' S? '[' ignoreSectContents* ']]>'
   [VC: Proper Conditional Section/PE Nesting]
   [64] ignoreSectContents ::= Ignore ('<![' ignoreSectContents ']]>' Ignore)*
   [65] Ignore ::= Char* - (Char* ('<![' | ']]>') Char*) */
ignore_sect :
 | Ignore { $1 }
;

/* Document/Element recursively read */
/* [39] element ::= EmptyElemTag | STag content ETag */
element :
    /* [44] EmptyElemTag ::= '<' Name (S Attribute)* S? '/>' */
  | EmptyElemTag
    { make_empty_element_tag $1 }
    /* [40] STag ::= '<' Name (S Attribute)* S? '>' */
  | STag content ETag
    { let start_pos = Parsing.rhs_start_pos 1 in
      let end_pos = Parsing.rhs_end_pos 3 in
      make_content_element_tag start_pos end_pos $1 $2 $3 }
;

/* Document/Element read after prolog parsing completed. */
element_after_prolog :
    /* [00] FirstEmptyElemTag ::= Name (S Attribute)* S? '/>' */
  | FirstEmptyElemTag { make_empty_element_tag $1 }
    /* [01] FirstSTag ::= Name (S Attribute)* S? '>' */
  | FirstSTag content ETag
    { let start_pos = Parsing.rhs_start_pos 1 in
      let end_pos = Parsing.rhs_end_pos 3 in
      make_content_element_tag start_pos end_pos $1 $2 $3 }
;

/* [43] content ::=
   CharData?
   ((element | Reference | CDSect | PI | Comment)
    CharData?)* */
content :
  | char_data_intercal_opt content_tag_star
    { make_element_content ($1, $2) }
;

char_data_intercal_opt  :
  | { None }
  | CharDataIntercal { Some $1 }
;

content_tag_star :
  | { [] }
  | content_tag_star_element content_tag_star { $1 :: $2 }
;

content_tag_star_element :
  | content_element char_data_intercal_opt
    {
      $1, $2
    }
;

content_element :
    /* Start by Empty_elem_tag or STag */
  | element { Child_element $1 }
  | CData { Child_cd_sect $1 }
    /* Start by <? */
  | PI { Child_pi (make_pi $1) }
    /* Start by <-- */
  | Comment { Child_comment $1 }
;

/* Document/misc_star */
misc_star :
  | { [] }
  | misc misc_star { $1 :: $2 }
;

/* [27] Misc ::= Comment | PI | S */
misc :
  | Comment { Misc_comment $1 }
  | PI { Misc_pi (make_pi $1) }
  | Space { Misc_space $1 }
;
/*

 Local Variables:
  compile-command: "cd ../../..; make"
  End:

*/
