(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, Clément Franchini                    *)
(*                   INRIA Rocquencourt                                *)
(*                                                                     *)
(*  Copyright 2013-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: slx_to_xml_ast.ml,v 1.1 2019-05-10 14:00:17 jpc Exp $ *)

type annex_name =
  | Block_diagram_defaults
  | Config_set_info
  | Graphical_interface
  | Stateflow
  | Annex_file_name of Path.file_name
;;

type annexes = (annex_name * Xml.document) list;;

type xml_model = {
  model_ident : Mdl_ast.model_ident;
  blockdiagram : Xml.document;
  annexes : annexes;
}
;;

let debug s = Say.debug (Printf.sprintf "slx_to_xml_ast: %s" s);;

let fdebug f =
  Say.fdebug
    (fun epf ->
     Format.fprintf epf "slx_to_xml_ast: %a"
       (fun epf () -> f epf)
       ())
;;

let fatal_error s =
  Say.fatal_error (Printf.sprintf "slx_to_xml_ast: %s" s)
;;

let xml_parse_explicit_file_name efname =
  Xml.parse (Path.of_explicit_file_name efname)
;;

let read_xml_file efname =
  let slx_model_ident = Configuration.get_slx_model_ident () in
  {
    model_ident = slx_model_ident;
    blockdiagram = xml_parse_explicit_file_name efname;
    annexes = [];
  }
;;

let simulink_sub_dir_base_name =
  Configuration.default_slx_simulink_dir_name
;;

let blockdiagram_base_name =
  Configuration.default_slx_blockdiagram_base_name
and block_diagram_defaults_annex_base_name =
  Configuration.default_slx_block_diagram_defaults_base_name
and config_set_info_annex_base_name =
  Configuration.default_slx_config_set_info_base_name
and graphical_interface_annex_base_name =
  Configuration.default_slx_graphical_interface_base_name
and stateflow_annex_base_name =
  Configuration.default_slx_stateflow_base_name
;;

let read_slx_xml_files new_dir =
  fdebug (fun epf -> Format.fprintf epf "read_slx_xml_files %S" new_dir);
  let simulink_sub_dir =
    Filename.concat new_dir simulink_sub_dir_base_name in
  let efnames =
    Path.list_files_with_extension
      Configuration.xml_source_file_extension
      simulink_sub_dir in
  fdebug
    (fun epf ->
     Format.fprintf epf "@[<v>%a@]"
       (fun epf ->
        List.iteri
          (fun i efname ->
           if i > 0 then Format.fprintf epf "@ ";
           Format.fprintf epf "entry: %s"
             (Path.of_explicit_file_name efname)))
       efnames);

  let read_annex annex_efname =
    fdebug
      (fun epf ->
       Format.fprintf epf "Slx_to_xml_ast: read_annex %S"
         (Path.of_explicit_file_name annex_efname));
    let annex_name =
      let base_name = Path.base_name annex_efname in
      if base_name = block_diagram_defaults_annex_base_name then
        Block_diagram_defaults else
      if base_name = config_set_info_annex_base_name then
        Config_set_info else
      if base_name = graphical_interface_annex_base_name then
        Graphical_interface else
      if base_name = stateflow_annex_base_name then
        Stateflow
      else
        Annex_file_name (
          Filename.concat simulink_sub_dir_base_name base_name) in
    annex_name, xml_parse_explicit_file_name annex_efname in

  let rec classify blockdiagram_opt annexes efnames =
    match blockdiagram_opt, efnames with
    | None, [] -> fatal_error "Missing blockdiagram xml file"
    | None, efname :: efnames ->
      if Path.base_name efname = blockdiagram_base_name then
        let blockdiagram = xml_parse_explicit_file_name efname in
        classify (Some blockdiagram) annexes efnames
      else
        let annex = read_annex efname in
        classify None (annex :: annexes) efnames
    | Some blockdiagram, [] -> blockdiagram, annexes
    | Some _blockdiagram, efname :: efnames ->
      if Path.base_name efname = blockdiagram_base_name then
        fatal_error (Printf.sprintf
          "two %s file in the simulink extracted directory"
          blockdiagram_base_name)
      else
        let annex = read_annex efname in
        classify blockdiagram_opt (annex :: annexes) efnames in

  let model_ident = Configuration.get_slx_model_ident () in
  let blockdiagram, annexes = classify None [] efnames in
  {
    model_ident = model_ident;
    blockdiagram = blockdiagram;
    annexes = annexes;
  }
;;

let rec mkdir dir =
  debug (Printf.sprintf "mkdir: %S" dir);
  try
    if Sys.is_directory dir then
      (* the directory exists, we do not need to create it *)
      debug (Printf.sprintf "mkdir: directory %S exists" dir)
    else
      (* there is already something with the same name, but it is not a
         directory *)
      fatal_error (Printf.sprintf "can not create directory %S" dir)
  with
  | Sys_error _ ->
    (* the directory does not exist *)
    debug (Printf.sprintf "mkdir: directory %S does not exist" dir);
    mkdir (Filename.dirname dir);
    debug (Printf.sprintf "mkdir: creating %S" dir);
    Unix.mkdir dir 0o777
;;

let extract_zip_file dirname efname =
  let extract_entry ifile entry =
    debug (Printf.sprintf
      "extract_zip_file: extract_entry: <- %S" entry.Zip.filename);
    let entry_efname = Filename.concat dirname entry.Zip.filename in
    if entry.Zip.is_directory then
      begin
        debug (Printf.sprintf
          "extract_zip_file: extract_entry: directory entry %S"
          entry.Zip.filename);
        mkdir entry.Zip.filename
      end
    else begin
      debug (Printf.sprintf
        "extract_zip_file: extract_entry: copying %s to %s"
        entry.Zip.filename
        entry_efname);
      mkdir (Filename.dirname entry_efname);
      try
        Zip.copy_entry_to_file ifile entry entry_efname
      with
      | x ->
        fatal_error (Printf.sprintf
          "extract_zip_file: extract_entry: error occured when creating file \
           %S: %s"
          entry.Zip.filename
          (Printexc.to_string x))
    end in

  let extract zipfile =
    let ic = Zip.open_in zipfile in
    try
      List.iter (extract_entry ic) (Zip.entries ic);
      Zip.close_in ic
    with
    | x ->
      Zip.close_in ic;
      raise x in

  debug (Printf.sprintf
    "extract_zip_file: unzipping %s" (Path.of_explicit_file_name efname));
  extract (Path.of_explicit_file_name efname)
;;

let configuration_slx_extension = Path.to_file_extension ".slx";;

let debug_rmdir dir =
  if not (Configuration.get_debug_slx ()) then
    Path.rmdir dir
;;

let extract_slx_file efname =
  let dname = Path.dir_name efname in
  let bname =
    Path.chop_extension (Path.base_name efname) configuration_slx_extension in
  let new_dir = Filename.concat dname bname in
  begin
    try
      mkdir new_dir
    with
    | Unix.Unix_error (Unix.EEXIST, _, _) -> ()
  end;
  try
    extract_zip_file new_dir efname;
    new_dir
  with
  | x ->
    debug_rmdir new_dir;
    raise x
;;

let read_slx_file efname =
  let new_dir = extract_slx_file efname in
  debug (Printf.sprintf "slx: unzip phase done");
  let r =
    try
      read_slx_xml_files new_dir
    with
    | x ->
      debug_rmdir new_dir;
      raise x in
  debug_rmdir new_dir;
  debug "slx: read xml phase done";
  r
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
 End:
*)
