(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, Clément Franchini                    *)
(*                   INRIA Rocquencourt                                *)
(*                                                                     *)
(*  Copyright 2013-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: xml_explicit_ast_to_mdl_ast.ml,v 1.1 2019-05-10 14:00:17 jpc Exp $ *)

open Binding;;
open Ast_types;;
open Mdl_ast;;
open Xml_explicit_ast;;
open Xml_translating_basics;;

type vendor_config_object =
  | Vendor_config_object_binding of Mdl_ast.vendor_binding
  | Vendor_config_object_object of Mdl_ast.object_description
;;

let module_name = "Xml_explicit_ast_to_mdl_ast";;

let not_yet = Xml_translating_basics.not_yet_warning module_name;;

let fatal_error = Xml_translating_basics.fatal_error module_name;;

let warning = Xml_translating_basics.warning module_name;;

let fdebug = Xml_translating_basics.fdebug module_name;;

let check_opt_is_empty m = function
  | None -> ()
  | Some _ -> fatal_error m
;;

let sid_to_block_names () =
  let block_names =
    Hashtbl.create
      (Unatural.to_int (Configuration.get_mean_blocks_by_system ())) in
  let get_block_name sid =
    if Hashtbl.mem block_names sid then
      Hashtbl.find block_names sid
    else
      fatal_error (
        Format.fprintf Format.str_formatter
          "block with SID %a was not found"
          Mdl_ast_print.print_block_sid sid;
        Format.flush_str_formatter ()) in
  let add_block_name sid name = Hashtbl.add block_names sid name in
  get_block_name, add_block_name
;;

let unclassified_text_to_parse = function
  | Ddv text -> text
  | Ddv_classified (parameter_class, text) ->
    fatal_error
      (Format.fprintf Format.str_formatter
         "trying to get an unclassified text, but found %S classified as %a"
         text
         Xml_explicit_ast_print.print_parameter_class parameter_class;
       Format.flush_str_formatter ())
;;

let check_bool_value = function
  | Bool bool -> bool
  | Float _
  | Int _
  | String _
  | Ident _ as simple_value ->
    fatal_error
      (Format.fprintf Format.str_formatter
         "a boolean was expected but %a was read"
         Mdl_ast_print.print_simple_value simple_value;
       Format.flush_str_formatter ())
;;

let check_int_value non_negative = function
  | Int s ->
    if non_negative then
      begin
        try
          ignore (Unatural.of_string s)
        with
        | Failure _ ->
          fatal_error (Printf.sprintf
            "waiting for an unsigned integer, \
             found a negative or invalid value %s"
            s)
      end;
    s
  | Bool _
  | Float _
  | String _
  | Ident _ as simple_value ->
    fatal_error
      (Format.fprintf Format.str_formatter
         "an integer was expected but %a was read"
         Mdl_ast_print.print_simple_value simple_value;
       Format.flush_str_formatter ())
;;

let mdl_bool bound_value =
  let text_to_parse =
    match bound_value with
    | Ddv text
    | Ddv_classified (Logical, text) -> text
    | Ddv_classified
        ((Char | Double | Int32 | Uint32) as parameter_class,
         text) ->
      fatal_error
        (Format.fprintf Format.str_formatter
           "trying to get a bool value from value %S classified as %a"
           text
           Xml_explicit_ast_print.print_parameter_class parameter_class;
         Format.flush_str_formatter ()) in
  let bool =
    check_bool_value (Mdl_entry_parsers.simple_value_parser text_to_parse) in
  mdl_bool_of_string bool
;;

let text = function
  | Ddv text
  | Ddv_classified (Char, text) -> text
  | Ddv_classified
      ((Logical | Double | Int32 | Uint32) as parameter_class,
       text) ->
    fatal_error
      (Format.fprintf Format.str_formatter
         "trying to get a text value from value %S classified as %a"
         text
         Xml_explicit_ast_print.print_parameter_class parameter_class;
       Format.flush_str_formatter ())
;;

let simple_value_of_custom_property_value custom_property_type bound_value =
  let text = unclassified_text_to_parse bound_value in
  match custom_property_type with
  | Custom_property_type_uint8
  | Custom_property_type_uint16
  | Custom_property_type_uint32
  | Custom_property_type_uint64 ->
    let simple_value = Mdl_entry_parsers.simple_value_parser text in
    ignore (check_int_value true simple_value);
    simple_value
  | Custom_property_type_int8
  | Custom_property_type_int16
  | Custom_property_type_int32
  | Custom_property_type_int64 ->
    let simple_value = Mdl_entry_parsers.simple_value_parser text in
    ignore (check_int_value false simple_value);
    simple_value
  | Custom_property_type_slbool ->
    let simple_value = Mdl_entry_parsers.simple_value_parser text in
    ignore (check_bool_value simple_value);
    simple_value
  | Custom_property_type_string ->
    Mdl_entry_parsers.string_simple_value_parser text
  | Custom_property_type_custom _ -> Mdl_ast.String text
;;

let workspace_data_source_type bound_value =
    Mdl_entry_parsers.workspace_data_source_type_parser (text bound_value)
;;

(*
let wstring bound_value =
  Configuration.convert_source_file_string (text bound_value)
;;
*)

let natural bound_value =
  let non_negative, text_to_parse =
    match bound_value with
    | Ddv text -> false, text
    | Ddv_classified (Int32, text) -> false, text
    | Ddv_classified (Uint32, text) -> true, text
    | Ddv_classified
        ((Char | Logical | Double) as parameter_class,
         text) ->
      fatal_error
        (Format.fprintf Format.str_formatter
           "trying to get a natural value from value %S classified as %a"
           text
           Xml_explicit_ast_print.print_parameter_class parameter_class;
         Format.flush_str_formatter ()) in
  match Mdl_entry_parsers.simple_value_parser text_to_parse with
  | Int s ->
    let i = int_of_string s in
    begin
      try
        Unatural.of_int i
      with
      | Failure _ as x ->
        if non_negative then
          fatal_error
            "negative int %d was found for a value classified as UInt32"
        else
          raise x
    end
  | Bool _
  | Float _
  | String _
  | Ident _ as simple_value ->
    fatal_error
      (Format.fprintf Format.str_formatter
         "a natural was expected but %a was read"
         Mdl_ast_print.print_simple_value simple_value;
       Format.flush_str_formatter ())
;;

let port_number bound_value =
  Ast_types.mdl_port_number_of_natural (natural bound_value)
;;

let position bound_value =
  let text_to_parse = unclassified_text_to_parse bound_value in
  Mdl_entry_parsers.position_parser text_to_parse
;;

let block_ports_description bound_value =
  let text_to_parse = unclassified_text_to_parse bound_value in
  Mdl_entry_parsers.natural_vector_parser text_to_parse
;;

let block_source_type document_value =
  mdl_name_of_string (unclassified_text_to_parse document_value)
;;

let block_source_path document_value =
  mdl_path_of_string (unclassified_text_to_parse document_value)
;;

let zorder = unclassified_text_to_parse;;

let mirror = mdl_bool;;

let name_placement = text;;

let orientation = text;;

let color = unclassified_text_to_parse;;

let font_name = unclassified_text_to_parse;;

let mdl_name dv = mdl_name_of_string (text dv);;

let mdl_text dv = mdl_text_of_string (unclassified_text_to_parse dv);;

let matstruct_type = mdl_name;;
let matstruct_name = mdl_name;;

let object_type = mdl_name;;

let object_name explicit_document_value =
  let mdl_name = mdl_name explicit_document_value in
  fdebug
    (fun epf ->
     Format.fprintf epf "object_name: %S"
       (Ast_types.string_of_mdl_name mdl_name));
  mdl_name
;;

let port_name = mdl_name;;

let line_name = mdl_name;;

let line_type bound_value =
  let text_to_parse = unclassified_text_to_parse bound_value in
  Mdl_entry_parsers.line_type_parser text_to_parse
;;

let integer_matrix bound_value =
  let text_to_parse = unclassified_text_to_parse bound_value in
  Mdl_entry_parsers.integer_matrix_parser text_to_parse
;;

let labels = integer_matrix;;

let horizontal_alignment bound_value =
  let text_to_parse = unclassified_text_to_parse bound_value in
  Mdl_entry_parsers.horizontal_alignment_parser text_to_parse
;;

let vertical_alignment bound_value =
  let text_to_parse = unclassified_text_to_parse bound_value in
  Mdl_entry_parsers.vertical_alignment_parser text_to_parse
;;

let points = integer_matrix;;

let dimension (i, j as dimension) =
  let dimension =
    if i = Unatural._1_ then j else
    if j = Unatural._1_ then i else
      fatal_error
        (Format.fprintf Format.str_formatter
           "unable to convert new dimension %a to old dimension"
           Xml_explicit_ast_print.print_explicit_dimension dimension;
         Format.flush_str_formatter ()) in
  Int (Unatural.to_string dimension)
;;

let font_item = function
  | {
      Binding.bound_ident = bound_ident;
      Binding.bound_value = bound_value;
    } ->
    match bound_ident with
    | Ident.Ident "FontName" ->
      Some (Font_name (font_name bound_value))
    | Ident.Ident "FontSize" ->
      Some (Font_size (natural bound_value))
    | Ident.Ident "FontWeight" ->
      Some (Font_weight (text bound_value))
    | Ident.Ident "FontAngle" ->
      Some (Font_angle (text bound_value))
    | Ident.Ident _ -> None
;;

let font_and_color = function
  | {
      Binding.bound_ident = bound_ident;
      Binding.bound_value = bound_value;
    } as explicit_binding ->
    match Ast_types.string_of_ident bound_ident with
    | "BackgroundColor" ->
      Some (Background (color bound_value))
    | "DropShadow" ->
      Some (Drop_shadow (mdl_bool bound_value))
    | "ForegroundColor" ->
      Some (Foreground (color bound_value))
    | _ ->
      match font_item explicit_binding with
      | None -> None
      | Some font_item -> Some (Font_item font_item)
;;

let block_parameter_default = function
  | {
      explicit_block_parameter_default_type = ty;
      explicit_block_parameter_default_bindings = explicit_bindings;
    } ->
    let parameter_defaults =
      Xml_translating_basics.bindings Mode_property explicit_bindings in
    {
      block_type = ty;
      parameter_defaults = parameter_defaults;
    }
;;

let annotation_graphical_default = function
  | {
      bound_ident = bound_ident;
      bound_value = bound_value;
    } as explicit_binding ->
    match string_of_ident bound_ident with
    | "HorizontalAlignment" ->
      Some (Horizontal_alignment (horizontal_alignment bound_value))
    | "VerticalAlignment" ->
      Some (Vertical_alignment (vertical_alignment bound_value))
    | "UseDisplayTextAsClickCallback" ->
      Some (Text_as_click_cb (mdl_bool bound_value))
    (* FIXME: How to get it?
       | _ -> Click_fcn
       | _ -> Tex_mode
       | _ -> Interpreter *)
    | _ -> match font_and_color explicit_binding with
      | None -> None
      | Some font_and_color -> Some (Font_and_color font_and_color)
;;

let line_graphical_default = function
{
  Binding.bound_ident = bound_ident;
  Binding.bound_value = bound_value;
} as explicit_binding ->
  match string_of_ident bound_ident with
  | "ZOrder" ->
    Some (Line_zorder_item (zorder bound_value))
  | _ ->
    match font_item explicit_binding with
    | None -> None
    | Some font_item ->
      Some (Line_font_item font_item)
;;

let mask_graphical_default = binding Mode_property;;

let mask_parameter_default = binding Mode_property;;

let connection_type bound_value =
  let text_to_parse = unclassified_text_to_parse bound_value in
  Mdl_entry_parsers.connection_type_parser text_to_parse
;;

let line_source, line_target =
  let line_connector get_block_name bound_value =
    let text_to_parse = unclassified_text_to_parse bound_value in
    fdebug
      (fun ppf ->
       Format.fprintf ppf
         "trying to parse slx port description %S"
         text_to_parse);
    let block_sid, port_value =
      Mdl_entry_parsers.slx_port_connector_parser text_to_parse in
    let block_name = get_block_name block_sid in
    {
      line_connector_block = block_name;
      line_connector_port = port_value;
    } in
  let line_source get_block_name bound_value =
    Line_source_connector
      (line_connector get_block_name bound_value)
  and line_target get_block_name bound_value =
    Line_target_connector
      (line_connector get_block_name bound_value) in
  line_source, line_target
;;

let rec line_target_branch get_block_name explicit_branch =
  let branch_items =
    line_target_branch_items
      get_block_name explicit_branch.explicit_branch_items in
  let branch_items =
    match explicit_branch.explicit_branch_connection_type with
    | None -> branch_items
    | Some connection_type ->
      Line_target_branch_connection_type connection_type :: branch_items in
  Line_target_branch branch_items

and line_target_branch_items get_block_name explicit_branch_items =
  List.map (line_target_branch_item get_block_name) explicit_branch_items

and line_target_branch_item get_block_name = function
  | Explicit_branch_binding explicit_binding ->
    line_target_branch_item_of_explicit_binding get_block_name explicit_binding
  | Explicit_branch_branch explicit_branch ->
    Line_target_branch_target
      (line_target_branch get_block_name explicit_branch)

and line_target_branch_item_of_explicit_binding get_block_name = function
  | {
      Binding.bound_ident = bound_ident;
      bound_value = bound_value;
    } ->
    match string_of_ident bound_ident with
    | "Type" ->
      Line_target_branch_connection_type (connection_type bound_value)
    | "Labels" ->
      Line_target_branch_labels (labels bound_value)
    | "Points" ->
      Line_target_branch_points (points bound_value)
    (* Can a branch have a specified Src? *)
    | "Src" ->
      Line_target_branch_source (line_source get_block_name bound_value)
    | "Dst" ->
      Line_target_branch_target (line_target get_block_name bound_value)
    | "ZOrder" ->
      Line_target_branch_zorder (zorder bound_value)
    | bound_string ->
      fatal_error (Printf.sprintf
        "unknown %S binding inside a branch" bound_string)
;;

let line_item_of_explicit_binding get_block_name = function
  | {
      Binding.bound_ident = bound_ident;
      bound_value = bound_value;
    } as explicit_binding ->
    match string_of_ident bound_ident with
    | "Name" ->
      Line_name (line_name bound_value)
    | "Src" ->
      Line_source (line_source get_block_name bound_value)
    | "Dst" ->
      Line_target (line_target get_block_name bound_value)
    | "Type" ->
      Line_type (line_type bound_value)
    | "Labels" ->
      Line_labels (labels bound_value)
    | "Points" ->
      Line_points (points bound_value)
    | s ->
      match line_graphical_default explicit_binding with
      | None ->
        fatal_error (Printf.sprintf
          "unknown binding %S for line item" s)
      | Some line_graphical_default ->
        Line_graphical_default line_graphical_default
;;

let line_item get_block_name = function
  | Explicit_line_binding explicit_binding ->
    line_item_of_explicit_binding get_block_name explicit_binding
  | Explicit_line_branch explicit_branch ->
    Line_target (line_target_branch get_block_name explicit_branch)
;;

let line get_block_name = function
  | {
      explicit_line_type = explicit_line_type;
      explicit_line_items = explicit_line_items;
    } ->
    let line_items = List.map (line_item get_block_name) explicit_line_items in
    let line_items =
      match explicit_line_type with
      | None -> line_items
      | Some explicit_line_type ->
        Line_type explicit_line_type :: line_items in
    Line line_items
;;

let lines get_block_name = List.map (line get_block_name);;

let annotation_item = function
  | {
      bound_ident = bound_ident;
      bound_value = bound_value;
    } as explicit_binding ->
    match string_of_ident bound_ident with
    | "Name" -> Annotation_name (mdl_text bound_value)
(*
    | "Tag" -> Annotation_tag (wstring bound_value)
*)
    | "Position" -> Annotation_position (position bound_value)
    | "AutoSize" -> Annotation_auto_size (mdl_bool bound_value)
    | "WordWrap" -> Annotation_word_wrap (mdl_bool bound_value)
    | "ZOrder" -> Annotation_zorder (zorder bound_value)
    | _ ->
      match annotation_graphical_default explicit_binding with
      | None -> Annotation_binding (binding Mode_property explicit_binding)
      | Some annotation_graphical_default ->
        Annotation_graphical_default annotation_graphical_default
;;

let annotation = function
  | {
    explicit_annotation_sid = annotation_sid;
    explicit_annotation_bindings = explicit_bindings;
  } ->
    let annotation_sid_item = Annotation_sid annotation_sid in
    let annotation_items = List.map annotation_item explicit_bindings in
    Annotation (annotation_sid_item :: annotation_items)
;;

let annotations = List.map annotation;;

let system_connector explicit_system_connector =
  bindings Mode_ignore explicit_system_connector
;;
let system_connectors = List.map system_connector;;

let port_binding explicit_binding =
(* TODO: there are vendor
   | Block_port_binding_array of array_description
   | Block_port_binding_vendor of vendor_binding
   but we do not know where to retrieve them *)
  Block_port_binding (binding Mode_property explicit_binding)
;;

let binding_of_explicit_list = function
{
  explicit_list_type = explicit_list_type;
  explicit_list_bindings = explicit_bindings;
} ->
  (* TODO: is it really the bound_ident? *)
  let bindings = bindings Mode_ignore explicit_bindings in
  let bound_value = Value_bindings bindings in
  {
    bound_ident = Ast_types.ident_from_string explicit_list_type;
    bound_value = bound_value;
  }
;;

let object_item = function
  | {
      Binding.
      bound_ident = bound_ident;
      bound_value = bound_value;
    } as explicit_binding ->
  match string_of_ident bound_ident with
  | "Type" ->
    Object_type (object_type bound_value)
  | "Name" ->
    fdebug
      (fun epf ->
       Format.fprintf epf "xml_explicit_ast_to_mdl_ast.object_item:@ %a"
         Xml_explicit_ast_print.print_explicit_document_value bound_value);
    Object_name (object_name bound_value)
  (* FIXME: Not sure about the binding
  | "Vendor" ->
     Object_vendor_binding (vendor_item bound_value) *)
  | _ -> Object_binding (binding Mode_property explicit_binding)
;;

let vendor_and_config_name =
  Mdl_entry_parsers.vendor_configuration_longident_parser
;;

let matstruct_item = function
  | Explicit_matstruct_field ({
      Binding.
      bound_ident = explicit_bound_ident;
      bound_value = explicit_bound_value;
    } as explicit_binding) ->
    begin
      match string_of_ident explicit_bound_ident with
      | "Type" ->
        Some (Matstruct_type (matstruct_type explicit_bound_value))
      | "Name" ->
        Some (Matstruct_name (matstruct_name explicit_bound_value))
      | _ ->
        let binding = binding Mode_property explicit_binding in
        Some (Matstruct_binding binding)
    end
  | Explicit_matstruct_array _array ->
    not_yet "matstruct_item: Explicit_matstruct_array";
    None
;;

let custom_property_value custom_property_type = function
  | {
      Binding.
      bound_ident = explicit_bound_ident;
      bound_value = explicit_bound_value;
    } ->
    match string_of_ident explicit_bound_ident with
    | "Value" ->
      simple_value_of_custom_property_value
        custom_property_type explicit_bound_value
    | pname ->
      fatal_error (Printf.sprintf
        "unknown keyword %s to introduce property value definition" pname)
;;

let enum_of_binding property_name = function
  | {
    Binding.
    bound_ident = bound_ident;
    bound_value = bound_value;
  } ->
    match string_of_ident bound_ident with
    | "Value" ->
      let text = unclassified_text_to_parse bound_value in
      let enum =
        try
          Unatural.of_string text
        with
        | Failure _ ->
          fatal_error
            (Printf.sprintf
               "invalide value in EnumStrPairs for custom property %s"
               property_name) in
      enum
    | pname ->
      fatal_error
        (Printf.sprintf
           "unknown keyword %s to introduce EnumStrPairs value \
            definition for custom property %s" pname property_name)
;;

let mdl_text_of_binding property_name = function
  | {
    Binding.
    bound_ident = bound_ident;
    bound_value = bound_value;
  } ->
    match string_of_ident bound_ident with
    | "String" -> mdl_text bound_value
    | pname ->
      fatal_error
        (Printf.sprintf
           "unknown keyword %s to introduce EnumStrPairs string \
            definition for custom property %s" pname property_name)
;;

let enum_str_pairs_of_bindings property_name bindings =
  (* EnumStrPairs is an alternance of Value and String. Example:
     <EnumStrPairs>
       <P Name="Value">0</P>
       <P Name="String">Ignore</P>
       <P Name="Value">1</P>
       <P Name="String">Stop model with an error</P>
       <P Name="Value">2</P>
       <P Name="String">Print assertion</P>
     </EnumStrPairs>
   *)
  let rec read_pairs accu = function
    | [] -> List.rev accu
    | [ _ ] ->
      fatal_error
        (Printf.sprintf
           "dandling binding in EnumStrPairs for custom property %s"
           property_name)
    | value_binding :: string_binding :: bindings ->
      let accu =
        let enum = enum_of_binding property_name value_binding in
        let text = mdl_text_of_binding property_name string_binding in
        (enum, text) :: accu in
      read_pairs accu bindings in
  read_pairs [] bindings
;;

let enum_str_pairs_of_custom_property_items
      property_name property_data_type custom_property_items =
  let rec loop value_opt enum_str_pairs_opt = function
    | [] -> value_opt, enum_str_pairs_opt
    | item :: items ->
      let value_opt, enum_str_pairs_opt =
        match item with
        | Explicit_custom_property_item_binding binding ->
          check_opt_is_empty
            (Printf.sprintf
               "Value defined twice for custom property %S"
               property_name)
            value_opt;
          let value =
            custom_property_value property_data_type binding in
          Some value, enum_str_pairs_opt
        | Explicit_custom_property_item_enum_str_pairs bindings ->
          check_opt_is_empty
            (Printf.sprintf
               "EnumStrPairs defined twice for custom property %S"
               property_name)
            enum_str_pairs_opt;
          let enum_str_pairs =
            enum_str_pairs_of_bindings property_name bindings in
          value_opt, Some enum_str_pairs in
      loop value_opt enum_str_pairs_opt items in
  let value_opt, enum_str_pairs_opt =
    loop None None custom_property_items in
  match value_opt, enum_str_pairs_opt with
  | None, _
  | _, None ->
    fatal_error
      (Printf.sprintf
         "incomplete EnumStrPairs defintion for custom property %S"
         property_name)
  | Some value, Some enum_str_pairs ->
    value, enum_str_pairs
;;

let custom_property = function
  | {
    explicit_custom_property_name = custom_property_name;
    explicit_custom_property_data_type = custom_property_data_type;
    explicit_custom_property_items = custom_property_items;
  } ->
    let value, enum_str_pairs =
      match custom_property_items with
      | [ custom_property_item; ] ->
        let value =
          match custom_property_item with
          | Explicit_custom_property_item_binding binding ->
            custom_property_value custom_property_data_type binding
          | Explicit_custom_property_item_enum_str_pairs _ ->
            fatal_error
              (Printf.sprintf
                 "standalone EnumStrPairs for custom property %S"
                 (Ast_types.string_of_ident custom_property_name)) in
        value, []
      | custom_property_items ->
        let property_name =
          Ast_types.string_of_ident custom_property_name in
        enum_str_pairs_of_custom_property_items
          property_name custom_property_data_type
          custom_property_items in
    {
      custom_property_type = custom_property_data_type;
      custom_property_enum_str_pairs = enum_str_pairs;
      custom_property_ident = custom_property_name;
      custom_property_value = value;
    }
;;

let list_map = List.map;;

let list_map_opt f list =
  let rec loop accu = function
    | [] -> List.rev accu
    | x :: xs ->
      match f x with
      | None -> loop accu xs
      | Some y -> loop (y :: accu) xs in
  loop [] list
;;

let rec object_description = function
  | {
      explicit_object_id = object_id;
      explicit_object_class_name = object_class_name_opt;
      explicit_object_prop_name = prop_name_opt;
      explicit_object_backup_class = backup_class_name_opt;
      explicit_object_version = explicit_version_opt;
      explicit_object_reference = explicit_reference_opt;
      explicit_object_items = explicit_object_items;
    } ->
    let object_items = list_map object_object explicit_object_items in
    begin
      match
        backup_class_name_opt,
        explicit_version_opt,
        explicit_reference_opt with
      | None, None, None -> ()
      | _ , _, _ ->
        warning (Printf.sprintf
                   "fields backup_class, version, and reference omitted")
    end;
    let object_items =
      Object_dollar_binding (Dollar_object_id object_id) :: object_items in
    let object_items =
      match object_class_name_opt with
      | None -> object_items
      | Some object_class_name ->
        Object_dollar_binding (Dollar_class_name object_class_name) ::
          object_items in
    let object_items =
      match prop_name_opt with
      | None -> object_items
      | Some prop_name ->
        Object_dollar_binding (Dollar_prop_name prop_name) :: object_items in
    object_items

and object_object = function
  | Explicit_object_binding explicit_binding ->
    object_item explicit_binding
  | Explicit_object_array explicit_array ->
    Object_array (array_description explicit_array)
  | Explicit_object_object explicit_object ->
    Object_object (object_description explicit_object)
  | Explicit_object_custom_property explicit_custom_property ->
    Object_custom_property (custom_property explicit_custom_property)

and array_description = function
  | {
      explicit_array_type = array_type;
      explicit_array_dimension = explicit_dimension;
      explicit_array_prop_name = prop_name;
      explicit_array_items = explicit_array_items;
    } ->
    let dimension = dimension explicit_dimension in
    let array_items = list_map array_item explicit_array_items in
    {
      array_type = array_type;
      array_dimension = dimension;
      array_prop_name = prop_name;
      array_items = array_items;
    }

and array_item = function
  | Explicit_array_item_object explicit_object ->
    begin
      match try_vendor_and_config_name explicit_object with
      | Vendor_config_object_binding {
          vendor_ident = vendor_ident;
          vendor_configuration_longident = configuration_longident;
          vendor_items = vendor_items;
        }
        when configuration_longident =
             Mdl_ast_functions.config_set_longident ->
        Array_vendor_configset (vendor_ident, vendor_items)
      | Vendor_config_object_binding vendor_binding ->
        Array_vendor_binding vendor_binding
      | Vendor_config_object_object object_description ->
        Array_object object_description
    end
  (*
      (fun vendor_binding -> )
      (fun object_description -> Array_object object_description)
      (Some explicit_object)
   begin
      match explicit_object.explicit_object_class_name with
      | None ->
        let object_description = object_description explicit_object in
        Array_object object_description
      | Some (Mdl_name class_string) ->
        begin
          try
            let vendor_ident, configuration_ident =
              vendor_and_config_name class_string in
            let vendor_items =
              List.map vendor_item explicit_object.explicit_object_items in
            let vendor_items =
              match explicit_object.explicit_object_prop_name with
              | None -> vendor_items
              | Some prop_name ->
                Vendor_dollar_binding (Dollar_prop_name prop_name) ::
                  vendor_items in
            let vendor_items =
              match explicit_object.explicit_object_backup_class with
              | None -> vendor_items
              | Some backup_class ->
                Vendor_dollar_binding (Dollar_backup_class backup_class) ::
                  vendor_items in
            let vendor_items =
              match explicit_object.explicit_object_class_name with
              | None -> vendor_items
              | Some class_name ->
                Vendor_dollar_binding (Dollar_class_name class_name) ::
                  vendor_items in
            let vendor_items =
              Vendor_dollar_binding
                (Dollar_object_id explicit_object.explicit_object_id) ::
                vendor_items in
            match Ast_types.string_of_ident configuration_ident with
            | "ConfigSet" ->
              Array_vendor_configset (vendor_ident, vendor_items)
            | _ ->
              let vendor_binding =
                {
                  vendor_ident = vendor_ident;
                  vendor_configuration_ident = configuration_ident;
                  vendor_items = vendor_items;
                } in
              Array_vendor_binding vendor_binding
          with
          | Mdl_parse_error.Error _ ->
            let object_description = object_description explicit_object in
            Array_object object_description
        end
    end
  *)
  | Explicit_array_item_cell
      {
        explicit_cell_class = explicit_cell_class_opt;
        explicit_cell_value = text;
      } ->
    let v = model_value_of_optional_classified explicit_cell_class_opt text in
    let binding =
      {
        bound_ident = Ast_types.ident_from_string "Cell";
        bound_value = v;
      } in
    Array_binding binding
  | Explicit_array_matstruct explicit_matstruct_fields ->
    let matstruct_items =
      list_map_opt matstruct_item explicit_matstruct_fields in
    Array_matstruct matstruct_items
  | Explicit_array_array explicit_array ->
    Array_sub_array (array_description explicit_array)

and vendor_item = function
  | Explicit_object_binding explicit_binding ->
    begin
      match explicit_binding with
      | {
          bound_ident = Ident.Ident "Name";
          bound_value = bound_value;
        } -> Vendor_binding (Object_name (object_name bound_value))
      | {
          bound_ident = Ident.Ident _ident;
          bound_value = _bound_value;
        } ->
        Vendor_binding
          (Object_binding (binding Mode_property explicit_binding))
    end
  | Explicit_object_array explicit_array ->
    Vendor_binding (Object_array (array_description explicit_array))
  | Explicit_object_object explicit_object ->
    let object_description = object_description explicit_object in
    Vendor_binding (Object_object object_description)
  | Explicit_object_custom_property explicit_custom_property ->
    Vendor_binding
      (Object_custom_property
         (custom_property explicit_custom_property))

and try_vendor_and_config_name explicit_object =
  match explicit_object.explicit_object_class_name with
  | None ->
    let object_description = object_description explicit_object in
    Vendor_config_object_object object_description
  | Some (Mdl_name class_string) ->
    try
      let vendor_ident, configuration_longident =
        vendor_and_config_name class_string in
      let vendor_items =
        list_map vendor_item explicit_object.explicit_object_items in
      let vendor_items =
        match explicit_object.explicit_object_prop_name with
        | None -> vendor_items
        | Some prop_name ->
          Vendor_dollar_binding (Dollar_prop_name prop_name) ::
            vendor_items in
      let vendor_items =
        match
          explicit_object.explicit_object_backup_class
        with
        | None -> vendor_items
        | Some backup_class ->
          Vendor_dollar_binding (Dollar_backup_class backup_class) ::
            vendor_items in
      let vendor_items =
        match explicit_object.explicit_object_class_name with
        | None -> vendor_items
        | Some class_name ->
          Vendor_dollar_binding (Dollar_class_name class_name) ::
            vendor_items in
      let vendor_items =
        Vendor_dollar_binding
          (Dollar_object_id
             explicit_object.explicit_object_id) ::
          vendor_items in
      let vendor_binding =
        {
          vendor_ident = vendor_ident;
          vendor_configuration_longident = configuration_longident;
          vendor_items = vendor_items;
        } in
      Vendor_config_object_binding vendor_binding
    with
    | Mdl_parse_error.Error _ ->
      let object_description = object_description explicit_object in
      Vendor_config_object_object object_description
    | x ->
      prerr_endline (Printf.sprintf "ERROR : %s" (Printexc.to_string x));
      raise x
;;

let block_port explicit_port =

  let classify_explicit_port_item (num_opt, nam_opt, bindings) = function
    | Explicit_port_array explicit_array ->
      let binding = Block_port_binding_array (array_description explicit_array) in
      let bindings = binding :: bindings in
      num_opt, nam_opt, bindings
    | Explicit_port_binding
        ({
          bound_ident = bound_ident;
          bound_value = bound_value;
        } as explicit_binding) ->
      match string_of_ident bound_ident with
      | "PortNumber" ->
        check_opt_is_empty "PortNumber for BlockPort is defined twice" num_opt;
        let num_opt = Some (port_number bound_value) in
        num_opt, nam_opt, bindings
      | "Name" ->
        check_opt_is_empty "Name for BlockPort is defined twice" nam_opt;
        let nam_opt = Some (port_name bound_value) in
        num_opt, nam_opt, bindings
      | _ ->
        let bindings = port_binding explicit_binding :: bindings in
        num_opt, nam_opt, bindings in

  let num_opt, nam_opt, bindings =
    List.fold_left
      classify_explicit_port_item (None, None, []) explicit_port in

  let num =
    match num_opt with
    | None -> fatal_error "missing PortNumber for BlockPort"
    | Some num -> num in
  {
    block_port_number = num;
    block_port_name = nam_opt;
    block_port_bindings = bindings;
  }
;;

let graphical_default = function
  | {
      Binding.
      bound_ident = bound_ident;
      bound_value = bound_value;
    } as explicit_binding ->
    match string_of_ident bound_ident with
    | "BlockMirror" -> Some (Block_mirror (mirror bound_value))
    | "NamePlacement" ->
      Some (Block_name_placement (name_placement bound_value))
    | "Orientation" -> Some (Block_orientation (orientation bound_value))
    | "BlockRotation" -> Some (Block_rotation (natural bound_value))
    | "ShowName" -> Some (Block_show_name (mdl_bool bound_value))
    | "HideAutomaticName" ->
      Some (Block_hide_automatic_name (mdl_bool bound_value))
    | "ZOrder" -> Some (Block_zorder (zorder bound_value))
    | bound_ident ->
      match font_and_color explicit_binding with
      | None ->
        warning (Printf.sprintf
          "graphical binding for ident %s was dropped" bound_ident);
        None
      | Some font_and_color -> Some (Block_font_and_color font_and_color)
;;

let mdl_comma_bools bound_value = Comma_bools (mdl_text bound_value)
and mdl_comma_texts bound_value = Comma_texts (mdl_text bound_value)
and mdl_pipe_expressions bound_value = Pipe_expressions (mdl_text bound_value)
and mdl_pipe_texts bound_value = Pipe_texts (mdl_text bound_value)
and mdl_semi_texts bound_value = Semi_texts (mdl_text bound_value)
and ident_of_bound_value bound_value =
  Ast_types.ident_from_string (text bound_value)
;;

let block_mask_item = function
  | {
      Binding.
      bound_ident = bound_ident;
      bound_value = bound_value;
    } ->
    match string_of_ident bound_ident with
    | "MaskHideContents" ->
      Some (Mask_hide_contents (mdl_bool bound_value))
    | "MaskType" ->
      Some (Mask_type (mdl_text bound_value))
    | "MaskDescription" ->
      Some (Mask_description (mdl_text bound_value))
    | "MaskPromptString" ->
      Some (Mask_prompt_string (mdl_pipe_texts bound_value))
    | "MaskStyleString" ->
      Some (Mask_style_string (mdl_comma_texts bound_value))
    | "MaskVariables" ->
      Some (Mask_variables (mdl_semi_texts bound_value))
    | "MaskTunableValueString" ->
      Some (Mask_tunable_value_string (mdl_comma_bools bound_value))
    | "MaskCallbackString" ->
      Some (Mask_callback_string (mdl_pipe_expressions bound_value))
    | "MaskEnableString" ->
      Some (Mask_enable_string (mdl_comma_bools bound_value))
    | "MaskVisibilityString" ->
      Some (Mask_visibility_string (mdl_comma_bools bound_value))
    | "MaskToolTipString" ->
      Some (Mask_tooltip_string (mdl_comma_bools bound_value))
    | "MaskIconFrame" ->
      Some (Mask_icon_frame (mdl_bool bound_value))
    | "MaskIconOpaque" ->
      Some (Mask_icon_opaque (mdl_bool bound_value))
    | "MaskIconRotate" ->
      Some (Mask_icon_rotate (ident_of_bound_value bound_value))
    | "MaskPortRotate" ->
      Some (Mask_port_rotate (ident_of_bound_value bound_value))
    | "MaskIconUnits" ->
      Some (Mask_icon_units (ident_of_bound_value bound_value))
    | "MaskValueString" ->
      Some (Mask_value_string (mdl_pipe_texts bound_value))
    | _ -> None
;;

let bindings_of_dialog_parameters = function
  | {
    explicit_dialog_parameters_block_name = explicit_block_name;
    explicit_dialog_parameters_bindings = explicit_bindings;
  } ->
    let sub_bindings = bindings Mode_ignore explicit_bindings in
    let sub_bound_value = Value_bindings sub_bindings in
    let dialog_parameters_binding = {
      bound_ident = Ast_types.ident_from_string "DialogParameters";
      bound_value = sub_bound_value;
    } in
    let block_name_binding = {
      bound_ident = Ast_types.ident_from_string "BlockName";
      bound_value =
        Xml_translating_basics.model_value_of_text
          (Ast_types.string_of_mdl_name explicit_block_name);
    } in
    block_name_binding, dialog_parameters_binding
;;

let binding_of_link_data link_data =
  let dialog_parameters_bindings =
    List.map bindings_of_dialog_parameters link_data in
  let bindings =
    let rec loop bindings = function
      | [] -> List.rev bindings
      | (block_name_binding, dialog_parameters_binding)
        :: dialog_parameters_bindings ->
        let bindings =
          dialog_parameters_binding :: block_name_binding :: bindings in
        loop bindings dialog_parameters_bindings in
    loop [] dialog_parameters_bindings in
  let bound_value = Value_bindings bindings in
  {
    bound_ident = Ast_types.ident_from_string "LinkData";
    bound_value = bound_value;
  }
;;

(* TODO: we are doing the opposite of xml_explicit, because mdl_ast is
   less typed than xml_explicit *)
let mk_binding_of_text ident_str txt =
  {
    bound_ident = Ast_types.ident_from_string ident_str;
    bound_value = model_value_of_text txt;
  }
;;

let binding_of_icon_frame = mk_binding_of_text "IconFrame"
and binding_of_icon_units = mk_binding_of_text "IconUnits"
and binding_of_icon_opaque = mk_binding_of_text "IconOpaque"
and binding_of_icon_rotate = mk_binding_of_text "IconRotate"
and binding_of_run_init_for_icon_redraw =
  mk_binding_of_text "RunInitForIconRedraw"
and binding_of_explicit_mask_help = mk_binding_of_text "Help"
;;

let mask_display = function
  | {
      explicit_mask_display_icon_frame = icon_frame;
      explicit_mask_display_icon_units = icon_units;
      explicit_mask_display_icon_opaque = icon_opaque;
      explicit_mask_display_icon_rotate = icon_rotate;
      explicit_mask_display_run_init_for_icon_redraw =
        run_init_for_icon_redraw;
      explicit_mask_display_script = script;
    } ->
    let mask_display_source = Uoption.map mdl_code_of_string script in
    let mask_display_bindings = [] in
    let mask_display_bindings =
      Uoption.cons_map
        binding_of_icon_frame icon_frame mask_display_bindings in
    let mask_display_bindings =
      Uoption.cons_map
        binding_of_icon_units icon_units mask_display_bindings in
    let mask_display_bindings =
      Uoption.cons_map
        binding_of_icon_opaque icon_opaque mask_display_bindings in
    let mask_display_bindings =
      Uoption.cons_map
        binding_of_icon_rotate icon_rotate mask_display_bindings in
    let mask_display_bindings =
      Uoption.cons_map
        binding_of_run_init_for_icon_redraw
        run_init_for_icon_redraw
        mask_display_bindings in
    {
      mask_display_source = mask_display_source;
      mask_display_bindings = mask_display_bindings;
    }
;;

let mask_parameter = function
  | {
      explicit_mask_parameter_name = mask_parameter_name;
      explicit_mask_parameter_type = mask_parameter_type;
      explicit_mask_parameter_prompt = mask_parameter_prompt;
      explicit_mask_parameter_callback = mask_parameter_callback;
      explicit_mask_parameter_tunable = mask_parameter_tunable;
      explicit_mask_parameter_value = mask_parameter_value;
      explicit_mask_parameter_evaluate = mask_parameter_evaluate;
      explicit_mask_parameter_enabled = mask_parameter_enabled;
      explicit_mask_parameter_visible = mask_parameter_visible;
      explicit_mask_parameter_internal = mask_parameter_internal;
      explicit_mask_parameter_never_save = mask_parameter_never_save;
      explicit_mask_parameter_read_only = mask_parameter_read_only;
    } ->
    let parameter_ident =
      Ast_types.ident_from_mdl_name mask_parameter_name in
    let parameter_style = mdl_text_of_string mask_parameter_type in
    let parameter_prompt = mdl_text_of_string mask_parameter_prompt in
    let parameter_callback =
      Uoption.map
        mdl_code_of_string mask_parameter_callback in
    let parameter_tunable =
      Uoption.map
        mdl_bool_of_string mask_parameter_tunable in
    let parameter_value = mdl_text_of_string mask_parameter_value in
    let parameter_evaluation =
      Uoption.map
        mdl_bool_of_string mask_parameter_evaluate in
    let parameter_enable =
      Uoption.map
        mdl_bool_of_string mask_parameter_enabled in
    let parameter_visible =
      Uoption.map
        mdl_bool_of_string mask_parameter_visible in
    let parameter_internal =
      Uoption.map
        mdl_bool_of_string mask_parameter_internal in
    let parameter_never_save =
      Uoption.map
        mdl_bool_of_string mask_parameter_never_save in
    let parameter_read_only =
      Uoption.map
        mdl_bool_of_string mask_parameter_read_only in
    {
      mask_parameter_ident = parameter_ident;
      mask_parameter_style = parameter_style;
      mask_parameter_prompt = parameter_prompt;
      mask_parameter_callback = parameter_callback;
      mask_parameter_tunable = parameter_tunable;
      mask_parameter_value = parameter_value;
      mask_parameter_evaluation = parameter_evaluation;
      mask_parameter_enable = parameter_enable;
      mask_parameter_visible = parameter_visible;
      mask_parameter_internal = parameter_internal;
      mask_parameter_never_save = parameter_never_save;
      mask_parameter_read_only = parameter_read_only;
    }
;;

let mask_description = function
  | {
      explicit_mask_type = explicit_mask_type_opt;
      explicit_mask_display = explicit_mask_display;
      explicit_mask_description = explicit_mask_description_opt;
      explicit_mask_initialisation = explicit_mask_initialisation_opt;
      explicit_mask_help = explicit_mask_help_opt;
      explicit_mask_parameters = explicit_mask_parameters;
    } ->
    let mask_type_opt =
      Uoption.map
        mdl_text_of_string explicit_mask_type_opt in
    let mask_display = mask_display explicit_mask_display in
    let mask_description_opt =
      Uoption.map mdl_text_of_string explicit_mask_description_opt in
    let mask_initialisation_opt =
      Uoption.map mdl_code_of_string explicit_mask_initialisation_opt in
    let mask_parameters =
      List.map mask_parameter explicit_mask_parameters in
    let mask_bindings =
      (* todo: add spurious attributes like Tunable in bindings *)
      let mask_bindings = [] in
      Uoption.cons_map
        binding_of_explicit_mask_help explicit_mask_help_opt
        mask_bindings in
    {
      mask_type = mask_type_opt;
      mask_display = mask_display;
      mask_description = mask_description_opt;
      mask_initialisation = mask_initialisation_opt;
      mask_parameters = mask_parameters;
      mask_bindings = mask_bindings;
    }
;;

let rec system get_block_name add_block_name system_name = function
  | {
      explicit_system_position = explicit_position;
      explicit_system_bindings = explicit_bindings;
      explicit_system_blocks = explicit_blocks;
      explicit_system_connections = explicit_lines;
      explicit_system_annotations = explicit_annotations;
      explicit_system_connectors = explicit_system_connectors;
    } ->
    let position = Mdl_entry_parsers.position_parser explicit_position in
    let bindings = bindings Mode_property explicit_bindings in
    let blocks = blocks get_block_name add_block_name explicit_blocks in
    let lines = lines get_block_name explicit_lines in
    let annotations = annotations explicit_annotations in
    let connectors = system_connectors explicit_system_connectors in
    {
      system_name = system_name;
      system_position = position;
      system_bindings = bindings;
      system_blocks = blocks;
      system_lines = lines;
      system_annotations = annotations;
      system_connectors = connectors;
    }

and block get_block_name add_block_name = function
  | {
      explicit_block_type = block_type;
      explicit_block_name = explicit_block_name;
      explicit_block_sid = sid;
      explicit_block_items = explicit_block_items;
    } ->
  add_block_name sid explicit_block_name;
  let block_type_item = Block_type block_type
  and block_name = Block_name explicit_block_name
  and block_sid = Block_sid sid in
  let block_items =
    block_items get_block_name add_block_name
      explicit_block_name explicit_block_items in
  fdebug
    (fun epf ->
     Format.fprintf epf "explicit xml to mdl ast: making block %S"
       (Ast_types.string_of_mdl_name explicit_block_name));
  Block (block_type_item :: block_name :: block_sid :: block_items)

and blocks get_block_name add_block_name blks =
  List.map (block get_block_name add_block_name) blks

and block_item_of_explicit_binding = function
  | {
      bound_ident = bound_ident;
      bound_value = bound_value;
    } as explicit_binding ->
    match string_of_ident bound_ident with
    | "SourceBlock" ->
      Block_source_path (block_source_path bound_value)
    | "SourceType" ->
      Block_source_type (block_source_type bound_value)
    | "Ports" ->
      Block_ports_description (block_ports_description bound_value)
    | "Position" ->
      Block_position (position bound_value)
    | "Tag" ->
      Block_tag (mdl_text bound_value)
    (* | _ -> Block_state _ *)
    (* | _ -> Block_array _ *)
    | _ ->
      match graphical_default explicit_binding with
      | None ->
        begin
          match block_mask_item explicit_binding with
          | None -> Block_binding (binding Mode_property explicit_binding)
          | Some block_mask_item -> Block_mask_item block_mask_item
        end
      | Some graphical_default -> Block_graphical_default (graphical_default)

and block_item get_block_name add_block_name block_name = function
  | Explicit_block_binding explicit_binding ->
    block_item_of_explicit_binding explicit_binding
  | Explicit_block_object explicit_object ->
    Block_object (object_description explicit_object)
  | Explicit_block_system explicit_system ->
    Block_system
      (system get_block_name add_block_name block_name explicit_system)
  | Explicit_block_list explicit_list ->
    Block_binding (binding_of_explicit_list explicit_list)
  | Explicit_block_port explicit_port ->
    Block_port (block_port explicit_port)
  | Explicit_block_array explicit_array ->
    Block_array (array_description explicit_array)
  | Explicit_block_instance_data instance_data ->
    Block_instance_data (bindings Mode_ignore instance_data)
  | Explicit_block_link_data link_data ->
    Block_binding (binding_of_link_data link_data)
  | Explicit_block_mask block_mask ->
    Block_mask (mask_description block_mask)
and block_items get_block_name add_block_name block_name items =
  List.map (block_item get_block_name add_block_name block_name) items
;;

let block_graphical_default explicit_binding =
  match graphical_default explicit_binding with
  | None ->
    fatal_error (Printf.sprintf
      "unknown binding %S for block graphical default"
      (string_of_ident explicit_binding.bound_ident))
  | Some graphical_default -> graphical_default
;;

let block_graphical_defaults = List.map block_graphical_default;;

let model_workspace_item explicit_binding =
  let bound_value = explicit_binding.bound_value in
  match string_of_ident explicit_binding.bound_ident with
  | "WSDataSource" ->
    Some (
      Workspace_data_source_type (
        workspace_data_source_type bound_value))
  | "WSMATLABCode" ->
    Some (Workspace_data_mcode (String (text bound_value)))
  | "WSSourceFileName" ->
     Some (Workspace_data_source_file_name (text bound_value))
  | bound_string ->
    not_yet (Printf.sprintf
      "model_workspace_item: %S %S" bound_string (text bound_value));
    None
;;

let model_workspace = function
  | None -> None
  | Some explicit_model_workspace_items ->
    Some (list_map_opt model_workspace_item explicit_model_workspace_items)
;;

(*
let configuration_set_parameter_of_explicit_object = function
  | None -> None
  | Some explicit_object ->
    match explicit_object.explicit_object_class_name with
    | None ->
      let object_description = object_description explicit_object in
      Some (Configuration_object object_description)
    | Some (Mdl_name class_string) ->
      try
        let vendor_ident, configuration_ident =
          vendor_and_config_name class_string in
        let vendor_items =
          List.map
            vendor_item explicit_object.explicit_object_items in
        let vendor_items =
          match explicit_object.explicit_object_prop_name with
          | None -> vendor_items
          | Some prop_name ->
            Vendor_dollar_binding (Dollar_prop_name prop_name) ::
              vendor_items in
        let vendor_items =
          match
            explicit_object.explicit_object_backup_class
          with
          | None -> vendor_items
          | Some backup_class ->
            Vendor_dollar_binding (Dollar_backup_class backup_class) ::
              vendor_items in
        let vendor_items =
          match explicit_object.explicit_object_class_name with
          | None -> vendor_items
          | Some class_name ->
            Vendor_dollar_binding (Dollar_class_name class_name) ::
              vendor_items in
        let vendor_items =
          Vendor_dollar_binding
            (Dollar_object_id
               explicit_object.explicit_object_id) ::
            vendor_items in
        let vendor_binding =
          {
            vendor_ident = vendor_ident;
            vendor_configuration_ident = configuration_ident;
            vendor_items = vendor_items;
          } in
        Some (Configuration_vendor_binding vendor_binding)
      with
      | Mdl_parse_error.Error _ ->
        let object_description = object_description explicit_object in
        Some (Configuration_object object_description)
;;
*)

let configuration_binding = function
  | {
      Binding.
      bound_ident = bound_ident;
      bound_value = bound_value;
    } ->
    let text = unclassified_text_to_parse bound_value in
    let simple_value = Mdl_ast.String text in
    Say.fdebug
      (fun epf ->
       Format.fprintf epf "configuration_binding: ident %s is bound to %a"
         (Ast_types.string_of_ident bound_ident)
         Mdl_ast_print.print_simple_value simple_value);
    Configuration_binding {
        Binding.
        bound_ident = bound_ident;
        bound_value = simple_value;
      }
;;

let model_section model_name model_version = function
  | {
      explicit_model_workspace = explicit_model_workspace;
      explicit_model_configuration = explicit_model_configuration;
      explicit_configuration_set = explicit_configuration_set;
      explicit_system_defaults = _explicit_system_defaults;
      explicit_block_defaults = explicit_block_defaults;
      explicit_annotation_defaults = explicit_annotation_defaults;
      explicit_line_defaults = explicit_line_defaults;
      explicit_mask_defaults = explicit_mask_defaults;
      explicit_mask_parameter_defaults = explicit_mask_parameter_defaults;
      explicit_block_parameter_defaults = explicit_block_parameter_defaults;
      explicit_system = explicit_system;
    } ->
    let model_ident = ident_from_mdl_name model_name in
    let model_workspace = model_workspace explicit_model_workspace in
    let model_configuration_set_parameters =
      match explicit_configuration_set with
      | {
          explicit_configuration_set_object =
            explicit_configuration_set_object;
          explicit_configuration_set_bindings =
            explicit_configuration_set_bindings;
        } ->
        let bindings =
          List.rev
            (List.rev_map
               configuration_binding explicit_configuration_set_bindings) in
        match explicit_configuration_set_object with
        | None -> bindings
        | Some explicit_object ->
          match try_vendor_and_config_name explicit_object with
          | Vendor_config_object_binding vendor_binding ->
            Configuration_vendor_binding vendor_binding :: bindings
          | Vendor_config_object_object object_description ->
            Configuration_object object_description :: bindings in
    let model_configurations =
      List.map array_description explicit_model_configuration in
    let block_graphical_defaults =
      block_graphical_defaults explicit_block_defaults in
    let block_parameter_defaults =
      List.map block_parameter_default explicit_block_parameter_defaults in
    let annotation_graphical_defaults =
      let annotation_graphical_default explicit_binding =
        match annotation_graphical_default explicit_binding with
        | None ->
          let binding = binding Mode_property explicit_binding in
          Annotation_graphical_binding binding
        | Some annotation_default -> annotation_default in
      List.map annotation_graphical_default explicit_annotation_defaults in
    let line_graphical_defaults =
      let line_graphical_default explicit_binding =
        match line_graphical_default explicit_binding with
        | None ->
          fatal_error (Printf.sprintf
            "unrecognized binding %S for line_default"
            (string_of_ident explicit_binding.bound_ident))
        | Some line_default -> line_default in
      List.map line_graphical_default explicit_line_defaults in
    let mask_graphical_defaults =
      List.map mask_graphical_default explicit_mask_defaults in
    let mask_parameter_defaults =
      List.map mask_parameter_default explicit_mask_parameter_defaults in
    let get_block_name, add_block_name = sid_to_block_names () in
    let system =
      system get_block_name add_block_name
        model_name explicit_system in
    {
      model_ident = model_ident;
      model_version = model_version;
      model_configuration_parameters = model_configuration_set_parameters;
      model_configurations = model_configurations;
      model_set_configuration_option =
        None (* FIXME: those data may be present in explicit_xml_ast *);
      model_workspace = model_workspace;
      block_graphical_defaults_section = block_graphical_defaults;
      block_parameter_defaults_section = block_parameter_defaults;
      annotation_graphical_defaults_section = annotation_graphical_defaults;
     line_graphical_defaults_section = line_graphical_defaults;
      mask_graphical_defaults_section = mask_graphical_defaults;
      mask_parameter_defaults_section = mask_parameter_defaults;
      model_system = system;
    }
;;

let stateflow_section = Xml_stateflow_to_mdl_stateflow.stateflow_section;;

let mdl_encoding_name_of_explicit_encoding = function
  | ASCII -> "US-ASCII"
  | Latin1 -> "ISO-8859-15"
  | Latin9 -> "ISO-8859-9"
  | UTF16 ->
    Say.fatal_error "slz file translation: utf16 encoding not supported"
  | UTF8 -> "UTF-8"
  | Encoding e -> e
;;

let fix_translated_file_encoding explicit_prolog =
  let explicit_encoding = explicit_prolog in
  let encoding_name =
    mdl_encoding_name_of_explicit_encoding explicit_encoding in
  Configuration.fix_source_encoding encoding_name;
;;

let model_version explicit_model_version =
  Mdl_ast.Slx_model_version explicit_model_version
;;

let implementation_section model_name = function
  | {
      explicit_prolog = explicit_prolog;
      explicit_model_sections = explicit_model_sections;
      explicit_model_version = explicit_model_version;
    } ->
  (*
     Must fix the encoding of the resulting mdl ast as the encoding
     of its XML file source origin, since all characters and strings
     in the explicit_xml_ast tree are encoded using the declared encoding of
     the XML file containing the block diagram.
     In particular the SAVED_CHARACTER_ENCODING binding of the resulting MDL
     ast should be ignored.
  *)
    let model_version = model_version explicit_model_version in
    let specific_model_section = model_section model_name model_version in
    fix_translated_file_encoding explicit_prolog;
    let rec loop accu = function
      | [] -> List.rev accu
      | explicit_implementation_section_item ::
          explicit_implementation_section_items ->
        let implementation_section_item =
          match explicit_implementation_section_item with
          | Explicit_implementation_library explicit_section ->
            Implementation_library (specific_model_section explicit_section)
       (* | Explicit_implementation_matdata of matdata_section *)
          | Explicit_implementation_model explicit_section ->
            Implementation_model (specific_model_section explicit_section)
          | Explicit_implementation_stateflow explicit_stateflow_section ->
            let state_flow_section_items =
              stateflow_section explicit_stateflow_section in
            Implementation_stateflow state_flow_section_items in
        loop
          (implementation_section_item :: accu)
          explicit_implementation_section_items in
    loop [] explicit_model_sections
;;

let explicit_xml_ast_to_mdl_ast = function
  | {
       explicit_model_ident = model_ident;
       explicit_blockdiagram = explicit_blockdiagram;
    } ->
    Say.fdebug
      (fun ppf ->
       Xml_explicit_ast_print.print_explicit_blockdiagram ppf
         explicit_blockdiagram);
    let model_name = mdl_name_of_ident model_ident in
    (* Must return a simulink_file_contents *)
    let implementation_section =
      implementation_section model_name explicit_blockdiagram in
    Implementation implementation_section
 (* Interface of interface_section
    there is no interface in ".slx", it's a simport addition to mdl files
    semantics. *)
;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
 End:
*)
