(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, Clément Franchini                    *)
(*                   INRIA Rocquencourt                                *)
(*                                                                     *)
(*  Copyright 2013-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: xml_stateflow_to_mdl_stateflow.ml,v 1.1 2019-05-10 14:00:17 jpc Exp $ *)

(*
open Binding;;
*)
open Ast_types;;
open Mdl_ast;;
open Xml_explicit_ast;;
open Xml_translating_basics;;

type children_classifier =
  {
    children_targets : stateflow_targets;
    children_charts : stateflow_charts;
    children_states : stateflow_states;
    children_transitions : stateflow_transitions;
    children_datas : stateflow_datas;
    children_junctions : stateflow_junctions;
    children_events : stateflow_events;
    children_items : stateflow_items;
  }
;;

type state_item_classifier =
  {
    stateflow_datas : stateflow_datas;
    stateflow_events : stateflow_events;
    stateflow_junctions : stateflow_junctions;
    stateflow_states : stateflow_states;
    stateflow_transitions : stateflow_transitions;
    stateflow_items : stateflow_items;
  }
;;

type machine_item_classifier =
  {
    explicit_stateflow_machine_charts : stateflow_charts;
    explicit_stateflow_machine_datas : stateflow_datas;
    explicit_stateflow_machine_events : stateflow_events;
    explicit_stateflow_machine_targets : stateflow_targets;
    explicit_stateflow_machine_items : stateflow_items;
  }
;;

type instance_items_classifier =
  {
    explicit_stateflow_instance_name_opt : mdl_name option;
    explicit_stateflow_instance_machine_ref_opt : stateflow_ref option;
    explicit_stateflow_instance_chart_ref_opt : stateflow_ref option;
  }
;;

let empty_instance_items_classifier =
  {
    explicit_stateflow_instance_name_opt = None;
    explicit_stateflow_instance_machine_ref_opt = None;
    explicit_stateflow_instance_chart_ref_opt = None;
  }
;;

let empty_children_classifier =
  {
    children_targets = [];
    children_charts = [];
    children_states = [];
    children_transitions = [];
    children_datas = [];
    children_junctions = [];
    children_events = [];
    children_items = [];
  }
;;

let empty_state_item_classifier as empty_chart_item_classifier =
  {
    stateflow_datas = [];
    stateflow_events = [];
    stateflow_junctions = [];
    stateflow_states = [];
    stateflow_transitions = [];
    stateflow_items = [];
  }
;;

let empty_machine_item_classifier =
  {
    explicit_stateflow_machine_charts = [];
    explicit_stateflow_machine_datas = [];
    explicit_stateflow_machine_events = [];
    explicit_stateflow_machine_targets = [];
    explicit_stateflow_machine_items = [];
  }
;;

(** {6 Miscellaneous functions.} *)

let module_name = "Xml_stateflow_to_mdl_stateflow";;

(*
let not_yet = Xml_translating_basics.not_yet module_name;;
*)

let fatal_error = Xml_translating_basics.fatal_error module_name;;

let warning = Xml_translating_basics.warning module_name;;

let fdebug = Xml_translating_basics.fdebug module_name;;

let binding_of_explicit_binding =
  Xml_translating_basics.binding Mode_property
;;

let unatural_of_string s =
  try
    Unatural.of_string s
  with
  | _ ->
    fatal_error (Printf.sprintf
      "%S is not a natural" s)
;;

let ssid_ident = Ast_types.ident_from_string "SSID"
and sub_view_ident = ident_from_string "subviewS"
and range_ident = ident_from_string "range"
and active_state_output_ident = ident_from_string "activeStateOutput"
and size_ident = Mdl_ast_functions.lsize_ident
and slide_ident = ident_from_string "slide"
and fixpt_ident = ident_from_string "fixpt"
and simulink_ident = ident_from_string "simulink"
and name_ident = Mdl_ast_functions.lname_ident
and chart_ident = Mdl_ast_functions.lchart_ident
and machine_ident = Mdl_ast_functions.lmachine_ident
;;

let ident_name_of_binding explicit_binding =
  Ast_types.string_of_ident (explicit_binding.Binding.bound_ident)
;;

let fatal_error_of_explicit_binding context explicit_binding =
  let bound_ident = ident_name_of_binding explicit_binding in
  fatal_error (Printf.sprintf "spurious binding for %s in %s"
    bound_ident context)
;;

let stateflow_reference_of_txt_value = Unatural.of_string;;

let explicit_binding_translator_table =
  Hash_table.of_bindings [
    (* Bound ident is "executionOrder" *)
    Mdl_ast_functions.lexecution_order_ident,
    (fun txt_value ->
     let n = unatural_of_string txt_value in
     Stateflow_execution_order n);
    (* Bound ident is "labelString" *)
    Mdl_ast_functions.llabel_string_ident,
    (fun txt_value ->
     let txt_value = Ast_types.mdl_text_of_string txt_value in
     Stateflow_label_string txt_value);
    (* Bound ident is "chartFileNumber" *)
    Mdl_ast_functions.lchart_file_number_ident,
    (fun txt_value ->
     let n = unatural_of_string txt_value in
     Stateflow_chart_file_number n);
    (* Bound ident is "name" *)
    Mdl_ast_functions.lname_ident,
    (fun txt_value ->
     let mdl_name = Ast_types.mdl_name_of_string txt_value in
     Stateflow_name mdl_name);
    (* Bound ident is "scope" *)
    Mdl_ast_functions.lscope_ident,
    (fun txt_value ->
     Stateflow_scope (Ast_types.ident_from_string txt_value));
    (* Bound ident is "ssIdNumber" *)
    Mdl_ast_functions.lss_id_number_ident,
    (fun txt_value ->
     let n = unatural_of_string txt_value in
     Stateflow_ssid_number n);
    (* Bound ident is "trigger" *)
    Mdl_ast_functions.ltrigger_ident,
    (fun txt_value ->
     let ident = Ast_types.ident_from_string txt_value in
     Stateflow_trigger ident);
    (* Bound ident is "type" *)
    Mdl_ast_functions.ltype_ident,
    (fun txt_value ->
     let ident = Ast_types.ident_from_string txt_value in
     Stateflow_type ident);
]
;;

(* Build a stateflow_binding from an ident and a model_value. *)
let make_stateflow_binding ident model_value =
  Stateflow_binding (Mdl_ast_functions.make_binding ident model_value)
;;

(* Build a stateflow_binding from an explicit_binding. *)
let default_explicit_binding_translator =
  (fun explicit_binding ->
   Stateflow_binding (binding_of_explicit_binding explicit_binding))
;;

let bindings_of_explicit_bindings = List.map binding_of_explicit_binding;;

let model_value_of_explicit_bindings explicit_bindings =
  let bindings = bindings_of_explicit_bindings explicit_bindings in
  Value_bindings bindings
;;

let binding_of_explicit_bindings ident explicit_bindings =
  let model_value = model_value_of_explicit_bindings explicit_bindings in
  Mdl_ast_functions.make_binding ident model_value
;;

let txt_value_of_explicit_bound_value = function
  | Ddv text -> text
  | Ddv_classified (_, text) -> text
;;

let explicit_binding_converter
      binding_translator_table default_binding_translator explicit_binding =
  let bound_ident = explicit_binding.Binding.bound_ident in
  if not (Hashtbl.mem binding_translator_table bound_ident) then
    default_binding_translator explicit_binding else
  let txt_value =
    txt_value_of_explicit_bound_value explicit_binding.Binding.bound_value in
  Hashtbl.find binding_translator_table bound_ident txt_value
;;

let stateflow_eml_item_of_explicit_binding =
  let eml_explicit_binding_translator_table =
    Hash_table.of_bindings [
      (* Bound ident is "name" *)
      Mdl_ast_functions.lname_ident,
      (fun txt_value ->
       let mdl_name = Ast_types.mdl_name_of_string txt_value in
       Stateflow_eml_name mdl_name);
      (* Bound ident is "script" *)
      Mdl_ast_functions.lscript_ident,
      (fun txt_value ->
       let mdl_code = Ast_types.mdl_code_of_string txt_value in
       Stateflow_eml_script mdl_code);
    ] in
  let eml_default_explicit_binding_translator =
    (fun explicit_binding ->
     Stateflow_eml_binding (binding_of_explicit_binding explicit_binding)) in
  explicit_binding_converter
    eml_explicit_binding_translator_table
    eml_default_explicit_binding_translator
;;

let stateflow_item_of_explicit_binding =
  explicit_binding_converter
    explicit_binding_translator_table default_explicit_binding_translator
;;

(* Ident id is no more translates to a binding: it is a true stateflow_item *)
let stateflow_item_of_name name =
  Stateflow_name (Ast_types.mdl_name_of_string name)
;;

(** {6 Complex converting functions.} *)

let stateflow_binding_of_explicit_bindings ident explicit_bindings =
  let model_value = model_value_of_explicit_bindings explicit_bindings in
  make_stateflow_binding ident model_value
;;

let stateflow_items_of_explicit_target_items =
  List.map stateflow_item_of_explicit_binding
;;

let stateflow_target_of_explicit_target = function
  | {
      explicit_target_id = target_id;
      explicit_target_name = target_name_opt;
      explicit_target_items = target_items;
    } ->
    let stateflow_items =
      stateflow_items_of_explicit_target_items target_items in
    let stateflow_items =
      match target_name_opt with
      | None -> stateflow_items
      | Some target_name ->
        stateflow_item_of_name target_name :: stateflow_items in
    {
      stateflow_target_id = target_id;
      stateflow_target_items = stateflow_items;
    }
;;

let split_explicit_slide = stateflow_binding_of_explicit_bindings slide_ident;;

let stateflow_src_dst_translator_table =
  Hash_table.of_bindings [
    (* Bound ident is "SSID" *)
    ssid_ident,
    (fun txt_value ->
     let n = unatural_of_string txt_value in
     Stateflow_src_dst_ref n);
    (* Bound ident is "intersection" *)
    Mdl_ast_functions.lintersection_ident,
    (fun txt_value ->
     let numbers = Mdl_entry_parsers.intersection_parser txt_value in
     Stateflow_intersection numbers);
]
;;

let stateflow_src_dst_item_of_explicit_binding context =
  explicit_binding_converter
    stateflow_src_dst_translator_table
    (fatal_error_of_explicit_binding context)
;;

let stateflow_src_dst_items_of_explicit_bindings context =
  let converter = stateflow_src_dst_item_of_explicit_binding context in
  List.map converter
;;

let stateflow_src_dst_items_of_explicit_src =
  stateflow_src_dst_items_of_explicit_bindings "src"
;;

let stateflow_src_dst_items_of_explicit_dst =
  stateflow_src_dst_items_of_explicit_bindings "dst"
;;

let stateflow_item_of_explicit_transition_item stateflow_items = function
  | Explicit_transition_item_binding binding ->
    fdebug (fun ppf ->
      Format.fprintf ppf "\
        @[<v>\
        while splitting a transition stateflow item@,\
        encountering binding for %S\
        @]"
        (Ast_types.string_of_ident binding.Binding.bound_ident));
    stateflow_item_of_explicit_binding binding :: stateflow_items
  | Explicit_transition_item_src src ->
    let stateflow_src_dst_items =
      stateflow_src_dst_items_of_explicit_src src in
    Stateflow_src stateflow_src_dst_items :: stateflow_items
  | Explicit_transition_item_dst dst ->
    let stateflow_src_dst_items =
      stateflow_src_dst_items_of_explicit_dst dst in
    Stateflow_dst stateflow_src_dst_items :: stateflow_items
  | Explicit_transition_item_slide slide ->
    split_explicit_slide slide :: stateflow_items
  | Explicit_transition_item_autogen _autogen ->
    warning "Dropping Stateflow autogen item";
    stateflow_items
;;

let stateflow_items_of_explicit_transition_items =
  List.fold_left stateflow_item_of_explicit_transition_item []
;;

let stateflow_transition_of_explicit_transition = function
  | {
      explicit_transition_id = transition_id;
      explicit_transition_items = transition_items;
    } ->
    let stateflow_items =
      stateflow_items_of_explicit_transition_items transition_items in
    {
      stateflow_transition_id = transition_id;
      stateflow_transition_items = stateflow_items;
    }
;;

let split_fixpt = binding_of_explicit_bindings fixpt_ident;;

let stateflow_props_type_item_of_explicit_type_item = function
  | Explicit_type_item_binding explicit_binding ->
    binding_of_explicit_binding explicit_binding
  | Explicit_type_item_fixpt explicit_fixpt ->
    split_fixpt explicit_fixpt
;;

let stateflow_props_type_items_of_explicit_type_items explicit_type_items =
  List.rev
    (List.rev_map
       stateflow_props_type_item_of_explicit_type_item
       explicit_type_items)
;;

let stateflow_props_type_items_of_explicit_type explicit_type =
  stateflow_props_type_items_of_explicit_type_items explicit_type
;;

let stateflow_array_translator_table =
  Hash_table.of_bindings
    [
      (* Bound ident is "size" *)
      size_ident,
      (fun txt_value ->
       let mdl_text = Ast_types.mdl_text_of_string txt_value in
       Stateflow_props_array_item_size mdl_text);
      (* Bound ident is "firstIndex" *)
      Mdl_ast_functions.lfirst_index_ident,
      (fun txt_value ->
       let mdl_text = Ast_types.mdl_text_of_string txt_value in
       Stateflow_props_array_item_first_index mdl_text);
    ]
;;

let stateflow_props_array_item_of_explicit_binding =
  let props_default_explicit_binding_translator =
    (fun explicit_binding ->
     Stateflow_props_array_item_binding
       (binding_of_explicit_binding explicit_binding)) in
  explicit_binding_converter
    stateflow_array_translator_table
    props_default_explicit_binding_translator
;;

let stateflow_props_array_items_of_explicit_stateflow_array
      explicit_stateflow_array =
  List.rev
    (List.rev_map
       stateflow_props_array_item_of_explicit_binding explicit_stateflow_array)
;;

let stateflow_props_item_of_explicit_props_item = function
  | Explicit_props_type explicit_type ->
    let stateflow_props_type_items =
      stateflow_props_type_items_of_explicit_type explicit_type in
    Stateflow_props_type stateflow_props_type_items
  | Explicit_props_binding explicit_binding ->
    let binding = binding_of_explicit_binding explicit_binding in
    Stateflow_props_binding binding
  | Explicit_props_stateflow_array explicit_stateflow_array ->
    let stateflow_props_array_items =
      stateflow_props_array_items_of_explicit_stateflow_array
        explicit_stateflow_array in
    Stateflow_props_array stateflow_props_array_items
  | Explicit_props_range explicit_range ->
    let binding = binding_of_explicit_bindings range_ident explicit_range in
    Stateflow_props_binding binding
;;

let stateflow_props_type_items_of_explicit_props_items explicit_props_items =
  List.rev
    (List.rev_map
       stateflow_props_item_of_explicit_props_item explicit_props_items)
;;

let stateflow_props_items_of_explicit_props explicit_props =
  stateflow_props_type_items_of_explicit_props_items explicit_props
;;

let stateflow_item_of_data_item = function
  | Explicit_data_binding binding ->
    fdebug
      (fun epf ->
       Format.fprintf
         epf "Found a binding for data item with ident %s"
         (Ast_types.string_of_ident binding.Binding.bound_ident));
    stateflow_item_of_explicit_binding binding
  | Explicit_data_props props ->
    let stateflow_props_items =
      stateflow_props_items_of_explicit_props props in
    Stateflow_props stateflow_props_items
;;

let stateflow_items_of_data_items = List.map stateflow_item_of_data_item;;

let stateflow_data_of_explicit_data = function
  | {
      explicit_data_id = id_opt;
      explicit_data_ssid = ssid_opt;
      explicit_data_name = data_name;
      explicit_data_items = data_items;
    } ->
    let id, ssid_opt =
      match id_opt, ssid_opt with
      | Some (Explicit_data_id id), ssid_opt -> id, ssid_opt
      | None, Some (Explicit_data_ssid ssid) -> ssid, None
      | None, None -> fatal_error "missing id for data in stateflow" in
    let stateflow_items = stateflow_items_of_data_items data_items in
    let stateflow_items =
      stateflow_item_of_name data_name :: stateflow_items in
    let stateflow_items =
      match ssid_opt with
      | None -> stateflow_items
      | Some (Explicit_data_ssid ssid) ->
        Stateflow_ssid_number ssid :: stateflow_items in
    {
      stateflow_data_id = id;
      stateflow_data_items = stateflow_items;
    }
;;

let stateflow_item_of_explicit_junction_item = function
  | Explicit_junction_item_binding binding ->
    Some (stateflow_item_of_explicit_binding binding)
  | Explicit_junction_item_autogen _autogen ->
    (* FIXME: enhance mdl_stateflow ast to support autogen. *)
    warning "Dropping Stateflow autogen item";
    None
;;

let stateflow_items_of_explicit_junction_items junction_items =
  let rec loop accu = function
    | [] -> List.rev accu
    | junction_item :: junction_items ->
      let junction_item_opt =
        stateflow_item_of_explicit_junction_item junction_item in
      let accu =
        match junction_item_opt with
        | None -> accu
        | Some junction_item -> junction_item :: accu in
      loop accu junction_items in
  loop [] junction_items
;;

let stateflow_junction_of_explicit_junction = function
  | {
      explicit_junction_id = junction_id;
      explicit_junction_items = junction_items;
    } ->
    let stateflow_items =
      stateflow_items_of_explicit_junction_items junction_items in
    {
      stateflow_junction_id = junction_id;
      stateflow_junction_items = stateflow_items;
    }
;;

let stateflow_items_of_explicit_event_items =
  List.map stateflow_item_of_explicit_binding
;;

let stateflow_event_of_explicit_event = function
  | {
      explicit_event_id = event_id;
      explicit_event_items = event_items;
    } ->
    let stateflow_items =
      stateflow_items_of_explicit_event_items event_items in
    {
      stateflow_event_id = event_id;
      stateflow_event_items = stateflow_items;
    }
;;

let stateflow_eml_items_of_explicit_bindings =
  List.map stateflow_eml_item_of_explicit_binding
;;

let stateflow_item_of_explicit_eml explicit_bindings =
  let stateflow_eml_items =
    stateflow_eml_items_of_explicit_bindings explicit_bindings in
  Stateflow_eml stateflow_eml_items
;;

let stateflow_binding_of_explicit_simulink =
  stateflow_binding_of_explicit_bindings simulink_ident
;;

let add_stateflow_item_to_classifier chart_item_classifier stateflow_item =
  {
    chart_item_classifier with
    stateflow_items = stateflow_item :: chart_item_classifier.stateflow_items;
  }
;;

let check_empty items_type container_type container_id = function
  | [] -> ()
  | _ ->
    fatal_error (Printf.sprintf "encountering %s items in %s %s"
      items_type container_type (Unatural.to_string container_id))
;;

let check_empty_targets = check_empty "target";;

let check_empty_charts = check_empty "chart";;

let check_empty_states = check_empty "state";;

let check_empty_transitions = check_empty "transition";;

let check_empty_junctions = check_empty "junction";;

(* The recursive part of stateflow tree *)
let rec classify_chart_item chart_id chart_item_classifier = function
  | Explicit_chart_item_binding binding ->
    let stateflow_item = stateflow_item_of_explicit_binding binding in
    add_stateflow_item_to_classifier chart_item_classifier stateflow_item
  | Explicit_chart_item_children explicit_children ->
    let classified_children = classify_explicit_children explicit_children in
    begin
      match classified_children with
      | {
          children_targets = stateflow_targets;
          children_charts = stateflow_charts;
          children_states = stateflow_states;
          children_transitions = stateflow_transitions;
          children_datas = stateflow_datas;
          children_junctions = stateflow_junctions;
          children_events = stateflow_events;
          children_items = stateflow_items;
        } ->
        let container_type = "chart" in
        check_empty_targets container_type chart_id stateflow_targets;
        check_empty_charts container_type chart_id stateflow_charts;
        {
          stateflow_datas =
            stateflow_datas @ chart_item_classifier.stateflow_datas;
          stateflow_events =
            stateflow_events @ chart_item_classifier.stateflow_events;
          stateflow_junctions =
            stateflow_junctions @ chart_item_classifier.stateflow_junctions;
          stateflow_states =
            stateflow_states @ chart_item_classifier.stateflow_states;
          stateflow_transitions =
            stateflow_transitions @
              chart_item_classifier.stateflow_transitions;
          stateflow_items =
            stateflow_items @ chart_item_classifier.stateflow_items;
        }
    end
  | Explicit_chart_item_sub_view_s sub_view_s ->
    let stateflow_item =
      stateflow_binding_of_explicit_bindings sub_view_ident sub_view_s in
    add_stateflow_item_to_classifier chart_item_classifier stateflow_item
  | Explicit_chart_item_eml explicit_eml ->
    let stateflow_item = stateflow_item_of_explicit_eml explicit_eml in
    add_stateflow_item_to_classifier chart_item_classifier stateflow_item
  | Explicit_chart_item_active_state_output active_state_output ->
    let stateflow_item =
      stateflow_binding_of_explicit_bindings
        active_state_output_ident active_state_output in
    add_stateflow_item_to_classifier chart_item_classifier stateflow_item
  | Explicit_chart_item_plant_modeling_info ->
    let stateflow_item = Stateflow_plant_modeling_info in
    add_stateflow_item_to_classifier chart_item_classifier stateflow_item

and classify_chart_items chart_id chart_items =
  List.fold_left
    (classify_chart_item chart_id) empty_chart_item_classifier chart_items

and stateflow_chart_of_explicit_chart = function
  | {
      explicit_chart_id = chart_id;
      explicit_chart_items = chart_items;
    } ->
    let chart_item_classifier =
      classify_chart_items chart_id chart_items in
    match chart_item_classifier with
    | {
        stateflow_datas = stateflow_datas;
        stateflow_events = stateflow_events;
        stateflow_junctions = stateflow_junctions;
        stateflow_states = stateflow_states;
        stateflow_transitions = stateflow_transitions;
        stateflow_items = stateflow_items;
      } ->
      {
        stateflow_chart_id = chart_id;
        stateflow_chart_datas = List.rev stateflow_datas;
        stateflow_chart_events = List.rev stateflow_events;
        stateflow_chart_junctions = List.rev stateflow_junctions;
        stateflow_chart_states = List.rev stateflow_states;
        stateflow_chart_transitions = List.rev stateflow_transitions;
        stateflow_chart_items = List.rev stateflow_items;
      }

and classify_child children_classifier = function
  | Explicit_child_target explicit_target ->
    let stateflow_target =
      stateflow_target_of_explicit_target explicit_target in
    {
      children_classifier with
      children_targets =
        stateflow_target :: children_classifier.children_targets;
    }
  | Explicit_child_chart explicit_chart ->
    let stateflow_chart = stateflow_chart_of_explicit_chart explicit_chart in
    {
      children_classifier with
      children_charts =
        stateflow_chart :: children_classifier.children_charts;
    }
  | Explicit_child_state explicit_state ->
    let stateflow_state =
      stateflow_state_of_explicit_state explicit_state in
    {
      children_classifier with
      children_states =
        stateflow_state :: children_classifier.children_states;
    }
  | Explicit_child_transition explicit_transition ->
    let stateflow_transition =
      stateflow_transition_of_explicit_transition explicit_transition in
    {
      children_classifier with
      children_transitions =
        stateflow_transition :: children_classifier.children_transitions;
    }
  | Explicit_child_data explicit_data ->
    let stateflow_data =
      stateflow_data_of_explicit_data explicit_data in
    {
      children_classifier with
      children_datas =
        stateflow_data :: children_classifier.children_datas;
    }
  | Explicit_child_junction explicit_junction ->
    let stateflow_junction =
      stateflow_junction_of_explicit_junction explicit_junction in
    {
      children_classifier with
      children_junctions =
        stateflow_junction :: children_classifier.children_junctions;
    }
  | Explicit_child_event explicit_event ->
    let stateflow_event =
      stateflow_event_of_explicit_event explicit_event in
    {
      children_classifier with
      children_events =
        stateflow_event :: children_classifier.children_events;
    }
  | Explicit_child_table _table ->
    (* FIXME: enhance mdl_stateflow ast to support table. *)
    warning "Dropping Stateflow table: please report";
    children_classifier
  | Explicit_child_state_cell _state_cell ->
    (* FIXME: enhance mdl_stateflow ast to support state_cell. *)
    warning "Dropping Stateflow state_cell: please report";
    children_classifier
  | Explicit_child_transition_cell _transition_cell ->
    (* FIXME: enhance mdl_stateflow ast to support transition_cell. *)
    warning "Dropping Stateflow transition_cell: please report";
    children_classifier

and classify_explicit_children explicit_children =
  List.fold_left classify_child empty_children_classifier explicit_children

and classify_state_item state_id state_item_classifier = function
  | Explicit_state_item_binding binding ->
    let stateflow_item = stateflow_item_of_explicit_binding binding in
    add_stateflow_item_to_classifier state_item_classifier stateflow_item
  | Explicit_state_item_children explicit_children ->
    let classified_children = classify_explicit_children explicit_children in
    begin
      match classified_children with
      | {
          children_targets = stateflow_targets;
          children_charts = stateflow_charts;
          children_states = stateflow_states;
          children_transitions = stateflow_transitions;
          children_datas = stateflow_datas;
          children_junctions = stateflow_junctions;
          children_events = stateflow_events;
          children_items = stateflow_items;
        } ->
        let container_type = "state" in
        check_empty_targets container_type state_id stateflow_targets;
        check_empty_charts container_type state_id stateflow_charts;
        {
          stateflow_datas =
            stateflow_datas @ state_item_classifier.stateflow_datas;
          stateflow_events =
            stateflow_events @ state_item_classifier.stateflow_events;
          stateflow_junctions =
            stateflow_junctions @ state_item_classifier.stateflow_junctions;
          stateflow_states =
            stateflow_states @ state_item_classifier.stateflow_states;
          stateflow_transitions =
            stateflow_transitions @
              state_item_classifier.stateflow_transitions;
          stateflow_items =
            stateflow_items @ state_item_classifier.stateflow_items;
        }
    end
  | Explicit_state_item_sub_view_s sub_view_s ->
    let stateflow_item =
      stateflow_binding_of_explicit_bindings sub_view_ident sub_view_s in
    add_stateflow_item_to_classifier state_item_classifier stateflow_item
  | Explicit_state_item_eml explicit_eml ->
    let stateflow_item = stateflow_item_of_explicit_eml explicit_eml in
    add_stateflow_item_to_classifier state_item_classifier stateflow_item
  | Explicit_state_item_active_state_output active_state_output ->
    let stateflow_item =
      stateflow_binding_of_explicit_bindings
        active_state_output_ident active_state_output in
    add_stateflow_item_to_classifier state_item_classifier stateflow_item
  | Explicit_state_item_simulink explicit_simulink ->
    let stateflow_item =
      stateflow_binding_of_explicit_simulink explicit_simulink in
    add_stateflow_item_to_classifier state_item_classifier stateflow_item
  | Explicit_state_item_autogen _autogen ->
    (* FIXME: enhance mdl_stateflow ast to support autogen. *)
    warning "Dropping Stateflow autogen";
    state_item_classifier
  | Explicit_state_item_truth_table _truth_table ->
    (* FIXME: enhance mdl_stateflow ast to support truth_table. *)
    warning "Dropping Stateflow truth_table";
    state_item_classifier

and classify_state_items state_id state_items =
  List.fold_left
    (classify_state_item state_id) empty_state_item_classifier state_items

and stateflow_state_of_explicit_state = function
  | {
      explicit_state_id = state_id;
      explicit_state_items = state_items;
    } ->
    let state_item_classifier =
      classify_state_items state_id state_items in
    match state_item_classifier with
    | {
        stateflow_datas = stateflow_datas;
        stateflow_events = stateflow_events;
        stateflow_junctions = stateflow_junctions;
        stateflow_states = stateflow_states;
        stateflow_transitions = stateflow_transitions;
        stateflow_items = stateflow_items;
      } ->
      {
        stateflow_state_id = state_id;
        stateflow_state_datas = List.rev stateflow_datas;
        stateflow_state_events = List.rev stateflow_events;
        stateflow_state_junctions = List.rev stateflow_junctions;
        stateflow_state_states = List.rev stateflow_states;
        stateflow_state_transitions = List.rev stateflow_transitions;
        stateflow_state_items = List.rev stateflow_items;
      }
;;

let classify_machine_item machine_id machine_item_classifier = function
  | Explicit_machine_binding explicit_binding ->
    let stateflow_item = stateflow_item_of_explicit_binding explicit_binding in
    {
      machine_item_classifier with
      explicit_stateflow_machine_items =
        stateflow_item ::
        machine_item_classifier.explicit_stateflow_machine_items;
    }
  | Explicit_machine_debug _explicit_debug ->
    warning "Dropping Stateflow debug";
    machine_item_classifier
  | Explicit_machine_children explicit_children ->
    let classified_children = classify_explicit_children explicit_children in
    match classified_children with
    | {
        children_targets = stateflow_targets;
        children_charts = stateflow_charts;
        children_states = stateflow_states;
        children_transitions = stateflow_transitions;
        children_datas = stateflow_datas;
        children_junctions = stateflow_junctions;
        children_events = stateflow_events;
        children_items = stateflow_items;
      } ->
      let container_type = "machine" in
      check_empty_states container_type machine_id stateflow_states;
      check_empty_transitions container_type machine_id stateflow_transitions;
      check_empty_junctions container_type machine_id stateflow_junctions;
      {
        explicit_stateflow_machine_charts =
          stateflow_charts @
          machine_item_classifier.explicit_stateflow_machine_charts;
        explicit_stateflow_machine_datas =
          stateflow_datas @
          machine_item_classifier.explicit_stateflow_machine_datas;
        explicit_stateflow_machine_events =
          stateflow_events @
          machine_item_classifier.explicit_stateflow_machine_events;
        explicit_stateflow_machine_targets =
          stateflow_targets @
          machine_item_classifier.explicit_stateflow_machine_targets;
        explicit_stateflow_machine_items =
          stateflow_items @
          machine_item_classifier.explicit_stateflow_machine_items;
      }
;;

let classify_machine_items machine_id machine_items =
  List.fold_left
    (classify_machine_item machine_id) empty_machine_item_classifier
    machine_items
;;

let stateflow_machine_of_explicit_machine = function
  | {
      explicit_machine_id = machine_id;
      explicit_machine_items = machine_items;
    } ->
    let machine_item_classifier =
      classify_machine_items machine_id machine_items in
    match machine_item_classifier with
    | {
        explicit_stateflow_machine_charts = stateflow_charts;
        explicit_stateflow_machine_datas = stateflow_datas;
        explicit_stateflow_machine_events = stateflow_events;
        explicit_stateflow_machine_targets = stateflow_targets;
        explicit_stateflow_machine_items = stateflow_items;
      } ->
      {
        stateflow_machine_id = machine_id;
        stateflow_machine_charts = List.rev stateflow_charts;
        stateflow_machine_datas = List.rev stateflow_datas;
        stateflow_machine_events = List.rev stateflow_events;
        stateflow_machine_targets = List.rev stateflow_targets;
        stateflow_machine_items = List.rev stateflow_items;
      }
;;

let classify_instance_item instance_id instance_items_classifier = function
  | {
      Binding.
      bound_ident = bound_ident;
      bound_value = bound_value;
    } ->
    let txt_value = txt_value_of_explicit_bound_value bound_value in
    if bound_ident = name_ident then
      let name = Ast_types.mdl_name_of_string txt_value in
      {
        instance_items_classifier with
        explicit_stateflow_instance_name_opt = Some name;
      }
    else if bound_ident = chart_ident then
      let chart_ref = stateflow_reference_of_txt_value txt_value in
      {
        instance_items_classifier with
        explicit_stateflow_instance_chart_ref_opt = Some chart_ref;
      }
    else if bound_ident = machine_ident then
      let machine_ref = stateflow_reference_of_txt_value txt_value in
      {
        instance_items_classifier with
        explicit_stateflow_instance_machine_ref_opt = Some machine_ref;
      }
    else
      fatal_error (Printf.sprintf "unknown binding %s for instance %s"
        (Ast_types.string_of_ident bound_ident)
        (Unatural.to_string instance_id))
;;

let classify_instance_items instance_id instance_items =
  List.fold_left
    (classify_instance_item instance_id) empty_instance_items_classifier
    instance_items
;;

let stateflow_instance_of_explicit_instance = function
  | {
      explicit_instance_id = instance_id;
      explicit_instance_items = instance_items;
    } ->
    let classified_instance_items =
      classify_instance_items instance_id instance_items in
    match classified_instance_items with
    | {
        explicit_stateflow_instance_name_opt = instance_name_opt;
        explicit_stateflow_instance_machine_ref_opt = instance_machine_ref_opt;
        explicit_stateflow_instance_chart_ref_opt = instance_chart_ref_opt;
      } ->
      let instance_name, instance_machine_ref, instance_chart_ref =
        match
          instance_name_opt, instance_machine_ref_opt, instance_chart_ref_opt
        with
        | Some instance_name,
          Some instance_machine_ref,
          Some instance_chart_ref ->
          instance_name, instance_machine_ref, instance_chart_ref
        | _, _, _ ->
          fatal_error (Printf.sprintf "instance %s is incomplete"
            (Unatural.to_string instance_id)) in
      {
        stateflow_instance_id = instance_id;
        stateflow_instance_name = instance_name;
        stateflow_instance_machine_ref = instance_machine_ref;
        stateflow_instance_chart_ref = instance_chart_ref;
      }
;;

let stateflow_section_item_of_explicit_stateflow_section_item = function
  | Explicit_stateflow_section_machine explicit_machine ->
    let stateflow_machine =
      stateflow_machine_of_explicit_machine explicit_machine in
    Stateflow_section_machine stateflow_machine
  | Explicit_stateflow_section_instance explicit_instance ->
    let stateflow_instance =
      stateflow_instance_of_explicit_instance explicit_instance in
    Stateflow_section_instance stateflow_instance
;;

let stateflow_section_of_explicit_stateflow_section = function
  | Explicit_stateflow_section explicit_stateflow_section_items ->
    let stateflow_section_items =
      List.map
        stateflow_section_item_of_explicit_stateflow_section_item
        explicit_stateflow_section_items in
    Stateflow_section stateflow_section_items
;;

let stateflow_section = stateflow_section_of_explicit_stateflow_section;;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
 End:
*)
