(***********************************************************************)
(*                                                                     *)
(*                               Simport                               *)
(*                                                                     *)
(*                   Pierre Weis, Clément Franchini                    *)
(*                   INRIA Rocquencourt                                *)
(*                                                                     *)
(*  Copyright 2013-2019,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved.                                               *)
(*                                                                     *)
(*  This file is distributed under the terms of the BSD License.       *)
(*                                                                     *)
(***********************************************************************)

(* $Id: xml_translating_basics.ml,v 1.1 2019-05-10 14:00:17 jpc Exp $ *)

open Xml_explicit_ast;;

type module_name = string;;

type property_conversion_mode =
  | Mode_ignore
  (* | Mode_force *)
  | Mode_property
;;

type parsed_property =
  | BlockMirror
  | BlockRotation
  | DropShadow
  | FontSize
  | Location
  | MaskHideContents
  | Opaque
  | Ports
  | Position
  | LimitOutput
  | SaturateOnIntegerOverflow
  | ShowName
  | TreatAsAtomicUnit
  | ZOrder
(** The set of properties bound to a model value and that model value
  text must be parsed with [slx_model_value_parser]. *)
;;

type parsed_parameter =
  | BlockType
  | ConcatenateDimension
  | IndexMode
  | IndexOptions
  | InheritSampleTime
  | LinearizeAsDelay
  | LinearizeMemory
  | Mode
  | OutputDataType
  | PhaseDelay
  | RTWStateStorageClass
  | SampleTimeType
  | ShowOutputPort
  | StateMustResolveToSignalObject
  | StatesWhenEnabling
  | ZeroCross
(** The set of parameters bound to a model value and that model value
  text must be parsed with [slx_model_value_parser]. *)
;;

(* By default all unknown parameter/property are unparsed
type unparsed_parameter =
  | ActionType
  | Amplitude
  | Period
  | PulseWidth
  | SampleTime
  | TriggerType
  | WhileBlockType
  | X0
  | Indices
;;
*)

let not_yet module_name s =
  failwith (Printf.sprintf "%s: not yet: %s" module_name s)
;;

let fatal_error module_name s =
  Say.fatal_error (Printf.sprintf "%s: %s" module_name s)
;;

let warning module_name s =
  Say.warning (Printf.sprintf "%s: %s" module_name s)
;;

let not_yet_warning module_name s =
  warning module_name (Printf.sprintf "not yet: %s" s)
;;

let fdebug module_name f =
  Say.fdebug
    (fun ppf -> Format.fprintf ppf
      "%s: %a"
      module_name
      (fun ppf () -> f ppf) ())
;;

let basics_module_name = "Xml_translating_basics";;

(*
let _basics_not_yet = not_yet basics_module_name;;

let _basics_fatal_error = fatal_error basics_module_name;;

let _basics_warning = warning basics_module_name;;
*)

let basics_fdebug = fdebug basics_module_name;;

(** Basics functions. *)

let string_binding_value text = Mdl_ast.Value_simple (Mdl_ast.String text);;

let simple_value_of_natural natural =
  Mdl_ast.Int (Unatural.to_string natural)
;;

let model_value_of_simple_value simple_value =
  Mdl_ast.Value_simple simple_value
;;

let model_value_of_natural natural =
  model_value_of_simple_value (simple_value_of_natural natural)
;;

let model_value_of_optional_classified parameter_class_opt text =
  let parse_model_value text = Mdl_entry_parsers.slx_model_value_parser text in
  let model_value_of_classified parameter_class text =
    match parameter_class with
    | Char -> string_binding_value text
    | Double
    | Int32
    | Uint32
    | Logical -> parse_model_value text in
  match parameter_class_opt with
  | None ->
    basics_fdebug
      (fun ppf ->
       Format.fprintf ppf
         "model_value_of_optional_classified: unclassified text %S"
         text);
    parse_model_value text
  | Some parameter_class -> model_value_of_classified parameter_class text
;;

let model_value_of_text text = model_value_of_optional_classified None text;;

let rec string_of_block_sid bl_sid =
  match bl_sid with
  | Mdl_ast.Sid_compound sc -> string_of_block_sid_compound sc
  | Mdl_ast.Sid_string s -> s

and string_of_block_sid_compound = function
  | Mdl_ast.Sid_natural n -> Unatural.to_string n
  | Mdl_ast.Sid_ref (n, sc) ->
    Printf.sprintf "%s:%s"
      (Unatural.to_string n) (string_of_block_sid_compound sc)
  | Mdl_ast.Sid_child (n, sc) ->
    Printf.sprintf "%s::%s"
      (Unatural.to_string n) (string_of_block_sid_compound sc)
;;

let model_value_of_block_ssid block_sid =
  let str = string_of_block_sid block_sid in
  Mdl_ast.Value_simple (Mdl_ast.String str)
;;

let parsed_property_of_string = function
  | "BlockMirror" -> BlockMirror
  | "BlockRotation" -> BlockRotation
  | "DropShadow" -> DropShadow
  | "FontSize" -> FontSize
  | "Location" -> Location
  | "MaskHideContents" -> MaskHideContents
  | "Opaque" -> Opaque
  | "Ports" -> Ports
  | "Position" -> Position
  | "LimitOutput" -> LimitOutput
  | "SaturateOnIntegerOverflow" -> SaturateOnIntegerOverflow
  | "ShowName" -> ShowName
  | "TreatAsAtomicUnit" -> TreatAsAtomicUnit
  | "ZOrder" -> ZOrder
  | _ -> raise Not_found
;;

let is_parsed_property property_name =
  try
    ignore (parsed_property_of_string property_name);
    true
  with
  | Not_found -> false
;;

let parsed_parameter_of_string = function
  | "BlockType" -> BlockType
  | "ConcatenateDimension" -> ConcatenateDimension
  | "IndexMode" -> IndexMode
  | "IndexOptions" -> IndexOptions
  | "InheritSampleTime" -> InheritSampleTime
  | "LinearizeAsDelay" -> LinearizeAsDelay
  | "LinearizeMemory" -> LinearizeMemory
  | "Mode" -> Mode
  | "OutputDataType" -> OutputDataType
  | "PhaseDelay" -> PhaseDelay
  | "RTWStateStorageClass" -> RTWStateStorageClass
  | "SampleTimeType" -> SampleTimeType
  | "ShowOutputPort" -> ShowOutputPort
  | "StateMustResolveToSignalObject" -> StateMustResolveToSignalObject
  | "StatesWhenEnabling" -> StatesWhenEnabling
  | "ZeroCross" -> ZeroCross
  | _ -> raise Not_found
;;

(* By default all unknown parameter/property are unparsed
let unparsed_parameter_of_string = function
  | "ActionType" -> ActionType
  | "Amplitude" -> Amplitude
  | "Indices" -> Indices
  | "Period" -> Period
  | "PulseWidth" -> PulseWidth
  | "SampleTime" -> SampleTime
  | "TriggerType" -> TriggerType
  | "WhileBlockType" -> WhileBlockType
  | "X0" -> X0
  | _ -> raise Not_found
;;
*)

let is_parsed_parameter parameter_name =
  try
    ignore (parsed_parameter_of_string parameter_name);
    true
  with
  | Not_found -> false
;;

(* By default all unknown parameter/property are unparsed
let is_unparsed_parameter parameter_name =
  try
    ignore (unparsed_parameter_of_string parameter_name);
    true
  with
  | Not_found -> false
;;
*)

let binding cmode = function
  | {
      Binding.
      bound_ident = (Ident.Ident bound_name) as  bound_ident;
      bound_value = bound_value;
    } ->
    let value =
      match bound_value with
      | Xml_explicit_ast.Ddv text ->
        Say.fdebug (fun epf -> Format.fprintf epf
          "making binding for %s" bound_name);
        begin
          match cmode with
          (* | Mode_force -> model_value_of_optional_classified None text *)
          | Mode_property
              when is_parsed_property bound_name
                || is_parsed_parameter bound_name ->
            Say.fdebug (fun epf -> Format.fprintf epf
              "use mode property for parsed %s" bound_name);
            model_value_of_optional_classified None text
          (* By default all unknown parameter/property are unparsed
          | Mode_property
              when is_unparsed_parameter bound_name ->
            Say.fdebug (fun epf -> Format.fprintf epf
              "use mode property for unparsed %s" bound_name);
            string_binding_value text
           *)
          | Mode_ignore ->
            Say.fdebug (fun epf -> Format.fprintf epf
              "use mode ignore for %s" bound_name);
            string_binding_value text
          | Mode_property ->
            Say.fdebug (fun epf -> Format.fprintf epf
              "use mode property for unknown %s" bound_name);
            string_binding_value text
        end
      | Ddv_classified (parameter_class, text) ->
        model_value_of_optional_classified (Some parameter_class) text in
    {
      Binding.bound_ident = bound_ident;
      bound_value = value;
    }
;;

let bindings cmode = List.map (binding cmode);;

(*
 Local Variables:
  compile-command: "cd ../../..; make"
 End:
*)
